//? cpmdefinitionswrc.h
//? C+- by Ulrich Mutze. Status of work 2022-03-28.
//? Copyright (c) 2022 Ulrich Mutze, www.ulrichmutze.de
//? All rights reserved

#ifndef CPM_DEFINITIONS_H_
#define CPM_DEFINITIONS_H_
/*************************************************************************
   cpmdefinitions.h
   Description: In this file one can place defines that are valid in all
   translation units working with Cpm classes
   Cpm = Classes for Physics and Mathematics || C+- (C plus minus)
         -           -           -                   - -    -
************************************************************************/
// This file will never be included by a C+- program since its name is
// cpmdefinitionswrc.h instead of cpmdefinitions.h. Upon renaming, it can
// function as cpmdefinitions.h.
// The name appendix 'wrc' stands for 'with reference comments'.
// We dont't use the more readable form cpmdefinitions_wrc.h in order
// not to disturb pdflatex-based Ruby script listing generation.
// The role of this file is to hold valid and complete explanations of the
// macros which can be set in any file cpmdefinitions.h.
// A file with similar scope is cpmsystemdependencieswrc.h.

#define CPM_NAMEOF
   // If this is defined, the C+- naming scheme is used instead of
   // C++ typeinfo. Recommended since for template types, native
   // typeinfo creates inconveniently long names.

#define CPM_RANGE_CHECK
   // If this is defined, all indexed access to arrays is index-
   // checked.
   // Recommended, increases safty.

#define CPM_USECOUNT
   // If this is defined, reference counting and 'copy on write'
   // are activated in class template CpmArrays::V<T>. Modern
   // implementation based on std::shared_ptr/ as outlined in
   // B. Stroustrup C++11, p. 512.
   // Recommended, increases speed.

#undef CPM_Fn
//#define CPM_Fn
   // if this is defined we use my own tool classes F1,F2,...
   // for binding variables of functions. Otherwise we use
   // std::bind and std::placeholders enabled by interfacing F<X,Y>
   // with std::function<Y(X)>. The latter method, provided with
   // C++11, is more flexible and simpler to use.

/////// controling numerical precision of floating point numbers /////////

/* There are 5 macros which deal with numerical precision to be set at
 compile time: CPM_FLOAT, CPM_DOUBLE, CPM_LONG, CPM_QUAD, CPM_MP.
 Only one or no of these macros may be set. Setting none has the same
 effect as setting CPM_DOUBLE, i.e. as saying
   #define CPM_DOUBLE
 If more than one are set behavior is undefined.
*/
//#define CPM_FLOAT
   // mainly for experimentation. 4 bytes per number
   // See www.ulrichmutze.de/rubystuff/doc_rnum/index.html
   // open AppMath::R and search for 'industrial project'

#define CPM_DOUBLE
   // 8 bytes per number
   // meaningfull default for real world simulations

//define CPM_LONG
   // 10 bytes per number
   // If this is or CPM_QUAD, or CPM_MP is defined, the types Z and N are
   // long int and unsigned long int, otherwise simply int and unsigned
   // int.
   // If this is defined type R is long double.

//#define CPM_QUAD
   // 16 bytes per number
   // Unfortunately many functions of the Eigen library do not yet
   // cooperate with QUAD. On my system QUAD is half as fast as DOUBLE
   // which is perfectly fair. It seems equal in speed as CPM_LONG which
   // indicates an technology advantage in the implementation of CPM_QUAD.
   // See cpm/tut1/source/tut1.cpp, function main, for a speed comparison
   // of the various number sizes.
   // Neeeds linkage -lgmp -lquadmath

//#define CPM_MP 32
   // This macro enables usage of multiple precision mathematics based on
   // boost/multiprecision. The number gives approximately the valid
   // decimal digits of what is known as the significand.
   // Regretfully I decided no longer to build also
   // on mpreal.h by Pavel Holoborodko. This was my first enabler of
   // multiprecision studies. It seems to be more difficult to use
   // together with the Eigen library than the boost implementation.
   // The boost-based implementation needs linkage  -lgmp. In order to
   // avoid compiler errors one needs the compiler option -fpermissive.
   // Even then lots of warnings are created unless the option -w prevents
   // this. Further, there is a probably deeper issue: The boost-based
   // implementation works only if for the multiprecision floating point
   // data type an option has been set which disables expression templates
   // and expression templates are said to be responsible for the
   // efficiency of Eigen.

#include <cpmcompdef.h>
   // This allows the defines given to the compiler via options of type
   // -D to override the definitions of CPM_FLOAT, CPM_DOUBLE, CPM_LONG,
   // CPM_QUAD, and CPM_MP given previously in this file. For instance, a compiler option
   // -DCPM_RMP_E (to be entered in e.g. Code::Blocks or in a makefile)
   // asks for real numbers of 256 decimal places. The defines
   // -DCPM_RMP_A, -DCPM_RMP_B, -DCPM_RMP_C, -DCPM_RMP_D, -DCPM_RMP_E,
   // provide 16, 32, 64, 128, 256 decimal places for real numbers of
   // type R.
   //
   // The active version of file cpmcompdef.h is placed at cpm0/include.
   // It should not be changed.

#define CPM_TEMPLATE_TESTS
   // having this defined increases code size
   // it helps to detect inconsistencies in interfaces of template classes
   // should be disabled for release compilation

//////////////////////////// do not use, considered obsolete /////////////
   //#define CPM_USE_MPI
   // MPI = message passing interface (for parallel computing).
   // See cpmmpi.h and cpmmpi.cpp.
   // This functionality was virtually unused for 7 years and was
   // reanimated in 2009-05-27. After dislocation of a single
   // b(road)cast statement in the main constructor of RecordHandler
   // a 2D example with gravitating and elastic repelling particles
   // worked fine and created a movie on which all particles were
   // to be seen in a common field of view. The command for
   // running was:
   //    mpirun -localonly -np 4 palaMS.exe
   // However, the C+- system favors a style of programming which
   // abstracts from technical aspects of the computing environment.
   // So questions of how to distribute a computational task over
   // various processors or various computers are not considered
   // of interest here. They  w e r e  of interest in the industrial
   // toning modeling project which contributed significantly to
   // the experiences on which the C+- system is based.

#endif
