//? cpmfo.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FO_H_
#define CPM_FO_H_
/*

Purpose: Define classes describing function-like objects with
   order operations

*/
#include <cpmf.h>

namespace CpmFunctions{

// Base class for classes describing binary operations

namespace auxiliary{

template <class X, class Y>
class BINOPFUO: public FncObj<X,Y>{ // BINOP stands for Binary
// Operation
protected:
   const F<X,Y>  f1;
   const F<X,Y>  f2;

public:
   BINOPFUO(  const F<X,Y> & f1In, const F<X,Y> & f2In ):
   f1(f1In),f2(f2In){}
};

// Base class for classes describing monary operations

template <class X, class Y>
class MONOPFUO: public FncObj<X,Y>{

protected:
   const F<X,Y>  f1;

public:

    MONOPFUO( const F<X,Y> &  f1In): f1(f1In){}
};

// order-related operations
// Now it will be assumed that Y is ordered ( Y < Y and Y > Y)

// class MINFUO<>

template <class X, class Y>
class MINFUO: public BINOPFUO<X,Y>{

public:
   Y operator()(const X& x)const
   {
      Y y1=f1(x);
      Y y2=f2(x);
      return (y1 < y2 ? y1 : y2);
   }

   MINFUO(const F<X,Y> & f1In, const F<X,Y> & f2In):
      BINOPFUO<X,Y>(f1In,f2In){}
};

// class MAXFUO<>

template <class X, class Y>
class MAXFUO: public BINOPFUO<X,Y>{

public:
   Y operator()(const X& x)const
   {
      Y y1=f1(x);
      Y y2=f2(x);
      return (y1 > y2 ? y1 : y2);
   }

   MAXFUO(const F<X,Y>& f1In, const F<X,Y>& f2In):
      BINOPFUO<X,Y>(f1In,f2In){}
};

// setting function values

template <class X, class Y>
class SETVALUEFUO : public FncObj<X,Y>{   // changing function values
   // assumes that '==' is defined in X

   const F<X,Y>  f1;
   const X xs;
   const Y ys;

public:
   Y operator()(const X& x)const
   {
      if (x==xs) return ys; else return f1(x);
   }

   SETVALUEFUO(const F<X,Y>& g1, const X& x1, const Y& y1):
   f1(g1),xs(x1),ys(y1){}
};

} // auxiliary

///////////////////////////// class Fo<X,Y> //////////////////////////////

template <class X, class Y>
class Fo: public F<X,Y> { // version of F with order-related operations

   typedef Fo<X,Y> Type;

public:

   Fo(const F<X,Y>& g):F<X,Y>(g){}
      // upcast constructor

   Fo(const Fo<X,Y>& g):F<X,Y>(g){}
      // copy constructor

   Fo(void):F<X,Y>(){}
      // default constructor

   Fo(FncObj<X,Y>* fop):F<X,Y>(fop){}
      // constructor from pointers to FncObj

// Construction from function pointers

   Fo( Y (*f)(const X& )):F<X,Y>(f){}
   Fo( Y (*f)(X)):F<X,Y>(f){}

   Fo( std::function<Y(X)> f):F<X,Y>(f){}

// Construtor for constant function

   explicit Fo(const Y& y1):F<X,Y>(y1){}
      // Functionality:
      // Y y=...; F<X,Y> f(y); X x=...; f(x)==y;
      // last statement always evaluates to 1, independent of x

   void setValue(const X& x0, const Y& y0);
      // non-constant member function which modifies the function value
      // at x=x0 to become y0
   virtual F<X,Y>* clone()const{ return new Fo(*this);}
   virtual Word nameOf()const
   {
      Word wi="Fo<";
      Word wx=CpmRoot::Name<X>()(X());
      Word wy=CpmRoot::Name<Y>()(Y());
      return wi&wx&","&wy&">";
   }
   CPM_ORDER
};

template <class X, class Y>
   // assumption Y < Y defined

Fo<X,Y> inf( const Fo<X,Y>& f1, const Fo<X,Y>& f2)
{
   using namespace auxiliary;
   return Fo<X,Y>(new MINFUO<X,Y>(f1,f2));
}

template <class X, class Y>
   // assumption Y < Y defined

Fo<X,Y> sup( const Fo<X,Y>& f1, const Fo<X,Y>& f2)
{
   using namespace auxiliary;
   return Fo<X,Y>(new MAXFUO<X,Y>(f1,f2));
}

// Order interface consistent but in a sense trivial.
// No assumtions on X,Y
// Does not take into account the function values but
// only the storage position.

template <class X, class Y>
Z Fo<X,Y>::com(const Fo<X,Y>& f)const
{
   return F<X,Y>::com(f);
}

} // namespace

#endif
