//? cpmm2.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_M_H_
#define CPM_M_H_
/*

   Description: Defines a template class M<X,Y> of mappings
      X-->Y comparable to std::map<X,Y>
*/

#include <cpms.h>

////////////////////// class M<X,Y> ////////////////////////////////////

namespace CpmArrays{

   using namespace CpmStd;
   using CpmFunctions::F;
   using CpmArrays::S;

namespace aux{ // a technical construct

template <class X, class Y>
class  PairO1{  // Pair with order according to first element only.
   // Especially all pairs differing in only the second component
   // are equal by ==. This allows an elegant access to the x-values
   // of (x,y) pairs held in M<X,Y>::rep_. Since this is a 'heavy-duty
   // technical class' it is no longer implemented as derived from
   // X2<X,Y>.
   X x_;
   Y y_;
   typedef PairO1<X,Y> Type;
public:
   Word nameOf()const
   {
      return Word("PairO1< "&
          CpmRoot::Name<X>()(X())&
          ","&
          CpmRoot::Name<Y>()(Y())&
          " >");
   }
   CPM_ORDER
   CPM_IO
   PairO1(X const& x, Y const& y):x_(x),y_(y){}
   explicit PairO1(X const& x):x_(x),y_(){}
   PairO1():x_(),y_(){}
   X& c1(void){ return x_;}
   Y& c2(void){ return y_;}
   X const& c1(void)const{ return x_;}
   Y const& c2(void)const{ return y_;}
};

template <class X, class Y>
Z PairO1<X,Y>::com(PairO1<X,Y> const& s)const
{
   if (x_<s.x_) return 1;
   else if (x_>s.x_) return -1;
   else return 0;
}

template <class X, class Y>
bool PairO1<X,Y>::prnOn(ostream& str)const
{
   cpmp(x_);
   cpmp(y_);
   return true;
}

template <class X, class Y>
bool PairO1<X,Y>::scanFrom(istream& str)
{
   cpms(x_);
   cpms(y_);
   return true;
}

} // aux

template <class X, class Y>
   class M{ // map, associative array, dictionary, hash
   // it is assumed that X<X and X>X is defined
   Y y0_;
      // default Y, allows to define
      // Y const& operator[](const X const& x)const;
      // which STL is omitting
   S< aux::PairO1<X,Y> > rep_;
   Z locate(X const& x)const{ return rep_.locate(aux::PairO1<X,Y>(x));}

public:
   typedef M<X,Y> Type;
   typedef aux::PairO1<X,Y> RepType;
   CPM_IO
   CPM_ORDER
   M(void):y0_(),rep_(){}
      // constructor for the void set of X,Y pairs.
      // Non-trivial instances can be obtained by applying
      // mutating methods, such as function set(...) on the
      // default intance

   RepType const& operator[](Z i)const{ return rep_[i];}

   M(S<X> const& sx, F<X,Y> const& f);
      // Setting the values by an algorithm.

   M(S<X> const& sx, Y const& y);
      // Setting a single value y for all elements of sx.

   void merge(V<Type> const& mapList, bool reverse=false);
      // combining *this and a list of maps into a single one.
      // This constructs a map that is formed out of all
      // (x,y)-pairs of *this and the components of mapList.
      // If for a x, there are more pairs (x,y), (x,y'), ...
      // The pair coming from mapList[i] with highest i
      // will actually be taken (if we would take more than one
      // we would not get a map). Thus we work through the
      // mapList in ascending order of their indexes and use it
      // in an overwriting mode. If the second argument differs
      // from the default value, we use reversed order so that
      // mapList[i] has highest priority for smallest i.

   Y operator()(X const& x)const;
      // in analogy to F<X,Y>; returns the 'value of *this at x'

   S<Y> operator()(S<X> const& sx)const;
      // The common extension of functions from single argument values
      // to sets of argument values

   Y& operator[](X const& x);
      // setting values in array-style (which also is the STL-style)
      // for M<X,Y> f, X x, Y y the statement
      // f[x]=y has the same effect as f.set(x,y)

   Y const& operator[](X const& x)const;
      // getting to the values as references.
      // Let Z g(const& Y), M<X,Y> f, X x, Y y . Then
      // g(f[x]) does the same as
      // Y y1; f.get(x,y1); g(y1);
      // So, operator[] provides slightly more convenient grammar than
      // get(...) in this case.
      // This is the most efficient non-mutating access function

   void set(X const& x, Y const& y){(*this)[x]=y;}
      //: set
      //  M<X,Y> f, X x, Y y ==>
      // ( f.set(x,y) ==> (y==f(x) evaluates to true)). Interesting:
      // it is possible to express this mixture of logic notation
      // and C++ much clearer in pure C++:
      // template <class X, class Y>
      // bool setTest(X x, Y y, M<X,Y> f){ f.set(x,y); return y==f(x);}
      // returns true for all arguments.

   bool get(X const& x, Y& y)const;
      //: get
      // template <class X, class Y>
      // bool getTest(X x, M<X,Y> f){ Y y; f.get(x,y); return y==f(x);}
      // returns true for all arguments.

// cardinality access: number of (x,y) pairs stored in *this
   Z car()const{ return rep_.dim();}
      //: cardinality
   Z dim()const{ return rep_.dim();} // for uniformity with other array
      //: dimension
   // types
   Z size()const{ return rep_.dim();} // for uniformity with STL code
   bool isVoid(void)const{ return rep_.isVoid();}
      //: is void
   Z b()const{ return rep_.b();}
      //. begin
   Z e()const{ return rep_.e();}
      //. end

// accessing arguments, values, and (x,y)-pairs by an index ranging from
// 1 to car()

   X x(Z i)const{ return rep_(i).c1();}
   Y y(Z i)const{ return rep_(i).c2();}
   X2<X,Y> xy(Z i)const{ return X2<X,Y>(x(i),y(i));}
      // template <class X, class Y>
      // bool xyTest(Z i, M<X,Y> f){ return y(i)==f(x(i));}
      // returns true for all arguments.
   bool defined(X const& x)const{ return locate(x)!=0;}
      // f.defined(x) iff before a statement f[x]=... or f.set(x,...)
      // happened.
   void clear_(X const& x){ Z i=locate(x); if (i>0) rep_.eliminate(i);}
      // after f.clear_(x), we have f.defined(x)==false

   S<X> dom()const;
      //: domain
      // domain of the function X-->Y associated with object *this
      // returns {x \in X | defined(x)==true}

   S<Y> ran()const;
      //: range
      // range of the function X-->Y associated with object *this
      // returns {y \in Y | y==operator[](x) for some x \in X}

   M<X,Y> select(F<X,bool> const& sel)const;
      // returns the M<X,Y> objects consisting of all those
      // (x,y)-pairs of *this for which sel(x)==true

   template <class T>
   M<X,T> operator()(F<Y,T> const& g)const;

   template <class T>
   M<X,T> operator&(M<Y,T> const& m)const;

   template< class T>
   M<T,Y> circ(M<T,X> const& m)const{ return m&(*this);}
      //: circ (LaTeX-command)
      // concatenation of mappings

   Word toWord()const;
      //: to word

   Word nameOf()const;
      //: name of
};

///////////////////////// Implementation ///////////////////////////////

template <class X, class Y>
template<class T>
M<X,T> M<X,Y>::operator()(F<Y,T> const& g)const
{
   M<X,T> res();
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      res.set(xyi.c1(),g(xyi.c2()));
   }
   return res;
}

template <class X, class Y>
template <class T>
M<X,T> M<X,Y>::operator&(M<Y,T> const& m)const
{
   M<X,T> res;
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      bool bi=m.defined(xyi.c2());
      if (bi) res.set(xyi.c1(),m[xyi.c2()]);
   }
   return res;
}

template <class X, class Y>
Word M<X,Y>::nameOf()const
{
   Word wi="M<";
   Word wx=CpmRoot::Name<X>()(X());
   Word wy=CpmRoot::Name<Y>()(Y());
   return wi&wx&","&wy&">";
}

template <class X, class Y>
Word M<X,Y>::toWord()const
{
   Word res="{ ";
   for (Z i=b();i<=e();++i){
      X2<X,Y> xyi=xy(i);
      Word wxi=CpmRoot::toWord<X>(xyi.c1());
      Word wyi=CpmRoot::toWord<Y>(xyi.c2());
      res&=" ( "&wxi&" , "&wyi&" ) ";
      if (i!=e()) res&=" , ";
   }
   res&=" }";
   return res;
}

template <class X, class Y>
M<X,Y>::M(S<X> const& sx, F<X,Y> const& f):y0_(),rep_()
{
   for (Z i=sx.b();i<=sx.e();++i){
      X xi=sx[i];
      set(xi,f(xi));
   }
}

template <class X, class Y>
M<X,Y>::M(S<X> const& sx, Y const& y):y0_(),rep_()
{
   for (Z i=sx.b();i<=sx.e();++i){
      X xi=sx[i];
      set(xi,y);
   }
}

template <class X, class Y>
Y M<X,Y>::operator()(X const& x)const
{
   Z i=locate(x);
   if (i==0) return Y();
   else return rep_[i].c2();
}

/*******************************
template <class X, class Y>
Y& M<X,Y>::operator[](X const& x)
{
   cpmcerr<<"Y& M<X,Y>::operator[](X const& x) calling"<<endl;
   aux::PairO1<X,Y> fx(x);
   cpmcerr<<" aux::PairO1<X,Y> fx(x); done"<<endl;
 //  Z i=rep_.loc3(fx);

   cpmcerr<<" first i = "<<i<<endl;
   if (i==0){
      cpmcerr<<"rep_.add(fx) calling"<<endl;
      rep_.add(fx);
   }
   i=rep_.locate(fx);
   cpmcerr<<" second i = "<<i<<endl;

   if (i<=0) cpmerror("Y& M<X,Y>::operator[](X const& x): i<=0");
   cpmcerr<<"Y& M<X,Y>::operator[](X const& x) nearly done"<<endl;

   rep_.enq_(fx);
   Z i=rep_.loc3(fx).c1();
   return rep_.ref(i).c2();
      // since the ordering of rep_ only relies on rep_.ref(i).c1()
      // there is no danger in exposing rep_.ref(i).c2() to
      // changes from outside
}
*********************************/

template <class X, class Y>
Y& M<X,Y>::operator[](X const& x)
{
   aux::PairO1<X,Y> fx(x);
   rep_<<fx;
   Z i=rep_.locate(fx);
   return rep_.ref(i).c2();
}

template <class X, class Y>
Y const& M<X,Y>::operator[](X const& x)const
{
   aux::PairO1<X,Y> fx(x);
   Z i=rep_.locate(fx);
   if (i<=0){
      return y0_;
   }
   else{
      return rep_[i].c2();
   }
}

template <class X, class Y>
bool M<X,Y>::get(X const& x, Y& y)const
{
   Z i=locate(x);
   if (i==0){
      return false;
   }
   else{
      y=rep_[i].c2();
      return true;
   }
}

template <class X, class Y>
S<Y> M<X,Y>::operator()(S<X> const& sx)const
{
   Z n=sx.car();
   V<Y> val(n);
   for (Z i=1;i<=n;i++){
      val[i]=y((sx[i]));
   }
   return S<Y>(val);
}

template <class X, class Y>
void M<X,Y>::merge(V<Type> const& mapList, bool reversed)
{
   Z ni,i,j,d=mapList.dim();
   if (!reversed){
      for (i=1;i<=d;i++){
         ni=mapList[i].car();
         for (j=1;j<=ni;j++){
            X x=mapList[i].x(j);
            Y y=mapList[i].y(j);
            set(x,y);
         }
      }
   }
   else{
      for (i=d;i>=1;i--){
         ni=mapList[i].car();
         for (j=1;j<=ni;j++){
            X x=mapList[i].x(j);
            Y y=mapList[i].y(j);
            set(x,y);
         }
      }
   }
}

template <class X, class Y>
S<X> M<X,Y>::dom()const
{
   S<X> res;
   for (Z i=1;i<=dim();++i) res.add(x(i));
   return res;
}

template <class X, class Y>
S<Y> M<X,Y>::ran()const
{
   S<Y> res;
   for (Z i=1;i<=dim();++i) res.add(y(i));
   return res;
}

template <class X, class Y>
M<X,Y>  M<X,Y>::select(F<X,bool> const& sel)const
{
   M<X,Y> res;
   for (Z i=1;i<=dim();++i){
      X2<X,Y> xyi=xy(i);
      X xi=xyi.c1();
      if (sel(xi)) res[xi]=xyi.c2();
   }
   return res;
}

template <class X, class Y>
bool M<X,Y>::prnOn(ostream& str)const
{
   Z mL=3;
   static Word loc("M<>::prnOn(ostream&)");
   CPM_MA
   Z n=car();
   cpmwat;
   cpmp(n);
   if (n==0){
      cpmmessage(mL, "no values there");
      goto label;
   }
   for (Z i=1;i<=n;i++){
      if (CpmRoot::wrtTit){
         Word wi("// i=");
         wi&=cpm(i);
         bool bi=wi.prnOn(str);
         cpmassert(bi==true,loc);
      }
      X2<X,Y> val(x(i),y(i));
      if (!val.prnOn(str)){
         CPM_MZ
         return false;
      }
   }
label:
   cpmwet;
   CPM_MZ
   return true;
}

template <class X, class Y>
bool M<X,Y>::scanFrom(istream& str)
{
   Z mL=3;
   Z mL2=5;
   static Word loc("M<>::scanFrom(istream&)");
   CPM_MA
   Z n;
   if (!CpmRoot::scanT<Z>(n,str)){
      cpmwarning(loc&": can't read dimension");
      CPM_MZ
      return false;
   }
   else if (n<0){
      cpmwarning(loc&": negative dimension: n="&cpm(n));
      CPM_MZ
      return false;
   }
   else if (n>dimMax)
      cpmwarning(loc&": dimension read as "&cpm(n)&
      ": probably too large");
   else if (n==0){
      cpmwarning(loc&" dimension read as 0");
      if (cpmverbose>mL2){ // copying the stream beyond the dubious point
         // can be useful
         char c;
         while (str.get(c)) cpmdata.put(c);
      }
   }
   else{
      cpmmessage(mL,loc&": dimension read as "&cpm(n));
   }
   M<X,Y> res;
   X2<X,Y> val;
   for (Z i=1;i<=n;i++){
      if (!val.scanFrom(str)){
         cpmwarning(loc&": reading failed for i="&cpm(i)&" of "&cpm(n));
         CPM_MZ
         return false;
      }
      res.set(val.c1(),val.c2());
   }
   *this=res;
   CPM_MZ
   return true;
}

template <class X, class Y >
Z M<X,Y>::com(M<X,Y> const& s)const
{
   Z d1=dim(), d2=s.dim();
   if (d1<d2) return 1;
   else if (d1>d2) return -1;
   else{ // then d1==d2
      for (Z i=1;i<=d1;i++){
         if (xy(i)<s.xy(i))   return 1;
         if (xy(i)>s.xy(i))   return -1;
      }
      return 0;
   }
}

} // CpmArrays

#endif
