//? cpmv_.h
//? C+- by Ulrich Mutze. Status of work 2022-10-08.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_V__H_
#define CPM_V__H_
/*

Purpose:
With C++11 'move semantics' was added to C++ and seduced me to base
the C+- array template V<> on std::vector<> which provided the
move facility 'under the hood'and no longer relies on
'copy on write' (there are still C+- classes that do). The present class template is
a near to minimal vector class which works with copy on write (no move semantics)
and does indexing by the same syntax as std::vector<> for staightforward speed
comparisons. In my project tut1 there is done such a comparison with a
program that does several calls to copy constructors and assignments in chain.
Here copy on write works around 700 times faster than 'move semantics'.
This test case may well be far from practice but it looks wise not to
eliminate copy on write completely from C+-.
*/


#include <vector>

/////////////////////////////// class V_<> ///////////////////////////////
// Array without index check, reference counting, and copy on write, and
// 'value semantics'
// Generalized from Koenig's class Handle p. 72-73.
// This is a remnant from an earlier version of the C+- class system.
// It is a minimal and functioning prototype of a template array class
// with a reference counted pointer to its data.

namespace CpmArrays{

   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::N;
   using namespace std;

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.

class V_{ // vector template, indexing starts with 0 .

public:

explicit V_(N n=0) ;
      // Has n components. Gives an error for n<0 and n>dimMax.
      // The components are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitly or implicitly).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.
      // If n>0, the first valid index is 1, which reflects the
      // normal behavior of class V.

    V_(std::vector<T> const& v);
      // Construction from a standard library vector. This is useful
      // for interaction with the standard containers (which don't
      // implement reference counting).

   V_(N n, T const& t) ;
      // Has n components all initialized as t.

   V_(V_<T> const& h);
      // copy constructor

// constructors from explicit lists
   explicit V_(std::initializer_list<T> il );
      // requires C++11
      // constructors from explicit lists such as
      // V_<Z> v{1,2,4,8};

   ~V_();
      // destructor, we intend not to derive from this, so not virtual

   V_<T>& operator=(V_<T> const& h);
      // assignment

   N dim()const { return sz_;}
      //: dimension
      // Returns the number of components of the vector *this

   N size()const { return sz_;}
      //: size
      // for uniformity with STL

   std::vector<T> std()const;
      //: standard
      // Returns a std::vector which holds all components of *this.

   Word nameOf()const;
      //: name of
      // returns a name of the type

// constant access functions

   const T& operator[](N i)const{ return *(p_+i);}


   T& operator[](N i){ cow_(); return *(p_+i);}

private:

   void cow_(void);
      //. copy on write
      // The first '*this-changing' statement in the
      // body of a non-constant member function has to be
      //   cow_();

   T* copy()const;
      //: copy
      // returns a pointer to a storage area that contains
      // a copy of the components of *this

// data members

   UseCount u_;

   // u_ becomes initialized by the default constructor and thus gets
   // a count 1, saying that the piece of free store pointed to by p_
   // is in use by *this

   N sz_;
      // number of valid indexes (depends on iv_, equals iv_.car())

   T* p_;
      // pointer to beginning of the array.
};

////////////////////// Implementation ///////////////////////////////

template <class T>
inline   V_<T>::~V_(){
   if (u_.only()){ delete[] p_;}
}

template <class T>
Word V_<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="V_<";
   return wi&nt&">";
}

template <class T>
T* V_<T>::copy()const
{
   T* itp=p_;
   T* q=new T[sz_];
   T* it=q;
   N i=sz_;
   while (i--) *it++=*itp++;
   return q;
}

template <class T>
void V_<T>::cow_(void)
{
   if (u_.makeOnly()){ // If we come here, we are not the only users of the data provided by *this
      // so, when we change the data of *this we offend other users. To avoid conflicts we
      // have to make a copy of *this's actual data and take ownership over this copy.
      // If this is done we are (so far) the only user of these new data, so ne state of this
      // is characterized by p_ pointing to the copy of data, for which a single owner
      // is registered. If we don't come here, we are the only user of the data and may
      // modify them at will. If the function which called cow_ has finished the unchanged data
      // are still there since other users refer to them and see the data and u and its count
      // Since we are not longer among these users (we work on ouer own privat copy) this count
      // has to be reduced by one (what makeOnly does if it returns 'true')
      p_= (sz_==0 ? 0 : copy());
      u_.startNew_();
   }
}

// constructors

template <class T>
V_<T>::V_(N n):sz_(n)
{
   p_= n==0 ? nullptr : new T[n];
   for (N i=0;i<sz_;++i) p_[i] = T();
}

template <class T>
V_<T>::V_(N n, T const& t):sz_(n)
{
   p_= n==0 ? nullptr : new T[n];
   for (N i=0;i<sz_;++i) p_[i] = t;
}

template <class T>
V_<T>::V_(std::initializer_list<T> il ):
sz_(il.size())
{
   p_= sz_==0 ? nullptr : new T[sz_];
   std::uninitialized_copy(il.begin(),il.end(),p_);
      // see BS4, p. 498
}

template <class T>
V_<T>::V_(std::vector<T> const& v):sz_(v.size())
{
   p_= sz_==0 ? nullptr : new T[sz_];
   typename std::vector<T>::const_iterator i;
   T* q=p_;
   for (i=v.begin();i!=v.end();++i) *q++=*i;
}

template <class T>
std::vector<T> V_<T>::std()const
{
   std::vector<T> res(sz_);
   typename std::vector<T>::iterator i;
   T* q=p_;
   for (i=res.begin();i!=res.end();++i) *i=*q++;
   return res;
}

template <class T>
V_<T>::V_(V_<T> const& h) :sz_(h.sz_),u_(h.u_),p_(h.p_){}

template <class T>
V_<T>& V_<T>::operator=(V_<T> const& h)
{
   if (sz_==0 && h.sz_==0) return *this;
   if (this==&h) return *this;
   sz_=h.sz_;
   if (u_.reattach_(h.u_)) delete[] p_;
   p_=h.p_;
   return *this;
}

//////////////////////////// experiment with replacing T* by std::vector<T>* ////////////////////////

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.

class Vv{ // vector template, indexing starts with 0 .

public:

explicit Vv(N n=0) ;
      // Has n components. Gives an error for n<0 and n>dimMax.
      // The components are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitly or implicitly).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.
      // If n>0, the first valid index is 1, which reflects the
      // normal behavior of class V.

    Vv(std::vector<T> const& v);
      // Construction from a standard library vector. This is useful
      // for interaction with the standard containers (which don't
      // implement reference counting).

   Vv(N n, T const& t) ;
      // Has n components all initialized as t.

   Vv(Vv<T> const& h);
      // copy constructor

// constructors from explicit lists
   explicit Vv(std::initializer_list<T> il );
    //   requires C++11
     //  constructors from explicit lists such as
    //   Vv<Z> v{1,2,4,8};

   ~Vv();
      // destructor, we intend not to derive from this, so not virtual

   Vv<T>& operator=(Vv<T> const& h);
      // assignment

   N dim()const { return sz_;}
      //: dimension
      // Returns the number of components of the vector *this

   N size()const { return sz_;}
      //: size
      // for uniformity with STL

   std::vector<T> std()const;
      //: standard
      // Returns a std::vector which holds all components of *this.

   Word nameOf()const;
      //: name of
      // returns a name of the type

// constant access functions

   const T& operator[](N i)const{ return (*p_)[i];}


   T& operator[](N i){ cow_(); return (*p_)[i];}

private:

   void cow_(void);
      //. copy on write
      // The first '*this-changing' statement in the
      // body of a non-constant member function has to be
      //   cow_();

// data members

   UseCount u_;

   // u_ becomes initialized by the default constructor and thus gets
   // a count 1, saying that the piece of free store pointed to by p_
   // is in use by *this

   N sz_;
      // number of valid indexes (depends on iv_, equals iv_.car())

   vector<T> *p_;
      // pointer to beginning of the array.
};

////////////////////// Implementation ///////////////////////////////

template <class T>
inline   Vv<T>::~Vv(){
   if (u_.only()){ p_->~vector<T>();}
}

template <class T>
Word Vv<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="Vv<";
   return wi&nt&">";
}

template <class T>
void Vv<T>::cow_(void)
{
   if (u_.makeOnly()){
      p_=new vector<T>(*p_);
      u_.startNew_();
   }
}

// constructors

template <class T>
Vv<T>::Vv(N n):sz_(n),u_()
{
   p_=new vector<T>(sz_,T());
}

template <class T>
Vv<T>::Vv(N n, T const& t):sz_(n),u_()
{
   p_=new vector<T>(sz_,t);
}

template <class T>
Vv<T>::Vv(std::initializer_list<T> il ):
sz_(il.size()) // 137 is dummy argument
{
   p_= new vector<T>(sz_);
   std::uninitialized_copy(il.begin(),il.end(),p_);
      // see BS4, p. 498
}

template <class T>
Vv<T>::Vv(std::vector<T> const& v):sz_(v.size()),u_()
{
   p_= new vector<T>(v);

}

template <class T>
std::vector<T> Vv<T>::std()const
{
   return *p_;
}

template <class T>
Vv<T>::Vv(Vv<T> const& h) :sz_(h.sz_),u_(h.u_),p_{h.p_}{}

template <class T>
Vv<T>& Vv<T>::operator=(Vv<T> const& h)
{
   if (sz_==0 && h.sz_==0) return *this;
      // added 2002-03-07, should be OK
   if (this==&h) return *this;
   sz_=h.sz_;
   if (u_.reattach_(h.u_))  p_->~vector<T>();
   p_=h.p_;
   return *this;
}

/////////////////////////////// Vsp ////////////////////////////////////

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.

class Vsp{ // vector template, indexing starts with 0 .

public:

explicit Vsp(N n=0) ;
      // Has n components. Gives an error for n<0 and n>dimMax.
      // The components are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitly or implicitly).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.
      // If n>0, the first valid index is 1, which reflects the
      // normal behavior of class V.

    Vsp(std::vector<T> const& v);
      // Construction from a standard library vector. This is useful
      // for interaction with the standard containers (which don't
      // implement reference counting).

   Vsp(N n, T const& t) ;
      // Has n components all initialized as t.

   Vsp(Vsp<T> const& h):sz_(h.sz_),p_(h.p_){}
      // copy constructor

// constructors from explicit lists
   explicit Vsp(std::initializer_list<T> il );
    //   requires C++11
     //  constructors from explicit lists such as
    //   Vsp<Z> v{1,2,4,8};

   ~Vsp() = default;
      // destructor, we intend not to derive from this, so not virtual

   Vsp<T>& operator=(Vsp<T> const& h)=default;
      // assignment

   N dim()const { return sz_;}
      //: dimension
      // Returns the number of components of the vector *this

   N size()const { return sz_;}
      //: size
      // for uniformity with STL

   std::vector<T> std()const;
      //: standard
      // Returns a std::vector which holds all components of *this.

   Word nameOf()const;
      //: name of
      // returns a name of the type

// constant access functions

   const T& operator[](N i)const{ return (*p_)[i];}


   T& operator[](N i){ /*cow_(); */return (*p_)[i];}

private:

   vector<T>* clone(void){ return p_;}


   void cow_(void);
      //. copy on write
      // The first '*this-changing' statement in the
      // body of a non-constant member function has to be
      //   cow_();

// data members

   N sz_;
      // number of valid indexes (depends on iv_, equals iv_.car())

   shared_ptr<vector<T>> p_;
      // pointer to beginning of the array.
};

////////////////////// Implementation ///////////////////////////////

template <class T>
Word Vsp<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="Vsp<";
   return wi&nt&">";
}

template <class T>
void Vsp<T>::cow_(void)
{
   if (p_.use_count()>1){
      p_=shared_ptr<vector<T>>(new vector<T>(*p_));
   }
}

// constructors

template <class T>
Vsp<T>::Vsp(N n):sz_(n)
{
   p_=shared_ptr<vector<T>>(new vector<T>(sz_,T()));

}

template <class T>
Vsp<T>::Vsp(N n, T const& t):sz_(n)
{
   p_=shared_ptr<vector<T>>(new vector<T>(sz_,t));
}

template <class T>
Vsp<T>::Vsp(std::initializer_list<T> il ):
sz_(il.size()) // 137 is dummy argument
{
   p_= shared_ptr<vector<T>>(new vector<T>(sz_));
   std::uninitialized_copy(il.begin(),il.end(),p_);
      // see BS4, p. 498
}

template <class T>
Vsp<T>::Vsp(std::vector<T> const& v):sz_(v.size())
{
   p_= shared_ptr<vector<T>>(new vector<T>(v));

}

template <class T>
std::vector<T> Vsp<T>::std()const
{
   return *p_;
}

}  // namespace CpmArrays

#endif
