//? cpmv_.h
//? C+- by Ulrich Mutze. Status of work 2022-10-08.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_VSP_H_
#define CPM_VSP_H_
/*

Purpose:
With C++11 'move semantics' was added to C++ and seduced me to base
the C+- array template V<> on std::vector<> which provided the
move facility 'under the hood'and no longer relies on
'copy on write' (there are still C+- classes that do). The present class template is
a near to minimal vector class which works with copy on write (no move semantics)
and does indexing by the same syntax as std::vector<> for staightforward speed
comparisons. In my project tut1 there is done such a comparison with a
program that does several calls to copy constructors and assignments in chain.
Here copy on write works around 700 times faster than 'move semantics'.
This test case may well be far from practice but it looks wise not to
eliminate copy on write completely from C+-.
*/

#include <memory>
//#include <vector>

/////////////////////////////// class Vsp<> ///////////////////////////////
// Array without index check, reference counting, and copy on write, and
// 'value semantics'
// Generalized from Koenig's class Handle p. 72-73.

namespace CpmArrays{

//   using namespace CpmStd;

   using CpmRoot::N;

template <typename T>
class DynamicArray
{
private:
	T* m_array;
	N m_length;

public:
	DynamicArray(N length=0)
		: m_array(new T[length]), m_length(length)
	{
	}


	~DynamicArray()
	{
		delete[] m_array;
	}

	// Copy constructor
	DynamicArray(const DynamicArray &arr) = default;

	// Copy assignment
	DynamicArray& operator=(const DynamicArray &arr) = default;

	// Move constructor
	DynamicArray(DynamicArray &&arr) noexcept
		:  m_array(arr.m_array), m_length(arr.m_length)
	{
		arr.m_length = 0;
		arr.m_array = nullptr;
	}

	// Move assignment
	DynamicArray& operator=(DynamicArray &&arr) noexcept
	{
		if (&arr == this)
			return *this;

		delete[] m_array;

		m_length = arr.m_length;
		m_array = arr.m_array;
		arr.m_length = 0;
		arr.m_array = nullptr;

		return *this;
	}

	N getLength() const { return m_length; }
	T& operator[](N index) { return m_array[index]; }
	const T& operator[](N index) const { return m_array[index]; }

};


}  // namespace CpmArrays

#endif
