//? cpmv_.h
//? C+- by Ulrich Mutze. Status of work 2022-10-08.
//? Copyright (c) 2016 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_VSP_H_
#define CPM_VSP_H_
/*

Purpose:
With C++11 'move semantics' was added to C++ and seduced me to base
the C+- array template V<> on std::vector<> which provided the
move facility 'under the hood'and no longer relies on
'copy on write' (there are still C+- classes that do). The present class template is
a near to minimal vector class which works with copy on write (no move semantics)
and does indexing by the same syntax as std::vector<> for staightforward speed
comparisons. In my project tut1 there is done such a comparison with a
program that does several calls to copy constructors and assignments in chain.
Here copy on write works around 700 times faster than 'move semantics'.
This test case may well be far from practice but it looks wise not to
eliminate copy on write completely from C+-.
*/

#include <memory>
#include <vector>

/////////////////////////////// class Vsp<> ///////////////////////////////
// Array without index check, reference counting, and copy on write, and
// 'value semantics'
// Generalized from Koenig's class Handle p. 72-73.

namespace CpmArrays{

   using namespace CpmStd;

   using CpmRoot::Word;
   using CpmRoot::N;

template <class T>
   // We assume that T provides (explicitly or implicitly)
   // copy constructor, and assignment or that T is a built-in type.

class Vsp{ // vector template, indexing starts with 0 .

public:

   Vsp()=default;

   explicit Vsp(N n);
      // Has n components. Gives an error for n<0 and n>dimMax.
      // The components are initialized by the default constructor
      // of T if T is a class for which such a constructor is defined
      // (explicitly or implicitly).
      // If T is a built-in type, initialization is done as 0.
      // See BS3, p. 131 for initialization of built-in types via
      // formal constructor calls.
      // If n>0, the first valid index is 1, which reflects the
      // normal behavior of class V.

    Vsp(std::vector<T> const& v);
      // Construction from a standard library vector. This is useful
      // for interaction with the standard containers (which don't
      // implement reference counting).

   Vsp(N n, T const& t) ;
      // Has n components all initialized as t.

   Vsp(Vsp<T> const& h)=default;
      // copy constructor

// constructors from explicit lists
   explicit Vsp(std::initializer_list<T> il );
      // requires C++11
      // constructors from explicit lists such as
      // Vsp<Z> v{1,2,4,8};

   ~Vsp()= default;
      // destructor, we intend not to derive from this, so not virtual

   Vsp<T>& operator=(Vsp<T> const& h)=default;
      // assignment

   N dim()const { return sz_;}
      //: dimension
      // Returns the number of components of the vector *this

   N size()const { return sz_;}
      //: size
      // for uniformity with STL

   std::vector<T> std()const;
      //: standard
      // Returns a std::vector which holds all components of *this.

   Word nameOf()const;
      //: name of
      // returns a name of the type

// constant access functions

   const T& operator[](N i)const{ return p_.get()[i];}


   T& operator[](N i){ return p_.get()[i];}

private:

//   void cow_(void);
//      //. copy on write
//      // The first '*this-changing' statement in the
//      // body of a non-constant member function has to be
//      //   cow_();
//
//   T* copy()const;
//      //: copy
//      // returns a pointer to a storage area that contains
//      // a copy of the components of *this

// data members

 //  UseCount u_;

   // u_ becomes initialized by the default constructor and thus gets
   // a count 1, saying that the piece of free store pointed to by p_
   // is in use by *this

   N sz_;
      // number of valid indexes (depends on iv_, equals iv_.car())

   std::shared_ptr<T> p_;
      // pointer to beginning of the array.
};

////////////////////// Implementation ///////////////////////////////

//template <class T>
//inline   Vsp<T>::~Vsp(){
//   if (u_.only()){ delete[] p_;}
//}

template <class T>
Word Vsp<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="Vsp<";
   return wi&nt&">";
}

//template <class T>
//T* Vsp<T>::copy()const
//{
//   T* itp=p_;
//   T* q=new T[sz_];
//   T* it=q;
//   N i=sz_;
//   while (i--) *it++=*itp++;
//   return q;
//}

//template <class T>
//void Vsp<T>::cow_(void)
//   // body is void if CPM_USECOUNT is not defined
//{
//   if (u_.makeOnly()){
//      p_= (sz_==0 ? 0 : copy());
//      u_.startNew_();
//   }
//}

// constructors

//template <class T>
//Vsp<T>::Vsp(N n):sz_(n)
//{
//   p_= n==0 ? nullptr : new T[n];
//   for (N i=0;i<sz_;++i) p_[i] = T();
//}

template <typename T>
Vsp<T>::Vsp(N n):sz_(n)
{
    p_=std::shared_ptr<T>(new T[n]);
   //for (N i=0;i<sz_;++i) p_[i] = T();
}

template <class T>
Vsp<T>::Vsp(N n, T const& t):sz_(n)
{
   p_= n==0 ? nullptr : new T[n];
   for (N i=0;i<sz_;++i) p_[i] = t;
}

template <class T>
Vsp<T>::Vsp(std::initializer_list<T> il ):
sz_(il.size()) // 137 is dummy argument
{
   p_= sz_==0 ? nullptr : new T[sz_];
   std::uninitialized_copy(il.begin(),il.end(),p_);
      // see BS4, p. 498
}

template <class T>
Vsp<T>::Vsp(std::vector<T> const& v):sz_(v.size())
{
   p_= sz_==0 ? nullptr : new T[sz_];
   typename std::vector<T>::const_iterator i;
   T* q=p_;
   for (i=v.begin();i!=v.end();++i) *q++=*i;
}

template <class T>
std::vector<T> Vsp<T>::std()const
{
   std::vector<T> res(sz_);
   typename std::vector<T>::iterator i;
   T* q=p_;
   for (i=res.begin();i!=res.end();++i) *i=*q++;
   return res;
}

//template <class T>
//Vsp<T>::Vsp(Vsp<T> const& h) :sz_(h.sz_),u_(h.u_),p_(h.p_){}

//template <class T>
//Vsp<T>& Vsp<T>::operator=(Vsp<T> const& h)
//{
//   if (sz_==0 && h.sz_==0) return *this;
//      // added 2002-03-07, should be OK
//   if (this==&h) return *this;
//   sz_=h.sz_;
//   if (u_.reattach_(h.u_)) delete[] p_;
//   p_=h.p_;
//   return *this;
//}

}  // namespace CpmArrays

#endif
