//? cpmmpi.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmmpi.cpp

   Description: see cpmmpi.h
      can be made part of the source file set even if
      CPM_USE_MPI is not defined
*/
#include <cpmmpi.h>

int CpmMPI::mpiverbose=1; // experiment, 1 is normal

namespace{
   const int mL1=10;
   const int mL2=20;   
}

   // limits against which mpiverbose will be tested
CpmMPI::Com CpmMPI::Cpm_com;
   // initialized by default constructor

#if defined(CPM_USE_MPI)

#include <cpmsystem.h>

using namespace CpmMPI;

using CpmRoot::Word;

using std::ostringstream;
using std::endl;

void CpmMPI::initialize(int* a, char*** b){ MPI_Init(a,b);}

void CpmMPI::finalize(){ MPI_Finalize();}

CpmMPI::Com::Com(MPI_Comm comm):mc(comm)
{
   int mpiflag;
   MPI_Initialized(&mpiflag);
   if(mpiflag) { // 01-09-25: mpiflag introduced by John Zollweg,
      // Cornell university, to avoids warnings
      MPI_Comm_size(comm,&size);
      MPI_Comm_rank(comm,&rank);
      ++rank;
   }
}

int CpmMPI::Com::cyc(int i)const
{
   while(i>size) i-=size;
   while(i<1) i+=size;
   return i;
}

// not working code for non-blocking sends and receives is commentarized
bool CpmMPI::Com::sendStr(const string& s, int dest, int tag)const
{
   static int shL=100; // show length, 100 is normal
   static int counter=0;
   counter++;
   int n=1+(int)s.size();
   if (mpiverbose>mL1){
      ostringstream mes;
      mes<<"sendStr( dest="<<dest<<" tag="<<tag<<") started n="
         <<n<<"; send counter="<<counter;
  //    if (counter==21) mes<<" the string to send is "<<s; // experiment
  // ad hoc
      cpmmessage(Word(mes.str()));
   }
   int tag1=tag;
   int tag2=tag+1;
   int* np=new int(n);
 //  MPI_Request request;
 //  MPI_Isend(np,1, MPI_INT, dest-1, tag1, mc,&request);
   MPI_Ssend(np,1, MPI_INT, dest-1, tag1, mc);
   if (mpiverbose>mL2){
      ostringstream mes;
      mes<<"sending the number done, tag="<<tag1;
      mes<<" next send will use tag="<<tag2;
      cpmmessage(Word(mes.str()));
   }
   delete np;
   if (mpiverbose>mL2){
      ostringstream mes;
      if (n<=shL)
         mes<<endl<<"string to send is: "<<s;
      else
         mes<<endl<<"string to send not shown since longer than "<<shL;
      mes<<endl<<"end of string to send";
      cpmmessage(Word(mes.str()));
   }

  int res= MPI_Ssend((void*)s.c_str(), n, MPI_CHAR, dest-1, tag2, mc);
  // int res= MPI_Isend((void*)s.c_str(), n, MPI_CHAR, dest-1, tag2,
  // mc,&request);

   if (mpiverbose>mL2){
      ostringstream mes;
      mes<<"sending the string done";
      cpmmessage(Word(mes.str()));
   }

   bool b=(res==MPI_SUCCESS);
   if (mpiverbose>mL1){
      ostringstream mes;
      mes<<"sendStr( dest="<<dest<<" tag="<<tag<<"), n="<<n<<
         " done; send counter="<<counter;
      cpmmessage(Word(mes.str()));
   }
   if (!b) cpmerror("error in CpmMPI::sendStr");
   return b;
}

bool CpmMPI::Com::recStr(string& s, int source, int tag)const
{
   static int counter=0;
   counter++;
   if (mpiverbose>mL1){
      ostringstream mes;
      mes<<"recStr( source="<<source<<" tag="<<tag<<") started"
         <<"; receive counter="<<counter;
      cpmmessage(Word(mes.str()));
   }
   int tag1=tag;
   int tag2=tag+1;
   MPI_Status status;
  // MPI_Request request;
   int* np=new int;
   MPI_Recv(np,1,MPI_INT,source-1,tag1,mc,&status);
 //  MPI_Irecv(np,1,MPI_INT,source-1,tag1,mc,&request);
   // The non-blocking version does not work in my program, not clear why
   int n=*np;
   if (mpiverbose>=mL1){ // >= intentionally to have a control mode
      // in which only the most important events get reported
      // for all other tests we have mpiverbose>mL
      cpmmessage("string received of length n="&cpmwrite(n));
   }
   char* data=new char[n];
 //  int out=MPI_Irecv(data,n,MPI_CHAR,source-1,tag2,mc,&request);
   int out=MPI_Recv(data,n,MPI_CHAR,source-1,tag2,mc,&status);
   ostringstream ost;
   ost<<data;
   s=ost.str();
   if (mpiverbose>mL1){
      ostringstream mes;
      mes<<"recStr( source="<<source<<" tag="<<tag<<") done; n="
         <<n<<"; receive counter="<<counter;
      if (mpiverbose>mL2){
         mes<<endl<<"begin of string received "<<endl;
         mes<<s;
         mes<<endl<<"end of string received";
      }
      cpmmessage(Word(mes.str()));
   }
   bool b=(out==MPI_SUCCESS);
   delete[] data;
   delete np;
   if (!b) cpmerror("error in CpmMPI::recStr");
   return b;
}

//////// some collective communication /////////////////////////////////
bool CpmMPI::Com::bcastStr(string& s)const
{
   if (mpiverbose>mL1){
      ostringstream mes;
      mes<<"bcastStr() started";
      cpmmessage(Word(mes.str()));
   }
   bool b=true;
   if (rank==1){
      for (int i=2; i<=size;i++){
         b=b&&sendStr(s,i,size+i);
      }
   }
   else{
      b=b&&recStr(s,1,size+rank);
   }
   if (!b) cpmerror("error in CpmMPI::bcastStr");
   if (mpiverbose>mL1){
      cpmmessage("bcastStr() done");
   }
   return b;
}

#else // !defined(CPM_USE_MPI)

void CpmMPI::initialize(int* a, char*** b){a;b;}

void CpmMPI::finalize(){}

#endif  // #if defined(CPM_USE_MPI)
