//? cpmnumbers.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmnumbers.h>
#include <cpmsystem.h>
#include <cpmsystemdependencies.h>
   // defines wrtPrc, wrtTit
#include <cpmmacros.h>
   // CPM_MA, CPM_MZ

using namespace CpmStd;
using CpmRoot::R;
using CpmRoot::Z;
using CpmRoot::N;
using CpmRoot::L;
using CpmRoot::Word;

#if defined(CPM_MP)
   Z CpmRoot::numPrc=CPM_MP;
#else
   Z CpmRoot::numPrc=0;
#endif

Z tenToTwo(Z p){ return (Z)(p*3.3219281 + 0.5);}
Z twoToTen(Z d){ return (Z)(d/3.3219281 + 0.5);}

void CpmRoot::prec(Z p)
{
;
}

Z CpmRoot::getPrec()
{
#if defined(CPM_MP)
   return CPM_MP;
#else
   return twoToTen(8*sizeof(R));
#endif
}

Z CpmRoot::getPrec(R const& r)
{
#if defined(CPM_MP)
   return CPM_MP;
#else
   return twoToTen(8*sizeof(R));
#endif
}

namespace{
// template tool for implementation
template <class T>
inline bool isValFunc(T x) // inline is needed
{
   if (cpmdbg>0){ // so calling this function can be switched
   // to idle from a central place.
      ostringstream ost;
      ost<<x;
      string s=ost.str();
      bool b1=s.find("I")==string::npos; // true if not found!
      bool b2=s.find("N")==string::npos;
      bool b3=s.find("n")==string::npos;
      //return b1||b2||b3; // corrected 2014-01-13
      return b1&&b2&&b3; // needs to give true for all options
   }
   else return true;
}

} // namespace

bool CpmRoot::isVal(L x){return true;}
bool CpmRoot::isVal(Z x){return isValFunc(x);}
bool CpmRoot::isVal(N x){return isValFunc(x);}

bool CpmRoot::isVal(R const& x)
{
#if defined(CPM_MP)
   if (cpmdbg>1){
      return !(isnan(x)||isinf(x));
   }
   else return true;
#else
   return isValFunc(x);
#endif
}

R CpmRoot::inv(R const& t)
{
   const Z maxMes=100;
   static Z mes=1;
   R z(0.);
   Word loc("inv(R): argument is 0");
   if (t==z){
      if (mes==maxMes){
         mes++;
         cpmmessage(loc&" ... messages discontinued");
      }
      if (mes<maxMes){
         mes++;
         cpmwarning(loc&", 0 returned");
      }
      return z;
   }
   R res=R(1.)/t;
   if (isVal(res)) return res;
   else {
      loc="inv(R): 1/argument is not valid";
      if (mes==maxMes){
         mes++;
         cpmmessage(loc&" ... messages discontinued");
      }
      if (mes<maxMes){
         mes++;
         cpmwarning(loc&", 0 returned");
      }
      return z;
   }
}

R CpmRoot::randomR(Z j)
{
   static N j0{137};
   static const R ranFac{1.e6};
   R x = j!=0 ? R(j) : R(++j0); // no overflow possible due to
      // using unsigned integers as N
   R y = cpmsin(x)*ranFac;
   return y -= cpmfloor(y);
}

R CpmRoot::ran(R const& r, Z j)
{
   R fac = randomR(j);
   return r*(fac*2.0 - 1.);
}

Z CpmRoot::hash(R const& r)
{
   const R reg=100000;
   R a=cpmabs(r);
   return cpmrnd(r+reg/(a*reg+1));
}

R CpmRoot::test(R const& r, Z complexity)
{ return R(1.123456789)*complexity;}

Z CpmRoot::toZ(R const& r, bool toZero)
{
   if (toZero){
      Z s = cpmsgn(r);
      R x = cpmfloor(cpmabs(r));
      Z xi;
#if defined(CPM_MP)
      xi=lround(x);
#else
      xi=(Z)x;
#endif
      return s*xi;
   }
   else{
#if defined(CPM_MP)
       return (Z)lround(cpmfloor(r));
#else
   return (Z)cpmfloor(r);
#endif
   }
}

R CpmRoot::dis(R const& r1, R const& r2)
{
   R a1 = cpmabs(r1);
   R a2 = cpmabs(r2);
   R a3 = cpmabs(r1-r2);
   return disDefFun(a1,a2,a3);
}

Z CpmRoot::randomZ(Z n, Z j)
{
   if (n<1) cpmerror("n<1 in CpmRoot::randomZ(Z n, Z j)");
   R y=n*randomR(j);
   Z res=toZ(y);
   return ((res==n) /* should not happen */ ? n : res+1);
}

bool CpmRoot::apprInt(R x, R tol)
{
   R d=x-rnd(x);
   return ( d>=0 ? d<=tol : -d<=tol);
}

namespace{
   L getbits(N x, N p, unsigned n)
   // Kernighan Ritchie: Programming in C, section 2.9
   { return (L)((x>>(p+1-n))& ~(~0<<n));}
}

L CpmRoot::getByte(Z z, Z n)
{
   N x=(N)z;
   return getbits(x, (n-1)*4,4);
}

#if defined(CPM_WRITE_PRECISION)
   Z CpmRoot::wrtPrc=CPM_WRITE_PRECISION;
#else
   Z CpmRoot::wrtPrc=10;
#endif

bool CpmRoot::writeTitle(string const& s , ostream& out)
{
   if (wrtTit) CpmRoot::write(string("//")+s,out);
   //return (out!=0);
   return !out ? false : true;
}

string CpmRoot::skipLeadingWhitespace(string const& str)
{
   Z n=(Z)str.size();
   if (n==0) return str;
   char cTest=str[0];
   if ( !isspace(cTest)) return str;
      // fast part of the process, no leading whitespace there
// now the first character is space
   Z i=0;
   while(i<n && isspace(str[i])) i++;
// eliminate the i leading characters
   string res=str;
   res.erase(0,i);
   return res;
}

string CpmRoot::skipTrailingWhitespace(string const& str)
   // made 2004-10-12 in analogy to the previous function
{
   Z n=(Z)str.size();
   if (n==0) return str;
   char cTest=str[n-1];
   if ( !isspace(cTest)) return str;
      // fast part of the process, no trailing whitespace there
// now the last character is space
   Z iTest=n-1;
   Z found=0;
   while(iTest>=0 && isspace(str[iTest])){
      iTest--;
      found++;
   }
   // now iTest is -1 or the index of a character
   // so iTest++ is the insex of the first of the trailing
   // whitespaces
   iTest++;
// eliminate the found trailing characters
   string res=str;
   res.erase(iTest,found);
   return res;
}

bool CpmRoot::startsComment(char c)
{ return c=='/' || c==';'|| c=='*'|| c=='#';}

bool CpmRoot::eatComments(istream& in)
{
   Z mL=3;
   Word loc("CpmRoot::eatComments(istream&)");
   CPM_MA
   char c=0;
   string buff;
   if (!in){
      cpmmessage(mL,loc&": bad stream before reading");
      CPM_MZ
      return false;
   }
LOOP:
   if (!(in>>c)){
      ostringstream ost;
      ost<<loc<<":  bad stream after reading character c="<<(Z)c;
      cpmmessage(mL,Word(ost.str()));
      CPM_MZ
      return false;
   }
   if (startsComment(c)){
      std::getline(in,buff); // reads the line till end of line
         // including the '\n' into the self-adjusting buff
      goto LOOP;
   }  // now c is no longer the potential beginning of a comment
   in.putback(c);
   if (!in){
      cpmmessage(mL,loc&": bad stream after reading");
      CPM_MZ
      return false;
   }
   CPM_MZ
   return true;
}

bool CpmRoot::strVal(istream& str)
{ if (!str||str.bad()) return false; else return str.good(); }

bool CpmRoot::strVal(ostream& str)
{ if (!str||str.bad()) return false; else return str.good(); }

bool CpmRoot::readLine(string& line , istream& in)
{
   if  (!eatComments(in)) return false;
   std::getline(in,line);
   line=skipTrailingWhitespace(line);
   return in ? true : false;
}

template <>
bool CpmRoot::read(L& x, istream& in)
{
   Z z;
   bool suc=CpmRoot::read<Z>(z,in);
   x=(L)z;
   return suc;
}

template <>
bool CpmRoot::write(L const& x, ostream& out)
{
   out<<(Z)x<<endl;
   return strVal(out);
}

template <>
bool CpmRoot::write(R const& r, ostream& out)
// experimental avoidance of writing too small numbers
// which caused trouble under Windows
{
   const R wrTiny=1e-304;
      // 1e-305 works, 1e-310 works not
   out.precision(wrtPrc);
   R x=r;
   if (x>0){
      if (x<wrTiny) x=wrTiny;
   }
   else if (x<0){
      if (x>-wrTiny) x=-wrTiny;
   }
   else{
      x=0;
   }
   out<<x<<endl; return strVal(out);
}

N CpmRoot::operator""_N(unsigned long long int n){ return (N)n; }
Z CpmRoot::operator""_Z(unsigned long long int z){ return (Z)z; }
R CpmRoot::operator""_R(long double r){ return (R)r; }
R CpmRoot::operator""_R(unsigned long long int r){ return (R)r; }

