//? cpmtypes.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <stdlib.h>
#include <sstream>

#include <cpmtypes.h>

using namespace CpmSystem;
using namespace CpmRoot;
using CpmArrays::X2;

//////////////////////////////////////////////////////////////////

void CpmRootX::copyTextFile(ifstream& from, ofstream& to)
{
   char ch;
   while (from.get(ch)) to.put(ch);
   if (!from.eof() || to.bad())
   cpmerror("copyTextFile(const ifstream&, ofstream& to) failed");
     // Stroustrup Section 10.5 p. 351 top of page
}

void CpmRootX::copyTextFile(const Word& nameOfSource,
   const Word& nameOfCopy)
{
   Z mL=1;
   Word loc("copying "&nameOfSource&" to "&nameOfCopy);
   cpmassert(nameOfSource!="cpmcerr.txt",loc);
      // copying "cpmcerr.txt" caused problems, which are hard to fully
      // understand
   cpmmessage(mL,loc&" started");
   IFileStream from(nameOfSource);
   OFileStream to(nameOfCopy);
   CpmRootX::copyTextFile(from(),to());
   cpmmessage(mL,loc&" done");
}

string CpmRootX::toString(ifstream& from)
{
   if (!from) cpmerror("toString(): cannot open input file");
   ostringstream to;
   if (!to) cpmerror("toString(): cannot create output file stream");
   char ch;
   while(from.get(ch)) to.put(ch);
   if (!from.eof() || !to) cpmerror("failure in toString");
   return to.str();
}

///////////////////////// class Z1 ////////////////////////////////

const Z maxZ(std::numeric_limits<Z>::max());
const Z minZ(std::numeric_limits<Z>::min());

Z CpmRootX::sig(Z i)
{
   if (i==0) return Z(0);
   else if (i>0) return Z(1);
   else return Z(-1);
}

R CpmRootX::sig(R x)
{
   if (x==R(0.)) return R(0.);
   else if (x>R(0.)) return R(1.);
   else return R(-1.);
}

R CpmRootX::krn(Z i, Z j)
{
   return i==j ? R(1.) : R(0.);
}

Z CpmRootX::parity(Z k)
{
   return k%2==0 ? Z(1) : Z(-1);
}

Z CpmRootX::fac(Z n)
{
   cpmassert(n>0,"Z fac(Z)");
   Z res=1;
   for (Z i=1; i<=n; i++) res*=i;
   return res;
}

Z CpmRootX::pow(Z const& x, Z p)
{
   cpmassert(p>=0,"Z powerZ(Z,Z)");
   if (p==0) return Z(0);
   else if (p==1) return x;
   else if (p==2) return x*x;
   else{
      Z res=x;
      Z nMult=p-1;
      for (Z i=1;i<=nMult;i++) res*=x;
      return res;
   }
}

bool CpmRootX::isPowOfTwo(Z const& x)
{
   Z fac{2};
   if (x<fac) return false;
   Z f=fac;
   while (f<x) f*=fac;
   return f==x;
}

Z CpmRootX::nextPowOfTwo(Z const& x)
{
   if (x<2) return Z(2);
   Z f=2;
   while (f<x) f*=2;
   return Z(f);
}

Z CpmRootX::nextLog2(Z const& x)
{
   Z res{1};
   Z f{2};
   while (f<x){ // after the loop f>=x1
      res++;
      f*=2; // guarantees that f grows over all bounderies
   }
   return res;
}

#if !defined(CPM_MP) && !defined(CPM_QUAD)
///////////////////////// class R1 ////////////////////////////////

using CpmRootX::R1;
using CpmRootX::Z1;

const R1 R1::max(std::numeric_limits<R>::max());
const R1 R1::eps1(std::numeric_limits<R>::epsilon());
const R1 R1::eps0(std::numeric_limits<R>::min());

const R1 R1::min(-R1::max);
const R1 R1::logmax(::log(R1::max));
const R1 R1::log10max(::log10(R1::max));
const R1 R1::pi(cpmpi);
const R1 R1::piInv(R(1)/cpmpi);

bool R1::prnOn(ostream& out)const
{
   return CpmRoot::write(x1,out);
}

bool R1::scanFrom(istream& in)
{
   return CpmRoot::read(x1,in);
}

Z R1::com(R1 a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   return 0;
}

X2<R,R1 > R1::polDec()const
{
   if (x1>=0) return X2<R,R1 >(x1,R1(1.));
   else return X2<R,R1 >(-x1,R1(-1.));
}

R R1::dis(const R1& x)const
{
   return CpmRoot::dis(x1,x.x1);
}

// functions referring to R as a discrete 'lattice'

R1& R1::next_(void)
{
   R eps_=eps0.x1;
   R xMem=x1;
   while (x1==xMem){
      x1+=eps_;
      eps_*=2;
   }
   return *this;
}

R1& R1::prv_(void)
{
   R eps_=eps0.x1;
   R xMem=x1;
   while (x1==xMem){
      x1-=eps_;
      eps_*=2;
   }
   return *this;
}

namespace{

   void step_p(R& x, R& e)
   {
      R xMem=x;
      Label:
      x+=e;
      if (x==xMem){
         e*=2;
         goto Label;
      } // thus we end up with x!=xMem
   }

   void step_m(R& x, R& e)
   {
      R xMem=x;
      Label:
      x-=e;
      if (x==xMem){
         e*=2;
         goto Label;
      } // thus we end up with x!=xMem
   }
}

Z R1::dz(R a)const
{
   Z res=0;
   R e=R1::eps0.x1;
   if (x1>=a){
      while (x1>a){
         step_p(a,e);
         ++res;
      }
   }
   else{
      while (x1<a){
         step_m(a,e);
         --res;
      }
   }
   return res;
}

R R1::eps(Z dir)const
{
   R res= (x1>=1 || x1<=-1) ? eps1.x1 : eps0.x1;
   if (dir<0) res*=-1;
   R xc=x1;
   Label:
   xc+=res;
   if (xc==x1){
      res*=2;
      goto Label;
   }
   return res;
}

R R1::dr(R a)const
{
   if (a==x1) return 0;
   R dis=x1-a;
   if (dis<0) dis*=-1;
   if (x1>a){
      R ex=eps(-1);
      R ea=R1(a).eps(1);
      return R(2)*dis/(ea-ex);
   }
   else{
      R ex=eps(1);
      R ea=R1(a).eps(-1);
      return R(2)*dis/(ex-ea);
   }
}

Z R1::run(R final, ostream& str, Z writePeriod)const
{
   R xMem=x1;
   R xRun=x1;
   R e=R1::eps0.x1;
   Z wrt=0;
   Z count=0;
   if (final>xRun){
      while(xRun<final){
         step_p(xRun,e);
         count++;
         if (count==writePeriod){
            str<<"xRun-xMem="<<xRun-xMem<<" e="<<e<<endl;
            ++wrt;
            count=0;
         }
      }
   }
   else if (final<xRun){
      while(xRun>final){
         step_m(xRun,e);
         count++;
         if (count==writePeriod){
            str<<"xRun-xMem="<<xRun-xMem<<" e="<<e<<endl;
            ++wrt;
            count=0;
         }
      }
   }
   else;
   return wrt;
}

// save functions

R R1::log()const
{
   if (x1>0) return ::log(x1);
   else if (x1<0) return ::log(-x1);
   else return -logmax;
}

R R1::log10()const
{
   if (x1>0) return ::log10(x1);
   else if (x1<0) return ::log10(-x1);
   else return -log10max;
}

R R1::exp()const
{
   if (x1>logmax) return max;
   else if (x1<-logmax) return 0;
   else return ::exp(x1);
}

R1 R1::inv()const
{
   R e0=eps0;
   if (x1>e0 || x1<-e0) return R1(1./x1);
   else if (x1>=0) return R1(max);
   else return R1(min);
}

using CpmRootX::Z1;
///////////////////////// class Z1 ////////////////////////////////

const Z1 Z1::max(std::numeric_limits<Z>::max());
const Z1 Z1::min(std::numeric_limits<Z>::min());

Z1 Z1::pow(Z p)const
{
   cpmassert(p>=0,"Z1 Z1::power(Z p)");
   if (p==0) return Z1();
   else if (p==1) return Z1(x1);
   else if (p==2) return Z1(x1*x1);
   else{
      Z res=x1;
      Z nMult=p-1;
      for (Z i=1;i<=nMult;i++) res*=x1;
      return Z1(res);
   }
}

bool Z1::prnOn(ostream& out)const
{
   return CpmRoot::write(x1,out);
}

bool Z1::scanFrom(istream& in)
{
   return CpmRoot::read(x1,in);
}

Z Z1::com(const Z1& a)const
{
   if (x1<a.x1) return 1;
   if (x1>a.x1) return -1;
   return 0;
}

X2<Z,Z1 > Z1::polDec()const
{
   if (x1>=0) return X2<Z,Z1 >(x1,Z1(1));
   else return X2<Z,Z1 >(-x1,Z1(-1));
}

R Z1::dis(const Z1& x)const
{
   return CpmRoot::disT<Z>(x1,x.x1);
}

bool Z1::isPowOfTwo()const
      // Let us write x for *this, and let N={1,2,3,...} (no zero!).
      // We return true iff x=2^n with some n \in N
{
   if (x1<2) return false;
   Z f=2;
   while (f<x1) f*=2;
   return f==x1;
}

Z1 Z1::nextPowOfTwo()const
{
   if (x1<2) return Z1(2);
   Z f=2;
   while (f<x1) f*=2;
   return Z1(f);
}

Z1 Z1::nextLog2()const
{
   Z res=1;
   Z f=2;
   while (f<x1){ // after the loop f>=x1
      res++;
      f*=2; // guarantees that f grows over all bounderies
   }
   return Z1(res);
}

#endif

