//? cpmalgorithms.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_ALF_H
#define CPM_ALF_H
/*
   Description: This namespace will be populated by functions which
    express algorithms of general interest, which don't fit well into
    one of the classes already introduced.
*/
#include <cpmx.h>
#include <cpmvo.h>
#include <cpminterval.h>
#include <cpmf.h>
#include <cpmrvector.h>
#include <cpmcmatrix.h>
#include <cpmrecordhandler.h>
#include <concepts>
#include <functional>

namespace CpmAlgorithms{

   using namespace CpmStd;
   using namespace CpmRoot;

   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::Vo;
   using CpmArrays::X2;
   using CpmArrays::X3;
   using CpmArrays::R_Vector;
   using CpmArrays::C_Matrix;
   using CpmFunctions::FncObj;
   using CpmFunctions::F;
#ifdef CPM_Fn
   using CpmFunctions::F2;
   using CpmFunctions::F3;
#endif
   using CpmGeo::Iv;
   using CpmLinAlg::Z2;
   using CpmRootX::RecordHandler;

////////////////////// class ALF /////////////////////////////////////////

using std::function;

// concepts which describe the minimal requiremsnts for the template
// arguments of ALF

template<typename T>
concept rLinear =
requires(T t1, T t2)
{
   {t1+t2};
   {t1-t2};
   {t1+=t2};
}
&& requires(T t, R r)
{
   {t*r};
};

template<typename T>
concept measurable =
requires(T t)
{
   {t.abs()};
};

template<typename T>
concept hilbert =
requires(T t1, T t2, C z, R r)
{
   {t1+t2};
   {t1-t2};
   {t1+=t2};
   {t1*z};
   {t1*r};
   { t1|t2};
};

template<typename X, typename Y>
concept affin =
requires(X x, Y y)
{
   {x+=y};
};

class StpCon{ // step control
public:
// independent quantities
   R dt0_{1.e-3};
      // initial value for dt

   R crit_{1.e-3};
      // a critial value for the internal quantity 'kink'. The algorithm
      // depends on kink > (a1 = crit) or kink < (a2 = crit*red)

   R red_{0.5};
      // reduction factor, 0 <= red <= 1. crit*red is a second critical
      // value for kink.
   R frac_{0.2};
      // A factor for increasing the step size is written as 1+frac
      // and one for decreasing the step size is written as 1-frac .
      // The case frac = 0
   B autoStep_{B(1)};
   B dalf_{B(1)};
   Z docPeriod_{100};
   R tw1_{0_R};
   R tw2_{0_R};
   R tw3_{0_R};
   R tf_{0_R};
// dependent quantities, computed by function ini_()
   R a1_, a2_, f1_, f2_;

   void ini_(){
      a1_=crit_;
      a2_=a1_*red_;
      f1_=1_R-frac_;
      f2_=1_R+frac_;
   }

   StpCon(R dt0, R crit, R red, R frac ): dt0_(dt0),
      crit_(crit), red_(red), frac_(frac){ ini_();}
   StpCon(){ini_();}
//   StpCon(RecordHandler const& rch, Word const& sec)
//      {
//         cpmrhf(dt0_);
//         cpmrhf(act_);
//         cpmrhf(crit_);
//         cpmrhf(red_);
//         cpmrhf(frac_);
//         ini_();
//      }

   R dt0()const{ return dt0_;}
   R crit()const{ return crit_;}
   R red()const{ return red_;}
   R a1()const{ return a1_;}
   R a2()const{ return a2_;}
   R f1()const{ return f1_;}
   R f2()const{ return f2_;}
   B autoStep()const{ return autoStep_;}
   B dalf()const{ return dalf_;}
   R tw1()const{ return tw1_;}
   R tw2()const{ return tw2_;}
   R tw3()const{ return tw3_;}
   Z docPeriod()const{ return docPeriod_;}
   R tf()const{ return tf_;}

   void set_dt0_(R const& dtNew){ dt0_=dtNew; ini_();}
   void set_crit_(R const& critNew){ crit_=critNew; ini_();}
   void set_red_(R const& redNew){ red_=redNew; ini_();}
   void set_frac_(R const& fracNew){ frac_=fracNew; ini_();}
   void set_autoStep_(B const& autoStepNew){ autoStep_=autoStepNew; ini_();}
   void set_dalf_(B const& dalfNew){ dalf_=dalfNew; ini_();}
   void set_tw1_(R const& tw1New){ tw1_=tw1New; ini_();}
   void set_tw2_(R const& tw2New){ tw2_=tw2New; ini_();}
   void set_tw3_(R const& tw3New){ tw3_=tw3New; ini_();}
   void set_docPeriod_(Z const& docPeriodNew){ docPeriod_=docPeriodNew; ini_();}
   void set_tf_(R const& tfNew){ tf_=tfNew; ini_();}
};

//template<typename X, rLinear Y>
//requires affin<X,Y>
//class DALFx{
//
//   R t_;
//   X x_;
//   function<Y(R,X)> dot_;
//   Y y_;
//
//public:
//   DALFx(R const& t, X const& x, function<Y(R,X)> const& dot):
//      t_(t),x_(x),dot_(dot),y_(dot(t_,x_)){}
//
//   void step_(R const& dt)
//   //: (time) step
//   {
//      R dth=dt*0.5_R;  // half of dt
//      R dtq=dth*0.5_R; // quarter of dt
//      t_+=dtq;
//      x_+=y_*dtq;
//      y_+=(dot_(t_,x_)-y_)*2_R;
//      t_+=dth;
//      x_+=y_*dth;
//      y_+=(dot_(t_,x_)-y_)*2_R;
//      x_+=y_*dtq;
//      t_+=dtq;
//   }
//
//   void evo_(R const& t, R const& dt)
//   //: evolution (from present state *this as initial state)
//   // t: time span for evolution, dt: tendative time step. The time step
//   // dtn in actual use is such that t = dtn * n, where n is a natural
//   // number
//   {
//      R ti=t_;
//      R tf=ti+t;
//      Z n = cpmtoz(t/dt);
//      R dtn=t/R(n); // actual step size
//      for (Z i=1;i<=n;++i){
//         step_(dtn);
//      }
//   }
//
//   void evo_(R const& t, R const& dt)
//
//   R t()const{ return t_;}
//   X x()const{ return x_;}
//   Y y()const{ return y_;}
//};
//

// The natural idea to make also dot a template argument:
// template<typename X, rLinear Y, function<Y(R,X)> dot>
// does not work since typed template arguments need to be typed
// 'structural' which is a subtle restriction which turns out to be not
// satisfied in the present case.

// Version in which dot is a data member.
// minimal environment for the ALF evolution algorithm
template<typename X, typename Y>
//requires affin<X,Y> && rLinear<Y> && measurable<Y>
requires affin<X,Y> && hilbert<Y>
class ALF{

   StpCon sc_;
   R dt_;
   R t_;
   X x_;
   Y y_;
   V<Y> b_; // basis in Y
   function<Y(R,X)> dot_;

public:
   ALF(StpCon const& sc, R const& t, X const& x,
      function<Y(R,X)> const& dot): sc_(sc), dt_(sc_.dt0()),
      t_(t), x_(x), dot_(dot), y_(dot(t_,x_)){
         Graph gr;
         x_.mark(gr);
         gr.vis();
         cpmwait(0.,2);
      }

   ALF(){}

   void steph0_( )
   //: (time) half step
   {
      R dt=dt_*0.5;
      R dth=dt*0.5_R;  // half of dt
      t_+=dth;
      x_+=y_*dth;
      Y dy=(dot_(t_,x_)-y_)*2_R;
      y_+=dy;
      x_+=y_*dth;
      t_+=dth;
   }

   void step0_( )
   //: (time) half step
   {
      steph0_( );
      steph0_( );
   }

   void step_()
   //: (time) step
   {

      B autoStep=sc_.autoStep();
      B dalf=sc_.dalf();
      R n1,n2,n12;
      ALF stateMem;
      if (autoStep) stateMem=*this;
         // may be needed if the step ahead turns out to be unvalid
      if (dalf){
         Y dy12;
         R dtq=dt_*0.25_R;  // quarter of dt_
         R dth=dt_*0.5_R; // half of dt_
         t_+=dtq;
         if (autoStep) n1=y_.abs();
         x_+=y_*dtq;
         Y dy=(dot_(t_,x_)-y_)*2_R;
         if (autoStep) dy12=dy;
         y_+=dy;
         x_+=y_*dth;
         t_+=dth;
         dy=(dot_(t_,x_)-y_)*2_R;
         if (autoStep) dy12+=dy;
         if (autoStep) n12=dy12.abs();
         y_+=dy;
         if (autoStep) n2=y_.abs();
         x_+=y_*dtq;
         t_+=dtq;
      }
      else{
         R dth=dt_*0.5_R;  // half of dt
         t_+=dth;
         if (autoStep) n1=y_.abs();
         x_+=y_*dth;
         Y dy=(dot_(t_,x_)-y_)*2_R;
         if (autoStep) n12=dy.abs();
         y_+=dy;
         if (autoStep) n2=y_.abs();
         x_+=y_*dth;
         t_+=dth;
      }
      if (!autoStep){ // running with constant time step
         return;
      }
      else{ // running with adaptive time step
         R kink=n12*cpminv(n1+n2);
         if (kink > sc_.a1()){ // need to redo the step (i.e. go back to stateMem) with smaller step width
            t_=stateMem.t();
            x_=stateMem.x();
            y_=stateMem.y();
            dt_*=sc_.f1();// f1 = 1-frac
            cout<<" kink1 = "<<kink<<", dt_ = "<<dt_<<endl;
            return; // the next step then will possibly be better
         }
         if (kink < sc_.a2()){ // no need to go back to stateMem
            dt_*=sc_.f2();// f2 = 1*frac
            cout<<" kink2 = "<<kink<<", dt_ = "<<dt_<<endl;
            return;
         }
      }
   }

//   void step_( )
//   //:
//   {
//      steph0_( );
//      steph0_( );
//   }

//   void step_() // method no DALF option
//   //: (time) step
//   {
//      ALF stateMem;
//      if (sc_.act()>0_Z) stateMem=*this;
//         // may be needed if the step ahead turns out to be unvalid
//      R dth=dt_*0.5_R;  // half of dt
//      t_+=dth;
//      R n1=y_.abs();
//      x_+=y_*dth;
//      Y dy=(dot_(t_,x_)-y_)*2_R;
//      R n12=dy.abs();
//      y_+=dy;
//      R n2=y_.abs();
//      x_+=y_*dth;
//      t_+=dth;
//      if (sc_.act()==0_Z){ // running with constant time step
//         return;
//      }
//      else{ // running with adaptive time step
//         R kink=n12*cpminv(n1+n2);
//         if (kink > sc_.a1()){ // need to redo the step (i.e. go back to stateMem) with smaller step width
//            t_=stateMem.t();
//            x_=stateMem.x();
//            y_=stateMem.y();
//            dt_*=sc_.f1();// f1 = 1-frac
//            return; // the next step then will possicly be better
//         }
//         if (kink < sc_.a2()){ // no need to go back to stateMem
//            dt_*=sc_.f2();// f2 = 1*frac
//            return;
//         }
//      }
//   }


   R abs()const{ return x().abs() + y().abs();}

   void evo_(R const& tf)
   //: evolution
   // time evolution from ti=t() to tf
   {
      Z clock=0;
      Graph gr;
      Z imgPer=1000;
      while (t_<tf){
         step_();
         clock++;
         if (clock%imgPer==0){
            cpmvalues3("t_, dt_, tf",t_,dt_,tf,3);
            x().mark(gr);
            gr.vis();
         }
      }
   }

   void evoCQ_()
   //: evolution + conserved quantities
   // time evolution from ti=t() to sc_.tf_
   {
      Z docPeriod=sc_.docPeriod();

      R tf=sc_.tf_;
      R tw1=sc_.tw1_;
      R tw2=sc_.tw2_;
      R tw3=sc_.tw3_;
      dt_=sc_.dt0_;
      C e0=(x_|y_)*C::I;
      R n0=(x_|x_)[1];

      Graph gr;
      vector<R> vt;
      vector<R> vdt;
      vector<R> vdn;
      vector<R> vder;
      vector<R> vdei;
      R dn=(x_|x_)[1]-n0;
      C de=(x_|y_)*C::I - e0;
      vt.push_back(t_);
      vdt.push_back(dt_);
      vdn.push_back(dn);
      vder.push_back(de[1]);
      vdei.push_back(de[2]);
      Z clock=0;
      while (t_<tf){
         step_();
         clock++;
         if (clock%docPeriod==0){
            cpmvalues3("t_, dt_, tf",t_,dt_,tf,3);
            dn=(x_|x_)[1]-n0;
            de=(x_|y_)*C::I-e0;
            vt.push_back(t_);
            vdt.push_back(dt_);
            vdn.push_back(dn);
            vder.push_back(de[1]);
            vdei.push_back(de[2]);
            x().mark(gr);
            gr.vis();
         }
      }
      Z n=vt.size();
      R_Vector rvt(n);
      R_Vector rvdt(n);
      R_Vector rvdn(n);
      R_Matrix rmde(2,n);
      Z i,j;
      for (i=1,j=0; i<=n; ++i,++j){
         rvt[i]=vt[j];
         rvdt[i]=vdt[j];
         rvdn[i]=vdn[j];
         rmde[1][i]=vder[j];
         rmde[2][i]=vdei[j];
      }

      gr.mark(rvt,rvdt);
      gr.vis();
      cpmwait(tw1,2);

      gr.mark(rvt,rvdn);
      gr.vis();
      cpmwait(tw2,2);

      gr.mark(rvt,rmde);
      gr.vis();
      cpmwait(tw3,2);
   }

   void safeEvo_(R const& tSpan)
      //: safe evolution
      // final evolution of initia state *this over a
      // time span of duration tSpan
   {
      auto absMax=1.e12_R;
      R tf=t()+tSpan;
      R tDone=0_R;
      ALF s0=*this;
      auto clock=1_Z;
      cout<<"dt_ = "<<dt_<<endl;
      while (tDone < tf){
         dt_*=0.5_R;
         cout<<"dt_ = "<<dt_<<endl;
         while (abs()< absMax && t_< tf){
            step_();
         }
         if (t_< tf){
            tDone=t_;
            R dtMem=dt_;
            cout<<"tDone = "<<tDone<<endl;
            cpmvalues3("tDone, tSpan, dt_",tDone,tSpan,dt_,2);
            cpmvalue("tDone",tDone,3);
            *this=s0;
            dt_=dtMem;

         }
      }
   }

   void safeEvo2_(R const& tSpan)
      //: safe evolution
      // final evolution of initia state *this over a
      // time span of duration tSpan
   {
      Word loc("safeEvo2_(R)");
      Z mL=1;
      CPM_MA
      Graph gr;
      auto absMax=1.e12_R;
      R tf=t()+tSpan;
      R tDone=0_R;
      ALF s0=*this;
      Z clock=0;
      while (tDone < tf){
         while (abs()< absMax && t_< tf){
            step_();
            clock++;
            if (clock%1000==0){
               cpmvalues3("t_, dt_, tf",t_,dt_,tf,3);
               x().mark(gr);
               gr.vis();
            }
         }
//         if (t_< tf){
//            tDone=t_;
//            R critMem=sc_.crit_;
//           // cout<<"tDone = "<<tDone<<endl;
//            cpmvalues3("tDone, tSpan, crit_",tDone,tSpan,sc_.crit_,2);
//            *this=s0;
//            sc_.crit_=critMem;
//         }
      }
      cpmcerr<<"while loop left"<<endl;
      gr.clr_();
      CPM_MZ
      return;
   }

   StpCon sc()const{ return sc_;}
   R dt()const{ return dt_;}
   void set_dt_(R const dt){ dt_=dt;}
   R t()const{ return t_;}
   X const& x()const{ return x_;}
   X& x(){ return x_;}
   Y const& y()const{ return y_;}
   Y& y(){ return y_;}

   C_Matrix jac(R eps)const{
      Z n=b_.dim();
      C_Matrix res(n,n);
      for (Z i=1;i<=n;++i){
         V<Y> dyi=f(t_, x_+eps*b_[i])-f(t_,x_);
         for (Z j=1;j<=n;++j){
            res[i][j] = (dyi|b_[j]);
         }
      }
      return res;
   }

};

///////////////////////////// test and application:///////////////////////
///// the finite dimensional quantum mechanical harmonic oscillator //////

// prepare the data needed to create ALF<_Vector,C_Vector> which is an
// harmonic oscillator.

//Z n=200;
//R L=2;
//R m=1.;
//R frac=0.125;
//R_Vector pos=R_Vector(Iv(-0.5*L,0.5*L).centers(n));
//R hw=frac*L; // half width
//R omega=R(1.)/(m*hw*hw);
//R beta=R(0.5)*m*omega*omega;
//C_Matrix pot(n);
//for (Z i=1; i<=n;++i){ R xi=pos[i]; pot[i][i]=beta*xi*xi;}
//R d=L/R(n);
//C_Matrix lapl(n_,"Delta",CYCLIC);
//C_Matrix ham=lapl*(R(-1.)/(R(2.)*d*d*m)) + pot;
//C_Matrix dotMat=C::Ibar*ham;
//function<C_Vector(R,C_Vector)> dot=[&,dotMat](R t, C_Vector const& state){ t;return dotMat*state;};
//StpCon sc();
//R t=0.;
//SpcObj so(ham);
////R_Vector spec=so.val();
//C_Vector x1=so.vec()[1];
//
//
//ALF<C_Vector,C_Vector> ho(sc,t,x1,dot);
//ho.safeEvo2_(0.1);


//class HarmOsc: public ALF<C_Vector,C_Vector>{
//
//
//   N n_; // number of discrete positions in biotope
//   R L_;  // length of biotope
//   R m_;  // mass of the oscillator particle
//   R frac_; // half-with of the ground state as fraction of L_
//   R_Vector pos_; // the discrete positions
//   R hw=frac*L; // half width
//   R omega=R(1.)/(m*hw*hw);
//      R beta=R(0.5)*m*omega*omega;
//      C_Matrix pot(n_);
//      for (Z i=1; i<=n_;++i){ R xi=pos_[i]; pot[i][i]=beta*xi*xi;}
//      R d=L_/R(n_);
//      C_Matrix lapl(n_,"Delta",CYCLIC);
//      C_Matrix ham=lapl*(R(-1.)/(R(2.)*d*d*m_)) + pot;
//      C_Matrix dotMat=C::Ibar*ham;
//
//public:
//
//   HarmOsc(N const& n, R const& L, R const& m, R const& frac, Z const& level, StpCon const& sc):
//      n_{n}, L_{L}, m_{m}, frac_{frac},
//      pos_{R_Vector(Iv(-0.5*L_,0.5*L_).centers(n_))}
//   {
//      R hw=frac_*L_; // half width
//      R omega=R(1.)/(m*hw*hw);
//      R beta=R(0.5)*m*omega*omega;
//      C_Matrix pot(n_);
//      for (Z i=1; i<=n_;++i){ R xi=pos_[i]; pot[i][i]=beta*xi*xi;}
//      R d=L_/R(n_);
//      C_Matrix lapl(n_,"Delta",CYCLIC);
//      C_Matrix ham=lapl*(R(-1.)/(R(2.)*d*d*m_)) + pot;
//      C_Matrix dotMat=C::Ibar*ham;
//      SpcObj so(ham);
//      R_Vector spec=so.val();
//      x_=so.vec()[level];
//
//      dot_=[&,dotMat](R t, C_Vector const& state){ t;return dotMat*state;};
//         // function = lambda; is the correct initialization of std::function
//      y_= dot_(t_,x_);
//
//      sc_=sc;
//   }
//
//    virtual ALF<C_Vector,C_Vector>* clone(void)const{ return new HarmOsc(*this);}
//
//};



} // namespace

#endif
