//? cpmalgorithms.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_ALF_H
#define CPM_ALF_H
/*
   Description: This namespace will be populated by functions which
    express algorithms of general interest, which don't fit well into
    one of the classes already introduced.
*/
#include <cpmx.h>
#include <cpmvo.h>
#include <cpminterval.h>
#include <cpmf.h>
#include <cpmrvector.h>
#include <cpmcmatrix.h>
#include <cpmrecordhandler.h>
#include <concepts>
#include <functional>

namespace CpmAlgorithms{

   using namespace CpmStd;
   using namespace CpmRoot;
   using namespace CpmRootX;

   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::Vo;
   using CpmArrays::X2;
   using CpmArrays::X3;
   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;
   using CpmArrays::C_Matrix;
   using CpmFunctions::FncObj;
   using CpmFunctions::F;
   using CpmGraphics::Graph;
   using std::vector;
#ifdef CPM_Fn
   using CpmFunctions::F2;
   using CpmFunctions::F3;
#endif
   using CpmGeo::Iv;
   using CpmLinAlg::Z2;
   using CpmRootX::RecordHandler;

////////////////////// class ALF /////////////////////////////////////////

using std::function;

// concepts which describe the minimal requiremsnts for the template
// arguments of ALF

template<typename T>
concept rLinear =
requires(T t1, T t2)
{
   {t1+t2};
   {t1-t2};
   {t1+=t2};
}
&& requires(T t, R r)
{
   {t*r};
};

template<typename T>
concept measurable =
requires(T t)
{
   {t.abs()};
};

template<typename T>
concept hilbert =
requires(T t1, T t2, C z, R r)
{
   {t1+t2};
   {t1-t2};
   {t1+=t2};
   {t1*z};
   {t1*r};
   { t1|t2};
};

template<typename X, typename Y>
concept affin =
requires(X x, Y y)
{
   {x+=y};
};

class StpCon{ // step control and run control
public:
// independent quantities
   R dt0_{1.e-3};
      // initial value for dt

   R crit_{1.e-3};
      // a critial value for the internal quantity 'kink'. The algorithm
      // depends on kink > (a1 = crit) or kink < (a2 = crit*red)

   R red_{0.5};
      // reduction factor, 0 <= red <= 1. crit*red is a second critical
      // value for kink.
   R frac_{0.2};
      // A factor for increasing the step size is written as 1+frac
      // and one for decreasing the step size is written as 1-frac .
      // The case frac = 0
   B autoStep_{B(1)};
   B dalf_{B(0)};
   B adalf_{B(1)};
   Z docPeriod_{100};
   R tw1_{0_R};
   R tw2_{0_R};
   R tw3_{0_R};
   R tf_{0_R};
   B writeToFiles_{B(0)};
   B writeMovie_{B(0)};
   B noDiff_{B(0)};
// dependent quantities, computed by function ini_()
   R a1_, a2_, f1_, f2_;

   void ini_(){
      a1_=crit_;
      a2_=a1_*red_;
      f1_=1_R-frac_;
      f2_=1_R+frac_;
   }

   StpCon(R dt0, R crit, R red, R frac ): dt0_(dt0),
      crit_(crit), red_(red), frac_(frac){ ini_();}
   StpCon(){ini_();}

   R dt0()const{ return dt0_;}
   R crit()const{ return crit_;}
   R red()const{ return red_;}
   R a1()const{ return a1_;}
   R a2()const{ return a2_;}
   R f1()const{ return f1_;}
   R f2()const{ return f2_;}
   B autoStep()const{ return autoStep_;}
   B dalf()const{ return dalf_;}
   B adalf()const{ return adalf_;}
   R tw1()const{ return tw1_;}
   R tw2()const{ return tw2_;}
   R tw3()const{ return tw3_;}
   Z docPeriod()const{ return docPeriod_;}
   R tf()const{ return tf_;}
   B writeToFiles()const{ return writeToFiles_; }
   B writeMovie()const{ return writeMovie_; }
   B noDiff()const{ return noDiff_; }

   void set_dt0_(R const& dtNew){ dt0_=dtNew; ini_();}
   void set_crit_(R const& critNew){ crit_=critNew; ini_();}
   void set_red_(R const& redNew){ red_=redNew; ini_();}
   void set_frac_(R const& fracNew){ frac_=fracNew; ini_();}
   void set_autoStep_(B const& autoStepNew){ autoStep_=autoStepNew; ini_();}
   void set_dalf_(B const& dalfNew){ dalf_=dalfNew; ini_();}
   void set_adalf_(B const& adalfNew){ adalf_=adalfNew; ini_();}
   void set_tw1_(R const& tw1New){ tw1_=tw1New;}
   void set_tw2_(R const& tw2New){ tw2_=tw2New;}
   void set_tw3_(R const& tw3New){ tw3_=tw3New;}
   void set_docPeriod_(Z const& docPeriodNew){ docPeriod_=docPeriodNew; ini_();}
   void set_tf_(R const& tfNew){ tf_=tfNew; ini_();}
   void set_writeToFiles_(B const& writeToFilesNew)
      { writeToFiles_=writeToFilesNew;}
   void set_writeMovie_(B const& writeMovieNew)
      { writeMovie_=writeMovieNew;}
   void set_noDiff_(B const& noDiffNew){ noDiff_=noDiffNew;}
};

// The natural idea to make also dot a template argument:
// template<typename X, rLinear Y, function<Y(R,X)> dot>
// does not work since typed template arguments need to be typed
// 'structural' which is a subtle restriction which turns out to be not
// satisfied in the present case.

// Version in which dot is a data member.
// minimal environment for the ALF evolution algorithm
template<typename X, typename Y>
requires affin<X,Y> && hilbert<Y>
class ALF{

   StpCon sc_;
   R dt_;
   R t_;
   X x_;
   Y y_;
   V<Y> b_; // basis in Y, not needed so far
   function<Y(R,X)> dot_;

public:
   ALF(StpCon const& sc, R const& t, X const& x,
      function<Y(R,X)> const& dot): sc_(sc), dt_(sc_.dt0()),
      t_(t), x_(x), dot_(dot), y_(dot(t_,x_)){
//         Graph gr;
//         x_.mark(gr);
//         gr.vis();
//         cpmwait(0.,2);
      }

   ALF(){}

   void step_()
   //: (time) step
   {
      bool autoStep=sc_.autoStep_;
      bool dalf=sc_.dalf_;
      bool adalf=sc_.adalf_;
      R n1,n2,n12,kink1,kink2;
      ALF stateMem;
      Y ydot,dy,y1;
      auto tiny=R(1.e-12);
      if (autoStep) stateMem=*this;
         // may be needed if the step ahead turns out to be unvalid
      if (dalf || adalf){ // then DALF or ADALF
         // ADALF has a bit of a dissipative character. That it creates a large time-contant
         // imaginary part of the energy is surprising. For a long time I considered it an artifact
         // of a programming error. For my present project ADALF seems not top be useful.
         R dtq=dt_*0.25_R;  // quarter of dt_
         R dth=dt_*0.5_R; // half of dt_
         t_+=dtq;
         x_+=y_*dtq;
         ydot=dot_(t_,x_);
         dy=(ydot-y_)*2_R;
         if (autoStep) kink1=dy.abs()*cpminv(ydot.abs()+y_.abs()+tiny);
         y_+=dy;
         if (adalf){
             y1=y_;
         }
         x_+=y_*dth;
         t_+=dth;
         ydot=dot_(t_,x_);
         dy=(ydot-y_)*2_R;
         if (autoStep) kink2=dy.abs()*cpminv(ydot.abs()+y_.abs()+tiny);
         y_+=dy;
         x_+=y_*dtq;
         if (adalf) {
            y_=(y_+y1)*0.5_R;
         }
         t_+=dtq;
      }
      else{ // then ALF
         R dth=dt_*0.5_R;  // half of dt
         t_+=dth;
         if (autoStep) n1=y_.abs();
         x_+=y_*dth;
         Y dy=(dot_(t_,x_)-y_)*2_R;
         if (autoStep) n12=dy.abs();
         y_+=dy;
         if (autoStep) n2=y_.abs();
         x_+=y_*dth;
         t_+=dth;
      }
      if (!autoStep){ // running with constant time step
         return;
      }
      else{ // running with adaptive time step
         R kink= dalf || adalf ? sup<R>(kink1,kink2) : n12*cpminv(n1+n2+tiny);
         if (kink > sc_.a1()){ // need to redo the step (i.e. go back to stateMem) with smaller step width
            t_=stateMem.t();
            x_=stateMem.x();
            y_=stateMem.y();
            dt_*=sc_.f1();// f1 = 1-frac
            return; // the next step then will possibly be better
         }
         if (kink < sc_.a2()){ // no need to go back to stateMem
            dt_*=sc_.f2();// f2 = 1+frac
            return;
         }
      }
   }

   R abs()const{ return x().abs() + y().abs();}

   void evoCQ_()
   //: evolution + conserved quantities
   // time evolution from ti=t() to sc_.tf_
   {
      Z docPeriod=sc_.docPeriod_;
      bool writeToFiles=sc_.writeToFiles_;
      bool writeMovie=sc_.writeMovie_;
      bool noDiff=sc_.noDiff_;
      R tf=sc_.tf_;
      R tw1=sc_.tw1_;
      R tw2=sc_.tw2_;
      R tw3=sc_.tw3_;
      dt_=sc_.dt0_;

      C e0=(x_|y_)*C::I;
      R n0=(x_|x_)[1];

      Graph gr;
      gr.setWithOrigin(1);
      vector<R> vt;
      vector<R> vdt;
      vector<R> vdn;
      vector<R> vder;
      vector<R> vdei;
      R dn= noDiff ? (x_|x_)[1] : (x_|x_)[1]-n0 ;
      C de= noDiff ? (x_|y_)*C::I : (x_|y_)*C::I-e0;
      vt.push_back(t_);
      vdt.push_back(dt_);
      vdn.push_back(dn);
      vder.push_back(de[1]);
      vdei.push_back(de[2]);
      Z clock=0;
      while (t_<tf){
         step_();
         clock++;
         if (clock%docPeriod==0){
            cpmvalues3("t_, dt_, tf",t_,dt_,tf,3);
            dn= noDiff ? (x_|x_)[1] : (x_|x_)[1]-n0 ;
            de= noDiff ? (x_|y_)*C::I : (x_|y_)*C::I-e0;
            vt.push_back(t_);
            vdt.push_back(dt_);
            vdn.push_back(dn);
            vder.push_back(de[1]);
            vdei.push_back(de[2]);
            x().mark(gr);
            gr.vis(writeMovie);
         }
      }
      Z n=vt.size();
      R_Vector rvt(n);
      R_Vector rvdt(n);
      R_Vector rvdn(n);
      R_Matrix rmde(2,n);
      Z i,j;
      for (i=1,j=0; i<=n; ++i,++j){
         rvt[i]=vt[j];
         rvdt[i]=vdt[j];
         rvdn[i]=vdn[j];
         rmde[1][i]=vder[j];
         rmde[2][i]=vdei[j];
      }
      gr.setText(V<Word>{"time step dt"});
      gr.mark(rvt,rvdt);
      gr.vis(writeToFiles);
      cpmwait(tw1,2);

      gr.clearText();
      gr.setText(V<Word>{"norm"});
      gr.mark(rvt,rvdn);
      gr.vis(writeToFiles);
      cpmwait(tw2,2);

      gr.clearText();
      gr.setText(V<Word>{"energy"});
      gr.mark(rvt,rmde);
      gr.vis(writeToFiles);
      cpmwait(tw3,2);
   }

   StpCon sc()const{ return sc_;}
   R dt()const{ return dt_;}
   void set_dt_(R const dt){ dt_=dt;}
   R t()const{ return t_;}
   X const& x()const{ return x_;}
   X& x(){ return x_;}
   Y const& y()const{ return y_;}
   Y& y(){ return y_;}

   C_Matrix jac(R eps)const{ // useful for non-linear DGLs not used so far
      Z n=b_.dim();
      C_Matrix res(n,n);
      for (Z i=1;i<=n;++i){
         V<Y> dyi=f(t_, x_+eps*b_[i])-f(t_,x_);
         for (Z j=1;j<=n;++j){
            res[i][j] = (dyi|b_[j]);
         }
      }
      return res;
   }
};

// harmonic oscillator as example: see CodeBlocks project
// ~/Desktop/e/cpm/quantumscatteringX1/quantumscattering.cbp

} // namespace

#endif
