//? cpmapplication.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_APPLICATION_H
#define CPM_APPLICATION_H

///////////////// Introductory section ///////////////////////////////////
//
//  Description: Provides an application framework that makes
//    convenient graphical and messaging capabilities available.
//
//  The user has to provide the definition of function
//     Z CpmApplication::main_(void).
//  The definition of the function
//     int main(int argc, char *argv[]),
//  which each C++ program has to define, is provided by having
//  cpmapplication.cpp among the source files of the C++ project.
//  The context in which main_() is called in cpmapplication.cpp
//  provides all C+- communication capabilities by having done all
//  necessary initializations.
//
//  The following code gives a simple, though realistic example.
//  It is ment as content of a source file smallexperiments.cpp. In
//  reading this code, please ignore the leading comment indicators '//'.
//  If we would switch here to the C comment style, my simple
//  documentation scripts are likely to get fooled.
//
//#include <cpmc.h>
//#include <cpmdigraph.h>
//#include <cpmfa.h>
//#include <cpmapplication.h>
//#include <cpmtests.h>
//#include <cpmrecordhandler.h>
//#include <cpmmultiimage.h>
//   // not all needed for test3()
//
//using namespace CpmRoot;
//using namespace CpmRootX;
//using namespace CpmSystem;
//using namespace CpmFunctions;
//using namespace CpmArrays;
//using namespace CpmGraphics;
//using namespace CpmImaging;
//using namespace CpmApplication;
//using namespace CpmGeo;
//
//R tWait=0;
//   ...
//void test3()
//{
//   // test of composing several images into one
//   Word dr=Message::ext("images");
//   Word n11=glue(dr,"bussard.ppm");
//   Word n12=glue(dr,"gnom.ppm");
//   Word n21=glue(dr,"thomas_meier.ppm");
//   Word n22=glue(dr,"kohlmeise1.ppm");
//
//   VV<Img24> vv(2,2); // 2 by 2 matrix of images
//   vv[1][1]=Img24(n11);
//   vv[1][2]=Img24(n12);
//   vv[2][1]=Img24(n21);
//   vv[2][2]=Img24(n22);
//   Z m=750, n=1000; // height and width of combined
//      // image
//   ImgR ir(vv,m,n);
//      // combined image, auto-layout
//   Graph g;
//   ir.mark(g);
//   g.dis(true);
//      // shows the composed image and writes a ppm
//   cpmwait(tWait,2); // remaining wait time to be displayed
//       // on pane 2 of the status bar. There are 4 panes,
//       // the first one is for news from CpmSystem::Message.
//}
//// define  functions test1(), test2()
//   ....
//
//Z CpmApplication::main_(void)
//{
//   Z mL=1;
//   Z sel=args_[2].toZ();
//   Word loc("CpmApplication::main_(), sel="&cpm(sel));
//     // if sel is not given in the comand-line it will be read from
//     // an inifile the name of which is shortNameOfExecutable.ini
//     // In the present example I name the executable
//     // smallexperimentsMS.exe
//     // and the ini file is smallexperiments.ini. This is unrelated to
//     // title_ which only by chance is the same.
//   CPM_MA
//   title_="smallexperiments";
//   tWait=args_[6].toR();
//   if (sel==1) test1();
//   else if (sel==2) test2();
//   else if (sel==3) test3();
//   else test4();
//   CPM_MZ
//   return 0;
//}
//////////////////////////////////////////////////////////////////////////

#include <cpmv.h>

namespace CpmSystem{

   void updFromConfigFile();
   // update from file config.ini
   // Tries to access a file named cpmconfig.ini at directory positions
   // which are set in file systemdependencies.h. Most of the quantities
   // that are defined by define macros in systemdependencies.h can be
   // read from this file and thus redefined. This redefinition affects
   // static data of class Message and thus does not require recompilation
   // (which changing a define in cpmsystemdependencies.h would).
   // The most important quantities to be set in cpmconfig.ini are the
   // size of the graphical application window, the directory locations
   // or application specific ini-files and the directory location for
   // output from the application.
   // Numerical precision can no longer be set by an entry in this file
   // (as it once could under control of CPM_USE_MPREAL).
}

namespace CpmApplication{

   void inspectCpmdefinitions();
   void inspectCpmsystemdependencies();

   // Provides a framework for defining  c o n s o l e  applications which
   // offer convenient graphical and messaging capabilities.
   //
   // Both capabilities can be eliminated at compile time by enabling or
   // disabling the lines
   //    #define CPM_NOGRAPHICS
   // and
   //    #define CPM_NOLOG
   // in file cpmsystemdependencies.h.
   // If CPM_NOGRAPHICS is defined, no graphics or images can be shown on
   // screen. But they still can be created in memory and written to file
   // in the portable pix-map format (*.ppm). The size of these images is
   // not limited to screen size. With this setting of CPM_NOGRAPHICS,
   // the program needs not to be linked with its screen graphics enabling
   // library, which is OpenGL and (Open)GLUT. It then also needs no
   // access to to the gl and glut header files. Of course, this is only
   // possible since all graphical primitives, including fonts (greek
   // included) are implemented within C+-.
   // Disabling message streams to the console and to log-files (as done
   // by #define CPM_NOLOG) was essential in a large industrial program
   // where my C+- program translated the intended motions of a complex
   // wire bending tool into stepper motor control files.
   // To the IT-professional, who had finally to put all these control
   // files together and to download them to the the wire-O bending robot,
   // the messages from my program were disturbing and he was happy to
   // hear that there was no problem switching them off. During my
   // development and testing of many C+- programs, these messages played
   // such a vital role that I failed to anticipate that in a different
   // mode of usage, there are different priorities and needs.
   // Many of the design decisions in C+- are actually revisions of old
   // decisions which turned out to support adaption to new needs not
   // as efficiently as they should.

   // A  C + -  a p p l i c a t i o n  (not just a C++ application which
   // uses some C+- classes) has cpmapplication.cpp among its source
   // files and defines the function
   //    Z CpmApplication::main_(void)
   // in its highest-level source files. This highest level source file
   // (which typically has a name that characterizes the application,
   // and for which a 'cpm' in the name would be misleading) has to
   // include the present file cpmapplication.h. The function
   //    int main(int argc, char *arv[])
   // which a console application needs to define, is defined in
   // cpmapplication.cpp in a way which depends on CPM_NOGRAPHICS and
   // CPM_NOLOG as explained above.

   // See cpminifilebasapp.h for an extensions of the C+- application
   // framework in applications based on reading ini-files.

   // Presently the usability of the C+- application framework
   // for parallel computing is not yet clarified. The PaLa program
   // (which is a C+- application) compiles with defined(CPM_USE_MPI) but
   // fails during run.

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmArrays::V;

// data

   extern Word title_;
      // Title of the whole program. Most conveniently to be defined
      // within function main_() as in the example of the Introductory
      // section.

   extern V<Word> args_;
      // An array that waits to get initialized by commandline arguments
      // and by additional arguments which may be provided by an ini-file.
      // In a C+- application (see definition above) args_[1] holds always
      // the full name (including directory location) of the program's
      // executable.
      // In calling such a program from the command line, one may enter
      // additonal parametes, which main(...) transforms into components
      // of args_. Such additional parameters do not affect any function
      // defined in cpmapplication.cpp. They can be used however in
      // programming the following function Z main_(), see the
      // introductory section of the present file for an example.
      // Function main, as implemented in cpmapplication.cpp always
      // checks for a file cpmconfig.ini for getting additional data.
      // From data which are both given on the command line and in the
      // cpmconfig.ini, the command line always wins. In addition
      // to cpmconfig.ini a file, here symbolically named
      // 'shortNameOfExecutable'.ini is tried to read for further input.
      // See function
      //       V<Word> CpmRootX::extComLine(int argc, char *argv[]);
      // in files cpmrecordhandler.(h,cpp).

      // The class IniFileBasedApplication introduces a particular form of
      // args_ as a list consisting of:
      //    selection,
      //    verbosity,
      //    initial wait time,
      //    relative path of further input data,
      //    final wait time.
      // This structure and meaning needs, however, not be followed in an
      // implementation of of main_().
      // V<Word> is the most flexible among the lower-level data types
      // and it should not, at this place, be replaced by a more specific
      // one.

   extern Word iniFile_;

  // functions

   Z main_(void);
      // Defines the main work of the program.
      // The function has available data args_.
      // Defining code is allowed to rely on correct initialization of
      // graphical, font-related, and directory-related matters.
      // The details of this initialization mechanism are described in
      // cpmapplications.cpp as comments to function
      //    int main(int argc, char *argv[])

} // namespace

#endif
