//? cpmfile.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FILE_H_
#define CPM_FILE_H_
/*

Description:
   Making a class File that satisfies the value interface
   along using smart pointers in close analogy to F<X,Y>
   Assignments between File instances
   don't yet behave as the strict value interface requires
   (not a very natural requirement for files)
   Saying
   File f1(name);
   File f2=f1;
   cout<<f1.readLine();
   cout<<f2.readLine();
   will output the first and the second line of f1 and not twice
   the first line as would be the requirement from the
   strict value interface.
   This effect comes from hidden data in files: the 'access-mark'
   is not accessible to me, so I cant make the copy of a file
   that also copies the position of this mark.
   More important:
      File f1(name);
      File f2(name);
      cout<<f1.readLine();
      cout<<f2.readLine();
   will output the first and the second line twice
   since f1 and f2 are completely independent, allthough
   they are fed by the same data on file.

  However one hardly will need files in a context that would
  require the strict value interface.

  The main benefit from the exercise so far is that we can
  use data of type File in classes in a manner that no additional
  default initialisations and destructors have to be given.
  Also the effective initialization has not to be done in a constructor
  initialization list (where the information for proper
  initialization may not yet be available) but when needed
  e.g.
         file=IFile("prog.ini");
  where file is a data member of the class the constructor of
  which we are just writing

  class CpmEP..::Extractor uses this style so far.

  All definitions are short and inline, thus no cpmfile.cpp
  exists.

  The formal test of the strict value interface in cpmtests.h
  by class ValueBehavior<IFile> works however since here as a test
  of strictness we don't use the critical non-constant read
  functions but only assignments. See cpmsmallexperiments test20().
*/
#include <cpmv.h>

namespace CpmSystem{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmArrays::P;
   using CpmArrays::V;

//////////////////////// class IFile //////////////////////////

   class IFile:public P<IFileStream>{ // input file
       typedef P<IFileStream> Base;
   public:
      Word nameOf()const{ return "IFile";}
      IFile():Base(){}
      explicit IFile(Word const& w, bool safe=true, bool binary=true):
         Base(new IFileStream(w,safe,binary)){}
      ifstream& operator()(void){ return p_->operator()();}

      bool isVal()const
         //: is valid
      { if (p_==0) return false; else return p_->isVal();}

      char readChar(){return p_->readChar();}
      string readWord(){ return p_->readWord();}
      string readLine(){ return p_->readLine();}
      Word getName()const
         { if (isVal()) return p_->getName(); else return Word();}
      bool getChar(char& ch)
         { p_->operator()().get(ch); return p_->isVal();}
// equivalent to CPM_ORDER, here inline implementation, since we
// want to avoid the need for a file cpmfile.cpp
      bool operator==(IFile const& obj)const
         { return p_==obj.p_;}
      bool operator!=(IFile const& obj)const
         { return p_!=obj.p_;}
      bool operator<(IFile const& obj)const
         { return p_<obj.p_;}
      bool operator>(IFile const& obj)const
         { return p_>obj.p_;}
      bool operator<=(IFile const& obj)const
         { return p_<=obj.p_;}
      bool operator>=(IFile const& obj)const
         { return p_>=obj.p_;}

      V<Word> readAll()const
      {
         Word fn=getName();
         if (fn.dim()==0) return V<Word>(0);
         else{
            std::vector<Word> res;
            IFile w(fn);
            while(w.isVal()) res.push_back(Word(w.readLine()));
            return V<Word>(1,res);
         }
      }

      friend std::ostream& operator<<(std::ostream& out, IFile const& obj)
      {
         Word fn=obj.getName();
         if (fn.dim()>0){
            IFile w(fn);
            while(w.isVal()) out<<w.readLine();
         }
         return out;
      }

      bool isEqlTo(IFile const& obj)const
         //: is equal to
      {
         return readAll()==obj.readAll();
      }
   };

//////////////////////// class OFile //////////////////////////

   class OFile:public P<OFileStream>{ // output file
       typedef P<OFileStream> Base;
   public:
      Word nameOf()const{ return "OFile";}
      OFile(){}
      explicit OFile(Word const& w, bool safe=true, bool binary=true):
         Base(new OFileStream(w,safe,binary)){}
      ofstream& operator()(void){ return p_->operator()();}
      bool isVal()const
         //: is valid
      { if (p_==0) return false; else return p_->isVal();}

      Word getName()const
         { if (isVal()) return p_->getName(); else return Word();}
      bool putChar(char& ch)
         { p_->operator()().put(ch); return p_->isVal();}
   // equivalent to CPM_ORDER, here inline implementation
      bool operator==(OFile const& obj)const
         { return p_==obj.p_;}
      bool operator!=(OFile const& obj)const
         { return p_!=obj.p_;}
      bool operator<(OFile const& obj)const
         { return p_<obj.p_;}
      bool operator>(OFile const& obj)const
         { return p_>obj.p_;}
      bool operator<=(OFile const& obj)const
         { return p_<=obj.p_;}
      bool operator>=(OFile const& obj)const
         { return p_>=obj.p_;}
   };


} // namespace

#endif
