//? cpmfontgeneration.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FONTGENERATION_H_
#define CPM_FONTGENERATION_H_
/*

    Description:
      classes for extracting a consistent font from a
      set of non-uniformely scaled bitmap patterns
      available in ppm format. It is assumed, however that
      the orientation is correct so that no rotation of the
      originals is needed

   Recent history: CPM_ORDER implemented for class Font

*/

#include <cpmimg24.h>
#include <cpmm.h>

namespace CpmFonts{

   using namespace CpmStd;

   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmRoot::L;
   using CpmRoot::R;

   using CpmArrays::IvZ;
   using CpmArrays::V;
   using CpmArrays::LEAN;
   using CpmArrays::M;

   using CpmGraphics::Img24;
   using CpmGraphics::xy;
   using CpmGraphics::ColRef;

/////////////////// class PrlLet ///////////////////////////////////////

class PrlLet{ // preliminary concept of a letter
   // preliminary letter, where letter means a printable symbol with
   // emphasis on its shape and proper placement in relation to a
   // 'printing box'.
   // Preliminarity refers to the meaning of h_,w_,d_ which do not yet
   // refer to a consistently defined printing box.
   V< V<IvZ> > r_; // representation
      // as an array of arrays of horizontal intervals. The main (outer)
      // index refers to the vertical direction. That there is a need for
      // a second (inner) index is clear from letters like A or B where a
      // horicontal cut results in either one or two segments.
   Z h_{0}; // Total actual height of the symbol (this is
      // a shift-invariant characteristic of the symbol).
      // It is very small for e.g. the minus sign '-' and
      // large for 'Q', particularly if it is given a descending part.
   Z w_{0};
      // width of the data
   Z d_{0};
      // descender (Unterlaenge), says how many pixels of the total
      // hight are to be placed below the text baseline.
      // Initialized as 0, which is convenient since most
      // letters don't descend
   L c_;
      // character, meaning of the letter. This comes into action if a
      // character string such as "x=10; y=20;" is to be translated into
      // letters.
   PrlLet(V< V<IvZ> > const& r, Z h, Z w):
      r_(r),h_(h),w_(w),c_('x'){}
      // private since internal consistency of data is not assured.
      // From h and w one can't give a meaningful setting of pw_.
      // So we let it stay as zero
   typedef PrlLet Type;
public:
   CPM_IO
   PrlLet():r_(V< V<IvZ> >(0, V<IvZ>(),0)),c_('x'){}
      // since all other members are of class type
      // this is OK
   PrlLet(Word const& fileName);
      // fileName is the name of a ppm-image
      // complete name whith ".ppm" at the end

   PrlLet(Img24 const& imr, ColRef sep=ColRef(Z(127),Z(127),Z(127)));
      // generation from a versatile image format that can be
      // get by transformation from ppm image files which, in turn,
      // may be made from electronic camera exposures.

   V<IvZ> const& operator[](Z i)const{return r_[i]; }

   PrlLet min()const; // minimize
      // returns a version of *this that has no white
      // rims arround the letter
   void setDes(Z d){ d_=d;}
      // set descender
   void setMen(L c){ c_=c;}
      // set meaning
   void setMen(Z n){ c_=(L)n;}
      // input of c by its decimal code

   void setWid(R w){ w_=cpmtoz(w);}

   Img24 toImg24()const;
      // returns a image of type Img24 based on the data of *this
   Img24 toImg24(Z hl, Z hu, Z r)const;
      // Returns an image of height=hl+hu and a width that is larger
      // at both the left and the right side by a rim of width r.
      // In the vertical direction the image is placed such that its
      // lowest line is at hl-d_ above the lower rim of the image,
      // or, put differently, the upper line of the character is
      // line hu-(h_-d_) (counted - as is standard - from the upper
      // end)

   Z h()const{ return h_;}
   Z w()const{ return w_;}
   Z d()const{ return d_;}
   L c()const{ return c_;}
   void print(Z x, Z y, Img24& im)const;
      // prints the letter *this to im such that
      // the 'origin' of the letter is placed on (x,y)
   Word nameOf()const{ return "PrlLet";}
};

///////////////////// class PrlFont ////////////////////////////////////
class PrlFont{
// preliminary font is a list of preliminary letters
   V<PrlLet> l_;
      // letters
   V<Img24> im_;
      // properly placed maximum resolution images
   Z h_; // maximum of l_[i].h()
   Z hu_;
      // 'upper hight': maximum extend above the main line
   Z hl_;
      // 'lower hight': maximum extend below the main line
   typedef PrlFont Type;
   Z getIntNum(L c)const;
      // get internal number, returns for c an
      // index i such that l_[i].getMen()==c
   void makeReady();
      // setting all descenders and printing with
public:
   CPM_IO
   PrlFont(){}
   PrlFont( V<PrlLet> const& l);
   PrlFont(Word const& fileName);
      // fileName is the complete name
      // (including extension) of the file
      // from which to read the preliminary
      // font data
   Z dim()const{ return l_.dim();}
   Z h()const{ return h_;}
   Z hu()const{ return hu_;}
   Z hl()const{ return hl_;}

   const Img24& getImg(Z i)const{ return im_[i];}
   const PrlLet& getLet(Z i)const{ return l_[i];}
   L getMen(Z i)const{ return l_[i].c();}

   const PrlLet& getLet(L c)const{ return l_[getIntNum(c)];}
      // access to the letters by meaning

  // static V<PrlLet> fromPPM();
      // It is assumed that there is a directory in which the raw
      // information for font creation is placed in the form of a
      // collection of ppm-images named in a way that lets deduce
      // from each file name an ASCII symbol which is represented
      // by the image encoded in this file. All symbols are assumed
      // to be represented according to the same (otherwise arbitrary
      // scale) so that they have to be demagnified all by a common
      // factor for creation of a scaled font. Presently the directory
      // and the file names are coded within this function and don't
      // enter as arguments.

   static V<PrlLet> fromPPM(Word const& dirIn, Word const& fnOut=Word());
      // It is assumed that there is a directory dirIn in which the raw
      // information for font creation is placed in the form of a
      // collection of ppm-images named in a way that lets deduce
      // from each file name an ASCII symbol which is represented
      // by the image encoded in this file. All symbols are assumed
      // to be represented according to the same (otherwise arbitrary
      // scale) so that they have to be demagnified all by a common
      // factor for creation of a scaled font. The return value of
      // type V<PrlLet> gets written on a file the name of which is
      // provided as the second argument.
};

///////////////////// class Let ///////////////////////////////////////

class Let{ // letter, where letter means a printable symbol
   // with emphasis on its shape and proper placement
   // in relation to a 'printing box'
   Img24 r_; // representation
   L c_;
      // character, meaning of the letter
   typedef Let Type;
public:
   CPM_IO

   Let(){} // since all members are of class type this is OK

   Let(PrlLet const& pl, Img24 const& im, R frac=0.1);
      // creates a letter from a preliminary letter and
      // a de-magnification factor (typically < 1)

   Img24 toImg24()const{ return r_;}
      // returns a image of type Img24 based on the data of *this

   Z h()const{ return r_.h();}
   Z w()const{ return r_.w();}
   L c()const{ return c_;}
   void print(Z x, Z y, Img24& im , ColRef cr=ColRef())const;
   Word nameOf()const{ return "Let";}
};

class Font{ // a font is a list of letters
   V<Let> l_;
      // letters
   Z h_{0};
      // uniform height of letters
   IvZ w_;
      // width range
   R mw_;
      // mean width
   M<L,Z> st_;
      // assigns to a letter its index in list l_
      // see function getLet
   typedef Font Type;
 public:
   CPM_IO
   CPM_ORDER
      // ordering according to h_
   Font():mw_(0.){}
   Font(Word const& fileName);
      // fileName is the complete name
      // (including extension) of the file
      // from which to read the final font data
      // (one file for one size !)
   Font(Word const& fileName, Z h);
      // fileName is the complete name
      // (including extension) of the file
      // from which to read the preliminary
      // font data. From these a font of
      // height of h pixels gets created
   Font(PrlFont const& pf, Z h);
      // same as previous function with the preliminary
      // font data as direct input instead of read from file
   Z dim()const{ return l_.dim();}
      // number of letters
   Z h()const{ return h_;}
      // hight of each letter box
   IvZ w()const{ return w_;}
      // minimum and maximum printing width
   R mw()const{ return mw_;}
      // mean printing width

   const Let& getLet(Z i)const{ return l_[i];}
   L c(Z i)const{ return l_[i].c();}
   const Let& getLet(L c)const{ return l_[st_[c]];}
      // access to the letters by meaning
   void show(Img24& img, ColRef cr )const;
      // write the font *this in a single text-line
      // (with proper line breaking if the size of img requires this)
      // onto img.
   void wrt(Word const& w, xy& p, Img24& img,
      ColRef cr=ColRef())const;
      // write
      // p acts as a writing cursor: Consider
      // Word w1=..., w2=...; xy p=...; Img24 img=...; Font f;
      // Then
      // f.wrt(w1,p,img); f.wrt(w2,p,img);
      // lets the Word w2 be written in direct succession to w1.
      // p is the left-upper corner of the rectangular 'box'
      // associated with the first letter of w. Notice that all
      // such boxes have the same height (not width!) within
      // one font.
   Z len(Word const& w)const;
      //: length
      // returns the actual length of w, when written in font *this.
   Word nameOf()const{ return "CpmFonts::Font";}
   static Font factory(Z h);
      // creates a font of hight f from data in presently valid
      // location

   static V<Font> massFactory(Word const& fnIn,
      Word const& fnOut=Word("fonts/cpmfont1.txt"));
      // returns a family of fonts al differing only in
      // height and which all are made from a single set
      // of letter shapes by interpolation.
      // Presently there are 6 fonts of heights
      // 10, 12, 14, 16, 18, 20 pixels
      // The function also creates a file with name fn to which
      // it writes the result.
};

}

#endif
