//? cpmframes.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FRAMES_H_
#define CPM_FRAMES_H_
/*

Purpose: making a class of matrix-like subdivided frames
A member function of class Frames outputs a matrix of
Graph objects. Those are most useful in vizualising states and
processes in simulations.

*/

#include <cpmframe.h>
#include <cpmgraph.h>

namespace CpmGraphics{

   class Frames{ // array of Frame objects

      Frame fr_;
      CpmArrays::VV<Frame> vv_;
      typedef Frames Type;
public:
   CPM_NAM("Frames")
      //: name of
   CPM_IO
   CPM_ORDER

   Frames(){}

   Frames(Frame const& f, Z m, Z n, R fill=0.98);
      // The frame f will be subdivided into a (m,n)-matrix
      // of non-overlapping subframes. In order to avoid overlap
      // of the subframes factor fill may be set to values slightly
      // below 1

   Frames(Frame const& f, const CpmArrays::Vo<R>& horizDiv,
      const CpmArrays::Vo<R>& vertDiv, R fill);
      // The frame f will be subdivided into a matrix
      // of non-overlapping subframes where the subdivision is given by
      // the real lists horizDiv, vertDiv
      // 0<= horizDiv[i]<=1; 0<= vertDiv[i]<=1;
      // List need not be ordered since done internally

   Frames(Frame const& f, Z n, R fill=0.95);
      // system of n frames reasonably distributed over
      // the viewing rectangle of f

   Z linDim()const{ return vv_.dim1()*vv_.dim2();}
      //: linear dimension

   Frame& operator()(Z i);
// defining and explaining code:
/*
   { Z2 mtl=linearToMatrix(i,m,n);
   return (*this)[mtl.x1][mtl.x2];}
*/
   // The valid indexes range from 1 to linDim();

   Frame const& operator()(Z i)const;

   Frame& operator()(Z i, Z j){ return vv_[i][j];}
   Frame const& operator()(Z i, Z j)const{ return vv_[i][j];}

 //  Frame& at(Z i, Z j){ return vv_[i][j];}
  // Frame const& at(Z i, Z j)const{ return vv_[i][j];}

   Frame toFrm()const;
      //: to Frame
      // The returned Frame contains all the graphic content of *this.

   VV<Frame> frsVV()const{ return vv_;}

   // getting a matrix of Graphs
   VV<Graph> grsVV()const{
      Z m=vv_.dim1();
      Z n=vv_.dim2();
      VV<Graph> res(m,n);
      for (Z i=1;i<=m;++i){
         for (Z j=1;j<=n;++j){
            res[i][j]=Graph(vv_[i][j]);
         }
      }
      return res;
   }

   void vis()const;
      //: visualize

/***** a recent (2021) application of Frames and Frames::grsVV() **********

   Graph gr;
   Z ni=2,nj=1;
   Frames frs(gr,ni,nj);
   VV<Graph> grs=frs.grsVV();
   wf1.mark(grs[1][1]);
   wf2.mark(grs[2][1]);
   grs[1][1].vis();
   grs[2][1].vis();
   cpmwait(4.,2);

******** a typical application of the previous two functions ********
   Frame fr;
   fr.paint(Color(LIGHTGREEN));
   Frames frs(fr,n);
   V<Graph> vgr(n);
   for (Z i=1;i<=n;i++){
      vgr[i]=Graph(frs(i));
      vgr[i].setWithOrigin(1);
      vgr[i].show(res[i]);
      vgr[i].vis(true);
   }
*/

/******** distributing Img24 instances over the screen **************
   V<Img24> v= ... ;
   Z n=v.dim();
   Frame fr;
   Frames frs(fr,n); // arrange matrix-like according to a simple
      // algorithm
   for (Z i=1;i<=v;++i) frs(i).fitImg(v[i]);
      // conserves the aspect ratio and gives maximum fillfactor
   frs.vis();

   Now there is also a way of visualization which works together with
   writing as a ppm file:

   bool wos=true;
   frs.toFrm().vis(wos);
   Here the function toFrm creates out of a Frames object a Frame object
   which has the same image context as the Frames object.
**********************************************************************/
};

} // namespace

#endif

