//? cpmgaz.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_GAZ_H_
#define CPM_GAZ_H_
/*
   Description: Group algebra of Z (the latter considered as its
   additive group). Can also be interpreted as the class of
   Laurent polynomials.
*/

#include <cpmvr.h>

namespace CpmAlg{ // algebra
   using namespace CpmStd;
   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmRoot::Root;

   using CpmArrays::IvZ;
   using CpmArrays::V;
   using CpmArrays::Vr;

/////////////////////// class GaZ<> //////////////////////////////////

template <class T>
class GaZ{
   // Group algebra of the additive group Z.
   // Can also be viewed as the  ring of Laurent polynomials over T,
   // where natural choices for T are  C, R or Z. See
   // http://en.wikipedia.org/wiki/Laurent_polynomial
   // Other classes may be interesting. Presently we assume from
   // T that for all t \in T Root<T>(t).f(...) is defined
   // for all memberfunctions f of Root except for inv).
   // Further we assume that T*T, T+T, -T, are defined.
   // In order that Test_rm<GaZ<T>,T>(...) does not report errors
   // * needs to be commutative (see cpmtests.h).
   Vr<T> rep_;
      //: representation
   // meaningful components of *this. Leading and trailing
   // components of value T() are not considered meaningful
   // and get eliminated by calling the function adj_().
   // For a natural implementation of GaZ it is important that we
   // have arrays in which the indexes of the components are allowed to be
   // negative. This is now the case for V (and thus also for Vr),
   // and most of the implementation (with the exception of
   // multiplication) can be taken from Vr. It would be unwise to derive
   // from Vr since then multiplication had to be virtual in Vr (actually
   // in its base Va, where arithmetics is defined).
   Z adj_();
      //: adjust
      // Eliminates leading and trailing T()-valued components
      // and returns the number of components that were eliminated.
   typedef GaZ<T> Type;
public:
// constructors
   GaZ():rep_(){} // no adj_() needed
      // zero element algebraically
   explicit GaZ(T t):rep_(V<T>(IvZ(0,0),t)){if (t==0) adj_();}
      // t times the unit element algebraically
   explicit   GaZ(V<T> const& v):rep_(v){adj_();}
      // construction from components

// declaration macros
   CPM_IO
   CPM_ORDER
   CPM_SUM_C
   CPM_DIFFERENCE
   CPM_PRODUCT_C // here the product is a concolution

// multiplication with 'scalars'
   Type operator*(T const& s)const{ return Type(rep_*s);}
   Type operator/(T const& s)const{ return Type(rep_/s);}
   friend Type operator*(T const& s,Type const& x){ return x*s;}
   Type& operator*=(T const& s){ return *this=*this*s;}
   Type& operator/=(T const& s){ return *this=*this/s;}

// infra structure, all functions of Root<Type> except of
// Root<Type>::inv()
   Type con()const{ return Type(rep_.con());}
      //: conjugate
   R absSqr()const{ return rep_.absSqr();}
      //: absolute (value) squared
   R abs()const{ return rep_.abs();}
      //: absolute (value)
   Word nameOf()const;
      //: name of
   Z dim()const{ return rep_.dim();}
      //: dimension
   Type net(Z i=0)const{ return Type(rep_.net(i));}
      //: neutrals
   Type ran(Z i=0)const;
      //: random (generator)
   Type test(Z cpl)const{ return Type(rep_.test(cpl));}
      //: test (element generator)
   Z hash()const{ return Type(rep_.hash());}
      //: hash (number)
   R dis(Type const& l)const{ return rep_.dis(l.rep_);}
      //: distance
   Word toWord()const;
      //: to Word
};

// private
template <class T>
Z GaZ<T>::adj_()
{
   Z count=0;
// E for 'end'
   bool workE=!rep_.isVoid()&&rep_.last()==T();
   while (workE){
      count++;
      rep_=rep_.eliLast();
      workE=!rep_.isVoid()&&rep_.last()==T();
   } // We start with cutting the end since this is
   // more economic. If all components of *this vanish, this
   // makes a difference since then the first while loop
   // does the whole work.
// B for 'beginning'
   bool workB=!rep_.isVoid()&&rep_.fir()==T();
   while (workB){
      count++;
      rep_=rep_.eliFirst();
      workB=!rep_.isVoid()&&rep_.fir()==T();
   }
   return count;
}

//public
// declared by macros
template <class T>
bool GaZ<T>::prnOn(ostream& str)const
// declaration in CPM_IO
{
   cpmwat;
   cpmp(rep_);
   return true;
}

template <class T>
bool GaZ<T>::scanFrom(istream& str)
// declaration in CPM_IO
{
   cpms(rep_);
   return true;
}

template <class T >
Z GaZ<T>::com(GaZ<T> const& v)const
// declaration in CPM_ORDER
{
   return rep_.com(v.rep_);
}

template <class T>
GaZ<T> GaZ<T>::neg()const
// declaration in CPM_DIFFERENCE
{
   return GaZ<T>(rep_.neg());
}

// declared explicitely in class
template <class T>
GaZ<T> GaZ<T>::ran(Z i)const
{
   V<T> rr=rep_.ran(i);
   IvZ iv=rr.dom();
   Z i0=iv.ranSel();
   IvZ iv0=iv-i0; // shift the domain stochastically
   V<T> rr0(iv0);
   for (Z j0=iv0.b(),j=iv.b();j0<=iv0.e();++j0,++j) rr0[j0]=rr[j];
   return GaZ<T>(rr0);
}

template <class T>
GaZ<T> GaZ<T>::operator+(GaZ<T> const& v)const
{
   IvZ iv=(rep_.dom()|v.rep_.dom());
   V<T> vr(iv);
   for (Z i=iv.b();i<=iv.e();++i){
      vr[i]=rep_(i)+v.rep_(i);
      // If rep_[i] is not defined, rep_(i) is defined as T().
      // This is just what is needed here.
   }
   return GaZ<T>(vr);
}

template <class T>
GaZ<T> GaZ<T>::operator*(GaZ<T> const& v)const
{
   IvZ iv1=rep_.dom();
   IvZ iv2=v.rep_.dom();
   IvZ iv=iv1+iv2;
   V<T> vr(iv);
   for (Z k=iv.b();k<=iv.e();++k){
      T sk=T();
      for (Z i=iv1.b();i<=iv1.e();++i) sk+=rep_(i)*v.rep_(k-i);
      // If rep_[i] is not defined, rep_(i) is defined as T().
      // This is just what is needed here.
      vr[k]=sk;
   }
   return GaZ<T>(vr);
}

template <class T>
Word GaZ<T>::nameOf()const{
   Word nt=Root<T>(T()).nameOf();
   Word wi="GaZ<";
   return wi&nt&">";
}

template <class T >
Word GaZ<T>::toWord()const
{
   Word header("Laurent polynomial determined \
by the following list of index-coefficient-pairs: ");
   return header&rep_.toWord();
}

} // namespace

#endif
