//? cpmhistogr.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_HISTOGR_H_
#define CPM_HISTOGR_H_

/*
  cpmhistogr.h
  Description: This class includes functions to make multimage-histograms,
      show multimage-histograms
      and find min and max values of the histogramms

*/
#include <cpmrmatrix.h>
#include <cpmframe.h>
#include <cpmrfunc.h>

namespace CpmImaging{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmLinAlg::R2;
   using CpmArrays::VV;
   using CpmFunctions::R_Func;
   using CpmGraphics::Frame;
   using CpmGeo::Iv;

class Histogram { // histograms: generation, visualization, analysis
   Z nPlanes;
      // number of planes of the histogram
   Iv range;
      // Iv(valMin,valMax); helps if available also in this form
   R valMin;
      // Minimum of Histogram (normaly 0)
   R valMax;
      // Maximum of Histogram (normaly 255)

   V<R> binValues;
      // has nBins components of equi-distant x-values

   Z nBins;
      // number of sectors of the histogram

   VV<Z> values;
      // Matrix contains the values of the histogram
      // values[i][j] is the number of values belonging to
      // plane i and bin number j

   R binWidth;
      // width of the sectors of the histogram

   Word topic;
      // a title

public:
   Histogram( Z nPlanesI=1, Z nBinsI=1, R valMinI=0, R valMaxI=255,
      const Word& theme=Word());
      //constructor for a multi-histogram
      // nPlanes: Number of distributions curves
      // nBins:   Number of sectors in each distribution
      // valMin:  value belonging to the left boundary of the lowest
      // sector
      // valMax:   value belonging to the right boundary of the highest
      // sector

   R2 minMax(R min=0.01,R max=0.01)const;
      // Let res be the return value.
      // res.x1 is the smallest x for which ...
      // 2005-08-29 not clear what the function really does.
      // Has to be revisited. Action !
      // for minimum: find the first bin of the histogram containing
      // the fraction min of all Pixels.
      // for maximum: find the last sector of the histogram including
      // a special rate of pixels (declared in max) max = 0.01   means
      // find the last sector include 1% of all Pixels
      // the function conforms this values from sector numbers to absolut
      // Pixelvalues (normaly between 0 and 255) and returns this values

   R2 minMax(Z plane, R min=0.01,R max=0.01)const;
      // for minimum: find the fist sector of the histogram including
      // a special rate of pixels (declared in min) min = 0.01   means
      // find the first sector include 1% of all Pixels
      // for maximum: find the last sector of the histogram including
      // a special rate of pixels (declared in max) max = 0.01   means
      // find the last sector include 1% of all Pixels
      // the function conforms this values from sector numbers to absolut
      // Pixelvalues (normaly between 0 and 255) and returns this values

   void incrementValue(Z h,Z SectNum);
      // increments a sector of a histogram
      // h:   Plane of the Histogram (RGB)
      // SectNum:   Number of Sector whitch should be incremented

   void mark(Frame& fr, bool cumulative=true)const;
      //: mark
      // Stores the histogram *this in cumulative form to g.
      // No display unless g.vis() is called.
      // As pointed out to me by Thomas Dera, the cumulative form
      // is much more suitable for screen display ('more quiet curves')
      // than a direct display of bins.
      // fr: Screen area in which the histogram is shown

   void show(Frame& fr, bool cumulative=true)const
      //: show
      // shows the histogram *this in cumulative form.
      // As pointed out to me by Thomas Dera, the cumulative form
      // is much more suitable for screen display ('more quiet curves')
      // than a direct display of bins.
      // fr: Screen area in which the histogram is shown
   { mark(fr,cumulative); fr.vis();}

   void put(Z h, R value, Z Method=0);
      // find the sector of a histogram to which value belongs, and
      // increase the count of this sector by one.
      // h:         Plane (RGB)
      // value:   signal value (normaly between 0 and 255)
      // you can select with the parameter Method how to handle out of
      // range values.
      // if Method=0, for all values out of borders the function put the
      // value to the borders
      // if Method=1 all values out of borders the function return 0
      // in this case the function put will forget this values

   R_Func getBrightValues(Z pl, R fracL=0.9, R fracU=0.001, R p=4)const;
      // a tone scale transformation function which makes the brightest
      // pixels white and all significantly darker pixels black.
      // Considers only plane pl for this procedure.
      // The pixel values that shall be identified as bright are by
      // definitiuon such that a fraction fracL of all pixels is
      // darker and a fraction fracU is brighter as the 'bright pixels'.
      // The tone scale transformation is such that the 'bright pixels'
      // get values near 255. The larger p, the harder the contrast of
      // the
      // scaled image.

   Iv firstPeak(Z pl=1, R frac=0.5, Z nL=1, Z nR=1);
      // returns the first bin (according to the normal numbering of
      // bins) that is a local maximum and has at least a hight
      // frac*total maximum.
      // Further it includes nL bins to the left and nR bins to the right
      // All these bins are united to a value range that is returned as
      // an intervalo in value space.
      // Considers only plane pl for this procedure.
};

} // namespace

#endif
