//? cpminifilebasapp.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_INIFILEBASAPP_H
#define CPM_INIFILEBASAPP_H

/*

Purpose: Provide tools for building application programs

   2007-08-07 Usage of path-names changed. Cooperation with Ruby
      programs suggested to use valid directory names x (for which
      Dir.chdir(x) works in situations were the present program had
      path names ending in '/'. Now the harmonization is done:
      pathname = directory name.
      This caused major changes in ini-files.

*/
#include <cpmf.h>
#include <cpmp.h>
#include <cpmapplication.h>
#include <cpmrecordhandler.h>

namespace CpmApplication{
   // Building an application framework within which one has
   // convenient graphical and messaging capabilities available.

   using CpmRoot::Word;
   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRootX::RecordHandler;
   using CpmArrays::V;
   using CpmArrays::Vp;
   using CpmFunctions::F;
#ifdef CPM_Fn
   using CpmFunctions::F1;
#else
   using std::bind;
   using std::placeholders::_1;
#endif

   extern Z dimApp;
      // 2 or 3 represents a dimension selection

////////////////////////  IniFileBasedApplication  //////////////////////
// A project which has cpminifilebasapp.cpp among its source files needs
// also cpmapplication.cpp. The latter defines the standard function
//             int main(int argc, char *argv[])
// in a way that graphical and stream reading and writing facilities are
// provided when a function declared as
//             Z CpmApplication::main_(void)
// gets called. The class IniFileBasedApplication to be considered here
// helps to define this function main_ in a comfortable way. Here is a code
// example from the C+- tutorial project chebyshev:
/**************************************************************************
Z CpmApplication::main_(void)
// In this form the C++ main function
//       int main(int argc, char *argv[])
// (defined in file cpmapplication.cpp) expects to see the basic
// functionality of the program.
// Notice that int main(...) in cpmapplication.cpp provides the program
// with its graphical window and a status bar.
// The fact that class ChebyApp is derived from class
// IniFileBasedApplication lets the following short code, although it
// looks like abstract nonsense, do very specific things.
{
   Word loc("Z CpmApplication::main_(void)");
   Z mL=1;
   CPM_MA
   title_=Word("chebyshev");
   ChebyApp app;
      // creates the application class by reading and deploying
      // chebyshev.ini
   app.run();
      // brings the application class to live and action
   CPM_MZ
   return 0;
}
// Thus, the structure of Z CpmApplication::main_(void) is as follows:
// It calls the constructor of a class derived from class
// IniFileBasedApplication and then calls the member function run of
// this class.
**************************************************************************/

// This is a proven tool for creating non-interactive applications which
// get all input from files and create output to files (probably along
// with dynamical screen output) including proper run documentation.
//
// Derive your application class from IniFileBasedApplication and
// re-define the virtual function void doTheWork(void). Primary data
// input is provided by a text file capable of access based on sections
// and keys.
//
// Supported data types for input are bool, Z, R, Word and homotypic
// lists of these.
//
// Originally this file had to be a Windows configuration file which
// always come with the file extension 'ini' which explains that I
// still refer to this primary input file as 'the ini-file'. The
// ini-file is the unique source to all data that the class can
// hold and which determine the 'state' of a class instance.
// In technical terms it is the attribute
//
//    RecordHandler rch;
//
// which holds the state. This object rch was created by reading
// the ini-file. Here RecordHandler is a flexible container class
// and RecordHandler itself is a direct descendant of a container class
// Record. Notice, that the state information may contain file names or
// URLs. These then open access to additional information.
// And also this information may enter the computations organized as
// member functions of the class.
//
// The syntax of ini-files will be explained in cpmrecord.h together with
// class Record, see particularly the explanation to the default
// constructor
//
//      Record(){};
//
// This is natural since this syntax can be considered as being
// indirectly defined in terms of a member function of this class:
// a text file named fn is a valid ini-file if and only if
//       iniFileTest(fn)==true
// for the following function definition:
//
/*
      bool iniFileTest(Word const& aFile)
      {
         IFileStream ofs(aFile);
         Record rc;
         return rc.readFormatted(ofs());
      }
*/
//
// According to my experience over years the 'ini-files' are very
// convenient to work with in programs that are subject to continuous
// improvement and adaption to new requirements.
// Since August 2003 there is the possibility to organize ini-files
// hierarchically: Any ini-file may contain enties of the form
// F anIniFile
// This helps to keep the input for complex programs organized,
// even if these need thousands of input parameters (which would
// be a challenge for GUI-based programs).

// A convenient method for creating a scientific program (application)
// is to derive a class from the class IniFileBasedApplication and to
// re-define the  virtual function doTheWork in a way that unfolds and
// utilizes the information given in rch. The actual execution of the core
// part of the program (the raison d'être of the program) is then
// initiated by calling the function run() of this class. This  function
// is not affected by the derivation - it is already defined in the
// present class .
// In a way it is indirectly virtual since it calls the virtual
// function doTheWork(). This is a common pattern in non-trivial C++
// programs.

Z setDimApp(Word const& prog, Word const& path="");
// sets dimApp from the arguments that can also be used to create an
// application. The reason is that one sometimes needs the 'mode'
// parameter from the ini file in order to start the right application.
// returns the value of 'selection.mode' of the ini-file and
// sets dimApp equal to this value if it is 2 or 3

class IniFileBasedApplication{
//creating file-based applications which document their input and output

protected:

   Z size;
   Z rank;
      // to support parallelism
      // Actually, no longer active. I did not purge all remnants of the
      // 'parallel aera' of my programming since the project pala would
      // then loose some interesting structures. These deal with
      // distributing a large system of particles over a large piece of
      // space which is partictioned in smaller pieces. The algorithm for
      // particle dynamics is organized in a way that the particles
      // located in any of these smaller pieces can be handled in one
      // of many separate processes (in parallel).

   RecordHandler rch;
      // The only non-trivial constructor of the present class
      // reads data from a file that is given by providing
      // its name and (optionally path) as argument. These data
      // load the storage area of data element rch which is capable
      // of holding data of types B,Z1,R1,Word,V<B>,V<Z1>,V<R1>,V<Word>
      // and provides bidirectional access to data of the related types
      // bool,Z,R,string,V<bool>,V<Z>,V<R>,V<string>.

      // The value (state) of rch is the essential determinant of
      // any meaningful instantiation of the class.

      // Due to the polytypic nature of the data type RecordHandler
      // rch can easily hold all information needed to determine the
      // run of an arbitrarily complex program.

      // So, as already stated, a proven design pattern for a scientific
      // application is to derive an application class from the present
      // class and re-define the virtual function doTheWork in a way that
      // unfolds and utilizes the information given in rch.

    Word programTitle_;
      // Typically a descriptive name for the fuctionality of the
      // present instance of the class. A whole 'program' such as
      // my 'PaLa' combines several classes derived from
      // class IniFileBasedApplication and is referred to by the
      // title held in CpmApplication::title_.
      // The present attribute does not get initialized by calling a
      // constructor but using member function setTitle(Word).

// dependent data: values will be initialized during calls to the
// constructor or to function run() from information in rch.

   V<Word> inFiles;
      // List of file names that define input values. The constructor
      // of the present class (and also no other member function) does
      // open any of the inFiles. This opening will typically be done
      // in the redefined function doTheWork. It may be convenient to
      // extend the filename under control of file cpmconfig.ini in the
      // style:
      // Word fiForOpening = Message::ext(inFiles[i]);
      // IFileStream ifs(fiForOpening);

   V<Word> outFiles;
      // list of files that are created on output
      // The names of the files that actually get created have runName
      // appended

// From here on we follow the good practice to alienate the names by an
// underscore in order to not unintentionally use the same name in
// programming member functions. The previous 5 members occur so often
// in existing code that it would be a major activity to change the
// names.

   Z sel_;
      // a selector for the 'depth' of auto-generated documentation
      // files.
      // Control of auto-generated documentation files is based on the
      // three entries sel_, mode_, runName_. These are the rules:
      // 1. for sel_<0 no documentation files will be created.
      // 2. for sel_==0 a light documentation file is created that
      //    covers all the input so that the program run can be repeated.
      // 3. for sel_>0, in addition to the light documentation file
      //    a more detailed one will be created that also can made to
      //    list computational results.
      // 4. All documentation files have names which contain runName_
      //    appended to the name of the calling program
      // 5. Also the parameter mode_ will be appended to the
      //    names of the documentation files if sel_>0

   Z mode_;
      // allows the application run in different modes, which can be
      // discerned by looking to their runname. The model for this
      // parameter is a dimension parameter which switches in a modeling
      // program from 2D to 3D modeling.

   Word runName_;
      // identifier for the program run that will be appended to
      // the names of documentation files.
      // It will be read from section 'selection' and key 'runName'
      // of the ini-file. For instance, the following lines may
      // ocurr in the ini-file
/*************************************************************************
   selection
   W runName=090615mp50
   // refers to date and precision 50 of a multiple precision computation
*************************************************************************/
      // If the value of this quantity is "auto" this value will
      // be replaced by an autogenerated name.
      // Depending on the value of
      // static bool CpmSystem::Message::appRunId
      // runName_ will be changed by appending the same run identifier
      // to it which was used to index the log files cpmcerr and cpmdata.

// end of documentation control

   Word programName_;
      // a name of the program, here only a tag. We assume that a
      // ini-file exists the name of which is
      // path&dirSep&programName_&".ini"

   V<Z> appendInFiles_;
      // a list with as many entries as inFiles; controls
      // whether an inputfile will be appended to documentation file
      // programName&"-"&runName&"_docu.txt"
      // 0 for not append, 1 for append
      // For most inputfiles, documentation is recommended. For programs
      // that simply view large moviefiles (which in this application are
      // inputfiles) it is essential however to be in a position to
      // disable the time-consuming copying of the input file

   V<Z> appendOutFiles_;
      // a list with as many entries as outFiles; controls
      // whether an outputfile - which will be created anyway -
      // additionally will be appended to file
      // programName&"-"&runName&"_docu.txt"
      // 0 for not append, 1 for append

   R tStart_;
      // time as given by cpmtime() at program start

   R tStop_;
       // time as given by cpmtime() at program stop

   R tWaitAnnounceScreen_;
      // time for reading the introduction screen

   R tWaitWorkScreen_;
      // wait time after doTheWork. Typically available for
      // ispecting the last screen produced by
      // function doTheWork.
      // For this to work one needs to have a statement
      //    cpmwait(tWaitWorkScreen_);
      // after the last screen display statement of doTheWork().

   void documentTheWork();// not const !
      // could, for instance, change runName by appending a code for the
      // execution time

   virtual void announceTheWork();
      // typically a start screen

   virtual void doTheWork(void) // was const till 2001-10-03
       // re-defining this function in a derived class
       // defines the specific behavior of the application
       // thus defined. Opening inFiles and reading them
       // is a typical preparatory task of this function.
   { cpmwarning(
      "IniFileBasedApplication: trivial version of doTheWork() called");
   }

public:
   typedef IniFileBasedApplication Type;

   virtual IniFileBasedApplication* clone()const
   { return new IniFileBasedApplication(*this);}

   IniFileBasedApplication(void):sel_(0), mode_(0){}
      // notice that most data have default constructor

   explicit IniFileBasedApplication(Word const& prog);
      // Creates an instance of IniFileBasedApplication from data
      // that are obtained by reading the file named
      // Message::ext(prog&".ini")=: fn which will be referred to as the
      // ini-file of the application.
      // Finally invokes Record::readFromFile(fn,true);
      // and this, in turn, invokes Record::readFormatted(istream& str)
      // with str obtained by opening file fn. So the
      // function Record::readFormatted(istream& str) finally determines
      // the syntax to be followed by the ini-file.

      // Example of the structure of this file
/****************** example for ini file ********************************
// text for explanation

about
   W file=...
   Ws line1=What is done here
   Ws line2=...
   Ws line3=...

selection
   Z sel=1
   Z cpmverbose=1
   Z mpiverbose=0
   Z cpmdbg=1
   Z mode=2
   W runName=021211a
   ...
input
   Z numberOfFiles=3
   W files1=movie1_station_xx.txt
   // irrelevant
   W files2=movie1_station_xx.txt
   // irrelevant
   W files3=movie3_station_xx.txt
   // irrelevant
   Z doc1=0
   Z doc2=0
   Z doc3=0
output
   Z numberOfFiles=3
   W files1=movie1
   W files2=movie2
   W files3=movie3
   Z doc1=0
   Z doc2=0
   Z doc3=0
   ...
****************************** end of file example *****************/

   void setTitle(Word const& w){ programTitle_=w;}
      // see explanation to data member programTitle_
   Word getTitle()const{ return programTitle_;}
      // see explanation to data member programTitle_
   Word getRunName()const{ return runName_;}
   Word getProgramName()const{ return programName_;}
   Z nInFiles()const{ return inFiles.dim();}
   Z nOutFiles()const{ return outFiles.dim();}
   void setModeFromInFiles();
      // In some programs based on IniFileBasedApplication the inFiles
      // are assumed to be the outFiles of other programs also based
      // on IniFileBasedApplication. These outfiles have the mode
      // parameter normally appended, and thus allow the mode-value
      // which fits the nature of these files to be read from
      // the file names of the inFiles under consideration. This
      // mechanism (which changes the values of mode and runName)
      // will be executed by calling the present function
   void set_tWaitWorkScreen_(R t){tWaitWorkScreen_=t;}

   void run(void);
      // runs the application and documents input and output

   void sizeAndRank(){
      size=1;
      rank=1;
#if defined(CPM_USE_MPI)
      size=CpmMPI::Cpm_com.getSize();
      rank=CpmMPI::Cpm_com.getRank();
#endif
   }
};

////////////////////////// class Task //////////////////////////////////

class Task{ // tool for implementing a menu-like application structure
   // A class that makes it easy to define function main_() of a program
   // so that it allows a menu-like structure
   Word name_;
      // device that allows to call f_ as an autonomous
      // action within class Task
   F<Word,Z> f_;
      // a representation of potentially extreme complicated
      // code that is to be executed in performing the task
 public:
   typedef Task Type;
   Word nameOf()const{ return Word("Task");}
   CPM_IO
   Task(){}
      // does nothing
   Task(Word const& name, F<Word,Z> const& f):name_(name),f_(f){}
      // creates a task from the attributes
   Task(Word const& name, Z (*f)(Word const&) ):name_(name),f_(f){}
      // creates a task from the attributes
   Z work()const{ return f_(name_);}
      // executes the task, for which the constructors
      // provided the code
   F<Word,Z> get_f()const{ return f_;}
      // helps to use proven earlier code that works with
      // the functions instead of the full data of a Task
      // (see implementation of comAsRoot)
   static Task subMen(Word const& name, F<Word,Z> const& f);
      // making a menu by providing a list of ini-files for
      // the code provided by function f. The list of ini-files
      // is assumed to be defined in a file name&".ini". This
      // will be searched in directories ../control, ./control, ./,
      // (tested in this order) relative to the execution directory ./
      // The structure of the expected ini-file is as in the following
      // example:
/*
// file pala2.ini
// Particle system enclosed by walls

topics
   Z begin=1
   Z end=15
      // selecting the first and the last index of the
      // ini-file that will be done one after the other
      // (which is particularly convenient for demo programs)

   Ws topics01=danceOfGrains  danceOfRubberGrains  magneticChains
   Ws topics02=vertical_momentum  thermalization  allInteractions
   Ws topics03=allInteractionsEtot  larmor_cycles  plasma
   Ws topics04=crystallization  crystallization2  brownian_motion
   Ws topics05=osmosis1  osmosis2  FaradayCage
   Ws topics06=top2Dev
      // a list of names of ini-files that can easily be extended
      // and for which the access index (starting with 1) is obvious
      // from formatting if one chooses to place always the same number
      // of names in a line
      // (The maximum number of topics is 16 so far)

path
   W path=./control
   // path saying where the ini-files danceOfGrains.ini,
   // danceOfRubberGrains.ini, ... the names of which are given
   // in topics1, topics2, ... will be taken from

control

   R tWait=2
      // wait time after
*/
   static Task subMen(Word const& name, Z (*f)(Word const&) )
      { return subMen(name, F<Word,Z>(f));}
      // see previous function
   static Task men(Word const& name, V<Task> const& tasks);
      // Creates a root-level menu from a ini-file that must have the
      // structure of a commandline as in the following example
/*
// top.ini
command_line
   // obligatory header

// 1: Rigid body dynamics
// 2: Particle system with walls
// 3: Movie viewer

Ws args=2 2 2 ./control/ 4
   // after the '=' write just as you would write on the command line
   // after the name of the program to be called.
   // e.g. 3 2 10 c:/pala/control/ 15
   // where
   // 3  is the subprogram selector,
   // 2  is the value of cpmverbose until it becomes reset by
   //    additional input from a ini-file
   // 10 is the value of tWait, the time to see the final graphics screen
   //    in seconds.
   // c:/pala/control/ is a location descriptor where to find
   //    top1.ini, top2.ini, ...    that should be used
   // after the '=' write just as you would write on the command line
   // after the name of the program to be called.
   // e.g. 3 2 10 c:/pala/control/ 15
   // where
   // 3  is the subprogram selector,
   // 2  is the value of cpmverbose until it becomes reset by
   //    additional input from a ini-file
   // 10 is the value of tWait, the time to see the final graphics screen
   //    in seconds.
   // c:/pala/control/ is a location descriptor where to find
   //    top1.ini, top2.ini, ...    that should be used
   // 15 is a further wait time, now for the introduction screen
   //    that a user may want to read accurately for the first time
   //    but that shine up only for very few seconds in routine usage.
*/
};

/*
// Typical application (assuming
// Z top1(Word const& path),
// Z top2(Word const& iniFile),
// Z top3(Word const& path)   )

Z main_()
{
// we define tasks and combine them
   Task top2=Task::subMen("top2",top2c);
   Z nT=3;
   V< Task > vTask(nT);
   vTask[1]=Task("top1",top1);
   vTask[2]=top2;
   vTask[3]=Task("top3",top3);
   Task rt=Task::men("top",vTask);
// finally we execute them (by executing their
// combined version)
   return rt.work();
}

as an example for the complex code imported by the
functions topi I give top2:

Z top2(Word const& iniFile)
{
   using namespace CpmApplication;
   SysControl();
   setDimApp(iniFile);
   if (dimApp==2){
      CpmTop2::Sys2Appl a(iniFile);
      a.run();
   }
   else{
      CpmTop3::Sys2Appl a(iniFile);
      a.run();
   }
   return 2;
}
*/

////////////////////////////////////////////////////////////////////////

//////////////////// menueGenerator() ///////////////////////////////////
/*                               Example of usage:
int main(void)
{
   Z nMenue=4;
   Word title="Hilfsprogramme zur fotografischen Belichtung. Ulrich\
    Mutze\
    1999";

   V<Word> items(nMenue);
   items[1]="Berechnung der Belichtungszeit fuer fotografische\
    Aufnahmen"\
   ;
   items[2]="Berechnung der Leuchtdichte aus fotografischen\
    Belichtungsdaten";
   items[3]="Berechnung des Logarithmus zur Basis 2";
   items[4]="Ablesung logarithmischer Skalen";

   V< F<Z,Z> > f(nMenue);
   f[1]=F<Z,Z>(test3b); // functions to be called on menue item.
   // Functions
   // where of
      // type void f(void) before, change to Z f(Z) is trivial
   f[2]=F<Z,Z>(test3a);
   f[3]=F<Z,Z>(test0);
   f[4]=F<Z,Z>(test2);

   return menueGenerator(title,items,f,1);
}
*/

   Z menueGenerator(Word const& title, const V<Word>& items,
      const V< F<Z,Z> >& f, Z lang=2 );
// title holds a title for the program part generated by means of this
// function items holds the titles of the main menue items f holds
// the functions to be executed for the various menue items. If
// these are actually coded as normal functions of prototype Z g(Z),
// then e.g. f[3]=F<Z,Z>(g) is a valid initialization of component
// of f.
// lang=1 für deutsche Texte, lang=2 for English texts

} // namespace

#endif
