//? cpmlinalg.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_LINALG_H_
#define CPM_LINALG_H_
/*

   Description: classes R1, R2, R3, R4, RR1, RR2, RR3, RR4
   Basis for effective geometry in 1,2,3,4-dimensional spaces.
   Defines linear spaces as non-template classes without virtual
   functions.
   Only standard C++.   C++ types used:
      void, bool, istream&, ostream&
   Metric properties of linear spaces like scalar product, and absolute
   value are provided. Dimension dependent concepts like vector product
   are provided for R2 and R3.

   Component access via indexes is save but out-of-range indexes silently
   access the nearest component.

*/

#include <cpmtypes.h>
#include <cpmv.h>
#include <cpmc.h>

namespace CpmLinAlg{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::C;
   using CpmRoot::Word;
 //  using CpmRootX::R1;
   using CpmArrays::V;
   using CpmArrays::X2;
   using CpmFunctions::F;

////////// Tuples of 2,3,4 components of type R, defining linear algebra.

///////////////////////////// class R2 /////////////////////////////////

class R2{  //  tuples of two real elements
public:
   R x1;
   R x2;
   typedef R2 Type;

// constructors
   R2():x1(0),x2(0){}
   R2(const R& a, const R& b):x1(a),x2(b){}
   R2(const R& a, const R& b, const R& c):x1(a),x2(b){c;}
      // for formal uniformity to R3, third argument is
      // passive. Introduced 2004-04-29
explicit R2(const V<R>&);
      // construction from dimension-independent lists
      // no restiction to the argument
// access
explicit R2(C z):x1(z[1]),x2(z[2]){}
   C toC()const{ return C(x1,x2);}
   R& operator[](Z i){ if (i>=2) return x2; else return x1;}
   const R& operator[](Z i)const{ if (i>=2) return x2; else return x1;}
      // 'else' should not be needed. Was intoduced in 2001-7-27 in order
      // to avoid malfunction in Dim2::Spc::toC() which, then, was based
      // on Dim2::Vec::operator[], Dim2::Vec derived from R2.
      // now also operator[] for Z2,R3,R4,C2,C3,C4 is defined by a
      // complete if, else if, ...else if,else selection statement
   virtual R operator()(Z i)const
      {if (i<1 || i>2) return 0; else return (*this)[i]; }
   void set(R t){ x1=t; x2=t;}
// mutating arithmetics (linear algebra)
   R2& operator +=(const R2& a)
   { x1+=a.x1; x2+=a.x2; return *this;}
   R2& operator -=(const R2& a)
   { x1-=a.x1; x2-=a.x2; return *this;}
   R2& operator *=(const R& r)
   { x1*=r; x2*=r; return *this;}
   R2& neg(void)
   { x1=-x1; x2=-x2; return *this;  }
   R2 operator-()const{ return R2(-x1,-x2);}

   R2 pos()const{ return R2(cpmabs(x1),cpmabs(x2));}
      // returns a positive version of *this

   R operator |(const R2& a)const
   { return x1*a.x1+x2*a.x2;}
   R crsPrd(const R2& a)const{ return x1*a.x2-x2*a.x1;}
      // related to cross product (vector product)
   R abs(bool careful=true)const;
   R absSqr()const{return x1*x1+x2*x2;}
   R meanValue()const{ return (x1+x2)*0.5;}
   R dist(R2 const& a)const;
      // distance
   R distSquare(R2 const& a)const;
      // square of distance
   R dis(R2 const&)const;
   R sup()const{ return x1<x2 ? x2:x1;}
   R inf()const{ return x1<x2 ? x1:x2;}
   R mod()const{ R a=abs(); return a==0 ? R(0) : (sup()-inf())/a;}
      // modulation, measures the deviation from constancy
   R2 operator~()const{ return R2(x2,x1);}
   X2<R,R2 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0)
   R2 carToPol()const;
      // Cartesian coordinates (x1,x2) to polar coordinates.
      // Returns R2(r,phi) with r=abs(), phi=::atan2(x2,x1)).
      // thus the phi-range is that given by the systems atan2 function.
      // Needless to say that angles are here treated as arcs of
      // unit circles of type R. The class CpmGeo::Angle is not
      // used in this file.
   R arg()const{ return cpmatan2(x2,x1);}
   R2 polToCar()const;
      // Polar coordinates (r,phi) to  Cartesian coordinates (x1,x2).
      // Returns R2(x1*cos(x2),x1*sin(x2))
   R angle(R2 const& a)const;
      // Angle between *this and a (eqals that between a and *this). Ranges
      // from 0. to pi.
   R2 rot90()const{ return R2(-x2,x1);}
      // rotation 90 (degrees)
      // Returns a vector which results from *this by a rotation
      // of 90 degrees in positive direction. I came to consider this
      // an important operation by studying lecture notes of Professor
      // Engelbert Schlipf (Fachhochschule Ulm) on technical mechanics.
      // January 2008, added here 2015-01-12
   R normalize()
      { X2<R,R2> p=polDec(); *this=p.c2();  return p.c1();}
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information
   R normalize_(R r=1);
      // turns *this into a vector of length r and returns the norm of
      // the original in order not to loose information. If *this was
      // the zero vector, it will be turned into  a 'standard vector'
      // of length r.
   R nor_(R r=1){ return normalize_(r);}
      //: normalize
      // C+- conformant name
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 2;}
   bool isVal()const
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2);}

   R2 ran(Z j=0)const;
   virtual Word nameOf()const{ return Word("R2");}
   static V<R2> canBas(){ return V<R2>{R2(1,0),R2(0,1)};}
      // canonical basis
};

inline R2 operator +(const R2& x1, const R2& x2)
{ R2 res=x1; return res+=x2;}


inline R2 operator -(const R2& x1, const R2& x2)
{ R2 res=x1; return res-=x2;}

inline R2 operator *(const R2& x1, const R& s)
{ R2 res=x1; return res*=s;}

inline R2 operator *(const R& s, const R2& x1)
{ R2 res=x1; return res*=s;}

inline R R2::distSquare(R2 const& a)const{ return (*this-a).absSqr();}

inline R R2::dist(R2 const& a)const{ return (*this-a).abs();}

///////////////////////////// class Z2 /////////////////////////////////

class Z2{  //  tuples of two integer elements
public:
   Z x1;
   Z x2;
public:
   typedef Z2 Type;
// constructors
   Z2():x1(0),x2(0){}
   Z2(const Z& a, Z const& b):x1(a),x2(b){}
explicit Z2(const V<Z>&);
      // construction from dimension-independent lists
      // no restiction to the argument
// access
   Z& operator[](Z i){ if (i>=2) return x2; else return x1;}
   Z const& operator[](Z i)const{ if (i>=2) return x2; else return x1;}
   void set(Z t){ x1=t; x2=t;}
// mutating arithmetics (linear algebra)
   Z2& operator +=(const Z2& a)
   { x1+=a.x1; x2+=a.x2; return *this;}
   Z2& operator -=(const Z2& a)
   { x1-=a.x1; x2-=a.x2; return *this;}
   Z2& operator *=(Z const& r)
   { x1*=r; x2*=r; return *this;}
   Z2& neg(void)
   { x1=-x1; x2=-x2; return *this;  }
   Z operator |(const Z2& a)const
   { return x1*a.x1+x2*a.x2;}
   R abs(bool careful=true)const;
   Z absSqr()const{return x1*x1+x2*x2;}
   Z2 operator~()const{ return Z2(x2,x1);}
   Z sup()const{ return x1<x2 ? x2:x1;}
   Z inf()const{ return x1<x2 ? x1:x2;}
   R dis(const Z2&)const;
   X2<R,R2 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0)
   CPM_IO
   CPM_ORDER
   CPM_WORD
   Z dim()const{ return 2;}
   //: dimension
   Word nameOf()const{ return Word("Z2");}
   //: name of
   bool isVal()const
   //: is valid
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2);}

// Two functions that were suggested by a graph-theoretic application.
   bool intSect(Z2 const& b)const
   //: intersect
   // The return value says whether the sets {x1,x2} and {b.x1,b.x2}
   // have elements in common.
   {  return (x1==b.x1||x1==b.x2||x2==b.x1||x2==b.x2);}

   bool fit(Z2 const& b)const
   //: fit (together)
   // a.fit(b) means that a and b 'can be chained together':
   // (x1,x2)(b.x1,b.x2) no gap since x2==b.x1
   {  return x2==b.x1;}

};

inline Z2 operator +(const Z2& x1, const Z2& x2)
{ Z2 res=x1; return res+=x2;}

inline Z2 operator -(const Z2& x1, const Z2& x2)
{ Z2 res=x1; return res-=x2;}

inline Z2 operator -(const Z2& x)
{ Z2 res=x; return res.neg();}

inline Z2 operator *(const Z2& x1, Z const& s)
{ Z2 res=x1; return res*=s;}

inline Z2 operator *(Z const& s, const Z2& x1)
{ Z2 res=x1; return res*=s;}

///////////////////////////// class R3 /////////////////////////////////

class R3{  //  tuples of 3 real elements
public:
   typedef R3 Type;
   R x1;
   R x2;
   R x3;
//public:
// constructors
   R3():x1(0.),x2(0.),x3(0.){}
   R3(const R& a, const R& b, const R& c):x1(a),x2(b),x3(c){}
   R3(const R& a, const R& b):x1(a),x2(b),x3(0.){}
      // added 02-09-07 to have common constructor for R2 and R3
      // which in some code is Rdim
explicit R3(const V<R>&); // construction from dimension-independent
// lists, no restiction to the argument.
// access
   R& operator[](Z i){ if (i>=3) return x3; else if (i==2) return x2;
    else return x1;}
   const R& operator[](Z i)const{ if (i>=3) return x3; else if (i==2)
    return x2; else return x1;}
   virtual R operator()(Z i)const
      {if (i<1 || i>3) return 0; else return (*this)[i]; }
   void set(R t){ x1=t; x2=t; x3=t;}
// mutating arithmetics (linear algebra)
   R3& operator +=(const R3& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; return *this;}
   R3& operator -=(const R3& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; return *this;}
   R3& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   R3& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; return *this;  }
   R3 operator-()const{ return R3(-x1,-x2,-x3);}
   R operator |(const R3& a)const
   { return x1*a.x1+x2*a.x2+x3*a.x3;}

   R3 crsPrd(R3 const& a)const
      // cross product (vector product)
   { return  R3(x2*a.x3-x3*a.x2,
                x3*a.x1-x1*a.x3,
                x1*a.x2-x2*a.x1); }

   R3 pos()const{ return R3(cpmabs(x1),cpmabs(x2),cpmabs(x3));}
      // returns a positive version of *this

   R abs(bool careful=true)const;
   R absSqr()const{return x1*x1+x2*x2+x3*x3;}
   R meanValue()const{ return (x1+x2+x3)/3;}
   R dist(R3 const& a)const;
      // distance
   R distSquare(R3 const& a)const;
      // square of distance
   R dis(R3 const&)const;
   R sup()const{ return CpmRootX::sup<R>(x1,x2,x3);}
   R inf()const{ return CpmRootX::inf<R>(x1,x2,x3);}
   R mod()const{ R a=abs(); return a==0 ? R(0) : (sup()-inf())/a;}
      // modulation, measures the deviation from constancy
   X2<R,R3> polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0)

   R3 carToPol()const;
      // Cartesian coordinates (x1,x2,x3) to polar coordinates
      // (r,phi,delta) where delta is the angle relative to the 1-2-plane
      // as is declination with respect to the equator plane in
      // astronomy.
      // This is in most circumstances more convenient and symmetric
      // than a polar angle theta measuring the angle against the
      // 3-direction. The precise definition is:
      // r=abs(),
      // phi=R2(x1,x2).carToPol()[2]
      // delta=R2(rho,x3).carToPol()[2]
      // where  rho:=R2(x1,x2).carToPol()[1]
      // It turned out that taken this as implementation causes dark
      // behavior with MS C++ 6.0. Now we express the same scheme
      // in a form that does not call R2::functions.

   R3 polToCar()const;
      // Polar coordinates to  Cartesian coordinates.
      // Returns R3(rho*cos(x2),rho*sin(x2),x1*sin(x3)),
      // where rho:=x1*cos(x3)

   R angle(R3 const& a)const;
      // Angle between *this and a (eqals that between a and *this). Ranges
      // from 0. to pi.

   R normalize()
      { X2<R,R3> p=polDec(); *this=p.c2();  return p.c1();}
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information

   R normalize_(R r=1);
      // turns *this into a vector of length r and returns the norm of
      // the original in order not to loose information. If *this was
      // the zero vector, it will be turned into  a 'standard vector'
      // of length r is

   R nor_(R r=1){ return normalize_(r);}
      //: normalize
      // C+- conformant name

   CPM_IO
   CPM_ORDER
   CPM_WORD
   Z dim()const{ return 3;}
   bool isVal()const
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2)
      &&CpmRoot::isVal(x3);}

   R3 ran(Z j=0)const;
      // returns a random element where the range of the components
      // is that by *this
   R3 ran(R r)const;
      // Returns a random element in the ball of radius r and center *this.
      // The result is determined by the present value of the static
      // counter variable j0 in function R CpmRoot::randomR(Z j). Of
      // course, this variable becomes incremented (at least trice)
      // by a call to ran(R).

   virtual Word nameOf()const{ return Word("R3");}
   static V<R3> canBas()
      { return V<R3>{R3(1,0,0),R3(0,1,0),R3(0,0,1)};}
      // canonical basis

// operations as used in my publication: 'Rigidly connected spherical
// particles: a versatile grain model' Granular Matter 2006
   R3 ast(R3 const& x)const; // ast for asterisk, \ast in LaTex
      // returns (*this)*x according to equation (6)
   R3 circ(R3 const& r)const; // circ for \circ of LaTex
      // returns (*this)o r according to equation (7)
   R3 exp(bool exact=true)const // see equation (11) of paper cited above
      // For exact==false the approximation for infinitesimal
      // transformations is used.
   {
      R3 res=*this;
      if (!exact) return res*=0.5;
      R av=abs();
      if (av==0) return res;
      R fac=cpmtan(av*0.5)/av;
      return res*=fac;
   }

   R3 ln()const // inverse function to exp
   {
      R3 res=*this;
      R av=abs();
      if (av==0) return res;
      R fac=(cpmatan(av)/av)*2;
      return res*=fac;
   }

   R3 toEulerAng()const;
      // Here we interpret *this as a rotation vector (list of
      // Euler-Rodrigues parameters) and return the Euler angles
      // phi, theta, psi.
      // Source: my notes UM/Astro 87-11-4/6

   R3 dig(R3 const& diag)const
      //: diagonal
      // Action of a diagonal matrix (represented by the diagonal
      // as a R3 object) on *this.
   { return R3(x1*diag.x1,x2*diag.x2,x3*diag.x3);}

};

inline R3 operator +(const R3& x1, const R3& x2)
{ R3 res=x1; return res+=x2;}

inline R3 operator -(const R3& x1, const R3& x2)
{ R3 res=x1; return res-=x2;}

inline R3 operator *(const R3& x1, const R& s)
{ R3 res=x1; return res*=s;}

inline R3 operator *(const R& s, const R3& x1)
{ R3 res=x1; return res*=s;}

inline R R3::dist(R3 const& a)const{ return (*this-a).abs();}

inline R R3::distSquare(R3 const& a)const{ return (*this-a).absSqr();}

///////////////////////////// class Z3 /////////////////////////////////
// placed behind R3 since type R3 is needed in X2<R,R3 > polDec()const;

class Z3{  // tuples of three integer elements
public:
   Z x1;
   Z x2;
   Z x3;
public:
   typedef Z3 Type;

// constructors
   Z3():x1(0),x2(0),x3(0){}
   Z3(const Z& a, const Z& b, const Z& c):x1(a),x2(b),x3(c){}
   Z3(const Z& a, const Z& b):x1(a),x2(b),x3(0){}
      // added 02-09-07 treat Z3 like R3
explicit Z3(const V<Z>&); // construction from dimension-independent
// lists
      // no restiction to the argument
// access
   Z& operator[](Z i){ if (i>=3) return x3; else if (i==2) return x2;
    else return x1;}
   const Z& operator[](Z i)const{ if (i>=3) return x3; else if (i==2)
    return x2; else return x1;}
   void set(Z t){ x1=t; x2=t; x3=t;}
// mutating arithmetics (linear algebra)
   Z3& operator +=(const Z3& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; return *this;}
   Z3& operator -=(const Z3& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; return *this;}
   Z3& operator *=(const Z& r)
   { x1*=r; x2*=r; x3*=r; return *this;}
   Z3& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; return *this;}
   Z operator |(const Z3& a)const
   { return x1*a.x1+x2*a.x2+x3*a.x3;}
   R abs(bool careful=true)const;
   Z absSqr()const{return x1*x1+x2*x2+x3*x3;}
   R dis(const Z3&)const;
   X2<R,R3 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value)
      // is the absolute value r and the second component a unit vector n
      // such that
      // *this = n*r. If *this is the zero vector, n is arbitrarily
      // chosen
      // as (1,0)
   CPM_IO
   CPM_ORDER
   Z dim()const{ return 3;}
   bool isVal()const
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2)
      &&CpmRoot::isVal(x3);}
   Word nameOf()const{ return Word("Z3");}
};

inline Z3 operator +(const Z3& x1, const Z3& x2)
{ Z3 res=x1; return res+=x2;}

inline Z3 operator -(const Z3& x1, const Z3& x2)
{ Z3 res=x1; return res-=x2;}

inline Z3 operator -(const Z3& x)
{ Z3 res=x; return res.neg();}

inline Z3 operator *(const Z3& x1, const Z& s)
{ Z3 res=x1; return res*=s;}

inline Z3 operator *(const Z& s, const Z3& x1)
{ Z3 res=x1; return res*=s;}

///////////////////////////// class R4 /////////////////////////////////

class R4{  //  tuples of 4 real elements
public:
   typedef R4 Type;
   R x1;
   R x2;
   R x3;
   R x4;
public:
// constructors
   R4():x1(0),x2(0),x3(0),x4(0){}
   R4(const R& a, const R& b, const R& c, const R& d):x1(a),x2(b),
      x3(c),x4(d){}
   R4(R2 const& a, R2 const& b):x1(a.x1),x2(a.x2),x3(b.x1),x4(b.x2){}
explicit R4(const V<R>&); // construction from dimension-independent
// lists
      // no restiction to the argument
// access
   R& operator[](Z i)
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   const R& operator[](Z i)const
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   virtual R operator()(Z i)const
      {if (i<1 || i>4) return 0; else return (*this)[i]; }
   void set(R t){ x1=t; x2=t; x3=t; x4=t;}
// mutating arithmetics (linear algebra)
   R4& operator +=(const R4& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; x4+=a.x4; return *this;}
   R4& operator -=(const R4& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; x4-=a.x4; return *this;}
   R4& operator *=(const R& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   R4& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; x4=-x4; return *this; }
   R4 operator-()const{ return R4(-x1,-x2,-x3,-x4);}
   R4 pos()const{ return R4(cpmabs(x1),cpmabs(x2),cpmabs(x3),cpmabs(x4));}
      // returns a positive version of *this

   R operator |(const R4& a)const
   { return x1*a.x1+x2*a.x2+x3*a.x3+x4*a.x4;}
   R abs(bool careful=true)const;
   R absSqr()const{return x1*x1+x2*x2+x3*x3+x4*x4;}
   R meanValue()const{ return (x1+x2+x3+x4)*0.25;}
   R dist(R4 const& a)const;
      // distance
   R distSquare(R4 const& a)const;
      // square of distance
   R dis(const R4&)const;
   R sup()const{ return CpmRootX::sup<R>(x1,x2,x3,x4);}
   R inf()const{ return CpmRootX::inf<R>(x1,x2,x3,x4);}
   R mod()const{ R a=abs(); return a==0 ? R(0) : (sup()-inf())/a;}
      // modulation, measures the deviation from constancy
   X2<R,R4 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0,0)
   R normalize(){ X2<R,R4> p=polDec(); *this=p.second();
      return p.first();}
      // turns *this into a unit vector and returns the norm of the
      // original in order not to loose information

   R normalize_(R r=1);
      // turns *this into a vector of length r and returns the norm of
      // the original in order not to loose information. If *this was
      // the zero vector, it will be turned into  a 'standard vector'
      // of length r is

   CPM_IO
   CPM_ORDER
   CPM_WORD
   Z dim()const{ return 4;}
   bool isVal()const
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2)
      &&CpmRoot::isVal(x3)&&CpmRoot::isVal(x4);}

   R4 ran(Z j=0)const;
   virtual Word nameOf()const{ return Word("R4");}
   static V<R4> canBas()
   { return V<R4>{R4(1,0,0,0),R4(0,1,0,0),R4(0,0,1,0),R4(0,0,0,1)};}
      // canonical basis
};

inline R4 operator +(const R4& x1, const R4& x2)
{ R4 res=x1; return res+=x2;}

inline R4 operator -(const R4& x1, const R4& x2)
{ R4 res=x1; return res-=x2;}

inline R4 operator *(const R4& x1, const R& s)
{ R4 res=x1; return res*=s;}

inline R4 operator *(const R& s, const R4& x1)
{ R4 res=x1; return res*=s;}

inline R R4::dist(R4 const& a)const{ return (*this-a).abs();}

inline R R4::distSquare(R4 const& a)const{ return (*this-a).absSqr();}

///////////////////////////// class Z4 /////////////////////////////////
// placed behind R4 since type R4 is needed in X2<R,R4> polDec()const;

class Z4{  //  tuples of 4 integer elements
public:
   typedef Z4 Type;
   Z x1;
   Z x2;
   Z x3;
   Z x4;
public:
// constructors
   Z4():x1(0),x2(0),x3(0),x4(0){}
   Z4(const Z& a, const Z& b, const Z& c, const Z& d):x1(a),x2(b),
      x3(c),x4(d){}
explicit Z4(const V<Z>&); // construction from dimension-independent
// lists
      // no restiction to the argument
// access
   Z& operator[](Z i)
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   const Z& operator[](Z i)const
   { if (i>=4) return x4; else if (i==3) return x3;
      else if (i==2) return x2; else return x1;}
   void set(Z t){ x1=t; x2=t; x3=t; x4=t;}
// mutating arithmetics (linear algebra)
   Z4& operator +=(const Z4& a)
   { x1+=a.x1; x2+=a.x2; x3+=a.x3; x4+=a.x4; return *this;}
   Z4& operator -=(const Z4& a)
   { x1-=a.x1; x2-=a.x2; x3-=a.x3; x4-=a.x4; return *this;}
   Z4& operator *=(const Z& r)
   { x1*=r; x2*=r; x3*=r; x4*=r; return *this;}
   Z4& neg(void)
   { x1=-x1; x2=-x2; x3=-x3; x4=-x4; return *this; }
   Z operator |(const Z4& a)const
   { return x1*a.x1+x2*a.x2+x3*a.x3+x4*a.x4;}
   R abs(bool careful=true)const;
   Z absSqr()const{return x1*x1+x2*x2+x3*x3+x4*x4;}
   R dis(const Z4&)const;
   X2<R,R4 > polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // a unit vector n such that *this = n*r. If *this is the zero
      // vector, n is arbitrarily chosen as (1,0,0,0)
   CPM_IO
   CPM_ORDER
   CPM_WORD
   Z dim()const{ return 4;}
   bool isVal()const
   { return CpmRoot::isVal(x1)&&CpmRoot::isVal(x2)
      &&CpmRoot::isVal(x3)&&CpmRoot::isVal(x4);}
   Word nameOf()const{ return Word("Z4");}
};

inline Z4 operator +(const Z4& x1, const Z4& x2)
{ Z4 res=x1; return res+=x2;}

inline Z4 operator -(const Z4& x1, const Z4& x2)
{ Z4 res=x1; return res-=x2;}

inline Z4 operator -(const Z4& x)
{ Z4 res=x; return res.neg();}

inline Z4 operator *(const Z4& x1, const Z& s)
{ Z4 res=x1; return res*=s;}

inline Z4 operator *(const Z& s, const Z4& x1)
{ Z4 res=x1; return res*=s;}

/*************** lean square matrices (1*1) - (4*4) ********************/

#define CPM_MAT_ALG(X,XX)\
friend X operator*(const XX&, const X&);\
friend XX operator*(const XX& , const XX&);\
friend XX operator+(const XX& , const XX& );\
friend XX operator~(const XX& );\
friend X operator*(const X& x, const XX& a){ return ~a*x;}\
friend XX operator-(const XX& a , const XX& b){ return a + (-b);}

// Comment to the last function where matrices are acting on vectors
// from the right side. This kind of action on a vector space gives a
// representation of the multiplicative monoid of matrices if the product
// law for endomorphisms of linear spaces is defined as contravariant.
// In our specific notation here:
// x*(a*b)=(x*a)*b
// Proof: x*(a*b)=~(a*b)*x=((~b)*(~a))*x=(~b)*((~a)*x)=((~a)*x)*b=(x*a)*b

///////////////////////////// class RR2 /////////////////////////////////

class RR2{  //  real 2*2 matrices
   // (*this)[i][k] is just xi.xk
   // the 2*2 matrices get special treatment in order to do
   // justice to the group SL(2,R)
public:
   typedef RR2 Type;
   R2 x1;
   R2 x2;
public:
   static RR2 zero;
   static RR2 unit;
// constructors
   RR2(){}
   RR2(const R2& a1, const R2& a2):x1(a1),x2(a2){}
   RR2(R z11, R z12, R z21, R z22):x1(z11,z12),x2(z21,z22){}
   RR2(R2 const& x, F<R2,R2> const& f, R h=1e-3);
      // Generates the Jacobian of f at point x. Method error is O(h^4).

// access
   R2& operator[](Z i){ if (i>=2) return x2; return x1;}
   const R2& operator[](Z i)const{ if (i>=2) return x2; return x1;}
   R det()const{ return x1.x1*x2.x2-x1.x2*x2.x1;}
   RR2 operator !()const;
      // inverse matrix can be expressed easily in this case
   R brkLinTrn(R z)const
      // broken linear transformation
   { return (x1.x1*z+x1.x2)*CpmRoot::inv(x2.x1*z+x2.x2);}
      // RR2 acts on R as a 'broken linear transformation'
   RR2 operator*(R r)const{ return RR2(x1*r,x2*r);}
   RR2 operator-()const{ return RR2(-x1,-x2);}
   R abs(bool careful=true)const{ return R4(x1,x2).abs(careful);}
      //: absolute (value)
      // Hilbert Schmidt norm
   R absSqr()const{ return x1.absSqr() + x2.absSqr();}
      //: absolute (value) squared
   F<R2,R2> toF()const;
      //: to F
      // Returns the function R2 --> R2, x |--> (*this)*x
      // Thus (RR2(x,mat.toF())-mat).abs()~~0 (~~ means
      // '== up to numerical noise') for all R2 x, RR2 mat.
   CPM_IO
   CPM_WORD
   CPM_MAT_ALG(R2,RR2)
   Word nameOf()const{ return Word("RR2");}
};

R2 operator*(const RR2&, const R2&);
RR2 operator*(const RR2& , const RR2&);
RR2 operator+(const RR2& , const RR2&);
RR2 operator~(const RR2&);
R2 operator*(const R2&, const RR2&);
RR2 operator-(const RR2&, const RR2&);

///////////////////////////// class RR3 /////////////////////////////////

class RR3{  //  real 3*3 matrices
public:
   typedef RR3 Type;
   R3 x1;
   R3 x2;
   R3 x3;
public:
   static RR3 zero;
   static RR3 unit;
// Various generating mechanisms as static functions.
   static RR3 hom(R r)
      //: homothety
      // hom(r)[i][j] =delta[i][j]*r
   {
      RR3 res;
      res.x1.x1=r;
      res.x2.x2=r;
      res.x3.x3=r;
      return res;
   }
   static RR3 digMat(R3 const& a)
      //: diagonal matrix
      // digMat(a)[i][j] =delta[i][j]*a[i]
   {
      RR3 res;
      res.x1.x1=a.x1;
      res.x2.x2=a.x2;
      res.x3.x3=a.x3;
      return res;
   }

   static RR3 hat(R3 const& a)
      //: hat matrix
      // hat(a)[i][j] := sum over k of eps[i][k][j]*a[k]
      // where eps is the well known totally antisymmetric tensor
   {
      RR3 res;
      res.x1.x2=-a.x3;
      res.x1.x3=a.x2;
      res.x2.x3=-a.x1;
      res.x2.x1=a.x3;
      res.x3.x1=-a.x2;
      res.x3.x2=a.x1;
      return res;
   }
   static RR3 dyad(R3 const& a, R3 const& b)
      //: hat matrix
      // hat(a)[i][j] := sum over k of eps[i][k][j]*a[k]
      // where eps is the well known totally antisymmetric tensor
   {
      RR3 res;
      res.x1.x1=a.x1*b.x1;
      res.x1.x2=a.x1*b.x2;
      res.x1.x3=a.x1*b.x3;

      res.x2.x1=a.x2*b.x1;
      res.x2.x2=a.x2*b.x2;
      res.x2.x3=a.x2*b.x3;

      res.x3.x1=a.x3*b.x1;
      res.x3.x2=a.x3*b.x2;
      res.x3.x3=a.x3*b.x3;

      return res;
   }
// constructors
   RR3(){}
   RR3(const R3& a1, const R3& a2, const R3& a3):x1(a1),x2(a2),x3(a3){}
   explicit RR3(R3 const& r);
      // construction from rotation vector (Euler-Rodrigues parameter)
      // Result is a SO(3) matrix.
   RR3(R3 const& x, F<R3,R3> const& f, R h=1e-3);
      // Generates the Jacobian of f at point x. Method error is O(h^4).

// access
   R3& operator[](Z i)
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   const R3& operator[](Z i)const
   { if (i>=3) return x3; if (i==2) return x2; return x1;}

   R det()const;
      //: determinant

   RR3 operator !()const;
      // inverse matrix can be expressed rather easily in this case

   RR3 conGrd()const;
      // contra-gredient

   RR3 operator*(R r)const{ return RR3(x1*r,x2*r,x3*r);}
   RR3 operator-()const{ return RR3(-x1,-x2,-x3);}

   X2<R,R3> toAngAndAxis()const;
      //: to angle and axis
      // For being meaningful, *this should be (at least approximately) a
      // SO(3)-matrix. The axis is a unit vector.

   R3 toRotVec()const
   { X2<R,R3> y=toAngAndAxis(); return y.c2()*cpmtan(y.c1()*0.5);}
      //: to angle and axis
      // For being meaningful, *this should be (at least approximately) a
      // SO(3)-matrix. The axis is a unit vector.

   R absSqr()const{ return x1.absSqr() + x2.absSqr()+ x3.absSqr();}
      //: absolute (value) squared

   R abs()const{ return cpmsqrt(absSqr());}
      //: absolute (value)
      // Hilbert Schmidt norm

   F<R3,R3> toF()const;
      //: to F
      // Returns the function R3 --> R3, x |--> (*this)*x
      // Thus (RR3(x,mat.toF())-mat).abs()~~0 (~~ means
      // '== up to numerical noise') for all R3 x, RR3 mat.

   CPM_IO
   CPM_WORD
   CPM_MAT_ALG(R3,RR3)
   Word nameOf()const{ return Word("RR3");}
};

R3 operator*(const RR3&, const R3&);
RR3 operator*(const RR3& , const RR3&);
RR3 operator+(const RR3& , const RR3&);
RR3 operator~(const RR3&);
R3 operator*(const R3&, const RR3&);
RR3 operator-(const RR3&, const RR3&);

///////////////////////////// class RR4 /////////////////////////////////

class RR4{  //  real 4*4 matrices
public:
   typedef RR4 Type;
   R4 x1;
   R4 x2;
   R4 x3;
   R4 x4;
public:
   static RR4 zero;
   static RR4 unit;
   static RR4 g;
      // Minkowski-metric
// constructors
   RR4(){}
   RR4(const R4& a1, const R4& a2, const R4& a3, const R4& a4):
      x1(a1),x2(a2),x3(a3),x4(a4){}
// access
   R4& operator[](Z i)
   {
      if (i>=4) return x4;
      if (i==3) return x3;
      if (i==2) return x2;
      return x1;
   }

   const R4& operator[](Z i)const
   {
      if (i>=4) return x4;
      if (i==3) return x3;
      if (i==2) return x2;
      return x1;
   }

   RR4 operator*(R r)const{ return RR4(x1*r,x2*r,x3*r,x4*r);}
   RR4 operator-()const{ return RR4(-x1,-x2,-x3,-x4);}
   CPM_IO
   CPM_WORD
   CPM_MAT_ALG(R4,RR4)
   Word nameOf()const{ return Word("RR4");}
};

R3 operator*(const RR4&, const R3&);
RR4 operator*(const RR4& , const RR4&);
RR4 operator+(const RR4& , const RR4&);
RR4 operator~(const RR4&);
R3 operator*(const R3&, const RR4&);
RR4 operator-(const RR4&, const RR4&);

} // namespace

#endif
