//? cpmmultiimage.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_MULTIIMAGE_H_
#define CPM_MULTIIMAGE_H_
/*
   Description: Defines template class MultiImage<T>
   Particularly functional is MultiImage<R>
   to which the typedef-alias ImgR refers.
*/
#include <cpmimage.h>
#include <cpmdiscretespace.h>
#include <cpmframes.h>

namespace CpmImaging{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::L;
   using CpmRoot::Conv;

   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::VVV;
   using CpmArrays::X2;
   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;

   using CpmFunctions::F;

   using CpmLinAlg::R2;
   using CpmLinAlg::R3;

   using CpmGraphics::Img24;
   using CpmGraphics::rgb;
   using CpmGraphics::Frame;
   using CpmGraphics::Frames;
   using CpmGraphics::Graph;

//////////////////////// class MultiImage<> //////////////////////////////

template<class T>
class MultiImage: private V<Image<T> >, public ImagingTools {
// real-valued multi-plane image,  typedef MultiImage<R> ImgR
// images with potentially more than one plane as vectors of images.
// T is the type of pixel values. No positivity of the pixel values
// is assumed. Main application for T = R, for details see
// class Image<T> in file cpmimage.h
// Wide range of methods for processing color images. Convenient
// conversion back and forth with C+- basic color image type Img24.

   typedef V< Image<T> > Base;

public:

   MultiImage( const V< Image<T> >& vimg):V< Image<T> >(vimg){}
      // upcast, no new member introduced in deriving MultiImage<T>
      // from V< Image<T> >

// constructors which do not provide pixel values

   MultiImage(Z p, Z m, Z n ):
      // p: number of image planes (e.g. 3 for RGB images)
      // m: number of lines
      // n: number of colums
   V<Image<T> >(p,Image<T>(m,n)){}

   explicit MultiImage( Z p ):
      // p: number of image planes (e.g. 3 for RGB images)
      V<Image<T> >(p,Image<T>()){}

   MultiImage(void):
      // default constructor
   V<Image<T> >(1,Image<T>()){}

// constructors which import pixel values from other data structures

   MultiImage(Z p, const Image<T>& im):
      // constructs a p-fold repetition of im
      V<Image<T> >(p,Image<T>(im)){}

   template <class Typ2> // Typ2=L was the only option till 2011-07-17
   MultiImage(Z p, const Image<Typ2>& im, T (*f)(Z,Typ2));
      // constructs a p-fold repetition of the geometry of im
      // and coloring the pixels using im together with a look-up
      // function f. f(i,val) is the pixel value to which the pixel
      // value val of im is translated in the i-th image plane of
      // the MultiImage to be constructed

   MultiImage(Z p, const VV<T>& y):
      // constructs a p-fold repetition of a image constructed from
      // a Matrix<T,Z>
   V<Image<T> >(p,Image<T>(y)){}

   MultiImage( const VVV<T>& y);
      // constructs a MultiImage<T> from a VVV<T> (which is the
      // natural triple-indexed data structure

   MultiImage( const VV<T>& y);
      // constructs a MultiImage<T> from a VV<T> (which is the
      // natural double-indexed data structure

   MultiImage(Frame const& fr);
      // constructor from a Frame (graphics window)
      // and its pictorial content. This is a convenient method
      // if one wants to apply more sophisticated image processing
      // to an image created on screen. Works only for T=R

   MultiImage(Img24 const&);
      // construction native basic color image format.
      // Only for T=R.

   MultiImage(VV<Img24> const& vv, Z m, Z n);
      // construct a MultiImage of m lines an n columns
      // by composing a matrix of Img24-images into
      // it by suitable scaling.
      // Only for T=R

   MultiImage( Word w, VV<C> const& vv);
      // We consider the case that vv is obtained as fourierTransform
      // from a GrayImgR. We will transform it in a ImgR from which
      // finally an Img24 can be made which can finally stored
      // as a ppm-imagefile.



// trivial access functions

   Z getLin(void)const{ return ((*this)[1]).getLin();}
      // get number of lines. This refers to the first image plane.
      // Presently all constructors generate multi-images
      // for which pixel count is independent of the image plane
      // Notice that ((*this)[1]).getLin() would be terribly
      // inefficient since calculating (*this)[1] implies a return
      // and thus a copy constuctor for images!.
      // 96-1-29: this is no longer true, since now (*this)[1] would
      // be a reference

   Z getCol(void)const{ return ((*this)[1]).getCol();}
      // get number of columns

   Z h()const{ return getLin();}
   Z w()const{ return getCol();}

   Z dim1()const{ return getLin();}
   Z dim2()const{ return getCol();}

   Z getPla(void)const{ return Base::dim();}
      // get number of 'planes'

// color plane access
   Image<T> getPla(Z i)const{ return (*this)[i];}

// pixel-value access

   T getPix(Z h, Z i, Z j )const{ return ((*this)[h]).getPix(i,j);}
      // acessing pixel values outside the image possible.
      // continuation as constant

   T getPix(Z i, Z j )const;
      // same as previous function but mean value over all channels
      // returned

   void putPix(Z h, Z i, Z j, T const& t){((*this)[h]).putPix(i,j,t);}

   void putPix(Z i, Z j, Color const& c){
      ((*this)[1]).putPix(i,j,c.red());
      ((*this)[2]).putPix(i,j,c.green());
      ((*this)[3]).putPix(i,j,c.blue());
   }

   T getLocalMean(Z h, Z i, Z j, const V<Mask>& a)const
      // forming a local mean value by masks which are allowed to
      // vary from image plane to image plane
   { return ((*this)[h]).getLocalMean(i,j,a[h]);}

   T operator()(Z h, R x, R y, Z method=1)const
      // returns the value at a point in the h-th plane with respect to a
      // normal coordinate system (origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1. This system
      // will be referred to as base system. Pixelation, or
      // bilinear interpolation according to 'method' as in function
      // interpolate()
   { return ((*this)[h])(x,y,method);}

   T operator()(R x, R y, Z method=1)const;
      // same as previous function but mean value over all channels
      // returned

   R2 minMax(R xL=0, R xU=1, R yL=0, R yU=1)const;
      // Let R2 lim = minMax(); Then lim.x1 is the lowest of all pixel
      // values (after conversion to R, all image components)
      // taken into account) and lim.x2 is the highest.

// image transformations

   MultiImage<T>& unVeil_()
      //:: unveil
      // subtracts the largest constant so that the result is a
      // positive-valued image. When applied to an RGB photo, the result
      // is such that at least in one pixel one of the color colorplanes
      // gives a zero-value. If the darkest areas in an photo are gray,
      // they are black after this transformation. Since for each pixel
      // the subtracted value does not depend on the channel, we 'subtract
      // white' and thus change (increase) saturation but not hue. So, the
      // transformation should look natural even for photos in which the
      // darkest areas are not really dark (as when shadows are lightened
      // up by haze.
   { return operator +=(-minMax().x1);}

   MultiImage<T>& setRan_(R valMin=0, R valMax=255,
      R gamma=Color::getGamma(),bool hueFix=true);
      // We assume as nearly always that T can be converted to R.
      // We assume further that all pixel values are >= 0 under this
      // conversion.
      // 1. Functionality for hueFix==true:
      // Here the hue of no pixels is not changed by the gamma
      // transformation.
      // We form the monochrome image max() image and perform
      // on it a gamma transformation which throws its maximum value to
      // valMax. This transformation is extended to all channels such that
      // for each pixel the all channels get multiplied by the same
      // positive real factor. Thus the (generalized) hue and the
      // saturation remains unchanged in any pixel. Further, valMax is the
      // maximum of all pixel values. In order to make valMin the minimum
      // value, we add or subtract white. This operation does not change
      // hue but it does change saturation. Anyway, it does change the
      // maximum of the pixel values. We restore this by a multi-
      // pliation of all channels and all pixels by a common factor. This
      // operation changes the minimum value a bit which is considered OK.
      //
      // So, valMax is guaranteed to be the maximum pixel value of the
      // transformed image and valMin is only a tendative prescription of
      // the minimum pixel value.
      // 2. Functionality for hueFix!=true:
      // Here each imageplane is transformed with
      // Image<T>::setRan_(T,T,R).
      // This looks very natural but the hue of pixels gets sometimes a
      // bit shifted. Till 2011-09-01 this was the only existing mode.

   MultiImage<T> interpolate(Z m, Z n, Z method)const;
      // returns a m*n MultiImage created from *this by a interpolation
      // method selected by the parameter method.
      // method =1 means bilinear interpolation and method=0 means
      // pixelation. The latter is useful if the number of pixels is to
      // be compared to expectations.

   Image<R> combine( R (*f)(const R_Vector& x) )const;
      // Returns a single-component image, obtained by using function
      // f to combine the pixel values in the various planes into one
      // pixel value.

   VVV<T> toVVV()const;
      // conversion to the natural mathematical type

   Img24 toImg24()const;
      // conversion to Img24, the image type of namespace CpmGraphics
      // This works for MultiImage<R> only.
      // If all pixel values are converted directly from R3 into ColRef
      // No adjustment since this is uniformly done by setRange(R,R,R)

   Image<R> locVar(bool useLocVar=false)const;
      //: local variation
      // Returns a real-valued monochrome image the brightness of which
      // is proportional to the local variation of image *this.
      // The intention is to let parts of electronic camera images, that
      // are out of focus appear as dark, whereas focussed parts are
      // bright. Presently this does not work really well.
      // With the default value of the argument, we use the Laplacian
      // (made positive) which seems a bit better, at least much simpler.

   Image<R> mono()const;
      // Returns a mono-chrome version defined as the mean value of
      // all image planes.

   Image<R> max()const;
      // Returns a mono-chrome version defined as the max value of
      // all image planes.

   Image<R> min()const;
      // Returns a mono-chrome version defined as the minimum value of
      // all image planes.

   void insertPattern(R X, R Y, R dX,
      const  F<C, Va<T> >& f, const F<C,B>& g,
      R x=0, R y=0, R dx=1, R dy=1, B average=B(false));
      // modifying a part of an image by values derived from function
      // values on the part
      //   X <= Re(z) <= X+dX
      //   Y <= Im(z) <= Y+dY
      // of the complex plane. dY is calculated from dX by aspect ratio
      // g(z)=trueUM indicates that f(z) will be calculated
      // and the values f(z)[1],... f(z)[p] will be stored to the pixel
      // corresponding to z in image planes 1,...p.
      // If average=trueUM, not only z belonging to the current pixel is
      // used but four positions and an average is formed.
      // This was designed to give alias-reduced gray-tone images for
      // binary patterns. In this case T has to such that automatic
      // type conversion from and to R is defined. L is a valid choice
      // for T.

   void insert(MultiImage<T> const& img, Rect const& rect, Z scale=0);
      // changes *this by inserting the image img 'into the subarea rect'
      // Since the format of img might not fit to this prescription
      // verbally, there are options, controlled by parameter scale which
      // give a precise meaning to this:
      // scale==0:
      // The image is inserted on a pixel by pixel basis, starting at the
      // left upper corner of rect. If img is larger (pixel-wise) than
      // rect, then  only the part in rect is changed.
      // scale==1:
      // The image img is scaled (respecting the aspect ratio) such that
      // it just can be held completely by rect. Scaling is done by
      // bilinear interpolation.

// image display on screen
    void mark(Frame& fr, R2 lim, Z method=0, Z faithful=1)const;
      // Writes the image *this onto the Frame fr.
      // This does not display the image. fr.vis() needs to
      // be called for showing the image on screen.
      // A copy of *this is acted on by
      // setRan_(lim[1],lim[2],Color::getGamma()) causing a proper
      // tonescale transformation. Giving lim one or two negative
      // components avoids this step.
      // The Frame fr also determines the number of pixels to
      // be displayed. Since this may not coincide with the number of
      // pixels of the image, a suitable transformation has to be done.
      // This is controlled via parameters method and faithful.
      // The parameter 'method' controls pixel rendition: 0 normal
      // pixelation 1 bilinear interpolation. If the parameter 'faithful'
      // is 1, the image will be placed in such a manner to the upper
      // left position that its aspect ratio is correctly reproduced on
      // the screen. Otherwise it is made filling the whole frame.
      // Addition 2004-10-17: For method=2, the averaging interpolator
      // is used which however for images with many pixels is very very
      // slow. The method could be of value for images with very small
      // number of pixels that nevertheless have to be shown on a
      // large screen area without annoyingly obvious pixel structure.
      // This function may also be used as a testbed for the
      // averaging interpolator (see cpmalgorithms.h)

   R2 mark(Frame& fr,Z method=0, Z faithful=1)const
      // The actual range of the image will be mapped to
      // display range.
    {
      R2 lim=minMax();
      mark(fr,lim,method,faithful);
      return lim;
    }

    void markDirect(Frame& fr, Z method=0, Z faithful=1)const
      // The pixel value range [0,255] will be mapped to
      // display range. This is the is the adequate functionality
      // for images got by reading TARGA files or other file formats
      // which adopt 1 BYTE per pixel and image plane.
    {
      R2 lim(-1,-1); // so function setRan_ will not be called
      mark(fr,lim,method,faithful);
    }

   void print(void);
      // outputs pixel count to screen

// multiplication of images for flat field correction

   MultiImage<T>& operator *=(T const& t);

   MultiImage<T>& operator +=(T const& t);

   MultiImage<T>& operator *=(const Image<T>& im);

   MultiImage<T>& operator *=(const MultiImage<T>& im);

   MultiImage<T>& operator /=(const Image<T>& im);

   MultiImage<T>& operator /=(const MultiImage<T>& im);

   MultiImage<T> cut(const Rect& a)const;
      // reduces the size of the image by cutting it down to the
      // rectangular subframe defined by Rect a.

   MultiImage<T> cut(R xL, R xU, R yL, R yU)const;
      // reduces the size of the image by cutting the borders. All
      // arguments have to be in the interval [0,1].
      // The part of the image which is left after cut has normal
      // coordinates (x,y) with respect to the original which satisfy
      //   xL <= x <= xU
      //   yL <= y <= yU
      // Normal coordinate system: origin lower-left, x-axis
      // to the right, y-axis vertically ascending. The upper-right
      // image corner has coordinates x=1, y=1.

   MultiImage<T> crop(Iv const& ivx, Iv const& ivy)const;
      //: crop
      // Creates a rectangular cut-out of the image by giving the
      // subinterval of the x-axis and the subinterval of the
      // y-axis to be used. Taking the whole image corresponds
      // to ivx=Iv(0,1), ivy=Iv(0,1). The origin (x,y)=(0,0) is
      // the upper-left corner of the image. Notice that for images with
      // an aspect ratio different from 1, the aspect ratio of the croped
      // image differs from the quotient ivx.abs()/ivy.abs().
      // Same style as Img24::crop().

   MultiImage<T> cut(Z2 cLL, Z2 cUR)const;
      // similar to the previous function. Here the lower left corner
      // and the upper right corner are given as pixel positions in
      // graphics format.

   MultiImage<T> cut(IvZ const& iRange, IvZ const& jRange)const;

   MultiImage<T> clip(R clipL, R clipU)const;
      // x.clip(a,b) takes values between a*x.min() and b*x.max()

   MultiImage<T>& operator -=(const MultiImage<T>& im);
      // returns *this-im, the pixelwise difference of images.
      // *this and im are assumed to coincide in numbers of planes,
      // lines, and columns

   MultiImage<T>& apply(const MultiImage<T>& im,
      const F< X2<T,T>, T>& f);
      // general operation of one multi-image onto another  multi-image
      // of the same pixel type

    MultiImage<T>& apply(const MultiImage<T>& im,
       const V<F<X2<T,T>,T> >& f);
      // Analogous to previous function. Uses different 'functions' for
      // combining the pixel values in the various planes of the
      // multi-images. These 'functions' are the components of the
      // function array f.

   MultiImage<T> operator -(const MultiImage<T>& im1)const
    { MultiImage<T> res=*this; return res-=im1;}
      // difference of images in normal notation

    MultiImage<T> operator +(const MultiImage<T>& im1)const
       // sum of images in normal notation
    {
       MultiImage<T> im2=im1;
       T fac(-1);
       im2*=fac;
       MultiImage<T> res=*this; return res-=im2;
    }

    MultiImage<T> overLay(MultiImage<T> const& im1, R fac=0.5)const
       //: over-lay
       // adds *this and a 'dimmed' (or intensified) version of im1
       // and returns the result.
    {
       MultiImage<T> im2=im1;
       T factor(-fac);
       im2*=factor;
       MultiImage<T> res=*this; return res-=im2;
    }

// comparison of images

   R norm1(void)const;
      // sum over the absolute values over all pixel values, divided by
      // number of pixels

   R norm2(void)const;
      // sum over the squares of all pixel values, divided by number of
      // pixels.
      // Square root of all of that. (Normalization is such, that the
      // norm of a constant image is equal to the pixel value)

   X2<R_Vector,Z2> hotspot(Z dm=0, Z dn=0);
      // returns the pixel value for all planes at the point where the
      // sum of all pixel values is maximum
      // actually the mean of a (1+2*dm)*(1+2*dn) neighborhood is given

   MultiImage<T> conv(R_Matrix const& a, Z si=0, Z sj=0)const;
      // Matrix a is taken as a convolution kernel with a(ci+si,cj+sj)
      // considered the center of the kernel.
      // Here
      // ci := 1+dim1(a)/2
      // cj := 1+dim2(a)/2
      // which is the actual center if the dimensions of a both are odd
      // which is the normal case.
      // No normalization will be done on a, since this can be done more
      // flexibly in advance.

   MultiImage<T> conv(R_Vector const& a, R_Vector const& b)const;
      // does the same as conv(R_Matrix(a,b)) but much more
      // efficiently. Makes use of the separability of the
      // kernel.

   MultiImage<T> smooth(Z n=1)const;
      // convolution with {{1,2,1},{2,4,2},{1,2,1}}/16

   MultiImage<T> conv(const R& sigma, Z dL=5);
      // The result is the convolution of *this with a numerical
      // approximation to a rotationally symmetric Gaussian distribution
      // where sigma is the standard deviation of center distance r.
      // The distribution is approximated as a (2d+1)(2d+1)-matrix with
      // maximum values arround the center, where d is reasonably
      // calculated from sigma. dL is a limiter for d. d<=dL is
      // guaranteed.

   MultiImage<T> conv(const V<Mask>& a)const;
      // modern form of convolution, defined by a family of masks,
      // one for each image plane

   MultiImage<T> conv(const Mask& a)const;
      // modern form of convolution, defined by one mask,
      // to be applied to each image plane

   MultiImage<T> deScreen(R lim=0.1, R rim=0.1, Z p=3)const;
      // Not yet mature algorithm. Good exercize to work with
      // Fourier transformed images. Showed the need for defining
      // a few new functions.
      // 0<=lim<=1, the smaller lim the more changes will be made
      // 0<=rim<=1, the smaller rim the lower the frequencies that can
      // be affected.
      // The masks for detecting artificial maxima in momentum space
      // has total width 2*p+1 pixels

   Img24 deScreen24(R lim=0.1, R rim=0.1, Z p=3)const;
      // Not yet mature algorithm. Good exercize to work with
      // Fourier transformed images. Showed the need for defining
      // a few new functions.
      // 0<=lim<=1, the smaller lim the more changes will be made
      // 0<=rim<=1, the smaller rim the lower the frequencies that can
      // be affected.
      // The masks for detecting artificial maxima in momentum space
      // has total width 2*p+1 pixels

   MultiImage<T> fourierFilter(F<R2,R> const& resp)const;
      // multiplies the fourier transform by function resp which
      // takes meaningful arguments in [0,1]x[0,1], where 0 corresponds
      // to frequency 0 and 1 to Nyquist frequency. The negative
      // frequencies (which are not independent) are treated
      // automatically in a consistent manner.
      // T is assumed to allow meaningful two-way conversion to R.
      // The resulting image has lines and columns equal to powers of
      // 2. The input image is not required to have this property.
      // For calculating its fourier transform we fill the missing pixels
      // in by 0.
      // After multiplication of the fourier transformed image with resp
      // (even if resp is positive) the back-transformed image needs not
      // to be positive. This is no problem since the display functions
      // shift the values anyway into a representable range.

   Img24 fourierFilter24(F<R2,R> const& resp)const;
      // Similar to the previous function. But value range transformed
      // to [0,255] using Color::gamma. The pixel count is that of the
      // original *this and not the power-of-two version of the
      // intermediate fourier transform.

   Img24 fourierFilter24(R_Func const& f)const
      // Special case of previous construct, where the damping function
      // is the same for both x-direction and y-direction of the image.
      { return fourierFilter24(f.tenPrd(f));}

   Img24 fourierFilter24(R x0=0.75, R x1=1.0)const;
      // Special case of previous construct. The damping of high
      // frequencies is here parametrized by parameters
      // 0 <= x0 < x1 <= 1.
      // or spatial frequencies larger as x0*fNyquist the
      // smooth damping starts and is complete (100%) for
      // frequencies larger or equal x1*fNyquist.
      // This produces a very pleasing image structure also
      // for noisy originals.

   MultiImage<T> fftFilter(F<C,C> const& resp)const;
      // Similiar to fourierFilter(F<R2,R>). It uses the
      // square [-0.5,0.5]x[-0.5,0.5] instead of [0,1]x[0,1]
      // and operates on the amplitude instead of the intensity.
      // Assumes T=R.

   Img24 fftFilter24(F<C,C> const& resp)const;
      // Similiar to fftFilter(F<C,C>). But value range transformed
      // to [0,255] using Color::gamma. The pixel count is that of the
      // original *this and not the power-of-two version of the
      // intermediate fourier transform.

   R getPixConv(Z h, Z i, Z j, R_Matrix const& a, Z si=0, Z sj=0);
      // makes convolution ('Faltung') with R_Matrix in pixel plane
      // h: Colorplane, i,j: pixel

   R_Vector getPixConvPla( Z i, Z j, R_Matrix const& a);
      // makes convolution ('Faltung') with R_Matrix in color plane
      // and returns a vector with all new colorvalues of the pixel i,j


   Histogram histogram(Z nBins=16,R valMin=0, R valMax=255);
      // makes a histogram of an Image
      // nBins:   number of Sectors
      // valMin:   minimal value (normally 0)
      // valMax:     maximal value (normally 255)

   VV<T> cut(Z h, Rect const& rect)const;
      // creates a rectangular cut-ut of the h'th image plane of
      // *this as a a matrix.
      // See class Rect for details on the description of the
      // rectangular area.
   bool isVal()const;
      //: is valid
};

typedef MultiImage<R> ImgR;
   // in analogy to Img24

namespace{
   R absFunc(C const& z){ return z.abs();}
}

template<class T>
MultiImage<T>::MultiImage( Word w, VV<C> const& vvc):
V<Image<T> >(3,Image<T>(vvc.dim1(),vvc.dim2()))
{
   Z mL=1;
   Word loc("MultiImage(Word, VV<C>)");
   CPM_MA
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   VV<R> vvr=vvc(absFunc);
   w;
   R_Matrix rm(vvr);
   Iv iv=rm.infSup();
   Z i,j;
   for (i=1;i<=m;++i){
      for (j=1;j<=n;++j){
         C cij=vvc[i][j];
         Color col(cij,iv);
         putPix(1,i,j,col.red());
         putPix(2,i,j,col.green());
         putPix(3,i,j,col.blue());
      }
   }
   CPM_MZ
 }

template<class T>
template<class Typ2>
MultiImage<T>::MultiImage(Z p, const Image<Typ2>& im, T (*f)(Z,Typ2)):
   V<Image<T> >(p,Image<T>(im))
{
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   for (h=1;h<=l;h++) for (i=1;i<=m;i++) for (j=1;j<=n;j++) {
      putPix(h,i,j,f(h,im.getPix(i,j)));
   }
}

template <class T>
void MultiImage<T>::print(void)
{
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   T t;
   Word w=nameOf(t);

   cout<<endl<<" This is a MultiImage<> with values of type "<<w;
   cout<<endl<<l<<" planes, "<<m<<" lines, "<<n<<" columns";
}

template<class T>
bool MultiImage<T>::isVal()const
{
   for (Z h=1;h<=getPla();++h){
      bool b=((*this)[h]).isVal();
      if (!b) return false;
   }
   return true;
}

template <class T>
MultiImage<T>::MultiImage(VV<Img24> const& vv,  Z m, Z n ):
   V<Image<T> >(3,Image<T>(m,n))
{
   Z ny=vv.dim1();
   Z nx=vv.dim2();
   Iv ivx(0.,1.);
   Iv ivy(0.,1.);
   V<Iv> vivx=ivx.divide(nx);
   V<Iv> vivy=ivy.divide(ny);
   Z i,j,scale=1;
   // notice that Rect has a (x,y) understanding of layout,
   // whereas vv is to be understood as matrix layout
   for (i=1;i<=ny;++i){
      for (j=1;j<=nx;++j){
         MultiImage<T> mij(vv[i][j]);
         insert(mij,Rect(vivx[j],vivy[1+ny-i]),scale);
      }
   }
}

template <class T>
MultiImage<T>::MultiImage( const VVV<T>& y):
V<Image<T> >( y.dim1(),Image<T>(y.dim2(), y.dim3()))
{
   Z l=getPla();  // notice initialization
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   T val;
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"MultiImage<>(Tensor3&): plane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"MultiImage<>(Tensor3&): line "<<i;}
         for (j=1;j<=n;j++){
            val=y(h,i,j);
            putPix(h,i,j,val);
         }
      }
   }
}

template <class T>
MultiImage<T>::MultiImage( const VV<T>& y):
V<Image<T> >(1,Image<T>(y.dim1(), y.dim2()))
{

   Z m=getLin();
   Z n=getCol();
   Z i,j;
   T val;
   for (i=1;i<=m;i++){
      if (verbose>9){ cout<<endl<<"MultiImage<>(Matrix&): line "<<i;}
      for (j=1;j<=n;j++){
         val=y(i,j);
         putPix(1,i,j,val);
        }
    }
}

template<class T>
MultiImage<T> MultiImage<T>::interpolate(Z mI, Z nI, Z method)const

{
   Z l=getPla(),h;
   MultiImage<T> res(l); // trivial multi-image with l components
   // redefining the components.
   for (h=1;h<=l;h++){
      res[h]=((*this)[h]).interpolate( mI, nI, method);
   }
   return res;
}

template<class T>
R2 MultiImage<T>::minMax(R xL, R xU, R yL, R yU)const
{
   Z l=getPla(),h;
   R_Vector minis(l);
   R_Vector maxis(l);
   for (h=1;h<=l;++h){
      R2 r2=getPla(h).minMax(xL,xU,yL,yU);
      minis[h]=r2.x1;
      maxis[h]=r2.x2;
   }
   return R2(minis.inf(),maxis.sup());
}

template<class T>
VVV<T> MultiImage<T>::toVVV()const
{
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   VVV<T> res(l,m,n);
   Z h,i,j;
   for (h=1;h<=l;h++) for (i=1;i<=m;i++) for (j=1;j<=n;j++)
      res[h][i][j]=((*this)[h])[i][j];

   return res;
}

/************************
template<class T>
void MultiImage<T>::mark(Frame& fr, R2 lim, Z method, Z faithful)const
{
   Z mL=2;
   Word loc("MultiImage<>::mark(Frame,R2,Z,Z)");
   CPM_MA
   MultiImage<T> temp=*this;
   if (lim.x1>=0 && lim.x2>=0)
      temp.setRan_(lim.x1,lim.x2,Color::getGamma());

   Z l=getPla();
   cpmassert(l==3,loc);
   cpmassert(method<=1,loc);

   Z m=getLin(), n=getCol();
   Z h,i,j;
   R width=1., hight=1.;
   if (faithful){ // we assume square pixels on screen
      R alpha_=((*this)[1]).getAspRat(); // aspect ratio of image
      R alpha=fr.getAspRat(); // aspect ratio of frame
      if (alpha_ > alpha){ // image has to be placed full width
         width=1.;
         hight=alpha/alpha_;
      }
      else{ // image has to be placed full hight
         width=alpha_/alpha;
         hight=1.;
      }
   }
   V<R> lattice1=ImagingTools::pixelCenters(0.,width,n);
   V<R> lattice2=ImagingTools::pixelCenters(0.,hight,m);

   Z2 index;
   R2 p;
   R3 vals;
   R pI, qI;
   Z left,right,lower,upper;
   for (i=1;i<=fr.getLin();i++){
      index.x1=i;
      for (j=1;j<=fr.getCol();j++){
         index.x2=j;
         p=fr.getxFyF(index); // point in frame system
         if (p.x2>hight || p.x1>width) continue;
         ImagingTools::find(lattice1, p.x1, pI, left, right);
         ImagingTools::find(lattice2, p.x2, qI, lower, upper);
         // left ... upper refer to discretized xF, yF
         lower=1+m-lower;
         upper=1+m-upper;
         // now left ... upper refer to the matrix index grid
         // pixel values are converted to R for (optional)
         // interpolation and input to ImagingTools::response
         R a11,a12,a21,a22;
         for (h=1;h<=3;h++){
            a11=temp.getPix(h,upper,left);
            a12=temp.getPix(h,upper,right);
            a21=temp.getPix(h,lower,left);
            a22=temp.getPix(h,lower,right);
            vals[h]=ImagingTools::int2(a11,a12,a21,a22,pI,qI,method);
         }
         rgb c(vals.x1,vals.x2,vals.x3);
         fr.put(index,Color(c));
      }
   }
   CPM_MZ
}
*******************/

template<class T>
void MultiImage<T>::mark(Frame& fr, R2 lim, Z method, Z faithful)const
{
   Z mL=2;
   Word loc("MultiImage<>::mark(Frame,R2,Z,Z)");
   CPM_MA
   R ga=Color::getGamma();
   MultiImage<T> temp=*this;
   if (lim.x1>=0 && lim.x2>=0  && method<2)
      temp.setRan_(lim.x1,lim.x2,ga);

   Z l=getPla();
   cpmassert(l==3,loc);
   cpmassert(method<=2,loc);

   Z m=getLin(), n=getCol();
   Z h,i,j;
   R width=1., hight=1.;
   if (faithful){ // we assume square pixels on screen
      R alpha_=((*this)[1]).getAspRat(); // aspect ratio of image
      R alpha=fr.getAspRat(); // aspect ratio of frame
      if (alpha_ > alpha){ // image has to be placed full width
         width=1.;
         hight=alpha/alpha_;
      }
      else{ // image has to be placed full hight
         width=alpha_/alpha;
         hight=1.;
      }
   }
   V<R> lattice1=ImagingTools::pixelCenters(0.,width,n);
   V<R> lattice2=ImagingTools::pixelCenters(0.,hight,m);

   Z2 index;
   R2 p;
   R3 vals;
   R pI, qI;
   Z left,right,lower,upper;
   if (method<=1){
      for (i=1;i<=fr.getLin();i++){
         index.x1=i;
         for (j=1;j<=fr.getCol();j++){
            index.x2=j;
            p=fr.getxFyF(index); // point in frame system
            if (p.x2>hight || p.x1>width) continue;
            ImagingTools::find(lattice1, p.x1, pI, left, right);
            ImagingTools::find(lattice2, p.x2, qI, lower, upper);
            // left ... upper refer to discretized xF, yF
            lower=1+m-lower;
            upper=1+m-upper;
            // now left ... upper refer to the matrix index grid
            // pixel values are converted to R for (optional)
            // interpolation and input to ImagingTools::response
            R a11,a12,a21,a22;
            for (h=1;h<=3;h++){
               a11=temp.getPix(h,upper,left);
               a12=temp.getPix(h,upper,right);
               a21=temp.getPix(h,lower,left);
               a22=temp.getPix(h,lower,right);
               vals[h]=ImagingTools::int2(a11,a12,a21,a22,pI,qI,method);
            }
            rgb c(vals.x1,vals.x2,vals.x3);
            fr.put(index,Color(c));
         }
      }
   }
   else if (method==2){ // very slow experimental method,
      // 0 and 1 are the well established methods
      R lx1=lim.x1;
      R lx2=lim.x2;
      Z nGrid=m*n, k=1;
      V<R2> x(nGrid);
      V<R3> y(nGrid);
      for (i=1;i<=m;++i){
         Z iy=m-i+1;
         R ly=lattice2[iy];   // 2 <--> y
         for (j=1;j<=n;++j){
            R lx=lattice1[j]; // 1 <--> x
            x[k]=R2(lx,ly);
            // this is the position to which the pixel value
            // getPix(*,i,j); belongs, where position is what it
            // is for function Frame::show(F<R2,R3> const&,...)
            R3 val;
            R v1=getPix(1,i,j);
            R v2=getPix(2,i,j);
            R v3=getPix(3,i,j);
            val[1]=ImagingTools::response(lx1,lx2,0,255,ga,v1);
            val[2]=ImagingTools::response(lx1,lx2,0,255,ga,v2);
            val[3]=ImagingTools::response(lx1,lx2,0,255,ga,v3);
            y[k++]=val;
         }
      } // these two loops arn't the bottle neck.
      CpmGeo::DisField<R2,R3>::setGrid(x);
      CpmGeo::DisField<R2,R3> df;
      df.setY(y);
         // now df knows x and y
      F<R2,R3> fc=df.contField();
         // for this kind of object we have a show-method in class
         // Frame. Each evaluation of fc is hard work even for small
         // images. So this is only an experimental method.
      bool noDisplay=true;
      Graph gr(fr);
      gr.show(fc,0,width,0,hight,noDisplay);
   }
   else{
      cpmwarning(loc&": invalid argument method");
   }
   CPM_MZ
}

template <class T>
MultiImage<T>::MultiImage(const Img24& img)
{
   cpmmessage("MultiImage(Img24) started");
   Z m=img.getLin();
   Z n=img.getCol();
   MultiImage<T> res(3,m,n);
   Z i,j;
   R prfrac=1./m;
   for (i=1;i<=m;i++){
      cpmprogress(" putPix()-loop in constructor",R(i)*prfrac);
      for (j=1;j<=n;j++){
         rgb pixel=img.getPel(j-1,i-1);
         res.putPix(1,i,j,(T)pixel.r);
         res.putPix(2,i,j,(T)pixel.g);
         res.putPix(3,i,j,(T)pixel.b);
      }
   }
   *this=res;
   cpmmessage("MultiImage(Img24) done");
}

template <class T>
MultiImage<T>::MultiImage(const Frame& fr)
{
   *this=MultiImage<T>(fr.getImg());
}

template<class T>
MultiImage<T>& MultiImage<T>::operator *=(T const& t)
{

   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   if (verbose){ cout<<endl<<"Multiplying an image by a scalar started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            putPix(h, i, j, t*getPix(h,i,j) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::operator +=(T const& t)
{

   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   if (verbose){ cout<<endl<<"Adding a scalar to an image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            putPix(h, i, j, t+getPix(h,i,j) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::operator *=(const Image<T>& im)
{
   Word loc("MultiImage<T>::operator *=(Image<T>&)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   if (verbose){ cout<<endl<<"Multiplication of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            putPix(h, i, j, getPix(h,i,j)*im.getPix(i,j) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::operator *=(const MultiImage<T>& im)
{
   Word loc("MultiImage<T>::operator *=(MultiImage<T>&)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(l==im.getPla(),loc);
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   if (verbose){ cout<<endl<<"Multiplication of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            putPix(h, i, j, getPix(h,i,j)*im.getPix(h,i,j) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::operator /=(const Image<T>& im)
{
   Word loc("MultiImage<T>::operator /=(Image<T>&)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;

   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   T val;
   if (verbose){ cout<<endl<<"Division of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            val=im.getPix(i,j);
            if (val==0){
               putPix(h, i, j, 0 );
            }
            else{
               putPix(h, i, j, getPix(h,i,j)/val );
            }
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::operator /=(const MultiImage<T>& im)
{
   Word loc("MultiImage<T>::operator /=(MultiImage<T>&)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   T val;
   if (verbose){ cout<<endl<<"Division of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            val=im.getPix(h,i,j);
            if (val==0){
               putPix(h, i, j, 0 );
            }
            else{
               putPix(h, i, j, getPix(h,i,j)/val );
            }
         }
      }
   }
   if (verbose){ cout<<endl<<"Division of image finished";}
   return *this;
}

template<class T>
MultiImage<T> MultiImage<T>::cut(R xL, R xU, R yL, R yU)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z i=1;i<=l;i++){
      res[i]=((*this)[i]).cut(xL,xU,yL,yU);
   }
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::crop(Iv const& ivx, Iv const& ivy)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z i=1;i<=l;i++){
      res[i]=((*this)[i]).crop(ivx,ivy);
   }
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::cut(Z2 cLL, Z2 cUR)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z i=1;i<=l;i++){
      res[i]=((*this)[i]).cut(cLL,cUR);
   }
   return res;
}

template<class T>
Image<R> MultiImage<T>::mono()const
{
   return combine(ImagingTools::fMono);
}

template<class T>
Image<R> MultiImage<T>::max()const
{
   return combine(ImagingTools::fMax);
}

template<class T>
Image<R> MultiImage<T>::min()const
{
   return combine(ImagingTools::fMin);
}

template<class T>
Image<R> MultiImage<T>::locVar(bool useLocVar)const
{
   return mono().locVar(useLocVar);
}

template<class T>
MultiImage<T>& MultiImage<T>::operator -=(const MultiImage<T>& im)
   // returns *this-im, the pixelwise difference of images.
   // *this and im are assumed to coincide in numbers of planes, lines,
   // and columns
{
   Word loc("MultiImage<T>::operator -=");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(l==im.getPla(),loc);
   if (verbose){ cout<<endl<<"Subtraction of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            putPix(h,i,j, getPix(h,i,j)-im.getPix(h,i,j) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::apply(const MultiImage<T>& im,
   const F<X2<T,T>,T>& f)
{
   Word loc("MultiImage<T>::apply(...)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(l==im.getPla(),loc);
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   if (verbose){ cout<<endl<<"apply() of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            X2<T,T> arg( getPix(h,i,j), im.getPix(h,i,j) );
            putPix(h, i, j, f(arg) );
         }
      }
   }
   return *this;
}

template<class T>
MultiImage<T>& MultiImage<T>::apply(const MultiImage<T>& im,
   const V<F<X2<T,T>,T> >& f)
{
   Word loc("MultiImage<T>::apply(MultiImage<T>&,V<F<X2<T,T>,T> >&)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   cpmassert(l==im.getPla(),loc);
   cpmassert(m==im.getLin(),loc);
   cpmassert(n==im.getCol(),loc);
   if (verbose){ cout<<endl<<"apply() of image started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            X2<T,T> arg( getPix(h,i,j), im.getPix(h,i,j) );
            putPix(h, i, j, (f[h])(arg) );
         }
      }
   }
   return *this;
}

template<class T>
R MultiImage<T>::norm1(void)const
   // sum over the absolute values over all pixel values, divided by
   // number of pixels
{
   Word loc("MultiImage<T>::norm1(void)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   R sum=0.;   // even if T is a integer type !
   R val;
   if (verbose){ cout<<endl<<"MultiImage<>::norm1 started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            val=getPix(h,i,j);
            R valr=CpmRoot::absT<T>(val);
            sum+=valr;
         }
      }
   }
   R nPix=1.*l*m*n;
   cpmassert(nPix>0,loc);
   sum/=nPix;
   return sum;
}

template<class T>
R MultiImage<T>::norm2(void)const
   // sum over the squares of all pixel values, divided by number of
   // pixels.
   // Square root of all of that. (Normalization is such, that the norm
   // of a constant image is equal to the pixel value)
{
   Word loc("MultiImage<T>::norm2(void)");
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   R sum=0., val;
   if (verbose){ cout<<endl<<"MultiImage<>::norm2 started";}
   for (h=1;h<=l;h++){
      if (verbose){ cout<<endl<<"treating pane "<<h;}
      for (i=1;i<=m;i++){
         if (verbose>9){ cout<<endl<<"treating line "<<i;}
         for (j=1;j<=n;j++){
            val=getPix(h,i,j);
            sum+=val*val;
         }
      }
   }
   R nPix=1.*l*m*n;
   cpmassert(nPix>0,loc);
   sum/=nPix;
   cpmassert(sum>=0,loc);
   sum=cpmsqrt(sum);
   return sum;
}

template <class T>
MultiImage<T>& MultiImage<T>::setRan_(R valMin, R valMax, R gamma,
                                      bool hueFix )
{
   Z mL=2;
   Word loc("MultiImage<T>::setRan_(R,R,R)");
   CPM_MA
   Z l=getPla(),h;
   if (!hueFix){
      for (h=1;h<=l;++h){
         ((*this)[h]).setRan_(valMin,valMax,gamma);
      }
   }
   else{

   Image<R> maxMono=max();
   R2 mm=maxMono.minMax();
   R vmax=mm.x2;

   R gi=R(1)/gamma;
   R vi=cpmpow(valMax,gi)/vmax;// (vmax*vi)^gamma = valMax

   Z m=getLin();
   Z n=getCol();
   Z i,j;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R s=maxMono[i][j];
         R fac= s==0. ? 1. : cpmpow(s*vi,gamma)/s;
         for (h=1;h<=l;h++){
            R v=getPix(h,i,j);
            v*=fac;
            putPix(h,i,j,Conv<R,T>(v)());
         }
      }
   }
   // Now we have an image with the intented maximum value. What is the
   // minimum value (over all channels)?
   Image<R> minMono=min();
   R vmin=minMono.minMax().x1;
   R shift=valMin-vmin;
   R valMaxNew=valMax+shift;
   R fac2=valMax*cpminv(valMaxNew);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
        for (h=1;h<=l;h++){
            R vhij=getPix(h,i,j);
            vhij+=shift;
            vhij*=fac2;
            putPix(h,i,j,Conv<R,T>(vhij)());
         }
      }
   }
   }
   CPM_MZ
   return *this;
}

template <class T>
Img24 MultiImage<T>::toImg24()const
{
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   Img24 img(n,m);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         R3 vals;
         for (h=1;h<=3;h++){
            vals[h]=(R)getPix(h,i,j);
         }
         rgb c(vals.x1,vals.x2,vals.x3);
         img.putPel(j-1,i-1,c);
      }
   }
   return img;
}

template<class T>
X2<R_Vector,Z2> MultiImage<T>::hotspot(Z di, Z dj)
{
   R hu=1e12;

   R valMax=-hu;
   Z iMax,jMax;

   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;

   R val;
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         val=0.;
         for (h=1;h<=l;h++){
            val+=getPix(h,i,j);
               // sum over all channels defines total brightness
         }
         if (val>valMax){
            valMax=val;
            iMax=i;
            jMax=j;
         }
      }
   }
   R normfac=(1+2*di)*(1+2*dj);
      // number of terms in summation
   normfac=R(1)/normfac;
   R_Vector res1(l);
   Z si,sj;
   Z Di=di;
   Z Dj=dj;
   for (h=1;h<=l;h++){
      val=0.;
      for (si=-Di;si<=Di;si++){ // does not work for di,dj instead of
      // Di,Dj
         for (sj=-Dj;sj<=Dj;sj++){
            val+=((*this)[h]).getPixCont(iMax+si,jMax+sj);
               // no problem with leaving the boundary due to getPixCon
         }
      }
      res1[h]=val*normfac;
   }
   Z2 res2(iMax,jMax);
   return X2<R_Vector,Z2>(res1,res2);
}

template<class T>
Image<R> MultiImage<T>::combine( R (*f)(R_Vector const&) )const
{
   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j;
   R_Vector x(l);
   Image<R> res(m,n);
   for (i=1;i<=m;i++){
      for (j=1;j<=n;j++){
         for (h=1;h<=l;h++){
            x[h]=/*Conv<T,R>(*/getPix(h,i,j)/*)()*/;
         }
         res.putPix(i,j,f(x));
      }
   }
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::conv(R_Matrix const& a, Z si, Z sj)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).conv(a,si,sj);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::conv(R_Vector const& a, R_Vector const& b\
)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).conv(a,b);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::smooth(Z n)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).smooth(n);
   return res;
}

static R gauss(R r2, R sigma)
   // 2-dimensional radially symmetric Gauss distribution as a function
   // of r*r (in our particular case r*r is easier to calculate than r)
{
   const static R PiInv=R(1)/cpmpi;
   cpmassert(sigma>0.,"gauss(R,R)");
   R factorArg=R(1)/(sigma*sigma*2.);
   R factor=cpmsqrt(factorArg*PiInv);
   return factor*cpmexp(-r2*factorArg);
}

template<class T>
MultiImage<T> MultiImage<T>::conv(const R& sigma, Z dL)
{
   static const R t=1./sqrt(3.);
   // see Abramowitz Stegun p. 892, second formula of 25.4.62
   // for the numerical integration formula to be used here
   // calculation of the convolution kernel as a matrix
   static const R tiny=1e-3;
   if (sigma<tiny) return *this;
      // no convolution required
   Z d=cpmtoz(sigma*2);
   if (d>dL) d=dL;
   Z n=2*d+1;
   R_Matrix a(n,n);
   Z i,j;
   R x, x0=0.5*n, x1;
   R y, y0=x0, y1;
   R r2, w, sum=0.;
   for (i=1;i<=n;i++){
      x=R(i)-x0-0.5;
      for (j=1;j<=n;j++){
         // we calculate the weight of a pixel by numerical
         // integration using a 4-point formula ( Abramowitz Stegun
         // as cited above)
         y=R(j)-y0-0.5;
         x1=x+t;
         y1=y+t;
         r2=x1*x1+y1*y1;
         w=gauss(r2,sigma);

         x1=x+t;
         y1=y-t;
         r2=x1*x1+y1*y1;
         w+=gauss(r2,sigma);

         x1=x-t;
         y1=y+t;
         r2=x1*x1+y1*y1;
         w+=gauss(r2,sigma);

         x1=x-t;
         y1=y-t;
         r2=x1*x1+y1*y1;
         w+=gauss(r2,sigma);
         w*=0.25;
         sum+=w;
         a[i][j]=w;
      }
   }
   sum=R(1)/sum;
   a*=sum;   // normalizing to sum over all pixels = 1

   return conv(a);
 }

template<class T>
MultiImage<T> MultiImage<T>::conv(const V<Mask>& a)const
{
   Word loc("MultiImage<T>::conv(V<Mask>&)");
   Z l=getPla();
   Z d=a.dim();
   cpmassert(d==l,loc);
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).conv(a[h]);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::cut(const Rect& a)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).cut(a);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::cut(IvZ const& iRange, IvZ const& jRange\
)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).cut(iRange,jRange);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::conv(const Mask& a)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).conv(a);
   return res;
}

template<class T>
MultiImage<T> MultiImage<T>::clip(R clipL, R clipU)const
{
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++) res[h]=((*this)[h]).clip(clipL, clipU);
   return res;
}

static Z part(Z m, const R& x)
{
   Z res=cpmtoz(R(m)*x+1);
   return (res>m ? m : res);
}

template<class T>
void MultiImage<T>::insertPattern(
   R X, R Y, R dX,   const F<C,Va<T> >& f, const F<C,B>& g,
   R x, R y, R dx, R dy, B average )
      // as previous function, but pixels are set only if g(z)=trueUM
      // otherwise pixel value is not changed
{
   Z mL=1;
   Word loc("MultiImage<T>::insertPattern(...)");
   CPM_MA
   const Z nTacit=4;
      // controls frequency of messages in verbose mode
   R ar=((*this)[1]).getAspRat();
   cpmassert(ar!=0.,loc);
   R dY=dX/ar;

   R XU=X+dX;
   R YU=Y+dY;

   Z l=getPla();
   Z m=getLin();
   Z n=getCol();
   Z h,i,j,i_;

   R u,v;
   R du=dX/(4*n);
   V<R> iLattice=ImagingTools::pixelCenters(Y,YU,m);
      // matrix index i varies in y direction
   V<R> jLattice=ImagingTools::pixelCenters(X,XU,n);
      // matrix index j varies in x direction

   C z,w;
   Z lineCounter=0;
   R mInv=1./m;
   B valid;

   Z iU=m-part(m,y)+1;
   Z iL=m-part(m,y+dy)+1;
   Z jL=part(n,x);
   Z jU=part(n,x+dx);
   Va<R> sum(l);
   Va<T> fval;
   T val;
   Z count;

   for (i=iL;i<=iU;i++){
      i_=m-i+1;
      v=iLattice[i_];
      if (verbose==1)
      {
         lineCounter++;
         if (lineCounter==nTacit){
            lineCounter=0;
            R fr=R(100*i)*mInv;
            cout<<endl<<"Image "<<fr<<" % completed";
         }
      }

      for (j=jL;j<=jU;j++){
         u=jLattice[j];
         if (average){
            sum.set_(0);
            count=0;
            for (Z eu=-1;eu<=1;eu+=2)for (Z ev=-1;ev<=1;ev+=2){
               z=C(u+R(eu)*du,v+R(ev)*du);
               valid=g(z);
               if (valid){
                  count++;
                  fval=f(z); // transformation C to Va<T,Z>
                  for (h=1; h<=l; h++){
                     sum[h]+=fval[h];
                  }
               }
            }
            if (count>0){
               for (h=1; h<=l; h++){
                  val=sum[h]/count;
                  putPix(h,i,j,val);
               }
            }
         }
         else{
            z=C(u,v);
            valid=g(z);
            if (valid){
               fval=f(z);
               for (h=1; h<=l; h++){
                  putPix(h,i,j,fval[h]);
               }
            }
         }

      }  // j loop
   }   // i loop
   CPM_MZ
}

template<class T>
MultiImage<T> MultiImage<T>::fourierFilter(F<R2,R> const& resp)const
{
   Z mL=1;
   static Word loc("MultiImage<>::fourierFilter(F<R2,R>)");
   CPM_MA
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++){
      res[h]=((*this)[h]).fourierFilter(resp);
      cpmstatus("plane "&cpm(h)&" filtered",2);
   }
   CPM_MZ
   return res;
}

template<class T>
Img24 MultiImage<T>::fourierFilter24(F<R2,R> const& resp)const
{
   Z mL=1;
   static Word loc("MultiImage<>::fourierFilter24(F<R2,R>)");
   CPM_MA
   MultiImage<T> fftImg=fourierFilter(resp);
   fftImg.setRan_(0,255,Color::getGamma());
   Z m=getLin();
   Z n=getCol();
   return fftImg.cut(IvZ(1,m),IvZ(1,n)).toImg24();
}

template<class T>
Img24 MultiImage<T>::fourierFilter24(R x0, R x1)const
{
   return fourierFilter24(R_Func::smoothTransit(R2(x0,x1),R2(1,0)));
}

template<class T>
MultiImage<T> MultiImage<T>::fftFilter(F<C,C> const& resp)const
{
   Z mL=1;
   static Word loc("MultiImage<>::fftFilter(F<C,C>)");
   CPM_MA
   Z l=getPla();
   MultiImage<T> res(l);
   for (Z h=1;h<=l;h++){
      res[h]=((*this)[h]).fftFilter(resp);
      cpmstatus("plane "&cpm(h)&" filtered",2);
   }
   CPM_MZ
   return res;
}

template<class T>
Img24 MultiImage<T>::fftFilter24(F<C,C> const& resp)const
{
   Z mL=1;
   static Word loc("MultiImage<>::fftFilter24(F<C,C>)");
   CPM_MA
   Z m=getLin();
   Z n=getCol();
   MultiImage<T> fftImg=fftFilter(resp);
   fftImg.setRange(0,255,Color::getGamma());
   return fftImg.cut(IvZ(1,m),IvZ(1,n)).toImg24();
}

template<class T>
MultiImage<T> MultiImage<T>::deScreen(R lim, R rim, Z p)const
{
   Z mL=1;
   static Word loc("MultiImage<>::deScreen()");
   CPM_MA
   Z l=getPla(),i,j,k;
   V< VV<C> > sp(l);
   sp[1]=((*this)[1]).fourierTransform();
   Z m=sp[1].dim1(), n=sp[1].dim2();
   Image<R> mono(m,n,0.);
   for (i=1;i<=m;++i) for (j=1;j<=n;++j)
      mono[i][j]=(sp[1][i][j]).abs();
   for (k=2;k<=l;++k){
      sp[k]=((*this)[k]).fourierTransform();
      for (i=1;i<=m;++i) for (j=1;j<=n;++j)
         mono[i][j]+=(sp[k][i][j]).abs();
   }
   Z nMask=2*p+1;
   V<R> w(nMask,1./nMask);
   V<Z> da(nMask,0);
   V<Z> db(nMask);
   Z p_=-p;
   for (i=1;i<=nMask;++i){
      db[i]=p_;
      p_++;
   }
   Mask mas1(da,db,w);
   Mask mas2(db,da,w);
   Image<R> mono1=mono.conv(mas1);
   Image<R> mono2=mono.conv(mas2);
   R gamma=0.25;
   VV<R> fac(m,n,1);
   Z ml=cpmrnd(R(m)*rim);
   Z mu=m-ml;
   Z nl=cpmrnd(R(n)*rim);
   Z nu=n-nl;
   R red=0.33;
   for (i=ml;i<=mu;++i) for (j=nl;j<=nu;++j){
      R v0=mono[i][j];
      R v1=mono1[i][j];
      R v2=mono2[i][j];
      R d1=v0-v1;
      R d2=v0-v2;
      R dr=lim*0.5*(v1+v2);
      if (d1>dr) fac[i][j]*=red;
      if (d2>dr) fac[i][j]*=red;
   }
   for (i=1;i<=m;++i) for (j=1;j<=n;++j){
      R fij=fac[i][j];
      for (k=1;k<=l;++k) sp[k][i][j]*=fac[i][j];
   }
   MultiImage<T> res(l);
   for (k=1;k<=l;++k){
      res[k]=Image<R>(sp[k]);
   }
   CPM_MZ
   return res;
}

template<class T>
Img24 MultiImage<T>::deScreen24(R lim, R rim, Z p)const
{
   Z mL=1;
   static Word loc("MultiImage<>::deScreen24(R,R,Z)");
   CPM_MA
   Z m=getLin();
   Z n=getCol();
   MultiImage<T> deScreenImg=deScreen(lim,rim,p);
   deScreenImg.setRange(0,255,Color::getGamma());
   return deScreenImg.cut(IvZ(1,m),IvZ(1,n)).toImg24();
}

template<class T>
R MultiImage<T>::getPixConv(Z h, Z i, Z j, const R_Matrix& a, Z si, Z sj)
{
   Z ma=a.dim1(),na=a.dim2(), ia, ja;
   Z ci=1+ma/2, cj=1+na/2;
   Z i0=ci+si, j0=cj+sj;
   R val=0.;

   for (ia=1;ia<=ma;ia++){
      for (ja=1;ja<=na;ja++){
            val+=(R)((*this)[h]).getPixCont(i+ia-i0,j+ja-j0)*a[ia][ja];
      }
   }
   return val;
}

template<class T>
R_Vector MultiImage<T>::getPixConvPla( Z i, Z j, const R_Matrix& a)
{
   Z l=getPla();
   Z h_;
   R_Vector res(l);

   for (h_=1;h_<=l;h_++)
   {
      res[h_]=getPix(h_,i,j);
   }
   return a*res;
}

template<class T>
Histogram MultiImage<T>::histogram(Z nBins, R valMin, R valMax )
{
   Z mL=1;
   Word loc("MultiImage<T>::histogram(Z,R,R)");
   CPM_MA

   Z l=getPla();
   Z m=getLin();
   Z n=getCol();

   Z h,i,j;
   R Val;
   Histogram b(l,nBins,valMin,valMax);
   for(i=1;i<=m;i++)
   {
      for(j=1;j<=n;j++)
      {
         for(h=1;h<=l;h++)
         {
            Val=getPix(h,i,j);
            b.put(h,Val);
         }
      }
   }
   CPM_MZ
   return b;
}

template<class T>
VV<T> MultiImage<T>::cut(Z h, const Rect& rect)const
{
   Word loc("MultiImage<T>::cut(Z, Rect&)");
   Z l=getPla();
   cpmassert(h>=1 && h<=l,loc);
   return ((*this)[h]).cut(rect);
}

template<class T>
void MultiImage<T>::insert(MultiImage<T> const& img,
   Rect const& a, Z scale)
{
   Z mL=4;
   Word loc("MultiImage<T>::insert(MultiImage<T>,...)");
   CPM_MA
   const Z method=1;
      // use bilinear interpolation when scaling
//   cpmmessage(mL,loc&" started");
   Z l=getPla();
   cpmassert(l==img.getPla(),loc);
   Z m=getLin();
   Z n=getCol();
   Z mi=img.getLin();
   Z ni=img.getCol();
   R xL=a.getxL();
   R xU=a.getxU();
   R yL=a.getyL();
   R yU=a.getyU();
   Roi ra(m,n,xL,xU,yL,yU);
   Z ma=ra.m();
   Z na=ra.n();
   Z h,i,j;
   if (scale==0){
      Z m1=CpmRootX::inf(mi,ma);
      Z n1=CpmRootX::inf(ni,na);
      for (h=1;h<=l;h++){
         for (i=1;i<=m1;i++){
            for (j=1;j<=n1;j++){
               putPix(h,ra.i(i),ra.j(j),img.getPix(h,i,j));
            }
         }
      }
   }
   else{
      R ara=R(na)/ma;
      R ari=R(ni)/mi;
      cpmassert(ara>0,loc);
      cpmassert(ari>0,loc);
      Z niNew, miNew;
      if (ara<=ari){   // then img is 'more landscape' than a
         // and thus ni has to be adjusted to na
         niNew=cpmtoz(na);
         miNew=cpmtoz(R(niNew)/ari);
      }
      else{   // then img is 'more portrait' than a
         // and thus mi has to be adjusted to ma
         miNew=cpmtoz(ma);
         niNew=cpmtoz(R(miNew)*ari);
      }
      MultiImage<T> imgNew=img.interpolate(miNew,niNew,method);
      Z m2=CpmRootX::inf(miNew,ma);
      Z n2=CpmRootX::inf(niNew,na);
      for (h=1;h<=l;h++){
         for (i=1;i<=m2;i++){
            for (j=1;j<=n2;j++){
               putPix(h,ra.i(i),ra.j(j),imgNew.getPix(h,i,j));
            }
         }
      }
   }
   CPM_MZ
}

template<class T>
T MultiImage<T>::getPix(Z i, Z j )const
{
   Z l=getPla();
   R res=0.;
   for (Z h=1;h<=l;h++) res+=getPix(h,i,j);
   res/=l;
   return r2t<T>(res);
}

template<class T>
T MultiImage<T>::operator()(R x, R y, Z method)const
{
   Z l=getPla();
   R res=0.;
   for (Z h=1;h<=l;h++) res+=operator()(h,x,y,method);
   res/=l;
   return r2t<T>(res);
}

} // namespace

#endif
