//? cpmperm.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_PERM_H_
#define CPM_PERM_H_
/*
Class for permutations of the numbers
{1,2,...,n} for n arbitrary and set by a constructor of the class.
*/

#include <cpmvo.h>

namespace CpmArrays{

class R_Vector;

class Perm{ // group of permutations

   Vo<Z> rep;

   Perm(const Vo<Z>& rep_, bool b):rep(rep_){b;}
      // efficient but completely unsave, since permutation
      // nature of argument not checked.
      // The second argument avoids interference with the
      // public Perm constructor from lists
public:
   typedef Perm Type;
   CPM_INVERSION
   CPM_PRODUCT_C
   CPM_IO
   CPM_ORDER
   CPM_NAM(Type)

   template <class T>
   explicit Perm(const Vo<T>& v):
      rep(v.permutationForIncreasingOrder(),true){}
      // 'king's road to permutations'. Let the argument be any array of
      // orderable objects. Then the permutation which brings them in
      // increasing order will be created.

   explicit Perm(R_Vector const& v);
    //  :rep(v.permutationForIncreasingOrder()){}
      // special case of the above

   explicit Perm(Z n=1);
      // creates the identity permutation of the set {1,2,...,n}

   Perm( Z n, Z i, Z j);
      // Constructor for a transposition i <---> j in the set {1,2,...,n}

   Perm( Z n, Z i);
      // Constructor for a cyclic permutation k-->k+i mod(n)+1
      // the set {1,2,...,n}

   Z par()const;
      // parity

   V<Z> cycDec()const;
      // cycle decomposition

   // The resulting array iv characterizes the decomposition of v
   // into a product of cycles (it does not output the factors
   // themselves).
   // iv.dim() is the number of cycles and iv[i] is the length of the
   // i-th cycle. The cycles are orderd as they arrise when always the
   // smallest possible starting value is chosen for a new cycle.

   Z operator[](Z i)const{ return rep[i];}
      // no reference returned, since changing the components is stricly
      // forbidden. This would in most cases destroy the property of
      // being a permutation!

   Z dim()const{ return rep.dim();}
      // if the return value is n, the set {1,2,...,n} is the set
      // for which *this is a permutation (it may not all elements
      // of this set really change; actually the indentity permutation
      // changes none of them).

   V<Z> getRep()const{ return rep;}

   Perm ran(Z j=0)const;
      // The return value is a Perm res such that res.dim()==dim()=:n.
      // It is 'random' since it is obtained by ordering a
      // random-generated list of n real numbers

   static V<Perm> all(Z n);
      // returns all permutations of {1,2,...,n} in lexicographic order
      // Implemened with the function next_permutation of the standard
      // library of C++. Added 2016-05-21
};

} // namespace

#endif
