//? cpmrvector.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_RVECTOR_H_
#define CPM_RVECTOR_H_

/*
   Purpose: see class R_Vector
*/
#include <cpmvr.h>
#include <cpmperm.h>
#include <cpmlinalg.h>
#include <cpminterval.h>

#include <vector>
#include <valarray>

namespace CpmArrays{

   using std::vector;
   using std::valarray;
   using CpmRoot::Z;
   using CpmLinAlg::Z2;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmArrays::Perm;
   using CpmLinAlg::R2;
   using CpmLinAlg::R3;
   using CpmLinAlg::R4;
   using CpmGeo::Iv;

   typedef Vr<Z> I_Vector;

   class R_Vector{
// vectors with R-valued components, rich interface including arithmetic
      // arbitrary length, dynamic allocation

   Vr<R> rep;
      // we choose composition instead of inheritance, so the
      // interface of Vr<R> needs to be known only to the implementer
      // of this class, the user finds the interface of R_Vector
      // in the present file together with the macro definitions
      // in cpminterfaces.h

   R_Vector(const Vr<R>& x):rep(x){}

   static R addFunc(const R& r1, const R& r2, const R& t)
      { return r1+r2*t;}

public:
   typedef R_Vector Type;
   typedef R ScalarType;
   CPM_IO
   bool scanFromFreeFor(istream&);
      // FreeFor stands for free format
      // reads numbers from a stream as long as this can successfully
      // be done and changes *this into the R_Vector having all these
      // numbers as components. So, one doesn't need to know the
      // number of these numbers in advance
   CPM_ORDER
   CPM_TEST
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_DOT_PRODUCT
   CPM_SUM_M
   CPM_PRODUCT_M
   CPM_SCALAR_M
   CPM_DIFFERENCE
   CPM_DIVISION
   Z dim()const{ return rep.dim();}
   Z b()const{ return rep.b();}
   Z e()const{ return rep.e();}
   R_Vector& add(const R_Vector& v, R t)
      // adds v*t to *this and returns *this.
   { rep.fg_(addFunc,v.rep,t); return *this;}

   static R etaInv;
// constructors
   explicit R_Vector(Z n=0):rep(n,0.){}
   explicit R_Vector(const V<R>& v):rep(v){}
   explicit R_Vector(vector<R> const& v):rep(v.size()){
      for (N i=0,j=1;i<v.size();++i,++j) rep[j]=v[i];
   }
   explicit R_Vector(valarray<R> const& v):rep(v.size()){
      for (N i=0,j=1;i<v.size();++i,++j) rep[j]=v[i];
   }
   explicit R_Vector(const R2& v):rep(V<R>{v[1],v[2]}){}
   explicit R_Vector(C z):rep(V<R>{z[1],z[2]}){}
   explicit R_Vector(const R3& v):rep(V<R>{v[1],v[2],v[3]}){}
   explicit R_Vector(const R4& v):rep(V<R>{v[1],v[2],v[3],v[4]}){}
   explicit R_Vector(IvZ ivz):rep(ivz){}
   R_Vector(IvZ ivz, R const& x):rep(ivz,x){}

   R_Vector(std::initializer_list<R> il):rep(V<R>(il)){}
      // requires C++11

   R_Vector(Z n, R x):rep(n,x){}
      // constructor for a vector with n comonents equal to x

   R_Vector(R xStart, R step, R xEnd,
    bool multiplicative = false, bool precautious = true);
// Chain constructor:
// Let here '&' denote '+' or '*' according to:
//           multiplicative ? * : +
// Creates a vector which holds a regular chain of real values
// starting with xStart, xStart & step, xStart & step & step,
// ...
// This sequence will finally grow out of the interval Iv(xStart, xEnd).
// The last value included as a component is the one which is contained in
// a slightly enlarged version of the interval Iv(xStart, xEnd).
// We do not assume that xStart < xEnde.
// For the trivial case xStart==xEnd we return R_Vector(1,xStart).
// If step does not match the order of xStart and xEnd we take 1./step
// in the multiplicative case and -step in the additive one.
// If the last argument (i.e. precautious) is set to true the termination
// condition as formulated above is tested via cpmassert.

   R_Vector(R xStart, R step, Z n);
// Second form of a chain constructor.
// Constructing an R_Vector x such that x.dim()==n, x[1]=xStart,
// v[i+1]=v[i]+step, (i=1, ... n-1, if n>1). Assumed is n>=0. For n=0 we
// construct R_Vector().

   R_Vector(Z n, Iv iv, Z j=0);
      // Random constructor:
      // Creates a vector of index range 1,...,n and components
      // which range randomly over the interval iv. The parameter
      // controls the seed of the random generator.


   R_Vector(Iv iv, Z n);
      // n equidistant components filling the interval iv (boundaries
      // included, (*this)[1]=iv.inf(), (*this)[n]=iv.sup() )
      // For n=1 (*this)[1]=(iv.inf()+iv.sup())/2
      // n=0 calls cpmassert
      // The result is always an ascending list.


   R_Vector(R2 p, Z n);
      // List of n equidistant components starting with p[1] and ending
      // exactly with p[2]
      // For n=1 (*this)[1]=(p[1]+p[2])/2
      // n=0 calls cpmassert
      // The result is always monotone, be it ascending or descending. A very
      // useful constructor.

  // end of constructors
   operator V<R>()const{ return rep;}
      // the opposite conversion is not automatic !

   IvZ dom()const{ return rep.dom();}

   vector<R> std()const{ return rep.std();}
      // returns a std::vector with the same components
      // Let res be the return value, then clearly
      // res[0]=rep[1]

   const R& cyc(Z i)const{ return rep.cyc(i);}
      //: cyclic
   R& cyc(Z i){ return rep.cyc(i);}
      //: cyclic

   R& operator[](Z i){ return rep[i];}
   const R& operator[](Z i)const{ return rep[i];}

   R& cui(Z i){ return rep.cui(i);}
      //. component (with) unchecked index
   const R& cui(Z i)const{ return rep.cui(i);}
      //. component (with) unchecked index

   R& fir(){ return rep.fir();}
      //: first

   R const& fir()const{ return rep.fir(); }
      //: first

   R& last(){ return rep.last(); }
      //: last

   R const& last()const{ return rep.last(); }
      //: last

//   R const& read(Z i, Outside mode)const{ return rep.read(i,mode);}
//      // reading components in an efficient (as &'s), safe,
//      // and flexible manner
//      // mode 0: values outside assumed to be given as T()
//      // mode 1: values outside assumed to be given by cyclic
//      // continuation
//      // else: values outside assumed to be given by constant
//      // continuation

   // improved version of the next (commented) function, ZERO reintroduced
   R  read(Z i, Outside mode)const{
      if (dom().hasElm(i)) return rep[i];
      else if (mode==ZERO) return R(0.);
      else return rep.read(i,mode);
   }

// same as previous function, uniformity with R_Matrix
   R  operator()(Z i, Outside mode)const{
      if (dom().hasElm(i)) return rep[i];
      else if (mode==ZERO) return R(0.);
      else return rep.read(i,mode);
   }

   R const& r(Z i)const{ return rep.r(i);}
      //. read


   R in_(R const& t, bool reverse=false){ return rep.in_(t,reverse);}
      // This operations treats *this as a shift register:
      // All components get shifted by one position 'to the right' and
      // the total length (dim) of the vector remains the same. So what
      // was the last component prior to the operation has to be removed
      // from the vector, in order to not wasting information, this
      // removed component will shine up as the return value of the
      // function. After the operation, the first component of the vector
      // will be t.
      // If reverse==true, t gets inputted at the end, and all shift
      // operations go 'to the left'.

   R_Vector eliFirst(Z nEli=1)const{ return R_Vector(rep.eliFirst(nEli));}

   R_Vector eliLast()const{ return R_Vector(rep.eliLast());}

   R_Vector eli(Z i, Z nEli=1)const{ return R_Vector(rep.eli(i,nEli));}

   R_Vector eli(IvZ const& iv)const{ return R_Vector(rep.eli(iv));}
      //:  eliminate
      // Eliminate the components the indexes of which belong to iv.

   R_Vector cut(Z i)const{ return R_Vector(rep.cut(i));}
      //: cut
      // returned is a R_Vector which results from *this by removing i
      // components from the end

   V<Z> permutationForIncreasingOrder()const
      {return rep.permutationForIncreasingOrder();}

   R_Vector permute(const V<Z>& p)
      { return R_Vector(rep.permute(p));}

   R_Vector operator*(const Perm& p)const
      { return R_Vector(rep.permute(p.getRep()));}

   R_Vector increasingCopy()const
      { return R_Vector(rep.increasingCopy());}

   R_Vector decreasingCopy()const
      { return R_Vector(rep.decreasingCopy());}

   R_Vector select(const V<B>& sel)const
      { return R_Vector(rep.select(sel));}

   void order_(){ rep.order_();}

   void sort_(){ rep.order_();}

   void set_(R const& t){ rep.set_(t);}
      // changes all components to the value t

   R_Vector& b_(Z i){ rep.b_(i); return *this;}
      // setting the beginning of the valid index range

   R_Vector& e_(Z i){ rep.e_(i); return *this;}
      // setting the end of the valid index range

   bool isVal()const;
      // returns false iff for any of the components
      // x[i] we have CpmRoot::isVal(x[i])==false

   bool valInd(Z i)const{ return rep.valInd(i);}
      //: valid index
      // returns the validity of i as an index of *this

   R inf(Z i=0)const;
      // returns the 'ith-smallest' component. No ordering
      // for i=0. For i>0 it is the component indexed 1+i
      // of the increasingly ordered version of rep

   R sup(Z i=0)const;
      // returns the 'ith-largest' largest component.
      // For i=0 fast operation, no ordering
      // For i>0 it is the component indexed rep.dim()-i
      // of the increasingly ordered version of rep

   R supAbs()const;
      // returns sup{ |rep[i]|}

   Z indSup()const{ return rep.indSup();}
      // returns the smallest i for which (*this)[i]=(*this).sup()

   Z indInf()const{ return rep.indInf();}
      // returns the smallest i for which (*this)[i]=(*this).inf()

   R sum()const{ return rep.sum();}
      // returns the sum of all components

   R mean()const{Z n=rep.dim(); return n==0 ? R(0) : rep.sum()/n;}
      // arithmetic mean value of all components, fast

   R med()const // median value of all components
      { return infMedSup()[2];}

   Iv infSup()const
   { T2<R> prr=rep.infSup(); return Iv(prr.c1(), prr.c2());}
   // returns the smallest and the largest component. Fast operation,
   // no ordering

   R3 infMeanSup()const;
   // returns the smallest component, the arithmetic mean value,
   // and the largest component of *this. Fast operation, no ordering

   R3 infMedSup()const;
   // returns the smallest component, the median value,
   // and the largest component of *this. Achieved by ordering the
   // components. If one should also need the mean value, don't call
   // the previous function in addition, but simply the cheap function
   // mean().

   R_Vector pseudoInverse(R acc)const;
// Gives the pseudoinverse of the vector *this.
// This means that all components get inverted except those that are
// close to zero in comparison to the component with the largest
// absolute value.
// For small components c ( |c| ~ acc*sup |(*this)[i]| ) a continuous
// transition between the inverse and zero becomes operational.

    R_Vector scaleTo(const R& r)const;
// returns a vector of absolute value |r| and the direction of
// *this for positive r and antidirection for negative r.
// same function is available for complex n

   R_Vector resize(Z newDim)const
   { return R_Vector(rep.resize(newDim));}

   R_Vector subMean()const{ R s=-mean(); return R_Vector(rep+s);}
      //: subtract mean

   R subMean_(){ R m=mean(); rep-=m; return m;}
      //: changes *this by subtractng the mean from all components
      // and returns the mean

   R_Vector posCopy()const;
      // positive copy
      // returns a R_Vector obtained from *this by replacing all
      // components by their absolute Value

   R_Vector cutToPosCopy()const;
      // cut to positive copy
      // returns a R_Vector obtained from *this by replacing all
      // negative components by 0.


   R_Vector log10()const;
      // log10
      // returns a R_Vector obtained from *this by replacing all
      // components
      // by the log10 of their absolute value. The safe version of log10
      // which is a member function of R1 is employed for this.

   R_Vector operator()(F<R,R> const& f)const{ return R_Vector(rep(f));}

   R_Vector operator()(std::function<R(R)> f)const
      { return R_Vector(rep(f));}


   R_Vector cutCopy(Iv iv)const;
      // Returns a vector with selected components, where the
      // selection is determined by iv.
      // Let ivr the section of iv and [0,1]. The interval [0,1] is
      // mapped in a natural way to the full index set {1,2,...,dim()}
      // and ivr then determines a contiguous subset of the full index
      // set. Returned is a vector which has just the components the
      // original indexes of which belong to this subset. In effect
      // the returned vector differs from *this by deleting some
      // (0 included) components at the beginning and some at the end.

   R3 variation()const;
      // returns (a,b,c) such that a=Min{ x[i+1]-x[i] : i=1,...d-1}
      // c=Max{ x[i+1]-x[i] : i=1,...d-1}
      // b=Mean{ |x[i+1]-x[i]| : i=1,...d-1}.
      // Thus a<=c, b<=Max(|a|,|c|). If *this is monotone ascending,
      // a is never negative. If *this consists of empirical data
      // which in the absence of errors and uncertainties would
      // be monotone ascending, a slightly negative a may be OK
      // if |a|<<b. Implemented as a single pass operation
      // without ordering or copying.
      // Notice that the meanvalue of x[i+1]-x[i] (thus without the abs)
      // is simply (x[n]-x[1])/(n-1) and thus has not to be determined by
      // a loop over all components.
      // Notice also that b*(d-1) is the sum of all |x[i+1]-x[i]|
      // which comes close to a polygon length if the spacing in
      // i-direction is neglegiable compared to the swing in x-direction.

   R_Vector der(R h=1, bool periodic=false) const;
// The vector a.der() results from the vector a by numerical
// differentiation of a function, the equi-distantly sampled version
// (sampling step = h) of which the vector a is considered to be.
// Formulas: Bronstein Semendiajew, Taschenbuch der Mathematik S. 768
// Tabelle 7.13 for generic points and Abramowitz Stegun p. 914
// Table 25.2 for the boundary points.
// If the second argument is true the function is periodic so that
// for i outside the allowed range we use this->cyc(i) as the component

   R_Vector rot(Z i,Outside mode=CYCLIC) const
//: rotate
// Name as the list transformation function Rotate of Mathematica.
// v.rot(s)[i] == v(i-s,mode)
// This means that we shift the component i of the original vector
// into the new position i+s. See V<>::rot(Z,Outside) in cpmv.h.
   { return R_Vector(rep.rot(i,mode));}

   R_Vector disLap(Outside mode=CYCLIC) const;
//: discrete Laplacean
// Here mode controls how to interprete indexes out of range. No reference
// to the lattice spacing is made.

   R_Vector& operator &=(const R_Vector& x1)
   { rep&=x1.rep; return *this;}
   // mutating form of concatenation

   R_Vector& operator &=(R r)
   { rep&=r; return *this;}

   R_Vector operator & (const R_Vector& x1)const
   { return R_Vector(rep&x1.rep);}
   // concatenation

   void operator<<(R const& r){ rep<<r;}
      // changes *this by appending a component, Ruby style

   void operator<<(R_Vector const& x1){ rep<<x1.rep;}
      // changes *this by appending a R_Vector, Ruby style

   R_Vector operator & (R const& r)const
   { return R_Vector(rep&r);}

   R_Vector prepend(R const& r)const
   { return R_Vector(rep.prepend(r));}

   void printLineOn(ostream& out , const Word& separator="  " )const
      // prints the components of *this, separated by the character
      // string separator. In default mode all numbers are in a line
      // which is usefull for creating data that later have to be given
      // to a plot program such as gnuplot.
   { for (Z i=1;i<=rep.dim();i++) out<<separator<<rep[i];}

   R_Vector crsPrd(const R_Vector& x)const;
      // cross product (vector product) defined for
      //    dim()==2 && x.dim()>=2
      // and
      //    dim()==3 && x.dim()>=3
      // else error, since then presumably something went severely wrong

   Z indOfNerCom(R const& r)const;
      //: index of nearest component
      // Assumes that *this is ordered ascendingly.
      // Returns i such that (*this)[i] is nearer to r
      // than any other component (*this)[j].

   Z2 indCon(Iv const& iv )const{
      Z i1=indOfNerCom(iv.inf());
      Z i2=indOfNerCom(iv.sup());
      return Z2(i1,i2);
   }
      //: indexes of confinement
      // Assumes that *this is ordered ascendingly.
      // Returns a pair (i,j),such that (*this)[i] is nearer to iv.inf()
      // than any other component (*this)[i'] and that (*this)[j] is nearer to iv.sup()
      // than any other component (*this)[j'].

   R_Vector confine(Z2 const& z2)const
      // Assumes z2[1] <= z2[2].
      // The components rep[z2[1]],...rep[z2[2]] are the components
      // of the resulting R_Vector (the first valid index of which is 1).
   {
      Z mL=4;
      Word loc("R_Vector confine(Z2 const& z2)const");
      CPM_MA
      Z n=z2[2]-z2[1]+1;
      R_Vector res(n);
      Z z1=z2[1];
      for (Z i=1,j=z1;i<=n;++i,++j){
         res[i]=rep[j];
      }
      CPM_MZ
      return res;
   }

};

// inline reduction of most functions to those of rep in Vr<R>
// in order not to loose speed

inline bool R_Vector::prnOn(ostream& out)const
{ return rep.prnOn(out);}

inline bool R_Vector::scanFrom(istream& in)
{ return rep.scanFrom(in);}

inline Z R_Vector::com(const R_Vector& x)const
{ return rep.com(x.rep);}

inline R_Vector& R_Vector::operator +=(const R_Vector& x)
{ rep+=x.rep; return *this;}

inline R_Vector& R_Vector::operator *=(const R& r)
{ rep*=r; return *this;} // exp

inline R_Vector& R_Vector::operator *=(const R_Vector& x)
{ rep*=x.rep; return *this;}

inline R_Vector R_Vector::neg()const
{ return R_Vector(-rep);}

inline R_Vector R_Vector::ran(Z j)const
{ return R_Vector(rep.ran(j));}

inline Z R_Vector::hash()const
{ return rep.hash();}

inline R_Vector R_Vector::test(Z j)const
{ return R_Vector(rep.test(j));}

inline Word R_Vector::nameOf()const
{ return Word("R_Vector");}

inline Word R_Vector::toWord()const
{ return rep.toWord();}

inline R_Vector R_Vector::net(Z i)const
{ return R_Vector(rep.net(i));}

inline R_Vector R_Vector::inv()const
{ return pseudoInverse(etaInv);}


inline R_Vector R_Vector::con()const
{ return *this;}

inline R R_Vector::operator |(const R_Vector& x )const
{ return rep|x.rep;}

} // namespace

#endif
