//? cpmtensor.h
//? C+- by Ulrich Mutze. Status of work 2022-12-08.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.

#ifndef CPM_TENSOR_H
#define CPM_TENSOR_H

#include <cpmc.h>
#include <cpmfa.h>
#include <cpmm.h>
#include <cpmgraph.h>
#include <cpmimagingtools.h>
#include <cpmlinalg.h>
#include <cpmcvector.h>
#include <cpmapplication.h>
#include <cpmalgorithms.h>
#include <cpmrecordhandler.h>
#include <cpmfile.h>
#include <cpmrfunc.h>
#include <functional>
#include <set>
#include <algorithm>
#include <valarray>


using namespace CpmRoot;
using namespace std;

namespace CpmArrays{

using VN = valarrayr<N>;

// Here a Tensor is an multi-indexed array of objects of T. The number of indexes
// is np_ and each of the indexes has valid values from 0 to m-1_. It has thus
// m_^np_ components. In a sense, this is a narrower concept than that of a matrix:
// In a non-square matrix the first and the second index don't have the same range.
// There are, however many important cases for which the present concept of tensor
// is the right one. Most important for us: discretized n-particle wave functions.
template<typename T>
class Tensor{
   N m_; // range of indexes
   N np_; // number of indexes
   N sz_; // number of components
   VN ind_; // each valid multiindex is an instance of VN
   valarray<T> rep_;
public:
   Tensor(N m, N np, T const& t): m_{m}, np_{np},sz_(pow(m_,np_)),
      ind_{np_}, rep_(t,sz_)
   {
      N i=1,k;
      VN prel(1,np_);
      while (i<np_){
         prel[i]=prel[i-1]*m_; i++;
      }
      for (i=0,k=np_-1;i<np_;++i,--k) ind_[i]=prel[k];
   }

   T const& operator()(VN ind)const{
      N s=0;
      for (N i=0;i<np_;++i) s+=ind[i]*ind_[i];
      return rep_[s];
   }

   T& operator()(VN ind){
      N s=0;
      for (N i=0;i<np_;++i) s+=ind[i]*ind_[i];
      return rep_[s];
   }
};

template<typename T>
class Tensor1{
   N sz_;
   valarray<T> rep_;
public:
   Tensor1(N n, T const& t): sz_{n}, rep_(t,sz_){}
   Tensor1(N n=0): sz_{n}, rep_(T(),sz_){}

   T const& operator()(N i)const{
       return rep_[i];
   }

   T& operator()(N i){
       return rep_[i];
   }

   T cyc(Z i)const{ // We will encounter together with i also i+1 and i-1
      if (i >= sz_) return rep_[i-sz_];
      if (i < 0) return rep_[i+sz_];
      return rep_[i];
   }

   T def(Z i)const{ // We will encounter together with i also i+1 and i-1
      if (i >= sz_) return T();
      if (i < 0) return T();
      return rep_[i];
   }

   N sz()const{ return sz_;}
};

// instances of Tensor2 are m_ x m_ (square) matrices
template<typename T>
class Tensor2{
   N m_; // range of indexes
   N np_{2}; // number of indexes
   N sz_; // number of components
   VN ind_; // each valid multiindex is an instance of VN
   valarray<T> rep_;
public:
   Tensor2(N m=0): m_{m}, np_{2},sz_(pow(m_,2)),ind_{2},rep_(T(),sz_){}

   T const& operator()(N i, N j)const{
       return rep_[VN{i,j}];
   }

   T& operator()(N i, N j){
       return rep_[VN{i,j}];
   }

   T cyc(Z i, Z j)const{ // We will encounter together with i also i+1 and i-1
      Z ir=i,jr=j;
      if (i >= sz_) ir-=sz_;
      if (i < 0) ir+=sz_;
      if (j >= sz_) jr-=sz_;
      if (j < 0) jr+=sz_;
      return rep_[VN{ir,jr}];
   }

   T def(Z i, Z j)const{ // We will encounter together with i also i+1 and i-1
      if (i >= sz_) return T();
      if (i < 0) return T();
      if (j >= sz_) return T();
      if (j < 0) return T();
      return rep_[VN{i,j}];
   }

   N sz()const{ return sz_;}
};

}

#endif
