//? cpmcvector.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#include <cpmcvector.h>
#include <cpmrfunc.h>
#include <cpms.h>

using namespace CpmRoot;
using namespace CpmSystem;
using namespace CpmArrays;

R C_Vector::etaInv=1e-12;

V<C_Vector> C_Vector::canBas(IvZ const& ivz) // canonical basis
{
   C z0=C(R(0.));
   C z1=C(R(1.));
   V<C_Vector> res(ivz,C_Vector(ivz,z0));
   for (Z i=res.b(); i<=res.e(); ++i){
      res.cui(i).cui(i) = z1;
   }
   return res;
}

void C_Vector::mark(CpmGraphics::Graph & g)const
{
   Z n=dim();
   R_Vector x(n);
   for (Z i=b(),i0=1; i<=e(); ++i,++i0) x[i0]=i;
   R_Matrix mat(3);
   mat[1]=real();
   mat[2]=imag();
   mat[3]=absC();
   using namespace CpmImaging;
   V<Color> vc{GREEN,RED,BLACK};
   g.mark(x,mat,vc);
}


bool C_Vector::isVal()const
{
   Z n=dim();
   for (Z i=1;i<=n;i++){
      if (!CpmRoot::isVal(rep[i])) return false;
   }
   return true;
}

#if defined(CPM_USE_EIGEN)
using namespace Eigen;
using namespace std;
C_Vector::C_Vector(
      Matrix<complex<R>,Dynamic,1> const& vec)
   {
      Z m=vec.size();
      V<C> res(m);
      for (Z i=1;i<=m;++i){
            res[i] = C(vec(i));
      }
      *this=C_Vector(res);
   }

   Matrix<complex<R>,Dynamic,1> C_Vector::toEigen()const
   {
      Z m=dim(),i;
      Matrix<complex<R>,Dynamic,1> res(m);
      for (i=1;i<=m;++i)
         res(i)=rep.cui(i).std();
      return res;
   }

#endif

C_Vector::C_Vector(R_Vector const& re, R_Vector const& im)
{
   Z d=re.dim(), d2=im.dim();
   Word loc("C_Vector(R_Vector, R_Vector)");
   cpmassert(d==d2,loc);
   rep = Vr<C>(d);
   for (Z i=1; i<=d;++i) rep[i] = C(re[i],im[i]);
}

 T2<R_Vector> C_Vector::prv()const
{
   Z d = dim();
   R_Vector re(d);
   R_Vector im(d);
   for (Z i=1;i<=d;++i){
      C ci = rep[i];
      re[i] = ci.real();
      im[i] = ci.imag();
   }
   return T2<R_Vector>(re,im);
}

 R_Vector C_Vector::real()const
{
   IvZ iv = dom();
   R_Vector re(iv);
   for (Z i=b();i<=e();++i){
      re[i] = rep[i].real();
   }
   return re;
}

R_Vector C_Vector::imag()const
{
   IvZ iv = dom();
   R_Vector im(iv);
   for (Z i=b();i<=e();++i){
      im[i] = rep[i].imag();
   }
   return im;
}

R_Vector C_Vector::absC()const
{
   IvZ iv = dom();
   R_Vector abc(iv);
   for (Z i=b();i<=e();++i){
      abc[i] = rep[i].abs();
   }
   return abc;
}

C_Vector::C_Vector(Iv2 iv, Z n):
   C_Vector(R_Vector(iv.c1(),n),R_Vector(iv.c2(),n)){}

C_Vector::C_Vector(Z n, Iv2 iv, Z j):
   C_Vector(R_Vector(n,iv.c1(),j),R_Vector(n,iv.c2(),j)){}

C_Vector::C_Vector(C2 p, Z n)
{
   C z1 = p.x1;
   C z2 = p.x2;
   R2 pRe = R2(z1.real(),z2.real());
   R2 pIm = R2(z1.imag(),z2.imag());
   R_Vector vRe(pRe,n);
   R_Vector vIm(pIm,n);
   *this = C_Vector(vRe,vIm);
}

C3 C_Vector::infMeanSup()const
{
   T2<R_Vector> vri = prv();
   R3 r1 = vri.c1().infMeanSup();
   R3 r2 = vri.c2().infMeanSup();
   return C3(r1,r2);
}

C3 C_Vector::variation()const
{
   T2<R_Vector> vri = prv();
   R3 r1 = vri.c1().variation();
   R3 r2 = vri.c2().variation();
   return C3(r1,r2);
}

C_Vector C_Vector::pseudoInverse(R acc) const
{
   C_Vector d=*this;
   Z i, n=dim();
   acc=cpmabs(acc);
   if (acc==0.){ // most common case, thus first and without ordering
   // overhead
      for (i=1;i<=n;i++) d[i]=cpminv((*this)[i]);
      return d;
   }
   R_Vector absthis(n);
   for (i=1;i<=n;i++){
      R ri=(*this)[i].abs();
      absthis[i]=ri;
   }
   R value, maxValue=-1;
   for (i=1;i<=n;i++){ // finding the maximum of the absolute values of
   // the
         // components as fast as possible, thus without ordering
      value=absthis[i];
      if (value>maxValue) maxValue=value;
   }
   R eta=maxValue*acc;
   eta*=0.5;
   eta*=eta;
   R nom;
   for (i=1;i<=n;i++){
      C xx=(*this)[i];
      nom=xx.abs2()+eta;
      cpmassert(nom>0,"C_Vector::pseudoInverse(R acc)");
      d[i]=xx/nom;
   }
   return d;
}

//C_Vector C_Vector::pseudoInverse(R acc) const
//{
//   C_Vector d=*this;
//   Z i, n=dim();
//   acc=cpmabs(acc);
//   if (acc==0.){ // most common case, thus first and without ordering
//   // overhead
//      for (i=1;i<=n;i++) d[i]=cpminv((*this)[i]);
//      return d;
//   }
//   R_Vector absthis(n);
//   for (i=1;i<=n;i++){
//      R ri=(*this)[i].abs();
//      absthis[i]=ri;
//   }
//   R value, maxValue=-1;
//   for (i=1;i<=n;i++){ // finding the maximum of the absolute values of
//   // the
//         // components as fast as possible, thus without ordering
//      value=absthis[i];
//      if (value>maxValue) maxValue=value;
//   }
//   R eta=maxValue*acc;
//   eta*=0.5;
//   eta*=eta;
//   R nom;
//   for (i=1;i<=n;i++){
//      C xx=(*this)[i];
//      nom=xx.abs2()+eta;
//      cpmassert(nom>0,"C_Vector::pseudoInverse(R acc)");
//      d[i]=xx/nom;
//   }
//   return d;
//}

C_Vector C_Vector::operator*(R const& t)const
{
   C_Vector res(*this);
   for (Z i=b();i<=e();++i) res[i]=rep[i]*t;
   return res;
}

C_Vector& C_Vector::operator*=(R const& t)
{
   for (Z i=b();i<=e();++i) rep[i]*=t;
   return *this;
}

C_Vector C_Vector::scaleTo(const R& r)const
{
   R rOrig=abs();
   C_Vector res=*this;
   if (rOrig!=0.){
      R fac=(r/rOrig);
      res*=C(fac);
   }
   return res;
}

C_Vector C_Vector::posCopy()const
{
   T2<R_Vector> vri = prv();
   R_Vector r1 = vri.c1().posCopy();
   R_Vector r2 = vri.c2().posCopy();
   return C_Vector(r1,r2);
}

C_Vector C_Vector::log10()const
{
   Z d=dim();
   C_Vector res(d);
   for (Z i=1;i<=d;i++) res[i]=CpmRoot::log10(rep[i]);
   return res;
}

C_Vector C_Vector::cutCopy(Iv iv)const
{
   R x1=iv[1];
   R x2=iv[2];
   if (x1<0) x1=0;
   if (x2>1) x2=1;
   Z i,n=dim();
   Z span=n-1;
   Z i1=cpmrnd(x1*span+1);
   Z i2=cpmrnd(x2*span+1);
   Z nRes=1+i2-i1;
   Z i1_=i1-1;
   C_Vector res(nRes);
   for (i=1;i<=nRes;++i){
      res.rep[i]=rep[i1_+i];
   }
   return res;
}

C3 C_Vector::infMedSup()const
{
   T2<R_Vector> vri = prv();
   R3 r1 = vri.c1().infMedSup();
   R3 r2 = vri.c2().infMedSup();
   return C3(r1,r2);
}

C C_Vector::sup(Z i)const
{
   if (i==0) return rep.sup();
   static Word loc("C_Vector::sup(Z i)");
   cpmassert(i>0,loc);
   Z n=dim();
   cpmassert(n>=1,loc);
   if (n==1) return rep[1];
   else{
      Z iRes=n-i;
      cpmassert(iRes>=1,loc);
      V<Z> indNew=rep.permutationForIncreasingOrder();
      return rep[indNew[iRes]];
   }
}

C C_Vector::inf(Z i)const
{
   if (i==0) return rep.inf();
   static Word loc("C_Vector::inf(Z i)");
   cpmassert(i>0,loc);
   Z n=dim();
   cpmassert(n>=1,loc);
   if (n==1) return rep[1];
   else{
      Z iRes=1+i;
      cpmassert(iRes<=n,loc);
      V<Z> indNew=rep.permutationForIncreasingOrder();
      return rep[indNew[iRes]];
   }
}

R C_Vector::supAbs()const
{
   R res=0;
   for (Z i=1;i<=dim();++i){
      R xi=rep[i].abs();
      if (xi>res) res=xi;
   }
   return res;
}

C_Vector C_Vector::der(R h, bool periodic) const
// The vector a.der() results from the vector a by numerical
// differentiation of a function, the equi-distantly sampled image
// (sampling step = h) of which the vector a is considered to be.
// Formulas: Bronstein Semendiajew, Taschenbuch der Mathematik S. 768
// Tabelle 7.13 for generic points and Abramowitz Stegun p. 914 Table
// 25.2 for the boundary points.
// Notice: D2x:=x[i+1]-x[i-1]/2
// D4x:=x[i+2]-x[i-2]/4, then the formula to be used can be written as
// D := (4/3)D2-(1/3)D4 which is OK since for D2==D4 we get D=D2.
{
   Z mL=4;
   static Word loc("C_Vector::der(R,bool)");
   cpmmessage(mL,loc&" started");
   Z i,n=dim();
   C_Vector res(n);
   if (n<=1) return res;
   cpmassert(h!=0,loc);
   R s;
   if (!periodic){
      if (n==2){
         s=R(1)/h;
         res[1]=((*this)[2]-(*this)[1])*s;
         res[2]=res[1];
      }
      else{  // n>=3
         s=R(1)/(h*12);
         R s1=R(1)/(h*2);
         res[1]=((*this)[1]*(-3)+(*this)[2]*4-(*this)[3])*s1;
         res[2]=(-(*this)[1]+(*this)[3])*s1;
         res[n-1]=(-(*this)[n-2]+(*this)[n])*s1;
         res[n]=((*this)[n-2]-(*this)[n-1]*4+(*this)[n]*3)*s1;
         for (i=3;i<=n-2;i++){
            res[i]=((*this)[i-2]-(*this)[i-1]*8+
               (*this)[i+1]*8-(*this)[i+2])*s;
         }
      }
   }
   else{
      s=R(1)/(h*12);
      for (i=1;i<=n;i++){
         C a1=this->cyc(i-2);
         C a2=this->cyc(i-1);
         C a3=this->cyc(i+1);
         C a4=this->cyc(i+2);
         res[i]=(a1-a2*8+a3*8-a4)*s;
      }
   }
   return res;
}

C_Vector C_Vector::disLap(Outside mode)const
{
   return (*this).rot(1,mode)+(*this).rot(-1,mode)+(*this)*C(-2.0);
}

C_Vector C_Vector::laplace(Z os)const
{
   return C_Vector(rep.laplace(os));
}

R C_Vector::var()const
{
   Z d=dim();
   R res=0;
   for (Z i=1;i<d;i++) res+=(rep[i+1]-rep[i]).abs();
   return res;
}

void C_Vector::adjPhase_(R eps)
{
   Z mL=3;
   Word loc("C_Vector::adjPhase(R)");
   CPM_MA
   Z i,d=dim();
   Vo<R> absv(d);
   for (i=1;i<=d;i++) absv[i]=rep[i].abs();
   V<Z> per=absv.permutationForIncreasingOrder();
//   Z iMax=per[d];
   R absMax=absv[per[d]];
   if (absMax==0) return;
   S<Z> is;
   for (i=1;i<=d;i++){
      if (CpmRoot::dis(absv[i],absMax)<=eps) is.add_(i);
   }
   cpmassert(is.car()>0,loc);
   Z iMaxEff=is[1]; // sets are always ordered !
   C cEff=(*this)[iMaxEff];
   R r=cEff.abs();
   cpmassert(r>0,loc);
   C phase=cEff/r;
   phase=phase.con();
   for (i=1;i<=d;i++) rep[i]*=phase;
   CPM_MZ
}

V<C_Vector> C_Vector::E_Schmidt(V<C_Vector>& x, R eps)
   // returns an orthonormal family which spans the same space
   // as family x. Made by E. Schmidt's orthogonalization. Notice that
   // we do not assume that the family x be linear independent.
   // The procedure goes through family x in the order of indexing.
   // It is evident that the procedure is most stable if the states
   // are ordered in the sense of increasing complexity (structured-ness;
   // e.g. number of peaks, or total variation.
   // Built-in are features specific to the situation of
   // spectralRepresentation
   // No longer needed for this purpose!
{
   static const Z nCrit=8;
      // if the degeneracy of an eigenvalue is larger than this, the
      // standard orthonormalization procedure begins to become very
      // susceptible to numerical errors. We then do reordering.

   //Z mL=1;
   Z mL=3;
   Word loc="C_Vector::E_Schmidt(V<C_Vector>,R)";
   CPM_MA
   Z d=x.dim();
   Z dh=d/2;
   cpmassert(2*dh==d,loc);
   if (dh>nCrit){
      Vo<R> var(d);
      for (Z i=1;i<=d;i++) var[i]=-x[i].var(); // '-' important
      V<Z> per=var.permutationForIncreasingOrder();
         // due to the '-' in the definition of var, we order for
         // decreasing variation. My intuition was to do the opposite.
         // When this was observed not to work, I tried the present
         // version without any expectation for making things better.
         // they became much better, though.
       x=x.permute(per);
   }

   V<C_Vector> u(dh);
   Z iTest=1,iOK=0;
   while (iTest<=d && iOK<dh){
      C_Vector vTest=x[iTest];
      C_Vector next=vTest;
      for (Z i=1;i<=iOK;i++) next-=u[i]*(u[i]|vTest);
      iTest++;
      R rn=next.nor_();
      if (rn>eps){
         iOK++;
         u[iOK]=next;
      }
   }
   if (iOK<dh){
      Z iMissing=dh-iOK;
      Word mes="C_Vector::E_Schmidt(...) "&cpm(iMissing);
      mes&=" states missing of "&cpm(dh);
      cpmmessage(mes);
   }
   CPM_MZ
   return u;
}

C_Vector C_Vector::fromRealForm(R_Vector const& x)
{
   Z i,j,d2=x.dim();
   Z d=d2/2;
   cpmassert(d*2==d2,"");
   C_Vector res(d);
   for (i=1,j=1+d;i<=d;i++,j++) res[i]=C(x[i],x[j]);
   return res;
}

 R_Vector C_Vector::toRealForm()const
 {
   Z d = dim(),i,j;
   Z d2 = 2*d;
   R_Vector res(d2);
   for (i=1, j=d+1; i<=d; i++,j++){
      C zi=rep[i];
      res[i]=zi.real();
      res[j]=zi.imag();
   }
   return res;
 }
      // returns a real 2*d-vector
