//? cpmfft.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// 2011-08-06 'old' data structures eliminated, now
// everything is C+-

#include <cpmfft.h>
#include <cpmx.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;

using CpmArrays::X2;
using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::VVV;

namespace{

inline void swp(R& a, R& b){R t=a;a=b;b=t;}

void fourn(V<R>& data, V<Z> const& dims, Z isign)
{
   Z idim;
   Z i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
   Z ibit,k1,k2,n,nprev,nrem,ntot;
   R tempi,tempr;
   R theta,wi,wpi,wpr,wr,wtemp;
   Z ndim=dims.dim();

   for (ntot=1,idim=1;idim<=ndim;idim++)
      ntot *= dims[idim];
   nprev=1;
   for (idim=ndim;idim>=1;idim--) {
      n=dims[idim];
      nrem=ntot/(n*nprev);
      ip1=nprev*2;
      ip2=ip1*n;
      ip3=ip2*nrem;
      i2rev=1;
      for (i2=1;i2<=ip2;i2+=ip1) {
         if (i2 < i2rev) {
            for (i1=i2;i1<=i2+ip1-2;i1+=2) {
               for (i3=i1;i3<=ip3;i3+=ip2) {
                  i3rev=i2rev+i3-i2;
                  swp(data[i3],data[i3rev]);
                  swp(data[i3+1],data[i3rev+1]);
               }
            }
         }
         ibit=ip2 >> 1;
         while (ibit >= ip1 && i2rev > ibit) {
            i2rev -= ibit;
            ibit /=2;
         }
         i2rev += ibit;
      }
      ifp1=ip1;
      while (ifp1 < ip2) {
         ifp2=ifp1*2;
         theta=R(isign)*cpm2pi/R(ifp2/ip1);
         wtemp=cpmsin(theta*0.5);
         wpr = wtemp*wtemp*(-2.0);
         wpi=cpmsin(theta);
         wr=1.0;
         wi=0.0;
         for (i3=1;i3<=ifp1;i3+=ip1) {
            for (i1=i3;i1<=i3+ip1-2;i1+=2) {
               for (i2=i1;i2<=ip3;i2+=ifp2) {
                  k1=i2;
                  k2=k1+ifp1;
                  tempr=wr*data[k2]-wi*data[k2+1];
                  tempi=wr*data[k2+1]+wi*data[k2];
                  data[k2]=data[k1]-tempr;
                  data[k2+1]=data[k1+1]-tempi;
                  data[k1] += tempr;
                  data[k1+1] += tempi;
               }
            }
            wr=(wtemp=wr)*wpr-wi*wpi+wr;
            wi=wi*wpr+wtemp*wpi+wi;
         }
         ifp1=ifp2;
      }
      nprev *= n;
   }
}

// New systematic treatment of fast Fourier transform
// including the inverse transform in a transparent manner.
// Due to the marvellous properties of function fourn of Press et al.
// The generalization to higher dimensions is trivial.
// According to the new style, only function fourn from Press et al. is
// used.
// All the subtilities of function rlft3 are avoided. If we use the
// method for real data, some wasting of storage is to be admitted since
// we have (artificially in this case), to store also imaginary parts 0.

Z po2(Z n)
   // simple method to get the power of two needed for re-dimensioning
   // Returns the smallest power of two which is equal or larger than n.
{
   Z k=1;
   while (k<n) k*=2;
   return k;
}

X2<V<R>,V<Z> > linearize(const V<C>& x)
{
   Z m=po2(x.dim());
   V<R> res(2*m);
   V<Z> dims{m};
   Z i,k=1;
   for (i=1;i<=m;i++){
      C val=x(i); // returns C() if i is out of range
      res[k++]=val[1];
      res[k++]=val[2];
   }
   return X2<V<R>,V<Z> >(res,dims);
}

X2<V<R>,V<Z> > linearize(const VV<C>& x)
   // See Press et al. Numerical Recipes in C, second edition
   // p.523, Fig 12.4.1, data arrangement on input.
   // The V<Z> component of the return value remembers the
   // matrix frame of the extended (to dimensions that are
   // powers of 2) version of x.
{
   Z m1=po2(x.dim1());
   Z m2=po2(x.dim2());
   V<Z> dims{m1,m2};
   V<R> res(2*m1*m2);
   Z i1,i2,k=1;
   for (i1=1;i1<=m1;i1++) for (i2=1;i2<=m2;i2++){
      C val=x(i1,i2);// returns C() if i1 or i2 is out of range
      res[k++]=val[1];
      res[k++]=val[2];
   }
   return X2<V<R>,V<Z> >(res,dims);
}

X2<V<R>,V<Z> > linearize(const VVV<C>& x)
{
   Z m1=po2(x.dim1());
   Z m2=po2(x.dim2());
   Z m3=po2(x.dim3());
   V<Z> dims{m1,m2,m3};
   V<R> res(2*m1*m2*m3);
   Z i1,i2,i3,k=1;
   for (i1=1;i1<=m1;i1++) for (i2=1;i2<=m2;i2++) for (i3=1;i3<=m3;i3++){
      C val=x(i1,i2,i3);
      res[k++]=val[1];
      res[k++]=val[2];
   }
   return X2<V<R>,V<Z> >(res,dims);
}

// since the 3 versions of deLinearize differ only in the
// return value they can't have exactly the same name.
// We use namespaces to differentiate between them. This
// allows to define CPM_SC uniformly.
namespace aux1{
V<C> deLinearize(V<R> const& x, V<Z> const& dims)
{
   Z mL=3;
   static Word loc("aux1::deLinearize(V<R>,V<Z>)");
   CPM_MA
   Z m=dims[1];
   V<C> res(m);
   R norm=R(m);
   cpmassert(norm>0,loc);
   norm=R(1)/cpmsqrt(norm);
   Z k=1;
   for (Z i=1;i<=m;i++){
      R valRe=x(k++)*norm;
      R valIm=x(k++)*norm;
      res[i]=C(valRe,valIm);
   }
   CPM_MZ
   return res;
 }
} // aux1

namespace aux2{

VV<C> deLinearize(V<R> const& x, V<Z> const& dims)
{
   Z mL=3;
   static Word loc("aux2::deLinearize(V<R>,V<Z>)");
   CPM_MA
   Z m1=dims[1],m2=dims[2];
   VV<C> res(m1,m2);
   R norm=R(m1*m2);
   cpmassert(norm>0,loc);
   norm=R(1)/cpmsqrt(norm);
   Z k=1,i1,i2;
   for (i1=1;i1<=m1;i1++) for (i2=1;i2<=m2;i2++){
      R valRe=x(k++)*norm;
      R valIm=x(k++)*norm;
      res[i1][i2]=C(valRe,valIm);
   }
   CPM_MZ
   return res;
}

} // aux2

namespace aux3{
VVV<C> deLinearize(V<R> const& x, V<Z> const& dims)
{
   Z mL=3;
   static Word loc("aux3::deLinearize(V<R>,V<Z>)");
   CPM_MA
   Z m1=dims[1],m2=dims[2],m3=dims[3];
   VVV<C> res(m1,m2,m3);
   R norm=R(m1*m2*m3);
   cpmassert(norm>0,loc);
   norm=R(1)/cpmsqrt(norm);
   Z k=1,i1,i2,i3;
   for (i1=1;i1<=m1;i1++) for (i2=1;i2<=m2;i2++) for (i3=1;i3<=m3;i3++){
      R valRe=x(k++)*norm;
      R valIm=x(k++)*norm;
      res[i1][i2][i3]=C(valRe,valIm);
   }
   CPM_MZ
   return res;
}
} // aux3

} // namespace

#define CPM_SC\
   X2<V<R>,V<Z> > x2=linearize(x);\
   fourn(x2.c1(),x2.c2(),exponent);\
   return deLinearize(x2.c1(),x2.c2());

V<C> CpmFourier::fft(const V<C>& x, Z exponent)
{
   using aux1::deLinearize;
   CPM_SC
}

VV<C> CpmFourier::fft(const VV<C>& x, Z exponent)
{
   using aux2::deLinearize;
   CPM_SC
}

VVV<C> CpmFourier::fft(const VVV<C>& x, Z exponent)
{
   using aux3::deLinearize;
   CPM_SC
}
#undef CPM_SC
