//? cpmimg24.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmimg24.h>
#include <cpmsystemdependencies.h>
#include <cpmsystem.h>
#include <cpmviewport.h>
#include <cpmgreg.h>
#include <cpmimagingtools.h>
#include <stdlib.h> // for system()

using namespace CpmSystem;
using namespace CpmRoot;
using CpmSystem::IFileStream;
using CpmSystem::OFileStream;
using CpmGraphics::Rec;
using CpmGraphics::Viewport;
using CpmArrays::IvZ;
using CpmArrays::V;
using CpmArrays::Vo;
using CpmArrays::X2;
using CpmFunctions::F;
#ifdef CPM_Fn
using CpmFunctions::F1;
using CpmFunctions::F4;
#else
using std::bind;
using std::placeholders::_1;
#endif
using CpmImaging::ImagingTools;
using CpmLinAlg::Z2;
using CpmLinAlg::R2;
using CpmLinAlg::R3;

namespace CpmGraphics {

///////////////////////// class TriAng ///////////////////////////////////

   R TriAng::area()const
   {
      R res=(p1_.x2+p3_.x2)*(p3_.x1-p1_.x1)
           +(p3_.x2+p2_.x2)*(p2_.x1-p3_.x1)
           +(p1_.x2+p2_.x2)*(p1_.x1-p2_.x1);
      return res*0.5;
   }

   namespace{
      R signR(R const& x)
      { return (x > R(0.) ? R(1.) : (x==0. ? R(0.) : R(-1.)));}
   }

   bool TriAng::out(R2 const& p)const
      // fast implementation which needs not to compute
      // full barycentric coordinates and also needs not always
      // consider all points
   {
      R a1=TriAng(p,p2_,p3_).area();
      R a2=TriAng(p1_,p,p3_).area();
     // R s1=R1(a1).sign();
     // if (s1!= R1(a2).sign()) return true;
      R s1=signR(a1);
      if (s1!= signR(a2)) return true;
      R a3=TriAng(p1_,p2_,p).area();
      if (s1!= signR(a3)) return true;
      return false;
   }

   R3 TriAng::barCor(R2 const& p)const
   {
      Word loc("TriAng::baryCenCor(R2)");
      R a=TriAng(p1_,p2_,p3_).area();
      cpmassert(a!=0,loc);
      R aInv=cpminv(a);
      R a1=TriAng(p,p2_,p3_).area();
      R a2=TriAng(p1_,p,p3_).area();
      R a3=TriAng(p1_,p2_,p).area();
      return R3(a1*aInv,a2*aInv,a3*aInv);
   }

   IvZ TriAng::i1()const
   {
      return IvZ(p1_.x1,p2_.x1)|IvZ(p2_.x1,p3_.x1);
   }

   IvZ TriAng::i2()const
   {
      return IvZ(p1_.x2,p2_.x2)|IvZ(p2_.x2,p3_.x2);
   }

////////////////////////// class Img24 /////////////////////////////////


ColRef Img24::miX(ColRef const& cB, ColRef const& c,
   ColSty const& s, R const& g, R const& w)
{
   if (s==SET_COLOR) return c;
   if (s==STEREO) return cB.add(c);
   R gx=g;
   R wx=w;
   R w1=1,w2=1;
   if (s!=ADD_COLOR){ w1=1-wx; w2=wx;}
   // w is the weight of the requested color not of
   // the background! Remember that 1 is the default
   // for w, which would be nonsense if this would
   // be the other way round.
   return rgb::mix((rgb)cB,(rgb)c,w1,w2,gx);
}

// We set gamma 1 per default (and not equal to
// Viewport::gamma) since this is much faster

Img24::Img24():
   width(0),height(0),c0(),sty(SET_COLOR),gamma(1){}

Img24::Img24(ColRef c, R fracX, R fracY):
   c0(c),sty(SET_COLOR),gamma(Viewport::getGamma())
{
   Z mL=3;
   static Word loc("Img24(ColRef,R,R)");
   width=cpmtoz(R(Viewport::win().w())*fracX);
   height=cpmtoz(R(Viewport::win().h())*fracY);
   if (width && height){
      V<ColRef> vw(width,c0,0);
      bh_=V<V<ColRef> >(height,vw,0);
   }
   cpmmessage(mL,loc&" done");
}

Img24::Img24(Z x, Z y, rgb c, ColSty cs):
   c0(c),sty(cs),gamma(1)
{
   Z mL=3;
   static Word loc("Img24(Z,Z,rgb,ColSty)");
   cpmmessage(mL,loc&" started: x="&cpmwrite(x)&" y="&cpmwrite(y));
   if (x<0){
      x=0;
      cpmwarning("Img24(Z x,Z y,rgb,ColSty): input x<0 replaced by x=0");
   }

   if (y<0){
      y=0;
      cpmwarning("Img24(Z x,Z y,rgb,ColSty): input y<0 replaced by y=0");
   }

   width=x;
   height=y;
   if (width && height){
      bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   }
      // notice that mem_ does not hold all the same byte, so that a
      // specially simple initialization of mem_ is not at hand
   cpmmessage(mL,loc&" done");
}

Img24::Img24(const VVV<L>& vv):width(vv.dim3()),height(vv.dim2()),
   sty(SET_COLOR),gamma(1)
{
   Z mL=4;
   cpmmessage(mL,"Img24(const VVV<L>& ) started");
   if (vv.dim1()!=3) cpmerror("Img24(const VVV<L>& vv): vv.dim()!=3");
   bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
         L r=vv[1][i][j];
         L g=vv[2][i][j];
         L b=vv[3][i][j];
         ColRef c(r,g,b);
         bh_[i-1][j-1]=c; // matrix style
      }
   }
   cpmmessage(mL,"Img24(const VVV<L>& ) done");
}

namespace {

   rgb con(R x, R y, R z, R alpha, R beta)
   {
      R xn=x*alpha+beta;
      R yn=y*alpha+beta;
      R zn=z*alpha+beta;
      return rgb(cpmtoz(xn),cpmtoz(yn),cpmtoz(zn));
   }
}

/******************
Img24::Img24(VVV<R> const& vv):width(vv.dim3()),height(vv.dim2()),
   sty(SET_COLOR),gamma(1),c0()
{
   Z mL=3;
   Word loc("Img24::Img24(VVV<R>)");
   CPM_MA
   Z nP=vv.dim1();
   cpmassert(nP>0,loc);
   if (nP<3) nP=1; else nP=3;
   R vHuge=1e64; //
   R vMin=vHuge, vMax=-vMin;
   Z i,j,k;
   for (i=1;i<=height;++i){
      for (j=1;j<=width;++j){
         for (k=1;k<=nP;++k){
            R v=vv[k][i][j];
            if (v>vMax) vMax=v;
            if (v<vMin) vMin=v;
         }
      }
   }
   bool direct=vMin>-0.5 && vMax<255.5;
   R alpha=1, beta=0;
   if (direct){
      if (vMin<0) beta=-vMin;
      // here we only shift a bit to eliminate tiny negative values
      // which may arrise from image processing algorithms with
      // floating point arithmetics.
   }
   else{ // image is out of range [0,255]
      R diff=vMax-vMin;
      if (diff==0.){ // image is constant. It will be
         // scaled to gray
         alpha=128./vMax;
         // if vMax would be 0, we had vMax=vMin=0. All values would
         // be zero, and we would have the case 'direct'
      }
      else{ // scaled to exact range [0,255]
         alpha=255./diff;
         beta=-alpha*vMin;
      }
   }

   bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   for (i=1;i<=height;++i){
      for (j=1;j<=width;++j){
         R v1=vv[1][i][j],v2=v1,v3=v1;
         if (nP==3){
            v2=vv[2][i][j];
            v3=vv[3][i][j];
         }
         rgb c=con(v1,v2,v3,alpha,beta);
         bh_[i-1][j-1]=c; // matrix style
      }
   }
   CPM_MZ;
}
*******************/

Img24::Img24(VVV<R> const& vv):width(vv.dim3()),height(vv.dim2()),c0(),
   sty(SET_COLOR),gamma(1)
{
   Z mL=3;
   Word loc("Img24::Img24(VVV<R>)");
   CPM_MA
   Z nP=vv.dim1();
   cpmassert(nP>0,loc);
   if (nP<3) nP=1; else nP=3;
/**************************
   R vHuge=1e64; //
   R vMin=vHuge, vMax=-vMin;
   Z i,j,k;
   for (i=1;i<=height;++i){
      for (j=1;j<=width;++j){
         for (k=1;k<=nP;++k){
            R v=vv[k][i][j];
            if (v>vMax) vMax=v;
            if (v<vMin) vMin=v;
         }
      }
   }
   bool direct=vMin>-0.5 && vMax<255.5;
   R alpha=1, beta=0;
   if (direct){
      if (vMin<0) beta=-vMin;
      // here we only shift a bit to eliminate tiny negative values
      // which may arrise from image processing algorithms with
      // floating point arithmetics.
   }
   else{ // image is out of range [0,255]
      R diff=vMax-vMin;
      if (diff==0.){ // image is constant. It will be
         // scaled to gray
         alpha=128./vMax;
         // if vMax would be 0, we had vMax=vMin=0. All values would
         // be zero, and we would have the case 'direct'
      }
      else{ // scaled to exact range [0,255]
         alpha=255./diff;
         beta=-alpha*vMin;
      }
   }
**************/
   bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   for (Z i=1;i<=height;++i){
      for (Z j=1;j<=width;++j){
         R v1=vv[1][i][j],v2=v1,v3=v1;
         if (nP==3){
            v2=vv[2][i][j];
            v3=vv[3][i][j];
         }
         rgb c(v1,v2,v3); // new mode rgb directly from R-valued arguments
         bh_[i-1][j-1]=c; // matrix style
      }
   }
   CPM_MZ;
}

Img24::Img24(const VV<L>& vv):width(vv.dim2()),height(vv.dim1()),
   sty(SET_COLOR),gamma(1)
{
   Z mL=4;
   Word loc("Img24(VV<L>)");
   CPM_MA
   bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
       //  L r=vv[i][j];
         Z r(vv[i][j]);
         rgb c(r,r,r);
         bh_[i-1][j-1]=c; // matrix style
      }
   }
   CPM_MZ
}

Img24::Img24(const VV<ColRef>& vv):width(vv.dim2()),height(vv.dim1()),
   sty(SET_COLOR),gamma(1)
{
   Z mL=4;
   Word loc("Img24(VV<ColRef>)");
   CPM_MA
   bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
         bh_[i-1][j-1]=vv[i][j]; // matrix style
      }
   }
   CPM_MZ
}

void Img24::resize(Z x, Z y)
{
   Z mL=3;
   static Word loc("Img24::resize(Z,Z)");
   cpmmessage(mL,loc&" started: x="&cpm(x)&" y="&cpm(y));
   if (x<0){
      x=0;
      cpmwarning(loc&": input x<0 replaced by x=0");
   }
   if (y<0){
      y=0;
      cpmwarning(loc&": input y<0 replaced by y=0");
   }
   width=x;
   height=y;
   if (width && height){
      bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   }
   readyForDisplay=B();
   CPM_MZ
}

Img24 Img24::trn()const
{
   Z mL=3;
   static Word loc("Img24::trn()");
   CPM_MA
   Z i,j,w=width,h=height;
   Img24 res(h,w);
   for (i=0;i<h;i++){
      for (j=0;j<w;j++){
         res.rep_()[j][i]=bh_[i][j];
      }
   }
   CPM_MZ
   return res;
}

Img24 Img24::crop(Iv const& ivx, Iv const& ivy)const
{
   Z mL=3;
   static Word loc("Img24::cut(Iv,Iv)");
   CPM_MA
   Iv xR=Iv(ivx.inf()*width,ivx.sup()*width);
   Iv yR=Iv(ivy.inf()*height,ivy.sup()*height);
      // now the y coordinates are those used in function disY
   Z i0=disX(xR.inf());
   Z j0=disY(yR.inf());
   Z i1=disX(xR.sup());
   Z j1=disY(yR.sup());
   Z w=i1-i0;
   Z h=j1-j0;
   Z i,j,i2,j2;
   Img24 res(w,h);
   for (i=0,i2=i0;i<w;i++,i2++){
      for (j=0,j2=j0;j<h;j++,j2++){
         res.rep_()[j][i]=bh_[j2][i2];
      }
   }
   CPM_MZ
   return res;
}

Img24 Img24::ref(Z p)const
{
   Z mL=3;
   static Word loc("Img24::ref(Z)");
   CPM_MA

   if (p==0){
      CPM_MZ
      return *this;
   }
   Z i,j,ir,jr,w=width,h=height;
   Img24 res(w,h);
   if (p==1){
      for (i=0,ir=h-1;i<h;++i,--ir){
         for (j=0;j<w;j++){
            res.rep_()[ir][j]=bh_[i][j];
         }
      }
   }
   else if (p==2){
      for (i=0;i<h;++i){
         for (j=0,jr=w-1;j<w;++j,--jr){
            res.rep_()[i][jr]=bh_[i][j];
         }
      }
   }
   else {
      for (i=0,ir=h-1;i<h;++i,--ir){
         for (j=0,jr=w-1;j<w;++j,--jr){
            res.rep_()[ir][jr]=bh_[i][j];
         }
      }
   }
   CPM_MZ
   return res;
}

void Img24::frame(rgb c)
{
   line(0,0,width-1,0,c);
   line(0,height-1,width-1,height-1,c);
   line(0,0,0,height-1,c);
   line(width-1,0,width-1,height-1,c);
   readyForDisplay=B();
}

VV<ColRef> Img24::toVVColRef()const
{
   VV<ColRef> res(height,width);
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
         res[i][j]=bh_[i-1][j-1];
      }
   }
   return res;
}

VVV<L> Img24::toVVV()const
{
   VVV<L> res(3,height,width);
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
         rgb c=getPel(j-1,i-1);
         res[1][i][j]=(L)c.r;
         res[2][i][j]=(L)c.g;
         res[3][i][j]=(L)c.b;
     }
   }
   return res;
}

VV<R> Img24::toVVR(R wR, R wG, R wB, Iv val)const
{
   VV<R> res(height,width);
   R wInv=cpminv((wR+wG+wB)*255);
   R val1=val.inf();
   R val2=val.sup();
   R dval=val2-val1;
   for (Z i=1;i<=height;i++){
      for (Z j=1;j<=width;j++){
         rgb c=getPel(j-1,i-1);
         R vR=R(c.r);
         R vG=R(c.g);
         R vB=R(c.b);
         R v=wR*vR+wG*vG+wB*vB;
         v*=wInv;
         if (val1==0. && val2==1.)
            res[i][j]=v;
         else
            res[i][j]=val1+dval*v;
     }
   }
   return res;
}

Img24::Img24(const Word& fileName):c0(),sty(SET_COLOR),gamma(1)
{
   Z mL=1;
   Word loc("Img24(Word)");
   CPM_MA
   Word filename=fileName;
   Word jpg1(".jpg");
   Word jpg2(".JPG");
   Word tail=filename.tail(4);
   if (tail==jpg1||tail==jpg2){
      Word fn=filename.cut(3)&"ppm";
      Word cl=Word("gm convert ")&filename&" "&fn;
      Z res=system(-cl);
      filename=fn;
      cpmmessage(loc&": image file converted to "&filename
                 &" returning "&cpm(res));
   }
   IFileStream in(filename);
   if (!in()){
      cpmerror("Img24(Word): bad stream opening "&filename);
      return;
   }
   string w1;
   string indicator1("P3");
   string indicator2("P6");
   in()>>w1;
   cpmmessage(loc&": file type read as "&Word(w1));
   in.skipComments();
      // several programs that create ppm-files introduce comment-lines
      // which have to be overread
   if (w1!=indicator1 && w1!=indicator2){
      ostringstream ost;
      ost<<"Img24("<<filename<<
         "): image file is not of type ppm P6 or P3";
      ost<<" instead we read "<<w1;
      cpmerror(Word(ost.str()));
   }
   in()>>width;
   in.skipComments();
   in()>>height;
   in.skipComments();
   Z depth;
   in()>>depth;
   in.skipComments();
   if (depth!=255){
      ostringstream ost;
      ost<<"Img24("<<filename<<"): image file has depth="<<depth;
      cpmerror(Word(ost.str()));
   }
   bh_=V<V<ColRef> >(height,V<ColRef>(width,ColRef(),0),0);

   if (w1==indicator2) in().unsetf(std::ios_base::skipws);
      // P6
   if (w1==indicator1){ // P3, non-binary reading
      unsigned short r,g,b;
      for (Z i=0;i<height;i++){
         for (Z j=0;j<width;j++){
            in()>>r;
            in()>>g;
            in()>>b;
            bh_[i][j]=rgb(Z(r),Z(g),Z(b));
         }
      }
   }
   else{ // P6
      L r,g,b;
      for (Z i=0;i<height;i++){
         for (Z j=0;j<width;j++){
            in()>>r;
            in()>>g;
            in()>>b;
            bh_[i][j]=rgb(Z(r),Z(g),Z(b));
         }
      }
   }
   CPM_MZ
}

void Img24::insert(xy p, Img24 const& img , ColRef cr)
// fast insert of image lines as whole
{
   Z mL=3;
   Word loc("Img24::insert(xy,Img24)");
   CPM_MA
   xy tp(0,0);
   xyxy tr(tp,width,height);
   Z iw=img.w(), ih=img.h();
   xyxy ir(p,iw,ih);
   ir=tr&ir; // set section, now ir fits the frame
      // of *this or it is void
   if (ir.isVoid()&& cpmverbose>1){
      cpmwarning(loc&": nothing to insert");
      return;
   }
   iw=ir.w();ih=ir.h();
   for (Z j=0;j<ih;++j){
      horIns(p,img[j],cr);
      p.y++;
   }
   readyForDisplay=B();
   CPM_MZ
}

void Img24::adjDisBuf()const
{
   if (!readyForDisplay){
      displayBuffer=Rec(bh_);
      Z dw=displayBuffer.w();
      Z dh=displayBuffer.h();
      Z excess_x=dw-Viewport::win().w();
      Z excess_y=dh-Viewport::win().h();
     Z sx=0, sy=0;
      if (excess_x>0){
         sx-=excess_x;
         if (sx<0) sx=0;
      }
      if (excess_y>0){
         sy-=excess_y;
         if (sy<0) sy=0;
      }
      // if the properly shifted image does not fit
      // Viewport::win() we reduce the shift to zero, so that
      // as much as possible can be seen
      displayBuffer+=xy(sx,sy);
      readyForDisplay=B(true);
   }
}


Img24& Img24::operator+=(xy const& s)
{
   displayBuffer=Rec(bh_);
   Z dw=displayBuffer.w();
   Z dh=displayBuffer.h();
      //Z excess_x=sx+displayBuffer.x1()+dw-Viewport::win().w();
      //Z excess_y=sy+displayBuffer.y1()+dh-Viewport::win().h();
         // unsuccessful attempt to let the display area directly related
         // to displayBuffer.rec_
   Z sx=s.x;
   Z sy=s.y;
   Z excess_x=sx+dw-Viewport::win().w();
   Z excess_y=sy+dh-Viewport::win().h();
   if (excess_x>0){
      sx-=excess_x;
      if (sx<0) sx=0;
   }
   if (excess_y>0){
      sy-=excess_y;
      if (sy<0) sy=0;
   }
      // if the properly shifted image does not fit
      // Viewport::win() we reduce the shift to zero, so that
      // as much as possible can be seen
   xy shift(sx,sy);
   displayBuffer+=shift;
   readyForDisplay=B(true);
   return *this;
}

void Img24::vis()const
// important that this is fast since Frame::display is based on it
{
   Z mL=3;
   static Word loc("Img24::vis()");
   CPM_MA
   adjDisBuf();
   displayBuffer.vis();
   CPM_MZ
}

Img24::Img24(Rec const& rec):width(rec.w()),height(rec.h()),
   sty(SET_COLOR),gamma(1)
{
   xy p1=rec[1];
   displayBuffer=rec;
   displayBuffer-=p1;
   bh_=rec.fold();
   readyForDisplay=B(true);
}

Word Img24::write(const Word& filename, bool binary)
{
   Z mL=3;
   Word loc("Img24::write(");
   loc&=(filename&","&cpm(binary)&")");
   CPM_MA
   Word ffn=filename&".ppm";
   // writing a portable pixmap; very simple
   const Z maxVal=255;
   cpmmessage(mL,"width="&cpmwrite(width)&", height="&cpmwrite(height));
   if (binary){
      OFileStream out(ffn);
      out()<<"P6"<<'\n'<<width<<' '<<height<<'\n'<<maxVal<<'\n';
      for (Z i=0; i<height;i++){
         for (Z j=0;j<width;j++){
            ColRef val=bh_[i][j];
            out()<<val.r<<val.g<<val.b;
         }
      }
   }
   else{ // not binary, explicit "publication mode"
      const string indicator="P3";
      // indicator for ASCII portable pixmap
      Word title="# Portable pix map made by CpmGraphics::Img24:write(";
      title&=filename;
      title&=") ";
      CpmTime::Greg greg;
      greg.now();
      Word date="# Date: ";
      date&=greg.getDate(false);
      Word textOut="#";
      bool safe=true;
      binary=false;
      OFileStream out(ffn,safe,binary);
      out()<<indicator<<endl;
      out()<<title<<endl;
      out()<<date<<endl;
      out()<<"# Text:"<<endl;
      out()<<textOut<<endl;
      out()<<"# Image data: width, height, maximum code value"<<endl;
      out()<<width<<" "<<height<<endl;
      out()<<maxVal<<endl;
      Z countMax=5;
      Z count=0;
      for (Z i=0; i<height;i++){
         for (Z j=0;j<width;j++){
            ColRef val=bh_[i][j];
            out()<<Z(val.r)<<" "<<Z(val.g)<<" "<<Z(val.b)<<"   ";
            count++;
            if (count==countMax){
               count=0;
               out()<<endl;
            }
         }
      }
      out()<<endl;
   }
   CPM_MZ
   return ffn;
}

bool Img24::clip(R& ax, R& ay, R& bx, R& by)const
// Cohen-Sutherland algorithm,modified from two internet sources
{
   static L X_TOO_SMALL=8;
   static L X_TOO_LARGE=4;
   static L Y_TOO_SMALL=2;
   static L Y_TOO_LARGE=1;
   L c1 = outcode( ax, ay);
   L c2 = outcode( bx, by);
   L c;
   R minusDeltaY, deltaX, d, x{0.}, y{0.};
   while((c1 | c2) != 0){
      if ((c1 & c2) != 0) return false ;
      minusDeltaY = ay-by;
      deltaX = bx-ax;
      d = ax*by-ay*bx;
      if (c1 == 0)
         c = c2;
      else
         c = c1;
      if ((c & X_TOO_SMALL) != 0) {
         x = 0;
         y = -(minusDeltaY * x + d) / deltaX;
      }
      else if ((c & X_TOO_LARGE) != 0) {    //c ist rechts
         x = width;
         y = -(minusDeltaY * x + d) / deltaX;
      }
      else if ((c & Y_TOO_LARGE) != 0) {      //c ist oben
         y = height;
         x = -(deltaX*y + d) / minusDeltaY;
      }
      else if ((c & Y_TOO_SMALL) != 0) {   //c ist unten
         y = 0;
         x = -(deltaX * y + d)/minusDeltaY;
      }
      if (c == c1) {
         ax = x;
         ay = y;
         c1 = outcode(ax, ay);
      }
      else {
         bx = x;
         by = y;
         c2 = outcode(bx, by);
      }
   }
   return true;
}

void Img24::lineThin(R x1, R y1, R x2, R y2, rgb c)
{
   if (newLines) return lineThinNew(x1,y1,x2,y2,c);
   if (!CpmRoot::isVal(x1)) return;
   if (!CpmRoot::isVal(y1)) return;
   if (!CpmRoot::isVal(x2)) return;
   if (!CpmRoot::isVal(y2)) return;
      // input may be irregular; then it is OK to do nothing
   static R tiny=0.001;
   bool b=clip(x1,y1,x2,y2);
   if (!b) return;

   ColSty styMem=sty;
   R wMixMem=wMix;
   if (antiAliasLines) sty=MIX_COLOR;

   Z i,j,i1=disX(x1), i2=disX(x2), j1=disY(y1), j2=disY(y2);
      // x1,x2,y1,y2 no longer needed
   if (i1==i2 && j1==j2){
      putPel(i1,j1,c); // had direct index access till 2010-08-17,
      // caused rare failures.
   }
   else{
      Z pdi=i2>=i1 ? i2-i1 : i1-i2;
      Z pdj=j2>=j1 ? j2-j1 : j1-j2;
      if (pdj<=pdi){ // flat case. Since at least one of pdi, pdj is
         // !=0, we know pdi>0, thus i1!=i2. We want to work over
         // increasing i, this can be achieved by exchanging the
         // role of 1 and 2:
         if (i2<i1){
            cpmswp<Z>(i1,i2);
            cpmswp<Z>(j1,j2);
         } // no 'else' needed
         // now a loop 'for (i=i1;i<=i2;++i)' is OK
         R m=R(j2-j1)/(i2-i1);
            // the definition m=(y2-y1)/(x2-x1) gives bad curves
         R yC=antiAliasLines ? R(j1)+0.5 : R(j1)+tiny;
         for (i=i1;i<=i2;++i){
            Z jC=disY(yC);
            if (antiAliasLines){
               R frac=yC-jC-0.5; // ranges in [-0.5,0.5)
             //  cpmassert(frac>=-0.5 && frac<0.5,loc); //tested to hold
               Z jC1= frac>=0 ? disY(yC+1) : disY(yC-1);
               if (frac<0) frac*=-1;
               wMix=R(1)-frac;
               putPel(i,jC,c);
               wMix=frac;
               putPel(i,jC1,c);

            }
            else{
               if (sty==SET_COLOR) bh_[jC][i]=c; else putPel(i,jC,c);
            // fast access
            }
            yC+=m;
         } // i-loop
      } // flat case
      else{ // steep case
         if (j2<j1){
            cpmswp<Z>(i1,i2);
            cpmswp<Z>(j1,j2);
         } // no 'else' needed
         // now a loop 'for (j=j1;j<=j2;++j)' is OK
         R m=R(i2-i1)/(j2-j1);
         R xC=antiAliasLines ? R(i1)+0.5 : R(i1)+tiny;
         for (j=j1;j<=j2;++j){
            Z iC=disX(xC);
            if (antiAliasLines){
               R frac=xC-iC-0.5;
              // cpmassert(frac>=-0.5 && frac<0.5,loc);//tested to hold
               Z iC1 = frac>=0 ? disX(xC+1) : disX(xC-1);
               if (frac<0) frac*=-1;
               wMix=frac;
               putPel(iC1,j,c);
               wMix=R(1)-frac;
               putPel(iC,j,c);
            }
            else{
              if (sty==SET_COLOR) bh_[j][iC]=c; else putPel(iC,j,c);
            }
            xC+=m;
         } // j-loop
      } // steep case
   } // if (i1==i2 && j1==j2){...}else{...}
   if (antiAliasLines){
      sty=styMem;
      wMix=wMixMem;
   }
   readyForDisplay=B();
}

namespace{
R floorA(R x){
    const R tiny=0.001;
    static R near1=R(1)-tiny;
    R y=cpmfloor(x);
    return x-y > near1 ? y+1 : y;
}
}// namespace

void Img24::lineThinNew(R x1, R y1, R x2, R y2, rgb c)
{
   const R tiny=0.001;
   const R large=1000000.;
   static R r2Inv=1./cpmsqrt(2.);
   if (!CpmRoot::isVal(x1)) return;
   if (!CpmRoot::isVal(y1)) return;
   if (!CpmRoot::isVal(x2)) return;
   if (!CpmRoot::isVal(y2)) return;
      // input may be irregular; then it is OK to do nothing
   bool b=clip(x1,y1,x2,y2);
   if (!b) return;

   R2 e(x2-x1,y2-y1);
   R le=e.nor_();
   R exInv=e.x1!=0 ? R(1)/e.x1 : R(0);
   R eyInv=e.x2!=0 ? R(1)/e.x2 : R(0);
   R2 pa,pb,pe;
   R lCum=0./*,lAct=0*/;
   bool ready=false;
   pa=R2(x1,y1);
   Z ip,jp;
   pe=R2(floorA(pa.x1),floorA(pa.x2)); // point on grid
   while(!ready){
       R ni=e.x1>0 ? pe.x1+1 :
          (cpmabs(pa.x1-pe.x1)<tiny ? pe.x1-1: pe.x1);
       R nj=e.x2>0 ? pe.x2+1 :
          (cpmabs(pa.x2-pe.x2)<tiny ? pe.x2-1: pe.x2);
       R lx = exInv!=0 ? (ni-pa.x1)*exInv : large;
       R ly = eyInv!=0 ? (nj-pa.x2)*eyInv : large;
       R lAct=std::min<R>(lx,ly);
       lCum+=lAct;
       if (lCum>le){
           R diff=lCum-le;
           lAct-=diff;
           ready=true;
       }
       pb=pa+e*lAct;
       R lr=lAct*r2Inv;
       wMix=lr; // for style SET_COLOR the value of wMix has
        // no influence.
       pe=R2(floorA(pb.x1),floorA(pb.x2));
       ip=cpmrnd(pe.x1);
       jp=cpmrnd(pe.x2);
       if (lr>tiny) putPel(ip,jp,c);
       pa=pb;
   }
   readyForDisplay=B();
}

namespace{

ColRef fStereo(ColRef const& cB, ColRef const& c){ return cB.add(c);}

} // namespace

bool Img24::horLine(Z x1, Z x2, Z y, rgb c)
{
   if (!i2().hasElm(y)) return false;
      // nothing to do
   IvZ iv=i1()&IvZ(x1,x2);
      // restriction of line to width
   ColRef cr=c;
   if (sty!=STEREO){
#ifdef CPM_Fn
      F<ColRef,ColRef> fc=F4<ColRef,ColRef,ColSty,R,R,ColRef>
         (cr,sty,gamma,wMix)(miX);
#else
   F<ColRef,ColRef> fc(bind(miX,_1,cr,sty,gamma,wMix));
#endif
      bh_[y].fa_(fc,iv);
   }
   else{ // should be fast, so we don't hand-over arguments
      // that are not needed
#ifdef CPM_Fn
      F<ColRef,ColRef> fc=F1<ColRef,ColRef,ColRef>(cr)(fStereo);
#else
      F<ColRef,ColRef> fc(bind(fStereo,_1,cr));
#endif
      bh_[y].fa_(fc,iv);
   }
   readyForDisplay=B();
   return true;
}

namespace{

ColRef fSet(ColRef const& cB, ColRef const& c)
{
   return c;
}

ColRef cWrite;
   // notice that we can't make cWrite a part
   // of the argument list of fSet2, so we pass
   // it 'over the head of this function by initializing
   // it in the function body of horIns; very
   // powerful trick which looks quite acceptable to me
   // when encasulated in a namespace as done here.

ColRef fSet2(ColRef const& cB, ColRef const& c)
{
   if (c.r<255){ // c is assumed to come from
      // a monochrome line to be inserted
      double w1=c.r/255., w2=1-w1;
      return ColRef((Z)(w1*cB.r+w2*cWrite.r),
                    (Z)(w1*cB.g+w2*cWrite.g),
                    (Z)(w1*cB.b+w2*cWrite.b));
   }
   else return cB;
}

}// namespace

void Img24::horIns(xy p, V<ColRef> const& lineA, ColRef cr)
// isn't this elegant ?
// safety features added 2005-09-12
// Argument cr added 2005-09-15
{
   V<ColRef> line=lineA;
   line.b_(0);
   if (p.y<0 || p.y>=height || p.x>=width) return;
      // readyForDisplay needs no change
   Z ld=line.dim();
   if (ld==0) return;
   IvZ rl=IvZ(p.x,p.x+ld-1); // range of line
   IvZ i=i1()&rl;
   if (i.isVoid()) return;
      // end of trivial returns
   cWrite=cr;
   if (cr==ColRef()){ // use fSet
      if (p.x>=0)
         bh_[p.y].ff_(fSet,line,p.x);
         // even if the line is too long for the image width
         // no cutting is needed
      else // line manipulation only if needed
         bh_[p.y].ff_(fSet,line.eliFirst(-p.x),0);
   }
   else{ // use fSet2
      if (p.x>=0)
         bh_[p.y].ff_(fSet2,line,p.x);
      else
         bh_[p.y].ff_(fSet2,line.eliFirst(-p.x),0);
   }
   readyForDisplay=B();
}

void Img24::clear(void)
{
   Z mL=3;
   Word loc("Img24::clear(void)");
   CPM_MA
   if(width && height){
      bh_=V<V<ColRef> >(height,V<ColRef>(width,c0,0),0);
   }
   if (c0.isGray()){
      L gb=c0.r;
      xyxy r=displayBuffer.rec();
      displayBuffer=Rec(r,gb);
      readyForDisplay=B(true);
   }
   else readyForDisplay=B();
   CPM_MZ
}

void Img24::fillRectangle(R x1, R y1, R x2, R y2, rgb c)
{
   CpmRootX::order<R>(x1,x2);
   CpmRootX::order<R>(y1,y2);
   Z zx1=cpmtoz(x1);
   Z zx2=cpmtoz(x2);
   Z zy1=cpmtoz(y1);
   Z zy2=cpmtoz(y2);
   for (Z j=zy1;j<=zy2;j++) horLine(zx1,zx2,j,c);
}

void Img24::fill(TriAng const& tri, rgb c)
{
   IvZ iv1=tri.i1();
   iv1=iv1&i1();
   if (iv1.isVoid()) return;
   IvZ iv2=tri.i2();
   iv2=iv2&i2();
   if (iv2.isVoid()) return;
   R shift=0.0;
   for (Z i=iv1.b();i<=iv1.e();++i){
      for (Z j=iv2.b();j<=iv2.e();++j){
         R2 p(shift+i,shift+j);
         bool out=tri.out(p);
         if (!out){
            putPel(i,j,c);
         }
      }
   }
}

void Img24::fillEllipse(R x1, R y1, R x2, R y2, rgb c)
// one should at least mark the center, that very small ellipses don't
// rely on numerical noise for getting shown
{
   Z mL=3;
   Word loc("Img24::fillEllipse(R,R,R,R,rgb)");
   CPM_MA
   CpmRootX::order<R>(x1,x2);
   CpmRootX::order<R>(y1,y2);
   xy z1=disXY(x1,y1), z2=disXY(x2,y2);
   Z zx1=z1.x, zy1=z1.y, zx2=z2.x, zy2=z2.y;

   if ((IvZ(zx1,zx2)&i1()).isVoid()) return;
   if ((IvZ(zy1,zy2)&i2()).isVoid()) return;

   R w2=(x2-x1)*0.5;
   R w22=w2*w2;
   R h2=(y2-y1)*0.5;
   R h22inv=1.;
   if (y1!=y2){
      R h22=h2*h2;
      h22inv=R(1)/h22;
   }
   R x0=w2+x1;
   R y0=h2+y1;

   bool marked=false;

   for(Z j=zy1; j<=zy2; j++){
      R rz=R(j)-y0;
      R rz2=rz*rz*h22inv;
      R wj=(R(1)-rz2)*w22;
      if (wj<0) continue;
      wj=cpmsqrt(wj);
      Z iL=cpmtoz(x0-wj);
      Z iU=cpmtoz(x0+wj);
      bool b=horLine(iL,iU,j,c);
      if (b==true) marked=true;
   }
   if (!marked){
      R xm=(x1+x2)*0.5, ym=(y1+y2)*0.5;
      putPel(xm,ym,c); // these 2 lines added 2006-11-09
   }
   CPM_MZ
}

bool Img24::prnOn(ostream& str)const
{
   cpmwt("Img24 begin");
   cpmp(width);
   cpmp(height);
   cpmp(bh_);
   cpmp(c0);
   cpmp((Z)(sty));
   cpmp(antiAliasLines);
   cpmp(gamma);
   if (sty==MIX_COLOR) cpmp(wMix);
   cpmwt("Img24 end");
   return true;
}

bool Img24::scanFrom(istream& str)
{
   Z mL=3;
   Word loc("Img24::scanFrom(istream)");
   CPM_MA
   readyForDisplay=B();
      // since we read only the primary image data
      // and not the derived data displayBuffer
   cpms(width);
   cpmmessage(mL,"reading width done");
   cpms(height);
   cpmmessage(mL,"reading height done");
   cpms(bh_);
   cpmmessage(mL,"reading bh_ done");
   cpms(c0);
   cpmmessage(mL,"reading c0 done");
   Z styAux;
   cpms(styAux);
   cpmmessage(mL,"reading styAux done");
   sty=(ColSty)styAux;
   cpms(antiAliasLines);
   cpmmessage(mL,"reading antiAliasLines done");
   cpms(gamma);
   cpmmessage(mL,"reading gamma done");
   if (sty==MIX_COLOR){
      cpms(wMix);
      cpmmessage(mL,"reading wMix done");
   }
   CPM_MZ
   return true;
}

Z Img24::com(Img24 const& obj)const
{
   cpmc(width); // new macro cpmc asks for name 'obj'
      // of the argument
   cpmc(height);
   cpmc(bh_);
   cpmc(c0);
   cpmc(antiAliasLines);
   cpmc(gamma);
   return 0;
}

ColRef Img24::biLin(ColRef f11, ColRef f12, ColRef f21, ColRef f22,
   R w11, R w12, R w21, R w22, Z method)
{
   double v11=cpmtod(w11);
   double v12=cpmtod(w12);
   double v21=cpmtod(w21);
   double v22=cpmtod(w22);
   if (method==1){
      double p11=v11*v21;
      double p12=v11*v22;
      double p21=v12*v21;
      double p22=v12*v22;
      Z r=cpmtoz(f11.r*p11 + f12.r*p12 + f21.r*p21 + f22.r*p22);
      Z g=cpmtoz(f11.g*p11 + f12.g*p12 + f21.g*p21 + f22.g*p22);
      Z b=cpmtoz(f11.b*p11 + f12.b*p12 + f21.b*p21 + f22.b*p22);
      return ColRef(r,g,b);
   }
   else{
      Z i= w11<=w12 ? 2 : 1;
      Z j= w21<=w22 ? 2 : 1;
      if (i==1 && j==1) return f11;
      else if (i==1 && j==2) return f12;
      else if (i==2 && j==1) return f21;
      else return f22;
   }
}

Img24 Img24::intPolAux(Z nI, Z mI, Z method)const
   // See function int2
{
   Z mL=3;
   Word loc("Img24::intPolAux(Z,Z,Z)");
   CPM_MA
   Z m=h();
   Z n=w();
   if (m==mI && n==nI){
      cpmmessage(mL,loc&": no size change, nothing to do");
      return *this;
   }
   Z i,j;
   ColRef f11,f12,f21,f22;
   Iv iv(0,1);
   V<R> ci=iv.centers(m);
   V<R> cj=iv.centers(n);
   V<R> cIi=iv.centers(mI);
   V<R> cIj=iv.centers(nI);

   Img24 res(nI,mI);
   for (i=1;i<=mI;i++){
      R xiI=cIi[i];
      X2<R2,Z2> qi=iv.intPolInfo(xiI,m);
      R2 wi=qi.c1();
      Z2 zi=qi.c2();
      R wi1=wi.x1;
      R wi2=wi.x2;
      Z i1=zi.x1-1;
      Z i2=zi.x2-1;
      for (j=1;j<=nI;j++){
         R xjI=cIj[j];
         X2<R2,Z2> qj=iv.intPolInfo(xjI,n);
         R2 wj=qj.c1();
         Z2 zj=qj.c2();
         R wj1=wj.x1;
         R wj2=wj.x2;
         Z j1=zj.x1-1;
         Z j2=zj.x2-1;
         f11=bh_[i1][j1];
         f12=bh_[i1][j2];
         f21=bh_[i2][j1];
         f22=bh_[i2][j2];
         res.bh_[i-1][j-1]=biLin(f11,f12,f21,f22,wi1,wi2,wj1,wj2,method);
      }
   }
   CPM_MZ
   return res;
}

Img24 Img24::interpolate(Z nI, Z mI, Z method)const
   // See function int2
{
   Z mL=3;
   Word loc("Img24::interpolate(Z,Z,Z)");
   CPM_MA
   Z m=h();
   Z n=w();
   if (mI>m/2 && nI>n/2){ // traditional case, interpolation to
      // similar or larger resolution
      cpmmessage(mL,loc&" no averaging requested");
      return Img24::intPolAux(nI,mI,method);
   }
   else{ // new case: interpolating to lower resolution
         // based on averaging; works well for characters
         // in fonts
      R mIinv=1./mI, nIinv=1./nI;
      Z mf=cpmrnd(R(m)*mIinv), nf=cpmrnd(R(n)*nIinv);
      Z mA=mI*mf, nA=nI*nf;
      Img24 resA=Img24::intPolAux(nA,mA,method);
         // image of similiar resolution than the original
         // but allowing schematic averaging
      Img24 res(nI,mI);
      Z i,j,iR,jR,k,l;
      R nmInv=1./(mf*nf);
      for (i=1;i<=mI;++i){
         iR=mf*(i-1);
         for (j=1;j<=nI;++j){
            R3 sum;
            jR=(j-1)*nf;
            for (k=0;k<mf;++k){
               for (l=0;l<nf;++l){
                  ColRef val=resA[iR+k][jR+l];
                  sum.x1+=val.r;
                  sum.x2+=val.g;
                  sum.x3+=val.b;
               }
            }
            sum*=nmInv;
            Z s1=cpmrnd(sum.x1);
            Z s2=cpmrnd(sum.x2);
            Z s3=cpmrnd(sum.x3);
            (res.bh_)[i-1][j-1]=ColRef(s1,s2,s3);
         }
      }
      cpmmessage(mL,loc&" averaging done");
      CPM_MZ
      return res;
   }
}

Img24 Img24::pair(Img24 const& img, Z gap)const
{
   Z mL=3;
   Word loc("Img24::pair(Img24)");
   CPM_MA
   Z hn=h();
   cpmassert(hn==img.h(),loc);
   Z wn=w()+img.w()+gap;
   Img24 res(wn,hn);
   xy p1(0,0);
   xy p2(w()+gap,0);
   res.insert(p1,*this);
   res.insert(p2,img);
   CPM_MZ
   return res;
}

} // namespace
