//? cpmmini.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmmini.h>
#include <cpmperm.h>

using CpmAlgorithms::Mini;
using CpmRoot::R;
using CpmRoot::Z;
using CpmRoot::Word;
using CpmFunctions::F;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmArrays::Perm;

// initialization of static members

Z Mini::iterMaxBrent=200;
Z Mini::iterMaxPowell=200; // was 50

R Mini::GOLD=1.618034;
R Mini::GLIMIT=100; // is 100 in Press et al.
R Mini::TINY=1.0e-20;
R Mini::EPS=1.0e-15;  // machine floating point precision was 1e-16
// before
         // 97-2-20
R Mini::TOL=cpmsqrt(EPS)*5.;  // factor was 2 before 97-2-20
R Mini::CGOLD=0.3819660;
R Mini::ZEPS=1.0e-10;


namespace{

   // functions helping the implementation

R MAX(R a, R b){return a > b ? a : b;}
R SIGN(R a, R b){return b > 0. ? cpmabs(a):-cpmabs(a);}
void SHFT(R& a, R& b, R& c, R& d){a=b;b=c;c=d;}
R SQR(R a){return a*a;}

} // namespace

bool Mini::testBrak(R& ax, R& bx, R& cx, R& fa, R& fb, R& fc)
{
   static const Z countmax=3;
   Z counter=0;
Start:
   Z good=((ax<=bx && bx<=cx)||(cx<=bx && bx<=ax)) && (fb<=fa && fb<=fc);
   if (good){
      return true;
   }
   else{
      // trying to fix the problem by reordering
     // if (verbose>0){
      if (verbose>0){
         cpmwarning("unexpected reordering needed in Mini testBrak()");
      }
      counter++;
      R_Vector xVal(3);
      xVal[1]=ax;
      xVal[2]=bx;
      xVal[3]=cx;
      Perm p(xVal);
      xVal=xVal*p;
      R_Vector fVal(3);
      fVal[1]=fa;
      fVal[2]=fb;
      fVal[3]=fc;
      fVal=fVal*p;
      if (fVal[2]<=fVal[1] && fVal[2]<=fVal[3]){
         ax=xVal[1];
         bx=xVal[2];
         cx=xVal[3];
         fa=fVal[1];
         fb=fVal[2];
         fc=fVal[3];
         if (counter<=countmax){
            goto Start;   // testing wether were sucessfull
         }
         else{
            goto BadEnd;
         }
      }
      else{
         goto BadEnd;
      }
   }

BadEnd:
   {
      cpmcerr<<endl<<"bad return in testBrak"<<endl<<
      "ax="<<ax<<" bx="<<bx<<" cx="<<cx<<endl<<
      "fa="<<fa<<" fb="<<fb<<" fc="<<fc<<" counter="<<counter<<endl;
      cpmwarning("bad return in testBrak");
      return false;
   }
}


Mini::Mini(R fracTol):nEval_(0)
{
   done=0;
   iter=0;
   iterBrent=0;
   callsBrent=0;
   ftol=fracTol;
   ff_Defined=0;
}

Mini::Mini(const F<R_Vector,R>& f, R fracTol):ff_(f),ftol(fracTol),
   nEval_(0)
{
   done=0;
   iter=0;
   iterBrent=0;
   callsBrent=0;
   ff_Defined=1;
}

void Mini::setAcc(R fracTol)
{
   ftol=fracTol;
}

void Mini::setF(const F<R_Vector,R>& f)
{
   ff_=f;
   ff_Defined=1;
   nEval_=0;
}

R_Vector Mini::getPoint(const R_Vector& xStartInput)
{
   if (verbose) cpmmessage("Mini::getMin() started");
   Word loc("Mini::getMin()");
   cpmassert(ff_Defined==1,loc);
   xStart=xStartInput;
   xMin=xStart;
   iterBrent=0;
   iter=0;  // not needed due to the unconventional usage of iter in
            // Mini::powell
   yMin=minimize(xMin);
   done=1;
   if (verbose>=4) show();
   if (verbose) cpmmessage("Mini::getMin() done");
   return xMin;
}

R Mini::getValue(void)const
   { cpmassert(done==1,"Mini::getValue()"); return yMin;}

Z Mini::getIter(void)const
   { cpmassert(done==1,"getIter()"); return iter;}

Z Mini::getIterBrent(void)const
   { cpmassert(done==1,"getIterBrent()"); return iterBrent;}

bool Mini::brak(R& ax, R& bx, R& cx, R& fa, R& fb, R& fc)
/*
See Press et al. mnbrak(), p. 400
Given distinct initial points
ax and bx, this routine searches in the downhill direction
(defined by the function as evaluated at the initial points)
and returns new points ax,bx,cx which bracket a minimum of the
function. Also returned are the function values at the three
points , fa, fb, fc.
*/
{
   R ulim,u,r,q,fu,dum;

   fa=f1Powell(ax);
   fb=f1Powell(bx);
   if (fb > fa) {
      SHFT(dum,ax,bx,dum);
      SHFT(dum,fb,fa,dum);
   }
   cx=bx+GOLD*(bx-ax);
   fc=f1Powell(cx);
   while (fb > fc) {
      if (verbose>2){
         cpmvalue("fb",fb,2);
         cpmvalue("fc",fc,3);
      }
      r=(bx-ax)*(fb-fc);
      q=(bx-cx)*(fb-fa);
      u=(bx)-((bx-cx)*q-(bx-ax)*r)/
         (R(2)*SIGN(MAX(cpmabs(q-r),TINY),q-r));
      ulim=(bx)+GLIMIT*(cx-bx);
      if ((bx-u)*(u-cx) > 0.0) {
         fu=f1Powell(u);
         if (fu < fc) {
            ax=(bx);
            bx=u;
            fa=(fb);
            fb=fu;
            return true;
         } else if (fu > fb) {
            cx=u;
            fc=fu;
            return true;
         }
         u=cx+GOLD*(cx-bx);
         fu=f1Powell(u);
      } else if ((cx-u)*(u-ulim) > 0.0) {
         fu=f1Powell(u);
         R fu1=fu, cxx=cx+GOLD*(cx-bx);
         if (fu < fc) {
            SHFT(bx,cx,u,cxx);
            SHFT(fb,fc,fu,fu1);
         }
      } else if ((u-ulim)*(ulim-cx) >= 0.0) {
         u=ulim;
         fu=f1Powell(u);
      } else {
         u=(cx)+GOLD*(cx-bx);
         fu=f1Powell(u);
      }
      SHFT(ax,bx,cx,u);
      SHFT(fa,fb,fc,fu);
   }
   return testBrak(ax, bx, cx, fa, fb, fc);
}


R Mini::brent(R ax, R bx, R cx, R& xmin)
/*
Given a bracketing triplet of abscissas ax,bx,cx
(such that bx is between ax and cx, and f(bx) is less than both f(ax) and
f(cx), this routine isolates the minimum to a fractional precision of
about TOL using Brent's method. The abscissa of the minimum is returned
as xmin, and the minimum function value is returned as brent, the
returned function value.
*/
{
   Word loc("Mini::brent(R,R,R)");
   Z iterB;
   R a,b,d=0,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
   R term;
   callsBrent++;
   R e=0.0;
   a=((ax < cx) ? ax : cx);
   b=((ax > cx) ? ax : cx);
   x=w=v=bx;
   fw=fv=fx=f1Powell(x);
   for (iterB=1;iterB<=iterMaxBrent;iterB++) {
      if (verbose>2) cpmvalue("iterB",iterB,2);
      xm=(a+b)*0.5;
      tol1=TOL*cpmabs(x)+ZEPS;
      tol2=tol1*2.0;
      term=tol2-(b-a)*0.5-cpmabs(x-xm);
      if (verbose>2){
         if ( cpmabs(term+1.)<=1.e-4 ){
            cpmcerr<<endl<<"term=-1"<<endl<<"tol1="<<tol1<<" tol2="<<tol2
            <<" b="<<b<<" a="<<a<<" x="<<x<<" xm="<<xm;
         }
      }
      if (verbose>20){ cpmcerr<<endl<<"Mini:brent(), term="<<term;}
      if ( term>=0.) {
         xmin=x;
         iterBrent=CpmRootX::sup(iterBrent,iterB);
         return fx;
      }
      if (cpmabs(e) > tol1) {
         r=(x-w)*(fx-fv);
         q=(x-v)*(fx-fw);
         p=(x-v)*q-(x-w)*r;
         q=(q-r)*2.0;
         if (q > 0.0) p = -p;
         q=cpmabs(q);
         etemp=e;
         e=d;
         if (cpmabs(p) >= cpmabs(R(0.5)*q*etemp) || p <= q*(a-x) ||
            p>=q*(b-x))
            d=CGOLD*(e=(x >= xm ? a-x : b-x));
         else {
            cpmassert(q!=0.,loc);
            d=p/q;
            u=x+d;
            if (u-a < tol2 || b-u < tol2)
               d=SIGN(tol1,xm-x);
         }
      } else {
         d=CGOLD*(e=(x >= xm ? a-x : b-x));
      }
      u=(cpmabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
      fu=f1Powell(u);
      if (fu <= fx) {
         if (u >= x) a=x; else b=x;
         SHFT(v,w,x,u);
         SHFT(fv,fw,fx,fu);
      } else {
         if (u < x) a=u; else b=u;
         if (fu <= fw || w == x) {
            v=w;
            w=u;
            fv=fw;
            fw=fu;
         } else if (fu <= fv || v == x || v == w) {
            v=u;
            fv=fu;
         }
      }
   }
   if (verbose>1){
      cpmwarning("Mini::brent() stopped after"&cpmwrite(iterMaxBrent)&
      "iterations");
   }
   xmin=x;
   iterBrent=CpmRootX::sup(iterBrent,iterB);
   return fx;
}

/**************** function powell *************************************/

bool Mini::linMinPowell(R_Vector& p, R_Vector& xi, R& fret)
{
   R ax=0.1, xx=1., bx=2.;
   R xmin,fx,fb,fa;
   pPowell=p;
   xiPowell=xi;
   bool b=brak(ax,xx,bx,fa,fx,fb);
   if (b==false) return false;
   fret=brent(ax,xx,bx,xmin);
   xi *= xmin;
   p += xi;
   return true;
}

bool Mini::powell(R_Vector& x, R_Matrix xi, R& fret)
/* Powell's direction set method in multidimensions.
Minimization of a function func of n variables. Input consists of an
initial starting point p[1,...,n]; an initial matrix xi[1,...n][1,...n]
whose columns contain the initial set of directions (usually the n unit
vectors); and ftol, the fractional tolerance in the function value such
that failure to decrease by more than this ammount on one iteration
signals doneness. On output, p is set to the best point found, xi is
the then-current direction set, fret is the returned function value at p,
and iter is the number of iterations taken.
*/
{
   static Word loc("Mini::powell(...)");
   cpmassert(x.isVal(),loc);
   Z i,ibig,n=x.dim();
   R t,fptt,fp,del;
   R_Vector pt=x;
   R_Vector ptt, xit;
   fret=ff_(x);
   nEval_++;
   for (iter=1;;iter++) {
      if (verbose>2) cpmvalue("iter",iter,2);
      fp=fret;
      ibig=0;
      del=0.0;
      for (i=1;i<=n;i++) {
         xit=xi[i];
         fptt=fret;
         bool lmpRes=linMinPowell(x,xit,fret);
         if (lmpRes==false) return false;
         if (cpmabs(fptt-fret) > del) {
            del=cpmabs(fptt-fret);
            ibig=i;
         }
      }
      if (cpmabs(fp-fret)*2.0 <= ftol*(cpmabs(fp)+cpmabs(fret))) {
         return true;
      }

      if (iter == iterMaxPowell){
         Word messText="More than "&cpmwrite(iterMaxPowell)
         &" iterations in function Mini::powell()";
         cpmwarning(messText);
         return false;
      }
      ptt=x*2-pt;
      xit=x-pt;
      pt=x;
      fptt=ff_(ptt);
      nEval_++;
      if (fptt < fp) {
         t=R(2)*(fp-R(2)*fret+fptt)*SQR(fp-fret-del)-del*SQR(fp-fptt);
         if (t < 0.0) {
            linMinPowell(x,xit,fret);
            xi[ibig]=xit;
         }
      }
   }
}

R Mini::minimize(R_Vector& x)
/*
x is a vector which the function ff_ accepts as argument. After calling
function minimize it will be transformed to a vector for which f attains
a minimum value. If f has several minimum values, the one which is thus
obtained is the one which is got by Powells method.
*/
{
   Word loc("Mini::minimize(R_Vector& x)");
   Z mL=3;
   CPM_MA
   Z dx=x.dim();
   R_Matrix xi(dx); // dx*dx zero-matrix (was unit matrix
      // in the KAGUM version)
   xi=xi.net(1);  // unit matrix
   R res=0;
   bool b=powell(x,xi,yMin);
   if (b){
      res=yMin;
      suc_=B(true);
   }
   else{
      cpmwarning(loc&" no useful value found, 0 returned");
      // suc_=false by initialization
   }
   CPM_MZ
   return res;
}

void Mini::show()const
{
   cpmcerr<<endl<<"diagnosis of an instance of Mini";
   cpmcerr<<endl<<"verbose= "<<verbose;
   cpmcerr<<endl<<"iterMaxBrent= "<<iterMaxBrent;
   cpmcerr<<endl<<"iterMaxPowell= "<<iterMaxPowell;
   cpmcerr<<endl<<"GOLD= "<<GOLD;
   cpmcerr<<endl<<"GLIMIT= "<<GLIMIT;
   cpmcerr<<endl<<"TINY= "<<TINY;
   cpmcerr<<endl<<"EPS= "<<EPS;
   cpmcerr<<endl<<"TOL= "<<TOL;
   cpmcerr<<endl<<"CGOLD= "<<CGOLD;
   cpmcerr<<endl<<"ZEPS= "<<ZEPS;
   cpmcerr<<endl<<"ftol= "<<ftol;
   cpmcerr<<endl<<"yMin= "<<yMin;
   cpmcerr<<endl<<"nEval_= "<<nEval_;
   cpmcerr<<endl<<"iter= "<<iter;
   cpmcerr<<endl<<"iterBrent= "<<iterBrent;
   cpmcerr<<endl<<"callsBrent= "<<callsBrent;
   cpmcerr<<endl<<"done= "<<done;
   cpmcerr<<endl<<"ff_Defined= "<<ff_Defined;
   cpmcerr<<endl<<"xStart= "<<xStart;
   cpmcerr<<endl<<"xMin= "<<xMin;
}

