//? cpmr2func.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmr2func.h>
#include <cpmvo.h>

using namespace CpmArrays;
using namespace CpmFunctions;
#ifndef CPM_Fn
using std::bind;
using std::placeholders::_1;
#endif

// constructors from functions of two real variables

// auxiliary function objects

namespace{
class R2_FuncAux1: public FncObj<RR,R>{

protected:

   R (*fptr)(const R&, const R& );

public:
   R operator()(const RR& x)const{ return (*fptr)(x[1],x[2]);}

   R2_FuncAux1( R (*g)(const R&, const R&)):fptr(g){}
};

class R2_FuncAux2: public FncObj<RR,R>{

protected:

   R (*fptr)(R,R);

public:
   R operator()(const RR& x)const{ return (*fptr)(x[1],x[2]);}

   R2_FuncAux2( R (*g)(R,R)):fptr(g){}
};

class R2_FuncAux3: public FncObj<RR,R>{

   R_Func f1,f2;

public:
   R operator()(const RR& x)const{ return f1(x[1])*f2(x[2]);}

   R2_FuncAux3(const R_Func f, const R_Func& g):f1(f),f2(g){}
};

class R2_FuncAux4: public FncObj<RR,R>{

   R_Func f1;

public:
   R operator()(const RR& x)const{ return f1(x[1]);}

   R2_FuncAux4(const R_Func f):f1(f){}
};

}
// main thing

R2_Func::R2_Func( R (*f)(const R&, const R&) ):
Fa<RR,R>(new R2_FuncAux1(f)){}

R2_Func::R2_Func( R (*f)(R,R) ):
Fa<RR,R>(new R2_FuncAux2(f)){}

R2_Func::R2_Func(const R_Func& f1, const R_Func& f2):
Fa<RR,R>(new R2_FuncAux3(f1,f2)){}

R2_Func::R2_Func(const R_Func& f1):
Fa<RR,R>(new R2_FuncAux4(f1)){}


// constructors from matrices and tensors by interpolation, to be
// applied to images

namespace{ // anon

template <class T>
   // automatic conversion from T to R, and R to T has to be available
class FunctionFromMatrix: public CpmFunctions::FncObj<RR,R>{
   //
   VV<T> mat;
   Z m,n,method;
      // important to declare the dependent members m,n later than mat
   Iv ivx,ivy;
   bool periodic;
   static void dec(Z& i){ if (i>1) i--;}
   void inc_i(Z& i)const{ if (i<m) i++;}
   void inc_j(Z& j)const{ if (j<n) j++;}
public:
   R operator()(const RR& x)const;
      // defines values on the whole area. The unit square 0<=x[1]<=1,
      // 0<=x[2]<=1 corresponds directly to the matrix as pixel values.
      // On the rest of the plane the function is defined by periodic
      // repetition
   FunctionFromMatrix(const VV<T>& matI, Z methodI,
      const Iv& ivxI, const Iv& ivyI,bool per):mat(matI),
      m(mat.dim1()),n(mat.dim2()),method(methodI),
      ivx(ivxI),ivy(ivyI),periodic(per)
   {
      cpmassert( method==1 || method==0,"FunctionFromMatrix(...)");
   }
      // method=1: pixelation
      // method=2: bilinear interpolation
};

template <class T>
R FunctionFromMatrix<T>::operator()(const RR& xx)const
{
   Z mL=4;
   Word loc("FunctionFromMatrix<T>::operator()(RR)");
   cpmmessage(mL,loc&" started, exit not recorded since multiple");
   R x=xx[1];
   R y=xx[2];
   if (!periodic){
      if (!ivx.contains(x) || !ivy.contains(y)) return 0;
   }
   x=ivx.periodic(x);
   y=ivy.periodic(y);
   x*=n;
   y*=m;
      // 0<=x<n, 0<=y<m,
   Z xd=cpmtoz(x);
      // varies from 0 to n-1
   Z yd=cpmtoz(y);
      // varies from 0 to m-1
   Z i=m-yd;   // varies from m to 1
   Z j=xd+1;   // varies from 1 to n
   if (method==0){
       return mat[i][j];
   }
   else if (method==1){
      // unfortunately every case seems to need special consideration:
      // see my pages UM/CTL 98-3-6 in the Folder UM 98, CTL, KAGUM
      x-=xd;
      y-=yd;
      R p1=0,p2=0,f00=0,f01=0,f10=0,f11=0;
      Z iNext=i;
      Z jNext=j;
         // these will be changed by reference by dec and incUM
         // this will happen only once
      if (x==0.5){
         if (y==0.5) return mat[i][j] ;

         if (y>0.5) {
            f00=mat[i][j];
            dec(iNext);
            f01=mat[iNext][j];
            p2=y-0.5;
         }
         else{ // y<0.5
            inc_i(iNext);
            f00=mat[iNext][j];;
            f01=mat[i][j];
            p2=y+0.5;
         }
         return f00*(R(1)-p2)+f01*p2;
      }

      if (x>0.5){
         if (y>0.5){  // OK
            dec(iNext);
            inc_j(jNext);
            f00=mat[i][j];
            f10=mat[i][jNext];
            f01=mat[iNext][j];
            f11=mat[iNext][jNext];
            p1=x-0.5;
            p2=y-0.5;
         }
         else if (y<0.5){   // 0K
            inc_i(iNext);
            inc_j(jNext);
            f00=mat[iNext][j];
            f10=mat[iNext][jNext];
            f01=mat[i][j];
            f11=mat[i][jNext];
            p1=x-0.5;
            p2=y+0.5;
         }
         else if (y==0){
            inc_j(jNext);
            f00=mat[i][j];
            f10=mat[i][jNext];
            p1=x-0.5;
            return f00*(R(1)-p1)+f10*p1;
         }
      }
      else{ // then x<0.5, since == was carried to return earlier
         if (y>0.5){  // OK
            dec(iNext);
            dec(jNext);
            f00=mat[i][jNext];
            f10=mat[i][j];
            f01=mat[iNext][jNext];
            f11=mat[iNext][j];
            p1=x+0.5;
            p2=y-0.5;
         }
         else if (y<0.5){
            inc_i(iNext);
            dec(jNext);
            f00=mat[iNext][jNext];
            f10=mat[iNext][j];
            f01=mat[i][jNext];
            f11=mat[i][j];
            p1=x+0.5;
            p2=y+0.5;
         }
         else if (y==0){
            inc_j(jNext);
            f00=mat[i][j];
            f10=mat[i][jNext];
            p1=x-0.5;
            return f00*(R(1)-p1)+f10*p1;
         }
      }
      R b=f10-f00;
      R c=f01-f00;
      R d=f11-f00-b-c;
      return f00+b*p1+(c+d*p1)*p2;
         // f(x,y):=f00+b*x+c*y+d*x*y has to obey
         // f(0,0)=f00, f(0,1)=f01, f(1,0)=f10, f(1,1)=f11
         // indexes and arguments refer to the (x,y)-system for which
         // lower left corner is (0,0) and right upper corner is (1,1)
   }
   else{
      cpmerror(
         "FunctionFromMatrix<T>::operator(): unvalid parameter method");
      return 0;
   }
}

template <class T>
   // automatic conversion from T to R, and R to T has to be available
class FunctionFromTensor: public FncObj< RR, V<R> >{
   //
protected:
   VVV<T> ten;
   Z l,m,n;
   Iv ivx, ivy;
   Z method;
public:
   V<R> operator()(const RR& x)const;
      // defines values on the whole area. The unit square 0<=x[1]<=1,
      // 0<=x[2]<=1 corresponds directly to the matrix as pixel values.
      // On the rest of the plane the function is defined by periodic
      // repetition
      FunctionFromTensor(const VVV<T>& tenI,Z methodI,
         const Iv& ivxI, const Iv& ivyI):
         ten(tenI),l(ten.dim1()),m(ten.dim2()), n(ten.dim3()),
         ivx(ivxI), ivy(ivyI), method(methodI){}
      // method=1: pixelation
      // method=2: bilinear interpolation
};

template <class T>
V<R> FunctionFromTensor<T>::operator()(const RR& x)const
{
   V<R> res(l);
   for (Z h=1;h<=l;h++) res[h]=FunctionFromMatrix<L>
      (ten[h],method,ivx,ivy,true)(x);
   return res;
}

} // anon

R2_Func::R2_Func(const VV<L>& mat, Z method,
   const Iv& ivx, const Iv& ivy):
   Fa<RR,R>(new FunctionFromMatrix<L>(mat,method,ivx,ivy,true)){}


R2_Func::R2_Func(const VV<R>& mat, Z method,
   const Iv& ivx, const Iv& ivy, bool per):
Fa<RR,R>(new FunctionFromMatrix<R>(mat,method,ivx,ivy,per)){}

F<RR,V<R> > CpmFunctions::toFRRVR(const VVV<L>& ten, Z method,
   const Iv& ivx, const Iv& ivy)
{
   return
   F<RR,V<R> >(new FunctionFromTensor<L>(ten,method,ivx,ivy));
}

V<R2_Func> CpmFunctions::toVR2_Func(const VVV<L>& ten, Z method,
   const Iv& ivx, const Iv& ivy)
{
   Z mL=3;
   Word loc("CpmFunctions::toVR2_Func(VVV<L>& ...)");
   cpmmessage(mL,loc&" started");
   Z l=ten.dim1();
   V<R2_Func> res(l);
   for (Z h=1;h<=l;h++){
      res[h]=R2_Func(ten[h],method,ivx,ivy);
   }
   cpmmessage(mL,loc&" done");
   return res;
}

namespace{

R twoDimensiolalInterpolation(R x, R y, const V<R>& ySeq,
    const V<R_Func>& fSeq, const IpolControl& c)
// Write ySeq as y1,...yn. Then we consider the functions fSeq obtained
// from a two-variable function f by
// fi(x,y) = f(x,yi). (Restriction of f on lines which are parallel to
// the x-axis). The present function tries to reconstruct f as
// f(x,y) = twoDimensiolalInterpolation(x, y, (y1,...yn), (f1,...fn)).
// The parameters m and method specify the one-dimensional interpolation
// method used in the implementation.
{
    Z n=ySeq.dim();
    V<R> fVal(n);
    for (Z i=1;i<=n;i++) fVal[i]=(fSeq[i])(x);
    return R_Func(ySeq,fVal,c)(y);
        // interpolation in y-direction
}

R twoDimIntFunc(const RR& xy, const V<R>& ySeq, const V<R_Matrix>& fVal,
    const IpolControl& c)
// Transforming value arrays into function arrays.

{
    R x=xy[1];
    R y=xy[2];
    Z i,n=ySeq.dim();
    V<R_Func> fSeq(n);
    R_Vector xv,yv;
    for (i=1;i<=n;i++){
        xv=fVal[i][1];
        yv=fVal[i][2];
        fSeq[i]=R_Func(xv,yv,c);
            // function list from data list
    }
    return twoDimensiolalInterpolation(x, y, ySeq, fSeq, c);
}

R  indRectangle(const RR& z, const R& xL, const R& xU, const R& yL,
   const R& yU )
// the return value is the distance from z to the boundary of the
// rectangle positive if the point is outside and negative if it is
// inside.
{
   RR center((xL+xU)*0.5,(yL+yU)*0.5);
   R a=(xU-xL)*0.5;
   R b=(yU-yL)*0.5;
   cpmassert(a>0 && b>0,"indRectangle(...)");
   RR zRel=z-center;
   R x=zRel[1];
   R y=zRel[2];
// We go through the various cases with imediate return values. This
// avoids nested conditions
   if (x>=a){
      if (y>b) return (R)(zRel-RR(a,b)).abs();
         // then the corner RR(a,b) is the boundary point nearest to z
      if (y<-b) return (R)(zRel-RR(a,-b)).abs();
         // then the corner RR(a,-b) is the boundary point nearest to z
      return x-a;
         // then the line x=a is the boundary segment nearest to z
   }
   if (x<=-a){
      if (y>b) return (R)(zRel-RR(-a,b)).abs();
      if (y<-b) return (R)(zRel-RR(-a,-b)).abs();
      return -x-a;
   }
   // now -a<x<a
   if (y>=b) return y-b;
   if (y<=-b) return -y-b;
   // now -b<y<b
// Now (x,y) fulfills -a<x<a, -b<y<b. It is inside the rectangle
// all distances of (x,y) from the walls of the rectangle (all positive)
   Vo<R> dxy(4,0);
   dxy[1]=a-x;
   dxy[2]=x+a;
   dxy[3]=b-y;
   dxy[4]=y+b;
   dxy.order_();
// The distance to the wall is the smallest of these values and returned
// with a minus sign since we are inside the area
   return -dxy[1];
}

R indCircle(const RR& z, const R& xC, const R& yC, const R& r )
{
   return (R)((z-RR(xC,yC)).abs()-r);
}


} // namespace

R2_Func CpmFunctions::indfuncRectangle(const RR& cornerLL,
   const RR& cornerUR)
{
   R xL=cornerLL[1];
   R xU=cornerUR[1];
   R yL=cornerLL[2];
   R yU=cornerUR[2];
#ifdef CPM_Fn
   return F4<RR,R,R,R,R,R>(xL,xU,yL,yU)(indRectangle);
#else
   return F<RR,R>(bind(indRectangle,_1,xL,xU,yL,yU));
#endif
}


R2_Func CpmFunctions::indfuncCircle(const RR& center, R radius)
{
   R xC=center[1];
   R yC=center[2];
#ifdef CPM_Fn
   return F3<RR,R,R,R,R>(xC,yC,radius)(indCircle);
#else
   return F<RR,R>(bind(indCircle,_1,xC,yC,radius));
#endif

}

R2_Func CpmFunctions::twoDimInt(const V<R>& ySeq,
   const V<R_Matrix>& fVal, const IpolControl& control)
{
    Z i,d,n=ySeq.dim();
    cpmassert(n==fVal.dim(),"twoDimInt(...)");
    for (i=1;i<=n;i++){
        d=(fVal[i]).dim1();
        if (d!=2) cpmerror(
        "twoDimensiolalInterpolation(): dimension mismatch in value\
 list");
    }
#ifdef CPM_Fn
   return F3<RR,V<R>,V<R_Matrix>,IpolControl,R>
       (ySeq,fVal,control)(twoDimIntFunc);
#else
   return F<RR,R>(bind(twoDimIntFunc,_1,ySeq,fVal,control));
#endif
}
