//? cpmroot.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmroot.h>

using CpmAlgorithms::FindRoot;
using CpmRoot::R;
using CpmRoot::Z;
using CpmFunctions::F;

FindRoot::FindRoot(const F<R,R>& f, R x1, R x2, R eps):ff(f)
{
   setLimits(x1,x2);
   setAcc(eps);
   ffDefined=1;
}

FindRoot::FindRoot(R x1, R x2, R eps)
{
   setLimits(x1,x2);
   setAcc(eps);
   ffDefined=0;
}

void FindRoot::setLimits(R x1, R x2)
{
   if (x1<x2){ xL=x1; xU=x2;}
   else { xL=x2; xU=x1;}
}

void FindRoot::setAcc(R eps){ tol=eps;}

void FindRoot::setF(const F<R,R>& f)
{
   ff=f;
   ffDefined=1;
}

R FindRoot::getRoot()
{
   if (ffDefined!=1) cpmerror("FindRoot::getRoot(): ffDefined!=1");
   solve();
   done=1;
   return x0;
}

R FindRoot::getRootCrit(Z& success)
{
   if (ffDefined!=1) cpmerror("FindRoot::getRootCrit(): ffDefined!=1");
   solveCrit(success);
   done=1;
   return x0;
}

R FindRoot::getValue()const
{
   if (done!=1) cpmerror("FindRoot::getValue(): done!=1");;
   return y0;
}

namespace{

R SIGN(R a, R b){ return  b >= 0.0 ? cpmabs(a) : -cpmabs(a) ;}

}

Z FindRoot::brac(R* x1, R* x2)
//modified from Press et al. Numerical Recipes in C
{
   const R FACTOR=1.6;
   const Z NTRY=50;
   Z j;
   R f1,f2;
   if (*x1 == *x2) cpmerror("FindRoot::brac(...): *x1 == *x2");
   f1=ff(*x1);
   f2=ff(*x2);
   for (j=1;j<=NTRY;j++) {
      if (f1*f2 < 0.0) return 1;
      if (cpmabs(f1) < cpmabs(f2))
         f1=ff(*x1 += FACTOR*(*x1-*x2));
      else
         f2=ff(*x2 += FACTOR*(*x2-*x1));
   }
   return 0;
}

R FindRoot::brent(R x1, R x2)
//modified from Press et al. Numerical Recipes in C
{
   const Z ITMAX=100;
   const R EPS=1.e-12; // also influences accuracy, 3.e-8 in original
   Z iter;
   R a=x1,b=x2,c=x2,d=0,e{0.},min1,min2;
   R fa=ff(a),fb=ff(b),fc,p,q,r,s,tol1,xm;

   if((fa > 0. && fb > 0.) || (fa < 0. && fb < 0.))
      cpmerror(
       "R FindRoot::brent(R x1, R x2): x1 and x2 do not bracket a root");
   fc=fb;
   for (iter=1;iter<=ITMAX;iter++) {
      if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
         c=a;
         fc=fa;
         e=d=b-a;
      }
      if (cpmabs(fc) < cpmabs(fb)) {
         a=b;
         b=c;
         c=a;
         fa=fb;
         fb=fc;
         fc=fa;
      }
      tol1=EPS*cpmabs(b)*2+tol*0.5;
      xm=(c-b)*0.5;
      if (cpmabs(xm) <= tol1 || fb == 0.0) return b;
      if (cpmabs(e) >= tol1 && cpmabs(fa) > cpmabs(fb)) {
         s=fb/fa;
         if (a == c) {
            p=xm*s*2;
            q=R(1)-s;
         } else {
            q=fa/fc;
            r=fb/fc;
            p=s*(xm*q*(q-r)*2-(b-a)*(r-1));
            q=(q-1)*(r-1)*(s-1);
         }
         if (p > 0.0) q = -q;
         p=cpmabs(p);
         min1=xm*q-cpmabs(tol1*q)*3;
         min2=cpmabs(e*q);
         if (p*2 < (min1 < min2 ? min1 : min2)) {
            e=d;
            d=p/q;
         } else {
            d=xm;
            e=d;
         }
      } else {
         d=xm;
         e=d;
      }
      a=b;
      fa=fb;
      if (cpmabs(d) > tol1)
         b += d;
      else
         b += SIGN(tol1,xm);
      fb=ff(b);
   }
   return 0.0;
}

void FindRoot::solve(void)
{
   Z i=brac(&xL,&xU);
   if (i==0) cpmerror("FindRoot::solve(): brac() failed");
   x0=brent(xL,xU);
   y0=ff(x0);
}

void FindRoot::solveCrit(Z& success)
{
   static Z warningCounter=0;
   const Z warningMax=10;
   success=brac(&xL,&xU);
   if (success){
      x0=brent(xL,xU);
      y0=ff(x0);
   }
   else{
      x0=0;
      y0=0;
      if (warningCounter<warningMax){
         cpmwarning("FindRoot: no root found, garbage returned");
         warningCounter++;
      }
      else if (warningCounter==warningMax){
         cpmwarning(
            "FindRoot: ... garbage returned, messages dicontinued");
         warningCounter++;
      }
      else{
         ;
      }
   }
}
