//? cpmsmoothlocalinterpol.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmsmoothlocalinterpol.h>

using CpmRootX::parity;

namespace CpmArrays{

void IpolControl::adjust()
{
   if (m<=0) cpmerror("IpolControl(...) m<=0");
   if (m>=20) cpmerror("IpolControl(...) m>=20");
   if (method<0) cpmerror("IpolControl(...) method<0");
   if (method>3) cpmerror("IpolControl(...) method>3");
      // limit 20 is ad hoc but large values of
      // m make no sense
}

IpolControl::IpolControl(const Word& w):method(0),monoTest(1),eps(1e-6)
 // w=polygon,first_order,second_order,third_order,...sixth_order
{

   if (w=="polygon"){
      m=1; method=3;
   }
   else if (w=="first_order"){
      m=2;
   }
   else if (w=="second_order"){
      m=3;
   }
   else if (w=="third_order"){
      m=4;
   }
   else if (w=="fourth_order"){
      m=5;
   }
   else if (w=="fifth_order"){
      m=6;
   }
   else if (w=="sixth_order"){
      m=7;
   }
   else{
      cpmerror("IpolControl(const Word& w): unvalid value for w");
   }
}

bool IpolControl::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("IpolControl",str);
   cpmp(m);
   cpmp(method);
   cpmp(monoTest);
   cpmp(eps);
   return true;
}

bool IpolControl::scanFrom(istream& str)
{
   cpms(m);
   cpms(method);
   cpms(monoTest);
   cpms(eps);
   return true;
}

namespace{  // anonymus

const R HUGE_UM=1.e25;
const R TINY=R(1)/HUGE_UM;

/*
Z parity(Z k)
{
   if (k%2) return -1; else return 1;
}
*/

void polint(V<R> const& xa, V<R> const& ya, R x, R& y, R& dy)
{
   Z n=xa.dim();
   cpmassert(n==ya.dim(),"polint(...)");
   Z i,m,ns=1;
   R den,dif,dift,ho,hp,w;
   dif=cpmabs(x-xa[1]);
   V<R> c(n);
   V<R> d(n);
   for (i=1;i<=n;i++) {
      if ( (dift=cpmabs(x-xa[i])) < dif) {
         ns=i;
         dif=dift;
      }
      c[i]=ya[i];
      d[i]=ya[i];
   }
   y=ya[ns--];
   for (m=1;m<n;m++) {
      for (i=1;i<=n-m;i++) {
         ho=xa[i]-x;
         hp=xa[i+m]-x;
         w=c[i+1]-d[i];
         if ( (den=ho-hp) == 0.0) cpmerror("Error in routine POLINT");
         den=w/den;
         d[i]=hp*den;
         c[i]=ho*den;
      }
      y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
   }
}

void ratint(V<R> const& xa, V<R> const& ya, R x, R& y, R& dy)
{
   Z n=xa.dim();
   cpmassert(n==ya.dim(),"ratint(...)");
   Z m,i,ns=1;
   R w,t,hh,h,dd;
   V<R> c(n);
   V<R> d(n);
   hh=cpmabs(x-xa[1]);
   for (i=1;i<=n;i++) {
      h=cpmabs(x-xa[i]);
      if (h == 0.0) {
         y=ya[i];
         dy=0.0;
         return;
      } else if (h < hh) {
         ns=i;
         hh=h;
      }
      c[i]=ya[i];
      d[i]=ya[i]+TINY;
   }
   y=ya[ns--];
   for (m=1;m<n;m++) {
      for (i=1;i<=n-m;i++) {
         w=c[i+1]-d[i];
         h=xa[i+m]-x;
         t=(xa[i]-x)*d[i]/h;
         dd=t-c[i+1];
         if (dd == 0.0){
            y=HUGE_UM;
            dy=HUGE_UM;
             cpmwarning("Problem in routine RATINT");
            return;
         // this is changed 91-11-12 for improved flexibility
         }
         dd=w/dd;
         d[i]=c[i+1]*dd;
         c[i]=t*dd;
      }
      y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
   }
   return;
}

/********************** my functions ******************************/

R s_log(R x)
// s_log stands for save log: argument is allows to be negative
{
   static R x_min=1e-50;
   static R y_min=cpmlog(x_min);
   if (x<x_min) return y_min;
   return cpmlog(x);
}

R s_exp(R x)
// s_exp stands for save exponential function, argument may be
// arbitrarily
// large
{

   static R y_max=1e50;
   static R x_max=cpmlog(y_max);
   if (x>x_max) return y_max;
   return cpmexp(x);
}

R hill(R x, R p)
// For any positive parameter p, the function x |---> hill(x,p)
// is a smooth (i.e. derivatives of any order exist) non-increasing
// function. It has the constant value 1 for x<=0 and the constant value
// 0 for x>=1.  So, it is a smooth down-hill connecting level 1 with
// level 0.
// The whole curve is radially symmetrical to the point (x,y)=(0.5,0.5).
// p=1 gives a nice shape. p=0.5 seems to give smaller absolut
// interpolation error but finer 'ripple' of the error curve.
// The idea behind is the following: The negative hyperbolic tangent
// function is already similar to such a hill. A transformation of the
// argument brings the x-values 0 and 1 to infinity before they are
// inserted into the hyperbolic tangent. The normal expression for the
// latter is transformed in order to call exp only once. Futher, to avoid
// overflow riscs in exp, this function is called only for negative
// arguments.p=0.810747 minimizes the L2-Norm of the second derivative.
{
   if (x<=0) return(1.);
   if (x>=1) return (0.);
   R y=p*(x-0.5)*2/(x*(x-1));
   if (y<=0) {
      y=cpmexp(y); y=y/(y+1);
   }
   else {
      y=cpmexp(-y); y=R(1)/(y+1);
   }
   // so the argument of exp is always negativ
   // hence no overflow errors can result by calling exp.
   // Since the values of exp are never negative, 1+y is never
   // zero.
   return(y);
}

/*

R hill(R x, R p)
// For any positive parameter p, the function x |---> hill(x,p)
// is a smooth (i.e. derivatives of any order exist) non-increasing
// function. It has the constant value 1 for x<=0 and the constant value
// 0 for x>=1.  So, it is a smooth down-hill connecting level 1 with
// level 0.
// The whole curve is radially symmetrical to the point (x,y)=(0.5,0.5).
// p=1 gives a nice shape.
// experimental simple replacement
{
   if (x<=0) return(1.);
   if (x>=1) return (0.);
   R x_=(1-x);
   R x2_=x_*x_;
   return x2_/(x2_+x*x);
}
*/

Z lonei(V<R> const& x, R x_){ return x.locAsc(x_);}

void range(R k, Z n, Z m, Z& l, Z& r, Z method)
/* k is integer or half-integer=m of my paper, m is P of my paper
l is L(m,P), r is R(m,P)
*/
{

   Z ir=cpmtoz(k+0.5*m);      // primarily (Z)(k+0.5*(m-1)+0.5)
   Z il=ir-m+1;

   if (method==2){ // variable order when approaching boundary
      Z flag=0;
      if (il<0) {il=0; flag=1;}
      if (ir>=n) {ir=n-1; flag=2;}
      if (parity(ir-il+1)!=parity(m)){
         if (flag==1) ir--;   // 93-1-19 changed ir++ in ir--
         if (flag==2) il++;   // 93-1-19 changed ir-- in ir++
            // this preferrs using less points near the boundary
            // compared with using more points; seems to be better for
            // robust extrapolation properties
      }                         // insure that ir-il+1 remains of
                          // the same parity as m    91-11-8
   }
   else {
      if (il<=0){ il=0; ir=m-1;}
      if (ir>=(n-1)){ ir=n-1; il=n-m;}       // constant order
      // since variable order is not useful for rational interpolation
      // (since especially used for extrapolation)
   }
   if (ir < il) ir=il;  // no real reason, should not happen anyway
   r=ir; l=il;
}

R pu(V<R> x, Z n, R x_, R k, R w_0, R w_1)
/* pu stands for partition of unity. x is the list of tabulated points.
x_ is the x-value for which the function pu is to be evaluated (value=
return value). k is the index, integer or half-integer, which indicates
what member of the partition of unity is considered. w_0, w_1 are weights
which modify the original pu in order to account for the expected error
of the local interpolation functions. This will be used if local
interpolation functions are rational and either due to poles or to
particularities of the Bulirsch Stoer algorithm not always give
contributions of equal reliability. If both weights are set equal,
no modification of the originally designed pu results. w_0=w_1=0. gives
fastest execution.
*/
{
   static R hillPar=0.810747;     // seems to be better than P=1
   x.b_(0); // x was Vl<R>
   Z kc,ku,kl;
   R xc,xn,y,y_;
   kc=cpmtoz(k+0.1);
   if( k==(R)kc ){ // then k is integer
      xc=x[kc];
      if (x_>xc){
         kc++;
         if ( kc>=n )  return 1;
      }
      else {
         kc--;
         if ( kc<0) return 1;
      }
      xn=x[kc];
   }
   else { // then in our application k is half-integer
      ku=cpmtoz(k+0.6); kl=ku-1;
      if ( ku>=n || kl<0) return 1;
      xc=(x[ku]+x[kl])*0.5;
      if (x_>xc){
         ku++; kl++;
         if (ku>=n) return 1;
      }
      else {
         ku--; kl--;
         if (kl<0 || ku<0 ) return 1;
      }
      xn=(x[ku]+x[kl])*0.5;
   }
   y=(x_-xc)/(xn-xc);
   if (w_0==0. && w_1==0.)  goto re;
   y_=y*w_1+(R(1)-y)*w_0;
   if (y_==0.) goto re;
   y*=(w_1/y_);
re:   return hill(y,hillPar);
}

R lip(V<R> x, V<R> y, Z n, Z m, R x_, R k, R& acc, Z method)
/* lip stands for local interpolation. It returns the value of the local
interpolating function, indexed by k (integer or half-integer) at x_.
x,y are the global interpolation data of length n. m of them are used
for local interpolation (according to 'method' this number may change
near the boundary). acc is a measure for the 'expected error'in the
interpolation, see Press, Flannery,..p.90....
method is a parameter which determines whether polynomial or rational
interpolation is to be used. Further, it determines the way how
interpolation near the boundary is handled.
*/
{
   x.b_(0); y.b_(0);
   R y_;
   Z j,l,r,n_;
   range(k,n,m,l,r,method);
   n_=1+r-l;
   if (n_==1) return y[l]; // this is the trivial case of constant
      // interpolation
   V<R> a(n_);
   V<R> b(n_);
   Z i_c=1;
   for (j=l; j<=r; j++){
      a[i_c]=x[j]; b[i_c]=y[j]; i_c++;
   }
   if (method == 1){
      ratint(a,b,x_,y_,acc);
   }
   else {
      polint(a,b,x_,y_,acc);
   }
   return y_;
}

R lip_pos(V<R> x, V<R> y, Z n, Z m, R x_, R k, R& acc, Z method)
/* lip_pos is in every respect identical to lip with the only exception
that it interpolates the logarithms of the function (by the
method selected by parameter method) and gives back the exp of the
logarithms. Since s_log is used for the logarithm, the values are allowed
to be negative, but in this case the result is not meaningful.
*/
{
   x.b_(0); y.b_(0);
   R y_;
   Z j,l,r,n_;
   range(k,n,m,l,r,method);
   n_=1+r-l;
   if (n_==1) return y[l]; // this is the trivial case of constant
      // interpolation
   V<R> a(n_);
   V<R> b(n_);
   Z i_c=1;
   for (j=l; j<=r; j++){
      a[i_c]=x[j]; b[i_c]=s_log(y[j]); i_c++;
   }
   if (method == 1){
      ratint(a,b,x_,y_,acc);
   }
   else {
      polint(a,b,x_,y_,acc);
   }
   return s_exp(y_);
}

} // anonymus

void unique(V<R>& x, V<R>& y, R eps=1e-6)
// tests whether the list of x's is stricly increasing as expected for
// interpolation purposes. If two x values are very close together they
// get merged and the y-value replaced by the mean

{
   Z mL=4;
   static Word loc("unique(Vl<R>& x, Vl<R>& y)");
   cpmmessage(mL,loc&" started");
   x.b_(0); y.b_(0);
   Z nx=x.dim();
   Z ny=y.dim();
   cpmassert(nx==ny,loc);
   Z i;
   R lastx=x[0];
   R lasty=y[0];
   V<B> select(nx,B{true});
   select.b_(0);
   for (i=1;i<nx;i++){
      R valx=x[i];
      R valy=y[i];
      R diffx=CpmRoot::dis(valx,lastx);
      R diffy=CpmRoot::dis(valy,lasty);
      if (diffx<eps){
         select[i]=false;
         x[i-1]=(valx+lastx)*0.5;
         y[i-1]=(valy+lasty)*0.5;
         if (diffx>0 || diffy>0){
            cpmwarning(loc&" points merged: "&
               " diffx="&cpmwrite(diffx)&" diffy="&cpmwrite(diffy));
         }
      }
      else if(valx<lastx){
         cpmerror(loc&" valx<lastx holds for valx="&cpmwrite(valx)&
            " and lastx="&cpmwrite(lastx));
      }
      lastx=valx;
      lasty=valy;
   }
   x=x.select(select);
   y=y.select(select);
   cpmmessage(mL,loc&" done");
}

void prepIpol(V<R>& x, V<R>& y, R eps)
{
   unique(x,y,eps);
}

R slint(R x_, V<R> const& xa, V<R> const& ya, Z m, Z method)
// sli stands for smooth local interpolation
// x is the array of tabular x-values, indexed 0,1,...n-1, strictly
// increasing y is the array of tabular y-values, indexed 0,1,...n-1
// x_ is the x-value for which the interpolated value ( return value of
// the function) has to be calculated. n =1,2,3,   is the number of
// tabular points. m is the 'order of the method' (m=1,2,..., m<=n).
// (apart from working near the boundary it is the number of points
// used in interpolation). method is 0 or 2 for polynomial interpolation
// and 1 for Bulirsch Stoer rational diagonal interpolation.
// The number of points used in interpolation is m if x_ is
// more than m/2 points away from the boundary. If method=2, things are
// arranged such, that coming closer to the boundary
// reduces the number of points which are used. This is only useful
// polynomial interpolation, there better extrapolation properties are
// achieved. For universal use there is included normal (not smooth)
// linear interpolation as method=3.
{
   const R epsilon=2.e-2;        // determines from what absolut
   // inaccuracy, reported from ratint on we begin to distrust the
   // result of rational interpolation.
   // should be an argument of the function, but only used if method=1

   V<R> x=xa; x.b_(0);
   V<R> y=ya; y.b_(0);
   Z n=x.dim();

   if (n!=y.dim()){
      cpmerror("slint(): x.dim() != y.dim()");
      return 0;
   }
   if (n==0){
      cpmwarning("slint(): x.dim() = 0");
      return 0;
   }
   if (n==1) return y[0];

   if (m<1) m=1;
   if (m>n) m=n;
   R w,ip_l,ip_r,acc_l,acc_r,k_;
   Z k=lonei(x,x_);
   if (method==3){
      if (k>n-2) {
         k=n-2;
      }
      else if (k==-1){
         k=0;
      }
      else{
         ;
      }
      R p=(x_-x[k])/(x[k+1]-x[k]);
      return y[k]+(y[k+1]-y[k])*p;
   }
   if ( parity(m)==-1 ){ // then m is odd
      if (k==-1) return lip(x,y,n,m,x_,0.,acc_l,method);
      if (k==n) return lip(x,y,n,m,x_,(R)(n-1),acc_l,method);
      k_=(R)k;
   }
   else { // then m is even
      if (x_<=(x[0]+x[1])*0.5)
         return lip(x,y,n,m,x_,0.5,acc_l,method);
      if (x_>=(x[n-2]+x[n-1])*0.5)
         return lip(x,y,n,m,x_,n-1.5,acc_l,method);
      // "=" in if (...)   introduced 91-11-8
      R center=(x[k]+x[k+1])*0.5;
      if (x_<= center){
         k_=k-0.5;
      }
      else {
         k_=k+0.5;
      }
   }

   ip_l=lip(x,y,n,m,x_,k_, acc_l, method);
   ip_r=lip(x,y,n,m,x_,k_+1,acc_r,method);
   if (method==1){ // to reduce the influence of poles
      R ip_i=y[k]+(y[k+1]-y[k])*(x_-x[k])/(x[k+1]-x[k]);
         // this is a linearly interpolated value
      R w_l=pu(x,n,x_,k_,0.,0.);
      R w_r=R(1)-w_l;
      R err_r=cpmabs(acc_r)/epsilon;
      R err_l=cpmabs(acc_l)/epsilon;
//      R err_i=fabs(ip_l-ip_r)/epsilon;
//      w_l+=err_l; w_r+=err_r;
      R w_i=(err_l+err_r)*0.5; w_i*=w_i; // seems to work well
      R norm=R(1)/(w_l+w_i+w_r);
      w_l*=norm; w_i*=norm; w_r*=norm;
      return (w_l*ip_l+w_i*ip_i+w_r*ip_r);
   }
      // If the expected error (acc as accuracy is misleading, but will
      // not be changed) for the left function is large, this increases
      // the weight ('credibility') of the right function. The epsilon
      // mechanism is introduced such that only considerable expected
      // errors influence the weights of left and right functions.
      // Some power, or other function of the a_l, a_r might be
      // useful to form.
   w=pu(x,n,x_,k_,1.0,1.0);
   return (w*ip_l-ip_r*(w-1));
}

R slint_pos(R x_, const V<R>& xa, const V<R>& ya, Z m, Z method)
/*
sli_pos is in every respect identical to sli with the only exception
that it interpolates the logarithms of the function (by the method
selected by parameter method) and gives back the exp of the logarithms.
IMPORTANT: Since s_log is used for the logarithm, the values are allowed
to be zero or negative, but in this case the result is  n o t
meaningful.
Also, the only partially sucessful provisions against the influence of
poles in case of method==1 are eliminated.
*/

{
   V<R> x=xa; x.b_(0);
   V<R> y=ya; y.b_(0);
   Z n=x.dim();
   if (n!=y.dim()){
      cpmerror("slint(): x.dim() != y.dim()");
      return 0;
   }
   if (n==0){
      cpmwarning("slint(): x.dim() = 0");
      return 0;
   }
   if (n==1){
      R res=y[0];
      if (res<0) cpmerror(
         "slint_pos(...): y.dim()=1, y[0]="&cpmwrite(res)&" <0."
         );
      return res;
   }
   R w,ip_l,ip_r,acc_l,acc_r,k_;
   Z k=lonei(x,x_);
    if (m<1) m=1;
    if (m>n) m=n;
   if (method==3){
      if (k>n-2) {
         k=n-2;
      }
      else if (k==-1){
         k=0;
      }
      R y_=s_log(y[k]), y__=s_log(y[k+1]);
      return s_exp((y_+(y__-y_)*(x_-x[k])/(x[k+1]-x[k])));
   }
   if ( parity(m)==-1 ){ // then m is odd
      if (k==-1) return lip_pos(x,y,n,m,x_,0.,acc_l,method);
      if (k==n) return lip_pos(x,y,n,m,x_,(R)(n-1),acc_l,method);
      k_=(R)k;
   }
   else { // then m is even
      if (x_<=(x[0]+x[1])*0.5)
          return lip_pos(x,y,n,m,x_,0.5,acc_l,method);
      if (x_>=(x[n-2]+x[n-1])*0.5)
         return lip_pos(x,y,n,m,x_,n-1.5,acc_l,method);
      // "=" in if (...)   introduced 91-11-8
      R center=(x[k]+x[k+1])*0.5;
      if (x_<= center){
         k_=k-0.5;
      }
      else {
         k_=k+0.5;
      }
   }
   ip_l=lip_pos(x,y,n,m,x_,k_, acc_l, method);
   ip_r=lip_pos(x,y,n,m,x_,k_+1,acc_r,method);
   w=pu(x,n,x_,k_,1.0,1.0);
   return (w*ip_l-ip_r*(w-1));
}

// 'modern' form

R locIpol(R x_, V<R> const& x, V<R> const& y,
   IpolControl const& c)
{
   return slint(x_,x,y,c.get_m(),c.get_method());
}

R locIpolPos(R x_, V<R> const& x, V<R> const& y,
   IpolControl const& c)
{
   return slint_pos(x_,x,y,c.get_m(),c.get_method());
}

R locIpol(R ic, V<R> const& yList, IpolControl const& c)
{
   Z i,d=yList.dim();
   V<R> xList(d);
   for (i=1;i<=d;++i) xList[i]=i;
   return locIpol(ic,xList,yList,c);
}

} // namespace
