//? cpmcamera2.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
     Description: class Camera achieves visualization of 3-dimensional
      objects via anaglyph-technique. Here we have the less
      sophisticated 2D version of it.
*/

namespace CpmCamera2{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::Word;
//   using CpmRoot::Pi;

   using CpmGeo::Angle;
   using CpmGeo::Iv; // Interval

   using CpmRootX::B;

   using CpmLinAlg::R2;
   using CpmLinAlg::R3;
   using CpmLinAlg::RR2;
   using CpmLinAlg::RR3;

   using CpmArrays::T2;
   using CpmArrays::X2;

   using CpmFunctions::F;
#ifdef CPM_Fn
   using CpmFunctions::F1;
#endif
   using CpmArrays::V;
   using CpmGraphics::VecGraph;
   using CpmGraphics::VecGraphs;

   using CpmImaging::Response;

using namespace CpmColCam;
using namespace CpmDim2;

////////////////////////// class Camera /////////////////////////////////

   class Camera{ // sees a rectangular part of infinite 2D object space
      // Since we are in 2D, this is not a stereo camera.

// independent data

      Color color1;
         // color for image corresponding to image 1 (left eye)

      Color color3;
         // color for painting the background. It may be better to have
         // this not completely black in order to reduce visibility of
         // reflexes.

      mutable Graph gr;
         // viewing window in the film plane. Defines the position and
         // size of the 'film gate'.

      Spc poi;
         // 'point of interest'= center of field of view in object space
         // all construction methods of Camera make this point the origin
         // of Spc().
         // geometrical transformations enable this to change after
         // construction.

      mutable Response rsp;
         // a quantity which controls the conversion from R to
         // pixel brightness control values

 // dependent data

      Vec v;
         // unit vector in the vertical direction in the image plane

      Vec h;
         // unit vector in the horizontal direction in the image plane

     void complete(void);
         // assumes that the indpendent data have their valid values
         // and initializes the dependent data accordingly

     B stereo_;
         // Will be used only in CpmCamera3::Camera. So there is no
         // problem with the initialization as false

   public:
      typedef Camera Type;
      CPM_IO

      Word nameOf()const{ return Word("CpmCamera2::Camera");}

      Camera(R imageHeight=1, R aspectRatio=1.4,
         const Angle& viewAngle=Angle(50), R frac=1.,
         const Angle& conv=Angle(5), const Frame& fr=Frame());
         // Arguments viewAngle, frac, and conv have no influence.
         // Point of interest (poi) is the origin of Spc().
         // Camera position in space different from that can be
         // achieved by operator *=(const Group&) or by functions
         // shift and roll.

      Camera(Graph const& g);
         // Here the object space is given in a very flexible way
         // by g

      C captureMono(const Spc& p)const
         {Vec d=p-poi; return C((d|h),(d|v));}
         // basic imaging function, inline for efficiency

      T2< X2<C,R> > capture(const Spc& p,
         const R& r)const;
         // returns the captureMono-result twice.
         // Only for uniformity of interface.

      X2<C,R> captureMono(const Spc& p, const R& r)const;
         // same as the previous function, but activating only the 1-part
         // of the stereo camera

      T2< C > capture(const Spc& p)const;
         // Same a previous function, but ignoring the radius and thus
         // returning two graphical positions only

      Spc inverse(const C& x, const R& mag )const
      { mag; return poi+h*x[1]+v*x[2];}
         // returns the position p in space of a point which on the first
         // stereo-image gets the 'film-plane-coordinates' x and for
         // which an small extended lateral object near p of linear size
         // s gets imaged as one linear size mag*s on the film.
         // abstract definition:
         // captureMono(inverse(x,mag),r)=X2<R2,R>(x,r*mag)  for all R2
         // x and R r, mag

      AxVec getAxis()const{ return AxVec(1.);}
         // returns a unit-vector in viewing direction as a data
         // type which makes sense in 2D and 3D

      void fillCircle(const C& center, R radius,
         const Color& color )const
         {gr.fillCircle(R2(center[1],center[2]), radius, color);}
         // shows a circular disk

      Camera& operator *=(const Group& g);
         // moving the whole camera

      void shift(const R& sx, const R& sy, const R& sz=0.);
         // translation of the whole camera. sx is the shift in direction
         // of h (horizontal image direction)
         // and sy is the shift in direction v (vertical image direction)
         // sz is the shift in direction of e , the direction of view

// in the following 3 operations the 'point of interest' is held fixed
// The terminology for the next three functions follows the book
// Isaac Victor Kerlow: The Art of 3-D Computer Animation and Imaging,
// Van Nostrand Reinhold, 1996
      void tilt(const Angle& a);
         // acts trivially in 2D
      void pan(const Angle& a);
         // acts trivially in 2D
      void roll(const Angle& a);
         // positive a lets the 'horizon' (horizontal line before roll)
         // appear ascending from the left to the right
         // (positive rotation arround e)

      void adjustFieldOfView( R enlarge, R sx=0, R sy=0, R sz=0);
      // enlarge is a factor by which the viewing window gets enlarged
      // compared to the one which is presently valid. (sx,sy) describes
      // a shift of the viewing window such that (1,1) shifts the lower
      // left corner of the enlarged window to the position previously
      // held by the upper-right corner.
      // sz describes a shift in direction of the mean viewing direction
      // in units of the image diagonal. The latter parameter, as
      // enlarge, is capable of changing the visual size of objects,
      // using sz changes the perspective too.

      void shiftGate(R sx=0, R sy=0){gr.shift(sx,sy);}
         // shifts the 'filmgate' in a manner that is clear from th code

      Color getColor1()const{ return color1;}
      Color getColor2()const{ return Color(BLACK);}
         // useful to have uniform interface for 2D and 3D
      Color getColor3()const{ return color3;}

      bool is3D()const{ return false;}
      bool is2D()const{ return true;}
      void setBW_(){ color1=BLACK; color3=WHITE;}

      void mark(const RefSys& rs, Z i)const;
         // marks the positive coordinate i-axis by a chain of points

      #include <cpmcamerax.h>
         // declarations which do not depend on dimension
    };

} // namespace
