//? cpmcamerax.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
Description: function declarations common for 2D and 3D
*/

//////////////// class Camera //////////////////////////////////////////

   Camera(RecordHandler const& rch,
      Word const& sec, CpmGraphics::Frame& fr);
      // constructor from RecordHandler

   void vis(bool lineArtProcessing=false)const
      //: visualize
   {
      Z mL=3;
      Word loc("Camera::vis(B)");
      CPM_MA
      bool file= gr.getWriteOnShow();
      gr.vis(file,lineArtProcessing);
      CPM_MZ
   }

   void clear()const{gr.paint(color3);}
         // paints the whole viewing rectangle of gr with a suitable
         // background color

   void clr()const{clear();}
      //: clear

   void setWriteOnShow(bool wos)const{ gr.setWriteOnShow(wos);}
      // setting gr.writeOnShow. If gr.writeOnShow==true
      // a call to gr.show(...) implies a call to gr.img.write()

   bool getWriteOnShow()const{ return gr.getWriteOnShow();}
      // getting gr.getWriteOnShow()

   void setColBac(Color c){color3=c;}

   void mono_()
      //: mono
      // mutating function
   {
      if (is3D()){
         stereo_=B(false);
         setBW_();
      }
   }

   bool isMono()const{ return !stereo_;}
   bool isStereo()const{ return stereo_;}

   // getting the object space data of gr

   R getxLeft()const{ return gr.getxLeft();}
   R getxRight()const{ return gr.getxRight();}
   R getyLower()const{ return gr.getyLower();}
   R getyUpper()const{ return gr.getyUpper();}
   R getdX()const{ return gr.getdX();}
   R getdY()const{ return gr.getdY();}
   Z dim1()const{ return gr.dim1();}
   Z dim2()const{ return gr.dim2();}

   void write(const Word& filename)const{ gr.write(filename);}

   void write(void)const{ gr.write();}

   void rot(const AxVec& axv);
      // rotation of the whole camera such that the 'point of interest'
      // is held fixed

   V<Vec> getVectors()const;
      // returns the vectors h,v,-e (last only in 3D) for getting info on
      // the orientation of the camera

   Vec get_h()const{ return h;}
   Vec get_v()const{ return v;}

   Vec getDir()const{ return -getVectors()[3];}
      // main indicator for the direction of the camera
      // returns a unit vector

   Spc getPoint()const{ return poi;}
      // returns the 'point of interest'

   Spc get_x()const{ return poi;}
      // for uniformity with class Shape

   RefSys getRefSys()const;
      // returns the reference system, which naturally belongs to the
      // camera:
      // axes are h,v,-e (we need a positivly oriented system, so the
      // third axis is against the direction of vision.
      // So, the 1 and 2 axes are the normal x and y axes of the
      // film plane considered as a drawing board

   void scale(R enlarge);
      // multiplies 'film size' and s by factor enlarge and adjusts all
      // other quantities accordingly. The viewing angle thus remains
      // unchanged.

   bool isVisible(const Spc& x)const;
      // returns true if the point x can be represented within the
      // graphics window of gr

   void setScreenArea( const Graph& g)const;
      // This function redefines gr such that it gets its display area on
      // screen from g, without changing the the objectspace-meaning
      // of gr. Thus, if g and gr do not match in aspect ratio, we get
      // distortion.

   Graph& getGraph()const{ return gr;}
      // retrieving the Graph object gr; return value was Graph
      // (instead of Graph&) till 2003-07-31 which had bad influence
      // in function ScalarField::mark()

   void captureGraph()const{/* if (!gr.getNewColors()) gr.capture();*/}

   Response& getResponse()const{return rsp;}
      // 2003-07-31 same return strategy as in the previous function
      // implemented

   Iv getRangeX()const{ return rsp.getRangeX();}

   void mark(const Spc& x, R r)const;
         // shows the sphere with center x and radius r in anaglyphic
         // manner (as seen by camera *this)

   void mark(const Spc& x, R r, const Color& c)const;
         // shows the sphere with center x and radius r in Color c
         // manner (as seen by camera *this)

   void mark(const Spc& x1, const Spc& x2, R rFoot=0.)const;
         // shows a line from x1 to x2 in anaglyphic manner (as seen by
         // camera *this). If rFoot>0, the position of x1 will be marked
         // by the image of a sphere of a radius rFoot. This allows to
         // represent vectors as oriented line pieces.

   bool mark(const Spc& x, const Color& c)const;
         // 1. c.stereo()==true
         // shows the point x in  in anaglyphic representation else.
         // Returns false if the mean of the points as seen from the
         // two stereo directions point falls outside the graphics
         // window defined by gr.
         // 2. c.visible()==true.
         // shows the point in color c and returns false if the point
         // as seen from the 'mono-direction' e falls outside the
         // graphics window defined by gr.

   bool mark(const Spc& x)const{ return mark(x,Color(INVISIBLE));}
         // shows the point x in anaglyphic manner (as seen by camera
         // *this)

   void mark(const Spc& x1, const Spc& x2, const Color& color,
         R rFoot=0.)const;
         // shows a line from x1 to x2 in color. If color is invisible,
         // anaglyphic representation is done.
         // If rFoot>0, the position of x1 will be marked by the
         // image of a sphere of a radius rFoot where rFoot refers to
         // the metric in the camera's Graph-object.

   void mark(V<Spc> const& vs, Color const& color)const
         // showing a polygon (list of points)
   {
      for (Z i=1;i<vs.dim();i++) mark(vs[i],vs[i+1],color);
   }

   void mark(Path const& p, Color const& c)const{ mark(p(),c);}

   void mark( V<X2<Spc,Vec> > const& svl, R fac, Color const& color)const
   {
      Z i,svld=svl.dim();
      for (i=1;i<=svld;++i){
         X2<Spc,Vec> xi=svl[i];
         Spc pi=xi.c1();
         Vec vi=xi.c2();
         Spc pf=pi+fac*vi;
         mark(pi,pf,color,0);
      }
   }

   void mark(const SubsetOfSpace&, SitRec i, const Color& c )const;
      // visualisation of extended bodies.

   // The function calculates a list of points on the surface of the
   // body and marks the camera-image of these points as a polygon.
   // This list of surface points lies in the 'film plane' of the camera.
   // It is created by a 'virtual curtain of laser test rays' emanating
   // perpendicularly from site i of the filmplane and being directed
   // inwards.
   // These parallel test rays have a distance ~ds. Where they hit
   // the surface of the SubsetOfSpace, there are the points mentioned
   // above.
   // The sites denoted i correspond to the rims of the monitor
   // viewing window. LOW: lower rim, RIG: right, UPP: upper, LEFT left.
   // dr gives the distance error allowed in localizing the surface of
   // the body. If the section of the contour with the film rectangle
   // is a line that may be considered a function graph over one of the
   // sites, then marking with this side as second argument gives a
   // perfect representation of this contour. For more complicated
   // contours marking from more than one side (e.g. from all sites k)
   // may be necessary for a representation without gaps. See the
   // redefinition of this function in class Shell

    void markCen(const SubsetOfSpace&, Z nP, const Color& c )const;
      // mark central
      // Visualisation of extended bodies. Here the 'virtual curtain
      // of laser test rays' emanates from the center of the camera's
      // film gate. It fills the full circle with nP points.
      // Could become generalized to a fan of rays emanating from
      // an adjustable point.

    void markFan(const SubsetOfSpace& sos, Spc const& p, Vec const& e1,
       AxVec const& n, R ang, Z nP, Color const& c)const;
      // mark (by) fan
      // visualisation of extended bodies. Here the 'virtual curtain
      // of laser test rays' emanates from the point p filling the
      // directions starting with e and consisting of nP directions
      // which originate from e by rotation arround axis ax (only
      // the direction of ax counts). Bewteen two adjacent directions
      // the angle is ang/nP. For nP and ang sufficienly large,
      // the 'fan' may cover an 'angle' much larger than one complete
      // turn. All exceptional arguments are treated reasonably and
      // create warnings (no errors).

   void show(const SubsetOfSpace&, const Texture& tx=Texture(),
      bool bodyFix=false)const;
      // areal visualisation of extended bodies. If the last argument
      // is true, the texture tx is not considered fixed in space but
      // moving with the SubsetOfSpace.

   void mark(SubsetOfSpace const&, Texture const& tx=Texture(),
      bool bodyFix=false)const;
      // areal visualisation of extended bodies. If the last argument
      // is true, the texture tx is not considered fixed in space but
      // moving with the SubsetOfSpace.

   void dirTo_(Spc const& x, R const& r, R fitFactor=1.2,
      Angle tilt=Angle(0), Angle pan=Angle(0),
      RefSys const& rs=RefSys());
      // The camera gets changed such that the sphere determined by x and
      // r just fits the gap between lower and upper boundary of a
      // 'virtual' display frame, the sites of which are those of
      // the actual display frame divided by fitFactor. Thus, if
      // fitFactor=2 the sphere S(x,r) will fill only one half of
      // the display window's height.The viewing angle and the parameter
      // frac of the camera remain unchanged. The angular orientation in
      // space of the camera after function call is given by angles tilt
      // and pan with respect to reference system rs. Calling this
      // function does not cause the object *this to show the object sos,
      // it only adjusts geometrical data of the camera.

   void dirTo_(SubsetOfSpace const& sos, R fitFactor, Angle tilt,
      Angle pan, RefSys const& rs=RefSys())
      { dirTo_(sos.get_x(),sos.get_r(),fitFactor,tilt,pan,rs);}
      // The camera gets changed such that the objects's bounding sphere
      // just fits the gap between lower and upper boundary of an
      // 'virtual' the display frame, the sites of which are those of
      // the actual display frame divided by fitFactor. Thus, if
      // fitFactor=2 the bounding sphere of sos will fill only one half of
      // the display window's height.The viewing angle and the parameter
      // frac of the camera remain unchanged. The angular orientation in
      // space of the camera after function call is given by angles tilt
      // and pan with respect to reference system rs. Calling this
      // function does not cause the object *this to show the object sos,
      // it only adjusts geometrical data of the camera.

   void lookTo(const SubsetOfSpace& sos, Angle tilt,
      Angle pan, const Texture& tx=Texture(),
      const RefSys& rs=RefSys(),bool bodyFix=false, R fitFactor=1);
      // This function adjusts the camera as described in the previous
      // function and also shows object sos textured by tx on the
      // display.
      // This was the first form of such a function and the previous
      // function was created by separating the code in geometrical
      // adjustment and imaging. For compatibility with old programs
      // the argument structure was not changed with exception of the
      // insertion of the last argument, which can't break old code.

   void mark(const Skeleton& sk, const Color& c)const;
      // visualisation of extended bodies represented as Skeletons.

   void mark(const WireFrame& wf, const Color& c)const;
      // visualisation of extended bodies represented as
      // arrays of Skeletons.

   void mark(Triangle const& tri, Color const& c)const;

   void mark(const VV<Spc>& vvs, const Color& c)const
      { mark(Skeleton(vvs,true),c);}

   void mark(const RefSys& rs, const Color& c, R lengthOfAxes)const;

   VecGraph captMono(const Skeleton& sk)const;
      // generating an instance of class VecGraph from a Skeleton.
      // This is the main tool to graphical representations
      // of WireFrames

   V<C> captMono(const V<Spc>& vs)const;
      // mono-image of a list of points

   T2< V<C> > capt(const V<Spc>& vs);
      // stereo-image of a list of points

   T2<VecGraph> capt(const Skeleton& sk)const;
      // generating an instance of class VecGraphs from a
      // Skeleton.

   VecGraphs captMono(const WireFrame& wf)const;
      // generating an instance of class VecGraphs from a
      // WireFrame.

   T2<VecGraphs> capt(const WireFrame& wf)const;
      // stereo version of the previous function

   void addText(const Word& text, R x=0.05, R y=0.15,
      Color c=CpmImaging::WRITECOLOR, R h=0)const
   { gr.addText(text,x,y,c,h);}

   V<Spc> grid(Z ny=8)const;
      // greates a family of space points that fill the
      // image window as a 2D cubic lattice of a width determined
      // by the fact that there are by lines per image height
      // These points are give as a one-dimensional array  since
      // the present mode of using the data cooperates with this format.
