//? cpmdim2.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
    Description: see cpmdim3.h. The present file handles the
      2-dimensional case:
*/

//doc namespace CpmDim2{

   class Group;
   class RefSys;
   class Vec;
   class Plane;

 //////////////////// class AxVec ////////////////////////////////////////

   class AxVec{
// axial vectors, rotation descriptors, 1-dimensional in this case
      // space of 'axial vectors' i.e. rotations in
      // 2-dimensional space and their infinitesimal generators.
      // Interestingly, this is the autonomous class
      // (i.e. needs no support from other classes)
      // Vector space structure over R is in this case not inherited from
      // base class.
      // If we interprete an axial vector e.g. as an angular velocity the
      // the absolute value of the AxVec is dphi/dt. If it is a finite
      // rotation, the length is phi.
      // The finite rotation belonging to *this with data element x1
      // is the one (x,y)|--> (x',y') such that
      // x'=cos(x1)*x-sin(x1)*y
      // y'=sin(x1)*x+cos(x1)*y
      // see function actionOfRotationVectors(...) in cpmdim2cpp.h

      friend class Group;
      friend class Vec;

      R x3;
         // a data element of this name would result from
         // class AxVec:public R1{ ...};. Since meanwhile I defined R1 in
         // a way that it allows automatic conversion to R and misses
         // many functions which R2 and R3 define, I can't use R1 any
         // longer in this way.

   public:
      typedef AxVec Type;
      CPM_IO
      CPM_ORDER
      Word nameOf()const{ return Word("CpmDim2::AxVec");}

      static const Z dimension;
         // in our case 1

      AxVec():x3(0){};
         // vector with all components equal to zero

      AxVec(const R& a, const RefSys& rs);

      explicit AxVec(R r):x3(r){}
         // constructor from components

       AxVec(Angle const& alpha, Z i=3):x3(i==3 ? alpha.toR() : R(0)){}
         // constructor from rotation angle and master frame axis
         //
         // Rotation arround 3-axis with angle phi. The skrew sense of
         // the rotation is the standart one: Looking against the axis
         // vector, a positive phi indicates counter-clock-wise rotation.
         // Notice that there are also negative Angles a
         // (a.idPos()==false)
         // These thus describe clock-wise rotations. This implies that
         // a rotation that is close to 360 degrees, all
         // counter-clock-wise is a kind of being not describable by an
         // AxVec. If Argument i is different from 3 we return the
         // same element as for phi=0 and i==3. This is most likely
         // to make programs written with 3D in mind also valid for 2D.
         //
         // Note on problems encountered:
         // If there is a 0 instead of 0., both the MS compiler and
         // gnu-compiler compile without warning but in running I found x3
         // initialized as 1 in a case where alpha=Angle(90,DEG).
         // Reason is unclear. Consequence: avoid expressions that rely
         // on type conversions in a constructor initialization list.
         // The second pair of brackets, also helps and than allows to
         // write  0 instead of 0. ! curious!
         // On the other hand: to be able to do some basic computation
         // inside initializers is extremely useful.

      AxVec(Angle const& phi, Vec const& axis);
         // same as previous function; the second argument has no
         // influence; it is there only for reasons of uniformity to the
         // 3D case. Defined later in the presnt file as inline

      AxVec(Angle const& phi, AxVec const& axis):x3(phi){axis;}
         // same as previous function; the second argument has no
         // influence.

      AxVec(R r1, R r2, R r3):x3(r3){}
         // Constructor from components respect to the master system of
         // reference. Only the third component has an inflence.
         // The function is introduced for uniformity with the 3D case.

      explicit AxVec(V<R> const& v):x3(v[3]){}
         // For uniformity with Vec

       AxVec(const V<Vec>& b1, const V<Vec>& b2, bool careful=true);
         // constructs the finite rotation which transforms the
         // orthonormal basis b1 into the orthonormal basis b2.
         // If the third parameter has value true, one actually works
         // with potentially modified versions of b1 and b2 which are
         // guaranteed to be orthonormal bases.

      R& operator[](Z i);
      R const& operator[](Z i)const;
// simple access
      R operator()(void)const{ return x3;}

// group structure, composition of rotations.
      friend AxVec operator *(const AxVec& a1, const AxVec& a2 );
         // see  Vec operator *(const Vec& x, const AxVec& d);
         // for a conceptual definition of this product.
      friend AxVec operator !(const AxVec& a){ return AxVec(-a.x3);}

      AxVec& operator *=(const AxVec& a){ return *this=(*this)*a;}

      AxVec& operator *=(Group const& g){ &g; return *this;}
         // AxVec is perpendicular to the plane which g transforms.
         // The action thus is trivial. (was wrong till 2006-03-28)

      AxVec& ref_(Vec const& n){ &n; return *this*=-1;}

      AxVec& ref_(Plane const& p);

      AxVec pow(R p)const;

// operations for infinitesimal rotations and angular velocities

   AxVec exp(const R& t, bool accurate=true)const;
         // the standard exponential function from the Lie algebra
         // (to which *this belongs here) and the Lie group (return
         // value).
         // If omega is a constant angular velocity, then omega.exp(t) is
         // the finite rotation which results from having this angular
         // velocity be effective over the time span t.
         // Since in 2D exp is the identity, there is no need for an
         // fast approximate representation. Thus the second argument
         // has no influence in this case (3D is different!)

   AxVec adjointAction(const AxVec& d)const;
         // returns *this since rotating the 1-dimensional AxVecs is
         // the trivial operation

// vector product
   AxVec cross(const AxVec& a)const{ a; return AxVec();}
      // cross product or vector product of axial vectors. Since all
      // AxVecs are parallel, all cross products vanish
   Vec cross(Vec const& v)const;

   Angle phi()const{ return Angle(x3,RAD);}
      // returns the rotation angle belonging to *this.

   AxVec& phi_(Angle const& phi)
   { if (x3>=0.) x3=phi.toR(); else x3=-phi.toR(); return *this;}
   //: phi
   // Changes *this into a rotation vector with a rotation angle
   // phi*sign(x3). So, in a sense, the function takes the sense of
   // rotation from *this and the rotation angle from the argument.
   // See the implementation of Path::quarterEllipse for a beneficial
   // usage.

   R nor_(){R res=cpmabs(x3); x3*=cpminv(res); return res;}
      //: normalize
   R normalize(){return nor_();} // old form

   AxVec ran(Z j=0)const;
      // abs of the result is <= abs of *this
   R dis(const AxVec& a)const;
// linear algebra
   AxVec& operator +=(const AxVec& a){ x3+=a.x3; return *this;}
   AxVec& operator -=(const AxVec& a){ x3-=a.x3; return *this;}
   AxVec& operator *=(const R& r){ x3*=r; return *this;}
   AxVec& neg(void){ x3=-x3; return *this; }
   Z dim()const{ return 1;}
   R abs()const{ return cpmabs(x3);}
   R absSqr()const{return x3*x3;}
   friend AxVec operator +(const AxVec& a, const AxVec& b)
      { return AxVec(a.x3+b.x3);}
   friend AxVec operator -(const AxVec& a, const AxVec& b)
      { return AxVec(a.x3-b.x3);}
   friend AxVec operator -(const AxVec& a){ return AxVec(-a.x3);}
   friend AxVec operator *(const AxVec& a, const R& r)
      { return AxVec(a.x3*r);}
   friend AxVec operator *(const R& r, const AxVec& a)
      { return AxVec(r*a.x3);}
   friend R operator |(const AxVec& a, const AxVec& b)
      { return  a.x3*b.x3;}
   X2<R,R> polDec()const;
      // polar decomposition: First component (res.first if res is the
      // return value) is the absolute value r and the second component
      // is sign(x3)

   X2<Angle,Vec> toAngAxs()const;
      // For uniformity with the corresponding function in cpmdim3.h
      // The returned vector is useless in this case and is chosen
      // as the zero vector
   bool isVal()const{ return CpmRoot::isVal(x3);}

   static V<AxVec> canBas()
         // canonical basis
      { return V<AxVec>{AxVec(1.)};}

   };

   RR2 toMat(const AxVec& d);
      // rotation matrix belonging to the axial vector. I.e.
      // x*toMat(d)=x*d, where on the left-hand side we have R2*RR2
      // (see cpmlinalg.h).

////////////////////// class Vec ////////////////////////////////////////

   class Vec { // space of translations and velocities
      friend class Group;
      Rdim x_;
      typedef Vec Type;
   public:
      static const Z dimension;
      CPM_IO
      CPM_ORDER
      // also function R normalize()
      Word nameOf()const{ return Word("CpmDim2::Vec");}
      Vec():x_(){}
      explicit Vec(const Rdim& a):x_(a){}
      Vec(const R2& a, const RefSys& rs);
      Vec(R r1, R r2, R r3=0.):x_(r1,r2){r3;}
         // to enable code unification for 2 and 3 dimensions
      explicit Vec(C const& a):x_(a[1],a[2]){}
      explicit Vec(V<R> const& v):x_(v[1],v[2]){}
      explicit Vec(AxVec const& ax):x_(){ax;}
         // no meaningful initialization from AxVec in 2D
      R2 toRdim()const{ return x_;}
      C toC()const{ return C(x_.x1,x_.x2);}
      friend Vec operator *(Vec const& x, AxVec const& d);
         // x*d is the image of x under the active transformation
         // represented by d. This is the 'master representation' which
         // defines the geometrical meaning of d. The group
         // multiplication in AxVec is defined just such that
         // x*(d1*d2) = (x*d1)*d2.
// These are the functions which I once got from Rdim, when Vec was
// derived fro Rdim.
      Vec& operator *=(R const& r){ x_*=r; return *this;}
      Vec& operator +=(Vec const& v){ x_+=v.x_; return *this;}
      Vec& operator -=(Vec const& v){ x_-=v.x_; return *this;}
      Vec operator *(R const& r)const{ return Vec(x_*r);}
      Vec operator +(Vec const& v)const{ return Vec(x_+v.x_);}
      Vec operator -(Vec const& v)const{ return Vec(x_-v.x_);}
      Vec operator -(void)const{ return Vec(-x_);}
      friend Vec operator *(R const& r, Vec const& v){ return v*r;}
      R operator |(Vec const& v)const{ return (x_|v.x_);}
      R nor_(R r=1){ return x_.nor_(r);}
    //  R normalize(){ return x_.nor_();} // old form
      R abs()const{ return x_.abs();}
      R absSqr()const{ return x_.absSqr();}
      Rdim operator()(void)const{ return x_;}
      R dis(Vec const& v)const{ return x_.dis(v.x_);}
      bool isVal()const{ return x_.isVal();}
      R& operator[](Z i){ return x_[i];}
      const R& operator[](Z i)const{ return x_[i]; }
      Z dim()const{ return x_.dim();}
      void neg(){ x_.neg();}
      X2<R,Vec> polDec()const{
         X2<R,Rdim> y=x_.polDec();
         return X2<R,Vec>(y.c1(),Vec(y.c2()));
      }

// other topics:

      Vec& operator *=(const AxVec& d){ return *this=(*this)*d;}

      AxVec cross(const Vec& a)const
         // vector product of vectors is an axial vector
      { return AxVec(x_.crsPrd(a.x_));}

      Vec cross(const AxVec& a)const
         // vector product of a vector with an axial vector is a vector
      { return Vec(x_.x2*a[3],-x_.x1*a[3]);}

      Vec crossExc(Vec const& v)const{ v;return Vec();}
         //: cross exceptional

      Vec rot90()const{ return Vec(-x_.x2,x_.x1);}
         // rotation 90 (degrees)
         // Returns a vector which results from *this by a rotation
         // of 90 degrees in positive direction. I came to consider this
         // an important operation by studying lecture notes of Professor
         // Engelbert Schlipf (Fachhochschule Ulm) on technical mechanics.
         // January 2008. 2015-01-12 overall sign of the result changed.
         // Only now the rotation is in the positive direction.

      R spat(const Vec& a, const Vec& b)const{a; b; return 0;}
         // for uniformity with 3D

      Vec& operator *=(const Group&);

      X2<Vec,Vec> tanNormDec(const Vec& b)const;
         // the first component of the return pair is the component of b
         // in the direction of *this, and the second component is the
         // normal component

      Vec ref(Vec const& n)const{Vec d=n*((*this|n)*2); return *this-d;}
         //: reflection
         // Returns a vector which is obtained from *this by a reflection
         // from a line which has the normal vector n, which is assumed
         // to be a unit vector.

      Vec& ref_(Vec const& n)
         {Vec d=n*((*this|n)*2); *this-=d; return *this;}
         //: reflection
         // Changes *this into the vector which is obtained from *this
         // by a reflection from a line which has the normal vector n,
         // which is assumed to be a unit vector.

      Vec& ref_(Plane const& p);
         //: reflection
         // see cpmdim3.h

      V<Vec> base(const Vec& b=Vec(0,1) )const;
         // returns a positively oriented pair of orthogonal unit vectors
         // e[1],e[2].
         // The argument b has no effect and is there only to have the
         // same interface as in the 3-dimensional case, where we need
         // two vectors for determining a ortho-normal base.
         // If *this is zero, we have e[1]=Vec(1,0), e[2]=Vec(0,1).
         // Else, e[1] is the unit vector (*this)/|*this| and
         // e[2]=e[1]*AxVec(Angle(90)), which is the uniquely determined
         // supplement, for forming a positively oriented ortho-normal
         // base.

      void setDir_(const Vec& v){*this=Vec(v.x_.polDec().second()*abs());}
         //: set direction
         // takes he direction from the argument and does not change the
         // absolute value

      AxVec rotTo(const Vec& v)const;
         // if the return value is interpreted as rotation, it is the
         // most direct one carrying *this into v.

      Angle angTo(const Vec& v)const{ return rotTo(v).phi();}
         // angle 'from *this to v'

      AxVec rotParOrAntiPar(const Vec& v);
         // name: rotate parallel or anti-parallel
         // The smallest rotation (i.e. the one with the absolute value
         // of the rotation angle) which makes *this parallel to the
         // the lines determined by {v,-v}

      Vec randomize(R dv, Angle const& a, Z j)const;
         // Returns a vector which is a stochastic
         // modification of a copy of *this. The length of the
         // result is uniformly distributed (as a function of j)
         // in the interval [max(0,|*this|-|dv|),|*this|+|dv|], and
         // the direction is obtained from the direction of *this
         // by rotating arround a random axis by an angle which is
         // uniformely distributed between -a and a.

      static Vec ranUnitVec(Z j=0);
         // random unit vector

      static V<Vec> canBas()
         // canonical basis
      { return V<Vec>{Vec(1,0),Vec(0,1)};}

      static V<Vec> homUnitVec(Z n, Z nSteps=10, Z stopPeriod=10,
         R dtFac=0.07, R eps=0.1);
         //: homogeneous unit vectors
         // generates a system of n unit vectors which are
         // first selected at random and then distributed more
         // homogeneously by applying nSteps steps of a
         // repulsive dynamical process. This is more important in
         // 3D than here, but for uniformity of methods its done
         // in 2D in the same manner. Implementation is identical
         // for 2D and 3D and is in cpmhomunitvecxpp.h.
     };

   inline AxVec::AxVec(const Angle& phi, const Vec& axis):x3(phi){axis;}
      // in order to avoid warnings from unreferenced formal parameters
      // one has to add the effect-less statement axis; this is not
      // allowed by the MS compiler to be placed directly in class AxVec,
      // since there Vec is not yet defined

//////////////////// class Group ////////////////////////////////////////
// See cpmdim3.h for explanation the concept and of most functions.

   class Spc;
   class InTens;

   class Group{
// group of active Euclidian transformations (translations and rotations)

      friend class AxVec;
      friend class Vec;
      friend class Spc;
      friend class InTens;
      friend Spc operator *(const Spc& x, const Group& g);

      AxVec A;
      Vec a;

   public:
      typedef Group Type;
      CPM_INVERSION
      CPM_PRODUCT_M
      CPM_IO
      CPM_ORDER
      Word nameOf()const{ return Word("CpmDim2::Group");}
      Group(){}
         // unit element
      Group(const Spc& x0, const AxVec& A1);
         // constructor from fixed point and axial vector for
         // rotation arround that point. See cpmdim3.h for more.

      Group( Spc const& x0, AxVec const& A1, Vec const& a1);
         // constructor from fixed point and axial vector for
         // rotation arround that point and a polar vector for
         // a final translation.

      Group(const Spc& x0, const Angle& phi);
         // see cpmdim3.h

      Group(const AxVec& A1, const Vec& a1):A(A1),a(a1){}

      explicit Group(const AxVec& A1):A(A1),a(){}

      explicit Group(Angle const& alpha):A(AxVec(alpha)),a(){}
         // constructor from angle; this is usefull in context where
         // Group is available and AxVector not (as in my ParDynSys
         // template)

      explicit Group(const Vec& a1):A(),a(a1){}

      Vec shift(void)const{ return a;}

      AxVec rot(void)const{ return A;}

      friend Vec operator * (const Vec& x, const Group& g)
      { return x*g.A;}

      static Rdim act(const Rdim& x, const Group& g)
         // transforming a point via the fundamental transformation
         // group. Here the point is represented by its coordinate-
         // tuple with respect to the master system of reference
      { return (Vec(x)*g.A+g.a)();}

      F<Rdim,Rdim> actRdim()const
      {
#ifdef CPM_Fn
         return F1<Rdim,Group,Rdim>(*this)(act);
#else
         return F<Rdim,Rdim>(bind(act,_1,*this));
#endif
      }

      R dis(const Group& g)const
         { return CpmRootX::sup(A.dis(g.A),a.dis(g.a));}

      Group ran(Z j=0)const{ return Group(A.ran(j));}
         // see cpmdim3.h for explanation

      Group ran(Spc const& x0, Z j=0)const
      { return Group(x0, A.ran(j));}
         // random rotations arround x0 with rotation angle up to that of
         // *this. Useful in context where AxVec (which defines the
         // corresponding random sequence) is not available
   };

   inline Group& Group::operator *=(const Group& g)
   {
      A*=g.A; (a*=g.A)+=g.a; return *this;
   }

   inline Group Group::operator !()const
   {
      return Group(!A,(-a)*(!A));
   }

////////////////////// class Spc ////////////////////////////////////////
// see cpmdim3.h for explanation of most functions

   class Spc{
// affine Euclidian space, instances are points and not vectors
      friend class Group;
      friend class RefSys;
      Vec p;
      explicit Spc(const Vec& p1):p(p1){}
      static R refLen;

   public:
      typedef Spc Type;
      CPM_IO
      CPM_ORDER
      Spc():p(){}
      Spc(R x1, R x2, R x3=0.):p(x1,x2){x3;}
         // x3 not active, for uniformity with cpmdim3.h

      explicit Spc(const Rdim& x):p(x[1],x[2]){}
      explicit Spc(const V<R>& x):p(x[1],x[2]){}

      Spc(const Rdim& x, const RefSys& rs);

      explicit Spc(const C& a):p(a[1],a[2]){}

      Spc(const V<R>& x, const RefSys& rs);
         // construction from components with respect
         // to the reference system rs, where the components
         // are represented by a dimension-independent type

      C toC()const{ return p.toC();}

      Rdim toRdim()const{ return p.toRdim();}

      bool isVal()const{ return p.isVal();}

      static Rdim toRdimStatic(const Spc& x)
         { return x.p.toRdim();}

      static F<Spc,Rdim> FSR;

      R& operator[](Z i){return p[i];}
      const R& operator[](Z i)const{return p[i];}

      Z dim()const{ return p.dim();}
      Word nameOf(void)const{ return "CpmDim2::Spc";}

      void setRefLen(R rl)
      { cpmassert(rl>0,"Spc::setRefLen()");refLen=rl;}
      R getRefLen(){ return refLen;}

      friend Spc operator * (const Spc& x, const Group& g)
      { return Spc(x.p*g.A+g.a);}

      Spc& operator *=(const Group& g)
         // action of Group as a mutating operation
      { p=p*g.A+g.a; return *this;}

      Spc& operator /=(const Group& g)
         // inverse action of Group as a mutating operation
         // for 'utmost efficiency' not written as *=!g
       { p-=g.a; p*=!g.A; return *this;}

      Spc& rot_(Spc const& x0, AxVec const& a)
         //: rotation
         // rotation by a, which leaves the point x0 fixed
      { Vec pr=p-x0.p; p=x0.p+pr*a; return *this;}

      Spc& rot(Spc const& x0, AxVec const& a){ return rot(x0,a);}
         // old name

      friend Spc operator + (const Spc& x, const Vec& a)
         // shifting position by adding a vector
       { return Spc(x.p+a);}

      friend Spc operator - (const Spc& x, const Vec& a)
         // shifting position by adding the vector -a
      { return Spc(x.p-a);}

      Spc& operator +=(const Vec& a)
         // adding a vector as a mutating operation
      { p+=a; return *this;}

      Spc& operator -=(const Vec& a)
         // subtracting a vector as a mutating operation
      { p-=a; return *this;}

      friend Vec operator - (const Spc& x, const Spc& y)
         // getting a vector from subtracting two points
      { return Vec(x.p-y.p);}

      Spc ref(Vec const& n, Spc const& s0)const
         //: reflection
         // Returns the image of *this under a reflection from
         // a plane which goes through point s0 and has the
         // normal unit vector n.
      { return s0+(*this-s0).ref(n);}

      Spc& ref_(Vec const& n, Spc const& s0)
         //: reflection
         // Changes *this into its image under a reflection from
         // a plane which goes through point s0 and has the
         // normal unit vector n.
      { return *this = ref(n,s0);}

      Spc& ref_(Plane const& p);
         //: reflection
         // Changes *this into its image under a reflection from
         // the plane p.

      Vec vecTo(const Spc& goal)const{ return Vec(goal.p-p);}
         // Vector which shifts *this into goal

      Spc scl(Spc const& fix, R fac)const{ return fix+(*this-fix)*fac;}
         //: scale
         // Returns the point resulting from *this by
         // multiplying the position vector relative to
         // the point 'fix' by factor fac.

      Spc& scl_(Spc const& fix, R fac){ return *this=scl(fix,fac);}

      Spc cen(Spc const& x)const{ return Spc((p+x.p)*0.5);}
         // center (midpoint) between *this and x

      static Spc convex(const V<Spc>& x, V<R>& w, bool careful=true);

      static X2<Spc,R> envSph(V<Spc> const& vs);
         //: enveloping sphere
         // center and radius of an envelopping sphere of an array
         // of points

      void convexAsMember(const V<Spc>& x, V<R>& w, bool careful=true)
      { *this=convex(x,w,careful);}
         // changes *this into the convex linear combination of the x

      Spc versus(const Spc& a, R frac)const
         { return Spc(p+(a.p-p)*frac);}
         // returns a point which lies on the line from *this to a.
         // frac==0 gives *this, frac==1 gives a;  for values
         // outside [0,1] the meaning is an obvious generalization of
         // this

      R dis(const Spc& a)const;

      R distance(const Spc& a)const // p1.distance(p2):=|p1-p2|
      {  return (p-a.p).abs();}

      R distSquare(const Spc& a)const // p1.distSquare(p2):=|p1-p2|^2
      {  return (p-a.p).absSqr();}

      static R vol(R r){ return cpmpi*r*r;}
         // 'volume' of the 'sphere' (radius=r) in the space under
         // consideration. If CpmDim2::Spc is considered together with
         // CpmDim3::Spc, it is more  natural to speak of area of a
         // circular disk

      static R sur(R r){ return cpmpi*r*2;}
         // 'surface area' of the 'sphere' (radius=r) in the space under
         // consideration. If CpmDim2::Spc is considered together with
         // CpmDim3::Spc, it is more natural to speak of the length of
         // the circumference of a circular disk

      static R radFrmVol(R ar)
      { return ar<=0 ? R(0) : cpmpow(ar/cpmpi,R(0.5));}
         // radius from 'volume' (of a 'sphere')

      static R radFrmSur(R len ){ return len<=0 ? R(0) : len/(cpmpi*2);}
         // radius from 'surface' (of a 'sphere')

      static Z dimension(void){ return 2;}
         // dimension of the space. Avoids to have to say:
         // Spc x; Z d=x.dim()

      static X4<Spc,R,R,R> meet(const Spc& p1, const Vec& v1,
         const Spc& p2, const Vec& v2);
         // Returns a point p such that there are real numbers x1,x2
         // such that p=p1+v1*x1=p2+v2*x2
         // Actually (p,0,x[1],x[2]) are returned as a quadruple.
         // The second less informative argument is
         // |p1+v1*x1-(p2+v2*x2)|/2 and is only numerical noise. In
         // the 3D case it is more interesting.

      static X2<Spc,R> circle(const V<Spc>& x);
         // let x be a list consisting of 3 mutually different points.
         // We return center and radius of a circle on which
         // x[1], x[2], and x[3] lie.
         // The radius (=inverse of curvature) is given as positive
         // if the direction of x[3]-x[2] results from x[2]-x[1] by
         // a positve rotation (< 180 degrees absolute) and negative if
         // one has to rotate negatively (also <180 degrees in absolute
         // measure)

      static V<Spc> tet();
         //: tetraeder
         // flat tetraeder, tetraeder vertices projected to the plane
   };

   inline Group::Group(const Spc& x0, const AxVec& A1):A(A1)
   { Vec x0p=x0.p; a=x0p; a-=(x0p*=A);}

   inline Group::Group(Spc const& x0, AxVec const& A1, Vec const& a1):
      A(A1){ Vec x0p=x0.p; a=x0p; a-=(x0p*=A); a+=a1;}

   inline Group::Group(const Spc& x0, const Angle& phi):A(AxVec(phi))
   { Vec x0p=x0.p; a=x0p; a-=(x0p*=A);}

//////////////////////// class RefSys ///////////////////////////////////
// orthonormal, positively oriented coordinate system in Euclidean affine
// 2-space

class RefSys{
//orthonormal, positively oriented reference system in space

      friend class AxVec;
      friend class Vec;
      friend class Spc;

// inependent data
      Group g0;

// dependent data for efficiency
      Group g0Inv;
      Vec v1,v2;
      Spc x0;

      void update();

      static Vec u1,u2;
      static Spc orig;

   public:
      typedef RefSys Type;
      CPM_IO
      CPM_ORDER
      Word nameOf()const{ return Word("CpmDim2::RefSys");}
      RefSys();

      explicit RefSys(const Group& g0_):g0(g0_){update();}

      RefSys(const Spc& x0_, const AxVec& av);

      explicit RefSys(const V<Spc>& vs);
         // assumption: vs.dim()>=Rdim
         // only the first Rdim components of vs get used.
         // vs[1] is the origin
         // e1:=(vs[2]-vs[1]), normalized, is the 1-axis
         // e2:=e1*AxVec(Angle(90,DEG)) is the 2-axis
         // If vs is too special for this to work, cpmassert will signal
         // the situation. In case that cpmassert is set not to stop but
         // only to warn (so that the function does not terminate) the
         // default RefSys gets constructed

      RefSys& operator *=(const Group g)
         { g0*=g; update(); return *this;}

      RefSys& operator +=(const Vec& a)
         { g0*=Group(a); update(); return *this;}

      RefSys& operator *=(const AxVec& A)
         { g0*=Group(A); update(); return *this;}

      RefSys& rotByEulAng(const Angle& phi, const Angle& theta,
         const Angle& psi);
         // Rotate by Euler angles. Theta acts as if it were 0.
         // Defined only for uniformity with the 3D case

      RefSys& ref_(Plane const& p);
         //: reflect
         // Changes *this in a way that the origin and the two
         // axis directions get reflected by plane p.

      RefSys per(const Word& p)const{p; return *this;}
      // Permutation not useful in 2-dim, to be defined in order to have
      // the same interface as in the 3-dimensional case.
      // Therefore we return *this without any change.

// getting the components of objects with respect to the reference system
// *this
      R2 operator()(const Spc&)const;
      R2 operator()(const Vec&)const;
      R operator()(const AxVec&)const;
      Spc operator()(Rdim x)const{ return x0+x[1]*v1+x[2]*v2; }
      Vec operator[](Rdim x)const{ return x[1]*v1+x[2]*v2;}

      Group frmMasSys()const{ return g0;}  // from master system
      AxVec rot()const{ return g0.rot();}
      Vec shift()const{ return g0.shift();}

      Group toMasSys()const{ return g0Inv;} // to master system

      Group operator /(const RefSys&)const;

      Group toSys(const RefSys& rs)const{ return rs/(*this);}
         // More suggestive (at least at a first glance) notation
         // for the / operator.
         // RefSys rs1=...,rs2=...;
         // Group g=rs1.toSys(rs2);
         // bool b=(rs2==(rs1*=g)); // b==true always

// getting axes and origin associated with the RefSys
      Vec e1()const{ return v1;}
         // e1 and e2 are orthogonal and lie in the RefSys; they
         // determine the orientation of the RefSys
      Vec e2()const{ return v2;}
      Vec n()const{ return v2;}

      Spc o()const{ return x0;}
      V<Vec> getVectors()const{ return V<Vec>{v1,v2};}
      CpmFunctions::F<Spc,Rdim> toFSR()const;
         // creates a CPM-function which transforms from Spc to
         // coordinates with respect to system *this
      CpmFunctions::F<Rdim,Spc> toFRS()const;
         // creates a CPM-function which transforms from coordinates
         // (with respect to system *this) to Spc

      Plane pln(Z i)const;
         //: plane
         // rs.pln(i) is the plane Plane(rs.ei(),rs.o()) if i=1,2
         // and Plane() else.
      };

//////////////////////// class Plane ///////////////////////////////////
// oriented affine hyperplane, which in this case is a straight line

   class Plane{
// oriented affine hyperplane, which in this case is a line

      RefSys rs;
         // the reference system in which *this is the 1-plane
   public:
      typedef Plane Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmDim2::Plane");}
      Plane();
         // creates the 1-plane of the master system of reference
      explicit Plane(const RefSys& rs_):rs(rs_){}
         // creates a plane, which is the 1-plane of RefSys rs_

      Plane(const Vec& n, const Spc& x0);
         // creates a plane containing point x0 and having a normal unit
         // vector parallel to n (irrespective of normalization of n)
      R dist(const Spc& x)const{ return rs(x)[2];}
         // returns the signed distance of y from the plane. Positive
         // distance means positive 2-component with respect to a
         // positively oriented 'Zweibein' the 1-direction which is in
         // the plane
      Spc reflect(const Spc& x)const
         { Spc res = x; return res.ref_(*this);}
          // returns the mirror point of x with respect to the plane *this
      Vec reflect(const Vec& x)const
         { Vec res = x; return res.ref_(*this);}
         // returns the mirror image of vector x with respect to the
         // plane *this
      AxVec reflect(const AxVec& x)const
         { AxVec res = x; return res.ref_(*this);}
         // returns the mirror image of axial vector x with
         // respect to the plane *this
      Spc project(const Spc& x)const{ R2 y=rs(x); y[2]=0;
         return Spc(y,rs);}
         // returns the projected image of x on the plane *this
      Vec project(const Vec& x)const{ R2 y=rs(x); y[2]=0;
         return Vec(y,rs);}
         // returns the projected image of x on the plane *this
      Plane& operator *=(const Group g){ rs*=g; return *this;}
      Plane& operator +=(const Vec& a){ rs*=Group(a); return *this;}
      Plane& operator *=(const AxVec& A){ rs*=Group(A); return *this;}
      R2 operator()(const Spc& x){ return rs(x);}
      R2 operator()(const Vec& v){ return rs(v);}
      R operator()(const AxVec& av){ return rs(av);}

// getting directions associated with the plane
      Vec e1()const{ return rs.e1();}
      Vec e2()const{ return rs.e2();}
      Vec n()const{ return rs.e2();}
         // normal direction
      Spc o()const{ return rs.o();}
      RefSys getRefSys()const{ return rs;}
      Word des()const;
         // description as a Word for diagnosis
   };

   class SubsetOfPlane{
// subset of a Plane  defined via an indicator function
         // stream I/O makes no sense for this class due to f being not
         // serializable
      RefSys rs;
         // the plane under consideration is the 12-plane of rs
      F<Rdim_,bool> f;
         // f(x1)==true says that a point x with rs-coordinates (x1,0)
         // belongs to the subset to be defined here.

   public:
      Word nameOf()const{ return Word("CpmDim2::SubsetOfPlane");}
      SubsetOfPlane(const RefSys& rs_, const F<Rdim_,bool>& f_):
         rs(rs_),f(f_){}
         // the default constructor thus creates the whole 12-plane of
         // the master system of reference

      SubsetOfPlane():rs(),f(true){}
      explicit SubsetOfPlane(const RefSys& rs_):rs(rs_),f(true){}

      X2<R,bool> ind(const Spc& x)const{ R2 y=rs(x);
         return X2<R,bool>(y[2],f(Rdim_(y[1])));}
         // returns the distance of x from the plane and the truth value
         // of the 'belonging to set'-relation for the projection of x
         // onto the plane

      bool indLean(const Spc& x)const
         { R2 y=rs(x); return f(Rdim_(y[1]));}
         // lean indicator
         // returns the truth value of the 'belonging to set'-relation
         // for the projection of x onto the plane

      SubsetOfPlane& operator *=(const Group g)
         { rs*=g; return *this;}
      SubsetOfPlane& operator +=(const Vec& a)
         { rs*=Group(a); return *this;}
      SubsetOfPlane& operator *=(const AxVec& A)
         { rs*=Group(A); return *this;}
   };

//////////////////// class InTens////////////////////////////////////////
// inertial tensor, tensor of inertia

   class InTens{
// tensor of inertia, reduces in 2D to a scalar, the moment of inertia

      R I; // moment of inertia

   public:
      typedef InTens Type;
      CPM_IO
      Word nameOf()const{ return Word("CpmDim2::InTens");}

      InTens():I(0){}

      explicit InTens(R I_):I(I_){}

      explicit InTens(const RR3& tens):I(tens[3][3]){}
         // construction from a matrix representation, only the
         // 3,3-component counts

      AxVec operator * (const AxVec& a )const
         // inertia tensors transform axial vectors (eg. applied to
         // angular velocity to give angular momentum.
          {  return AxVec(a[3]*I); }

      void balance();
         // changes the inertia tensor *this into one which has equal
         // eigenvalues and the euclidean length of eigVal remains the
         // same. This is useful for plausibility controls.

      InTens& operator *=(const AxVec& a );
         // The definition is such that (I*a)*(x*a)=(I*x)*a

      InTens& operator *=(const Group& );

      InTens ref(Vec const& n)const{ n;return *this;}
         //: reflection
         // Reflection with respect to a line acts trivially here.

      InTens& ref_(Vec const& n){ n;return *this;}
         //: reflection

      InTens& ref_(Plane const& p){ return ref_(p.n());}
         //: reflection

      InTens adjointAction(AxVec const& a)const;

      void scale(R fac){I*=fac;}
         // multiplying the eigenvalues by a common factor fac. No
         // dimension reflecting power of fac is taken.

      friend InTens operator !(const InTens& x);
         // inverse matrix, here implemented by inverting the eigen
         // values

      Vec e1()const{ return Vec();}
      Vec e2()const{ return Vec();}
      Vec e3()const{ return Vec();}
         // should not be relevant in 2D

      R i1()const{ return I;}
      R i2()const{ return I;}
      R i3()const{ return I;}
         // should not be relevant in 2D
   };

   InTens rotate(const InTens&, const AxVec& );
