//? cpmdim2cpp.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using namespace CpmStd;
//////////////////////////// class AxVec ///////////////////////////////

R& AxVec::operator[](Z i)
{
   if (i==3) return x3;
   else {
      cpmerror("CpmDim2::AxVec: index i="&cpmwrite(i)&
      " out of range");
      return x3;
   }
}

const R& AxVec::operator[](Z i)const
{
   if (i==3) return x3;
   else {
      cpmerror("CpmDim2::AxVec: index i="&cpmwrite(i)&
      " out of range");
      return x3;
   }
}

X2<Angle,Vec> AxVec::toAngAxs()const
{
   return X2<Angle,Vec>(Angle(x3,RAD),Vec());
}

AxVec operator *(const AxVec& a1, const AxVec& a2)
// addition of angles
{
   Angle phi1=a1.phi();
   Angle phi2=a2.phi();
   return AxVec(phi1+phi2);
}

namespace{
   R2 actionOfRotationVectors(R2 const& x, R const& a)
   {
      R co=cpmcos(a);
      R si=cpmsin(a);
      return R2(co*x.x1-si*x.x2,si*x.x1+co*x.x2);
   }
}

AxVec AxVec::adjointAction(const AxVec& d)const
{
   d; return *this; // AxVec is Abelian, thus adjoint action trivial
}

AxVec AxVec::exp(const R& t, bool accurate)const
{
   accurate;
   return AxVec(x3*t);
}

AxVec AxVec::pow(R p)const
{
   return AxVec(x3*p);
}

R AxVec::dis(const AxVec& a)const
// method free of potential singularities
{
   Angle phi1=phi();
   Angle phi2=a.phi();
   return phi1.dis(phi2);
}

RR2 toMat(const AxVec& d)
{
   R arc=d[3];
   R co=cpmcos(arc);
   R si=cpmsin(arc);
   RR2 res;
   res.x1.x1=co;
   res.x1.x2=-si;
   res.x2.x1=si;
   res.x2.x2=co;
   return res;
}

Z AxVec::com(const AxVec& a)const
{ return CpmRoot::Root<R>(x3).com(a.x3);}

bool AxVec::scanFrom(istream& in)
{
   return CpmRoot::read(x3,in);
}

bool AxVec::prnOn(ostream& out)const
{
   if (!CpmRoot::writeTitle("AxVec",out)) return false;
   if (!CpmRoot::write(x3,out)) return false;
   return true;
}

AxVec::AxVec(const V<Vec>& b1, const V<Vec>& b2, bool careful)
{
   V<Vec> e1=b1;
   V<Vec> e2=b2;
   if (careful){
      e1=e1[1].base(e1[2]);
      e2=e2[1].base(e2[2]);
   }
   Vec vn=b2[1];
   C z(vn|b1[1],vn|b1[2]);
      // components of the new direction in the old basis
   *this=AxVec(Angle(z));
}

AxVec AxVec::ran(Z j)const
{
   R av=CpmRoot::ran(x3,j);
   return AxVec(av);
}

Vec AxVec::cross(Vec const& v)const
{
   return Vec(-x3*v[2],x3*v[1]);
}

///////////////////// class Vec ///////////////////////////////////////

Vec& Vec::operator *=(const Group& g )
   // The definition is such that (I*a)*(x*a)=(I*x)*a
{
   return *this*=g.A;
}

Vec operator *(const Vec& x, const AxVec& d)
// see actionOfRotationVectors()
{
   return Vec(actionOfRotationVectors(x(),d()));
}

X2<Vec,Vec> Vec::tanNormDec(const Vec& a)const
{
   X2<R,Vec> pdt=polDec();
   Vec t=pdt.second();
   Vec tanComp=t*(t|a);
   Vec normComp=a-tanComp;
   return X2<Vec,Vec>(tanComp,normComp);
}

V<Vec> Vec::base(const Vec& a)const
{
   // a not needed in 2 dimensions
   a;
   X2<R,Vec> p=polDec();
   V<Vec> res(2);
   if (p.first()==0.){
      res[1]=Vec(1,0);
      res[2]=Vec(0,1);
   }
   else{
      Vec v1=p.second();
      res[1]=v1;
      res[2]=v1;
      AxVec rot90(Angle(90.));
      res[2]*=rot90;
   }
   return res;
}

AxVec Vec::rotTo(const Vec& v)const
{
   X2<R,Vec> p1=polDec();
   if (p1.first()==0.) return AxVec();
   Vec e1=p1.second();
   AxVec rot90(Angle(90.));
   Vec e2=e1;
   e2*=rot90;
   C z((v|e1),(v|e2));
   return AxVec(Angle(z));
}

AxVec Vec::rotParOrAntiPar(const Vec& v)
{
  AxVec res1=rotTo(v);
  AxVec res2=rotTo(-v);
  return (res1.abs()<=res2.abs() ? res1 : res2);
}

Vec Vec::ranUnitVec(Z j)
{
   R phi=CpmRoot::randomR(j)*cpmpi*2;
   return Vec(cpmcos(phi),cpmsin(phi));
}

#include <cpmhomunitvecxcpp.h>
   // a piece of implementation which is common to 2D and 3D

////////////////////////// class Spc ////////////////////////////////////

R Spc::refLen=1e-3;
   // a mm, this suitable for small particle applications

static R f(R x)
//f: [0,Infinity]-->[0,1], f'(0)=1
{
   static R PiHalf=cpmpi*0.5;
   static R PiHalfInv=R(1)/PiHalf;

   return PiHalfInv*cpmatan(PiHalf*x);
}

F<Spc,Rdim> Spc::FSR=F<Spc,Rdim>(toRdimStatic);

R Spc::dis(const Spc& a)const
   // p1.sistValue(p2)=(2/Pi)arctan((Pi/2)*|p1-p2|/refLen)
   // translation invariant function
   // Was a bad construct having a hidden reference to the
   // conventional length 1
   // (corresponding to 1 m in my applications till 2000-01-27)
{
   R x=(p-a.p).abs()/refLen;
   return f(x);
}

Spc Spc::convex(const V<Spc>& x, V<R>& w, bool careful)
{
   Z n=x.dim(),i;
   cpmassert(n==w.dim(),"CpmDim2::convex");
   if (careful){
      R sum=0,val;
      for (i=1;i<=n;i++){
         val=w[i];
         if (val<0) val=-val;
         sum+=val;
         w[i]=val;
      }
      if (sum>0){
         sum=R(1)/sum;
         for (i=1;i<=n;i++) w[i]*=sum;
      }
      else{
         cpmwarning("CpmDim2::convex: sum=0");
         R wi=1./n;
         for (i=1;i<=n;i++) w[i]=wi;
      }
   }
   Vec pSum;
   for (i=1;i<=n;i++) pSum+=x[i].p*w[i];
   return Spc(pSum);
}

X2<Spc,R> Spc::envSph(V<Spc> const& vs)
{
   Z n=vs.dim();
   if (n==0){
      Spc o;
      return X2<Spc,R>(o,0);
   }
   R nInv=1./n;
   Vec pSum;
   Z i;
   for (i=1;i<=n;i++) pSum+=vs.cui(i).p*nInv;
   Spc cen(pSum);
   Vo<R> rs(n);
   for (i=1;i<=n;i++) rs.cui(i)=(vs.cui(i)-cen).abs();
   R r=rs.sup();
   return X2<Spc,R>(cen,r);
}

X2<Spc,R> Spc::circle(const V<Spc>& x)
{
   using namespace CpmArrays;
   Z mL=3;
   static Word loc("Spc::circle(const V<Spc>& x)");
   cpmmessage(mL,loc&" started");
   Z n=x.dim();
   cpmassert(n==3,loc);
   Angle w90(90.);
   AxVec rot90(w90);
   Spc p1=x[1];
   Spc p2=x[2];
   Spc p3=x[3];
   Vec v21=p2-p1;
   Vec v32=p3-p2;
   bool pos=v21.rotTo(v32).phi().isPos();
   Spc m21=p1+v21*0.5;
   Spc m32=p2+v32*0.5;
   v21*=rot90;
   v32*=rot90;
   X4<Spc,R,R,R> meet=Spc::meet(m21,v21,m32,v32);
   R_Vector d(3);
   Spc pRes=meet.c1();
   d[1]=(p1-pRes).abs();
   d[2]=(p2-pRes).abs();
   d[3]=(p3-pRes).abs();
   R r= ( pos==true ? d.mean() : -d.mean() );
   cpmmessage(mL,loc&" done");
   return X2<Spc,R>(pRes,r);
}

V<Spc> Spc::tet()
// as in 3 dimensions but ignoring all z-coordinates
{
   const R h=::sqrt(3.)*0.5;
   const R hs=h/3;
   const R hl=hs*2;
   Rdim p1(hl,0);
   Rdim p2(-hs,0.5);
   Rdim p3(-hs,-0.5);
   Rdim p4(0,0);
   return V<Spc>{Spc(p1),Spc(p2),Spc(p3),Spc(p4)};
}

//////////////////////////// class Group ///////////////////////////////

//////////////////////// class RefSys ///////////////////////////////////

Vec RefSys::u1(1,0);
Vec RefSys::u2(0,1);

Spc RefSys::orig(0,0);

RefSys::RefSys() // g0,g0Inv initialized as unity, x0 as (0,0)
{
   v1=u1; // calling update() here would be unnecessarily complicated
   v2=u2;
}

RefSys::RefSys(const Spc& x0_, const AxVec& av)
// construct the RefSys which has x0_ as origin and the axes of
//which result from the master axes by applying the rotation
// described by the axial vector av
{
   Vec shift=x0_-orig;
   g0=Group(av,shift);
   update();
}

void RefSys::update()
{
   v1=u1*g0;
   v2=u2*g0;
   x0=orig*g0;
   g0Inv=!g0;
}

R2 RefSys::operator()(const Spc& x)const
{
   Vec v=x-x0;
   return R2(v|v1,v|v2);
}

R2 RefSys::operator()(const Vec& v)const
{
   return R2(v|v1,v|v2);
}

R RefSys::operator()(const AxVec& v)const
{
   return v[1];
}

Group RefSys::operator /(const RefSys& rs)const
{
   return rs.g0Inv*g0; // notice that g0 = rs.g0 * ( rs.g0Inv*g0) is the
   // relation to be satisfied by the expression in brackets
}

AxVec::AxVec(const R& a, const RefSys& rs)
{
   rs;
   x3=a;
}

Vec::Vec(const R2& a, const RefSys& rs)
{
   *this=a[1]*rs.v1+a[2]*rs.v2;
}

Spc::Spc(const R2& x, const RefSys& rs)
{
   *this=rs.x0+x[1]*rs.v1+x[2]*rs.v2;
}

Spc::Spc(const V<R>& x, const RefSys& rs)
{
   R2 a(x);
   *this=rs.x0+a[1]*rs.v1+a[2]*rs.v2;
}

RefSys& RefSys::rotByEulAng(const Angle& phi, const Angle& theta,
         const Angle& psi)
{
   theta; // no influence in 2D
   AxVec av(phi+psi);
   return operator *=(av);
}

RefSys& RefSys::ref_(Plane const& p)
{
   V<Spc> vs(3,o());
   vs[2]+=e1();
   vs[3]+=e2();
   for (Z i=1; i<=3; ++i) vs[i].ref_(p);
   return *this=RefSys(vs);
}

Plane RefSys::pln(Z i)const
{
   if (i==1) return Plane(e1(),o());
   else if (i==2) return Plane(e2(),o());
   else return Plane();
}

//////////////////////// class Plane ///////////////////////////////////

Plane::Plane(){}

Plane::Plane(const Vec& n, const Spc& x)
{
   Spc x0;
   Vec a=x-x0;
   Vec e2(0,1);
   AxVec A=e2.rotTo(n);
   Group g(A,a);
   rs=RefSys(g);
}

Word Plane::des()const
{
   ostringstream ost;
   ost<<endl<<"Description of a instance of Plane:";
   ost<<endl<<"Origin is"<<rs.o();
   ost<<endl<<"e1() is"<<e1();
   ost<<endl<<"n() is"<<n();
   string res=ost.str();
   return Word(res);
}

////////////////////////// class InTens ////////////////////////////////

InTens& InTens::operator *=(const AxVec& a )
   // The definition is such that (I*a)*(x*a)=(I*x)*a
{
   a; // no influence
   return *this;
}

InTens& InTens::operator *=(const Group& g )
   // The definition is such that (I*a)*(x*a)=(I*x)*a
{
   g; // no influence
   return *this;
}

InTens operator !(const InTens& x)
{
   R xI=x.I;
   cpmassert(xI>0.,"CpmDim2::operator !(InTens)");
   return InTens(R(1)/xI);
}

InTens rotate(const InTens& I, const AxVec& a )
// far from optimum at least for symmetric I's
{
   InTens res=I;
   res*=a;
   return res;
}

//////////////// I/O for all classes ///////////////////////////////////

bool Spc::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Spc",out);
   return p.prnOn(out);
}

bool Spc::scanFrom(istream& in)
{
   return p.scanFrom(in);
}

bool Group::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Group",out);
   return (A.prnOn(out)&&a.prnOn(out));
}

bool Group::scanFrom(istream& in)
{
   return (A.scanFrom(in)&&a.scanFrom(in));
}

bool RefSys::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("RefSys",out);
   return g0.prnOn(out);
}

bool RefSys::scanFrom(istream& in)
{
   bool res=g0.scanFrom(in);
   update();
   return res;
}

bool Plane::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Plane",out);
   return rs.prnOn(out);
}

bool Plane::scanFrom(istream& in)
{
   return rs.scanFrom(in);
}

bool InTens::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("InTens",out);
   return CpmRoot::write(I,out);
}

bool InTens::scanFrom(istream& in)
{
   return CpmRoot::read(I,in);
}

void InTens::balance(){}
   // changes the inertia tensor *this into one which has equal
   // eigenvalues and the euclidean length of eigVal remains the same.
   // This is useful for plausibility controls

InTens InTens::adjointAction(AxVec const& a)const
{
   InTens res=*this;
   res*=a;
   return res;
}
