//? cpmdim3cpp.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmGeo;
   // using declarations/directives seem to work
   // only reliable if not separated from its scope by include mechanisms
   // so we did this not in cpmdim.cpp

//////////////////////////// class AxVec ///////////////////////////////
// some local tools:

namespace {

   const R hugeVal=1e12;
   // rotation vectors should never be longer than this
   // hugeVal=1e6 says that instead of rotating by 180 degrees (which has a
   // singular representation) we rotate by an angle which deviates by
   // 2/hugeVal rad = 2e-6 rad = 0.41" from 180 degrees. This is OK for
   // scientific applications.
   // The closer the rotation angle to 180 degrees, the larger the roundoff
   // errors in calculating the action of this transformation on space.
   // action: This should depend on the precision if multiple precision is
   // enabled!

   const R hugeValInv=R(1)/hugeVal;
   const R piHalfAppr=cpmatan(hugeVal);
   const R piAppr=piHalfAppr*2;

R tanHalf(R x)
// safe way of forming nonsingular tan(x/2)
{
   if (!(x<=cpmpi)) cpmerror("tanHalf(R x):x<=Pi required, x="&cpm(x));
   if (!(x>=-cpmpi)) cpmerror("tanHalf(R x):x>=-Pi required x="&cpm(x));
      // was x>-Pi till 2005-06-28. Running PaLa under Linux once
      // resulted in calling the function for x=-Pi which under Windows
      // never happened
   if (x==0.) return 0.;
   R xh=x*0.5;
   if (xh>piHalfAppr) xh=piHalfAppr;
   if (xh<-piHalfAppr) xh=-piHalfAppr;
   return cpmtan(xh);
}

} // namespace

AxVec::AxVec(const Angle& phi, const R3& n)
{
   R3 unitAxis=n.polDec().second();
   R arc=phi;
      // now arc in (-pi,pi] , only for this range, forming the
      // half-angle has the desired effect
   R da=tanHalf(arc);
   x_.x1=da*unitAxis.x1;
   x_.x2=da*unitAxis.x2;
   x_.x3=da*unitAxis.x3;
}

AxVec::AxVec(const Angle& phi, Z i)
{
   cpmassert(i>0 && i<4,"AxVec(Angle,Z)");
   R arc=phi;
   R da=tanHalf(arc);
   if (i==1){ x_.x1=da; x_.x2=0; x_.x3=0; return;}
   if (i==2){ x_.x1=0; x_.x2=da; x_.x3=0; return;}
   if (i==3){ x_.x1=0; x_.x2=0; x_.x3=da; return;}
}

AxVec::AxVec(const Angle& phi, const Vec& axis)
// since we describe the object to be constructed by components
// with respect to the master system, it is OK to use the components
// of axis with respect to the master system
{
   R3 ax=axis();
   R3 unitAxis=ax.polDec().second();
   R arc=phi;
      // now arc in (-pi,pi] , only for this range, forming the
      // half-angle has the desired effect
   R da=tanHalf(arc);
   x_=unitAxis*da;
}

AxVec::AxVec(Angle const& phi, AxVec const& axis)
{
   R3 ax=axis();
   R3 unitAxis=ax.polDec().second();
   R arc=phi;
      // now arc in (-pi,pi] , only for this range, forming the
      // half-angle has the desired effect
   R da=tanHalf(arc);
   x_=unitAxis*da;
}

Angle AxVec::phi()const
{
  R av=abs(); // tan(phi/2)
  if (av>=hugeVal) av=piAppr; else if (av>0) av=cpmatan(av)*2; else ;
  return Angle(av,RAD); // i.e. don't interpret the argument as degrees
   // RAD was missing till 2000-2-28
}

AxVec AxVec::log()const
{
   R3 x=x_;
   R av=x.nor_(); // tan(phi/2)
   if (av>=hugeVal) av=piAppr; else if (av>0) av=cpmatan(av)*2; else ;
   return AxVec(x*av);
}

AxVec& AxVec::phi_(Angle const& phi)
{
   nor_();
   R arc=phi;
   R da=tanHalf(arc);
   return operator*=(da);
}

X2<Angle,Vec> AxVec::toAngAxs()const
{
   R3 x=x_;
   R av=x.nor_();
   if (av>=hugeVal) av=piAppr; else if (av>0) av=cpmatan(av)*2; else ;
   return X2<Angle,Vec>(Angle(av,RAD),Vec(x));
}

AxVec AxVec::operator *(const AxVec& a)const
{
   R nom=R(1)-x_.x1*a.x_.x1-x_.x2*a.x_.x2-x_.x3*a.x_.x3;
      // Count of operations: 3M+3A
   R nomAbs=(nom>=0 ? nom : -nom);
      // A
   R fac=(nomAbs<=hugeValInv ? hugeVal : R(1)/nom);
      // D
   AxVec res;
   res.x_.x1=(x_.x1+a.x_.x1-x_.x2*a.x_.x3+x_.x3*a.x_.x2)*fac;
   res.x_.x2=(x_.x2+a.x_.x2-x_.x3*a.x_.x1+x_.x1*a.x_.x3)*fac;
   res.x_.x3=(x_.x3+a.x_.x3-x_.x1*a.x_.x2+x_.x2*a.x_.x1)*fac;
      // 9M+9A
      // Total count: 12M+D+13A
   return res;
}

AxVec& AxVec::operator *=(const Group& g )
{
   return *this*=g.A;
}

Vec AxVec::cross(Vec const& v)const
{ return Vec(x_.crsPrd(v()));}

/*
Vec CpmDim3::cross(const AxVec& a, const Vec& v)
{ return Vec(a.crsPrd(v));}
*/

namespace{ // anon

/*
R3 actionOfRotationVectors(const R3& x, const R3& d)
// The result is just R(d)x in my old terminology. For |d|<<1 this
// implies:
// x*d=x+dx=x+2d.cross(x)=x+dphi*n.cross(x)  for d=tan(dphi/2)*n,  |n|=1.
// Thus if d has only a 3-component and x only a 1-component, dx has only
// the component d[3]*x[1]. This is what a dphi-rotation arround the
// 3-axis in the positive sense (the one given in the (1,2) plane as
// going from the 1-axis to the 2-axis).
// The comments to the following expressions give the 'big-O' behavior
// for large |d|; for small |d|, there is not even a quasi-problem.
// This algorithmic block will be used for AxVec(d) and x and return
// value being either Vec or AxVec. In order not to dublicate code,
// it appears here as a auxiliary function private to this file.
// Without introducing ad hoc subexpressions one has
// x*d := [(1-(d|d))*x +2*(x|d)*d + 2*vectorProduct(d,x)]/[1+(d|d)]
{
   R ha=d|d;         // O(|d|^2) for large |d|
      // multiplication (M), division (D) and addition (A) count:
      // 3M+2A
   if (ha==0.) return x;
   R hb=x|d;         // O(|d|)
      // 3M+2A
   R hc=2./(1.+ha);  // O(|d|^-2)
      // D+A
   R hd=0.5*(1.-ha); // O(d^2)
      // M+A
   R he=hc*hd;       // O(|d|^0)
      // M
   R3 ds=d*hc;    // O(|d|^-1)
      // 3M
   R r1=he*x.x1 + hb*ds.x1 + ds.x2*x.x3 - ds.x3*x.x2;
      // O(|d|^0)+O(|d|^0)+O(|d|^-1)-O(|d|^-1); thus no large terms for
      // large |d|
   R r2=he*x.x2 + hb*ds.x2 + ds.x3*x.x1 - ds.x1*x.x3;
   R r3=he*x.x3 + hb*ds.x3 + ds.x1*x.x2 - ds.x2*x.x1;
   // each: 4M+3A thus whole group: 12M+9A
   // Total count: 23M+1D+15A
   return R3(r1,r2,r3);
}
*/

R3 actionOfRotationVectors(const R3& x, const R3& d)
// This implementation is 6 Multiplications less effective
// as the optimum but saves 16M+1D+13A if the same
// d gets applied to one vector more. If the same d is applied
// to more vectors in succession with each vector this saving
// occurs again. For instance, rotating a whole basis is much more
// efficient this way.
{
   static bool firstrun=true;
   static R b11,b12,b13,b21,b22,b23,b31,b32,b33;
   static R d1,d2,d3,fac;
   bool setStatics; // will get initialized with certainty
   if (firstrun){
      setStatics=true;
      firstrun=false;
   }
   else{
      setStatics=d1!=d.x1||d2!=d.x2||d3!=d.x3;
   }
   if (setStatics){
      // all this stuff is saved if d is the same as in the last
      // call of this function
      d1=d.x1;
      d2=d.x2;
      d3=d.x3;
      R a11=d1*d1;
      R a12=d1*d2;
      R a13=d1*d3;
      R a22=d2*d2;
      R a23=d2*d3;
      R a33=d3*d3;
   // 6M

      R rr=a11+a22+a33;
      fac=R(2)/(rr+1);
   // D+3A
      R c=(-rr+1)*0.5;
   // M+A

      b11=a11+c;
      b12=a12+d3;
      b13=a13-d2;

      b21=a12-d3;
      b22=a22+c;
      b23=a23+d1;

      b31=a13+d2;
      b32=a23-d1;
      b33=a33+c;
      // 9A

      // Total in this block: 7M+1D+13A, only to be done if
      // d has a new value
   }
   return R3(
      (b11*x.x1+b21*x.x2+b31*x.x3)*fac,
      (b12*x.x1+b22*x.x2+b32*x.x3)*fac,
      (b13*x.x1+b23*x.x2+b33*x.x3)*fac
      );
      // 12M+6A
   // Total 19M+1D+19A (in worst case that d has a new value)
}

} // namespace

AxVec AxVec::adjointAction(const AxVec& d)const
{
   return AxVec(actionOfRotationVectors(x_,d()));
}

AxVec AxVec::exp(const R& t, bool accurate)const
{
   if (!accurate){// return (*this)*0.5*t;
      AxVec res=*this;
      return res*=(t*0.5); // t was missing here till 2005-02-14
   }
   else{
      X2<R,R3> pd=x_.polDec();
      R phiDot=pd.first();
      R3 axis=pd.second();
      Angle dphi(phiDot*t,RAD); // not in degrees!
      return AxVec(dphi,axis);
         // notice that n*r |--> n*tan(r*t/2) so that for small r
         // we have n*r |--> (n*r)*(t/2) which is the special
         // version for accurate==false
   }
}

AxVec AxVec::pow(R p)const
{
   X2<R,R3> pd=x_.polDec();
   R3 axis=pd.second();
   Angle phi_=phi();
   phi_*=p;
   return AxVec(phi_,axis);
}

R AxVec::dis(const AxVec& a)const
// method free of potential singularities
{
   static const Vec e1(1,0,0);
   static const Vec e2(0,1,0);
   static const Vec e3(0,0,1);
   static const R inv6=1./6.;
   Vec v1=e1*(*this); // R3 could be replaced by Vec
   Vec v2=e2*(*this);
   Vec v3=e3*(*this);
   Vec w1=e1*a;
   Vec w2=e2*a;
   Vec w3=e3*a;
   return
      ((v1-w1).abs()+(v2-w2).abs()+(v3-w3).abs())*inv6;
      // <=1,
      // g1=*this=id and g2=total inversion e1|-->-e1, e2|-->-e2,
      // e3|-->-e3,
      // would give g1.disue(g2)=1; however the total inversion is not
      // created by a AxVec so the value 1 will presumably not reached
      // within AxVec by considering (*this)*!a we could do with v1,v2,v3
      // instead of v1,v2,v3,w1,w2,w3 at the price of relaying on
      // AxVec*AxVec in addition to AxVec*Vec. Presently AxVec*AxVec has
      // to do a polar decomposition so that the present method should be
      // slightly faster (no big concern anyway!)
}

RR3 toMat(const AxVec& ad)
      // rotation matrix belonging to the axial vector
      // looks quite efficient. The result is R(-d) in my old terminology
{
   R3 d=ad();
   R d11=d.x1*d.x1;
   R d22=d.x2*d.x2;
   R d33=d.x3*d.x3;
   R d2=d11+d22+d33;
   R ha=(R(1)-d2)*0.5;
   R hb=R(2)/(d2+1); // OK without test for 0

   RR3 res;
// diagonal
   res.x1.x1=hb*(ha+d11);
   res.x2.x2=hb*(ha+d22);
   res.x3.x3=hb*(ha+d33);
// off-diagonal
   R dxy=d.x1*d.x2;
   res.x1.x2=hb*(dxy+d.x3);
   res.x2.x1=hb*(dxy-d.x3);
   dxy=d.x1*d.x3;
   res.x1.x3=hb*(dxy-d.x2);
   res.x3.x1=hb*(dxy+d.x2);
   dxy=d.x2*d.x3;
   res.x2.x3=hb*(dxy+d.x1);
   res.x3.x2=hb*(dxy-d.x1);
   return res;
}

namespace{

AxVec fromMat(const RR3& mat)
// natural return value would be R3
// Theory see: Werner Graeub: Lineare Algebra, Springer
// (Grundlehren...Band XCVII) 1958
// p. 147-148
{
   R asy23=mat[2][3]-mat[3][2];
   R asy13=mat[1][3]-mat[3][1];
   R asy12=mat[1][2]-mat[2][1];
   R3 a(-asy23,asy13,-asy12);
   R trace=mat[1][1]+mat[2][2]+mat[3][3];
   R cs2=trace-1; // 2*cos(phi)
   R si2=a.abs(); // 2*sin(phi)
   Angle phi(C(cs2,si2));
   return AxVec(phi,a);
      // a needs not to be normalized
}

} // namespace

AxVec::AxVec(const V<Vec>& b1, const V<Vec>& b2, bool careful)
{
   V<Vec> e1=b1;
   V<Vec> e2=b2;
   if (careful){
      e1=e1[1].base(e1[2]);
      e2=e2[1].base(e2[2]);
   }
   RR3 mat;
   for (Z i=1;i<=3;i++){
      for (Z j=1;j<=3;j++){
         mat[i][j]=e1[i]|e2[j];
      }
   }
   AxVec prel=fromMat(mat);
   x_.x1=prel[1]*e1[1][1]+prel[2]*e1[2][1]+prel[3]*e1[3][1];
   x_.x2=prel[1]*e1[1][2]+prel[2]*e1[2][2]+prel[3]*e1[3][2];
   x_.x3=prel[1]*e1[1][3]+prel[2]*e1[2][3]+prel[3]*e1[3][3];
}

AxVec AxVec::ran(Z j)const
{
   static unsigned long int j0=137;
   Z i=(j!=0 ? j : j0++);
   R alpha=phi();
   Iv iv(alpha,-alpha);
   R alphai=iv.ran(i);
   Vec ni=Vec::ranUnitVec(i);
   return AxVec(Angle(alphai,RAD),ni);
}

AxVec& AxVec::ref_(Vec const& n)
{
   R3 nr = n.toRdim();
   R3 xr = toR3();
   R sp2 = (xr|nr)*2;
   R3 d = nr*sp2;
   *this = AxVec(d-xr);
   return *this;
}

bool AxVec::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(x_);
   return true;
}

bool AxVec::scanFrom(istream& str)
{
   cpms(x_);
   return true;
}

Z AxVec::com(AxVec const& v)const
{ return x_.com(v.x_);}

///////////////////// class Vec ///////////////////////////////////////

Vec& Vec::operator *=(const Group& g )
   // The definition is such that (I*a)*(x*a)=(I*x)*a
{
   return *this*=g.A;
}

Vec operator *(const Vec& x, const AxVec& d)
// see actionOfRotationVectors()
{
   return Vec(actionOfRotationVectors(x(),d()));
      // was AxVec(...) till 2000-2-28
}

X2<Vec,Vec> Vec::tanNormDec(const Vec& a)const
{
   X2<R,Vec> pdt=polDec();
   Vec t=pdt.second();
   Vec tanComp=t*(t|a);
   Vec normComp=a-tanComp;
   return X2<Vec,Vec>(tanComp,normComp);
}

V<Vec> Vec::base(const Vec& a)const
   // returns a positively oriented triplet of orthogonal unit vectors
   // e[1],e[2],e[3].
   // If *this is not zero, e[1] is parallel to *this
   // else e[1]=Vec(1,0,0).
   // If e[1] and b determine a plane, then e[1],b and e[1],e[2]
   // determine the same oriented plane.
   // Otherwise, e[2] is the first vector from the list
   // Vec(1,0,0), Vec(0,1,0), Vec(0,0,1) which is not proportional to
   // e[1].
{
   static bool firstrun=true;
   static V<Vec> test(4,Vec(),0); // valid indexes are 0,1,2,3
   static const R small=1e-9;
      // two vectors are considered to define a plane with sufficient
      // precision if the tangent of the Angle which they form is larger
      // than this
   if (firstrun){
      test[1]=Vec(1,0,0);
      test[2]=Vec(0,1,0);
      test[3]=Vec(0,0,1);
      firstrun=false;
   }
   V<Vec> e(3);
   X2<R,Vec> p=polDec();
   if (p.first()>0) e[1]=p.second(); else e[1]=test[1];
   Z iTest=0;
   test[iTest]=a;
   while (iTest<4){
      X2<Vec,Vec> pa=e[1].tanNormDec(test[iTest]);
      R pafa=pa.first().abs(); // length of part of test[iTest]
         // which is parallel to e[1]

      X2<R,Vec> p2=pa.second().polDec(); // part of test[iTest]
         // which is normal to e[1]; we start with test[iTest]=a
         // and replace a by canonical basis vectors if it does not fit

      if (p2.first()>small*pafa){
         e[2]=p2.second(); // is a unit vector
         e[3]=e[1].crossExc(e[2]);
         return e;
      }
      iTest++;
   }
// never reach  this point
   cpmerror("unexpected behavior in Vec::tripod(...)");
   return e;
}

AxVec Vec::rotTo(const Vec& v)const
{
   X2<R,Vec> p1=polDec();
   if (p1.first()==0.) return AxVec();
   Vec e1=p1.second();
   Vec e3(e1.crossExc(v));
   X2<R,Vec> p3=e3.polDec();
   if (p3.first()==0.) return AxVec();
   e3=p3.second();
   Vec e2(e3.crossExc(e1));
   C z((v|e1),(v|e2));
   return AxVec(Angle(z),e3);
}

AxVec Vec::rotParOrAntiPar(const Vec& v)
{
  AxVec res1=rotTo(v);
  AxVec res2=rotTo(-v);
  return (res1.abs()<=res2.abs() ? res1 : res2);
}

Vec Vec::ranUnitVec(Z j)
{
   static bool jIni=false;
   static CpmRoot::N j0;
   if (!jIni){
      j0=137;
      jIni=true;
   }
   R r1,r2,r3;
   R3 r;
   if (j==0){
Again:
      r1=CpmRoot::randomR();
      r2=CpmRoot::randomR();
      r3=CpmRoot::randomR();
      r1-=0.5;
      r2-=0.5;
      r3-=0.5;
      r=R3(r1,r2,r3);
      R len2=r.absSqr();
      if (len2>0.25 || len2==0) goto Again; else r.nor_();
   }
   else{
      Z jAct=j0*3+j;
   Again2:
      r1=CpmRoot::randomR(jAct++);
      r2=CpmRoot::randomR(jAct++);
      r3=CpmRoot::randomR(jAct++);
      j0+=3;
      r1-=0.5;
      r2-=0.5;
      r3-=0.5;
      r=R3(r1,r2,r3);
      R len2=r.absSqr();
      if (len2>0.25 || len2==0) goto Again2; else r.nor_();
   }
   return Vec(r);
}

#include <cpmhomunitvecxcpp.h>
   // a piece of implementation which is common to 2D and 3D

//////////////////////// class LieAlg ////////////////////////////////////
Group LieAlg::exp(R t)const
{
   AxVec axis=A_;
   R phi=axis.nor_();
   phi*=t;
   Vec axisVec(axis); // need a Vec
   Vec ap=axisVec*(a_|axisVec);
   Vec an=a_-ap;
   AxVec ARes=(axis*phi).exp(1.);
   Vec aRes=ap+((an*ARes-an)*(-axis))*(R(1)/phi);
   return Group(ARes,aRes);
}

namespace{
   Vec complexAction(C z, Vec const& v)
   {
      C vc(v[1],v[2]);
      vc*=z;
      return Vec(vc[1],vc[2],v[3]);
   }

   AxVec fromTo(Vec const& v1, Vec const& v2)
      // assumes that v1 and v2 are unit vectors.
      // Returs the shortest rotation which moves
      // v1 into v2
   {
      return v1.cross(v2)*(R(1)/((v1|v2)+1));
   }
} // namespace

LieAlg Group::log()const
{
   AxVec ARes=A.log();
   AxVec axis=ARes;
   R na=axis.nor_();
   Vec axisVec(axis);
   Vec e3(0.,0.,1.);
   AxVec rot3=fromTo(axisVec,e3);
   Vec aRes=a;
   aRes*=rot3;
   C z=C(0.,na)/(C::expi(na)-C(1.));
   aRes=complexAction(z,aRes);
   aRes*=-rot3;
   return LieAlg(ARes,aRes);
}

////////////////////////// class Spc ////////////////////////////////////

R Spc::refLen=1e-3;
   // a mm, this suitable for small particle applications

static R f(R x)
//f: [0,Infinity]-->[0,1], f'(0)=1
{
   static R PiHalf=cpmpi*0.5;
   static R PiHalfInv=R(1)/PiHalf;

   return PiHalfInv*cpmatan(PiHalf*x);
}

R Spc::dis(const Spc& a)const
   // p1.sistValue(p2)=(2/Pi)arctan((Pi/2)*|p1-p2|/refLen)
   // translation invariant function
   // Was a bad construct having a hidden reference to the conventional
   // length 1
   // (corresponding to 1 m in my applications till 2000-01-27)
{
   R x=(p-a.p).abs()/refLen;
   return f(x);
}

Spc Spc::convex(const V<Spc>& x, V<R>& w, bool careful)
{
   Z n=x.dim(),i;
   cpmassert(n==w.dim(),"CpmDim3::convex");
   if (careful){
      R sum=0,val;
      for (i=1;i<=n;i++){
         val=w[i];
         if (val<0) val=-val;
         sum+=val;
         w[i]=val;
      }
      if (sum>0){
         sum=R(1)/sum;
         for (i=1;i<=n;i++) w[i]*=sum;
      }
      else{
         cpmwarning("CpmDim3::convex: sum=0");
         R wi=1./n;
         for (i=1;i<=n;i++) w[i]=wi;
      }
   }
   Vec pSum;
   for (i=1;i<=n;i++) pSum+=x[i].p*w[i];
   return Spc(pSum);
}

F<Spc,Rdim> Spc::FSR=F<Spc,Rdim>(toRdimStatic);

X2<Spc,R> Spc::envSph(V<Spc> const& vs)
{
   Z n=vs.dim();
   if (n==0){
      Spc o;
      return X2<Spc,R>(o,0);
   }
   R nInv=1./n;
   Vec pSum;
   Z i;
   for (i=1;i<=n;i++) pSum+=vs.cui(i).p*nInv;
   Spc cen(pSum);
   Vo<R> rs(n);
   for (i=1;i<=n;i++) rs.cui(i)=(vs.cui(i)-cen).abs();
   R r=rs.sup();
   return X2<Spc,R>(cen,r);
}

V<Spc> Spc::tet()
{
   const R h=::sqrt(3.)*0.5;
   const R hz=::sqrt(2./3);
   const R hs=h/3;
   const R hl=hs*2;
   Rdim p1(hl,0,0);
   Rdim p2(-hs,0.5,0);
   Rdim p3(-hs,-0.5,0);
   Rdim p4(0,0,hz);
   return V<Spc>{Spc(p1),Spc(p2),Spc(p3),Spc(p4)};
}

//////////////////////// class RefSys ///////////////////////////////////

Vec RefSys::u1(1,0,0);
Vec RefSys::u2(0,1,0);
Vec RefSys::u3(0,0,1);
Spc RefSys::orig(0,0,0);

RefSys::RefSys() // g0,g0Inv initialized as unity, x0 as (0,0,0)
{
   v1=u1; // calling update() here would be unnecessarily complicated
   v2=u2;
   v3=u3;
}

RefSys::RefSys(const Spc& x0_, const AxVec& av)
{
   Vec shift=x0_-orig;
   g0=Group(av,shift);
   update();
}

void RefSys::update()
{
   v1=u1*g0;
   v2=u2*g0;
   v3=u3*g0;
   x0=orig*g0;
   g0Inv=!g0;
}

/************
RefSys& RefSys::ref_(Plane const& p)
{
   V<Spc> vs(4,o());
   vs[2]+=e1();
   vs[3]+=e2();
   vs[4]+=e3();
   for (Z i=1; i<=4;++i) vs[i].ref_(p);
   return *this=RefSys(vs);
}
************/

Plane RefSys::pln(Z i)const
{
   if (i==1) return Plane(e1(),o());
   else if (i==2) return Plane(e2(),o());
   else if (i==3) return Plane(e3(),o());
   else return Plane();
}

R3 RefSys::operator()(const Spc& x)const
{
   Vec v=x-x0;
   return R3(v|v1,v|v2,v|v3);
}

R3 RefSys::operator()(const Vec& v)const
{
   return R3(v|v1,v|v2,v|v3);
}

R3 RefSys::operator()(const AxVec& v)const
{
   Vec vv(v());
   return R3(vv|v1,vv|v2,vv|v3);
}

Group RefSys::operator /(const RefSys& rs)const
{
   return rs.g0Inv*g0;
      // notice that g0 = rs.g0 * ( rs.g0Inv*g0) is the relation
      // to be satisfied by the expression in brackets
}

RefSys RefSys::per(const Word& p)const
{
   static const R fac3=1./sqrt(3.);

   if (p=="123") return *this;
   R angleDeg;
   Vec axis;
   if (p=="231"){
      axis=(e1()+e2()+e3())*fac3;
      angleDeg=120.;
   }
   else if (p=="312"){
      axis=(e1()+e2()+e3())*fac3;
      angleDeg=-120.;
   }
   else if (p=="1-32"){ // transposition 3<-->2, 1 fixed
      axis=e1();
      angleDeg=90.;
   }
   else if (p=="32-1"){ // transposition 3<-->1, 2 fixed
      axis=e2();
      angleDeg=90.;
   }
   else if (p=="-213"){// transposition 1<-->2, 3 fixed
      axis=e3();
      angleDeg=90.;
   }
   else{
      Word mess="RefSys::per(...) argument ";
      cpmerror(mess&p&" is not valid");
   }
   Angle alpha(angleDeg);
   AxVec av(alpha,axis);
   Group grot(x0,av);
   RefSys res=*this;
   return res*=grot;
}

AxVec::AxVec(const R3& a, const RefSys& rs)
{
   x_.x1=a[1]*(rs.v1)[1]+a[2]*(rs.v2)[1]+a[3]*(rs.v3)[1];
   x_.x2=a[1]*(rs.v1)[2]+a[2]*(rs.v2)[2]+a[3]*(rs.v3)[2];
   x_.x3=a[1]*(rs.v1)[3]+a[2]*(rs.v2)[3]+a[3]*(rs.v3)[3];
}

Vec::Vec(const R3& a, const RefSys& rs)
{
   *this=a[1]*rs.v1+a[2]*rs.v2+a[3]*rs.v3;
}

Spc::Spc(const R3& x, const RefSys& rs)
{
    *this=rs.x0+x[1]*rs.v1+x[2]*rs.v2+x[3]*rs.v3;
}

Spc::Spc(const V<R>& x, const RefSys& rs)
{
   R3 a(x);
   *this=rs.x0+a[1]*rs.v1+a[2]*rs.v2+a[3]*rs.v3;
}

RefSys& RefSys::rotByEulAng(const Angle& phi, const Angle& theta,
         const Angle& psi)
{
   AxVec phiv3(phi,v3);
   operator *=(Group(x0,phiv3));
   AxVec thetav1(theta,v1);
   operator *=(Group(x0,thetav1));
   AxVec psiv3(psi,v3);
   return operator *=(Group(x0,psiv3));
}

RefSys& RefSys::rot(const Angle& phi, Z i)
{
   AxVec phiv(phi,getVectors()[i]);
   return operator *=(Group(x0,phiv));
}

//////////////////////// class Plane ///////////////////////////////////

Plane::Plane(){}

Plane::Plane(const Vec& n, const Spc& x)
{
   Spc x0;
   Vec a=x-x0;
   Vec e3(0,0,1);
   AxVec A=e3.rotTo(n);
   Group g(A,a);
   rs=RefSys(g);
}

Word Plane::des()const
{
   ostringstream ost;
   ost<<endl<<"Description of a instance of Plane:";
   ost<<endl<<"Origin is"<<rs.o();
   ost<<endl<<"e1() is"<<e1();
   ost<<endl<<"n() is"<<n();
   string res=ost.str();
   return Word(res);
}

////////////////////////// class InTens ////////////////////////////////

InTens& InTens::operator *=(const AxVec& a )
// The definition is such that (I*a)*(x*a)=(I*x)*a
{
   eigVec.x1=actionOfRotationVectors(eigVec.x1,a());
   eigVec.x2=actionOfRotationVectors(eigVec.x2,a());
   eigVec.x3=actionOfRotationVectors(eigVec.x3,a());
   return *this;
}

InTens& InTens::ref_(Vec const& n)
{
   eigVec.x1=e1().ref(n)();
   eigVec.x2=e2().ref(n)();
   eigVec.x3=e3().ref(n)();
   return *this;
}

AxVec InTens::operator * (const AxVec& a )const
{
   R a1=eigVal.x1*(eigVec.x1|a());
   R a2=eigVal.x2*(eigVec.x2|a());
   R a3=eigVal.x3*(eigVec.x3|a());
   R3 res=eigVec.x1*a1+eigVec.x2*a2+eigVec.x3*a3;
   return AxVec(res);
}

InTens InTens::operator !()const
{
   InTens res=*this;
   res.eigVal.x1=R(1)/eigVal.x1;
   res.eigVal.x2=R(1)/eigVal.x2;
   res.eigVal.x3=R(1)/eigVal.x3;
   return res;
}

InTens rotate(const InTens& I, const AxVec& a )
// far from optimum at least for symmetric I's
{
   InTens res=I;
   return res*=a;
}

InTens::InTens(const RR3& g):eigVec()
// construction from a matrix representation
{
   Z mL=3;
   Word loc("InTens(const RR3&)");
   CPM_MA
   bool testing=false;
      // no verbose testing
   using CpmArrays::R_Matrix;
   R_Matrix s(3,3);
   R_Matrix eigenVectors(3,3);
   Z i,j;
   for (i=1;i<=3;i++) for (j=1;j<=3;j++) s[i][j]=g[i][j];
   s=0.5*(s+s.con());
   R_Vector d=eigen(s, eigenVectors, testing);
   cpmassert(d[3]>0,loc);
   cpmassert(d[1]>=d[2],loc);
   cpmassert(d[2]>=d[3],loc);
   for (i=1;i<=3;i++) eigVal[i]=d[i];
   for (i=1;i<=3;i++) for (j=1;j<=3;j++) eigVec[i][j]=eigenVectors[i][j];
   CPM_MZ
}

R3 InTens::volume(R m)const
   // Let res be the return value. Then res[1], res[2], res[3] are
   // the sites of a homogeneous Quader of total mass m which has the
   // inertial tensor *this (res[i] is along axis eigVec[i] i=1,2,3)
   // see my pages UM/Ariton 99-6-8/4-4
{
   Word loc("InTens::volume(R)");
   cpmassert(m>0,loc);
   R ev1=eigVal[1];
   R ev2=eigVal[2];
   R ev3=eigVal[3];
   R a=-ev1+ev2+ev3;
   cpmassert(a>0,loc);
   R b= ev1-ev2+ev3;
   cpmassert(b>0,loc);
   R c= ev1+ev2-ev3;
   cpmassert(c>0,loc);
   R fac=R(6)/m;
   a*=fac;
   b*=fac;
   c*=fac;
   a=cpmsqrt(a);
   b=cpmsqrt(b);
   c=cpmsqrt(c);
   return R3(a,b,c);
}

void InTens::balance()
{
   R eps=eigVal.polDec().first();
   eps/=3.;
   eigVal.set(eps);
}

//////////////// I/O for all classes ///////////////////////////////////

bool Spc::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Spc",out);
   return p.prnOn(out);
}

bool Spc::scanFrom(istream& in)
{
   return p.scanFrom(in);
}

bool Group::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Group",out);
   return (A.prnOn(out)&&a.prnOn(out));
}

bool Group::scanFrom(istream& in)
{
   return(A.scanFrom(in)&&a.scanFrom(in));
}

bool RefSys::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("RefSys",out);
   return g0.prnOn(out);
}

bool RefSys::scanFrom(istream& in)
{
   bool res=g0.scanFrom(in);
   update();
   return res;
}

bool Plane::prnOn(ostream& out)const
{
   CpmRoot::writeTitle("Plane",out);
   return rs.prnOn(out);
}

bool Plane::scanFrom(istream& in)
{
   return rs.scanFrom(in);
}

bool InTens::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(eigVal);
   cpmp(eigVec);
   return true;
}

bool InTens::scanFrom(istream& str)
{
   cpms(eigVal);
   cpms(eigVec);
   return true;
}

bool LieAlg::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(A_);
   cpmp(a_);
   return true;
}

bool LieAlg::scanFrom(istream& str)
{
   cpms(A_);
   cpms(a_);
   return true;
}

