//? cpmdimxcpp.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   Description: dimension-independent part of implementation
   of geometrical objects
*/
using namespace CpmFunctions;
using namespace CpmRoot;
using namespace CpmStd;
using CpmGeo::Iv;

// symmetric version
// second order, seems to be as accurate as fourth order in my
// applications

Vec grad(Spc const& x, const F<Spc,R>& f, const R& h0)
{
   //const R hTiny=1e-12;
   const R hTiny=1e-9;
   R h=(h0>=hTiny ? h0 : hTiny);
   Z n=x.dim();
   Vec res;
   R hInv=R(0.5)/h;
   for (Z i=1;i<=n;i++){
      Spc a=x;
      Spc b=x;
      a[i]+=h; // shifted on i-axis
      b[i]-=h;
      res[i]=(f(a)-f(b))*hInv;
   }
   return res;
}
// symmetric version 4th order
// elegant employment of function type, presently not used

/*
namespace{
   R gradFunc(R const& x, Z const& i, Spc const& p, F<Spc,R> const& f)
   {
      Vec h;
      h[i]=x;
      return f(p+h);
   }
}

Vec grad(Spc const& p, F<Spc,R> const& f, R const& h0)
{
   Vec res;
   Z n=res.dim();
   for (Z i=1;i<=n;++i){
      R_Func g=F3<R, Z,Spc,F<Spc,R>, R>(i,p,f)(gradFunc);
      res[i]=g.der(1,h0)(0.);
   }
   return res;
}
*/

//////////////////////////////// class AxVec /////////////////////////////

AxVec& AxVec::ref_(Plane const& p)
{
   return AxVec::ref_(p.n());
}

////////////////////// class Vec ///////////////////////////////////////

Vec& Vec::ref_(Plane const& p){ return ref_(p.n());}

Vec Vec::randomize(R dv, Angle const& a, Z j)const
{
   Vec res=*this;
   if (dv!=0.){ // then the result should have a length
      // in the interval [max(0,|v|-|dv|),|v|+|dv|]
      R avL=0,avU=0;
      R adv=cpmabs(dv); // |dv|
      R av=res.nor_();
      if (av>0.){
         avU=av+adv;
         avL=av-adv;
         if (avL<0) avL=0.;
      }
      else{
         res=Vec::ranUnitVec(j++);
         avU=adv; // notice avL=0
      }
      // now res is a unit-vector
      Iv iv(avL,avU);
      av=iv.ran(j++);
      res*=av;
   }
   AxVec ax(a);
   ax=ax.ran(j);
   res*=Group(ax);
   return res;
}

bool Vec::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(x_);
   return true;
}

bool Vec::scanFrom(istream& str)
{
   cpms(x_);
   return true;
}

Z Vec::com(Vec const& v)const
{ return x_.com(v.x_);}

////////////////////// class RefSys ///////////////////////////////////

RefSys::RefSys(const V<Spc>& vs)
{
   Z mL=3;
   Word loc("RefSys::RefSys(const V<Spc>& vs)");
   CPM_MA
   Z d=vs.dim();
   Z ns=orig.dim();
   R a1,a2,a3;
   bool b1,b2,b3,b4;

   b1=(d>=ns);
   cpmassert(d>=ns,loc);
   AxVec phi;
   if (!b1) goto trivial;
   x0=vs[1];
   v1=vs[2]-x0;
   a1=v1.nor_();
   b2=(a1>0);
   cpmassert(a1>0,loc);
   if (!b2) goto trivial;

   if (ns==2){
      v2=v1*AxVec(Angle(90.,DEG));
      phi=u1.rotTo(v1);
   }
   else{
      v2=vs[3]-x0;
      a2=v2.nor_();
      b3=(a2>0);
      cpmassert(a2>0,loc);
      if (!b3) goto trivial;
      Vec v2Ortho=v1.tanNormDec(v2).second();
      a3=v2Ortho.nor_();
      b4=(a3>0);
      cpmassert(a3>0,loc);
      if (!b4) goto trivial;
      Vec v2=v2Ortho;
      AxVec v3_=v1.cross(v2);
      AxVec u3_=u1.cross(u2);
      Vec v3;
      Vec w3; // in order not to re-declare u3 which is
         // already defined for ns==3; however, the compiler does
         // not know that we come here only in this case
      for (Z i=1;i<=ns;++i){
         v3[i]=v3_[i];
         w3[i]=u3_[i];
      }
      V<Vec> tripodMaster{u1,u2,w3};
      V<Vec> tripodThis{v1,v2,v3};
      bool careful=false;
      phi=AxVec(tripodMaster,tripodThis,careful);
   }
   g0=Group(phi,x0-orig);
   update();
   CPM_MZ
   return;
trivial:
   cpmwarning("Trivial return in "&loc);
   *this=RefSys();
}

/////////////////////// class Group ///////////////////////////////////

Z Group::com(const Group& s)const
{
   if (A<s.A) return 1;
   if (A>s.A) return -1;
   if (a<s.a) return 1;
   if (a>s.a) return -1;
   return 0;
}

/////////////////////// class Spc ///////////////////////////////////

Spc& Spc::ref_(Plane const& p){ return ref_(p.n(),p.o());}

Z Spc::com(const Spc& s)const
{
   return p.com(s.p);
}

X4<Spc,R,R,R> Spc::meet(Spc const& p1, Vec const& v1,
         Spc const& p2, Vec const& v2)
{
   R v11=(v1|v1);
   R v12=(v1|v2); // v21=v12, not needed
   R v22=(v2|v2);
   Vec diff=p1-p2;
   R a1=(diff|v1);
   R a2=(diff|v2);
   R_Matrix m(2,2);
   m[1][1]=-v11;
   m[1][2]=v12;
   m[2][1]=-v12;
   m[2][2]=v22;
   R_Vector a(2);
   a[1]=a1;
   a[2]=a2;
   R_Vector sol=(!m)*a;
   R s1=sol[1], s2=sol[2];
   Spc q1=p1+v1*s1;
   Spc q2=p2+v2*s2;
   R dev=(q1-q2).abs()*0.5;
 //  cpmdebug(dev);
   return X4<Spc,R,R,R>(q1.cen(q2),dev,s1,s2);
}

/////////////////////// class RefSys ////////////////////////////////////

namespace{

   Rdim srFunc(const Spc& x, const RefSys& rs){ return rs(x);}
   Spc rsFunc(const Rdim& x, const RefSys& rs){ return rs(x);}

}

F<Spc,Rdim> RefSys::toFSR()const
{
#ifdef CPM_Fn
   return F1<Spc,RefSys,Rdim>(*this)(srFunc);
#else
   return F<Spc,Rdim>(bind(srFunc,_1,*this));
#endif
}

F<Rdim,Spc> RefSys::toFRS()const
{
#ifdef CPM_Fn
   return F1<Rdim,RefSys,Spc>(*this)(rsFunc);
#else
   return F<Rdim,Spc>(bind(rsFunc,_1,*this));
#endif

}

/////////////////////// BndSph ////////////////////////////////

BndSph BndSph::operator|(const BndSph& bs)const
{
   Vec n=bs.x-x;
   R d=n.nor_();
   R rMax=CpmRootX::sup<R>(r,bs.r);
   if (d==0) return BndSph(x,rMax);
   Iv iv1(-r,r,true);
   Iv iv2(d-bs.r,d+bs.r,true);
   Iv iv=iv1|iv2;
   R rRes=iv.radius();
   Spc xRes=x+n*iv.mean();
   return BndSph(xRes,rRes);
}

BndSph BndSph::operator&(const BndSph& bs)const
{
   Vec n=bs.x-x;
   R d=n.nor_();
   R r1=r;
   R r2=bs.r;
   if (d>=r1+r2){ // void result
      return BndSph(); // 'exit 1'
   }
   else if (d>=r1 && d>=r2){ // section is a biconvex lens
      R rRes=cpmsqrt((-d+r1+r2)*(d-r1+r2)*(d+r1-r2)*(d+r1+r2))/(d*2);
         // notice: d>0 since d==0 here, implies r1==r2==0.
         // In this case the first exit has been taken.
         // All factors under the root are >=0:
         // the first since exit 1 would have been taken else
         // the second since d>=r1
         // the third since d>=r2 and the fourth obviously.
         // rRes is found from the obvious relation
         // d = sqrt(r1^2-rRes^2)+sqrt(r2^2-rRes^2)
         // See my Mathematica file boundingshere3.nb.
      R aRes=cpmsqrt(r1*r1-rRes*rRes);
      Spc xRes=x+n*aRes;
      return BndSph(xRes,rRes);
   }
   else{  // result contains a hemi-sphere
      if (r1<r2) return *this;
      else return bs;
   }
}

namespace{

   R sqrtS(R x) // safe form of function sqrt
      // the following algorithm failed frequently when working on
      // data read from file (limited accuracy !) by feeding
      // slightly negative arguments ( ~ -4e-16) into sqrt.
      // Here we treat such values as 0

   {
      const Z mL=3;
      const R limit=-1e-12;
      if (x>=0) return cpmsqrt(x);
      else if (x>=limit){
         if (cpmverbose>=mL){
            Word mes="sqrtS(R): 0>x>=limit, x="&cpmwrite(x);
            cpmmessage(mes);
         }
         return 0;
      }
      else{
         Word mes="sqrtS(R): limit>x, x="&cpmwrite(x);
         cpmerror(mes);
         return 0;
      }
   }
}

Iv BndSph::hit(const Spc& x0, const Vec& n)const
// n unit vector
{
   using CpmAlgorithms::ssqe;
      // positive solutions of quadratic equations
   if (r==0){
      cpmwarning("Iv BndSph::hit(...): unexpected r==0");
      return Iv();
   }
   Vec v=x-x0;
   R d=v.abs();
   if (d==0) return Iv(0,r);
// now d>0
   R sp=(n|v);
   R d1=d-r;
   if (d1>0){ // x0 outside *this
         // here analytic geometry works simpler than trigonometry
      if (sp<=0) return Iv(); // no chance for contact
         // dInv not needed in this case
      R dInv=R(1)/d;
      R cs=sp*dInv; // 0<cs<=1
      R si=sqrtS(-cs*cs+1);
      if (si>=r*dInv) return Iv();
         // now cs farer away from 0
      R csInv=R(1)/cs;
      R tg=csInv*sqrtS(-cs*cs+1); // sin/cos
// Now we solve y=tg*x and y^2+(x-d)^2=r^2 for x
      R a=tg*tg+1;
      R b=-d*2;
      R d2=d+r;
      R c=d1*d2;
      R x1=ssqe(a,b,c,d1,d2);
      R x2=c/(a*x1); // Vieta
      return Iv(x1*csInv,x2*csInv,true);
   }
   else{ // x0 inside, using sine theorem and cosine theorem of
         // trigonometry implies 3 transcendental function calls
      R cs=sp/d;
      R siAlpha=sqrtS(-cs*cs+1);
      R siGamma=d*siAlpha/r; // sine theorem; notice d<r
      R beta=cpmpi-cpmasin(siAlpha)-cpmasin(siGamma);
      R t=sqrtS(d*d+r*r-R(2)*d*r*cpmcos(beta));
      return Iv(0,t,true);
   }
}

X2<V<Spc>,Z> BndSph::ran(Z n, const F<Spc,bool>& cond, Z tryMax)const
{
   V<Spc> res(n);
   Z ir=137;
   using CpmRoot::ran;
   Rdim a;
   Spc xTest;
   Z d=Spc::dimension();
   Z iTested=0;
   Z iFound=0;
   R a0=1;
   while(iFound<n && iTested<tryMax){
      bool aValid=false;
      while (!aValid){
         for (Z j=1;j<=d;j++){
            a[j]=ran(a0,ir++);
               //
               // between -1 and 1
         }
         aValid=(a.abs()<1);
      }
      xTest=x+r*Vec(a);

      if (cond(xTest)){
         iFound++;
         res[iFound]=xTest;
      }
      iTested++;
   }
   if (iFound<n) res=res.resize(iFound);
   return X2<V<Spc>,Z>(res,iTested);
}

X2<Z,Z> BndSph::ranLean(Z n, const F<Spc,bool>& cond, Z tryMax)const
// doublicated code from previous function, all statements refering to
// array res are eliminated, return structure modified.
// Keep in touch with ran, if this should be modified!!!
{
   Z ir=137;
   using CpmRoot::ran;
   Rdim a;
   Spc xTest;
   Z d=Spc::dimension();
   Z iTested=0;
   Z iFound=0;
   R a0=1;
   while(iFound<n && iTested<tryMax){
      bool aValid=false;
      while (!aValid){
         for (Z j=1;j<=d;j++){
            a[j]=ran(a0,ir++);
               // between -1 and 1
         }
         aValid=(a.abs()<1);
      }
      xTest=x+r*Vec(a);
      if (cond(xTest)==true){
         iFound++;
      }
      iTested++;
   }
   // at the end either iTested==tryMax or iFound==n
   return X2<Z,Z>(iFound,iTested);
}

////////////////// struct DisPar ////////////////////////////////

bool DisPar::prnOn(ostream& str)const
{
   cpmwt("DisPar");
   cpmp(dr);
   cpmp(ds);
   cpmp(nFind);
   cpmp(nMax);
   return true;
}

bool DisPar::scanFrom(istream& str)
{
   cpms(dr);
   cpms(ds);
   cpms(nFind);
   cpms(nMax);
   return true;
}

//////////////////// SubsetOfSpace ///////////////////////////////

namespace{ // some tools

 Rdim fss(const Rdim& x, const F<Spc,Spc>& f, const RefSys& rs)
   {
      return rs(f(rs(x)));
   }

 R fsr(const Rdim& x, const F<Spc,R>& f, const RefSys& rs)
   { return f(rs(x));}

 R tf(const Spc& x, const F<Rdim,R>& fr, const RefSys& rs)
   {
      return fr(rs(x));
   }

 R supFunc(const Rdim& x, const F<Spc,R>& f1, const F<Spc,R>& f2)
   {
      Spc sp(x);
      return CpmRootX::sup(f1(sp),f2(sp));
   }

 R infFunc(const Rdim& x, const F<Spc,R>& f1, const F<Spc,R>& f2)
   {
      Spc sp(x);
      return CpmRootX::inf(f1(sp),f2(sp));
   }

 R comFunc(const Rdim& x, const F<Spc,R>& f1, const F<Spc,R>& f2)
   {
      Spc sp(x);
      return CpmRootX::sup(f1(sp),-f2(sp));
   }

 R symDiffFunc(const Rdim& x, const F<Spc,R>& f1, const F<Spc,R>& f2)
   {
      Spc sp(x);
      R f1s=f1(sp);
      R f2s=f2(sp);
      R val1=CpmRootX::sup(f1s,-f2s);
      R val2=CpmRootX::sup(-f1s,f2s);
      return CpmRootX::inf(val1,val2);
   }

} // namespace

SubsetOfSpace::SubsetOfSpace(const F<Spc,R>& f_, const RefSys& rs_,
   const Spc& x_, R r_, const DisPar& dp_):
   f(f_),rs(rs_),x(x_),r(r_),dp(dp_)
{
#ifdef CPM_Fn
   fr=F2<Rdim,F<Spc,R>,RefSys,R>(f,rs)(fsr);
#else
   fr=F<Rdim,R>(bind(fsr,_1,f,rs));
#endif
}

void SubsetOfSpace::update()
{
#ifdef CPM_Fn
   f=F2<Spc,F<Rdim,R>,RefSys,R>(fr,rs)(tf);
#else
   f=F<Spc,R>(bind(tf,_1,fr,rs));
#endif
}

SubsetOfSpace SubsetOfSpace::operator &(const SubsetOfSpace& sos)const
{
   BndSph bs(x,r);
   BndSph bssos(sos.x,sos.r);
   BndSph res=bs&bssos;
#ifdef CPM_Fn
   F<Rdim,R> f1r=F2<Rdim,F<Spc,R>,F<Spc,R>,R>(f,sos.f)(supFunc);
#else
   F<Rdim,R> f1r(bind(supFunc,_1,f,sos.f));
#endif
   return SubsetOfSpace(f1r,RefSys(),res.get_x(),res.get_r());
}

SubsetOfSpace SubsetOfSpace::operator |(const SubsetOfSpace& sos)const
{
   BndSph bs(x,r);
   BndSph bssos(sos.x,sos.r);
   BndSph res=bs|bssos;
#ifdef CPM_Fn
   F<Rdim,R> f1r=F2<Rdim,F<Spc,R>,F<Spc,R>,R>(f,sos.f)(infFunc);
#else
   F<Rdim,R> f1r(bind(infFunc,_1,f,sos.f));
#endif
   return SubsetOfSpace(f1r,RefSys(),res.get_x(),res.get_r());
}

SubsetOfSpace SubsetOfSpace::operator ^(const SubsetOfSpace& sos)const
{
   BndSph bs(x,r);
   BndSph bssos(sos.x,sos.r);
   BndSph res=bs|bssos;
#ifdef CPM_Fn
   F<Rdim,R> f1r=F2<Rdim,F<Spc,R>,F<Spc,R>,R>(f,sos.f)(symDiffFunc);
#else
   F<Rdim,R> f1r(bind(symDiffFunc,_1,f,sos.f));
#endif
   return SubsetOfSpace(f1r,RefSys(),res.get_x(),res.get_r());
}

SubsetOfSpace SubsetOfSpace::operator -(const SubsetOfSpace& sos)const
      // forming *this\sos i.e. the set of all those elements of
      // *this which don't belong to sos (relative complement)
{
#ifdef CPM_Fn
   F<Rdim,R> f1r=F2<Rdim,F<Spc,R>,F<Spc,R>,R>(f,sos.f)(comFunc);
#else
   F<Rdim,R> f1r(bind(comFunc,_1,f,sos.f));
#endif
   return SubsetOfSpace(f1r,RefSys(),x,r);
}

SubsetOfSpace SubsetOfSpace::combine(const V<SubsetOfSpace>& vs)const
         // here we consider *this as a region of interest and return a
         // body which is formed out of all intersections vs[i]&*this
         // Those terms for which this intersection is void, are left
         // out of the result for efficiency.
{
   Z d=vs.dim();
   V<B> vb(d);
   V<SubsetOfSpace> vsi(d);
   for (Z i=1;i<=d;i++){
      SubsetOfSpace soi=vs[i]&(*this);
      soi.improveBoundingSphere();
         // to see wether it is void, this can't be seen from the
         // defining function directly
      vsi[i]=soi;
      vb[i]=(soi.get_r()>R(0.));
   }
   V<SubsetOfSpace> res=vsi.select(vb);
   return unionOfArray(res);
}

SubsetOfSpace unionOfArray(const V<SubsetOfSpace>& vs)
{
   Z i,d=vs.dim();
   if (d<1) return SubsetOfSpace(); // void set
   SubsetOfSpace res=vs[1];
   for (i=2;i<=d;i++) res=res|vs[i];
   return res;
}

SubsetOfSpace sectionOfArray(const V<SubsetOfSpace>& vs)
{
   Z i,d=vs.dim();
   if (d<1) return SubsetOfSpace();
   SubsetOfSpace res=vs[1];
   for (i=2;i<=d;i++) res=res&vs[i];
   return res;
}

void SubsetOfSpace::operator *=(const Group& g)
{
   rs*=g;
   x*=g;
   update();
}

void SubsetOfSpace::operator *=(const F<Spc,Spc>& f_)
{
   f.leftCombine(f_);
// fr_: action of f_ on coordinates with respect to rs
#ifdef CPM_Fn
   F<Rdim,Rdim> fr_=F2<Rdim,F<Spc,Spc>,RefSys,Rdim>(f_,rs)(fss);
#else
   F<Rdim,Rdim> fr_(bind(fss,_1,f_,rs));
#endif
   fr.leftCombine(fr_);
      // function update would only set f according to fr
      // in our case setting directly is much cheaper
}

namespace{

   bool fComb(const R& t, const Spc& x0, const Vec& v,
      const F<Spc,R>& f, const Z& k)
      // the most natural implementation turns out to be the fastest!
      // So, no reason to avoid the semantic complexity to have a
      // F<>-object as an argument of a function to be used for creating
      // another F<>-object.
   {
      return R(k)*f(x0+t*v)>0;
   }

    R fComb2(const R& t, const Spc& x0, const Vec& v,
      const F<Spc,R>& f)
      // the most natural implementation turns out to be the fastest!
      // So, no reason to avoid the semantic complexity to have a
      // F<>-object as an argument of a function to be used for creating
      // another F<>-object.
   {
      return f(x0+t*v);
   }

} // namespace

namespace{
   Spc ray(const R& r, const Spc& x0, const Vec& v)
   { return x0+r*v;}

   bool fInterior(R x){ return x<0;}
}

X2<Spc,bool> SubsetOfSpace::hit(const Spc& xi, const Vec& v,
                                   R sFine, R sCoarse)const
{
   Z mL=3;
   Word loc("SubsetOfSpace::hit(Spc,Vec,R,R)");
   CPM_MA
   cpmassert(sFine>0,loc);
   cpmassert(sCoarse>=sFine,loc);
//   cpmassert(r>0,loc);
   cpmdebug(r);
   BndSph bs(get_x(),get_r());
   Iv iv=bs.hit(xi,v);
   if (iv.isVoid()){
      cpmdata<<endl<<"iv.isVoid(), trivial result returned";
      return X2<Spc,bool>(Spc(),false);
   }
   R xMin=iv.inf()-sFine;
   R xMax=iv.sup()+sCoarse+sFine;
      // not understood why '+sCoarse' is not always sufficient
      // numerics???
#ifdef CPM_Fn
   F<R,Spc> fs=F2<R,Spc,Vec,Spc>(xi,v)(ray);
#else
   F<R,Spc> fs(bind(ray,_1,xi,v));
#endif
   bool silent=true;
   X2<R,bool> prr=CpmAlgorithms::root(fs&f,xMin,xMax,sFine,sCoarse,
      silent);
      // for silent==false one would get always warnings if *this
      // would not be in the field of view of the object providing
      // xi and v
   Spc xLast=ray(prr.c1(),xi,v);
   CPM_MZ
   return X2<Spc,bool>(xLast,prr.c2());
}

R SubsetOfSpace::vol()const
{
   Z nMax=dp.nMax;
   Z nFind=dp.nFind;
   BndSph bs(x,r);
   X2<Z,Z> pzz=bs.ranLean(nFind,f&F<R,bool>(fInterior),nMax);
   return (bs.vol()*pzz.first())/pzz.second();
}

V<Spc> SubsetOfSpace::ran()const
{
   Z nMax=dp.nMax;
   Z nFind=dp.nFind;
   BndSph bs(x,r);
   X2<V<Spc>,Z> pvz=bs.ran(nFind,f&F<R,bool>(fInterior),nMax);
      // here we use limitation by nLarge
   return pvz.c1();
}

R SubsetOfSpace::improveBoundingSphere(Z n,Z tryMax)
{
   const R saftyFactor=1.1;
      // muliplies the radius that we found. Takes into account that
      // by definition, the true radius is >= the one we can find by the
      // following method
   using CpmArrays::Vo;
   BndSph bs(x,r);
   X2<V<Spc>,Z> pvz=bs.ran(n,f&F<R,bool>(fInterior),tryMax);
   V<Spc> xa=pvz.c1(); // x-array
   Z m=xa.dim();
   if (m==0){
      r=0; return 0;
         // no point found which belongs to the set *this.
   }
   R ret=((R)(m))/pvz.c2();
   Z mp=m*(m-1);
   mp/=2;
   Vo<R> dis(mp);
   Z jp=1,i,j;
   for (i=1;i<=m;i++){
      for (j=i+1;j<=m;j++){
         dis[jp++]=(xa[i]-xa[j]).abs();
      }
   }
   Z jpMax=dis.indInfSup().c2();
   r=dis[jpMax]*saftyFactor*0.5;
// retrieving i and j in a simple and safe manner. The obvious lack
// of utmost efficiency dosn't matter compared to the workload
// of the other operations
   jp=1;
   for (i=1;i<=m;i++){
      for (j=i+1;j<=m;j++){
         if (jp==jpMax) goto done;
         jp++;
      }
   }
done:  // now we know i and j belonging to jp
   V<Spc> xp{xa[i],xa[j]};
   V<R> wp{0.5,0.5};
   x=Spc::convex(xp,wp,false);
   return ret;
}

Plane SubsetOfSpace::tangentialPlane(const Spc& x)const
{
   R d=f(x);
   Vec n=grad(x,f,dp.dr);
   Spc xBase=x-n*d;
   return Plane(n,xBase);
}

X2<R,Vec> SubsetOfSpace::distAndDirection(const Spc& x)const
// Re-implementation 2006-02-15 in order to avoid that the axis of
// a Drum is seen by particles as an elastic obstacle
{
   Word loc("SubsetOfSpace::distAndDirection(Spc)");
   const R hDef=1e-6; // should never be used, using it
   // creates a warning and a failing assert
   const R gradLengthLimit=3.;
 //  static Z nWarn=0;

   R h=dp.dr;
   if (h==0) h=dp.ds;
   cpmassert(h>0,loc);
   if (h==0){
      h=hDef;
      cpmwarning(loc&": dp.dr==0, dp.ds==0, thus h set to "&cpm(h));
   }
   Vec n=grad(x,f,h);
   R gradLength=n.nor_();
   Spc xTest=x;
   Z wc=0;
   Z iTest=1;
   Z d; V<Vec> b;
   while (gradLength==0.){ // searching a point nearby
      // for which the grad is not 0
      wc++;
    //  if (nWarn<100){
    //     nWarn++;
         cpmwarning(loc&": gradLength==0, whileCount="&cpm(wc));
   //   }
      if (wc==1){
         Vec v;
         v[1]=1;
         b=v.base();
         d=b.dim();
      }
      Spc p=xTest+b[iTest]*h;
      if (iTest==d){
         iTest=1;
         xTest=p;
      }
      else{
         iTest++;
      }
      n=grad(p,f,h);
      gradLength=n.nor_();
   }
   if (gradLength>gradLengthLimit) gradLength=gradLengthLimit;
      // if gradLength is very large, the value of
      // f(x)*cpminv(gradLength) is very small and thus simulates
      // a point near the surface, which might not be correct
   return X2<R,Vec>(f(x)*cpminv(gradLength),n);
}

void SubsetOfSpace::shiftExternalPointToSurface4(Spc& x, R& dOut,
   R& fxBefore, R& fxAfter)const
{
   fxBefore=f(x);
   if (fxBefore<=0){ // then the point is internal
      dOut=0;
      fxAfter=fxBefore;
   }
   else{
      X2<R,Vec> prv=distAndDirection(x);
      dOut=prv.c1();
      Vec shift=prv.c2()*(-dOut);
      x+=shift;
      fxAfter=f(x);
   }
}

bool SubsetOfSpace::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("SubsetOfSpace",str)) return false;
   if (!rs.prnOn(str)) return false;
   if (!x.prnOn(str)) return false;
   if (!Root<R>(r).prnOn(str)) return false;
   if (!dp.prnOn(str)) return false;
   return true;
}

namespace{

   R bandFunc(const Rdim& x, const Z& i, const Iv& iv){ return iv(x[i]);}

   R boxFunc(const Spc& x, const Rdim x0, const Rdim& s0,
      const F<Spc,Rdim>& f )
   {
      Rdim y=f(x);
      return SubsetOfSpace::box(y,x0,s0);
   }

   R sphereFunc(Spc const& p, Spc const& p0, R const& r)
   { return (p-p0).abs()-r;}
}

R SubsetOfSpace::box(const Rdim& x, const Rdim& x0, const Rdim& s0)
{
   Z d=Spc::dimension();
   Rdim val;
   for (Z i=1;i<=d;i++){
      R hw=cpmabs(s0[i]);
      R c=x0[i];
      Iv iv(c-hw,c+hw);
      val[i]=iv(x[i]);
   }
   return val.sup();
}

// static generating function
SubsetOfSpace SubsetOfSpace::band(Z i, const Iv& iv, const RefSys& rf)
{
#ifdef CPM_Fn
   F<Rdim,R> f=F2<Rdim,Z,Iv,R>(i,iv)(bandFunc);
#else
   F<Rdim,R> f(bind(bandFunc,_1,i,iv));
#endif
   R r0=iv.abs()*3;
   return SubsetOfSpace(f,rf,rf.o(),r0);
}

SubsetOfSpace SubsetOfSpace::boxoid(const Rdim& x0, const Rdim& s0,
   const F<Spc,Rdim>& f )
{
   const R tol=1e-5;
   const R sec=4; // security
   Z mL=3;
   Word loc("SubsetOfSpace::boxoid(...)");
   CPM_MA
#ifdef CPM_Fn
   F<Rdim,R> f1=F2<Rdim,Rdim,Rdim,R>(x0,s0)(box);
#else
   F<Rdim,R> f1(bind(box,_1,x0,s0));
#endif
   F<Spc,R> f2=f&f1;
   SubsetOfSpace res(f2);
   Z i;
   for (i=1;i<=dimVec;++i){
      if (s0[i]==0.){
         CPM_MZ
         return res;
            // res has radius 0 by definition
      }
   }
// In order to set r properly, we test whether f determines
// fixed magnification factors along the master axes
/******************************/
   Rdim trueAxes;
   V<Rdim> vv=Rdim::canBas();
   RefSys rs=res.getRefSys();
   bool distortion=false;
   for (i=1;i<=dimVec;++i){
      R si=cpmabs(s0[i]);
      cpmassert(si>0,loc);
      R siInv=R(1)/si;
      Rdim v1=vv[i]*si;
      Rdim v2=v1*2;
      Spc p1=rs(v1);
      Spc p2=rs(v2);
      Spc m1=rs(-v1);
      Spc m2=rs(-v2);
      R2 mag;
      mag[1]=f(p1).dist(f(m1))*siInv*0.5;
      mag[2]=f(p2).dist(f(m2))*siInv*0.25;
      R fi=mag.meanValue();
      cpmassert(fi>0,loc);
      trueAxes[i]=si/fi;
      if (mag.mod()>tol) distortion=true;
   }
   R radius=trueAxes.abs();

   if (!distortion){
      res.set_r(radius);
   }
   else{
      res.set_r(sec*radius);
      res.improveBoundingSphere();
   }
/***************************************/
   CPM_MZ
   return res;
}

SubsetOfSpace SubsetOfSpace::sphere(Spc const& x0, R r0, R accRel)
{
#ifdef CPM_Fn
   F<Spc,R> f=F2<Spc,Spc,R,R>(x0,r0)(sphereFunc);
#else
   F<Spc,R> f(bind(sphereFunc,_1,x0,r0));
#endif
   RefSys rs;
   R dis=r0*accRel;
   DisPar dp(dis,dis);
   return SubsetOfSpace(f,rs,x0,r0,dp);
}

///////////////// class Texture /////////////////////////////////////////

namespace {

R fAux1(const Spc& sp, const V<R>& omega, const V<R>& delta,
   const R& p, const Z& method)
{
   Z i,d=Spc::dimension();
   R dInv=1./d;
   R y=0;
   Rdim yy;
   for (i=1;i<=d;i++){
      yy[i]=(cpmcos(delta[i]+sp[i]*omega[i])+1)*0.5;
   }
   if (method==1){ // additive method
      for (i=1;i<=d;i++) y+=yy[i];
      y*=dInv;
   }
   else if (method==2){ // extended additive method
      for (i=1;i<=d;i++) y+=cpmpow(yy[i],p);
      y*=dInv;
      y=cpmpow(y,R(1)/p);
   }
   else { // multiplicative method
      y=1;
      for (i=1;i<=d;i++) y*=yy[i];
   }

   return y-cpmfloor(y);
}

R fAux2(const Spc& sp, const CpmFunctions::R2_Func& r2f, const Iv& iv1,
        const Iv& iv2)
{
   C z(sp[1],sp[2]);
   R a1=iv1.periodic(sp[1]);
   R a2=iv2.periodic(sp[2]);
   return r2f(C(a1,a2));
}

R fAux3(const Spc& sp, const F<C,R>& fCR)
{
   return fCR(C(sp[1],sp[2]));
}

} // anonymous namespace

Texture::Texture(const V<R>& omega, const V<R>& delta, R p, Z method):
#ifdef CPM_Fn
   Base(F4<Spc,V<R>,V<R>,R,Z,R>(omega,delta,p,method)(fAux1))
#else
   Base(F<Spc,R>(bind(fAux1,_1,omega,delta,p,method)))
#endif
{}

Texture::Texture(R width)
{
   R p=1;
   Z method=1;
   R omegaR=cpmpi*cpminv(width)*2;
   R deltaR=0;
   Z d=Spc::dimension();
   V<R> omega(d,omegaR);
   V<R> delta(d,deltaR);
#ifdef CPM_Fn
   *this=
      Type(Base(F4<Spc,V<R>,V<R>,R,Z,R>(omega,delta,p,method)(fAux1)));
#else
   *this=Type(Base(F<Spc,R>(bind(fAux1,_1,omega,delta,p,method))));
#endif
}

Texture::Texture(const VV<R>& mat, Iv xI, Iv yI, Z method):
#ifdef CPM_Fn
Base(F3<Spc,R2_Func,Iv,Iv,R>(R2_Func(mat,method),xI,yI)(fAux2))
#else
Base(F<Spc,R>(bind(fAux2,_1,R2_Func(mat,method),xI,yI)))
#endif
{}

Texture::Texture(const F<C,R>& fIn):
#ifdef CPM_Fn
Base(F1<Spc,F<C,R>,R>(fIn)(fAux3))
#else
Base(F<Spc,R>(bind(fAux3,_1,fIn)))
#endif
{}

/////////////////////////////// class Triangle ///////////////////////////

Triangle& Triangle::operator *=(Group const& g)
{
   p1_*=g;
   p2_*=g;
   p3_*=g;
   return *this;
}

Triangle& Triangle::operator +=(Vec const& a)
{
   p1_+=a;
   p2_+=a;
   p3_+=a;
   return *this;
}

bool Triangle::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(p1_);
   cpmp(p2_);
   cpmp(p3_);
   return true;
}

bool Triangle::scanFrom(istream& str)
{
   cpms(p1_);
   cpms(p2_);
   cpms(p3_);
   return true;
}

namespace{

V<C> circlePoints(R r, Z n)
// generating a regular n-gon ( a polygon, poly=n)
// in the complex plane. Let res be the result, then
// res[1] and res[n] are different; res[1] follows
// res[n] just as res[2] follows res[1]. The case n=2
// is useful for clarification.
{
   R phi=cpmpi*2/R(n); // phi=cpmpi for n==2
   C z1(r,0.);
   C fac(1.,phi,"polar"); //e^(i*phi)
   V<C> p(n);
   p[1]=z1;
   Z i;
   for (i=1;i<n;i++){
      p[i+1]=p[i]*fac;
   }
   return p;
}

} //anon

////////////////////// Path ////////////////////////////////////////

R Path::update()
{
   n_=x_.dim();
   y_=Vo<R>(n_,0.);
   if (n_<=1){
      length_=0;
      l_=V<R>(0);
      e_=V<Vec>(0);
      return 0;
   }
   else { // here n_-1 >=1
      R gap=CpmRootX::hugeNumber;
      length_=0;
      Z n=n_-1;
      l_=V<R>(n);
      e_=V<Vec>(n);
      for (Z i1=1,i2=2;i2<=n_;++i1,++i2){
         Vec ei=x_[i2]-x_[i1];
         R li=ei.nor_();
         if (li<gap) gap=li;
         e_[i1]=ei;
         l_[i1]=li;
         length_+=li;
         y_[i2]=length_;
      }
      return gap;
   }
}

bool Path::isSmooth()const
{
   if (n_<=2) return true;
// now n_>=3
   for (Z i=1;i<=(n_-2);++i){
      R scp=(e_[i]|e_[i+1]);
      if (scp<=0){
         cpmdebug(x_[i]);
         cpmdebug(x_[i+1]);
         cpmdebug(x_[i+2]);
            // to help diagnosis
         return false;
      }
   }
   return true;
}

Z Path::adjust(R acc)
{
   const R reducFac=0.25;
      // should be <0.5. By this factor the smallest inter-point distance
      // is multiplied for obtaining acc
   Z np=x_.dim();
   if (np==0){
      update();
      return 0;
   }
      // then length and n are initialized to 0 anyway
   if (acc>0){
      acc_=acc;
      V<B> sel(np,B(false));
      sel[1]=true;
      length_=0;
      for (Z i=2;i<=np;++i){
         R dis=(x_[i]-x_[i-1]).abs();
         if (dis>=acc_){
            sel[i]=true;
            length_+=dis;
         }
      }
      x_=x_.select(sel);
      R gap=update();
      Z leftOut=np-n_;
      return leftOut;
   }
   else{
      acc_=update();
      acc_*=reducFac;
      return 0; // no point left out
   }
}

namespace{
   Word mesDir(
   "result vector probably too short for defining a useful direction");
}

Vec Path::firstDir()const
{
   Word loc("Path::firstDir()");
   if (n_<2) return Vec();
   else return e_[1];
}

Vec Path::lastDir()const
{
   Word loc("Path::lastDir()");
   if (n_<2) return Vec();
   else return e_[n_-1];
}

Vec Path::dir(Z i)const
{
   Word loc("Path::dir()");
   Vec res;
   if (n_<2 || i<1 || i>n_) return res;
   else if (i==1) return firstDir();
   else if (i==n_) return lastDir();
   else{
      Z ic1=i-1;
      Z ic2=i+1;
      R dis;
again:
      res=x_[ic2]-x_[ic1];
      dis=res.abs();
      if (dis<acc_){
         bool suc=false;
         if (ic1>1){ ic1--; suc=true;}
         if (ic2<n_){ ic2++; suc=true;}
         if (suc==false){ // there are no components available
            cpmwarning(loc,mesDir);
            return res;
         }
         goto again;
      }
      return res;
   }
   return res; // to avoid warning
}

Path& Path::operator *=(Group const& g)
{
   for (Z i=1;i<=n_;i++) x_[i]*=g;
   update(); // may change directions
   return *this;
}

Path& Path::operator +=(Vec const& a)
{
   for (Z i=1;i<=n_;i++) x_[i]+=a;
      // does not change directions
   return *this;
}

Path& Path::operator *=(AxVec const& A)
{
   Spc fix=x_[1];
   for (Z i=2;i<=n_;i++) x_[i].rot_(fix,A);
   update();
   return *this;
}

Path& Path::operator *=(R fac)
{
   Spc fix=x_[1];
   for (Z i=2;i<=n_;i++) x_[i]=x_[i].scl(fix,fac);
   update();
   return *this;
}

Path& Path::rot_(Spc const& fix, AxVec const& A)
{
   for (Z i=1;i<=n_;i++) x_[i].rot_(fix,A);
   update();
   return *this;
}

Path Path::scl(Spc const& fix, R fac)const
{
   V<Spc> x(n_);
   for (Z i=1;i<=n_;i++) x[i]=x_[i].scl(fix,fac);
   return Path(x);
}

V<Path> Path::thc(R r, Z n)const
{
   if (r<=0 || n<=1) return V<Path>{*this};
   Spc o;
   Path disk=Path::circle(o,r,n);
   V<Spc> cir=disk();
   Z nc=cir.dim(); // paranoia, should be n
   V<Vec> vcir(nc);
   Z i;
   for (i=1;i<=nc;++i) vcir[i]=cir[i]-o;
   V<Path> res(nc,*this);
   for (i=1;i<=nc;++i){
      res[i]+=vcir[i];
   }
   return res;
}

V<Path> Path::brkLine(R d, R frac)const
{
   Z mL=3;
   Word loc("Path::brkLine(R,R)");
   CPM_MA
   if (d<=0 || frac<=0 || frac >=1){
      CPM_MZ
      return V<Path>{*this};
   }
   R l=len();
   R dAdj=d;
   Z n=CpmAlgorithms::getNerFrc_1(dAdj,l);
   // now the total length consists of n pieces, each of length dAdj.
   // We get two space points for each of these n pieces which we store
   // for later use.
   R s1=dAdj*(-frac+1)*0.5;
   R s2=dAdj-s1;
   // designed such that s2-s1=dAdj*frac. The interval
   // from s1 to s2 is the part of a segment of length dAdj
   // where the line is invisible (i.e. not drawn)
   cpmassert(s1>0,loc);
   cpmassert(s2>0,loc);
   Z n2=n*2;
   V<Spc> vs(n2);
   R t=0; // arc length parameter for defining proper division points
   Z i,j=1;
   typedef X3<Spc,Vec,Z> x3; // return type of function fromArc
   for (i=1;i<=n;++i){
      R t1=t+s1;
      R t2=t+s2;
      x3 a1=fromArc(t1);
      vs[j++]=a1.c1();
      x3 a2=fromArc(t2);
      vs[j++]=a2.c1();
      t+=dAdj;
   }
   V<Path> res(n+1);
   j=1;
   res[1]=Path(firstPoint(),vs[j]);
   for (i=2;i<=n;++i){
      Spc p1=vs[++j];
      Spc p2=vs[++j];
      res[i]=Path(p1,p2);
   }
   res[n+1]=Path(vs[n2],lastPoint());
   CPM_MZ
   return res;
}

VecGraph Path::flatten()const
{
   V<C> vr2(n_);
   for (Z i=1;i<=n_;i++) vr2[i]=C(x_[i][1],x_[i][2]);
   Z connect=2;
   return VecGraph(vr2,connect);
}

Path Path::arc(const Spc& p1, const Spc& center,
   const AxVec& rot, Z n)
{
   X2<Angle,Vec> pav=rot.toAngAxs();
   Angle phi=pav.c1();
   Vec axis=pav.c2();
   V<Spc> ps(n);
   ps[1]=p1;
   cpmassert(n>1,"Path::arc");
   R nInv=1./(n-1);
   Angle dPhi=phi*nInv;
   Vec v=p1-center;
   Z i;
   for (i=2;i<=n;i++){
      v*=AxVec(dPhi,axis);
      ps[i]=center+v;
   }
   return Path(ps);
}

Path Path::arc2(const Spc& p1, const Spc& aimPoint,
      const AxVec& rot, Z n)
{
   Z mL=3;
   Word loc("Path::arc2(...)");
   CPM_MA
   Path res;
   Angle phi=rot.toAngAxs().c1();
   if (phi==Angle()){
      res=line(p1,aimPoint,n);
   }
   else{
      Vec dir=aimPoint-p1;
      R length=dir.nor_();
      Vec dir2=dir;
      dir2*=rot;
      X2<Vec,Vec> pvv=dir.tanNormDec(dir2);
      Vec orth=pvv.c2();
      orth.nor_();
      R r=cpmabs(length/phi.toR()); // abs important for negative angles
      Spc center=p1+orth*r;
      res=arc(p1,center,rot,n);
   }
   CPM_MZ
   return res;
}

Path::Path(Spc const& p1, Vec const& v1, Spc const& p2,
      Vec const& v2, Z n)
{
   Z mL=3;
   Word loc("Path(Spc,Vec,Spc,Vec,Z)");
   CPM_MA
   X4<Spc,R,R,R> meet=Spc::meet(p1,v1,p2,v2);
   R a1=meet.c3();
   R a2=meet.c4();
   cpmassert(a1>0,loc);
   cpmassert(a2<0,loc);
   Spc cp=meet.c1(); // control point
   *this=bezier(V<Spc>{p1,cp,p2},n);
   CPM_MZ
}

Path::Path(Spc const& p1, Vec const& v1, Spc const& p2,
      Angle const& phi, Z n)
{
   Z mL=3;
   Word loc("Path(Spc,Vec,Spc,AxVec,Z)");
   CPM_MA
   AxVec av=v1.cross(p2-p1);
   av.phi_(phi);
   Vec v2=v1*av;
   *this=Path(p1,v1,p2,v2,n);
   CPM_MZ
}

Path::Path(Spc const& p1, Spc const& p2, Vec const& v2, R s, Z n)
{
   Z mL=3;
   Word loc("Path(Spc,Spc,Vec,R,Z)");
   CPM_MA
   R vv=(v2|v2);
   cpmassert(vv>0,loc);
   Vec dp=p2-p1;
   R pv=(dp|v2);
   R x=pv/vv;
   Spc q=p2-v2*x;
   Spc pc=p2.versus(q,s);
   *this=bezier(V<Spc>{p1,pc,p2},n);
   CPM_MZ
}

Path Path::bezier(V<Spc> const& vs, Z np)
// essentially taken from function Skeleton::bezier
{
   Z mL=3;
   Word loc("Path::bezier(V<Spc>,Z)");
   CPM_MA
   Z i,j,vsd=vs.dim();
   V<R_Vector> vvr(dimVec, R_Vector(vsd));
   for (i=1;i<=dimVec;++i) for (j=1;j<=vsd;++j) vvr[i][j]=vs[j][i];
   V<R_Func> vrf(dimVec);
   R2 iv(0,1);
   for (i=1;i<=dimVec;++i) vrf[i]=R_Func::Bernstein(vvr[i],iv);
   V< Spc > ps(np);
   R du=1./(np-1);
   R u=0;
   for (i=1;i<=np;++i){
      for (j=1;j<=dimVec;++j){
         ps[i][j]=(vrf[j])(u);
      }
      u+=du;
   }
   CPM_MZ
   return Path(ps);
}

Path Path::locInt(V<Spc> const& vs, Z n)
{
   Z mL=3;
   Word loc("Path::locInt(V<Spc>,Z)");
   CPM_MA
   bool d3=(Spc::dimension()==3);
   Z i,vd=vs.dim();
   V<R> x(vd);
   V<R> y(vd);
   V<R> z= d3 ? V<R>(vd) : V<R>(0);
   V<R> r(vd);
   for (i=1;i<=vd;++i){
      Spc pi=vs[i];
      x[i]=pi[1];
      y[i]=pi[2];
      if (d3) z[i]=pi[3];
      r[i]=i;
   }
   V<Spc> res(n);
   cpmassert(n>1,loc);
   R fac=(vd-1.)/(n-1.);
   IpolControl c("second_order");
   for (i=1;i<=n;++i){
      R ic=R(i-1)*fac+1;
      R xi,yi,zi;
      xi=CpmArrays::locIpol(ic,r,x,c);
      yi=CpmArrays::locIpol(ic,r,y,c);
      if (d3) zi=CpmArrays::locIpol(ic,r,z,c);
      Spc pi;
      pi[1]=xi;
      pi[2]=yi;
      if (d3) pi[3]=zi;
      res[i]=pi;
   }
   CPM_MZ
   return Path(res);
}

namespace{
//See Aumann,Spitzmueller:Computerorientierte Geometrie, BI 1993
// p. 330, (11.7) for this formula:
   Spc cub(R const& x, Spc const& p1, Vec const& v1,
      Spc const& p2, Vec const& v2) // 0 <= x <= 1 assumed
   {
      Vec v0=p2-p1;
      R x2=x*x;
      R xm=x-1;
      R f0=x2*(R(3)-x*2);
      R f1=x*xm*xm;
      R f2=x2*xm;
      // f0=F^2(u), f1=F^3(u), f2=F^4(u) in the notation of the
      // book mentioned above
      return p1+v0*f0+v1*f1+v2*f2;
   }
}

Path Path::cubCur(Spc const& p1, Vec const& v1,
      Spc const& p2, Vec const& v2, Z n)
{
   Z mL=3;
   Word loc("Path::cubCur(Spc,Vec,Spc,Vec,Z)");
   CPM_MA
   cpmassert(n>1,loc);
   V<Spc> ps(n);
   ps[1]=p1;
   ps[n]=p2;
   if (n>2){ // this is the perfectly normal case
      R dx=1./n;
      R xi=dx;
      for (Z i=2;i<n;++i){
         ps[i]=cub(xi,p1,v1,p2,v2);
         xi+=dx;
      }
   }
   CPM_MZ
   return Path(ps);
}

Path Path::cubCur(Spc const& p1, Vec const& v1, R k1,
      Spc const& p2, Vec const& v2, R k2, Z n)
{
   Vec e1=v1, e2=v2;
   e1.nor_();
   e2.nor_();
   R s=(p2-p1).abs();
   Vec w1=e1*(s*k1);
   Vec w2=e2*(s*k2);
   return cubCur(p1,w1,p2,w2,n);
}

Path Path::waveLine(Spc const& p1, Spc const& p2, R ampl,
   R omega, Vec const& tr, Z n)
{
   if (n<=2) return Path(p1,p2);
   Word loc("Path::waveLine(Spc,Spc,R,R,Vec,Z)");
   Vec e=p2-p1;
   R l=e.nor_();
   X2<Vec,Vec> xvv=e.tanNormDec(tr);
   Vec tr1=xvv.c2();
   tr1.nor_();
   V<Spc> ps(n);
   R dt=l/(n-1);
   R t=0;
   cpmassert(omega!=0,loc);
   R halfWave=cpmpi/omega;
   Z nW=CpmAlgorithms::getNerFrc_1(halfWave,l);
   cpmassert(nW>0,loc);
   R omegaAct=cpmpi/halfWave;
   for (Z i=1;i<=n;++i){
      Spc pi=p1+e*t;
      R phase=omegaAct*t;
      R ai=ampl*cpmsin(phase);
      Vec tri=tr1*ai;
      ps[i]=pi+tri;
      t+=dt;
   }
   return Path(ps);
}

Path Path::line(const Spc& pi, const Spc& pf, Z n)
{
   static Word loc("Path::line(Spc,Spc,Z)");
   cpmassert(n>1,loc);
   V<Spc> ps(n);
   ps[1]=pi;
   ps[n]=pf;
   if (n>2){
      R frac=1./(n-1);
      Vec sh=(pf-pi)*frac; // translation vector from point to point
      for (Z i=2;i<n;i++) ps[i]=ps[i-1]+sh;
   }
   return Path(ps);
}

Path Path::parabola(const Spc& pi, const Spc& pc, const Spc& pf,
   Z n)
{
   static Word loc("Path::parabola(Spc,Spc,Spc,Z)");
   cpmassert(n>2,loc);
   V<Spc> ps(n);
   ps[1]=pi;
   ps[n]=pf;
   Vec v=(pc-pi)*2;
   Vec aHalf=(pf-pi)-v;
   R tStep=1./(n-1);
   R t=0;
   for (Z i=2;i<n;i++){
      t+=tStep;
      ps[i]=pi+(v+aHalf*t)*t;
   }
   return Path(ps);
}

Path Path::circle(const Spc& pc, R r, Z n)
{
   if (r==0.){
      cpmerror("Path::circle(...) vanishing radius");
   }
   V<C> p=circlePoints(r,n);
   Vec shift=pc-Spc();
   Z i;
   V<Spc> ps(n+1);
   for (i=1;i<=n;i++){
      ps[i]=Spc(p[i])+shift;
   }
   ps[n+1]=ps[1];
   return Path(ps);
}

Path Path::ellipse(const Spc& pc, R a1, R a2, Z n)
{
   if (a1==0.|| a2==0.){
      cpmerror("Path::ellipse(...) vanishing half-axis");
   }
   V<C> p=circlePoints(a1,n);
   Vec shift=pc-Spc();
   Z i;
   if (a2!=a1){
      R f=a2/a1;
      for (i=1;i<=n;i++) p[i][2]*=f;
   }
   V<Spc> ps(n+1);
   for (i=1;i<=n;i++){
      ps[i]=Spc(p[i])+shift;
   }
   ps[n+1]=ps[1];
   return Path(ps);
}

Path Path::quarterEllipse(Spc const& p1, Vec const& v1,
   Spc const& p2, Z n)
{
   Z mL=3;
   Word loc("Path::quarterEllipse(Spc,Vec,Spc,Z)");
   CPM_MA
   Vec dp=p2-p1;
   R sp=(dp|v1);
   cpmassert(sp!=0,loc);
   Vec v=v1;
   if (sp<0) v=-v;
   Vec vl=v;
   vl.nor_(); // longitudinal unit vector
   Vec dpl=vl*(dp|vl); // longitudinal component
   Vec dpt=dp-dpl; // transversal component
      // now dp=dpl+dpt
   R r=dpt.abs();
   cpmassert(r>0,loc);
   Spc center=p1+dpt;
   R phi=cpmpi/(2*n);
   AxVec av;
   av=v.cross(dpt);
   av.phi_(Angle(phi,RAD)); // works also in 2D !
   V<Vec> vv(n);
   vv[1]=p1-center;
   Z i;
   for (i=2;i<=n;++i){
      vv[i]=vv[i-1]*av;
   }
   R fac=(p2-center).abs()/r;
   Z np=n+1;
   V<Spc> ps(np);
   ps[1]=p1;
   for (i=2;i<=n;++i){
      Vec vi=vv[i];
      Vec vil=vl*(vl|vi);
      Vec vit=vi-vil;
      vil*=fac;
      ps[i]=center+vil+vit;
   }
   ps[np]=p2;
   Path res(ps);
   CPM_MZ
   return res;
}

Path Path::operator &(const Path& p)const
{
   R acc=Root<R>(acc_).inf(p.acc_);
   return Path(x_&p.x_,acc);
}

Path Path::moveAndMerge(const Path& p)const
{
   Z mL=3;
   Word loc("Path::moveAndMerge(Path)");
   CPM_MA
   Z d=dim();
   Z dp=p.dim();
   Path res;
   if (d==0) res=p;
   else if (dp==0){
       res=*this;
   }
   else if (d==1){
      Vec shift=lastPoint()-p.firstPoint();
      Path p1=p;
      p1*=Group(shift);
      res=p1;
   }
   else if (dp==1){
      res=*this;
   }
   else{
      Vec e=lastDir();
      Vec ep=p.firstDir();
      AxVec rot=ep.rotTo(e);
      Spc fP=p.firstPoint();
      Path p1=p;
      p1*=Group(fP,rot);
      Vec shift=lastPoint()-fP;
      p1+=shift;
      res=(*this)&p1;
   }
   CPM_MZ
   return res;
}

X2<Path,Path> Path::divideByArc(R t)const
{
   Path res1,res2; // void paths
   if (t<=0) res2=*this;
   else if (t>=length_) res1=*this;
   else{
      X3<Spc,Vec,Z> xsvz=fromArc(t);
      Z i1=xsvz.c3();
      Spc sp1=xsvz.c1();
      V<Spc> x1=x_.resize(i1)&sp1;
      res1=Path(x1,acc_);
      V<Spc> x2=x_.eli(1,i1).ins(1,sp1);
      res2=Path(x2,acc_);
   }

   return X2<Path,Path>(res1,res2);
}

Path Path::replace(const Spc& s1, const Spc& s2, const Path& p)const
{
   Z mL=3;
   Word loc("Path::replace(Spc,Spc,Path)");
   CPM_MA
   cpmassert(isSmooth(),loc);

   X3<Spc,Vec,R> eval1=project(s1);
   X3<Spc,Vec,R> eval2=project(s2);
   R t1=eval1.c3();
   R t2=eval2.c3();
   CpmRootX::order<R>(t1,t2);

   X2<Path,Path> xpp=divideByArc(t2);

   Path p1=xpp.c1(); // process further
   Path p2=xpp.c2(); // last part

   X2<Path,Path> xpp2=p1.divideByArc(t1);
   Path p0=xpp2.c1();  // first part

   Path resPrel=p0.moveAndMerge(p);
   Path res=resPrel.moveAndMerge(p2);

   CPM_MZ
   return res;
}

bool Path::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Path",str);
   cpmp(x_);
   cpmp(acc_);
   return true;
}

bool Path::scanFrom(istream& str)
{
   cpms(x_);
   cpms(acc_);
   update();
   return true;
}

namespace{
   R dist2FromSegment1(Spc& s, const Spc& x1, const Vec& e, R d)
      // name end ...1 means that there are 1 mutable argument(s)
   {
      Vec vs=s-x1;
      R scp=(vs|e);
      Spc sProx;
      if (scp<=0){
         sProx=x1;
      }
      else if (scp>=d){
         sProx=x1+e*d;
      }
      else{
         sProx=x1+e*scp;
      }
      R res=(s-sProx).absSqr();
      s=sProx;
      return res;
   }
}

X3<Spc,Vec,R> Path::project(const Spc& s)const
{
   Z mL=3;
   Word loc("Path::project(Spc)");
   CPM_MA
   Spc sRes=s;
   Vec eRes;
   R lRes=0;
   if (n_==0){
      ;
      // then the path gives no information
      // that allows to create a non-trivial projected image from s
   }
   else if (n_==1){
      sRes=x_[1];
         // since the path has only one point, the projected image of
         // s has to be this point.
   }
   else{
      Z m=n_-1;
      V<Spc> vsProx(m);
      Vo<R> dis2(m); // list of squares of distances of s from
         // seqment x_[i],x_[i+1]
      for (Z i=1;i<=m;++i){
         Spc pi=s;
         dis2[i]=dist2FromSegment1(pi,x_[i],e_[i],l_[i]);
         vsProx[i]=pi;
      }
      Z im=dis2.indInf();
      sRes=vsProx[im];
      eRes=e_[im];
      lRes=y_[im]+(sRes-x_[im]).abs();
   }
   CPM_MZ
   return X3<Spc,Vec,R>(sRes,eRes,lRes);
}

X3<Spc,Vec,Z> Path::fromArc(R p)const
{
   Z mL=3;
   Word loc("Path::fromArc(R)");
   CPM_MA
   cpmassert(n_>=2,loc);

   Spc sp;
   Vec e;
   Z i0;

   if (p>=length_){
      i0=n_;
      e=lastDir();
      e.nor_();
      sp=x_[n_]+e*(p-length_);
   }
   else if (p<=0){
      i0=0;
      e=firstDir();
      e.nor_();
      sp=x_[1]+e*p; // notice p<0
   }
   else{
      i0=y_.locate(p);
      R pExc=p-y_[i0];
      sp=x_[i0]+e_[i0]*pExc;
      e=e_[i0];
   }
   CPM_MZ
   return X3<Spc,Vec,Z>(sp,e,i0);
}

R Path::turn()const
{
   Z mL=3;
   Word loc("Path::turn()");
   CPM_MA
   R sum=0.;
   Z i1,i2;
   for (i1=1,i2=2; i2<n_; ++i1,++i2){
      sum+=(e_[i1].angTo(e_[i2])).toR();
   }
   CPM_MZ
   return sum;
}

Group Path::con()const
{
   Z mL=3;
   Word loc("Path::con()");
   CPM_MA
   Group g;
   Z i1,i2;
   for (i1=1,i2=2; i2<n_; ++i1,++i2){
      Group g1=Group(x_[i2]-x_[i1]);
      Group g2=Group(x_[i2], e_[i1].rotTo(e_[i2]));
      g*=g1;
      g*=g2;
   }
   g*=Group(x_[n_]-x_[n_-1]);
   CPM_MZ
   return g;
}

R Path::turn2(R t1, R t2)const
{
   Z mL=3;
   Word loc("Path::turn2(R,R)");
   CPM_MA
   R u1,u2,sigma;
   if (t1<=t2){
      u1=t1;
      u2=t2;
      sigma=1;
   }
   else{
      u1=t2;
      u2=t1;
      sigma=-1;
   }
   X2<Path,Path>  pp=divideByArc(u1);
   R l1=pp.c1().length();
   Path p1=pp.c2();
   Path p2=p1.divideByArc(u2-l1).c1();
   R res=sigma*p2.turn();
   CPM_MZ
   return res;
}

Group Path::con2(R t1, R t2)const
{
   Z mL=3;
   Word loc("Path::con2(R,R)");
   CPM_MA
   R u1,u2;
   Z sigma;
   if (t1<=t2){
      u1=t1;
      u2=t2;
      sigma=1;
   }
   else{
      u1=t2;
      u2=t1;
      sigma=-1;
   }
   X2<Path,Path>  pp=divideByArc(u1);
   R l1=pp.c1().length();
   Path p1=pp.c2();
   Path p2=p1.divideByArc(u2-l1).c1();
   Group res=p2.con();
   if (sigma==-1) res=!res;
   CPM_MZ
   return res;
}

void Path::test(Z k)
{
   Z mL=1;
   Word loc("Path::test(Z)");
   CPM_MA
   k; // not used
   Z nd=2001;
   Spc s(0.789,-1000.3);
   R ang=cpmdeg*45.4;
      // only positive
   R r=1.78;
   R dis=-1.679999;
   R rd=r+dis;
   Vec shift(rd*cpmcos(ang),rd*cpmsin(ang));
   Spc st=s+shift;

   Path circ=Path::circle(s,r,nd);
   X3<Spc,Vec,R> xsr=circ.project(st);
   Spc xRes=xsr.c1();
   R lenRes=xsr.c3();
   R rRes=(xRes-s).abs();
   R rError=r-rRes;
   R lengthError=r*ang-lenRes;
   cpmcerr<<endl<<"rError="<<rError;
   cpmcerr<<endl<<"lengthError="<<lengthError;

   X3<Spc,Vec,Z> xsv=circ.fromArc(lenRes);
   Spc sp1=xsv.c1();
   Spc sp2=xRes;
   R locError=(sp1-sp2).abs();
   cpmcerr<<endl<<"locError="<<locError;
   Z i,iS=10370;
   R n=2.387654;
   R phiTot=cpmpi*n*2;
   R phiStep=phiTot/R(iS-1);
   R rhoStep=1./(iS-1);
   V<Spc> pq(iS);
   R phi=0;
   R rho=1;
   for (i=1;i<=iS;++i){
      Spc sp(rho*cpmcos(phi),rho*cpmsin(phi));
      phi+=phiStep;
      rho+=rhoStep;
      pq[i]=sp;

   }

   Path pRound=circ;
   R turns=pRound.turn();
   cpmdebug(turns);
   cpmdebug(phiTot);

   Group g=pRound.con();
   Vec vFirst=pRound.firstDir();
   Vec vLast=pRound.lastDir();
   Vec vTrans=vFirst;
   vTrans*=g;
   cpmdebug(vFirst);
   cpmdebug(vLast);
   cpmdebug(vTrans);
   cpmdebug(g);

   CPM_MZ
}

////////////////////// Skeleton ////////////////////////////////////////

void Skeleton::connectAsPath()
{
   Z mL=3;
   Word loc("Skeleton::connectAsPath()");
   CPM_MA
   for (Z i=1;i<x_.dim();i++) addEdge(i,i+1);
   CPM_MZ
}

Skeleton& Skeleton::operator *=(const Group& g)
{
   for (Z i=1;i<=x_.dim();i++) x_[i]*=g;
   return *this;
}

Skeleton& Skeleton::operator +=(const Vec& a)
{
   for (Z i=1;i<=x_.dim();i++) x_[i]+=a;
   return *this;
}

Skeleton& Skeleton::operator *=(const AxVec& A)
{
   for (Z i=1;i<=x_.dim();i++) x_[i]*=Group(A);
   return *this;
}

Skeleton& Skeleton::scl_(Spc const& fix, R fac)
{
   for (Z i=1;i<=x_.dim();i++) x_[i].scl_(fix,fac);
   return *this;
}

VecGraph Skeleton::flatten(Z connect)const
{
   Z nx=x_.dim();
   V<C> vr2(nx);
   for (Z i=1;i<=nx;i++) vr2[i]=C(x_[i][1],x_[i][2]);
   return VecGraph(vr2,connect);
}

Skeleton Skeleton::arc(const Spc& p1, const Spc& center,
   const AxVec& rot, Z n)
{
   X2<Angle,Vec> pav=rot.toAngAxs();
   Angle phi=pav.c1();
   Vec axis=pav.c2();
   V<Spc> ps=Path::arc(p1,center,rot,n)();
   Skeleton res(ps);
   for (Z i=1;i<n;i++){
      res.addEdge(i,i+1);
   }
   return res;
}

Skeleton Skeleton::line(const Spc& pi, const Spc& pf, Z n)
{
   static Word loc("Skeleton::line(Spc,Spc,Z)");
   cpmassert(n>1,loc);
   V<Spc> ps=Path::line(pi,pf,n)();
   Skeleton res(ps);
   res.connectAsPath();
   return res;
}

Skeleton Skeleton::parabola(const Spc& pi, const Spc& pc, const Spc& pf,
   Z n)
{
   static Word loc("Skeleton::parabola(Spc,Spc,Spc,Z)");
   cpmassert(n>2,loc);
   return Skeleton(Path::parabola(pi,pc,pf,n));
}

Skeleton Skeleton::circle(const Spc& pc, R r, Z n, bool connect)
{
   if (r==0.){
      cpmerror("Skeleton::circle(...) vanishing radius");
   }
   Z i;
   V<Spc> ps=Path::circle(pc,r,n)();
   Skeleton res(ps);
   if (connect){
      for (i=1;i<n;i++){
         res.addEdge(i,i+1);
      }
      res.addEdge(n,1); // close the polygon
   }
   return res;
}

Skeleton Skeleton::ellipse(const Spc& pc, R a1, R a2, Z n, bool connect)
{
   if (a1==0.|| a2==0.){
      cpmerror("Skeleton::ellipse(...) vanishing half-axis");
   }
   Z i;
   V<Spc> ps=Path::ellipse(pc,a1,a2,n)();
   Skeleton res(ps);
   if (connect){
      for (i=1;i<n;i++){
         res.addEdge(i,i+1);
      }
      res.addEdge(n,1); // close the polygon
   }
   return res;
}

Skeleton Skeleton::bezier(V<Spc> const& vs, Z np)
// see R_Func::Bezier for the pattern followed here
// Notice the seamless cooperation of V, R_Vector, and R_Func
// which allows dimVec treated like a variable.
{
   Z mL=3;
   Word loc("Skeleton::bezier(V<Spc>,Z)");
   CPM_MA
   Z i,j,vsd=vs.dim();
   V<R_Vector> vvr(dimVec, R_Vector(vsd));
   for (i=1;i<=dimVec;++i) for (j=1;j<=vsd;++j) vvr[i][j]=vs[j][i];
   V<R_Func> vrf(dimVec);
   R2 iv(0,1);
   for (i=1;i<=dimVec;++i) vrf[i]=R_Func::Bernstein(vvr[i],iv);
   V< Spc > ps(np);
   R du=1./(np-1);
   R u=0;
   for (i=1;i<=np;++i){
      for (j=1;j<=dimVec;++j){
         ps[i][j]=(vrf[j])(u);
      }
      u+=du;
   }
   Skeleton res(ps);
   for (i=1;i<np;i++) res.addEdge(i,i+1);
   CPM_MZ
   return res;
}

Skeleton Skeleton::cone(R r, R h, Z n)
{
   Spc ori;
   Skeleton res=ellipse(ori,r,r,n,true);
   if (dimVec==2) return res;
   Spc sp(0,h,0);
   res.addVertex(sp);
   Z m=res.dim();
   for (Z i=1;i<=n;i++) res.addEdge(i,m);
   return res;
}

Skeleton Skeleton::cylinder(R r, R h, Z n)
{
   Spc ori;
   Skeleton res=circle(ori,r,n,true);
      // circle in the 1-2 plane
   if (dimVec==2) return res;
   Vec shift;
   shift[3]=h;
   Skeleton res2=res;
   res2+=shift;
   res&=(res2);
   for (Z i=1;i<=n;i++) res.addEdge(i,i+n);
   return res;
}

Skeleton Skeleton::dblPyr(R length, R d, Z n)
{
   R r=d*0.5;
// p is the polygon in the complex plane
// bringing the polygon to the 2,3 plane for 3D and the 2-axis
// for 2D
   V<C> p=circlePoints(r,n);
   V<Spc> sp(n);
   Z i;
   for (i=1;i<=n;i++){
      R py=p[i][2];
      R pz=p[i][1];
      sp[i]=Spc(0.,py,pz); // works also for 2D
   }
// adding the double pyramide end points on the 1-axis
   R h=length*0.5;
   Spc spL(-h,0,0);
   Spc spU(h,0,0);
   V<Spc> spExt=sp&spL&spU;
// making the Skeleton
   Skeleton res(spExt);
   for (i=1;i<n;i++)  res.addEdge(i,i+1);
   res.addEdge(n,1);
      // the previous two lines are not essential
   for (i=1;i<=n;i++){
      res.addEdge(n+1,i);
      res.addEdge(i,n+2);
   }
   return res;
}

Skeleton Skeleton::con(R w)const
{
   using CpmArrays::S;
   S<Z2> edg=edges();
   Z i,j,n=edg.dim();
   if (n==0) return Skeleton();
   V<Spc> xRes(n);
   for (i=1;i<=n;i++){
      Z2 zi=edg[i];
      Spc p1=x_[zi[1]];
      Spc p2=x_[zi[2]];
      xRes[i]=p1+(p2-p1)*w;
   }
   Skeleton res(xRes);
   for (i=1;i<=n;i++){
      Z2 zi=edg[i];
      for (j=1;j<=n;j++){
         Z2 zj=edg[j];
         if (zi[2]==zj[1]) res.neg_[i].add_(j);
      }
   }
   return res;
}

Skeleton Skeleton::rec(R a, R b)
{
   Spc t1(0,0);
   Spc t2(a,0);
   Spc t3(a,b);
   Spc t4(0,b);
   V<Spc> vtb{t1,t2,t3,t4};
   Skeleton res(vtb,true);
   res.clsPath_();
   return res;
}

///////////////////// class WireFrame //////////////////////////////////

WireFrame WireFrame::con(R w)const
{
   Z n=rep.dim();
   V<Skeleton> vc(n);
   for (Z i=1;i<=n;i++) vc[i]=rep[i].con(w);
   return WireFrame(vc);
}

WireFrame& WireFrame::operator *=(Group const& g)
{
   Z n=rep.dim();
   for (Z i=1;i<=n;i++) rep[i]*=g;
   return *this;
}

WireFrame& WireFrame::operator +=(Vec const& a)
{
   Z n=rep.dim();
   for (Z i=1;i<=n;i++) rep[i]+=a;
   return *this;
}

WireFrame& WireFrame::operator *=(AxVec const& A)
{
   Z n=rep.dim();
   for (Z i=1;i<=n;i++) rep[i]*=A;
   return *this;
}

WireFrame& WireFrame::operator *=(const V<Group>& g)
{
   Z n=cpminf<Z>(rep.dim(),g.dim());
   for (Z i=1;i<=n;i++) rep[i]*=g[i];
   return *this;
}

WireFrame& WireFrame::operator +=(const V<Vec>& a)
{
   Z n=cpminf<Z>(rep.dim(),a.dim());
   for (Z i=1;i<=n;i++) rep[i]+=a[i];
   return *this;
}

WireFrame& WireFrame::operator *=(const V<AxVec>& A)
{
   Z n=cpminf<Z>(rep.dim(),A.dim());
   for (Z i=1;i<=n;i++) rep[i]*=A[i];
   return *this;
}

WireFrame WireFrame::tripod(R length, R relWidth, R relGap, Z n)
{
   R fill=(-relGap+1);
   R L1=length, L2=L1*0.5;
   R l1=L1, l2=L2*fill;

   R d=l1*relWidth;
   Z ns=(dimVec==3 ? 6:3);
   V<Skeleton> res(ns);

// 1-axis made out of one Skeleton
   res[1]=Skeleton::dblPyr(l1,d,n);
   res[1]+=Vec(L2,0,0);

// 2-axis made of two Skeletons
   res[2]=Skeleton::dblPyr(l2,d,n);
   res[2]*=AxVec(Angle(90.));
   res[2]+=Vec(0,l2*0.5,0);
   res[3]=res[2];
   res[3]+=Vec(0,L2,0);
      // works for 2D and 3D

   if (dimVec==3){
// 3-axis made of three Skeletons
      R L3=L1*0.333333333;
      R l3=L3*fill;
      res[4]=Skeleton::dblPyr(l3,d,n);
      AxVec rot2(Angle(-90),2);
      res[4]*=rot2;
      Vec sh1(0,0,l3*0.5);
      res[4]+=sh1;
      res[5]=res[4];
      Vec sh2(0,0,L3);
      res[5]+=sh2;
      res[6]=res[5];
      res[6]+=sh2;
   }
   return WireFrame(res);
}

WireFrame WireFrame::globe(R r, R f, Z nLong, Z nLat, Z nCirc, bool axes)
{
   const R refSysEnlarge=1.2;
   R f_=-f+1;
   Spc ori;
   Skeleton sk1=Skeleton::ellipse(ori,r,r*f_,nCirc);
   if (dimVec==2) return WireFrame(V<Skeleton>(1,sk1));
// now we are in 3D, where we are interested in exclusively
   sk1*=AxVec(Angle(90),1);// now in 1-3-plane
   Angle dLong(180./nLong);
   V<Skeleton> res(nLong);
   res[1]=sk1;
   Z i;
   for (i=2;i<=nLong;i++){
      res[i]=res[i-1];
      res[i]*=AxVec(dLong,3);
   }
   Skeleton eq=Skeleton::ellipse(ori,r,r,nCirc);
   res&=eq;
   Angle dLat(90./nLat);
      // nLat is the number of latitude circles
      // equator included but pole not counting as a
      // latitude circle.
   Angle beta=dLat;
   for (i=1; i<nLat; i++){
      R cb=cpmcos(beta);
      R rLat=r*cb;
      R hLat=r*cpmsin(beta)*f_;
      Skeleton circLat=Skeleton::ellipse(ori,rLat,rLat,nCirc);
      Skeleton skNorth=circLat;
      skNorth+=Vec(0,0,hLat);
      res&=skNorth;
      Skeleton skSouth=circLat;
      skSouth+=Vec(0,0,-hLat);
      res&=skSouth;
      beta+=dLat;
   }
   if (axes)
      return WireFrame(res)&=tripod(r*refSysEnlarge);
   else
      return WireFrame(res);
}

WireFrame WireFrame::spline(const V<Spc>& vs, const Spc& pc1, Z n)
{
   static Word loc("WireFrame::spline(V<Spc>,Spc,Z)");
   Z i,nr=vs.dim()-1;
   cpmassert(nr>0,loc);
   V<Skeleton> res(nr); // p-1 connecting parabolic arcs between
      // p points
   res[1]=Skeleton::parabola(vs[1],pc1,vs[2],n);
   if (nr>1){
      V<Spc> pc(nr); // list of all control points
         // logic is easier for doing them in one pass and
         // storing them
      pc[1]=pc1;
      for (i=2;i<=nr;i++) pc[i]=vs[i]+(vs[i]-pc[i-1]);
         // making a list of control points
      for (i=2;i<=nr;i++)
         res[i]=Skeleton::parabola(vs[i],pc[i],vs[i+1],n);
            // last i+1 is vs.dim() as it should be
   }
   return WireFrame(res);
}
