//? cpmfield.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_FIELD_H_
#define CPM_FIELD_H_
/*
Purpose: Define classes describing function-like objects
   with linear algebra arithmetic operations and transformation
   properties under a Group. Most fields in physics fall under this
   catagories. The space X on which the functions are defined is only
   assumed to transform under G. If we assume for G a structure that
   allows to define local translations, we can derive from that
   gradients and other differential operations.
*/
#include <cpmf.h>


namespace CpmFields{

   using CpmFunctions::F;
#ifdef CPM_Fn
   using CpmFunctions::F1;
   using CpmFunctions::F2;
   using CpmFunctions::F3;
#endif
   using CpmArrays::V;
   using CpmRoot::R;
   using CpmRoot::Z;
   using namespace CpmStd;

////////////////////////// class ScalarField<> /////////////////////////

template <class X, class S, class G>
// assumptions X*=G, S+S, S-S, S*S
class ScalarField: public F<X,S> { // scalar fields
   // function space with linear structure and
   // transformation properties under Group G
   typedef F<X,S> Base;
public:

   ScalarField(void):Base(){}
      // default constructor

   ScalarField(const Base& g):Base(g){}
      // downcast constructor

// Construtor for constant function

   ScalarField(const S& s1):Base(s1){}

   ScalarField operator +(const ScalarField& x)const;
   ScalarField& operator +=(const ScalarField& x){ return *this=*this+x;}

   ScalarField operator -(const ScalarField& x)const;
   ScalarField& operator -=(const ScalarField& x){ return *this=*this-x;}

   ScalarField operator *(const S& s)const;
   ScalarField& operator *=(const S& s){ return *this=*this*s;}

   ScalarField operator *(const G& g)const;
   ScalarField& operator *=(const G& g){ return *this=*this*g;}
   const F<X,S>& toBase()const{ return *this;}

   static ScalarField sum(const V< ScalarField >& vf);
      // adding many fields together by a single operation

private:

   // the algorithmic content of the class should be available in
   // unbiased form
   // notice that these are member functions, so that the arguments need
   // no template arguments <...> which makes the matter quite compact

   static S plus(const X& x, const ScalarField& f1,
      const ScalarField& f2){ return f1(x)+f2(x);}

   static S diff(const X& x, const ScalarField& f1,
      const ScalarField& f2){ return f1(x)-f2(x);}

   static S mult(const X& x, const ScalarField& f, const S& s)
   { return f(x)*s;}

   static S transform(const X& x, const ScalarField& f, const G& g)
   { X xv=x; return f(xv*=!g); }

   static S sumOfArray(const X& x, const V<ScalarField>& vf)
   { S s0=S(); for (Z i=1;i<=vf.dim();i++) s0+=(vf[i])(x); return s0; }

};

template <class X, class S, class G>
ScalarField<X,S,G> ScalarField<X,S,G>::operator +
(const ScalarField& f)const
{
#ifdef CPM_Fn
   return F2<X,ScalarField,ScalarField,S>(*this,f)(plus);
#else
   return F<X,S>(bind(plus,_1,*this,f));
#endif
}

template <class X, class S, class G>
ScalarField<X,S,G> ScalarField<X,S,G>::operator -
(const ScalarField& f)const
{
#ifdef CPM_Fn
   return F2<X,ScalarField,ScalarField,S>(*this,f)(diff);
#else
   return F<X,S>(bind(diff,_1,*this,f));
#endif
}

template <class X, class S, class G>
ScalarField<X,S,G> ScalarField<X,S,G>::operator * (const S& s)const
{
#ifdef CPM_Fn
   return F2<X,ScalarField,S,S>(*this,s)(mult);
#else
   return F<X,S>(bind(mult,_1,*this,s));
#endif
}

template <class X, class S, class G>
ScalarField<X,S,G> ScalarField<X,S,G>::operator * (const G& g)const
{
#ifdef CPM_Fn
   return F2<X,ScalarField,G,S>(*this,g)(transform);
#else
   return F<X,S>(bind(transform,_1,*this,g));
#endif
}

template <class X, class S, class G>
ScalarField<X,S,G> ScalarField<X,S,G>::sum(
   const V< ScalarField<X,S,G> >& vf)
{
#ifdef CPM_Fn
   return F1<X,V<ScalarField>,S>(vf)(sumOfArray);
#else
   return F<X,S>(bind(sumOfArray,_1,vf));
#endif
}

//////////////////////////////// class Field<> /////////////////////////

template <class X, class Y, class S, class G>
// assumptions: X*=G, Y*=G, S Y[Z], Y+Y, Y-Y, Y*S
class Field: public F<X,Y> { // fields on which a group acts
// see ScalarField; now the group acts also
// on the values of the field (vector field, spinor field, ..

   typedef F<X,Y> Base;
   typedef ScalarField<X,S,G> FieldS;

public:

   Field(void):Base(){}
      // default constructor

   Field(const F<X,Y>& g):Base(g){}
      // downcast constructor

// Construtor for constant function

   Field(const Y& y1):F<X,Y>(y1){}

   Field operator +(const Field& x)const;
   Field& operator +=(const Field& x){ return *this=*this+x;}

   Field operator -(const Field& x)const;
   Field& operator -=(const Field& x){ return *this=*this-x;}

   Field operator *(const S& s)const;
   Field& operator *=(const S& s){ return *this=*this*s;}

   Field operator *(const G& g)const;
   Field& operator *=(const G& g){ return *this=*this*g;}

   void setComponent(const FieldS& fs, Z i);
      // makes the ith component of *this equal to the scalar field fs

   FieldS getComponent(Z i)const;
      // returns a selected field component as a scalar field

   static Field sum(const V< Field >& vf);
      // adding many fields together by a single operation

private:  // algorithmic content, see ScalarField
   static Y plus(const X& x, const Field& f1, const Field& f2)
   { return (Y)(f1(x)+f2(x));}

   static Y diff(const X& x, const Field& f1, const Field& f2)
   { return (Y)(f1(x)-f2(x));}

   static Y mult(const X& x, const Field& f, const S& s)
   { return (Y)(f(x)*s);}

   static Y transform(const X& x, const Field& f, const G& g)
   { X xv=x; return f(xv*=!g)*=g; }
      // see UM/Ariton 99-7-22/1 for a proof that this is the correct
      // operation of a group on a field

   static Y setcomp(const X& x, const Field& f, const FieldS& fs,
      const Z& i ){ Y y=f(x); y[i]=fs(x); return y;}

   static R getcomp(const X& x, const Field& f, const Z& i )
      {return f(x)[i];}

   static Y sumOfArray(const X& x, const V<Field>& vf)
   { Y y0=Y(); for (Z i=1;i<=vf.dim();i++) y0+=(vf[i])(x); return y0; }
};

template <class X, class Y, class S, class G>
void Field<X,Y,S,G>::setComponent(const FieldS& fs, Z i)
  // makes the ith component of *this equal to the scalar field fs
{
#ifdef CPM_Fn
   (*this)=F3<X,Field,FieldS,Z,Y>(*this,fs,i)(setcomp);
#else
   (*this)=F<X,Y>(bind(setcomp,_1,*this,fs,i));
#endif
}

template <class X, class Y, class S, class G>
ScalarField<X,S,G> Field<X,Y,S,G>::getComponent(Z i)const
{
#ifdef CPM_Fn
   return FieldS(F2<X,Field,Z,R>(*this,i)(getcomp));
#else
   return FieldS(F<X,R>(bind(getcomp,_1,*this,i)));
#endif
}

template <class X, class Y, class S, class G>
Field<X,Y,S,G> Field<X,Y,S,G>::operator + (const Field& f)const
{
#ifdef CPM_Fn
   return F2<X,Field,Field,Y>(*this,f)(plus);
#else
   return F<X,Y>(bind(plus,_1,*this,f));
#endif
}

template <class X, class Y, class S, class G>
Field<X,Y,S,G> Field<X,Y,S,G>::operator - (const Field& f)const
{
#ifdef CPM_Fn
   return F2<X,Field,Field,Y>(*this,f)(diff);
#else
   return F<X,Y>(bind(diff,_1,*this,f));
#endif
}

template <class X, class Y, class S, class G>
Field<X,Y,S,G> Field<X,Y,S,G>::operator * (const S& s)const
{
#ifdef CPM_Fn
   return F2<X,Field,S,Y>(*this,s)(mult);
#else
   return F<X,Y>(bind(mult,_1,*this,s));
#endif
}

template <class X, class Y, class S, class G>
Field<X,Y,S,G> Field<X,Y,S,G>::operator * (const G& g)const
{
#ifdef CPM_Fn
   return F2<X,Field,G,Y>(*this,g)(transform);
#else
   return F<X,Y>(bind(transform,_1,*this,g));
#endif
}

template <class X, class Y, class S, class G>
Field<X,Y,S,G> Field<X,Y,S,G>::sum(const V< Field<X,Y,S,G> >& vf)
{
#ifdef CPM_Fn
   return F1<X,V<Field>,Y>(vf)(sumOfArray);
#else
   return F<X,Y>(bind(sumOfArray,_1,vf));
#endif
}

} // namespace
#endif
