//? cpmlineartx.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
Description: 'lineart' has nothing to do with 'linear': it's 'line art'.
   Defining classes Wrt (Writing) and LineArt
   which both represent spatial graphical objects which use a
   Camera-object for visualization. Intended to allow us to
   create figures to be placed in scientific articles.
   Includes the definition of a vector font (latin and greek).
   The size of code needed to define the letters and numbers was the
   reason to remove the classes from the files of the dim family.
*/

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;

   using CpmArrays::V;

   using CPM_DIM::Group;
   using CPM_DIM::Spc;
   using CPM_DIM::Vec;
   using CPM_DIM::AxVec;
   using CPM_DIM::Skeleton;
   using CPM_DIM::Path;

   using CPM_CAMERA::Camera;

//doc namespace CpmDimx{

//////////////////////// struct IndCon ///////////////////////////////////

struct IndCon{ // indexing control
   // Allows to shift indexes relative to their default position. No
   // limitation in range, so right indexes may be changed to left ones.
   // What is meant here are indexes to symbols (mostly letters), e.g.
   // what
   // in HTML would be x<sub>1</sub>.
   R fac; // Scaling factor to be applied to the index
   R dh; // dh>0 shifts vertically upwards
   R dw; // dw>0 shifts horizontally to the right
   IndCon():fac(1),dh(0),dw(0){}
   IndCon(Word const& w){
      if (w=="sub"){ fac=0.5; dh=-0.4; dw=0;}
      else if (w=="sup"){ fac=0.5; dh=-0.2; dw=0;}
      else{ fac=1; dh=0; dw=0;}
   }
   // A constructor IndCon(R,R,R) is considered error prone:
   // It is easy to forget whether dh is the second or the third
   // argument. By contrast, saying ic.dh=-0.3; ic.dw=-0.1 is
   // perfectly clear.
};

///////////////////////// class Wrt //////////////////////////////////////
/*********** Here is a list of all names w for which
   Wrt(Word const& w);
is defined:

V<Word> Wrt::allNam()
{
   V<Word> res(123);
   res[1]="A"; res[2]="Au"; res[3]="B"; res[4]="C"; res[5]="D";
   res[6]="E"; res[7]="F"; res[8]="G"; res[9]="H"; res[10]="I";
   res[11]="J"; res[12]="K"; res[13]="L"; res[14]="M"; res[15]="N";
   res[16]="O"; res[17]="Ou"; res[18]="P"; res[19]="Q"; res[20]="R";
   res[21]="S"; res[22]="T"; res[23]="U"; res[24]="Uu"; res[25]="V";
   res[26]="W"; res[27]="X"; res[28]="Y"; res[29]="Z"; res[30]="a";
   res[31]="au"; res[32]="b"; res[33]="c"; res[34]="d";res[35]="e";
   res[36]="f"; res[37]="g"; res[38]="h"; res[39]="i";res[40]="j";
   res[41]="k"; res[42]="l"; res[43]="m"; res[44]="n";res[45]="o";
   res[46]="ou"; res[47]="p"; res[48]="q"; res[49]="r";res[50]="s";
   res[51]="t"; res[52]="u"; res[53]="uu"; res[54]="v"; res[55]="w";
   res[56]="x"; res[57]="y"; res[58]="z"; res[59]="n1"; res[60]="n2";
   res[61]="n3"; res[62]="n4"; res[63]="n5"; res[64]="n6"; res[65]="n7";
   res[66]="n8"; res[67]="n9"; res[68]="n0"; res[69]="alpha";
   res[70]="beta";
   res[71]="gamma"; res[72]="delta"; res[73]="epsilon"; res[74]="theta";
   res[75]="phi";
   res[76]="psi"; res[77]="omega"; res[78]="rho"; res[79]="lambda";
   res[80]="chi";
   res[81]="sigma"; res[82]="mu"; res[83]="nu"; res[84]="upsilon";
   res[85]="eta";
   res[86]="tau"; res[87]="pi"; res[88]="zeta"; res[89]="xi";
   res[90]="kappa";
   res[91]="omicron"; res[92]="iota"; res[93]="Gamma"; res[94]="Lambda";
   res[95]="Delta";
   res[96]="Pi"; res[97]="Sigma"; res[98]="Theta"; res[99]="Xi";
   res[100]="Omega";
   res[101]="Phi"; res[102]="Psi"; res[103]="space"; res[104]="halfspace";
   res[105]="backspace";
   res[106]="backspace2"; res[107]="period"; res[108]="colon";
   res[109]="comma"; res[110]="semicolon";
   res[111]="exclamation_mark"; res[112]="question_mark"; res[113]="dash";
   res[114]="equal"; res[115]="l_bracket";
   res[116]="r_bracket"; res[117]="slash"; res[118]="plus";
   res[119]="quote"; res[120]="dquote";
   res[121]="tilde"; res[122]="schlipf"; res[123]="ast";
   return res;
}
*************************************************************************/

class Wrt{
   // Wrt=writing, actually printable symbols
   // If CPM_DIM is defined as CpmDim3 all Wrt objects 'live' in the 1-2
   // plane upon construction from Word objects. The function
   //    Wrt& operator *=(Group const& g);
   //  allows to move them freely arround within 3-space.
   V<Path> v_; // array of lines which make up the symbol
   Spc f_; // first
   Spc l_; // last,
      // the intention is that the next letter should be placed
      // with its first point just at the last point of its
      // predecessor. Also connection point for lower
      // indexes. Thus these are points on the writing line.
   Spc u_;
      // upper right corner;
      // connection point for upper indexes
   R h_; // maximum hight over base line (writing line)

   R w_; // maximum width of the visible object
      // not including the reserved space to the next
      // Wrt-object.
   R d_; // descender, underlength

   void operator&=(Path const& p){ v_&=p;}
      // adding to the path list without considering
      // influence on f_,l_,h_,d_. Thus to be used with care,
      // thus private
   void operator&=(V<Path> const& vp){ v_&=vp;}
      // previous function extended to arrays

   typedef Wrt Type;
public:
   CPM_IO
   explicit Wrt(Word const& w);
      // creating the letters of Latin and Greek alphabets together
      // with punctuation and some special characters.
      // examples Wrt("gamma"), Wrt("Gamma"), Wrt("ou") (o as 'Umlaut'),
      // Wrt("hs") (half-space), Wrt(n5) (5 , 'number 5')
      // All implemented cases can be infered from the implementation in
      // cpmlineartcpp.h
      // The main fact as to size of the symbols is that the height of
      // A (i.e. capital a) is 7. Size can be varied continuously using
      // function scl_(Spc,R) or *=(R).
      // See also the copied code above.
   Word nameOf()const{ return "Wrt";}
   Wrt():v_(),f_(),l_(),h_(0),w_(0),d_(0){}
   Spc fir()const{ return f_;}
      //: first
      // left-most point of the writing box which is
      // on the writing line
   Spc last()const{ return l_;}
      //: last
      // right-most point of the writing box which is
      // on the writing line
   Spc upp()const{ return u_;}
      //: upper
      // right upper corner
   Spc rigLowCor()const{ return f_+e1()*w_;}
      //: right lower corner
   Spc rigUppCor()const{ return u_;}
      //: right upper corner
   Spc uppMid()const{ return u_-e1()*(w_*0.5);}
      //: upper midle
   Spc leftLowCor()const{ return f_;}
      //: left lower corner
   Spc leftUppCor()const{ return u_-e1()*w_;}
      //: left upper corner
   Vec spc()const{ return l_-rigLowCor();}
      //: space
      // space between the right lower corner of the symbol box
      // and the next symbol box
   Vec e1()const{ Vec e=l_-f_; e.nor_(); return e;}
      //: e1
      // unit vector in writing direction
   Vec e2()const{ Vec t=(u_-f_); Vec ex=e1(); Vec tp=ex*(ex|t);
      Vec tn=t-tp; tn.nor_(); return tn; }
      //: e2
      // unit vector from bottom to top of the symbols, perpendicular
      // to e1.
   V<Vec> bas()const;
      //: basis
   CpmArrays::V<Path> const& v()const{ return v_;}
      //: v
      // read access to main data v_

   Path box()const
      //: box
      // Enclosing box for the symbol.
   {
      Vec ex=e1(), ey=e2();
      Spc ll=f_-ey*d_;
      Spc lr=ll+ex*w_;
      Spc ur=u_;
      Spc ul=f_+ey*h_;
      return Path(ll,lr,ur,ul,ll);
   }

   R wTot()const{ return (f_-l_).abs();}
      //: w total
      // with of the object together with space to the
      // next Wrt-object
   R h()const{ return h_;}
      //: h (height)
   R w()const{ return w_;}
      //: w (width)
   Wrt& operator *=(Group const& g);
      // general Euclidian transformation
   Wrt& operator +=(Vec const& v);
      // translation
   Wrt& operator *=(AxVec const& av);
      // rotation arround point fir_
   Wrt& operator *=(R scl);
      // scaling such that point fir remains fix
   Wrt& scl_(Spc const& fix, R fac);
      //: scale
   Wrt& thc_(R r, Z n);
      //: thick
      // This function is active only if r>0 and n >=2. If this is
      // satisfied, v_ gets replaced by the union of n shifted
      // copies, where the n shift vectors have length r and
      // result from equispaced points on a circle (not disk!).
      // This is a brut force method to simulate a line thickness
      // of 2*r. Works fine. But the suitable n has to be found by
      // experimentation for given r. If n is too small, there remain
      // voids, if it is too large, the function takes too much time.
   Wrt& vec_();
      //: vector
      // endowes *this with a vector arrow above it
   Wrt& dot_();
      //: dot
      // endowes *this with a dot (Newton's notion for the derivative)
      // above it
   Wrt& ddot_();
      //: ddot
      // endowes *this with a double dot (Newton's notion for the
      // second derivative) above it
   Wrt& sub_(V<Wrt> const& vw, IndCon const& ic=IndCon("sub"));
      //: subscript
      // Changes *this into a combined symbol which carries a
      // subscript (placed to the right of the main symbol).
      // The connection point for an upper (right) index of
      // this combined symbol is the same as for the original
      // symbol. Thus by applying sup_ after sub_ it is possible
      // to endow *this with and upper and a lower index such
      // that both indexes start at the same horizontal position.
   Wrt& sup_(V<Wrt> const& vw, IndCon const& ic=IndCon("sup"));
      //: superscript
      // See ecplanation to function sub_. See preceeding
      // struct IndCom for an explanation of the argument ic.
      //
   Wrt& to_(Spc const& p);
      //: to
      // Moves the point f_ of *this to point p by translation.
      // This is the method for designing Words. The word "good"
      // will be created like this:
      // Wrt a1("g");
      // Wrt a2("o"); a2.to_(a1);
      // Wrt a3(a2); a3.to_(a2);
      // Wrt a4("d"); a4.to_(a3);
      // Now the letters a2,a3,a4 are placed relative to a1 as to be
      // expected from the word "good". To make a graphical object
      // out of this word, we define
      // LineArt x; x&=a1; x&=a2; x&=a3; x&=a4;
      // Making use of the member functions of class LineArt,
      // we can shift, rotate, and scale the "graphical word" x
      // as a whole. Also creation of x from the Word list
      // V<Word> vw("","g","o","o","d") can be done by the
      // constructor LineArt(V<Word>,R,Z).
      // The function CpmCamerax::Camera::mark(LineArt)
      // (Camerax stands for Camera2 or Camera3) allows to map the word
      // into the C+- frame buffer from which it can be displayed on
      // screen and written into a ppm-image-file.
      // Single Wrt-objects can be treated alike by the function
      // CpmCamerax::Camera::mark(Wrt,Color).
   void mark(Camera const& ca, Color const& color,
      bool points=false)const;
      //: mark
      // If the last argument is true, we show also the characteristic
      // points

   static V<Word> allNam();
      //: all names
      // returns a list of all allowed arguments of the constructor
      //    explicit Wrt(Word const& w);

   static void showAll(Camera const& ca, Color const& color,
      bool points=false, R tWait=12);
      // shows the whole set of characters from the list generated by the
      // previous funcion. In interpreting the output be aware that there
      // are three backspace characters present. The last argument
      // sets the inspection time for the screen image.
};

///////////////////////// class LineArt //////////////////////////////////
class LineArt{ // a variety of graphical objects in combination
   // similiar to WireFrame, but containing additional
   // objects of graphical relevance. See topic combining LineArts
   // for typical usage.

   V<Skeleton> s_;
   V<Triangle> t_;
   V<Path> p_;
   V<Wrt> w_; // writing

   typedef LineArt Type;
public:
// infrastructure
   CPM_IO
   Word nameOf()const{ return "LineArt";}

// constructing LineArts
LineArt(V<Skeleton> const& s, V<Triangle> const& t,
   V<Path> const& p, V<Wrt> const& w ):
   s_(s),t_(t),p_(p),w_(w){}

LineArt():s_(),t_(),p_(),w_(){}

explicit LineArt( V<Skeleton> const& s):
   s_(s),t_(),p_(),w_(){}

// construction from arrays
explicit LineArt( V<Triangle> const& t):
   s_(),t_(t),p_(),w_(){}

explicit LineArt( V<Path> const& p):
   s_(),t_(),p_(p),w_(){}

explicit LineArt( V<Wrt> const& w, bool chainIt=false);
// If the last argument is 'true', the components of
// w will be put in a line so that it is easy to
// generate a line of text from the Wrt-object representing
// the letters.

// construction from individuals
explicit LineArt( Triangle const& t):
   s_(),t_(V<Triangle>{t}),p_(),w_(){}

explicit LineArt( Skeleton const& s):
   s_(V<Skeleton>{s}),t_(),p_(),w_(){}

explicit LineArt( Path const& p):
   s_(),t_(),p_(V<Path>{p}),w_(){}

explicit LineArt( Wrt const& w):
   s_(),t_(),p_(),w_(V<Wrt>{w}){}

explicit LineArt( V<Word> const&  vw, R r=0, Z n=32);
// a list of names of Wrt-object will form a graphical word by
// proper concatenation, and optional thickening of the lines.

// transforming LineArts
LineArt& operator *=(Group const& g);

LineArt& operator +=(Vec const& a);

LineArt& operator *=(AxVec const& A)
{ return operator*=(Group(A));}

Spc fir()const{ return w_.dim()>0 ? w_.fir().fir() : Spc();}
//: first
// position of the first among the letters, Spc() if
// there is none

Spc last()const{ return w_.dim()>0 ? w_.last().last() : Spc();}
//: last
// position of the last among the letters, Spc() if
// there is none

Spc cen()const{ return fir().cen(last());}
//: center

LineArt& to_(Spc const& p){ return (*this)+=(p-fir());}
//: to
// functionality just as that of the corresponding
// function Wrt::to_(Spc)

LineArt& scl_(Spc const& fix, R fac);
//: scale
// It is important to fix here a fixed point since an instance
// of LinArt does not have a distinguished point as instances
// of Wrt do.

LineArt& thc_(R r, Z n);
//: thick
// This function is active only if r>0 and n >=2. If this is
// satisfied, v_ gets replaced by the union of n shifted
// copies, where the n shift vectors have length r and
// result from equispaced points on a circle (not disk!).
// This is a brute force method to simulate a line thickness
// of 2*r. Works fine. But the suitable n has to be found by
// experimentation for given r. If n is too small, there remain
// voids, if it is too large, the function takes too much time.
// Small void will be filled when the capuring camera is asked
// to perform line art processing (see function Camera::dis).

// combining LineArts
//
// This allows to create parts in isolation, move them around,
// transform them at will, and then chain them together
// to a final piece that then can be displayed or/and filed
// e.g. as an illustration for a scientific article.
//    LineArt la1=...;
//    LineArt la2=...;
//       ...
//    LineArt la;
//    la&=la1; la&=la2; ...
//    Camera ca=...;
//    la.mark(ca);
//    ca.setWriteOnShow(true);
//    bool lineArtProcessing=true;
//    ca.dis(lineArtProcessing);
//    cpmwait(30,2);

LineArt& operator &=(LineArt const& la)
   { s_&=la.s_; t_&=la.t_; p_&=la.p_; w_&=la.w_; return *this;}

LineArt& operator &=(Skeleton const& s)
   { s_&=s; return *this;}

LineArt& operator &=(Triangle const& t)
   { t_&=t; return *this;}

LineArt& operator &=(Path const& p)
   { p_&=p; return *this;}

LineArt& operator &=(Wrt const& w)
   { w_&=w; return *this;}

LineArt& operator &=(V<Skeleton> const& s)
   { s_&=s; return *this;}

LineArt& operator &=(V<Triangle> const& t)
   { t_&=t; return *this;}

LineArt& operator &=(V<Path> const& p)
   { p_&=p; return *this;}

LineArt& operator &=(V<Wrt> const& w)
   { w_&=w; return *this;}

// access
const V<Skeleton>& cs()const{ return s_;}
V<Skeleton>& cs_(){ return s_;}

const V<Triangle>& ct()const{ return t_;}
V<Triangle>& ct_(){ return t_;}

const V<Path>& cp()const{ return p_;}
V<Path>& cp_(){ return p_;}

const V<Wrt>& cw()const{ return w_;}
V<Wrt>& cw_(){ return w_;}

R h()const;
//: h
// Text height. Maximum height of a Wrt in w_, i.e.
// maximum of w_[i].h(), 1 <= i <= w_.dim().
// Notice thus that h refers only to the height of the letters involved
// and not to the height of other geometrical structures.

Z dimS()const{ return s_.dim();}
Z dimT()const{ return t_.dim();}
Z dimP()const{ return p_.dim();}
Z dimW()const{ return w_.dim();}

void mark(Camera const& cam)const;
//: mark
// writing *this to the frame store. The Camera object cam controls
// the conversion to 2D, or a 2D transformation if we are in 2D
// already. Display to screen is under the responsibility of
// the Camera cam and gets invoked by cam.dis();
// See also comment to topic header 'combining LineArts'.

void mark(Camera const& cam, Color const& col, bool points=false)const;

static LineArt disk(Spc const& p, Vec const& h, R r, Z n=16);
//: disk
// p is the center of the disk area which is a regular n-polygon
// of radius r. h is a normal vector of this area (length
// does not matter).

static LineArt ellDisk(Spc const& p, Vec const& h, Vec const& h1,
   R r1, R r2, Z n=16);
//: elliptic disk
// p is the center of the disk area which is a deformed version of
// a regular n-polygon of radius r1.
// h is a normal vector of this area (length does not matter),
// and h1 is the vector which defines the 1-direction,to which
// r1 refers.

static LineArt sph(Spc const& p, R r, Z n=16);
//: sphere
// made of only three disks, which should be sufficient

static LineArt cone(Spc const& p,  Vec const& h, R r, Z n=16);
//: cone
// similar as Skeleton::cone but with more geometric parameters.
// p is the center of the base area and p+h is the top of the
// cone. In 3D the base plane is a regular n-polygon of radius r.
// In 2D the object actually is a triangle, so that n has no influence.

static LineArt arr(Spc const& p1, Spc const& p2, R h, R r, Z n=16);
//: arrow
// The arrow extends from p1 to p2. The arrowhead is a cone, the top of
// which is at p2. The height of the conic arrowhead is h (no vector in
// this case) and the radius of its base plane is r.

static LineArt text(V<Wrt> const& vw);
//: text
// Takes the position and orientation from the first Wrt in the argument
// list and forms the whole chain by shifting following symbols to the
// fitting positions. No rotation of symbols will be done.
};
