//? cpmshapesx.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using CpmFunctions::FncObj;

using CpmRoot::R;
using CpmRoot::Z;
using CpmFunctions::F;
using CpmArrays::V;

//doc namespace CpmDimx{

/////////////////// class Form<> /////////////////////////////////////////

template <class Dat>
class Form: public FncObj<Rdim,R>{ // shape determining function objects
   // a more direct version of F1<Rdim,Dat,R> to be used for
   // initialization in Shp<Dat>
   Dat dat;
       // streamable data
 public:
   Form(const Dat& dat_):dat(dat_){}
   R operator()(const Rdim& sp)const{ return dat.indFnc(sp);}
};

///////////////////// class Shp<> ///////////////////////////////////

template <class Dat>
class Shp:public SubsetOfSpace{ // shapes

protected:

   Dat dat;

public:

   typedef SubsetOfSpace Base;
   typedef Shp<Dat> Type;
   CPM_IO_V

   virtual SubsetOfSpace* clone()const{ return new Shp(*this);}

   Shp(const Dat& dat_=Dat(), const RefSys& rs=RefSys()):
      Base( F<Rdim,R>(new Form<Dat>(dat_)), rs ), dat(dat_)
   {
      set_x(dat.getCenter());
      set_r(dat.getRadius());
   }

   Dat getDat()const{ return dat;}
};

template <class Dat>
bool Shp<Dat>::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("Shp<>",str)) return false;
   if (!Base::prnOn(str)) return false;
   if (!dat.prnOn(str)) return false;
   return true;
}

template <class Dat>
bool Shp<Dat>::scanFrom(istream& str)
{
   RefSys rs_;
   if (!rs_.scanFrom(str)) return false;
   Spc x_;
   if (!x_.scanFrom(str)) return false;
   R r_;
   cpms(r_);
      // x and r not used, but they have to be read
   DisPar dp_;
   if (!dp_.scanFrom(str)) return false;
   Dat dat_;
   if (!dat_.scanFrom(str)) return false;
   *this=Shp<Dat>(dat_,rs_);
   set_dp(dp_);
   return true;
}

////////////////////////// class Cylinder ///////////////////////////


struct CylinderDat{ // data which determine a cylinder
      // Cylinder
   Rdim x0;
         // coordinate triplet of center
   R rho{0.};
         // radius of the cylinder
   R hh{0.};
         // half-height of the cylinder along the cylinder axis
   CylinderDat(){}
   CylinderDat(Rdim x0_,R rho_,R hh_):
      x0(x0_),rho(rho_),hh(hh_){}
   typedef CylinderDat Type;
   CPM_IO
   R getRadius()const{ return cpmsqrt(hh*hh+rho*rho);}
   Rdim getCenter()const{ return x0;}
   R vol()const{
      R arBase=(Spc::dimension()==3 ? cpmpi*rho*rho : R(2.)*rho);
      return arBase*hh;
   }
   R indFnc(const Rdim&)const;
};


class Cylinder:public Shp<CylinderDat>{ // cylinder
public:
   typedef Shp<CylinderDat> Base;
   Cylinder(Rdim x0=Rdim(),R rho=0,R hh=0):Base(CylinderDat(x0,rho,hh)){}
   Cylinder(const Base& b):Base(b){}
   R vol()const{ return dat.vol();}
};

///////////////////////// class Boxoid /////////////////////////////////
// intermediate object between box and ellipsoid

struct BoxoidDat{ // data of which determine a boxoid

   Rdim x0;
      // coordinate list of center
   Rdim s;
      // list of half-axes
   R p{0.};
      // for p=2 we get an ellipsoid, for larger p we approach the box
      // for p=1 one has a romboid
   BoxoidDat(){}
      // void Boxoid
   BoxoidDat(Rdim x0_, Rdim s_, R p_):x0(x0_),s(s_),p(p_){}
      // Boxoid with prescribed center and sites of the enclosing box.
      // Further, p_ is the form factor p described above
   typedef BoxoidDat Type;
   CPM_IO
   R getRadius()const;
   Rdim getCenter()const{ return x0;}
   R indFnc(const Rdim&)const;
};

class Boxoid:public Shp<BoxoidDat>{// objects between box and ellipsoid

public:
   typedef Shp<BoxoidDat> Base;
   Boxoid(Rdim x0=Rdim(),Rdim s=Rdim(),R p=0):Base(BoxoidDat(x0,s,p)){}
   Boxoid(const Base& b):Base(b){}
      // no explicit volume formula
};

////////////////// class Ellipsoid ///////////////////////////////////////

struct EllipsoidDat{ // data which determine an ellipsoid

   Rdim x0;
      // coordinate list of center
   Rdim s;
      // list of half-axes

   EllipsoidDat(){}
      // void Ellipsoid
   EllipsoidDat(Rdim x0_, Rdim s_):x0(x0_),s(s_){}
      // Ellipsoid with prescribed center and sites of the enclosing box.
      // Further, p_ is the form factor p described above
   typedef EllipsoidDat Type;
   CPM_IO
   R getRadius()const;
   Rdim getCenter()const{ return x0;}
   R vol()const;
   R indFnc(const Rdim&)const;
};

class Ellipsoid:public Shp<EllipsoidDat>{ // ellipsoids
public:
   typedef Shp<EllipsoidDat> Base;
   Ellipsoid(Rdim x0=Rdim(),Rdim s=Rdim()):Base(EllipsoidDat(x0,s)){}
   Ellipsoid(const Base& b):Base(b){}
   R vol()const{ return dat.vol();}
};

/////////////////////////////// class Box //////////////////////////////
// rectangular parallelepiped

struct BoxDat{ // data which determine a box

   Rdim x0;
      // coordinate list of center
   Rdim s;
      // list of half-axes (=half-sites)

   BoxDat(){}
      // void Box
   BoxDat(Rdim x0_, Rdim s_):x0(x0_),s(s_){}
      // Box with prescribed center and sites of the enclosing box.
      // Further, p_ is the form factor p described above
   typedef BoxDat Type;
   typedef BoxDat CloneBase;
   CPM_IO_V
      // I/0 functions are virtual in order to allow overwriting in
      // derived classes
   R getRadius()const;
   Rdim getCenter()const{ return x0;}
   R vol()const;
   R indFnc(const Rdim& )const;
};

class Box:public Shp<BoxDat>{ // rectangular parallelepipeds
public:
   typedef Shp<BoxDat> Base;
   Box(Rdim x0=Rdim(),Rdim s=Rdim()):Base(BoxDat(x0,s)){}
      // construction from center and half-sites such that
      // these data refer to the master system of refereence
   Box(const Base& b):Base(b){}
   Box(const V<Spc>& vs, const RefSys& rs);
      // a box with axes parallel to those of rs and enclosing
      // the points of vs in a minimum way.
   R vol()const{ return dat.vol();}
   V<Spc> ran(Z n)const;
   V<Spc> lattice(Z n, R site);
      // divides the box into identically sized cells
      // which are nearly cubic (or quadratic for Spc::dimension=2).
      // Each site s[i] satisfies  s[i] >= site and is as close to size
      // as a genuine fraction of the corresponding site of the Box
      // can be. Returned are nRes centers of cells, where nRes is the
      // minimum of n and the total number of cells. If n is smaller than
      // the total number of cells, the sequence of cell centers starts
      // with the the left lower corner and gets filled according to the
      // scheme (for 3D):
      //  x=corner[1];
      //  for (i=1;i<=ns[1];i++){
      //    x+=dx;
      //    xs[1]=x;
      //    y=corner[2];
      //    for (j=1;j<=ns[2];j++){
      //       y+=dy;
      //       xs[2]=y;
      //       z=corner[3];
      //       for (k=1;k<=ns[3];k++){
      //          l++;
      //          z+=dz;
      //          xs[3]=z;
      //          res[l]=refSys(xs);
      //       }
      //    }
      //  }
};

////////////////// class Sphere //////////////////////////////////

struct SphereDat{ // data which determine a sphere
      // intermediate object between box and ellipsoid
   Rdim x0;
      // coordinate list of center
   R r{0.};
      // radius (= 'half-axis')

   SphereDat(){}
      // void Sphere
   SphereDat(Rdim x0_, R r_):x0(x0_),r(r_){}
      // Sphere with prescribed center and radius.

   typedef SphereDat Type;
   CPM_IO
   R getRadius()const{return r;}
   Rdim getCenter()const{ return x0;}
   R vol()const{return Spc::vol(r);}
   R indFnc(const Rdim& x)const{ return (x-x0).abs()-r;}
};

class Sphere:public Shp<SphereDat>{ // spheres
public:
   typedef Shp<SphereDat> Base;

explicit Sphere(Rdim x0=Rdim(),R r=0):Base(SphereDat(x0,r)){}
      // constructor of the same design as all before
explicit  Sphere(const BndSph& bs):
   Base(SphereDat(refSys(bs.get_x()),bs.get_r())){}
      // special constructor from a bounding sphere
   Sphere(const Spc& sp, R r):
      Base(SphereDat(sp.toRdim(),cpmabs(r))){}
      // special constructor from center in Spc and radius
   Sphere(const Base& b):Base(b){}
   R vol()const{ return dat.vol();}
};
