//? cpmbiotope.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_BIOTOPE_H_
#define CPM_BIOTOPE_H_
/*
   Description: Development and test of a class
      which represents space as one-dimensional finite equidistant\
       lattice.
*/


#include <cpminterval.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmGeo::Iv;

using namespace std;
using namespace CpmArrays;

namespace CpmQuantumMechanics{

//////////////////// class Bio ////////////////////////////////////////////

class Bio{ // 'biotope', the substrate where the spinor-valued wave
   // functions 'live'. Recall: physical space is 1-dimensional here.
   typedef Bio Type;
   Z d_; // number of lattice points
   Iv iv_; // interval of continuous space
public:
   CPM_IO
   CPM_NAM(Bio)
   Bio():d_(0){}
   explicit Bio(Z d, Iv const& iv=Iv(-0.5,0.5)):d_(d),iv_(iv){}
   Z d()const{ return d_;}
   Z dd()const{ return d_+d_;}
      // dimension of the space of Dirac-wavefunctions on biotope *this.
      // The spinors living on the lattice points are not really part
      // of the concept 'biotope'. This is the only member function
      // of the present class which refers to the spinors (the inhabitants
      // of the biotope).
   Iv iv()const{ return iv_;}
   Iv range()const{ return iv_;}
   Iv frqRange()const{ R fN=R(0.5)/gap(); return Iv(-fN,fN);}
   R gap()const{ return iv_.abs()/d_;}
      //: gap
      // Distance between adjacent lattice points. Notice that the lattice
      // points are the centers of the d_ sub-intervals of iv_. If end
      // points of iv_ would be lattice points, the gap would be
      // iv_.abs()/(d_-1).
   R gapInv()const{ return R(d_)/iv_.abs();}
      //: gap inverse
   R frqGap()const{ return frqRange().abs()/d_;}
      //: frequency gap
      // Definition is chosen to show that the frequency biotope is treated
      // on equal footing with the position biotope. Of course the final
      // formula is frqGap()=1/(iv_.abs()). And to memorize:
      // gap()*frqGap() = 1/d_
   V<R> points()const{ return iv_.centers(d_);}
      // returns the lattice points as an array
   V<R> frequencies()const{ return frqRange().centers(d_);}

   Bio operator*(R const& r)const{ return Bio(d_,iv_*r);}
      // scaling the biotope, leaving its center fixed
   Bio& operator*=(R const& r){ iv_*=r; return *this;}
      // scaling as a mutating operation
   bool operator==(Bio const& b)const{ return d_==b.d_ && iv_==b.iv_;}
      // natural equality relation
};

} // namespace CpmQuantumMechanics

#endif
