//? cpmh.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_H_H_
#define CPM_H_H_
/*

   Description: class H of quaternions.
      H could be used for a more conventional treatment of rotations.
      In a final form, the treatment of two-dimensional and
      3-dimensional geometry should probably be based on the field C
      and H respectively. Finally also 4-dimensional Minkowski geometry
      had to be added.

   History: Started 02-1-23

*/

#include <cpmlinalg.h>

namespace CpmLinAlg{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmArrays::T2;
   using CpmArrays::X2;

/////////////////////////////// class H /////////////////////////////////
// Quaternions ( H in honor of Hamilton). Arithmetics is implemented
// less schematically than in R4. It aims at utmost efficiency just as
// in was done for complex numbers C. However, no reference to complex
// arithmetics occurs here.

class H{  //  H  quaternions
   typedef H Type;
   R a0;
   R3 a;

public:
// constructors

   H():a0(0.),a(){}
   // default constructor

explicit H(R a0_):a0(a0_),a(){}
   // embedding R

explicit H(const R3& a_):a0(0.),a(a_){}
   // embedding R3

   H(R a0_, const R3& a_):a0(a0_),a(a_){}
   // general constructor

   H(R a0_, R a1, R a2, R a3):a0(a0_),a(a1,a2,a3){}
   // constructor for very efficient code (which avoids to
   // call R3-functions

explicit H(const R4& r):a0(r.x1),a(r.x2,r.x3,r.x4){}
   // conversion from R4

// access
   R& operator[](Z i);
      // no exceptions; always the best match of i in {1,2,3,4}
      // is actually used

   const R& operator[](Z i)const;
      // no exceptions; always the best match of i in {1,2,3,4}
      // is actually used

   R4 toR4()const
      { return R4(a0,a.x1,a.x2,a.x3);}
      // conversion to quadruples of R's

// arithmetics within H
// operator +
   H& operator +=(const H& h)
   { a0+=h.a0; a+=h.a; return *this;}

   H operator +(const H& h)const
   { return H(a0+h.a0, a+h.a);}

// operator -
   H& operator -=(const H& h)
   { a0-=h.a0; a-=h.a; return *this;}

   H operator -(const H& h)const
   { return H(a0-h.a0, a-h.a);}

   H operator -(void)const
   { return H(-a0,-a); }

// operator *
   H operator *(const H& h)const;
      // returns  H(a0*h.a0-(a|h.a), a*h.a0+h.a*a0+crsPrd(a,h.a)).
      // Hunting for efficiency to present implementation does only
      // one function call: constructor H(R,R,R,R).

   H& operator*=(const H& h)
   {  return *this=operator*(h);}

// operator /

   H inv(void)const;

   H operator !(void)const
      { return inv();}

   H operator /(const H& h)const
      { return operator*(h.inv());}

   H& operator /=(const H& h)
      { return operator*=(h.inv());}

//  conjugation
   H con(void)const
      { return H(a0,-a);}

   H operator~(void)const
      { return H(a0,-a);}

// arithmetics with R

   H& operator *=(R r)
   { a0*=r; a*=r; return *this;}

   H operator *(R r)const
   { return H(a0*r,a*r);}

// quaternions and rotations

   R3 operator *(const R3& x)const;
      // action of quaternions on R3 vectors. See e.g.
      // E. Peschl: Analytische Geometrie, BI Hochschultaschenbuecher
      // 15/15a, 1961 p. 121 Satz 69.
      // There h*x:=h*(0,x)*h.con() assuming h.norm()=1.
      // Replacing con() by inv(), as is done in the present code,
      // eliminates the need for this assumption.

  static H fromRot(R phi, const R3& axis);
      // generating a normalized quaternion (the norm of which is 1)
      // by rotation angle and rotation axis.
      // Notice that a constructor of the natural signature is already
      // present, so that the most usual implementation as a constructor
      // is not viable. The axis needs not to be a unit vector, but only
      // the corresponding unit vector determines the axis to be taken.
      // The 'angle' phi will be cut to [0,2*Pi) prior to usage

   X2<R,R3> rot()const;
      // returns angle and axis. In order to find these quantities
      // efficiently, *this needs not to be normalized. We just consider
      // the rotation which belongs to (*this)/abs() without
      // actually
      // doing this normalization

// metric quantities

   R abs(bool careful=true)const
      { return toR4().abs(careful);}

   R norm()const
      // this 'norm' plays an important role in quaternions;
      // one has h1.norm()*h2.norm()=(h1*h2).norm()
      // Not a norm in the sense of normed linear spaces
      // could be written as a0*a0+(a|a)
      { return a0*a0+a.x1*a.x1+a.x2*a.x2+a.x3*a.x3;}

   R normalize();
      // turns *this into a normalized quaternion
      // i.e. abs()==norm()==1.
      // and returns the abs() of the original in order not to
      // loose information. Notice that normalizing has to be done
      // by deviding by abs(); dividing by norm() would not
      // work.

   R dis(const H& h)const
      { return toR4().dis(h.toR4());}

// standard interface

   CPM_IO
   CPM_ORDER
   Z dim()const{ return 4;}
   Word nameOf(void) const{ return Word("H");}
};

} // namespace

#endif

