//? cpmnr.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_NR_H_
#define CPM_NR_H_
/*
   Description: Gives a definition of natural numbers based on
   recursion. This is closer to usefulness than my first
   implementation in class NatNum (see cpmmathfound.h)
*/

#include <string>
#include <iostream>
#include <cpmbasictypes.h>

namespace CpmMathFound{ // foundation of mathematics

   using std::string;
   using std::ostream;
   using CpmRoot::Z;

class Nr{ // natural numbers, based on recursion
   // Here 0 is not a member of Nr (actually 0 is the first
   // addition of human abstraction and imagination to the
   // 'really natural' system of positive numbers).
   // Implementing numbers with more digits than 0 and 1
   // (e.g. 0,1,2,3,4,5,6,7,8,9) can easily be done along
   // the same lines. However, then a slight complication
   // arrises from the fact that then multiplication of single-
   // digit numbers can't be represented without 'carry'.
   // Notice that using a single allowed digit ('stone age coding')
   // would make the definition of all the functions here
   // a bit simpler. However, this representation needs such ammounts
   // of 'paperspace' for large numbers that it is not of
   // practial value.
   // When starting this project 4 days ago, I intended only to
   // make sure that normal recursive C++ functions (general recursive
   // functions in the sense of mathematics) are sufficient
   // to implement natural number arithmetics. Now it ended up
   // with an implementation that seems to be even practicable
   // in the sense that it mutiplies numbers of 1000 binary
   // digits within seconds.

   string rep;
      // string of symbols \in {0,1} that constitute the 'state'
      // of the quantity (number) and that allows to compute
      // basic functions such as Peano's successor function
      // by an algorithm acting on the state descriptors.
      // This is a much more powerful, and according to
      // my judgement also more natural approach than the one prefered
      // in mathematical logic, which is to charcterize the individuality
      // of a particular number by the arithmetic (normalized) expression
      // that creates the number.
      // Based on this state description, the basic arithmetic
      // functions +,*,== can be coded as recursions over the length
      // of the state descriptor. This is much faster than the
      // standard recursion over the series of numbers as
      // defined by the successor function. In present recursion
      // scheme the append function app() plays much the same role as
      // the successor function suc() in the 'naive' recursive
      // definition of number arithmetics.

public:

   explicit Nr(string const& s="1");
      // constructor from a string of 1's and 0's where the leading
      // digit has to be a 1 (if s does not respect these rules
      // the call creates an error. Notice that Nr does not contain 0,
      // so 1 is the only distinquished element, which here is
      // created by the default constructor, so a static Nr::unit
      // or Nr::one is not needed.

   explicit Nr(Z i);
      // so far definition by counting, not by directly setting rep from
      // the bits of i. So this is very slow for large numbers.
      // Notice that large random numbers are available from function
      // random.

   Nr suc()const;
      // successor function of Peano
      // defined as a recursive function (the implementation code
      // calls suc)

   Nr cut()const;
      // If the string is at least two symbols long we returns the string
      // with the last digit removed. If there is only one symbol
      // we return the string unchanged (i.e. Nr("1").cut()==Nr("1")).

   Nr pre()const;
      // predecessor function
      // defined as a recursive function (the implementation code
      // calls pre). Defined such that Nr("1").pre()==Nr("1").
      // This function is given here for symmetry with respect to
      // function suc(). It is not needed for an recursive definition
      // of + and *. (Needed, however, for the non-recursive definition
      // of *)

   char last()const{ return rep[rep.length()-1];}
      // the corresponding 'first' is not needed to define

   Nr app(char x)const;
      // appending a digit at the end, error if x is not '0' or '1'

   bool sin()const{ return rep.length()==1;}
      // says whether the string has a single component

   bool operator==(Nr const& n)const;
      // equality defined as a recursive function
      // not making use of the corresponding function defined by
      // the system for string

   bool operator<(Nr const& n)const;
      // relation '<' defined as a recursive function.
      // not making use of the corresponding function defined by
      // the system for string

   Nr operator+(Nr const& n)const;
      // recursive definition, which is much faster than the one based on
      // 'countdown'

   Nr operator*(Nr const& n)const;
      // recursive definition which is much faster than the one
      // based on successive addition

   bool prnOn(ostream&)const;
      // prints the representation string res

   friend ostream& operator<<(ostream& out, Nr const& n)
      { n.prnOn(out); return out;}

   static Z test(Z z=137);
      // returns the number of errors made, so it has to come out
      // as 0. z gives the maximum number of digits of the random
      // test numbers under consideration

   static Nr random(Z z);
      // random binary number with at most z digits, successive calls
      // yield a random sequence (periodicity is that of built-in
      // data type unsigned long int)
};

} // namespace

#endif
