//? cpmqma.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_QMA_H_
#define CPM_QMA_H_

//   Description: An attempt to reconstruct the treatment of spin chains
//      as given as an example in my article on separated quantum mechanics
//      from 2008.
//   spin chain quantum system, soll mit der ALF Dynamik arbeiten

#include <cpmvlin.h>
#include <cpmx.h>
#include <cpmc.h>
#include <cpmgraph.h>
#include <cpmrecordhandler.h>
#include <cpmalgorithms.h>
#include <cpmclinalg.h>
#include <functional>

using CpmRoot::B;
using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;
using CpmRoot::randomR;


using CpmRootX::RecordHandler;

using CpmGeo::Iv;

using CpmArrays::V;
using CpmArrays::Vlin;
using CpmArrays::X2;
using CpmArrays::X3;
using CpmArrays::T3;
using CpmArrays::T4;
using CpmArrays::R_Vector;
using CpmArrays::Outside;

using CpmFunctions::F;

using CpmGraphics::Graph;

using CpmLinAlg::R2;
using CpmLinAlg::R3;
using CpmLinAlg::C2;
using CpmLinAlg::C4;
using CpmLinAlg::CC2;
using CpmLinAlg::CC4;
using CpmLinAlg::CCCC2;

using CpmAlgorithms::StpCon;
using CpmAlgorithms::ALF;

using std::vector;
using std::function;
using  namespace CpmRoot;

namespace CpmQMa{

// Within this namespace we alway use units in which the numerical values
// of hDirac and c are 1. hPlanck thus has value 2*pi (=CpmRoot::Pi2).
// Three structs for compact input from an ini-file

////////////////////////////// class SysCon //////////////////////////////

   struct SysCon{ // sys control
      Z np;
      R h;
      R qXY;
      R qZ;
      R dotFac;
      R fracRim;
      R phi0;
      R theta0;
      R flipProbability;
      B random;

   public:
      SysCon(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(np);
         cpmrhf(h);
         cpmrhf(qXY);
         cpmrhf(qZ);
         cpmrhf(dotFac);
         cpmrhf(fracRim);
         cpmrhf(phi0);
         cpmrhf(theta0);
         cpmrhf(flipProbability);
         cpmrhf(random);
      }
   };

////////////////////////////// class RunCon //////////////////////////////
// A collection of control parameters which makes up a part of the state of
// instances of class SpinNa.
   struct RunCon{ // run control
      R dt;
         // time step
      Z nSteps;
         // Function run_(Graph, DisConx) will run from t=0 till
         // t=dt*nSteps where dt is the value of rc_.dt at the point of
         // invocation of SpinNa::run(Graph,DisConx).
      B twoRuns;
         // If true, we run runX_ instead of run_. This allows to verify
         // the order of the integration method.
      Z dtAdjMethod;
         // control of manipulation that dt is subjected to when inserted
         // into the constructor of class SpinNa. These manipulations
         // determine the value of dt which is active in the first
         // invocation of function stepLean_()-
         // Presently the valid values are 0, 1, 2, else
         // 0: no change
         // 1: multiplication with a factor which accounts for particle
         //    number and total energy.
         // 2: invoce adj_dt_()
      Z nIter;
      R eps; // if this is > 0, automatic time step control is in action
         // and eps controls the sensitivity to norm deviations.
         // The auto step mode does not depend on enfNor and only slightly
         // on adalf. If the state descriptor SpinNa.vx_ gets a complicated
         // structure during evolution, the auto-selected time step may
         // become rather short, so that it may become very time-consuming
         // to do the step number set in advance.

      R facRed;
      R facInc;
      B autoStep;
      B adalf;
         // sets the ADALF mode instead of the DALF mode of the basic
         // integrator stepLean_()
      B rigid;
      B enfNor;
         // enforce normalization in the basic integrator stepLean_()

      Z viewPeriod;
         // Let n=viewPeriod. The each nth step will be shown on screen.
      Z docuPeriod;
         // Let n=docuPeriod. The each nth step will enter the
         // documentation graphics.
      R tWaitIni;
         // time for inspection of the image made by
         // the first invocation of SpinNa::mark(Graph,DisConx)
      R tWaitView;
         // time for inspection of the image made by
         // SpinNa::mark(Graph,DisConx)
      // tWaitXXX >= 0 activates documentation graphics concerning XXX
      // and gives time for viewing for this time in seconds.
      R tWaitDocuNormError;
      R tWaitDocuEnergies;
      R tWaitDocuEnergyError;
      R tWaitBlochAngles;
      R tWaitSigma3;
      R tWaitNorm2;
      R tWaitEnergy2;

    public:
      RunCon(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(dt);
         cpmrhf(nSteps);
         cpmrhf(twoRuns);
         cpmrhf(dtAdjMethod);
         cpmrhf(nIter);
         cpmrhf(eps);
         cpmrhf(facRed);
         cpmrhf(facInc);
         
         cpmrhf(autoStep);
         cpmrhf(adalf);
         cpmrhf(rigid);
         cpmrhf(enfNor);
         cpmrhf(viewPeriod);
         cpmrhf(docuPeriod);
         cpmrhf(tWaitIni);
         cpmrhf(tWaitView);
         cpmrhf(tWaitDocuNormError);
         cpmrhf(tWaitDocuEnergies);
         cpmrhf(tWaitDocuEnergyError);
         cpmrhf(tWaitBlochAngles);
         cpmrhf(tWaitSigma3);
         cpmrhf(tWaitNorm2);
         cpmrhf(tWaitEnergy2);
      }
   };

////////////////////////////// class DisCon //////////////////////////////
   struct DisCon{ // display control
      Z colMet;
      R gamma;
      B wrtOnShow;
         //: write on show
   public:
      DisCon():colMet(1_Z),gamma(1_R),wrtOnShow(0_B){}
      DisCon(RecordHandler const& rch, Word const& sec)
      {
         cpmrhf(colMet); // colorizing Method
         cpmrhf(gamma);
         cpmrhf(wrtOnShow);
      }
      DisCon(DisCon const&)=default;
   };

/////////////////////////// class Ia /////////////////////////////////////

class Ia{ // interaction descriptor for spin one half
   // Interaction descriptor for spin 1/2
      CC2 h1_;
         // interaction with external field
      CCCC2 h12_;
         // pair interaction

   public:
      Ia(R h, R qXY, R qZ): h1_(CC2::sig3*h),
      h12_(CCCC2::spinSpinXYZ(qXY, qZ)){}

      Z dim()const{ return 2;}
         //: dimension
      C2 operator()(C2 const& x)const{ return h1_*x;}
      C2 operator()(C2 const& x, C2 const& y)const
      { return h12_(x,y).trc1(x);}

      C avr(C2 const& x, C2 const& y)const
         //: average
      { return (y|operator()(x,y));}

   };

//////////////////////////// class SpinNa ////////////////////////////////

using VC2=Vlin<C2,R>;

   class SpinNa{ // spin chain quantum system, implemented such that its
   // dynamics is given by the ALF integrator
      Z np_;
      Ia ia_;     
      RunCon rc_;
      DisCon dc_;
      ALF<VC2,VC2> rep_;
      typedef SpinNa Type;

   public:
      CPM_NAM(SpinNa)

      static R nvx(VC2 const& vx){
         Word loc("SpinNa::nvx(V<C2>)");
         Z mL=3;
         CPM_MA
         R r=1_R;
         for (Z i=1;i<=vx.dim();++i){ //r*=vx.cui(i).abs();
            R ri=vx.cui(i).abs();
            if (ri==0.) cpmwarning("ri=0, i="&cpm(i));
            r*=ri;
         }
         CPM_MZ
         return r;
      }

      static VC2 iniState(SysCon const& sc){
       Word loc("SpinNa::iniState(SysConx const&)");
         Z mL=1;
         CPM_MA
         Z n=sc.np;
         if (sc.random){
            C z0(1.,1.);            
            VC2 res(n);
            for (Z i=1;i<=n;++i){
               C2 ci=C2(z0.ran(),z0.ran());
               ci.nor_();
               res.cui(i)=ci;
            }
            CPM_MZ
            return res;
         }
         else{
            VC2 res(n,C2(sc.theta0,sc.phi0));
            Z dRim=cpmrnd(n * sc.fracRim);
            R flip=sc.flipProbability;
            for (Z i=1;i<=n;++i){
               if (randomR()<flip && i>dRim && i+dRim<=sc.np){
                  res[i].swap_();
               }
            }
            CPM_MZ
            return res;
         }
      }

      SpinNa(StpCon const& sc, VC2 const& vx, R h, R qXY, R qZ, 
         RunCon const& rc, DisCon const& dc);

      StpCon sc()const{ return rep_.sc();}
      R dt()const{ return rep_.dt();}
      R t()const{ return rep_.t();}
      VC2 x()const{ return rep_.x();}
      VC2 y()const{ return rep_.y();} 

      void set_dt_(R dt){ rep_.set_dt_(dt);}

      Ia ia()const{ return ia_;}

      Z parDim()const{ return 2;}
         //: particle dimension

      R n()const
         //: norm
         // Notice that vx_ is the tensor product of the states
         // rep_.x().cui(i). This implies the product structure of the norm.
      {
         R r=1_R;
         for (Z i=1_Z;i<=np_;++i){ //r*=vx.cui(i).abs();
            R ri=x().cui(i).abs();
            if (ri==0.) cpmwarning("ri=0, i="&cpm(i));
            r*=ri;
         }
         return r;
      }

      C sclPrd(VC2 const& vx)const
         //: scalar product
      {
         C z=C::one;
         for (Z i=1;i<=np_;++i) z*=(x().cui(i)|vx.cui(i));
         return z;
      }

      void renormalize_()
         // rescale the vectors rep_.x().cui(i) to unit vectors and change the
         // velocities vv_.cui(i) accordingly
      {
         if (rc_.eps < 0_R) return; // do nothing
         for (Z i=1;i<=np_;++i){
            R ni=x().cui(i).abs();
            if (cpmabs(ni-1_R) > rc_.eps ){
               // thus for rc_.eps == 0 we enforce full normalization
               R niInv=1./ni;
               x().cui(i)*=niInv;
               y().cui(i)*=niInv;
            }
         }
      }

      

      C hInt()const;
         //: h interaction (h for Hamiltonian)
         // More precisely: The expectation value of the particle-particle
         // interaction Hamiltonian in the state vx_

      // Energy contributions of the individual particles to the energy
      // expectation value. For reasons of efficiency this will not be used
      // to built up the total energy.
      C hInt_i(Z i)const;
         // gives the interaction of particle i with its direct neighbors.
         // hInt() = sum from i=1 to i= np_ of hInt_i(i).No factor 0.5 !

      C eSep_i(Z i)const;
         // gives the interaction of particle i with external field
      C eTot_i(Z i)const;
         // total energy of particle i

      R test2()const;
         // if hInt() and hInt_i cooperate correctly the return value
         // differs from 0 by numerical noise.

      // State descriptors of the individual particles.
      R theta_i(Z i)const{ return x().cui(i).blcAng()[1];}
      R phi_i(Z i)const{ return x().cui(i).blcAng()[2];}
      R2 blc_i(Z i)const{ return x().cui(i).blcAng();}

      // mean value of Pauli matrix sigma 3 over the chain
      R sig3()const
      {
         R res=0;
         for (Z i=1;i<=np_;++i){
            C2 ci=x().cui(i);
            res+=(ci|CC2::sig3*ci)[1];
         }
         return res/np_;
      }

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      VC2 col()const;
         //: color

      R2 blcAng()const
         //: Bloch angles
         // returns mean values of Bloch angles over the chain
      {
         R r1=0, r2=0;
         for (Z i=1;i<=np_;++i){
            R2 ri=x().cui(i).blcAng();
            r1+=ri[1];
            r2+=ri[2];
         }
         R npInv=1./np_;
         return R2(r1*npInv,r2*npInv);
      }

      C eSep()const;
         // total interaction energy with external field

     C eTot()const{ return eSep()+hInt();}
         // total interaction energy with external field and neighbor
         // particles

      void step_();
         //: step
         // implemented by means of stepLean

      void safeEvo_(R tf){
         rep_.safeEvo_(tf);
      } 

      void safeEvo2_(R tf){
         rep_.safeEvo2_(tf);
      }   

      void mark(Graph& g, Z met=1)const;
         //: mark

      V<V<R2>> run_(Graph& g );
         //: run

  //    void runX_(Graph& g);
         //: run extended

  //    void adj_dt_();
         //: adjust dt
         // Assumes that dt_ as set by the constructor. Successively halves
         // the steps till the rc_.nIter steps run fine. For security the
         // step is then halved again to give it the final value.

  }; // class SpinNa end

//////////////////////////// class SpinNaa ////////////////////////////////

class WC2: public Vlin<C2,R>{

   typedef Vlin<C2,R> Base;

public:
   WC2(Vlin<C2,R> const& x):Vlin<C2,R>(x){} 
   WC2(Z n=0):Vlin<C2,R>(n){}
   WC2(Z n, C2 const& c2):Vlin<C2,R>(n,c2){}

   C operator|(WC2 const& x)const
      //. my notation for scalar products 
   {
      C res=C::one;
      for (Z i=1; i<=dim(); ++i){
         res*=(cui(i)|x.cui(i));
      }
      return res;
   }

   WC2 times(C const& c)const
   // More natural would be to define 
   // WC2 operator *(C const& c)const.
   // But this would spoil usages of operator*(R const&)
   // from the base class Vlin(C2,R). I consider this
   // an annoying bug of C++. Why not several operators * 
   // can't be discerned on basis of the type of their second argument
   // ( R and C in the present case)? Of course, the problem only
   // arrises if the second operator * comes in with forming a
   // derived class! In defining the class C of complex numbers one
   // certainly defines operator z1*z2 and z1*r for complex z1,z2 and
   // real r. 
   // We use the function 'times' only for an experiment, that turned out
   // not to work.
   {
      WC2 res(*this);
      for (Z i=1; i<=dim(); ++i){
         res.cui(i)*=c;
      }
      return res;      
   }

   void nor_(){ 
      //: normalize
      for (Z i=1; i<=dim(); ++i) cui(i).nor_();
   }

   R exc()const
      //: excess (of norm relative to a unit vector)
   {
      auto res=0_R;
      for (Z i=1;i<=dim();++i){ 
         R exci=cpmabs(cui(i).abs()-1_R);
         if (exci>res){
            res=exci;
         }
      }
      return res;
   }  

};

   class SpinNaa{ // spin chain quantum system, implemented such that its
   // dynamics is given by the ALF integrator
      SysCon sc_;
      RunCon rc_;
      DisCon dc_;
      Ia ia_;
      Z np_;
      R t_;
      WC2 x_;
      WC2 v_;
      R dt_;
      function<WC2(R,WC2)> dot_;
    
      typedef SpinNaa Type;

   public:
      CPM_NAM(SpinNaa)
  
      static WC2 iniState(SysCon const& sc);

      SpinNaa(SysCon const& sc, RunCon const& rc, DisCon const& dc, 
         R const& t, WC2 const& x);

      SysCon sc()const{ return sc_;}
      R dt()const{ return dt_;}
      R t()const{ return t_;}
      WC2 x()const{ return x_;}
      WC2 v()const{ return v_;} 
      Z dim()const{ np_;}

      void set_dt_(R dt){ dt_=dt;}

      Ia ia()const{ return ia_;}

      Z parDim()const{ return 2;}
         //: particle dimension

      R n()const{ return x_.abs();}
         //: norm

      void renormalize_(R const& tol);
         // normalizes x_
         // and also applies the normalizing factor to v_
 
       
      C hInt()const;
         //: h interaction (h for Hamiltonian)
         // More precisely: The expectation value of the particle-particle
         // interaction Hamiltonian in the state vx_

      // Energy contributions of the individual particles to the energy
      // expectation value. For reasons of efficiency this will not be used
      // to built up the total energy.
      C hInt_i(Z i)const;
         // gives the interaction of particle i with its direct neighbors.
         // hInt() = sum from i=1 to i= np_ of hInt_i(i).No factor 0.5 !

      C eSep_i(Z i)const;
         // gives the interaction of particle i with external field
      C eTot_i(Z i)const;
         // total energy of particle i

      R test2()const;
         // if hInt() and hInt_i cooperate correctly the return value
         // differs from 0 by numerical noise.

      // State descriptors of the individual particles.
      R theta_i(Z i)const{ return x().cui(i).blcAng()[1];}
      R phi_i(Z i)const{ return x().cui(i).blcAng()[2];}
      R2 blc_i(Z i)const{ return x().cui(i).blcAng();}

      // mean value of Pauli matrix sigma 3 over the chain
      R sig3()const
      {
         R res=0;
         for (Z i=1;i<=np_;++i){
            C2 ci=x().cui(i);
            res+=(ci|CC2::sig3*ci)[1];
         }
         return res/np_;
      }

      R3 sig(Z i)const;
         //: sigma
         // expectation values of the three Pauli matrices
         // at chain position i

      VC2 col()const;
         //: color

      R2 blcAng()const
         //: Bloch angles
         // returns mean values of Bloch angles over the chain
      {
         R r1=0, r2=0;
         for (Z i=1;i<=np_;++i){
            R2 ri=x().cui(i).blcAng();
            r1+=ri[1];
            r2+=ri[2];
         }
         R npInv=1./np_;
         return R2(r1*npInv,r2*npInv);
      }

      C eSep()const;
         // total interaction energy with external field

     C eTot()const{ return eSep()+hInt();}
         // total interaction energy with external field and neighbor
         // particles. Deviates from constancy only by terms of
         // second (?) order in the time step.
       
      C eTotA()const{
         C res;
         for (Z i=1;i<=np_;++i){ res+=(x_.cui(i)|v_.cui(i));}
         return C::I*res;
      }

      C eTotB()const{ return eSep()+hInt()*2_R;}
         // eTotA == eTotB up to numerical noice. It was quite unexpected
         // that it is significantly far from the effective constant of
         // motion eTot.
// example for eTotA == eTotB != eTot
//
// eTotA =//C
// 55.487019878483451001
// -1.2442913995165671981e-18
//
// eTotB =//C
// 55.487019878483451005
// -1.3662641439211864469e-18
//
// eTot =//C
// 34.413137409608558152
// -6.7508525896167737024e-19

      C eTotC()const{ return (x_|v_.times(C::I));}
//       naturally looking but bad experiment: 
// example
// eTotC =//C
// 9.5792664649990281341e-237
// -5.9517630546828562524e-254

      void stepLean_();

      void step_();
         //: step
         // implemented by means of stepLean
     
      void mark(Graph& g, Z met=1)const;
         //: mark

      V<V<R2>> run_(Graph& g );
         //: run

  }; // class SpinNaa end

} // namespace

#endif
