//? cpmqmtemplates.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_QMTEMPLATES_H_
#define CPM_QMTEMPLATES_H_
/*
   Description: Template classes for constructing quantum systems
      in a manner illustrating the midpoint integrator publication
*/
#include <cpmrvector.h>
#include <cpmc.h>
#include <cpmf.h>
namespace CpmQuantumMechanics{

   using namespace CpmRoot;
   using namespace CpmRootX;
   using CpmArrays::R_Vector;
  
   template <class Op, class St>
   // Op and St need to implement some of the functions, which
   // classes Observable and State implement. Inspection of the simple
   // code says precisely what these few functions are.
   class Chebyshev{ //The Chebyshev one-step integraton method
      // The Chebyshev one-step integraton method as advocated by
      // Hans De Raedt.
      // As Hans stated: very simple to implement.
      Op a_; // skew-symmetric operator satisfying |a_|<1.
         // Made from the operator under primary consideration
         // by multiplication with a real factor.
      St s_;
         // State, the evolution of which is under consideration
      R norm_; 
         // The operator under consideration is a_*norm_
      R eps_; // epsilon
      R sec_; // security factor
      void ini_();
         // initializes norm_ and tests integrity of a_ and s_
   public:
      Chebyshev(Op const& A, St const& s, R sec=1.2, R eps=1e-3):
         a_(A),s_(s),sec_(sec),eps_(eps){ini_();}
         // A should be a skew-symmetric linear operator,
         // and s the state vector on which exp(tA) is to operate.
         // sec and eps are method parametes which control the
         // determination
         // of the operator norm i.e. norm_=|A|. Notice that for the
         // properly initialized operator a_ we have |a_| < 1.
      St step(R t)const;
         // returns the evolved state exp(t*a_*norm_)*s_
      void step_(R t);
         // changes the state into the evolved one
      St getState()const{ return s_;}
         // returns the state, e.g. for analysis
   };

template <class Op, class St>
void Chebyshev<Op,St>::ini_(){
   static Word loc("Chebyshev<>::ini_()");
   Z mL=3;
   CPM_MA
   const Z iterMax=200;
   Z n=s_.dim();
   St st0(n);
   St st=st0.ran();
   R res=-1e20;
   bool work=true;
   Z iter=0;
   while (work){
      iter++;
      st=a_*st;
      R resNew=st.normalize_();
      R diff=CpmRoot::dis(res,resNew);
      if (diff<eps_){
         work=false;
         res=0.5*(res+resNew);
      }
      else if (iter>iterMax){
         work=false;
         cpmerror(loc&": stopped after "&cpm(iter)&" iterations");
         res=resNew;
      }
      else{
         res=resNew;
      }
   }
   norm_=res*sec_;
   R normInv=cpminv(norm_);
   a_*=normInv;
   CPM_MZ
}

template <class Op, class St>
St Chebyshev<Op,St>::step(R t)const
// needs a fast and accurate implementation of
// the Bessel functions of integer order.
{
   static Word loc("Chebyshev<>::step(R)");
   Z mL=3;
   CPM_MA
   R tRel=t*norm_;
   R fac=1.4;
     // factor arrises as e/2, bit larger does not hurt
   Z kMax=1+cpmtoz(fac*cpmabs(tRel)); 
   St res=s_*besselJ(0,tRel);
   St s0=s_, s1=a_*s_;
   for (Z k=1;k<=kMax;++k){
      res+=s1*(besselJ(k,tRel)*2);
      St temp=s1;
      s1=(a_*temp)*2+s0;
      s0=temp;
   }
   CPM_MZ
   return res;
}

template <class Op, class St>
void Chebyshev<Op,St>::step_(R t)
{
   s_=step(t);
}

//////////////////////////// class MidPoint ////////////////////////////

template < class Op, class St>
   class MidPoint{ //  direct midPoint integraton method
      Op D;
      St psi;
      R t;
      St phi;
   public:
      MidPoint(Op const& D0, St const& psi0, R t0):
         D(D0),psi(psi0),t(t0),phi(D(psi,t)){}
 
      void step(R dt)
         // changes the state into the evolved one over a small time span
      {   
         R tau=dt*0.5;
         t+=tau;
         psi+=phi*tau;
         phi+=D(D(psi,t),t)*dt;
         if (!D.timeIndependent()){ 
            phi+=D(psi,t+tau)-D(psi,t-tau); 
         }
         psi+=phi*tau;
         t+=tau;
         if (D.enforceNormalization()){
            R fac=psi.normalize_();
            R facInv=1./fac;
            phi*=facInv;
         }
      }
         
      St getPsi()const{ return psi;}
      St getPhi()const{ return phi;}
         // returns the state, e.g. for analysis
      R getTime()const{ return t;}
      R getNorm()const{ return psi.abs();}
      C getEnergy()const{ return (psi|phi);}

      void revert(){phi*=-1;}

      void evolve(R t1, Z n){
         R dt=(t1-t)/n;
         for (Z i=1;i<=n;++i) step(dt);
      }

      static R testOfComposition(Op const& D0, St const& psi0, R t0,
         R t1, R t2, Z n1, Z n2)
      {
         MidPoint m0(D0,psi0,t0);
         m0.evolve(t1,n1);
         St psi1=m0.getPsi();

         MidPoint m1(D0,psi1,t1);
         m1.evolve(t2,n2);
         St psi2= m1.getPsi();

         MidPoint m12(D0,psi0,t0);
         m12.evolve(t2,n1+n2);
         St psi12=m12.getPsi();

         return (psi2-psi12).norm();
      }

      static R testOfReversibility(Op const& D0, St const& psi0, R t0,
         R t1, Z n) 
      {
         MidPoint m0(D0,psi0,t0);
         St phi0=m0.getPhi();
         m0.evolve(t1,n);
         m0.evolve(t0,n);
         St psif=m0.getPsi();
         St phif=m0.getPhi();
         return psi0.dis(psif)+phi0.dis(phif);
      }

      static R testOfNormConservation(Op const& D0, St const& psi0, R t0,
       R t1, Z n) 
      {
         MidPoint m0(D0,psi0,t0);
         R n0=psi0.norm();
         m0.evolve(t1,n);
         R nf=m0.getPsi().norm();
         return (nf-n0)/(n0+nf);
      }

      static R testOfConservation(Op const& D0, St const& psi0, R t0,
         R t1, Z n) 
      {
         R_Vector epsReal(n);
         R_Vector epsImag(n);
         R_Vector nu(n);
         R_Vector y(n);
         R dt=(t1-t0)/n;
         MidPoint m0(D0,psi0,t0);
         for (Z i=1;i<=n;++i){
            m0.step(dt);
            C vc=m0.getEnergy();
            R vr=m0.getNorm();
            epsReal[i]=vc[1];
            epsImag[i]=vc[2];
            nu[i]=vr*vr;
            y[i]=cpmabs(vc[1])/vc.abs();
         }
         R diff1=nu.der(dt).dis(epsReal*2);
         R diff2=nu.dis(R_Vector(n,1.));
         R diff3=y.sup();
         R diff4=n*epsImag.variation()[2]/epsImag.mean();
         R error=diff1+diff2+diff3;
         if (D0.timeIndependent()) error+=diff4;
         return error;
      }
   };

////////////////////////////// class GMPM ////////////////////////////

template < class Op, class In, class St>
   class GMPM{ // general midpoint method
      // The .
      Op D;
      In f;
      St psi;
      R t;
      St phi;
      static Z method;
   public:
      GMPM(Op const& D0, In const& f0, St const& psi0, R t0):
         D(D0),f(f0),psi(psi0),t(t0),phi(D(psi,t)+f(t))
         {cpmmessage("GMPM constructed");}
 
      void step(R dt)
         // changes the state into the evolved one over a small time span
      {   
         R ti=t, tf=t+dt;
         R tau=dt*0.5;
         t+=tau;
         psi+=phi*tau;
         if (method==0)
            phi=(D(psi,t)+f(t))*2-phi; // Leap-frog, not too bad
         else
             phi+=(D(D(psi,t),t)+D(f(t),t))*dt+D(psi,tf)+
               f(tf)-D(psi,ti)-f(ti);  
               // full method
            // phi+=D(phi,t)*dt+D(psi,tf)+f(tf)-D(psi,ti)-f(ti);
            // not approved
         psi+=phi*tau;
         t+=tau;
      }

      void setMethod(Z m){ method=m;}
         
      St getPsi()const{ return psi;}
      St getPhi()const{ return phi;}
         // returns the state, e.g. for analysis
      R getTime()const{ return t;}
      R getNorm()const{ return psi.abs();}
      C getEnergy()const{ return (psi|phi);}

      void revert(){phi*=-1;}

      void evolve(R t1, Z n){
         R dt=(t1-t)/n;
         for (Z i=1;i<=n;++i) step(dt);
      }

      static R testOfComposition(Op const& D0, In const& f0,
         St const& psi0, R t0, R t1, R t2, Z n1, Z n2)
      {
         GMPM m0(D0,f0,psi0,t0);
         m0.evolve(t1,n1);
         St psi1=m0.getPsi();

         GMPM m1(D0,f0,psi1,t1);
         m1.evolve(t2,n2);
         St psi2= m1.getPsi();

         GMPM m12(D0,f0,psi0,t0);
         m12.evolve(t2,n1+n2);
         St psi12=m12.getPsi();

         return (psi2-psi12).norm();
      }

      static R testOfReversibility(Op const& D0, In const& f0,
         St const& psi0, R t0, R t1, Z n) 
      {
         GMPM m0(D0,f0,psi0,t0);
         St phi0=m0.getPhi();
         m0.evolve(t1,n);
         m0.evolve(t0,n);
         St psif=m0.getPsi();
         St phif=m0.getPhi();
         return psi0.dis(psif)+phi0.dis(phif);
      }

      static R testOfNormConservation(Op const& D0, In const& f0,
         St const& psi0, R t0, R t1, Z n) 
      {
         GMPM m0(D0,f0,psi0,t0);
         R n0=psi0.norm();
         m0.evolve(t1,n);
         R nf=m0.getPsi().norm();
         return (nf-n0)/(n0+nf);
      }

      static R testOfConservation(Op const& D0, In const& f0,
         St const& psi0, R t0, R t1, Z n) 
      {
         R_Vector epsReal(n);
         R_Vector epsImag(n);
         R_Vector nu(n);
         R_Vector y(n);
         R dt=(t1-t0)/n;
         GMPM m0(D0,f0,psi0,t0);
         for (Z i=1;i<=n;++i){
            m0.step(dt);
            C vc=m0.getEnergy();
            R vr=m0.getNorm();
            epsReal[i]=vc[1];
            epsImag[i]=vc[2];
            nu[i]=vr*vr;
            y[i]=cpmabs(vc[1])/vc.abs();
         }
         R diff1=nu.der(dt).dis(epsReal*2);
         R diff2=nu.dis(R_Vector(n,1.));
         R diff3=y.sup();
         R diff4=n*epsImag.variation()[2]/epsImag.mean();
         R error=diff1+diff2+diff3;
         if (D0.timeIndependent()) error+=diff4;
         return error;
      }

   static R testOfOrder(Op const& D0, In const& f0, St const& psi0,
      R t0, R t1, Z n , Z nRuns) 
      {
         V< V<St> > res(nRuns, V<St>(n));
         R dt=(t1-t0)/n;
         CyclicAlarm readAlarm(1);
         Z nMem=n, readPeriod=1;
         Z i,j;
         for (Z k=1;k<=nRuns;++k){
            readAlarm.restartWithNewPeriod(readPeriod);
            cpmdebug(readPeriod);
            readPeriod*=2;
            Z ia=1;
            GMPM m0(D0,f0,psi0,t0);
            for (i=1;i<=n;++i){
               m0.step(dt);
               if (readAlarm()){
                  res[k][ia]=m0.getPsi();
                  ia++;
               }
            }
            n*=2;
            dt*=0.5;
         }
         R_Vector dis(nRuns-1);
         R nMemInv=1./nMem;
         R fac=1;
         for (i=1;i<nRuns;++i){
            R sum=0;
            for (j=1;j<=nMem;++j){
               sum+=(res[nRuns][j]-res[i][j]).abs();
            }
            dis[i]=sum*nMemInv*fac;
          //  fac*=2;
         }
         Graph gr;
         gr.show(dis);
         cpmwait(20);
         
         return 1;
      }
   };

   template < class Op, class In, class St>
   Z GMPM<Op,In,St>::method=1;

} // namespace
#endif
