//? cpmqp1.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_QP1_H_
#define CPM_QP1_H_
/*************************************************************************

QP1 stands for quantumparticle in one spatial dimension

Unlike the treatments in cpmstate1.h, cpmstate1.cpp
we introduce here a method which allows to follow the motion of
travelling and broadening wave packets. The clue is to use
the associative container 'map' as a representationof the wave
function.

**************************************************************************/
#include <map>
#include <cpmbas1.h> // includes cpmm.h
#include <cpmcmatrix.h>
#include <cpmgraph.h>

using namespace CpmRoot;
using namespace CpmFunctions;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;
using namespace CpmImaging;
using namespace std;

namespace CpmQuantumMechanics{

using Wf_ = map<Z, C>;
   // wave function representation tool

/////////////////// class Wf /////////////////////////////////////////////
// wave function of quantum particle in one spatial dimension

class Wf
{
   Wf_ rep_;
public:
   typedef Wf Type;
   CPM_IO
   Wf():rep_(){}
   Wf(Wf_ const& rep):rep_(rep){}
   Wf(F<R, C> const& f, IvZ const& ivz, R const& d):rep_()
   {
      for (Z i=ivz.b();i<=ivz.e();++i) rep_[i]=f(d*R(i));
   }

   Word nameOf()const{ return "Wf";}

   Wf con()const;
      //: conjugate
      // Returns the complex conjugate of *this.
   Wf re()const;
   Wf im()const;
   C operator|(Wf const& psi)const;
   R abs2()const{ return (*this|*this).abs(); }
   R abs()const{ return cpmsqrt((*this|*this).abs()); }
   Z size()const{ return rep_.size();}

   Wf operator + (Wf const& psi)const;
   Wf operator - (Wf const& psi)const;
   Wf operator * (C const& z)const;
   Wf operator * (R const& r)const;
   Wf operator + (Z const& i)const;
      // shifting the keys
   Wf mult(F<R,R> const& f, R const& d=1.)const;
      // for multiplication operstors

   Wf& operator += (Wf const& psi);
   Wf& operator -= (Wf const& psi);
   Wf& operator *= (C const& z);
   Wf& operator *= (R const& r);

   R nor_();
      //: normalise.
      // Returns the norm of original *this and changes
      // *this into its normalised version.

   void mark(Graph& gr, R const& d=1.)const;

   Wf dot(R const& d, R const& m)const;
      //: dot
      // Time derivative for free evolution

   Wf der(R const& d)const;
      //: der
      // Spatial derivation

   static Wf ran(IvZ const& ivz, Z j=0);

   void cut_(R const& eps);

   static R opNor(F<Wf,Wf> const& op, IvZ const& ivz);

   M<Z,C> toM()const;
};

/////////////////// class Qp1 /////////////////////////////////////////////
// quantum particle in one spatial dimension
class Qp1{
   Wf psi_;
   Wf phi_;
   R t_;
   R d_;
   R m_;
   R eps_;
   R norm_;
   R avr_; //: average, expectation value of the hamiltonian or
      // the momentum.
   void ini_(){
      phi_=psi_.dot(d_,m_);
      norm_=psi_.abs();
      avr_=(C::I*(psi_|phi_)).real();
   }
public:
   Qp1(Wf const& psi, R const& d, R const& m ):psi_(psi), t_(0.), d_(d),
      m_(m)
   {
      psi_.nor_();
      ini_();
   }
   void step_(R dt);
   void step_(R dt, F<Wf,Wf> const& f);
      // Step with any hermitean operator f as generator.
   void shift_(R ds);

      // ds>0 shifts to the right. The naive discrete derivation
      // operator seems not to be a good generator of translations.
   void setEps_(R const eps){ eps_=eps;}
   Wf psi()const{ return psi_;}
   Z size()const{
      Z s1=psi_.size();
//      Z s2=phi_.size();
   //   cpmassert(s1==s2,"Qp1::size");
      return s1;
   }
   R norm()const{ return norm_;}
   R avr()const{ return avr_;}
   R t()const{ return t_;}
   R d()const{ return d_;}
};

} // namespace CpmQuantumMechanics

#endif
