//? cpmqp2.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_QP2_H_
#define CPM_QP2_H_
/*************************************************************************

QP2 stands for two quantum particles each of which moves in a one
dimensional discrete space.

Unlike the treatments in cpmstate1.h, cpmstate1.cpp
we introduce here a method which allows to follow the motion of
travelling and broadening wave packets. The clue is to use
the associative container 'map' as a representation of the wave
function.

**************************************************************************/
#include <climits>
#include <map>
#include <functional>
#include <cpmbas1.h> // includes cpmm.h
#include <cpmcmatrix.h>
#include <cpmgraph.h>

using namespace CpmRoot;
using namespace CpmFunctions;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;
using namespace CpmImaging;
using namespace std;

namespace CpmQuantumMechanics{

// a collection of control parameters to be used in Wf2::mark
struct GraphControl{
   Z hw_{0}; // half-width of cross mark
   Z ha_{0}; // hue angle of cross color ( 0 <= ha_ <= 360)
   Z im_{0}; // interpolation method
   Z ac_{0}; // active if ac_ !=  0
   GraphControl(){}
   GraphControl(Z hw, Z ha, Z im, Z ac):hw_(hw),ha_(ha),im_(im),ac_(ac){}
};

using Wf2_ = map<Z2, C>;
   // wave function representation tool

/**************************************************************************
The allowed two-particle positions are represented as points points in the 'unlimited' lattice Z2. Of course, Z3 would be used for three particles ...
Consider an arbitrary point p=(i1,i2) in Z2 (i.e. Z2 p(i1,i2) as a valid
C++ declaration). To associate a real world situation with entity we
consider a Euclidean plane with an orthogonal reference system in it. Let
A1 and A2 be the axes of this reference system. On these axes we choose
physical lengths d1 and d2. Then we consider the situation that particle 1
is located at a point p1 of the plane with coordinates (d1*i1,0) and
particle 2 is located at a point p2 with coordinates (0,d2*i2). Associating
a complex number with each conceivable situation of this type is what
quantum mechanics teaches us to be the minimal structure which allows formulating a dynamical law (the Schroedinger equation for the wave function
which our association of complex numbers forms).
The present framework is that described in
   http://www.ma.utexas.edu/mp_arc/c/06/06-356.pdf
with the exception that the linear 'biotopes' of the particles are no tori
but unlimited chains. Further, for the stepwise solution of the Scgroedinger
equation we need the densified asynchronous leapfrog integrator instead of
the direct midpoint method.

**************************************************************************/

/////////////////// class Wf2 /////////////////////////////////////////////
// wave function of the crossway system which consists  of 2 quantum
// particles each of which lives in one spatial dimension.

class Wf2
{
   Wf2_ rep_;
public:
   typedef Wf2 Type;
   CPM_IO
   Wf2():rep_(){}
   Wf2(Wf2_ const& rep):rep_(rep){}
   Wf2(F<R2,C> const& f, IvZ const& ivz1, IvZ const& ivz2, R2 const& d):
   rep_()
   {
      for (Z i1=ivz1.b();i1<=ivz1.e();++i1){
         for (Z i2=ivz2.b();i2<=ivz2.e();++i2){
            rep_[Z2(i1,i2)]=f(R2(d.x1*R(i1),d.x2*R(i2)));
         }
      }
   }

   Wf2(C_Vector const& cv1, C_Vector const& cv2):rep_()
   {
      for (Z i1=cv1.b();i1<=cv1.e();++i1){
         for (Z i2=cv2.b();i2<=cv2.e();++i2){
            rep_[Z2(i1,i2)]=cv1.cui(i1)*cv2.cui(i2);
         }
      }
   }

   Word nameOf()const{ return "Wf2";}

   Wf2 con()const;
      //: conjugate
      // Returns the complex conjugate of *this.
   Wf2 smooth()const;
      // returns a smoothed version of *this
   Wf2 expand()const;
      // returns an expanded version of *this
   Wf2 re()const;
   Wf2 im()const;
   C operator|(Wf2 const& psi)const;
   R abs2()const{ return (*this|*this).abs(); }
   R abs()const{ return cpmsqrt((*this|*this).abs()); }
   Z size()const{ return rep_.size();}

   Wf2 operator + (Wf2 const& psi)const;
   Wf2 operator - (Wf2 const& psi)const;
   Wf2 operator * (C const& z)const;
   Wf2 operator * (R const& r)const;
   Wf2 operator + (Z2 const& i)const;
      // shifting the keys
   Wf2 mult(F<R2,R> const& f, R2 const& d=R2(1.,1.))const;
      // for multiplication operstors

   Wf2& operator += (Wf2 const& psi);
   Wf2& operator -= (Wf2 const& psi);
   Wf2& operator *= (C const& z);
   Wf2& operator *= (R const& r);

   R nor_();
      //: normalise.
      // Returns the norm of original *this and changes
      // *this into its normalised version.

   void mark(Graph& gr, GraphControl const& gc)const;
      //: mark
      // graphical representation as a color districution on
      // a 2-dimensional lattice the lattice points of which have integer
      // coordinates. The point with coordinates (0,0) gets marked
      // by a line cross (line thickness 1 pixel) under the control
      // of the second argument.

   Wf2 dot(R2 const& d, R2 const& m, F<R,R> const& f12,
      F<R,R> const& f2)const;
      //: dot
      // Time derivative for free evolution

   Wf2 der1(R2 const& d)const;
      //: der1
      // Spatial derivation in 1-direction

   Wf2 der2(R2 const& d)const;
      //: der2
      // Spatial derivation in 2-direction

   static Wf2 ran(IvZ const& ivz1, IvZ const& ivz2, Z j=0);

   static Wf2 hamFunc(Wf2 const& psi, R2 const& d, R2 const& m,
      F<R,R> const& f12, F<R,R> const& f2){
      return psi.dot(d,m,f12,f2)*C::I;}

   static Wf2 p1Func(Wf2 const& psi, R2 const& d){
      return psi.der1(d)*C::I;}

   static Wf2 p2Func(Wf2 const& psi, R2 const& d){
      return psi.der2(d)*C::I;}

   static Wf2 multFunc(Wf2 const& psi, F<R2,R> const& f, R2 const& d){
   return psi.mult(f,d);}

   static C wf(R2 const& x, R2 const& x0, R2 const& w, R2 const& p){
      R2 y(x.x1-x0.x1, x.x2-x0.x2);
      return C::expi(p.x1*y.x1+p.x2*y.x2)*
         exp(-0.5*(y.x1*y.x1/w.x1 + y.x2*y.x2/w.x2));}

   static C bell(R const& x, R const& x0, R const& w, R const& p){
      R y=x-x0;
      R yr=y/w;
      return C::expi(p*y)*exp(-0.5*(yr*yr));}

   void cut_(R const& eps);

   static R opNor(F<Wf2,Wf2> const& op,
      IvZ const& ivz1, IvZ const& ivz2,
         Z const& nSteps, bool graph=false);
      //. operator norm

   static R test1(IvZ const& ivz1, IvZ const& ivz2);
      // test of linear algebra in Wf2
      // returns an error sum which should nothing but
      // numerical noise.

   X3<IvZ,IvZ,R> biotope()const;
      //. a rectangular subset of the lattice of points in the
      // plane which have integer coordinates. This is the minimal
      // such rectangle that contains the keys of rep_.
      // As the third component we give the maximum absolute value
      // of the wavefunction.

   M<Z2,C> toM()const;
      //: to M (map class of C+-)

};

/////////////////// class Qp2 /////////////////////////////////////////////
// Two quantum particles in one spatial dimensions
class Qp2{
   Wf2 psi_; // state vector, wave function
   Wf2 phi_; // time derivative of the state vector
  // F<Wf2,Wf2> ho_;
   F<Wf2,Wf2> ho_;
      // hermitean operator
      // in applications often the Hamilton operator
   R t_{0.}; // time
   R dt_{0.}; // time step in use
   R dtRel_{1.}; // dt_/dtCrit
   Z count_{0};
   R eps_{0.}; // controlling small values of the wave function
   R eta_{0.}; // controlling norm groth. Only if etaValid_==true
      // eta_ may trigger a call to adjustStep_()
   bool etaValid_{false};
   bool graph_{false};
   Z normSteps_{100};
   R norm_; // norm of psi_
   R norm0_; // for symmetry, actually, by definition of nor_(),
             // always norm0_ == 1.
   R avr_;
      //: average, expectation value of the operator ho_
   R avr0_; // initial value of av_
   R dRelNorm_; //  (norm_-norm0_)/norm0_
   R dRelAvr_; // (avr_ - avr0_)/avr0;
   R hoNorm_; // operator norm of ho_

   void ini_(){
    //  t_=0.;
      psi_.nor_(); // turn psi_into a unit vector
      phi_=ho_(psi_)*C::Ibar;
         // phi_ = time derivative of psi_
//      Graph gr;
//      phi_.mark(gr,GraphControl());
//      gr.vis();
//      cpmwait(12,2);
      norm_=psi_.abs();
      norm0_=norm_;
      cpmassert(norm0_!=0.,"Qp2::ini_()");
      dRelNorm_=(norm_-norm0_)/norm0_;
      avr_=(C::I*(psi_|phi_)).real();
      avr0_=avr_;
   //   cpmcerr<<"crit:"<<C::I*(psi_|phi_)<<endl;
     // cpmwait(30,2);
      cpmassert(avr0_!=0.,"Qp2::ini_()");
      dRelAvr_=(avr_-avr0_)/avr0_;
      auto bio=psi_.biotope();
      hoNorm_ = Wf2::opNor(ho_,bio.c1(),bio.c2(),normSteps_,graph_);
      dt_=dtRel_*2.*cpminv(hoNorm_);
      cpmcerr<<"t_: "<<t_<<", hoNorm_: "<<hoNorm_<<endl;
    }
public:

 Qp2(Wf2 const& psi, F<Wf2,Wf2> const& ho,
   R const& dtRel, Z const& normSteps = 100, bool graph=false):
      psi_(psi), ho_(ho), dtRel_(dtRel), graph_(graph),
      normSteps_(normSteps)
   {
      ini_(); // initialize all the rest
   }
   void step_();
      // Step with any the hermitean operator ho_ as generator.
      // redetermines the operator-norm upon growth of the biotope.
      // Notice that with growth of the biotope regions of high potential
      // energy may come into scope which lets the operator norm of the
      // amiltonian grow.
   void adjustStep_();
      // encapsulates dt_ updating in step_()

   void setEps_(R const eps){ eps_=eps;}
   void setEta_(R const eta){ eta_=eta;}
   void setNormSteps_(R const steps){ normSteps_=steps; ini_();}
   Wf2 psi()const{ return psi_;}
   Wf2 phi()const{ return phi_;}
   Z size()const{ return psi_.size();}
   R norm()const{ return norm_;}
   R avr()const{ return avr_;}
   R t()const{ return t_;}
};

} // namespace CpmQuantumMechanics

#endif
