//? cpmsi.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_SI_H
#define CPM_SI_H
/*
   Description: Physical quantities carrying 'dimension'
*/

#include <cpmq.h>
   // for class Q of rational numbers
   // (coded as numerator/denominator)
#include <cpmvr.h>
   // for forming Vr<Q> := Q_Vector
#include <cpmtypes.h>

namespace CpmPhysics{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Q;
   using CpmRoot::Word;

enum TYPE { SI, NATCONST, UNSPECIFIED};

class Si{  // class for description of physical quantities
   // (including dimensions) especially with respect to the SI system.
   // Some exploratory methods to deal also with other systems of units,
   // e.g. Planck's units bases on natural constants c,h,G

   typedef Si Type;
   typedef R ScalarType;
public:
   // supportive type
    typedef CpmArrays::Vr<Q> Q_Vector;
private:
// independent data
   R val;   // dimensionless number
   Q_Vector qv;
      // x = val * m^d1 * kg^d2 * s^d3 * A^d4 * K^d5 * mol^d6 * cd^d7
      // for di=qv[i]
   TYPE type; // sets the meaning of the dimensional exponents

// dependent data
   bool hasDim;   // false if the quantity is 'dimension-less'
   void adjust();
      // sets the dependent data according to the independent ones
   Q& operator[](Z i){ return qv[i];} // after change one may have to
   // call
      // adjust, so this function should not be public

public:

// constructors
   Si(R valIn, Q_Vector dimensionsIn, TYPE t=SI);
      // dimensionsIn is cut to length nDim if too long and filled by
      // zero if too short

   explicit Si(R v=1): val(v),qv(nDim),type(SI),hasDim(0){}
      // default Si is the dimension-less 1

   Si( R in_val, Q in_m, Q in_kg,
      Q in_s, Q in_A=Q(), Q in_K=Q(), Q in_mol=Q(), Q in_cd=Q() );
      // constructor from numerical value and fractional dimensions for
      // each of the basic units

   Si( R in_val, Z in_m, Z in_kg,
      Z in_s, Z in_A=0, Z in_K=0, Z in_mol=0, Z in_cd=0 );
      // constructor from numerical value and integer dimensions for each
      // of the basic units
// interfaces (very similar to those of Q)
   CPM_IO
   CPM_ORDER
   CPM_TEST_X
   CPM_DESCRIPTORS
   CPM_CONJUGATION
   CPM_SUM_M
   CPM_PRODUCT_M
   CPM_PRODUCT_SCALAR_M
    // this does not incude Type + ScalarType
   CPM_DIFFERENCE
   CPM_DIVISION

//overloading pow() for type SI
   Si pow(Q y)const;
   Si pow(Z y)const;
   Si sqrt()const{ return pow(Q(Z(1),Z(2)));}

// access methods
   TYPE getType()const{ return type;}
   R operator()(void)const{ return val;}
      // numerical value
   const Q& operator[](Z i)const{ return qv[i];}
// For Si x=...;
// and x.getType()==SI we have
// x=x()*m^x[1]*kg^x[2]*s^x[3]*A^x[4]*K^x[5]*mol^x[6]*cd^x[7]
//
// for x.getType()==NATCONST we presently don't seriously consider
// quantities of dimension of temperature, amount of substance,
// and luminous intensity so that we did not fix natural constants
// describing them. We thus put:
// x=x()*c^x[1]*h^x[2]*G^x[3]*e^x[4]*K^x[5]*mol^x[6]*cd^x[7]

// express a quantiy in terms of other base quantities

   Si convert(const CpmArrays::V<Si>& newBase, TYPE newType)const;
      // given a array of SI quantities we return a quantity
      // which is formal SI but actually give value and dimension
      // product referring to the quantities from newBase. Tool for
      // implementing the next function

   Si fromSiToNat()const;
      // applying the convert function for converting SI to NATCONST
      // meaning

   static void PlanckUnits(const Word& fileName="natunits.txt");
      // nice demonstration of function Si fromSiToNat()const;

   void test(Z stop=0);
      // old test function, not particularly well designed

   static Z nDim; // always 7 for SI

   static Si m;
   static Si kg;
   static Si s;
   static Si A;
   static Si K;
   static Si mol;
   static Si cd;

   static Si c;
   static Si h;
   static Si h_; // h divided by 2*Pi
   static Si e;
   static Si G;
   static Si u;
   static Si eps0;
   static Si mu0;
};

// derived classes for simple introduction of physical quantities
// Example: terrestial acceleration g may be introduced as
// ACCELERATION g(9.81);
// This fixes the value in SI units as 9.81 and the dimension as that
// of an acceleration: m^1*s^-2

// *************************** L E N G T H  Class ***********************

class LENGTH : public Si{      // LENGTH() = m
public:
   explicit   LENGTH(R valIn=1):Si(valIn,1,0,0,0,0,0,0){}
};

// *************************** M A S S  Class ***************************
class MASS : public Si{    // MASS() = kg
public:
   explicit   MASS(R valIn=1):Si(valIn,0,1,0,0,0,0,0){}
};

// *************************** T I M E Class ***************************
class TIME : public Si{    // TIME() = s
public:
   explicit   TIME(R valIn=1):Si(valIn,0,0,1,0,0,0,0){}
};

// ************** (Electric) C U R R E N T  Class *********************

class CURRENT : public Si{ // CURRENT() = A
public:
   explicit   CURRENT(R valIn=1):Si(valIn,0,0,0,1,0,0,0){}
};

// ************* (Thermodynamic) T E M P E R A T U R E  Class   *********
class TEMPERATURE : public Si{ //  TEMPERATURE() = K
public:
   explicit   TEMPERATURE(R valIn=1):Si(valIn,0,0,0,0,1,0,0){}
};

// *************   (Amount of) S U B S T A N C E Class  *************
class SUBSTANCE : public Si{  // SUBSTANCE() = mol
public:
   explicit   SUBSTANCE(R valIn=1):Si(valIn,0,0,0,0,0,1,0){}
};

// *************** (Luminous) I N T E N S I T Y Class  ***************
class INTENSITY : public Si{  // INTENSITY = cd
public:
   explicit   INTENSITY(R valIn=1):Si(valIn,0,0,0,0,0,0,1){}
};

// *************** (Plane) A N G L E Class  ***************
class ANGLE : public Si{ // unit rad = radian = 1
public:
   explicit   ANGLE(R valIn=1):Si(valIn,0,0,0,0,0,0,0){}
};

// *************** S O L I D  A N G L E Class  ************************
class SOLID_ANGLE : public Si{ // unit sterrad = steradian = 1
public:
   explicit   SOLID_ANGLE(R valIn=1):Si(valIn,0,0,0,0,0,0,0){}
};

// ************************* V E L O C I T Y   Class ********************

class VELOCITY : public Si{         // m*s^-1
public:
   explicit   VELOCITY(R valIn=1):Si(valIn,1,0,-1,0,0,0,0){}
};

// ************************* A C C E L E R A T I O N   Class ************

class ACCELERATION : public Si{      // m*s^-2
public:
   explicit   ACCELERATION(R valIn=1):Si(valIn,1,0,-2,0,0,0,0){}
};

// ************************** A R E A  **********************************

class AREA : public Si{      // m^2
public:
   explicit   AREA(R valIn=1):Si(valIn,2,0,0,0,0,0,0){}
};

// ************************** V O L U M E *******************************

class VOLUME : public Si{      // m^3
public:
   explicit   VOLUME(R valIn=1):Si(valIn,3,0,0,0,0,0,0){}
};

// ************************* D E N S I T Y   Class **********************

class DENSITY : public Si{              // kg*m^-3
public:
   explicit   DENSITY(R valIn=1):Si(valIn,-3,1,0,0,0,0,0){}
};

// ************************* F O R C E   Class **************************

class FORCE : public Si{      // N == m * kg * s^-2
public:
   explicit   FORCE(R valIn=1):Si(valIn,1,1,-2,0,0,0,0){}
};

// ************************* E N E R G Y  Class **********************

class ENERGY : public Si{      // J = N*m == m^2 * kg * s^-2
public:
   explicit   ENERGY(R valIn=1):Si(valIn,2,1,-2,0,0,0,0){}
};


// ************************* C H A R G E   Class ************************

class CHARGE : public Si{   // electrical charge == Coulomb [C] = s*A
public:
   explicit   CHARGE(R valIn=1):Si(valIn,0,0,1,1,0,0,0){}
};

// ************************* R E S I S T A N C E   Class ****************

class RESISTANCE : public Si{ // electrical resistance Ohm = V/A
public:
   explicit   RESISTANCE(R valIn=1):Si(valIn,2,1,-3,-2,0,0,0){}
};

// ************************* V O L T A G E   Class **********************

class VOLTAGE : public Si{ // V
public:
   explicit   VOLTAGE(R valIn=1):Si(valIn,2,1,-3,-1,0,0,0){}
};

} // namespace

#endif
