//? cpmstate2.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STATE2_H_
#define CPM_STATE2_H_
/*
   Description: Development and test of a class that represent quantum
      mechanical pure states of 2-particle systems, where each particle
      lives on a 1-dimensional toroidal space.
*/

#include <cpmstate.h>
#include <cpmtrj.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;

namespace CpmQuantumMechanics{

   struct State2Data{ // data that define a State2 object
      R alpha{0.};
      R beta{0.};
      R mu{0.};
      R eps{0.};
      R p{0.};
      R p1{0.},p2{0.},b1{0.},b2{0.};
      R m1{0.};
      R m2{0.};
      R size1{0.};
      R size2{0.};
      Z box1{0};
      Z box2{0};
      B crossway;
         // If this is true, the discrete positions of the particles are
         // arranged perpendicular to each other (in this case one may
         // interprete the state also as referring to a single particle
         // in 2-dimensional space and the distance between particles
         // replaced by the distance from the origin.
         // If crossway==false, the discrete positions are arranged on
         // the same line, with the centers of the respective intervals
         // coinciding.
      B barrierForm;
         // new version with !crossway dealing with two particles in a
         // single 1-dimensional discrete biotope which can be separated
         // in two halfs by a growing potential barrier.
      B autoStep;
      R_Func ft;
         // Function which defines time dependence of of potentials;
         // active if wellForm==true.
      R_Func f,g;
      State2Data():m1(1.),m2(1.),size1(1.),size2(1.){}
         // notice R1's always initialized as 0
   };

   struct RunControl{
      R tTot{0.};
      Z nDoc{0};
      Z nImg{0};
      Z nView{0};
      R tWait{0.};
      Z ipolMethod{0};
      R gamma{0.};
   };

class State2{//fixed-dimensional Quantum mechanical 2-particle state space
   // Fixed-dimensional Quantum mechanical 2-particle state space with
   // Hilbert space structure, selectable boundary conditions,
   // and parameterized potential
protected:
   Z m_,n_;
      // numbers of lattice points
   R sm_,sn_; // lattice spacing in m_-direction and n_-direction
      // respectively
 //  R lpm_,lpn_;
      // lpm_: lowest position in m_-direction
      // lpn_: lowest position in n_-direction
      // see function ini() for the exact definition
      // used for a fast implementation of functions x and y.
   V< V<C> > rep_;
      // Standard double indexed array with indexes starting
      // with 1 the range of the first index will be mapped to the
      // x-range where particle 1 lives, and the range of the second index
      // will be mapped to the y-range, where the second particle lives
      // (See UM 2006-05-16 p. 1 of my yellow concept booklet).
      // It is the function void mark(Graph& f, Z ipolMeth=1)const;
      // which does this mapping.
   Z box1_,box2_;
      // 0: values of rep_ for out-of range indexes give as default
      // object
      // 1: values given by cyclic continuation
      // else  : values given by constant continuation

   Iv iv1_,iv2_; // biotopes in 1-direction and in 2-direction
   V<R> x_,y_; // lattice positions in 1-direction and in 2-direction

   static State2Data d_;
      // Thus all instances of State2 share the data collected in d_

   void ini_()
   {
      R ds1=R(d_.size1),ds2=R(d_.size2);
      sm_=ds1/R(m_); sn_=ds2/R(n_);
   //   lpm_=(sm_-ds1)*0.5; lpn_=(sn_-ds2)*0.5;
      // thus lpm_=-d_.size1/2 + 0.5*sm_, i.e. one half-space above the
      // lower end of the biotope [-d_.size1/2,d_.size1/2]
      box1_=d_.box1;
      box2_=d_.box2;
      iv1_=Iv(-ds1*0.5,ds1*0.5);
      iv2_=Iv(-ds2*0.5,ds2*0.5);
      x_=iv1_.cen(m_);
      y_=iv1_.cen(n_);
   }

   State2(V<C> const& psi1, V<C> const& psi2);
      // creating a tensor-product of two one-particle states
   State2(Z d1, Z d2):m_(d1),n_(d2),rep_(m_,V<C>(n_,C())){ini_();}
public:
   State2(V< V<C> > const& psi):rep_(psi){ini_();}
   typedef State2 Type;
   CPM_IO
   CPM_NAM(State2)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)

   State2():m_(0),n_(0),rep_(m_,V<C>(n_)){ini_();}
      // lean constructor which does no allocation for components

   State2(State const& psi1, State const& psi2);

   State red1(State const& psi)const;
      // reduce first index via psi

   State red2(State const& psi)const;
      // reduce second index via psi

   V< V<C> > getRep()const{ return rep_;}

   C operator()(Z i, Z j)const{ return rep_.cui(i).cui(j);}
      // efficient but unsafe reading

   void put_(Z i, Z j, C v){ rep_.cui(i).cui(j)=v;}
   // efficient but unsafe writing

   Z dim1()const{ return m_;}
   Z dim2()const{ return n_;}
   Z d1()const{ return m_;}
   Z d2()const{ return n_;}
   Iv iv1()const{ return iv1_;}
   Iv iv2()const{ return iv2_;}

   State2 dot(R t=0.)const;
      // hamilton operator divided by ihDirac. Applied to a state
      // it gives the time derivative of this state. This is what
      // the name tries to suggest.

   State2 V12()const;
      // returns the state resulting from applying the
      // interaction energy operator defined by v12
      // This makes it easy to compute expectation values
      // of this interaction operator, and this helpd to
      // set the coupling constant d_.alpha meaningfully.

   R dis(R xi, R yj)const
      // particle distance if particle 1 is at xi and particle 2 is at yj
      // Depends on wether the biotops of the particles are orthogonal
      // to each other (d_.crossway==true) or lie on the same line
      // In both cases the centers of the biotopes are assumed to
      // coincide.
   {
      return d_.crossway ? R2(xi,yj).abs() : cpmabs(xi-yj);
   }

   R v12(R r)const
   {
      return cpmexp(-r*R(d_.mu)*cpmpow(r+R(d_.eps),R(-1)-R(d_.p)));
   }

   R pot12(R r)const
   {
      return R(d_.alpha)==0 ? R(0.): v12(r)*R(d_.alpha);
   }

// if d_.barrierForm==true the only potentials that enter the definition
// of dot are pot1 and pot2
   R pot1(R x, R t=0.)const
   {
      if (d_.barrierForm){
         return d_.f(x)+d_.ft(t)*d_.g(x);
      }
      else{
         R x1=iv1().x(0.25);
         return R(d_.b1)==0 ? R(0) :cpmpow(cpmabs(x-x1),R(d_.p1))*R(d_.b1);
      }
   }

   R pot2(R y, R t=0.)const
   {
      if (d_.barrierForm){
         return d_.f(y)+d_.ft(t)*d_.g(y);
      }
      else{
         R y2=iv2().x(0.75);
         return R(d_.b2)==0 ? R(0):cpmpow(cpmabs(y-y2),R(d_.p2))*R(d_.b2);
      }
   }

   R V1(R r)const
   {
      return d_.alpha==0 ? R(0) : v12(r)*d_.alpha;
   }

   R V2(R r2)const{ return d_.beta==0 ? R(0) : r2*d_.beta;}
      // for efficiency this function gets input of the form y^2

   void mark(Graph& f, Z ipolMeth=0, R gamma=0.5, R secFac=1.1)const;
      // showing the wave function in a way that the x-axis refers to
      // particle 1 and the y-axis ( oriented vertically upward!)
      // refers to particle 2.
      // If the last argument is >0 it triggers an automatic
      // bright-level adjustment only in the first call and then
      // holds this level so that changes in brightness of individual
      // pixels can't occur as the result of a new level adjustment.

// basic linear structure in most efficient form
   State2& operator *=(C z);
   State2& operator *=(R r);
   State2& operator +=(State2 const& s);
   State2& operator -=(State2 const& s);

   State2 operator *(C z)const
      { State2 res(*this); return res*=z;}
   State2 operator *(R r)const
      { State2 res(*this); return res*=r;}
   State2 operator +(State2 const& s )const
      { State2 res(*this); return res+=s;}
   State2 operator -(State2 const& s )const
      { State2 res(*this); return res-=s;}

   R nor_(R r=1);
      // turns *this into a vector of length r and returns the norm of
      // the original in order not to loose information. If *this was
      // the zero vector, it will not be changed
   R norm();
      // returns the length of *this
   R sup()const;
      // max of rep_[i][j].abs()
   void smooth_();
   // applies a next neighbor local averaging operation
   // to rep_. Does not conserve the norm although the weights
   // are optimum for doing so.
   void concentrate_();
   // multiplies each component of rep_ with the square
   // of its absolute value and renormalizes to unit norm.
   // This amplifies these parts of the wave function where
   // the values are close to the maximum value.
   // No change of the phase.
   void con_();
      // complex conjugation, makes motion reversal
   void backGrn_(R level=1e-6);
      // background
      // changes the state in a new one for which to each component
      // the real level gets added. Finally normalizing
// Accessing static member d_
   static void set_d(State2Data const& d){ d_=d;}
   static const State2Data& get_d(){ return d_;}
      // reading access to internal data
   static void setAlpha(R alpha){ d_.alpha=alpha;}
   static void setBeta(R beta){ d_.beta=beta;}
   static R alpha(){ return d_.alpha;}
   static R beta(){ return d_.beta;}
   R ds1()const{ return sm_;}
      // discretization length for particle 1
   R ds2()const{ return sn_;}
      // discretization length for particle 2
   R m1()const{ return d_.m1;}
      // mass of particle 1
   R m2()const{ return d_.m2;}
      // mass of particle 2

   R dtSup(Z p=100, R tWait=0, R gamma=0.1, R t=0.)const;
      // upper limit for the time step according to
      // the new insights from the preparation of the
      // publication. Tries to estimate the largest
      // energy eigenvalue (better: absolute value of this)
      // Seems to work now after many frustrations. (2006-10-26)
      // 2014-10-05: The limit dtSup satisfies dtSup * ||H|| = 2 and thus
      // is the stability limit of the direct midpoint method
      // and the DALF method. Notice that a stable time step dt for the
      // ALF method requires dt * ||H|| <= 1.
      // Since the Hamiltonian ( or equivalently dot() is now allowed to
      // depend on time, the time is introduced as the last argument.
      // This is not optimal since it thus is placed behind less essential
      // technical parameters.

   State2 ran()const;
      //: random
      // returns a normalized random vector. Notice: the usual Z-argument
      // is not implemented here (would not be easily implemented in a
      // reasonable way and will hardly be needed).
};

class State2Trj{ // State2 trajectory
protected:
   R t_;
      // time
   R dt_;
      // time step
   R dtL_;
      // natural time step
   State2 x_;
      // Schroedinger wave function at time t_
   State2 v_;
      // time derivative of the Schroedinger wave function at time t_
   R jAmp1_{0.}, jAmp2_{0.};
      // jitter amplitudes (active if >0) for jAmpi_==1 the phase of the
      // wavefunction becomes randomized completely.
      // Initialized as 0!
public:
   bool enfNor_;
      // enforce normalization
   bool autoStep_;
   R normTol_;
      // if <=0 no action. Otherwise norm conservation within
      // this tolerance will be enforced
   Z stepMet_;
      // step method (3 for the one based on acceleration)
      // 2, and 1 for simplifications of increasing level
      // of naivity
   Z revMet_;
      // reversal method (0 bad method based on re-initialization
      // 1 exact method

   void setPhaseJitter_(R a1, R a2){jAmp1_=a1; jAmp2_=a2;}

   State2Trj(State2 const& psi, R dtRel=0.75, bool rel=true, R ti=0.);
   virtual void step_();
   void run_(RunControl const& rc, Graph& gr);
      // creates images on screen and a selectable series of ppm files
      // in the output directory of the calling application
   C enr()const{ return (x_|v_)*QMBase::get_ihDirac();}
      // energy expectation value
   C xv()const{ return (x_|v_);}
      // similar to energy expectation value
   R xx()const{ return x_.absSqr();}
   State2 x()const{ return x_;}
   State2 v()const{ return v_;}
  // State2 a()const{ return a_;}
   R t()const{ return t_;}
   void set_t(R t){ t_=t;}
   R dt()const{ return dt_;}
   void set_dt_(R dt){ dt_=dt;}
   R dtLimit(bool light=false)const;
      // working values for dt in step_(dt) are equal or smaller
      // as this value. dtLimit=2*ds*ds*m/hPlanck for mass m,
      // discretization length ds. The reason:
      // ds/dtLimit = QMBase::vMax(n).

   void smooth_(Z n=1){
      for (Z i=0;i<n;i++) x_.smooth_();
      if (n>0){
         x_.nor_();
         v_=x_.dot();
      }
   }

   void concentrate_(){
      x_.concentrate_();
      v_=x_.dot();
   }

   void con_(){
      x_.con_();
      if (revMet_==1){ // exact method
        v_.con_();
        v_*=C(-1);
      }
      else  v_=x_.dot();
         // approximate method, only rough motion reversal
   }

   void phaseJitter_(R relAmp1=1e-3,R relAmp2=1e-3);
      // multiplies each component by exp(i*phi) with -phi0<=phi<=phi0
      // and phi0=2*pi*relAmp1; not yet final
};

class State2Trj2: public State2Trj{ // State2 trajectory 2
   TrjDMI<State> x1_,x2_;
public:
   State2Trj2(State const& x1, State const& x2);
   virtual void step_();
   void step1_(){ x1_.step_(dt_);}
   void step2_(){ x2_.step_(dt_);}
   State x1()const{ return x1_.x();}
   State x2()const{ return x2_.x();}
  // X2<State, State> dot()const;
   void mark(Graph& f, Z ipolMeth=0, R gamma=0.5, R secFac=1.1)const;
};

} // namespace

#endif
