//? cpmstatedirac1.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STATED1_H_
#define CPM_STATED1_H_
/*
   Description: Development and test of a class StateD1
      which represents quantum mechanical pure electron states on a
      one-dimensional equidistant lattice. Contray to the old class
      StadeD in the present class the electron has a 2-dimensional
      spinor space and not a 4-dimensional one. This is in line with my
      Wolfram Demonstrations concerning electron motion in 1D.
*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmtypes.h>
#include <cpminterfaces.h>
#include <cpmrfunc.h>
#include <cpmconstphys.h>
#include <cpmgraph.h>
#include <cpmobservable.h> // contains cpmstate.h
#include <cpmclinalg.h>
#include <cpmfile.h>
#include <cpmbiotope.h> // defines class Bio

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRootX::B;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGeo::Iv;
using CpmLinAlg::Z2;
using CpmLinAlg::R2;
using CpmLinAlg::C2;
using CpmLinAlg::CC2;
using CpmSystem::OFileStream;
//using CpmSystem::OFile;

using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;

namespace CpmQuantumMechanics{

//class Observable;
class OpD1;
class SpectralObjectD1;

typedef X2<C2,R> SX1; //. spinor x
   // 2 spinor components together with a 1D position
   // Notice that for 1D psition space 2 spinor components is
   // sufficient due to decoupling. My Wolfram demonstrations make use of
   // this. For dynamics based on ALF there is no important difference
   // but spectral representation becomes much faster with replacing
   // 2n x 2n matrices by n x n matrices
   // Comparison of execution timing for application CpmDirac::App16 when
   // realized with StateD (A) and StateD1 (B) for various values of n:

   // n :   8        16       32       64       128         256
   // t_A   0.32     2.69     21.28    219.7
   // t_B   0.185    0.422    2.625    20.663   190.463     1874.75
   // The spectral representation (this is what App16 analyses) is done
   // ~ 10 times faster if the 4-spinors are replaced by 2-spinors.
   // Mathematica was by a similar but a bit larger problem for n=256
   // by a factor 139 faster than (B) (on a bit faster computer, though).

////////////////// class StateD1 /////////////////////////////////////////
// D for Dirac

class StateD1 { // spinor-valued wave functions on a finite 1-D lattice
   // Finite-dimensional Quantum mechanical state space with Hilbert
   // space structure and 'periodic boundary conditions'. Notice however,
   // that a fixed value of the dimension of the Hilbert space is not part
   // of the present concept.

   typedef StateD1 Type;

   B fft_; // true if the values of rep_[i].c1() are spatial
      // frequencies and false if these are spatial positions
      // Notice: auto-initialized as false.
   Iv s_;
      //. size
      // a range in 1-dimensional x-space ('the biotope')

   Iv sup_;
      // support (i.e. the maximum domain outside of which the wave
      // function vanishes)

   R xc_{0.};
      // Center point of the biotope in position (not frequency) space.
      // Notice: auto-initialized as 0.

   V<SX1> rep_;
      //: representation

   StateD1(V<SX1> const& v, Iv const& s):s_(s),rep_(v){}
      // private constructor. To be used with care since
      // matching of objects not tested

   V<C> val1()const;
   V<C> val2()const;
      // val, separated in its 2 components

   void setVal_(V<C2> const& v)
   {
      Z d=rep_.dim();
      cpmassert(v.dim()==d,"StateD1::setVal_");
      for (Z i=1;i<=d;++i) rep_[i].c1()=v[i];
   }

   static Z bc_;
      //. boundary conditions
      // 1 periodic, 2 reflecting, 3 constant

   static Z verbose_;
      //.. verbose.
      // So far controlling only if the computation of a operatornorm
      // should be accompanied by a graphical representation.

   static bool useDALF_;
      //: use DALF
      // method switch for the implementation of boosts

   static bool enfNor_;
      //: enforce normalization
      // correction for the small unitarity deficit of the boost
      // implementations

public:

   CPM_IO
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
      // Hilbert space with anti-unitary operator con()
   CPM_NAM(StateD1)

   StateD1(){}
      // default constructor

   StateD1(V<C2> const& v, Iv const& s);
      // constructor from spinor values and spatial interval

   explicit StateD1(Bio const& b);
      // Constructing the zero-vector. All components are zero spinors.

   StateD1(C const& z, Bio const& b);
      // Constructing the wave function for which all spinor components
      // have the value z.

   StateD1(Iv const& iv, Bio const& b, Z i=1);
      // Constructing the wave function on b which differs from the zero-
      // vector only within iv, where the i-th component has the value
      // C(1).

   StateD1(F<R,C2> const& f, Bio const& b);
      // Here the values of the wavefunction are given by a function f.

   StateD1(Z i, Z a, Bio const& b, bool frqSpc=false);
      // A state for which the wavefunction is different from zero only
      // at place i. This single value has 2 spinor components and from
      // those only that indexed a is different from zero. It has value 1.
      // The states constructed this way form an orthonormal basis.
      // If i is outside the valid index range of rep_ (as set by b) then
      // i is understood as its cyclic image within the valid range.
      // If the state is to be defined in frequency space, the last
      // argument has to be set to 'true'.

   StateD1(Bio const& b, C2 const& c, R const& x0, R const& sigma0,
          R const& k, bool normalize=true);
      // Constructs the wave function
      // psi(x) = c*exp(-0.5*((x-x0)/sigma0)^2)*exp(i*k*x).
      // The values of x0, sigma0, and k refer to the same units as do the
      // values of b. So, they are not relative to, say, the length of the
      // biotope b. If the last argument is true, the constructed object
      // is a unit vector.

   StateD1(Bio const& b, R const& xPeak, R const& sigma , R const& p,
   R const& a1, R const& a2, R const& phase1, R const& phase2,
   bool compact);

   static CC2 unit_;
   static CC2 alpha1_;
   static CC2 beta_;

   CC2 boost1(R const& chi)const;
   // tanh(chi)=v/c

   static R2 lorentz1(R2 const& y, R const& chi);
      // y=(x0,x1)=(t,x)

   static void setBonCon(Z bc){ bc_=bc;}
      //: set boundary conditions

   static Z getBonCon(){ return bc_;}
      //: get boundary conditions

   static void setVer(Z ver){verbose_=ver;}
      //: set verbosity

   static Z getVer(){return verbose_;}
      //: get verbosity

   static void setUseDALF(bool b){useDALF_=b;}
      //: set useDALF

   static bool getUseDALF(){return useDALF_;}
      //: get useDALF

    static void setEnfNor(bool b){enfNor_=b;}
      //: set enforce normalization

   static bool getEnfNor(){ return enfNor_;}
      //: get enforce normalization

   static StateD1 nyq(C2 const& c1, C2 const& c2, Bio const& b);
      //: Nyquist
      // Returns a state res such that res.dim()==d, res.s()==s,
      // res.val()[1]==res.val()[3]==res.val()[5]==...==c1
      // res.val()[2]==res.val()[4]==res.val()[6]==...==c2
      // Since we have alternating values, the dominating frequency is
      // Nyquist frequency.

   R_Vector pos()const;
      //: positions
      // returns the positions or rep_ with the values removed

   V<C2> val()const;
      //: values
      // returns the values or rep_ with the positions removed

   R valSup2()const;
      // maximum of val(i).abs2()

   R valInf2()const;
      // minimum of val(i).abs2()

   R_Matrix valMat(bool logScale=false)const;
      //: values (as)  matrix
      // the 2 components in real representation by dividing into real and
      // imaginary part. Together with absolute value as the first row of
      // the resulting matrix.
      // Actually all these enumerated real quantities are not given by
      // their value but by the value of their squares. These are
      // probabilities and have a more direct meaning than the amplitudes.

   C_Vector lin()const{ return C_Vector(val1()) & C_Vector(val2());}
      //: linear
      // returning the spinor components as a linear array

   void setVal_(C_Vector const& v);
   //: set values
   // setting the spinor comonents from an linear array of complex
   // numbers. In this linear order all first components come earlier
   // than any second component.

   void setSup_(Iv const& iv){ sup_=iv;}
      //: set support

   StateD1 toFrqSpc()const;
      //: to frequency space
      // Unitary transformation, implemented by FFT

   StateD1 fromFrqSpc()const;
      //: from frequency Space
      // Inverse of toFrqSpc, implemented by inverse FFT

   bool mat(Type const& psi)const
   //: match
   { return s_==psi.s_ && rep_.sameDom(psi.rep_);}
   // returns true if *this and psi belong to the same state space

   StateD1 applyCC2(CC2 const&)const;
   //:: apply CC4
   // multiplies all spinor values with the same CC4-matrix

   StateD1 rev()const;
   //: reversed
   // returns the space-inverted version of *this

   StateD1 cc()const;
   //. complex conjugate
   // returns the version of *this with all spinor components changed into
   // their complex conjugates

   StateD1 tr()const{ return cc().applyCC2(beta_);}
   //. time reversal

   StateD1 timeRev()const{ return cc().applyCC2(beta_);}
      //: time reversal

   StateD1 apply(R_Func const& f)const;
   //:: apply
   // multiplies all spinor values with the values of a real function on
   // s_

   StateD1 win(Iv const& iv, R const& softParameter=0)const;
   //: window
   // Restricts the wave function to the interval iv in a way that depends
   // on softParameter. s=0 makes hard edges, s=0.01 steep edges and flat
   // plateau, s=1 no plateau at all. s<0 and >1 not recommended.

   StateD1 frqWin(Iv const& iv, R const& softParameter=0)const;
   //: frequency window
   // Restricts the wave function, which is assumed to be given in
   // position space, by a windowing operation which acts on the FFT
   // image. The argument iv which belongs to the full frequency range
   // is Iv(-1,1) in order to be independent of the discretization under
   // consideration.

   StateD1 applyF(R_Func const& f)const;
   //:: apply to discrete Fourier transform.
   // The argument of f ranges from -1 to 1, where 1 corresponds to the
   // spatial Nyquist frequency 1/(2*gap()) and -1 to -1/(2*gap().
   // It is to be noted however that this last most negative value no
   // longer belongs to the indexing range of the discrete Fourier
   // transform.

   StateD1 applyPot(R_Vector const& pot)const;
   //:: apply
   // Multiplies all spinor values with the values of the corresponding
   // values of -i*pot.

   StateD1 x1()const{ return apply(R_Func::x1);}
   //. x1
   // The return value is the StateD resulting from multiplying each
   // amplitude of *this by its position coordinate.

   StateD1 dx()const;
   //. d/dx
   // The return value is the StateD resulting from differentiation of
   // *this with respect to position. Depends on the static variable bc_
   // which sets the boundary conditions.

   StateD1 mom()const
   //: momentum
   // momentum operator
   { return dx()*C(0.,-1.);}

   StateD1 dot(R const& m, R_Vector const& pot=R_Vector())const;
      //: dot
      // returns the time derivative of *this under Dirac evolution where
      // m is the mass of the Dirac particle and pot an electric potential.

   inline StateD1 dot(OpD1 const& op)const;
      //: dot

   StateD1 ham(R const& m, R_Vector const& pot=R_Vector())const
      //: Hamiltonian
      // psi.ham(m,pot) is the state which from state psi results by
      // applying the Dirac Hamiltonian of mass m and potential pot to it.
      // m should be in the range given by m=mRed/gap(), where 0<=mRed<=1
   { return dot(m,pot)*C::I;}

   static OpD1 h(Bio const& b, R const& m, R_Vector const& pot=R_Vector());
      //. Hamilton operator

   static OpD1 p(Bio const& b);
      //. momentum operator

   static OpD1 x(Bio const& b);
      //. position operator

   static OpD1 shift(Bio const& b, R const& a);
      // Approximately unitary representation operator for finite
      // shift a.

   static OpD1 boostL(Bio const& b, R const& m, R const& chi,
                     R const& dtRel=0.1);
      // tanh(chi) = v/c
      // boost of a free Dirac particle. Thus no potential as argument.
      // boostL means "local" in the sense of being not too far away from
      // the unit element of the Lorentz group. The argument dtRel
      // controls the evaluation of the time translations which are
      // involved. It should always satisfy 0<dtRel<1. dtRel=1 is the
      // stability limit, where exponential grow of the state may happen.

   static OpD1 boost(Bio const& b, R const& m, R const& chi,
                    R const& deltaChi=0.0125, R const& dtRel=0.05);
      // tanh(chi) = v/c
      // boost of a free Dirac particle. Thus no potential as argument.
      // chi may have any value, but large values translate to slow
      // execution.
      // There are two implementations which depend on the value of the
      // static boolean variable useDALF_.
      // useDALF==false:
      // The finite boost is represented as a product of some integer
      // number n of identical boostL(b,m,chiSmall,dtRel), where chiSmall
      // is as close to deltaChi as possible. Actually,
      // n*chiSmall=chi, n=|round(chi/deltaChi)|.
      // useDALF==true:
      // Here the finite boost is represented as an exponential of an
      // infinitesimal boost, and the exponential is implemented by solving
      // the corresponding differential equation by means of the DALF
      // integrator. In this case deltaChi controls the definition of
      // the infinitesimal boost as a difference quotient. And
      // dtRel controls the step size on two the exponentials: the
      // exponential from 0 to chi and the exponential involved in the
      // definition of the infinitesimal boost.
      // The default values for deltaChi and dtRel are good compromises
      // between speed and accuracy in both cases.

   static OpD1 dBoost(Bio const& b, R const& m, R const& deltaChi=0.0125,
                   R const& dtRel=0.05);
      // infinitesimal boost
      // Formed as a differece quotient of small finite boosts. Then
      // arbitrary finite boosts can be formed by exponentiation which is
      // implemented by the DALF integrator. This method is unexpectedly
      // slow and changes the norm more than expected. Therefore I added
      // a renormalisation step to the implementation which can be
      // enabled/disabled by means of the static function setEnfNor(bool).

   static SpectralObjectD1 dBoostExact(Bio const& b, R const& m);
      // exact form of infinitesimal boost

   static SpectralObjectD1 boostExact(Bio const& b, R const& m,
      R const& chi);
      // exact form of finite boost

   C enr(R const& m, R_Vector const& pot=R_Vector())const;
      //: energy

   Z d()const{return rep_.dim();}
      // returns the cardinality of the biotope

   Z dim()const{return rep_.dim();}
      // returns the cardinality of the biotope (for homogeneousity with
      // State)

   Iv s()const{ return s_;}
      //. space

   Bio bio()const{ return Bio(rep_.dim(),s_);}
      //: biotope

   StateD1& nul_()
      //: nullify, non-constant function
      // sets all the spinor components of *this to zero
   {
      C2 c0;
      for (Z i=rep_.b();i<=rep_.e();++i) rep_[i].c1()=c0;
      return *this;
   }

   Z b()const{ return rep_.b();}
      //. begin (of valid indexes)

   Z e()const{ return rep_.e();}
      //. end (of valid indexes)

   R xc(Z i)const{ return rep_[i].c2();}
      //. x coordinate
      // read access to the position coordinate by indexes

   C2 const& operator[](Z i)const{ return rep_(i,Outside(bc_)).c1();}
      // read access to the spinor components by indexes, valid for all i
      // where the meaning depends on the boundary conditions.

   C2& operator[](Z i){ return rep_[i].c1();}
      // write access to the spinor components by indexes

   C2 at(R const& x)const;
      // getting the spinor value by interpolation for a position which
      // may not shine up in the list pos()

   C& operator()(Z i)
   { Z j=i-1, k=1+j/2, l=1+j%2; return rep_[k].c1()[l];}
      // write access to the complex components by indexes

   C const& operator()(Z i)const
   { Z j=i-1, k=1+j/2, l=1+j%2; return rep_[k].c1()[l];}
      // read access to the complex components by indexes

   R gap()const{ return s_.abs()/rep_.dim();}
      //: gap
      // distance between neighbor-points. 1/gap() is a mass.
      // It is the largest mass which can be used in the Dirac Hamiltonian.

   R lambdaMin()const{ return gap()*R(2);}

   R pMax()const{ return R(rep_.dim())/s_.abs();}
      // pMax := 1/gap and -pMax are the largest and the smallest
      // eigenvalue
      // of the momentum operator mom(). It can best be seen from the
      // unitary transform momFrq() which is a multiplication operator.
      // Notice that this differs from the self-suggesting ansatz
      // pMax = h/lambdaMin (since hDirac=1, we have hPlanck=2*pi)
      // = 2*pi/(2*gap) = pi/gap.

   R mMax()const{ return pMax();}
      // mMax = pMax = 1./gap
      // introduced for ease of use

   R normHa(R const& m)const{ R p=pMax(); return cpmsqrt(m*m+p*p);}
      // Exact value of normH(m).

   R nor_(R r=1);
      //: normalize
      // returns the norm of *this and changes *this into
      // a StateD of norm r by multiplying with a real factor.
      // If *this was the zero StateD, it will be turned into
      // a 'standard StateD' of norm r

   StateD1 ran(Z j=0)const;
      //: random

   R dis(Type const& psi)const;
      //: distance

   void mark(Graph& g, bool logScale=false)const;
      //: mark
      // auto-scaled polygon representation of the four complex valued
      // as real and complex part together with the absolute value of the
      // wave function

   void mark2(Graph& g, StateD1 const& sc, R const& scale=1)const;
      // Shows *this together with a comparison state
      // allows to boost the difference etween the two by a factor.

   static void markArr(V<StateD1> const& vs, Graph& g);
      // mark an array of states, where the autoscale function
      // takes all array components into account

   void gp(ostream& out, bool logScale=false)const;
      // writes data to out from which gnuplot can construct the
      // showAsCurves representation.

   void smt_(Z n=1)
      //: smooth
      // Changes *this into a smoothed, normalized version.
      // The argument gives the number of times the basic
      // smoothing operation (kernel 1/4 1/2 1/4) is performed.
   {
      V<C2> v2=val();
      for (Z i=1;i<=n;++i) v2.fht_<R>(0.25,0.5,0.25,CYCLIC);
      setVal_(v2);
      nor_();
   }

   bool isVal()const;
      //: is valid
      // is true if none of the spinor components has exceptional values
      // such as infinity

   StateD1 momFrq()const;
      //: momentum (in terms of) frequency

   StateD1 hamFrq(R mass)const;
      //: hamiltonian in frequency space
 };

//////////////////// class OpD1 //////////////////////////////////////////

class OpD1{ // operators in Dirac space as functions
   typedef OpD1 Type;
   Bio b_;
   Fa<StateD1,StateD1> f_;
public:
   CPM_NAM(OpD1)
   OpD1(){}
   OpD1(Bio const& b, F<StateD1,StateD1> const& f):b_(b),f_(f){}
   explicit OpD1(Observable const& obs, Bio const& bio);
   Observable toObs()const;
      //: to observable
   bool mat(OpD1 const& op)const{ return b_==op.b_;}
      //: match
   bool mat(StateD1 const& psi)const{ return b_==psi.bio();}
      //: match
   F<StateD1,StateD1> f()const{ return f_;}

   Bio bio()const{ return b_;}
      // biotope

   static OpD1 ham(Bio const& b, R const& m, R_Vector const& pot)
   { return StateD1::h(b,m,pot);}

   OpD1 operator+(OpD1 const& op)const
   { cpmassert(mat(op),"OpD1 +"); return OpD1(b_,f_+op.f_);}

   OpD1 operator-(OpD1 const& op)const
   { cpmassert(mat(op),"OpD1 -"); return OpD1(b_,f_-op.f_);}

   OpD1 operator-()const
   { return OpD1(b_,-f_);}

   OpD1 operator*(C const& z)const
   { return OpD1(b_,multR<StateD1,StateD1,C>(z,f_));}

   OpD1 operator*(R const& r)const
   { return OpD1(b_,multR<StateD1,StateD1,R>(r,f_));}

   OpD1 operator*(OpD1 const& op)const
   { cpmassert(mat(op),"OpD1 *"); return OpD1(b_,f_.circ(op.f_));}

   OpD1& operator*=(OpD1 const& op)
   { cpmassert(mat(op),"OpD1 *="); return *this=OpD1(b_,f_.circ(op.f_));}

   StateD1 operator*(StateD1 const& psi)const
   { cpmassert(b_==psi.bio(),"OpD1 * StadeD1");return f_(psi);}

   StateD1 operator()(StateD1 const& psi)const
   { cpmassert(b_==psi.bio(),"OpD1(StadeD1)");return f_(psi);}

   R norm()const; // operator norm
   R dtLim()const{return R(2)/norm();}
};

inline StateD1 StateD1::dot(OpD1 const& op)const{ return op(*this)*C::Ibar\
;}
// d/dt psi = H psi/i => i d/dt psi = H psi => psi(t) = exp(-i t H) psi(0)

///////////////////// class SpectralObjectD ///////////////////////////////

class SpectralObjectD1{//eigenvectors/values in combination
      // All eigen vectors match StateD1(b_)
      Bio b_;
      Z l_; // = b_.d()*2 dimension of the linear space formed by all
         // states that match StateD1(b_)
      V<C> spectrum; // such that spectrum.dim() == l_
      V<StateD1> eigenVectors;
         // set of l_ orthonormal states of type StateD which all match
         // StateD(b_)
      SpectralObjectD1(V<C> const& eval, V<StateD1> const& evec):
         b_(evec[1].bio()),l_(b_.d()*2),spectrum(eval),
         eigenVectors(evec){}
            // private since not safe
   public:
      typedef SpectralObjectD1 Type;
      CPM_IO
      CPM_NAM(SpectralObjectD1)
      friend class Observable;
      Bio bio()const{ return b_;}
      Z d()const{ return b_.d();}
      Z l()const{ return l_;}
      V<C> c1()const{ return spectrum;}
      V<StateD1> c2()const{ return eigenVectors;}

      const C& operator()(Z i)const{ return spectrum[i];}
         // eigenvalues by (), notice that i has to be in
         // {1,...,l_}.
      const StateD1& operator[](Z i)const{ return eigenVectors[i];}
         // eigenvectors by [], notice that i has to be in
         // {1,...,l_}, the order is such that the ground-state of
         // the harmonic oscillator is obtained for i=l_.

      SpectralObjectD1():b_(),l_(0){}

      SpectralObjectD1(Bio const& b);
         // has the 'canonical vectors' as eigenvectors
         // and the canonical count index 1,2,....4*b.d() for spectrum
         // and eigenvectors.

      SpectralObjectD1(OpD1 const& op, bool test=false);
         // spectral representation of the 'operator' op. Only if op
         // is a linear function and gets applied to states which
         // match with op the result makes sense. The observable to
         // which op belongs needs to be Hermitean.

      SpectralObjectD1(Observable const& obs , Bio const& bio,
         bool test=false);
         // spectral representation of the 'operator' op. Only if op
         // is a linear function and gets applied to states which
         // match with op the result makes sense. The observable to
         // which op belongs needs to be Hermitean.

      bool mat(StateD1 const& a)const{ return a.mat(StateD1(b_));}
         //: match
         // returns true if a has the same biotope as the eigenVectors

      R_Vector getSpectrum()const
      {
         R_Vector res(l_);
         for (Z i=1;i<=l_;i++) res[i]=spectrum[i].real();
         return res;
      }

      R supEigenVectors()const
      {
         R_Matrix rs(l_,l_);
         for (Z i=1;i<=l_;i++)
            for (Z j=1;j<=l_;j++)
               rs[i][j]=eigenVectors[i](j).abs();
         return rs.sup();
      }

      R spectralRadius()const
      {
         R_Vector rs(l_);
         for (Z i=1;i<=l_;i++) rs[i]=spectrum[i].abs();
         return rs.sup();
      }

      void setSpectrum_(C z){ for (Z i=1;i<=l_;i++) spectrum[i]=z;}

      void mark(Graph& gr, Z ipolMethod=0, R gamma=0.5)const;
         // modern implementation, writes on image memory.
         // Visible only upon gr.vis().

      void markSpectrum(Graph& gr)const;
         // real part of eigenvalues, as autoscaled graph

      bool writeSpectrum(ostream& str)const;
         // writes the eigenvalues to str, returns true
         // if writing was successful

      void map_(const F<C,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=l_;i++) spectrum[i]=f(spectrum[i]);}

      void map_(const F<R,C>& f)
         // f acts on the spectrum
      {for (Z i=1;i<=l_;i++) spectrum[i]=f(spectrum[i].real());}

      void expi_(R t);
         // changes *this into exp(i*t*(*this))

      void map_(R_Func const& f)
         // real function f acts on the real part of the spectrum
      {
         for (Z i=1;i<=l_;i++){
            C ci=spectrum[i];
            R ri=ci[1];
            R ii=ci[2];
            spectrum[i]=C(f(ri),ii);
         }
      }

      void nor_()
         { for (Z i=1;i<=l_;i++) eigenVectors[i].nor_();}
      // normalizes all eigenvectors

      StateD1 operator *(StateD1 const& state)const;
      // SpectralObjectD acts on state just as the underlying observable.

      V<C> rep(StateD1 const& state)const;
         //: representation
         // Returns the list of development coefficients of state
         // in the basis given by the list eigenVectors.

      SpectralObjectD1 order()const;
         // returns a new SpectralObject which results form *this
         // by common re-ordering of eigenvalues and eigenvectors
         // such that the real parts of the eigenvalues are in increasing
         // order with the index (not in decreasing as originally). Thus
         // the ground state corresponds to index 1.

   };

///////////////////////// class TrjBas1 ///////////////////////////////////
// abstract base class for trajectories
// Needed for polymorphic container Vp<TrjBas1>

class TrjBas1{
protected: // allows to reuse code from a version in which no derivation
   // was used
   R t_;
   R dt_;
   StateD1 psi_;
public:
   CPM_NAM_V(TrjBas1)
   TrjBas1(){}
   virtual TrjBas1* clone()const{ return new TrjBas1(*this);}

   TrjBas1(StateD1 const& psi, R const& dt):t_(0.),dt_(dt),psi_(psi){}

   virtual void step_(){};
      //: step

   StateD1 psi()const{ return psi_;}
   virtual StateD1 phi()const{ return StateD1();}
   R t()const{ return t_;}
   R dt()const{ return dt_;}
   R norm()const{ return psi_.norm();}
      //: norm
   virtual C expVal()const{ return C(0.);}
      //: expectation value
   R gap()const{ return psi_.gap();}
      //: gap
   virtual void evl_(R tf){}
      //: evolve
};

//////////////////////////// class TrjD ///////////////////////////////////
// Trajectory for Dirac particles based on the leapfrog dynamic

class TrjD1: public TrjBas1{
   OpD1 op_;
   R tau_;
   R tau2_;
   StateD1 phi_; // velocity-like auxiliar quantity for the DALF integrator

public:
   CPM_NAM(TrjD1)
   TrjBas1* clone()const{ return new TrjD1(*this);}

   TrjD1(OpD1 const& op, StateD1 const& psi, R const& dt);
      // dt is the true stepsize. A safe setting is
      // dt=dtLim()*dtRel with 0 < dtRel < 1. It turned out to be more
      // convenient not to build this procedure into the constructor.

   void step_();
      //: step

   StateD1 phi()const{ return phi_;}
   void set_dt_(R h){ dt_=h; tau_= dt_*R(0.25); tau2_ = tau_*R(2);}
      // needed to implement evl_(R)
   C expVal()const{ return (psi_|op_(psi_));}
      //: expectation value
   void evl_(R tf);
      //: evolve
};

//////////////////////////// class SpTrjD /////////////////////////////////
// Spectral trajectory for Dirac particles

class SpTrjD1: public TrjBas1{
   SpectralObjectD1 so_;
      // spectral representation of the evolution generator (e.g. the
      // Hamiltonian).
   SpectralObjectD1 dso_;
      // evolution operator for a step
public:
   CPM_NAM(SpTrjD1)
   TrjBas1* clone()const{ return new SpTrjD1(*this);}

   SpTrjD1(SpectralObjectD1 const& so, StateD1 const& psi, R const& dt):
      TrjBas1(psi,dt),so_(so),dso_(so_){dso_.expi_(-dt_);}

   void step_(){t_+=dt_; psi_=dso_*psi_;}
      //: step

   StateD1 phi()const{ return (so_*psi_)*C(0.,-1.);}

   C expVal()const{ return (psi_|so_*psi_);}
      //: expectation value

   void evl_(R tf);
      //: evolve
};

} // namespace

#endif
