//? cpmtrj.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_TRJ_H_
#define CPM_TRJ_H_
/*
   Description: tools for a unified description of evolution
   resulting from an initial condition
*/

#include <cpmobservable.h>

namespace CpmQuantumMechanics{

   using CpmRootX::B;
   using CpmRootX::Record;
   using CpmAlgorithms::PeriodicAlarm;
   using CpmGraphics::Graph;
   using CpmLinAlg::Z4;

//////////////////////// struct TrjCon ////////////////////////////

struct TrjCon{ // trajectory control
// Here I use the uninitialized native C types (instead of B,R1,Z1)
// since these presently cooperate better with class Record

   B autoAdj_{true};
      // auto-adjust (graphics to curve size)
      // If autoAdj (auto-adjust) is true, the y-range gets reset
      // if needed to let the state not leave the range of the graphics
      // window. If secFac_ is only a bit larger than 1, this adjustment
      // will take place often, which could be annoying; so something
      // like secFac_=1.25 should be chosen.
   B autoStep_{false};
      // if autoStep_==false, dt_ is the time step that will be applied
      // to the state of *this. For autoStep_==true the time step to be
      // done will be computed (under control of QMBase::phiStep),
      // and not taken as the present dt_
   R dt_;
      // the prescribed time step of the trajectory that
      // is considered valid only if autoStep_==false
   R tTot_;
      // total time of the run, after the elapse of which it
      // considered finished. Dependent quantity
   B enfNor_{false};
      // enforce normalization (of state to 1) after each
      // integration step
   B ignPhs_{false};
      // ignore phases, i.e. only the absolute value of the wave
      // function (state) will be represented graphically
   Z nSteps_;
      // number of steps that make up the trajectory
   Z disMode_;
      // display mode
      // selects the graphical representation mode
   R secFac_;
      // security factor e.g. 1.25 controls with which
      // limits the graphics frame adapts to the curve.
      // if 1 adustment is done with any step that
      // changes the curve range by whatsoever small an
      // ammount
   R tWait_;
      // wait time for showing frames in function TrjB::run(Graph)
      // So better rather short such as 0.02
   Z vwnPer_;
      // viewing period. Only each vwnPer_'th computed step will
      // be shown on screen
   Z filPer_;
      // filing period. Only each filPer_'th computed step will
      // be written to file
   Z imgPer_;
      // filing period. Only each imgPer_'th computed step will
      // be written to image (bitmap)
   R dph1_,dph2_;
      // parameters controling an optional dephasing mechanism
   Z dphMethod_;
      // method parameter controling an optional dephasing mechanism
      // If this is 0, no dephasing is active

// parameters which define coarse graining of states after a dynamical
// step
   Z corGrnMet_;
      //: coarse graining method
      // Presently only one method which is switched off if this
      // parameter is zero
   Z nPhi_;
      // number of phi-values
      // all phases of state components will be changed to the
      // nearest of nPhi_ equi-spaced and equi-distributed
      // discrete values
      // see State::corGrn(Z,Z), used as first argument there
   Z nBins_;
      // number of bins
      // see State::corGrn(Z,Z), used as second argument there
   R gam_;
      //: gamma
      // display gamma
   Z intMet_;
      //: interpolation method
      // 0 for pixelization
      // 1 for bilinear interpolation

   typedef TrjCon Type;
   CPM_IO
   CPM_NAM(TrjCon)
   TrjCon():vwnPer_(0),filPer_(0),dph1_(0),dph2_(0),dphMethod_(0),
      corGrnMet_(0), gam_(1),intMet_(0){}
   TrjCon(Record const& rc);
      // constructor from objects that can conveniently be initialized
      // from text
};

//////////////////////// TrjBase //////////////////////////////////////
// Basis for trajectory classes i.e. classes dealing with the
//'initial value problem'.
// In order to be able to hold various derivatives of this class
// as components of polymorphic arrays of type Vp<TrjBase>
// (see class Trj2) it is important to define the virtual function clone.

class TrjBase{ // basis for trajectory classes
   friend class Trj2;
protected:
    TrjCon tc_;
      // trajectory control
    Z d_{0};
      // dimension of state
    Z stepCon_{0};
      // step count, initialized as 0
    R t_,dt_;
      // time and time step. Notice t_ initialized as 0.
    R n_1_;
      // norm of the wavefunction (state) - 1
    C enr_;
      // energy (may due to integrator particularities have a small
      // imaginary component

    PeriodicAlarm vwn_;
    PeriodicAlarm fil_;
    PeriodicAlarm img_;

    CyclicAlarm vwns_;
    CyclicAlarm fils_;
    CyclicAlarm imgs_;

    void adjustBase_()
    {
       t_=0.;
       cpmcerr<<"adjustBase_() started"<<endl;
       if (tc_.dt_< 0) tc_.autoStep_=true;
         // this should not apply to tc_.dt_==0 since in the presence of
         // dephasing methods also dt_==0 may be an interesting situation
       vwn_=PeriodicAlarm(R(tc_.vwnPer_)*tc_.dt_);
       fil_=PeriodicAlarm(R(tc_.filPer_)*tc_.dt_);
       img_=PeriodicAlarm(R(tc_.imgPer_)*tc_.dt_);
       vwns_=CyclicAlarm(tc_.vwnPer_); // viewing
       fils_=CyclicAlarm(tc_.filPer_); // storing on movie file
       imgs_=CyclicAlarm(tc_.imgPer_); // storing as ppm image
       // the following two statements may not be useful if tc_.dt_<0
       tc_.tTot_=R(tc_.nSteps_)*tc_.dt_;
       dt_=tc_.dt_; // experiment
       cpmcerr<<"adjustBase_() done"<<endl;
    }
    virtual void adjust_(){adjustBase_();}
public:
   typedef TrjBase Type;
   virtual TrjBase* clone()const{ return new TrjBase(*this);}
   CPM_IO_V
   CPM_NAM(TrjBase)
   TrjBase(){}
   virtual ~TrjBase(){}
   TrjBase(TrjCon const& tc):tc_(tc){
       cpmmessage("calling constructor of TrjBase");
       adjustBase_();
       }
      // not calling ajust_() here since the call of final adjust_
      // will be made prior to first serious usage of any instance
      // of this class or a derived class

   virtual void step_(R h){}

   Z dim()const{ return d_;}

   virtual State getRep()const{ return State(); }
      // the return type is State which here is used as the
      // adequate data type supporting visualization

   void showAsLine(Graph& g, Z4 const& z4)const
      { getRep().show(g,z4[1],z4[2],z4[3],z4[4]); g.vis();}

   virtual void showAsArea(Graph& g)const{cpmmessage("showAsArea");}

   virtual Word text()const{ return Word("TrjBase");}

   R t()const{ return t_;}
   R& t(){ return t_;}
   // Cygwin g++ 3.4.4 does not allow to use dt_ directly in class TrjDMI
   // which is derived from a class which in turn is derived from the
   // present class.
   // (although all derivations are public). Functions  c a n  be used,
   // only with explicit scope resolution.
   // I don't agree with this rape of the C++ spirit.

   R dt()const{ return dt_;}
   R& dt(){ return dt_;}

   R sup()const{ return getRep().sup();}

   R n_1()const{ return n_1_;}
      // should suggest norm - 1
      // The object the norm of which has to be taken is not
      // yet available in this class

   C enr()const{ return enr_;}
   C& enr(){ return enr_;}

   TrjCon trjCon()const{ return tc_;}

   bool vwnAlarm_()
   { if (tc_.dt_>0) return vwn_(t_); else return vwns_();}
      // the logic is: if tc_.dt_ has a useful value it will
      // be used (together with nSteps_) to determine a time span,
      // that controls function run,  otherwise all control is
      // based on nSteps_ alone.

   bool filAlarm_()
   { if (tc_.dt_>0) return fil_(t_); else return fils_();}

   bool imgAlarm_()
   { if (tc_.dt_>0) return img_(t_); else return imgs_();}

   bool keepGoing()const
      // Notice that we ideally plan the points in time
      // the passing of which should trigger the various actions
      // for which the counters take care. In this form the actual
      // triggering depends on numerical chance events. The shift
      // by tol avoids this.
   {
      const R tolFac=0.25;
      R tol=tolFac*R(dt_);
      if (R(tc_.dt_)>0) return tol+R(t_) <= tc_.tTot_;
      else return stepCon_ < tc_.nSteps_;
   }

   Z stepCon()const{ return stepCon_;}
   Z& stepCon(){ return stepCon_;}

// when trajectories get written to file and then retreived from there
// one should be able to create easily the diagrams concerning
// norm-1 and eTot that get greated at the end of the run-functions.
// The most systematic method would be to store system state sufficiently
// completely so that the values of n1_() and eTot() can be computed
// from them. Unfortunately these data are quite different in the various
// derived classes: in TrjSR they include even the huge (pre-computed)
// constant object so of type SpectralObject. So, in a sense we fake in
// this place and write and read these dependent quantities to and from
// which gives the read-function an ad hoc argument structure.

   virtual bool fileState_(ostream& str){ return false;}

   virtual bool readState_(istream& str){ return false;}

   bool autoStep()const{ return tc_.autoStep_;}
      // says whether the auto stepping option is valid

   void setAutoStep_(bool b){ tc_.autoStep_=b;}

   Z runAsLine(Graph& g, ostream& str);
      // Running n steps of length n and representing this as a color
      // image.
      // The run begins with x0, i.e. t=0 and displays in succession
      // the states at times 0,dt,2*dt,......n*dt. These are (n+1) states
      // which correspond to n time steps

   virtual Z runAsArea(Graph& g, ostream& str);
      // represents the wave function as a colored d_ times d_ image
      // which is natural only for two-particle states of FieldSpace.
      // Here it is the most detailed visualization of the state that
      // allows one to get a feeling for the difference of fermionic
      // and bosonic wave function evolution.

   Z runAsCur(Graph& g, ostream& str);
      // running tc_.nSteps steps of length n and representing this as a
      // series of curves. Time range as in previous function.
      // Each frame is at least shown for a time of waitTime seconds.
      // The y-range is set so that if the initial state would be
      // multiplied by secFac (security factor) it would fill the frame.
      // So setFac>1 is needed to see the whole state *this.
      // Implemented in this class, see cpmtrj.cpp
      // Shows finally norm and energy, very instructive!

   void run(Graph& g, ostream& str)
      // Notice: this function is not virtual. It thus organizes the run
      // of all classes derived from TrjBase (directly, or over
      // intermediate steps
   {
      Z mL=1;
      Word loc("TrjBase::run(Graph,ostream)");
      CPM_MA
      R t0=cpmtime();
      cpmcerr<<"adjust_ ahead"<<endl;
      adjust_(); // is virtual and does non-trivial things
         // in some derived classes
      Z steps=0;
      cpmmessage("TrjBase::run adjust_() done");
      cpmdebug(tc_.disMode_);
      cpmdebug(nameOf());
      if (tc_.disMode_==1) steps=runAsCur(g,str);
      else if (tc_.disMode_==2) steps=runAsLine(g,str);
      else{
         cpmcerr<<" runAsArea ahead"<<endl;
         steps=runAsArea(g,str);
      }
      R t1=cpmtime();
      R tTot=t1-t0;
      str<<"Total number of steps = "<<steps<<endl;
      str<<"Total time of TrjBase::run(g,ostream) was "<<tTot<<endl;

      CPM_MZ
   }

   void playAsLine(Graph& g, istream& str);
   void playAsArea(Graph& g, istream& str);
   void playAsCur(Graph& g, istream& str);

   void play(Graph& g,istream& str)
   {
      Z mL=1;
      Word loc("TrjBase::play(Graph,ostream)");
      CPM_MA
      if (tc_.disMode_==1) playAsCur(g,str);
      else if (tc_.disMode_==2) playAsLine(g,str);
      else playAsArea(g,str);
      CPM_MZ
   }
};

//////////////////////// TrjT<> //////////////////////////////////////
// Trajectory class which adds to TrjBase the essential state
// descriptor x_ and thus allows a final implementation of
// some of the virtual functions in TrjBase

template <class StateT>
class TrjT: public TrjBase{//adds to TrjBase the state descriptor x_
protected:
    StateT x_;
      // state
    virtual void adjust_()
    {
       cpmcerr<<"TrjT::adjust_() started"<<endl;
       TrjBase::adjust_();
       d_=x_.dim();
       R n1=x_.norm();
       n_1_=n1-1;
       cpmcerr<<"TrjT::adjust_() done"<<endl;
    }
public:
   typedef TrjT<StateT> Type;
   virtual TrjBase* clone()const{ return new TrjT(*this);}
   CPM_IO_V
   TrjT():TrjBase(){}
   TrjT(StateT const& x, TrjCon const& tc):TrjBase(tc),x_(x)
   {cpmmessage("calling ctor of TrjT");}

   void setState_(StateT const& x){  x_=x; adjust_();}
      // setting a new x_ and adjusting the rest. since adjust_
      // is virtual, this definition is OK for all derived classes

   State getRep()const{ return x_.toState(tc_.ignPhs_); }
      // the return type is State and not the template argument StateT
      // Here State is a data type supporting visualization
      // also in cases where StateT is not of fixed dimension

   StateT& x(){return x_;}

   StateT const& x()const{return x_;}

   void norIfEnf_()
       // normalize if enforced, update n_1_ always
   {
      if (tc_.enfNor_) x_.nor_();
      n_1_=x_.norm()-1.;
   }

   virtual Word text()const{ return Word("TrjT<>");}

   bool fileState_(ostream& str){
      cpmp(d_);
      cpmp(t_);
      cpmp(n_1_);
      cpmp(enr_);
      cpmp(x_);
      return true;
   }

   bool readState_(istream& str)
   {
      Z mL=3;
      Word loc="Trj<>::readState_(istream)";
      CPM_MA
      cpms(d_);
      cpms(t_);
      cpms(n_1_);
      cpms(enr_);
      cpms(x_);
      stepCon_+=1;
         // we dont write the step counter to file
      CPM_MZ
      return true;
   }
};

template <class StateT>
bool TrjT<StateT>::prnOn(ostream& str)const
{
   cpmwt("TrjT<>");
   if (! TrjBase::prnOn(str)) return false;
   cpmp(x_);
   return true;
}

template <class StateT>
bool TrjT<StateT>::scanFrom(istream& str)
{
   if (! TrjBase::scanFrom(str)) return false;
   cpms(x_);
   adjust_();
   return true;
}

//////////////////////// TrjDMI<> ////////////////////////////////////////
// Trajectory class which adds to TrjT<> the essential state
// change descriptor v_ and thus allows a final implementation of
// all virtual functions in TrjT<>

template <class StateT>
class TrjDMI: public TrjT<StateT>{ // adding velocity to TrjT

   B initialized;
   StateT v_; //velocity
   virtual StateT op(StateT const& x, R t)const{ t; return StateT();}

//protected:
public:
   typedef TrjT<StateT> Base;
   C eTot()const{
     // cpmcerr<<"eTot() entered"<<endl;
      auto res=(Base::x_|v_)*QMBase::get_ihDirac();
     // cpmcerr<<"eTot() done"<<endl;
      return res;
   }
      // expectation value of the operator which (in the sense of Stone's
      // theorem) is the generator of the trajectory.

protected:
   virtual void adjust_(){
      cpmcerr<<"TrjDMI::adjust_() started"<<endl;
      Base::adjust_();
      cpmcerr<<"v_ ahead"<<endl;
      v_=op(Base::x_,Base::t_)*QMBase::get_ihDiracInv();
      cpmcerr<<"v_ done"<<endl;
      TrjBase::enr()=eTot();
      cpmcerr<<"TrjDMI::adjust_() done"<<endl;
   }

public:
   typedef TrjDMI<StateT> Type;
   virtual TrjBase* clone()const{ return new TrjDMI(*this);}
   CPM_IO_V
   TrjDMI():TrjT<StateT>(){}
   TrjDMI(StateT const& x, TrjCon const& tc):TrjT<StateT>(x,tc)
   {cpmmessage("calling ctor of TrjDMI");}
   virtual Word text()const{ return Word("TrjDMI<>");}
   void step_(R h);
};

template <class StateT>
void TrjDMI<StateT>::step_(R h)
{
   static C fac=QMBase::get_ihDiracInv()*QMBase::get_ihDiracInv();
 //  static C fac2=QMBase::get_ihDirac();
   static Word loc("TrjDMI<>::step_(R h)");
   Z mL=3;
   CPM_MA
   if (!initialized){
      adjust_();
      initialized=B(true);
   }

   if (TrjBase::autoStep()){ // due to time dependence of
      // the generator, the proposed time step will vary with time
      // and has to be computed for each step (rather cheaply)
      R n1=Base::x_.norm();
      R n2=v_.norm();
      TrjBase::dt()=QMBase::step(n1,n2);
   }
   else TrjBase::dt()=h;

   R tau=R(0.5)*TrjBase::dt();
   StateT xi=Base::x_+v_*tau;
   R ti=tau+Base::t();
   StateT dv=op(op(xi,ti),ti)*(fac*TrjBase::dt());
   Base::x_+=(v_*TrjBase::dt()+dv*tau);
   v_+=dv;
   TrjBase::t()+=TrjBase::dt();
   TrjBase::stepCon()+=1;
   Base::norIfEnf_();
   TrjBase::enr()=eTot(); // very cheap function here
   CPM_MZ
}

template <class StateT>
bool TrjDMI<StateT>::prnOn(ostream& str)const
{
   cpmwt("TrjDMI<>");
   if (! TrjT<StateT>::prnOn(str)) return false;
   cpmp(v_);
   return true;
}

template <class StateT>
bool TrjDMI<StateT>::scanFrom(istream& str)
{
   if (! TrjT<StateT>::scanFrom(str)) return false;
   cpms(v_);
   return true;
}

//////////////////////// Trj2DMI<> //////////////////////////////////////
// Trajectory class which adds to TrjT<> the essential state
// change descriptor v_ , a Hamiltonian, and a decoherence operator.
// It allows a final implementation of all virtual functions in TrjT<>.
// Notice: in the present class the 2 in the name does not mean that
// two or more trajectories are simultaneously considered as it will be
// the case for the Trj2 at the end of this file.

template <class StateT, class ObsT, class DecT >
class Trj2DMI: public TrjT<StateT>{//adds velocity and hamiltonian to TrjT

   B initialized;
   StateT v_; // velocity
   ObsT ham_; // hamiltonian
   DecT dec_; // decoherence operator
   StateT op(StateT const& x, R t)const{ t; return ham_(x);}
   //void dePhase_(){x_=dec_.dePhase(x_,tc_.dph1_,tc_.dph2_,tc_.nBins_);
   // v_=op(x_,t());}
   //void dePhase_(){
   //   X2<R,StateT> xrs=dec_.reduction(x_,tc_.dph1_);
    //  x_=xrs.c2();v_=op(x_,t());
   //}
   void dePhase_(){
      Z i=40,j=41;
      bool act=Base::x_.col_(IvZ(i,j),Base::tc_.dph1_);
      if (act){
         Base::x_.smt_(4);
         v_=op(Base::x_,Base::t());
      }
   }
//protected:
public:
   typedef TrjT<StateT> Base;
   C eTot()const{ return (Base::x_|v_)*QMBase::get_ihDirac();}
      // expectation value of the operator which (in the sense of Stone's
      // theorem) is the generator of the trajectory.

protected:
   virtual void adjust_(){
      Base::adjust_();
      v_=op(Base::x_,Base::t_)*QMBase::get_ihDiracInv();
      TrjBase::enr()=eTot();
   }

public:
   typedef Trj2DMI<StateT,ObsT,DecT> Type;
   virtual TrjBase* clone()const{ return new Trj2DMI(*this);}
   CPM_IO_V
   Trj2DMI():TrjT<StateT>(){}
   Trj2DMI(StateT const& x, ObsT const& ham, DecT const& dec,
      TrjCon const& tc):TrjT<StateT>(x,tc),ham_(ham),dec_(dec)
   {cpmmessage("calling ctor of Trj2DMI");}
   Word text()const{ return Word("Trj2DMI<>");}
   void step_(R h);
};

template <class StateT, class ObsT, class DecT>
void Trj2DMI<StateT,ObsT,DecT>::step_(R h)
{
   static C fac=QMBase::get_ihDiracInv()*QMBase::get_ihDiracInv();
   static Word loc("Trj2DMI<>::step_(R h)");
   Z mL=3;
   CPM_MA
   if (!initialized){
      adjust_();
      initialized=B(true);
   }

   if (TrjBase::autoStep()){ // due to time dependence of
      // the generator, the proposed time step will vary with time
      // and has to be computed for each step (rather cheaply)
      R n1=Base::x_.norm();
      R n2=v_.norm();
      TrjBase::dt()=QMBase::step(n1,n2);
   }
   else TrjBase::dt()=h;

   R t=TrjBase::t();
   R dt=TrjBase::dt();
   R tau=dt*0.5;

   t+=tau;
   Base::x_+=v_*tau;
   v_+=op(op(Base::x_,t),t)*(fac*dt);
   Base::x_+=v_*tau;
   t+=tau;

   TrjBase::t()+=dt;
   TrjBase::stepCon()+=1;
   dePhase_();
   Base::norIfEnf_();
   TrjBase::enr()=eTot(); // very cheap function here
   CPM_MZ
}

template <class StateT, class ObsT, class DecT>
bool Trj2DMI<StateT,ObsT,DecT>::prnOn(ostream& str)const
{
   cpmwt("Trj2DMI<>");
   if (! TrjT<StateT>::prnOn(str)) return false;
   cpmp(v_);
   return true;
}

template <class StateT, class ObsT, class DecT>
bool Trj2DMI<StateT,ObsT,DecT>::scanFrom(istream& str)
{
   if (! TrjT<StateT>::scanFrom(str)) return false;
   cpms(v_);
   return true;
}

//////////////////////// TrjB //////////////////////////////////////
// Basis for trajectory classes i.e. classes dealing with the
//'initial value problem'.
// In order to be able to hold various derivatives of this class
// as components of polymorphic arrays of type Vp<TrjB>
// (see class Trj2) it is important to define the virtual function clone.

typedef TrjT<State> TrjB;

/////////////////////// class TrjSR ///////////////////////////
// a trajectory holds a SpectralObject from a Observable and an
// initial state. From the spectral data it calculates future states
// based on the spectral theorem and does not need iterated
// matrix multiplication.

class TrjSR: public TrjB{//trajectory from a SpectralObject
   SpectralObject so_;
   V<C> cof_;
protected:
   virtual void adjust_();
public:
   typedef TrjSR Type;
   virtual TrjBase* clone()const{ return new TrjSR(*this);}
   CPM_IO_V
   TrjSR(){}
   TrjSR(const Observable& h, const State& x, TrjCon const& tc );
      // creating a trajectory from hamiltonian and initial state
   TrjSR(const SpectralObject& so, const State& x, TrjCon const& tc );
      // creating a trajectory from the spectral representation of its
      // hamiltonian and an initial state

   C eTot()const{ return (x_|so_*x_);}
      // see declaration in TrjB

   virtual Word text()const{ return Word("TrjSR");}

   virtual void step_(R h);
   // if autoStep_==false, h is the time step that will be applied
   // to the state of *this. For autoStep_==true the time step to be
   // done is dt_ independent of the value of h.
};

/////////////////////// class TrjDMI_Obs /////////////////////////////////
class TrjDMI_Obs: public TrjDMI<State>{ // trajectory from observable

   Observable obs_;
   virtual State op(State const& x, R t)const{ return obs_*x;}
public:
   typedef TrjDMI_Obs Type;
   virtual TrjBase* clone()const{ return new TrjDMI_Obs(*this);}
   CPM_IO_V
   TrjDMI_Obs(){}
   TrjDMI_Obs(Observable const& obs, State const& x, TrjCon const& tc):
   TrjDMI<State>(x,tc),obs_(obs){}
   virtual Word text()const{ return Word("TrjDMI_Obs");}
};

/////////////////////// class Trj2 ///////////////////////////////////////
// This class holds two trajectories that have the same
// initial state and evolve in time according to the two
// methods considered above. The runAsCurves method aimes
// at making a comparison of the two methods easy. Notice that
// the SR-method is exact mathematically (which does not
// exclude numerical inaccuracies) whereas the DMI-method
// is a approximation (however, a simple,stable, and accurate one)
// After the reimplementation of 2004-09-26 we can compare all
// trajectory types that are derived from the newly introduced
// trajectory base class TrjB. Introduction of this class
// eliminated all major code doublications

class Trj2{ // array of trajectories from various methods for comparison

   Vp<TrjBase> rep_;

public:
   Trj2(){}
   Trj2(TrjBase const& trj1, TrjBase const& trj2 ):rep_(2)
   { rep_[1]=trj1; rep_[2]=trj2;}

   Trj2(Vp<TrjBase> const& rep):rep_(rep){}
   void step_(R h);
   Z dim()const{ return rep_(2).dim();}

   R n_1()const{ return rep_(2).n_1();}

   State getRep1()const{ return rep_(1).getRep();}
   State getRep2()const{ return rep_(2).getRep();}
      // State needs not to be the type of the state which varies
      // along the trajectory. It may be used as a visualization device

   Word text1()const{ return rep_(1).text();}
   Word text2()const{ return rep_(2).text();}

   R t()const{ return rep_(2).t();}
   R dt()const{return rep_(2).dt();}
   bool vwnAlarm_(){ return rep_(2).vwnAlarm_();}
   bool filAlarm_(){ return rep_(2).filAlarm_();}

   void fileState_(ostream& str1, ostream& str2)
   { rep_(1).fileState_(str1); rep_(2).fileState_(str2);}

   void run(Graph& g, ostream& str1, ostream& str2);
      // running n steps of length n and representing this as a series of
      // curves. Time range as in previous function. Each frame is at
      // least shown for a time of waitTime seconds.
      // The y-range is set so that if the initial state would be
      // multiplied by secFac (security factor) it would fill the frame.
      // So setFac>1 is needed to see the whole state *this.
      // If autoAdj (auto-adjust) is true, the y-range gets reset
      // if needed to let the state not leave the range of g. If secFac
      // is only a bit larger than 1, this adjustment will take place
      // often, which could be annoying; so something like secFac=1.25
      // should be chosen.
   void play(Graph& g, istream& str1, istream& str2);
};

} // namespace
#endif
