//? cpmvc.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_VC_H_
#define CPM_VC_H_
/*
    Description: Lean arithmtics for Vl<C>  and Vl< Vl<C> >
*/

#include <cpmv.h>
#include <cpmc.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmRoot::Word;

using CpmArrays::V;
using CpmArrays::Outside;
using CpmFunctions::F;

// interfaces as declaration macros

#define CPM_VC_PURE_VECTOR_SPACE_INTERFACE(X,S)\
public:\
   X operator - (void)const;\
   X& operator *=(S);\
   X& operator +=(X const&);\
   X& operator -=(X const&);\
   X operator *(S)const;\
   X operator +(X const&)const;\
   X operator -(X const&)const;

#define CPM_VC_PURE_SCALAR_INTERFACE(X,S)\
public:\
    X& operator *=(S);\
    X operator *(S)const;

#define CPM_VC_SCALAR_PRODUCT_INTERFACE(X,S)\
public:\
   X con(void)const;\
   S operator |(X const&)const;\
   R normSqr(void)const{ return (operator|(*this)).real();}\
   R norm(void)const{ return cpmsqrt(normSqr());}\
   R abs(void)const{ return norm();}\
   R absSqr(void)const{ return normSqr();}

namespace CpmArrays{

//////////////////////////// class VC ////////////////////////////////////

   class VC{
// vectors with C-valued components, rich interface including arithmetics
      V<C> rep_;
         //: representation
      typedef VC Type;
      typedef C ScalarType;
   public:
      CPM_NAM(VC)
      CPM_VC_PURE_VECTOR_SPACE_INTERFACE(Type,C)
      CPM_VC_SCALAR_PRODUCT_INTERFACE(Type,C)
      CPM_VC_PURE_SCALAR_INTERFACE(Type,R)
      CPM_NORMALIZE

      explicit VC(Z d=0): rep_(d){}
      VC(Z d, C z): rep_(d,z){}
      explicit VC(V<C> const& v): rep_(v){}
         // construction from data

      V<C> rep()const{ return rep_;}
         //: representation

      Z dim()const{ return rep_.dim();}
         //: dimension

      Z b()const{ return rep_.b();}
      Z e()const{ return rep_.e();}

      C const& operator[](Z i)const{ return rep_[i];}
      C& operator[](Z i){ return rep_[i];}

      C const& cui(Z i)const{ return rep_.cui(i);}
      C& cui(Z i){ return rep_.cui(i);}

      C operator()(Z i, Outside mode=DEFAULT)const{ return rep_(i,mode);}
      // Read access defined for all i.
      C const& read(Z i, Outside mode=DEFAULT)const
         { return rep_.read(i,mode);}
      //: read ; Reading components in an efficient (as &'s), safe,
      // and flexible manner.

      VC& makeReal_(R p=2);
      // returns a VC with all components of *this
      // replaced by their absolute value to power p

      VC ran(Z j=0)const;
         //: random values
   };

} // namespace

#endif
