//? cpmwaves.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_WAVES_H_
#define CPM_WAVES_H_

/*
   Description : wave equation in one dimension.
   Introducing a subdivision of the 1-dimensional discrete
   chain that is intended to help parallelization
*/

#include <cpmrfunc.h>
#include <cpmrmatrix.h>
#include <cpmlinalg.h>
#include <cpmv.h>

namespace CpmWaves{

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::Word;
//using CpmRootX::Z1;
//using CpmRootX::R1;
using CpmRoot::B;
using CpmLinAlg::R2;
using CpmArrays::R_Vector;
using CpmArrays::R_Matrix;
using CpmArrays::V;
using CpmFunctions::R_Func;
using CpmGeo::Iv;

/////////////////// class Wave ///////////////////////////////////////////

class Wave{ // states of a discrete string
   Z d{0};
      // number of discrete points in representing the state of the string
   R c{0.};
      // velocity in the wave equation (could be made position
      // dependent for more interesting applications)
   R dx{0.};
      // spatial separation of neighbor points.
   Z method{0};
      // method to compute a time step
      // presently 0 or 1
      // 0 Euler method, 1 explicit mid point method (see
      // implementation of Wave::step(R) in file cpmwaves.cpp)
   R_Vector x;
      // x[i] is the wave quantity (e.g. elongation of the elastic
      // string) at point i; i=1,2,...d. We treat the periodic problem in
      // which the right neighbor of i=d is i=1
   R_Vector v;
      // dx[i]/dt. Such a representation is superior to considering x
      // alone and let x[i,t],x[i,t-dt] be a representation of the
      // velocity

   R fac{0.};
      // dependent quantity (c/dx)^2; needed for efficient coding of
      // function acc

   R_Vector acc(const R_Vector& ra); // acceleration
      // function which determines the dynamics; in our case it is
      // c^2 * second spatial derivation. The discrete version of
      // the spatial derivatives are written in a way that the
      // periodicity mentioned above is taken into account.

   void finish(){ fac=c/dx; fac*=fac;}

   typedef Wave Type;
      // this typedef is expected by the following declaration
      // macro CPM_IO, which includes a unified definition of send
      // and receive for Wave objects
public:
   Word nameOf()const{ return "Wave";}
   CPM_IO
      // declaration and partial implementation of I/O by a macro
explicit   Wave(Z d0=1, R c0=1, R dx0=1 );
      // a basic constructor
   void step(R dt);
      // time stepping algorithm, depends on value of method
   R get_c()const{ return c;}
      // evident
   R get_dx()const{ return dx;}
      // evident
   R2 getLeft()const{ return R2(x[2],v[2]);}
      // returns the data that a 'left neighbor' needs to replace
      // his own unreliably computed boundary values
   R2 getRight()const{ return R2(x[d-1],v[d-1]);}
      // returns the data that a 'right neighbor' needs to replace
      // his own unreliably computed boundary values

   R2 to(Z j)const
  { if (j==-1) return getRight(); else return getLeft(); }

   void from(Z j, R2 xv)
  //    { if (j==1) setRight(xv); else setLeft(xv);}
  { if (j==1) setRight(xv); else setLeft(xv); }

   void setRight(R2 xv){ x[d]=xv[1]; v[d]=xv[2];}
      // replaces own unreliable boundary data by the
      // values given by the argument (these argument values have to
      // be obtained by calling getLeft() (sic!) for the right neighbor
   void setLeft(R2 xv){ x[1]=xv[1]; v[1]=xv[2];}
      // replaces own unreliable boundary data by the
      // values given by the argument (these argument values have to
      // be obtained by calling getRight() for the ledt neighbor
   void setLocalState(Z j, R ax, R vx)
      {x.cyc(j)=ax; v.cyc(j)=vx;}
      // evident
   R2 getLocalState(Z j)const{ return R2(x.cyc(j),v.cyc(j));}
      // evident
   Z dim()const{ return d;}
      // evident
   void setMethod(Z m){ method=m;}
      // evident
   R_Matrix rep()const;
     // graphical object representing the amplitude distribution
};

///////////////////// class Waves ////////////////////////////////////////

class Waves{ // array of waves on strata with one-point overlap

   Z d{0};
   R c{0.};
   R dx{0.};
   Z method{0};
   V<Wave> vw;
   typedef Waves Type;
public:
   CPM_IO
   Waves(){}
   Waves(const V<Z>& vz, R c0, R dx0);
   Waves(const V<Wave>& vw0);
   void step(R dt);
   void setState(const R_Func& ampl, const R_Func& vel);
      // setting amplitude and velocity from functions
      // defined on 1-dimensional position space
   void adjust();
      // sets the boundary values in a consistent manner
   R_Matrix rep()const;
      // graphical object representing the amplitude distribution
   Iv getSpace()const;
      // getting the lower and the upper end of the x-space covered
   void setMethod(Z m)
      { method=m; for (Z i=1;i<=d;i++) vw[i].setMethod(m);}
   Wave& operator[](Z i){ return vw[i];}
   const Wave& operator[](Z i)const{ return vw[i];}
   V<Wave> get_vw()const{ return vw;}
};

} // namespace

#endif
