//? cpmfieldstate.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmfieldstate.h>
#include <cpmimagingtools.h>
#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmx.h>
#include <cpmrvector.h>
#include <cpmfieldoperator.h>
#include <cpmstate2.h> // for using a graphics function from there

using namespace CpmQuantumMechanics;
using namespace CpmRoot;
using namespace CpmArrays;

/************************* class Ky *********************************/

Z Ky::lim_=1;
   // the default value should come into action only for unmanagable
   // high 'occupation numbers' such as 1000

Z Ky::eps_=1;

// not relying verbally on the implementation in Vo<Z>
// for utmost efficiency

Z Ky::com(const Ky& s)const
{
   return v_.com(s.v_);
}

bool Ky::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(v_);
   return true;
}

bool Ky::scanFrom(istream& str)
{
   cpms(v_);
   ini_();
   return true;
}

Z Ky::sup()const
{
   if (s_<1) return -1;
   Z res=v_.cui(1);
   for (Z i=2;i<=s_;i++){
      Z vi=v_.cui(i);
      if (vi>res) res=vi;
   }
   return res;
}

Z Ky::inf()const
{
   if (s_<1) return -1;
   Z res=v_.cui(1);
   for (Z i=2;i<=s_;i++){
      Z vi=v_.cui(i);
      if (vi<res) res=vi;
   }
   return res;
}

void Ky::sup_(Ky const& k)
{
   if (s_!=k.s_) cpmerror("dimension mismatch in Ky::sup_(Ky)");
   for (Z i=1;i<=s_;++i){
      Z n1=v_.cui(i);
      Z n2=k.v_.cui(i);
      v_.cui(i)=CpmRootX::sup<Z>(n1,n2);
   }
}

bool Ky::up_(Z i)
{
   Z n=v_.cui(i);
   if (n<lim_){
      v_.cui(i)+=1;
      return true;
   }
   else return false;
}

bool Ky::down_(Z i)
{
   if (s_==0) return false; // don't change v_ and don't call indexing
      // operator
 //  Z n=v_.cui(i);
   Z n=v_[i];
   if (n>0){ // change a component of v_, but not its length
   //   v_.cui(i)-=1;
      v_[i]-=1;
      return true;
   }
   else{
      return false;
   }
}

R Ky::sigma(Z i)const
{
   Z sum=0;
   for (Z k=1;k<=i;++k) sum+=v_.cui(k);
   return CpmRootX::parity(sum);
}

R Ky::up(Z i)const
{
   Z ni=v_.cui(i);
   if (eps_==-1)
      return ::sqrt(1.+ni);
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
      return sigma(i)*(1-ni);
   }
}

R Ky::down(Z i)const
{
   if (s_==0) return 0.;
   Z ni=v_.cui(i);
   if (eps_==-1){
      return cpmsqrt(R(ni));
   }
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
      return sigma(i-1)*ni;
   }
}

/******************* class FieldState *********************************/

bool FieldState::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("FieldState",str)) return false;
   cpmp(s_);
   cpmp(lim_);
   cpmp(pbc_);
   cpmp(m_);
   return true;
}

bool FieldState::scanFrom(istream& str)
{
   Z mL=3;
   static Word loc="FieldState::scanFrom(istream)";
   CPM_MA
   cpms(s_);
   cpms(lim_);
   cpms(pbc_);
   cpms(m_);
   CPM_MZ
   return true;
}

FieldState::FieldState(Z s, Z kind, Z lim, bool pbc):s_(s>0 ? s : 1),
   lim_(lim),pbc_(pbc),m_()
{
   Ky::setLim(lim_);
   if (kind==1) m_[Ky(s_)]=C::one;
}

FieldState::FieldState(State const& st, Z lim, bool pbc):s_(st.dim()),
   lim_(lim),pbc_(pbc)
{
   Ky::setLim(lim_);
   Ky k0(s_);
   for (Z i=1;i<=s_;++i){
      Ky k(k0);
      k.up_(i);
      m_[k]=st[i];
   }
}

FieldState::FieldState(V<State> const& vs, Z lim, bool pbc):
   lim_(lim),pbc_(pbc)
{
   Z mL=3;
   Word loc("FieldState::FieldState(V<State> const& vs)");
   CPM_MA
   Ky::setLim(lim_);
   Z s=(vs.dim()>0 ? vs[1].dim() : 0);
   Z kind=1;
   FieldState fs(s,kind,lim_,pbc_); // vacuum state
   for ( Z i=vs.b();i<=vs.e();++i) fs=fs.crt(vs[i]);
   fs.nor_();
   s_=fs.size();
   m_=fs.m_;
   CPM_MZ
}

C FieldState::get(Ky const& k)const
// notice that in the case of  std::map<Ky,C> the
// code {return m_[k];} would not be allowed
// for a const function get.
{
   return m_[k];
}

// notice the beautyful symmetry and structure in the
// implementation of crt and des.

FieldState FieldState::crt(Z j0)const
{
   Word loc("FieldState::crt(Z)");
   Z mL=3;
   CPM_MA
   Z j=j0;
   b_1(j);
   Ky::setLim(lim_);
   Mkc res;
   for (auto const& [key,value] : m_.p_){
      Ky kj{key};
      if (kj.up_(j)) res[kj]=value*key.up(j);
   }
   CPM_MZ
   return mod(res);
}

FieldState FieldState::crt(State const& st)const
{
   FieldState res(s_);
   for (Z i=1;i<=s_;++i) res+=crt(i)*st(i);
   return res;
}

FieldState FieldState::des(Z j0)const
{

   Word loc("FieldState::des(Z ");
   loc&=cpm(j0);
   loc&=" )";
   Z mL=3;
   CPM_MA
   Z j=j0;
   b_1(j);
   Ky::setLim(lim_);
   Mkc res;
   for (auto const& [key,value] : m_.p_){
      Ky kj{key};
      if (kj.down_(j)) res[kj]=value*key.down(j);
   }
   CPM_MZ
   return mod(res);
}

FieldState FieldState::des(State const& st)const
{
   FieldState res(s_);
   for (Z i=1;i<=s_;++i) res+=des(i)*st(i).con();
   return res;
}

FieldState FieldState::cook1(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldState::cook1(Observable)");
   CPM_MA
   FieldState res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
        res+=des(i).crt(j)*ob(i,j);
   CPM_MZ
   return res;
}

FieldState FieldState::cook2(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldState::cook2(Observable)");
   CPM_MA
   FieldState res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
       res+=des(j).des(i).crt(i).crt(j)*(ob(i,j)*0.5);
   CPM_MZ
   return res;
}

FieldState FieldState::manyParHam(R2_Func const& pot1,
   R2_Func const& pot2, R t)const
{
   Z mL=3;
   Word loc("FieldState::manParHam(R2_Func,R2_Func,R)");
   CPM_MA
   static R hD=QMBase::get_hDirac();
   static R hD2=hD*hD;
   Z i,j,d=s_;
   //R dxNorm=1./d;
   R dxInv2=R(d)/QMBase::getSize(); // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(QMBase::getMass()*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   FieldState res=zero();
   V<R> x0=posNor(d);
   for (i=1;i<=d;++i){
      FieldState fsi=des(i); // can also be written as fsi = (*this).des(i)
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.)*fac;
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   for (i=1;i<=d;++i){
      R xi=x0[i];
      C xit(xi,t);
      R pxit=pot1(xit);
      if (pxit!=0) res+=des(i).crt(i)*pxit; // efficiency !
      for (j=1;j<=d;++j){
         R xij=xi-x0[j];
         if (xij>0.5)  xij-=1;
         else if (xij<-0.5) xij+=1;
         else ;
         C xijt(xij,t);
         R pxijt=pot2(xijt)*0.5;
         if (pxijt!=0) res+=des(j).des(i).crt(i).crt(j)*pxijt;
            // efficiency !
      }
   }
   CPM_MZ
   return res;
}

FieldState FieldState::manyParHam_t(
   R2_Func const& pot1, R_Func const& tf1,
   R2_Func const& pot2, R_Func const& tf2, R t)const
{
   Z mL=3;
   Word loc("FieldState::manParHam_t(...)");
   CPM_MA
   static R hD=QMBase::get_hDirac();
   static R hD2=hD*hD;
   Z i,j,d=s_;
  // R dxNorm=1./d;
   R dxInv2=R(d)/QMBase::getSize(); // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(QMBase::getMass()*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   FieldState res=zero();
   V<R> x0=posNor(d);
   for (i=1;i<=d;++i){
      FieldState fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.)*fac;
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions. For places where the
   // potential is zero, no work should be done.
   R f1=tf1(t), f2=tf2(t)*0.5;
   if (f1!=0 || f2!=0){ // efficiency
      for (i=1;i<=d;++i){
         R xi=x0[i];
         if (f1!=0){ // efficiency !
            C xit(xi,t);
            R pxit=pot1(xit)*f1;
            if (pxit!=0) res+=des(i).crt(i)*pxit; // efficiency !
         }
         if (f2!=0){ // efficiency
            for (j=1;j<=d;++j){
               R xij=xi-x0[j];
               if (xij>0.5)  xij-=1;
               else if (xij<-0.5) xij+=1;
               else ;
               C xijt(xij,t);
               R pxijt=pot2(xijt)*f2;
               if (pxijt!=0) res+=des(j).des(i).crt(i).crt(j)*pxijt;
            // efficiency !
            }
         }
      }
   }
   CPM_MZ
   return res;
}

FieldState FieldState::manyParHam_w(
   R2_Func const& pot, R2_Func const& well, R_Func const& tf, R t)const
{
   Z mL=3;
   Word loc("FieldState::manParHam_w(...)");
   CPM_MA
   static R hD=QMBase::get_hDirac();
   static R hD2=hD*hD;
   Z i,d=s_;
   //R dxNorm=1./d;
   R dxInv2=R(d)/QMBase::getSize(); // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(QMBase::getMass()*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   FieldState res=zero();
   V<R> x0=positions(d);
   for (i=1;i<=d;++i){
      FieldState fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.)*fac;
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions. For places where the
   // potential is zero, no work should be done.
   R ft=tf(t);
   for (i=1;i<=d;++i){
      R xi=x0[i];
      C xit(xi,t);
      R pxit=pot(xit)+(ft==0. ? 0. : well(xit)*ft);
      if (pxit!=0) res+=des(i).crt(i)*pxit; // efficiency !
   }
   cpmvalue("ft",ft,2);
   CPM_MZ
   return res;
}

FieldState FieldState::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldState::hub(R,R,R)");
   CPM_MA
   t; // no role so far
   FieldState res=zero();
   for (Z i=1;i<=s_;++i){
      Z i1=i+1;
      FieldState f1i=des(i1).crt(i);
      FieldState f2i=des(i).crt(i1);
      res+=(f1i+f2i)*a;
      res+=des(i).des(i).crt(i).crt(i)*b;
   }
   CPM_MZ
   return res;
}

FieldState FieldState::reduce(Z d)const
{
   Mkc res;
   for ( auto const& [x,y] : m_.p_){
      Z max=x.sup();
      if (max<=d) res[x]=y;
   }
   return mod(res);
}

C FieldState::operator[](Z j)const
{
   C res;
   for ( auto const& [x,y] : m_.p_){
      if (x[j]>0) res+=y;
   }
   return res;
}

State FieldState::toState(bool ignorePhases)const
{
   ignorePhases;
   return den();
}

State FieldState::den()const
{
   State res(s_);
   for (Z i=1;i<=s_;++i){
      FieldState fsi=des(i);
      res[i]=C(fsi.absSqr());
   }
   return res;
}

Observable FieldState::toObservable(bool ignorePhases)const
{
   FieldState vac(s_,1,lim_);
   Observable res(s_);
   for (Z i=1;i<=s_;++i){
      for (Z j=1;j<=s_;++j){
         FieldState sij=des(i).des(j);
         C z=(vac|sij);
         if (ignorePhases) z=C(z.abs());
         res.setLean(i,j,z);
      }
   }
   return res;
}

X3<Ky,Z,R> FieldState::dgn()const
{
   Z res{0};
   Ky ks(s_);
   for ( auto const& [x,y] : m_.p_){
      ks.sup_(x);
      res+=1;
   }
   R n(norm());
   return X3<Ky,Z,R>(ks,res,n);
}

C FieldState::operator|(FieldState const& fs)const
// see Musser,Saini: STL Tutorial and Reference Guide,
// Addison-Wesley 1996, Example 7-6, p. 169,170
// for the pattern of this implementation.
{
   C res;
   for ( auto const& [x,y] : m_.p_){
      if (fs.m_.defined(x)){
         C z=y;
         C z2=fs[x];
         res+=z.con()*z2;
      }
   }
   return res;
}

FieldState FieldState::operator+(FieldState const& fs)const
{

   Mkc res(m_);
   for (auto const& [x,y] : fs.m_.p_){
      res[x]+=y;
   }
   return mod(res);
}

FieldState FieldState::operator-(FieldState const& fs)const
{
   Mkc res(m_);
   for (auto const& [x,y] : fs.m_.p_){
      res[x]-=y;
   }
   return mod(res);
}

FieldState FieldState::operator*(C z)const
{
   Mkc res(m_);
   for (auto const& [x,y] : res.p_) res[x]*=z;
   return mod(res);
}

FieldState FieldState::operator*(R r)const
{
   Z mL=3;
   Word loc("FieldState::operator*(R r="&cpm(r)&")");
   CPM_MA
   Mkc res(m_);
   for (auto const& [x,y] : res.p_) res[x]*=r;
   CPM_MZ
   return mod(res);
}

FieldState& FieldState::operator+=(FieldState const& fs)
{
   return *this=operator+(fs);
}

FieldState& FieldState::operator-=(FieldState const& fs)
{
   return *this=operator-(fs);
}

FieldState& FieldState::operator*=(C z)
{
   return *this=operator*(z);
}

FieldState& FieldState::operator*=(R r)
{
   return *this=operator*(r);
}

R FieldState::sup(void)const
{
   R res=0;
   for (auto const& [x,y] : m_.p_){
      R val=y.abs();
      if (val>res) res=val;
   }
   return res;
}

FieldState FieldState::con(void)const
{
   Mkc res;
   for (auto const& [x,y] : m_.p_){ res[x]=y.con();}
   return mod(res);
}

FieldState FieldState::operator-(void)const
{
   Mkc res;
   for (auto const& [x,y] : m_.p_){ res[x]=-y;}
   return mod(res);
}

R FieldState::nor_(R r)
{
   Z d=dim();
   if (d==0) return 0.;

   R r0=norm();
   if (r0==0){
      FieldState aux=vac();
         // ground state
      aux*=r;
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

void FieldState::show(Frame& f, Z mult_i, Z mult_j,
      Z iShift, Z jShift, bool ignorePhases)const
{
   toState(ignorePhases).show(f,mult_i,mult_j,iShift,jShift,ignorePhases);
}

void FieldState::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
// Represents the component of *this which belongs to the two-particle
// subspace of the Fock space.
{
   State s(s_);
   State2 s2(s,s);
   FieldState omega=vac();
   for (Z i=1;i<=s_;++i){
      FieldState fsi=omega.crt(i);
      for (Z j=1;j<=s_;++j){
         FieldState fsij=fsi.crt(j);
         C cij=(fsij|*this);
         s2.put_(i,j,cij);
      }
   }
   s2.mark(f,ipolMeth,gamma,secFac);
}

R FieldState::test(Z com)
{
   Z mL=1;
   Word loc("FieldState::test(");
   loc&=cpm(com);
   loc&=")";
   CPM_MA
   R tol=1e-12;
   Z d=com;
   State s0(d);
   Z i;
   Z lim=1;
   R errTot=0;
   R tStart=cpmtime();
back:
   cpmmessage(loc&": doing test for lim="&cpm(lim));
   State s1=s0.ran();
   V<State> v2(2);
   for (i=v2.b();i<=v2.e();++i) v2[i]=s0.ran();
   V<State> v3(3);
   for (i=v3.b();i<=v3.e();++i) v3[i]=s0.ran();
   FieldState fs1(s1,lim);
   FieldState fs2(v2,lim);
   FieldState fs3(v3,lim);
   V<R> c(6);
   for (i=c.b();i<=c.e();++i) c[i]=CpmRoot::randomR();

   FieldState sr=fs1*c[1]+fs2*c[2]+fs3*c[3];
   FieldState sl=fs1*c[4]+fs2*c[5]+fs3*c[6];
   R nr=sr.abs();
   R nl=sl.abs();
   cpmmessage(loc&": nr="&cpm(nr)&", nl="&cpm(nl));
   R res=0;
   R val=0;
   for (i=1;i<=d;++i){
      //cpmcerr<<" loop starts with i="<<i<<endl;
      FieldState pr=sr.des(i);
      FieldState pl=sl.crt(i);
      C c1=(sl|pr);
      C c2=(pl|sr);
      val+=c1.abs();
      res+=(c1-c2).abs();
   }
   errTot+=res;
// Now testing the commutation relations according to the statement
// For R eps=Ky::getEps();
// and State s1,s2; FieldState fs; s1.dim()==s2.dim()==fs.size();
// fs.crt(s2).des(s1)+fs.des(s1).crt(s2)*eps-fs*(s1|s2)==0
   R eps=Ky::getEps();
   State st1=v2[1]*0.4+v2[2]*0.6;
   State st2=v3[1]*0.5+v3[3]*0.5;
   st1.nor_();
   st2.nor_();
   FieldState fl=sl.crt(st2).des(st1)+sl.des(st1).crt(st2)*eps;
   FieldState fr=sl*(st1|st2);
   R diffRel=(fl-fr).abs()/sl.abs();
   cpmcerr<<"Error in commutation relation is "<<diffRel<<endl;
   errTot+=diffRel;
   cpmmessage(loc&": error sum was "&cpm(res)&
      ", value sum was "&cpm(val));
   if (lim==1){
      lim+=3;
      goto back; // same for bosons
   }
   cpmmessage(loc&": total error="&cpm(errTot));
   if (errTot>tol) cpmwarning("error too large,  something is wrong");
   else cpmmessage("no problem found");
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   CPM_MZ
   return errTot;
}

R FieldState::graphicsTest(Z d)
{
   Z mL=1;
   Word loc("FieldState::graphicsTest(");
   loc&=cpm(d);
   loc&=")";
   CPM_MA
   R tStart=cpmtime();
   State s0(d);
   Z smooth=4;
   State s1=s0.ran(1);
   s1.smt_(smooth);
   State s2=s0.ran(2);
   s2.smt_(smooth);
   V<State> v2{s1,s2};
   Z lim=1;
   R tWait=8;
   Graph gr;
   s1.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   s2.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   gr.clr_();
back:
   FieldState fs(v2,lim);
   fs.mark(gr,1,0.5);
   gr.vis();
   cpmwait(tWait*2,2);
   if (lim==1){
      lim=4;
      goto back; // same for bosons
   }
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   CPM_MZ
   return tTot;
}
