//? cpmfieldstatedirac1.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmfieldstatedirac1.h>
#include <cpmimagingtools.h>
#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmx.h>
#include <cpmrvector.h>
#include <cpmfieldoperator.h>
#include <cpmstate2.h> // for using a graphics function from there

using namespace CpmQuantumMechanics;
using namespace CpmRoot;
using namespace CpmArrays;

/************************* class KyD *********************************/

Z KyD1::lim_=1;
   // the default value should come into action only for unmanagable
   // high 'occupation numbers' such as 1000

Z KyD1::eps_=1;

// not relying verbally on the implementation in Vo<Z>
// for utmost efficiency

Z KyD1::com(const KyD1& s)const
{
   return v_.com(s.v_);
}

bool KyD1::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(v_);
   return true;
}

bool KyD1::scanFrom(istream& str)
{
   cpms(v_);
   ini_();
   return true;
}

Z KyD1::sup()const
{
   if (d_<1) return -1;
   Z res=v_.cui(1);
   for (Z i=2;i<=d_;i++){
      Z vi=v_.cui(i);
      if (vi>res) res=vi;
   }
   return res;
}

Z KyD1::inf()const
{
   if (d_<1) return -1;
   Z res=v_.cui(1);
   for (Z i=2;i<=d_;i++){
      Z vi=v_.cui(i);
      if (vi<res) res=vi;
   }
   return res;
}

void KyD1::sup_(KyD1 const& k)
{
   if (d_!=k.d_) cpmerror("dimension mismatch in KyD1::sup_(KyD1)");
   for (Z i=1;i<=d_;++i){
      Z n1=v_.cui(i);
      Z n2=k.v_.cui(i);
      v_.cui(i)=CpmRootX::sup<Z>(n1,n2);
   }
}

bool KyD1::up_(Z i)
{
   Z n=v_.cui(i);
   if (n<lim_){
      v_.cui(i)+=1;
      return true;
   }
   else return false;
}

bool KyD1::down_(Z i)
{
   Z n=v_.cui(i);
   if (n>0){
      v_.cui(i)-=1;
      return true;
   }
   else{
      return false;
   }
}

R KyD1::sigma(Z i)const
{
   Z sum=0;
   for (Z k=1;k<=i;++k) sum+=v_.cui(k);
   return CpmRootX::parity(sum);
}

R KyD1::up(Z i)const
{
   Z ni=v_.cui(i);
   if (eps_==-1)
      return cpmsqrt(1.+ni);
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
      return sigma(i)*(1-ni);
   }
}

R KyD1::down(Z i)const
{
   Z ni=v_.cui(i);
   if (eps_==-1){
      return cpmsqrt(R(ni));
   }
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
      return sigma(i-1)*ni;
   }
}


/******************* class FieldStateD1 *********************************/

bool FieldStateD1::prnOn(ostream& str)const
{
   if (!CpmRoot::writeTitle("FieldStateD1",str)) return false;
   cpmp(b_);
   cpmp(lim_);
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      cpmp(i->first);
      cpmp(i->second);
   }
   cpmp(KyD1(0));
   C2 zClosing;
   cpmp(zClosing);
#else
   cpmp(m_);
#endif
   return true;
}

bool FieldStateD1::scanFrom(istream& str)
{
   Z mL=3;
   static Word loc="FieldStateD1::scanFrom(istream)";
   CPM_MA
   cpms(b_);
   cpms(lim_);
#if defined(CPM_M)
   cpms(m_);
   CPM_MZ
   return true;
#else
   m_=Mkc();
 //  bool active=true;
   bool res=true;
   KyD1 k;
   C val;
   while (true){
      bool succ=k.scanFrom(str);
      if (!succ){
         cpmwarning("can't read a KyD1!");
         res=false;
         break;
      }
      succ=val.scanFrom(str);
      if (!succ){
         cpmwarning("can't read a C!");
         res=false;
         break;
      }
      if (k.d()==0){
         cpmmessage(mL,loc&": stopping trigger found");
         break;
         // then the function should return true since
         // one is open to read the next FieldStateD from file
      }
      m_[k]=val;
         // only done if k and val were found as valid
         // and k was not a stopping trigger
   }
   CPM_MZ
   return res;
#endif
}

FieldStateD1::FieldStateD1(Bio const& b, Z kind, Z lim):b_(b),
   lim_(lim),m_()
{
   KyD1::setLim(lim_);
   if (kind==1) m_[KyD1(b_.d())]=C(1.); //i.e. the only occupied state is
      // that with all occupation numbers 0.
}

FieldStateD1::FieldStateD1(StateD1 const& st, Z lim):b_(st.bio()),
   lim_(lim)
{
   KyD1::setLim(lim_);
   Z d=b_.d();
   Z d2=d*2;
   KyD1 k0(d);
   for (Z i=1;i<=d2;++i){
      KyD1 k(k0);
      Z2 ij=k.toZ2(i);
      k.up_(i);
      m_[k]=st[ij.x1][ij.x2];
   }
}

FieldStateD1::FieldStateD1(V<StateD1> const& vs, Z lim):lim_(lim)
{
   Z mL=3;
   Word loc("FieldStateD1::FieldStateD1(V<StateD> const& vs)");
   CPM_MA
   Ky::setLim(lim_);
   Bio b=(vs.dim()>0 ? vs.fir().bio() : Bio());
   Z kind=1;
   FieldStateD1 fs(b,kind,lim_); // vacuum state
   for ( Z i=vs.b();i<=vs.e();++i) fs=fs.crt(vs[i]);
   fs.nor_();
   b_=fs.b_;
   m_=fs.m_;
   CPM_MZ
}

C FieldStateD1::get(KyD1 const& k)const
// notice that in the case of  std::map<Ky,C> the
// code {return m_[k];} would not be allowed
// for a const function get.
{
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   i=m_.find(k);
   return i!=m_.end() ? i->second : C();
#else
   return m_[k];
#endif
}

// notice the beautyful symmetry and structure in the
// implementation of crt and des.

/***************
FieldStateD FieldStateD::crt(Z j0)const
{
   Z j=j0;
   b_1(j);
   Ky::setLim(lim_);
   Mkc res;
#if !defined(CPM_M)
   std::map<KyD,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      Ky kj(k);
      if (kj.up_(j)) res[kj]=(i->second)*k.up(j);
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      Ky k=m_.x(i);
      Ky kj(k);
      if (kj.up_(j)) res[kj]=m_.y(i)*k.up(j);
   }
#endif
   return mod(res);
}
**********************/

FieldStateD1 FieldStateD1::crt(StateD1 const& st)const
{
   KyD1::setLim(lim_);
   Mkc res;
   Z d=st.dim();
   cpmassert(d==b_.d(),"FieldStateD1::crt(StateD)");
   Z d2=d*2;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      for (Z j=1;j<=d2;++j){
         KyD1 kj(k);
         Z2 j2=k.toZ2(j);
         if (kj.up_(j)) res[kj]+=(i->second)*k.up(j)*st[j2.x1][j2.x2];
            // notice that all indexes of st are defined according to bc_
      }
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      for (Z j=1;j<=d2;++j){
         KyD1 kj(k);
         if (kj.up_(j)){
             Z2 j2=k.toZ2(j);
             res[kj]+=m_.y(i)*k.up(j)*st[j2.x1][j2.x2];
            // notice that all indexes of st are defined according to bc_
         }
      }
   }
#endif
   return mod(res);
}

/********* less efficient, but shows the logic nicely
FieldStateD FieldStateD::crt(State const& st)const
{
   FieldStateD res(s_);
   for (Z i=0;i<s_;++i) res+=crt(i)*st(i);
   return res;
}
*/

/***************
FieldStateD FieldStateD::des(Z j0)const
{
   Z j=j0;
   b_1(j);
   Ky::setLim(lim_);
   Mkc res;
#if !defined(CPM_M)
   std::map<KyD,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      Ky kj(k);
      if (kj.down_(j)) res[kj]=(i->second)*k.down(j);
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      Ky k=m_.x(i);
      Ky kj(k);
      if (kj.down_(j)) res[kj]=m_.y(i)*k.down(j);
   }
#endif
   return mod(res);
}
********************/

/***************** less efficient, but shows the logic nicely
FieldStateD FieldStateD::des(State const& st)const
{
   FieldStateD res(s_);
   for (Z i=0;i<s_;++i) res+=des(i)*st(i).con();
   return res;
}
*/

FieldStateD1 FieldStateD1::des(StateD1 const& st)const
{
   KyD1::setLim(lim_);
   Mkc res;
   Z d=st.dim();
   cpmassert(d==b_.d(),"FieldStateD1::des(StateD)");
   Z d2=d*2;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      for (Z j=1;j<=d2;++j){
         KyD1 kj(k);
         if (kj.down_(j)){
            Z2 j2=k.toZ2(j);
            res[kj]+=(i->second)*k.down(j)*(st[j2.x1][j2.x2].con());
         }
      }
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      for (Z j=1;j<=d2;++j){
         KyD1 kj(k);
         Z2 j2=k.toZ2(j);
         if (kj.down_(j)) res[kj]+=m_.y(i)*k.down(j)*(st[j2.x1][j2.x2].con\
         ());
      }
   }
#endif
   return mod(res);
}

/************************
FieldStateD FieldStateD::cook1(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldStateD::cook1(Observable)");
   CPM_MA
   FieldStateD res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
        res+=des(i).crt(j)*ob(i,j);
   CPM_MZ
   return res;
}


FieldStateD FieldStateD::cook2(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldStateD::cook2(Observable)");
   CPM_MA
   FieldStateD res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
       res+=des(j).des(i).crt(i).crt(j)*(ob(i,j)*0.5);
   CPM_MZ
   return res;
}

FieldStateD FieldStateD::manyParHam(R2_Func const& pot1,
   R2_Func const& pot2, R t)const
{
   Z mL=3;
   Word loc("FieldStateD::manParHam(R2_Func,R2_Func,R,bool)");
   CPM_MA
   static R hD=QMBase::get_hDirac();
   static R hD2=hD*hD;
   Z i,j,d=s_;
   R dxNorm=1./d;
   R dxInv2=R(d)/QMBase::getSize(); // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(QMBase::getMass()*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   FieldStateD res=zero();
   V<R> x0=posNor(d);
   for (i=1;i<=d;++i){
      FieldStateD fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.)*fac;
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   for (i=1;i<=d;++i){
      R xi=x0[i];
      C xit(xi,t);
      R pxit=pot1(xit);
      if (pxit!=0) res+=des(i).crt(i)*pxit; // efficiency !
      for (j=1;j<=d;++j){
         R xij=xi-x0[j];
         if (xij>0.5)  xij-=1;
         else if (xij<-0.5) xij+=1;
         else ;
         C xijt(xij,t);
         R pxijt=pot2(xijt)*0.5;
         if (pxijt!=0) res+=des(j).des(i).crt(i).crt(j)*pxijt;
            // efficiency !
      }
   }
   CPM_MZ
   return res;
}
**********************/

FieldStateD1 FieldStateD1::aux(CC2 const& mat1, CC2 const& mat2,
   R const& potj, Z j)const
{
   FieldStateD1 res=zero();
   for (Z a=1;a<=2;++a){
      C2 m1a=mat1[a];
      C2 m2a=mat2[a];
      for (Z b=1;b<=2;++b){
         C m1ab=m1a[b];
         if (a==b) m1ab+=potj;
         C m2ab=m2a[b];
         res+=des(j,b).crt(j,a)*m1ab;
         res-=des(j,b).crt(j+1,a)*m2ab;
         res+=des(j,b).crt(j-1,a)*m2ab;
      }
   }
   return res;

}

FieldStateD1 FieldStateD1::ham(R const& m, R_Vector const& pot)const
{
   Z mL=3;
   Word loc("FieldStateD1::ham(R,R_Vector)");
   CPM_MA
   C fac=C(0.,-0.5)/b_.gap(); // 1/(2 i delta)
   CC2 beta_m=CC2(C(m),C(0.),C(0.),-C(m));
   CC2 alpha1_fac=CC2(C(0.),C(fac),C(fac),C(0));
   Z d=b_.d();
   FieldStateD1 res=zero();
   for (Z j=1;j<=d;++j){
      R pj=(d==pot.dim() ? pot.cui(j) : R(0));
      res+=aux(beta_m,alpha1_fac,pj,j);
   }
   CPM_MZ
   return res;
}

/************************
FieldStateD FieldStateD::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldStateD::hub(R,R,R)");
   CPM_MA
   FieldStateD res(s_,0,lim_);
   FieldStateD psi=*this;
   for (Z i=0;i<s_;++i){
      FieldStateD fi=des(i);
      res+=des(i-1).crt(i);
      res+=fi.crt(i+1);
      res+=des(i+1).crt(i);
      res+=fi.crt(i-1);
      res-=fi.crt(i)*4;
   }
   CPM_MZ
   return res;
}
*/
/******************** works for fermions and bosons why? *****
FieldStateD FieldStateD::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldStateD::hub(R,R,R)");
   CPM_MA
   FieldStateD res(s_,0,lim_);
   Z i;
   for (i=0;i<s_;++i){
      FieldStateD fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.);
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   CPM_MZ
   return res;
}
*/

/**********************
FieldStateD FieldStateD::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldStateD::hub(R,R,R)");
   CPM_MA
   t; // no role so far
   FieldStateD res=zero();
   for (Z i=1;i<=s_;++i){
      Z i1=i+1;
      FieldStateD f1i=des(i1).crt(i);
      FieldStateD f2i=des(i).crt(i1);
      res+=(f1i+f2i)*a;
      res+=des(i).des(i).crt(i).crt(i)*b;
   }
   CPM_MZ
   return res;
}
******************************/

FieldStateD1 FieldStateD1::reduce(Z d)const
{
   Mkc res;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      Z max=k.sup();
      if (max<=d) res[k]=i->second;
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      Z max=k.sup();
      if (max<=d) res[k]=m_.y(i);
   }
#endif
   return mod(res);
}

/***************
C FieldStateD::operator[](Z j)const
{
   C res;
#if !defined(CPM_M)
   std::map<KyD,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      if (k[j]>0) res+=i->second;
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      Ky k=m_.x(i);
      if (k[j]>0) res+=m_.y(i);
   }
#endif
   return res;
}
******************/

/*
State FieldStateD::toState(bool ignorePhases)const
{
   State res(s_);
   std::map<Ky,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      C val=i->second;
      for (Z j=0;j<s_;++j){
         if (ignorePhases){
            if (k[j]>0) res[j]+=val.absSqr();
         }
         else{
            if (k[j]>0) res[j]+=val;
         }
      }
   }
   return res;
}
*/
/************
StateD FieldStateD::toState(bool ignorePhases)const
{
   ignorePhases;
   return den();
}
******************/

/****************
StateD FieldStateD::den()const
{
   StateD res(b_);
   Z d=b_.d();
   for (Z i=1;i<=d;++i){
      FieldStateD fsi=des(i);
      res[i]=C(fsi.absSqr());
   }
   return res;
}
******************/
/*************
Observable FieldStateD::toObservable(bool ignorePhases)const
{
   FieldStateD vac(s_,1,lim_);
   Observable res(s_);
   for (Z i=1;i<=s_;++i){
      for (Z j=1;j<=s_;++j){
         FieldStateD sij=des(i).des(j);
         C z=(vac|sij);
         if (ignorePhases) z=C(z.abs());
         res.setLean(i,j,z);
      }
   }
   return res;
}
*****************/

/**************
X3<Ky,Z1,R1> FieldStateD::dgn()const
{
   Z1 res;
   Ky ks(s_);
#if !defined(CPM_M)
   std::map<Ky,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      ks.sup_(k);
      res+=1;
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      Ky k=m_.x(i);
      ks.sup_(k);
      res+=1;
   }
#endif
   R1 n(norm());
   return X3<Ky,Z1,R1>(ks,res,n);
}
************************/

C FieldStateD1::operator|(FieldStateD1 const& fs)const
// see Musser,Saini: STL Tutorial and Reference Guide,
// Addison-Wesley 1996, Example 7-6, p. 169,170
// for the pattern of this implementation.
{
   C res;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i,i2;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      i2=fs.m_.find(k);
      if (i2!=fs.m_.end()){
         C z=i->second;
         C z2=i2->second;
         res+=(z|z2);
      }
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      if (fs.m_.defined(k)){
         C z=m_.y(i);
         C z2=fs[k];
         res+=(z|z2);
      }
   }
#endif
   return res;
}

FieldStateD1 FieldStateD1::operator+(FieldStateD1 const& fs)const
{

   Mkc res(m_);
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=fs.m_.begin();i!=fs.m_.end();++i){
      KyD1 k=i->first;
      res[k]+=(i->second);
   }
#else
   for (Z i=fs.m_.b();i<=fs.m_.e();++i){
      KyD1 k=fs.m_.x(i);
      res[k]+=fs.m_.y(i);
   }
#endif
   return mod(res);
}

FieldStateD1 FieldStateD1::operator-(FieldStateD1 const& fs)const
{
   Mkc res(m_);
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=fs.m_.begin();i!=fs.m_.end();++i){
      KyD1 k=i->first;
      res[k]-=(i->second);
   }
#else
   for (Z i=fs.m_.b();i<=fs.m_.e();++i){
      KyD1 k=fs.m_.x(i);
      res[k]-=fs.m_.y(i);
   }
#endif
   return mod(res);
}

FieldStateD1 FieldStateD1::operator*(C z)const
{
   Mkc res(m_);
#if !defined(CPM_M)
   std::map<KyD1,C>::iterator i;
   for (i=res.begin();i!=res.end();++i) (i->second)*=z;
#else
   for (Z i=res.b();i<=res.e();++i) res[m_.x(i)]*=z;
#endif
   return mod(res);
}

FieldStateD1 FieldStateD1::operator*(R r)const
{
   Z mL=3;
   Word loc("FieldStateD1::operator*(R r="&cpm(r)&")");
   CPM_MA
   Mkc res(m_);
#if !defined(CPM_M)
   std::map<KyD1,C>::iterator i;
   for (i=res.begin();i!=res.end();++i) (i->second)*=r;
#else
   for (Z i=res.b();i<=res.e();++i) res[m_.x(i)]*=r;
#endif
   CPM_MZ
   return mod(res);
}

FieldStateD1& FieldStateD1::operator+=(FieldStateD1 const& fs)
{
   return *this=operator+(fs);
}

FieldStateD1& FieldStateD1::operator-=(FieldStateD1 const& fs)
{
   return *this=operator-(fs);
}

FieldStateD1& FieldStateD1::operator*=(C z)
{
   return *this=operator*(z);
}

FieldStateD1& FieldStateD1::operator*=(R r)
{
   return *this=operator*(r);
}

R FieldStateD1::sup(void)const
{
   R res=0;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      R val=(i->second).abs();
      if (val>res) res=val;
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      R val=m_.y(i).abs();
      if (val>res) res=val;
   }
#endif
   return res;
}

FieldStateD1 FieldStateD1::con(void)const
{
   Mkc res;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      res[k]=(i->second).con();
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      res[k]=m_.y(i).con();
   }
#endif
   return mod(res);
}

FieldStateD1 FieldStateD1::operator-(void)const
{
   Mkc res;
#if !defined(CPM_M)
   std::map<KyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KyD1 k=i->first;
      res[k]=-(i->second);
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KyD1 k=m_.x(i);
      res[k]=-m_.y(i);
   }
#endif
   return mod(res);
}

R FieldStateD1::nor_(R r)
{
   Z d=dim();
   if (d==0) return 0.;

   R r0=norm();
   if (r0==0){
      FieldStateD1 aux=vac();
         // ground state
      aux*=r;
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

/*************
void FieldStateD::show(Frame& f, Z mult_i, Z mult_j,
      Z iShift, Z jShift, bool ignorePhases)const
{
   toState(ignorePhases).show(f,mult_i,mult_j,iShift,jShift,ignorePhases);
}
*****************/

/*****************
void FieldStateD::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
{
   V<FieldStateD> vf(s_);
   Z i,j;
   for (i=1;i<=s_;++i){
      vf[i]=des(i);
   }
   State s(s_);
   State2 s2(s,s);
   for (i=1;i<=s_;++i){
      for (j=i;j<=s_;++j){
         C cij=(vf[i]|vf[j]);
         s2.put_(i,j,cij);
         if (i!=j) s2.put_(j,i,cij.con());
      }
   }
   s2.mark(f,ipolMeth,gamma,secFac);
}
****************/

/*******************
void FieldStateD::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
{
   State s(s_);
   State2 s2(s,s);
   FieldStateD omega=vac();
   for (Z i=1;i<=s_;++i){
      FieldStateD fsi=omega.crt(i);
      for (Z j=1;j<=s_;++j){
         FieldStateD fsij=fsi.crt(j);
         C cij=(fsij|*this);
         s2.put_(i,j,cij);
      }
   }
   s2.mark(f,ipolMeth,gamma,secFac);
}
************************/
/*****************
R FieldStateD::test(Z com)
{
   Z mL=1;
   Word loc("FieldStateD::test(");
   loc&=cpm(com);
   loc&=")";
   CPM_MA
   R tol=1e-12;
   Z d=com;
   State s0(d);
   Z i;
   Z lim=1;
   R errTot=0;
   R tStart=cpmtime();
back:
   cpmmessage(loc&": doing test for lim="&cpm(lim));
   State s1=s0.ran();
   V<State> v2(2);
   for (i=v2.b();i<=v2.e();++i) v2[i]=s0.ran();
   V<State> v3(3);
   for (i=v3.b();i<=v3.e();++i) v3[i]=s0.ran();
   FieldStateD fs1(s1,lim);
   FieldStateD fs2(v2,lim);
   FieldStateD fs3(v3,lim);
   V<R> c(6);
   for (i=c.b();i<=c.e();++i) c[i]=CpmRoot::randomR();

   FieldStateD sr=fs1*c[1]+fs2*c[2]+fs3*c[3];
   FieldStateD sl=fs1*c[4]+fs2*c[5]+fs3*c[6];
   R nr=sr.abs();
   R nl=sl.abs();
   cpmmessage(loc&": nr="&cpm(nr)&", nl="&cpm(nl));
   R res=0;
   R val=0;
   for (i=1;i<=d;++i){
      //cpmcerr<<" loop starts with i="<<i<<endl;
      FieldStateD pr=sr.des(i);
      FieldStateD pl=sl.crt(i);
      C c1=(sl|pr);
      C c2=(pl|sr);
      val+=c1.abs();
      res+=(c1-c2).abs();
   }
   errTot+=res;
// Now testing the commutation relations according to the statement
// For R eps=Ky::getEps();
// and State s1,s2; FieldStateD fs; s1.dim()==s2.dim()==fs.size();
// fs.crt(s2).des(s1)+fs.des(s1).crt(s2)*eps-fs*(s1|s2)==0
   R eps=Ky::getEps();
   State st1=v2[1]*0.4+v2[2]*0.6;
   State st2=v3[1]*0.5+v3[3]*0.5;
   st1.nor_();
   st2.nor_();
   FieldStateD fl=sl.crt(st2).des(st1)+sl.des(st1).crt(st2)*eps;
   FieldStateD fr=sl*(st1|st2);
   R diffRel=(fl-fr).abs()/sl.abs();
   cpmcerr<<"Error in commutation relation is "<<diffRel<<endl;
   errTot+=diffRel;
   cpmmessage(loc&": error sum was "&cpm(res)&
      ", value sum was "&cpm(val));
   if (lim==1){
      lim+=3;
      goto back; // same for bosons
   }
   cpmmessage(loc&": total error="&cpm(errTot));
   if (errTot>tol) cpmwarning("error too large,  something is wrong");
   else cpmmessage("no problem found");
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   Word impl="Implementation based on ";
#if defined(CPM_M)
   impl&="CpmArrays::M<>";
#else
   impl&="std::map<>";
#endif
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   cpmcerr<<impl;
   CPM_MZ
   return errTot;
}


R FieldStateD::graphicsTest(Z d)
{
   Z mL=1;
   Word loc("FieldStateD::graphicsTest(");
   loc&=cpm(d);
   loc&=")";
   CPM_MA
   R tStart=cpmtime();
   State s0(d);
   Z smooth=4;
   State s1=s0.ran(1);
   s1.smt_(smooth);
   State s2=s0.ran(2);
   s2.smt_(smooth);
   V<State> v2("",s1,s2);
   Z lim=1;
   R tWait=8;
   Graph gr;
   s1.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   s2.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   gr.clr_();
back:
   FieldStateD fs(v2,lim);
   fs.mark(gr,1,0.5);
   gr.vis();
   cpmwait(tWait*2,2);
   if (lim==1){
      lim=4;
      goto back; // same for bosons
   }
   Word impl="Implementation based on ";
#if defined(CPM_M)
   impl&="CpmArrays::M<>";
#else
   impl&="std::map<>";
#endif
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   cpmcerr<<impl;
   CPM_MZ
   return tTot;
}

*************************/

namespace{

FieldStateD1 crtFunc(FieldStateD1 const& fs, StateD1 const& st)
{
    return fs.crt(st);
}

FieldStateD1 desFunc(FieldStateD1 const& fs, StateD1 const& st)
{
    return fs.des(st);
}

FieldStateD1 hamFunc(FieldStateD1 const& fs, R const& m, R_Vector const&\
 pot)
{
    return fs.ham(m,pot);
}

}

FieldStateD1 opFunc(FieldStateD1 const& st)
{
   return st.zero();
}

OpFD1::OpFD1(Bio const& b):b_(b),
f_(F<FieldStateD1,FieldStateD1>(opFunc)){}

OpFD1 OpFD1::crt(StateD1 const& st)
{
   return OpFD1(st.bio(),F1<FieldStateD1,StateD1,FieldStateD1>(st)(crtFunc\
   ));
}

OpFD1 OpFD1::crt(Z i, Z a, Bio const& b)
{
   return OpFD1::crt(StateD1(i,a,b));
}

OpFD1 OpFD1::des(StateD1 const& st)
{
   return OpFD1(st.bio(),F1<FieldStateD1,StateD1,FieldStateD1>(st)(desFunc\
   ));
}

OpFD1 OpFD1::des(Z i, Z a, Bio const& b)
{
   return OpFD1::des(StateD1(i,a,b));
}

OpFD1 OpFD1::ham(Bio const& b, R const& m, R_Vector const& pot)
{
   return OpFD1(b,F2<FieldStateD1,R,R_Vector,FieldStateD1>(m,pot)(hamFunc)\
   );
}

OpFD1 OpFD1::cook(OpD1 const& op)
{
   Bio bi=op.bio();
   Z d=bi.d();
   OpFD1 res(bi);
   for (Z i=1;i<=d;++i){
      for (Z a=1;a<=2;++a){
         StateD1 sia(i,a,bi);
         StateD1 oia=op*sia;
         for (Z j=1;j<=d;++j){
            for (Z b=1;b<=2;++b){
               StateD1 sjb(j,b,bi);
               OpFD1 oijab=OpFD1::crt(sjb)*OpFD1::des(sia)*(sjb|oia);
               res=res+oijab;
            }
         }
      }
   }
   return res;
}

