//? cpmfieldsttd1.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmsystemdependencies.h>
#if defined(CPM_USE_EIGEN)

#include <cpmfieldsttd1.h>
#include <cpmimagingtools.h>
#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmx.h>
#include <cpmrvector.h>
#include <cpmfieldoperator.h>

using namespace CpmQuantumMechanics;
using namespace CpmRoot;
using namespace CpmArrays;

/************************* class KeyD1D *********************************/

Z KeyD1::lim_=1;
   // the default value should come into action only for unmanagable
   // high 'occupation numbers' such as 1000

Z KeyD1::eps_=1;

// not relying verbally on the implementation in Vo<Z>
// for utmost efficiency

Z KeyD1::sup()const
{
   if (d_<1) return -1;
   Z res=v_(0);
   for (Z i=1;i<d_;i++){
      Z vi=v_(i);
      if (vi>res) res=vi;
   }
   return res;
}

Z KeyD1::inf()const
{
   if (d_<1) return -1;
   Z res=v_(0);
   for (Z i=1;i<d_;i++){
      Z vi=v_(i);
      if (vi<res) res=vi;
   }
   return res;
}

void KeyD1::sup_(KeyD1 const& k)
{
   if (d_!=k.d_) cpmerror("dimension mismatch in KeyD1::sup_(KeyD1)");
   for (Z i=0;i<d_;++i){
      Z n1=v_(i);
      Z n2=k.v_(i);
      v_(i)=CpmRootX::sup<Z>(n1,n2);
   }
}

bool KeyD1::up_(Z i)
{
   Z ni=v_(i);
   if (ni<lim_){
      v_(i)+=1;
      return true;
   }
   else return false;
}

bool KeyD1::down_(Z i)
{
   Z ni=v_(i);
   if (ni>0){
      v_(i)-=1;
      return true;
   }
   else{
      return false;
   }
}

R KeyD1::sigma(Z i)const
{
   Z sum=0;
   for (Z k=0;k<i;++k) sum+=v_(k);
   return CpmRootX::parity(sum);
}

R KeyD1::up(Z i)const
{
   Z ni=v_(i);
   if (eps_==-1)
      return cpmsqrt(1.+ni);
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
      return sigma(i)*(1-ni);
   }
}

R KeyD1::down(Z i)const
{
   Z ni=v_(i);
   if (eps_==-1){
      return cpmsqrt(R(ni));
   }
   else{
      if (ni>1) cpmerror("ni>1: not a fermion state");
     // return sigma(i-1)*ni;
     return sigma(i)*ni; // this is the right factor
   }
}

bool KeyD1::operator<(KeyD1 const& k)const
{
   for (Z i=0; i<d_; ++i){
      if (v_(i)<k.v_(i)) return true;
      else if (v_(i)>k.v_(i)) return false;
      else;
   }
   return false;
}


/******************* class FieldSttD1 *********************************/


FieldSttD1::FieldSttD1(Bio const& b, Z kind, Z lim):b_(b),
   lim_(lim),m_()
{
   KeyD1::setLim(lim_);
   if (kind==1) m_[KeyD1(b_.dd())]=C(1.); //i.e. the only occupied state is
      // that with all occupation numbers 0.
}

FieldSttD1::FieldSttD1(SttD1 const& st, Z lim):b_(st.bio()),
   lim_(lim)
{
   KeyD1::setLim(lim_);
 //  Z d=b_.d();
   Z d2=b_.dd();
   KeyD1 k0(d2);
   for (Z i=0;i<d2;++i){
      KeyD1 k(k0);
      k.up_(i);
      m_[k]=st(i);
   }
}

FieldSttD1::FieldSttD1(V<SttD1> const& vs, Z lim):lim_(lim)
{
   Z mL=3;
   Word loc("FieldSttD1::FieldSttD1(V<SttD1D> const& vs)");
   CPM_MA
   KeyD1::setLim(lim_);
   Bio b=(vs.dim()>0 ? vs.fir().bio() : Bio());
   Z kind=1;
   FieldSttD1 fs(b,kind,lim_); // vacuum state
   for ( Z i=vs.b();i<=vs.e();++i) fs=fs.crt(vs[i]);
   fs.nor_();
   b_=fs.b_;
   m_=fs.m_; // the main thing
   CPM_MZ
}

C FieldSttD1::get(KeyD1 const& k)const
// notice that in the case of  std::map<KeyD1,C> the
// code {return m_[k];} would not be allowed
// for a const function get.
{
   std::map<KeyD1,C>::const_iterator i;
   i=m_.find(k);
   return i!=m_.end() ? i->second : C();
}

// notice the beautyful symmetry and structure in the
// implementation of crt and des.

FieldSttD1 FieldSttD1::crt(SttD1 const& st)const
{
   KeyD1::setLim(lim_);
   Mkc res;
   Z d=st.dim();
   cpmassert(d==dim1(),"FieldSttD1::crt(SttD1)");
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      for (Z j=0;j<d;++j){
         KeyD1 kj(k);
         if (kj.up_(j)) res[kj]+=(i->second)*k.up(j)*st(j);
      }
   }
 //  return mod(res);
   return FieldSttD1(b_,res,lim_);
}

FieldSttD1 FieldSttD1::des(SttD1 const& st)const
{
   KeyD1::setLim(lim_);
   Mkc res;
   Z d=st.dim();
   cpmassert(d==dim1(),"FieldSttD1::des(SttD1)");
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      for (Z j=0;j<d;++j){
         KeyD1 kj(k);
         if (kj.down_(j)){
            res[kj]+=(i->second)*k.down(j)*(st(j).con());
         }
      }
   }
 //  return mod(res);
   return FieldSttD1(b_,res,lim_);
}

/************************
FieldSttD1 FieldSttD1::cook1(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldSttD1::cook1(Observable)");
   CPM_MA
   FieldSttD1 res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
        res+=des(i).crt(j)*ob(i,j);
   CPM_MZ
   return res;
}


FieldSttD1 FieldSttD1::cook2(Observable const& ob)const
{
   Z mL=3;
   Word loc("FieldSttD1::cook2(Observable)");
   CPM_MA
   FieldSttD1 res=zero();
   Z i,j;
   for (i=1;i<=s_;++i)
     for (j=1;j<=s_;++j)
       res+=des(j).des(i).crt(i).crt(j)*(ob(i,j)*0.5);
   CPM_MZ
   return res;
}

FieldSttD1 FieldSttD1::manyParHam(R2_Func const& pot1,
   R2_Func const& pot2, R t)const
{
   Z mL=3;
   Word loc("FieldSttD1::manParHam(R2_Func,R2_Func,R,bool)");
   CPM_MA
   static R hD=QMBase::get_hDirac();
   static R hD2=hD*hD;
   Z i,j,d=s_;
   R dxNorm=1./d;
   R dxInv2=R(d)/QMBase::getSize(); // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(QMBase::getMass()*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   FieldSttD1 res=zero();
   V<R> x0=posNor(d);
   for (i=1;i<=d;++i){
      FieldSttD1 fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.)*fac;
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   for (i=1;i<=d;++i){
      R xi=x0[i];
      C xit(xi,t);
      R pxit=pot1(xit);
      if (pxit!=0) res+=des(i).crt(i)*pxit; // efficiency !
      for (j=1;j<=d;++j){
         R xij=xi-x0[j];
         if (xij>0.5)  xij-=1;
         else if (xij<-0.5) xij+=1;
         else ;
         C xijt(xij,t);
         R pxijt=pot2(xijt)*0.5;
         if (pxijt!=0) res+=des(j).des(i).crt(i).crt(j)*pxijt;
            // efficiency !
      }
   }
   CPM_MZ
   return res;
}
**********************/

FieldSttD1 FieldSttD1::aux(CC2 const& mat1, CC2 const& mat2,
   R const& potj, Z j)const
{
   FieldSttD1 res=zero();
   for (Z a=1;a<=2;++a){
      C2 m1a=mat1[a];
      C2 m2a=mat2[a];
      for (Z b=1;b<=2;++b){
         C m1ab=m1a[b];
         if (a==b) m1ab+=potj;
         C m2ab=m2a[b];
         res+=des(j,b).crt(j,a)*m1ab;
         res-=des(j,b).crt(j+1,a)*m2ab;
         res+=des(j,b).crt(j-1,a)*m2ab;
      }
   }
   return res;

}

FieldSttD1 FieldSttD1::ham(R const& m, R_Vector const& pot)const
{
   Z mL=3;
   Word loc("FieldSttD1::ham(R,R_Vector)");
   CPM_MA
   C fac=C(0.,-0.5)/b_.gap(); // 1/(2 i delta)
   CC2 beta_m=CC2(C(m),C(0.),C(0.),-C(m));
   CC2 alpha1_fac=CC2(C(0.),C(fac),C(fac),C(0));
   Z d=b_.d();
   FieldSttD1 res=zero();
   for (Z j=1;j<=d;++j){
      R pj=(d==pot.dim() ? pot.cui(j) : R(0));
      res+=aux(beta_m,alpha1_fac,pj,j);
   }
   CPM_MZ
   return res;
}

/************************
FieldSttD1 FieldSttD1::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldSttD1::hub(R,R,R)");
   CPM_MA
   FieldSttD1 res(s_,0,lim_);
   FieldSttD1 psi=*this;
   for (Z i=0;i<s_;++i){
      FieldSttD1 fi=des(i);
      res+=des(i-1).crt(i);
      res+=fi.crt(i+1);
      res+=des(i+1).crt(i);
      res+=fi.crt(i-1);
      res-=fi.crt(i)*4;
   }
   CPM_MZ
   return res;
}
*/
/******************** works for fermions and bosons why? *****
FieldSttD1 FieldSttD1::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldSttD1::hub(R,R,R)");
   CPM_MA
   FieldSttD1 res(s_,0,lim_);
   Z i;
   for (i=0;i<s_;++i){
      FieldSttD1 fsi=des(i);
      fsi=(fsi.crt(i-1)+fsi.crt(i+1)-fsi.crt(i)*2.);
      res+=fsi;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   CPM_MZ
   return res;
}
*/

/**********************
FieldSttD1 FieldSttD1::hub(R a, R b, R t)const
{
   Z mL=3;
   Word loc("FieldSttD1::hub(R,R,R)");
   CPM_MA
   t; // no role so far
   FieldSttD1 res=zero();
   for (Z i=1;i<=s_;++i){
      Z i1=i+1;
      FieldSttD1 f1i=des(i1).crt(i);
      FieldSttD1 f2i=des(i).crt(i1);
      res+=(f1i+f2i)*a;
      res+=des(i).des(i).crt(i).crt(i)*b;
   }
   CPM_MZ
   return res;
}
******************************/

FieldSttD1 FieldSttD1::reduce(Z d)const
{
   Mkc res;
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      Z max=k.sup();
      if (max<=d) res[k]=i->second;
   }
   return mod(res);
}

/***************
C FieldSttD1::operator[](Z j)const
{
   C res;
#if !defined(CPM_M)
   std::map<KeyD1D,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      if (k[j]>0) res+=i->second;
   }
#else
   for (Z i=m_.b();i<=m_.e();++i){
      KeyD1 k=m_.x(i);
      if (k[j]>0) res+=m_.y(i);
   }
#endif
   return res;
}
******************/

/*
SttD1 FieldSttD1::toSttD1(bool ignorePhases)const
{
   SttD1 res(s_);
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      C val=i->second;
      for (Z j=0;j<s_;++j){
         if (ignorePhases){
            if (k[j]>0) res[j]+=val.absSqr();
         }
         else{
            if (k[j]>0) res[j]+=val;
         }
      }
   }
   return res;
}
*/
/************
SttD1D FieldSttD1::toSttD1(bool ignorePhases)const
{
   ignorePhases;
   return den();
}
******************/


SttD1 FieldSttD1::den()const
{
   SttD1 res(b_);
   Z d=b_.d();
   for (Z a=1;a<=2;++a){
      for (Z i=1;i<=d;++i){
         FieldSttD1 fsi=des(i,a);
         res.setVal_(i,a,C(fsi.absSqr()));
      }
   }
   return res;
}


X3<KeyD1,Z,R> FieldSttD1::dgn()const
{
   Z res{0};
   Z d=b_.d();
   KeyD1 ks(2*d);
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      ks.sup_(k);
      res+=1;
   }
   R nr(norm());
   return X3<KeyD1,Z,R>(ks,res,nr);
}


C FieldSttD1::operator|(FieldSttD1 const& fs)const
// see Musser,Saini: STL Tutorial and Reference Guide,
// Addison-Wesley 1996, Example 7-6, p. 169,170
// for the pattern of this implementation.
{
   C res;
   std::map<KeyD1,C>::const_iterator i,i2;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      i2=fs.m_.find(k);
      if (i2!=fs.m_.end()){
         C z=i->second;
         C z2=i2->second;
         res+=(z|z2);
      }
   }
   return res;
}

FieldSttD1 FieldSttD1::operator+(FieldSttD1 const& fs)const
{

   Mkc res(m_);
   std::map<KeyD1,C>::const_iterator i;
   for (i=fs.m_.begin();i!=fs.m_.end();++i){
      KeyD1 k=i->first;
      res[k]+=(i->second);
   }
   return mod(res);
}

FieldSttD1 FieldSttD1::operator-(FieldSttD1 const& fs)const
{
   Mkc res(m_);
   std::map<KeyD1,C>::const_iterator i;
   for (i=fs.m_.begin();i!=fs.m_.end();++i){
      KeyD1 k=i->first;
      res[k]-=(i->second);
   }
   return mod(res);
}

FieldSttD1 FieldSttD1::operator*(C z)const
{
   Mkc res(m_);
   std::map<KeyD1,C>::iterator i;
   for (i=res.begin();i!=res.end();++i) (i->second)*=z;
   return mod(res);
}

FieldSttD1 FieldSttD1::operator*(R r)const
{
   Z mL=3;
   Word loc("FieldSttD1::operator*(R r="&cpm(r)&")");
   CPM_MA
   Mkc res(m_);
   std::map<KeyD1,C>::iterator i;
   for (i=res.begin();i!=res.end();++i) (i->second)*=r;
   CPM_MZ
   return mod(res);
}

FieldSttD1& FieldSttD1::operator+=(FieldSttD1 const& fs)
{
   return *this=operator+(fs);
}

FieldSttD1& FieldSttD1::operator-=(FieldSttD1 const& fs)
{
   return *this=operator-(fs);
}

FieldSttD1& FieldSttD1::operator*=(C z)
{
   return *this=operator*(z);
}

FieldSttD1& FieldSttD1::operator*=(R r)
{
   return *this=operator*(r);
}

R FieldSttD1::sup(void)const
{
   R res=0;
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      R val=(i->second).abs();
      if (val>res) res=val;
   }
   return res;
}

FieldSttD1 FieldSttD1::con(void)const
{
   Mkc res;
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      res[k]=(i->second).con();
   }
   return mod(res);
}

FieldSttD1 FieldSttD1::operator-(void)const
{
   Mkc res;
   std::map<KeyD1,C>::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      KeyD1 k=i->first;
      res[k]=-(i->second);
   }
   return mod(res);
}

R FieldSttD1::nor_(R r)
{
   Z d=dim1();
   if (d==0) return 0.;

   R r0=norm();
   if (r0==0){
      FieldSttD1 aux=vac();
         // ground state
      aux*=r;
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

/*************
void FieldSttD1::show(Frame& f, Z mult_i, Z mult_j,
      Z iShift, Z jShift, bool ignorePhases)const
{
   toSttD1(ignorePhases).show(f,mult_i,mult_j,iShift,jShift,ignorePhases);
}
*****************/

/*****************
void FieldSttD1::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
{
   V<FieldSttD1> vf(s_);
   Z i,j;
   for (i=1;i<=s_;++i){
      vf[i]=des(i);
   }
   SttD1 s(s_);
   SttD12 s2(s,s);
   for (i=1;i<=s_;++i){
      for (j=i;j<=s_;++j){
         C cij=(vf[i]|vf[j]);
         s2.put_(i,j,cij);
         if (i!=j) s2.put_(j,i,cij.con());
      }
   }
   s2.mark(f,ipolMeth,gamma,secFac);
}
****************/

/*******************
void FieldSttD1::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
{
   SttD1 s(s_);
   SttD12 s2(s,s);
   FieldSttD1 omega=vac();
   for (Z i=1;i<=s_;++i){
      FieldSttD1 fsi=omega.crt(i);
      for (Z j=1;j<=s_;++j){
         FieldSttD1 fsij=fsi.crt(j);
         C cij=(fsij|*this);
         s2.put_(i,j,cij);
      }
   }
   s2.mark(f,ipolMeth,gamma,secFac);
}
************************/

R FieldSttD1::test(Z com)
{
   Z mL=1;
   Word loc("FieldSttD1::test(");
   loc&=cpm(com);
   loc&=")";
   CPM_MA
   cout<<"FieldSttD1::test( "<<com<<" ) started"<<endl;
   R tol=1e-12;
   Z d=com;
   Bio b(d);
   SttD1 s0(b);
   Z i;
   Z lim=1;
   R errTot=0;
   R tStart=cpmtime();
   cpmmessage(loc&": doing test for lim="&cpm(lim));
   SttD1 s1=s0.ran();
   SttD1 s2=s0.ran();
   V<SttD1> v2(2);
   for (i=v2.b();i<=v2.e();++i) v2[i]=s0.ran();
   V<SttD1> v3(3);
   for (i=v3.b();i<=v3.e();++i) v3[i]=s0.ran();
   FieldSttD1 fs1(s1,lim); // a 1-particle state
   FieldSttD1 fs2(v2,lim); // a 2-particle state
   FieldSttD1 fs3(v3,lim); // a 3-particle state
   V<R> c(6);
   for (i=c.b();i<=c.e();++i) c[i]=CpmRoot::randomR();
   FieldSttD1 sr=fs1*c[1]+fs2*c[2]+fs3*c[3];
   FieldSttD1 sl=fs1*c[4]+fs2*c[5]+fs3*c[6];
      // two rather general states with particle content of 1,2,3 particles
   R nr=sr.abs();
   R nl=sl.abs();
   cpmmessage(loc&": nr="&cpm(nr)&", nl="&cpm(nl));
   cout<<"nr = "<<nr<<", nl = "<<nl<<endl;
   // now we test the hermitean conjugates

   R diff1 = ((sl|sr.crt(s2))-(sl.des(s2)|sr)).abs();
   R diff2 = ((sl|sr.des(s2))-(sl.crt(s2)|sr)).abs();

   cout<<" diff1 = "<<diff1<<endl;
   cout<<" diff2 = "<<diff2<<endl;
   errTot+=diff1;
   errTot+=diff2;
// Now testing the commutation relations according to the statement
// For R eps=KeyD1::getEps();
// and SttD1 s1,s2; FieldSttD1 fs; s1.dim()==s2.dim()==fs.size();
// fs.crt(s2).des(s1)+fs.des(s1).crt(s2)*eps-fs*(s1|s2)==0
   R eps=KeyD1::getEps();
   SttD1 st1=v2[1]*0.4+v2[2]*0.6;
   SttD1 st2=v3[1]*0.5+v3[3]*0.5;
   st1.nor_();
   st2.nor_();
   R resC=(sl.crt(st2).crt(st1)+sl.crt(st1).crt(st2)*eps).abs();
   R resD=(sl.des(st2).des(st1)+sl.des(st1).des(st2)*eps).abs();
   cout<<" resC = "<<resC<<endl;
   cout<<" resD = "<<resD<<endl;
   FieldSttD1 fl=sl.crt(st2).des(st1)+sl.des(st1).crt(st2)*eps;
   FieldSttD1 fr=sl*(st1|st2);
   R diffRel=(fl-fr).abs()/sl.abs();
   cpmcerr<<"Error in commutation relation is "<<diffRel<<endl;
   cout<<"Error in commutation relation is "<<diffRel<<endl;
   errTot+=resC;
   errTot+=resD;
   errTot+=diffRel;
   cpmmessage(loc&": total error="&cpm(errTot));
   if (errTot>tol) cpmwarning("error too large,  something is wrong");
   else cpmmessage("no problem found");
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   CPM_MZ
   cout<<" errTot = "<<errTot<<endl;
   return errTot;
}

/**********************************
R FieldSttD1::graphicsTest(Z d)
{
   Z mL=1;
   Word loc("FieldSttD1::graphicsTest(");
   loc&=cpm(d);
   loc&=")";
   CPM_MA
   R tStart=cpmtime();
   SttD1 s0(d);
   Z smooth=4;
   SttD1 s1=s0.ran(1);
   s1.smt_(smooth);
   SttD1 s2=s0.ran(2);
   s2.smt_(smooth);
   V<SttD1> v2("",s1,s2);
   Z lim=1;
   R tWait=8;
   Graph gr;
   s1.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   s2.mark(gr);
   gr.vis();
   cpmwait(tWait,2);
   gr.clr_();
back:
   FieldSttD1 fs(v2,lim);
   fs.mark(gr,1,0.5);
   gr.vis();
   cpmwait(tWait*2,2);
   if (lim==1){
      lim=4;
      goto back; // same for bosons
   }
   Word impl="Implementation based on ";
#if defined(CPM_M)
   impl&="CpmArrays::M<>";
#else
   impl&="std::map<>";
#endif
   R tEnd=cpmtime();
   R tTot=tEnd-tStart;
   cpmcerr<<"Total computation time was "<<tTot<<
      ", space dimension was "<<d<<endl;
   cpmcerr<<impl;
   CPM_MZ
   return tTot;
}

*************************/

namespace{

FieldSttD1 crtFunc(FieldSttD1 const& fs, SttD1 const& st)
{
    return fs.crt(st);
}

FieldSttD1 desFunc(FieldSttD1 const& fs, SttD1 const& st)
{
    return fs.des(st);
}

FieldSttD1 hamFunc(FieldSttD1 const& fs, R const& m, R_Vector const& pot)
{
    return fs.ham(m,pot);
}

}

FieldSttD1 opFunc(FieldSttD1 const& st)
{
   return st.zero();
}

OprFD1::OprFD1(Bio const& b):b_(b),
f_(F<FieldSttD1,FieldSttD1>(opFunc)){}

OprFD1 OprFD1::crt(SttD1 const& st)
{
   return OprFD1(st.bio(),F1<FieldSttD1,SttD1,FieldSttD1>(st)(crtFunc));
}

OprFD1 OprFD1::crt(Z i, Z a, Bio const& b)
{
   return OprFD1::crt(SttD1(i,a,b));
}

OprFD1 OprFD1::des(SttD1 const& st)
{
   return OprFD1(st.bio(),F1<FieldSttD1,SttD1,FieldSttD1>(st)(desFunc));
}

OprFD1 OprFD1::des(Z i, Z a, Bio const& b)
{
   return OprFD1::des(SttD1(i,a,b));
}

OprFD1 OprFD1::ham(R const& m, R_Vector const& pot, Bio const& b)
{
   return OprFD1(b,F2<FieldSttD1,R,R_Vector,FieldSttD1>(m,pot)(hamFunc));
}

OprFD1 OprFD1::cook(ObsD1 const& op, Bio const& bi)
{
   Z d=bi.d();
   OprFD1 res(bi);
   for (Z i=1;i<=d;++i){
      for (Z a=1;a<=2;++a){
         SttD1 sia(i,a,bi);
         SttD1 oia=op*sia;
         for (Z j=1;j<=d;++j){
            for (Z b=1;b<=2;++b){
               SttD1 sjb(j,b,bi);
               OprFD1 oijab=OprFD1::crt(sjb)*OprFD1::des(sia)*(sjb|oia);
               res=res+oijab;
            }
         }
      }
   }
   return res;
}

#endif
