//? cpmqm.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmqm.cpp
   History: Started 98-10-12

*/

#include <cpmqm.h>
#include <cpmvectormore.h>
#include <cpmobservable.h>
#include <cpmdistribution.h>
#include <cpmmultiimage.h>
#include <cpmframes.h>

using CpmQM::Lat1;
using CpmQM::RunCon;
using CpmQM::DisCon;
using CpmQM::DynDat;
using CpmQM::SysLat1;
using CpmQM::TrjLat1;
using CpmQM::IntActDat;
using CpmQM::SysLat2;
using CpmQM::SysLat3;
using CpmQM::TrjLat2;
using CpmQM::TrjLat3;
using CpmQM::Sep2;
using CpmQM::Sep3;
using CpmQM::Fun3;
using CpmQM::CM3;
using CpmQM::SpinN;
using CpmQM::Ia;
using CpmQM::Ky;
using CpmQM::Qreg;
using CpmQM::QregPar;
using CpmQM::TrjQreg;
using CpmQM::testEC;
using CpmQM::x2vv;

//using CpmSystem::R_;
using CpmArrays::VV;
using CpmArrays::VVV;
using CpmArrays::VVVV;
using CpmArrays::R_Vector;
using CpmArrays::DisR;
using CpmLinAlg::CC2;
using CpmLinAlg::CCCC2;
using CpmImaging::Color;
using CpmImaging::WHITE;
using CpmImaging::BLACK;
using CpmImaging::LIGHTGREEN;
using CpmImaging::ImgR;


////////////////// class Lat1 ////////////////////////////////////////////

bool Lat1::prnOn(ostream& str)const
{
   cpmwt("Lat1");
   cpmp(iv_);
   cpmp(d_);
   return true;
}

bool Lat1::scanFrom(istream& str)
{
   cpms(iv_);
   cpms(d_);
   ini_();
   return true;
}

Z Lat1::com(Lat1 const& lat)const
{
   if (iv_<lat.iv_) return 1;
   if (iv_>lat.iv_) return -1;
   if (d_<lat.d_) return 1;
   if (d_>lat.d_) return -1;
   return 0;
}

////////////////////////////// class DynDat /////////////////////////////
R DynDat::beta_=0;

// CPM_IO

bool DynDat::prnOn(ostream& str)const
{
   cpmwt("DynDat");
   cpmp(m_);
   cpmp(q_);
   cpmp(lam_);
   cpmp(x0_);
   cpmp(p_);
   cpmp(out_);
   return true;
}

bool DynDat::scanFrom(istream& str)
{
   cpms(m_);
   cpms(q_);
   cpms(lam_);
   cpms(x0_);
   cpms(p_);
   cpms(out_);
   return true;
}

////////////////////// struct RunCon /////////////////////////////////////
RunCon::RunCon(RecordHandler const& rch, Word const& sec):
tWaitDocu2(0),tWaitDocu3(0),reversed(true),fromDisk(false),imgSteps(0)
{
   cpmrh(dt);
   cpmrhf(nIter);// needs not to be there
   cpmrhf(tWaitIter);// needs not to be there
   cpmrhf(gammaIter); // needs not to be there
   cpmrh(nSteps);
   cpmrh(viewPeriod);
   cpmrh(docuPeriod);
   cpmrhf(imgPeriod); // needs not to be there
   cpmrh(tWaitIni);
   cpmrh(tWaitView);
   cpmrh(tWaitDocu);
   cpmrhf(tWaitDocu2); // needs not to be there
   cpmrhf(tWaitDocu3); // needs not to be there
   cpmrh(tWaitRes);
   cpmrhf(reversed); // needs not to be there
   cpmrhf(fromDisk); // needs not to be there
   cpmrh(sel);
   cpmrhf(imgSteps); // needs not to be there
}

////////////////////////////// class SysLat1 /////////////////////////////
// CPM_IO

bool SysLat1::prnOn(ostream& str)const
{
   cpmwt("SysLat1");
   cpmp(lat_);
   cpmp(rep_);
   return true;
}

bool SysLat1::scanFrom(istream& str)
{
   cpms(lat_);
   cpms(rep_);
   return true;
}

// construction

namespace{
   R potAux(R const& x, DynDat const& dd, Lat1 const& lat)
   {
      R xs=x+lat.iv().mean(); // shifted
      return dd.pot(xs);
   }
      // the involved argument shift is needed, since
      // QMBase understands the biotope always with the
      // origin at the center
}

SysLat1::SysLat1(Lat1 const& lat, DynDat const& dd, Z n):lat_(lat)
{
   Z mL=2;
   Word loc("SysLat1(Lat1,DynDat,Z)");
   CPM_MA
   using namespace CpmQuantumMechanics;
   QMBase::setMass(dd.m());
   QMBase::setSize(lat.size());
   Z d=lat_.d();
   Observable T1(Word("eKinFd"),d);
#ifdef CPM_Fn
   R_Func pot=F2<R,DynDat,Lat1,R>(dd,lat_)(potAux);
#else
   R_Func pot=F<R,R>(bind(potAux,_1,dd,lat_));
#endif
   Observable V1("x",d,pot);
   Observable H1=T1+V1;
   SpectralObject so;
   {  // encapsulating the heavy initialization of so
      bool enforceSymmetry=true;
         // our H2 is known to be symmetric
      bool test=false;
         // true would consume a lot of time but worked in my test cases
         // perfectly
      Z iterMax=128;
      R eps0=0;
      R eps1=1e-7,eps2=1e-6; // test values from the past
      so=H1.spectralRepresentation(enforceSymmetry,
      test,iterMax,eps0,eps1,eps2); // this is a big beast
      so=so.order();
   }
   rep_=so[n].getRep();
}

#define CPM_SC\
   Z d=lat_.d();\
   rep_=V<C>(d);\
   R xi=lat_.fir();\
   for (Z i=1;i<=d;++i){\
      rep_.cui(i)=C(f(xi));\
      lat_.next_1(xi);\
   }\
   nor_();

void SysLat1::ini_(F<R,C> const& f)
{
   CPM_SC
}

void SysLat1::ini_(F<R,R> const& f)
{
   CPM_SC
}

#undef CPM_SC

// arithmetics

namespace{
   void accFunc(C& z, C const& zAdd){ z+=zAdd;}
   C conjFunc(C const& z){ return z.con();}
   C negFunc(C const& z1){  return -z1;}
   C sumFunc(C const& z1, C const& z2){ return z1+z2;}
   C diffFunc(C const& z1, C const& z2){ return z1-z2;}
   C prodFunc(C const& z1, C const& z2){ return z1*z2;}
   C prodFuncConj1(C const& z1, C const& z2){return z1.con()*z2; }
   C prodFuncConj2(C const& z1, C const& z2){ return z1*z2.con();}
   C prodFuncReal(C const& z1, C const& z2){ return z1*z2.real();}
}

C SysLat1::operator |(SysLat1 const& x2)const
{
   return rep_.fAcc2(prodFuncConj1,accFunc,x2.rep_);
}

SysLat1 SysLat1::con(void)const
{
   V<C> v=rep_.fa(conjFunc);
   return SysLat1(lat_.iv(),v);
}

SysLat1 SysLat1::operator - (void)const
{
   V<C> v=rep_.fa(negFunc);
   return SysLat1(lat_.iv(),v);
}

namespace{
   C fpFunc(C const& z, R const& p)
   {
      return p==2 ? C(z.absSqr()) : C(cpmpow(z.abs(),p));
   }
}

SysLat1& SysLat1::makeReal_(R p)
{
#ifdef CPM_Fn
   F<C,C> fp=F1<C,R,C>(p)(fpFunc);
#else
   F<C,C> fp(bind(fpFunc,_1,p));
#endif
   rep_.fa_(fp,rep_.dom());
   return *this;
}

SysLat1& SysLat1::operator +=(SysLat1 const& x)
{
   rep_.ff_(sumFunc,x.rep_);
   return *this;
}

SysLat1& SysLat1::operator -=(SysLat1 const& x)
{
   rep_.ff_(diffFunc,x.rep_);
   return *this;
}

SysLat1& SysLat1::operator *=(C z)
{
   rep_.fc_(prodFunc,z);
   return *this;
}

SysLat1& SysLat1::operator *=(R r)
{
   rep_.fc_(prodFuncReal,C(r));
   return *this;
}

SysLat1 SysLat1::operator +(SysLat1 const& x)const
{
  SysLat1 res(*this);
  return res+=x;
}

SysLat1 SysLat1::operator -(SysLat1 const& x)const
{
   SysLat1 res(*this);
   return res-=x;
}

SysLat1 SysLat1::operator *(C x2)const
{
  SysLat1 res(*this);
  return res*=x2;
}

SysLat1 SysLat1::operator *(R x2)const
{
  SysLat1 res(*this);
  return res*=x2;
}

R SysLat1::nor_()
{
   R n=norm();
   if (n>0){
      R nInv=R(1)/n;
      operator*=(nInv);
   }
   return n;
}

// graphics

void SysLat1::mark(Graph& g, DisCon const& dc)const
{
   using CpmLinAlg::R2;
   using CpmArrays::V;
   using namespace CpmImaging;
   V< V<R2> > dat(3);
   Z d=lat_.d();
   R xi=lat_.fir();
   V<R2> re(d);
   V<R2> im(d);
   V<R2> ab(d);
   for (Z i=0;i<d;++i){
      C zi=rep_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re.cui(ip)=rei;
      im.cui(ip)=imi;
      ab.cui(ip)=abi;
      lat_.next_1(xi);
   }
   dat[1]=re;
   dat[2]=im;
   dat[3]=ab;
   V<Color> cl(3);
   cl[1]=RED;  // color for real part
   cl[2]=GREEN;   // color for imaginary part
   cl[3]=BLACK; // color for absolute value
   g.setAutoScl_(dc.autoScl);
   g.mark(dat,cl,dc.secFac);
}
// dynamics

void SysLat1::dot_(DynDat const& dd)
// mutating function to avoid copying of state data
{
   Z d=lat_.d();
   if (d<2) cpmerror("SysLat1::dot_(DynDat): d<2");

   R ds=lat_.ds();
   R fr=dd.m()*ds*ds*2;
   if (fr==0.) cpmerror("SysLat1::dot_(DynDat): fr==0");

   fr=R(1)/fr;
   C f1=fr*C::I;
   C f2=-C::I;
   V<C> mem=rep_;
   R xi=lat_.fir();
//   Z dm=d-1;
// handling of boundary conditions
   CpmArrays::Outside out=dd.out();
   C valLeft=mem(-1,out);
   C valRight= mem(d,out);
// first term of the loop
   rep_.cui(1)=(valLeft - mem.cui(1)*2 + mem.cui(2))*f1 +
      mem.cui(1)*f2*dd.pot(xi);
   lat_.next_1(xi); // shifts xi to the next position
// core terms of the loop
   for (Z i=2;i<d;++i){ // part not affected by boundary conditions
      rep_.cui(i)=(mem.cui(i-1) - mem.cui(i)*2 + mem.cui(i+1))*f1 +
         mem.cui(i)*f2*dd.pot(xi);
      lat_.next_1(xi);
   }
// last term of the loop
// xi is now at the last position
   rep_.cui(d)=(mem.cui(d-1) - mem.cui(d)*2 + valRight)*f1 +
      mem.cui(d)*f2*dd.pot(xi);
      // up to positve constants we have psi^dot= -i H psi =
      // -i (T + V) psi =
      // (-i) T psi - i V psi = (-i) -psi" - i V psi
      // = i psi" - i V psi . In finite differences:
      // psi" = psi1 -2 psi0 +psi(-1).
 }

R SysLat1::enrHomFil(DynDat const& dd)const
{
   R b=dd.q()*DynDat::beta();
   R res=0;
   R xi=lat_.fir();
   for (Z i=rep_.b();i<=rep_.e();++i){
      res+=rep_.cui(i).absSqr()*b*xi;
      lat_.next_1(xi); // shifts xi to the next position
   }
   return res;
 }

R SysLat1::dtSup(DynDat const& dd, Z p, R tWait)const
{
   Z mL=1;
   Word loc("SysLat1::dtSup(Z)");
   CPM_MA
   cpmassert(p>0,loc);
   SysLat1 psi=*this;
   psi.nor_();
   R_Vector mem(p);
   Graph gr;
   R pInv=1./p;
   for (Z i=1;i<=p;++i){
      psi.dot_(dd);
      R mi=psi.nor_(); // the value is the norm!
      mem[i]=mi; // the value is the norm!
      cpmvalue("enr",mi,2);
      if (tWait>=0){
         psi.mark(gr,DisCon());
         gr.vis();
         cpmwait(tWait);
      }
      if (i==1){
         R n1=mem[i];
         cpmdebug(n1);
      }
      cpmprogress("x",R(i)*pInv,3);
   }
   R_Vector stat=CpmArrays::statistics(mem);
   R sigma=stat[3];
   R maxVal=stat[10];
   cpmdebug(maxVal);
   R epsMax=maxVal+sigma;
   cpmdebug(epsMax);
   cpmvalue("eMax",epsMax,3);
   R res=R(2)/epsMax;
   CPM_MZ
   return res;
}

//////////////////////////// class TrjLat1 ///////////////////////////////

C TrjLat1::hInt()const
{
   Lat1 lat=x_.lat();
   Z d=lat.d();
   SysLat1 ax(x_);
   ax.makeReal_();
   C res;
   for (Z i=1;i<=d;++i){
      R xi=lat.pos(i);
      res+= ax.cui(i)*dd_.pot(xi);
   }
   return res;
}

#define CPM_SC_RUN\
   using CpmArrays::R_Matrix;\
   using CpmArrays::R_Vector;\
   using std::vector;\
   g.setSclMem_(false);\
   R n0=n();\
   R n0Inv=R(1)/n0;\
   C e0=eTot();\
   R eTotR=e0[1];\
   C ei0=hInt();\
   R eInt=ei0[1];\
   cpmvalues2("eTot,eInt",eTotR,eInt,4);\
   cpmdata<<endl<<"eTot="<<e0<<" eInt="<<ei0<<endl;\
   R e0Inv=cpminv(e0.abs());\
   R e01=e0[1];\
   R e02=e0[2];\
   R dt=rc.dt;\
   Z nSteps=rc.nSteps;\
   R nInv=R(1)/nSteps;\
   R tWaitDocu=rc.tWaitDocu;\
   R tWaitRes=rc.tWaitRes;\
   CpmAlgorithms::CyclicAlarm img(rc.imgPeriod);\
   CpmAlgorithms::CyclicAlarm view(rc.viewPeriod);\
   CpmAlgorithms::CyclicAlarm docu(rc.docuPeriod);\
   if (rc.tWaitIni>0){\
      mark(g,dc);\
      g.setText("initial state");\
      g.vis();\
      cpmwait(rc.tWaitIni,2);\
   }\
   g.clr_();\
   vector<R> tv(1,0.), nv(1,0.), e1v(1,0.), e2v(1,0.);

#define CPM_SC_RUN2\
   R ti=t();\
   R ni=n();\
   C ei=eTot();\
   tv.push_back(ti);\
   nv.push_back((ni-n0)*n0Inv);\
   e1v.push_back((ei[1]-e01)*e0Inv);\
   e2v.push_back((ei[2]-e02)*e0Inv);

#define CPM_SC_RUN3\
   R_Matrix doc(4);\
   doc[1]=R_Vector(tv);\
   doc[2]=R_Vector(nv);\
   doc[3]=R_Vector(e1v);\
   doc[4]=R_Vector(e2v);\
   g.clr_();\
   g.setAutoScl_(true);

V< V<R2> > TrjLat1::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=x_.dtSup(dd_,rc.nIter,rc.tWaitIter);
      dt*=dtMax;
   }
   vector<R2> s1v;
   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R si=s();
            s1v.push_back(R2(ti,si));
         }
      }
      if (view()){
         mark(g,dc);
         g.vis();
         cpmwait(rc.tWaitView);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("TrjLat1: run documentation");
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(1);
   res[1]=V<R2>(1,s1v);
   if (tWaitRes>0){
      g.clr_();
      g.setText("TrjLat1: res documentation");
      g.mark(res);
      g.vis();
      cpmwait(tWaitRes,2);
   }
   return res;
}

//////////////////////////// class IntActDat /////////////////////////////
// CPM_IO

bool IntActDat::prnOn(ostream& str)const
{
   cpmwt("IntActDat");
   cpmp(alpha_);
   cpmp(mu_);
   cpmp(eps_);
   cpmp(p_);
   return true;
}

bool IntActDat::scanFrom(istream& str)
{
   cpms(alpha_);
   cpms(mu_);
   cpms(eps_);
   cpms(p_);
   return true;
}

///////////////////////// class SysLat2 ////////////////////////////////
// CPM_IO

bool SysLat2::prnOn(ostream& str)const
{
   cpmwt("SysLat2");
   cpmp(lat1_);
   cpmp(lat2_);
   cpmp(rep_);
   return true;
}

bool SysLat2::scanFrom(istream& str)
{
   cpms(lat1_);
   cpms(lat2_);
   cpms(rep_);
   return true;
}

// construction

SysLat2::SysLat2(SysLat1 const& s1, SysLat1 const& s2):
   lat1_(s1.lat()), lat2_(s2.lat()),rep_(lat1_.d(),V<C>(lat2_.d()))
{
   Z d1=lat1_.d(),d2=lat2_.d();
   for (Z i=1;i<=d1;++i){
      C si=s1.cui(i);
      for (Z j=1;j<=d2;++j){
         C sij=si*s2.cui(j);
         rep_.cui(i).cui(j)=sij;
      }
   }
}

SysLat2& SysLat2::operator *=(C z)
{
   Z d1=lat1_.d();
   for (Z i=1;i<=d1;++i){
      rep_.cui(i).fc_(prodFunc,z);
   }
   return *this;
}

SysLat2& SysLat2::operator *=(R r)
{
   Z d1=lat1_.d();
   for (Z i=1;i<=d1;++i){
      rep_.cui(i).fc_(prodFuncReal,C(r));
   }
   return *this;
}

SysLat2& SysLat2::operator +=(const SysLat2& x)
{
   Z d1=lat1_.d();
   for (Z i=1;i<=d1;i++){
      rep_.cui(i).ff_(sumFunc,x.rep_.cui(i));
   }
   return *this;
}

SysLat2& SysLat2::operator -=(const SysLat2& x)
{
   Z d1=lat1_.d();
   for (Z i=1;i<=d1;i++){
      rep_.cui(i).ff_(diffFunc,x.rep_.cui(i));
   }
   return *this;
}

C SysLat2::operator |(SysLat2 const& x)const
{
   C res;
   Z d1=lat1_.d();
   for (Z i=1;i<=d1;i++){
      res+=rep_.cui(i).fAcc2(prodFuncConj1, accFunc, x.rep_.cui(i));
   }
   return res;
}

SysLat2 SysLat2::operator +(SysLat2 const& x)const
{
  SysLat2 res(*this);
  return res+=x;
}

SysLat2 SysLat2::operator -(SysLat2 const& x)const
{
   SysLat2 res(*this);
   return res-=x;
}

SysLat2 SysLat2::operator *(C x2)const
{
  SysLat2 res(*this);
  return res*=x2;
}

SysLat2 SysLat2::operator *(R x2)const
{
  SysLat2 res(*this);
  return res*=x2;
}

R SysLat2::nor_()
{
   R n=norm();
   if (n>0){
      R nInv=R(1)/n;
      operator*=(nInv);
   }
   return n;
}

void SysLat2::dot_(DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12)
{
   Z mL=2;
   Word loc("SysLat2::dot_(DynDat,DynDat,IntActDat)");
   CPM_MA
   R m1=dd1.m(), m2=dd2.m();
   R q1=dd1.q(),q2=dd2.q();
   R q12=q1*q2;
   R ds1=lat1_.ds(), ds2=lat2_.ds();
   R f1=R(1)/ds1;
   f1*=(-f1);
   R f2=R(1)/ds2;
   f2*=(-f2);
   R fac1=f1/(m1*2);
   R fac2=f2/(m2*2);
   R facs=-(fac1+fac2)*2;
   C fac=-C::I;
   Z d1=lat1_.d(), d2=lat2_.d();
   Outside o1=dd1.out(), o2=dd2.out();
   V< V<C> > vv=rep_;
   for (Z i=1;i<=d1;++i){
      R xi=lat1_.pos(i);
      R v1=dd1.pot(xi);
      for (Z j=1;j<=d2;++j){
         R xj=lat2_.pos(j);
         C vvij=vv.cui(i).cui(j);
         C cij=vvij*facs+
         (vv.read(i-1,o1).read(j,o2)+vv.read(i+1,o1).read(j,o2))*fac1+
         (vv.cui(i).read(j-1,o2)+vv.cui(i).read(j+1,o2))*fac2;
         R rij=dd12.dis(xi,xj);
         R v12=q12*dd12.pot(rij);
         R v2=dd2.pot(xj);
         C vij=vvij*(v1+v12+v2);
         cij+=vij;
         cij*=fac;
         rep_.cui(i).cui(j)=cij;
      }
   }
   CPM_MZ
}

SysLat1 SysLat2::red1(SysLat1 const& s)const
{
   Z mL=2;
   Word loc("SysLat2::red1(SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   cpmassert(s.dim()==d1,loc);
   SysLat1 res(lat2_);
   for (Z j=1;j<=d2;++j){
      C cj;
      for (Z i=1;i<=d1;++i){
         C si=s.cui(i).con();
         cj+=si*rep_.cui(i).cui(j);
      }
      res.cui(j)=cj;
   }
   CPM_MZ
   return res;
}

SysLat1 SysLat2::red2(SysLat1 const& s)const
{
   Z mL=2;
   Word loc("SysLat2::red2(SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   cpmassert(s.dim()==d2,loc);
   SysLat1 res(lat1_);
   for (Z i=1;i<=d1;++i){
      C ci;
      for (Z j=1;j<=d2;++j){
         C sj=s.cui(j).con();
         ci+=sj*rep_.cui(i).cui(j);
      }
      res.cui(i)=ci;
   }
   CPM_MZ
   return res;
}

R SysLat2::dtSup(DynDat const& dd1, DynDat const& dd2,
         IntActDat const& dd12, Z p, R tWait, R gamma)const
{
   Z mL=1;
   Word loc("SysLat2::dtSup(DynDat,DynDat,IntActDat,Z,R,R)");
   CPM_MA
   cpmassert(p>0,loc);
   SysLat2 psi=*this;
   psi.nor_();
   R_Vector mem(p);
   Graph gr;
   R pInv=1./p;
   for (Z i=1;i<=p;++i){
      psi.dot_(dd1,dd2,dd12);
      R mi=psi.nor_(); // the value is the norm!
      mem[i]=mi; // the value is the norm!
      cpmvalue("enr",mi,2);
      if (tWait>0){
         psi.mark(gr,0,gamma);
         gr.vis();
         cpmwait(tWait);
      }
      if (i==1){
         R n1=mem[i];
         cpmdebug(n1);
      }
      cpmprogress("x",R(i)*pInv,3);
   }
   R_Vector stat=CpmArrays::statistics(mem);
   R sigma=stat[3];
   R maxVal=stat[10];
   cpmdebug(maxVal);
   R epsMax=maxVal+sigma;
   cpmdebug(epsMax);
   cpmvalue("eMax",epsMax,3);
   R res=R(2)/epsMax;
   CPM_MZ
   return res;
}

R_Vector SysLat2::stt(DynDat const& dd1, DynDat const& dd2)const
{
   Z mL=2;
   Word loc("SysLat2::stt(DynDat)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   R res1=0;
   R res2=0;
   R res12=0;
   for (Z i=1;i<=d1;++i){
      R s1=lat1_.pos(i)-dd1.x0();
      for (Z j=1;j<=d2;++j){
         R s2=lat2_.pos(j)-dd2.x0();
         R wij=rep_.cui(i).cui(j).absSqr();
         res1+=wij*s1;
         res2+=wij*s2;
         res12+=wij*s1*s2;
      }
   }
   CPM_MZ
   return R_Vector{res1,res2,res12};
}

void SysLat2::mark(Graph& g, Z ipolMeth, R gamma, R secFac)const
{
   CpmImaging::Color::setGamma(gamma);
   bool aspRatFromState=false;
   bool periodic=true;
   bool matrixstyle=true;
   g.markVVC(rep_,matrixstyle,Iv(),ipolMeth,aspRatFromState,periodic,
      secFac);

}
///////////////////// class SysLat3 ////////////////////////////////////
// CPM_IO

bool SysLat3::prnOn(ostream& str)const
{
   cpmwt("SysLat3");
   cpmp(lat1_);
   cpmp(lat2_);
   cpmp(lat3_);
   cpmp(rep_);
   return true;
}

bool SysLat3::scanFrom(istream& str)
{
   cpms(lat1_);
   cpms(lat2_);
   cpms(lat3_);
   cpms(rep_);
   return true;
}

// construction

SysLat3::SysLat3(SysLat1 const& s1, SysLat1 const& s2, SysLat1 const& s3):
   lat1_(s1.lat()), lat2_(s2.lat()), lat3_(s3.lat())
{
   Z d1=lat1_.d(),d2=lat2_.d(),d3=lat3_.d();
   V<C> rep3(d3);
   V< V<C> > rep23(d2, rep3);
   rep_=V< V < V <C> > >(d1,rep23);
   for (Z i=1;i<=d1;++i){
      C si=s1.cui(i);
      for (Z j=1;j<=d2;++j){
         C sij=si*s2.cui(j);
         for (Z k=1;k<=d3;++k){
            C sijk=sij*s3.cui(k);
            rep_.cui(i).cui(j).cui(k)=sijk;
         }
      }
   }
}

SysLat3& SysLat3::operator *=(C z)
{
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         for (Z k=1;k<=d3;++k){
            rep_.cui(i).cui(j).cui(k)*=z;
         }
      }
   }
   return *this;
}

SysLat3& SysLat3::operator *=(R r)
{
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         for (Z k=1;k<=d3;++k){
            rep_.cui(i).cui(j).cui(k)*=r;
         }
      }
   }
   return *this;
}

SysLat3& SysLat3::operator +=(const SysLat3& x)
{
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         for (Z k=1;k<=d3;++k){
            rep_.cui(i).cui(j).cui(k)+=x.rep_.cui(i).cui(j).cui(k);
         }
      }
   }
   return *this;
}

SysLat3& SysLat3::operator -=(const SysLat3& x)
{
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         for (Z k=1;k<=d3;++k){
            rep_.cui(i).cui(j).cui(k)-=x.rep_.cui(i).cui(j).cui(k);
         }
      }
   }
   return *this;
}

C SysLat3::operator |(SysLat3 const& x)const
{
   C res;
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         for (Z k=1;k<=d3;++k){
            res+=rep_.cui(i).cui(j).cui(k).con()*
               x.rep_.cui(i).cui(j).cui(k);
         }
      }
   }
   return res;
}

SysLat3 SysLat3::operator +(SysLat3 const& x)const
{
  SysLat3 res(*this);
  return res+=x;
}

SysLat3 SysLat3::operator -(SysLat3 const& x)const
{
   SysLat3 res(*this);
   return res-=x;
}

SysLat3 SysLat3::operator *(C x2)const
{
  SysLat3 res(*this);
  return res*=x2;
}

SysLat3 SysLat3::operator *(R x2)const
{
  SysLat3 res(*this);
  return res*=x2;
}

R SysLat3::nor_()
{
   R n=norm();
   if (n>0){
      R nInv=R(1)/n;
      operator*=(nInv);
   }
   return n;
}

void SysLat3::dot_(DynDat const& dd1, DynDat const& dd2,
   DynDat const& dd3, IntActDat const& dd12)
{
   Z mL=2;
   Word loc("SysLat3::dot_(DynDat,DynDat,DynDat,IntActDat)");
   CPM_MA
   R m1=dd1.m(), m2=dd2.m(), m3=dd3.m();
   R q1=dd1.q(), q2=dd2.q(), q3=dd3.q();
   R ds1=lat1_.ds(), ds2=lat2_.ds(), ds3=lat3_.ds();

   R f1=R(1)/ds1;
   f1*=(-f1);
   R f2=R(1)/ds2;
   f2*=(-f2);
   R f3=R(1)/ds3;
   f3*=(-f3);
   R fac1=f1/(m1*2);
   R fac2=f2/(m2*2);
   R fac3=f3/(m3*2);

   R facs=(fac1+fac2+fac3)*(-2);
   C fac=-C::I;
   Z d1=lat1_.d();
   Z d2=lat2_.d();
   Z d3=lat3_.d();
   Outside o1=dd1.out();
   Outside o2=dd2.out();
   Outside o3=dd3.out();
   V< V< V<C> > > vv=rep_;
   for (Z i=1;i<=d1;++i){
      R xi=lat1_.pos(i);
      R v1=dd1.pot(xi);
      for (Z j=1;j<=d2;++j){
         R xj=lat2_.pos(j);
         R v2=dd2.pot(xj);
         R rij=dd12.dis(xi,xj);
         R pij=q1*q2*dd12.pot(rij);
         for (Z k=1;k<=d3;++k){
            R xk=lat3_.pos(k);
            R v3=dd3.pot(xk);
            R rik=dd12.dis(xi,xk);
            R pik=q1*q3*dd12.pot(rik);
            R rjk=dd12.dis(xj,xk);
            R pjk=q2*q3*dd12.pot(rjk);
            C vvijk=vv.cui(i).cui(j).cui(k);
            C cijk=vvijk*facs
               +(vv.read(i-1,o1).read(j,o2).read(k,o3)
                  +vv.read(i+1,o1).read(j,o2).read(k,o3))*fac1
               +(vv.cui(i).read(j-1,o2).read(k,o3)
                  +vv.cui(i).read(j+1,o2).read(k,o3))*fac2
               +(vv.cui(i).cui(j).read(k-1,o3)
                  +vv.cui(i).cui(j).read(k+1,o3))*fac3;
            C vijk=vvijk*(v1+v2+v3+pij+pik+pjk);
               // here was v2+v2 instead of v2+v3 till 2008-02-11, had bad
               // effects
            cijk+=vijk;
            cijk*=fac;
            rep_.cui(i).cui(j).cui(k)=cijk;
         }
      }
   }
   CPM_MZ
}

SysLat2 SysLat3::red3(SysLat1 const& s)const
{
   Z mL=2;
   Word loc("SysLat3::red1(SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s.dim()==d3,loc);
   SysLat2 res(lat1_,lat2_);
   for (Z i=1;i<=d1;++i){
      for (Z j=1;j<=d2;++j){
         C ck;
         for (Z k=1;k<=d3;++k){
            C sk=s.cui(k).con();
            ck+=sk*rep_.cui(i).cui(j).cui(k);
         }
         res.cui(i).cui(j)=ck;
      }
   }
   CPM_MZ
   return res;
}

SysLat1 SysLat3::red12(SysLat1 const& s1, SysLat1 const& s2)const
{
   Z mL=3;
   Word loc("SysLat3::red12(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s1.dim()==d1,loc);
   cpmassert(s2.dim()==d2,loc);
   SysLat1 s3(lat3_);
   for (Z k=1;k<=d3;++k){
      C ck;
      for (Z i=1;i<=d1;++i){
         C ci=s1.cui(i).con();
         for (Z j=1;j<=d2;++j){
            C cj=s2.cui(j).con();
            C cij=ci*cj;
            ck+=cij*rep_.cui(i).cui(j).cui(k);
         }
      }
      s3.cui(k)=ck;
   }
   CPM_MZ
   return s3;
}

SysLat1 SysLat3::red23(SysLat1 const& s2, SysLat1 const& s3)const
{
   Z mL=3;
   Word loc("SysLat3::red23(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s2.dim()==d2,loc);
   cpmassert(s3.dim()==d3,loc);
   SysLat1 s1(lat1_);
   for (Z k=1;k<=d1;++k){
      C ck;
      for (Z i=1;i<=d2;++i){
         C ci=s2.cui(i).con();
         for (Z j=1;j<=d3;++j){
            C cj=s3.cui(j).con();
            C cij=ci*cj;
            ck+=cij*rep_.cui(k).cui(i).cui(j);
         }
      }
      s1.cui(k)=ck;
   }
   CPM_MZ
   return s1;
}


SysLat1 SysLat3::red31(SysLat1 const& s3, SysLat1 const& s1)const
{
   Z mL=3;
   Word loc("SysLat3::red31(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s1.dim()==d1,loc);
   cpmassert(s3.dim()==d3,loc);
   SysLat1 s2(lat2_);
   for (Z k=1;k<=d2;++k){
      C ck;
      for (Z i=1;i<=d1;++i){
         C ci=s1.cui(i).con();
         for (Z j=1;j<=d3;++j){
            C cj=s3.cui(j).con();
            C cij=ci*cj;
            ck+=cij*rep_.cui(i).cui(k).cui(j);
         }
      }
      s2.cui(k)=ck;
   }
   CPM_MZ
   return s2;
}

SysLat1 SysLat3::red12(SysLat2 const& s12)const
{
   Z mL=3;
   Word loc("SysLat3::red12(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s12.dim1()==d1,loc);
   cpmassert(s12.dim2()==d2,loc);
   SysLat1 s3(lat3_);
   for (Z k=1;k<=d3;++k){
      C ck;
      for (Z i=1;i<=d1;++i){
       //  C ci=s1.cui(i).con();
         for (Z j=1;j<=d2;++j){
        //    C cj=s2.cui(j).con();
            C cij=s12.cui(i).cui(j).con();
            ck+=cij*rep_.cui(i).cui(j).cui(k);
         }
      }
      s3.cui(k)=ck;
   }
   CPM_MZ
   return s3;
}

SysLat1 SysLat3::red23(SysLat2 const& s23)const
{
   Z mL=3;
   Word loc("SysLat3::red23(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s23.dim1()==d2,loc);
   cpmassert(s23.dim2()==d3,loc);
   SysLat1 s1(lat1_);
   for (Z k=1;k<=d1;++k){
      C ck;
      for (Z i=1;i<=d2;++i){
       //  C ci=s2.cui(i).con();
         for (Z j=1;j<=d3;++j){
         //   C cj=s3.cui(j).con();
          //  C cij=ci*cj;
            C cij=s23.cui(i).cui(j).con();
            ck+=cij*rep_.cui(k).cui(i).cui(j);
         }
      }
      s1.cui(k)=ck;
   }
   CPM_MZ
   return s1;
}


SysLat1 SysLat3::red31(SysLat2 const& s31)const
{
   Z mL=3;
   Word loc("SysLat3::red31(SysLat1,SysLat1)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim3();
   cpmassert(s31.dim1()==d3,loc);
   cpmassert(s31.dim2()==d1,loc);
   SysLat1 s2(lat2_);
   for (Z k=1;k<=d2;++k){
      C ck;
      for (Z i=1;i<=d1;++i){
          for (Z j=1;j<=d3;++j){
            C cij=s31.cui(i).cui(j).con();
            ck+=cij*rep_.cui(i).cui(k).cui(j);
         }
      }
      s2.cui(k)=ck;
   }
   CPM_MZ
   return s2;
}


R SysLat3::dtSup(DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
         IntActDat const& dd12, Z p, R tWait, R gamma)const
{
   Z mL=1;
   Word loc("SysLat3::dtSup(DynDat,DynDat,DynDat,IntActDat,Z,R,R)");
   CPM_MA
   cpmassert(p>0,loc);
   SysLat3 psi=*this;
   psi.nor_();
   R_Vector mem(p);
   Graph gr;
   R pInv=1./p;
   cpmdata<<endl<<loc<<": Iterative approach to the norm of the\
    Hamiltonian:";
   for (Z i=1;i<=p;++i){
      psi.dot_(dd1,dd2,dd3,dd12);
      R mi=psi.nor_(); // the value is the norm!
      mem[i]=mi; // the value is the norm!
      cpmdata<<endl<<"mi="<<cpm(mi);
      if (tWait>0){
         DisCon dc; dc.ipolMet=0; dc.gamma=gamma;
         psi.mark(gr,dc);
         gr.vis();
         cpmwait(tWait);
         cpmprogress("x",R(i)*pInv,3);
      }
   }
   R_Vector stat=CpmArrays::statistics(mem);
   R sigma=stat[3];
   R maxVal=stat[10];
   cpmnote(sigma);
   cpmnote(maxVal);
   R epsMax=maxVal+sigma;
   cpmdata<<endl<<"Norm as estimated from statistcs: epsMax= "<<epsMax;
   R dt=R(2)/epsMax;
   cpmdata<<endl<<"dt=2./epsMax="<<dt;
   CPM_MZ
   return dt;
}

R_Vector SysLat3::stt(DynDat const& dd1, DynDat const& dd2,
   DynDat const& dd3)const
{
   Z mL=2;
   Word loc("SysLat3::s123(DynDat,DynDat,DynDat)");
   CPM_MA
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim2();
   R res1=0;
   R res2=0;
   R res3=0;
   R res12=0;
   R res13=0;
   R res23=0;
   R res123=0;
   for (Z i=1;i<=d1;++i){
      R s1=lat1_.pos(i)-dd1.x0();
      for (Z j=1;j<=d2;++j){
         R s2=lat2_.pos(j)-dd2.x0();
         for (Z k=1;k<=d3;++k){
            R s3=lat3_.pos(k)-dd3.x0();
            R wijk=rep_.cui(i).cui(j).cui(k).absSqr();
            res1+=wijk*s1;
            res2+=wijk*s2;
            res3+=wijk*s3;
            res12+=wijk*s1*s2;
            res13+=wijk*s1*s3;
            R s23=s2*s3;
            res23+=wijk*s23;
            res123+=wijk*s1*s23;
         }
      }
   }
   R rest=res1+res2+res3;
   CPM_MZ
   return R_Vector{res1,res2,res3,res12,res13,res23,res123,rest};
}

void SysLat3::mark(Graph& g, DisCon const& dc)const
{
   CpmImaging::Color::setGamma(dc.gamma);
   bool aspRatFromState=true;
   bool matrixstyle=true;
   bool periodic=true;
   V< V<C> > rep2=CpmArrays::tenToMat(rep_);
   g.markVVC(rep2,matrixstyle,Iv(),dc.ipolMet,aspRatFromState,periodic,
      dc.secFac);
}

/************* experimental version which makes the data
// stuctures of function transform visible
void SysLat3::mark(Graph& g, DisCon const& dc)const
{
   CpmImaging::Color::setGamma(0);
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim2();
   Z d12=d1+d2;
   Z d123=d12+d3;
   Z d=d1*d2*d3;
   Z nH=g.dim2()-1;
   Z nV=g.dim1()-1-d123;
   Z i,j,k,l;
   Z p=1;
   static Z ii=0,jj=0;
   static bool firstrun=true;
   if (firstrun){
      g.clr_(BLACK);
      firstrun=false;
   }
   for (i=0;i<d1;++i){
      for (j=0;j<d2;++j){
         for (k=0;k<d3;++k){
            C val=rep_[i][j][k];
            Color c(val);
            R w=val.absSqr();
            for (l=0;l<p;++l){
               R rv=CpmRoot::randomR();
               if (rv<w*d){
                  g.putPixel(jj+i,ii,c);
                  g.putPixel(jj+d1+j,ii,c);
                  g.putPixel(jj+d12+k,ii,c);
               }
               if (ii<nH)
                  ii++;
               else{
                  ii=0;
                  if (jj<nV){
                     jj+=d123;
                  }
                  else{
                     jj=0;
                     g.clr_(BLACK);
                  }
               }
            }
         }
      }
   }
}
*/

vector< vector<C> > SysLat3::transform(Graph& g, DisCon const& dc)const
/*
This provides an alternative form of the concept of a 3-particle wave
function (3 is the situation in this specific function, generalization
to general n is obvious). For each of the d1+d2+d3 configurations
(not d1*d2*d3 !) there is stream of data which, when taken together,
describe the present state. Notice that the length of each of these
streams is d1*d2*d3. An index running simultaneously for the
d1+d2+d3 streams from 0 to d1*d2*d3-1 is a kind of time parameter
within the 'point in time' associated with the state as a whole.
The state thus can be considered an evolution process in this
new time parameter (which we call 'fast time').
As already mentioned, the state is completely determined by the
the data arrays res[i]. This suggests that the primary data are
the arrays res[i], i=1,2,... d1+d2+d3 which provides a 'local view'.
For each of streams it is normal to 'have value 0' i.e. res[i][j]=C(0)
is normal. For given j, there are always three indexes i1, i2, i3
such that res[i1][j], res[i2][j], res[i3][j] differ from C(0) and
are equal. This shows that retreiving the state information from the
arrays res[i] is possible only if one inspect 'time-slices' in
fast time.
The information which makes up a three-particle wave function is
thus contained in 'coincidences' in parallel streams. This should
be a model for my old project to understand many-particle wave functions
as correlation functions of a fast local 'Funkel-Process'.
There is an obvious paralellism to the concept of the late Fritz Bopp
who studied local appearance and dis-appearance of particles as a
process which may underlies quantum mechanical wave functions.
*/
{
   Z d1=dim1();
   Z d2=dim2();
   Z d3=dim2();
   Z d12=d1+d2;
   Z d123=d12+d3;
   Z i,j,k,l;

   vector< vector<C> > res(d123);

   for (i=1;i<=d1;++i){
      for (j=1;j<=d2;++j){
         for (k=1;k<=d3;++k){
            C val0,val=rep_[i][j][k];
            for (l=0;l<d123;++l){ // each res[l] gets an new
               // component appended
               if (l==i) res[l].push_back(val);
               else if (l==j+d1) res[l].push_back(val);
               else if (l==k+d12) res[l].push_back(val);
               else res[l].push_back(val0);
            }
         }
      }
   }
   return res;
}
////////////////////////////// class TrjLat2 /////////////////////////////

C TrjLat2::hInt()const
{
   C enr=eTot();
   IntActDat dd120=dd12_;
   dd120.free_();
   R betaMem=DynDat::beta();
   DynDat::beta()=0;
   SysLat2 v0(x_);
   v0.dot_(dd1_,dd2_,dd120);
   C enr0=C::I*(x_|v0);
   DynDat::beta()=betaMem;
   return enr-enr0;
}

V< V<R2> > TrjLat2::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
      CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=dtSup(rc.nIter,rc.tWaitIter,rc.gammaIter);
      dt*=dtMax;
      rc.dt=-dt; // then call another run function subsequently will use
         // dt
   }
   vector<R2> s1v, s2v, s12v;
   g.clr();
   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R_Vector rv=stt();
            s1v.push_back(R2(ti,rv[1]));
            s2v.push_back(R2(ti,rv[2]));
            s12v.push_back(R2(ti,rv[3]));
         }
      }
      if (view()){
         mark(g,dc);
         g.vis();
         cpmwait(rc.tWaitView);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("TrjLat2: run documentation");
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(3);
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s12v);
   if (tWaitRes>0){
      g.clr_();
      g.setAutoScl_(true);
      g.setText("TrjLat2: res documentation");
      g.mark(res);
      g.vis();
      cpmwait(tWaitRes,2);
   }
   return res;
}

////////////////////////////// class TrjLat3 /////////////////////////////

C TrjLat3::hInt()const
{
   C enr=eTot();
   IntActDat dd120=dd12_;
   dd120.free_();
   R betaMem=DynDat::beta();
   DynDat::beta()=0;
   SysLat3 v0(x_);
   v0.dot_(dd1_,dd2_,dd3_,dd120);
   C enr0=C::I*(x_|v0);
   DynDat::beta()=betaMem;
   return enr-enr0;
}

x2vv TrjLat3::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=dtSup(rc.nIter,rc.tWaitIter,rc.gammaIter);
      dt*=dtMax;
      rc.dt=-dt; // then calling another run-function (e.g. one for
         // separated QM, i.e. Sep3::run_()) subsequently with
         // argument rc will use dt, and will disable its own time
         // step adjustment method. So far the stability properties
         // of Sep3 allowed this taking over of a  time step determined
         // by TrjLat3::run_.
   }
   vector<R2> s1v;
   vector<R2> s2v;
   vector<R2> s3v;
   vector<R2> s12v;
   vector<R2> s13v;
   vector<R2> s23v;
   vector<R2> s123v;
   vector<R2> stv;
   vector<Word> vw;
   cpmdebug(rc.imgSteps);
   Vo<Z> imgSteps=rc.imgSteps;
   cpmmessage("before sorting");
   imgSteps.sort_();
   cpmmessage("sorting done");
   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R_Vector rv=stt();
            s1v.push_back(R2(ti,rv[1]));
            s2v.push_back(R2(ti,rv[2]));
            s3v.push_back(R2(ti,rv[3]));
            s12v.push_back(R2(ti,rv[4]));
            s13v.push_back(R2(ti,rv[5]));
            s23v.push_back(R2(ti,rv[6]));
            s123v.push_back(R2(ti,rv[7]));
            stv.push_back(R2(ti,rv[8]));
         }
      }
      if (view()){
         mark(g,dc);
         Word fn=g.vis(dc.wrtOnShow);
         if (fn.dim()>0) vw.push_back(fn);
         cpmwait(rc.tWaitView);
      }
      if (img()||imgSteps.find(i)!=0){
         g.clr_();
         mark(g,dc);
         g.setText(Word("Lat3: t=")&cpm(t_));
         g.vis(true);
     }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("TrjLat3: run documentation");
      g.show(doc);
      doc.toGnuPlot("TrjLat3_run_documentation.txt");
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(8);
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s3v);
   res[4]=V<R2>(1,s12v);
   res[5]=V<R2>(1,s13v);
   res[6]=V<R2>(1,s23v);
   res[7]=V<R2>(1,s123v);
   res[8]=V<R2>(1,stv);
   if (tWaitRes>0){
      g.clr_();
      g.setAutoScl_(true);
      Word txt1("TrjLat3: expectation values and correlations");
      Word txt2(" of particle shifts");
      g.setText(txt1&txt2);
      g.mark(res);
      g.vis(dc.wrtOnShow);
      cpmwait(tWaitRes,2);
   }
   V<Word> res2(1,vw);
   return x2vv(res,res2);
}

x2vv TrjLat3::run_(Graph& g, RecordHandler const& rch)
{
   RunCon rc(rch,"run control");
   DisCon dc(rch,"display control");
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=dtSup(rc.nIter,rc.tWaitIter,rc.gammaIter);
      dt*=dtMax;
   }
   vector<R2> s1v;
   vector<R2> s2v;
   vector<R2> s3v;
   vector<R2> s12v;
   vector<R2> s13v;
   vector<R2> s23v;
   vector<R2> s123v;
   vector<R2> stv;
   vector<Word> vw;
   cpmdebug(rc.imgSteps);
   Vo<Z> imgSteps=rc.imgSteps;
   cpmmessage("before sorting");
   imgSteps.sort_();
   cpmmessage("sorting done");
   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R_Vector rv=stt();
            s1v.push_back(R2(ti,rv[1]));
            s2v.push_back(R2(ti,rv[2]));
            s3v.push_back(R2(ti,rv[3]));
            s12v.push_back(R2(ti,rv[4]));
            s13v.push_back(R2(ti,rv[5]));
            s23v.push_back(R2(ti,rv[6]));
            s123v.push_back(R2(ti,rv[7]));
            stv.push_back(R2(ti,rv[8]));
         }
      }
      if (view()){
         mark(g,dc);
         Word fn=g.vis(dc.wrtOnShow);
         if (fn.dim()>0) vw.push_back(fn);
         cpmwait(rc.tWaitView);
      }
      if (img()||imgSteps.find(i)!=0){
         g.clr_();
         mark(g,dc);
         g.setText(Word("Lat3: t=")&cpm(t_));
         g.vis(true);
     }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("TrjLat3: run documentation");
      g.show(doc);
      doc.toGnuPlot("TrjLat3_run_documentation.txt");
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(8);
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s3v);
   res[4]=V<R2>(1,s12v);
   res[5]=V<R2>(1,s13v);
   res[6]=V<R2>(1,s23v);
   res[7]=V<R2>(1,s123v);
   res[8]=V<R2>(1,stv);
   if (tWaitRes>0){
      g.clr_();
      g.setAutoScl_(true);
      Word txt1("TrjLat3: expectation values and correlations");
      Word txt2(" of particle shifts");
      g.setText(txt1&txt2);
      g.mark(res);
      g.vis(dc.wrtOnShow);
      cpmwait(tWaitRes,2);
   }
   V<Word> res2(1,vw);
   return x2vv(res,res2);
}

////////////////////////////////////// class Sep2 ////////////////////////


SysLat1 Sep2::vInt(SysLat1 const& y1, SysLat1 const& y2,
   DynDat const& dyn1, DynDat const& dyn2, IntActDat const& ia12)
{
   Lat1 lat1=y1.lat(), lat2=y2.lat();
   Z d1=lat1.d(), d2=lat2.d();
   R q1=dyn1.q(),q2=dyn2.q();
   R q12=q1*q2;
   SysLat1 res(lat2); // zero vector
   C fac=C::Ibar;
   for (Z j=1;j<=d2;++j){
      R xj=lat2.pos(j);
      R resj=0;
      for (Z i=1;i<=d1;++i){
         R xi=lat1.pos(i);
         R rij=ia12.dis(xi,xj);
         R pij=y1.cui(i).absSqr()*q12*ia12.pot(rij);
         resj+=pij;
      }
      res.cui(j)=fac*y2.cui(j)*resj;
   }
   return res;
}

X2<SysLat1, SysLat1> Sep2::dot(SysLat1 const& x1, SysLat1 const& x2,
   DynDat const& dd1, DynDat const& dd2,IntActDat const& dd12)
      // returns the velocities determined by the arguments
      // very transparent form for generalisation to more particles.
{
   SysLat1 v1(x1),v2(x2);
   v1.dot_(dd1);
   v2.dot_(dd2);
   SysLat1 dv1=vInt(x2,x1,dd1,dd2,dd12);
   SysLat1 dv2=vInt(x1,x2,dd1,dd2,dd12);
   v1+=dv1;
   v2+=dv2;
   return X2<SysLat1, SysLat1>(v1,v2);
}

C Sep2::hInt()const
//: h interaction (h for hamiltonian)
{
   R res=vIntScl(x1_,x2_,dd1_,dd2_,dd12_);
   res+=x1_.enrHomFil(dd1_);
   res+=x2_.enrHomFil(dd2_);
   return C(res);
}

void Sep2::stepGam_(R dt)
{
   R tau=dt*0.5;
   t_+=tau;
   x1_+=v1_*tau;
   x2_+=v2_*tau;
   X2<SysLat1,SysLat1> vv=dot(x1_,x2_,dd1_,dd2_,dd12_);
   v1_=vv.c1()*2-v1_;
   v2_=vv.c2()*2-v2_;
   t_+=tau;
   x1_+=v1_*tau;
   x2_+=v2_*tau;
}

void Sep2::step_(R dt)
{
   // fine, avoids oscillatory behavior
   // of the diagnostic quantities norm and energy
      R tau=dt*0.5;
      stepGam_(tau);
      stepGam_(tau);
}

// master implementation of run_ functions
V< V<R2> > Sep2::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=SysLat2(x1_,x2_).
         dtSup(dd1_,dd2_,dd12_,rc.nIter,rc.tWaitIter,rc.gammaIter);
      dt*=dtMax;
   }
   vector<R2> s1v, s2v, s12v;

   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R_Vector rv=stt();
            s1v.push_back(R2(ti,rv[1]));
            s2v.push_back(R2(ti,rv[2]));
            s12v.push_back(R2(ti,rv[3]));
         }
      }
      if (view()){
         mark(g,dc);
         g.vis();
         cpmwait(rc.tWaitView);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("Sep2: run documentation");
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(3);
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s12v);
   if (tWaitRes>0){
      g.clr_();
      g.setAutoScl_(true);
      g.setText("Sep2: res documentation");
      g.mark(res);
      g.vis();
      cpmwait(tWaitRes,2);
   }
   return res;
}

void Sep2::mark(Graph& g, DisCon const& dc)const
{
   using CpmLinAlg::R2;
   using CpmArrays::V;
   using namespace CpmImaging;

   V< V<R2> > dat(6);
   Lat1 lat1=x1_.lat();
   Z d1=lat1.d();
   R xi=lat1.fir();
   V<R2> re1(d1);
   V<R2> im1(d1);
   V<R2> ab1(d1);
   Z i;
   for (i=1;i<=d1;++i){
      C zi=x1_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re1.cui(ip)=rei;
      im1.cui(ip)=imi;
      ab1.cui(ip)=abi;
      lat1.next_1(xi);
   }
   dat[1]=re1;
   dat[2]=im1;
   dat[3]=ab1;

   Lat1 lat2=x2_.lat();
   Z d2=lat2.d();
   xi=lat2.fir();
   V<R2> re2(d2);
   V<R2> im2(d2);
   V<R2> ab2(d2);
   for (i=1;i<=d2;++i){
      C zi=x2_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re2.cui(ip)=rei;
      im2.cui(ip)=imi;
      ab2.cui(ip)=abi;
      lat2.next_1(xi);
   }
   dat[4]=re2;
   dat[5]=im2;
   dat[6]=ab2;

   V<Color> cl(6);
   cl[1]=RED;  // color for real part
   cl[2]=GREEN;   // color for imaginary part
   cl[3]=BLACK; // color for absolute value
   cl[4]=LIGHTRED;  // color for real part
   cl[5]=LIGHTGREEN;   // color for imaginary part
   cl[6]=LIGHTGRAY; // color for absolute value
   g.setAutoScl_(dc.autoScl);
   g.mark(dat,cl,dc.secFac);
   g.addText(Word("t=")&cpm(t_));
}

////////////////////////////////////// class Sep3 ////////////////////////

X3<SysLat1, SysLat1, SysLat1> Sep3::dot(SysLat1 const& x1,
   SysLat1 const& x2, SysLat1 const& x3,
   DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
   IntActDat const& dd12)
{
   SysLat1 v1(x1),v2(x2),v3(x3);
   v1.dot_(dd1);
   v2.dot_(dd2);
   v3.dot_(dd3);
   SysLat1 dv1=Sep2::vInt(x2,x1,dd1,dd2,dd12)
      +Sep2::vInt(x3,x1,dd1,dd3,dd12);
   SysLat1 dv2=Sep2::vInt(x1,x2,dd1,dd2,dd12)
      +Sep2::vInt(x3,x2,dd2,dd3,dd12);
   SysLat1 dv3=Sep2::vInt(x1,x3,dd1,dd3,dd12)
      +Sep2::vInt(x2,x3,dd2,dd3,dd12);
   v1+=dv1;
   v2+=dv2;
   v3+=dv3;
   return X3<SysLat1, SysLat1,SysLat1>(v1,v2,v3);
}

C Sep3::hInt()const
{
   R res=0;
   res+=Sep2::vIntScl(x1_,x2_,dd1_,dd2_,dd12_);
   res+=Sep2::vIntScl(x1_,x3_,dd1_,dd3_,dd12_);
   res+=Sep2::vIntScl(x2_,x3_,dd2_,dd3_,dd12_);
   res+=x1_.enrHomFil(dd1_);
   res+=x2_.enrHomFil(dd2_);
   res+=x3_.enrHomFil(dd3_);
   return C(res);
}

void Sep3::stepGam_(R dt)
{
   R tau=dt*0.5;
   t_+=tau;
   x1_+=v1_*tau;
   x2_+=v2_*tau;
   x3_+=v3_*tau;
   X3<SysLat1,SysLat1,SysLat1> vvv=dot(x1_,x2_,x3_,dd1_,dd2_,dd3_,dd12_);
   v1_=vvv.c1()*2-v1_;
   v2_=vvv.c2()*2-v2_;
   v3_=vvv.c3()*2-v3_;
   t_+=tau;
   x1_+=v1_*tau;
   x2_+=v2_*tau;
   x3_+=v3_*tau;
}

void Sep3::step_(R dt)
{
   const bool correctNorm=false;
      // we choose false so that norm conservation of the method
      // can not arise as an artifact of implementation

   const bool twoSteps=true; // true is proper
      // clean way of verifying the oscillatory behavior of 2 steps
      // period if the combination of two steps int one are missing

   if (twoSteps){
      R tau=dt*0.5;
      stepGam_(tau);
      stepGam_(tau);
   }
   else stepGam_(dt);

   if (correctNorm){
      R eps=1e-5; // controls the rigidity of control
      R n1=x1_.abs(), n2=x2_.abs(), n3=x3_.abs();
      R d1=cpmabs(n1-1), d2=cpmabs(n2-1), d3=cpmabs(n3-1);
      if (d1>eps){
         R corr=R(1)/n1; x1_*=corr; v1_*=corr;
      }
      if (d2>eps){
         R corr=R(1)/n2; x2_*=corr; v2_*=corr;
      }
      if (d3>eps){
         R corr=R(1)/n3; x3_*=corr; v3_*=corr;
      }
   }
}

x2vv Sep3::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      dt*=dtSup( rc.nIter,rc.tWaitIter,rc.gammaIter);
         // now based on SysLat3 as a whole
   }
   Vo<Z> imgSteps=rc.imgSteps;
   imgSteps.sort_();
   cpmdebug(imgSteps);
   vector<Word> vw;
   vector<R2> s1v, s2v, s3v, s12v, s13v, s23v, s123v,stv; //st = s total
      // s1+s2+s3, 't' is used similar to '123'
   vector<R2> s1vc, s2vc, s3vc, s12vc, s13vc, s23vc, s123vc, stvc;
      // means for storing output from stt()
   Z i;
//   Color cGridMem=g.getGridColor();
   for (i=1;i<=nSteps;++i){
      step_(dt); // promotes the wave functions
      cs_.step_(dt,dd1_,dd2_,dd3_,dd12_);
         // promotes the classical particles
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
           CPM_SC_RUN2
           R_Vector rv=stt();
           s1v.push_back(R2(ti,rv[1]));
           s2v.push_back(R2(ti,rv[2]));
           s3v.push_back(R2(ti,rv[3]));
           s12v.push_back(R2(ti,rv[4]));
           s13v.push_back(R2(ti,rv[5]));
           s23v.push_back(R2(ti,rv[6]));
           s123v.push_back(R2(ti,rv[7]));
           stv.push_back(R2(ti,rv[8]));

           R_Vector rvc=sttCls(); // classical motion
           s1vc.push_back(R2(ti,rvc[1]));
           s2vc.push_back(R2(ti,rvc[2]));
           s3vc.push_back(R2(ti,rvc[3]));
           s12vc.push_back(R2(ti,rvc[4]));
           s13vc.push_back(R2(ti,rvc[5]));
           s23vc.push_back(R2(ti,rvc[6]));
           s123vc.push_back(R2(ti,rvc[7]));
           stvc.push_back(R2(ti,rvc[8]));
         }
      }
      if (view()){
         mark(g,dc);
         Word fn=g.vis(dc.wrtOnShow);
         if (fn.dim()>0) vw.push_back(fn);
         cpmwait(rc.tWaitView);
      }
      if (img()||imgSteps.find(i)!=0){
         g.clr_(WHITE);
         DisCon dcMod=dc;
         dcMod.areaSty=false;
         g.setThcLin_(true);
         mark(g,dcMod);
         g.vis(true);
         g.clr_();
         g.setThcLin_(false);
         dcMod.areaSty=true;
         mark(g,dcMod);
         g.setText(Word("Sep3: t=")&cpm(t_));
         g.vis(true);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      Word txt1("Sep3: relative changes of norm(black), real energy");
      Word txt2("(violet), imaginary energy(blue)");
      g.setText(txt1&txt2);
      g.show(doc);
      doc.toGnuPlot("Sep3_run_documentation.txt");
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(16);
// quantum data
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s3v);
   res[4]=V<R2>(1,s12v);
   res[5]=V<R2>(1,s13v);
   res[6]=V<R2>(1,s23v);
   res[7]=V<R2>(1,s123v);
   res[8]=V<R2>(1,stv);
// classical data
   res[9]=V<R2>(1,s1vc);
   res[10]=V<R2>(1,s2vc);
   res[11]=V<R2>(1,s3vc);
   res[12]=V<R2>(1,s12vc);
   res[13]=V<R2>(1,s13vc);
   res[14]=V<R2>(1,s23vc);
   res[15]=V<R2>(1,s123vc);
   res[16]=V<R2>(1,stvc);
   if (tWaitRes>0){
      g.clr();
      g.setAutoScl_(true);
      if (dc.showCor){
         Word txt1("Sep3: expectation values and correlations");
         Word txt2(" of particle shifts");
         g.setText(txt1&txt2);
         g.mark(res);
      }
      else{
         V< V<R2> > resSmall(2);
         resSmall[1]=res[8];
         resSmall[2]=res[16];
         Word txt1("Sep3: sum of particle shifts, QM and classical");
         g.setText(txt1);
         g.mark(resSmall);
      }
      g.vis(dc.wrtOnShow);
      cpmwait(tWaitRes,2);
   }
   if (!dc.showCls) res=res.resize(8);
      // then TrjLat3 and Sep3 return data of equal format which in some
      // applications is required
   V<Word> res2(1,vw);
   return x2vv(res,res2);
}

x2vv Sep3::run_(Graph& g, RecordHandler const& rch)
{
   RunCon rc(rch,"run control");
   DisCon dc(rch,"display control");
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      dt*=dtSup( rc.nIter,rc.tWaitIter,rc.gammaIter);
         // now based on SysLat3 as a whole
   }
   Vo<Z> imgSteps=rc.imgSteps;
   imgSteps.sort_();
   cpmdebug(imgSteps);
   vector<Word> vw;
   vector<R2> s1v, s2v, s3v, s12v, s13v, s23v, s123v,stv; //st = s total
      // s1+s2+s3, 't' is used similar to '123'
   vector<R2> s1vc, s2vc, s3vc, s12vc, s13vc, s23vc, s123vc, stvc;
      // means for storing output from stt()
   Z i;
//   Color cGridMem=g.getGridColor();
   for (i=1;i<=nSteps;++i){
      step_(dt); // promotes the wave functions
      cs_.step_(dt,dd1_,dd2_,dd3_,dd12_);
         // promotes the classical particles
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
           CPM_SC_RUN2
           R_Vector rv=stt();
           s1v.push_back(R2(ti,rv[1]));
           s2v.push_back(R2(ti,rv[2]));
           s3v.push_back(R2(ti,rv[3]));
           s12v.push_back(R2(ti,rv[4]));
           s13v.push_back(R2(ti,rv[5]));
           s23v.push_back(R2(ti,rv[6]));
           s123v.push_back(R2(ti,rv[7]));
           stv.push_back(R2(ti,rv[8]));

           R_Vector rvc=sttCls(); // classical motion
           s1vc.push_back(R2(ti,rvc[1]));
           s2vc.push_back(R2(ti,rvc[2]));
           s3vc.push_back(R2(ti,rvc[3]));
           s12vc.push_back(R2(ti,rvc[4]));
           s13vc.push_back(R2(ti,rvc[5]));
           s23vc.push_back(R2(ti,rvc[6]));
           s123vc.push_back(R2(ti,rvc[7]));
           stvc.push_back(R2(ti,rvc[8]));
         }
      }
      if (view()){
         mark(g,dc);
         Word fn=g.vis(dc.wrtOnShow);
         if (fn.dim()>0) vw.push_back(fn);
         cpmwait(rc.tWaitView);
      }
      if (img()||imgSteps.find(i)!=0){
         if (imgSteps.find(i)!=0){
            g.clr_(WHITE);
            DisCon dcMod=dc;
            dcMod.areaSty=false;
            g.setThcLin_(true);
            mark(g,dcMod);
            g.vis(true);
            g.clr_();
            g.setThcLin_(false);
           // dcMod.areaSty=true;
         }
    //     mark(g,dcMod);
         mark(g,dc);
         g.setText(Word("Sep3: t=")&cpm(t_));
         g.vis(true);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      Word txt1("Sep3: relative changes of norm(black), real energy");
      Word txt2("(violet), imaginary energy(blue)");
      g.setText(txt1&txt2);
      g.show(doc);
      doc.toGnuPlot("Sep3_run_documentation.txt");
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(16);
// quantum data
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s3v);
   res[4]=V<R2>(1,s12v);
   res[5]=V<R2>(1,s13v);
   res[6]=V<R2>(1,s23v);
   res[7]=V<R2>(1,s123v);
   res[8]=V<R2>(1,stv);
// classical data
   res[9]=V<R2>(1,s1vc);
   res[10]=V<R2>(1,s2vc);
   res[11]=V<R2>(1,s3vc);
   res[12]=V<R2>(1,s12vc);
   res[13]=V<R2>(1,s13vc);
   res[14]=V<R2>(1,s23vc);
   res[15]=V<R2>(1,s123vc);
   res[16]=V<R2>(1,stvc);
   if (tWaitRes>0){
      g.clr();
      g.setAutoScl_(true);
      if (dc.showCor){
         Word txt1("Sep3: expectation values and correlations");
         Word txt2(" of particle shifts");
         g.setText(txt1&txt2);
         g.mark(res);
      }
      else{
         V< V<R2> > resSmall(2);
         resSmall[1]=res[8];
         resSmall[2]=res[16];
         Word txt1("Sep3: sum of particle shifts, QM and classical");
         g.setText(txt1);
         g.mark(resSmall);
      }
      g.vis(dc.wrtOnShow);
      cpmwait(tWaitRes,2);
   }
   if (!dc.showCls) res=res.resize(8);
      // then TrjLat3 and Sep3 return data of equal format which in some
      // applications is required
   V<Word> res2(1,vw);
   return x2vv(res,res2);
}


void Sep3::mark(Graph& g, DisCon const& dc)const
{
   if (dc.areaSty){
      SysLat3(x1_,x2_,x3_).mark(g,dc);
      return;
   }
   using CpmLinAlg::R2;
   using CpmArrays::V;
   using namespace CpmImaging;

   V< V<R2> > dat(9);
   Lat1 lat1=x1_.lat();
   Lat1 lat2=x2_.lat();
   Lat1 lat3=x3_.lat();

   Z d1=lat1.d();
   Z d2=lat2.d();
   Z d3=lat3.d();

   R xi=lat1.fir();
   V<R2> re1(d1);
   V<R2> im1(d1);
   V<R2> ab1(d1);
   Z i;
   for (i=1;i<=d1;++i){
      C zi=x1_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re1.cui(ip)=rei;
      im1.cui(ip)=imi;
      ab1.cui(ip)=abi;
      lat1.next_1(xi);
   }
   dat[1]=re1;
   dat[2]=im1;
   dat[3]=ab1;

   xi=lat2.fir();
   V<R2> re2(d2);
   V<R2> im2(d2);
   V<R2> ab2(d2);
   for (i=1;i<=d2;++i){
      C zi=x2_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re2.cui(ip)=rei;
      im2.cui(ip)=imi;
      ab2.cui(ip)=abi;
      lat2.next_1(xi);
   }
   dat[4]=re2;
   dat[5]=im2;
   dat[6]=ab2;

   xi=lat3.fir();
   V<R2> re3(d3);
   V<R2> im3(d3);
   V<R2> ab3(d3);
   for (i=1;i<=d3;++i){
      C zi=x3_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re3.cui(ip)=rei;
      im3.cui(ip)=imi;
      ab3.cui(ip)=abi;
      lat3.next_1(xi);
   }
   dat[7]=re3;
   dat[8]=im3;
   dat[9]=ab3;

   V<Color> cl(9);
   cl[1]=RED;  // color for real part
   cl[2]=GREEN;   // color for imaginary part
   cl[3]=BLACK; // color for absolute value
   cl[4]=LIGHTRED;  // color for real part
   cl[5]=LIGHTGREEN;   // color for imaginary part
   cl[6]=LIGHTGRAY; // color for absolute value
   cl[7]=ORANGE;  // color for real part
   cl[8]=CYAN;   // color for imaginary part
   cl[9]=DARKGRAY; // color for absolute value
   g.setAutoScl_(dc.autoScl);
   g.mark(dat,cl,dc.secFac);

   R x1=cs_.x1(), x2=cs_.x2(), x3=cs_.x3();
   V<R> xList{x1,x2,x3}; // classical positions
   V<Color> xColors{RED,GREEN,BLUE};
   R y1=dd1_.x0(), y2=dd2_.x0(), y3=dd3_.x0();
   V<R> yList{y1,y2,y3}; // equilibrium positions
   V<Color> yColors{DARKGRAY,DARKGRAY,DARKGRAY};
   R z1=x1_.s(), z2=x2_.s(), z3=x3_.s();
   V<R> zList{z1,z2,z3}; // quantum positions
   V<Color> zColors{LIGHTRED,LIGHTGREEN,LIGHTBLUE};
   g.mark(xList,xColors);
   g.mark(zList,zColors);
   g.mark(yList,yColors);
// show the time on the graph
   Word tText=Word("t=")&cpm(t_);
   R xT=g.getGridTextPos()[1], yT=0, hT=g.getGridTextHeg();
      // yT=0 for automatic y
   g.addText(tText,xT,yT,BLACK,hT);
}

////////////////////////////////////// class Fun3 ////////////////////////
bool Fun3::simplify_=false;
R Fun3::epsilon_=0;

void Fun3::step_(R dt)
{
//   bool correctNorm=true;
//   bool simplification=false;
   // no simplification should be allowed
   Z mL=2;
   Word loc("Fun3::step_(R)");
   CPM_MA

   SysLat3 x(x1_,x2_,x3_);
   TrjLat3 tx(dd1_,dd2_,dd3_,dd12_,x,t_);
   tx.step_(dt);
   SysLat3 xt=tx.x();

   TrjLat2 tx12(dd1_,dd2_,dd12_,x1_,x2_,t_);
   TrjLat2 tx23(dd2_,dd3_,dd12_,x2_,x3_,t_);
   TrjLat2 tx31(dd3_,dd1_,dd12_,x3_,x1_,t_);

   if (!simplify_){
      // surprising that this step has nearly no influence
      tx12.step_(dt);
      tx23.step_(dt);
      tx31.step_(dt);
   }

   SysLat2 x12=tx12.x();
   SysLat2 x23=tx23.x();
   SysLat2 x31=tx31.x();

   x1_=xt.red23(x23);
   x2_=xt.red31(x31);
   x3_=xt.red12(x12);

   if (epsilon_>=0){
      R eps=epsilon_; // controls the rigidity of control
      // to do the correction always seems to be not worse
      // So one should have a small value, e.g. 1e-9 here
      R n1=x1_.abs(), n2=x2_.abs(), n3=x3_.abs();
      R d1=cpmabs(n1-1), d2=cpmabs(n2-1), d3=cpmabs(n3-1);
      if (d1>eps){
         R corr=R(1)/n1; x1_*=corr;
      }
      if (d2>eps){
         R corr=R(1)/n2; x2_*=corr;
      }
      if (d3>eps){
         R corr=R(1)/n3; x3_*=corr;
      }
   }
   t_+=dt;
   CPM_MZ
}

x2vv Fun3::run_(Graph& g, RunCon& rc, DisCon const& dc)
// many additions and modifications made to Sep3::run_
// are not carried out here (2008-03-19)
{
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dt12=SysLat2(x1_,x2_).
         dtSup(dd1_,dd2_,dd12_,rc.nIter,rc.tWaitIter,rc.gammaIter);
      R dt13=SysLat2(x1_,x3_).
         dtSup(dd1_,dd3_,dd12_,rc.nIter,rc.tWaitIter,rc.gammaIter);
      R dt23=SysLat2(x2_,x3_).
         dtSup(dd2_,dd3_,dd12_,rc.nIter,rc.tWaitIter,rc.gammaIter);
      R dtMax=CpmRootX::inf<R>(dt12,dt13,dt23);
      dt*=dtMax;
      cpmdata<<"dt12="<<dt12<<", dt13="<<dt13<<", dt23="<<dt23<<
         ", dtMax="<<dtMax<<endl;
   }
   vector<Word> vw;
   vector<R2> s1v, s2v, s3v, s12v, s13v, s23v, s123v, stv;
   vector<R2> s1vc, s2vc, s3vc, s12vc, s13vc, s23vc, s123vc, stvc;
      // means for storing output from stt()
   Z i;
   for (i=1;i<=nSteps;++i){
      step_(dt); // promotes the wave functions
      cs_.step_(dt,dd1_,dd2_,dd3_,dd12_);
         // promotes the classical particles
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
           CPM_SC_RUN2
           R_Vector rv=stt();
           s1v.push_back(R2(ti,rv[1]));
           s2v.push_back(R2(ti,rv[2]));
           s3v.push_back(R2(ti,rv[3]));
           s12v.push_back(R2(ti,rv[4]));
           s13v.push_back(R2(ti,rv[5]));
           s23v.push_back(R2(ti,rv[6]));
           s123v.push_back(R2(ti,rv[7]));
           stv.push_back(R2(ti,rv[8]));

           R_Vector rvc=sttCls(); // classical motion
           s1vc.push_back(R2(ti,rvc[1]));
           s2vc.push_back(R2(ti,rvc[2]));
           s3vc.push_back(R2(ti,rvc[3]));
           s12vc.push_back(R2(ti,rvc[4]));
           s13vc.push_back(R2(ti,rvc[5]));
           s23vc.push_back(R2(ti,rvc[6]));
           s123vc.push_back(R2(ti,rvc[7]));
           stvc.push_back(R2(ti,rvc[8]));
         }
      }
      if (view()){
         stt();
         mark(g,dc);
         Word fn=g.vis(dc.wrtOnShow);
         if (fn.dim()>0) vw.push_back(fn);
         cpmwait(rc.tWaitView);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      Word txt1("Fun3: relative changes of norm(black), real energy");
      Word txt2("(violet), imaginary energy(blue)");
      g.setText(txt1&txt2);
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(16);
// quantum data
   res[1]=V<R2>(1,s1v);
   res[2]=V<R2>(1,s2v);
   res[3]=V<R2>(1,s3v);
   res[4]=V<R2>(1,s12v);
   res[5]=V<R2>(1,s13v);
   res[6]=V<R2>(1,s23v);
   res[7]=V<R2>(1,s123v);
   res[8]=V<R2>(1,stv);
// classical data
   res[9]=V<R2>(1,s1vc);
   res[10]=V<R2>(1,s2vc);
   res[11]=V<R2>(1,s3vc);
   res[12]=V<R2>(1,s12vc);
   res[13]=V<R2>(1,s13vc);
   res[14]=V<R2>(1,s23vc);
   res[15]=V<R2>(1,s123vc);
   res[16]=V<R2>(1,stvc);
   if (tWaitRes>0){
      g.clr();
      g.setAutoScl_(true);
      if (dc.showCor){
         Word txt1("Fun3: expectation values and correlations");
         Word txt2(" of particle shifts");
         g.setText(txt1&txt2);
         g.mark(res);
      }
      else{
         V< V<R2> > resSmall(2);
         resSmall[1]=res[8];
         resSmall[2]=res[16];
         Word txt1("Fun3: sum of particle shifts, QM and classical");
         g.setText(txt1);
         g.mark(resSmall);
      }
      g.vis();
      cpmwait(tWaitRes,2);
   }
   if (!dc.showCls) res=res.resize(8);
      // then TrjLat3 and Fun3 return data of equal format which in some
      // applications is required
   V<Word> res2(1,vw);
   return x2vv(res,res2);
}

void Fun3::mark(Graph& g, DisCon const& dc)const
{
   if (dc.areaSty){
    SysLat3(x1_,x2_,x3_).mark(g,dc);
    return;
   }
   using CpmLinAlg::R2;
   using CpmArrays::V;
   using namespace CpmImaging;

   V< V<R2> > dat(9);
   Lat1 lat1=x1_.lat();
   Lat1 lat2=x2_.lat();
   Lat1 lat3=x3_.lat();

   Z d1=lat1.d();
   Z d2=lat2.d();
   Z d3=lat3.d();

   R xi=lat1.fir();
   V<R2> re1(d1);
   V<R2> im1(d1);
   V<R2> ab1(d1);
   Z i;
   for (i=0;i<d1;++i){
      C zi=x1_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re1.cui(ip)=rei;
      im1.cui(ip)=imi;
      ab1.cui(ip)=abi;
      lat1.next_1(xi);
   }
   dat[1]=re1;
   dat[2]=im1;
   dat[3]=ab1;

   xi=lat2.fir();
   V<R2> re2(d2);
   V<R2> im2(d2);
   V<R2> ab2(d2);
   for (i=0;i<d2;++i){
      C zi=x2_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re2.cui(ip)=rei;
      im2.cui(ip)=imi;
      ab2.cui(ip)=abi;
      lat2.next_1(xi);
   }
   dat[4]=re2;
   dat[5]=im2;
   dat[6]=ab2;

   xi=lat3.fir();
   V<R2> re3(d3);
   V<R2> im3(d3);
   V<R2> ab3(d3);
   for (i=0;i<d3;++i){
      C zi=x3_.cui(i);
      R2 rei(xi,zi.real());
      R2 imi(xi,zi.imag());
      R2 abi(xi,zi.abs());
      Z ip=i+1;
      re3.cui(ip)=rei;
      im3.cui(ip)=imi;
      ab3.cui(ip)=abi;
      lat3.next_1(xi);
   }
   dat[7]=re3;
   dat[8]=im3;
   dat[9]=ab3;

   V<Color> cl(9);
   cl[1]=RED;  // color for real part
   cl[2]=GREEN;   // color for imaginary part
   cl[3]=BLACK; // color for absolute value
   cl[4]=LIGHTRED;  // color for real part
   cl[5]=LIGHTGREEN;   // color for imaginary part
   cl[6]=LIGHTGRAY; // color for absolute value
   cl[7]=ORANGE;  // color for real part
   cl[8]=CYAN;   // color for imaginary part
   cl[9]=DARKGRAY; // color for absolute value
   g.setAutoScl_(dc.autoScl);
   g.mark(dat,cl,dc.secFac);

   R x1=cs_.x1(), x2=cs_.x2(), x3=cs_.x3();
   V<R> xList{x1,x2,x3}; // classical positions
   V<Color> xColors{RED,GREEN,BLUE};
   R y1=dd1_.x0(), y2=dd2_.x0(), y3=dd3_.x0();
   V<R> yList{y1,y2,y3}; // equilibrium positions
   V<Color> yColors{DARKGRAY,DARKGRAY,DARKGRAY};
   R z1=x1_.s(), z2=x2_.s(), z3=x3_.s();
   V<R> zList{z1,z2,z3}; // quantum positions
   V<Color> zColors{LIGHTRED,LIGHTGREEN,LIGHTBLUE};
   g.mark(xList,xColors);
   g.mark(zList,zColors);
   g.mark(yList,yColors);
}

////////////////////////////////// class CM3 /////////////////////////////

void CM3::step(R dt, R& x1, R& x2, R& x3, R& v1, R& v2, R& v3,
   DynDat const& dd1, DynDat const& dd2, DynDat const& dd3,
   IntActDat const& dd12)
{
   R q1=dd1.q(), q2=dd2.q(), q3=dd3.q();
   R m1=dd1.m(), m2=dd2.m(), m3=dd3.m();
   R tau=dt*0.5;
   x1+=v1*tau; x2+=v2*tau; x3+=v3*tau;
   R f1=dd1.force(x1) + q1*q2*dd12.fi(x1-x2) + q1*q3*dd12.fi(x1-x3);
   R f2=dd2.force(x2) + q2*q1*dd12.fi(x2-x1) + q2*q3*dd12.fi(x2-x3);
   R f3=dd3.force(x3) + q3*q1*dd12.fi(x3-x1) + q3*q2*dd12.fi(x3-x2);
   R a1=f1/m1, a2=f2/m2, a3=f3/m3;
   v1+=a1*dt; v2+=a2*dt; v3+=a3*dt;
   x1+=v1*tau; x2+=v2*tau; x3+=v3*tau;
}

/////////////////////////// class SpinN /////////////////////////////////

Z4 SpinN::neg(Z i)const
{
   Z i1=i/n_, i2=i%n_; // 0 <= i1 < m_, 0 <= i2 < n_
      // i=i1*n_+i2
   Z i1p = i1==m_-1 ? 0 : i1+1;
   Z i1m = i1==0 ? m_-1 : i1-1;

   Z i2p = i2==n_-1 ? 0 : i2+1;
   Z i2m = i2==0 ? n_-1 : i2-1;

   Z k1=i1p*n_+i2;
   Z k2=i1m*n_+i2;
   Z k3=i1*n_+i2p;
   Z k4=i1*n_+i2m;
   return Z4(k1,k2,k3,k4);
}

void SpinN::dot_()
{
   Word loc("SpinN::dot_()");
   Z mL=2;
   CPM_MA
 //  Outside os=CYCLIC;
   Z np=vx_.dim();
   vv_=V<C2>(np);
   for (Z i=1;i<=np;++i){
      C2 ci=vx_.cui(i);
      Z4 ni=neg(i);
      C2 ri=ia_(ci) + ia_(vx_(ni.x1),ci)+ia_(vx_(ni.x2),ci)
         +ia_(vx_(ni.x3),ci)+ia_(vx_(ni.x4),ci);
      vv_.cui(i)=ri*C::Ibar*rf_;
   }
   CPM_MZ
}

C SpinN::hInt()const
{
   C res;
//   Outside os=CYCLIC;
   Z np=vx_.dim();
   for (Z i=1;i<=np;++i){
      Z4 ni=neg(i);
      C2 ci=vx_.cui(i);
      res+= (ia_.avr(vx_(ni.x1),ci)+ia_.avr(vx_(ni.x3),ci)
        +ia_.avr(vx_(ni.x2),ci)+ia_.avr(vx_(ni.x4),ci));
   }
   return res*0.5;
}

R SpinN::hInt_i(Z i)const
{
   Outside os=CYCLIC;
   C2 ci=vx_[i];
   C z=ia_.avr(vx_(i-1,os),ci)
      + ia_.avr(vx_(i+1,os),ci);
   return z[1];
}

R SpinN::eSep_i(Z i)const
{
   C2 ci=vx_[i];
   return (ci|ia_(ci))[1];
}

R SpinN::eTot_i(Z i)const
{
   return eSep_i(i)+hInt_i(i);
}

C SpinN::eSep()const
{
   C res;
   Z np=vx_.dim();
   for (Z i=1;i<=np;++i){
      C2 ci=vx_.cui(i);
      res+=(ci|ia_(ci));
   }
   return res;
}

R3 SpinN::sig(Z i)const
{
   C2 ci=vx_.cui(i);
   C v1=(ci|CC2::sig1*ci);
   C v2=(ci|CC2::sig2*ci);
   C v3=(ci|CC2::sig3*ci);
   return R3(v1[1],v2[1],v3[1]);
}

namespace{ // anonymous
   const R xg=-0.5;
   const R yg=0.5*sqrt(3.);
   const R2 er(1,0);
   const R2 eg(xg,yg);
   const R2 eb(xg,-yg);
      // er+eg+eb=0,
      // er.abs()==eg.abs()==eb.abs()==1
   C transform(R3 const& r){
      R xr=(r[1]+1);
      R xg=(r[2]+1);
      R xb=(r[3]+1);
      R2 res=er*xr+eg*xg+eb*xb;
      return C(res[1],res[2]);
   }
}

V<C2> SpinN::col()const
{
   Z np=vx_.dim();
   if (np==0) return V<C2>(0);
   V<C2> res(np);
   for (Z i=1;i<=np;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res[i]=C2(z,z);
   }
   return res;
}

void SpinN::stepGam_(R dt)
{
   Z i,np=parNum();
   R tau=dt*0.5;

   t_+=tau;
   for (i=1;i<=np;++i) vx_.cui(i)+=vv_.cui(i)*tau;

   V<C2> vMem=vv_;
   dot_();
   for (i=1;i<=np;++i) vv_.cui(i)=vv_.cui(i)*2-vMem.cui(i);

   for (i=1;i<=np;++i) vx_.cui(i)+=vv_.cui(i)*tau;
   t_+=tau;
}


void SpinN::step_()
{
//   cout<<"SpinN::step started"<<endl;
   const R facRed=0.75;
   const R facInc=1.1;
   const R epsLow=0.1;
   V<C2> xMem=vx_;
   R tau=dt_*0.5;
   stepGam_(tau);
   stepGam_(tau);
   if (eps_>0){ // should not be active in the normal case ???
      // seems to be needed nevertheless since otherwise even harmless
      // configurations explode after having evolved rather regularly over
      // a long time. Together with time step adjustment this is much more
      // acceptable than before.
      // Conserved reversibility for eps_=1e-6 for the helicopter image
      // and seemed to fail reversibility for eps_=1e-4. In this case
      // the initial image was nevertheless recovered, but already in the
      // first third of the run.
      Z i;
      R sup=0;
      for (i=vx_.b();i<=vx_.e();++i){
         R ni=vx_[i].abs();
         R si=cpmabs(ni-1);
         if (si>sup) sup=si;
      }
      if (sup>eps_){ // do the step again with smaller dt
         t_-=dt_;
         dt_*=facRed;
         vx_=xMem;
         for (i=vx_.b();i<=vx_.e();++i) vx_[i].nor_();
         dot_();
         tau=dt_*0.5;
         stepGam_(tau);
         stepGam_(tau);
      }
      else if (sup<epsLow*eps_){ dt_*=facInc; }
      else ;
   } // eps_>0
   sc_++;
//   cout<<"SpinN::step done"<<endl;

}

namespace{

void lineMark_1(Graph& g, V<C2> const& vx, DisCon const& dc, Z met=1)
{
   using CpmImaging::Color;
   static Z jShift=1;
   Z gw=g.w();
   Z gh=g.h();
   Z np=vx.dim();
   Z pix=gh/np;
   Z mult_i=pix;
   Z mult_j=pix;
   Z js=pix,is=pix;
   Z i,j;
   Z d_=np*mult_i;
   jShift+=js;

  // Z iState=0;
   Z iState=1;
   Z iCounter=0;
   Z iEff=is;
   Color c;

   for (i=1;i<=d_;i++){
      iEff++;
      if (iCounter==0){
         C val;
         if (met==1){
            val=vx.cui(iState).blc();
         }
         else{
            val=vx.cui(iState)[1];
         }
         c=Color(val);
      }
      iCounter++;
      if (iCounter==mult_i){
         iCounter=0; // triggers new evaluation of c in the next loop
         iState++;
      }
      for (j=0;j<mult_j;j++){
         g.put(Z2(iEff,(j+jShift)%gw),c);
      }
   }
}

}

void SpinN::mark(Graph& g, DisCon const& dc)const
{
   ImgR im(3,m_,n_);
   for (Z i=1;i<=m_;++i){
      for (Z j=1;j<=n_;++j){
         C2 cij=vx_[(i-1)*n_+j];
         //C2 cij=vx_[i*n_+j];
         R3 yij=cij.toR3();
         im.putPix(i,j,Color(yij));
      }
   }
   im.markDirect(g,dc.ipolMet);
}

void SpinN::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinN::run_(Graph,RunCon,DisCon)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot();
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eInt=hInt()[1];
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs());
   R e01=e0[1];
   R e02=e0[2];
   cpmvalue("dtRel",dt_*dt0Inv_,3);

   Z nSteps=rc.nSteps;
   R tTot=R(nSteps)*dt_;
   R tTotInv=R(1)/tTot;
//   R tWaitView=rc.tWaitView;
   R tWaitDocu=rc.tWaitDocu;
//   R tWaitBoltzmann=rc.tWaitDocu2;
//   R tWaitDocu3=rc.tWaitDocu3;
   R tWaitRes=rc.tWaitRes;
   CpmAlgorithms::PeriodicAlarm view(R(rc.viewPeriod)*dt_);
   CpmAlgorithms::PeriodicAlarm docu(R(rc.docuPeriod)*dt_);
   if (rc.tWaitIni>0){
      mark(g,dc);
      g.setText("initial state");
      g.vis();
      cpmwait(rc.tWaitIni,2);
   }
   g.clr_();
   vector<R> tv(1,0.), nv(1,0.), e1v(1,0.), e2v(1,0.),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,0);
 //  Z np=vx_.dim();
   bool dtVar=getEps()>0;
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
      if (dtVar){
         cpmvalue("dtRel",dt_*dt0Inv_,3);
      }
      cpmprogress("run",tTotInv*ti);
      if (docu(ti)){
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];
         tv.push_back(ti);
         nv.push_back((ni-n0)*n0Inv);
         e1v.push_back((ei[1]-e01)*e0Inv);
         e2v.push_back((ei[2]-e02)*e0Inv);
         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
      }
      if (view(ti)||ti>=tTot){
         mark(g,dc);
         if (ti>=tTot) g.vis(true);
         else g.vis(dc.wrtOnShow);
         cpmwait(rc.tWaitView);
      }
   }
//   R fac=10;
 //  R facInv=R(1)/fac;
   if (tWaitDocu>0){
      R_Matrix doc(4);
      doc[1]=R_Vector(tv);
      doc[2]=R_Vector(nv);
      doc[3]=R_Vector(e1v);
      doc[4]=R_Vector(e2v);
      Graph g1; // it is neither important nor always useful
      // if diagnostic graphics is of the same size as
      // the evolving image. Hence we use a new frame here.
      g1.clr_();
      g1.setText("SpinN: evolution of norm and energy errors");
      g1.show(doc);
      cpmwait(tWaitDocu,2);
   }
   if (tWaitRes>0){
      R_Matrix doc(4);
      doc[1]=R_Vector(tv);
      doc[2]=R_Vector(thetav);
      doc[3]=R_Vector(phiv);
      doc[4]=R_Vector(sig3v);
      Graph g2;
      g2.clr_();
      g2.setColorActive(LIGHTGREEN);
      g2.setText("SpinN: evolution of mean Bloch angles and sig3");
      g2.show(doc);
      cpmwait(tWaitRes,2);
   }
   CPM_MZ
}

void SpinN::runSim_(Graph& g, RunCon& rc, DisCon const& dc)
{
 //  using CpmArrays::R_Matrix;
 //  using CpmArrays::R_Vector;
 //  using std::vector;
   Word loc("SpinN::runSim_(Graph,RunCon,DisCon)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   Z nSteps=rc.nSteps;
   R tTot=R(nSteps)*dt_;
   R tTotInv=R(1)/tTot;
//   R tWaitView=rc.tWaitView;
//   R tWaitRes=rc.tWaitRes;
   CpmAlgorithms::PeriodicAlarm view(R(rc.viewPeriod)*dt_);
   if (rc.tWaitIni>0){
      mark(g,dc);
      g.setText("initial state");
      g.vis();
      cpmwait(rc.tWaitIni,2);
   }
   g.clr_();
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
      cpmprogress("run",tTotInv*ti);
      if (view(ti)||ti>=tTot){
         //mark(g,dc);
         lineMark_1(g,vx_,dc);
         if (ti>=tTot) g.vis(true);
         else g.vis(dc.wrtOnShow);
         cpmwait(rc.tWaitView);
      }
   }
   CPM_MZ
}
// clean version for publication. Basing reversal on making dt negative
// and not reversing v
void SpinN::runRevTest_(Graph& g, RunCon& rc, DisCon const& dc)
{
   Word loc("SpinN::runRevTest_(Graph,RunCon,DisCon)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   Z nSteps=rc.nSteps;
   bool rev=rc.reversed;
   bool fd=rc.fromDisk;
   Z nStepsAll=rev ? nSteps*2 : nSteps;
//  R tWaitView=rc.tWaitView;
//   R tWaitRes=rc.tWaitRes;
   Z vp=rc.viewPeriod;
   cpmassert(nSteps%vp==0,loc);
   R nInv=1./nStepsAll;
   Word prog("progress");
   Word imgName;
   Z progPane=3;
   Z i=0;
   cpmprogress(prog,nInv*i,progPane);
   mark(g,dc);
   imgName=g.vis(dc.wrtOnShow);
   cpmcerr<<endl<<"time of "<<imgName<<": "<<t();
   cpmwait(rc.tWaitView);
   for (i=1;i<=nStepsAll;++i){
      step_();
      if (i%vp==0){
         mark(g,dc);
         imgName=g.vis(dc.wrtOnShow);
         cpmcerr<<endl<<"time of "<<imgName<<": "<<t();
         cpmwait(rc.tWaitView);
      }
      if (rev && i==nSteps){
         if (fd){
            Img24 img=g.getImg();
            SpinN sn(img,getIa());
            sn.revDir_(); // changes the sign of dt_
            *this=sn;
         }
         else revDir_(); // changes the sign of dt_
      }
      cpmprogress(prog,nInv*i,progPane);
   }
   CPM_MZ
}

/////////////////////////////// class Qreg ///////////////////////////////

Qreg::Qreg(Z n, Z kind):n_(cpmpos(n)),m_()
{
   if (kind==1) m_[Ky(n_)]=C::one;
}

Qreg::Qreg(V<C2> const& v):n_(v.dim()),m_()
{
   if (n_==0) return;
   Ky k(n_);
   bool work=true;
   while (work){
      C val(1.);
      for (Z i=1;i<=n_;++i){
         val*=v[i].sel(k.get(i));
      }
      m_[k]=val;
      work=k.next_();
   }
}

R3 Qreg::sig(Z i)const
{
   Mkc::const_iterator it;
   C r1,r2,r3;
   for (it=m_.begin();it!=m_.end();++it){
      Ky k=it->first;
      C val=it->second;
      bool bi=k.get(i);
      Ky k0=k; k0.set_(i,0);
      Ky k1=k; k1.set_(i,1);
      C2 ci(get(k0),get(k1));
      C val1=CC2::sig1.app(ci,bi);
      C val2=CC2::sig2.app(ci,bi);
      C val3=CC2::sig3.app(ci,bi);
      r1+=(val|val1);
      r2+=(val|val2);
      r3+=(val|val3);
   }
   return R3(r1[1],r2[1],r3[1]);
}

Qreg Qreg::app(CC2 const& a, Z i)const
{
   Qreg res(*this);
   Mkc::const_iterator it;
   for (it=m_.begin();it!=m_.end();++it){
      Ky k=it->first;
      bool bi=k.get(i);
      Ky k0=k; k0.set_(i,0);
      Ky k1=k; k1.set_(i,1);
      C2 ci(get(k0),get(k1));
      C val=a.app(ci,bi);
      res.set_(k,val);
   }
   return res;
}

Qreg Qreg::app(CCCC2 const& a, Z i, Z j)const
{
   Qreg res(*this);
   Mkc::const_iterator it;
   for (it=m_.begin();it!=m_.end();++it){
      Ky k=it->first;
      bool bi=k.get(i);
      bool bj=k.get(j);
      Ky k00=k; k00.set_(i,0); k00.set_(j,0);
      Ky k01=k; k01.set_(i,0); k01.set_(j,1);
      Ky k10=k; k10.set_(i,1); k10.set_(j,0);
      Ky k11=k; k11.set_(i,1); k11.set_(j,1);
      CC2 cij(get(k00),get(k01),get(k10),get(k11));
      C val=a.app(cij,bi,bj);
      res.set_(k,val);
   }
   return res;
}

Qreg Qreg::ham(CC2 const& ci, CCCC2 const& cij)const
{
   Z mL=3;
   Word loc("Qreg::ham(QregPar)");
   CPM_MA
   Qreg res(n_,0); // zero vector
   for (Z i=0;i<n_;++i){ // loop over nearest neighbors
      res+=app(ci,i);
      Z j=i+1;
      if (j==n_) j=0;
      res+=app(cij,i,j);
   }
   CPM_MZ
   return res;
}

Qreg Qreg::operator+(Qreg const& q)const
{
   Qreg res(*this);
   Mkc::const_iterator i;
   for (i=q.m_.begin();i!=q.m_.end();++i) res.add_(i->first,i->second);
   return res;
}

Qreg& Qreg::operator+=(Qreg const& q)
{
   Mkc::const_iterator i;
   for (i=q.m_.begin();i!=q.m_.end();++i) add_(i->first,i->second);
   return *this;
}

Qreg& Qreg::operator-=(Qreg const& q)
{
   Mkc::const_iterator i;
   for (i=q.m_.begin();i!=q.m_.end();++i) sub_(i->first,i->second);
   return *this;
}

Qreg Qreg::operator*(R r)const
{
   Qreg res(*this);
   Mkc::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i)  res.mul_(i->first,r);
   return res;
}

Qreg& Qreg::operator*=(R r)
{
   Mkc::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i) mul_(i->first,r);
   return *this;
}

Qreg& Qreg::operator*=(C z)
{
   Mkc::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i) mul_(i->first,z);
   return *this;
}

C Qreg::operator|(Qreg const& q)const
// see Musser,Saini: STL Tutorial and Reference Guide,
// Addison-Wesley 1996, Example 7-6, p. 169,170
// for the pattern of this implementation.
{
   C res;
   Mkc::const_iterator i,i2;
   for (i=m_.begin();i!=m_.end();++i){
      Ky k=i->first;
      i2=q.m_.find(k);
      if (i2!=q.m_.end()){
         C z=i->second;
         C z2=i2->second;
         res+=z.con()*z2;
      }
   }
   return res;
}

V<C2> Qreg::den()const
{
   if (n_==0) return V<C2>(0);
   V<C2> res(n_);
   if (n_==1){
      Ky k(n_);
      C c1=get(k);
      k.next_();
      C c2=get(k);
      res[1]=C2(c1,c2);
      return res;
   }
   Z nm=n_-1;
   for (Z i=1;i<=n_;++i){
      R rif=0,rit=0;
      Ky k(nm);
      bool work=true;
      while (work){
         Ky kif=k;
         kif.ins_(i,false);
         Ky kit=k;
         kit.ins_(i,true);
         rif+=get(kif).absSqr();
         rit+=get(kit).absSqr();
         work=k.next_();
      }
      res[i]=C2(C(rif),C(rit));
   }
   return res;
}

V<C2> Qreg::col()const
{
   if (n_==0) return V<C2>(0);
   V<C2> res(n_);
   for (Z i=1;i<=n_;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res[i]=C2(z,z);
   }
   return res;
}

Qreg Qreg::con(void)const
{
   Qreg res(n_,0);
   Mkc::const_iterator i;
   for (i=m_.begin();i!=m_.end();++i){
      C val=(i->second).con();
      res.set_(i->first,val);
   }
   return res;
}

void Qreg::mark(Graph& g, DisCon const& dc)const
{
   //lineMark_1(g,den(),dc);
   lineMark_1(g,col(),dc,2);
}

R Qreg::dtSup(CC2 const& h0, CCCC2 const& hi, Z p, R tWait)const
{
   Z mL=1;
   Word loc("Qreg::dtSup(QregPar,Z,R)");
   CPM_MA
   cpmassert(p>0,loc);
   Qreg psi=*this;
   psi.nor_();
   R_Vector mem(p);
   Graph gr;
   R pInv=1./p;
   for (Z i=1;i<=p;++i){
      psi=psi.ham(h0,hi); // no matter of hamiltonian or dot
      R mi=psi.nor_(); // the value is the norm!
      mem[i]=mi; // the value is the norm!
      cpmvalue("enr",mi,2);
      if (tWait>=0){
         psi.mark(gr,DisCon());
         gr.vis();
         cpmwait(tWait);
      }
      cpmprogress("dtSup",R(i)*pInv,3);
   }
   R_Vector stat=CpmArrays::statistics(mem);
   R sigma=stat[3];
   R maxVal=stat[10];
   R epsMax=maxVal+sigma;
   cpmvalue("eMax",epsMax,3);
   R res=R(2)/epsMax;
   CPM_MZ
   return res;
}

R_Vector Qreg::stt()const
{
   R_Vector res(1);
   res[1]=mag();
   return res;
}

void TrjQreg::step_(R dt)
{
   R tau=dt*0.5;
   x_+=v_*tau;
   v_-=x_.ham(h0_,hi_).ham(h0_,hi_)*dt;
   x_+=v_*tau;
   t_+=dt;
}

V< V<R2> > TrjQreg::run_(Graph& g, RunCon& rc, DisCon const& dc)
{
   Word loc("TrjQreg::run_(Graph,RunCon,DisCon)");
   Z mL=1;
   CPM_MA
   CPM_SC_RUN
   if (dt<0) dt=-dt; // negative input allows us to use values without
      // energy dependent rescaling
   else{
      R dtMax=dtSup(rc.nIter,rc.tWaitIter);
      dt*=dtMax;
      rc.dt=-dt; // then call another run function subsequently will use
         // dt
   }
   cpmdebug(rc.dt);
   vector<R2> s1v;
   g.clr();
   for (Z i=1;i<=nSteps;++i){
      step_(dt);
      cpmprogress("run",nInv*i);
      if (tWaitDocu>0){
         if (docu()){
            CPM_SC_RUN2
            R_Vector rv=stt();
            s1v.push_back(R2(ti,rv[1]));
           // s2v.push_back(R2(ti,rv[2]));
          //  s12v.push_back(R2(ti,rv[3]));
         }
      }
      if (view()){
         mark(g,dc);
         g.vis();
         cpmwait(rc.tWaitView);
      }
   }
   if (tWaitDocu>0){
      CPM_SC_RUN3
      g.setText("TrjQreg: run documentation");
      g.show(doc);
      cpmwait(tWaitDocu,2);
   }
   V< V<R2> > res(1);

   res[1]=V<R2>(1,s1v);
/*
   res[2]=V<R2>(s2v);
   res[3]=V<R2>(s12v);
*/
   if (tWaitRes>0){
      g.clr_();
      g.setAutoScl_(true);
      g.setText("TrjQreg: res documentation");
      g.mark(res);
      g.vis();
      cpmwait(tWaitRes,2);
   }
   CPM_MZ
   return res;
}

#undef CPM_SC_RUN
#undef CPM_SC_RUN2
#undef CPM_SC_RUN3

// Studying matrix elements of general random-generated pair potentials
// to see whether they are real. Actually those combinations get tested
// the reality of which guarantees conservation of energy. Since
// rather complicated combinatorics is involved, I decided to convince
// myself of the correctness my conjecture before I try to complete my
// proof on paper.
// Result: the conjecture is correct, energy conservation holds for
// arbitrary pair potentials, not only those that are multiplication
// operators in position space.
// The proof on paper was completed 2008-02-26
// So this part of the exercise is no longer important, but it was
// helpful.
////////////////////// class PairStt /////////////////////////////////////
class PairStt{ // pair state
   Z d1_,d2_;
   VV<C> x_;
public:
   PairStt(Z d1, Z d2):d1_(cpmpos(d1)),d2_(cpmpos(d2)),x_(d1_,d2_){}
   Z dim1()const{ return d1_;}
   Z dim2()const{ return d2_;}
   C& operator()(Z i, Z j){ return x_[i][j];}
   C const& operator()(Z i, Z j)const{ return x_[i][j];}
   V<C> trc1(V<C> const& s1)const;
      //: trace 1
      // getting a one-particle state by tracing one state out
   V<C> trc2(V<C> const& s2)const;
      //: trace 2
      // getting a one-particle state by tracing one state out
};

V<C> PairStt::trc1(V<C> const& s1)const
{
   Word loc("PairStt::trc1(V<C>)");
   cpmassert(s1.dim()==d1_,loc);
   V<C> res(d2_);
   Z i,j;
   for (i=1;i<=d2_;++i){
      C ri;
      for (j=1;j<=d1_;++j){
         ri+=s1[j].con()*x_[j][i];
      }
      res[i]=ri;
   }
   return res;
}

V<C> PairStt::trc2(V<C> const& s2)const
{
   Word loc("PairStt::trc2(V<C>)");
   cpmassert(s2.dim()==d2_,loc);
   V<C> res(d1_);
   Z i,j;
   for (i=1;i<=d1_;++i){
      C ri;
      for (j=1;j<=d2_;++j){
         ri+=x_[i][j]*s2[j].con();
      }
      res[i]=ri;
   }
   return res;
}

////////////////////////// class Stt ////////////////////////////////////
class Stt{ // state
   // here state of a three-particle system, later we get also
   // Stt4 for 4-particle states which should be effectively the generic
   // case since all relations between 2-particle subsets are no
   // possible: equality, overlap, disjointness
   Z d1_,d2_,d3_;
   VVV<C> x_;
public:
   Stt(Z d1,Z d2, Z d3):d1_(cpmpos(d1)),d2_(cpmpos(d2)),d3_(cpmpos(d3)),
      x_(d1_,d2_,d3_){}

   Stt(V<C> const& p1, V<C> const& p2, V<C> const& p3):d1_(p1.dim()),
      d2_(p2.dim()),d3_(p3.dim()),x_(d1_,d2_,d3_)
   {
      Z i,j,k;
      for (i=1;i<=d1_;++i){
         C pi=p1[i];
         for (j=1;j<=d2_;++j){
            C pij=pi*p2[j];
            for (k=1;k<=d3_;++k){
               x_[i][j][k]=pij*p3[k];
            }
         }
      }
   }

   Stt(PairStt const& pp, V<C> const& s, Word const& w);
      // construction from interacting pair state
      // and a free state, important!

   Z dim1()const{ return d1_;}
   Z dim2()const{ return d2_;}
   Z dim3()const{ return d3_;}
   C& operator()(Z i, Z j, Z k){ return x_[i][j][k];}
   C const& operator()(Z i, Z j, Z k)const{ return x_[i][j][k];}

   Stt operator+(Stt const& stt)const;
   C operator|(Stt const& stt)const;

};

Stt Stt::operator+(Stt const& stt)const
{
   Word loc("Stt::operator+(Stt)");
   cpmassert(d1_==stt.dim1(),loc);
   cpmassert(d2_==stt.dim2(),loc);
   cpmassert(d3_==stt.dim3(),loc);
   Stt res(d1_,d2_,d3_);
   {
      Z i,j,k;
      for (i=1;i<=d1_;++i){
        for (j=1;j<=d2_;++j){
            for (k=1;k<=d3_;++k){
               res(i,j,k)=x_[i][j][k]+stt(i,j,k);
            }
         }
      }
   }
   return res;
}

C Stt::operator|(Stt const& stt)const
{
   Word loc("Stt::operator+(Stt)");
   cpmassert(d1_==stt.dim1(),loc);
   cpmassert(d2_==stt.dim2(),loc);
   cpmassert(d3_==stt.dim3(),loc);
   C res;
   {
      Z i,j,k;
      for (i=1;i<=d1_;++i){
        for (j=1;j<=d2_;++j){
            for (k=1;k<=d3_;++k){
               res+=x_[i][j][k].con()*stt(i,j,k);
            }
         }
      }
   }
   return res;
}

Stt::Stt(PairStt const& pp, V<C> const& s, Word const& w)
{
   Word loc("Stt::Stt(PairStt,V<C>,Word)");
   if (w=="12_3"){
      d1_=pp.dim1();
      d2_=pp.dim2();
      d3_=s.dim();
      x_=VVV<C>(d1_,d2_,d3_);
      Z i,j,k;
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d2_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d3_;++k){
               x_[i][j][k]=pij*s[k];
            }
         }
      }
   }
   else if (w=="31_2"){
      d3_=pp.dim1();
      d1_=pp.dim2();
      d2_=s.dim();
      x_=VVV<C>(d1_,d2_,d3_);
      Z i,j,k;
      for (i=1;i<=d3_;++i){
         for (j=1;j<=d1_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d2_;++k){
               x_[j][k][i]=pij*s[k];
            }
         }
      }
   }
   else if (w=="23_1"){
      d2_=pp.dim1();
      d3_=pp.dim2();
      d1_=s.dim();
      x_=VVV<C>(d1_,d2_,d3_);
      Z i,j,k;
      for (i=1;i<=d2_;++i){
        for (j=1;j<=d3_;++j){
           C pij=pp(i,j);
           for (k=1;k<=d1_;++k){
              x_[k][i][j]=pij*s[k];
           }
         }
      }
   }
   else{
      cpmerror(loc&": bad Word argument");
   }
}

////////////////////////// class Stt4 ////////////////////////////////////
class Stt4{ // state
   // here state of a four-particle system
   Z d1_,d2_,d3_,d4_;
   VVVV<C> x_;
public:
   Stt4(Z d1,Z d2, Z d3, Z d4):d1_(cpmpos(d1)),d2_(cpmpos(d2)),
      d3_(cpmpos(d3)), d4_(cpmpos(d4)), x_(d1_,d2_,d3_,d4_){}

   Stt4(V<C> const& p1, V<C> const& p2, V<C> const& p3, V<C> const& p4):
      d1_(p1.dim()),d2_(p2.dim()),d3_(p3.dim()),d4_(p4.dim()),
      x_(d1_,d2_,d3_,d4_)
   {
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
         C pi=p1[i];
         for (j=1;j<=d2_;++j){
            C pij=pi*p2[j];
            for (k=1;k<=d3_;++k){
               C pijk=pij*p3[k];
               for (l=1;l<=d4_;++l){
                  x_[i][j][k][l]=pijk*p4[l];
               }
            }
         }
      }
   }

   Stt4(PairStt const& pp, V<C> const& p1, V<C> const& p2, Word const& w);
   // construction from interacting pair state
   // and two free states, important!

   Z dim1()const{ return d1_;}
   Z dim2()const{ return d2_;}
   Z dim3()const{ return d3_;}
   Z dim4()const{ return d4_;}
   C& operator()(Z i, Z j, Z k, Z l){ return x_[i][j][k][l];}
   C const& operator()(Z i, Z j, Z k, Z l)const{ return x_[i][j][k][l];}

   Stt4 operator+(Stt4 const& stt)const;
   C operator|(Stt4 const& stt)const;

};

Stt4 Stt4::operator+(Stt4 const& stt)const
{
   Word loc("Stt4::operator+(Stt4)");
   cpmassert(d1_==stt.dim1(),loc);
   cpmassert(d2_==stt.dim2(),loc);
   cpmassert(d3_==stt.dim3(),loc);
   cpmassert(d4_==stt.dim4(),loc);
   Stt4 res(d1_,d2_,d3_,d4_);
   {
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
        for (j=1;j<=d2_;++j){
            for (k=1;k<=d3_;++k){
               for (l=1;l<=d4_;++l){
                  res(i,j,k,l)=x_[i][j][k][l]+stt(i,j,k,l);
               }
            }
         }
      }
   }
   return res;
}

C Stt4::operator|(Stt4 const& stt)const
{
   Word loc("Stt4::operator+(Stt4)");
   cpmassert(d1_==stt.dim1(),loc);
   cpmassert(d2_==stt.dim2(),loc);
   cpmassert(d3_==stt.dim3(),loc);
   cpmassert(d4_==stt.dim4(),loc);
   C res;
   {
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
        for (j=1;j<=d2_;++j){
            for (k=1;k<=d3_;++k){
               for (l=1;l<=d4_;++l){
                  res+=x_[i][j][k][l].con()*stt(i,j,k,l);
               }
            }
         }
      }
   }
   return res;
}

Stt4::Stt4(PairStt const& pp, V<C> const& p1,
   V<C> const& p2, Word const& w)
{
   Word loc("Stt4::Stt4(PairStt,V<C>,V<C>,Word)");
   if (w=="12_3_4"){
      d1_=pp.dim1();
      d2_=pp.dim2();
      d3_=p1.dim();
      d4_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d2_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d3_;++k){
               for (l=1;l<=d4_;++l){
                  x_[i][j][k][l]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else if (w=="13_2_4"){
      d1_=pp.dim1();
      d3_=pp.dim2();
      d2_=p1.dim();
      d4_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d3_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d2_;++k){
               for (l=1;l<=d4_;++l){
                  x_[i][k][j][l]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else if (w=="14_2_3"){
      d1_=pp.dim1();
      d4_=pp.dim2();
      d2_=p1.dim();
      d3_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d4_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d2_;++k){
               for (l=1;l<=d3_;++l){
                  x_[i][k][l][j]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else if (w=="23_1_4"){
      d2_=pp.dim1();
      d3_=pp.dim2();
      d1_=p1.dim();
      d4_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d2_;++i){
         for (j=1;j<=d3_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d1_;++k){
               for (l=1;l<=d4_;++l){
                  x_[k][i][j][l]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else if (w=="24_1_3"){
      d2_=pp.dim1();
      d4_=pp.dim2();
      d1_=p1.dim();
      d3_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d2_;++i){
         for (j=1;j<=d4_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d1_;++k){
               for (l=1;l<=d3_;++l){
                  x_[k][i][l][j]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else if (w=="34_1_2"){
      d3_=pp.dim1();
      d4_=pp.dim2();
      d1_=p1.dim();
      d2_=p2.dim();
      x_=VVVV<C>(d1_,d2_,d3_,d4_);
      Z i,j,k,l;
      for (i=1;i<=d3_;++i){
         for (j=1;j<=d4_;++j){
            C pij=pp(i,j);
            for (k=1;k<=d1_;++k){
               for (l=1;l<=d2_;++l){
                  x_[k][l][i][j]=pij*p1[k]*p2[l];
               }
            }
         }
      }
   }
   else{
      cpmerror(loc&": bad Word argument");
   }
}

////////////////////// class PairInt /////////////////////////////////////
class PairInt{ // pair interaction
   Z d1_,d2_;
   VVVV<C> x_;
public:
   PairInt(Z d1, Z d2):d1_(cpmpos(d1)),d2_(cpmpos(d2)),x_(d1_,d1_,d2_,d2_)
      {} // trivial matrix elements but right format
   Z dim1()const{ return d1_;}
   Z dim2()const{ return d2_;}
   PairStt evl(V<C> const& p1, V<C> const& p2)const;
      // returns the pair state which results from the
      // free state p1 \otimes p2 by applying the
      // pair interaction operator *this on it
   PairInt ran(Z i=0)const;
      // returns an hermitan pair interaction operator
      // with randomly selected matrix elements
   C& operator()(Z i, Z j, Z k, Z l){ return x_[i][j][k][l];}
   C const& operator()(Z i, Z j, Z k, Z l)const{ return x_[i][j][k][l];}
};

PairStt PairInt::evl(V<C> const& p1, V<C> const& p2)const
{
   Word loc("PairInt::evl(V<C>,V<C>)");
   Z d1=p1.dim(), d2=p2.dim();
   cpmassert(d1==d1_,loc);
   cpmassert(d2==d2_,loc);
   PairStt res(d1_,d2_);
   Z i,j,k,l;
   for (i=1;i<=d1_;++i){
      for (j=1;j<=d2_;++j){
         C rij;
         for (k=1;k<=d1_;++k){
            for (l=1;l<=d2_;++l){
               rij+=x_[i][k][j][l]*p1[k]*p2[l];
            }
         }
         res(i,j)=rij;
      }
   }
   return res;
}

PairInt PairInt::ran(Z p)const
{
   PairInt temp(d1_,d2_);
      // not hermitean
   C z(1.,1.);
   Z i,j,k,l;
   Z ir=p;
   if (ir!=0){
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d1_;++j){
            for (k=1;k<=d2_;++k){
               for (l=1;l<=d2_;++l){
                  temp(i,j,k,l)=z.ran(ir++);
               }
            }
         }
      }
   }
   else{
      for (i=1;i<=d1_;++i){
         for (j=1;j<=d1_;++j){
            for (k=1;k<=d2_;++k){
               for (l=1;l<=d2_;++l){
                  temp(i,j,k,l)=z.ran();
               }
            }
         }
      }
   }
   PairInt res(d1_,d2_);
      // hermitean result
   for (i=1;i<=d1_;++i){
      for (j=1;j<=d1_;++j){
         for (k=1;k<=d2_;++k){
            for (l=1;l<=d2_;++l){
               res(i,j,k,l)=temp(i,j,k,l)+temp(j,i,l,k).con();
            }
         }
      }
   }
   return res;
}

void CpmQM::testEC(Z d1, Z d2, Z d3)
{
   PairInt v12(d1,d2);
   PairInt v31(d3,d1);
   PairInt v23(d2,d3);

   v12=v12.ran();
   v31=v31.ran();
   v23=v23.ran();

   C z1(1.,1.);
   C z2=z1.ran();
   C z3=z1.ran();

   Vr<C> p1(d1,z1);
   Vr<C> p2(d2,z2);
   Vr<C> p3(d3,z3);
   p1=p1.ran();
   p2=p2.ran();
   p3=p3.ran();
   p1.nor_();
   p2.nor_();
   p3.nor_();

   PairStt p12=v12.evl(p1,p2);
   PairStt p31=v31.evl(p3,p1);
   PairStt p23=v23.evl(p2,p3);
   Stt p12_3(p12,p3,"12_3");
   Stt p31_2(p31,p2,"31_2");
   Stt p23_1(p23,p1,"23_1");
   Stt pL=p12_3+p31_2+p23_1;

   Stt q12(v12.evl(p1,p2).trc2(p2),p2,p3);
   Stt q13(v31.evl(p3,p1).trc1(p3),p2,p3);

   Stt q23(p1,v23.evl(p2,p3).trc2(p3),p3);
   Stt q21(p1,v12.evl(p1,p2).trc1(p1),p3);

   Stt q31(p1,p2,v31.evl(p3,p1).trc2(p1));
   Stt q32(p1,p2,v23.evl(p2,p3).trc1(p2));

   Stt pR=q12+q13+q23+q21+q31+q32;
   cpmdebug((pL|q12));
   cpmdebug((pL|q13));
   cpmdebug((pL|q21));
   cpmdebug((pL|q23));
   cpmdebug((pL|q31));
   cpmdebug((pL|q32));

   cpmdebug((p12_3|q12));
   cpmdebug((p12_3|q13));

   cpmdebug((p31_2|q12));
   cpmdebug((p31_2|q13));

   cpmdebug((p23_1|q12));
   cpmdebug((p23_1|q13));

   C res=(pL|pR);
   cpmnote(res);
   cpmdebug(res);
}

void CpmQM::testEC4(Z d1, Z d2, Z d3, Z d4)
{
   PairInt v12(d1,d2);
   PairInt v13(d1,d3);
   PairInt v14(d1,d4);
   PairInt v23(d2,d3);
   PairInt v24(d2,d4);
   PairInt v34(d3,d4);

   v12=v12.ran();
   v13=v13.ran();
   v14=v14.ran();
   v23=v23.ran();
   v24=v24.ran();
   v34=v34.ran();

   C z1(1.,1.);
   C z2=z1.ran();
   C z3=z1.ran();
   C z4=z1.ran();

   Vr<C> p1(d1,z1);
   Vr<C> p2(d2,z2);
   Vr<C> p3(d3,z3);
   Vr<C> p4(d4,z4);

   p1=p1.ran();
   p2=p2.ran();
   p3=p3.ran();
   p4=p4.ran();

   p1.nor_();
   p2.nor_();
   p3.nor_();
   p4.nor_();

   PairStt p12=v12.evl(p1,p2);
   PairStt p13=v13.evl(p1,p3);
   PairStt p14=v14.evl(p1,p4);
   PairStt p23=v23.evl(p2,p3);
   PairStt p24=v24.evl(p2,p4);
   PairStt p34=v34.evl(p3,p4);

   Stt4 p12_3_4(p12,p3,p4,"12_3_4");
   Stt4 p13_2_4(p13,p2,p4,"13_2_4");
   Stt4 p14_2_3(p14,p2,p3,"14_2_3");

   Stt4 p23_1_4(p23,p1,p4,"23_1_4");
   Stt4 p24_1_3(p24,p1,p3,"24_1_3");

   Stt4 p34_1_2(p34,p1,p2,"34_1_2");

   Stt4 pL=p12_3_4 + p13_2_4 + p14_2_3 + p23_1_4 + p24_1_3 + p34_1_2;

   Stt4 q12(v12.evl(p1,p2).trc2(p2),p2,p3,p4);
   Stt4 q13(v13.evl(p1,p3).trc2(p3),p2,p3,p4);
   Stt4 q14(v14.evl(p1,p4).trc2(p4),p2,p3,p4);

   Stt4 q21(p1,v12.evl(p1,p2).trc1(p1),p3,p4);
   Stt4 q23(p1,v23.evl(p2,p3).trc2(p3),p3,p4);
   Stt4 q24(p1,v24.evl(p2,p4).trc2(p4),p3,p4);

   Stt4 q31(p1,p2,v13.evl(p1,p3).trc1(p1),p4);
   Stt4 q32(p1,p2,v23.evl(p2,p3).trc1(p2),p4);
   Stt4 q34(p1,p2,v34.evl(p3,p4).trc2(p4),p4);

   Stt4 q41(p1,p2,p3,v14.evl(p1,p4).trc1(p1));
   Stt4 q42(p1,p2,p3,v24.evl(p2,p4).trc1(p2));
   Stt4 q43(p1,p2,p3,v34.evl(p3,p4).trc1(p3));


   Stt4 pR=q12+q13+q14+q21+q23+q24+q31+q32+q34
      +q41+q42+q43;

   C res=(pL|pR);
   cpmnote(res);
   cpmdebug(res);
      // real up to numerical noise
   cpmdebug((pL|(q12+q13+q14)));
      // real up to numerical noise
   cpmdebug((pL|(q21+q23+q24)));
      // real up to numerical noise
   cpmdebug((pL|(q31+q32+q34)));
      // real up to numerical noise
   cpmdebug((pL|(q41+q42+q43)));
      // real up to numerical noise

   cpmdebug((p12_3_4|q12));
   cpmdebug((p12_3_4|q13));
   cpmdebug((p12_3_4|q14));

   cpmdebug((p13_2_4|q12));
   cpmdebug((p13_2_4|q13));
   cpmdebug((p13_2_4|q14));

   cpmdebug((p14_2_3|q12));
   cpmdebug((p14_2_3|q13));
   cpmdebug((p14_2_3|q14));

   cpmdebug((p23_1_4|q12));
   cpmdebug((p23_1_4|q13));
   cpmdebug((p23_1_4|q14));

   cpmdebug((p24_1_3|q12));
   cpmdebug((p24_1_3|q13));
   cpmdebug((p24_1_3|q14));

   cpmdebug((p34_1_2|q12));
   cpmdebug((p34_1_2|q13));
   cpmdebug((p34_1_2|q14));
}
