//? cpmqma.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


/*
   cpmqma.cpp
   History: see cpmqmx.h

*/

#include <cpmqma.h>
#include <cpmvectormore.h>
#include <cpmobservable.h>
#include <cpmdistribution.h>
#include <cpmframes.h>

using CpmQMa::RunCon;
using CpmQMa::DisCon;
using CpmQMa::SpinNa;
using CpmQMa::SpinNaa;
using CpmQMa::Ia;
using CpmQMa::VC2;
using CpmQMa::WC2;

using CpmArrays::V;
using CpmArrays::VV;
using CpmArrays::VVV;
using CpmArrays::VVVV;
using CpmArrays::R_Vector;
using CpmArrays::DisR;
using CpmLinAlg::C2;
using CpmLinAlg::CC2;
using CpmLinAlg::CCCC2;
using CpmGraphics::Img24;
using CpmImaging::Color;
using CpmImaging::ColRef;
using CpmImaging::WHITE;
using CpmImaging::BLACK;
using CpmImaging::LIGHTGREEN;
using CpmImaging::LIGHTCYAN;
using CpmImaging::LIGHTBLUE;
using CpmImaging::FAINTYELLOW;


namespace{ // anonymous
   // color related stuff
   const R xg=-0.5;
   const R yg=0.5*sqrt(3.);
   const R2 er(1,0);
   const R2 eg(xg,yg);
   const R2 eb(xg,-yg);
   C transform(R3 const& r){
      R xr=(r[1]+1);
      R xg=(r[2]+1);
      R xb=(r[3]+1);
      R2 res=er*xr+eg*xg+eb*xb;
      return C(res[1],res[2]);
   }
}

void lineMark_1(Graph& g, VC2 const& vx, DisCon const& dc, Z met)
{
   using CpmImaging::Color;
   static Z jShift=1;
   Z gw=g.w();
   Z gh=g.h();
   Z np=vx.dim();
   Z pix=gh/np;
   Z mult_i=pix;
   Z mult_j=pix;
   Z js=pix,is=pix;
   Z i,j;
   Z d_=np*mult_i;
   jShift+=js;
   Z iState=1;
   Z iCounter=0;
   Z iEff=is;
   Color c;

   for (i=1;i<=d_;i++){
      if (iCounter==0){
         C2 ci=vx.cui(iState);
         if (dc.colMet==1){ // Bloch angles based coloring
            C val=ci.blc();
            c=Color(val);
   // if this is true we have rigid (reflective) boundary conditions,
   // periodic otherwiseal);
         }
         else if (dc.colMet==2){
            C val=ci[1];
            c=Color(val);
         }
         else if (dc.colMet==3){
            C val=ci[2];
            c=Color(val);
         }
         else if (dc.colMet==4){
            C val=ci[1]*ci[2];
            c=Color(val);
         }
         else if (dc.colMet==5){
            R3 r3i=ci.toR3();
            C c1=transform(r3i);
            c=Color(c1);
         }
         else if (dc.colMet==6){ // new natural method
            R r1=((ci|CC2::sig1*ci)[1]+1.)*0.5;
            R r2=((ci|CC2::sig2*ci)[1]+1.)*0.5;
            R r3=((ci|CC2::sig3*ci)[1]+1.)*0.5;
            c=Color(R3(r1,r2,r3));
         }
         else{ c=Color(ci.toR3());}
      }
      iCounter++;
      if (iCounter==mult_i){
         iCounter=0; // triggers new evaluation of c in the next loop
         iState++;
      }
      for (j=0;j<mult_j;j++){ g.put(Z2(iEff,(j+jShift)%gw),c);}
      iEff++;
   }
}

/////////////////////////// class SpinNa  /////////////////////////////////

SpinNa::SpinNa(StpCon const& sc, VC2 const& x, R h, R qXY, R qZ, 
   RunCon const& rc, DisCon const& dc):
   np_(x.dim()), ia_(h,qXY,qZ), rc_(rc), dc_(dc)
{  
// notice that all but one of the members became initialized by the 
// standard initialization list method.The missing member res_ will now 
// be initialized comfortably with a lambda expression which makes use of
// the already initialized members.
   auto lambda = [*this](R const& t, VC2 const& x)
   {
      Outside os=rc_.rigid ? DEFAULT : CYCLIC;
      VC2 res(np_);
      for (Z i=1;i<=np_;++i){
         C2 ci=x.cui(i);
         C2 ri=ia_(ci)+ia_(x(i-1,os),ci)+ia_(x(i+1,os),ci);
         res.cui(i)=ri*C::Ibar;
      }
      return res;
   };
      // Here ia_(ci) is the energy of particle i in an external field,
      // ia_(x(i-1,os),ci) adds the interaction energy from one
      // neighbor particle and ia_(x(i+1,os),ci) adds the interaction 
      // energy from the other neighbor particle. Final multiplication
      // by 1/i transforms energies into contdibutions to d/dt and hence
      // to dot.

   function<VC2(R,VC2)> dot{lambda};
      // this defines the dot-function which is 'the heart of dynamics'.
   rep_=ALF<VC2,VC2>(sc,0_R,x,dot); 
      // an initial value of dt comes from sc as sc.dt0()
      // an initial value of t is set here as 0
      // the initial value of x comes from the argument x of the
      // constructor function.
      // The initial value of y is dot(t,x) as formed by the
      // constructor of ALF. 
}

C SpinNa::hInt()const
{
   C res; // initialized as 0
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   for (Z i=1;i<=np_;++i){
      C2 ci=rep_.x().cui(i);
      res+= (ia_.avr(rep_.x()(i-1,os),ci)
             + ia_.avr(rep_.x()(i+1,os),ci));
   }
   return res*0.5_R; // In a summation over particle pairs
      // we have counted each pair twice and have to correct for this
      // by taking only one half.
}

C SpinNa::hInt_i(Z i)const
{
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   C2 ci=rep_.x().cui(i);
   C z=ia_.avr(rep_.x()(i-1,os),ci) + ia_.avr(rep_.x()(i+1,os),ci);
   return z*0.5_R;
}
   // hInt_i is defined such that hInt_i(1)+ ... + hInt_i(np) = hInt()
   // This requires the final multiplication by 0.5

R SpinNa::test2()const
{
   Word loc("SpinNa::test2()");
   Z mL=2;
   CPM_MA
   C res1=hInt();
   C res2;
   for (Z i=1;i<=np_;++i){
      res2+=hInt_i(i);
   }   
   R res=(res1-res2).abs();
   cout<<" res = "<<res<<", this should be 0 or numerical noise "<<endl;
   CPM_MZ
   return res; // 
}

C SpinNa::eSep_i(Z i)const
{
   C2 ci=rep_.x().cui(i);
   return (ci|ia_(ci));
}

C SpinNa::eTot_i(Z i)const
{
   return eSep_i(i)+hInt_i(i);
}

C SpinNa::eSep()const
{
   C res;
   for (Z i=1;i<=np_;++i){
      C2 ci=rep_.x().cui(i);
      res += (ci|ia_(ci));
   }
   return res;
}

R3 SpinNa::sig(Z i)const
{
   C2 ci=rep_.x().cui(i);
   C v1=(ci|CC2::sig1*ci);
   C v2=(ci|CC2::sig2*ci);
   C v3=(ci|CC2::sig3*ci);
   return R3(v1[1],v2[1],v3[1]);
}

VC2 SpinNa::col()const
{
   if (np_==0) return VC2(0);
   VC2 res(np_);
   for (Z i=1;i<=np_;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res.cui(i)=C2(z,z);
   }
   return res;
}

void SpinNa::step_()
{
   rep_.step_();
}

void SpinNa::mark(Graph& g, Z met)const
{
   lineMark_1(g,rep_.x(),dc_,met);
}

V<V<R2>> SpinNa::run_(Graph& g)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinNa::run_(Graph,DisConx)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot(); // energy has a small imaginary part in the present context
   C es=eSep(); // energies may have a small imaginary part in the present
      // context. Energy in the external field.
   C eint=hInt(); // contribution of spin-spin-interaction to total
      // energy
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eTotAbs=e0.abs();

   cpmvalue("eTotR = ",eTotR,3);
   R eInt=hInt()[1];
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs()); // e0.abs() == 0  does not harm.
   R e01=e0[1]; // real and imaginary part
   R e02=e0[2];
   R es1=es[1]; // real and imaginary part
   R es2=es[2];
   R eint1=eint[1]; // real and imaginary part
   R eint2=eint[2];
   Z nSteps=rc_.nSteps;
   R dt=rep_.dt();
   R tTot=R(nSteps)*dt;
   R tTotInv=1_R/tTot;
   R dt0Inv=cpminv(dt);
   R tWaitDocuNormError=rc_.tWaitDocuNormError;
   R tWaitDocuEnergies=rc_.tWaitDocuEnergies;
   R tWaitDocuEnergyError=rc_.tWaitDocuEnergyError;
   R tWaitBlochAngles=rc_.tWaitBlochAngles;
   R tWaitSigma3=rc_.tWaitSigma3;
   CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*dt);
   CpmAlgorithms::PeriodicAlarm docu(R(rc_.docuPeriod)*dt);
   if (rc_.tWaitIni>0){
      mark(g);
      g.setText("         initial state");
      g.vis();
      cpmwait(rc_.tWaitIni,2);
   }
   g.clr_();
   R r0=0_R;
   vector<R> tv(1,r0), nv(1,r0), e1v(1,r0), e2v(1,r0),
      de1v(1,r0), de2v(1,r0),
      es1v(1,r0), es2v(1,r0),eint1v(1,r0), eint2v(1,r0),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,r0);
   bool dtVar=rc_.autoStep;
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
      if (dtVar){
         cpmvalue("dtRel",dt*dt0Inv,3);
      }
      cpmprogress("run",tTotInv*ti);
      if (docu(ti)){
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         C esi=eSep();
         C einti=hInt();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];

         tv.push_back(ti);
            // points in time
         nv.push_back((ni-n0)*n0Inv);
            //

         e1v.push_back(ei[1]);
         e2v.push_back(ei[2]);
         es1v.push_back(esi[1]);
         es2v.push_back(esi[2]);
         eint1v.push_back(einti[1]);
         eint2v.push_back(einti[2]);

         de1v.push_back((ei[1]-e01)*e0Inv);
         de2v.push_back((ei[2]-e02)*e0Inv);

         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
      }
      if (view(ti)||ti>=tTot){
         mark(g);
         if (ti>=tTot) g.vis(true); // always conserve the last frame
         else g.vis(dc_.wrtOnShow);
         cpmwait(rc_.tWaitView);
      }
   }
   Z dv=tv.size();
   V<R2> nv2(dv);
   V<R2> e1v2(dv);
   V<R2> e2v2(dv);
   V<R2> es1v2(dv);
   V<R2> es2v2(dv);
   V<R2> eint1v2(dv);
   V<R2> eint2v2(dv);
   V<R2> de1v2(dv);
   V<R2> de2v2(dv);
   V<R2> thetav2(dv);
   V<R2> phiv2(dv);
   V<R2> sig3v2(dv);
   for (Z i=1;i<=dv;++i){
      Z j=i-1;
      R ti=tv[i-1];
      nv2[i]=R2(ti,nv[j]);
      e1v2[i]=R2(ti,e1v[j]);
      e2v2[i]=R2(ti,e2v[j]);
      es1v2[i]=R2(ti,es1v[j]);
      es2v2[i]=R2(ti,es2v[j]);
      eint1v2[i]=R2(ti,eint1v[j]);
      eint2v2[i]=R2(ti,eint2v[j]);
      de1v2[i]=R2(ti,de1v[j]);
      de2v2[i]=R2(ti,de2v[j]);

      thetav2[i]=R2(ti,thetav[j]);
      phiv2[i]=R2(ti,phiv[j]);
      sig3v2[i]=R2(ti,sig3v[j]);
   }

   V<V<R2>> doc1{nv2};
   V<V<R2>> doc2{e1v2,e2v2,es1v2,es2v2,eint1v2,eint2v2};
   V<V<R2>> doc3{thetav2,phiv2};
   V<V<R2>> doc4{sig3v2};
   V<V<R2>> doc5{de1v2,de2v2};

   Z dwl=4; // docu write limit
   bool fileIt=false;
   //Color cbg = FAINTYELLOW;
   Color cbg = LIGHTGREEN;
   if (tWaitDocuNormError>=0){
      Graph g1; // it is neither important nor always useful
      // if diagnostic graphics is of the same size as
      // the evolving image. Hence we use a new frame here.
      g1.clr_();
      fileIt=tWaitDocuNormError>=dwl;
      g1.setColorActive(cbg);
      g1.setText("SpinNa: evolution of norm error");
      g1.mark(doc1);
      g1.vis(fileIt);
      cpmwait(tWaitDocuNormError,2);
   }
   if (tWaitDocuEnergies>=0){
      Graph g2;
      g2.clr_();
      fileIt=tWaitDocuEnergies>=dwl;
      g2.setWithOrigin(0);
      g2.setColorActive(cbg);
      g2.setText("SpinNa: evolution of eTot, eSep and hInt");
      g2.mark(doc2);
      g2.vis(fileIt);
      cpmwait(tWaitDocuEnergies,2);
   }
   if (tWaitBlochAngles>=0){
      Graph g3;
      g3.clr_();
      fileIt=tWaitBlochAngles>=dwl;
      g3.setColorActive(cbg);
      g3.setText("SpinNa: evolution of mean Bloch angles");
      g3.mark(doc3);
      g3.vis(fileIt);
      cpmwait(tWaitBlochAngles,2);
   }
   if (tWaitSigma3>=0){
      Graph g4;
      g4.clr_();
      fileIt=tWaitSigma3>=dwl;
      g4.setColorActive(cbg);
      g4.setText("SpinNa: evolution of sig3");
      g4.mark(doc4);
      g4.vis(fileIt);
      cpmwait(tWaitSigma3,2);
   }
   if (tWaitDocuEnergyError>=0){
      Graph g5;
      g5.clr_();
      fileIt=tWaitDocuEnergyError>=dwl;
      g5.setWithOrigin(0);
      g5.setColorActive(cbg);
      g5.setText("SpinNa: evolution of eTot error");
      g5.mark(doc5);
      g5.vis(fileIt);
      cpmwait(tWaitDocuEnergyError,2);
   }
   CPM_MZ
   return V<V<R2>>{nv2,e1v2,e2v2,thetav2,phiv2,sig3v2};
}

//void SpinNa::runX_(Graph& g/*, DisConx const& dc*/)
//{
//   using CpmArrays::R_Matrix;
//   using CpmArrays::R_Vector;
//   using std::vector;
//   Word loc("SpinN::runX_(Graph,DisConx)");
//   Z mL=1;
//   CPM_MA
//   SpinNa stateMem(*this);
//   V<V<R2>> res1=run_(g/*,dc*/);
//
//   RunConx rc2=stateMem.rc_;
//   Z fac=2;
//   R facInv=cpminv(R(fac));
//   rc2.dt*=facInv;
//   rc2.nSteps*=fac;
//   rc2.docuPeriod*=fac;
//   R order=2.;
//   R collapsFactor=cpmpow(R(fac),order);
//   SpinNa state2(stateMem);
//   state2.rc_=rc2;
//   g.clr_();
//   V<V<R2>> res2=state2.run_(g/*,dc*/);
//
//#ifdef CPM_Fn
//   F<R2,R2> ff=F1<R2,R,R2>(collapsFactor)(fp);
//#else
//   F<R2,R2> ff(bind(fp,_1,collapsFactor));
//#endif
//
//   V<R2> no1=res1[1];
//   V<R2> no2=res2[1];
//   no2=no2(ff);
//
//   V<R2> en1=res1[2];
//   V<R2> en2=res2[2];
//   en2=en2(ff);
//
//   Graph g1;
//   g1.clr_();
//   g1.setText("SpinNa: evolution of norm error for dt and "
//      &cpm(fac)&" dt");
//   g1.mark(V<V<R2>>{no1,no2});
//   g1.vis(true);
//   cpmwait(rc_.tWaitNorm2,2);
//
//   Graph g2;
//   g2.clr_();
//   g2.setText("SpinNa: evolution of energy error for dt and "
//      &cpm(fac)&" dt");
//   g2.mark(V<V<R2>>{en1,en2});
//   g2.vis(true);
//   cpmwait(rc_.tWaitEnergy2,2);
//   CPM_MZ
//}

R abs2(VC2 const& v){
   R sum=0_R;
   for (Z i=1;i<=v.dim();++i) sum+=v.cui(i).abs2();
   return sum;
}

//void SpinNa::adj_dt_()
//{
//   Word loc("SpinNa::adj_dt_()");
//   Z mL=1;
//   CPM_MA
//   R huge=1.e12_R;
//   R fac=cpminv(4_R);
//      // 10 instead of 20 was my best guess from previous applications.
//      // For the present application the stable timestep seems to
//      // be a growing function of the time span tTot. Propably
//      // adaptive time stepping is what is needed
//   setEps_(rc_.eps); // in most cases eps<0 most useful
//   SpinNa s0=*this;
//   Graph gr;
//   gr.setAutoScale(true);
//   R dt0=rep_.dt();
//   R tTot=dt0*rc_.nSteps;
//   R tTotInv=cpminv(tTot);
//   Z nIni=100_Z;
//   R dt=tTot/nIni;
//   Z verbMem=cpmverbose;
//   cpmverbose=3;
//   cpmvalue("dt",get_dt(),2);
//   cpmverbose=verbMem;
//label:
//   *this=s0; // start always from the same initial state
// //  CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*dt);
//   set_dt_(dt); // give the present value of dt to *this
//   verbMem=cpmverbose;
//   cpmverbose=3;
//   cpmvalue("dt",get_dt(),2);
//   cpmverbose=verbMem;
//   vector<R> hv;
//   while (t_<tTot){
//      stepLean_();
//      cpmprogress("run",tTotInv*t_,3);
//      R a2=abs2(rep_.x())+abs2(vv_);
//      hv.push_back(a2);
//      if ( a2 > huge){
//         gr.clr_();
//         R_Vector hr(hv);
//         gr.mark(hr); 
//         gr.vis();
//         cpmwait(4.,2);
//         dt*=0.5_R;
//         goto label;
//      }
//   }
//   dt*=fac; // last safety
//   *this=s0;
//      // has valid vx_ and vv_
//   set_dt_(dt);
//   verbMem=cpmverbose;
//   cpmverbose=3;
//   rc_.nSteps=tTot/rc_.dt;
//   cpmvalues2("dtFinal, nStepsFinal",rc_.dt,rc_.nSteps,2);
//   rc_.dtAdjMethod=0;
//   cpmverbose=verbMem;
//   CPM_MZ
//}

/////////////////////////// class SpinNaa  ////////////////////////////////

SpinNaa::SpinNaa(SysCon const& sc, RunCon const& rc, DisCon const& dc, 
         R const& t, WC2 const& x ): sc_(sc), rc_(rc), dc_(dc),
         ia_(Ia(sc_.h,sc_.qXY,sc_.qZ)), np_(sc_.np), t_(t), dt_(rc_.dt),
         x_(x)
{  
// notice that all members up to dot_ and v_ are initialized by the 
// standard initialization list method.The missing member dot_ will now 
// be initialized comfortably with a lambda expression which makes use of
// the already initialized members.
   Word loc("SpinNaa(SysCon, RunCon, DisCon, R, WC2)");
   Z mL=1;
   CPM_MA
   cpmassert(np_==x_.dim(),"dimension mismatch");
   dot_ = [*this](R const& t, WC2 const& x){
      Outside os=rc_.rigid ? DEFAULT : CYCLIC;
      WC2 res(np_);
      for (Z i=1;i<=np_;++i){
         C2 ci=x.cui(i);
         C2 ri=ia_(ci)+ia_(x(i-1,os),ci)+ia_(x(i+1,os),ci);
         res.cui(i)=ri*C::Ibar;
      }
      return res;
   };
      // In this lambda expression ia_(ci) is the energy of particle i 
      // in an external field,
      // ia_(x(i-1,os),ci) adds the interaction energy from one
      // neighbor particle and ia_(x(i+1,os),ci) adds the interaction 
      // energy from the other neighbor particle. Final multiplication
      // by 1/i transforms energies into contributions to d/dt and hence
      // to dot.
   // Finally we are ready to initialize v_:
   v_=dot_(t_,x_);
   CPM_MZ
}

WC2 SpinNaa::iniState(SysCon const& sc){
   Word loc("SpinNaa::iniState(SysConx const&)");
   Z mL=1;
   CPM_MA
   Z n=sc.np;
   if (sc.random){
      C z0(1.,1.);            
      WC2 res(n);
      for (Z i=1;i<=n;++i){
         C2 ci=C2(z0.ran(),z0.ran());
         ci.nor_();
         res.cui(i)=ci;
      }
      CPM_MZ
      return res;
   }
   else{
      WC2 res(n,C2(sc.theta0,sc.phi0));
      Z dRim=cpmrnd(n * sc.fracRim);
      R flip=sc.flipProbability;
      for (Z i=1;i<=n;++i){
         if (randomR()<flip && i>dRim && i+dRim<=sc.np){
            res[i].swap_();
         }
      }
      CPM_MZ
      return res;
   }
}

C SpinNaa::hInt()const
{
   C res; // initialized as 0
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   for (Z i=1;i<=np_;++i){
      C2 ci=x_.cui(i);
      res+= (ia_.avr(x_(i-1,os),ci)
             + ia_.avr(x_(i+1,os),ci));
   }
   return res*0.5_R; // In a summation over particle pairs
      // we have counted each pair twice and have to correct for this
      // by taking only one half.
}

C SpinNaa::hInt_i(Z i)const
{
   Outside os=rc_.rigid ? DEFAULT : CYCLIC;
   C2 ci=x_.cui(i);
   C z=ia_.avr(x_(i-1,os),ci) + ia_.avr(x_(i+1,os),ci);
   return z*0.5_R;
}
   // hInt_i is defined such that hInt_i(1)+ ... + hInt_i(np) = hInt()
   // This requires the final multiplication by 0.5

void SpinNaa::renormalize_(R const& tol)
{
   for (Z i=1_Z;i<=np_;++i){
      R ni=x_.cui(i).abs();
      if (cpmabs(ni-1_R) > tol ){
         R niInv=1_R/ni;
         x_.cui(i)*=niInv;
         v_.cui(i)*=niInv;
      }
   }
}

R SpinNaa::test2()const
{
   Word loc("SpinNaa::test2()");
   Z mL=1;
   CPM_MA
   C res1=hInt();
   C res2;
   for (Z i=1;i<=np_;++i){
      res2+=hInt_i(i);
   }   
   R res=(res1-res2).abs();
   cout<<" res = "<<res<<", this should be 0 or numerical noise "<<endl;
   cpmcerr<<" res = "<<res<<", this should be 0 or numerical noise "<<endl;
   CPM_MZ
   return res; 
}

C SpinNaa::eSep_i(Z i)const
{
   C2 ci=x_.cui(i);
   return (ci|ia_(ci));
}

C SpinNaa::eTot_i(Z i)const
{
   return eSep_i(i)+hInt_i(i);
}

C SpinNaa::eSep()const
{
   C res;
   for (Z i=1;i<=np_;++i){
      C2 ci=x_.cui(i);
      res += (ci|ia_(ci));
   }
   return res;
}

R3 SpinNaa::sig(Z i)const
{
   C2 ci=x_.cui(i);
   C v1=(ci|CC2::sig1*ci);
   C v2=(ci|CC2::sig2*ci);
   C v3=(ci|CC2::sig3*ci);
   return R3(v1[1],v2[1],v3[1]);
}

VC2 SpinNaa::col()const
{
   if (np_==0) return VC2(0);
   VC2 res(np_);
   for (Z i=1;i<=np_;++i){
      R3 r3=sig(i);
      C z=transform(r3);
      res.cui(i)=C2(z,z);
   }
   return res;
}

void SpinNaa::stepLean_()
{
   R dth=dt_*0.5_R;
   t_+=dth;
   x_+=v_*dth;
   v_+=(dot_(t_,x_)-v_)*2_R;
   x_+=v_*dth;
   t_+=dth;
}
 

void SpinNaa::step_()
{  
   R tMem=t_; 
   VC2 xMem=x_;
   stepLean_();
   R eps=rc_.eps;
   if (eps==0_R) return;
   else if (eps<0_R){
      renormalize_(-eps);
      return;
   }
   else{
      R ex=x_.exc();
      if (ex > rc_.eps){
         t_=tMem;
         x_=xMem;
         x_.nor_();
         v_=dot_(t_,x_);
         dt_*=rc_.facRed;
      }
      else if (ex < rc_.eps*0.1_R){
         dt_*= rc_.facInc;
      }
      else{
         ;
      }
   }
}

void SpinNaa::mark(Graph& g, Z met)const
{
   lineMark_1(g,x_.base(),dc_,met);
}

V<V<R2>> SpinNaa::run_(Graph& g)
{
   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using std::vector;
   Word loc("SpinNaa::run_(Graph,DisConx)");
   Z mL=1;
   CPM_MA
   g.setSclMem_(false);
   R n0=n();
   R n0Inv=cpminv(n0);
   C e0=eTot(); // energy has a small imaginary part in the present context
   C es=eSep(); // energies may have a small imaginary part in the present
      // context. Energy in the external field.
   C eint=hInt(); // contribution of spin-spin-interaction to total
      // energy
   R sig30=sig3();
   R2 blc0=blcAng();
   R theta0=blc0[1];
   R phi0=blc0[2];
   R eTotR=e0[1];
   R eTotAbs=e0.abs();

   cpmvalue("eTotR = ",eTotR,3);
   R eInt=hInt()[1];
   cpmvalues2("eTot,eInt",eTotR,eInt,4);
   R e0Inv=cpminv(e0.abs()); // e0.abs() == 0  does not harm.
   R e01=e0[1]; // real and imaginary part
   R e02=e0[2];
   R es1=es[1]; // real and imaginary part
   R es2=es[2];
   R eint1=eint[1]; // real and imaginary part
   R eint2=eint[2];
   Z nSteps=rc_.nSteps;
   R dt=dt_;
   R tTot=R(nSteps)*dt;
   R tTotInv=1_R/tTot;
   R dt0Inv=cpminv(dt);
   R tWaitDocuNormError=rc_.tWaitDocuNormError;
   R tWaitDocuEnergies=rc_.tWaitDocuEnergies;
   R tWaitDocuEnergyError=rc_.tWaitDocuEnergyError;
   R tWaitBlochAngles=rc_.tWaitBlochAngles;
   R tWaitSigma3=rc_.tWaitSigma3;
   CpmAlgorithms::PeriodicAlarm view(R(rc_.viewPeriod)*dt);
   CpmAlgorithms::PeriodicAlarm docu(R(rc_.docuPeriod)*dt);
   if (rc_.tWaitIni>0){
      mark(g);
      g.setText("         initial state");
      g.vis();
      cpmwait(rc_.tWaitIni,2);
   }
   g.clr_();
   R r0=0_R;
   vector<R> tv(1,r0), nv(1,r0), e1v(1,r0), e2v(1,r0),
      de1v(1,r0), de2v(1,r0),
      es1v(1,r0), es2v(1,r0),eint1v(1,r0), eint2v(1,r0),
      thetav(1,theta0), phiv(1,phi0), sig3v(1,r0);
   bool dtVar=rc_.autoStep;
   R ti=0;
   while (ti<tTot){
      step_();
      ti=t();
//      if (dtVar){
//         cpmvalue("dtRel",dt_*dt0Inv,3);
//      }
      cpmprogress("run",tTotInv*ti);
      if (docu(ti)){
         R ni=n();
         R sig3i=sig3();
         C ei=eTot();
         C esi=eSep();
         C einti=hInt();
         R2 blc=blcAng();
         R thetai=blc[1];
         R phii=blc[2];

         tv.push_back(ti);
            // points in time
         nv.push_back((ni-n0)*n0Inv);
            //

         e1v.push_back(ei[1]);
         e2v.push_back(ei[2]);
         es1v.push_back(esi[1]);
         es2v.push_back(esi[2]);
         eint1v.push_back(einti[1]);
         eint2v.push_back(einti[2]);

         de1v.push_back((ei[1]-e01)*e0Inv);
         de2v.push_back((ei[2]-e02)*e0Inv);

         thetav.push_back(thetai);
         phiv.push_back(phii);
         sig3v.push_back(sig3i-sig30);
      }
      if (view(ti)||ti>=tTot){
         mark(g);
         if (ti>=tTot) g.vis(true); // always conserve the last frame
         else g.vis(dc_.wrtOnShow);
         cpmwait(rc_.tWaitView);
      }
   }
   Z dv=tv.size();
   V<R2> nv2(dv);
   V<R2> e1v2(dv);
   V<R2> e2v2(dv);
   V<R2> es1v2(dv);
   V<R2> es2v2(dv);
   V<R2> eint1v2(dv);
   V<R2> eint2v2(dv);
   V<R2> de1v2(dv);
   V<R2> de2v2(dv);
   V<R2> thetav2(dv);
   V<R2> phiv2(dv);
   V<R2> sig3v2(dv);
   for (Z i=1;i<=dv;++i){
      Z j=i-1;
      R ti=tv[i-1];
      nv2[i]=R2(ti,nv[j]);
      e1v2[i]=R2(ti,e1v[j]);
      e2v2[i]=R2(ti,e2v[j]);
      es1v2[i]=R2(ti,es1v[j]);
      es2v2[i]=R2(ti,es2v[j]);
      eint1v2[i]=R2(ti,eint1v[j]);
      eint2v2[i]=R2(ti,eint2v[j]);
      de1v2[i]=R2(ti,de1v[j]);
      de2v2[i]=R2(ti,de2v[j]);

      thetav2[i]=R2(ti,thetav[j]);
      phiv2[i]=R2(ti,phiv[j]);
      sig3v2[i]=R2(ti,sig3v[j]);
   }

   V<V<R2>> doc1{nv2};
   V<V<R2>> doc2{e1v2,e2v2,es1v2,es2v2,eint1v2,eint2v2};
   V<V<R2>> doc3{thetav2,phiv2};
   V<V<R2>> doc4{sig3v2};
   V<V<R2>> doc5{de1v2,de2v2};

   Z dwl=4; // docu write limit
   bool fileIt=false;
   //Color cbg = FAINTYELLOW;
   Color cbg = LIGHTGREEN;
   if (tWaitDocuNormError>=0){
      Graph g1; // it is neither important nor always useful
      // if diagnostic graphics is of the same size as
      // the evolving image. Hence we use a new frame here.
      g1.clr_();
      fileIt=tWaitDocuNormError>=dwl;
      g1.setColorActive(cbg);
      g1.setText("SpinNaa: evolution of norm error");
      g1.mark(doc1);
      g1.vis(fileIt);
      cpmwait(tWaitDocuNormError,2);
   }
   if (tWaitDocuEnergies>=0){
      Graph g2;
      g2.clr_();
      fileIt=tWaitDocuEnergies>=dwl;
      g2.setWithOrigin(0);
      g2.setColorActive(cbg);
      g2.setText("SpinNaa: evolution of eTot, eSep and hInt");
      g2.mark(doc2);
      g2.vis(fileIt);
      cpmwait(tWaitDocuEnergies,2);
   }
   if (tWaitBlochAngles>=0){
      Graph g3;
      g3.clr_();
      fileIt=tWaitBlochAngles>=dwl;
      g3.setColorActive(cbg);
      g3.setText("SpinNaa: evolution of mean Bloch angles");
      g3.mark(doc3);
      g3.vis(fileIt);
      cpmwait(tWaitBlochAngles,2);
   }
   if (tWaitSigma3>=0){
      Graph g4;
      g4.clr_();
      fileIt=tWaitSigma3>=dwl;
      g4.setColorActive(cbg);
      g4.setText("SpinNaa: evolution of sig3");
      g4.mark(doc4);
      g4.vis(fileIt);
      cpmwait(tWaitSigma3,2);
   }
   if (tWaitDocuEnergyError>=0){
      Graph g5;
      g5.clr_();
      fileIt=tWaitDocuEnergyError>=dwl;
      g5.setWithOrigin(0);
      g5.setColorActive(cbg);
      g5.setText("SpinNaa: evolution of eTot error");
      g5.mark(doc5);
      g5.vis(fileIt);
      cpmwait(tWaitDocuEnergyError,2);
   }
   CPM_MZ
   return V<V<R2>>{nv2,e1v2,e2v2,thetav2,phiv2,sig3v2};
}

