//? cpmqp2.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmqp2.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmFunctions;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;
using namespace CpmImaging;
using namespace CpmLinAlg;
using namespace CpmApplication;
using namespace CpmQuantumMechanics;
using namespace std;

namespace{
// Let us consider elements of Wf2_ as functions f with dom(f) a subset
// of Z2 and ran(f) subset C. Let us define the neighbors of (i,j) in
// Z2 as the following 4 points: (i+1,j),(i-1,j),(i,j+1),(i,j-1).
// Notice that, for instance, (i+1,j+1) is not considered a  neighbor of
// (i,j). The function expand_ returns an extension of the function wf
// by setting the extended function 0 at all points not belonging to
// the domain of wf but being a neighbor of some point in dom(wf).
//
   Wf2_ expand_(Wf2_ const& wf)
   {
      Word loc("expand(Wf2_)");
      Z mL=3;
      CPM_MA
      Wf2_ res=wf;
      Z szi=wf.size();
      C z(0.,0.);
      for (auto ix=wf.begin(); ix!=wf.end(); ++ix ){
         Z2 ij=(*ix).first;

         Z i=ij.x1;
         Z j=ij.x2;

         Z2 rjp(i,j+1);
         auto yjp=wf.find(rjp);
         if(yjp==wf.end()) res[rjp]=z;

         Z2 rjm(i,j-1);
         auto yjm=wf.find(rjm);
         if(yjm==wf.end()) res[rjm]=z;

         Z2 rip(i+1,j);
         auto yip=wf.find(rip);
         if(yip==wf.end()) res[rip]=z;

         Z2 rim(i-1,j);
         auto yim=wf.find(rim);
         if(yim==wf.end()) res[rim]=z;
      }
      Z szf=res.size();
      cpmassert(szf>szi,"expand_(Wf2_ const& wf)");
      CPM_MZ
      return res;
   }
}
/////////////////// Wf2 implementation ////////////////////////////////////

R Wf2::opNor(F<Wf2,Wf2> const& op, IvZ const& ivz1,
   IvZ const& ivz2, Z const& nSteps, bool graph)
{
   Z jRan=137;
   R nStepsInv=1./nSteps;
   Wf2 psi = Wf2::ran(ivz1, ivz2,jRan);
   Wf2 chi = op(psi);
   Graph gr;
   GraphControl gc;
   R eta=chi.abs();
   R_Vector etaVec(nSteps);
   for (Z i=1;i<=nSteps;++i){
      cpmprogress("opNor",i*nStepsInv);
      psi=chi*cpminv(eta);
      if (graph){
         gr.clear();
         psi.mark(gr,gc);
         gr.vis();
      }
      chi=op(psi);
      eta=chi.abs();
      etaVec[i]=eta;
   }
   if (graph){
      gr.clear();
      gr.mark(etaVec);
      gr.vis();
      cpmwait(10,2);
   }
   return etaVec.sup();
}

Wf2 Wf2::con()const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      res[i]=(*ix).second.con();
   }
   return res;
}

Wf2 Wf2::re()const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      res[i]=(*ix).second.real();
   }
   return res;
}

Wf2 Wf2::im()const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      res[i]=(*ix).second.imag();
   }
   return res;
}

C Wf2::operator|(Wf2 const& psi)const
{
   C sum(0.);
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      auto iy=psi.rep_.find(i);
      if(iy!=psi.rep_.end()) sum+=(*ix).second.con()*(*iy).second;
   }
   return sum;
}

Wf2 Wf2::mult( F<R2,R> const& f, R2 const& d)const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      C z=(*ix).second;
      res[i]=z*f(R2(d.x1*i.x1, d.x2*i.x2));
   }
   return res;
}

Wf2 Wf2::operator + (Wf2 const& psi)const
{
   map<Z2,C> res;
   for (auto p=rep_.begin(); p!=rep_.end(); ++p ){
      Z2 i=(*p).first;
      res[i]+=(*p).second;
   }
   for (auto q=psi.rep_.begin(); q!=psi.rep_.end(); ++q ){
      Z2 i=(*q).first;
      res[i]+=(*q).second;
   }
   return res;
}

Wf2 Wf2::operator - (Wf2 const& psi)const
{
   map<Z2,C> res;
   for (auto p=rep_.begin(); p!=rep_.end(); ++p ){
      Z2 i=(*p).first;
      res[i]+=(*p).second;
   }
   for (auto q=psi.rep_.begin(); q!=psi.rep_.end(); ++q ){
      Z2 i=(*q).first;
      res[i]-=(*q).second;
   }
   return res;
}

Wf2 Wf2::operator * (C const& z)const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      res[i]=(*ix).second * z;
   }
   return res;
}

Wf2 Wf2::operator * (R const& r)const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      res[i]=(*ix).second * C(r);
   }
   return res;
}

Wf2 Wf2::operator + (Z2 const& i)const
{
   map<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 j=(*ix).first;
      res[j+i]=(*ix).second ;
   }
   return res;
}

// shorter code for '+=' than for '+' and similar
// operations.
Wf2& Wf2::operator += (Wf2 const& psi)
{
   for (auto q=psi.rep_.begin(); q!=psi.rep_.end(); ++q ){
      Z2 i=(*q).first;
      rep_[i]+=(*q).second;
   }
   return *this;
}

Wf2& Wf2::operator -= (Wf2 const& psi)
{
   for (auto q=psi.rep_.begin(); q!=psi.rep_.end(); ++q ){
      Z2 i=(*q).first;
      rep_[i]-=(*q).second;
   }
   return *this;
}

Wf2& Wf2::operator *= (C const& z)
{
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      (*ix).second *= z;
   }
   return *this;
}

Wf2& Wf2::operator *= (R const& r)
{
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      (*ix).second *= r;
   }
   return *this;
}

R Wf2::nor_()
{
   R r=abs();
   R rInv=cpminv(r);
   (*this)*=rInv;
   return r;
}

X3<IvZ,IvZ,R> Wf2::biotope()const
{
   Word loc("Wf2::biotope");
   Z mL=3;
   CPM_MA
   Z2 cL(INT_MAX,INT_MAX);// lower corner 'paradox' initialisation
   Z2 cU(INT_MIN,INT_MIN); // upper corner  'paradox' initialisation
   R absMax=-1.;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      Z i1=i.x1;
      Z i2=i.x2;
      R a=(*ix).second.abs();
      if (i1<cL.x1) cL.x1=i1;
      if (i1>cU.x1) cU.x1=i1;
      if (i2<cL.x2) cL.x2=i2;
      if (i2>cU.x2) cU.x2=i2;
      if (a > absMax) absMax=a;
   }
   CPM_MZ
   return X3<IvZ,IvZ,R>(IvZ(cL.x1,cU.x1),IvZ(cL.x2,cU.x2),absMax);
}

void Wf2::mark(Graph& g, GraphControl const& gc)const
{
   Word loc("Wf2::mark");
   Z mL=3;
   CPM_MA
   auto ivzs=biotope();
   IvZ ivz1=ivzs.c1();
   IvZ ivz2=ivzs.c2();
   R absMax=ivzs.c3();
   Z2 cL(ivz1.inf(),ivz2.inf());
   Z2 cU(ivz1.sup(),ivz2.sup());

   Z n1=cU.x1-cL.x1+1;
   Z n2=cU.x2-cL.x2+1;

   Z s1=cL.x1-1;
   Z s2=cL.x2-1;

   V<C> column(n2,C());
   V<V<C>> mat(n1,column);

   Z delta=gc.hw_;
   Z d2=delta*delta;
   C cc = absMax*C::expi(gc.ha_*cpmdeg);

   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      Z i1=i.x1;
      Z i2=i.x2;
      R q1=i1*i1;
      R q2=i2*i2;
      if (gc.ac_==0){
         mat[i.x1-s1][i.x2-s2] = (*ix).second;
      }
      else{
         mat[i.x1-s1][i.x2-s2] = ( i1==0 && q2<d2)||( i2==0 && q1<d2) ?
          cc : (*ix).second;
      }
   }
   g.markVVC(mat,false,Iv(),gc.im_);
   g.vis();
   CPM_MZ
}

M<Z2,C> Wf2::toM()const
{
   M<Z2,C> res;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      C v=(*ix).second;
      res[i]=v;
   }
   return res;
}

bool Wf2::prnOn( ostream& str )const
{
   return toM().prnOn(str);
}

bool Wf2::scanFrom( istream& str )
{
   return toM().scanFrom(str);
}

//Wf2 Wf2::dot(R2 const& d, R2 const& m, F<R,R> const& f12,
//      F<R,R> const& f2)const{
//   Word loc("WF2::dot(...)");
//   Z mL=3;
//   CPM_MA
//   R g1=cpminv(-m.x1*d.x1*d.x1*2.);
//   R g2=cpminv(-m.x2*d.x2*d.x2*2.);
//   Wf2_ res;
//   Wf2_ rep=expand_(rep_);
//  // Wf2_ rep=rep_;
//   for (auto p = rep.begin(); p != rep.end(); ++p){
//      Z2 pij=(*p).first;
//      Z i=pij.x1;
//      Z j=pij.x2;
//      Z ip=i+1;
//      Z im=i-1;
//      Z jp=j+1;
//      Z jm=j-1;
//      C wij=(*p).second;
//      C wip,wim,wjp,wjm; // all initialized as C(0)
//      Z2 zip(ip,j);
//      Z2 zim(im,j);
//      Z2 zjp(i,jp);
//      Z2 zjm(i,jm);
//
//      auto yip=rep.find(zip);
//      if(yip!=rep.end()) wip=(*yip).second;
//
//      auto yim=rep.find(zim);
//      if(yim!=rep.end()) wim=(*yim).second;
//
//      auto yjp=rep.find(zjp);
//      if(yjp!=rep.end()) wjp=(*yjp).second;
//
//      auto yjm=rep.find(zjm);
//      if(yjm!=rep.end()) wjm=(*yjm).second;
//
//      R rj=j*d.x2;
//      R rij=R2(i*d.x1,rj).abs();
//
//      C hij = (f12(rij) + f2(rj))*wij +
//            g1*(wip + wim - wij*2.) + g2*(wjp + wjm - wij*2.);
//
//      res[Z2(i,j)] = C::Ibar*hij;
//   }
//   CPM_MZ
//   return Wf2(res);
//}

Wf2 Wf2::dot(R2 const& d, R2 const& m, F<R,R> const& f12,
      F<R,R> const& f2)const{
   Word loc("WF2::dot(...)");
   Z mL=3;
   CPM_MA
   R g1=cpminv(-m.x1*d.x1*d.x1*2.);
   R g2=cpminv(-m.x2*d.x2*d.x2*2.);
   C g1c=C::Ibar*g1;
   C g2c=C::Ibar*g2;
   R g12=(g1+g2)*R(-2);
   Wf2_ res;
   for (auto p = rep_.begin(); p != rep_.end(); ++p){
      Z2 pij=(*p).first;
      Z i=pij.x1;
      Z j=pij.x2;
      Z ip=i+1;
      Z im=i-1;
      Z jp=j+1;
      Z jm=j-1;
      C wij=(*p).second;
    //  C wipj,wimj,wijp,wijm; // all initialized as C(0)
      Z2 zij(i,j);
      Z2 zipj(ip,j);
      Z2 zimj(im,j);
      Z2 zijp(i,jp);
      Z2 zijm(i,jm);

      R rj=j*d.x2;
      R rij=R2(i*d.x1,rj).abs();

      C dij=C::Ibar*(g12 + f12(rij) + f2(rj))*wij;
      C dipj=g1c*wij;
      C dimj=g1c*wij;
      C dijp=g2c*wij;
      C dijm=g2c*wij;

      res[zij]  += dij;
      res[zipj] += dipj;
      res[zimj] += dimj;
      res[zijp] += dijp;
      res[zijm] += dijm;
   }
   CPM_MZ
   return Wf2(res);
}

Wf2 Wf2::smooth()const
// not useful so far
{
   Word loc("WF2::smooth()");
   Z mL=3;
   CPM_MA
   Wf2_ res;
   Wf2_ rep=expand_(rep_);
   for (auto p = rep.begin(); p != rep.end(); ++p){
      Z2 pij=(*p).first;
      Z i=pij.x1;
      Z j=pij.x2;
      Z ip=i+1;
      Z im=i-1;
      Z jp=j+1;
      Z jm=j-1;
      C wij=(*p).second;
      C wip,wim,wjp,wjm;
      Z2 zip(ip,j);
      Z2 zim(im,j);
      Z2 zjp(i,jp);
      Z2 zjm(i,jm);

      auto yip=rep.find(zip);
      if(yip!=rep.end()) wip=(*yip).second;

      auto yim=rep.find(zim);
      if(yim!=rep.end()) wim=(*yim).second;

      auto yjp=rep.find(zjp);
      if(yjp!=rep.end()) wjp=(*yjp).second;

      auto yjm=rep.find(zjm);
      if(yjm!=rep.end()) wjm=(*yjm).second;

      res[Z2(i,j)] = (wip+wim+wjp+wjm)*0.125 + wij*0.5;
   }
   CPM_MZ
   return Wf2(res);
}

Wf2 Wf2::expand()const
{
   return Wf2(expand_(rep_));
}

Wf2 Wf2::der1(R2 const& d)const
{
   R g=cpminv(d.x1*2.);
   Wf2_ res;
   for (auto p = rep_.begin(); p != rep_.end(); ++p){
      Z2 i=(*p).first;
      C w=g*(*p).second;
      Z2 i1p=Z2(i.x1+1,i.x2);
      Z2 i1m=Z2(i.x1-1,i.x2);
      res[i1p]+=w;
      res[i1m]-=w;
   }
   return Wf2(res);
}

Wf2 Wf2::der2(R2 const& d)const
{
   R g=cpminv(d.x2*2.);
   Wf2_ res;
   for (auto p = rep_.begin(); p != rep_.end(); ++p){
      Z2 i=(*p).first;
      C w=g*(*p).second;
      Z2 i2p=Z2(i.x1,i.x2+1);
      Z2 i2m=Z2(i.x1,i.x2-1);
      res[i2p]+=w;
      res[i2m]-=w;
   }
   return Wf2(res);
}

Wf2 Wf2::ran(IvZ const& ivz1, IvZ const& ivz2, Z j)
{
   C z(1.,1.);
   Wf2 res;
   for (Z i1=ivz1.b();i1<=ivz1.e();++i1){
      for (Z i2=ivz2.b();i2<=ivz2.e();++i2){
         res.rep_[Z2(i1,i2)]=z.ran(j++);
      }
   }
   res.nor_();
   return res;
}

void Wf2::cut_(R const& eps)
{
   if (eps<=0.) return;
   R sup=-1.;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
//      Z2 i=(*ix).first;
      C v=(*ix).second;
      R s=v.abs();
      if (s > sup) sup=s;
   }
   R er=eps*sup;
   for (auto ix=rep_.begin(); ix!=rep_.end(); ++ix ){
      Z2 i=(*ix).first;
      C v=(*ix).second;
      R s=v.abs();
      if (s<er) rep_.erase(i);
   }
 }

 R Wf2::test1(IvZ const& ivz1, IvZ const& ivz2)
 {
   Z j=137;
   R errorSum=0.;
   Wf2 w1=ran(ivz1,ivz2,++j);
   w1=w1.expand();
   Wf2 w2=ran(ivz1,ivz2,++j);
   Wf2 w3=ran(ivz1,ivz2,++j);
   w3=w3.expand();
   w3=w3.expand();
   w3=w3.expand();
   Wf2 wL=w1+(w2+w3);
   Wf2 wR=(w1+w2)+w3;
   R e1=(wL-wR).abs();
   errorSum+=e1;
   wL=w1+w2;
   wR=w2+w1;
   e1=(wL-wR).abs();
   errorSum+=e1;
   Wf2 w4=w1-w2+w3;
   w4-=w1;
   w4+=w2;
   w4-=w3;
 //  w4-=w1+=w2-=w3;
   e1=w4.abs();
   errorSum+=e1;
   cpmcerr<<"errorSum: "<<errorSum<<endl;
   e1=((w1|(w2+w3))-(w1|w2)-(w1|w3)).abs();
   errorSum+=e1;
   cpmcerr<<"errorSum: "<<errorSum<<endl;
   e1=((w2|w3)-~(w3|w2)).abs();
   errorSum+=e1;
   cpmcerr<<"errorSum: "<<errorSum<<endl;
   C z;
   C z0=z.test(100);
   C z1=z0.ran();
   C z2=z0.ran();
   C z3=z0.ran();
   cpmcerr<<"z1: "<<z1<<"z2: "<<z2<<"z3: "<<z3<<endl;
   wL=(w1+w3)*z1;
   wR=w1*z1+w3*z1;
   e1=(wL-wR).abs();
   errorSum+=e1;
   cpmcerr<<"errorSum: "<<errorSum<<endl;
   return errorSum;
 }

////////////////////// class Qp2 //////////////////////////////////////////

void Qp2::adjustStep_()
{
   cpmcerr<<"Qp2::adjustStep_() called at dRelNorm_ = "<<dRelNorm_<<endl;
   R dtMem=dt_;
   R hoNormMem=hoNorm_;
   psi_.nor_();
   norm_=1.;
   auto bio=psi_.biotope();
   hoNorm_ = Wf2::opNor(ho_,bio.c1(),bio.c2(),normSteps_);
   dt_=dtRel_*2.*cpminv(hoNorm_);
   cpmcerr<<"||ho|| changed from "<<hoNormMem<<" to "<<hoNorm_<<endl;
   cpmcerr<<"dt_ changed from "<<dtMem<<" to "<<dt_<<endl;
   phi_=ho_(psi_)*C::Ibar;
   avr_=(C::I*(psi_|phi_)).real();
   dRelAvr_=(avr_-avr0_)/avr0_;
   cpmcerr<<"dRelAvr_: "<<dRelAvr_<<endl;
   dRelNorm_=0;
   cpmcerr<<"dRelNorm_ forced to "<<dRelNorm_<<endl;
}

void Qp2::step_()
{
   count_++;
   R tau=dt_*0.5;
   R tauh=tau*0.5;
   t_+=tauh;
   psi_+= phi_*tauh;
   phi_= ho_(psi_)*C(0.,-2.) - phi_;
 //  phi_.cut_(eps_);
   psi_+= phi_*C(tau);
   t_+=tau;
   phi_= ho_(psi_)*C(0.,-2.) - phi_;
//   phi_.cut_(eps_);
   psi_+= phi_*tauh;
   t_+=tauh;
   psi_.cut_(eps_);
   norm_=psi_.abs();
   avr_=(C::I*(psi_|phi_)).real();
   dRelNorm_=(norm_-norm0_)/norm0_;   // = cpmabs(norm_-1.)/1.
   eta_=dRelNorm_;
   if (!etaValid_){ // then eta_ does not yet have its final value
      R etaMem=eta_;
      eta_=dRelNorm_;
      if (eta_< etaMem){
         etaValid_=true;
         cpmcerr<<"final eta_: "<<eta_<<endl;
      }
         // eta_ decreases for the first time. If dRelNorm_ will again
         // grow to this value this will be taken as a trigger for
         // updating hoNorm_
   }
   dRelAvr_=(avr_-avr0_)/avr0_;
   if (etaValid_ && (cpmabs(dRelNorm_) > eta_)) adjustStep_();
}

C fFunc(R2 const& x, R2 const& x0, R2 const& w, R2 const& p){
   return C(1.);
//   R xr=(x-x0)/w;
//   return C::expi(x*p)*exp(-0.5*xr*xr);
}

Wf2 p1Func(Wf2 const& psi, R2 const& d){
   return psi.der1(d)*C::I;
}

Wf2 p2Func(Wf2 const& psi, R2 const& d){
   return psi.der2(d)*C::I;
}

