//? cpmstate4.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmstate4.cpp
   History: Started 2011-03-16

*/

#include <cpmstate4.h>

using CpmQuantumMechanics::Y;
using CpmQuantumMechanics::ShowCon;
using CpmQuantumMechanics::State4;
using CpmQuantumMechanics::Trj4;

using CpmQuantumMechanics::Observable;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGraphics::Z2;
using CpmImaging::Color;
using CpmAlgorithms::CyclicAlarm;
using namespace std;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;


//////////////////////////////// class State4 ////////////////////////////

using CpmFunctions::R_Func;

State4::State4(V<C> const& v1, V<C> const& v2, V<C> const& v3,
               V<C> const& v4, Iv const& iv):
   rep_(
      IvZ(CpmRootX::inf<Z>(v1.dim(),v2.dim(),v3.dim(),v4.dim()),1,137)
      ),
      iv_(iv),
      x_(iv_.cen(rep_.dim()))
{
   for (Z p=rep_.b(),i=v1.b(),j=v2.b(),k=v3.b(),l=v4.b();
      p<=rep_.e();++p,++i,++j,++k,++l){
      rep_.cui(p)=C4(v1.cui(i),v2.cui(j),v3.cui(k),v4.cui(l));
   }
}

   State4::State4(Z n, Iv const& iv):rep_(n),iv_(iv),
      x_(iv_.cen(rep_.dim()))
   {
      static R x0=1./sqrt(2.);
      C z0(x0,x0);
      for (Z i=1;i<=n;++i){
         C4 ci;
         ci.x1=z0.ran();
         ci.x2=z0.ran();
         ci.x3=z0.ran();
         ci.x4=z0.ran();
         rep_[i]=ci;
      }
   }

   State4::State4(Z n, Iv const& iv, T4< F<R,C> > const& f):
      rep_(n),iv_(iv),x_(iv_.cen(rep_.dim()))
   {
      for (Z i=1;i<=n;++i){
         R xi=x_[i];
         rep_[i]=C4(f.c1()(xi),f.c2()(xi),f.c3()(xi),f.c4()(xi));
      }
      nor_();
   }


bool State4::prnOn(ostream& str)const
{
   cpmwt("State4");
   cpmp(rep_);
   cpmp(iv_);
   return true;
}

bool State4::scanFrom(istream& str)
{
   cpms(rep_);
   cpms(iv_);
   return true;
}

namespace{ // clean encapsulation of implementation details

  void accFun(C& z, C const& zAdd){ z+=zAdd;}

   C scalProdFun(C4 const& x1, C4 const& x2){ return (x1|x2);}

   C4 conjFun(C4 const& z ){ return z.con();}

   C4 negFun(C4 const& z ){ return -z;}

   C4 diffFun(C4 const& z1, C4 const& z2){ return z1-z2;}

   C4 sumFun( C4 const& z1, C4 const& z2){ return z1+z2;}

   C4 multFun( C4 const& z1, C const& z){ return z1*z;}

   C4 multFunR( C4 const& z1, R const& r){ return z1*r;}

   C4 diracFun(C4 const& x_1, C4 const& x0, C4 const& x1, Y const& y, Z k)
   {
      R h=y.get_h();
      R m=y.get_m();
      bool nr=y.get_nr();
      R hi=R(0.5)/h;
      if (nr) hi/=(m*h); // non-relativistc dynamics; only the 1 component
      // moves
      R phi_k=y.get_phi()[k];
      if (!nr){
         return ((x_1-x1)*hi).rev()-(x0*C(0,m)).beta()+x0*C(0,phi_k);
      }
      else{
         C4 res;
         C z_1=x_1.x1,z0=x0.x1,z1=x1.x1;
         res.x1=z0*C(0,-phi_k) +((z_1-z0)+(z1-z0))*C(0,hi);
         return res;
      }
   }
}

State4 State4::con(void)const
{
   return State4(rep_.fa(conjFun),iv_);
}

C State4::operator |(State4 const& x2)const
{
   return rep_.fAcc3<C>(scalProdFun,accFun,x2.rep_);
}

State4 State4::operator - (void)const
{
   return State4(rep_.fa(negFun),iv_);
}

State4 State4::operator +(State4 const& x2)const
{
   return State4(rep_.fe(sumFun,x2.rep_),iv_);
}

State4 State4::operator -(State4 const& x2)const
{
   return State4(rep_.fe(diffFun,x2.rep_),iv_);
}

State4 State4::operator *(C const& z)const
{
   return State4(rep_.fb2<C>(multFun,z),iv_);
}

State4 State4::operator *(R const& r)const
{
   return State4(rep_.fb2<R>(multFunR,r),iv_);
}

/************/
State4 State4::dot(Y const& y)const
{
   return State4(rep_.fh<Y>(diracFun,y,CYCLIC),iv_);
}
/*********************/

State4 State4::rev()const
{
   Z n=dim();
   V<C4> rep(n);
   for (Z i=1;i<=n;++i) rep[i]=rep_[i].rev();
   return State4(rep,iv_);
}

State4 State4::beta()const
{
   Z n=dim();
   V<C4> rep(n);
   for (Z i=1;i<=n;++i) rep[i]=rep_[i].beta();
   return State4(rep,iv_);
}

State4 State4::setPhi(V<R> const& phi)const
{
   Z n=dim();
   V<C4> rep(n);
   for (Z i=1;i<=n;++i) rep[i]=rep_[i]*phi[i];
   return State4(rep,iv_);
}

 void State4::boost_(R frac)
 {
    if (frac==0.) return;
    Z n=dim();
    for (Z i=1;i<=n;++i){
       rep_[i]*=expI(frac*x_[i]);
    }
 }

R State4::enrNorm(Y const& y, R eps, Z iterMax, bool randomize)const
   {
      Word loc("State4::enrNorm(Y,R,Z)");
      R res=-1000;
      State4 x=randomize ? State4(dim(),iv_) : *this;
         // statistical object for randomize==true
      x.nor_();
      Z iter=0;
      bool work=true;
      while (work){
         iter++;
         x=x.dot(y)*C::I;
         R resNew=x.nor_();
         R diff=cpmabs(res-resNew)/(cpmabs(res)+cpmabs(resNew));
         if (diff<eps ||iter>iterMax) {
            work=false;
            res=(res+resNew)*0.5;
            if (diff>eps) cpmwarning(loc,"aim not reached");
         }
         else res=resNew;
      }
      return res;
   }

/************* working **************
State4 State4::dot(Y const& y)const
{
   R m=y.get_m();
   V<R> phi=y.get_phi();
   State4 st=*this;
   return -st.d().rev()+ (st*(m*C::Ibar)).beta()+st.setPhi(phi)*C::Ibar;
}
****************/

/*************
void State4::showAsCurves(Graph& g, ShowCon const& sc)const
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   Z nAmp=sc.ir.car();
   Z i,j,k,d=dim(),m=1+nAmp;
   V<Color> cl=colors9;
   R_Matrix res(m,d);
   for (i=1;i<=d;i++){
      res[1][i]=x_[i];
   }
   for (i=1;i<=d;++i){
      C4 val=rep_[i];
      for (j=sc.ir.b(),k=2;j<=sc.ir.e();++j,++k){
         res[k][i]=val[1+(j-1)/2][2-j%2];
      }
   }
   g.clear();
   bool asMem=g.getAutoScl();
   g.setAutoScale(sc.autoScale);
   if (!sc.autoScale){
      g.setX(sc.xr);
      g.setY(sc.yr);
   }
   Word mes("State4: ");
   mes&=(cpm(nAmp)&" real curves for "&cpm(nAmp/2)&" complex amplitudes");
   g.addText(mes);
   g.show(res,cl);
   g.setAutoScale(asMem);
}
**************/

void State4::showAsCurves(Graph& g, ShowCon const& sc)const
// simpler logic, shows all components together with absolute
// value. Easier to compare with Mathematica
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   Z i,j,d=dim(),m=10;
   V<Color> cl=colors9;
   R_Matrix res(m,d);
   for (i=1;i<=d;i++){
      res[1][i]=x_[i];
   }
   for (i=1;i<=d;++i){
      C4 val=rep_[i];
      for (j=1;j<=4;++j){
         C vj=val[j];
         Z j2=j*2;
         res[j2][i]=vj[1];
         res[j2+1][i]=vj[2];
      }
      res[10][i]=val.abs();
   }
   g.clear();
   bool asMem=g.getAutoScl();
   g.setAutoScale(sc.autoScale);
   if (!sc.autoScale){
      g.setX(sc.xr);
      g.setY(sc.yr);
   }
   Word mes("State4: ");
   mes&="9 real curves for 4 complex amplitudes + abs value";
   g.addText(mes);
   g.show(res,cl);
   g.setAutoScale(asMem);
}

State4 State4::d()const
{
   Z n=dim();
   R h=(x_.last()-x_.fir())/(n-1);
   R fac=R(0.5)/h;
   V<C4> rep(n);
   rep[1]=(rep_[2]-rep_[n])*fac;
   rep[n]=(rep_[1]-rep_[n-1])*fac;
   for (Z i=2;i<n;++i) rep[i]=(rep_[i+1]-rep_[i-1])*fac;
   return State4(rep,iv_);
}

/////////////////////////////// class Y //////////////////////////////////
void Y::showPhi(Graph g)const
{
   Z n=dim();
   R_Matrix res(2,n);
   for (Z i=1;i<=n;++i){
      res[1][i]=x_[i];
      res[2][i]=phi_[i];
   }
   g.show(res);
}

////////////////////////////////// class ShowCon /////////////////////////
ShowCon::ShowCon(RecordHandler const& rch, Word section)
{
   Word sec=section;
   cpmrh(autoScale);
   V<R> ivX,ivY;
   cpmrh(ivX);
   cpmrh(ivY);
   xr=Iv(ivX[1],ivX[2]);
   yr=Iv(ivY[1],ivY[2]);
   V<Z> szi;
   cpmrh(szi);
   ir=S<Z>(szi);
   cpmrh(tWait);
   cpmrh(tWait2);
   cpmrh(tWait3);
   cpmrhf(imgPeriod);
   cpmrhf(appendix);
   cpmrhf(fac);
   cpmrhf(secFac);
}

///////////////////////////////////// class Trj4 /////////////////////////

void Trj4::step_()
   // direct midpoint integrator or asynchronous leapfrog
{
   if (y_.get_dmi()){ // direct midpoint integrator (dmi)
      R tau=dt_*0.5;
      t_+=tau;
      x_+=(v_*tau);
      v_+=x_.dot(y_).dot(y_)*dt_;
      x_+=(v_*tau);
      t_+=tau;
   }
   else{ // asynchronous leap-frog (alf)
      R tau=dt_*0.25;
      t_+=tau;
      x_+=(v_*tau);
      State4 vm=x_.dot(y_);
      R tau2=tau+tau;
      v_=vm*2.0-v_;
      x_+=(v_*tau2); // here two operations go into one
      // compared to the previous commentarized form
      t_+=tau2;
      vm=x_.dot(y_);
      v_=vm*2.0-v_;
      x_+=(v_*tau);
      t_+=tau;
   }
}

void Trj4::run_(R tFinal, Graph g)
{
   Z n=cpmrnd(tFinal/dt_);
   R dtAct=tFinal/n;
   dt_=dtAct;
      // we adjust dt such that the final time will be reached exaxtly
   C enr0=get_e();
   R n20=(x_|x_)[1];
   vector<R> enr1;
   vector<R> enr2;
   vector<R> n2;
   vector<R> ts;
   C enrDiff;
   R n2Diff=0;
   if (sc_.tWait>0.){ // showing the initial state
      x_.showAsCurves(g,sc_);
      cpmwait(sc_.tWait,2);
   }
   CyclicAlarm ca(sc_.imgPeriod);
   R tFinalInv=cpminv(tFinal);
   for (Z i=1;i<=n;++i){
      if (ca()){
         x_.showAsCurves(g,sc_);
         enrDiff=get_e()-enr0;
         n2Diff=(x_|x_)[1]-n20;
         enr1.push_back(enrDiff[1]);
         enr2.push_back(enrDiff[2]);
         n2.push_back(n2Diff);
         ts.push_back(t_);
         cpmprogress("",t_*tFinalInv);
      }
      step_();
   }
   bool wosMem=g.getWrtOnShow();
   if (sc_.appendix.dim()>0){
      g.setWrtOnShow(true);
      CpmSystem::Message::setAppRunId(true);
   }
   if (sc_.tWait2>0.){ // showing the final state
      x_.showAsCurves(g,sc_);
      cpmwait(sc_.tWait2,2);
   }
   g.setWrtOnShow(wosMem);

   R_Matrix res(4);
   res[1]=R_Vector(ts);
   res[2]=get_dmi() ? R_Vector(enr1)*sc_.fac : R_Vector(enr1);
      // for the dmi-method Im(x_|v_) is constant up to numerical noise,
      // so sc_.fac needs to be large (>1e9) to let this appear as
      // non-constant
   res[3]=R_Vector(enr2);
   res[4]=get_dmi() ? R_Vector(n2) : R_Vector(n2)*sc_.fac;
      // res[4] (or res[2]) may be much smaller than res[3], (depending
      // on the the time stepping method used).
   if (sc_.tWait3>0.){
      g.clr_();
      g.show(res);
      cpmwait(sc_.tWait3,2);
   }
}

void Trj4::runLean_(R tFinal, Graph g)
{
    if (sc_.tWait>0.){ // showing the initial state
      x_.showAsCurves(g,sc_);
      cpmwait(sc_.tWait,2);
   }
   CyclicAlarm ca(sc_.imgPeriod);
   while ((get_t()<tFinal&&dt_>0)||(get_t()>tFinal&&dt_<0)){
      if (ca()){
         x_.showAsCurves(g,sc_);

      }
      step_();
   }
   if (sc_.tWait2>0.){ // showing the final state
      x_.showAsCurves(g,sc_);
      cpmwait(sc_.tWait2,2);
   }
}

R absFunc(C4 const& c4){ return c4.abs();}
R cut(R const& r){ return r <0.1 ? 0. : (r-0.1)/0.9;}

State4 conc(State4 const& x)
{
   V<C4> vc4=x.get_rep();
   F<C4,R> f(absFunc);
   V<R> vr=vc4(f);
   R_Vector rr(vr);
   R rrm=rr.sup();
   R rrmi=cpminv(rrm);
   rr*=rrmi;
   rr=rr(F<R,R>(cut));
   for (Z i=vc4.b();i<=vc4.e();++i){
      vc4[i]*=rr[i];
   }
   Iv iv=x.get_iv();
   State4 res(vc4,iv);
   res.nor_();
   return res;
}

void Trj4::concentrate_()
{
   State4 x=x_;
   x_=conc(x);
   adj_();
}
