//? cpmstateX.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmstate.cpp
   History: Started 98-10-12

*/

#include <cpmobservable.h>
#include <cpmbiotope.h>
#include <cpmimagingtools.h>
#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmv.h>
#include <cpms.h>
#include <cpmrvector.h>

/*************** struct QMBase *****************************************/

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::Bio;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGraphics::Z2;
using CpmImaging::Color;
using namespace std;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;

using CpmFunctions::R_Func;

namespace{
   R aux_=1./cpmsqrt(2.);
}

R QMBase::phiStep=0.025;
Z QMBase::verbose=1;

R QMBase::c0=1;
R QMBase::hDirac=1;
R QMBase::hDiracInv=R(1)/hDirac;
R QMBase::hPlanck=hDirac*cpm2pi;
C QMBase::ihDirac=C(0,hDirac);
C QMBase::ihDiracInv=C(0,-hDiracInv);
R QMBase::ihDiracInv2=-hDiracInv*hDiracInv;

void QMBase::adjust()
{
   hPlanck=hDirac*cpm2pi;
   hDiracInv=R(1)/hDirac;
   ihDirac=C(0,hDirac);
   ihDiracInv=C(0,-hDiracInv);
   ihDiracInv2=-hDiracInv*hDiracInv;
}

void QMBase::setUnits(Word const& w)
{
   if (w=="SI"){
      hDirac=CpmPhysics::h__PHYS;
      c0=CpmPhysics::c_PHYS;
   }
   else{
      hDirac=1;
      c0=1;
   }
   adjust();
}

R QMBase::mass=1;
R QMBase::size=1;

R QMBase::mMax(Z d)
{
   return hDirac*R(d)/(size*c0*2);
}

R QMBase::pMax(Z d)
{
  return hPlanck*R(d)/(size*2);
}

R QMBase::vMax(Z d)
{
  return hPlanck*R(d)/(size*mass*2);
}

R QMBase::eMax(Z d)
// eMax=(2/m)*(hDirac/ds)^2, ds=(size/d);
// maximum kinetic energy in free motion.
{
   R ds=size/d;
   R y=hDirac/ds;
   return (R(2)/mass)*y*y;
}

R QMBase::tRef(Z d)
{
   return (size*size*mass*2)/(hPlanck*d);
}

R QMBase::dtLimit(Z d)
{
   R ds=size/R(d);
   return ds*ds*mass*2/hPlanck;
}

R QMBase::dtStabilityLimit(Z d)
{
   return hDirac*2/eMax(d);
}

R QMBase::lambdaOsc(Z d)
      // (hPlanck*d/(size^2))^2 / (2*mass)
{
   R val=hPlanck*d/(size*size);
   return val*val*0.5/mass;
}

V<R> QMBase::positions( Z d)
{
   R sh=size*0.5;
   return Bio(d,Iv(-sh,sh)).points();
}

V<R> QMBase::posNor(Z d)
{
   R sh=0.5;
   return Bio(d,Iv(-sh,sh)).points();
}


V<R> QMBase::frequencies( Z d)
// This is the case where negative indexes would be natural.
// The new version of V supports this. Not yet done.
{
   Z mL=1;
   Word loc( "QMBase::V<R> frequencies(" );
   loc&=cpm(d);
   loc&=", ";
   loc&=(cpm(size)&")");
   CPM_MA
   cpmassert(d>1,loc);
   Z i,dh=d/2; // d half
   cpmassert(d==dh*2,loc);
   // assures that n is even
   R delta=size/d; // sampling spacing resulting from dividing an
     // interval of length size into d equal parts
   R fc=R(0.5)/delta; // Nyquist frequency (see Press et al. p. 500...)
   R fSpacing=fc*2/d;
   V<R> res(d);
   R val=0;
   for (i=1;i<=dh+1;i++){
      res[i]=val;
      val+=fSpacing;
   }
   val=-fc+fSpacing;
   for (i=2;i<=dh;i++){
      res[dh+i]=val;
      val+=fSpacing;
   }
   CPM_MZ
   return res;
}

R QMBase::rt(Z i){return i>=0 ? cpmsqrt(R(i)) : R(0);}
void QMBase::accFunc(C& z, const C& zAdd){ z+=zAdd;}
C QMBase::conjFunc(const C& z){ return z.con();}
C QMBase::negFunc(const C& z1){  return -z1;}
C QMBase::sumFunc(const C& z1, const C& z2){ return z1+z2;}
C QMBase::diffFunc(const C& z1, const C& z2){ return z1-z2;}
C QMBase::prodFunc(const C& z1, const C& z2){ return z1*z2;}
C QMBase::prodFuncConj1(const C& z1, const C& z2){return z1.con()*z2; }
C QMBase::prodFuncConj2(const C& z1, const C& z2){ return z1*z2.con();}
C QMBase::prodFuncReal(const C& z1, const C& z2){ return z1*z2.real();}
C QMBase::divFunc(const C& z1, const C& z2){ return z1/z2;}
C QMBase::facPhi=C(aux_,0); // 1/sqrt(2)
C QMBase::facPsi=C(0,aux_); // i/sqrt(2)

/************************* class State *********************************/

// State::shift(...), is defined in cpmobservables.cpp
// since we better have observables available as tools.

State::State(const R_Func& f, Z d):dL_(1),rep_(d)
{
   Iv iv(-0.5,0.5);
   V<R> xList=iv.centers(d);
   for (Z i=1;i<=d;i++) rep_[i]=C(f(xList[i]),0);
   normalize_();
}

State::State(const CpmFunctions::F<R,C>& f, Z d):dL_(1),rep_(d)
{
   R xL=-0.5, xR=0.5;
   V<R> xList=CpmImaging::ImagingTools::pixelCenters(xL,xR,d);
   for (Z i=1;i<=d;i++) rep_[i]=f(xList[i]);
   normalize_();
}

State::State(const R_Vector& arr, Z d)
{
   R_Func fb=R_Func::Bernstein(arr, R2(-0.5,0.5));
   *this=State(fb,d);
}

State::State(Z d, Z n1, Z n2, Z n3):dL_(1),rep_(d),n1_(n1),n2_(n2),n3_(n3)
{
   Z mL=3;
   Word loc("State(Z,Z,Z,Z)");
   CPM_MA
   cpmassert(n1_>=0,loc);
   cpmassert(n2_>=0,loc);
   cpmassert(n3_>=0,loc);
   cpmassert(d>=0,loc);
   Z s=n1_+n2_+n3_;
   cpmassert(s<=d,loc);
   CPM_MZ
}

void State::setBra_(Z n1, Z n2, Z n3)
{
   Z mL=3;
   Word loc("State::setBra_(Z,Z,Z)");
   CPM_MA
   n1_=n1;
   n2_=n2;
   n3_=n3;
   cpmassert(dL_==1,loc);
   cpmassert(n1_>=0,loc);
   cpmassert(n2_>=0,loc);
   cpmassert(n3_>=0,loc);
   Z d=rep_.dim();
   cpmassert(d>=0,loc);
   Z s=n1_+n2_+n3_;
   cpmassert(s<=d,loc);
   for (Z i=1;i<=s;++i) rep_.cui(i)=C(0); // clean the interferometer
   CPM_MZ
}

R State::normalize_(R r)
{
   Z d=rep_.dim();
   if (d==0) return 0.;

   R r0=norm();
   if (r0==0){
      State aux(d);
      aux[1]=C(r);
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

State State::makeReal()const
{
   Z d=rep_.dim();
   State res(d);
   for (Z i=1;i<=d;++i)
      res.rep_.cui(i)=C(rep_.cui(i).abs());
   return res;
}

State State::symmetrize()const
{
   Z d=rep_.dim();
   State res(d);
   for (Z i=1,ic=d;i<d;++i,--ic){
      res.rep_.cui(i)=(rep_.cui(i)+rep_.cui(ic))*0.5;
   }
   return res;
}

State State::smooth(bool normalizeRes)const
{
   Z d=rep_.dim();
   State res(d);
   for (Z i=1;i<=d;++i){
      res.rep_.cui(i)=
         (rep_.cyc(i-1)*0.25+rep_.cui(i)*0.5+rep_.cyc(i+1)*0.25);
   }
   if (normalizeRes) res.normalize_();
   return res;
}

// fast implementation using natural iteration over pointers

State State::con(void)const
{
   return State(dL_,rep_.fa(conjFunc),n1_,n2_,n3_);
}

C State::operator |(const State& x2)const
{
   return rep_.fAcc2(QMBase::prodFuncConj1,QMBase::accFunc,x2.rep_);
}

R State::variation()const
{
   Z d=dim();
   R res=0;
   for (Z i=1;i<d;i++) res+=(rep_[i+1]-rep_[i]).abs();
   return res;
}

State State::operator - (void)const
{
   return State(dL_,rep_.fa(negFunc),n1_,n2_,n3_);
}

State& State::operator +=(const State& x)
{
   return *this=*this+x;
}

State& State::operator -=(const State& x)
{
   return *this=*this-x;
}

State& State::operator *=(C z)
{
   rep_.fc_(prodFunc,z);
   return *this;
}

State& State::operator *=(R r)
{
   rep_.fc_(prodFuncReal,C(r));
   return *this;
}

State State::operator +(const State& x2)const
{
  return State(dL_,rep_.fe2(QMBase::sumFunc,x2.rep_),n1_,n2_,n3_);
}

State State::operator -(const State& x2)const
{
   return State(dL_,rep_.fe2(QMBase::diffFunc,x2.rep_),n1_,n2_,n3_);
}

State State::operator *(C x2)const
{
  State res(*this);
  return res*=x2;
}

State State::operator *(R x2)const
{
  State res(*this);
  return res*=x2;
}

bool State::prnOn(ostream& str)const
{
   cpmwt("State");
   cpmp(dL_);
   cpmp(rep_);
   cpmp(n1_);
   cpmp(n2_);
   cpmp(n3_);
   return true;
}

bool State::scanFrom(istream& str)
{
   cpms(dL_);
   cpms(rep_);
   cpms(n1_);
   cpms(n2_);
   cpms(n3_);
   return true;
}

State State::operator ||(const State& s2)const
// beautyful symmetry of computation
{
   Z d1=dim();
   Z d2=s2.dim();
   State res(d1*d2);
   Z k=0;
   for (Z i=1;i<=d1;i++){
      C vali=rep_[i];
      for (Z j=1;j<=d2;j++){
         res[k]=vali*s2.getLean(j);
         k++;
      }
   }
   return res;
}

State State::sinParHam(R2_Func const& pot, R t, R lambda)const
{
   Z mL=3;
   Word loc("State::sinParHam(R2_Func,R,R)");
   CPM_MA
   R hD2=hDirac*hDirac;
   Z i,d=dim();
   R dxNorm=1./d;
   R dxInv2=R(d)/size; // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(mass*2);
      // since the - is here, we have to take the second derivative and
      // n o t  its negative. So we take the coefficients directly from
      // Bronstein ... p. 768 Tabelle 7.13
   State res(d);
   Outside per=CYCLIC; // periodic
   for (i=1;i<=d;++i){
      C ri=(rep_(i+1,per)+rep_(i-1,per)-rep_(i)*2.)*fac;
      res[i]=ri;
   }
   // now res=Kinetic energy operator applied to *this.
   // What follows are the potential contributions.
   if (lambda!=0){
      V<R> x0=posNor(d);
      for (i=1;i<=d;++i){
         R xi=x0[i];
         C xit(xi,t);
         R pxit=pot(xit)*lambda;
         if (pxit!=0) res[i]+=rep_(i)*pxit;
      }
   }
   CPM_MZ
   return res;
}

/****************** old and good
State State::freeParHam()const
{
   Z mL=3;
   Word loc("State::freeParHam()");
   CPM_MA
   R hD2=hDirac*hDirac;
   Z i,d=dim();
   R dxNorm=1./d;
   R dxInv2=d/size; // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(2*mass);
   State res(d);
   Outside per=CYCLIC; // periodic
   for (i=0;i<d;++i){
      C ri=(rep_(i+1,per)+rep_(i-1,per)-rep_(i)*2.)*fac;
      res[i]=ri;
   }
   CPM_MZ
   return res;
}
*/

Z2 State::pos(Z i)const
// we assume 0<=i<dim()
{
// we assume n3_<=n2_ (not clear whether really needed).
// We have to shift all the points but the n_1+n_2 first ones.
// The shift is uniform for the next n3_ ones
// and uniform (but different) for all following ones
// condition for s1: n1_+n2_+s1=n_1, thus s1=-n2_
// condition for s2: n1_+n2_+n3_+s2=n_1+n_2, thus s1=-n3_
   if (n2_==0 || n3_==0) return Z2(i,1);
   Z l1=n1_+n2_, l2=l1+n3_, s1=-n2_, s2=-n3_; // +1 experiment
   if (i<l1) return Z2(i,1);
   else if (i<l2) return Z2(i+s1,2);
   else return Z2(i+s2,1);
}

State State::freeParHam()const
// trying to implement the branching logic
{
   Z mL=3;
   const R facL=3.;
      // factor in the Laplacian at the branching point.
      // 3 seems graph theoretically more canonical than 2
      // I had 2 first
   Word loc("State::freeParHam()");
   CPM_MA
   R hD2=hDirac*hDirac;
   Z i,d=dim();
   R dxNorm=R(1)/d;
   R dxInv2=R(d)/size; // (1/dx)^2, dx=size/d
   dxInv2*=dxInv2;
   R fac=-hD2*dxInv2/(mass*2);
   State res(d,n1_,n2_,n3_);
   Outside per=CYCLIC; // periodic
   if (n1_==0 || n2_==0){ // then no branching
      for (i=0;i<d;++i){
         C ri=rep_(i+1,per)+rep_(i-1,per)-rep_.cui(i)*2.;
         res.cui(i)=ri;
      }
      CPM_MZ
      return res*fac;
   }
// there are 5 special points:
   Z p1=n1_-1;
   Z p2=p1+n2_;
   Z p3=p2+1;
   Z p4=p2+n3_;
   Z p5=p4+1;
   for (i=0;i<p1;++i){
      C ri=rep_.cui(i+1)+rep_(i-1,per)-rep_.cui(i)*2.;
      res.cui(i)=ri;
   }
   for (i=p1+1;i<p2;++i){
      C ri=rep_.cui(i+1)+rep_.cui(i-1)-rep_.cui(i)*2.;
      res.cui(i)=ri;
   }
   for (i=p3+1;i<p4;++i){
      C ri=rep_.cui(i+1)+rep_.cui(i-1)-rep_.cui(i)*2.;
      res.cui(i)=ri;
   }
   for (i=p5+1;i<d;++i){
      C ri=rep_(i+1,per)+rep_.cui(i-1)-rep_.cui(i)*2.;
      res.cui(i)=ri;
   }
   res.cui(p1)=rep_.cui(p3)+rep_.cui(p1+1)+rep_.cui(p1-1)-rep_.cui(p1\
   )*facL;
   res.cui(p2)=rep_.cui(p2-1)+rep_.cui(p5)-rep_.cui(p2)*2.;
   if (n3_>1){
      res.cui(p3)=rep_.cui(p1)+rep_.cui(p3+1)-rep_.cui(p3)*2.;
      res.cui(p4)=rep_.cui(p5)+rep_.cui(p4-1)-rep_.cui(p4)*2.;
   }
   else{ // n3_==1 is slightly special
      res.cui(p3)=rep_.cui(p1)+rep_.cui(p5)-rep_.cui(p3)*2.;
      res.cui(p4)=rep_.cui(p5)+rep_.cui(p1)-rep_.cui(p4)*2.;
   }
   res.cui(p5)=rep_.cui(p5+1)+rep_.cui(p4)+rep_.cui(p2)-rep_.cui(p5)*facL;
   CPM_MZ
   return res*fac;
}

V< V<C> > State::factorize(Z d1, Z d2)const
{
   Z mL=3;
   Word loc("State::factorize(Z,Z)");
   cpmmessage(mL,loc&" started");
   Z i,j,d=dim();
   cpmassert(d==d1*d2,loc);
   Z k=0;
   V< V<C> > res(d1,V<C>(d2));
   for (i=1;i<=d1;i++){
      for (j=1;j<=d2;j++){
         res.cui(i)[j]=rep_[k];
         k++;
      }
   }
   cpmmessage(mL,loc&" done");
   return res;
}

State State::symPrd(const State& s2)const
{
   return ((*this)||s2 + s2||(*this))*C(0.5);
}

State State::antSymPrd(const State& s2)const
{
   return ((*this)||s2 - s2||(*this))*C(0.5);
}

State State::crt()const
{
   Z n=dim();
   Z nr=n+1;
   State res(nr);
   for (Z i=2;i<=nr;++i) res[i]=rep_[i-1]*rt(i);
   return res;
}

State State::des()const
{
   Z n=dim();
   if (n==0) return *this;
   Z nr=n-1;
   State res(nr);
   for (Z i=1;i<=nr;++i) res[i]=rep_[i+1]*rt(i+1);
   return res;
}

R_Vector State::toRealForm()const
{
   Z i,i2,d=dim();
   Z realDim=2*d;
   R_Vector res(realDim);
   C val;
   for (i=1;i<=d;i++){
      i2=i+d;
      val=rep_.cui(i);
      res[i]=val[1];
      res[i2]=val[2];
   }
   return res;
}

void State::fromRealForm(const R_Vector& r)
{
   Word loc("State::fromRealForm(R_Vector)");
   Z i,i2,realDim=r.dim();
   Z d=realDim/2;
   cpmassert(2*d==realDim,loc);
   rep_=V<C>(d);
   R valr,vali;
   for (i=1;i<=d;i++){
      i2=i+d;
      valr=r[i];
      vali=r[i2];
      rep_.cui(i)=C(valr,vali);
   }
}

State State::toMomSpc()const
{
   Z1 d(dim());
   cpmassert(dL_==1,"State::toMomSpc()");
   cpmassert(d.isPowOfTwo(),"State::toMomSpc()");
   V<C> temp=toArray();
   temp=CpmFourier::fft(temp,1);
   return State(dL_,temp);
}

State State::frmMomSpc()const
{
   Z1 d(dim());
   cpmassert(dL_==1,"State::frmMomSpc()");
   cpmassert(d.isPowOfTwo(),"State::frmMomSpc()");
   V<C> temp=toArray();
   temp=CpmFourier::fft(temp,-1);
   return State(dL_,temp);
}

R State::sup()const
{
   Z d=dim();
   R res=rep_.cui(1).abs();
   for (Z i=2;i<=d;i++){
      R r=rep_.cui(i).abs();
      if (r>res) res=r;
   }
   return res;
}

/*************** old and good **************************/
void State::show(Frame& f, Z mult_i, Z mult_j, Z is, Z js,
   bool ignorePhases)const
// rather fast implementation
{
   Z mL=1;
   Word loc("State::show(Frame,Z,Z,Z,Z,bool)");
   CPM_MA
   R valMax=sup();
   if (ignorePhases) valMax*=valMax;
   R valMaxInv=R(1)/valMax;
   Z i,j,d=dim();
   Z d_=d*mult_i;
   Z jShift=1+js;

   Z iState=1;
   Z iCounter=0;
   Z iEff=is;
   Color c;

   for (i=0;i<d_;i++){
      iEff++;
      if (iCounter==0){
         C val=rep_.cui(iState);
         if (ignorePhases) val=C(val.absSqr());
         val*=valMaxInv;
         c=Color(val);
      }
      iCounter++;
      if (iCounter==mult_i){
         iCounter=0; // triggers new evaluation of c in the next loop
         iState++;
      }
      for (j=0;j<mult_j;j++){
         f.put(Z2(iEff,j+jShift),c);
      }
   }
   CPM_MZ
}


// new version taking n1_,n2_,n3_ into account
/************************
void State::show(Frame& f, Z mult_i, Z mult_j, Z is, Z js,
   bool ignorePhases)const
{
   Z mL=1;
   Word loc("State::show(Frame,Z,Z,Z,Z,bool");
   CPM_MA
   bool lean=(n2_==0 || n3_==0);
   Z ff=lean? 1 : 4;
   mult_j*=ff;
   R valMax=sup();
   if (ignorePhases) valMax*=valMax;
   R valMaxInv=1./valMax;
   Z i,d=dim();
   Z d_=d*mult_i;
   Z jShift=(1+js)*ff;

   Z iState=0;
   Z iCounter=0;
   Z iEff=is;
   Color c;
   Z multjh= lean ? mult_j :  mult_j/2;
   Z mi= lean ? mult_i : (n2_*mult_i)/n3_;

   for (i=1;i<=d;i++){
      C val=rep_.cui(i);
      if (ignorePhases) val=C(val.absSqr());
      val*=valMaxInv;
      c=Color(val);
      Z2 z2=pos(i);
      Z sj=0,p1=z2[1],p2=z2[2];
      Z multAct;
      if (p2==1){
         multAct=mult_i;
         p1*=multAct;
      }
      else{
         multAct=mi;
         p1=n1_*mult_i+(p1-n1_)*multAct;
         sj=multjh;
      }
      Z pa=p1;
      for (Z ki=0;ki<multAct;++ki){
         for (Z kj=0;kj<multjh;++kj){
             f.put(Z2(pa,kj+sj+jShift),c);
         }
         pa++;
      }
   }
   CPM_MZ
}
***************************/

void State::show(Z d1, Z d2, Frame& f, Z mult_i, Z mult_j,
      Z iShift, Z jShift)const
{
   V< V<C> > vvc=factorize(d1,d2);
   for (Z i=1;i<=d1;i++){
      State s(dL_,vvc[i],n1_,n2_,n3_);
      s.show(f,mult_i,mult_j,iShift,jShift);
      jShift+=mult_j;
   }
}

bool State::orthogonal(V<State>& x, R eps)const
      // The return value indicates whether *this is orthogonal to all
      // components of x. If this value is true, x will be appended to
      // x, so that one can collect vectors into a orthogonal base.
{
   Z i,n=x.dim();
   R ti=norm();
   for (i=1;i<=n;i++){
      R xi=x[i].norm();
      if ((*this|x[i]).abs()>eps*xi*ti ) return false;
   }
   x&=*this;
   return true;
}

void State::adjPhase(R eps)
{
   Z mL=3;
   Word loc("State::adjPhase(R)");
   CPM_MA
   Z i,d=dim();
   Vo<R> absv(d);
   for (i=1;i<=d;i++) absv[i]=rep_.cui(i).abs();
   V<Z> per=absv.permutationForIncreasingOrder();
   Z iMax=per[d];
   R absMax=absv[per[d]];
   if (absMax==0) return;
   S<Z> is;
   for (i=1;i<=d;i++){
      if (CpmRoot::dis(absv[i],absMax)<=eps) is.add(i);
   }
   cpmassert(is.car()>0,loc);
   Z iMaxEff=is[1]; // sets are always ordered !
   C cEff=(*this)[iMaxEff];
   R r=cEff.abs();
   cpmassert(r>0,loc);
   C phase=cEff/r;
   phase=phase.con();
   for (i=1;i<=d;i++) rep_.cui(i)*=phase;
   CPM_MZ
}

V<State> State::E_Schmidt(V<State>& x, R eps)
   // returns an orthonormal family which spans the same space
   // as family x. Made by E. Schmidt's orthogonalization. Notice that
   // we do not assume that the family x be linear independent.
   // The procedure goes through family x in the order of indexing.
   // It is evident that the procedure is most stable if the states
   // are ordered in the sense of increasing complexity (structured-ness;
   // e.g. number of peaks, or total variation.
   // Built-in are features specific to the situation of
   // spectralRepresentation
{
   static const Z nCrit=8;
      // if the degeneracy of an eigenvalue is larger than this, the
      // standard orthonormalization procedure begins to become very
      // susceptible to numerical errors. We the do reordering

   Z mL=3;
   Word loc="State::E_Schmidt(V<State>,R)";
   CPM_MA
   Z d=x.dim();
   Z dh=d/2;
   cpmassert(2*dh==d,loc);
   if (dh>nCrit){
      Vo<R> var(d);
      for (Z i=1;i<=d;i++) var[i]=-x[i].variation(); // '-' important
      V<Z> per=var.permutationForIncreasingOrder();
         // due to the '-' in the definition of var, we order for
         // decreasing variation. My intuition was to do the opposite.
         // When this was observed not to work, I tried the present
         // version without any expectation for making things better.
         // they became much better, though.
       x=x.permute(per);
   }

   V<State> u(dh);
   Z iTest=1,iOK=0;
   while (iTest<=d && iOK<dh){
      State vTest=x[iTest];
      State next=vTest;
      for (Z i=1;i<=iOK;i++) next-=u[i]*(u[i]|vTest);
      iTest++;
      R rn=next.normalize_();
      if (rn>eps){
         iOK++;
         u[iOK]=next;
      }
   }
   if (iOK<dh){
      Z iMissing=dh-iOK;
      Word mes="State::E_Schmidt(...) "&cpm(iMissing);
      mes&=" states missing of "&cpm(dh);
      cpmmessage(mes);
   }
   CPM_MZ
   return u;
}

void State::show(Graph& g, V<State> const& x)
      // shows a familie of states as curves
{
   Z mL=3;
   Word loc("State::show(Graph,V<State>)");
   CPM_MA
   Z n=x.dim();
   cpmassert(n>0,"State::show(...)");
   Z d=x[1].dim();
   R_Vector xl(d);
   R_Matrix yl(n,d);
   Z i,j;
   for (j=1;j<=d;j++) xl[j]=j;
   for (i=1;i<=n;i++) for (j=1;j<=d;j++) yl[i][j]=x[i][j-1].abs();
   g.show(xl,yl);
   CPM_MZ
}

void State::showAsCurves(Graph& g)const
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   Z i,d=dim();
   V<Color> cl(3);
   cl[1]=BLACK; // color for absolute value
   cl[2]=RED;  // color for real part
   cl[3]=GREEN;   // color for imaginary part

   R_Matrix res(4,d);
   for (i=1;i<=d;i++){
      res[1][i]=i;
      C val=rep_[i];
      res[2][i]=val.abs();
      res[3][i]=val[1];
      res[4][i]=val[2];
   }
   g.clear();
   g.addText("a State: Real part red, Imaginary part green, Abs black");
   g.show(res,cl);
}


R State::orthoTest(const V<State>& x)
{
   Z d=x.dim();
   R_Matrix sp(d);
   for (Z i=1;i<=d;i++) for (Z j=1;j<=d;j++)
      sp[i][j]=(x[i]|x[j]).abs();
   R_Matrix unit=sp.net(1);
   R res=(sp-unit).abs();
   if (res>1e-3) cpmdata<<sp;
   return res;
}

namespace{

   R f_dePhase(R x)
   {
      return cpmpi*(-x*x+1)*2;
   }
}

void State::dePhase(R p1, R p2, Z meth )
{
   Z d=dim(),i;
   if (meth<=0) return;
   R refAngle=cpm2pi*p1;
   if (meth==1){ // stochastical phase jitter
      for (i=1;i<=d;i++){
         R phi_i=CpmRoot::ran(refAngle);
         C omega_i=C::expi(phi_i);
         rep_[i]*=omega_i;
      }
   }
   else if (meth==2){
      R alpha=p2*cpmsqrt(1./d);
      for (i=1;i<=d;i++){
         C psi_i=rep_[i];
         R phi_i=CpmRoot::ran(refAngle);
         C omega_i=C::expi(phi_i);
         rep_[i]+=omega_i*alpha; // adding a complex vector
            // of lenght controlled by p2
      }
      R rb=normalize_();
      if (rb==0)
         cpmerror("void State::dePhase(R,R,Z): unexpected rb==0");
   }
   else if (meth==3){
      C omega=C::expi(refAngle);
      // complex number of unit modulus
      V<R> w(d);
      for (i=1;i<=d;i++){
         C xi=omega*rep_[i];
         // each value of the wave function gets multiplied by omega
         R wi=xi.real();
         // we take the real part of the 'rotated' wave function.
         // Depending on the phase of xi, this may be swmall even
         // for large values of |xi|

         wi=cpmpow(cpmabs(wi),p2);
         rep_[i]*=wi;
      }
      R rb=normalize_();
      if (rb==0)
         cpmerror("void State::dePhase(R,R,Z): unexpected rb==0");
   }
   else{
      ;
   }
}

R_Vector State::probabilities()const
{
   Z d=dim();
   R_Vector res(d);
   for (Z i=1;i<=d;i++){
      R r=rep_[i].abs();
      res[i]=r*r;
   }
   return res;
}

State State::ran(Z j)const
{
   Z d=dim();
   C z(1,1);
   Vr<C> vc(d,z);
   V<C> resPrel=vc.ran(j);
   return State(dL_,resPrel);
}

State State::boost(R v)const
{
   Z d=dim();
   cpmassert(dL_==1,"State::boost(R v)");
   State res=*this;
   V<R> pos=QMBase::positions(d);
   R fac=v*QMBase::mass*QMBase::hDiracInv;
   for (Z i=1;i<=d;++i){
      R val=fac*pos[i];
      res.rep_[i]*=C(cpmcos(val),cpmsin(val));
   }
   return res;
}

State State::corGrn(Z nPhi, Z n)const
{
   using CpmGeo::RAD;
   R nInv=1./n;
   Z i,d=dim();
   V<C> r(d);
   for (i=1;i<=d;++i){
      R rho,phi;
      rep_.cui(i).toPolar(rho,phi);
      Angle a(phi,RAD); // degrees are default
      a.qun_(nPhi); // quantize the angle
      R ci=rho*rho;
      ci=nInv*cpmrnd(ci*n);
      R ri=cpmsqrt(ci);
      r.cui(i)=C(ri,a,"polar");
   }
   State res(dL_,r,n1_,n2_,n3_);
   res.nor_();
   return res;
}

bool State::col_(IvZ const& ivz, R p)
{
   static bool trapped=false;
   if (trapped) return false;
   R r=CpmRoot::randomR();
   if (r>p ) return false;
   IvZ iz=rep_.dom()&ivz;
   R w=0;
   Z i;
   for (i=iz.b();i<=iz.e();++i){
      w+=rep_.cui(i).absSqr();
   }
   bool res=false;
   r=CpmRoot::randomR();
   if (r<w){ // then the ivz-components survive
      // the others get to zero, 'state is trapped'
      for (i=rep_.b();i<=rep_.e();++i){
         if (!iz.hasElm(i)) rep_.cui(i)=C(0);
      }
      nor_();
      trapped=true;
      return true;
   }
   else return false;
}
