//? cpmstatedirac.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


/*
   cpmstate.cpp
   History: Started 2014-01-?

*/

#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmframes.h>
#include <cpmvr.h>
#include <cpms.h>
#include <cpmrvector.h>
#include <cpmstatedirac.h>
#include <cpmmultiimage.h>
#include <cpmfile.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::Bio;
using CpmQuantumMechanics::StateD;
using CpmQuantumMechanics::OpD;
using CpmQuantumMechanics::SpectralObjectD;
using CpmQuantumMechanics::TrjD;
using CpmQuantumMechanics::SpTrjD;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::SX;
using CpmLinAlg::R2;
using CpmGraphics::Frame;
using CpmGraphics::Frames;
using CpmGraphics::Graph;
using CpmGraphics::Z2;
using CpmGeo::Iv;
using CpmImaging::Color;
using CpmImaging::MultiImage;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmFunctions::R_Func;
using CpmSystem::OFile;
using namespace std;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;

////////////////////////////// class StateD ///////////////////////////////

Z StateD::bc_=1; // 1: periodic
Z StateD::verbose_=1; //
bool StateD::useDALF_=0; // usage of a DALF trajectory for
   // constructing boost operators
bool StateD::enfNor_=1; // enforcing normalization to correct unitarity
   // deficiencies of boost implementation

CC4 StateD::unit_=CpmLinAlg::Gamma(Word("unit")).rep();

CC4 StateD::gam0_=CpmLinAlg::Gamma(0).rep();
CC4 StateD::gam1_=CpmLinAlg::Gamma(1).rep();
CC4 StateD::gam2_=CpmLinAlg::Gamma(2).rep();
CC4 StateD::gam3_=CpmLinAlg::Gamma(3).rep();
CC4 StateD::gam5_=CpmLinAlg::Gamma(5).rep();
CC4 StateD::beta_=CpmLinAlg::Gamma::beta();
CC4 StateD::alpha1_=CpmLinAlg::Gamma::alpha(1);
CC4 StateD::alpha2_=CpmLinAlg::Gamma::alpha(2);
CC4 StateD::alpha3_=CpmLinAlg::Gamma::alpha(3);

void StateD::setWhichGamma(Word const& wg)
{
   gam0_=CpmLinAlg::Gamma(0,wg).rep();
   gam1_=CpmLinAlg::Gamma(1,wg).rep();
   gam2_=CpmLinAlg::Gamma(2,wg).rep();
   gam3_=CpmLinAlg::Gamma(3,wg).rep();
   gam5_=CpmLinAlg::Gamma(5,wg).rep();
   beta_=CpmLinAlg::Gamma::beta(wg);
   alpha1_=CpmLinAlg::Gamma::alpha(1,wg);
   alpha2_=CpmLinAlg::Gamma::alpha(2,wg);
   alpha3_=CpmLinAlg::Gamma::alpha(3,wg);
 }

CC4 StateD::boost1(R const& chi)const
{
   R chih=chi*0.5;
   return unit_*cpmcosh(chih) - alpha1_*cpmsinh(chih);
}

R2 StateD::lorentz1(R2 const& y, R const& chi)
{
   R sc=cpmsinh(chi), cc=cpmcosh(chi);
   R x0=y.x1, x1=y.x2;
   return  R2(x0*cc-x1*sc,x1*cc-x0*sc);
}

StateD::StateD(V<C4> const& v, Iv const& s ):
s_(s), xc_(s_.cen()),rep_(v.dim())
{
   Z d=rep_.dim();
   V<R> x=s_.centers(d);
   for (Z i=1; i<=d; ++i) rep_[i] = SX(v[i],x[i]);
}

StateD::StateD(Bio const& bio ): s_(bio.iv()),
   xc_(s_.cen()),rep_(bio.d())
{
   V<R> x=bio.points();
   C4 c4; // zero vector
   for (Z i=rep_.b(); i<=rep_.e(); ++i) rep_[i] = SX(c4,x[i]);
}

StateD::StateD(C const& z, Bio const& b):
   s_(b.iv()), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C4 c4(z);
   for (Z i=rep_.b(); i<=rep_.e(); ++i) rep_[i] = SX(c4,x[i]);
}

StateD::StateD(Iv const& iv, Bio const& b, Z j):
   s_(b.iv()), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C4 c0,c1;
   c1[1]=C(1.);
   for (Z i=rep_.b(); i<=rep_.e(); ++i){
      R xi=x[i];
      bool b=iv.contains(xi);
      rep_[i] = b ? SX(c1,xi) : SX(c0,xi);
   }
}

StateD::StateD(F<R,C4> const& f, Bio const& b): s_(b.iv()),
   xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   for (Z i=x.b();i<=x.e();i++){
      R xi=x[i];
      rep_[i]=SX(f(xi),xi);
   }
}

StateD::StateD(Z i, Z a, Bio const& b, bool frqSpc):
fft_(B(frqSpc)),s_(b.iv()), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C4 c0; // zero vector
   C4 ca;
   ca[a]=C(1.);
   Z ii=rep_.makeIndVal(i); // understood Cyclic, so i need not be in range
   for (Z j=x.b(); j<=x.e(); ++j){
      rep_[j] = (j==ii ? SX(ca,x[j]) : SX(c0,x[j]));
   }
}

StateD::StateD(Bio const& b, C4 const& c, R const& x0, R const& sigma0,
          R const& k, bool normalize):
         s_(b.iv()), xc_(s_.cen()),rep_(b.d())
{
   V<R> x=b.points();
   for (Z i=x.b();i<=x.e();i++){
      R xi=x[i];
      R yi=(xi-x0)/sigma0;
      yi*=yi;
      C fi=C::expi(xi*k)*exp(-0.5*yi);
      rep_[i]=SX(c*fi,xi);
   }
   if (normalize) nor_(); // constructs a unit vector
}

namespace{

SX negFunc(SX const& a)
{
   SX res=a;
   res.c1()=-a.c1();
   return res;
}

SX conFunc(SX const& a)
{
   SX res=a;
   res.c1()= a.c1().con();
//   res.c1()= StateD::gam0_*(a.c1().con()); // experiment
      // gam0 added 2014-03-25
   return res;
}

SX multFunc(SX const& a, C const& z )
{
   SX res=a;
   res.c1()*=z;
   return res;
}

SX multRFunc(SX const& a, R const& r )
{
   SX res=a;
   res.c1()*=r;
   return res;
}

SX addFunc(SX const& a, SX const& b )
{
   SX res=a;
   res.c1()+=b.c1();
   return res;
}

SX subFunc(SX const& a, SX const& b )
{
   SX res=a;
   res.c1()-=b.c1();
   return res;
}

C cumFunc(const SX& a, const SX& b)
{
   return (a.c1()|b.c1());
}

} // namespace

StateD StateD::operator-(void)const
{
   return StateD(rep_.fa(negFunc),s_);
}

StateD& StateD::operator*=(C z)
{
   rep_.fb2_<C>(multFunc,z);
   return *this;
}

StateD& StateD::operator+=(StateD const& psi)
{
   rep_.ff_(addFunc,psi.rep_);
   return *this;
}

StateD& StateD::operator-=(StateD const& psi)
{
   rep_.ff_(subFunc,psi.rep_);
   return *this;
}

StateD StateD::operator*(C z)const
{
   return StateD(rep_.fb2<C>(multFunc,z),s_);
}

StateD StateD::operator +(StateD const& psi)const
{
   return StateD(rep_.fe(addFunc,psi.rep_),s_);
}

StateD StateD::operator -(StateD const& psi)const
{
   return StateD(rep_.fe(subFunc,psi.rep_),s_);
}

StateD& StateD::operator*=(R r)
{
   rep_.fb2_<R>(multRFunc,r);
   return *this;
}

StateD StateD::operator*(R r)const
{
   return StateD(rep_.fb2<R>(multRFunc,r),s_);
}

namespace{
   SX mFunc(SX const& sx, CC4 const& m){ return SX(m*sx.c1(),sx.c2());}
   SX wFunc(SX const& sx, R_Func const& f)
   {
      C4 s=sx.c1(); R x=sx.c2(); return SX(s*f(x),x);
   }

   SX vFunc(SX const& sx, R const& px)
   {
      SX res=sx;
      res.c1()*=C(0.,-px);
      return res;
   }
   C4 conFunc(C4 const& x){ return x.con();}
}

StateD StateD::applyCC4(CC4 const& mat)const
{
   return StateD(rep_(F1<SX,CC4,SX>(mat)(mFunc)),s_);
}

StateD StateD::rev()const
{
   StateD res=*this;
   res.setVal_(val().rev());
   return res;
}

StateD StateD::cc()const
{
   StateD res=*this;
   V<C4> v=val();
   V<C4> vc=v(conFunc);
   res.setVal_(vc);
   return res;
}

StateD StateD::apply(R_Func const& f)const
{
   return StateD(rep_(F1<SX,R_Func,SX>(f)(wFunc)),s_);
}

StateD StateD::win(Iv const& iv, R const& softParameter)const
{
   return apply(R_Func::charFunc(iv,softParameter));
}

StateD StateD::frqWin(Iv const& iv, R const& softParameter)const
   //: frequency window
   // Restricts the wave function, which is assumed to be given in
   // position space, by a windowing operation which acts on the FFT
   // image. The argument iv which belongs to the full frequency range
   // is Iv(-1,1) in order to be independent of the discretization under
   // consideration.
{
   StateD psi=toFrqSpc();
   Iv ivfr=psi.bio().iv();
   R fr=ivfr[2];
   R x1=iv[1]*fr;
   R x2=iv[2]*fr;
   return psi.win(Iv(x1,x2),softParameter).fromFrqSpc();
}

StateD StateD::applyF(R_Func const& f)const
   //:: apply to discrete Fourier transform.
   // The argument of f ranges from -1 to 1, where 1 corresponds to the
   // spatial Nyquist frequency 1/(2*gap()) and -1 to -1/(2*gap().
   // It is to be noted however that this last most negative value no
   // longer belongs to the indexing range of the discrete Fourier
   // transform.
{
   return *this; // preliminary nonsense
}

StateD StateD::applyPot(R_Vector const& pot)const
{
   Z rd=rep_.dim();
   cpmassert(pot.dim()==rd,"StateD::applyPot");
   return StateD(rep_.fet<R>(vFunc,pot),s_);
}

namespace
{
   SX func(SX const& c_1, SX const& c0, SX const& c1, R const& fac, Z i)
   {
      i;
      return SX((c1.c1()-c_1.c1())*fac,c0.c2());
   }
}

StateD StateD::dx()const
{
   Outside out=(Outside)bc_;
   return StateD(rep_.fh<R>(func,R(0.5)/gap(),out),s_);
}

StateD StateD::dot(R const& m, R_Vector const& pot)const
{
   if (d()==pot.dim()){
      return applyCC4(beta_)*C(0.,-m)
      -dx().applyCC4(alpha1_)
      +applyPot(pot);
   }
   else{
      return applyCC4(beta_)*C(0.,-m)
      -dx().applyCC4(alpha1_);
   }
}

/********************
StateD StateD::w2d()const
{
   R fac=R(1)/cpmsqrt(R(2));
   CC2 a=CC2::unit;
   CC4 u(a,a,-a,a);
   return this->applyCC4(u*fac);
}

StateD StateD::d2w()const
{
   R fac=R(1)/cpmsqrt(R(2));
   CC2 a=CC2::unit;
   CC4 u(a,-a,a,a);
   return this->applyCC4(u*fac);
}
**********************/

StateD StateD::wg2d(Word const& wg)const
{
   if (wg=="Dirac") return *this;
   CC4 U = CpmLinAlg::Gamma(wg).rep();
   return this->applyCC4(U.hc());
}

StateD StateD::d2wg(Word const& wg)const
{
   if (wg=="Dirac") return *this;
   CC4 U = CpmLinAlg::Gamma(wg).rep();
   return this->applyCC4(U);
}

namespace{
   StateD hFunc(StateD const& psi, R const& m, R_Vector const& pot)
   {
      return psi.ham(m,pot);
   }

   StateD xFunc(StateD const& psi){ return psi.x1();}

   StateD pFunc(StateD const& psi){ return psi.dx()*C::Ibar;}

   StateD shiftFunc(StateD const& psi, R const& a)
   {
      R g=psi.gap();
      R ar=a/g;
      Z ariL=cpmtoz(ar);
      Z ariR=ariL+1;
      R arf=ar-ariL;
      R wL=R(1)-arf;
      R wR=arf;
      Z d=psi.d();
      Iv s=psi.s();
      V<C4> v4=psi.val();
      V<C4> w4(d);
      Outside out=(Outside) StateD::getBonCon();
      for (Z i=1;i<=d;++i){
         w4[i]=v4(i-ariL,out)*wL+v4(i-ariR,out)*wR;
      }
      return StateD(w4,s);
   }

   StateD boostLFunc(StateD const& psi, OpD const& h, R const& chi,
                     R const& dt )
   {
      Z mL=StateD::getVer();
      Word loc("boostLFunc");
      CPM_MA
      Z d=psi.d();
      R psiN=psi.abs();
      CC4 cc4=psi.boost1(chi);
      R ch=cpmcosh(chi);
      R sh=cpmsinh(chi);
      R th=sh/ch;
      Bio b=psi.bio()*ch;
      StateD psiT(psi.val(),b.iv());
      TrjD trj(h,psiT,dt); // gives exp(-ith) just as in Domenico's
         // formula
      V<C4> val(d);
      for (Z i=1;i<=d;++i){
         R xi=psiT.xc(i);
         R ti=th*xi;
         trj.evl_(ti);
         StateD si=trj.psi();
         val[i]=cc4*si[i];
      }
      StateD res(val,psi.s());
      R resN=res.abs();
      res*=C(psiN/resN);
      CPM_MZ
      return res;
   }

   StateD boostFunc1(StateD const& psi, OpD const& db, R const& chi,
                    R const& chiStep)
   {
      Z mL=StateD::getVer();
      Word loc("boostFunc1");
      CPM_MA
      R psiN=0_R,resN;
      bool b=StateD::getEnfNor();
      if (b) psiN=psi.abs();
      TrjD trj(db,psi,chiStep);
      trj.evl_(chi);
      StateD res=trj.psi();
      if (b){
         resN=res.abs();
         res*=C(psiN/resN);
      }
      CPM_MZ
      return res;
   }

   StateD boostFunc2(StateD const& psi, OpD const& db, Z const& n)
   {
      Z mL=StateD::getVer();
      Word loc("boostFunc2");
      CPM_MA
      R psiN{0.},resN{0.};
      bool b=StateD::getEnfNor();
      if (b) psiN=psi.abs();
      StateD res=psi;
      for (Z i=1;i<=n;++i) res=db*res;
      if (b){
         resN=res.abs();
         res*=C(psiN/resN);
      }
      CPM_MZ
      return res;
   }
}

OpD StateD::h(Bio const& b, R const& m, R_Vector const& pot)
{
   return OpD(b,F2<StateD,R,R_Vector,StateD>(m,pot)(hFunc));
}

OpD StateD::x(Bio const& b)
{
   return OpD(b,F<StateD,StateD>(xFunc));
}

OpD StateD::p(Bio const& b)
{
   return OpD(b,F<StateD,StateD>(pFunc));
}

OpD StateD::shift(Bio const& b, R const& a)
{
   return OpD(b,F1<StateD,R,StateD>(a)(shiftFunc));
}

OpD StateD::boostL(Bio const& b, R const& m, R const& chi,
                   R const& dtRel)
{
   cpmassert(dtRel>0 && dtRel<1,"StateD::boostL");
   R ch=cpmcosh(chi);
   Bio bch=b*ch;
   OpD h=StateD::h(bch,m); // Lorentz transformed biotope
   R dt=h.dtLim()*dtRel; // important to place time-consuming norm
      // evaluation in the constructor (or static generating function) and
      // not in the workhorse function boostLFunc.
   return OpD(b,F3<StateD,OpD,R,R,StateD>(h,chi,dt)(boostLFunc));
}

OpD StateD::dBoost(Bio const& b, R const& m, R const& deltaChi,
                   R const& dtRel)
{
   R fac=R(0.5)/deltaChi;
   return
   (boostL(b,m,deltaChi,dtRel)-boostL(b,m,-deltaChi,dtRel))*C(0.,fac);
      // second order accurate expression for the derivative at unity
}

OpD StateD::boost(Bio const& b, R const& m, R const& chi,
                  R const& deltaChi, R const& dtRel)
{
   if (getUseDALF()){ // slow and accurate
      OpD db=dBoost(b,m,deltaChi,dtRel);
      R dchi=dtRel*db.dtLim();
      return OpD(b,F3<StateD,OpD,R,R,StateD>(db,chi,dchi)(boostFunc1));
   }
   else{ // fast and not too accurate
      R cs=cpmabs(deltaChi);
      if (cpmabs(chi)<=cs) return boostL(b,m,chi,dtRel);
      Z n=cpmpos(cpmrnd(chi/cs));
      R dchi=chi/n;
      OpD bL=boostL(b,m,dchi,dtRel);
      return OpD(b,F2<StateD,OpD,Z,StateD>(bL,n)(boostFunc2));
   }
}

StateD StateD::con(void)const
{
   return StateD(rep_.fa(conFunc),s_);
}

C StateD::operator |(StateD const& psi)const
{
   return rep_.fAcc3<C>(cumFunc,QMBase::accFunc,psi.rep_);
}

R StateD::nor_(R r)
{
   Z d=rep_.dim();
   if (d==0) return 0.;

   R r0=abs();
   if (r0==0){
      StateD aux=*this;
      (aux.rep_[1].c1())[1]=C(r);
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

C StateD::enr(R const& m, R_Vector const& pot)const
{
   StateD s=*this;
   s.nor_();
   return (s|s.ham(m,pot));
}

StateD StateD::nyq(C4 const& c1, C4 const& c2, Bio const& b)
{
   Z n=b.d();
   cpmassert(n>1,"StateD::nyq");
   V<C4> v(n,c1);
   for (Z i=2;i<=n; ++++i) v[i]=c2;
   return StateD(v,b.iv());
}

R_Vector StateD::valVec()const
{
   Z d_=d();
   R_Vector x(d_);
   for (Z i=1;i<=d_;i++){
      SX sxi=rep_[i];
      R  xi=sxi.c2();
      x[i]=xi;
   }
   return x;
}

R_Matrix StateD::valMat()const
{
   Z d_=d();
   R_Matrix y(9,d_);
   for (Z i=1;i<=d_;i++){
      SX sxi=rep_[i];
      C4 c4=sxi.c1();
      y[1][i]=c4.abs(); // for values close to 0, the more
         // sophisticated version is worse
      C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
      y[2][i]=s1[1];
      y[3][i]=s1[2];
      y[4][i]=s2[1];
      y[5][i]=s2[2];
      y[6][i]=s3[1];
      y[7][i]=s3[2];
      y[8][i]=s4[1];
      y[9][i]=s4[2];
   }
   return y;
}

/**************************
void StateD::showAsCurves(Graph& g)const
{
   using namespace CpmImaging;
   using namespace CpmArrays;

   V<Color> cl(9);
   cl[1]=BLACK; // color for absolute value
   cl[2]=RED;  // color for real part of v1
   cl[3]=LIGHTRED;   // color for imaginary part v1
   cl[4]=GREEN;
   cl[5]=LIGHTGREEN;
   cl[6]=BLUE;
   cl[7]=LIGHTBLUE;
   cl[8]=MAGENTA; // color for real part of v4
   cl[9]=LIGHTMAGENTA; // color for imaginary part v4

   Z d_=d();
   R_Matrix res(10,d_);
   for (Z i=1;i<=d_;i++){
      SX sxi=rep_[i];
      C4 c4=sxi.c1();
      R  xi=sxi.c2();
      res[1][i]=xi; // position
      res[2][i]=c4.abs(false); // for values close to 0, the more
         // sophisticated version is worse
      C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
      res[3][i]=s1[1];
      res[4][i]=s1[2];
      res[5][i]=s2[1];
      res[6][i]=s2[2];
      res[7][i]=s3[1];
      res[8][i]=s3[2];
      res[9][i]=s4[1];
      res[10][i]=s4[2];
   }
   g.clear();
   g.addText("StateD. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.show(res,cl);
}
************************/

void StateD::mark(Graph& g)const
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   X2<R_Vector,R_Matrix> xy=valVecMat();
   V<Color> cl(9);
   cl[1]=BLACK; // color for absolute value
   R fac=0.6;
   Z nL=4;
   V<R> list=Iv(0.,1.).div(nL);
   for (Z i=1;i<=nL;++i){
      Color c1=Color::hue(list[i]);
      Color c2=(c1*fac).addWhite(1.);
      Z j=2*i;
      cl[j]=c1;
      cl[j+1]=c2;
   }
   g.clear();
   g.addText("StateD. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.markGL(xy.c1(),xy.c2(),cl);
}

void StateD::markArr(V<StateD> const& vs, Graph& g)
{
   Z mL=3;
   Word loc("StateD::markArr(V<StateD>,Graph)");
   CPM_MA

   using namespace CpmImaging;
   using namespace CpmArrays;

   V<Color> cl(18);
   cl[1]=LIGHTGRAY;
   R fac=0.6;
   Z nL=4;
   V<R> list=Iv(0.,1.).div(nL);
   for (Z i=1;i<=nL;++i){
      Color c1=Color::hue(list[i]);
      Color c2=(c1*fac).addWhite(1.);
      Z j=2*i;
      cl[j]=c1;
      cl[j+1]=c2;
   }
// darker colors
   cl[10]=BLACK; // color for absolute value
   fac=0.75;
   cl[11]=cl[2]*fac;  // color for real part of v1
   cl[12]=cl[3]*fac;   // color for imaginary part v1
   cl[13]=cl[4]*fac;
   cl[14]=cl[5]*fac;
   cl[15]=cl[6]*fac;
   cl[16]=cl[7]*fac;
   cl[17]=cl[8]*fac; // color for real part of v4
   cl[18]=cl[9]*fac; // color for imaginary part v4

   Z i;
   StateD s1=vs.fir(); // making sure that the states match
   for (i=vs.b(); i<=vs.e(); ++i){
      cpmassert(s1.mat(vs[i]),loc);
   }

   R_Vector x=s1.valVec();
   R_Matrix y=s1.valMat();
   for (i=vs.b()+1; i<=vs.e(); ++i){
      y&=vs[i].valMat(); // appending the matrix part
   }
   g.clear();
   g.addText("StateD. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.markGL(x,y,cl);
   CPM_MZ
}

void StateD::mark2(Graph& g, StateD const& sc, R const& scale)const
      // Shows *this together with a comparison state
      // allows to boost the difference etween the two by a factor.
{
   StateD s1(*this);
   StateD s2(sc);
   StateD sm=(s1+s2)*0.5;
   R sh=scale*0.5;

   StateD sDiff=(s1-s2)*sh;
   StateD v1=sm+sDiff;
   StateD v2=sm-sDiff;
   StateD::markArr(V<StateD>{v1,v2},g);
}

/*********************
void StateD::gp(OFileStream& out)const
      // returns a data file from which gnuplot can construct the
      // showAsCurves representation.
{
   Z mL=StateD::getVer();
   Word loc("gp(Word)");
   CPM_MA
   for (Z i=1;i<=d();i++){
      SX sxi=rep_[i];
      C4 c4=sxi.c1();
      R  xi=sxi.c2();
      R s0=c4.abs(false);
      C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
      out()<<endl<<" "<<xi<<" "<<s0<<" "<<s1[1]<<" "<<s1[2]<<" "
      <<s2[1]<<" "<<s2[2]<<" "<<s3[1]<<" "<<s3[2]<<" "
      <<s4[1]<<" "<<s4[2];
   }
   CPM_MZ
}
***********************************/

void StateD::gp(ostream& out, Iv const& iv)const
      // returns a data file from which gnuplot can construct the
      // showAsCurves representation.
{
   //Z mL=StateD::getVer();
   Z mL=3;
   Word loc("gp(Word)");
   CPM_MA
   if (iv.isVoid()){
      for (Z i=1;i<=d();i++){
         SX sxi=rep_[i];
         C4 c4=sxi.c1();
         R  xi=sxi.c2();
         R s0=c4.abs();
         C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
         out<<endl<<" "<<xi<<" "<<s0<<" "<<s1[1]<<" "<<s1[2]<<" "
         <<s2[1]<<" "<<s2[2]<<" "<<s3[1]<<" "<<s3[2]<<" "
         <<s4[1]<<" "<<s4[2];
      }
   }
   else{
      for (Z i=1;i<=d();i++){
         SX sxi=rep_[i];
         C4 c4=sxi.c1();
         R  xi=sxi.c2();
         R s0=c4.abs();
         C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
         R r= iv.contains(xi) ? R(0.1) : R(0);
         out<<endl<<" "<<xi<<" "<<s0<<" "<<s1[1]<<" "<<s1[2]<<" "
         <<s2[1]<<" "<<s2[2]<<" "<<s3[1]<<" "<<s3[2]<<" "
         <<s4[1]<<" "<<s4[2]<<" "<<r;
      }
   }
   CPM_MZ
}

namespace{
   C4 valF( SX const& sx ){ return sx.c1();}
   C val1F( SX const& sx ){ return sx.c1()[1];}
   C val2F( SX const& sx ){ return sx.c1()[2];}
   C val3F( SX const& sx ){ return sx.c1()[3];}
   C val4F( SX const& sx ){ return sx.c1()[4];}
   R posF( SX const& sx ){ return sx.c2();}
}

/****************/
C4 StateD::at(R const& x)const
{
   Iv iv=s();
   Z n=d();
   X2<R2,Z2> x2=iv.intPolInfo2(x,n);
  // X2<R2,Z2> x2=iv.intPolInfo(x,n);
   R2 ww=x2.c1();
   Z2 ii=x2.c2();
   R w1=ww.x1, w2=ww.x2;
   Z i1=ii.x1, i2=ii.x2;
   return rep_[i1].c1()*w1+rep_[i2].c1()*w2;
}
/****************************/

/**************
C4 StateD::at(R const& x)const
{
   Iv iv=s();
   R p=(x-iv[1])/iv.abs();
   Z i1=cpmrnd(p*d());
   CpmArrays::Outside out=(Outside)bc_;
   return rep_(i1,out).c1();
}
******************/

V<C4> StateD::val()const
{
   F<SX,C4> fval(valF);
   return rep_(fval);
}

V<C> StateD::val1()const
{
   F<SX,C> fval1(val1F);
   return rep_(fval1);
}

V<C> StateD::val2()const
{
   F<SX,C> fval2(val2F);
   return rep_(fval2);
}

V<C> StateD::val3()const
{
   F<SX,C> fval3(val3F);
   return rep_(fval3);
}

V<C> StateD::val4()const
{
   F<SX,C> fval4(val4F);
   return rep_(fval4);
}

R_Vector StateD::pos()const
{
   F<SX,R> fpos(posF);
   return R_Vector(rep_(fpos));
}

StateD StateD::ran(Z j)const
{
   Vr<C4> vr(val());
   return StateD(vr.ran(j),s_);
}

R StateD::dis(StateD const& psi)const
{
   Z mL=3;
   Word loc("StateD::dis(StateD)");
   CPM_MA
   V<C4> v4=val(), v4psi=psi.val();
   V<R> x=pos(), xpsi=psi.pos();

   R_Vector d(3);
   d[1]=v4.dis(v4psi);
   d[2]=x.dis(xpsi);
   d[3]=s_.dis(psi.s_);
   R res=d.supAbs();
   CPM_MZ
   return res;
}

bool StateD::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(fft_);
   cpmp(xc_);
   cpmp(rep_);
   cpmp(s_);
   return true;
}

bool StateD::scanFrom(istream& str)
{
   cpms(fft_);
   cpms(xc_);
   cpms(rep_);
   cpms(s_);
   return true;
}

bool StateD::isVal()const
{
   for (Z i=rep_.b();i<=rep_.e();++i){
      if (!rep_.cui(i).c1().isVal()) return false;
   }
   return true;
}

StateD StateD::toFrqSpc()const
{
   Z n2=d();
   cpmassert(isPowOfTwo(n2),"StateD::toFrqSpc()");
   Z n=n2/2,i,j;
   R fN=R(0.5)/gap(); // Nyquist frequency
   Iv sF(-fN,fN);
   V<R> f=sF.centers(n2);
   V<C> temp1=val1();
   temp1=CpmFourier::fft(temp1,1);
   V<C> temp2=val2();
   temp2=CpmFourier::fft(temp2,1);
   V<C> temp3=val3();
   temp3=CpmFourier::fft(temp3,1);
   V<C> temp4=val4();
   temp4=CpmFourier::fft(temp4,1);

  V<SX> vF(n2);
   for (i=1,j=n+1;i<=n;++i,++j){
      vF[i]=SX(C4(temp1[j],temp2[j],temp3[j],temp4[j]),f[i]);
   }
   for (i=1,j=n+1;i<=n;++i,++j){
      vF[j]=SX(C4(temp1[i],temp2[i],temp3[i],temp4[i]),f[j]);
   }
   StateD res(vF,sF);
   res.fft_=B(true);
   res.xc_=s_.cen(); // The position center should not be lost
   return res;
}

StateD StateD::fromFrqSpc()const
{
   Z n2=d();
   cpmassert(isPowOfTwo(n2),"StateD::fromFrqSpc()");
   Z n=n2/2,i,j;
   R fN=s().e(); // end of frequency biotope
   R dx=R(0.5)/fN;
   R sh=dx*n;
   Iv s(-sh+xc_,sh+xc_); // length = dx*n2 = 2*dx*n
   V<R> x=s.centers(n2);
   V<C> temp1(n2);
   V<C> temp2(n2);
   V<C> temp3(n2);
   V<C> temp4(n2);

   for (i=1,j=n+1;i<=n;++i,++j){
      temp1[i]=val1()[j];
      temp2[i]=val2()[j];
      temp3[i]=val3()[j];
      temp4[i]=val4()[j];
   }

   for (i=1,j=n+1;i<=n;++i,++j){
      temp1[j]=val1()[i];
      temp2[j]=val2()[i];
      temp3[j]=val3()[i];
      temp4[j]=val4()[i];
   }

   temp1=CpmFourier::fft(temp1,-1);
   temp2=CpmFourier::fft(temp2,-1);
   temp3=CpmFourier::fft(temp3,-1);
   temp4=CpmFourier::fft(temp4,-1);

   V<SX> v(n2);
   for (i=1;i<=n2;++i){
      v[i]=SX(C4(temp1[i],temp2[i],temp3[i],temp4[i]),x[i]);
   }
   StateD res(v,s);
   res.fft_=B(false);
   return res;
}

/*****************************
R StateD::momShp(R x)
{
   static R alpha=4.*cpmpi*cpmpi/27.;
   static R xm=cpmsqrt(R(1)/(alpha*3.));
   R xp=cpmabs(x);
   R w=(xp-xm)/(R(1)-xm);
   R y= (xp<=xm ? cpm2pi*xp*(1.-xp*xp*alpha) : R(2)*(1.-w*w));
   return x>= 0. ? y : -y;
}

namespace{
C soiFunc(R const& x, R const& fm, R const& fp)
{
   return x<0 ? C(fm) : C(fp) ;
}
}
**************/

StateD StateD::momFrq()const
// We assume that *this is given in frequency space
{
   Word loc("StateD::momFrq()");
   cpmassert(fft_==true,loc);
   Z d_=d();
   R pmax=d_*gap(); // notice that gap() does not coincide with the
      // lattice spacing in position space; we are in frequency space!
   R fac=cpm2pi/d_;
   StateD res=*this;
   for (Z i=1;i<=d_;++i){
      res.rep_[i].c1()*=(cpmsin(fac*(i-1))*pmax);
   }
   return res;
}

StateD StateD::hamFrq( R mass)const
{
   Word loc("StateD::hamFrq(R)");
   cpmassert(fft_==true,loc);
   return momFrq().applyCC4(alpha1_)+applyCC4(beta_*mass);
}

//////////////////////// class OpD ////////////////////////////////////////

R OpD::norm()const
{
   Z mL=StateD::getVer();
  // Z mL=0;
   Word loc("OpD::norm");
   CPM_MA
   bool graphics=StateD::getVer()>=3;
   //bool graphics=true;
   Z n=500; // 200 was too inaccurate in some cases
   R tWait=12.;
   C c1(1.,1.);
   StateD s(c1,b_);
   s=s.ran(1370);
   s.nor_();
   if (graphics){
      Graph gg;
      s.mark(gg);
      gg.vis();
      cpmwait(tWait,2);
      gg.clr_();
   }
   R eta,res=0;

   R_Vector etav;
   if (graphics) etav=R_Vector(n);
   for (Z i=1;i<=n;++i){
      s=f_(s);
      eta=s.nor_();
      if (graphics) etav[i]=eta;
      if (eta>res) res=eta;
   }
   if (graphics){
      Graph gr;
      gr.show(etav);
      cpmwait(tWait,2);
      gr.clr_();
      s.mark(gr);
      gr.vis();
      cpmwait(tWait,2);
      gr.clr_();
   }
   CPM_MZ
   return res;
}

////////////////////// class SpectralObjectD ////////////////////////////

SpectralObjectD::SpectralObjectD(Bio const& b):b_(b),
   l_(b_.d()*4),spectrum(l_),eigenVectors(l_)
{
   for (Z i=0,j=1;i<l_;++i,++j){
      spectrum[j]=C(j);
      eigenVectors[j]=StateD(1+i/4,1+i%4,b_);
   }
}

SpectralObjectD::SpectralObjectD(OpD const& op, bool test):b_(op.bio()),
l_(b_.d()*4), spectrum(l_), eigenVectors(l_,StateD(b_))
{
   Z mL=1;
   Word loc("SpectralObjectD(OpD,bool)");
   CPM_MA
   V<StateD> cs(l_); // canonical ortho-normal system
   Z i,j;
   for (i=0,j=1;i<l_;++i,++j){
      cs[j]=StateD(1+i/4,1+i%4,b_);
   }
   Observable ob(l_); // made out of the matrix elements of op
      // with respect to the canonical orthonormal system
   for (i=1;i<=l_;++i){
      StateD opsi=op(cs[i]); // so op is applied only l_ times
         // modification 2015-07-05
      for (j=1;j<=l_;++j){
         ob.set(i,j,(opsi|cs[j]));
      }
   }
   //SpectralObject sob=ob.spectralRepresentation(false,false);
   SpectralObject sob=ob.spectralRepresentation(true,true);
   for (i=1;i<=l_;++i){
      spectrum[i]=sob(i);
      for (j=1;j<=l_;++j){
         eigenVectors[i]+=cs[j]*sob[i][j];
      }
   }
   if (test){
      R dInv=R(1.)/l_;
      R e1=0;
      for (i=1;i<=l_;i++){
         e1+=(op(eigenVectors[i])-eigenVectors[i]*spectrum[i]).norm();
      }
      e1*=dInv;
      cpmcerr<<endl<<loc<<" spectral error = "<<e1;
      R e2=0;
      for (i=1;i<=l_;i++){
         for (j=1;j<=l_;j++){
            e2+=((eigenVectors[i]|eigenVectors[j]) - krn(i,j)).abs();
               //krn = Kronecker's delta
         }
      }
      e2*=(dInv*dInv);
      cpmcerr<<endl<<loc<<" ortho error = "<<e2;
      //cpmassert(e1+e2<1.e-6,loc);
      cpmassert(e1+e2<1.e-2,loc);

   }
   CPM_MZ
}

StateD SpectralObjectD::operator *(StateD const& a)const
{
   Z mL=3;
   Word loc("SpectralObjectD::operator*(StateD)");
   CPM_MA
   StateD res(b_);
   cpmassert(a.mat(res),loc);
   for (Z i=1;i<=l_;i++){
      C ci=spectrum[i]*(eigenVectors[i]|a);
      res+=eigenVectors[i]*ci;
   }
   CPM_MZ
   return res;
}

V<C> SpectralObjectD::rep(StateD const& state)const
{
   Z mL=2;
   Word loc("SpectralObjectD::rep(StateD)");
   CPM_MA
   cpmassert(mat(state),loc);
   V<C> res(l_);
   for (Z i=1;i<=l_;i++){
      res[i]=(eigenVectors[i]|state);
   }
   CPM_MZ
   return res;
}

void SpectralObjectD::mark(Graph& gr, Z ipolMeth, R gamma)const
{
   Z mL=2;
   Word loc("SpectralObjectD::mark(Graph,Z,R)");
   CPM_MA
   R gammaMem=Color::getGamma();
   Color::setGamma(gamma);
   R maxSpectrum=spectralRadius();
   R maxVectors=supEigenVectors();
   R fac=maxVectors*cpminv(maxSpectrum);
   VVV<R> vvvr(3,l_,l_+1);
   for (Z i=1;i<=l_;++i){
      C valSpectrum=spectrum[i]*fac;
      Color col(valSpectrum);
      vvvr[1][i][1]=col.Red();
      vvvr[2][i][1]=col.Green();
      vvvr[3][i][1]=col.Blue();
      StateD si=eigenVectors[i];
      for (Z j=1;j<=l_;++j){
         C val=si(j);
         col=Color(val);
         vvvr[1][i][j+1]=col.Red();
         vvvr[2][i][j+1]=col.Green();
         vvvr[3][i][j+1]=col.Blue();
            // notice that the channel index is the first one
      }
   }
   CpmImaging::MultiImage<R> mi(vvvr);
   mi.mark(gr,ipolMeth);
   Color::setGamma(gammaMem);
   CPM_MZ
}

 void SpectralObjectD::markSpectrum(Graph& gr)const
 {
   Z mL=2;
   Word loc("SpectralObject::markSpectrum(Graph)");
   CPM_MA
   R_Vector sp(l_);
   for (Z i=1;i<=l_;i++){
      sp[i]=spectrum[i].real();
   }
   cpmmessage("sp defined");
   gr.mark(sp,RED,1);
   CPM_MZ
 }

bool SpectralObjectD::writeSpectrum(ostream& str)const
{
   CpmRoot::writeTitle("Eigenvalues of a SpectralObjectD:",str);
   cpmp(spectrum);
   return true;
}

bool SpectralObjectD::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(b_);
   cpmp(l_);
   cpmp(spectrum);
   cpmp(eigenVectors);
   return true;
}

bool SpectralObjectD::scanFrom(istream& str)
{
   cpms(b_);
   cpms(l_);
   cpms(spectrum);
   cpms(eigenVectors);
   return true;
}

namespace{

C f_exp(C const& z, R const& t){ return exp(C(0,t)*z) ;}

}

void  SpectralObjectD::expi_(R t)
{
   F<C,C> f=F1<C,R,C>(t)(f_exp);
   this->map_(f);
}

namespace{
   R makeReal(C const& z){return z.real();}
}

SpectralObjectD SpectralObjectD::order()const
{
   F<C,R> transf(makeReal);
      // going for the real part
   V<R> spr=spectrum(transf);
      // making an array that holds the real parts
   V<R> spr2(spr);
   Vo<R> sp(spr2);
      // making an auxiliary object that
      // allows to create an ordering permutation
   V<Z> per=sp.permutationForIncreasingOrder();
      // making the ordering permutation
// rest is simple
   V<C> spectrumTemp(spectrum);
   V<StateD> eigenVectorsTemp(eigenVectors);
   spectrumTemp=spectrumTemp.permute(per);
   eigenVectorsTemp=eigenVectorsTemp.permute(per);
   V<C> spectrumRes=spectrumTemp;
   V<StateD> eigenVectorsRes=eigenVectorsTemp;
   return SpectralObjectD(spectrumRes,eigenVectorsRes);
}

//////////////////////////// class TrjD ///////////////////////////////////

TrjD::TrjD(OpD const& op, StateD const& psi, R const& dt):
TrjBas(psi,dt),op_(op),tau_(dt_*0.25), tau2_(tau_*2.)
{
   Z mL=StateD::getVer();
   Word loc("TrjD(OpD,StateD,R)");
   CPM_MA
   phi_=StateD(psi_.dot(op_));
   CPM_MZ
}

void TrjD::step_() // DALF integrator
{
   t_+=tau_;
   psi_+=phi_*tau_;
   phi_+=(psi_.dot(op_)-phi_)*R(2);
   t_+=tau2_;
   psi_+=phi_*tau2_;
   phi_+=(psi_.dot(op_)-phi_)*R(2);
   psi_+=phi_*tau_;
   t_+=tau_;
}

/*********************************
   void TrjD::step_(){ // ADALF integrator
      t_+=tau_;
      psi_+=phi_*tau_;
      phi_+=(psi_.dot(op_) - phi_)*R(2);
      StateD phiMem=phi_;
      t_+=tau2_;
      psi_+=phi_*tau2_;
      phi_+=(psi_.dot(op_) - phi_)*R(2);
      psi_+=phi_*tau_;
      phi_=(phi_+phiMem)*0.5;
      t_+=tau_;
    }
*************************/

void TrjD::evl_(R tf)
{
   R span=tf-t_;
   R dtMem=dt_;
   Z steps=cpmtoz(cpmpos(cpmrnd(span/dt_)));
   if (steps==0) steps=1;
   set_dt_(span/steps);
   for (Z i=1;i<=steps;++i) step_();
   set_dt_(dtMem);
}

//////////////////////////// class SpTrjD /////////////////////////////////

void SpTrjD::evl_(R tf)
{
   R span=tf-t_;
   SpectralObjectD so(so_);
   so.expi_(-span);
   psi_=so*psi_;
   t_=tf;
}
