//? cpmstatedirac1.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   cpmstate.cpp
   History: Started 2014-01-?

*/

#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmframes.h>
#include <cpmvr.h>
#include <cpms.h>
#include <cpmrvector.h>
#include <cpmstatedirac1.h>
#include <cpmmultiimage.h>
#include <cpmfile.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::Bio;
using CpmQuantumMechanics::StateD1;
using CpmQuantumMechanics::OpD1;
using CpmQuantumMechanics::SpectralObjectD1;
using CpmQuantumMechanics::TrjD1;
using CpmQuantumMechanics::SpTrjD1;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::SX1;
using CpmLinAlg::R2;
using CpmGraphics::Frame;
using CpmGraphics::Frames;
using CpmGraphics::Graph;
using CpmGraphics::Z2;
using CpmGeo::Iv;
using CpmImaging::Color;
using CpmImaging::MultiImage;
using CpmFunctions::F;
using CpmFunctions::F1;
using CpmFunctions::R_Func;
using CpmSystem::OFile;
using namespace std;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;

////////////////////////////// class StateD1 //////////////////////////////

Z StateD1::bc_=1; // 1: periodic
Z StateD1::verbose_=1; //
bool StateD1::useDALF_=0; // usage of a DALF trajectory for
   // constructing boost operators
bool StateD1::enfNor_=1; // enforcing normalization to correct unitarity
   // deficiencies of boost implementation

//CC2 StateD1::unit_=CpmLinAlg::CC2::sig0;
//CC2 StateD1::beta_=CpmLinAlg::CC2::sig3;
//CC2 StateD1::alpha1_=CpmLinAlg::CC2::sig1;
   // Never initialize static data with static data as done in the
   // commented three lines! Whether this works or not depends on the
   // order of initialization and this is not fixed by the code.

CC2 StateD1::unit_=CpmLinAlg::CC2(C(1.),C(0.),C(0.),C(1.));
CC2 StateD1::beta_=CpmLinAlg::CC2(C(1.),C(0.),C(0.),C(-1.));
CC2 StateD1::alpha1_=CpmLinAlg::CC2(C(0.),C(1.),C(1.),C(0.));

CC2 StateD1::boost1(R const& chi)const
{
   R chih=chi*0.5;
   return unit_*cpmcosh(chih) - alpha1_*cpmsinh(chih);
}

R2 StateD1::lorentz1(R2 const& y, R const& chi)
{
   R sc=cpmsinh(chi), cc=cpmcosh(chi);
   R x0=y.x1, x1=y.x2;
   return  R2(x0*cc-x1*sc,x1*cc-x0*sc);
}

StateD1::StateD1(V<C2> const& v, Iv const& s ):
s_(s), sup_(s_), xc_(s_.cen()),rep_(v.dim())
{
   Z d=rep_.dim();
   V<R> x=s_.centers(d);
   for (Z i=1; i<=d; ++i) rep_[i] = SX1(v[i],x[i]);
}

StateD1::StateD1(Bio const& bio ): s_(bio.iv()), sup_(s_),
   xc_(s_.cen()),rep_(bio.d())
{
   V<R> x=bio.points();
   C2 c4; // zero vector
   for (Z i=rep_.b(); i<=rep_.e(); ++i) rep_[i] = SX1(c4,x[i]);
}

StateD1::StateD1(C const& z, Bio const& b):
   s_(b.iv()), sup_(s_), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C2 c2(z);
   for (Z i=rep_.b(); i<=rep_.e(); ++i) rep_[i] = SX1(c2,x[i]);
}

StateD1::StateD1(Iv const& iv, Bio const& b, Z j):
   s_(b.iv()), sup_(iv), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C2 c0,c1;
   c1[1]=C(1.);
   for (Z i=rep_.b(); i<=rep_.e(); ++i){
      R xi=x[i];
      bool b=iv.contains(xi);
      rep_[i] = b ? SX1(c1,xi) : SX1(c0,xi);
   }
}

StateD1::StateD1(F<R,C2> const& f, Bio const& b): s_(b.iv()), sup_(s_),
   xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   for (Z i=x.b();i<=x.e();i++){
      R xi=x[i];
      rep_[i]=SX1(f(xi),xi);
   }
}

StateD1::StateD1(Z i, Z a, Bio const& b, bool frqSpc):
fft_(B(frqSpc)),s_(b.iv()), sup_(s_), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   C2 c0; // zero vector
   C2 ca;
   ca[a]=C(1.);
   Z ii=rep_.makeIndVal(i); // understood Cyclic, so i need not be in range
   for (Z j=x.b(); j<=x.e(); ++j){
      rep_[j] = (j==ii ? SX1(ca,x[j]) : SX1(c0,x[j]));
   }
}

StateD1::StateD1(Bio const& b, C2 const& c, R const& x0, R const& sigma0,
   R const& k, bool normalize):
   s_(b.iv()), sup_(s_), xc_(s_.cen()), rep_(b.d())
{
   V<R> x=b.points();
   for (Z i=x.b();i<=x.e();i++){
      R xi=x[i];
      R yi=(xi-x0)/sigma0;
      yi*=yi;
      C fi=C::expi(xi*k)*exp(-0.5*yi);
      rep_[i]=SX1(c*fi,xi);
   }
   if (normalize) nor_(); // constructs a unit vector
}

StateD1::StateD1(Bio const& b, R const& xPeak, R const& sigma , R const& p,
   R const& a1, R const& a2, R const& phase1, R const& phase2,
   bool compact):s_(b.iv()), sup_(s_), xc_(s_.inf()), rep_(b.d())
{
   V<R> x=b.points();
   V<R> y=x(R_Func::bell(xPeak,sigma,compact));
   R phi1=phase1*cpmdeg;
   R phi2=phase2*cpmdeg;
   C2 c2(C::expi(phi1)*a1,C::expi(phi2)*a2);
   for (Z i=x.b();i<=x.e();i++){
      R xi=x[i];
      R yi=y[i];
      C fi=C::expi(xi*p)*yi;
      rep_[i]=SX1(c2*fi,xi);
   }
   if (compact){
      R hw=sigma*4; // half width of support
      sup_=Iv(xPeak-hw,xPeak+hw);
   }
}

void StateD1::setVal_(C_Vector const& v)
{
   Z i,k,d=rep_.dim();
   cpmassert(v.dim()==2*d,"StateD1::setVal_(C_Vector)");
   for (i=1,k=d+1;i<=d;++i,++k){
      rep_[i].c1()=C2(v[i],v[k]);
   }
}

namespace{

SX1 negFunc(SX1 const& a)
{
   SX1 res=a;
   res.c1()=-a.c1();
   return res;
}

SX1 conFunc(SX1 const& a)
{
   SX1 res=a;
   res.c1()= a.c1().con();
//   res.c1()= StateD::gam0_*(a.c1().con()); // experiment
      // gam0 added 2014-03-25
   return res;
}

SX1 multFunc(SX1 const& a, C const& z )
{
   SX1 res=a;
   res.c1()*=z;
   return res;
}

SX1 multRFunc(SX1 const& a, R const& r )
{
   SX1 res=a;
   res.c1()*=r;
   return res;
}

SX1 addFunc(SX1 const& a, SX1 const& b )
{
   SX1 res=a;
   res.c1()+=b.c1();
   return res;
}

SX1 subFunc(SX1 const& a, SX1 const& b )
{
   SX1 res=a;
   res.c1()-=b.c1();
   return res;
}

C cumFunc(const SX1& a, const SX1& b)
{
   return (a.c1()|b.c1());
}

} // namespace

StateD1 StateD1::operator-(void)const
{
   return StateD1(rep_.fa(negFunc),s_);
}

StateD1& StateD1::operator*=(C z)
{
   rep_.fb2_<C>(multFunc,z);
   return *this;
}

StateD1& StateD1::operator+=(StateD1 const& psi)
{
   rep_.ffl_(addFunc,psi.rep_);
   return *this;
}

StateD1& StateD1::operator-=(StateD1 const& psi)
{
   rep_.ffl_(subFunc,psi.rep_);
   return *this;
}

StateD1 StateD1::operator*(C z)const
{
   return StateD1(rep_.fb2<C>(multFunc,z),s_);
}

StateD1 StateD1::operator +(StateD1 const& psi)const
{
   return StateD1(rep_.fe(addFunc,psi.rep_),s_);
}

StateD1 StateD1::operator -(StateD1 const& psi)const
{
   return StateD1(rep_.fe(subFunc,psi.rep_),s_);
}

StateD1& StateD1::operator*=(R r)
{
   rep_.fb2_<R>(multRFunc,r);
   return *this;
}

StateD1 StateD1::operator*(R r)const
{
   return StateD1(rep_.fb2<R>(multRFunc,r),s_);
}

namespace{
   SX1 mFunc(SX1 const& sx, CC2 const& m){ return SX1(m*sx.c1(),sx.c2());}
   SX1 wFunc(SX1 const& sx, R_Func const& f)
   {
      C2 s=sx.c1(); R x=sx.c2(); return SX1(s*f(x),x);
   }

   SX1 vFunc(SX1 const& sx, R const& px)
   {
      SX1 res=sx;
      res.c1()*=C(0.,-px);
      return res;
   }
   C2 conFunc(C2 const& x){ return x.con();}
}

StateD1 StateD1::applyCC2(CC2 const& mat)const
{
   return StateD1(rep_(F1<SX1,CC2,SX1>(mat)(mFunc)),s_);
}

StateD1 StateD1::rev()const
{
   StateD1 res=*this;
   res.setVal_(val().rev());
   return res;
}

StateD1 StateD1::cc()const
{
   StateD1 res=*this;
   V<C2> v=val();
   V<C2> vc=v(conFunc);
   res.setVal_(vc);
   return res;
}

StateD1 StateD1::apply(R_Func const& f)const
{
   return StateD1(rep_(F1<SX1,R_Func,SX1>(f)(wFunc)),s_);
}

StateD1 StateD1::win(Iv const& iv, R const& softParameter)const
{
   return apply(R_Func::charFunc(iv,softParameter));
}

StateD1 StateD1::frqWin(Iv const& iv, R const& softParameter)const
   //: frequency window
   // Restricts the wave function, which is assumed to be given in
   // position space, by a windowing operation which acts on the FFT
   // image. The argument iv which belongs to the full frequency range
   // is Iv(-1,1) in order to be independent of the discretization under
   // consideration.
{
   StateD1 psi=toFrqSpc();
   Iv ivfr=psi.bio().iv();
   R fr=ivfr[2];
   R x1=iv[1]*fr;
   R x2=iv[2]*fr;
   return psi.win(Iv(x1,x2),softParameter).fromFrqSpc();
}

StateD1 StateD1::applyF(R_Func const& f)const
   //:: apply to discrete Fourier transform.
   // The argument of f ranges from -1 to 1, where 1 corresponds to the
   // spatial Nyquist frequency 1/(2*gap()) and -1 to -1/(2*gap().
   // It is to be noted however that this last most negative value no
   // longer belongs to the indexing range of the discrete Fourier
   // transform.
{
   return *this; // preliminary nonsense
}

StateD1 StateD1::applyPot(R_Vector const& pot)const
{
   Z rd=rep_.dim();
   cpmassert(pot.dim()==rd,"StateD1::applyPot");
   return StateD1(rep_.fet<R>(vFunc,pot),s_);
}

namespace
{
   SX1 func(SX1 const& c_1, SX1 const& c0, SX1 const& c1, R const& fac, Z\
    i)
   {
      i;
      return SX1((c1.c1()-c_1.c1())*fac,c0.c2());
   }
}

StateD1 StateD1::dx()const
{
   Outside out=(Outside)bc_;
   return StateD1(rep_.fh<R>(func,R(0.5)/gap(),out),s_);
}

StateD1 StateD1::dot(R const& m, R_Vector const& pot)const
{
   if (d()==pot.dim()){
      return applyCC2(beta_)*C(0.,-m)
      -dx().applyCC2(alpha1_)
      +applyPot(pot);
   }
   else{
      return applyCC2(beta_)*C(0.,-m)
      -dx().applyCC2(alpha1_);
   }
}

namespace{
   StateD1 hFunc(StateD1 const& psi, R const& m, R_Vector const& pot)
   {
      return psi.ham(m,pot);
   }

   StateD1 xFunc(StateD1 const& psi){ return psi.x1();}

   StateD1 pFunc(StateD1 const& psi){ return psi.dx()*C::Ibar;}

   StateD1 shiftFunc(StateD1 const& psi, R const& a)
   {
      R g=psi.gap();
      R ar=a/g;
      Z ariL=cpmtoz(ar);
      Z ariR=ariL+1;
      R arf=ar-ariL;
      R wL=R(1)-arf;
      R wR=arf;
      Z d=psi.d();
      Iv s=psi.s();
      V<C2> v2=psi.val();
      V<C2> w2(d);
      Outside out=(Outside) StateD1::getBonCon();
      for (Z i=1;i<=d;++i){
         w2[i]=v2(i-ariL,out)*wL+v2(i-ariR,out)*wR;
      }
      return StateD1(w2,s);
   }

   StateD1 boostLFunc(StateD1 const& psi, OpD1 const& h, R const& chi,
                     R const& dt )
   {
      Z mL=StateD1::getVer();
      Word loc("boostLFunc");
      CPM_MA
      Z d=psi.d();
      R psiN=psi.abs();
      CC2 cc2=psi.boost1(chi);
      R ch=cpmcosh(chi);
      R sh=cpmsinh(chi);
      R th=sh/ch;
      Bio b=psi.bio()*ch;
      StateD1 psiT(psi.val(),b.iv());
      TrjD1 trj(h,psiT,dt); // gives exp(-ith) just as in Domenico's
         // formula
      V<C2> val(d);
      for (Z i=1;i<=d;++i){
         R xi=psiT.xc(i);
         R ti=th*xi;
         trj.evl_(ti);
         StateD1 si=trj.psi();
         val[i]=cc2*si[i];
      }
      StateD1 res(val,psi.s());
      R resN=res.abs();
      res*=C(psiN/resN);
      CPM_MZ
      return res;
   }

   StateD1 boostFunc1(StateD1 const& psi, OpD1 const& db, R const& chi,
                    R const& chiStep)
   {
      Z mL=StateD1::getVer();
      Word loc("boostFunc1");
      CPM_MA
      R psiN,resN;
      bool b=StateD1::getEnfNor();
      if (b) psiN=psi.abs();
      TrjD1 trj(db,psi,chiStep);
      trj.evl_(chi);
      StateD1 res=trj.psi();
      if (b){
         resN=res.abs();
         res*=C(psiN/resN);
      }
      CPM_MZ
      return res;
   }

   StateD1 boostFunc2(StateD1 const& psi, OpD1 const& db, Z const& n)
   {
      Z mL=StateD1::getVer();
      Word loc("boostFunc2");
      CPM_MA
      R psiN{0.},resN{0.};
      bool b=StateD1::getEnfNor();
      if (b) psiN=psi.abs();
      StateD1 res=psi;
      for (Z i=1;i<=n;++i) res=db*res;
      if (b){
         resN=res.abs();
         res*=C(psiN/resN);
      }
      CPM_MZ
      return res;
   }
}

OpD1 StateD1::h(Bio const& b, R const& m, R_Vector const& pot)
{
   return OpD1(b,F2<StateD1,R,R_Vector,StateD1>(m,pot)(hFunc));
}

OpD1 StateD1::x(Bio const& b)
{
   return OpD1(b,F<StateD1,StateD1>(xFunc));
}

OpD1 StateD1::p(Bio const& b)
{
   return OpD1(b,F<StateD1,StateD1>(pFunc));
}

OpD1 StateD1::shift(Bio const& b, R const& a)
{
   return OpD1(b,F1<StateD1,R,StateD1>(a)(shiftFunc));
}

OpD1 StateD1::boostL(Bio const& b, R const& m, R const& chi,
                   R const& dtRel)
{
   cpmassert(dtRel>0 && dtRel<1,"StateD::boostL");
   R ch=cpmcosh(chi);
   Bio bch=b*ch;
   OpD1 h=StateD1::h(bch,m); // Lorentz transformed biotope
   R dt=h.dtLim()*dtRel; // important to place time-consuming norm
      // evaluation in the constructor (or static generating function) and
      // not in the workhorse function boostLFunc.
   return OpD1(b,F3<StateD1,OpD1,R,R,StateD1>(h,chi,dt)(boostLFunc));
}

OpD1 StateD1::dBoost(Bio const& b, R const& m, R const& deltaChi,
                   R const& dtRel)
{
   R fac=R(0.5)/deltaChi;
   return
   (boostL(b,m,deltaChi,dtRel)-boostL(b,m,-deltaChi,dtRel))*C(0.,fac);
      // second order accurate expression for the derivative at unity
}

SpectralObjectD1 StateD1::dBoostExact(Bio const& b, R const& m)
{
   OpD1 ham=StateD1::h(b,m);
   OpD1 x=StateD1::x(b);
   return SpectralObjectD1((ham*x+x*ham)*R(0.5));
}

SpectralObjectD1 StateD1::boostExact(Bio const& b, R const& m,
   R const& chi)
{
   SpectralObjectD1 res=StateD1::dBoostExact(b,m);
   res.expi_(chi);
   return res;
}

OpD1 StateD1::boost(Bio const& b, R const& m, R const& chi,
                  R const& deltaChi, R const& dtRel)
{
   if (getUseDALF()){ // slow and accurate
      OpD1 db=dBoost(b,m,deltaChi,dtRel);
      R dchi=dtRel*db.dtLim();
      return OpD1(b,F3<StateD1,OpD1,R,R,StateD1>(db,chi,dchi)(boostFunc1));
   }
   else{ // fast and not too accurate
      R cs=cpmabs(deltaChi);
      if (cpmabs(chi)<=cs) return boostL(b,m,chi,dtRel);
      Z n=cpmpos(cpmrnd(chi/cs));
      R dchi=chi/n;
      OpD1 bL=boostL(b,m,dchi,dtRel);
      return OpD1(b,F2<StateD1,OpD1,Z,StateD1>(bL,n)(boostFunc2));
   }
}

StateD1 StateD1::con(void)const
{
   return StateD1(rep_.fa(conFunc),s_);
}

C StateD1::operator |(StateD1 const& psi)const
{
   return rep_.fAcc3<C>(cumFunc,QMBase::accFunc,psi.rep_);
}

R StateD1::nor_(R r)
{
   Z d=rep_.dim();
   if (d==0) return 0.;

   R r0=abs();
   if (r0==0){
      StateD1 aux=*this;
      (aux.rep_[1].c1())[1]=C(r);
      *this=aux;
   }
   else{
      *this*=(r/r0);
   }
   return r0;
}

C StateD1::enr(R const& m, R_Vector const& pot)const
{
   StateD1 s=*this;
   s.nor_();
   return (s|s.ham(m,pot));
}

StateD1 StateD1::nyq(C2 const& c1, C2 const& c2, Bio const& b)
{
   Z n=b.d();
   cpmassert(n>1,"StateD1::nyq");
   V<C2> v(n,c1);
   for (Z i=2;i<=n; ++++i) v[i]=c2;
   return StateD1(v,b.iv());
}

namespace{
   R convertC2toR(C2 const& c2){ return c2.abs2();}
}

R StateD1::valSup2()const
{
   V<C2> vc=val();
   F<C2,R> fc(convertC2toR);
   V<R> vr=vc(fc);
   R_Vector amp(vr);
   return amp.sup();
}

R StateD1::valInf2()const
{
   V<C2> vc=val();
   F<C2,R> fc(convertC2toR);
   V<R> vr=vc(fc);
   R_Vector amp(vr);
   return amp.inf();
}

namespace{
   R const redFactor=0.25;
   // used not only in valMat
}

R_Matrix StateD1::valMat(bool logScale)const
{
   Z mL=3;
   Word loc("StateD1::valMat(bool)");
   CPM_MA
   Z d_=d();
   V<C2> vc=val();
   R_Func f=R_Func::log10;
   if (!sup_.contains(s_)){ // only then the support should be indicated
      R_Matrix y(6,d_);
      R_Vector x=pos();
      R vs=valSup2();
      R v0=valInf2();
      if (logScale){
         vs=f(vs); v0=f(v0);
      }
      for (Z i=1;i<=d_;i++){
         C2 c2=vc.cui(i);
         C s1=c2.x1, s2=c2.x2;
         R s11=s1[1], s12=s1[2], s21=s2[1], s22=s2[2];
         R q11=s11*s11,q12=s12*s12,q21=s21*s21,q22=s22*s22;
         R a2=q11+q12+q21+q22;
         if (logScale){ a2=f(a2); q11=f(q11); q12=f(q12); q21=f(q21);
            q22=f(q22);
         }
         y[1][i]=a2;
         y[2][i]=q11;
         y[3][i]=q12;
         y[4][i]=q21;
         y[5][i]=q22;
         y[6][i]=sup_.contains(x[i]) ? vs*redFactor : v0;
      }
      CPM_MZ
      return y;
   }
   else{
      R_Matrix y(5,d_);
      for (Z i=1;i<=d_;i++){
         C2 c2=vc.cui(i);
         C s1=c2.x1, s2=c2.x2;
         R s11=s1[1], s12=s1[2], s21=s2[1], s22=s2[2];
         R q11=s11*s11,q12=s12*s12,q21=s21*s21,q22=s22*s22;
         R a2=q11+q12+q21+q22;
         if (logScale){ a2=f(a2); q11=f(q11); q12=f(q12); q21=f(q21);
            q22=f(q22);
         }
         y[1][i]=a2;
         y[2][i]=q11;
         y[3][i]=q12;
         y[4][i]=q21;
         y[5][i]=q22;
      }
      CPM_MZ
      return y;
   }
}

/************************
R_Matrix StateD1::valMat(bool logScale)const
{
   R_Func f=R_Func::log10;
   Z d_=d();
   R_Matrix y(5,d_);
   for (Z i=1;i<=d_;i++){
      SX1 sxi=rep_[i];
      C2 c2=sxi.c1();
      C s1=c2.x1, s2=c2.x2;
      R s11=s1[1], s12=s1[2], s21=s2[1], s22=s2[2];
      R q11=s11*s11,q12=s12*s12,q21=s21*s21,q22=s22*s22;
      R a2=q11+q12+q21+q22;
      if (logScale){ a2=f(a2); q11=f(q11); q12=f(q12); q21=f(q21);
         q22=f(q22);
      }
      y[1][i]=a2;
      y[2][i]=q11;
      y[3][i]=q12;
      y[4][i]=q21;
      y[5][i]=q22;
   }
   return y;
}
***************************/

/**************************
void StateD::showAsCurves(Graph& g)const
{
   using namespace CpmImaging;
   using namespace CpmArrays;

   V<Color> cl(9);
   cl[1]=BLACK; // color for absolute value
   cl[2]=RED;  // color for real part of v1
   cl[3]=LIGHTRED;   // color for imaginary part v1
   cl[4]=GREEN;
   cl[5]=LIGHTGREEN;
   cl[6]=BLUE;
   cl[7]=LIGHTBLUE;
   cl[8]=MAGENTA; // color for real part of v4
   cl[9]=LIGHTMAGENTA; // color for imaginary part v4

   Z d_=d();
   R_Matrix res(10,d_);
   for (Z i=1;i<=d_;i++){
      SX sxi=rep_[i];
      C4 c4=sxi.c1();
      R  xi=sxi.c2();
      res[1][i]=xi; // position
      res[2][i]=c4.abs(false); // for values close to 0, the more
         // sophisticated version is worse
      C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
      res[3][i]=s1[1];
      res[4][i]=s1[2];
      res[5][i]=s2[1];
      res[6][i]=s2[2];
      res[7][i]=s3[1];
      res[8][i]=s3[2];
      res[9][i]=s4[1];
      res[10][i]=s4[2];
   }
   g.clear();
   g.addText("StateD. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.show(res,cl);
}
************************/

void StateD1::mark(Graph& g, bool logScale)const
{
   Z mL=3;
   Word loc("StateD1::mark(Graph,bool)");
   using namespace CpmImaging;
   using namespace CpmArrays;
   R_Vector x=pos();
   R_Matrix y=valMat(logScale);
   V<Color> cl{BLACK,GREEN,RED,LIGHTCYAN,LIGHTMAGENTA,LIGHTGRAY};
   if (sup_.contains(s_)) cl=cl.cut(1); // takes away the last component
   g.clear();
   g.markGL(x,y,cl);
   CPM_MZ
}

void StateD1::markArr(V<StateD1> const& vs, Graph& g)
{
   Z mL=3;
   Word loc("StateD1::markArr(V<StateD1>,Graph)");
   CPM_MA

   using namespace CpmImaging;
   using namespace CpmArrays;

   V<Color> cl(10);
   cl[1]=LIGHTGRAY;
   R fac=0.6;
   Z nL=2;
   V<R> list=Iv(0.,1.).div(nL);
   for (Z i=1;i<=nL;++i){
      Color c1=Color::hue(list[i]);
      Color c2=(c1*fac).addWhite(1.);
      Z j=2*i;
      cl[j]=c1;
      cl[j+1]=c2;
   }
// darker colors
   cl[6]=BLACK; // color for absolute value
   fac=0.75;
   cl[7]=cl[2]*fac;  // color for real part of v1
   cl[8]=cl[3]*fac;   // color for imaginary part v1
   cl[9]=cl[4]*fac;
   cl[10]=cl[5]*fac;
   Z i;
   StateD1 s1=vs.fir(); // making sure that the states match
   for (i=vs.b(); i<=vs.e(); ++i){
      cpmassert(s1.mat(vs[i]),loc);
   }
   R_Vector x=s1.pos();
   R_Matrix y=s1.valMat();
   for (i=vs.b()+1; i<=vs.e(); ++i){
      y&=vs[i].valMat(); // appending the matrix part
   }
   g.clear();
   g.addText("StateD1. Real part: full color, Imaginary part: light color,\
 Abs: black");
   g.markGL(x,y,cl);
   CPM_MZ
}

void StateD1::mark2(Graph& g, StateD1 const& sc, R const& scale)const
      // Shows *this together with a comparison state
      // allows to boost the difference between the two by a factor.
{
   StateD1 s1(*this);
   StateD1 s2(sc);
   StateD1 sm=(s1+s2)*0.5;
   R sh=scale*0.5;

   StateD1 sDiff=(s1-s2)*sh;
   StateD1 v1=sm+sDiff;
   StateD1 v2=sm-sDiff;
   StateD1::markArr(V<StateD1>{v1,v2},g);
}

/*********************
void StateD::gp(OFileStream& out)const
      // returns a data file from which gnuplot can construct the
      // showAsCurves representation.
{
   Z mL=StateD::getVer();
   Word loc("gp(Word)");
   CPM_MA
   for (Z i=1;i<=d();i++){
      SX sxi=rep_[i];
      C4 c4=sxi.c1();
      R  xi=sxi.c2();
      R s0=c4.abs(false);
      C s1=c4.x1, s2=c4.x2, s3=c4.x3, s4=c4.x4;
      out()<<endl<<" "<<xi<<" "<<s0<<" "<<s1[1]<<" "<<s1[2]<<" "
      <<s2[1]<<" "<<s2[2]<<" "<<s3[1]<<" "<<s3[2]<<" "
      <<s4[1]<<" "<<s4[2];
   }
   CPM_MZ
}
***********************************/
/*********************************
void StateD1::gp(ostream& out, Iv const& iv)const
      // returns a data file from which gnuplot can construct the
      // showAsCurves representation.
{
   Z mL=3;
   Word loc("StateD1::gp(Word)");
   CPM_MA
   if (iv.isVoid()){ // normal case, no additional indicator function
      for (Z i=1;i<=d();i++){
         SX1 sxi=rep_[i];
         C2 c2=sxi.c1();
         R  xi=sxi.c2();
         R q0=c2.abs2();
         C s1=c2.x1, s2=c2.x2;
         R r11=s1[1], r12=s1[2], r21=s2[1], r22=s2[2];
         R q11=r11*r11, q12=r12*r12, q21=r21*r21, q22=r22*r22;
         out<<endl<<" "<<xi<<" "<<q0<<" "<<q11<<" "<<q12<<" "
         <<q21<<" "<<q22;
      }
   }
   else{
      for (Z i=1;i<=d();i++){
         SX1 sxi=rep_[i];
         C2 c2=sxi.c1();
         R  xi=sxi.c2();
         R s0=c2.abs(false);
         C s1=c2.x1, s2=c2.x2;
         R r= iv.contains(xi) ? R(0.1) : R(0);
         out<<endl<<" "<<xi<<" "<<s0<<" "<<s1[1]<<" "<<s1[2]<<" "
         <<s2[1]<<" "<<s2[2]<<" "<<r;
      }
   }
   CPM_MZ
}
****************************/

void StateD1::gp(ostream& out, bool logScale)const
{
   Z mL=1;
   Word loc("SttD1::gp(Word,bool)");
   CPM_MA
   Z d_=d();
   R_Func f=R_Func::log10;
   V<C2> vc=val();
   R_Vector x=pos();

   if (sup_.contains(s_)){ // no additional indicator function
      for (Z i=1;i<=d_;i++){
         C2 c2=vc[i];
         R  xi=x[i];
         R q0=c2.abs2();
         C s1=c2.x1, s2=c2.x2;
         R r11=s1[1], r12=s1[2], r21=s2[1], r22=s2[2];
         R q11=r11*r11, q12=r12*r12, q21=r21*r21, q22=r22*r22;
         if (logScale){
            q0=f(q0); q11=f(q11); q12=f(q12); q21=f(q21); q22=f(q22);
         }
         out<<endl<<" "<<xi<<" "<<q0<<" "<<q11<<" "<<q12<<" "
         <<q21<<" "<<q22;
      }
   }
   else{
      R q0Sup=valSup2();
      R q0Inf=valInf2();
      if (logScale){
         q0Sup=f(q0Sup);
         q0Inf=f(q0Inf);
      }
      for (Z i=1;i<=d_;i++){
         C2 c2=vc[i];
         R  xi=x[i];
         R q0=c2.abs2();
         C s1=c2.x1, s2=c2.x2;
         R r= sup_.contains(xi) ? q0Sup*redFactor : q0Inf;
         R r11=s1[1], r12=s1[2], r21=s2[1], r22=s2[2];
         R q11=r11*r11, q12=r12*r12, q21=r21*r21, q22=r22*r22;
         if (logScale){
            q0=f(q0); q11=f(q11); q12=f(q12); q21=f(q21); q22=f(q22);
         }
         out<<endl<<" "<<xi<<" "<<q0<<" "<<q11<<" "<<q12<<" "
         <<q21<<" "<<q22<<" "<<r;
      }
   }
   CPM_MZ
}

namespace{
   C2 valF( SX1 const& sx ){ return sx.c1();}
   C val1F( SX1 const& sx ){ return sx.c1()[1];}
   C val2F( SX1 const& sx ){ return sx.c1()[2];}
   R posF( SX1 const& sx ){ return sx.c2();}
}

C2 StateD1::at(R const& x)const
{
   Iv iv=s();
   Z n=d();
   X2<R2,Z2> x2=iv.intPolInfo2(x,n);
  // X2<R2,Z2> x2=iv.intPolInfo(x,n);
   R2 ww=x2.c1();
   Z2 ii=x2.c2();
   R w1=ww.x1, w2=ww.x2;
   Z i1=ii.x1, i2=ii.x2;
   return rep_[i1].c1()*w1+rep_[i2].c1()*w2;
}

V<C2> StateD1::val()const
{
   F<SX1,C2> fval(valF);
   return rep_(fval);
}

V<C> StateD1::val1()const
{
   F<SX1,C> fval1(val1F);
   return rep_(fval1);
}

V<C> StateD1::val2()const
{
   F<SX1,C> fval2(val2F);
   return rep_(fval2);
}

R_Vector StateD1::pos()const
{
   F<SX1,R> fpos(posF);
   return R_Vector(rep_(fpos));
}

StateD1 StateD1::ran(Z j)const
// For a zero-vector psi,the vector psi.ran() should not be a zero-vector
{
   Z dd=d();
   C z(1.,0.5);
   Vr<C2> vr(dd,C2(z,z));
   StateD1 res(vr.ran(j),s_);
   res.nor_();
   return res;
}

R StateD1::dis(StateD1 const& psi)const
{
   Z mL=3;
   Word loc("StateD::dis(StateD)");
   CPM_MA
   V<C2> v2=val(), v2psi=psi.val();
   V<R> x=pos(), xpsi=psi.pos();

   R_Vector d(3);
   d[1]=v2.dis(v2psi);
   d[2]=x.dis(xpsi);
   d[3]=s_.dis(psi.s_);
   R res=d.supAbs();
   CPM_MZ
   return res;
}

bool StateD1::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(fft_);
   cpmp(xc_);
   cpmp(rep_);
   cpmp(s_);
   return true;
}

bool StateD1::scanFrom(istream& str)
{
   cpms(fft_);
   cpms(xc_);
   cpms(rep_);
   cpms(s_);
   return true;
}

bool StateD1::isVal()const
{
   for (Z i=rep_.b();i<=rep_.e();++i){
      if (!rep_.cui(i).c1().isVal()) return false;
   }
   return true;
}

StateD1 StateD1::toFrqSpc()const
{
   Z n2=d();
   cpmassert(CpmRootX::isPowOfTwo(n2),"StateD1::toFrqSpc()");
   Z n=n2/2,i,j;
   R fN=R(0.5)/gap(); // Nyquist frequency
   Iv sF(-fN,fN);
   V<R> f=sF.centers(n2);
   V<C> temp1=val1();
   temp1=CpmFourier::fft(temp1,1);
   V<C> temp2=val2();
   temp2=CpmFourier::fft(temp2,1);

   V<SX1> vF(n2);
   for (i=1,j=n+1;i<=n;++i,++j){
      vF[i]=SX1(C2(temp1[j],temp2[j]),f[i]);
   }
   for (i=1,j=n+1;i<=n;++i,++j){
      vF[j]=SX1(C2(temp1[i],temp2[i]),f[j]);
   }
   StateD1 res(vF,sF);
   res.fft_=B(true);
   res.xc_=s_.cen(); // The position center should not be lost
   return res;
}

StateD1 StateD1::fromFrqSpc()const
{
   Z n2=d();
   cpmassert(CpmRootX::isPowOfTwo(n2),"StateD1::fromFrqSpc()");
   Z n=n2/2,i,j;
   R fN=s().e(); // end of frequency biotope
   R dx=R(0.5)/fN;
   R sh=dx*n;
   Iv s(-sh+xc_,sh+xc_); // length = dx*n2 = 2*dx*n
   V<R> x=s.centers(n2);
   V<C> temp1(n2);
   V<C> temp2(n2);

   for (i=1,j=n+1;i<=n;++i,++j){
      temp1[i]=val1()[j];
      temp2[i]=val2()[j];
   }

   for (i=1,j=n+1;i<=n;++i,++j){
      temp1[j]=val1()[i];
      temp2[j]=val2()[i];
   }

   temp1=CpmFourier::fft(temp1,-1);
   temp2=CpmFourier::fft(temp2,-1);

   V<SX1> v(n2);
   for (i=1;i<=n2;++i){
      v[i]=SX1(C2(temp1[i],temp2[i]),x[i]);
   }
   StateD1 res(v,s);
   res.fft_=B(false);
   return res;
}

StateD1 StateD1::momFrq()const
// We assume that *this is given in frequency space
{
   Word loc("StateD1::momFrq()");
   cpmassert(fft_==true,loc);
   Z d_=d();
   R pmax=d_*gap(); // notice that gap() does not coincide with the
      // lattice spacing in position space; we are in frequency space!
   R fac=cpm2pi/d_;
   StateD1 res=*this;
   for (Z i=1;i<=d_;++i){
      res.rep_[i].c1()*=(cpmsin(fac*(i-1))*pmax);
   }
   return res;
}

StateD1 StateD1::hamFrq( R mass)const
{
   Word loc("StateD1::hamFrq(R)");
   cpmassert(fft_==true,loc);
   return momFrq().applyCC2(alpha1_)+applyCC2(beta_*mass);
}

////////////////////// class OpD1 ////////////////////////////////////////

R OpD1::norm()const
{
   Z mL=StateD1::getVer();
  // Z mL=0;
   Word loc("OpD1::norm");
   CPM_MA
   bool graphics=StateD1::getVer()>=3;
   //bool graphics=true;
   Z n=500; // 200 was too inaccurate in some cases
   R tWait=12.;
   C c1(1.,1.);
   StateD1 s(c1,b_);
   s=s.ran(1370);
   s.nor_();
   if (graphics){
      Graph gg;
      s.mark(gg);
      gg.vis();
      cpmwait(tWait,2);
      gg.clr_();
   }
   R eta,res=0;

   R_Vector etav;
   if (graphics) etav=R_Vector(n);
   for (Z i=1;i<=n;++i){
      s=f_(s);
      eta=s.nor_();
      if (graphics) etav[i]=eta;
      if (eta>res) res=eta;
   }
   if (graphics){
      Graph gr;
      gr.show(etav);
      cpmwait(tWait,2);
      gr.clr_();
      s.mark(gr);
      gr.vis();
      cpmwait(tWait,2);
      gr.clr_();
   }
   CPM_MZ
   return res;
}

Observable OpD1::toObs()const
{
   Z mL=1;
   Word loc("OpD1::toObs()");
   CPM_MA
   Z l=2*bio().d(); // each point of the biotope carries 2 C's
   V<StateD1> cs(l); // canonical ortho-normal system of dimension l
   Z i,j;
   for (i=0,j=1;i<l;++i,++j){
      cs[j]=StateD1(1+i/2,1+i%2,b_);
   }
   Observable ob(l); // made out of the matrix elements of op
      // with respect to the canonical orthonormal system
   for (i=1;i<=l;++i){
      StateD1 opsi=(*this)(cs[i]); // so op is applied only l times
         // modification 2015-07-05
      for (j=1;j<=l;++j){
         ob.set(i,j,(opsi|cs[j]));
      }
   }
   CPM_MZ
   return ob;
}

namespace{

StateD1 act(StateD1 const& psi, Observable const& obs)
{
   StateD1 res=psi;
   C_Vector cv=obs.c()*psi.lin();
   res.setVal_(cv);
   return res;
}

}

OpD1::OpD1(Observable const& obs, Bio const& bio):b_(bio),
f_(F1<StateD1,Observable,StateD1>(obs)(act)){}

////////////////////// class SpectralObjectD1 ////////////////////////////

SpectralObjectD1::SpectralObjectD1(Observable const& obs, Bio const& bio,
bool test):b_(bio), l_(b_.d()*2), spectrum(l_),
eigenVectors(l_,StateD1(b_))
{
   Z mL=1;
   Word loc("SpectralObjectD1(Observable, Bio, bool");
   CPM_MA
   OpD1 o(obs,b_);
   SpectralObject sob=obs.spectralRepresentation(false,false);
   Z i,j;
   V<StateD1> cs(l_); // canonical ortho-normal system
   for (i=0,j=1;i<l_;++i,++j){
      cs[j]=StateD1(1+i/2,1+i%2,b_);
   }

   for (i=1;i<=l_;++i){
      spectrum[i]=sob(i);
      for (j=1;j<=l_;++j){
         eigenVectors[i]+=cs[j]*sob[i][j];
      }
   }
   if (test){
      R dInv=R(1.)/l_;
      R e1=0;
      for (i=1;i<=l_;i++){
         e1+=(o(eigenVectors[i])-eigenVectors[i]*spectrum[i]).norm();
      }
      e1*=dInv;
      cpmcerr<<endl<<loc<<" spectral error = "<<e1<<endl;
      cout<<endl<<loc<<" spectral error = "<<e1<<endl;;
      R e2=0;
      for (i=1;i<=l_;i++){
         for (j=1;j<=l_;j++){
            e2+=((eigenVectors[i]|eigenVectors[j]) - krn(i,j)).abs();
               //krn = Kronecker's delta
         }
      }
      e2*=(dInv*dInv);
      cpmcerr<<endl<<loc<<" ortho error = "<<e2<<endl;;
      cout<<endl<<loc<<" ortho error = "<<e2<<endl;
      cpmassert(e1+e2<1.e-8,loc);
   }
   CPM_MZ
}

SpectralObjectD1::SpectralObjectD1(OpD1 const& op, bool test):b_(op.bio()),
l_(b_.d()*2), spectrum(l_), eigenVectors(l_,StateD1(b_))
{
   Z mL=1;
   Word loc("SpectralObjectD1(OpD1,bool)");
   CPM_MA
   V<StateD1> cs(l_); // canonical ortho-normal system
   Z i,j;
   for (i=0,j=1;i<l_;++i,++j){
      cs[j]=StateD1(1+i/2,1+i%2,b_);
   }
   Observable ob(l_); // made out of the matrix elements of op
      // with respect to the canonical orthonormal system
   for (i=1;i<=l_;++i){
      StateD1 opsi=op(cs[i]); // so op is applied only l_ times
         // modification 2015-07-05
      for (j=1;j<=l_;++j){
         ob.set(i,j,(opsi|cs[j]));
      }
   }
   SpectralObject sob=ob.spectralRepresentation(false,false);
   for (i=1;i<=l_;++i){
      spectrum[i]=sob(i);
      for (j=1;j<=l_;++j){
         eigenVectors[i]+=cs[j]*sob[i][j];
      }
   }
   if (test){
      R dInv=R(1.)/l_;
      R e1=0;
      for (i=1;i<=l_;i++){
         e1+=(op(eigenVectors[i])-eigenVectors[i]*spectrum[i]).norm();
      }
      e1*=dInv;
      cpmcerr<<endl<<loc<<" spectral error = "<<e1<<endl;
      cout<<endl<<loc<<" spectral error = "<<e1<<endl;;
      R e2=0;
      for (i=1;i<=l_;i++){
         for (j=1;j<=l_;j++){
            e2+=((eigenVectors[i]|eigenVectors[j]) - krn(i,j)).abs();
               //krn = Kronecker's delta
         }
      }
      e2*=(dInv*dInv);
      cpmcerr<<endl<<loc<<" ortho error = "<<e2<<endl;;
      cout<<endl<<loc<<" ortho error = "<<e2<<endl;
      cpmassert(e1+e2<1.e-8,loc);
   }
   CPM_MZ
}

StateD1 SpectralObjectD1::operator *(StateD1 const& a)const
{
   Z mL=3;
   Word loc("SpectralObjectD1::operator*(StateD1)");
   CPM_MA
   StateD1 res(b_);
   cpmassert(a.mat(res),loc);
   for (Z i=1;i<=l_;i++){
      C ci=spectrum[i]*(eigenVectors[i]|a);
      res+=eigenVectors[i]*ci;
   }
   CPM_MZ
   return res;
}

V<C> SpectralObjectD1::rep(StateD1 const& state)const
{
   Z mL=2;
   Word loc("SpectralObjectD1::rep(StateD1)");
   CPM_MA
   cpmassert(mat(state),loc);
   V<C> res(l_);
   for (Z i=1;i<=l_;i++){
      res[i]=(eigenVectors[i]|state);
   }
   CPM_MZ
   return res;
}

void SpectralObjectD1::mark(Graph& gr, Z ipolMeth, R gamma)const
{
   Z mL=2;
   Word loc("SpectralObjectD1::mark(Graph,Z,R)");
   CPM_MA
   R gammaMem=Color::getGamma();
   Color::setGamma(gamma);
   R maxSpectrum=spectralRadius();
   R maxVectors=supEigenVectors();
   R fac=maxVectors*cpminv(maxSpectrum);
   VVV<R> vvvr(3,l_,l_+1);
   for (Z i=1;i<=l_;++i){
      C valSpectrum=spectrum[i]*fac;
      Color col(valSpectrum);
      vvvr[1][i][1]=col.Red();
      vvvr[2][i][1]=col.Green();
      vvvr[3][i][1]=col.Blue();
      StateD1 si=eigenVectors[i];
      for (Z j=1;j<=l_;++j){
         C val=si(j);
         col=Color(val);
         vvvr[1][i][j+1]=col.Red();
         vvvr[2][i][j+1]=col.Green();
         vvvr[3][i][j+1]=col.Blue();
            // notice that the channel index is the first one
      }
   }
   CpmImaging::MultiImage<R> mi(vvvr);
   mi.mark(gr,ipolMeth);
   Color::setGamma(gammaMem);
   CPM_MZ
}

 void SpectralObjectD1::markSpectrum(Graph& gr)const
 {
   Z mL=2;
   Word loc("SpectralObjectD1::markSpectrum(Graph)");
   CPM_MA
   R_Vector sp(l_);
   for (Z i=1;i<=l_;i++){
      sp[i]=spectrum[i].real();
   }
   cpmmessage("sp defined");
   gr.mark(sp,RED,1);
   CPM_MZ
 }

bool SpectralObjectD1::writeSpectrum(ostream& str)const
{
   CpmRoot::writeTitle("Eigenvalues of a SpectralObjectD1:",str);
   cpmp(spectrum);
   return true;
}

bool SpectralObjectD1::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(b_);
   cpmp(l_);
   cpmp(spectrum);
   cpmp(eigenVectors);
   return true;
}

bool SpectralObjectD1::scanFrom(istream& str)
{
   cpms(b_);
   cpms(l_);
   cpms(spectrum);
   cpms(eigenVectors);
   return true;
}

namespace{

C f_exp(C const& z, R const& t){ return exp(C(0,t)*z) ;}

}

void  SpectralObjectD1::expi_(R t)
{
   F<C,C> f=F1<C,R,C>(t)(f_exp);
   this->map_(f);
}

namespace{
   R makeReal(C const& z){return z.real();}
}

SpectralObjectD1 SpectralObjectD1::order()const
{
   F<C,R> transf(makeReal);
      // going for the real part
   V<R> spr=spectrum(transf);
      // making an array that holds the real parts
   V<R> spr2(spr);
   Vo<R> sp(spr2);
      // making an auxiliary object that
      // allows to create an ordering permutation
   V<Z> per=sp.permutationForIncreasingOrder();
      // making the ordering permutation
// rest is simple
   V<C> spectrumTemp(spectrum);
   V<StateD1> eigenVectorsTemp(eigenVectors);
   spectrumTemp=spectrumTemp.permute(per);
   eigenVectorsTemp=eigenVectorsTemp.permute(per);
   V<C> spectrumRes=spectrumTemp;
   V<StateD1> eigenVectorsRes=eigenVectorsTemp;
   return SpectralObjectD1(spectrumRes,eigenVectorsRes);
}

////////////////////////// class TrjD1 ///////////////////////////////////

TrjD1::TrjD1(OpD1 const& op, StateD1 const& psi, R const& dt):
TrjBas1(psi,dt),op_(op),tau_(dt_*0.25), tau2_(tau_*2.)
{
   Z mL=StateD1::getVer();
   Word loc("TrjD1(OpD1,StateD1,R)");
   CPM_MA
   phi_=StateD1(psi_.dot(op_));
   CPM_MZ
}

void TrjD1::step_() // DALF integrator
{
   t_+=tau_;
   psi_+=phi_*tau_;
   phi_+=(psi_.dot(op_)-phi_)*R(2);
   t_+=tau2_;
   psi_+=phi_*tau2_;
   phi_+=(psi_.dot(op_)-phi_)*R(2);
   psi_+=phi_*tau_;
   t_+=tau_;
}

void TrjD1::evl_(R tf)
{
   R span=tf-t_;
   R dtMem=dt_;
   Z steps=cpmtoz(cpmpos(cpmrnd(span/dt_)));
   if (steps==0) steps=1;
   set_dt_(span/steps);
   for (Z i=1;i<=steps;++i) step_();
   set_dt_(dtMem);
}

//////////////////////////// class SpTrjD1 ////////////////////////////////

void SpTrjD1::evl_(R tf)
{
   R span=tf-t_;
   SpectralObjectD1 so(so_);
   so.expi_(-span);
   psi_=so*psi_;
   t_=tf;
}
