//? cpmbodycpp.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


using namespace CpmRoot;
using namespace CpmProperties;
using namespace CpmImaging;
using namespace CpmArrays;
using namespace CpmAlgorithms;
using namespace CpmPhysics; // for physical constants
using namespace CpmStd;

using CPM_DIM::Vec;
using CPM_DIM::AxVec;
using CPM_DIM::Spc;
using CPM_DIM::Group;
using CPM_DIM::RefSys;
using CPM_DIM::Plane;
using CPM_DIM::InTens;
using CPM_DIM::Rdim;
using CPM_DIM::Rdim_;
using CPM_DIM::Skeleton;
using CPM_DIM::WireFrame;

using CPM_CAMERA::Camera;

using CPM_RIGID_BODY::ComPar;
using CPM_RIGID_BODY::PointPar;
using CPM_RIGID_BODY::RigidObject;

using CpmFunctions::F;
#ifdef CPM_Fn
using CpmFunctions::F1;
using CpmFunctions::F2;
using CpmFunctions::F4;
using CpmFunctions::F5;
#endif
using CpmFunctions::R2_Func;
using CpmGeo::Angle;

/////////////////////// class ExtSys ///////////////////////////////////

ForTor ExtSys::interaction(const ComPar& cl , const R& dt)
{
   if (field_only) return ForTor();
   ForTor ft1; // felt by *this
   ForTor ft2; // felt by cl
   V< X4<Force,Spc,Force,Spc> > vx4=forceSpots(cl);
   Z i,j,n=vx4.dim();
   Spc x1=get_x();
   Spc x2=cl.get_x();
   for (i=1;i<=n;i++){
      X4<Force,Spc,Force,Spc> x4=vx4[i];
      Force f1=   x4.c1();
      Spc   p1=   x4.c2();
      Force f2=   x4.c3();
      Spc   p2=   x4.c4();
      Spc pc=p1.cen(p2);
      ft1.add(f1,pc,x1);
      ft2.add(f2,pc,x2);
   }
   if (!SysControl::noInteractionChargeToWall){
      // interaction with mirror charges
      if (SysControl::singleCharge){
         PointCharge q=cl.getPointCharge();
         Spc xq=q.get_x();
         V<PointCharge> vpc=influence(q);
         Z n2=vpc.dim();
         for (j=1;j<=n2;j++){
            PointCharge qj=vpc[j];
            Force f1j=qj.force(q);
            Spc xj=qj.get_x();
            ft1.add(f1j,xj,x1);
            ft2.add(-f1j,xq,x2);
         }
      }
      else{
         V<PointPar> vp=cl.getConstituents();
         for (i=1;i<=n;i++){
            PointPar pi=vp[i];
            Spc xi=pi.get_x();
            PointCharge qi=pi.getPointCharge();
            V<PointCharge> vpci=influence(qi);
            Z n3=vpci.dim();
            for (j=1;j<=n3;j++){
               PointCharge qij=vpci[j];
               Force f1ij=qij.force(qi);
               Spc xij=qij.get_x();
               ft1.add(f1ij,xij,x1);
               ft2.add(-f1ij,xi,x2);
            }
         }
      }
   }
   Vec fcl=ft1.get_force();
   Vec dMom=fcl*dt;
   force+=fcl;
   AxVec tcl=ft1.get_torque();
   torque+=tcl;
   AxVec dAngMom=tcl*dt;
   momentum+=dMom; // was + instead of += till
   // 2005-03-22
   angularMomentum+=dAngMom;

   return ft2;
}

Force ExtSys::interaction(const PointPar& p, const R& dt)
{
   if (field_only) return Force();
   X2<Force,Spc> fs=forceSpot(p);
   Force fa=fs.c1();// fa is the force on the system *this
   Spc xa=fs.c2();
   Vec fp=fa.get_f();
   force+=fp;
   Vec dMom=fp*dt;
   AxVec tp=(xa-get_x()).cross(fp);
   torque+=tp;
   AxVec dAngMom=tp*dt;
   momentum+=dMom;
   angularMomentum+=dAngMom;
   return -fa;
}

R ExtSys::dis(const ExtSys& b )const
{
   R_Vector a(6);
   a[1]=x_.dis(b.x_);
   a[2]=v_.dis(b.v_);
   a[3]=ori_.dis(b.ori_);
   a[4]=omega_.dis(b.omega_);
   a[5]=momentum.dis(b.momentum);
   a[6]=angularMomentum.dis(b.angularMomentum);
   return a.sup();
}

bool ExtSys::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("ExtSys",str);
   if (!RigidObject::prnOn(str)) return false;
   cpmp(active);
   cpmp(momentum);
   cpmp(angularMomentum);
   return true;
}

bool ExtSys::scanFrom(istream& str)
{
   if (!RigidObject::scanFrom(str)) return false;
   cpms(active);
   cpms(momentum);
   cpms(angularMomentum);
   return true;
}

//////////////////////////////// class Body ////////////////////////////
namespace{
const R facEl=cpmsqrt(2.)/3;
// factor appearing in Hertz' formula
const R facLarge=1e8;
} // anon namespace

X4<Spc,Spc,Vec,R> Body::clsApp(PointPar const& pa)const
// since only the functions get_x() and get_r() of the particle
// are needed this is given by the same code as for SphPar
{
   Spc px=pa.get_x();
   X2<R,Vec> prv=distAndDirection(px);
      // distAndDirection(...) defined in cpmbodyx.h
      // as shape.distAndDirection(x) which is defined in
      // cpmdimxcpp.h as
      // SubsetOfSpace::distAndDirection(const Spc& x)const
   R d=prv.first();
   R ra=pa.get_r();
   R gap=d-ra;
   Vec n=prv.second();
   Spc res1=px-d*n;
   Spc res2=px-ra*n;
   return X4<Spc,Spc,Vec,R>(res1,res2,n,gap);
}

R Body::gap(PointPar const& pa)const
{
   Spc px=pa.get_x();
   X2<R,Vec> prv=distAndDirection(px);
   R d=prv.first();
   R ra=pa.get_r();
   return d-ra;
}

R Body::gap(ComPar const& cp)const
{
   V<PointPar> vsp=cp.getConstituents();
   Z i,np=vsp.dim();
   R_Vector gaps(np);
   for (i=1;i<=np;++i) gaps[i]=gap(vsp[i]);
   return gaps.inf();
}

V< X4<Force,Spc,Force,Spc> > Body::forceSpots(const ComPar& cl)const
   // force which a particle feels from the Body and which is not
   // mediated by data member field.
{
   Word loc("Body::forceSpots(ComPar)");
   if (!active || fieldOnly()) return V< X4<Force,Spc,Force,Spc> >(0);
   V<PointPar> vsp=cl.getConstituents();
   Z i,np=vsp.dim();
   V< X4<Force,Spc,Force,Spc> > res(np);
   R rp=cl.get_r();
   if (rp<pinhole_) return res;

   ContactProperties cp1=bodyProp;
   for (i=1;i<=np;i++){
      PointPar pi=vsp[i];
      X4<Spc,Spc,Vec,R> x4=clsApp(pi);
      R h=x4.c4(); // for the gap and many following quantities
         // the dependence on i is suppressed in the notation
      ContactProperties cp2=pi.get_cp();
      R r2=pi.get_r();
      R r1=facLarge*r2;
         // the Body *this is considered flat as far as Hertz formula
         // for the elastic contribution is considered
      if (h>cl.get_r()+r2*2) return res;
      if (h>0) continue; // avoid adoo if not needed
      Spc x1=x4.c1();
      Spc x2=x4.c2();
      Vec n=x4.c3();
      Vec v1=get_v(x1);
      Vec v2=cl.get_v(x2);
      Force f=PointPar::forceFrictionAdhesionElastic(
         h,n,r1,r2,v1,v2,cp1,cp2);
      res[i]=X4<Force,Spc,Force,Spc>(f,x1,-f,x2);
   }
   return res;
}

R Body::potElastic(const ComPar& p)const
{
   Word loc("Body::pot(ComPar)");
   Z mL=3;
   CPM_MA
   R res=0;
   if (!active || SysControl::noElasticRepulsion
      || SysControl::ignoreExternalSystems || fieldOnly()) return res;
   R rp=p.get_r();
   if (rp<pinhole_) return res;
   V<PointPar> vsp=p.getConstituents();
   Z i,np=vsp.dim();
   for (i=1;i<=np;i++){
      PointPar pi=vsp[i];
      res+=Body::potElastic(pi);
   }
   CPM_MZ
   return res;
}


X2<Force,Spc> Body::forceSpot(const PointPar& p)const
{
   X2<Force,Spc> res;
   if (!active || fieldOnly()) return res;
      // added 2002-04-19 in order to be able to 'switch off'
      // the body effectively
   R rp=p.get_r();
   if (rp<pinhole_) return res;

   ContactProperties cp1=bodyProp;
   ContactProperties cp2=p.get_cp();

   X4<Spc,Spc,Vec,R> x4=clsApp(p);
   R h=x4.c4(); // for the gap and many following quantities
         // the dependence on i is suppressed in the notation
   R r2=p.get_r();
 //  const R largeFac=1e8;
   R r1=facLarge*r2;
      // the Body *this is considered flat as far as Hertz formula
      // for the elastic contribution is considered
   if (h>0) return res;
   Spc x1=x4.c1();
   Spc x2=x4.c2();
   Spc xc=x1.cen(x2);
   Vec n=x4.c3();
   Vec v1=get_v();
   Vec v2=p.get_v();
   Force f=PointPar::forceFrictionAdhesionElastic(
      h,n,r1,r2,v1,v2,cp1,cp2);
   return X2<Force,Spc>(f,xc);
}

R Body::potElastic(const PointPar& p)const
{
   R res=0;
   if (!active || fieldOnly()) return res;
      // added 2002-04-19 in order to be able to 'switch off'
      // the body effectively
   R rp=p.get_r();
   if (rp<pinhole_) return res;

   ContactProperties cp1=bodyProp;
   ContactProperties cp2=p.get_cp();

   X4<Spc,Spc,Vec,R> x4=clsApp(p);
   R h=x4.c4(); // for the gap and many following quantities
         // the dependence on i is suppressed in the notation
   R r2=p.get_r();
 //  const R largeFac=1e8;
   R r1=facLarge*r2;
      // the Body *this is considered flat as far as Hertz formula
      // for the elastic contribution is considered
   if (h>0) return res;
   Vec n=x4.c3();
   R pot=PointPar::elasticPotential(h,n,r1,r2,cp1,cp2);
   return pot;
}

bool Body::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Body",str);
   if (!ExtSys::prnOn(str)) return false;
   cpmp(bodyProp);
   return true;
}

bool Body::scanFrom(istream& str)
{
   if (!ExtSys::scanFrom(str)) return false;
   cpms(bodyProp);
   setColor();
      // color is a dependent quantity that gets not written to file
   return true;
}

/////////////////////////// class HomFil //////////////////////////

void HomFil::update()
{
   RigidObject::update();
   FieldSpace fsUpd(hfd_.fE_,hfd_.fBH_,hfd_.fG_);
   VectorField vf(fsUpd);
   setField(vf);
   ScalarField sf(fsUpd);
   setU(sf);
   Group g=frmMasSys();
   field*=g;
   fieldU*=g;
}

bool HomFil::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("HomFil::prnOn(...)");
   CPM_MA
   CpmRoot::writeTitle("HomFil",str);
   if (!ExtSys::prnOn(str)) return false;
   cpmp(hfd_);
   CPM_MZ
   return true;
}

bool HomFil::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("HomFil::scanFrom(...)");
   CPM_MA
   if (!ExtSys::scanFrom(str)) return false;
   cpms(hfd_);
   update();
   CPM_MZ
   return true;
}

///////////////////// class VibData ///////////////////////////////

bool VibData::prnOn(ostream& str)const
{
   cpmwat; // write automatic title
   cpmp(amp_);
   cpmp(omg_);
   cpmp(del_);
   return true;
}

bool VibData::scanFrom(istream& str)
{
   cpms(amp_);
   cpms(omg_);
   cpms(del_);
   return true;
}

Group VibData::vib(R t)const
{
   Vec shift;
   if (amp_[1]!=0) shift[1]=amp_[1]*cpmsin(omg_[1]*t+del_[1]);
   if (amp_[2]!=0) shift[2]=amp_[2]*cpmsin(omg_[2]*t+del_[2]);
   if (amp_[3]!=0 && Spc::dimension()>2)
      shift[3]=amp_[3]*cpmsin(omg_[3]*t+del_[3]);
   return Group(shift);
}

VibData::VibData(Record const& rec, Word const& sec)
{
   Z dim=Spc::dimension();
   Z ped=0; // we do not assume vibration data to be provided
   V<R> amp(dim);
   rec.get(sec,"vibAmp",amp,ped);
   V<R> frq(dim);
   rec.get(sec,"vibFrq",frq,ped);
   V<R> phs(dim);
   rec.get(sec,"vibPhs",phs,ped);
   amp_=Rdim(amp);
   omg_=Rdim(frq);
   del_=Rdim(phs);
   omg_*=cpm2pi;
   del_*=cpm2pi;
}

//////////////////////// class Pot ///////////////////////////////////

R potFunc(Rdim const& x, PotData const& vd)
// a combinations of half-space contributions, no attempt to smoothing
// out the edges
{
   bool sw=vd.getSideWalls();
   Rdim sh=vd.s*0.5;
   R res=x[2]+sh[2];
   if (sw){
      R h1L=x[1]+sh[1];
      R h1U=-x[1]+sh[1];
      res=CpmRootX::inf<R>(h1L,h1U,res);
   }
   if (Spc::dimension()==3){
      R h3L=x[3]+sh[3];
      R h3U=-x[3]+sh[3];
      res=CpmRootX::inf<R>(h3L,h3U,res);
   }
   return res;
}

void Pot::upd()
{
   Z nCorners=(Spc::dimension()==3 ? 8 : 4);
   corners=V<Spc>(nCorners);
   Rdim s=dat.s*0.5;
   R s1=s[1];
   R s2=s[2];
   R s3=s[3];
   if (nCorners==4){
      Rdim c1(-s1,s2);
      Rdim c2(-s1,-s2);
      Rdim c3(s1,-s2);
      Rdim c4(s1,s2);
      corners[1]=Spc(c1);
      corners[2]=Spc(c2);
      corners[3]=Spc(c3);
      corners[4]=Spc(c4);
   }
   else{
      Rdim c1(-s1,s2,-s3);
      Rdim c2(-s1,-s2,-s3);
      Rdim c3(s1,-s2,-s3);
      Rdim c4(s1,s2,-s3);
      Rdim c5(-s1,s2,s3);
      Rdim c6(-s1,-s2,s3);
      Rdim c7(s1,-s2,s3);
      Rdim c8(s1,s2,s3);
      corners[1]=Spc(c1);
      corners[2]=Spc(c2);
      corners[3]=Spc(c3);
      corners[4]=Spc(c4);
      corners[5]=Spc(c5);
      corners[6]=Spc(c6);
      corners[7]=Spc(c7);
      corners[8]=Spc(c8);
   }
   RigidObject::update();
#ifdef CPM_Fn
   F<Rdim,R> vsf=F1<Rdim,PotData,R>(dat)(potFunc);
#else
   F<Rdim,R> vsf(bind(potFunc,_1,dat));
#endif
   shape.setShape(vsf);
   shape.set_dr(SysControl::singRange); // factor experiment
   R shpr=dat.s.abs()*0.5;
   set_r(shpr);
   shape.set_ds(shpr*0.002);
   Group g=frmMasSys();
   Group gVib=vd.vib(get_t());
   g*=gVib;
   shape*=g;
   for (Z i=1;i<=nCorners;++i) corners[i]*=g;
   shape.set_x(x_);
   shape.set_r(r_);
}

bool Pot::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("Pot::prnOn(...)");
   CPM_MA
   cpmwat;
   if (!Body::prnOn(str)) return false;
   cpmp(dat);
   cpmp(vd);
   CPM_MZ
   return true;
}

bool Pot::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("Pot::scanFrom(...)");
   CPM_MA
   if (!Body::scanFrom(str)) return false;
   cpms(dat);
   cpms(vd);
   Pot::upd();
   CPM_MZ
   return true;
}

void Pot::mark(const Camera& sc)const
{
   if (SysControl::ignoreExternalSystems) return;
   if (active ){
      sc.mark(corners[1],corners[2],color);
      sc.mark(corners[2],corners[3],color);
      sc.mark(corners[3],corners[4],color);
      if (Spc::dimension()==3){
         sc.mark(corners[4],corners[1],color);
         sc.mark(corners[5],corners[6],color);
         sc.mark(corners[6],corners[7],color);
         sc.mark(corners[7],corners[8],color);
         sc.mark(corners[8],corners[5],color);
         sc.mark(corners[1],corners[5],color);
         sc.mark(corners[4],corners[8],color);
         sc.mark(corners[2],corners[6],color);
         sc.mark(corners[3],corners[7],color);
      }
   }
}

#define CPM_SC\
   cpmassert(i>0 && i<=3,loc);\
   if (i>Spc::dimension()) return true;\
   else{\
      Spc x=cl.get_x();\
      R h=get_h();\
      Iv ivScaled(h*iv[1],h*iv[2]);\
      return ivScaled.contains(x[i]);\
   }

bool Pot::boxSel(const ComPar& cl, Iv iv, Z i)
{
   static Word loc("Pot::boxSel(ComPar,Iv,Z)");
   CPM_SC
}

bool Pot::boxSel(const PointPar& cl, Iv iv, Z i)
{
   static Word loc("Pot::boxSel(PointPar,Iv,Z)");
   CPM_SC
}

#undef CPM_SC

//////////////////////// class Lid ///////////////////////////////////
namespace{
R lidFunc(Rdim const& x, R const& h0, B const& isBottom)
{
   if (isBottom) return x[2]-h0;
   else return h0-x[2] ;
}
} // namespace

void Lid::upd()
{
   Z mL=3;
   Word loc("Lid::upd()");
   CPM_MA
   Z nCorners=2;
   corners=V<Spc>(nCorners);
   R s=get_r();
   Rdim c1(-s,dat.y0,0);
   Rdim c2(s,dat.y0,0);
   corners[1]=Spc(c1);
   corners[2]=Spc(c2);

   RigidObject::update();
#ifdef CPM_Fn
   F<Rdim,R> vsf=F2<Rdim,R,B,R>(dat.y0,dat.isBottom)(lidFunc);
#else
   F<Rdim,R> vsf(bind(lidFunc,_1,dat.y0,dat.isBottom));
#endif
   shape.setShape(vsf);
   shape.set_dr(SysControl::singRange);
   R shpr=dat.iv1.radius();
   set_r(shpr);
   shape.set_ds(shpr*0.002);
   Group g=frmMasSys();
   Group gVib=vd.vib(get_t());
   g*=gVib;
   shape*=g;
   for (Z i=1;i<=nCorners;++i) corners[i]*=g;
 //  field=VectorField(FieldE(),FieldBH(),FieldG());
   shape.set_x(x_);
   shape.set_r(r_);
   CPM_MZ
}

bool Lid::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("Lid::prnOn(...)");
   CPM_MA
   cpmwt("Lid");
   if (!Body::prnOn(str)) return false;
   cpmp(dat);
   cpmp(vd);
   CPM_MZ
   return true;
}

bool Lid::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("Lid::scanFrom(...)");
   CPM_MA
   if (!Body::scanFrom(str)) return false;
   cpms(dat);
   cpms(vd);
   upd();
   CPM_MZ
   return true;
}

void Lid::mark(const Camera& sc)const
{
   Z mL=3;
   Word loc("Lid::mark(Camera)");
   CPM_MA
   if (SysControl::ignoreExternalSystems) return;
   if (active ){
      sc.mark(corners[1],corners[2],color);
   }
   CPM_MZ
}

//////////////////////// class Fen ///////////////////////////////////

R fenFunc(Rdim const& x, R const& r, V<R> const& cenX)
{
   Z n=cenX.dim();
   R_Vector dis(n);
   R dy=x[2];
   R dy2=dy*dy;
   for (Z i=1;i<=n;++i){
      R dxi=x[1]-cenX[i];
      dis[i]=cpmsqrt(dxi*dxi+dy2)-r;
   }
   return dis.inf();
}

void Fen::upd()
{
   Z mL=2;
   Word loc("Fen::upd()");
   CPM_MA
   RigidObject::update();
   //active=;
   R r=dat.r;
   R g=dat.g;
   Iv iv=dat.iv1;
   R l=iv.abs();
   Z nBars=cpmrnd(l/(r*2+g)+1);
   cenX=iv.div(nBars);
   r=(l/(nBars-1)-g)*0.5;
      // r gets adjusted so that the fence fits iv
      // (the endpoints of iv are centers of pickets
      // since this promisses smoother motion of particles
      // at the regions where the fence contacts the pot.
#ifdef CPM_Fn
   F<Rdim,R> vsf=F2<Rdim,R,V<R>,R>(r,cenX)(fenFunc);
#else
   F<Rdim,R> vsf(bind(fenFunc,_1,r,cenX));
#endif
   shape.setShape(vsf);
   shape.set_dr(SysControl::singRange);
   R shpr=l*0.5;
   set_r(shpr);
   shape.set_ds(shpr*0.002);
   Group gr=frmMasSys();
   shape*=gr;
   field=VectorField(FieldE(),FieldBH(),FieldG());
   shape.set_x(x_);
   CPM_MZ
}

bool Fen::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("Fen::prnOn(...)");
   CPM_MA
   CpmRoot::writeTitle("Fen",str);
   if (!Body::prnOn(str)) return false;
   cpmp(dat);
   CPM_MZ
   return true;
}

bool Fen::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("Fen::scanFrom(...)");
   CPM_MA
   if (!Body::scanFrom(str)) return false;
   cpms(dat);
   Fen::upd();
   CPM_MZ
   return true;
}

void Fen::mark(const Camera& sc)const
{
   if (SysControl::ignoreExternalSystems) return;
   if (active){
      Z i,n=cenX.dim();
      if (Spc::dimension()==2){
         for (i=1;i<=n;++i){
            Skeleton ski=Skeleton::circle(Spc(cenX[i],0),dat.r);
            sc.mark(ski,color);
         }
      }
      else{
         Z circPoints=10;
         R h=dat.iv3.abs();
         R z0=dat.iv3.inf();
         Skeleton cy0=Skeleton::cylinder(dat.r,h,circPoints);
         for (i=1;i<=n;++i){
            Skeleton cyi=cy0;
            Vec shi;
            shi[1]=cenX[i];
            shi[3]=z0;
            cyi+=shi;
            sc.mark(cyi,color);
         }
      }
   }
}

////////////////////// class SphTank ///////////////////////////////////

R sphFunc(Rdim const& x, SphTankData const& vd)
{
   return vd.r_-x.abs();
      // value increases from outside to inside since
      // the body is ouside its spherical surface
}

void SphTank::upd()
{
   Z mL=3;
   Word loc("SphTank::upd()");
   CPM_MA
   RigidObject::update();
#ifdef CPM_Fn
   F<Rdim,R> vsf=F1<Rdim,SphTankData,R>(dat_)(sphFunc);
#else
   F<Rdim,R> vsf(bind(sphFunc,_1,dat_));
#endif
   shape.setShape(vsf);
   shape.set_dr(SysControl::singRange); // factor experiment
   R shpr=dat_.r_;
   cpmassert(shpr>0,"SphTank::upd()");
   set_r(shpr);
   shape.set_ds(shpr*0.002);
   Spc p0;
   Group g=frmMasSys();
   shape*=g;
   p0*=g;
   shape.set_x(p0);
   shape.set_r(shpr);
   CPM_MZ
}

bool SphTank::prnOn(ostream& str)const
{
   Z mL=2;
   Word loc("SphTank::prnOn(...)");
   CPM_MA
   cpmwt("SphTank");
   if (!Body::prnOn(str)) return false;
   cpmp(dat_);
   CPM_MZ
   return true;
}

bool SphTank::scanFrom(istream& str)
{
   Z mL=2;
   Word loc("SphTank::scanFrom(...)");
   CPM_MA
   if (!Body::scanFrom(str)) return false;
   cpms(dat_);
   SphTank::upd();
   CPM_MZ
   return true;
}

/****************
void SphTank::mark(const Camera& sc)const
{
   if (SysControl::ignoreExternalSystems) return;
   if (active ){
      Spc x=get_x();
      Spc ori;
      Vec sh=x-ori;
      R r=dat_.r_;
      WireFrame gl=WireFrame::globe(r);
      gl*=Group(sh);
      sc.mark(gl,Color(STEREO));
   }
}

void SphTank::mark(const Camera& sc)const
{
   if (active && !SysControl::ignoreExternalSystems){
      AxVec ax=sc.getAxis();
      Spc p=shape.get_x();
      Z nP=200;
      sc.markFan(shape,p,sc.get_v(),ax,cpm2pi,nP,color);
   }
}
*********************************/

void SphTank::mark(const Camera& sc)const
{
   if (SysControl::ignoreExternalSystems) return;
   if (active ){
      Spc x=get_x();
      Spc ori;
      Vec sh=x-ori;
      R r=dat_.r_;
      Z nP=64;
      Skeleton gl=Skeleton::circle(ori,r,nP);
      gl*=Group(sh);
      sc.mark(gl,Color(STEREO));
   }
}


///////////////////////// some local tools /////////////////////

namespace{

   R profFunc(C const& z, R const& k1, R const& k2, R const& del1,
      R const& del2, R const& amp)
   { return amp*cpmcos(z[1]*k1+del1)*cpmcos(z[2]*k2+del2);}

// form function for class Drum
   R fDrum(Rdim const& x, R const& r0,
      B const& hollow, R2_Func const& prof)
   {
      C z(x[1],x[2]); // experiment
      R x2=Spc::dimension()==3 ? x[3] : R(0);
      R r,phi;
      z.toPolar(r,phi);
      C z1(phi*r0,x2);
      // first is a peripheral coordinate along the circum-
      // pherence of the cylinder
      R res=r-r0-prof(z1);
      if (hollow) res=-res;
      return res;
   }

} // namespace

DrumData::DrumData(RecordHandler const& rch, Word const& sec, Z ped)
{
   rch.read(sec,"amp",amp,ped);
   rch.read(sec,"pointsPerHalfwave",pointsPerHalfwave);
   rch.read(sec,"k1",k1,ped);
   rch.read(sec,"del1",del1,ped);
   rch.read(sec,"k3",k3,ped);
   rch.read(sec,"del3",del3,ped);
   rch.read(sec,"r0",r0,ped);
   rch.read(sec,"omg",omg,ped);
   rch.read(sec,"hollow",hollow,ped);
}

//////////////////////// class Drum ///////////////////////////////////

void Drum::upd()
{
   Z mL=3;
   Word loc("Drum::upd()");
   CPM_MA
   RigidObject::update();
   if (drd.r0<=R(0.)){
      active=0;
      CPM_MA
      return;
   }
// First we have to make the wavelength arround the circumference
// equal to an integer faction of it.
   Z nk1=cpmrnd(drd.r0*drd.k1);
   drd.k1=R(nk1)/drd.r0; // corrected value such that a integer number
      // of waves makes up the circumference of the circle of
      // radius drd.r0
   R nk3=drd.k3*drd.r0;
   R nk=cpmsup<R>(nk1,nk3);
   nP=cpmrnd(nk*drd.pointsPerHalfwave);
#ifdef CPM_Fn
   R2_Func prf=F5<C,R,R,R,R,R,R>(drd.k1,drd.k3,drd.del1,drd.del3,
      drd.amp)(profFunc);
   F<Rdim,R> fif=F3<Rdim,R,B,R2_Func,R>(drd.r0,drd.hollow,prf)(fDrum);
#else
   R2_Func prf=F<C,R>(bind(profFunc,_1,drd.k1,drd.k3,drd.del1,drd.del3,
      drd.amp));
   F<Rdim,R> fif(bind(fDrum,_1,drd.r0,drd.hollow,prf));
#endif
   shape.setShape(fif);
   R drFine=R(1e-3)*drd.r0;
   shape.set_ds(drFine);
   R drCoarse=drFine*R(20.)+drd.amp*R(0.5);
   shape.set_dr(drCoarse);
   Group g=frmMasSys();
   Group gVib=vd.vib(get_t());
   g*=gVib;
   shape*=g;
      // it is important that the data member rs_ of RigidObject does
      // not move as a consequence of the vibration. Otherwise the
      // vibration shift could not be defined relative to this frame.
   shape.set_x(x_);
   shape.set_r(r_);
   CPM_MZ
}

bool Drum::prnOn(ostream& str)const
{
   cpmwat;
   if (!Body::prnOn(str)) return false;
   cpmp(drd);
   cpmp(vd);
   return true;
}

bool Drum::scanFrom(istream& str)
{
   if (!Body::scanFrom(str)) return false;
   cpms(drd);
   cpms(vd);
   Drum::upd(); // sets nP
   return true;
}

void Drum::mark(const Camera& sc)const
{
   if (active && !SysControl::ignoreExternalSystems){
     AxVec ax=sc.getAxis();
     Spc p=shape.get_x();
     sc.markFan(shape,p,sc.get_v(),ax,cpm2pi,nP,color);
   }
}
