//? cpmgroupgraph.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_GROUPGRAPH_H_
#define CPM_GROUPGRAPH_H_
/*
    Description: directed graphs for organizing the
      communication between parallel processes.
*/
#include <cpmv.h>

namespace CpmGeo{

   using CpmArrays::V;

   template<class Group>
   class GroupGraph{ // directed graph with group elements attached
      // to edges. Has nothing to do with
      // CpmGraphics::Graph

      Z n{0};
         // number of vertices;

      V< V<Z> >c;
         // c[i][j]=1 iff (i,j) is a (directed) edge of the graph.
         // thus c is the connection matrix

      V< V<Group> > g;
         // g[i][j] should have a usefull value if (i,j) is an edge
         // of the graph

      static Z cutNeg(Z n){ return n>0 ? n : 0;}
         // trivial tool for safely defining the constructor

   public:
      typedef GroupGraph Type;
      virtual GroupGraph* clone()const{ return new  GroupGraph(*this);}
      CPM_IO_V
         // IO should be virtual in order to allow
         // additional attributes in derived classes
// all functions test the validity of their arguments and
// stop via cpmerror for unvalid ones.

explicit GroupGraph(Z n=0);
         // a graph with n vertices and no edges

      void addEdge(Z i, Z j, const Group& gr=Group() )
      { c[i][j]=1; g[i][j]=gr;}
          // makes (i->j) an edge and associates gr
          // with this edge

      void addEdgeSym(Z i,Z j, const Group& gr=Group() )
      { c[i][j]=1; g[i][j]=gr; c[j][i]=1; g[j][i]=!gr;}
         // makes (i->j) and (j->i) edges
         // Group has to define inversion

      void remEdge(Z i, Z j){ c[i][j]=0; g[i][j]=Group();}
         // 'removes' an edge

      void remEdgeSym(Z i, Z j)
         { c[i][j]=0; g[i][j]=Group(); c[j][i]=0; g[j][i]=Group();}
         // self-explanatory

      Z dim()const{ return n;}

      bool close(Z i,Z j)const{ return c[i][j]==1;}
      Group group(Z i, Z j)const { return g[i][j];}
   };

template<class Group>
GroupGraph<Group>::GroupGraph(Z n_):
n(cutNeg(n_)),c(n,V<Z>(n)), g(n,V<Group>(n)){}

template<class Group>
bool GroupGraph<Group>::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("GroupGraph",str);
   cpmp(n);
   cpmp(c);
   cpmp(g);
   return true;
}

template<class Group>
bool GroupGraph<Group>::scanFrom(istream& str)
{
   cpms(n);
   cpms(c);
   cpms(g);
   return true;
}

} // namespace

#endif
