//? cpmpardynsys.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_PARDYNSYS_H_
#define CPM_PARDYNSYS_H_
/*
Description: Heterogeneous dynamical systems prepared for
   parallel computing.
   Useful as a base class for a granular system, where
   a large number of particles, and a variety of confining systems
   such as walls, movable pistons, semi-permeable membranes ...
   form a dynamical system.

   Development version with speed increasing features.
   Some of the comments are not updated and thus refer to the old
   class HetDynSys.
*/
#include <set>
#include <sstream>

#include <cpmrmatrix.h>
#include <cpmgraph.h>
#include <cpmm.h>
#include <cpmp.h>
#include <cpmvectormore.h>
#include <cpmalgorithms.h>
#include <cpmimagingtools.h>
#include <cpmrecordhandler.h>
#include <cpmproperties.h>
#include <cpmdynaux.h>
#include <cpmdiscretespace.h>
#include <cpmparticleviewer.h>
#include <cpmmacros.h>
#include <cpmpardef.h>
#include <cpmmpi.h>

namespace CpmDynSys{

   // no quantities from namespaces CpmDim2 or CpmDim3
   // should be used here. The dimension comes in when template
   // arguments get selected e.g. Spc as CpmDim2::Spc or
   // CpmDim3::Spc. It was not always easy to keep the template arguments
   // general enough and not assuming the specific PaDyAn choice for
   // Spc, FieldSpace, VectorField, ....

   using namespace CpmStd;
   using std::set;
   using std::vector;
   using namespace CpmStd;

   using CpmRoot::read;
   using CpmRoot::write;
   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;

   using CpmSystem::Error;
   using CpmSystem::OFileStream;
   using CpmSystem::Timer;

   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;
   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::Vp;
   using CpmArrays::Pp;
   using CpmArrays::Vo;
   using CpmArrays::S;
   using CpmArrays::M;
   using CpmArrays::X2;
   using CpmArrays::T2;
   using CpmArrays::X3;


   using CpmFunctions::F;
   #ifdef CPM_Fn
   using CpmFunctions::F2;
   using CpmFunctions::F3;
   using CpmFunctions::F4;
   #endif
   using CpmFunctions::R2_Func;

   using CpmRootX::B;
   using CpmRootX::hugeNumber;
   using CpmRootX::Record;

   using CpmLinAlg::R2;

   using CpmGeo::Iv;
   using CpmGeo::Angle;
   using CpmGeo::DEG;

   using CpmAlgorithms::CyclicAlarm;
   using CpmAlgorithms::BurstAlarm;
   using CpmAlgorithms::LatticeBox;

   using CpmGraphics::Graph;
   using CpmGraphics::Frame;
   using CpmGraphics::Frames;

   using CpmImaging::Color;
   using CpmImaging::Palette;
   using CpmImaging::INVISIBLE;
   using CpmImaging::WHITE;
   using CpmImaging::LIGHTBLUE;

   using CpmProperties::SysControl;
   using CpmProperties::SysControlState;

   using CpmPhysics::ParticleViewer;

   using CpmMPI::Cpm_com;

namespace aux{
   void adjust_dt(R&,Z);
   R log10s(R const& x);
}// aux

/////////////////////// class ParDynSys ////////////////////////////////
// 'Parallel Dynamical System'

#define CPM_CN ParDynSys
   // this style of introducing the class name helps to define
   // new versions

#define CPM_TAL template <class Par, class Ext,\
   class Vec, class Spc, class Group, class Partition,\
   class For, class FieldSpace, class ScalarField, class VectorField,\
   class PointCharge, class SourceOfVectorField, class MagDipole,\
   class SphPart, class Camera>
// 'dependencies' list of template arguments
// a function template
// CPM_TAL
// func ...
// can be inserted into an argument list of another function as
// func CPM_DEP
// surch for CPM_DEP inthe present file.

#define CPM_DEP <Par,Ext,Vec,Spc,Group,Partition,\
   For,FieldSpace,ScalarField,VectorField,PointCharge,\
   SourceOfVectorField,MagDipole,SphPart,Camera>

#define CPM_FCN CPM_CN<Par,Ext,Vec,Spc,Group,Partition,\
   For,FieldSpace,ScalarField,VectorField,PointCharge,\
   SourceOfVectorField,MagDipole,SphPart,Camera>

/////////////////////////// class ParDynSys //////////////////////////////
CPM_TAL
class CPM_CN{// = ParDynSys dynamical system of particles, granular matter

   static bool fEqui(const Par& p1, const Par& p2)
   {
      return p1.get_id()==p2.get_id();
   }

protected:
// here we havn't grouped data into independent and dependent ones.
// The independent data are those which are handled in function
// prnOn(...). Function scanFrom(...), together with function
// prepareStep(...) defines the mechanism by which
// the dependent quantities get their values.

   StepControl stpCon;
      // book-keeping quantities for time stepping, these data also
      // include the time t, which, despite of its physical importance
      // is placed into one house together with some minor relatives

   V<Par> par;
      // here, polymorphism isn't essential, toner and carrier particles
      // can be reasonably described as instances of a single class

   Vp<Ext> ext;
      // here, the components may be a toning shell, a magnetic roller, a
      // photo-conductor .... In the old version these entites were all
      // combined in a single entity since polymorphic arrays were not at
      // hand. All these types of external systems have to be derived
      // from class Ext.

   Z np;
      // number of particles, not really independent, related to par
      // makes I/O easier if considered independent. The way of
      // initializing np and ne in the constructor asks np to be placed
      // here later than par and ne later than ext.

   Z ne;
      // number of external systems; see np

   V<Spc> xPar;
      // list of particle positions, needed to monitor motion
      // which helps to decide when certain arrays have to be updated

   vector< set<Z> > nearList;
      // nearList is an array of index sets.
      // since 2006-11-09 I use std::vector again here since V would
      // require that std::set would define nameOf.
      // nearList[i-1] holds all indexes of particles which
      // have distance from particle i smaller than nearLength.

   R nearLength;
      // see nearList

// data and functions supporting parallelism. Notice that the partition
// of the whole system into parts and describing the interaction between
// different parts by electromagnetic fields which can be represented
// by relatively few data is a contribution to speeding up the dynamical
// algorithm even if the various parts can't be computed in parallel.

   mutable Z rank;
   mutable Z size;
      // declared mutable for technical reasons: the function that
      // is suitable for initializing these happens to be favorably
      // declared const

   M<Z,Color> showColor;
      // Map (associative list) of particle colors to be used in marking
      // particles if different from default. The key is to be
      // understood as the particle ID number, not the index in the
      // array par.

   Partition partition;
      // partitions the available space into regions for which a process
      // takes responsibility. The whole space available to the particles
      // of the system is divided into disjoint subsets. And each such
      // subset (region) is associated with just one process. To each
      // region there is an extended region which contains its region so
      // that the extended regions cover the space available to the
      // particles in a manner that allows overlap. Let the regions be
      // denoted X(i), the extended regions Y(i) (see cpmpartitionx.h),
      // where i indexes the processes.

   B meanFieldAct;
      // initialized as false, which is OK

   V<Spc> pointList;
      // a list of points determined by partition and needed
      // for defining meanFieldPar

   //V<Z> nextPointIndexes(const Spc& x, R range);
   set<Z> nextPointIndexes(const Spc& x, R range);
      // returns the list of all i for which
      // |pointList[i]-x|<=range. If such points don't
      // exist, we return the void list.

   VV<Spc> repPointList;
       // an augmentation of pointList taking into account toroidal
       // topology

   VectorField meanFieldPar;
      // smooth average field created by particles not listed in
      // nearList according to my 'self-consistent far field method'.
      // Besides this, it contains contributions
      // from virtual periodic replications (if partition implies
      // toroidal topology) and contribution from course grained
      // electro-magnetic sources from the neighbor processes.
      // See function prepareStep()

   ScalarField meanPotPar;
      // electrical potential corresponding to meanFieldPar

   void redistribute(void);
      // Let be any state of par over the various processes.
      // After call of this function, par for each process is in
      // a form that each process has just the particles that he owns
      // and those that it has to use.

      // Till 01-11-08 I had a hard time to understand the deadlock
      // happening in earlier versions. The final solution follows the
      // suggestion 'different ordering of calls between tasks' in
      // Section 2 of
      // http://www.tc.cornell.edu/Services/Edu/Topics/MPI/Pt2pt2/

   void setTimesFromExtSys();
      // make all times of system identical and equal to the mean
      // value of the times defined by the external systems
      // (which all have a get_t() function). This is important for
      // full reconstruction of a state from the information
      // written on movie-files
public:
   void prepareStep(const Camera& cam=Camera());
      // sets the preconditions for function step() to work
      // properly. After this function did its work, the function step
      // runs 'autonomously' without communicating with other processes
      // till it calls redistribute(). Thus the responsibility for calling
      // step(dt) with a reasonable value of dt is with the user of step.
      // The idea is that all influences which a particle feels from
      // particles outside the near zone of radius nearLength should be
      // effectively constant during dt.
   V<Spc> posPar()const{
      V<Spc> res(np);
      for (Z i=1;i<=np;++i) res[i]=par[i].get_x();
      return res;
   }
protected:
   virtual void steplet(R t0, R dt, const Camera& cam);
      // time evolution over a fixed span dt of time. Implements my
      // direct mid-point method of time-discrete dynamics.
      // After function call, the whole system is in a state that is the
      // best guess for a system state at time t0+dt (based on the
      // information contained in *this at the point of function call).
      // In particular, for all particles par[i] we have after function
      // call par[i].get_t()==t0+dt. In parallel computing the various
      // processes don't interact during executing this function.

// constructors. The data elements tPar, xPar, nearList, filValList
// will not become initialized by the constructors, but only in preparing
// the execution of a time step.
public:
   CPM_CN(V<Par>  par_, Vp<Ext> ext_,
      Partition partition_=Partition()):
      par(par_),ext(ext_),np(par.dim()),ne(ext.dim()),showColor(),
      partition(partition_){}

//public:

   CPM_CN(V<Par>  par_, Z ne_):
      par(par_),ext(Vp<Ext>(ne_)),
      np(par.dim()),ne(ext.dim()),showColor()
      { }
      // simple constructor for filling in the external systems by
      // calling
      // void setExtComp(Z i, const X& x);
      // this is sufficient for enabling the visualization functionality
      // of the class.
      // For parallel computing also a proper setting of partition is
      // required.

   CPM_CN():par(V<Par>(0)),np(0),ne(0),showColor(){}
         // notice that the default for V<Par> would bring in one
         // particle

// I/O and documentation

   typedef CPM_CN Type;
   typedef CPM_CN CloneBase;
   CPM_IO_V // stream I/O, important that this is virtual
      // these are the functions scanFrom and prnOn which deserve some
      // clarifications.
      // Precondition of prnOn(str) is only that
      // str in valid. Then a complete data set for the system state gets
      // written on the stream, restricted only by the number of digits
      // chosen for data to ASCII conversion (set in cpmdefinitions.h
      // #define CPM_WRITE_PRECISION). By definition, the data set is
      // complete if it - together with the data of SysControl
      // determines the behavior of function step().
      // All 'bookkeeping quantities' that get initialized
      // anyway within function step() are not written
      // to str. These, however are copied by the copy constructor and
      // by the assignment operator. Unfortunately there is even a third
      // copying method in the game: the way in which states are filed
      // (movie-file production) Since these files tend to grow very
      // large, it is essential to have this process very data-efficient.
      // See comment to function file(...).
      // Precondition of scanFrom(str) is that str is valid and that the
      // data on str are such as if they had been written by
      // pardyn.prnOn(str) where pardyn is an instance of the present
      // template class with the same template arguments as *this
      // and with the  s a m e  n u m b e r  and  e x a c t  t y p e s
      // (derived from ExtSys) as this.
      // Postcondition is that *this is changed exactly into pardyn.
      // (By definition, a postcondition is met after function call if
      // the precondition held before function call; this seems to be a
      // useful notion, see e.g. Paul Taylor: Practical Foundations of
      // Mathematics, Cambridge 1999, p. 199)

   virtual Word nameOf()const{ return Word("ParDynSys<...>");}

   virtual CPM_FCN* clone()const{ return new CPM_CN(*this);}

 //  virtual ~CPM_FCN(){}  // is an error in c++20, fortunately easy to avoid
   virtual ~CPM_CN(){}

   void done(R dt_, R dis_){ stpCon.done(dt_,dis_);}
      // if a time step dt was effective in changing the dynamical
      // variables of the system, this is the change to the 'book-keeping
      // quantities'

// obvious 'read'-access functions
   R get_t()const{ return stpCon.t;}
   R get_dt()const{ return stpCon.dtMean;}
   R get_tLimit()const{ return stpCon.tLimit;}
   bool getRelaxFlag()const{ return stpCon.relaxFlag;}

   const Par& getPar(Z i)const{ return par[i];}
   V<Par> getPar()const{ return par;}
   Vp<Ext> const& getExt()const{ return ext;}
   Vp<Ext>& getExt(){ return ext;}

   template <class X>
   X getExtComp(Z i, const X& x)const{ return ext(i,x);}
      // template member function works!!! fine
      // Here the value of x is unimportant, it is needed only to tell
      // the function the precise type of the returned quantity

   Z get_np()const{ return np;}
   Z get_ne()const{ return ne;}
   const Partition& getPartition()const{ return partition;}
   void setPartition(const Partition& pa){ partition=pa;}

   virtual Z get_nTot()const{ return np;}
      // in a derived class that support parallel computing
      // this has to become re-defined to mean the total number
      // of particles over all processes

   Spc cenOfMas(const V<B>& vb)const;
      // returns the center-of-mass of those particles par[i]
      // for which vb[i]==true (actually we need the center-of-mass of
      // all carrier particles, but being a carrier particle is too
      // special a concept to be placed here. The default argument is set
      // such that the center-of-mass of all particles is calculated if
      // no vb-argument is given

   R get_m(const V<B>& vb)const;
      // returns the mass of those particles par[i]
      // for which vb[i]==true

   R kinEnr(const V<B>& vb)const;
      // returns the kinetic energy of those particles par[i]
      // for which vb[i]==true

   R potEnr(const V<B>& vb)const;
      // returns the potential pair energy of those particles par[i]
      // for which vb[i]==true
      // Included are only Coulomb forces of point charges and
      // gravitational forces of mass points (added 2004-05-03)
      // Useful for testing energy conversation in pure Coulomb
      // systems or gravitational mass point systems.

   R kinEnr()const;
      // returns the kinetic energy of all particles owned by rank

   R potEnr()const;
      // returns the potential pair energy of all particles owned by rank
      //
      // by calling  res+=par[i].pot(par[j]); within a suitable loop.
      // So, the pot memberfunction of class Par decides what
      // contributions are employed here.
      // Notice that this has np^2 complexity and thus should be called
      // with caution.
      // Useful for testing energy conversation in pure Coulomb
      // systems or gravitational mass point systems.

   R potElasticExt()const;
      // potential energy of all particles resulting from elastic
      // interaction with the external systems

   R potHomFil(FieldSpace const& fs)const
   {
      R res=0;
      for (Z i=1;i<=np;++i){
         if (par[i].getOwner()==rank) res+=par[i].pot(fs);
      }
      return res;
   }
      // Total potential energy of the particles in the homogeneous field
      // of (constant) strength fs

   R totEnr()const;
      //: total energy
      // An expression for parts of the energy which is the total energy
      // in simple conservative test systems.

   R totAngMomCom(Spc const& p, Vec const& dir)const
      //: total angular momentum component
   {
      R res=0;
      for (Z i=1;i<=np;++i){
         if (par[i].getOwner()==rank) res+=par[i].angMomCom(p,dir);
      }
      return res;
   }

   R get_m()const;
      // returns the mass of all particles

// 'write'-access to the most important substructure
   template <class X>
   void setExtComp(Z i, const X& x){ext.set(i,x);}
      // template member function works!!! fine. Class X has to be a
      // derivation of Ext.

   void setPar(const V<Par>& par_){ par=par_; np=par.dim();}
      // a combination of get and set allows to perform modifications of
      // *this for which there are no member functions provided in
      // ParDynSys<>. For instance, a list of np instances of Group can
      // be used to transform the particles, each particle by its own
      // transformation:
      // V<Par> vp=getPar(); // extract the particles for modification
      // V<Group> vg(np);
      // Z i;
      // for (i=1;i<=np;i++) vg[i]=...; // define the Group-instances
      // for (i=1;i<=np;i++) vp[i]*=vg[i]; // transform the extracted
      // particles
      // setPar(vp); // Put the particles back. Now *this houses as par
      // the transformed list of particles

   void selectParticles(const V<B>& vb)
   { par=par.select(vb); np=par.dim();}
      // allows to remove particles from the system under control of a
      // boolean array. Paricles par[i] for vb[i]==true are selected to
      // stay, not to get removed!
      // has to be redefined in a derived class

   void setActivityOfExternalSystems(Z iext, Z act)
      { ext(iext).setActivity(act); }
      // allows to disable the dynamical anf graphical activity of
      // selected external systems.

   void setVelocityOfExternalSystems(Z iext, const Vec& v)
      { ext(iext).set_v(v); }

   void setCol(V<Palette> const& c)
      {for (Z i=1;i<=np;i++) par[i].setCol(c[i]); }
         // setting the visualization color

   void setShowColor(Z i, Color c){ showColor[i]=c;}
      // setting the special Color (not only Palette). For the particle
      // with ID-number equal to i, color c is available for
      // visualization.

// 'write' access to simple control quantities
   void set_t(R t0);
      // make all times of system identical to the input
      // value t0;
   void set_tLimit(R dtL){ stpCon.tLimit=dtL;}
   void set_tSmall(R ts){ SysControl::tSmall=ts;}
      // see attribute tSmall
 //  void clear_dtLimit(void){stpCon.dtLimit=hugeNumber;}
   void clear()const{ VectorField::newScl(); }

   void setRelaxFlag(bool rf)
      // switching between 'relaxed dynamics' (a 'dynamics' which sets
      // all velocities to zero after each time step) and normal dynamics
      {
         stpCon.relaxFlag=rf;
         if (rf) for (Z i=1;i<=np;i++) par[i].stop();
      }

   void showParticleFields(bool b){SysControl::showPartFields=b;}

// kinematical transformations
   void boost(Vec const& v){ for (Z i=1;i<=np;i++) par[i].boost(v);}

   void ranBoost(Vec const& v, R dv, R angleRange, Z j=0)
   { for (Z i=1;i<=np;i++) par[i].ranBoost(v,dv,angleRange,++j);}

   void ranRotPar(R angleRange=180, Z j=0);
      // randomly rotate particles

   virtual void operator*=(const Group&);
      // geometrical transform of the whole system
      // if one adds data in derived classes, these have to be
      // transformed too; thus the function should be virtual

   void transform(Z i, const Group& g){ par[i]*=g;}
      // geometrical transform of particle i

   void transformExternalSystems(Z i, const Group& g){ ext(i)*=g;}
      // geometrical transform of external system i

   For force(const Par& par1, const Par& par2)const;
      // force which particle par1 feels from particle par2 and all the
      // mirror charges which get generated by par2 (not by par1 !)
      // Mirror charges come from the external systems individually
      // (e.g. from the Shell and the Target) and are defined in the
      // function influence() of the respective classes.
      // For force(par1,par1) returns the force felt by its own
      // mirror charge without an non-sensical direct self-force
      // since Clust::forces(const Clust&) takes care of this.

   FieldSpace field(const Spc& sp, const Par& par1,
      const bool& farAway=false)const;
      // field strength that particle par1 creates in point sp together
      // with all the mirror charges created in external systems. These
      // are defined by the same functions influence() called in the
      // previous function force().
      // If the last argument is true, the various mirror charges
      // created from the same external devise get collected into
      // one charge for computational efficiency.

   R potential(Spc const & sp, Par const& par1,
      bool const& farAway=false)const;
      // potential belonging to interaction type it that particle par1
      // creates in point sp ( together with all the mirror charges
      // created in external systems, if it==1)
      // it: interaction type
      // it=1: electric, it=2: magnetic, it=3 gravitational
      // Notice that the more important vector fields unify the
      // interaction types by using the return type FieldSpace
      // which unites electric, magnetic, and gravitational field.
      // The analog method here would be to use a return type R3.
      // However, too much of the existing code is based on
      // type ScalarField which would not harmonize with such a
      // choice.

// lambda abstractions of the previous

   VectorField fieldPar(const Par& par1, bool farAway=false)const;
    //  { return F2<Spc,Par,bool,FieldSpace>(par1,farAway)(field);}
    // This does not work since field is a member function and not a
    // plain function of signature Spc x Par x bool -->FieldSpace
    // However, the working version given below (not inline) is
    // also not much more complicated. It is a triumph of my lambda
    // abstraction mechanism

   ScalarField potentialPar(Par const& par1, bool farAway=false)const;

   void sizeAndRank()const{
      size=Cpm_com.getSize();
      rank=Cpm_com.getRank();
   }

   void step(R dt,const Camera& cam = Camera());
      // evolving the system over exactly the time step dt.
      // After call all subsystems refer exactly to the same time. The
      // working of the function will not be completed before it is
      // completed in each of all parallel processes. This function is
      // intended to be called for dt' s large enough that the change
      // of system state conveys information to the user. Since,
      // however, the particles which show a distance larger than
      // nearLength interact in a course-grained and constant manner
      // during dt (especially no inter-process communication takes place
      // during this span of time) dt has also not to be taken too large.
      // My present usage is to have dt=(time span between adjacent movie
      // frames)/5=1.88e-6 s. This time step is divided into
      // SysControl::fixedStep=20 steplets which implement the direct
      // mid-point method of integration.

   bool run(R dtStep, Z nSteps,
         Z filingPeriod, Z docuPeriod,
         Z viewingPeriod, Z stoppingPeriod, const Word& movieName,
         Camera& cam, const Timer& tm, const F<R,Z>& f,
         OFileStream& ofs, bool albuming=false);
      //: run
      // Calling function step(dtStep) nSteps times in sequence
      // and writing to file (under control by movieName) and visualize
      // on screen (under control by cam) with a period (in number of
      // calls of step(R)) of filingPeriod and viewingPeriod
      // respectively.
      // Exceptions: nSteps<0 or dtStep<0 renders the action of the
      // function trivial.
      // filingPeriod<1 disables all filing
      // viewingPeriod<1 disables all display activity
      // Thus for nSteps==0 steps(R) will not be called but *this
      // will be filed unless filingPeriod<1 and *this will be displayed
      // on screen unless viewingPeriod<1.
      // The argument f manipulates as a function of t the parameters
      // in SysControl.
      // The argument tm allows to set a limit to the computation time.
      // After a filing action tm gets asked whether the timeOut
      // condition holds and if this is the case the program is stopped
      // after the last state has been written to file completely. After
      // restarting the run from file one needs not to adjust nSteps
      // according to the computational progress already achieved.
      // The new functionality governed by tm is not implemented in the
      // present class but only in the derived class Station by suitable
      // re-definition of virtual function file.
      // Return value is true iff function was stopped early (earlyStop).
      // If 'albuming' is true, all images recorded in the movie file will
      // also be recorded as ppm images.

   R_Vector anlOrd(R dtStep, Z nSteps, Z nFrames, R fac, Word const& tag,
      Camera& cam, Iv const& iv, Z repLog, R repOrderD, R repOrderE,
      R repOrderL, bool makeMovie=true, bool storeGraphAsImage=false,
      Vec const& dir=Vec(V<R>{0.,1.,0.}));
      //: analyze order
      // documents the behaviour of a consecution of nFrames (>=2) runs
      // where the number of steplet calls within a step is
      // given as SysControl::fixedStep for the first run and increases
      // by the integer factor fac in each following run.
      // Both the trajectory , the total energy totEnr() along
      // trajectories, and a component of the total angular momentum
      // are displayed on screen and the corresponding data
      // written to various files, the name of which always contains tag.
      // For the determination of the order in dt only the fraction
      // of the trajectory given by iv is taken (Iv iv(0,1) corresponds
      // to the whole trajectory). If iv is void extracting the order from
      // data is disabled. The same is true for the orbit order if
      // repOrderD>0, and for the energy order if repOrderE>0,
      // and for the angular momentum order if repOrderL>0.
      // If iv is void and repOrderD<=0, then order is taken as 0, so
      // that no scaling of data according to an order takes place. The
      // same is true for repOrderE<=0 and repOrderL<=0.
      // If repLog!=0, the screen graphics is shown with y replaced by
      // log10(cpmabs(y)+1e-16). Data files tag&X&".txt" are created for
      // X = D,DL,E,EL,L,LL in a format that facilitates the creation of
      // order-collapsed families of curves for path variation and energy
      // variation in either normal (X=D,E,L) or logarithmic (X=DL,EL,LL)
      // form.
      // If the argument makeMovie has its default value
      // two movies fsMovie&tag&".txt" and midMovie&tag&".txt" are made:
      // The first gives the final states of the runs in direction of
      // largest steplet-step to smallest steplet-step
      // and the last one gives the whole trajectory for a run in the
      // middle (the larger if nSteps+1 is even !).
      // 2010-08-28: type of fac changed from Z to R, and last two argu-
      // ments added.
      // 2010-09-04: dir is the direction that defines the component of
      // the angular momentum to be monitored. Consideration of angular
      // momentum was added during the last week.

   T2<R_Vector> anlExc(R dtStep, Z nSteps, Z substeps, R fac,
                   bool logView)const;
      //: anlalize exactness (?)
      /*
         We compare two trajectories (CM-positions of all particles as a
         a function of time. These two trajectories use the same number and
         stepsize for function step() but not the same number of steplets
         for a single step. The number of steplets per step is given by the
         argument 'substeps' for the first trajectory and that times 'fac'
         (up to rounding) for the second one. Let res the return value.
         Then res.c1() is the list of the time values to which the position
         differences belong and res.c2() lists the differences themselves.
         The differences are the means over the contributions of all par-
         ticles and can be give as its logarithm if required.
       */

   V< X2<V<Spc>,R> > xParTrj(R dtStep, Z nSteps)const;
      //: xPar trajectory
      // returns the list xPar of the nSteps+1 states:
      // *this (initial state) and nSteps states resulting from applying
      // step(dtStep,cam) nSteps times. The xPar lists are given together
      // with the time to which they belong.

   Vp<CPM_FCN> trj(R dtStep, Z nSteps, Camera& cam, Z viewingPeriod=0);
      //: trajectory
      // returns the nSteps+1 states: *this (initial state) and nSteps
      // states resulting from applying step(dtStep,cam) nSteps times.
      // Visualization with cam can be enabled by giving the last
      // argument a value >0. The documentation channel OFileStream& ofs
      // would be redundant in this case since any application dependent
      // analysis of the state can be done with the result in a
      // transparent manner.
      // Notice that this does not work usefully with parallel execution
      // since one probably would need the array of total system states
      // in this case.
      // Due to the usage of the polymorhic array one gets the total
      // information of the states even when calling this function for a
      // suitably derived class. Example of usage:
/*
         Vp<Sys2Base> vsb=sys.trj(dtStep,nSteps,cam);
         Z dv=vsb.dim();
         for (Z i=1;i<=dv;++i){
            cam.clear();
            vsb(i).mark(cam,true);
            cpmwait(SysControl::tWait);
         }
*/
      // For long runs of large systems this needs a lot of memory,
      // wheras function run() is much less demanding with respect to this
      // point, it has a much more complicated interface, though.

   void movieFromTrj(Vp<CPM_FCN> const& trj, Word const& fileName,
      Camera& cam, Z viewingPeriod=1, bool showOnly=false);
      // If the last argument does not have its default value, no
      // writing to file, only screen display, takes place.
      // 2010-08-28: last argument added.

   void stopParticles()
      // sets the velocities of the particles to zero
   {
      for (Z j=1;j<=np; j++){
         par[j].stop();
#if !defined(CPM_POINT_PARTICLES)
         par[j].deElectrize();
#endif
      }
   }

   virtual void setVel(){}
      // allows for derived classes to implement methods that give the
      // particles suitable velocities e.g. as in a gas of some given
      // temperature. Since all velocities are initialized as zero, it is
      // appropriate to define this action as addition of velocities to
      // the existing one.

// descriptors

   VectorField getFieldExt()const;
      // returns the field got by adding all contributions from Env and
      // all external systems

   VectorField getFieldPar(bool mean=false)const;
      // returns the field got by adding the contributions of all
      // particles and the mirror charges associated with them
      // (mirror charges where not included here till 03-01-13)
      // Forming the sum is done on the level of fields so that
      // much stress is put on the stack.
      // If the argument mean is true, the field meanFieldPar
      // is returned. This may differ from the field for
      // mean==false due to the periodic replicas.

   VectorField getFieldTot(bool mean=false)const;
      // return the sum of the previously considered fields

   void markPot(const Camera& cam)const;
      // shows on sceen (or only in cam's image memory; and after
      // calling cam.vis() on screen) the potential
      // distribution in cam's image plane. What subsystems are included
      // in the potential is controled by
      // SysControl::showPartFields
      // SysControl::showMeanFil
      // SysControl::ignoreExternalSystems
      // SysControl::noInteractionChargeToWall

   void markField(const Camera& cam)const;
      // same logic as markPot(Camera)

   void markPotTot(const Camera& cam)const;

   void markPotExt(const Camera& cam)const;
      // same data types used as in previous function. For uniformity
      // with previous function

   void markPotPar(const Camera& cam)const;
      // before marking on screen, the particle fields will
      // be added as instances of DisField (a discrete version
      // of ScalarField) so that the number of fields that can be
      // is not limited by the stack.
      // The discrete set of points for the definition of DisField
      // is provided by the argument cam

   void markCharge(const Camera& cam)const;
      // shows the charge density derived from charges sorted
      // into a rectangular grid on the field of view of the camera

   void markScalarValue(const Word& w, const Camera& sc)const;
      // similar to charge density, for other properties indivated by w

   ScalarField getPotExt()const;
      // sum of ext(j).getU()

   ScalarField getPotPar(bool mean=false)const;
      // sum of the fields potentialPar(par[i]) Notice that this includes
      // the  contributions of mirror charges.
      // For mean=true, this is the
      // potential generated by particles in the form as
      // the actual dynamics (in function steplet) sees it
      // (actually, the dynamics only sees vector field
      // meanFieldPar. However, meanPotPar is built closely
      // parallel to meanFieldPar that it is obvious that it
      // actually is the potential of meanFieldPar.
      // Notice that in the present version of the program
      // gravitation is included and that the ScalarField returned
      // from this function is the gravitation potential in case
      // SysControl::noInteractionMassToMass==0.

   ScalarField getPotTot(bool mean=false)const;
      // sum of getPotExt() and getPotPar()

   R dis(const Type& )const;
      //: distance
      // Comparison function on which the step adjustment in step() is
      // based. This is not limited by 1 as in my numerical classes. See
      // the defining code for the meaning.

   V<R> dist(const Type& )const;
      // Returns distance values as individual values for the the
      // subsystems and not only the maximum over all.

   virtual void mark(Camera const& cam, bool display,
      Word const& infoText=Word(), bool fileOnShow=false)const;
      // visual representation of the configuration on which movie
      // production is based. What is shown (e.g. vector fields or
      // potentials) is controlled by
      // SysControl::bitmapToShow
      // SysControl::showParticles
      // SysControl::showMeanFil
      // SysControl::ignoreExternalSystems
      // SysControl::showPartition.
      // for details see implementation code
      // If the last argument has the value true, the screen image will
      // be written as a ppm file to disk.

   virtual bool file(const Word& fn, const Timer& tm)const
      {
         cpmwarning("Trivial version CPM_FCN::file(...) called");
         fn; tm;
         return false;
      }
      // Intended to writes the state into a suitable file, which should
      // be made visible by a suitable viewer program on a
      // single computer. Since the number of particles to be
      // considered is huge, it is rather evident that one should
      // file not too much redundant data. This suggests to divide
      // the data of an object in constant ones and those that change
      // with time. This is presently implemented in a non-trivial manner
      // only in a derived class Sys2.

   virtual X2<Z,Word> docuStep(OFileStream&)const
      { return X2<Z,Word>(0,"not useful");}
      // a function to be called in function run(....) just after step()
      // in order to allow problem dependent information gathering
      // In a useful implementation, the return value gives the number of
      // R's that are written to file in sequence in one call of the
      // function.
      // If active==false, the function will return this number without
      // acting on the argument stream.
};

CPM_TAL
void CPM_FCN::setTimesFromExtSys()
{
   if (ne==0) return;
   R sum=0;
   Z i;
   for (i=1;i<=ne; i++) sum+=ext(i).get_t();
   R tMean=sum/ne;
   for (i=1;i<=ne; i++) ext(i).set_t(tMean);
   stpCon.t=tMean;
}

CPM_TAL
void CPM_FCN::set_t(R t0)
{
   stpCon.t=t0;
   Z i;
   for (i=1;i<=ne; i++) ext(i).set_t(t0);
   for (i=1;i<=np; i++) par[i].set_t(t0);
      // par[i]-loop was missing till 02-06-28 and spoiled the
      // start of ctc7c run3
}

CPM_TAL
bool CPM_FCN::prnOn(ostream& str)const
{
   Z mL=1;
   static Word loc("CPM_FCN::prnOn(...)");
   CPM_MA
   cpmwt("ParDynSys<>");
   cpmp(stpCon);
   cpmp(np);
   cpmp(ne);
   Z i;
   R npInv=cpminv((R)np);
   Z count=0;
   Z countMax=50;
   for (i=1;i<=np;i++){
      count++;
      if (count==countMax){
         count=0;
         cpmmessage(mL,"fraction done: "&cpm(R(i)*npInv));
      }
      cpmp(par[i]);
   }
   V<Word> tL(ne);
   for (i=1;i<=ne;i++){
      Word ti=ext(i).nameOf();
      tL[i]=ti;
      cpmmessage("type of ext("&cpm(i)&") is "&ti);
   }
   for (i=1;i<=ne;i++) cpmp(ext(i));
   cpmp(partition);
   cpmwt("ParDynSys<> end");
   CPM_MZ
   return true;
}

CPM_TAL
bool CPM_FCN::scanFrom(istream& str)
{
   Z mL=1;
   static Word loc("CPM_FCN::scanFrom(...)");
   CPM_MA
   cpms(stpCon);
   cpms(np);
   cpms(ne);
   Z i;
   par=V<Par>(np);
   // needed since the calling instance is not assumed to have the
   // correct number of particles already.
   for (i=1;i<=np;i++) cpms(par[i]);
   cpmmessage(mL,"reading "&cpm(np)&" particles ready");

   // ext=Vp<Ext>(ne); is missing since we assume that the calling
   // instance already has right number and types of ext(i).
   // Introducing the above statement would even be wrong since it would
   // give the external system the type ExtSys and thus would spoil the
   // additional attributes of derived classes.

   cpmmessage(mL,"trying to read "&cpm(ne)&" external systems");
   for (i=1;i<=ne;i++){
      cpmmessage(mL,"reading external system "&cpm(i)&" ahead");
      cpms(ext(i));
      cpmmessage(mL,"reading external system "&cpm(i)&" ready");
   }
   cpmmessage(mL,"reading external systems ready");
   cpms(partition);
   cpmmessage(mL,"reading partition ready");
   CPM_MZ
   return true;
}

CPM_TAL
void CPM_FCN::operator*=(const Group& g)
{
   Z i;
   for (i=1;i<=np;i++) par[i]*=g;
   for (i=1;i<=ne;i++) ext(i)*=g;
}

CPM_TAL
void CPM_FCN::ranRotPar(R angleRange, Z j)
{
   Group g(Angle(angleRange,DEG));
   for (Z i=1;i<=np;i++){
      Spc xi=par[i].get_x();
      Group gi=g.ran(xi,++j);
      par[i]*=gi;
   }
}

CPM_TAL
R CPM_FCN::dis(const Type& hs)const
{
   Z i;
   R_Vector vp(np);
   for (i=1;i<=np;i++)
      vp[i]=par[i].dis((hs.par)[i]);
   R res_p=vp.abs();

   R_Vector ve(ne,0.);
   for (i=1;i<=ne;i++)
      if (ext(i).isActive()) ve[i]=ext(i).dis((hs.ext)(i));
   R res_e=ve.abs();

   R2 res2(res_p,res_e);
   R res=res2.abs();
   if (res==0.){ cpmwarning(
      "ParDynSys<>::dis=0; parabolic motion or numerical problem");
   }
   return res;
}

CPM_TAL
V<R> CPM_FCN::dist(const Type& hs)const
{
   V<R> res(np,0);
   for (Z i=1;i<=np;i++) res[i]=par[i].dis((hs.par)[i]);
   return res;
}

CPM_TAL
R CPM_FCN::get_m(const V<B>& vb)const
{

   R res=0;
   for (Z i=1;i<=np;i++){
      if (vb[i]) res+=par[i].get_m();
   }
   return res;
}

CPM_TAL
R CPM_FCN::get_m()const
{
   R res=0;
   for (Z i=1;i<=np;i++) res+=par[i].get_m();
   return res;
}

CPM_TAL
Spc CPM_FCN::cenOfMas(const V<B>& vb)const
{
   V<Spc> x(np);
   V<R>  w(np);
   for (Z i=1;i<=np;i++){
      x[i]=par[i].get_x();
      w[i]= vb[i] ? par[i].get_m() : 0. ;
   }
   Spc res;
   res.convexAsMember(x,w,true);
   return res;
}

CPM_TAL
R CPM_FCN::kinEnr(const V<B>& vb)const
{
   R res=0;
   for (Z i=1;i<=np;i++){
      if (vb[i]) res+=par[i].kinEnr();
   }
   return res;
}

CPM_TAL
R CPM_FCN::potEnr(const V<B>& vb)const
{
   R res=0;
   Z i,j;
   for (i=1;i<=np;i++){
      if (!vb[i]) continue;
      for (j=1;j<i;j++){
         if (!vb[i]) continue;
         res+=par[i].pot(par[j]);
      }
   }
   return res;
}

CPM_TAL
R CPM_FCN::kinEnr()const
{
   R res=0;
   for (Z i=1;i<=np;i++){
      if (par[i].getOwner()==rank) res+=par[i].kinEnr();
   }
   return res;
}

CPM_TAL
R CPM_FCN::potEnr()const
{
   R res=0;
   Z i,j;
   for (i=1;i<=np;i++){
      bool bi=(par[i].getOwner()==rank);
      if (bi){
         for (j=1;j<i;j++) res+=par[i].pot(par[j]);
      }
   }
   return res;
}

CPM_TAL
R CPM_FCN::potElasticExt()const
{
   R res=0;
   Z i,j;
   for (i=1;i<=np;i++){
      bool bi=(par[i].getOwner()==rank);
      if (bi){
         for (j=1;j<=ne;j++) res+=ext(j).potElastic(par[i]);
      }
   }
   return res;
}

CPM_TAL
R CPM_FCN::totEnr()const
{
   R res=kinEnr()+potEnr()+potElasticExt();
   for (Z i=1;i<=np;++i) res+=getExt()(1).potField(getPar(i));
   return res;
}

CPM_TAL
For CPM_FCN::force(const Par& par1, const Par& par2)const
{
   For res=par1.forces(par2);
      // direct force, vanishes for par1==par2
      // Since 2004-05-09 the dcision of equaliy is based on
      // id rather than on distance
   if (!SysControl::noInteractionChargeToWall){
      PointCharge q1=par1.getPointCharge();
         // feels the force
      PointCharge q2=par2.getPointCharge();
         // generates mirror charges
      for (Z j=1;j<=ne;j++){
         bool bj=ext(j).isActive();
         if (!bj) continue;
         R dj=ext(j).dist(q2.get_x());
         if (dj<=nearLength){
            V<PointCharge> extinfl=ext(j).influence(q2);
            res.addForce(q1.force(extinfl));
         }
      }
   }
   return res;
}

CPM_TAL
FieldSpace CPM_FCN::field(const Spc& sp, const Par& par1,
   const bool& farAway)const
{
   bool nogravity=(SysControl::noInteractionMassToMass!=0);
      // we don't want the particles to create gravitational fields
      // if masses don't interact

   FieldSpace res=(par1.getField(nogravity))(sp);
   if (!SysControl::noInteractionChargeToWall){
      PointCharge q=par1.getPointCharge();
      for (Z j=1;j<=ne;j++){
         V<PointCharge> extinfl=ext(j).influence(q);
         Z eid=extinfl.dim();
         if (eid==0) continue;
         else if(farAway){ // treat the mirror charges from one source
            // as placed at the same position of the first one
            PointCharge pc=extinfl[1];
            for (Z k=2;k<=eid;k++) pc.add(extinfl[k]);
            res.c1()+=PointCharge::electrostaticField(sp,pc);
         }
         else{
            res.c1()+=PointCharge::electrostaticFieldV(sp,extinfl);
         }
      }
   }
   return res;
}

CPM_TAL
R CPM_FCN::potential( Spc const& sp, Par const& par1,
   bool const& farAway)const
{
   R res=par1.getU()(sp);
   if (SysControl::potentialToShow==1 &&
      !SysControl::noInteractionChargeToWall){
      PointCharge q=par1.getPointCharge();
      for (Z j=1;j<=ne;j++){
         V<PointCharge> extinfl=ext(j).influence(q);
         Z eid=extinfl.dim();
         if (eid==0) continue;
         else if(farAway){ // treat the mirror charges from one source
            // as placed at the same position of the first one
            PointCharge pc=extinfl[1];
            for (Z k=2;k<=eid;k++) pc.add(extinfl[k]);
            res+=PointCharge::electrostaticPotential(sp,pc);
         }
         else{
            res+=PointCharge::electrostaticPotentialV(sp,extinfl);
         }
      }
   }
   return res;
}

////////////////////////// dynamics stuff //////////////////////////////

CPM_TAL
set<Z> CPM_FCN::nextPointIndexes(const Spc& x, R range)
{
   set<Z> res;
   Z pld=pointList.dim();
   for (Z j=1;j<=pld;++j){
      Spc xj=pointList[j];
      R dj=(x-xj).abs();
      if (dj<=range){
         res.insert(j);
      }
   }
   return res;
}

CPM_TAL
void CPM_FCN::prepareStep(const Camera& sc)
// to be called after redistribute, so np is the
// number of particles which are owned or used
{
   const Z mL=2;
      // normal message limit, proposal 2

   const Z mL2=4;
      // message limit for graphical activity, proposal 4

   static Word loc("CPM_FCN::prepareStep()");

   if (SysControl::noMeanFieldPar) return;
      // then all forces come directly from particles and
      // external systems.

   cpmmessage(mL,loc&" started");

   sizeAndRank();
   pointList=partition.getPoints(rank);
   Z pld=pointList.dim();
   Z i,j,k,l;
   xPar=V<Spc>(np);
   for (i=1;i<=np;i++) xPar[i]=par[i].get_x();
   nearLength=partition.getRes(rank);
   cpmmessage(mL,loc&": nearLength="&cpm(nearLength),-1);
   nearLength*=SysControl::nearFactor;
   cpmmessage(mL,loc&": nearLength after modification="
      &cpm(nearLength),-1);

   R covRad=partition.getCovRad(rank);
   R rAct=covRad+nearLength;
   vector< set<Z> > parListNeg(pld);
      // parListNeg[i] is a set {l[1],...l[n]} of particle
      // indexes such that
      // Par[l[k]], k=1,...n are the particles such that
      // | Par[l[k]].get_x()- poinList[i]|<= rAct.
      // The intention behind the choice of rAct is to implement
      // my algorithm (of 2004-05-07) that finds for each
      // x \in X all particles p for which
      // |p.get_x()-x|<= nearLength
   V< Z > nearPoint(np);
   for (i=1;i<=np;++i){
      Spc xi=xPar[i];
      R dMin=SysControl::intactRange;
      Z jMin=0;
      for (j=1;j<=pld;j++){ // save iteration
         // allows to use unchecked indexing for parListNeg
         Spc xj=pointList[j];
         R dij=(xi-xj).abs();
         if (dij<dMin){
            jMin=j;
            dMin=dij;
         }
         if (dij<=rAct) parListNeg[j-1].insert(i);
      }
      nearPoint[i]=jMin;
   }
   cpmmessage(mL,"particle loop for parListNeg done");

   R npInv=(np==0 ? 0 : 1./np);
   nearList=vector< set<Z> >(np);
   for (i=1;i<=np;++i){
      if (cpmverbose>=mL2) cpmprogress("nearList",R(i-1)*npInv);
      Spc xi=xPar[i];
      set<Z> neg=nextPointIndexes(xi,covRad);
      set<Z>::const_iterator k,l;
      for (k=neg.begin();k!=neg.end();++k){
         set<Z> negk=parListNeg[*k-1];
         for (l=negk.begin();l!=negk.end();++l){
            Spc xl=xPar[*l];
            R dil=(xi-xl).abs();
            if (dil<=nearLength) nearList[i-1].insert(*l);
         }
      }
   }
   cpmmessage(mL,"particle loop for nearList done");

// We represent the electro-magnetic field of our np particles
// by computing suitable sources located at the few points of pointList.
// Aha! (2003-07-08) only owned particles should contribute to
// meanFieldPar, since the other ones are taken into account by other
// processes so that double-counting would result.

   Z npOwned=0;
   for (i=1;i<=np;i++) if (par[i].getOwner()==rank) ++npOwned;
   V<SourceOfVectorField> vsvf(npOwned);
   V<Z> whereToGo(npOwned);
   Z iOwned=0;
   for (i=1;i<=np;i++){
      if (par[i].getOwner()==rank){
         ++iOwned;
         vsvf[iOwned]=par[i].source();
         whereToGo[iOwned]=nearPoint[i];
      }
   }
   vsvf=SourceOfVectorField::redistribute2(vsvf,pointList,whereToGo);
      // this function does not neglect contributions of particles.
      // since each source is taken and only shifted.
      // See the declaration of
      // SourceOfVectorField::redistribute(...) in file cpmforcesx.h
      // for the meaning of this function call.
      // now we have a much smaller list sources than before since
      // typically pointList.dim() is much smaller than np
   Z nk=vsvf.dim();
   if (cpmverbose>=mL2){ // to see what we got
      // no longer very useful since the nean field can now be
      // represented
      // more conveniently via function mark(...)
      for (i=1;i<=nk;i++){
         vsvf[i].field().mark(sc);
         sc.vis();
         sc.clear();
      }
   }
// it is economical to redistribute the actual sources before computing
// the mirror-charges. Notice that the mirrorcharges will never be
// redistributed so that we don't need to extend the grid into the area
// the mirror charges are located.

   if (!SysControl::noInteractionChargeToWall){
      // The point charges in vsvf should be augmented by the charges
      // which they influence in the external systems.
      for (j=1;j<=pld;j++){
         PointCharge pc=vsvf[j].getCharge();
         if (pc.get_q()==0.) continue;
            // this happens if there are points in the point list in the
            // neighborhood of which there are no particles. If these
            // charges are 0, we have not to care about mirror charges
         for (k=1;k<=ne;k++){
            V<PointCharge> pck=ext(k).influence(pc);
            Z pckd=pck.dim();
            if (pckd>0){
               PointCharge pcks=pck[1];
               for (l=2;l<=pckd;l++) pcks.add(pck[l]);
                  // it is certaily allowed to pack the induced
                  // charges (from same source charge and same external
                  // system !) together
               vsvf&=SourceOfVectorField(pcks); // array may become
               // longer!
            }
         } // k loop
      } // j loop
   }
   // now we may have much more sources than those at points of
   // pointList

   nk=vsvf.dim();

// Now we pack the information on sources in a manner
// suitable to sending to other processes.
   typedef  V<SourceOfVectorField> mesType;
   mesType valMy(vsvf);
// We should be prepared to receive quantities of type mesType from all
// other processes. So we have to have an array to receive these
// objects.
   V<mesType> valAll(size,mesType(V<SourceOfVectorField>(0)));
      // do not initialize vectors by default constructor
      // since these then have already one component which is
      // not what we want; was wrong till 02-04-04
   Cpm_com.sendAndRec(valMy,valAll);
      // responsibility for a consistent pairing of sends and
      // receives is now with class Com.
      // Now the present process has available the sources that create
      // the far field in the neighbor areas. Unlike the situation
      // for particles, these far field sources do not correspond only
      // to the physics happening in the overlap zone between the present
      // process and the neighbors but in the whole neighbor area.

   cpmmessage(mL,"communication done");

// initializing repPointList if not yet done

   if (!meanFieldAct){
      Z n2=partition.repNegDim();
      repPointList=VV<Spc>(pld,n2);
      for (i=1;i<=pld;i++){
         Spc pi=pointList[i];
         V<Spc> ppi=partition.repNeg(pi); // Note ppi[1]==pi
         for (j=1;j<=n2;j++) repPointList[i][j]=ppi[j];
      }
   }

   cpmmessage(mL,"defining meanField");

// now we have evaluate the contributions of all known sources
// not only those centered at some pointList[i] on the grid
// given by pointList.

   V<FieldSpace> filValList(pld);
   V<R> potValList(pld);

   for (i=1;i<=pld;i++){ // loop over the points where the
         // field values have to be defined
      FieldSpace vali; // starting with 0
      R valUi=0;
      Z n2=repPointList.dim2();
      for (j=1;j<=n2;j++){
         Spc xij=repPointList[i][j];
         // the virtual repetition copies for toroidal topology are taken
         // here
         for (k=1;k<=nk;k++){
            if (k!=i || j!=1){ // why j!=1
               vali+=vsvf[k].eval(xij);
               if (SysControl::showPartFields) valUi+=vsvf[k].evalU(xij);
            }
            // this excludes only the direct source at xi1
            // mirror charges from there are included, since these
            // are of the form vsvf[k] for k>pld.
            // adding fields from all sources (in rank) appart
            // the ones at xi1. Notice that it is natural to neglect
            // contributions from some sources since the nearest
            // neighbors of a particle are taken into account as direct
            // interaction partners and thus would be countet twice if
            // they would enter meanFieldPar. A reasonable definition of
            // 'nearLength' has made in the beginning aiming at making the
            // volume where sources are neglected equal to the volume in
            // with particles interact directly.
            // This is a canonical (though not trivial)
            // method of separating near field and far field.
            // This is what I call the 'self-consistent far field method'
            // In short the logic is as follows:
            // To extract a everywhere smooth field from a cloud of
            // point sources:
            // (i): chooce a lattice of points that will determine
            //    the resolution of the smooth field to be created
            // (ii) collect all sources at the lattice points by
            //    moving each source to the nearest lattice point
            // (iii) Compute the field strength in all lattice points
            //    by neglecting in each point just the contribution
            //    of the sources in just that point (which would be
            //    singular).
            // (iv) continue the field (not the sources!) from the
            //    lattice points to the whole space by whatever
            // reasonable interpolation method.
         }
      }
      filValList[i]=vali;
      if (SysControl::showPartFields) potValList[i]=valUi;
         // does neither know of other processes nor of
         // repetition
   }
   cpmmessage(mL,"defining meanField: first loop done");
// now the contribution from neighbor processes.
// No repetition copies here!

   for (i=1;i<=pld;i++){ // loop over the points where the
         // field values have to be defined
      FieldSpace exVali; // starting with 0
      R exValUi=0; // starting with 0
      Spc xi=pointList[i];
         // position where to compute the field values
         // no sum over repetition points
      for (j=1;j<=size;j++){ // loop over the contributing
            // neigbor processes
         if (j==rank) continue;
         Z nj=valAll[j].dim();
         for (k=1;k<=nj;k++){
            exVali+=valAll[j][k].eval(xi);
            if (SysControl::showPartFields)
               exValUi+=valAll[j][k].evalU(xi);
         }
      }
      filValList[i]+=exVali;
      if (SysControl::showPartFields) potValList[i]+=exValUi;
   }
   cpmmessage(mL,"defining meanField: second loop done");
   meanFieldPar=VectorField(pointList,filValList);
      // this is an interpolation constructor

   if (SysControl::showPartFields)
      meanPotPar=ScalarField(pointList,potValList);
   // meanPotPar is only needed for visualization and debugging,
   // so it should not be computed in serious production runs
   // These have to be done with
   // SysControl::showPartFields==false
   meanFieldAct=true;
   cpmmessage(mL,loc&" done");
}

///////////////////////// main dynamics stuff //////////////////////////

CPM_TAL
void CPM_FCN::steplet(R t0, R dt, const Camera& cam)
{
   const Z mL=2; // 2 usual
   const Z mL2=3;
   static Word loc("CPM_FCN::steplet(...)");
   CPM_MA
   static R dtMem=0;
   static Z npMem=0;
   cam; // presently not used
   if (dt==0.){ // this should be efficient
      cpmmessage(mL,loc&" trivially done");
      return;
   }
   bool b=false;
   if (dt!=dtMem){
      b=true;
      dtMem=dt;
   }
   if (np!=npMem){
      b=true;
      npMem=np;
   }
   if (b) cpmvalues2("steplet: dt, np",dt,np,2);

   cpmmessage(mL,"steplet(...), dt="&cpm(dt)&" np="&cpm(np));

   bool parValid=Par::isVal(par);
      // soundness check for particle array

   cpmassert(parValid==true,loc);
      // allows to stop by setting cpmdbg=2 (cpmdbg=1 is normal)

// linear evolution to the midpoint in time: direct midpoint method
// Linear means here that the velocities (linear velocity and angular
// velocity) are constant. Notice that constant angular velocity for
// a rigid body normally does not mean free motion.
   Z i,j;
   R tau=dt*0.5;
   R tc=t0+tau;
      // t center
   R tf=t0+dt;

   cpmassert(par.dim()==np,loc);

   for (i=1;i<=np;i++){
     par[i].freeMot(tc);
     par[i].phase1(tau);
   }

   V<R> te(ne,R(0));
   for (j=1;j<=ne;j++){
      te[j]=ext(j).get_t();
      ext(j).clear();
      ext(j).freeMot(tc);
         // see comment to ext(j).freeMot(tf)
   }

   R npInv=(np>0 ? R(1)/R(np) : 0);

// collecting the interaction terms at the mid-point in time
// don't move particles yet, otherwise dynamics would depend on
// numbering of particles

   V<For> forces(np);
      // forces[i]=force acting on particle par[i]
   V<FieldSpace> fValues(np);
   //fValues[i]=field values at place of particle par[i]

   for (i=1;i<=np;i++){ // first particle loop
      if (cpmverbose>=mL2) cpmprogress("force loop",R(i-1)*npInv);
      Par pi=par[i];
      For fi;
// forces by general fields
      if (SysControl::eta!=0 && ne>0 && !SysControl::noFriction)
       fi+=pi.forces(ext(1));
         // 'air friction' in the reference system of ext[1].
         // Here ext(i) enters through the base class RigidObject
      if (!SysControl::ignoreExternalSystems){
         for (j=1;j<=ne;j++){
            if (ext(j).isActive()){
              // cpmmessage("pi.forces(ext(j).getField()) reached");
               fi+=pi.forces(ext(j).getField());
            }
          }
      }
// forces by particle-generated fields
      FieldSpace fvi; // zero field
      Spc xi=pi.get_x();
      if (!SysControl::noMeanFieldPar){
         fvi=meanFieldPar(xi);
         fi+=pi.forces(fvi); //(***)
         // all influence of fvi on velocities are thus taken into
         // account; we need fvi later also as a quantity that
         // causes the electric polarization of pi to change
         // (see call of function react).
      }
// forces by interacting with particles directly
// including self-interaction:
// force on pi created by the mirror charges created by
// pi itself in external systems. Proper dependence on
// SysControl::noInteractionChargeToWall and ext(j).isActive()
// is assured by the implementation of function force()

      if (SysControl::noMeanFieldPar){ // forces from all particles
         for (j=1;j<=np;++j){
            fi+=force(pi,par[j]);
         }
      }
      else{ // forces from near particles

         set<Z>::const_iterator js;
         for (js=nearList[i-1].begin();js!=nearList[i-1].end();++js){
            fi+=force(pi,par[*js]);
            // includes forces from mirror charges, member of CPM_TAL, not
            // of Par. According to the new logic we have to include
            // mirror charges here only if we are near to a wall since
            // filValList includes all mirrorcharges on the average
         }
      }
      if (!SysControl::ignoreExternalSystems){
         for (j=1;j<=ne;j++){
            if (ext(j).fieldOnly()) continue;
            fi+=ext(j).interaction(pi,dt);
         }
      }
      forces[i]=fi;
      fValues[i]=fvi;
   } // end of first particle loop

// moving the particles according to interaction

   for (i=1;i<=np;i++){ // second particle loop
      par[i].intAct(dt,forces[i],fValues[i]);
         // Changes velocities according to forces.
         // For Par=Clust (as in class Station) this call includes
         // building up polarization of the particle through the
         // electric component of fValues[i]. See
         // Clust::intAct(R, const ForTor&, const FieldSpace&)
         // and void Clust::setElcDipole(R dt, const VecE& E);
      par[i].freeMot(tf);
         // executes the shift in position and angular orientation
         // according to the full acceleration and angular acceleration
         // that the particle feels (no regularization) at this place.
      par[i].reg();
         // regularization of the motion by reducing the velocity
         // if this is much larger than SysControl::vCrit
   } // end of second particle loop

   if (!SysControl::ignoreExternalSystems){
      for (j=1;j<=ne;j++) ext(j).freeMot(tf);
        // linear evolution at the end. Also
        // inactive systems should do this since
        // they should be in the right place if reactivated
   }
   CPM_MZ
}

CPM_TAL
void CPM_FCN::step(R dt, const Camera& cam)
{
   const Z mL=2;
   Z fS=SysControl::fixedStep;
   SysControl::tStep=dt;

   static Z stepCount=0;
   R tcStart=0,tcPrepareDone=0,tcFinal=0;
      // tc : computation time
   R tmStart=get_t(),tmFinal=0;
     // tm : modeling time = image of real time in our model
   Z nStart=get_np(), nFinal=0;

   if (cpmverbose>0){
      tcStart=cpmtime();
   }
   cpmmessage(mL,"CPM_FCN::step("&cpm(dt)&") started");
   redistribute();
   prepareStep(cam); // was missing till 02-06-14
      // we need to take into account that interaction with other
      // processes at the end of step changed everything in the present
      // process (e.g. the particle number) in a manner that could not be
      // predicted within the given process. We also would not rely on
      // a constructor to ensure the preconditions for function step().

   if (cpmverbose>0){
      tcPrepareDone=cpmtime();
   }

   if (nStart==0){
      cpmmessage("step() started with no particles");
      steplet(tmStart,dt,cam);
      done(dt,0);
   }
   else{
      R dt_fS=dt/fS;
      for (Z j=1;j<=fS;j++){
         if (fS>1 && cpmverbose>1){
            R val=(j-1.)/(fS-1.);
            cpmvalue("frac",val,4);
         }
         R t=get_t();
      //   R tf=t+dt_fS;
         steplet(t,dt_fS,cam);
         done(dt_fS,0); // faster than done(0)
      }
      if (getRelaxFlag()) stopParticles();
   }

   tcFinal=cpmtime();
   tmFinal=get_t();
   set_t(tmFinal); // to be sure that all internal times are equal
   if (cpmverbose>1){ // limit was 0 till 2005-09-19, however under Linux
      // writing long messages to file causes irregular pauses in the
      // video output which now can be avoided by setting cpmverbose to 1
      stepCount++;
      tcFinal=cpmtime();
      nFinal=get_np();
      ostringstream ost;
      ost<<endl<<"Performance data for step # "<<stepCount<<endl;
      ost<<"n: particle number: prior to step "
         <<nStart<<", after step "<<nFinal<<endl;
      R tcTotal=tcFinal-tcStart;
      R tcTotalInv=tcTotal>0 ? R(1)/tcTotal : R(0);
         // due to the limited resolution of the internal clock
         // tcTotal may well come out as 0, this is not to be
         // considered an situation to warn
      R tcPre=tcPrepareDone-tcStart;
      R percentPrepare=tcPre*tcTotalInv*100;
      R tmTotal=tmFinal-tmStart;
      ost<<"Point in model time at beginning of step: "<<tmStart<<endl;
      ost<<"tc: computation time/s for step: "<<tcTotal<<endl;
      ost<<"actual model time/s for step: "<<tmTotal<<endl;
      ost<<"tm: scheduled model time/s for step: "<<dt<<endl;

      if (tcTotal>0){
         ost<<"preparation phase in percent: "<<percentPrepare<<endl;
         R eff=(tmTotal*nStart)*tcTotalInv;
         ost<<"efficiency=(tm*n)/tc= "<<eff;
      }
      else{
         ost<<"preparation phase in percent undefined"<<endl;
         ost<<"efficiency undefined since computation time"
            <<" unmeasurably small";
      }
      cpmmessage(1,Word(ost.str()),-1); // don't write to the status bar
   }
   cpmmessage(mL,"CPM_FCN::step("&cpm(dt)&") done");
}

CPM_TAL
bool CPM_FCN::run(R dtStep, Z nSteps,
   Z filingPeriod, Z docuPeriod, Z viewingPeriod, Z stoppingPeriod,
   const Word& movieName, Camera& cam, const Timer& tm,
   const F<R,Z>& f, OFileStream& ofs, bool albuming)
      // was const F<R,V<R> >& f till 02-06-03
      // return value not needed (void does not work with my function
      // objects since 'unknown type' for the MS compiler
      // Presently f does some manipulations of SysControl
      // as a function of time.
{
   static SysControlState sc;
   static Word loc("CPM_FCN::run(...)");
   const Z mL=2;
   CPM_MA
   const Z mL2=3;
   bool earlyStop=false;
   if (nSteps<0 || dtStep<0){
      cpmmessage(mL,loc&" trivially done");
      return earlyStop;
   }
   sc.mem(); // memorizing the state of SysControl since
      // SysControl may be changed by calling f
   CyclicAlarm ca1(filingPeriod,filingPeriod);
   CyclicAlarm ca2(viewingPeriod,viewingPeriod);
   CyclicAlarm ca3(stoppingPeriod,stoppingPeriod);
   CyclicAlarm ca4(docuPeriod,docuPeriod); // introduced 2005-03-24
      // so the next call to operator() will give true
      // unless ...Period<1
   Z messagePeriod=nSteps/100; // don't overload cpmcerr
   CyclicAlarm ca5(messagePeriod,messagePeriod);
   Z count=0;
   bool display=true;
   Word imgTxt("");

   if (ca1()){ // filing (as movie) requested
      earlyStop=file(movieName,tm);
      cpmmessage("frame written at t="&cpm(get_t()));
   }
   if (ca2()){ // viewing is requested
      if (!meanFieldAct) prepareStep(cam);
         // otherwise the fields meanFieldPar and meanPotPar
         // would have zero-values in the first viewing frame that
         // preceeds all calls to step(). Notice that it is extremely
         // useful to get a display even for nSteps==0 since one
         // can so inspect particle-rich configurations for which a
         // step would take a prohibitively long computation time.
      mark(cam,display,imgTxt,albuming);
      cpmwait(SysControl::tWait);
      cam.clr();
   }
   if (nSteps==0 ){
      cpmmessage(mL,loc&" trivially done due to nSteps==0");
      return earlyStop;
   }
   if (earlyStop){
      cpmmessage(mL,loc&" trivially done due to earlyStop");
      return earlyStop;
   }
   if (!meanFieldAct) prepareStep(cam);
   R tFinal=dtStep*nSteps;
   R tFinalInv=R(1)/tFinal;
//   R nStepsInv=R(1)/R(nSteps);
   R tAct=get_t();
   for (Z is=1;is<=nSteps;++is){
      R glob_prog=tAct*tFinalInv;
      cpmprogress("Progress",glob_prog,3,3);
      if (cpmverbose>1||ca5()){
         cpmcerr<<endl<<loc<<" global progress ="<<glob_prog;
         // very important since tasks can be stopped early via time-out
         // should not interfere on the status bar with cpmprogress
      }
      f(tAct);
      cpmmessage(mL2,"vCrit="&cpm(SysControl::vCrit));
      cpmmessage(mL2,"aCrit="&cpm(SysControl::aCrit));
      step(dtStep,cam);
      tAct=get_t();
      count++;
      if (ca3()) stopParticles();
         // this is considered a part of the intended dynamics
         // thus it should preceed documentation
      if (ca4()) docuStep(ofs);
      bool b1=ca1(); // true only when movie-filing takes place
      if (b1){
         earlyStop=file(movieName,tm);
         cpmmessage(1,"**** movie frame written at t="&
            cpm(get_t())&" count="&cpm(count));
      }
      if (ca2()){
         bool b2=b1&&albuming;
         mark(cam,display,imgTxt,b2);
            // a viewed image gets filed as a ppm only if the frame was
            // also written to he movie-file and if albuming is requested
         cpmwait(SysControl::tWait);
         cam.clr();
      }
      if (earlyStop) break;
   }
   sc.act();
      // restoring the state of SysControl which may have been
      // changed by the action of f
   cpmmessage(mL,loc
      &" done, earlyStop="&cpm(earlyStop)
      &", tAct="&cpm(tAct)
      &", tFinal="&cpm(tFinal)
      &", count="&cpm(count)
   );
   return earlyStop;
}

CPM_TAL
Vp<CPM_FCN> CPM_FCN::trj(R dtStep, Z nSteps, Camera& cam,
   Z viewingPeriod)
{
   static Word loc("CPM_FCN::trj(...)");
   const Z mL=2;
   CPM_MA
   bool trivial=nSteps<0 || dtStep<0;
   Z resDim=trivial ? 0 : nSteps+1;
   Vp<CPM_FCN> res(resDim);
   if (trivial){
      cpmmessage(mL,loc&" trivially done");
      return res;
   }
   CyclicAlarm ca(viewingPeriod,viewingPeriod);
   if (ca()){ // viewing is requested
      if (!meanFieldAct) prepareStep(cam);
         // otherwise the fields meanFieldPar and meanPotPar
         // would have zero-values in the first viewing frame that
         // preceeds all calls to step(). Notice that it is extremely
         // useful to get a display even for nSteps==0 since one
         // can so inspect particle-rich configurations for which a
         // step would take a prohibitively long computation time.
      mark(cam,true);
      cpmwait(SysControl::tWait);
   }
   if (nSteps==0 ){
      cpmmessage(mL,loc&" trivially done due to nSteps==0");
      return res;
   }
   if (!meanFieldAct) prepareStep(cam);
   R prog=0;
   cpmprogress("Progress",prog,3,3);
   R dprog=1./nSteps;
   res[1]=*this;
   for (Z i=1;i<=nSteps;++i){
      step(dtStep,cam);
      res[i+1]=*this;
      prog+=dprog;
      cpmprogress("Progress",prog,3,3);
      if (ca()){
         mark(cam,true);
         cpmwait(SysControl::tWait);
      }
   }
   CPM_MZ
   return res;
}

CPM_TAL
V< X2<V<Spc>,R> > CPM_FCN::xParTrj(R dtStep, Z nSteps)const
{
   static Word loc("CPM_FCN::xParTrj(...)");
   const Z mL=2;
   CPM_MA
   bool trivial=nSteps<0 || dtStep<0;
   Z resDim=trivial ? 0 : nSteps+1;
   V< X2<V<Spc>,R> > res(resDim);
   if (trivial){
      cpmmessage(mL,loc&" trivially done");
      return res;
   }
   CPM_FCN tmp=*this;
   R prog=0;
   cpmprogress("Progress",prog,3,3);
   R dprog=1./nSteps;
   res[1]=X2<V<Spc>,R>(tmp.posPar(),tmp.get_t());
   for (Z i=1;i<=nSteps;++i){
      tmp.step(dtStep);
      res[i+1]=X2<V<Spc>,R>(tmp.posPar(),tmp.get_t());
      prog+=dprog;
      cpmprogress("Progress",prog,3,3);
   }
   CPM_MZ
   return res;
}

CPM_TAL
void CPM_FCN::movieFromTrj(Vp<CPM_FCN> const& trj, Word const& fileName,
   Camera& cam, Z viewingPeriod, bool showOnly)
{
   Timer tm;
   CyclicAlarm ca(viewingPeriod,viewingPeriod);
   Z i,n=trj.dim();
   Pp<CPM_FCN> sMem(*this);
   Word header("Progress");
   Z seg=3;
   Z acc=3;
   R prog=0;
   cpmprogress(header,prog,seg,acc);
   R incr=1./n;
   for (i=1;i<=n;++i){
      *this=trj(i);
      if (!showOnly) file(fileName,tm);
         // I presently don't understand why
         // trj(i).file(fileName,tm);
         // does not work; why the detour over *this is needed
         // (notice that only derived classes implement function
         // file(...)
         // non-trivially
      prog+=incr;
      cpmprogress(header,prog,seg,acc);
      if (ca()){
         cam.clear();
         mark(cam,true);
         cpmwait(SysControl::tWait);
      }
   }
   *this=sMem();
}

CPM_TAL
T2<R_Vector> CPM_FCN::anlExc(R dtStep, Z nSteps, Z substeps, R fac,
                         bool logView)const
{
   Z mL=2;
   Word loc("CPM_FCN::anlExc(...)");
   CPM_MA
   cpmassert(dtStep>0,loc);
   cpmassert(nSteps>=1,loc);
   Z fsMem=SysControl::fixedStep; // memorizing, will be restored at the
      // end
   cpmassert(fsMem>=1,loc);

   Z fixedStep1=cpmrnd(fac*substeps);
   Z fixedStep2=substeps;

   cpmdebug(fixedStep1);
   cpmdebug(fixedStep2);

   SysControl::fixedStep=fixedStep1;
   V< X2<V<Spc>,R> > vv1=xParTrj(dtStep,nSteps);

   SysControl::fixedStep=fixedStep2;
   V< X2<V<Spc>,R> > vv2=xParTrj(dtStep,nSteps);

   R_Vector xs(nSteps);
   R_Vector ys(nSteps);
   Z npz=np;
   R nI=R(1)/R(npz);
   for (Z i=1;i<=nSteps;++i) {
      X2<V<Spc>,R> vi=vv1[i];
      X2<V<Spc>,R> wi=vv2[i];
      R ri=0.;
      for (Z j=1;j<=np;++j){ // loop over particles
         ri+=vi.c1()[j].distance(wi.c1()[j]);
      }
      ri*=nI; // mean contribution of one particle
      ys[i]=logView ? aux::log10s(ri): ri;
      xs[i]=vi.c2(); // time that belongs to this contribution
   }
   SysControl::fixedStep=fsMem; // restauration
   CPM_MZ
   return T2<R_Vector>(xs,ys); // returning two R_Vectors
}

CPM_TAL
R_Vector CPM_FCN::anlOrd(R dtStep, Z nSteps, Z nFrames, R fac,
   Word const& runId, Camera& cam, Iv const& iv, Z repLog,
   R repOrderD, R repOrderE, R repOrderL, bool makeMovie,
   bool storeGraphAsImage, Vec const& dir)
{
   Z mL=2;
   Word loc("CPM_FCN::anlOrd(...)");
   CPM_MA
   cpmassert(dtStep>0,loc);
   cpmassert(nFrames>=2,loc);
   cpmassert(nSteps>=1,loc);
   Z fsMem=SysControl::fixedStep;
   cpmassert(fsMem>=1,loc);

   Vp<CPM_FCN> finalStates(nFrames);
   Pp<CPM_FCN> sMem(*this);
   Z i,j;
   V<Z> scfs(nFrames); // scfs: SystemControl::fixedStep
   scfs[1]=fsMem;
   R fi=fac;
   for (i=2;i<=nFrames;++i){
      Z scfsi=cpmrnd(R(fsMem)*fi);
      while (scfsi<=scfs[i-1]) scfsi++;
      scfs[i]=scfsi;
      fi*=fac;
   }
   scfs=scfs.rev(); // we want to proceed from the shortest time step
   // to the longest one.
   R_Vector dt(nFrames); // holds the time step for each trajectory
      // which is needed for the computation of the order
   for (i=1;i<=nFrames;++i) dt[i]=dtStep/scfs[i];
   SysControl::fixedStep=scfs[1]; // this is now the highest
      // value from which the following ones are obtained by
      // diminishing.
   Vp<CPM_FCN> vs1=trj(dtStep,nSteps,cam);
      // This is the first of nFrames trajectories. From it the others
      // will be built by reducing SysControl::fixedStep.
      // Each of these trajectories has nSteps states which belong to
      // an increasing sequence of time points.
      // vs1 will be referred to as the reference trajectory. Since it
      // uses the shortest time step of all trajectories under
      // consideration, it is the one which should come closest to the
      // 'exact' trajectory.
   Word header("Total");
   Z seg=4;
   Z acc=2;
   R eps=1e-16;
   R incr=1./nFrames;
   R glob_prog=incr;
   cpmprogress(header,glob_prog,seg,acc);
   Z vsd=vs1.dim();
   finalStates[1]=vs1[vsd];
      // dimension is nSteps+1 since initial state is vs1(1)
      // Notice that vs1 is a polymorphic array for which
      // the most regular component access is by normal brackets.
      // Since however finalStates will be instances of a class
      // derived from CPM_FCN, here the [] are correct.
      // 'Final' refers, of course, to the last time step, i.e.
      // to the final state of the trajectory under consideration.
   cpmassert(vsd==nSteps+1,loc);
   R eTot0=vs1(1).totEnr();
      // Total energy of initial state of the reference trajectory.
      // Notice that all trajectories under consideration have the
      // same initial state by design.
   cpmassert(eTot0!=0,loc);
   R eTot0Inv=R(1)/eTot0;
   R amcTot0=vs1(1).totAngMomCom(Spc(),dir);
      // component of interest of total angular momentum of the
      // initial state.
   //cpmassert(amcTot0!=0,loc);
   R amcTot0Inv= cpminv(amcTot0);

   R_Matrix enr(nFrames+1,nSteps);
      // For each trajectory there is a R_Vector of energy differences
      // relative to the initial state. In addition to that we have
      // a single R_Vector for the time values to which these differences
      // belong.
   R_Matrix amc(nFrames+1,nSteps);
      // Differences of the angular momentum component relative to the
      // initial state. Time values as in enr.

   R_Matrix dd(nFrames,nSteps);
      // Here we employ a non-trivial numerical characterization of the
      // relative attitude of trajectories: For two synchronous
      // trajectories (t1[j],j=1,...ns), (t2[j],j=1,...,ns) we consider
      // the R_Vector (t1[j].dis(t2[j]), j=1,...,ns ). In our case
      // we consider the attitude of each trajectory relative to the
      // reference trajectory and thus have a list of nFrames-1
      // R_Vectors for trajectory data and, again, one for the
      // time values. It is plausible to assume that in some manifold
      // t1[j]=t_ext[j]+v[j]*(dt1)^p and thus
      // t1[j].dis(t2[j]) =
      // t_ext[j].dis(t_ext[j]) + t_ext[j].dis(v[j])*(dt1)^p
      // + t_ext[j].dis(v[j])*(dt2)^p + v[j].dis(v[j])*(dt1)^p*(dt2)^p
      // = t_ext[j].dis(v[j])*(dt1)^p + t_ext[j].dis(v[j])*(dt2)^p
      // + v[j].dis(v[j])*(dt1)^p*(dt2)^p
      // ~ t_ext[j].dis(v[j])*(dt2)^p (for dt2 >> dt1)
      // notice t_ext[j].dis(t_ext[j]) == 0.
      // Especially if the reference trajectory t1 has a very small dt1
      // it is essentially the dependence on the much larger dt2 which
      // dominates the behavior of t1[j].dis(t2[j]) so that the order
      // analysis as it will done in the following for orderD in analogy
      // to orderE and orderL is well motivated. Specific to the orderD
      // analysis is the usage of a non-trivial weight vector wd in this
      // case.

   for (j=1;j<=nSteps;++j){

      R tj=vs1(j).get_t();
      enr[1][j]=tj;
      amc[1][j]=tj;
      dd[1][j]=tj;

      R e2j=(vs1(j+1).totEnr()-eTot0)*eTot0Inv;
      R amc2j=(vs1(j+1).totAngMomCom(Spc(),dir)-amcTot0)*amcTot0Inv;
      enr[2][j]=e2j;
      amc[2][j]=amc2j;
   }

// main loop that successively increases the number of steplet calls
// in a step
   Z iMid=cpmrnd((1.1+nFrames)*0.5); // the larger if nFrames is even
   Word fileMid("midMovie");
   Word graphMid("midGraph");
   fileMid&=runId;
   fileMid&=".txt";
   graphMid&=runId;
   graphMid&=".txt";
   Z viewingPeriod=4;
   for (i=2;i<=nFrames;++i){
      *this=sMem();
      SysControl::fixedStep=scfs[i];
      Vp<CPM_FCN> vsi=trj(dtStep,nSteps,cam);
      if (i==iMid){
         movieFromTrj(vsi,fileMid,cam,viewingPeriod,!makeMovie);
      }
      finalStates[i]=vsi[vsd];
      glob_prog+=incr;
      cpmprogress(header,glob_prog,seg,acc);
      for (j=1;j<=nSteps;++j){
         R xij=vs1(j).dis(vsi(j));
         dd[i][j]=xij;
         R eij=(vsi(j+1).totEnr()-eTot0)*eTot0Inv;
         enr[i+1][j]=eij;
         R amcij=(vsi(j+1).totAngMomCom(Spc(),dir)-amcTot0)*amcTot0Inv;
         amc[i+1][j]=amcij;
      }
   }

   Word fileFS("fsMovie");
   fileFS&=runId;
   fileFS&=".txt";

   finalStates=finalStates.rev();

   viewingPeriod=1;
   if (makeMovie){
      movieFromTrj(finalStates,fileFS,cam,viewingPeriod);
   }
   R orderE=0;
   R orderD=0;
   R orderL=0;
   bool ivv=iv.isVoid();
   if (!ivv){ // then we know from which piece of the trajectory to infer
      // an estimate for order
      if (repOrderD<=0){ // if there is no valid input, we have to
            // determine a reasonable order from data
         R_Vector xd(nFrames-1);
         R_Vector yd(nFrames-1);
         R_Vector wd(nFrames-1);
         for (i=1;i<nFrames;++i){
            R ti=dt[i+1];
            wd[i]=ti*ti; // The larger values of ti should
            // have the significantly higher influence. See
            // comment to R_Matrx dd.
            xd[i]=cpmlog10(ti);
            yd[i]=cpmlog10(eps+dd[i+1].cutCopy(iv).abs());
         }
         R2 lfd=CpmArrays::lineFit(xd,yd,wd);
            // weighted line fit, introduced here in 2010-09-05
         orderD=lfd.x2;
      }
      if (repOrderE<=0){
         R_Vector xe(nFrames);
         R_Vector ye(nFrames);
         R_Vector we(nFrames,1.);
         for (i=1;i<=nFrames;++i){
            xe[i]=cpmlog10(dt[i]);
            ye[i]=cpmlog10(eps+enr[i+1].cutCopy(iv).abs());
         }
         R2 lfe=CpmArrays::lineFit(xe,ye,we);
         orderE=lfe.x2;
      }
      if (repOrderL<=0){
         R_Vector xe(nFrames);
         R_Vector ye(nFrames);
         R_Vector we(nFrames,1.);
         for (i=1;i<=nFrames;++i){
            xe[i]=cpmlog10(dt[i]);
            ye[i]=cpmlog10(eps+amc[i+1].cutCopy(iv).abs());
         }
         R2 lfl=CpmArrays::lineFit(xe,ye,we);
         orderL=lfl.x2;
      }
   }
   if (repOrderD>0) orderD=repOrderD;
   if (repOrderE>0) orderE=repOrderE;
   if (repOrderL>0) orderL=repOrderL;
      // meaningful input values have the priority
   R_Vector facRepD(nFrames-1);
   R_Vector facRepE(nFrames);
   R_Vector facRepL(nFrames);
      // holds factors which help to collapse a family of E curves into
      // a single one if the order of the integrator has order orderE
      // of the energy error
   for (i=1;i<nFrames;++i){
      R fi=dt[1]/dt[i];
      if (orderD>0) facRepD[i]=cpmpow(fi,orderD);
   }
   for (i=1;i<=nFrames;++i){
      R fi=dt[1]/dt[i];
      if (orderE>0) facRepE[i]=cpmpow(fi,orderE);
      if (orderL>0) facRepL[i]=cpmpow(fi,orderL);
   }
   R_Matrix ddLog(nFrames,nSteps);
   R_Matrix enrLog(nFrames+1,nSteps);
   R_Matrix amcLog(nFrames+1,nSteps);

   for (j=1;j<=nSteps;++j){
      R ti=dd[1][j];
      ddLog[1][j]=ti;
      enrLog[1][j]=ti;
      amcLog[1][j]=ti;
   }
   for (i=2;i<=nFrames;++i){
      R fi=facRepD[i-1];
      for (j=1;j<=nSteps;++j){
         dd[i][j]*=fi;
         ddLog[i][j]=cpmlog10(cpmabs(dd[i][j])+eps);
      }
   }
   for (i=2;i<=nFrames+1;++i){
      R fi=facRepE[i-1];
      for (j=1;j<=nSteps;++j){
         enr[i][j]*=fi;
         enrLog[i][j]=cpmlog10(cpmabs(enr[i][j])+eps);
      }
   }
   for (i=2;i<=nFrames+1;++i){
      R fi=facRepL[i-1];
      for (j=1;j<=nSteps;++j){
         amc[i][j]*=fi;
         amcLog[i][j]=cpmlog10(cpmabs(amc[i][j])+eps);
      }
   }

// Writing data to file in a format suitable to make gnuplot diagrams

   Word fileD=runId&"D.txt";
   OFileStream ofsD(fileD);
   for (i=1;i<=nSteps;++i){
      ofsD()<<endl;
      for (j=1;j<=nFrames;++j)ofsD()<<" "<<dd[j][i];
   }

   Word fileDL=runId&"DL.txt";
   OFileStream ofsDL(fileDL);
   for (i=1;i<=nSteps;++i){
      ofsDL()<<endl;
      for (j=1;j<=nFrames;++j)ofsDL()<<" "<<ddLog[j][i];
   }

   Word fileE=runId&"E.txt";
   OFileStream ofsE(fileE);
   for (i=1;i<=nSteps;++i){
      ofsE()<<endl;
      for (j=1;j<=nFrames+1;++j)ofsE()<<" "<<enr[j][i];
   }

   Word fileEL=runId&"EL.txt";
   OFileStream ofsEL(fileEL);
   for (i=1;i<=nSteps;++i){
      ofsEL()<<endl;
      for (j=1;j<=nFrames+1;++j)ofsEL()<<" "<<enrLog[j][i];
   }

   Word fileL=runId&"L.txt";
   OFileStream ofsL(fileL);
   for (i=1;i<=nSteps;++i){
      ofsL()<<endl;
      for (j=1;j<=nFrames+1;++j)ofsL()<<" "<<amc[j][i];
   }

   Word fileLL=runId&"LL.txt";
   OFileStream ofsLL(fileLL);
   for (i=1;i<=nSteps;++i){
      ofsLL()<<endl;
      for (j=1;j<=nFrames+1;++j)ofsLL()<<" "<<amcLog[j][i];
   }

   Frame fra;
   fra.paint(Color(LIGHTBLUE));
   Frames frs(fra,Z(1),Z(3));
   VV<Frame> vvf=frs.frsVV();
   V<Frame> vf=vvf[1];
   Graph g11(vf[1]);
   Graph g12(vf[2]);
   Graph g13(vf[3]);
   g11.setWrtOnShow(storeGraphAsImage);
   g12.setWrtOnShow(storeGraphAsImage);
   g13.setWrtOnShow(storeGraphAsImage);

   g11.setText(nameOf()&": ordD="&cpm(orderD),0.5);
   if (repLog) g11.show(ddLog); else g11.show(dd);

   g12.setText(nameOf()&": ordE="&cpm(orderE),0.5);
   if (repLog) g12.show(enrLog); else g12.show(enr);

   g13.setText(nameOf()&": ordL="&cpm(orderL),0.5);
   if (repLog) g13.show(amcLog); else g13.show(amc);

   SysControl::fixedStep=fsMem;
   *this=sMem();
   R_Vector res(3);
   res[1]=orderD;
   res[2]=orderE;
   res[3]=orderL;
   CPM_MZ
   return res;
}


CPM_TAL
void  CPM_FCN::redistribute(void)
{
   Z mL=2;
   Z mL2=3;
   sizeAndRank();
   Word loc("CPM_FCN::redistribute()");
   bool isTorus=partition.torus();
   if (size==1 && !isTorus) return; // nothing to redistribute then
   // notice that toroidal topology needs some redistribution
   // even for size==1
      // added 2004-05-11 and 2004-05-18
   cpmmessage(mL,loc&" started, np="&cpm(np));
   const R slowDownFactor=0.9;
   Z i,j;
   V< V<Par> > groupsToSend(size,V<Par>(0));
   V< V<Par> > groupsToReceive(size,V<Par>(0));
   for (i=1;i<=np;i++){ // each rank loops over the particles that are
      // visible to it
      Z oi=par[i].getOwner();
         // ownership by presently valid assignment
      if (oi==rank){ // we dont collect the 'virtual' particles
         // moreover we collect only those particles which we own
         Par pi=par[i];
         Spc xi=pi.get_x();
         X3<Z,Group,B> xzg=partition.owner(xi);
         oi=xzg.c1();
            // owner based on present position, which is the one that
            // matters
         Group gi=xzg.c2();
         bool slowDown=xzg.c3();
         pi*=gi;
            // so far this is OK for toroidal topology. But also in
            // the non-toroidal case one should have means to bring a
            // particle
            // back to X if it has left it by action of steplet(...)
            // No problem, in this case the wall forces have to do the
            // job!
         if (slowDown){
            bool selective=true; // influences only velocity in
            // 3-direction
            pi.stop(slowDownFactor,selective);
         }
         pi.setOwner(oi);
         groupsToSend[oi].push_back(pi);
            // for oi==rank this is also OK
      }
   }
   // Thus the assigned owner decides about the new ownership (i.e. the
   // one after function call)

   if (cpmverbose>=mL2){ // diagnosis concerning groupsToSend and
      //groupsToReceive
      Z dimS=groupsToSend.dim(), dimR=groupsToReceive.dim();
      Z dimSumS=0, dimSumR=0;
      for (i=1;i<=size;++i){
         dimSumS+=groupsToSend[i].dim();
         dimSumR+=groupsToReceive[i].dim();
      }
      cpmmessage(loc&": size="&cpm(size));
      cpmmessage(loc&": dimS="&cpm(dimS));
      cpmmessage(loc&": dimR="&cpm(dimR));
      cpmmessage(loc&": dimSumS="&cpm(dimSumS));
      cpmmessage(loc&": dimSumR="&cpm(dimSumR));
   }
   Cpm_com.exchange(groupsToSend,groupsToReceive);
   par=groupsToReceive[rank];
      // these particles have positions that where held by particles
      // of rank before
   for (i=1;i<=size;i++){
      if (i!=rank) par&=groupsToReceive[i];
   }
      // this adds particles that might overlap with existing particles
      // and could cause desaster.
   np=par.dim();
   for (i=1;i<=np;i++) cpmassert(par[i].getOwner()==rank,loc);
// now each process has all the particles which he owns. It now waits
// for those that he has to use only.

   V< V<Par> > groupsToSend2(size,V<Par>(0));
   V< V<Par> > groupsToReceive2(size,V<Par>(0));

   typedef X2<Z,Group> ZG;

   for (i=1;i<=np;i++){
      Par pi=par[i];
      groupsToSend2[rank].push_back(pi);
         // needed in order for following statement (*) to work
      Spc pix=pi.get_x(); // only the position matters
      V<ZG> vzg=partition.toSend(rank,pix);
      Z vd=vzg.dim();
      for (j=1;j<=vd;j++){
         ZG zgj=vzg[j];
         Z ij=zgj.c1();
         Group gj=zgj.c2();
         Par pij=pi;
         pij*=gj;
         pij.setOwner(0);
         groupsToSend2[ij].push_back(pij);
      }
   }

   Cpm_com.exchange(groupsToSend2,groupsToReceive2);

   par=groupsToReceive2[rank]; // (*)
   for (i=1;i<=size;i++){
      if (i!=rank) par&=groupsToReceive2[i];
   }
      // again, this adds particles that might overlap with existing
      // particles and could cause desaster.
   np=par.dim();
   cpmmessage(mL,loc&" done, np="&cpm(np));
}

////////////////////////// end of main dynamics stuff //////////////////

// getting VectorField and ScalarField
// external fields

CPM_TAL
VectorField CPM_FCN::getFieldExt()const
{
   VectorField res;
   if (!SysControl::ignoreExternalSystems)
      for (Z i=1;i<=ne;i++) res+=ext(i).getField();
   return res;
}

CPM_TAL
ScalarField CPM_FCN::getPotExt()const
{
   ScalarField res;
   if (!SysControl::ignoreExternalSystems)
      for (Z i=1;i<=ne;i++) res+=ext(i).getU();
   return res;
}

// fields of particles. These have the mean field alternative.
// This alternative becomes active only for graphical
// actions. The internal dynamics does not use
// meanPotPar and uses meanFieldPar directly where needed.

CPM_TAL
VectorField CPM_FCN::getFieldPar(bool mean)const
{
   Z mL=3;
   Word loc("CPM_FCN::getFieldPar(mean)");
   CPM_MA
   VectorField res;
   if (mean){
      if (!meanFieldAct){
         cpmerror(loc&": mean field not initialized");
      }
      else
         res=meanFieldPar;
   }
   else{
      V<VectorField> vf(np);
      for (Z i=1;i<=np;i++) vf[i]=fieldPar(par[i]);
      res=VectorField::sum(vf);
   }
   CPM_MZ
   return res;
}

CPM_TAL
ScalarField CPM_FCN::getPotPar(bool mean)const
{
   Z mL=3;
   Word loc("CPM_FCN::getPotPar(bool)");
   CPM_MA
   ScalarField res;
   if (mean){
      if (!meanFieldAct){
         cpmerror(loc&": mean field not initialized");
      }
      else
         res=meanPotPar;
   }
   else{
      V<ScalarField> vf(np);
      for (Z i=1;i<=np;i++) vf[i]=potentialPar(par[i]);
      // for (Z i=1;i<=np;i++) vf[i]=par[i].getU(); is OK
      // if there are no mirror charges
      res=ScalarField::sum(vf);
   }
   CPM_MZ
   return res;
}

// combining external and particle fields

CPM_TAL
VectorField CPM_FCN::getFieldTot(bool mean)const
{
   return getFieldExt()+getFieldPar(mean);
}

CPM_TAL
ScalarField CPM_FCN::getPotTot(bool mean)const
{
   return getPotExt()+getPotPar(mean);
}

// marking fields

CPM_TAL
void CPM_FCN::markPotExt(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markPotExt(Camera)");
   CPM_MA
   getPotExt().mark(sc);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markPotPar(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markPotPar(Camera)");
   CPM_MA
   bool mean=SysControl::showMeanFil;
   getPotPar(mean).mark(sc);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markPotTot(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markPotPar(Camera)");
   CPM_MA
   bool mean=SysControl::showMeanFil;
   getPotTot(mean).mark(sc);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markPot(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markPot(Camera)");
   CPM_MA
   if(SysControl::showPartFields)
      getPotTot(SysControl::showMeanFil).mark(sc);
   else
      getPotExt().mark(sc);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markField(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markField(Camera)");
   CPM_MA
   if(SysControl::showPartFields)
      getFieldTot(SysControl::showMeanFil).mark(sc);
   else
      getFieldExt().mark(sc);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markScalarValue(const Word& w,const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markScalarValue(Camera)");
   CPM_MA
   ParticleViewer<Par,Spc,Camera> pv(par);
   Z ny=SysControl::nyScalarFields;
   bool pixelized=false;
   pv.showScalarValue(w,sc,ny,pixelized);
   CPM_MZ
}

CPM_TAL
void CPM_FCN::markCharge(const Camera& sc)const
{
   Z mL=3;
   Word loc("CPM_FCN::markCharge(Camera)");
   CPM_MA
   ParticleViewer<Par,Spc,Camera> pv(par);
   Z ny=SysControl::nyScalarFields;
   bool pixelized=false;
   pv.showCharge(sc,ny,pixelized);
   CPM_MZ
}

// marking the whole system

CPM_TAL
void CPM_FCN::mark(const Camera& sc, bool display,
   const Word& text, bool fileOnShow)const
{
   Z mL=3;
   cpmmessage(mL,"CPM_FCN::mark(Camera,bool,Word,Z) started, np="&
      cpm(np));
   Z i,id;
   Color cAux(WHITE);
   Z bms=SysControl::bitmapToShow;
   Z scd=showColor.dim();
// first graphical layer: area-filling bitmap for bms!=0
   if (bms==0) ; // no bitmap layer
   else if (bms==1) markPot(sc);
   else if (bms==3) markCharge(sc); // traditional
   else{
      Word w;
      if (bms==2) w="mass";
      else if (bms==4)  w="velocity";
      else if (bms==5)  w="angular_velocity";
      else if (bms==6)  w="kinetic_energy";
      else if (bms==7)  w="electric_dipole";
      else if (bms==8)  w="electric_polarization";
      else if (bms==9)  w="magnetic_dipole";
      else if (bms==10) w="volume";
      else if (bms==11) w="surface";
      else w="nonsense";
      markScalarValue(w,sc); // fails for 'nonsense'
   }
// second layer: vector graphics showing particles and field
// vectors
   if (SysControl::showParticles){ // showing the mean field
         // works better if the particles are not shown
#if !defined(CPM_FIGURE_FOR_PUBLICATION_1) // normal situation
      if (scd>0){
         for (i=1;i<=np;i++){
            id=par[i].get_id();
            Color ci=showColor(id) ;
            // If the particle manifold resulted from a larger one by
            // selection, id needs not to be in the range 1,...,np
            if (ci==Color()) par[i].mark(sc); else par[i].mark(sc,ci);
         }
      }
      else{
         for (i=1;i<=np;i++){
            Z ovi=par[i].getOwner();
            if (ovi<=0) par[i].mark(sc,cAux);
            else par[i].mark(sc);
         }
      }
#else // making a particular image for a publication,
      // All particles and the external system ext(2) get combined into
      // a SubsetOfSpace, which then is shown as a textured stereo image
      using CpmDim3::SubsetOfSpace;
      using CpmDim3::Texture;
      using CpmDim3::RefSys;
      SubsetOfSpace sos=par[1].toSubOfSpc();
      for (i=2;i<=np;i++) sos=sos|par[i].toSubOfSpc();

      SubsetOfSpace sos2=ext(2).toSubOfSpc();
      R r2=sos2.get_r();
      RefSys s2=sos2.getRefSys();
      SubsetOfSpace b1=SubsetOfSpace::band(3,Iv(0,-0.5*r2),s2);
      sos2=sos2&b1;
      sos=sos|sos2;
      sos.set_r(r2*1.1);
      sos.set_x(Spc());
      sc.mark(sos,Texture(0.025));
#endif
   }

   if (SysControl::markElectricField>=0
         || SysControl::markMagneticField>=0
         || SysControl::markGravityField>=0 ) markField(sc);

#if !defined(CPM_FIGURE_FOR_PUBLICATION_1)
   for (i=1;i<=ne;i++) ext(i).mark(sc);
#endif

   if (SysControl::showPartition) partition.mark(sc);

   R xT=0, yT=1;
   Color colT(Z(200),Z(200),Z(200));
   if (display){
      sc.addText(text,xT,yT,colT);
      bool wosMem=sc.getWriteOnShow();
      sc.setWriteOnShow(fileOnShow);
      sc.vis(); // here an argument controls thick lines!!
     // sc.vis(); // argument true lets write a ppm - image
      // xT=0, yT=1 sets the text to the left upper corner (i.e. so that
      // the left upper corner of the text box coincides with the
      // left upper corner of the frame rectangle) which is presumably
      // the least obstructing position
      sc.setWriteOnShow(wosMem);
   }

   cpmmessage(mL,"CPM_FCN::mark(Camera,...) done, np="&cpm(np));
}

// most advanced lambda abstraction within a template class:

//fieldPar(...)
CPM_TAL
FieldSpace fieldParFunc
(const Spc& x, const CPM_FCN &th, const Par& p, const B& b)
{ return th.field(x,p,b);}


CPM_TAL
VectorField CPM_FCN::fieldPar(const Par& par1, bool farAway)const
{
   return
#ifdef CPM_Fn
   F3<Spc,CPM_FCN,Par,B,FieldSpace>(*this,par1,B(farAway))
   (fieldParFunc CPM_DEP);
#else
   F<Spc,FieldSpace>(bind(fieldParFunc CPM_DEP,_1,*this,par1,B(farAway)));
#endif
}

// potentialPar(...)
CPM_TAL
R potentialParFunc
(const Spc& x, const CPM_FCN &th, const Par& p,
   const B& b)
{ return th.potential(x,p,b);}

CPM_TAL
ScalarField CPM_FCN::potentialPar(const Par& par1, bool farAway)const
{
   return
#ifdef CPM_Fn
   F3<Spc,CPM_FCN,Par,B,R>(*this,par1,B(farAway))
      (potentialParFunc CPM_DEP); // CPM_DEP is optional here
#else
   F<Spc,R>(bind(potentialParFunc CPM_DEP,_1,*this,par1,B(farAway)));
 // here CPM_DEP is mandatory for finding a resolution of the function
 // name 'potentialParFunc'
#endif
}

#undef CPM_CN
#undef CPM_TAL
#undef CPM_FCN

} // namespace

#endif
