//? cpmparticleviewer.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_PARTICLEVIEWER_H_
#define CPM_PARTICLEVIEWER_H_

#include <cpminterval.h>
#include <cpmlinalg.h>
#include <cpmnumbers.h>
#include <cpmimg24.h>
#include <cpmcamera.h>
#include <cpmimage.h>
#include <cpmmultiimage.h>
#include <cpmr2func.h>

namespace CpmPhysics{

///////////////////////////// class ParticleViewer /////////////////////

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;
   using CpmRoot::C;
   using CpmRoot::L;

   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::VVV;
   using CpmArrays::X2;
   using CpmArrays::R_Matrix;

   using CpmGeo::Iv;

   using CpmImaging::Image;
   using CpmImaging::MultiImage;
   using CpmImaging::Mask;
   using CpmImaging::Rect;
   using CpmImaging::Color;
   using CpmImaging::RED;
   using CpmImaging::GREEN;
   using CpmImaging::BLUE;
   using CpmImaging::Response;

   using CpmGraphics::Graph;
   using CpmGraphics::ColRef;
   using CpmGraphics::Img24;

   using CpmFunctions::R_Func;
   using CpmFunctions::R2_Func;



template <class Part, class Space, class Camera>
class ParticleViewer{ // particle viewer
// Visualizing distributed quantities ('densities' as
// contrasted to intensities) that arrise from a set of particles.
// Presently only the charge distribution is treated. The logic
// of parcellating the image plane of a camera to get a
// visual rendition of the density can easily be generalized to other
// quantities such as the kinetic energy.
// Addition 03-09-25: now all scalar value implemented in
// R Clust::sclVal(const Word& w) for some w (e.g. w="charge",
// or w="kinetic_energy" or w="velocity") are treated in the new function
// void showScalarValue(...)

   V<Part> vp;
public:
   ParticleViewer(const V<Part>& vpIn):
      vp(vpIn){}

   void showCharge( const Camera& cam, Z ny, bool pixelized=false)const;
   // notice that the mutable attribute cam.gr gets normally
   // changed, and that this is essential for the intended
   // functionality
   // special elaborate color coding for charge density
   // where a hypothetical 'unsigned total charge density'
   // (to which a particle charge q always contributes |q|)
   // controls the brightness of an area. Gray means the
   // neutrality with respect to the true (signed) charge
   // and blue saturation increases with net positive
   // charge density (all my movie-files show positive
   // (carrier) particles as blue) wheras red saturation increases
   // with negative charge density.

   void showScalarValue( const Word& prop,
      const Camera& cam, Z ny, bool pixelized=false)const;
   // shows on the screen window of cam a distributed (positive) property
   // which is determined by the behavior of Clust::sclVal(prop).
   // The distribution is constructed in the following manner.
   // All particles of vp get imaged by cam. The 'film plane'
   // gets pixelized into ny(height)*nx cells , where nx is determined
   // the aspect ratio of cam's Graph attribute. To each cell
   // a real value val is defined by initialization as 0
   // and updating any present value by the maximum of the
   // present value and the value of any particle the center-of-mass
   // lies in this cell.
   // This is especially suited for finding hot-spots of e.g. the
   // velocity. Not for all values of prop for which
   // Clust::sclVal(prop) is defined, this is expected to be useful.
   // For potential extensions to quantities not yet taken into
   // consideration, one has simply to extend the definition of
   // Clust::sclVal(Word).

};

template <class Part, class Space, class Camera>
void ParticleViewer<Part,Space,Camera>::showCharge(
   const Camera& cam, Z ny, bool pixelized)const
{
   bool styleTD=true;
      // if true, parts of space that are electrically neutral
      // by containing no charge at all get displayed as dark
      // Always: negative charge is read, positive one is blue

   using namespace CpmRoot;
   static bool firstrun=true;
//   static Z imgCount=0;
   Z mL=1;
   Word loc("ParticleViewer::showScalarValue(...)");
   cpmmessage(mL,loc&" started");
   using CpmGeo::Iv;
   using CpmLinAlg::R2;
   using CpmRoot::L;
   using CpmGraphics::Img24;
   using CpmImaging::Response;

   Z i,j,np=vp.dim();
   Z nx=cpmtoz(R(ny)*cam.getGraph().getAspRat());
   cpmassert(nx>0,loc);
   Iv ivx=cam.getGraph().getIvX();
   Iv ivy=cam.getGraph().getIvY();
   cpmassert(ivx.abs()>0,loc);
   cpmassert(ivy.abs()>0,loc);
   R valIni=0;
   VV<R> val0(ny,nx,valIni); // sic !
   VV<R> val1(ny,nx,valIni); // sic !
   VV<R> val2(ny,nx,valIni); // sic !

   bool restrict=true;
   for (i=1;i<=np;i++){
      Space pi=vp[i].get_x();
      C xi=cam.captureMono(pi);
      R qi=vp[i].sclVal("charge");
      Z ix=ivx.quant(xi[1],nx,restrict); // without restrict one gets
         // points outside the index range too
      Z iy=ivy.quant(xi[2],ny,restrict);
      Z iyc=ny-iy+1;
      val0[iyc][ix]+=qi; // cumulative channel
      if (qi>0){
         val1[iyc][ix]+=qi;
      }
      else{
         val2[iyc][ix]-=qi;
      } // registrating the value to the right box
   }
   // now we have registered each particle's contribution
   // in the cumulative channel and in the approriate signed channel
   R rMask=4.;
   R sigmaMask=1.5;
   static Mask mask;
   static Response rsp0;
   static Response rsp1;
   static Response rsp2;
   static R valAbs0;
   static R valAbs1;
   static R valAbs2;
   if (firstrun){
      // my monochrome image class already provides
      // flexible convolution
      mask=Mask(rMask,R_Func::Gauss(0,sigmaMask));

      Image<R> im0(val0);
      im0=im0.conv(mask);
      val0=im0.toVV();

      Image<R> im1(val1);
      im1=im1.conv(mask);
      val1=im1.toVV();

      Image<R> im2(val2);
      im2=im2.conv(mask);
      val2=im2.toVV();

      valAbs0=0;
      valAbs1=0;
      valAbs2=0;
      for (i=1;i<=ny;i++){
         for (j=1;j<=nx;j++){
            R v0=cpmabs(val0[i][j]);
            R v1=val1[i][j];
            R v2=val2[i][j];
            if (v0>valAbs0) valAbs0=v0;
            if (v1>valAbs1) valAbs1=v1;
            if (v2>valAbs2) valAbs2=v2;
         }
      }
      rsp0.setGamma(1.);
      rsp1.setGamma(1.);
      rsp2.setGamma(1.);
      rsp0=cam.getResponse();
      rsp1=cam.getResponse();
      rsp2=cam.getResponse();
      rsp0.setRangeY(0,255);
      rsp1.setRangeY(0,255);
      rsp2.setRangeY(0,255);
      R valAbs=cpmsup(valAbs1,valAbs2);
      R security=1.2;
      rsp0.setRangeX(-security*valAbs0,security*valAbs0);
      rsp1.setRangeX(0,security*valAbs);
      rsp2.setRangeX(0,security*valAbs);
      firstrun=false;
   }
   else{
      Image<R> im0(val0);
      im0=im0.conv(mask);
      val0=im0.toVV();
      Image<R> im1(val1);
      im1=im1.conv(mask);
      val1=im1.toVV();
      Image<R> im2(val2);
      im2=im2.conv(mask);
      val2=im2.toVV();
   }
   VV<ColRef> imDat(ny,nx);
   for (i=1;i<=ny;i++){
      for (j=1;j<=nx;j++){
         if (styleTD){
            R p1=rsp1.getRangeX().ramp(val1[i][j]);
            R p2=rsp2.getRangeX().ramp(val2[i][j]);
            Color col1(BLUE);
            Color col2(GREEN);
            Color col3(RED);
            R p12=cpminf(p1,p2);
            Color col=col1*p1+col2*p12+col3*p2;
            imDat[i][j]=col.toColRef();
         }
         else{
            R p0=rsp0.getRangeX().ramp(val0[i][j]);
            Color col(p0);
            imDat[i][j]=col.toColRef();
         }
      }
   }
   Img24 im(imDat);
   MultiImage<L> imm(im);
   Z method=(pixelized ? 0 : 1);
   Z faithful=1;
   imm.markDirect(cam.getGraph(),method,faithful);
   cam.captureGraph();
   cpmmessage(mL,loc&" done");
}

template <class Part, class Space, class Camera>
void ParticleViewer<Part,Space,Camera>::showScalarValue(
   const Word& prop, const Camera& ca, Z ny, bool pixelized)const
{
   Z mL=3;
   Word loc("ParticleViewer<>::showScalarValue(...)");
   CPM_MA
   pixelized; // not used
   Response rsp=ca.getResponse();
   rsp.setLevels(10);
   bool adjust=!rsp.normalizeGamma();
   Z nx=cpmtoz(R(ny)*ca.getGraph().getAspRat());
   Iv ivx=ca.getGraph().getIvX();
   Iv ivy=ca.getGraph().getIvY();
   V<R> vx=ivx.centers(nx);
   V<R> vy=ivy.centers(ny);
   Z np=vp.dim();
   Z i;
   Z count=0;
   R allInv=1./np;
   R_Matrix val(ny,nx); // indexed like an image matrix
   bool restrict=true; // true is OK
   for (i=1;i<=np;i++){
      Space pi=vp[i].get_x();
      C xi=ca.captureMono(pi);
      if ( ivx.contains(xi[1]) && ivy.contains(xi[2]) ){
         Z ix=ivx.quant(xi[1],nx,restrict); // without restrict one gets
         // points outside the index range too
         Z iy=ivy.quant(xi[2],ny,restrict);
         Z iyc=ny-iy+1;
         R propi=vp[i].sclVal(prop);
         val[iyc][ix]=CpmRootX::sup<R>(propi,val[iyc][ix]);
            // cumulation as for charges is much less attractive here
         count++;
         cpmprogress("density-loop",R(count)*allInv);
      }
   }
   Iv riv;
   if (adjust){
      Iv ivRange=rsp.getRangeX();
      X2<R_Matrix,Iv> xri=val.adjustRange(ivRange);
      val=xri.c1();
      riv=xri.c2();
   }
   Z iPol=1; // interpolates acording to my standard image
      // interpolation (simplex method).
   R2_Func r0(val.toVV(),iPol,ivx,ivy);
   ca.getGraph().show(r0,rsp,ivx,ivy);
   if (adjust){
      R vL=riv.inf();
      R vU=riv.sup();
      if (cpmverbose>0)
         ca.getGraph().addText(cpm(vL)&" <=val<= "&cpm(vU),0.25,0.1);
   }
   CPM_MZ
}
} // namespace
#endif
