//? cpmproperties.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_PROPERTIES_H_
#define CPM_PROPERTIES_H_
/*
   Description: Declares basic quantities describing
      properties of particles, confining structures, and the system as
      a whole.
*/

#include <cpmrecordhandler.h>
#include <cpmimagingtools.h>
#include <cpmrvector.h>

namespace CpmProperties{  // a namespace for managing common data

   using namespace CpmRoot;
   using CpmRootX::RecordHandler;
   using CpmRootX::Record;
   using CpmRootX::Recordable;
   using CpmRootX::B;
   using CpmImaging::Palette;
   using CpmImaging::Color;
   using CpmArrays::V;
   using CpmArrays::R_Vector;
   using CpmAlgorithms::CyclicAlarm;

R nonHarmonicMean(R x1, R x2);
// For combination of some physical quantities the property
// of the harmonic mean to let a 0 always dominate the result
// is inconvenient, so we use the simple mean instead. Could
// be changed to some L^p-mean if needed

R harmonicMean(R x1, R x2);
   // 1/res = (1/x1 + 1/x2)/2, => res=2*x1*x2/(x1+x2)
   // notice:
   //   harmonicMean(a,b) = harmonicMean(b,a)
   //   harmonicMean(a,a) = a
   //   harmonicMean(a,0) = 0
   //   harmonicMean(a,infinity) = 2*a
   // Important approximate relation:
   //   harmonicMean(a,b) '=' 2*a  for  a << b

R mixedMean(R x1, R x2);
   // harmonic mean normally but switches to nonHarmonicMean if at
   // least one argument is 0

R transitionFunction(R const& x, R const& xNear, R const& xFar);
   // makes a differentiable third-order polynomial transition from 1
   // valid for x<=xNear, and 0 valid for x>=xFar. Computationally very
   // effective.

///////////////////////// SysControl /////////////////////////////////////

struct SysControl{ // system control parameters
   // A collection of static variables and functions.
   // Thus no constructor.

   static Z nyScalarFields;

   static Z nxElectricField;
   static Z nyElectricField;
   static Z markElectricField;

   static Z nxMagneticField;
   static Z nyMagneticField;
   static Z markMagneticField;

   static Z nxGravityField;
   static Z nyGravityField;
   static Z markGravityField;

   static R fieldGamma;

   static Z verbose;

   static Z fixedStep;
   static R intactRange;
   static R singRange;
   static Z noAdhesion;
      // adhesion forces=0 in the present program
   static Z noElasticRepulsion;
   static Z noFriction;
      // if this is different from 0 all energy dissipating mechanisms
      // in interaction get disabled
   static Z noInteractionChargeToWall;
   static Z noInteractionChargeToCharge;
   static Z noMagneticInteractionChargeToCharge;
   static Z noInteractionMagnetToMagnet;
   static Z noInteractionMassToMass;
   static Z noInteractionElectricField;
   static Z noInteractionElectricDipole;
   static Z noInteractionMagneticField;
   static Z noForceMagneticField;
   static Z noInteractionGravitationField;
   static Z noMeanFieldPar;
   static Z ignoreExternalSystems;
      // if this is true, all external systems (typically
      // instances of classes derived from ExtSys have
      // the mark(...) function and the field evaluation
      // 'switched off'
   static Z coloredParticles;
   static Z singleCharge;
   static B accMem;
      // acceleration memory
   static B newRigBodyInt;
      // new rigid body integrator
   static R eta;
      // dynamical viscosity used for generating a damping force
      // on any particle according to Stokes' formula
   static R vCrit;
   static R aCrit;
   static B veryFar;
   static Z bitmapToShow;
   static Z potentialToShow;
      // selects between showing the electric potential (1)
      // scalar magnetical potential (2) and the gravitational
      // potential (3)
   static Z howToMarkParticles;
      // 0 for classical representation by a cross of axes
      // 1 via WireFrame::globe
   static B showMeanFil;
   static B showParticles;
   static B showPartFields;
   static B showPartition;
   static B showPrimaryChargesOnly;
   static R nearFactor;
   static R tSmall;
   static R tStep;
   static R tWait;
   static R dtMult;
      // v*dt, with
      // dt=SysControl::tStep*SysControl::dtMult
      // is the distance that is shown to indicate the velocity
      // of a particle a value <=0 disables graphical representation of
      // particle velocities

   static void read(RecordHandler const& rch, Word const& sec, Z ped=1);
      // sets the values from rch, which is a data type conveniently
      // created from ini-file input. ped means pedantry: value 1
      // creates warnings if an item was not found in rch, and
      // 2 stops on failure in this case.

   static void reset();
      // resets the state of the static variables to the one valid after
      // initialization
};

///////////////////////// SysControlState ////////////////////////////////

struct SysControlState{ // system control parameters as non-static data
   // holds the same data (attributes) as SysControl, however here
   // the data are non-static so that there are different instances
   // of the class. Non-trivial setting of the values is intended
   // to be achieved by calling mem().
   // Notice that all members are class instances and thus are
   // automatically initialized.

   typedef SysControlState Type;
   CPM_IO

   Z nyScalarFields{0};

   Z nxElectricField{0};
   Z nyElectricField{0};
   Z markElectricField{0};

   Z nxMagneticField{0};
   Z nyMagneticField{0};
   Z markMagneticField{0};

   Z nxGravityField{0};
   Z nyGravityField{0};
   Z markGravityField{0};

   R fieldGamma{0.};

   Z verbose{0};

   Z fixedStep{0};
   R intactRange{0.};
   R singRange{0.};
   Z noAdhesion{0};
   Z noElasticRepulsion{0};
   Z noFriction{0};
   Z noInteractionChargeToWall{0};
   Z noInteractionChargeToCharge{0};
   Z noMagneticInteractionChargeToCharge{0};
   Z noInteractionMagnetToMagnet{0};
   Z noInteractionMassToMass{0};
   Z noInteractionElectricField{0};
   Z noInteractionElectricDipole{0};
   Z noInteractionMagneticField{0};
   Z noForceMagneticField{0};
   Z noInteractionGravitationField{0};
   Z noMeanFieldPar{0};
   Z ignoreExternalSystems{0};
   Z coloredParticles{0};
   Z singleCharge{0};
   B  accMem;
   B  newRigBodyInt;
   R eta{0.};
   R vCrit{0.};
   R aCrit{0.};
   B  veryFar;
   Z bitmapToShow;
   Z potentialToShow;
   Z howToMarkParticles;
   B  showMeanFil;
   B  showParticles;
   B  showPartFields;
   B  showPartition;
   B  showPrimaryChargesOnly;
   R nearFactor{0.};
   R tSmall{0.};
   R tStep{0.};
   R tWait{0.};
   R dtMult{0.};

   SysControlState(void){}
      // trivial constructor

   void act()const;
      // copies the data of *this into the static equivalents and thus
      // controls the behavior of all functions depending on SysControl

   void mem();
      // copies the valid static values (of SysControl) into the
      // corresponding data of *this
};

////////////////////////////// ContactProperties ////////////////////////

struct ContactProperties{ // data describing particle contact

   typedef ContactProperties  Type;
   CPM_IO

// adhesion, not implemented in the present program
   R alphaAdhesion_;
   R relativeRangeAdhesion_;

// elasticity
   R E_;
      // a modified module of elasticity E (e.g. 3.2e9 for Polystyrol) in
      // SI units effectiveE = E/(1-mu*mu), mu=Poisson number (0.35 for
      // Polystyrol)
      // Thus effectiveE approximately equal to E
// tangential friction
   R mu_;
      // coefficient of friction
   R v0_;
      // a small regularizing velocity
// normal friction
   R scnr_;
      // square of coefficient of normal restitution

   ContactProperties(void):alphaAdhesion_(0.),relativeRangeAdhesion_(0.),
      E_(0.),mu_(0.),v0_(0.),scnr_(1.){}
        // default constructor should not contain I/O operations
        // for efficiency. For scnr_ is 1 the natural default value.
   ContactProperties(RecordHandler const& rch, Word const& section);
      // 'modern' input from a section of a ini-file
   ContactProperties combine(const ContactProperties& cp)const;
      // returns an instance, the components of which are the harmonic
      // mean of the components of *this and cp
};

/////////////////// ParticleProperties ///////////////////////////////////

struct ParticleProperties: public ContactProperties{
   // particle properties

   Palette color;
      // not a physical property; artificially assigned color for
      // visualization; We have in mind the constructor Color(Z i)
      // which creates INVISIBLE for i<0. The class Color needs too much
      // storage to be the right type to appear here
   R rho;
      // density in SI unit kg/(cubic meter)
   R sigma;
      // surface charge density in SI unit Coulomb/(square meter)
   R J;
      // magnetic polarization in SI unit Tesla
   R eps;
      // relative dielectric constant
   R fRelax;
      // inverse of a dielectric relax time

   typedef ParticleProperties Type;
   CPM_IO
   ParticleProperties():ContactProperties(),color(CpmImaging::STEREO),
      rho(0.),sigma(0.), J(0.),eps(1.), fRelax(0.){}
        // default constructor should not contain I/O operations
        // for efficiency. color is INVISIBLE for default

   explicit ParticleProperties(ContactProperties const& cp):
      ContactProperties(cp),color(CpmImaging::STEREO),
      rho(0.),sigma(0.), J(0.),eps(1.), fRelax(0.){}

   ParticleProperties(const RecordHandler& rch, const Word& section);
      // 'modern' input from a section of a ini-file

   void set_cp(const ContactProperties& cp)
     // {(ContactProperties)(*this)=cp;} dubious construction
      // setting the base class part of data from base class
   {
      alphaAdhesion_=cp.alphaAdhesion_;
      relativeRangeAdhesion_=cp.relativeRangeAdhesion_;
      E_=cp.E_;
      mu_=cp.mu_;
      v0_=cp.v0_;
      scnr_=cp.scnr_;
   }
};

//////////////////////// BodyProperties //////////////////////////////////

struct BodyProperties:public ContactProperties{//prop. surrounding bodies

   typedef BodyProperties  Type;
   typedef ContactProperties  Base;
   CPM_IO

// extendable list of material properties and method parameters

   R eps;
      // relative constant of dielectricity
   R d;
      // thickness of the dielectric layer

   BodyProperties(void):eps(1.),d(0.){}
        // default constructor should not contain I/O operations
        // for efficiency
   BodyProperties(const Base& b):
      Base(b),eps(1.),d(0.){}

   BodyProperties(const RecordHandler& rch, const Word& section);
      // 'modern' input from a section of a ini-file
};

//////////// NuggetData, DumbbellData, PairData //////////////////////////
//
// The following three structs serve as types for Clust constructor
// arguments. They do not only give a suitable means to handle the
// moderately large set of data going into the construction of a Clust,
// but also allow to distinquish by name several distinct methods of
// construction which can be considered to belong to different types
// of Clust's ('morphological types'). These types don't express
// themselves as different data types for describing a state but only by
// different structure of the values ('value-polymorphism'). To treat
// 'subspecies' of a type in this way evolved from my earlier method to
// use a Word-typed variable as a constructor argument to clearly
// indicate the nature of the construction process under consideration.
// To use different types instead of different string-values offers the
// essential advantage that the construction data may vary from method to
// method.

struct NuggetData{ // data of nugget-like grains
   // see Clust(R d, const ParticleProperties& pp, const NuggetData& pd);
   // or ComPar(R d, const ParticleProperties& pp, const NuggetData& pd);
   R overlapFac{0.};
   R isotropyFac{0.};
   Z np{0};
   Z trialMax{0};
   typedef  NuggetData Type;
   CPM_IO
   NuggetData(){}
   NuggetData(Z n, R overlap=1.75, R isotr=1.0):
   overlapFac(overlap),isotropyFac(isotr),np(n),trialMax(40000){}
   NuggetData(const RecordHandler& rch, const Word& section);
   NuggetData(Record const& rc, const Word& section);
};

struct DumbbellData{// data of dumbbell-shaped grains
   // see Clust(R d, const ParticleProperties& pp, const DumbbellData&
   // pd);
   R rc{0.};
   R re{0.};
   R phi{0.};
   typedef  DumbbellData Type;
   CPM_IO
   DumbbellData(){}
   DumbbellData(const RecordHandler& rch, const Word& section);
   DumbbellData(Record const& rc, const Word& section);
};

struct PairData{// data of pair-shaped grains
   // see Clust(R d, const ParticleProperties& pp, const PairData& pd);
   R r1{0.};
   R r2{0.};
   R phi{0.};
   typedef  PairData Type;
   CPM_IO
   PairData(){}
   PairData(const RecordHandler& rch, const Word& section);
   PairData(Record const& rc, const Word& section);
};

//////////////////////// ParEnsPrp ///////////////////////////////////////

class ParEnsPrp: public Recordable{ // Particle Ensemble Properties
   // Describes the design parameters of a particle ensemble using
   // Recordable which is a flexible new data type for such purposes.

public:

   ParEnsPrp(){}

   ParEnsPrp(const RecordHandler& rch, const Word& sec):
      Recordable(rch.getRecord(),sec){}

   ParEnsPrp(const Record& rc, const Word& sec):Recordable(rc,sec){}
   ParEnsPrp(const Record& rc):Recordable(rc){}
};

} // CpmProperties

#endif
