//? cpmactionprinciple.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmactionprinciple.h>
#include <cpmlinalg.h>
#include <cpmrfunc.h>
#include <cpmgraph.h>
#include <cpmimage.h>

using namespace CpmRoot;
using namespace CpmArrays;
using namespace CpmGraphics;
using namespace CpmFunctions;
using namespace CpmStd;
using CpmApplication::IniFileBasedApplication;

using CpmRootX::RecordHandler;
using CpmLinAlg::R3;

///////////////////// class Action /////////////////////////////////////
class Action{
// independent data
   R3 I_;
     // eigen values of the tensor of inertia
   R3 x0_;
   R3 v0_;
   R3 a0_;
      // initial values of the configuration
      // relative to which all variations will be done

   R3 v_;
      // initial angular velocity in a varied path
   R3 a_;
      // initial angular acceleration in a varied path
   R tau_{0.};
   R eps_{0.};
      // eps_*tau_ will be used as the integration step
      // for computing the action integral from 0 to
      // 2*tau_
   R3 Iinv_;
      // inverse eigen values of the tensor of inertia
// auxiliary functions
   B mod1_,mod2_;
   R lag1(R3 v)const
   {
      return (I_.x1*v.x1*v.x1+I_.x2*v.x2*v.x2+I_.x3*v.x3*v.x3)*0.5;
   }
   R3 J(R3 x, R3 v)const;
   R3 Jinv(R3 x, R3 v)const;
   R lag2(R3 x, R3 v)const{ return lag1((-x).ast(v));}
      // faster version
   //R lag2(R3 x, R3 v)const{ return (v|J(x,v))*0.5;}
      // also correct, but a bit slower.

   void ini_();
      // sets a0_ and a_ according of equation (52) of my Granular Matter
      // article. Also sets v_ equal to v0_.
      // To be used in the constructor only.
public:
   R tau()const{ return tau_;}
   R3 x0()const{ return x0_;}
   R3 v0()const{ return v0_;}
   R3 a0()const{ return a0_;}
   void setMod_(bool mod1, bool mod2){mod1_=mod1; mod2_=mod2;}
   Action(RecordHandler const& rch);
   Action(R3 I, R3 x, R3 v, R tau, R eps=1e-2):
      I_(I),x0_(x),v0_(v),tau_(tau),eps_(eps),
      Iinv_(cpminv(I.x1),cpminv(I.x2),cpminv(I.x3)){ini_();}

   X2<R3,R3> pp(R h)const;
      // returns the angular position and angular velocity at time h
      // along the 'parabolic path' given by x0_,v_,a_
   R lag(R h)const
   { X2<R3,R3> pph=pp(h); return lag2(pph.c1(),pph.c2());}
   R S()const;
      // Integral of lag(h) from h=0 to h=2*tau_
   void var_(R3 da);
      // changes a_ to a0_+da and also v_ to v0_-tau_*da.
      // This combined change of a_ and v_ lets pp(2*tau_) remain fixed.
      // As a mutating function, it has a name ending in '_'
};

X2<R3,R3> Action::pp(R h)const
{
   static Word loc("Action::pp(R)");
   R h4=h*0.25;
   R3 sh=v_*2+a_*h;
   R3 shh=sh*h4;
   R fac1=(shh|shh)+1;
   cpmassert(fac1!=0,loc);
   R3 v=(v_+(a_+v_.crsPrd(a_)*h4)*h)*(R(1)/fac1);

   R fac2=-(shh|x0_)+1;
   cpmassert(fac2!=0,loc);
   R3 x=(x0_+shh+shh.crsPrd(x0_))*(R(1)/fac2);
   return X2<R3,R3>(x,v);
}

void Action::var_(R3 da)
{
   a_=a0_+da;
   v_=v0_-da*tau_;
}

R3 Action::J(R3 x, R3 v)const
{
   R3 e1=x.ast(R3(1,0,0));
   R3 e2=x.ast(R3(0,1,0));
   R3 e3=e1.crsPrd(e2); // saves one evaluation of ast
   return (I_.x1*(e1|v))*e1+(I_.x2*(e2|v))*e2+(I_.x3*(e3|v))*e3;
}

R3 Action::Jinv(R3 x, R3 v)const
{
   R3 e1=x.ast(R3(1,0,0));
   R3 e2=x.ast(R3(0,1,0));
   R3 e3=e1.crsPrd(e2); // saves one evaluation of ast
   return (Iinv_.x1*(e1|v))*e1+(Iinv_.x2*(e2|v))*e2+(Iinv_.x3*(e3|v))*e3;
}

void Action::ini_()
{
   R3 x1= mod1_ ? x0_ :(v0_*tau_).exp().circ(x0_);
   R3 x2= mod2_ ? (v0_*tau_).exp().circ(x0_) : x0_;
   R3 r1=J(x1,v0_);
   R3 r2=r1.crsPrd(v0_);
   a0_=Jinv(x2,r2);
   a_=a0_;
   v_=v0_;

   cpmdebug(x0_);
   cpmdebug(x1);
   cpmdebug(x2);
   cpmdebug(v0_);
   cpmdebug(a0_);
   cpmdebug(tau_);
}

namespace{
R fS(R const& h, Action const& act){ return act.lag(h);}
}// namespace

R Action::S()const
{
#ifdef CPM_Fn
   R_Func f(F1<R,Action,R>(*this)(fS));
#else
   R_Func f(F<R,R>(bind(fS,_1,*this)));
#endif
   return f.antiDer(0,tau_*eps_)(tau_*R(2.));
      // notice that function antiDer returns the
      // anti-derivative of f (as a function, of course)
}

Action::Action(RecordHandler const& rch)
{
   Z mL=1;
   Word loc("Action::Action(RecordHandler)");
   CPM_MA
   R3 I,x,v,a;
   R tau,eps;

   Word sec="action data";
   V<R> eigenvalues;
   cpmrh(eigenvalues);
   I.x1=eigenvalues[1];
   I.x2=eigenvalues[2];
   I.x3=eigenvalues[3];

   V<R> pos;
   cpmrh(pos);
   x.x1=pos[1];
   x.x2=pos[2];
   x.x3=pos[3];

   V<R> vel;
   cpmrh(vel);
   v.x1=vel[1];
   v.x2=vel[2];
   v.x3=vel[3];

   cpmrh(tau);
   cpmrh(eps);

   *this=Action(I,x,v,tau,eps);
   CPM_MZ
}

CpmPalaAction::ActionAppl::ActionAppl(Word const& w):
   IniFileBasedApplication(w)
{}


void CpmPalaAction::ActionAppl::doTheWork()
// RecordHandler rch is available for reading data from it by
// means of the cpmrh-macro
{
   Word sec="topic under consideration";
   Word topic;
   cpmrh(topic);
   if (topic=="action_principle"){
      Action act(rch);
      Word sec="run control 1";
      Z i,nTest;
      cpmrh(nTest);
      V<R> limits;
      cpmrh(limits);
      R3 da(limits[1],limits[2],limits[3]);
      R aScaleRel;
      cpmrh(aScaleRel);
      R aScale=aScaleRel*act.a0().abs();
      da.normalize_(aScale);
      R_Matrix valS(2,nTest);
      R_Matrix valPos(4,nTest);
      R nInv=1./(nTest-1);
      for (i=1;i<=nTest;++i){
         R frac=R(2*(i-1))*nInv-1;
         R3 dai=da*frac;
         act.var_(dai);
         R si=act.S();
         R3 posi=act.pp(act.tau()*2).c1();
         valS[1][i]=frac;
         valPos[1][i]=frac;
         valS[2][i]=si;
         valPos[2][i]=posi.x1;
         valPos[3][i]=posi.x2;
         valPos[4][i]=posi.x3;
      }
      Graph g;
      g.setWithOrigin(false);
      g.show(valS);
      cpmwait(10);
    //  act.run();
   }
   else if (topic=="action_image"){
      Action act(rch);
      Word sec="run control 2";
      using namespace CpmImaging;
      B mod1,mod2,makeMovie;
      Word movieName;
      Z i,j,k,nx,ny,nz;
      R dsRel, aScaleRel,tWait,stretchFactor=1.1,fracH=1;
      cpmrh(mod1);
      cpmrh(mod2);
      cpmrh(makeMovie);
      cpmrh(movieName);
      cpmrh(fracH);
      act.setMod_(mod1,mod2);
      cpmrh(nx);
      cpmrh(ny);
      cpmrh(nz);
      cpmrh(dsRel);
      cpmrh(aScaleRel);
      cpmrh(stretchFactor);
      cpmrh(tWait);
      R aScale=aScaleRel*act.a0().abs();
      R ds=dsRel*act.a0().abs();
      Z m=ny,n=nx; // matrix style indexing
      Image<R> imr(m,n);
// geometry of scanning
      R3 e3=act.v0();
      e3.normalize_(1.);
      R3 iDir=act.x0().ast(R3(0,0,1));
         // this is the instantaneous location of the inertia axis
         // with the largest momentum
// now we form e2 as the component of iDir which is perpendicular to e3
      R3 e2=iDir-e3*(iDir|e3);
      e2.normalize_(1.);
      R3 e1=e2.crsPrd(e3);
      R3 pk;
      R mr=1./(m-1), nr=1./(n-1), zr=1./(nz-1);
// Generating the image 'best in focus' for setting the brightness range
      for (i=1;i<=m;++i){ // Image<> requires matrix indexing
         R fraci=R(2*(i-1))*mr-1;
         fraci*=aScale;
         R3 pi=pk+e2*fraci;
         for (j=1;j<=n;++j){
            R fracj=R(2*(j-1))*nr-1;
            fracj*=aScale;
            R3 pij=pi+e1*fracj;
            act.var_(pij);
            R as=act.S();
            imr.putPix(i,j,as);
         }
      }
      R2 mm=imr.minMax();
      cpmdebug(mm);
      R diff=mm[2]-mm[1];
      cpmdebug(diff);
      Iv iv1(mm[1],mm[2]);
      iv1=iv1.stretch(stretchFactor);
      Iv iv2(0,255);
      Response res(iv1,iv2,1,6);
      Frame frm0;
      Color cTopic(Z(200),Z(100),Z(80));
      frm0.paint(cTopic);
      frm0.vis();
      Frame frm;
      R dy=fracH;
      R yC=(-dy+1)*0.5;
      R aspRat=R(nx)/R(ny);
      frm.setxy(yC,yC,dy,dy);
      frm.setAspRatHoldHeight(aspRat);
      imr.mark(frm,res,1);
      frm.setText("cental frame which sets the color code",
         0.1,0.9,Color(Z(255),Z(255),Z(255)));
      frm.vis();
      cpmwait(tWait*2+2);
      for (k=1; k<=nz;++k){
         R frack=R(2*(k-1))*zr-1;
         frack*=ds;
         pk=e3*frack;
// now we have an orthogonal tripod with e3 having the direction of the
// angular velocity.
         for (i=1;i<=m;++i){ // Image<> requires matrix indexing
            R fraci=R(2*(i-1))*mr-1;
            fraci*=aScale;
            R3 pi=pk+e2*fraci;
            for (j=1;j<=n;++j){
               R fracj=R(2*(j-1))*nr-1;
               fracj*=aScale;
               R3 pij=pi+e1*fracj;
               act.var_(pij);
               R as=act.S();
               imr.putPix(i,j,as);
            }
         }
         mm=imr.minMax();
         cpmdebug(mm);
         diff=mm[2]-mm[1];
         cpmdebug(diff);

         imr.mark(frm,res,1);
         frm.setText("frame "&cpm(k));
         frm.vis();
         if (makeMovie) frm.write(movieName&"_"&cpm(k,2));
         cpmwait(tWait);
      }
   }
   else{ // may be extended
      cpmmessage("nothing to do");
   }
}
