//? cpmrigbodydyn.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#include <cpmrigbodydyn.h>

#if defined(CPM_DIMDEF3) // otherwise the file has void content
#include <cpmparens.h>
#include <cpmframes.h>
#include <cpmgreg.h>
#include <cpmvectormore.h>

using namespace CpmPala1_3;
using namespace CpmDim3;
using namespace CpmStd;

using CpmProperties::ParEnsPrp;
#ifdef CPM_Fn
using CpmFunctions::F4;
#endif
using CpmGraphics::Frames;
using CpmTime::Greg;
using CpmPhysics3::VecE;
using CpmPhysics3::VecBH;
using CpmPhysics3::VecG;

/////////////////////// struct DynCon //////////////////////////////////
DynCon::DynCon(RecordHandler const& rch, Word const& sec)
// written such that for each item not readable from file
// a useful default value is provided
{
   V<R> fieldE(3), fieldH(3), fieldG(3);
   // initialized
   Z sof=1;
   stepsForTypTime=200;
   rch.read(sec,"stepsForTypTime",stepsForTypTime,sof);
   relTotTime=1;
   rch.read(sec,"relTotTime",relTotTime,sof);
   fitFactor=1.05;
   rch.read(sec,"fitFactor",fitFactor,sof);
   tWait=0.01;
   rch.read(sec,"tWait",tWait,sof);
   viwPer=1;
   rch.read(sec,"viewingPeriod",viwPer,sof);
   filPer=0;
   rch.read(sec,"filingPeriod",filPer,sof);
   imgPer=0;
   rch.read(sec,"imgagingPeriod",imgPer,sof);
   trapPow=4;
   rch.read(sec,"trapPow",trapPow,sof);
   lambda=0;
   rch.read(sec,"lambda",lambda,sof);
   rch.read(sec,"fieldE",fieldE,sof);
   rch.read(sec,"fieldH",fieldH,sof);
   rch.read(sec,"fieldG",fieldG,sof);
   VecE fE(fieldE);
   VecBH fH(fieldH,"H");
   VecG fG(fieldG);
   fs=FieldSpace(fE,fH,fG);
   aCritRel=-1;
   vCritRel=-1;
      // these values are OK since they disable the
      // v,a-limitation mechanism
   rch.read(sec,"aCritRel",aCritRel,sof);
   rch.read(sec,"vCritRel",vCritRel,sof);
   nRunsForOrd=1;
   rch.read(sec,"nRunsForOrd",nRunsForOrd,sof);
   if (nRunsForOrd>1){
      // in this case it is essential to compute as fast as
      // possible
      filPer=0;
      imgPer=0; // reasonable
   }
}

/////////////////////// class RigBodyDyn ///////////////////////////////

namespace{

const Color cRigBodyDyn(Z(180),Z(225),Z(230));

Vec trap1Func(Spc const& sp, Spc const& s0, R const& r0, R const& p,
   R const& lambda)
{
   cpmassert(r0>0,"trap1Func(...)");
   R r0Inv=R(1)/r0;
   Vec x=(sp-s0);
   R r=x.nor_();
   r*=r0Inv;
   R fac=-p*r0Inv*cpmpow(r,p-1)*lambda;
   return x*fac;
}

R trap2Func(Spc const& sp, Spc const& s0, R const& r0, R const& p,
   R const& lambda)
{
   cpmassert(r0>0,"trap2Func(...)");
   R r0Inv=R(1)/r0;
   Vec x=(sp-s0);
   R r=x.abs()*r0Inv;
   return cpmpow(r,p)*lambda;
}

} // namespace

void RigBodyDyn::updConQun_()
{
   mom_=Vec();
   angMom_=AxVec();
   enr_=0;
   Spc ori;
   for (Z i=1;i<=n_;++i){
      ParT pi=vp_[i];
      R mi=pi.get_m();
      Spc xi=pi.get_x();
      mom_+=pi.mom();
      angMom_+=pi.angMom(ori);
      enr_+=pi.kinEnr();
      enr_+=pi.pot(dc_.fs);
         // contribution from a homogeneous field
      R trp=trapPot_(xi)*mi;
      enr_+=trp;
         // contribution from the trapping potential
      for (Z j=i+1;j<=n_;++j) enr_+=pi.pot(vp_[j]);
   }
}

BndSph RigBodyDyn::bndSph()const
{
   using CpmRoot::abs;
   R mass=0;
//   R energy=0;
   V<Spc> vs(n_);
   V<R> masses(n_);
   R rMax=0;
   Z i;
   for (i=1;i<=n_;++i){
      ParT pi=vp_[i];
      R ri=pi.get_r();
      if (ri>rMax) rMax=ri;
      R mi=pi.get_m();
      masses[i]=mi;
      mass+=mi;
      Spc xi=pi.get_x();
      vs[i]=xi;
   }
   bool careful=true;
   Spc xcm=Spc::convex(vs,masses,careful);

   R_Vector ri(n_); // certainly an even number
   for (i=1;i<=n_;++i) ri[i]=(vs[i]-xcm).abs();
 //  R length=ri.mean()+rMax;
      // here rMax is the maximum radius of a particle
      // without the rMax contribution
      // this would be 0 for a single particle
   rMax+=ri.sup();
      // now rMax is the one defining the bounding sphere
   return BndSph(xcm,rMax);
}

TypQun RigBodyDyn::typQun()const
   // typical energy,mass,length,time, velocity determined by the system
{
   using CpmRoot::abs;
   TypQun res;
   R mass=0;
   R energy=0;
   V<Spc> vs(n_);
   V<R> masses(n_);
   R rMax=0;
   Z i;
   for (i=1;i<=n_;++i){
      ParT pi=vp_[i];
      R ri=pi.get_r();
      if (ri>rMax) rMax=ri;
      R mi=pi.get_m();
      masses[i]=mi;
      mass+=mi;
      Spc xi=pi.get_x();
      vs[i]=xi;
      energy+=pi.kinEnr();
      R trp=trapPot_(xi)*mi;
      energy+=cpmabs(trp);
      R fld=pi.pot(dc_.fs);
      energy+=cpmabs(fld);
         // contribution from the trapping potential
      for (Z j=i+1;j<=n_;++j) energy+=cpmabs(pi.pot(vp_[j]));
   }
   bool careful=true;
   Spc xcm=Spc::convex(vs,masses,careful);

   R_Vector ri(n_); // certainly an even number
   for (i=1;i<=n_;++i) ri[i]=(vs[i]-xcm).abs();
   R length=ri.mean()+rMax;
      // here rMax is the maximum radius of a particle
      // without the rMax contribution
      // this would be 0 for a single particle
   rMax+=ri.sup();
      // now rMax is the one defining the bounding sphere

   cpmassert(mass>0,"RigBodyDyn::typQun()");
   R velocity=cpmsqrt(energy*2/mass);
   cpmassert(velocity>0,"RigBodyDyn::typQun()");
   R time=length/velocity;
   res.enr=energy;
   res.len=length;
   res.mass=mass;
   res.time=time;
   res.vel=velocity;
   res.cenOfMass=xcm;
   res.bonSph=SubsetOfSpace::sphere(xcm,rMax);
   return res;
}

void RigBodyDyn::adjust_()
{
   R p=dc_.trapPow;
   R lambda=dc_.lambda;
   n_=vp_.dim();
   for (Z i=1;i<=n_;++i) vp_[i].set_id(i);
   accPar_=V<ForT>(n_);
   BndSph bs=bndSph();
      // from the typical quantities we presently only
      // need the diameter
   R r0=bs.get_r();
   Spc x0=bs.get_x();
#ifdef CPM_Fn
   trapAcc_=F4<Spc,Spc,R,R,R,Vec>(x0,r0,p,lambda)(trap1Func);
   trapPot_=F4<Spc,Spc,R,R,R,R>(x0,r0,p,lambda)(trap2Func);
#else
   trapAcc_=F<Spc,Vec>(bind(trap1Func,_1,x0,r0,p,lambda));
   trapPot_=F<Spc,R>(bind(trap2Func,_1,x0,r0,p,lambda));
#endif
   viws_=CyclicAlarm(dc_.viwPer);
   fils_=CyclicAlarm(dc_.filPer);
   imgs_=CyclicAlarm(dc_.imgPer);
   updConQun_();
}

bool RigBodyDyn::prnOn(ostream& str)const
{
   str; return true;
}

bool RigBodyDyn::scanFrom(istream& str)
{
   str; return true;
}

void RigBodyDyn::step_(R h)
// Same logic as function void CPM_FCN::steplet(R t0, R dt, const Camera&
// cam) in  cpmpardynsys.h
{
   Z mL=2;
   static Word loc("RigBodyDyn::step_(R)");
   static R hMem=0;
   static Z nMem=0;
   bool b=false;
   if (h!=hMem){
      b=true;
      hMem=h;
   }
   if (n_!=nMem){
      b=true;
      nMem=n_;
   }
   if (b) cpmvalues2("step_: dt, np",h,n_,2);
   dt_=h;
   SysControl::tStep=dt_;
   if (dt_==R(0.)) return;
// Free evolution to the midpoint in time: direct midpoint method
   Z i,j;
   R tau1=R(0.5)*dt_;
   R tau2=dt_-tau1;
   Z n=n_;
   for (i=1;i<=n;i++){
      vp_[i].freeStep(tau1);
      // notice that freeMot, which is used in cpmpardynsys.h
      // only differs in the handling of the argument
      vp_[i].phase1(tau1);
   }
// collecting the interaction terms at the mid-point in time
// don't move particles yet, otherwise dynamics would depend on
// numbering of particles
   V<ForT> forces(n);
      // forces[i]=force acting on particle par[i]
   for (i=1;i<=n;i++){ // first particle loop
      ParT pi=vp_[i];
      forces[i]=pi.forces(dc_.fs);
      Spc xi=pi.get_x();
      R mi=pi.get_m();
      Vec trapi=trapAcc_(xi)*mi;
         // force from a trapping potential
      forces[i].addForce(trapi);
      for (j=1;j<=n;++j){
         if (j==i) continue;
         forces[i]+=pi.forces(vp_[j]);
      }
   } // end of first particle loop
// moving the particles according to interaction
   t_+=tau1;
   for (i=1;i<=n;i++){ // second particle loop
      vp_[i].intAct(dt_,forces[i]);
         // Changes velocities according to forces.
      vp_[i].freeStep(tau2);
         // executes the shift in position and angular orientation
         // according to the full acceleration and angular acceleration
         // that the particle feels (no regularization) at this place.
      vp_[i].reg();
         // regularization of the motion by reducing the velocity
         // if this is much larger than SysControl::vCrit
   } // end of second particle loop
   t_+=tau2;
   stepCon_+=1;
   updConQun_();
   CPM_MZ
}

namespace{

   R logS(R x){ return x<1e-137 ? -R(137) : cpmlog10(x);}
}

void RigBodyDyn::run(Camera& cam, ostream& str,Word const& name)
{
   using namespace CpmImaging;
   using namespace CpmArrays;
   using namespace CpmSystem;

   Z mL=1;
   Word loc="RigBodyDyn::run(Camera, ostream)";
   CPM_MA

   Z n=n_;
   R tWait=dc_.tWait;

   Z stepsForTypTime=dc_.stepsForTypTime;
   R relTotTime=dc_.relTotTime;
   Z nRunsForOrd=dc_.nRunsForOrd;
   Z i,nSteps=cpmrnd(R(stepsForTypTime)*relTotTime);

// showing the initial condition
   Graph g;
   g.paint(cRigBodyDyn);
   g.vis();
// size of the particle system

   TypQun tq=typQun();
   R typE=tq.enr;
   R typL=tq.len;
   R typM=tq.mass;
   R typT=tq.time;
   R typV=tq.vel;
   Greg date;
   date.now();
   str<<endl<<date.getDate(false);
   str<<endl<<"This is output from run "<<name<<endl;
   str<<"Number of particles: "<<n<<endl;
   str<<
   "These are the typical quantities of the particle system (in SI units)"
   <<endl;
   str<<"energy="<<typE<<endl;
   str<<"length="<<typL<<endl;
   str<<"mass="<<typM<<endl;
   str<<"time="<<typT<<endl;
   str<<"velocity="<<typV<<endl;

// adjusting the camera to the particle system
   RefSys rs=cam.getRefSys();
   R fitFactor=dc_.fitFactor;
   Angle tilt(0);
   Angle pan(0);

   SubsetOfSpace bs=tq.bonSph;
   cam.dirTo_(bs,fitFactor,tilt,pan,rs);
      // by putting in the orientation of cam as the last argument
      // we need no additional tilt and pan
   for (i=1;i<=n;++i) vp_[i].mark(cam);
   cam.vis();
   cpmwait(tWait*10);

   SysControl::vCrit=dc_.vCritRel*typV;
   R typAcc=typV/typT;
   SysControl::aCrit=dc_.aCritRel*typAcc;

   if (nSteps<1) return; // it is often useful to show the initial
   // condition without invoking dynamics

   RigBodyDyn stateMem(*this);

   R tTot=relTotTime*typT;

   R_Vector dT(nRunsForOrd);
   R_Vector devE(nRunsForOrd);
   R_Vector devM1(nRunsForOrd);
   R_Vector devM2(nRunsForOrd);
   R_Vector devM3(nRunsForOrd);
   R_Vector devA1(nRunsForOrd);
   R_Vector devA2(nRunsForOrd);
   R_Vector devA3(nRunsForOrd);

   R_Vector times, energies, mom1, mom2, mom3,
      amom1, amom2, amom3;
   Z countAct=0;
   R xRel=0.02;

   for (Z iRun=1;iRun<=nRunsForOrd;++iRun){
      *this=stateMem;
      R dt=tTot/nSteps;
      dT[iRun]=logS(dt);
      Z nStepsMod=nSteps+2;
      str<<"total time="<<tTot<<endl;
      str<<"dt="<<dt<<endl;
      str<<"vCrit="<<SysControl::vCrit<<
      " negative values have no effect"<<endl;
      str<<"aCrit="<<SysControl::aCrit<<
      " negative values have no effect"<<endl;

      times=R_Vector(nStepsMod);
      energies=R_Vector(nStepsMod);
      mom1=R_Vector(nStepsMod);
      mom2=R_Vector(nStepsMod);
      mom3=R_Vector(nStepsMod);
      amom1=R_Vector(nStepsMod);
      amom2=R_Vector(nStepsMod);
      amom3=R_Vector(nStepsMod);

      countAct=1;
      Z countImg=0;

      R e0=enr();
      Vec vec=mom();
      AxVec avec=angMom();

      R m1=vec[1];
      R m2=vec[2];
      R m3=vec[3];

      R am1=avec[1];
      R am2=avec[2];
      R am3=avec[3];

      R typEinv= R(1)/typE;
      R typMomInv= R(1)/(typM*typV);
      R typAngMomInv= typMomInv/typL;

      times[countAct]=t();
      energies[countAct]=0;
      mom1[countAct]=0;
      mom2[countAct]=0;
      mom3[countAct]=0;

      amom1[countAct]=0;
      amom2[countAct]=0;
      amom3[countAct]=0;

      R nStepsInv=1./nSteps;

      R tStart=cpmtime();

      while (stepCon_<nSteps){  // one state already registered
         // so we show nSteps more, making up nSteps+1 states and nSteps
         // integration steps
         R prog=R(stepCon_)*nStepsInv;
         cpmprogress("Progress",prog,3,3);
         step_(dt);
         cam.clear();
// graphics part
         if (viwAlarm_()){
            countAct++;
            times[countAct]=t();

            energies[countAct]=(enr()-e0)*typEinv;
            vec=mom();
            avec=angMom();
            mom1[countAct]=(vec[1]-m1)*typMomInv;
            mom2[countAct]=(vec[2]-m2)*typMomInv;
            mom3[countAct]=(vec[3]-m3)*typMomInv;

            amom1[countAct]=(avec[1]-am1)*typAngMomInv;
            amom2[countAct]=(avec[2]-am2)*typAngMomInv;
            amom3[countAct]=(avec[3]-am3)*typAngMomInv;

            cam.clear();
            SysControl::tStep*=viws_.per();
               // for showing the velocity in vp_[i].mark(cam)
            for (i=1;i<=n;++i) vp_[i].mark(cam);
            cam.vis();

            if(imgAlarm_()){
               countImg++;
               Word imf="RigBodyDyn_run";
               cam.getGraph().write(imf&cpm(countImg,3));
            }
            cpmwait(tWait,2);
         } // viwAlarm_
      }

      R tEnd=cpmtime();
      R tCom=tEnd-tStart;
      str<<"Main computation time ="<<tCom<<" seconds"<<endl;

      times=times.resize(countAct);
      energies=energies.resize(countAct);
      mom1=mom1.resize(countAct);
      mom2=mom2.resize(countAct);
      mom3=mom3.resize(countAct);
      amom1=amom1.resize(countAct);
      amom2=amom2.resize(countAct);
      amom3=amom3.resize(countAct);

      R mom1S=mom1.supAbs();
      R mom2S=mom2.supAbs();
      R mom3S=mom3.supAbs();
      R amom1S=amom1.supAbs();
      R amom2S=amom2.supAbs();
      R amom3S=amom3.supAbs();
      R energiesS=energies.supAbs();
      devE[iRun]=logS(energiesS);
      devM1[iRun]=logS(mom1S);
      devM2[iRun]=logS(mom2S);
      devM3[iRun]=logS(mom3S);
      devA1[iRun]=logS(amom1S);
      devA2[iRun]=logS(amom2S);
      devA3[iRun]=logS(amom3S);

    str<<endl<<"Data from run #"<<iRun<<" dt="<<dt<<endl
   <<"Maximum relative deviation of Px from initial value "<<mom1S<<endl
   <<"Max. relative deviation of Py from initial value "<<mom2S<<endl
   <<"Max. relative deviation of Pz from initial value "<<mom3S<<endl
   <<"Max. relative deviation of Lx from initial value "<<amom1S<<endl
   <<"Max. relative deviation of Ly from initial value "<<amom2S<<endl
   <<"Max. relative deviation of Lz from initial value "<<amom3S<<endl
   <<"Max. rel. deviation of energy from init. value "<<energiesS<<endl;

      g.clearText();
      g.paint(cRigBodyDyn);
      g.vis();
      g.setAutoScale(1);

      Frames frs(g,Z(2),Z(3));
      Graph g11(frs(1,1));
      Graph g12(frs(1,2));
      Graph g13(frs(1,3));

      Graph g21(frs(2,1));
      Graph g22(frs(2,2));
      Graph g23(frs(2,3));

      g11.setWithOrigin(0);
      g12.setWithOrigin(0);
      g13.setWithOrigin(0);

      g21.setWithOrigin(0);
      g22.setWithOrigin(0);
      g23.setWithOrigin(0);

      Word tit;
      Color c(RED);
      tit="Relative dev. of Px from initial value";
      g11.show(times,mom1,c,0,tit,xRel);

      tit="Rel. deviation of Py from init. value";
      g12.show(times,mom2,c,0,tit,xRel);

      c=Color(GREEN);
      tit="Rel. dev. of energy from init. value";
      g13.show(times,energies,c,0,tit,xRel);

      c=Color(CYAN);
      tit="Rel. dev. of Lx from init. value";
      g21.show(times,amom1,c,0,tit,xRel);

      tit="Rel. dev. of Ly from init. value";
      g22.show(times,amom2,c,0,tit,xRel);

      tit="Rel. dev. of Lz from init. value";
      g23.show(times,amom3,c,0,tit,xRel);

    //  g.write(name&"_run_"&cpm(iRun));
      nSteps*=2;
      stepCon_=0;
      cpmwait(SysControl::tWait,2);
   } // loop over iRun
// processing devE,....
   using CpmArrays::polyFit;
   if (nRunsForOrd>1){
      R_Vector fE=polyFit(dT,devE,1,0,1,false);

      R_Vector fM1=polyFit(dT,devM1,1,0,1,false);
      R_Vector fM2=polyFit(dT,devM2,1,0,1,false);
      R_Vector fM3=polyFit(dT,devM3,1,0,1,false);

      R_Vector fA1=polyFit(dT,devA1,1,0,1,false);
      R_Vector fA2=polyFit(dT,devA2,1,0,1,false);
      R_Vector fA3=polyFit(dT,devA3,1,0,1,false);

      R rE=fE[4];

      R rM1=fM1[4];
      R rM2=fM2[4];
      R rM3=fM3[4];

      R rA1=fA1[4];
      R rA2=fA2[4];
      R rA3=fA3[4];

   str<<endl<<
      "Order of dependence on dt for the following quantities:"
      <<endl
   <<"Maximum relative deviation of Px from initial value "<<rM1<<endl
   <<"Max. relative deviation of Py from initial value "<<rM2<<endl
   <<"Max. relative deviation of Pz from initial value "<<rM3<<endl
   <<"Max. relative deviation of Lx from initial value "<<rA1<<endl
   <<"Max. relative deviation of Ly from initial value "<<rA2<<endl
   <<"Max. relative deviation of Lz from initial value "<<rA3<<endl
   <<"Max. rel. deviation of energy from init. value "<<rE<<endl;
      g.clearText();
      g.paint(cRigBodyDyn);
      g.vis();
      Frames frs(g,Z(2),Z(3));
      Graph g11(frs(1,1));
      Graph g12(frs(1,2));
      Graph g13(frs(1,3));

      Graph g21(frs(2,1));
      Graph g22(frs(2,2));
      Graph g23(frs(2,3));

      g11.setWithOrigin(0);
      g12.setWithOrigin(0);
      g13.setWithOrigin(0);

      g21.setWithOrigin(0);
      g22.setWithOrigin(0);
      g23.setWithOrigin(0);

      Color c(RED);
      Word tit;
      tit="Relative dev. of Px from initial value";
      g11.show(dT,devM1,c,0,tit,xRel);

      tit="Rel. deviation of Py from init. value";
      g12.show(dT,devM2,c,0,tit,xRel);

      c=Color(GREEN);
      tit="Rel. dev. of energy from init. value";
      g13.show(dT,devE,c,0,tit,xRel);

      c=Color(CYAN);
      tit="Rel. dev. of Lx from init. value";
      g21.show(dT,devA1,c,0,tit,xRel);

      tit="Rel. dev. of Ly from init. value";
      g22.show(dT,devA2,c,0,tit,xRel);

      tit="Rel. dev. of Lz from init. value";
      g23.show(dT,devA3,c,0,tit,xRel);

      cpmwait(R(nRunsForOrd)*SysControl::tWait,2);
      // reasonable to have time to study this
   }
   CPM_MZ
}

////////////////////// class App1RigBodyDyn ////////////////////////////

namespace{
   const R rimPaLa=0.;
   const Color cPaLa(Z(150),Z(150),Z(255));
}

void CpmPala1_3::App1RigBodyDyn::doTheWork()
{
   Z mL=1;
   Word loc("CpmPala1_3::App1RigBodyDyn::doTheWork()");
   CPM_MA
   Graph g;
   g.setWithOrigin(0);
   Word sec="camera control";
   Camera cam(rch,sec,g);

   sec="syscontrol";
   SysControl::read(rch,sec);

   ParEnsPrp gp(rch.getRecord());
   sec="grain properties";
   R gpdC=gp.getR(sec,"dC");
   R gpdT=gp.getR(sec,"dT");
   R dgp=gpdC+gpdT;
////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////
   R facBox=1,facVol=0.1,ranBoo=0,ranSpin=0;
   Z cutTo=-1;
   B showInertiaAxes=B(false);
 //  B newRigBodyInt=B(true);
   B balance=B(false);

   cpmrh(facBox);
   cpmrh(facVol);

   cpmrhf(ranBoo);
   cpmrhf(ranSpin);
   cpmrhf(cutTo);
   cpmrhf(showInertiaAxes);
   cpmrhf(balance);
   CpmRigidBody3::markComParMec=showInertiaAxes;
   CpmRigidBody3::balance=balance;

   R3 b0;
   R3 bs(dgp,dgp,dgp);
   bs*=facBox;
   Box box(b0,bs);

   Z nLim=64;
      // in order allow making many particles

   CpmRigidBody3::ParEns<ParT> pe(gp,box,facVol,nLim);
   Iv v0(-ranBoo,ranBoo);
   Iv om(0,ranSpin);
   V<ParT> vcp=pe();

   Z i,np=vcp.dim();
   for (i=1;i<=np;++i){
      Vec vb;
      vb[1]=v0.ran();
      vb[2]=v0.ran();
      vb[3]=v0.ran();
      vcp[i].boost(vb);
      AxVec avs;
      avs[1]=om.ran();
      avs[2]=om.ran();
      avs[3]=om.ran();
      vcp[i].spin(avs);
   }
   if (cutTo>0) vcp=vcp.resize(cutTo);
   sec="run control";
   DynCon dc(rch,sec);
   R tWaitFinalScreen;
   cpmrhf(tWaitFinalScreen);
   set_tWaitWorkScreen_(tWaitFinalScreen);

   RigBodyDyn rbd(dc,vcp);
   OFileStream out1(outFiles[1]);
   rbd.run(cam,out1(),programName_&"_"&runName_);
   cpmwait(tWaitWorkScreen_,2);
   CPM_MZ
}
#endif // CPM_DIMDEF3

