//? cpmbranchstate.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_STATE_H_
#define CPM_STATE_H_
/*
   cpmstate.h
   version  98-10-11
   Author: Ulrich Mutze
      ulrichmutze@aol.com

   Description: Development and test of a class
      that represent quantum mechanical pure states.
   History: Started 97-10-29
      2000-1-30 StateSpace-->State
      2004-08-06 implementation of C operator()(const Z& i)const
         corrected.
         Implementation of creation and destruction operators
         finished and commutation relations tested.
      2004-09-22: function void State::showAsCurves(Graph) added
      2004-09-25: function adjPhase(...) reimplemented so that
         the previously present (obscure) second argument is eliminated
      2007-06-11 QMBase::dtStabilityLimit, QMBase::eMax introduced

*/

#include <cpmv.h>
#include <cpmc.h>
#include <cpmtypes.h>
#include <cpminterfaces.h>
#include <cpmrfunc.h>
#include <cpmconstphys.h>
#include <cpmgraph.h>

using CpmRoot::Z;
using CpmRoot::R;
using CpmRoot::C;
using CpmGraphics::Frame;
using CpmGraphics::Graph;
using namespace std;
using namespace CpmArrays;
using namespace CpmFunctions;

// interfaces as declaration macros

#define CPM_PURE_VECTOR_SPACE_INTERFACE(X,S)\
public:\
   X operator - (void)const;\
   X& operator *=(S);\
   X& operator +=(const X&);\
   X& operator -=(const X&);\
   X operator *(S)const;\
   X operator +(const X&)const;\
   X operator -(const X&)const;

#define CPM_PURE_SCALAR_INTERFACE(X,S)\
public:\
    X& operator *=(S);\
   X operator *(S)const;

#define CPM_SCALAR_PRODUCT_INTERFACE(X,S)\
public:\
   X conj(void)const;\
   S operator |(const X&)const;\
   R normSquared(void)const{ return (operator|(*this)).real();}\
   R norm(void)const{ return ::sqrt(normSquared());}\
   R abs(void)const{ return norm();}\
   R absSqr(void)const{ return normSquared();}

namespace CpmQuantumMechanics{

class QMBase{

protected:

// control quantities
   static Z verbose;
      // controls messages

   static R phiStep;
      // Limit to the phase angle that a single sep of the
      // direct midpoint integrator is allowed to achieve.
      // 0.01 seems to be a good value (corresponds to 0.572 angular
      // degrees. Or is it 3.6 degrees which would better agree with
      // experience in mechanical equations)

// independent physical quantities

   static R c;
      // velocity of light in vacuum

   static R hDirac;
      //=CpmPhysics::h__PHYS;
      // My reference for using the name 'hDirac' for h-bar
      // is Landau Lifschitz: Lehrbuch der Theoretischen Physik,
      // Band 3, Seite 20, Fussnote

   static R mass;
      //e.g. CpmPhysics::me_PHYS

   static R size;
      // length of the spatial periodicity interval

// dependent physical quantities

   static R hDiracInv;
      // 1/hDirac

   static R hPlanck;
      // hDirac*2*pi, the original of 1900

   static C ihDirac;
      // C(0,hDirac)

   static C ihDiracInv;
      // C(0,-1/hDirac)

   static R ihDiracInv2;
      // -1/(hDirac*hDirac);

// Dependent physical quantities that also depend on the number d
// of discretization-points. This number d could probably also be
// usefully a static data member of the present class. I hesitate to do
// treat it this way since this would reduce flexibility (see function
// State::resize(Z) and the requirement for d to be even in cases
// where the momentum operator is used) and would induce extended code
// revision.

   static V<R> positions(Z d);
      // discrete positions associated with an interval [-size/2,size/2]
      // subdivided into d congruent sub-intervals. The positions
      // returned here are the midpoints of these intervals. Notice that
      // the spacing of these mid-points is size/d.
      // Notice that this picture differs a bit from the naive one in
      // which one would consider d equidistant points where the first
      // and the last one coincide with the boundaries of the interval.
      // Then the d points would divide the interval into d-1 congruent
      // subintervals and the point spacing thus would be size/(d-1).
      // Wave functions are value distributions by nature and thus are
      // associated with intervals rather than points. Therefore we don't
      // use the naive picture.

   static V<R> posNor(Z d);
      // 'positions normalized'. As above with size==1

   static V<R> frequencies(Z d);
      // spatial frequencies associated with a linear lattice of the kind
      // described in the comment to function positions.
      // See Press et a. p. 503 between eq. (12.1.8) and (12.1.9).
      // This function requires d to be an even positive number.
      // The maximum value is the Nyquist frequency fc=d/(2*size).
      // The shortest 'representable wavelength' on the lattice is
      // 1/fc = (2*size)/d. There is a mass mc for which this smallest
      // wavelength just equals its Compton-wavelength (=hDirac/(m*c)).
      // (this actually is abuse of language; hPlank/(m*c) is the real
      // Compton wavelength (Weizel and Kaellen agree).
      // T psi = -(hDirac^2/(2*m)) psi"
      // If we use mc for m and the numerical difference formula for
      // step size = lattice spacing s (s=size/d), we get
      // (T psi)[i] =
      // (2*psi[i]-psi[i+1]-psi[i-1])*d^2*hDirac^2/(2*mc*size^2) =
      // (2*psi[i]-psi[i+1]-psi[i-1])*(hDirac*c/s).
      // The factor (hDirac*c/s)is a energy.
      // It is the energy uncertainty associated with a limitation
      // of any quantum system to a spatial interval of length s
      // hDirac ~ dE*dt = dE *(s/c) ==> dE  ~ hDirac*c/s.
      // In a system of units with hDirac=c=1 we thus have
      // T psi = s*(2*psi[i]-psi[i+1]-psi[i-1])/(s*s) =
      // -s*psi" = -(size/d)*psi". If our system also sets s=1
      // (discretization length as unit) then T psi = - psi".
      // Notice that masses and Compton wavelengths are inversely
      // proportional. Since we adjusted to the smallest possible
      // wave length, we also adjusted for the largest reasonable mass.
      // So all masses to be dealt with on the lattice of length size
      // and d points should be  s m a l l e r  than mc.
      //  m <= mc= d*hDirac/(2*size*c) = hDirac/(2*s*c).
      // Again, in a system of units in which s=1 (together with
      // hDirac=1,c=1) the maximum representable mass is 1/2. Using such
      // a system one has T psi=-psi" for the wave function of this maximum
      // mass. This canonical result would not be obtained if we
      // had related the mass to the minimum wave length via the
      // histotically 'true' Compton wavelength formula.

public: // the following four functions need to be public; this can't
   // do any harm since they don't allow to make changes

   static R mMax(Z d);
      // hDirac*d/(2*size*c) = pMax/(2*Pi*c). This mass has a better
      // relation to the motion on our lattice (by giving the
      // Schroedinger equation the form of the most simple form
      // in natural units) than the 6.28 times larger mass
      // pMax/c. Proposal for discretization: All masses should
      // be smaller than this value. See the extended comment to
      // static V<R> frequencies(Z d) for this issue.

   static R pMax(Z d);
      // pMax=hPlanck*d/(2*size)= (hPlank/2)/ds (ds=size/d)
      //  = pi*hDirac/ds.
      // The momentum associated with the highest spatial frequency fc
      // which can be  represented with equi-distant positions with
      // spacing size/d. Notice that fc=frequencies(d)[d/2]=d/(2*size).

   static R vMax(Z d);
      // vMax(d)=pMax(d)/mass
      // velocity associated with the momentum pMax(d)
      // Notice that a boost operation is essentially a shift operation
      // in momentum space so that state components that get shifted
      // (formally) to values larger than the upper end of the momentum
      // range will enter the momentum range at the other end.
      // Thus applying a boost with vMax to a Gaussian wave packet (at
      // rest) will result in a wave packet with momentum components
      // arround fc and -fc. Under time evolution this packet will split
      // into two beams; for boosts with velocities close to vMax, these
      // two beams are also obvious but their relative strength varies
      // with v.

   static R eMax(Z d);
      // eMax=(2/m)*(hDirac/ds)^2, ds=(size/d); maximum kinetic energy in
      // free motion if the Hamiltonian is realized by the wellknown
      // differencing scheme for the second derivative.
      // If one writes this as p^2/(2*m), we get p=2*hDirac/ds
      // This p, which also is a kind of maximal momentum turns out to be
      // smaller than pMax as defined above. It is smaller in the way
      // 2 is smaller than pi. This is the least-surprising direction ---
      // the opposit would say that the fastest possible motion would
      // have momenta exceeding pMax. For the Hamilton operator defined
      // in momentum space (by discrete Fourier transformation) no such
      // discrepency would occur.

   static R tRef(Z d);
      // size/vMax(d)
      // time needed for traversing the distance size when travelling at
      // speed vMax(d)

   static R dtLimit(Z d);
      // time needed for traversing the distance size/d when travelling
      // at speed vMax(d)

   static R dtStabilityLimit(Z d);
      // Stability limit for the direct midpoint method
      // dtStabilityLimit*eMax=2*hDirac
      // Actually here the actual norm of the Hamilton operator should
      // stand instead of the free contributions eMax.

   static R lambdaOsc(Z d);
      // (hPlanck*d/(size^2))^2 / (2*mass)
      // Consider the Hamiltonian of a harmonic oscillator:
      // Observable p("p",d), x("x",d);
      // Observable h=p*p/(2*mass)+x*x*lambda;
      // Then, for lambda=QMBase::lambdaOsc(d), a state of maximum
      // momentum will just have the kinetic energy lambda*(size/2)^2.
      // A fictitious state having this momentum and being localized near
      // x=0, would oscillate with an amplitude size*0.5.
      // Let us write  mass = xi*mMax, then one gets
      // lambdaOsc(Z d) = ((2*Pi)^2 * c * hDirac)/(xi * size^3),
      // which is independent of d. This is not too surprising since d is
      // inherent in our parameterization of the  mass as xi*mMax.

protected:

   static C facPhi;
   static C facPsi;

   static void adjust();
      // used to update all quantities that depend on hDirac

// Some functions needed to build expressions
public:
   static void accFunc(C&, const C&);
      // z1 -> z1+=z2

   static C conjFunc(const C& );
      // z.conj()

   static C negFunc(const C&);
      // -z

   static C sumFunc(const C&, const C&);
      // z1+z2

   static C diffFunc(const C&, const C&);
      // z1-z2

   static C prodFunc(const C&, const C&);
      // z1*z2

   static C prodFuncReal(const C&, const C&);
      // z1*Re(z2)

   static C prodFuncConj1(const C&, const C&);
      // z1.conj()*z2

   static C prodFuncConj2(const C&, const C&);
      // z1*z2.conj()

   static C divFunc(const C&, const C&);
      // z1/z2

   static R rt(Z i); //  i<0 ? 0 : ::sqrt(R(i))

   static void setUnits(Word const& w);
      // w ="SI" sets SI-Unites everything else sets "natural units"
      // with hDirac==c==1. All dependent data get set accordingly.

   static void setVerbose(Z v){ verbose=v;}
   static Z getVerbose(){ return verbose;}

   static void setSize(R s){ size=s;}
   static R getSize(){ return size;}

   static void setMass(R m){ mass=m;}
   static R getMass(){ return mass;}

   static void setPhiStep(R ps){ phiStep=ps;}
   static R getPhiStep(){ return phiStep;}

   static R get_c(){ return c;}
   static R get_hDirac(){ return hDirac;}
   static R get_hPlanck(){ return hPlanck;}
   static R get_hDiracInv(){ return hDiracInv;}
   static C get_ihDirac(){ return ihDirac;}
   static C get_ihDiracInv(){ return ihDiracInv;}
   static R get_ihDiracInv2(){ return ihDiracInv2;}

   static R step(R n1,R n2)
   { return n2==0 ? CpmRoot2::hugeNumber : 0.999*phiStep*n1/n2;}
   // a building block for auto-setting the step size for the
   // direct midpoint integrator
};

class Observable;

class State: public QMBase{
   // Finite-dimensional Quantum mechanical state space with Hilbert
   // space structure and 'periodic boundary conditions'. Notice however,
   //
   // that a fixed value of the dimension is not part of the present
   // concept.
   // So creation and destruction operators can be defined without the
   // artificial modification deviced by Fritz Bopp and nevertheless
   // satisfy their characteristic commutation relations exactly.
   // Implementation aims at best efficiency that can achieved with Cpm.
   // Here the direct pointer operations in V<> are essential.
   // Without aiming at efficiency one could build on C_Vector or on
   // Va<C>

   friend class Observable;

   Z adjust(Z i)const;
      // brings i ito the range 0<=i<dim();
      // for reasons of utmost efficiency we let indexes start with 0
      // in quantum mechanical context

   V<C> rep;
      // lean and efficient array with indexes starting with 0

   explicit State(const V<C>& rep_):rep(rep_){}
      // creating

   explicit State(const V<C>& v):rep(v.c()){}

   V<C> toArray()const{ return V<C>(rep);}

public:

   typedef State Type;
   CPM_IO
   CPM_PURE_VECTOR_SPACE_INTERFACE(Type,C)
   CPM_SCALAR_PRODUCT_INTERFACE(Type,C)
   CPM_PURE_SCALAR_INTERFACE(Type,R)
      // Hilbert space with anti-unitary operator conj()
   CPM_NAM(State)

   State():rep(){}

   explicit State(Z d):rep(d){}
      // creates zero vector of dimension d.

   State(Z d, Z i):rep(d){ rep[i]=C::one;normalize_();}

   State(Z d, C z):rep(d,z){ normalize_();}
      // creates a 'd-dimensional' normalized state
      // the components of which are all the same
      // positive real multiple of z

   State(Word const& idle, const V<C>& v):rep(v.c()){normalize_();}

   State(const R_Func& f, Z d);
      // f is a function defined on the interval [-1/2,1/2] and creates a
      // normalized 'wave function' such that f(-1/2) corresponds to
      // rep[0] and f(1/2) to rep[d-1].

   State(const F<R,C>& f, Z d);
      // f is a function defined on the interval [-1/2,1/2] and creates a
      // normalized 'wave function'

   State(const R_Vector& arr, Z d);
      // The array arr will be used for building a Bernstein
      // polynomial, which the will define the state

   State toMomSpc()const;
      // to momentum space (FFT)

   State frmMomSpc()const;
      // from momentum Space (inverse FFT)

   State shift(R s)const;
      // active transformation translates a wave function peaked at
      // 0 into one peaked at s. So for s>0 translation on
      // screen is to the right. (Was in the opposite direction till
      // 2004-10-13)

   State boost(R v)const;
      // active transformation that makes a particle initially
      // at rest moving with velocity v, where v>0 means
      // motion from left to right on screen.
      // (Was in the opposite direction till 2004-10-13)

   State splitBoost(R v)const
      // returns a superposition of states 'going in different
      // directions' 'beam-splitter states'
   {
      State res=boost(v)+boost(-v);
      res.normalize_();
      return res;
   }

   C getLean(Z i)const
      { return rep.cui(i);}
      // fast read access to components without index range check

   void setLean(Z i, const C& z)
      { rep.cui(i)=z;}
      // fast write access to matrix elements without index range check

   V<C> getRep()const{ return rep;}
      // fast return of the main data

   C operator()(const Z& i)const{ return rep[adjust(i)];}
      // lets a state be used as a function, which is defined on the
      // set Z

   C& operator[](Z i){ return rep[adjust(i)];}
      // Indexes start here with 0 as in rep for utmost efficiency.
      // standard component access, no limit on indexes since always
      // periodic continuation is enforced

   const C& operator[](Z i)const{ return rep[adjust(i)];}
      // standard component access, no limit on indexes since always
      // periodic continuation is enforced

   C& cui(Z i){ return rep.cui(i);}
      // fast and unsave index access

   const C& cui(Z i)const{ return rep.cui(i);}
      // fast and unsave index access

   Z dim(void)const{return rep.dim();}
      // returns the dimension of the state space

   void set_(R r){ rep.set_(C(r));}
      // setting all components of *this equal to C(r,0)

   void set_(C z){ rep.set_(z);}
      // setting all components of *this equal to z

   void reg_(){ Z iLast=rep.dim()-1;
      C val=(rep[0]+rep[iLast])*0.5; rep[0]=val;
      rep[iLast]=val;}

   Type resize(Z newDim)const{ return State(rep.resize(newDim));}
      // returns the State of dimension Max(0,newDim) which
      // is naturally defined by *this

   Type reduce(Z d=32)const{ return d<dim() ? resize(d) : *this;}
      // the result never exceeds *this by dimension

   Type makeReal()const;
      // returns a State with all components of *this
      // replaced vy their absolute value

   Type symmetrize()const;
      // returns a state which has each component i
      // replaced by the mean of i and dim()-1-i

   Type smooth(bool normalizeRes=true)const;
      // returns a smoothed  version of *this.
      // If the argument has value true, the returned state, in addition,
      // is normalized (to norm 1).
      // Smoothing is by the kernel 1/4 1/2 1/4 where periodic
      // treatment of the boundary is used

   void smt_(Z n=1)
      //:smooth
      // Changes *this into a smoothed, normalized version.
      // The argument gives the number of times the basic
      // smoothing operation (kernel 1/4 1/2 1/4) is performed.
   {
      for (Z i=1;i<=n;++i) rep.fh_(C(0.25),C(0.5),C(0.25),CYCLIC);
      nor_();
   }

   Type operator ||(const Type&)const;
      // tensor product; member function is preferred due to more
      // transparent behavior in the presence of namespaces

   Type symPrd(const Type&)const;
      // symmetrized tensor product (for bosons)
      // Normalization factor is chosen such, that a product
      // of unit vectors has a unit vector as a result.

   Type antSymPrd(const Type&)const;
      // anti-symmetrized tensor product (for fermions)
      // Normalization factor is chosen such, that a product
      // of unit vectors has a unit vector as a result.

   Type crt()const;
      // creator = creation operator

   Type des()const;
      // destructor (not in the sense of C++) = destruction operator.
      // The basic property of crt and des is:
      //    psi.crt().des()-psi.des().crt()==psi
      // which corresponds to
      //    [a,a*]=1
      // taking into account the order of application given by
      // postfix notation.

   Type phi()const
   { return (crt()+des())*facPhi;}

   Type phi2()const
      // normal ordered product
   {  return (crt().crt()+des().des())*0.5+des().crt();}

   Type psi()const
   { return (crt()-des())*facPsi;}

   Type psi2()const
      // normal ordered product
   {  return (crt().crt()+des().des())*(-0.5)+des().crt();}

   Type sinParHam(R2_Func const& pot, R t=0, R lambda=1.)const;
      // 'single particle Hamiltonian'
      // returns a State which results from *this by application of
      // a single-body hamiltonian with position and time dependent
      // potential pot. The potential is a functions of two variables
      // (instance of R2_Func), where the second argument is time.
      // The first argument is a function of particle position
      // and of time.
      // Particle position here refers to a coordinate system
      // in which the whole position range is [-1/2,+1/2].
      // This is the actual position range for QMBase::size=1, which thus
      // is the natural choice when the present function is to be used.
      // If the coupling constant lambda is 0, no computation of the
      // potential will be done. We then have the Hamiltonian of a
      // free particle. Since the implementation uses a finte difference
      // formula for the second derivative, the dimension of *this
      // is not restricted to powers of 2.

   Type freeParHam()const;
      // Like previous function but no potential

   R normalize_(R r=1);
      // returns the norm of *this and changes *this into
      // a State of norm r by multiplying with a real factor.
      // If *this was the zero State, it will be turned into
      // a 'standard State' of norm r

   R nor_(R r=1){ return normalize_(r);}
      //: normalize
      // 'modern' name

   R variation()const;
      // returns the curve length if *this is represented as
      // the polygon rep[0]->rep[1]->,...->rep[d-1] in C
      // It is a measure for the geometrical complexity of the state

   R_Vector toRealForm()const;
      // returns a 2*d vector

   void fromRealForm(const R_Vector& r);
      // changes *this into the one resuling from the real vector r

   void show(Frame& f, Z mult_i=8, Z mult_j=4,
      Z iShift=0, Z jShift=0, bool ignorePhases=false)const;
      // one component of rep will be represented as a rectangle of
      // mult_i*mult_j pixels where i refers to the vertical direction
      // and j to the horizontal (as in matrices). The whole image will
      // be shifted by (iShift,jShift) with respect to the left upper
      // corner of the graphics window f. Actually the components of rep
      // are not converted to a screen color directly but before get
      // renormalized by a real factor chosen such that the dominant
      // component of rep gets an absolute value of 1. If the true values
      // of the components of a state s are to be shown use
      // Graph g;
      // g.show(s.toRealForm())
      // If ignorePhases differs from 0, what is represented is the
      // probability density instead of the complex-valued wave function.

   void showAsCurves(Graph& g)const;
      // auto-scaled polygon representation of real part (red), imaginary
      // part (green) and absolute value of the wave function

   void show(Z d1, Z d2, Frame& f, Z mult_i=4, Z mult_j=4,
      Z iShift=0, Z jShift=0)const;
      // represents the array of states got from
      // factorize(Z d1, Z d2)

   R sup()const;
      // max of absolute value of components

   bool orthogonal(V<State>& x, R eps=1e-5)const;
      // The return value indicates whether *this is othogonal to all
      // components of x. If this value is true, x will be appended to
      // x, so that one can collect vectors into a orthogonal base.

   static V<State> E_Schmidt(V<State>& x, R eps=1e-6);
      // returns an orthonormal family which spans the same space
      // as family x. Made by E. Schmidt's orthogonalization;
      // eps controles the termination of the process and helps to
      // return vectors which are made 'out of numerical noise'.
      // More specifically we assume that x.dim() is an even number 2*n
      // and that there are just n complex-linearly independent members
      // in the family x. (all 2*n are real-orthonormal in our
      // application). Thus we try to return just n vectors. As is well
      // known, E. Schmidt's procedure tends to strongly accumulate
      // round-off errors for families of more than, say, 20 members.
      // So, we do a semi-understood reordering for larger families which
      // seems to cure all deseases

   static void show(Graph& g, V<State> const& x);
      // shows a family of states as curves

   static R orthoTest(const V<State>& x);
      // returns a very small value if the family is orthonormal,
      // and a much larger one else

   void adjPhase(R eps);
      // adjust phase such that the component of largest absolute
      // value is real and positive. Actually the component of
      // largest absolute value is not always uniquely determined
      // independent of numerical noise. Therefore we give a
      // control parameter to determine a set of components that
      // will be under consideration together with the one of
      // largest absolute value: This set consists of those components
      // c of *this such that  dis(|c|,|cMax|)<eps ,
      // cMax: largest component.
      // From all these components under consideration, we make
      // the one with the smallest index positive.

   void dePhase(R p1, R p2, Z meth);
      // experimental device to enforce collapse of a wave packet
      // meth=1 results in stochastical phase jitter with maximum phase
      // change 2*pi*p1. p2 has no influence then
      // method<=0: no action. meth: 1,2,3,4 defined, see
      // implementation in cpmstate.cpp

   void dePhase_(R p){ dePhase(p,0,1);}
      // 'modern' name, since is a non-constant method

   State corGrn(Z nPhi, Z nBins)const;
      //: coarse grained
      // Returns a state in which complex numbers only habe nPhi
      // different (and equi-spaced) values
      // and each |rep[i]|^2 is an integer multiple of nBins.
      // added 2007-09-08

   R_Vector probabilities()const;

   State toState(bool ignorePhases)const
   { return ignorePhases ? makeReal() : *this ;}

   V< V<C> > factorize(Z d1, Z d2)const;
      // returns the state as a matrix in a manner that a state
      // being a tensor product State(d1)||State(d2)
      // brings the factors into evidence

   State ran(Z j=0)const;
      // random generator. The calling instance gives as the only
      // information its dimension

   R dis(State const& st)const
   {
      R a=(*this).abs();
      R b=st.abs();
      R d=(*this-st).abs();
      return CpmRoot::distFunc(a,b,d);
   }

   bool col_(IvZ const& ivz, R p=1);
      //: collapse
      // If called, the intended action happens with probability p.
      // The intended action is collapsing the wave function
      // to either the domain given by ivz or to the complement
      // with a probability that is consistent with the state
};

} // namespace

#endif
