//? cpmdynsys.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_DYNSYS_H_
#define CPM_DYNSYS_H_
/*

Description: Heterogeneous dynamical systems not directly
   prepared for parallel computing since parallelism
   will be achieved by a rather general template construction.
   Experiment, not yet working

   History:
      started 2002-06-18

*/
#include <cpmrmatrix.h>
#include <cpmgraph.h>
#include <cpmv.h>
#include <cpms.h>
#include <cpmx.h>
#include <cpmalgorithms.h>
#include <cpmimagingtools.h>
#include <cpmrecordhandler.h>
#include <cpmproperties.h>
#include <cpmdynaux.h>
#include <sstream>

#undef CPM_USE_STL
   // uses class std::vector<> instead of CpmArrays::V<>
   // for some efficiency-critical interal workings. Causes
   // failure in parallel computing so far. Introduced
   // in order to enable studying the nature of these failures

#if defined(CPM_USE_STL)
   #include <vector>
   using std::vector;
#endif

namespace CpmDynSys{

   using std::ostringstream;

   using CpmRoot::read;
   using CpmRoot::write;
   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot::Word;

   using CpmSystem::Error;
   using CpmSystem::Timer;

   using CpmArrays::R_Vector;
   using CpmArrays::R_Matrix;
   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::Vp;
   using CpmArrays::Vo;
   using CpmArrays::S;
   using CpmArrays::Sio;
   using CpmArrays::X2;

   using CpmFunctions::F;
   using CpmFunctions::F2;

   using CpmRoot2::Z1;
   using CpmRoot2::R1;
   using CpmRoot2::B;
   using CpmRoot2::hugeNumber;
   using CpmRoot2::Record;

   using CpmGeo::Iv;

   using CpmAlgorithms::CyclicAlarm;

   using CpmGraphics::Graph;
   using CpmGraphics::Frame;
   using CpmGraphics::Frames;

   using CpmImaging::Color;
   using CpmImaging::Palette;
   using CpmImaging::INVISIBLE;

   using CpmProperties::SysControl;
   using CpmProperties::SysControlState;

/////////////////////// class DynSys ////////////////////////////////
// 'Parallel Dynamical System'

#define CPM_CN DynSys
   // this style of introducing the class name helps to define
   // new versions

#define CPM_TAL template <class Par, class Ext,\
   class Vec, class Spc, class Group,\
   class For, class FieldSpace, class ScalarField, class VectorField,\
   class PointCharge, class Charges, class MagDipole, class SphPart,\
   class Camera>

#define CPM_FCN CPM_CN<Par,Ext,Vec,Spc,Group,\
   For,FieldSpace,ScalarField,VectorField,PointCharge,\
   Charges,MagDipole,SphPart,Camera>

////////// class DynSys //////////////////////////////////////////////////
CPM_TAL
class CPM_CN{//dynamical system DynSys, granular matter confined by bodies

   static bool fEqui(const Par& p1, const Par& p2)
   {
      return p1.get_id()==p2.get_id();
   }

protected:
// here we havn't grouped data into independent and dependent ones.
// The independent data are those which are handled in function
// printOn(...). Function scanFrom(...), together with functions
// update(...) and updateVeryFar(...) defines the mechanism by which
// the dependent quantities get their values.

   StepControl stpCon;
      // book-keeping quantities for time stepping, these data also
      // include the time t, which, despite of its physical importance
      // is placed into one house together with some minor relatives

public:

   V<Par> par;
      // here, polymorphism isn't essential, toner and carrier particles
      // can be reasonably described as instances of a single class


   Vp<Ext> ext;
      // here, the components may be a toning shell, a magnetic roller, a
      // photo-conductor .... In the old version these entites were all
      // combined in a single entity since polymorphic arrays were not at
      // hand. All these types of external systems have to be derived
      // from class Ext.
protected:
   Z1 np;
      // number of particles, not really independent, related to par
      // makes I/O easier if considered independent. The way of
      // initializing np and ne in the constructor aks np to be placed
      // here later than par and ne later than ext.

   Z1 ne;
      // number of external systems; see np

   Parameters pr;
      // see struct Parameters;

   mutable B auxListsInitialized;
      // initialized as false, which is OK, so no further action needed
      // in constructors of the present class. If this is false, the
      // next call to update() will initialize the arrays/fields
      // xPar, tPar, active, nearList, medList, filFar

   mutable B filFarUpToDate;
      // if false, the values of the quantities mentioned in the comment
      // to auxListsInitialized need to be updated to become consistent
      // with the overall state of the system

   mutable V<FieldSpace> filFar;
      // allows to hold a field value for each particle. This field 
      // strength describes approximately the interaction of a particle
      // with the distant particles.
      // Here this replacement field strength  is an electromagnetic 
      // field strength, since all the other forces to be considered
      // here are short range. If we had no magnetic field, it would
      // be more convenient to store directly the force instead of 
      // the field. Since in magnetic case the force also depends on
      // the orientation of the feeling particle one had to 
      // pre-calculate forces for different orientations or build
      // a tensor magnetic force tensor instead a force vector as
      // I did in the Ariton program.
      // Updating filFar has not to be done very frequently since the
      // contribution of distant particles can't change rapidly. In
      // addition, it is weak so that inaccuracies should not be
      // critical. Instead of a field value, in earlier versions I had
      // here a VectorField (so that filFar was typed V<VectorField>).
      // This worked well and was not noticibly slower. however it
      // looks physically un-plausible to represent the distant
      // particle by anything more complicated than a homogeneous field.

   mutable V<bool> active;
      // The value of active[i] controls the treatment of par[i] in
      // function stepBasic(...).
      // Only if active[i]==true, we will calculate forces
      // acting on par[i]; otherwise the change of state with time is
      // considered as given by the function t |--> par[i].osc(t).

   mutable Vo<R> tPar;
      // a list that stores for each particle the point in time up
      // to which the trajectory of this particle is yet computed
      // on the basis of actual forces.
      // It is a particularity of the dynamical method to be implemented
      // here, that these times need not to be exactly the same for all
      // particles. This allows to compute the dynamics of particles in
      // a 'critical state of motion' (e.g. fierce collision) in smaller
      // time-steps than the those that move in free space.
      // Even for times t>tPar[i] we may consider the state
      // par[i].osc(t).
      // This state then is not infered not from a computed trajectory
      // but from extrapolating the computed trajectory based on a local
      // approximation similar to a 'osculating trajectory' of celestial
      // mechanics.

   mutable V<Spc> xPar;
      // list of particle positions, needed to monitor motion
      // which helps to decide when certain arrays have to be updated

#if defined(CPM_USE_STL)
   mutable V< std::vector<Z> > nearList;
   mutable V< std::vector<Z> > medList;
#else
   mutable V< V<Z> > nearList;
      // nearList is an array of index sets.
      // nearList[i] holds all indexes of particles which
      // have distance from particle i smaller than nearLimit.

   mutable V< V<Z> > medList;
      // medList[i] holds all indexes of particles which have distance
      // from particle i between nearLimit and farLimit
#endif

   void update(bool enforce=false);
      // updates the arrays xPar, nearList, medList, filFar, ....
      // Presently updating is a global action extending to all
      // particles at once. It is conceivable to update the neighborhood
      // of fast particles at a different rate from slow ones.
      // If the argument is true, update is done without calling
      // function updateNeeded(), which normally checks whether
      // this expensive action is really needed.

   void updateNeeded()const;
      // sets good values for filFarUpToDate.
      // Thus this function controls the update needs of the auxiliary
      // quantities referring to 'new fast dynamics'.

   void updateActive(R t)
   {
      R tc=t-pr.tSmall; // this way of getting rid of numerical
         // inaccuracies is not elegant, what else can be done ?
      for (Z i=1;i<=np;i++) active[i]=(tc>tPar[i]);
   }

   void setTimesFromExtSys();
      // make all times of system identical and equal to the mean
      // value of the times defined by the external systems
      // (which all have a get_t() function). This is important for
      // full reconstruction of a state from the information
      // written on movie-files

// constructors. The data elements active, tPar, xPar, nearList, filFar
// will not become initialized by the constructors, but only if they are
// needed to support time step control

   CPM_CN(V<Par>  par_, Vp<Ext> ext_, Parameters pr_): 
      par(par_),ext(ext_), np(par.dim()),ne(ext.dim()),pr(pr_)
   {}

public:

   void enforceGeneralUpdate()
      // sets the preconditions for function step() to work 
      // properly
   {
      auxListsInitialized=B(false);
      filFarUpToDate=B(false);
      update();
   }

   CPM_CN(V<Par>  par_, Ext ext_):
      par(par_),ext(Vp<Ext>(1))),
      np(par.dim()),ne(ext.dim()),pr(pr_)
   {
      ext.set(1,ext_);
   }

   CPM_CN():par(V<Par>(0)),np(0),ne(0){}
         // notice that the default for V<Par> would bring in one
         // particle

// I/O and documentation

   typedef CPM_CN Type;
   typedef CPM_CN CloneBase;
   CPM_IO_V // stream I/O, important that this is virtual
      // these are the functions scanFrom and printOn which deserve some
      // clarifications. 
      // Precondition of printOn(str) is only that
      // str in valid. Then a complete data set for the system state gets
      // written on the stream, restricted only by the number of digits
      // chosen for data to ASCII conversion (set in cpmdefinitions.h 
      // #define CPM_WRITE_PRECISION). By definition, the data set is
      // complete if it - together with the data of SysControl 
      // determines the behavior of function step().
      // All 'bookkeeping quantities' that get initialized
      // anyway within function step() are not written
      // to str. These, however are copied by the copy constructor and 
      // by the assignment operator. Unfortunately there is even a third 
      // copying method in the game: the way in which states are filed
      // (movie-file production) Since these files tend to grow very
      // large,
      // it is essential to have this process very data-efficient. 
      // See comment to function file(...).
      // Precondition of scanFrom(str) is that str is valid and thatthe 
      // data on str are such as if they had been written by 
      // pardyn.printOn(str) where pardyn is an instance of the present
      // template class with the same template arguments as *this
      // and with the  s a m e  n u m b e r  and  e x a c t  t y p e s
      // (derived from ExtSys) as this.
      // Postcondition is that *this is canged exactly into  pardyn.  
      // (By definition, a postcondition is met after function call if
      // the precondition held before function call; this seems to be a 
      // useful notion, see e.g. Paul Taylor: Practical Foundations of
      // Mathematics, Cambridge 1999, p. 199) 
   CPM_CLONE

   void done(R dt_, R dis_){ stpCon.done(dt_,dis_);}
      // if a time step dt was effective in changing the dynamical
      // variables of the system, this is the change to the 'book-keeping
      // quantities'

   void done(R dis_){ R t=tPar.inf(); stpCon.doneTill(t,dis_);}
      // if a time step dt was effective in changing the dynamical
      // variables of the system, this is the change to the 'book-keeping
      // quantities'

// obvious 'read'-access functions
   R get_t()const{ return stpCon.t;}
   R get_dt()const{ return stpCon.dtMean;}
   R get_tLimit()const{ return stpCon.tLimit;}
   bool getRelaxFlag()const{ return stpCon.relaxFlag;}
   R get_tSmall()const{ return tSmall;}
   R getFraction()const{ return fraction;}

   const Par& getPar(Z i)const{ return par[i];}
   Par& getPar(Z i){ return par[i];}

   V<Par> getPar()const{ return par;}
   Vp<Ext> getExt()const{ return ext;}

   template <class X>
   X getExtComp(Z i, const X& x)const{ return ext(i,x);}
      // template member function works!!! fine
      // Here the value of x is unimportant, it is needed only to tell
      // the function the precise type of the returned quantity

   Z get_np()const{ return np;}
   Z get_ne()const{ return ne;}

   Spc cenOfMas(const V<bool>& vb)const;
      // returns the center-of-mass of those particles par[i]
      // for which vb[i]==true (actually we need the center-of-mass of
      // all carrier particles, but being a carrier particle is too
      // special a concept to be placed here. The default argument is set
      // such that the center-of-mass of all particles is calculated if
      // no
      // vb-argument is given

   R kinEnr(const V<bool>& vb)const;
      // returns the kinetic energy of those particles par[i]
      // for which vb[i]==true

// 'write'-access to the most important substructure
   template <class X>
   void setExtComp(Z i, const X& x){ext.set(i,x);}
      // template member function works!!! fine. Class X has to be a
      // derivation of Ext.

   void setPar(const V<Par>& par_){ par=par_; np=par.dim();}
      // a combination of get and set allows to perform modifications of
      // *this for which there are no member functions provided in
      // ParDynSys<>. For instance, a list of np instances of Group can
      // be used to transform the particles, each particle by its own
      // transformation:
      // V<Par> vp=getPar(); // extract the particles for modification
      // V<Group> vg(np);
      // Z i;
      // for (i=1;i<=np;i++) vg[i]=...; // define the Group-instances
      // for (i=1;i<=np;i++) vp[i]*=vg[i]; // transform the extracted
      // particles
      // setPar(vp); // Put the particles back. Now *this houses as par
      // the transformed list of particles

   void delPar(){ par=V<Par>(0); np=0;}
   void addPar(const Par p){ par.push_back(p); np=par.dim();}
   void addPar(const V<Par>& p){ par&=p; np=par.dim();}

   void selectParticles(const V<bool>& vb)
      { par=par.select(vb); np=par.dim();}
      // allows to remove particles from the system under control of a
      // boolean array. Paricles par[i] for vb[i]==true are selected to
      // stay, not to get removed!

   void setActivityOfExternalSystems(Z iext, Z act)
      { ext[iext].setActivity(act); }
      // allows to disable the dynamical anf graphical activity of
      // selected external systems.

   void setVelocityOfExternalSystems(Z iext, const Vec& v)
      { ext[iext].set_v(v); }

   void setCol(const V<Palette>& c)
      {for (Z i=1;i<=np;i++) par[i].setCol(c[i]); }
         // setting the visualization color

// 'write' access to simple control quantities
   void set_t(R t0);
      // make all times of system identical to the input 
      // value t0;
   void set_tLimit(R dtL){ stpCon.tLimit=dtL;}
   void set_tSmall(R ts){ pr.tSmall=ts;}
      // see attribute tSmall
   void setNearFarLimits(R a1 , R a2)
      { pr.nearFarLimits=Iv(a1,a2); auxListsInitialized=false;}
   void setShiftLimits(R s1){pr.shiftFar=s1;}
   void setFraction(R fr){ pr.fraction=fr;}
   void setParameters(const Parameters& pr_){ pr=pr_;}
   void clear_dtLimit(void){stpCon.dtLimit=hugeNumber;}
   void clear()const{ VectorField::newScl(); }

   void setRelaxFlag(bool rf)
      // switching between 'relaxed dynamics' (a 'dynamics' which sets
      // all velocities to zero after each time step) and normal dynamics
      {
         stpCon.relaxFlag=rf;
         if (rf) for (Z i=1;i<=np;i++) par[i].stop();
      }

   void showParticleFields(bool b){pr.showPartFields=b;}

// kinematical transformations
   void boost(const Vec& v){ for (Z i=1;i<=np;i++) par[i].boost(v);}

   virtual void operator*=(const Group&);
      // geometrical transform of the whole system
      // if one adds data in derived classes, these have to be
      // transformed too; thus the function should be virtual

   void transform(Z i, const Group& g){ par[i]*=g;}
      // geometrical transform of particle i

   void transformExternalSystems(Z i, const Group& g){ ext[i]*=g;}
      // geometrical transform of external system i

   For force(const Par& par1, const Par& par2)const;
      // force which particle par1 feels from particle par2 and all the
      // mirror charges which get generated by par2 (not by par1 !)
      // Mirror charges come from the external systems individually
      // (e.g. from the Shell and the Target) and are defined in the
      // function influence() of the respective classes.

   FieldSpace field(const Par& par1, const Spc& sp, 
      bool farAway=false)const;
      // field that particle par1 creates in point sp together with
      // all the mirror charges created in external systems. These
      // are defined by the same functions influence() called in the
      // previous function force().
      // If the last argument is true, the various mirror charges 
      // created from the same external devise get collected into 
      // one charge for computational efficiency.

// functions supporting new fast dynamics

private:

   virtual void stepBasic(R t0, R dt,const VectorField& filVeryFar);
      // time evolution over a fixed span dt of time. Implements the
      // explicit mid-point method of time-discrete dynamics assuming
      // that some particles are active and others are not (they then are
      // called passive).
      // Passive particles contribute to the generation of fields but get
      // not deflected by forces. The status of being active
      // will not be changed by calling this function.
      // We assume that all active particles par[i] are in a state
      // such that par[i].osc(t0) is a reliable estimate of the
      // particle state at time t0. (Ideally, par[i].get_t()=t0).
      // After function call, the whole system is in a state that is the
      // best guess for a system state at time t0+dt (based on the
      // information contained in *this at the point of function call).
      // In particular, for all particles par[i](active or not)  we have
      // after function call par[i].get_t()==t0+dt.

   Z stepControled(R& dt,const VectorField& filVeryFar);
   // This is the main building block of the function step(R)
   // which implements automated time step adjustment for
   // the present multi-time (asynchronous) formalism.
   // To explain this function precisely, we need some definitions:
   // The particle par[i] is said to be active at time t if
   // t>tPar[i] (unless data are in preliminary state, we have
   // tPar[i]=par[i].get_t() ). The present function finds a time
   // step dt'=dt*pow(2,-n) ( 0<=n, integer n) such that a 'percentage'
   // (here I use 'percentage' as a synonym for fraction, implying
   // a value range from 0 to 1 and not 0 to 100; the reason for this
   // is that 'fraction' is an attribute of the present class)
   // frac'>=fraction of all particles active at tMin+dt',
   // tMin:=min{ tPar[i] | 1<=i<=np}, can do the time step dt' with
   // an step error nor exceeding limits set by SysControl::acceptable.
   // After calling the function the states of the particles belonging to
   // this successful fraction will be replaced by those resulting from
   // the successfully performed time step and tPar[i] for such a
   // successful particle i will be incremented by dt'. The value of dt
   // after call will be dt'/2 if frac'<1 and dt'*increment if frac'=1.
   // The return value of the function is n.
   // dt' will not be divided down till it becomes smaller than tSmall.
   // If this would be the consequence of the normal control logic,
   // a drastic measure is taken to allow further time-stepping
   // with normal size: all particle velocities are set to zero.
   // This will never occur, if the parameter SysControl::bad is
   // not given too large values (reasonable: SysControl::bad<=0.2)
   // This functionality is best seen as useful by looking into
   // function stepAsy and to see how it achieves the time evolution of
   // the whole system by 'promoting in time' just the particles which
   // are 'most behind'.

   void stepAsy(R& dt,const VectorField& filVeryFar);
      // 'asynchronous step'. Time evolution with automatic step
      // adjustment. Here dt is only a proposal. After the step is
      // done, dt shows the proposed value for the next step.
      // The value actually chosen is obtained as get_dtLast().
      // At the end of the step, the particle states need not to
      // refer to exactly the same time, thus the name asynchronous.
      // This function is intended not to be used for
      // SysControl::fixedStep==0. For this case the final function
      // step(R) is implemented without calling stepAsy(R&).

public:

   void step(R dt,const VectorField& filVeryFar);
      // evolving the system over exactly the time step dt.
      // After call all subsystems refer exactly to the same time. The
      // working of the function will not be completed before it is
      // completed in each of all parallel processes. This function is
      // intended to be called for dt' s large enough that the change
      // of system state conveys information to the user and has not
      // to be limited by numerical restrictions concernig numerical
      // modelling of collisions of particles. If SysControl::fixedStep
      // is zero, automated time step adjustnent will break down the
      // dt-step into many substeps which assure that the accuracy limits
      // set as described in Z stepControled(R& dt). There is however
      // one non-automated aspect of dt: exactly after each subsystem
      // evolved autonomously over the timespan dt it will be brought
      // into contact with all the other parallel processes for proper
      // exchange of the 'overlap particles' (which are handled by more
      // than one processes since they are near to a process domain
      // boundary) and for updating the electromagnetic fields which
      // each process domain creates in order to represent the action
      // of its own particles onto the particles of other processes.
      // Due to the distance between process domains and the property
      // of these fields to be an average of many weakly correlated
      // contributions it is not necessary to update these fields on the
      // time scale of particle collisions.

   void run(R dtStep, Z nSteps, Z filingPeriod, Z viewingPeriod,
         Z stoppingPeriod, const Word& movieName, Camera& cam,
         const Timer& tm,  const F<R,Z>& f=F<R,Z>(0));
      // Calling function step(dtStep) nSteps times in sequence
      // and writing to file (under control by movieName) and visualize
      // on screen (under control by cam) with a period (in number of
      // calls of step(R)) of filingPeriod and viewingPeriod
      // respectively.
      // Exceptions: nSteps<0 or dtStep<0 renders the action of the
      // function trivial.
      // filingPeriod<1 disables all filing
      // viewingPeriod<1 disables all display activity
      // Thus for nSteps==0 steps(R) will not be called but *this
      // will be filed unless filingPeriod<1 and *this will be displayed
      // on screen unless viewingPeriod<1.
      // The last argument manipulates as a function of t the parameters
      // in SysControl. 
      // The argument tm allows to set a limit to the computation time.
      // After a filing action tm gets asked whether the timeOut
      // condition
      // holds and if this is the case the program is stopped after 
      // the last state has been written to file completely. After 
      // restarting the run from file one needs not to adjust nSteps
      // according to the computational progress already achieved.
      // The new functionality governed by tm is not implemented in the
      // present class but only in the derived class Station by suitable
      // re-definition of virtual function file.

   void stopParticles()
      // sets the velocities of the particles to zero
   {
      for (Z j=1;j<=np; j++){
         par[j].stop();
         par[j].deElectrize();
      }
   }

// descriptors

   VectorField getFieldExt()const;
      // returns the field got by adding all contributions from Env and
      // all external systems

   VectorField getFieldPart()const;
      // returns the field got by adding the contributions of all
      // particles

   VectorField getFieldTot()const;
      // return the sum of the previously considered fields

   VectorField getFarField()const;
      // for defining the return value of this function the system is 
      // considered as a point charge and an electric moment and a
      // magnetic dipole moment placed at the arithmetic mean of the
      // positions of the particles. 
      // If charge to wall interaction is enabled, we let the 
      // above-mentioned point charge create mirror charges in 
      // the external systems and add the electric field generated
      // by these. In most cases the point charge will be small
      // and the mirror will be small too. This only shows that the
      // concetration of all charges into a single point is a 
      // very strong idealization

   R dis(const Type& )const;
      // comparison function on which the step adjustment in step() is
      // based. This is not limited by 1 as in my numerical classes. See
      // the defining code for the meaning.

   V<R> dist(const Type& )const;
      // returns distance values as individual values for the the
      // subsystems and not only the maximum over all

   virtual void mark(const Camera& cam)const;
      // visual representation of the configuration on which movie
      // production is based

   virtual bool file(const Word& fn, const Timer& tm)const
      { return false;}
      // writes the state into a suitable file, which should
      // be made visible by a suitable viewer program on a
      // single computer. Since the number of particles to be
      // considered is huge, it is rather evident that one should
      // file not too much redundant data. This suggests to divide
      // the data of an object in constant ones and those that change
      // with time. This is presently implemented in a non-trivial manner
      // only in a derived class Station. 
};

CPM_TAL
void CPM_FCN::setTimesFromExtSys()
{
   if (ne==0) return;
   R sum=0;
   Z i;
   for (i=1;i<=ne; i++) sum+=ext[i].get_t();
   R tMean=sum/ne;
   for (i=1;i<=ne; i++) ext[i].set_t(tMean);
   stpCon.t=tMean;
}

CPM_TAL
void CPM_FCN::set_t(R t0)
{
   stpCon.t=t0;
   Z i;
   for (i=1;i<=ne; i++) ext[i].set_t(t0);
   for (i=1;i<=np; i++) par[i].set_t(t0);
      // par[i]-loop was missing till 02-06-28 and spoiled the 
      // start of ctc7c run3
}

CPM_TAL
bool CPM_FCN::printOn(ostream& str)const
{
   CpmRoot::writeTitle("ParDynSys<>",str);
   cpmp(stpCon);
   cpmp(np);
   cpmp(ne);
   Z i;
   for (i=1;i<=np;i++) cpmp(par[i]);
   for (i=1;i<=ne;i++) cpmp(ext[i]);
   cpmp(pr);
   return true;
}

CPM_TAL
bool CPM_FCN::scanFrom(istream& str)
{
   cpms(stpCon);
   cpms(np);
   cpms(ne);
   Z i;
   par=V<Par>(np); 
   // needed since the calling instance is not assumed to have the
   // correct
   // number of particles already
   // ext=Vp<Ext>(ne); is missing since we assume that the calling
   // instance
   // 
   // already has right number and types of ext[i].
   // Introducing the above staement would even be wrong since it would
   // give the external system the type ExtSys and thus would spoil the
   // additional attributes of derived classes.
   for (i=1;i<=np;i++) cpms(par[i]);
   for (i=1;i<=ne;i++) cpms(ext[i]);
   cpms(pr);
   auxListsInitialized=false;
   return true;
}

CPM_TAL
void CPM_FCN::operator*=(const Group& g)
{
   Z i;
   for (i=1;i<=np;i++) par[i]*=g;
   for (i=1;i<=ne;i++) ext[i]*=g;
}

CPM_TAL
R CPM_FCN::dis(const Type& hs)const
{
   Z i;
   R_Vector vp(np);
   for (i=1;i<=np;i++) vp[i]=par[i].dis((hs.par)[i]);

   R_Vector ve(ne,0);
   for (i=1;i<=ne;i++) 
      if (ext[i].isActive()) ve[i]=ext[i].dis((hs.ext)[i]);

   R res=cpmsup<R>(vp.sup(),ve.sup());
   if (res==0.){ cpmwarning(
      "ParDynSys<>::dis=0; parabolic motion or numerical problem");
   }
   return res;
}

CPM_TAL
V<R> CPM_FCN::dist(const Type& hs)const
{
   V<R> res(np,0); // was not initialized by 0 till 02-04-26
   for (Z i=1;i<=np;i++)
      if (active[i]) res[i]=par[i].dis((hs.par)[i]);
   return res;
}

CPM_TAL
Spc CPM_FCN::cenOfMas(const V<bool>& vb)const
{
   V<Spc> x(np);
   V<R>  w(np);
   for (Z i=1;i<=np;i++){
      x[i]=par[i].get_x();
      w[i]= vb[i] ? par[i].get_m() : 0. ;
   }
   Spc res;
   res.convexAsMember(x,w,true);
   return res;
}

CPM_TAL
R CPM_FCN::kinEnr(const V<bool>& vb)const
{
   R res=0;
   for (Z i=1;i<=np;i++){
      if (vb[i]) res+=par[i].kinEnr();
   }
   return res;
}

CPM_TAL
For CPM_FCN::force(const Par& par1, const Par& par2)const
{
   For res=par1.forces(par2);
      // direct force
   if (!SysControl::noInteractionChargeToWall){
      PointCharge q1=par1.getPointCharge();
         // feels the force
      PointCharge q2=par2.getPointCharge();
         // generates mirror charges 
      for (Z j=1;j<=ne;j++){
         V<PointCharge> extinfl=ext[j].influence(q2);
         res.addForce(q1.force(extinfl));
      }   // was wrong till 02-07-22
   }
   return res;
}

CPM_TAL
FieldSpace CPM_FCN::field(const Par& par1, const Spc& sp, 
   bool farAway)const
{
   const bool nogravity=true;
      // we don't want the particles to create gravitational fields

   FieldSpace res=(par1.getField(nogravity))(sp);
   if (!SysControl::noInteractionChargeToWall){
      PointCharge q=par1.getPointCharge();
      for (Z j=1;j<=ne;j++){
         V<PointCharge> extinfl=ext[j].influence(q);
         Z eid=extinfl.dim();
         if (eid==0) continue;
         else if(farAway){ // treat the mirror charges from one source 
            // as placed at the same position of the first one
            PointCharge pc=extinfl[1];
            for (Z k=2;k<=eid;k++) pc.add(extinfl[k]);
            res.c1()+=PointCharge::electrostaticField(sp,pc);
         }
         else{
            res.c1()+=PointCharge::electrostaticField(sp,extinfl);
         }
      }
   }
   return res;
}

////////////////////////// dynamics stuff //////////////////////////////

CPM_TAL
void CPM_FCN::update(bool enforced)
{
   const Z mL=4;
      // message limit (normal value 4)
   const Z mL2=2;
      // message limit for important messages which in an earlier version
      // were issued always (normal value 2)

   cpmmessage(mL,"CPM_FCN::update(...) started");

   if (!auxListsInitialized){
      xPar=V<Spc>(np);
      R t0=get_t();
      tPar=V<R>(np,t0);
      active=V<bool>(np,true);
#if defined(CPM_USE_STL)
      nearList=V< vector<Z> >(np);
      medList=V< vector<Z> >(np);
#else
      nearList=V< V<Z> >(np);
      medList=V< V<Z> >(np);
#endif
      filFar=V<FieldSpace>(np);
      auxListsInitialized=true;
      filFarUpToDate=false;
   }
   else{
      updateNeeded();
   }

   if (!enforced && filFarUpToDate){
      cpmmessage(mL,"CPM_FCN::update(...) trivially done");
      return;
   }

   cpmmessage(mL2,"updating fields of distant particles");
      // does not happen too often, good to know when this happens

   R nearLimit=pr.nearFarLimits.inf();
   R farLimit=pr.nearFarLimits.sup();

   Z i,j;
   R npInv=(np==0 ? 0 : 1./np);
#if defined(CPM_USE_STL)
   vector<Z> voidSet;
#else
   V<Z> voidSet(0);
#endif
   for (i=1;i<=np;i++){
      if (cpmverbose>=mL2) cpmprogress("xPar",(i-1)*npInv);
      xPar[i]=par[i].get_x();
      nearList[i]=voidSet;
      medList[i]=voidSet;
         // old values have to be reset for correctly
         // building the new ones
   }

   cpmmessage(mL2,"updating fields of distant particles, xPar-loop done")\
   ;

   for (i=1;i<=np;i++){
      if (cpmverbose>=mL2){
         cpmprogress("nearList",(i-1)*npInv);
      }
      Spc xi=xPar[i];
      for (j=i+1;j<=np;j++){ // notice j>i, economy !
         Spc xj=xPar[j];
         R dis=(xi-xj).abs();
         if (dis<=nearLimit){
            nearList[j].push_back(i);
            nearList[i].push_back(j);
         }
         else if (dis<=farLimit){
            medList[j].push_back(i);
            medList[i].push_back(j);
         }
         else ;
      }
   }
   cpmmessage
      (mL2,"updating fields of distant particles, nearList-loop done");

// defining the array filFar of fieldstrengths from distant particles.
// Notice that with the parallel version, there are even farer particles
// namely those in the other processes. These are now represented
// by filVeryFar, which is the electromagnetic field which these
// particles create.

   bool farAway(true);

   for (i=1;i<=np;i++){
      if (cpmverbose>=mL2){
         cpmprogress("filFar",(i-1)*npInv);
      }
      Spc xi=xPar[i];
      FieldSpace yi;
     
      for (j=1;j<=np;j++){
         // particle j has to contribute to farFil[i] iff (j!=i and j
         // does neither appear in the list of near particles nor in that
         // of the medium distant particles. (Not excluding i==j caused
         // serious trouble in 2000-12-15)
         if (i==j) continue;
         // find contribution of particle j to the far field. If the
         // particle is in the near or medium zone, it does not
         // contribute
         R dij=(xi-xPar[j]).abs();
         if (dij>farLimit && dij< SysControl::intactRange){
            // this way of looping over the far particles is faster than
            // explicitely excluding particles from nearList and medList
            yi+=field(par[j],xi,farAway);
               // this includes the contributions from mirror charges of
               // par[j]
         }
      }
      filFar[i]=yi;
   }
   filFarUpToDate=true;
   cpmmessage
      (mL2,"updating fields of distant particles, filFar-loop done");
   cpmmessage(mL,"CPM_FCN::update(...) done");
}

CPM_TAL
void CPM_FCN::updateNeeded()const
// the update needs of the 'veryFar'-quantities are presently dealt in a
// simple manner: they get updated after every call of step(R) via
// function redistribute, which sets filVeryFarUpToDate=false and
// thus triggers the correct activity of updateFilVeryFar(...).
{
   const Z mL=3;
   cpmmessage(mL,"CPM_FCN::updateNeeded() started");

   if (np<=1){  // then no distant particles are there
      filFarUpToDate=true;
      cpmmessage(mL,"CPM_FCN::updateNeeded() trivially done for np==1");
      return;
   }
// test for maximum motion of particles in the near-far ring zone.
// Logic is similar to the update rule for Verlet neighbor lists
// See: Paul A. Fishwick: Simulation, Model Design, and Execution,
// Prentice-Hall 1995, p. 270-271
   Z i,j,k,ni;
   bool bFar=true;
   for (i=1;i<=np;i++){
      ni=medList[i].size();
#if defined(CPM_USE_STL)
      for (j=0;j<ni;j++){
#else
      for (j=1;j<=ni;j++){
#endif
         k=medList[i][j];
         Spc xNew=par[k].get_x();
         Spc xOld=xPar[k];
         R shift=(xOld-xNew).abs();
         if (shift>pr.shiftFar){
            bFar=false;
            goto outOfLoop;
               // was break till 02-04-19, thus did needless
               // work when udate need was detected
         }
      }
   }
outOfLoop:
   filFarUpToDate=bFar;
   cpmmessage(mL,"CPM_FCN::updateNeeded() done");
}

CPM_TAL
VectorField CPM_FCN::getFarField()const
{
   if (!SysControl::veryFar) return VectorField();

   Word loc("CPM_FCN::getFarField()const");
   const Z mL=3;
      // meassage limit (normal value 3)
   cpmmessage(mL,loc&" started");
   Spc xMean;
   V<R> weights(np,1./np);
   bool careful=false;
   xMean.convexAsMember(xPar,weights,careful); 
      // weights are obviously OK, so no care needed
      // now xMean is the mean value of the points

   Charges ch;
   MagDipole md;
   ch.set_x(xMean);
   md.set_x(xMean);
   Z i,k;
   for (i=1;i<=np;i++){
      Charges chi=par[i].getCharges();
      MagDipole mdi=par[i].get_j();
      ch.add(chi); // adding a Charge to Charges updates
         // also the electric dipole moment
      md.add(mdi);
   }
   VectorField res(ch.field(),md.field());

   if (!SysControl::noInteractionChargeToWall){
      // The point charges should be augmented by the charges which they
      // influence in the external systems.
      PointCharge pc=ch.getCharge();
      if (pc.get_q()!=0.){ 
            // this happens if there are points in the point list in the
            // neighborhood of which there are no particles. If these
            // charges are 0, we have not to care about mirror charges
         for (k=1;k<=ne;k++){
            V<PointCharge> pck=ext[k].influence(pc);
            Charges chk(pck); // we don't add the charges
               // adding the fields looks more natural
            res.add(chk.field());
         }
      }
   }
   cpmmessage(mL,loc&" done");
   return res;
}

///////////////////////// main dynamics stuff //////////////////////////

CPM_TAL
void CPM_FCN::stepBasic(R t0, R dt,const VectorField& filVeryFar)
{
   const Z mL=2;
   const Z mL2=2;
   if (dt==0.){ // this should be efficient
      cpmmessage(mL,"CPM_FCN::stepBasic(...) trivially done");
      return;
   }

   cpmvalue("dt",dt,2);
   cpmmessage(mL,"stepBasic(...), dt="&cpmwrite(dt)&" np="&cpmwrite(np));

// linear evolution to the midpoint in time: explicit midpoint method
   Z i,j;
   R tau=0.5*dt;
   R tc=t0+tau;
      // t center
   R tf=t0+dt;

   for (i=1;i<=np;i++){
     par[i].osc(tc);
   }

   V<R> te(ne,0);
   for (j=1;j<=ne;j++){
      te[j]=ext[j].get_t();
      ext[j].osc(tc);
         // see comment to ext[j].osc(tf)
   }

   R npInv=(np>0 ? 1./np : 0);

// collecting the interaction terms at the mid-point in time
// dont move particles yet, otherwise dynamics would depend on 
// numbering of particles

   V<For> forces(np);
   V<FieldSpace> fValues(np);
   Z act=0;

   for (i=1;i<=np;i++){
      if (!active[i]) continue;
      act++;
      if (cpmverbose>=mL2) cpmprogress("force loop",(i-1)*npInv);
      Par pi=par[i];
// forces by general fields
      For fi=pi.forces(ext[1]); // 'air friction' in the reference 
         // system of ext[1] (= PC in our present case)
         // Here ext[i] enters through the base class RigidObject
      for (j=1;j<=ne;j++){ 
         if (ext[j].isActive()) fi+=pi.forces(ext[j].getField());
      }
// forces by homogeneous fields
      Spc xi=pi.get_x();
      FieldSpace fvi=filFar[i];
      if (SysControl::veryFar) fvi+=filVeryFar(xi); // += was + till
      // 02-05-31
      fi+=pi.forces(fvi);
         // all influence of fvi on velocities are thus taken into
         // account; we need fvi later also as a quantity that 
         // causes the electric polarization of pi to change
         // (see call of function react).
// forces from particles
      Z nld=nearList[i].size();
#if defined(CPM_USE_STL)
      for (j=0;j<nld;j++) fi+=force(pi,par[nearList[i][j]]);
#else
      for (j=1;j<=nld;j++) fi+=force(pi,par[nearList[i][j]]);
#endif      
            // includes mirror charges, member of CPM_TAL, not 
            // of Par
      Z mld=medList[i].size();
#if defined(CPM_USE_STL)
      for (j=0;j<mld;j++) fi+=force(pi,par[medList[i][j]]);
#else
      for (j=1;j<=mld;j++) fi+=force(pi,par[medList[i][j]]);
#endif

      for (j=1;j<=ne;j++){
         if (ext[j].fieldOnly()) continue;
         fi+=ext[j].interaction(pi,tf-te[j]);
         // tf-te[j] has no influence on the returned force, but it
         // influences the state change of ext[j]
      }
      forces[i]=fi;
      fValues[i]=fvi;
   }

// moving the particles according to interaction

   for (i=1;i<=np;i++){
      if (!active[i]) continue;
  //    if (cpmverbose>=mL2) cpmprogress("interaction loop",(i-1)*npInv);
      par[i].react(dt,forces[i],fValues[i]);
         // Changes velocities according to forces.
         // For Par=Clust (as in class Station) this call includes
         // building up polarization of the particle through the
         // electric component of fValues[i]. See
         // Clust::react(R, const ForTor&, const FieldSpace&)
         // and void Clust::setElcDipole(R dt, const VecE& E);
         // I did not remember to have this included when scrutinizing
         // this part in 2001-11-20. A remark to Clust::ej says
         // 'addition 2001-4-1'.
      par[i].osc(tf);
   }

   for (j=1;j<=ne;j++) ext[j].osc(tf);
        // linear evolution at the end. Also
        // inactive systems should do this since
        // they should be in the right place if reactivated 

   cpmmessage(mL,"stepBasic(...) done, act= "&cpmwrite(act));
}

CPM_TAL
Z CPM_FCN::stepControled(R& dt,const VectorField& filVeryFar)
{
   const Z mL=2;
      // controls verbosity
   R eps=SysControl::acceptable;
   R shiftMax=SysControl::bad;
   R incr=SysControl::increase;
      // controls growing of step size under favorite conditions

   cpmmessage(mL,"void CPM_FCN::stepControled(...) started");

   R t=get_t();
   R tLimit=get_tLimit();
   R dtLimit=tLimit-t;
   if (dt>dtLimit) dt=dtLimit;
   R tf=t+dt;
   updateActive(tf);
   Type sMem=*this;
   Type sf=*this;
   sf.stepBasic(t,dt,filVeryFar);
      // full step
   Z i,res=0;
   bool success=false;
   while(!success){
      // we may arrive here after having gone many times through the
      // non-success path und never through a success path.
      // Form this we see that sMem has the same value as after
      // initialization, and further: *this=sMem.
      // sf also has the right value corresponding to a full step
      // of the new dt as step size.
      res++;
      tf=t+dt;
      R tau=0.5*dt;
      stepBasic(t,tau,filVeryFar);
      Type sh=*this;
      stepBasic(t+tau,tau,filVeryFar);
         // as a result of this operation we have
         // par[i].get_t()=tf for all particles i
// Loop that analyses whether the steps were acceptable in terms of
// accuracy
      Z nActive=0;
      Z nPromoted=0;
      R pd=0;
      for (i=1;i<=np;i++){
         if (!active[i]) continue;
         nActive++;
         R pdi=par[i].dis((sf.par)[i]);
            // difference between a full step and two consecutive half
            // steps
         R shi=par[i].dis((sMem.par)[i]);
            // shift of particle i during two consecutive half steps
         pd+=pdi;
            // forming a accuracy measure for documentation

         if(pdi<=eps && shi<=shiftMax){
               // then par[i] is considered acceptable as the
               // state of particle i at time tf
            nPromoted++;
            tPar[i]=tf;
            active[i]=false;
               // i has reached its near-term goal, thus needs no
               // longer to be active in the present process.
               // Notice that got promoted here is not yet final.
               // If not sufficiently many companions can
               // get promoted too, we have to set back also
               // this successful particle.
         }
         else{
            par[i]=(sMem.par)[i];
               // par[i] needs to stay in active mode.
               // This particle gets set back in time since attempt for
               // promotion did not work
         }
      }

      if (nActive) pd/=nActive;   // getting the mean value for storage

      if (nActive==0){
         success=true;
         done(dt,pd); // pd=0 in this case
         dt*=incr;
         if (cpmverbose>=mL){
            ostringstream mes;
            mes<<"success="<<success<<" nActive="<<nActive<<
            " new dt="<<dt;
            cpmmessage(Word(mes.str()));
         }
      }
      else if (nActive==nPromoted){
         success=true;
         done(dt,pd); // pd has meaningful value here
         dt*=incr;
         if (cpmverbose>=mL){
            ostringstream mes;
            mes<<"success="<<success<<" nActive=nPromoted="<<nActive<<
               " new dt="<<dt<<" pd="<<pd<<" res="<<res;
            cpmmessage(Word(mes.str()));
         }
      }
      else if (nPromoted >= CpmRoot2::sup<Z>(1,nActive*pr.fraction) ){
         success=true;
         done(0.,pd); // for not loosing the information on
         dt=tau; // to be able to promote the next fraction in
            // the next call of stepControled(...)
         if (cpmverbose>=mL){
            ostringstream mes;
            mes<<"success="<<success<<" nActive="<<nActive<<
             "nPromoted="<<nPromoted<<
               " new dt="<<dt<<" pd="<<pd<<" res="<<res;
            cpmmessage(Word(mes.str()));
         }
      }
      else{
         dt=tau; // this is a reduction by 0.5 since dt=2*tau is normal
         tf=t+dt;
         if (dt>pr.tSmall){
            success=false;
            sf=sh;
            // got from previous step, this would not work if we would
            // reduce the step size by another factor than by 0.5
            *this=sMem;
            // this includes the fields active and filFar. This implies
            // that after an unsucessful atempt one starts from scratch
            updateActive(tf);
            // tf changed !
            // tPar also taken from sMem

            if (cpmverbose>=mL){
               ostringstream mes;
               mes<<"success="<<success<<" nActive="<<nActive<<
                  " tried to promote="<<nPromoted<<
                  " new dt="<<dt<<" pd="<<pd;
               cpmmessage(Word(mes.str()));
            }
         }
         else{ // we do the step and stop all particles
               // and declare this workarround a success
            success=true;
            *this=sh;
            for (i=1;i<=np;i++){
               par[i].stop();
               par[i].osc(tf);
               tPar[i]=tf;
            }
            done(pd);
            dt=pr.tSmall;
            res=-1;
            if (cpmverbose>=mL){
               ostringstream mes;
               mes<<"success="<<success<<
               " new dt after stop ="<<dt<<" pd="<<pd<<" res="<<res;
               cpmmessage(Word(mes.str()));
            }
         }
      }
   } // while(...)
   return res;
}

CPM_TAL
void CPM_FCN::stepAsy(R& dt,const VectorField& filVeryFar)
{
   const Z mL=2;
   if (cpmverbose>=mL || ((cpmverbose>0)&&(dt<0)))
      cpmmessage("void CPM_FCN::stepAsy(R& dt,filVeryFar) started, dt="
         &cpmwrite(dt));
   if (dt<0) return;
Label:
   R t=get_t();
   R tLimit=get_tLimit();
   R tf=cpminf<R>(t+dt,tLimit);
   update();
   Z counter=0;
   while (get_t()<tf){
      Z res=stepControled(dt,filVeryFar);
      if (res<0) goto Label; // may only happen if
         // SysControl::bad is too large or forces are too strong
         // for time steps larger than tSmall
      else counter+=res;
   }
   adjust_dt(dt,counter);
   if (cpmverbose>=mL){
      ostringstream mes;
      mes<<"void CPM_FCN::stepAsy(R& dt) done, dt="<<dt<<
         "counter="<<counter;
      cpmmessage(Word(mes.str()));
   }
}

CPM_TAL
void CPM_FCN::step(R dt,const VectorField& filVeryFar)
{
   const Z mL=2;
   Z fS=SysControl::fixedStep;

   static Z stepCount=0;
   R tcStart,tcEndOfWork, tcFinal;
      // tc : computation time
   R tmStart=get_t(),tmFinal;
     // tm : modeling time = image of real time in our model
   Z nStart=get_np(), nFinal;

   if (cpmverbose>0){
      tcStart=cpmtime();
   }
   cpmmessage(mL,"CPM_FCN::step("&cpmwrite(dt)&") started");
 //  enforceGeneralUpdate(); // was missing till 02-06-14
      // we need to take into account that interaction with other
      // processes at the end of step changed everything in the present
      // process (e.g. the particle number) in a manner that could not be
      // predicted within the given process. We also would not rely on 
      // a constructor to ensure the preconditions for function step(). 

   if (nStart==0){
      cpmmessage("step() started with no particles");
      stepBasic(tmStart,dt,filVeryFar);
      done(dt,0);
   }
   else{
      if (fS){
         R dt_fS=dt/fS;
         for (Z j=1;j<=fS;j++){
            if (fS>0 && cpmverbose>1) cpmvalue("frac",(j-1.)/(fS-1.),3);
            R t=get_t();
            R tf=t+dt_fS;
            update();
            stepBasic(t,dt_fS,filVeryFar);
            for (Z i=1;i<=np;i++) tPar[i]=tf;
            done(dt_fS,0); // faster than done(0)
         }
      }
      else{
         const R red=0.25;
         R tIni=get_t();
         R tFinal=tIni+dt;
         R ts=tIni;
         R dtM=get_dt();
         if (dtM<=0.) dtM=red*dt;
         R dtProp=dtM; // 'dt proposed'
         set_tLimit(tFinal);
         while (ts<tFinal){
            ts=get_t();
            if (dt>0 && cpmverbose>1) cpmvalue("frac",(ts-tIni)/dt,3);
            stepAsy(dtProp,filVeryFar);
         }
      }
      if (getRelaxFlag()) stopParticles();
   }

   tcEndOfWork=cpmtime();

   if (cpmverbose>0){
      stepCount++;
      tcFinal=cpmtime();
      tmFinal=get_t();
      nFinal=get_np();
      ostringstream ost;
      ost<<endl<<"Performance data for step # "<<stepCount<<endl;
      ost<<"n: particle number: prior to step "
         <<nStart<<", after step "<<nFinal<<endl;
      R tcTotal=tcFinal-tcStart;
      R tcSynOverhead=tcFinal-tcEndOfWork;
      R percentOverhead=100*tcSynOverhead/tcTotal;
      R tmTotal=tmFinal-tmStart;
      ost<<"tc: computation time/s for step: "<<tcTotal<<endl;
      ost<<"actual model time/s for step: "<<tmTotal<<endl;
      ost<<"tm: scheduled model time/s for step: "<<dt<<endl;
      ost<<"synchronization overhead in percent: "<<percentOverhead<<endl\
      ;
      if (tcTotal>0){
         R eff=(tmTotal*nStart)/tcTotal;
         ost<<"efficiency=(tm*n)/tc= "<<eff;
      }
      cpmcerr<<ost.str()<<endl;
   }
   cpmmessage(mL,"CPM_FCN::step("&cpmwrite(dt)&") done");
}

CPM_TAL
void CPM_FCN::run(R dtStep, Z nSteps, Z filingPeriod, Z viewingPeriod,
   Z stoppingPeriod, const Word& movieName, Camera& cam,
   const Timer& tm, const F<R,Z>& f) 
      // was const F<R,V<R> >& f till 02-06-03
      // return value not needed (void does not work with my function
      // objects since 'unknown type' for the MS compiler
{
   static SysControlState sc;
   Word loc("CPM_FCN::run(...)");
   const Z mL=1;
   cpmmessage(mL,loc&" started");
   if (nSteps<0 || dtStep<0){
      cpmmessage(mL,loc&" trivially done");
      return;
   }
   sc.mem(); // memorizing the state of SysControl since 
      // SysControl may be changed by calling f 
   CyclicAlarm ca1(filingPeriod,filingPeriod);
   CyclicAlarm ca2(viewingPeriod,viewingPeriod);
   CyclicAlarm ca3(stoppingPeriod,stoppingPeriod);
      // so the next call to operator() will give true
      // unless ...Period<1
   bool earlyStop=false;
   if (ca1()) earlyStop=file(movieName,tm);
   if (ca2()) mark(cam);
   if (nSteps==0 || earlyStop) return;
   R tFinal=dtStep*nSteps;
   R tFinalInv=1./tFinal;
   R nStepsInv=1./nSteps;
   R tAct=get_t();
   while (tAct<tFinal){
      f(tAct);
      step(dtStep);
      tAct=get_t();
      cpmprogress("run-loop",tAct*tFinalInv);
      if (ca1()) earlyStop=file(movieName,tm);
      if (ca2()) mark(cam);
      if (ca3()) stopParticles();
      if (earlyStop) break;
   }
   sc.act();
      // restoring the state of SysControl which may have been
      // changed by the action of f

   cpmmessage(mL,loc&" done");
}


////////////////////////// end of main dynamics stuff //////////////////

/*
CPM_TAL
Z CPM_FCN::remove(const F<Spc,B>& f)
// no inter-process communication needed since each process
// can do this on its own. Oho! wat's about particle id's
{
   Z i,mL=0;
   Word loc("CPM_FCN::remove(const F<Spc,B>& f)");
   cpmmessage(mL,loc&" started");
//   if (np==0) return;
   V<bool> vb(np);
   for (i=1;i<=np;i++) vb[i]=((f(par[i].get_x())).not()).and(owned[i]);
   par=par.select(vb);
   np=par.dim();
   if (np==0) cpmwarning(" all particles removed");
   owned=V<B>(np,B(true));
   Z1 np1(np);
   Vio<Z1> npAll(size, Z1());
   Cpm_com.gather(np1,npAll);
      // now the np-values of all processes are available for rank 1.
   
   Z2 idIni_nTot;
   V<Z2> idIni_nTotAll(size,Z2());
   if (rank==1){
      Z ns=0;
      for (i=1;i<=size;i++) ns+=npAll[i];

      V<Z> res(size);
      res[1]=0;
      for (i=2;i<=size;i++){
         res[i]=res[i-1]+npAll[i-1];
      }

      for (i=2;i<=size;i++) 
         idIni_nTotAll[i]=Z2(res[i],ns);

      idIni_nTot.x1=0;
      idIni_nTot.x2=ns;
   }
   Cpm_com.scatter(idIni_nTot,idIni_nTotAll);
   Z idIni=idIni_nTot.x1;
   Z nTot= idIni_nTot.x2;

   for (i=1;i<=np;i++){
      par[i].cancel_id();
      Z id=i+idIni;
      par[i].set_id(id);
   }
   redistribute();
   cpmmessage(mL,loc&" done nTot="&cpmwrite(nTot));
   return nTot;
}
*/

CPM_TAL
VectorField CPM_FCN::getFieldExt()const
{
   VectorField res;
   for (Z i=1;i<=ne;i++){
       res+=ext[i].getField();
   }
   return res;
}

CPM_TAL
VectorField CPM_FCN::getFieldPart()const
{
   VectorField res;
   for (Z i=1;i<=np;i++){
       res+=par[i].getField();
   }
   return res;
}

CPM_TAL
VectorField CPM_FCN::getFieldTot()const
{
   return getFieldExt()+getFieldPart();
}


CPM_TAL
void CPM_FCN::mark(const Camera& sc)const
{
   Z mL=3;
   cpmmessage(mL,"CPM_FCN::mark(Camera) started, np="&cpmwrite(np));
   Z i;
try{
   for (i=1;i<=np;i++){
        par[i].mark(sc);
   }
   if (SysControl::ignoreExternalSystems) return;
      // then nothing to do
}
catch(...){
   Error("Problem in ParDynSys<>::mark(Graph& gr)const, Particle loop");
}
try{
   for (i=1;i<=ne;i++){
      ext[i].mark(sc);
   }
}
catch(...){
Error("Problem in ParDynSys<>::mark(Graph& gr)const, External systems\
 loop");
}
try{
   if (pr.showPartFields) getFieldTot().mark(sc); else
      getFieldExt().mark(sc);
}
catch(...){
   Error("Problem in ParDynSys<>::mark(Graph& gr)const, field-part");
}
   cpmmessage(mL,"CPM_FCN::mark(Camera) done, np="&cpmwrite(np));

}

#undef CPM_CN
#undef CPM_TAL
#undef CPM_FCN

} // namespace

#endif
