//? cpmgroup.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_GROUP_H_
#define CPM_GROUP_H_
/*
   Description:
      general framework for defining (mathematical) groups
      project on studying categories in the framework of C++

   History: Started 96-12-19
      by changing categ.c, devoted to studying categories
      in the framework of C++
      97-6-19 test function added and
      2001-4-20 changed from group.tum

   Status:
*/
// Giving the mathematical structure of a group in a form which can
// be coded as an abstract class. Addition of operations (e.g. in
// going from abelian groups to rings) can be done by forming new
// abstract classes by derivation. The C++ language forces us not
// to chose the usual form of operations (e.g. operator * (.,.) )
// but something different for which the next class shows a working
// method.
// What this class defines is a embryonic form of a group structure
// from which the final form is developed by forming a template Group<>.

////////////////////////////////////////////////////////////////////////
//                           Group Embryo or Pre-Group                //
////////////////////////////////////////////////////////////////////////

#include <cpmnumbers.h>
   // here only to write 'R' instead of 'double'

namespace CpmAlg{

   using CpmRoot::R;

class Group_ {   // abstract base class 'group embryo' or 'pre-group'
   // for construction of groups. Actually, we deal only with the syntax.
   // That, for instance, the binary operation is associative is in the
   // responsibility of the programmer who defines mult()
   // When deriving a class G_ from Group_ one has to implement the
   // default constructor in a way that it creates a unit element, i.e.
   // an element e  satisfying  g.mult(e)==g for all g in G_.

public:

   virtual void mult(const Group_& g)=0;
      // turns *this into the group product of *this and g.
      // (right product (*this)*g ).  Notice that
      //
      // Group_ operator * (const Group_& g);
      //
      // which corresponds to a normal notion of a product in C++
      // is not an allowed declaration for an abstract class.

      // When a derivation G:Group_ is formed, it is not sufficient
      // (and not necessary) to define mult(const G& g). Instead
      // one actually has to define mult(const Group_& g). Code could
      // start
      // { G g1=(G&)g; .... statements using g1 ...}
      // See Ellis,Stroustrup: The annotated C++ Reference Manual
      // Section 5.4 p.69 second part of the example block:
      // ' I know that g refers to a G'

   virtual void inv(void)=0;
      // turns *this into its group inverse
};

// To avoid the slight non-canonicity in redefining mult() one may choose
// not
// to derive from an abstract group but to verbally implement a interface
// which can be inserted into any code by the following macro. This
// technique
// is widely tested within KAGUM classes. Note that you have to put
// the final ';' youself. The correct usage is
//
// class Gp {
// ....
// public:
//  PRE_GROUP_INTERFACE(Gp);
// ....

#define PRE_GROUP_INTERFACE(X)\
   X(void);\
   X(const X&);\
   void mult(const X&);\
   void inv(void)

////////////////////////////////////////////////////////////////////////
//                           Group                                    //
////////////////////////////////////////////////////////////////////////

template <class X>   // X has to be a public derivation from Group_
   // this may be expressed formally as  X:Group_ .
   // Alternatively (actually, preferably) it may provide the
   // corresponding
   // structure without actually being declared as a derivation.
   // In addition, X has to define a default constructor
   // which generates a unit element (see Group_)
   //
   // Precisely:
   //
   // X(void);
   // X(const X&);
   // void mult(const X&);
   // void inv(void);
   //
   // Notice that not even an assignment operator is needed

class Group : public X {
   // contains all algorithms for convenient operations deduced
   // from rudiments mult, inv preformed in Group_.
   // The binary operation in the group is written multiplicatively;
   // the inverse of g is written as !g.

public:

   Group(void):X(){}
      // constructs the unit (neutral) element

   Group(const Group<X>& g):X(g){}
      // copy constructor

   Group(const X& xIn):X(xIn){}
      // 'upcast constructor'
      // constructor from X instances. X will normally have a constructor
      // X(const P& p) with an argument from some suitable parameter
      // space.
      // Notice, that althoug then no constructor Group<X>(const P& p) is
      // defined, we nevertheless may write code like
      //
      // P p=...;
      // Group<X> g(p);
      //
      // since p will automaticly be converted to an instance of X, which
      // then will be 'upcasted' to Group<X>.
      // This would not work if we would use two arguments in
      // paramatrizing:
      //
      // X(const P1& p1, const P2& p2).
      //
      // If such a constructor from p1 and p2 is desirable one has to
      // derive a class G from Group<X> without introducing new data
      // members and define the desired constructor there:
      //
      // G(const P1& p1, const P2& p2):Group<X>(X(p1,p2)){}
      //
      // Further, defining the  'upcast constructor'
      //
      // G(const Group<X>& g): Group<X>(g){}
      //
      // is needed

   Group<X>& operator *=(const Group<X>& g){ mult(g); return *this;}
      // g*=g1 is equivalent to g=g*g1;

   Group<X>& operator /=(const Group<X>& g)
      // g/=g1 is equivalent to g=g*(!g1);
   {  Group<X> gRes(g); gRes.inv(); mult(gRes); return *this;}

   Group<X> operator !(void)const
   { Group<X> g(*this); g.inv(); return g;}
      // !g is the inverse of g
      // notice that since inv() is not const, we should not apply inv()
      // directly to *this

friend Group<X> operator *(const Group<X>& g1,const Group<X>& g2)
   // group product
   { Group<X> g(g1); return g*=g2; }

friend Group<X> operator /(const Group<X>& g1,const Group<X>& g2)
   // group division
   { Group<X> g(g1); return g/=g2; }

// operator = has not to be redefined

friend R test(const Group<X>& g1, const Group<X>& g2, const Group<X>& g3)
{
   Group<X> g0;
      // should be the unit (neutral) element
   Group<X> g10=g1*g0;
      // should be equal to g1
   Group<X> g01=g0*g1;
      // should be equal to g1
   Group<X> g00=g1/g1;
      // should be equal to g0
   Group<X> g1i=!g1;
      // inverse of g1
   Group<X> g0L=g1*g1i;
      // should be the unit element
   Group<X> g0R=g1i*g1;
      // should be the unit element
   Group<X> g12=g1*g2;
   Group<X> g23=g2*g3;
   Group<X> g12_3=g12*g3;
   Group<X> g1_23=g1*g23;
      // associativity says that g12_3 and g1_23 are equal

   R sum=0;
   sum+=dis(g1,g10);
   sum+=dis(g1,g01);
   sum+=dis(g0,g00);
   sum+=dis(g0,g0L);
   sum+=dis(g0,g0R);
   sum+=dis(g12_3,g1_23);
   return sum;
}

};

} // namespace

#endif       
