//? cpmlatex.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPMLATEX_H_
#define CPMLATEX_H_
/*

Purpose: Declaration of classes for creating diagrams in LATEX
History: 2004-11-22 started

*/

#include <cpmrmatrix.h>
#include <cpmvectorgraphics.h>

namespace CpmLatex{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::C;
   using CpmRoot::Word;
   using CpmRootX::B;
   using CpmRootX::Z1;
   using CpmLinAlg::R2;
   using CpmLinAlg::R3;

   using CpmArrays::R_Matrix;
   using CpmArrays::R_Vector;
   using CpmArrays::V;
   using CpmGeo::Iv;

class Figure{

// independent data (all belong to a class with default constructor)
   Iv rx_;  // range in x
   Iv ry_;  // range in y

   Word nx_; // name of the quantity to which the x-axis refers
   Word ny_; // name of the quantity to which the y-axis refers

   Word tit_; // title of the diagram

   Z1 withOri_;
      // withOrigin!=0 controls autoscaling such that the origin
      // is always on the diagram
   Z1 newGrid_;
      // only =0 or !=0 matters. If it is 0 (should be an acception) one
      // let the grid remain unchanged and draw curves on the frame with
      // a meaning of object space wich is no longer in agreement with
      // the grid and the grid text
   Z1 autoScl_;
      // only =0 or !=0 matters. Influences show(..) such that CL CU are
      // adjusted to the data to be represented prior to drawing them.
      // 'autoscaling'
      // The change of CL and CU remains till new values are assigned
      // by new action
   Word gridText_;
   R2 gridTextPos_;
      // text position is understood in the Frame coordinate system
      // (xF,yF)
      // i.e. (0,0) is left lower corner, (1,1) is right upper corner

   VecGraphs vg_;

public:
   typedef Figure Type;

   CPM_IO
      // It is not the intention of this I/O operations
      // to save the content of video memory, and other content data
// constructors and assignement

   Figure(void):rx_(0,1),ry_(0,1),nx_("x"),ny_("y"),tit_("untitled"){}

// setting functions

   void setWithOrigin(Z);
      // setWithOrigin(1) controls autoscaling such that the origin
      // is always on the diagram
   void setNewGrid(Z);
   void setAutoScale(Z);
   void setGridText(Word a);
   void setGridTextPosition(const R2& pos);
   void setGridTextPosition(R xText, R yText);
      // 0<=xText<=1, 0<=yText<=1, sets position of left-lower corner
      // of grid text window within the frame of the Figure

// modern style making use of class Iv (interval)
   bool setX(Iv iv){ rx_=iv;}
      // sets the corners in object space X direction according to the
      // interval iv
   bool setY(Iv iv){ ry_=iv;}
      // sets the corners in object space Y direction according to the
      // interval iv. Also sets autoScale to 0.
   bool setXY(Iv ivX, Iv ivY){ rx_=ivX; ry_=ivY;}
      // sets the corners in object space according to the
      // intervals ivX, ivY. Also sets autoScale to 0.
   Iv getIvX()const{ return rx_;}
   Iv getIvY()const{ return ry_;}

   R getAspRat()const{ return rx_.abs()/ry_.abs();}
      // Landscape format gives values larger than 1
      // Portrait < 1;
      // image width / image height

// access fuctions to corner coordinates in object space

// changing the size in object space of the window

bool scale(R magnification);
   // argument >1 means that the rectangle in object space gets enlarged
   // by a factor 'magnification' in linear dimensions. The center of the
   // window remains fixed in object space. The display size is not
   // affected.
   // If the argument is < 1, the rectangle in object space gets reduced.

void shift(R sx, R sy);
   // shifts along X-axis and  Y-axis in object space. The display size
   // is not affected. sx and sy are to be understood as fractions of dX
   // and dY. Thus g.shift(0.5,0.5) will always make the old center the
   // new left-lower corner.

void shiftAbs(R sx, R sy);
   // shifts along X-axis and  Y-axis in object space. The display size
   // is not affected. sx and sy are to be understood as given relative
   // to units in object space.

void getObjectSpaceFrom(const Figure& gr );
   // takes the object space related data CL, CU, dX, dY from gr and
   // changes the corresponding data of *this as to become identical to
   // them.

// Setting the window according to the size of objects to be
// represented graphically. Overloading with differently
// typed arguments may be added later according to need

void setXY(const R_Vector& x, const R_Matrix& y);
/*
   (x,y) represents a family of curves. x contains the x-values
   and y[i][j] are the y-values belonging to x=x[i].
   The values of CL, CU such that these curves
    fill the window exactly
*/

void setXY(const R_Vector& x, const R_Vector& y);

void setXY(const R_Matrix& y);

void adjustXY(Z divX=10, Z divY=10);
// Enlarges the window such that the boundaries correspond to values
// which in scientific number format give short and convenient
// numbers. Also a grid is drawn which subdivides the window
// conveniently into approximately divX subintervals in the x-axis
// and divY subintervals in the y-axis. The actual grid may deviate
// from these numbers since only convenient values for the grid line
// coordinates are accepted (for instance, the unit interval will
// be devided into 2 or 4 subintervals but never in 3).

void adjustScaleY(void);
// The origin CL and the X-interval [CL.x1,CU.x1] are left unchanged
// and CU.x2 is adjusted such that a circle in object space (actually,
// object plane) will be mapped into a circle (as opposed to an
// ellipse) on screen. The essential part of this operation is to
// set an appropriate scale in Y direction, which explains the name
// of the function.

void adjustScaleX(void);
// The origin CL and the Y-interval [CL.x2,CU.x2] are left unchanged
// and CU.x1 is adjusted such that a circle in object space (actually,
// object plane) will be mapped into a circle (as opposed to an
// ellipse) on screen. The essential part of this operation is to
// set an appropriate scale in X direction, which explains the name
// of the function.


Z   adjust(R2& x)const;
// If x is very near to one of the 4 infinite lines defining the window,
// x is shifted to sit exactly on this line. This is sometimes necessary
// for making decisions like in function boundary meaningful. The return
// value counts the number of adjustments done. The return value ranges
// from 0 (no adjustment) to 2. (XU>XL,YU>YL assumed.)

bool indFunc(const R2& p)const;
// returns true if p belongs to the object space rectangle associated
// with *this and false else.


void  grid(Z n_x=4, Z n_y=4);
/*
Draws a grid by dividing x-range in n_x equally sized parts, and y-range
into n_y such parts.
*/

void writeGridText(void);
// writes the grid text


void addText(const Word& text, R x=0.05, R y=0.15, Z font=1,
   Z charSize=0);
// Writing a text string to the graph (x,y) Position of text starting
// point. (0,0) is left lower corner.
// font: see addText of class Viewport if charSize=0 size will be
// determined automatically from existing space. Detailed size control
// is beneficial for publication quality graphics

void addFrame(void);
// marks the boundary of the Graph by a line in GraphColor

bool show(Word const& fileName, R_Vector const& x, R_Vector const& y,
   Word const& title, R xPosRel);

};

} // namespace

#endif

