//? cpmmaxw1d_070604a.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_MAXW1D_H_
#define CPM_MAXW1D_H_
/*
   cpmmaxw1d.h
   version  2007-05-26
   Author: Ulrich Mutze   
      ulrichmutze@aol.com
   
   Description: Development and test of a class
      that represent a Maxwell field in a 1-dimensional cavity.
      The system under consideration is that introduced in
      Hans De Raedt et al. Computer Physics Communications 156 (2003)\
       43-61
      in section Implementation
   History: Started 2007-05-18
   
*/

#include <valarray>

#include <cpmgraph.h>
#include <cpmrfunc.h>
#include <cpmrvector.h>

using CpmRoot::Z; // int by typedef
using CpmRoot::R; // double by typedef
using CpmRoot::Word; // character strings
using CpmRoot2::R1; // character strings
using CpmLinAlg::R2; // pairs of R
using CpmGraphics::Graph; // graphics
using CpmFunctions::R_Func; // functions R --> R
using CpmArrays::V; // array template
using CpmArrays::X2; // Cartesian product of two factors
using CpmArrays::R_Vector;// R-valued array
using CpmGeo::Iv; // intervals

using namespace std;

namespace CpmWaves{

class TEM {
// named TEM according to 
// Jackson: Classical Electrodynamics, Wiley 1963, Sect. 8.2
// Actually there are two TEM modes: one with E and H perpendicular,
// and one with E and H parallel. The first one is treated in De Raedt et
// al.,
// and is the one to be treated here.
// Names of data members and of non-constant member functions end in '_'.
// 

   Z n_;
      // any positive integer
   Z d_;
      // the odd number 2*n_+1
   R L_;
      // distance from wall to wall
   R s_;
      // spacing between neigbors
      // s_=L_/(2*n_)=L_/(d_-1)
   R sInv_;
      // 1/*s_
   R s2Inv_;
      // 1./(s_*2)
// all fields depend only on x (0<=x<=L_), E is E_z, H is H_y
   valarray<R> x_;
      // equi-distant positions: x_[0]=0, x_[d_-1]=L_
   valarray<R> X_;
      // mixed magnetic and electric components according to the Yee
      // scheme.
      // 
      // See the 2003 Computer Physics Communications article by Hans De
      // Raedt
      // et al. Fig.2. I found it convenient to include in X_ the
      // vanishing
      // E-component at the boundaries. 
   valarray<R> dX_;
      // time derivative of the field components
   void ini_(){ s_=L_/(2*n_); sInv_=1./s_; s2Inv_=0.5*sInv_; }
      // setting the quantities depending on L_ and n_
   valarray<R> der(valarray<R> const& v)const
      // Time derivative of the field values v in Yee order. 
      // Expressed by Maxwell's equations as a spatial derivative.
      // Vanishes at the two boundary points.
   {
      valarray<R> res(0.,d_);
      for (Z i=1;i<d_-1;++i) res[i]=(v[i+1]-v[i-1])*s2Inv_;
      return res;
   }

   valarray<R> acc(valarray<R> const& v)const{ return der(der(v));}
      // acceleration as the second time derivative.

public:

   TEM(){}

   TEM(Z n, R L, R_Func const& fE, R_Func const& fH);
      // fE, and fH are defined on the interval [0,1].
      // n is the number of E-values and of H-values
      // The number of X_-values is 2*n+1. (n H,E-pairs and a H value at
      // the
      // end.
      // this is a very clever way to suggest a uniform definition of the
      // time 
      // derivative of the field values in a manner that the vanishing of
      // 
      // the electric field on the boundary is guaranteed

   TEM(Z n, R L, Iv const& iv);
      // E-field has stochastic values uniformly distributed in iv and H
      // is
      // zero. 

   void step_(R dt)
      // time step according to the direct midpoint method
   {
      R tau=0.5*dt;
      X_+=dX_*tau;
      dX_+=acc(X_)*dt;
      X_+=dX_*tau;
   }

   R energy()const{ return (X_*X_).sum();}
      // Total energy (some re-interpretation is involved since verbatim
      // the field energy of homogeneous fields in an infinite
      // volume is infinite. Turns out to be constant to good
      // approximation.

   R dissipation()const{ return (X_*dX_).sum()*2;}
      // Time derivative of total energy. Is small since the
      // total energy is close to being constant

// Connection to arrays of type R_Vector and V<R2> since all my
// graphical representation functions work with them. Their
// index range starts with 1 instead of the C-native 0.

   R_Vector getXr()const
   {
      R_Vector res(d_);
      for (Z i=0;i<d_;++i) res[i+1]=X_[i];
      return res;
   }

   R_Vector getxr()const
   {
      R_Vector res(d_);
      for (Z i=0;i<d_;++i) res[i+1]=x_[i];
      return res;
   }

   V<R2> getX()const
   {
      V<R2> res(d_);
      for (Z i=0;i<d_;++i) res[i+1]=R2(x_[i],X_[i]);
      return res;
   }

   V< V<R2> > getEH()const
      // extracting E and H-fields in a way which is suitable for 
      // graphical representation
   {
      V<R2> vE(n_+1); // boundary values included
      V<R2> vH(n_); // d_=n_+(n_+1)
      Z iE=1,iH=1;
      bool putE=true;
      for (Z i=0;i<d_;++i){
         if (putE){
            vE[iE++]=R2(x_[i],X_[i]);
            putE=!putE;
         }
         else{
            vH[iH++]=R2(x_[i],X_[i]);
            putE=!putE;
         }
      }
      return V< V<R2> >("",vE,vH);
   }
   
   void showAsCurves(Graph& g)const;
      // auto-scaled polygon representation of the fields.
      // E as a red curve, B as a green one. Based on 
      // the previous function.
};

namespace{

   class RV2{
      R_Vector E_;
      R_Vector H_;
      R_Vector J_;
      R1 fs_;
public:
      RV2(R_Vector const& E, R_Vector const& H, R fs):E_(E),H_(H),fs_(fs)
      {
         if (E_.dim()!=H_.dim()) 
         cpmerror("RV2(R_Vector,R_Vector): dimension mismatch");
      }
      RV2(){}
      explicit RV2(Z n):E_(n),H_(n),fs_(1){}
      RV2& operator+=(RV2 const& a){ E_+=a.E_; H_+=a.H_; return *this;}
      RV2& operator*=(R r){ E_*=r; H_*=r;return *this;}
      RV2 operator*(R r)const{ return RV2(E_*r,H_*r,fs_);}
      
      RV2 dot()const{ return RV2(derH(H_)*fs_-J_,derE(E_)*fs_);}

      RV2 acc()const{ return dot().dot();}
      R_Vector c1()const{ return E_;}
      R_Vector c2()const{ return H_;}
      void setE_(Z i, R val){E_[i]=val;}
      void setH_(Z i, R val){H_[i]=val;}
      void setJ_(R_Vector const& J){ J_=J;}
      static R_Vector derE(R_Vector const& v)
         // spatial derivative (up to a factor) of a field v which
         // vanishes at the ends (as the E-field does)
      {  
         Z n=v.dim();
         R_Vector res(n);
         for (Z i=1;i<=n;++i) res[i]=(v.read(i+1,0)-v.read(i-1,0));
         return res;
      }

      static R_Vector derH(R_Vector const& v)
        // spatial derivative (up to a factor) of a field v which needs
        // not vanish at the ends (as the E-field does)
      {
         Z n=v.dim();
         R_Vector res(n,0);
         for (Z i=1;i<=n;++i) res[i]=(v.read(i+1,2)-v.read(i-1,2));
         return res;
      }
    
   };
}

class TEM2 {
// class TEM in cpm style
// discretization does not make use of Yee scheme
// works fine

   Z n_;
      // any positive integer
   R L_;
      // distance from wall to wall
   R s_;
      // spacing between neigbors
   R sInv_,hsInv_, fac_;
      // 1/*s_
 
// all fields depend only on x (0<=x<=L_), E_ is E_z, H_ is H_y
   R_Vector x_;
      // equi-distant positions: x_[0]=0, x_[d_-1]=L_

   RV2 X_;

   RV2 dX_;

   R t_;
 
  // R_Vector dE_,dH_;
      // time derivative of the field components
   void ini_(){
      s_=L_/(n_-1); sInv_=1./s_; hsInv_=sInv_*0.5; fac_=hsInv_* hsInv_;
   }
      // setting the quantities depending on L_ and n_
public:

   TEM2(){}

   TEM2(Z n, R L, R_Func const& fE, R_Func const& fH);
      // fE, and fH are defined on the interval [0,1].
      // n is the number of E-values and of H-values
      // The number of X_-values is 2*n+1. (n H,E-pairs and a H value at
      // the
      // end.
      // this is a very clever way to suggest a uniform definition of the
      // time 
      // derivative of the field values in a manner that the vanishing of
      // 
      // the electric field on the boundary is guaranteed

   TEM2(Z n, R L, Iv const& iv);
      // E-field has stochastic values uniformly distributed in iv and H
      // is
      // zero. 

   void step_(R dt)
      // time step according to the direct midpoint method
   {
      R tau=0.5*dt;
      t+=tau;
      X_+=dX_*tau;
      X_.set(fJ(t));
      dX_+=X_.acc()*(dt*fac_);
      X_+=dX_*tau;
      t+=tau;
   }

   R energy()const{ return (X_.c1()|X_.c1())+(X_.c2()|X_.c2());}
      // Total energy (some re-interpretation is involved since verbatim
      // the field energy of homogeneous fields in an infinite
      // volume is infinite. Turns out to be constant to good
      // approximation.

   R dissipation()const{ return ((X_.c1()|dX_.c1())+(X_.c2()|dX_.c2()))*2\
   ;}
      // Time derivative of total energy. Is small since the
      // total energy is close to being constant

// Connection to arrays of type R_Vector and V<R2> since all my
// graphical representation functions work with them. Their
// index range starts with 1 instead of the C-native 0.

   R_Vector getE()const
   {
      return X_.c1();
   }

   R_Vector getH()const
   {
      return X_.c2();
   }

   R_Vector getxr()const{ return x_;}

   
   V< V<R2> > getEH()const
      // extracting E and H-fields in a way which is suitable for 
      // graphical representation
   {
      V<R2> vE(n_); 
      V<R2> vH(n_); 
      R_Vector E=getE();
      R_Vector H=getH();
      for (Z i=1;i<=n_;++i){
         R xi=x_[i];
         R Ei=E[i];
         R Hi=H[i];
         vE[i]=R2(xi,Ei);
         vH[i]=R2(xi,Hi);
      }
      return V< V<R2> >("",vE,vH);
   }
   
   void showAsCurves(Graph& g)const;
      // auto-scaled polygon representation of the fields.
      // E as a red curve, B as a green one. Based on 
      // the previous function.

   R_Vector getXr()const
   {
      return getE()&getH();
   }
};

} // namespace

#endif
