//? cpmparallel.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


#ifndef CPM_PARALLEL_H_
#define CPM_PARALLEL_H_
/*

    Description: Making a parallel 
    system out of a normal one. 
    I try a flexible implementation that allows to switch between
    single processor and mult-processor execution while
    using the same algorithm for the time step.
    This is not yet achieved!


   History: Started 2002-6-18 

*/
#include <cpmsystem.h>
#include <cpmgroupgraph.h>

namespace CpmDynSys{

   using CpmRoot::R;
   using CpmRoot::Z;
   using CpmRoot2::B;
   using CpmRoot::Word;

   using CpmSystem::Error;
   using CpmSystem::Timer;

   using CpmArrays::V;
   using CpmArrays::VV;
   using CpmArrays::Vio;

   using CpmGeo::GroupGraph;

#define CPM_CN Parallel
   // this style of introducing the class name helps to define 
   // new versions 
   // CN: class name

#define CPM_TAL template <class Sys,class Prc,\
class Spc,class Field,class Group,class Prt>
   // TAL: template argument list

#define CPM_FCN CPM_CN<Sys,Prc,Spc,Field,Group,Prt>
   // FCN: full class name


#if !defined(CPM_USE_MPI) // single processor version
   // of the 'parallel algorithm'

CPM_TAL
class CPM_CN{
   R1 t;
   Prt par;
   Z1 size;
   Z1 np;
   Vio<Sys> vsys;
   Vio< Vio<B> > owned;
      // for each rank owned[rank] is an array of bools
      // vsys[rank][i] is owned by rank iff owned[rank][i]==true 
   GroupGraph<Group> gg;
      // structure holding the geometrical transformations from one
      // rank to its neighbors
  
   void updateOwnership();
   void addUsedParticles();
   void redistribute();

public:

   typedef CPM_CN Type;
   typedef CPM_CN CloneBase; 
      // infrastructure for forming heterotypic arrays of
      // derived classes
   CPM_IO_V
      // stream I/O, important that this is virtual

   Parallel():par(),size(),np(),vsys(0),owned(0),gg(){} 
      // trivial default constructor     
   Parallel(const Prt& par, const Sys& sys);
      // constructor from basic ingredients
   Z dim()const{ return size;}
      // number of processes
   Z get_np()const{ return np;}
      // total number of particles
   Prt getPartition()const{ return par;} 
      // returns the object that organizes the assignment of particles
      // to processes
   R get_t()const{ return t;}
      // returns the point in time for which the data given in
      // vsys are valid
   void stopParticles()
      // setting the velocities of all particles to zero
   { for (Z i=1;i<=size;i++) vsys[i].stopParticles();}

   void step(R dt);
      // changing the state *this into the one which is valid at
      // the point t+dt in time. This function is the raison d'etre 
      // for class Parallel. The algorithm is best explained by the
      // actual code

   Sys& operator[](Z i){ return vsys[i];}
      // accessing the process data for changes
   const Sys& operator[](Z i)const{ return vsys[i];}
     // accessing the process data for changes for 'read only'

   virtual bool file(const Word& fn, const Timer& tm)const{ return true;}
      // writing on file not yet implemented
};

CPM_TAL
bool CPM_FCN::printOn(ostream& str)const
{
   CpmRoot::writeTitle("Parallel",str);
   cpmp(t);
   cpmp(par);
   cpmp(size);
   cpmp(np);
   cpmp(vsys);
   cpmp(owned);
   cpmp(gg);
   return true;
}

CPM_TAL
bool CPM_FCN::scanFrom(istream& str)
{
   cpms(t);
   cpms(par);
   cpms(size);
   cpms(np);
   cpms(vsys);
   cpms(owned);
   cpms(gg);
   return true;
}

CPM_TAL
void CPM_FCN::step(R dt)
{
   const Z mL=2;
   cpmmessage(mL,"Parallel::step(R dt) started, size="&cpmwrite(size));

   V<Field> f1(size); 
   Z i,j;
   for (i=1;i<=size;i++){
      f1[i]=vsys[i].getFarField();
   }
   V<Field> f2(size,Field());
   for (i=1;i<=size;i++){
      for (j=1;j<=size;j++){
         if (j==i) continue;
         if (gg.close(i,j)){
            Group gij=gg.group(i,j);
            Field fij=f1[j];
            fij*=gij;
            f2[i]+=fij;
         }
      }
      cpmvalue("proc",i,4);
      vsys[i].step(dt,f2[i]);
   }
   redistribute();
   t=R1(t+dt);
   cpmmessage(mL,"Parallel::step(R dt) done, size="&cpmwrite(size));
}


CPM_TAL
CPM_FCN::CPM_CN(const Prt& par_,const Sys& sys):par(par_)
{
   const Z mL=1;
   const Z mL2=mL+1;
   Word loc("Parallel(const Prt& par_,const Sys& sys)");
   cpmmessage(mL,loc&" started");
   size=par.card();
   np=sys.get_np();
   cpmmessage(mL2,"size="&cpmwrite(size));
   cpmmessage(mL2,"np="&cpmwrite(np));
   if (size>1){
      Sys sy0(sys);
      sy0.delPar();
      vsys=V<Sys>(size,sy0);
      owned=Vio< Vio<B> >(size,Vio<B>(0));
      Z i,rank;
      V<Spc> vs(np);
      for (i=1;i<=np;i++) vs[i]=sys.getPar(i).get_x();
      gg=par.groupgraph(vs);
      Z count=0;
      for (i=1;i<=np;i++){
         for (rank=1;rank<=size;rank++){
            if (par.owned(rank,vs[i])){
               count++;
               vsys[rank].addPar(sys.getPar(i));
               owned[rank].push_back(B(true));
               // the index of array to which the value is set,
               // is the same for vsys[rank] and owned[rank] 
               // On initialization, each 'process gets just those
               // particles which it needs.
            }
         }
      }
      cpmassert(count==np,loc);
      for (rank=1;rank<=size;rank++){
         Z npr=vsys[rank].get_np();
         cpmmessage(mL2,"rank "&cpmwrite(rank)&
            " owns "&cpmwrite(npr)&" particles");
      }
      redistribute();
   }
   else{
      vsys=V<Sys>(1,sys);
      owned=Vio< Vio<B> >(1,Vio<B>(1,B(true)));
      vsys[1].enforceGeneralUpdate();
      gg=GroupGraph<Group>(1);
   }
   cpmmessage(mL,loc&" done");
}


CPM_TAL
void CPM_FCN::updateOwnership(void)
{

   Z mL=2;
   Word loc("Parallel::updateOwnership()");
   cpmmessage(mL,loc&" started");

   if (size==1){ // nothing to do
      cpmmessage(mL,loc&" trivially done");
      return; 
   }

   VV< V<Prc>  > fromTo(size, size, V<Prc>(0));
      // square matrix of arrays, initialized with the void array
   Z i,rank,rank2,npr;

   for (rank=1; rank<=size; rank++){
      npr=vsys[rank].get_np();
      Z npr2=owned[rank].dim();
      cpmassert(npr==npr2,loc);
// process rank gives away all particles which left X(rank)
// during the last time step and passes them over to the uniquely
// determined new owner.
      for (i=1;i<=npr;i++){
         if (!owned[rank][i]) continue;
         // ownly the particles owned at the beginning of the time 
         // step are under consideration
         Prc pi=vsys[rank].getPar(i);
         Spc pix=pi.get_x();
         X2<Z,Group> zg=par.owner(pix);
         Z i1=zg.c1();
         Group g1=zg.c2();
         if (i1!=rank){
            pi*=g1;
            // if g1 is unity, the operator *= is 
            // implemented as the trivial operation, so it is cheap
            fromTo[rank][i1].push_back(pi);
            owned[rank][i]=false;
         }
      }
   }
   for (rank=1;rank<=size;rank++){
      npr=vsys[rank].get_np();
      V<bool> owr(npr);
      for (i=1;i<=npr;i++) owr[i]=owned[rank][i]; 
         // unelegant type conversion B -> bool
      vsys[rank].selectParticles(owr);
   }
// now we have only particles which we own 
// Process rank gets new particles that were previously owned by 
// other processes
   for (rank=1;rank<=size;rank++){
      for (rank2=1;rank2<=size;rank2++){
         if (rank==rank2) continue;
         vsys[rank].addPar(fromTo[rank2][rank]);
      }
   }
// now each rank has exactly the particles which he owns
   for (rank=1;rank<=size;rank++){
      npr=vsys[rank].get_np();
      owned[rank]=V<B>(npr,B(true));
      cpmmessage(mL+1,"rank "
         &cpmwrite(rank)&" owns "
         &cpmwrite(npr)&" particles"
      );
   }
   cpmmessage(mL,loc&" done");
}

CPM_TAL
void CPM_FCN::addUsedParticles(void)
{   
   
   Z mL=2;
   Word loc("Parallel::addUsedParticles()");
   cpmmessage(mL,loc&" started");

   Z rank;

   if (size==1){ // not much to do 
      cpmmessage(mL,loc&" trivially done");
      for (rank=1;rank<=size;rank++){
         vsys[rank].enforceGeneralUpdate();
      }
      cpmmessage(mL,loc&" done for size=1");
      return;
   }
 
// Process rank has now to look out for the particles which he has to 
// use (without owning them)   

   Z rank2,i,j,npr;

   VV< V<Prc>  > fromTo(size, size, V<Prc>(0));

   typedef X2oio<Z1,Group> ZG;
      // Z,Group pairs

   for (rank=1; rank<=size; rank++){
      npr=vsys[rank].get_np();
      for (i=1;i<=npr;i++){
         Prc pi=vsys[rank].getPar(i);
         Spc pix=pi.get_x();
         V<ZG> vzg=par.toSend(rank,pix);
         // returns a list of processes which need to know about the
         // present particle. The process index comes together with the
         // transformation which transforms the present particle into the
         // version needed by the receiver.

         Z vzgd=vzg.dim();
            // make i known also to processes who need it
         for (j=1;j<=vzgd;j++){
            ZG zg=vzg[j];
            Z k=zg.c1();
            Group gk=zg.c2();
            Prc pk=pi;
            pk*=gk;
            fromTo[rank][k].push_back(pk);
         }
      } 
   } 

   for (rank=1;rank<=size;rank++){
      for (rank2=1;rank2<=size;rank2++){
         // adding to the particles owned by rank all those
         // owned by different ranks but being close enough that their
         // direct interaction with rank-particles is considered
         // esssential for modeling their motion properly.
         if (rank==rank2) continue;
         vsys[rank].addPar(fromTo[rank2][rank]);
         Z nAdd=fromTo[rank2][rank].dim();
         owned[rank]&=V<B>(nAdd,B(false));
      }
      vsys[rank].enforceGeneralUpdate();
   }
   cpmmessage(mL,loc&" done");
}

CPM_TAL
void CPM_FCN::redistribute()
{
   updateOwnership();
   addUsedParticles();
}

#else // many-processor version

CPM_TAL
class CPM_CN{
};

#endif

#undef CPM_TAL
#undef CPM_CN
#undef CPM_FCN

}

#endif // guard




      







}

#endif // guard




      






el.h, version 2004-11-15             



