//? cpmpartitionx.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*
   Description: See cpmpartition.h

   History: Started 2001-4-5 for defining partition of a
   the area between toning shell and photo conductor in a way that
   allows the particles to distribute over several processes
   in parallel computing.
   2001-7-11: Consistent treatment of overlaps achieved
   2001-11-12: Old cpmpartition.h moved from namespace
      CpmGeo to both CpmDim2 and CpmDim3 in analogy to all
      other higher geometric classes. See cpmpartitionx.h
      for details. Most advanced geometric class with a
      completely generic (with respect to dim 2 or 3)
      implementation. As a template defining file it has
      simple file structure (no cpp's)
   2001-11-30 class StationDat now in one file with Station
   2002-1-10: now own() finds always an owner
   2002-2-22: failure mode of proLis (for particles out of the
      radial range) eliminated
   2002-3-11: comments completed
   2002-5-1: eps eliminated, function owner added
   2002-7-18: function inX introduced
   2002-12-06: epLattice modified in a way that toner particles form
      a zick-zack line
   2003-01-30 epLattice modified in a way that it works for 1,2,3,...
      particles without failure. 
   2003-04-17 epLattice modified so that the particle number can 
      approximately be prescribed
   2003-06-04 toSend, toSendi modified so that also a single process
      is allowed to be toroidal in a non-trivial manner
   2003-06-07 The X(i) made available as instances of class SubsetOfSpace
      ( data element sub)
   2003-08-22 also X made available as instances of class SubsetOfSpace
   2003-09-02 ownerSimple(...): error situation that occurred for p==1 
      eliminated. 
   2004-06-24 code for  ownerSimple(...) simlified based on Iv::quant
         ownerSimpleSub(...) added
*/

   using namespace CpmSystem;
   using CpmRoot::C;
   using CpmRoot::R;

   using CpmRoot2::R1;
   using CpmRoot2::Z1;
   using CpmRoot2::B;
   using CpmRoot2::Record;
   using CpmRoot2::RecordHandler;

   using CpmLinAlg::Z2;
   using CpmLinAlg::Z3;

   using CpmArrays::V;
   using CpmArrays::Vo;
   using CpmArrays::X2;
   using CpmArrays::X3;
   using CpmArrays::X4;
   using CpmArrays::R_Vector;

   using CpmFunctions::R_Func;
   using CpmFunctions::R2_Func;
   using CpmFunctions::F;
   using CpmFunctions::F5;

   using CpmAlgorithms::CyclicAlarm;
   using CpmAlgorithms::Mini;
 
   using CpmGraphics::Frame;
   using CpmGraphics::Graph;

   using namespace CpmImaging;
   
   using CpmGeo::Angle;
   using CpmGeo::Iv;
   using CpmGeo::GroupGraph;

   using CPM_PHYSICS::ScalarField;
   using CPM_CAMERA::Camera;

//doc namespace CpmDimx{
// belongs to the namespace for which Spc,Vec,SubsetOfSpc are free
// for usage without qualification
/////////////////////// class ParOfSpc<> ///////////////////////////////
   template <class Dat>
   class ParOfSpc{ // Partition of space
      // A construct for dividing a region in space (2D or 3D
      // according to the dimension set in cpmdimdef.h) in
      // regions in such a manner that particles living in
      // the same region can efficiently be handled by a single
      // processor in a parallel computing application.
      // The underlying abstraction is the following:
      // Let 1,...n enumerate the processes and let X be
      // the part of space in which 'the whole system lives'.
      // Let X(1),...X(n) be a partition of X. (i.e. the
      // X(i) are mutually disjoint and their union is X).
      // For each X(i), we consider a larger region Y(i) such
      // that the boundary of Y(i) has positive distance from
      // the boundary of X(i). A particle p with position x \in X(i)
      // is said to be owned by process i and used by all processes j
      // for which x \in Y(j). ('\in' is a notation borrowed from LaTeX
      // and denotes the 'is element of ...' relation symbol of set 
      // theory). Or put differently: Process i owns all
      // particles with position \in X(i) and uses all particles with
      // position \in Y(i)\X(i). The idea behind this is: giving
      // a process for some small span of time control over all 
      // the particles he owns and all the particles he uses
      // will allow it to take into account all the interactions that
      // control the fate of the owned particles. After
      // the time span has elapsed, the particles have to become
      // redistributed over the processes since they might have crossed
      // boundaries of the X(i) and Y(i), and thus have to be associated
      // with new owners and users for the next autonomous time step.
      // To provide the information on which this redistribution can
      // be efficiently based is the main functionality of this class.
      //
      // For systems with long-ranging forces, like charged and
      // magnetized particles, it is not sufficient to take the particles
      // in Y(i) into account in addition to the 'native' ones of X(i).
      // Actually also the most distant particles must not be forgotten
      // since they may - in their totality - add significantly to,say,
      // the electric field in area X(i). The important point is,
      // however,
      // that these distant particles need not enter the model as
      // individual particles but as more economic integrated quantities
      // with drastically less degrees of freedom and a lower rate of
      // change so that they can be considered constant during the time
      // span of the previous consideration.
      // The data nS1, nS2, nS3 of the present class will be used to
      // implement this "far field description" when constructing
      // class Station. For this construction the present template class
      // will be used with template argument class StationDat
      // (see cpmstationdatx.h) which specializes X to the region between
      // a plane, a cylinder, and two planes containing the cylinder
      // axis.

   typedef X2<Z1,Group> ZG;

   Dat dat;
      // data that determine the part of space to be partitioned
   Z1 nL1;
      // number of angle subfields
   Z1 nL2;
      // number of radial subfields
   Z1 nL3;
      // number of subfields i z direction
   R1 overlap1;
      // fractional overlap in angular direction
   R1 overlap2;
      // fractional overlap in radial direction
   R1 overlap3;
      // fractional overlap in z direction

   R1 ov1;
      // overlap1/nL1, is absolute overlap between adjacent regions
      // Y(i) and X(++i) if we denote by ++i the index of a neighbor
      // process to process i. Here the overlap is given in the
      // normalized parameter space to which dat.map() maps.
   R1 ov2;
      // overlap2/nL2
   R1 ov3;
      // overlap3/nL3

   Z1 nS1;
      // a number of points needed for course graining the sources of the
      // 
      // far field
   Z1 nS2;
      // see nS1
   Z1 nS3;
      // see nS1

   B isTorus1;
      // True if toroidal topology in 1-direction should hold.
      // Toroidal topology is the present implementation of
      // 'periodic boundary conditions' as proposed by Eric Stelter.
   B isTorus2;
      // true if toroidal topology in 2-direction should hold
   B isTorus3;
      // true if toroidal topology in 3-direction should hold

   Group g1;
      // Transformation that change a particle state near one
      // boundary of the box into the state which influences
      // the particles near the opposit boundary of the box.
      // Schematics of the situation:
      //
      // (......Y1.......)  (...... Y3......)  areas Y of usage
      // ..Y3..)  (.......Y2.......)  (..Y1..     with toroidal topology
      //  .s[.p)..(..][..)..(..][..)..(.r].q   overlap and
      //                                          particles p,q,r,s
      //  ..[...X1...][...X2...][...X3...]..   areas X of ownership
      //
      // Since p belongs to X1 and lives near the boundary to
      // the toroidal neighbor region X3, thus in the wrapped-arround
      // piece of Y3, p has to be messaged to process 3 as a transformed
      // object q=p*g1 which lives in the non-wrapped-arround part of
      // Y3\X3, where it is close to the particles owned by process 3.

      // In an completely analog manner:
      // Since r belongs to X3 and lives near the boundary to
      // the toroidal neighbor region X1,
      // r has to be messaged to process 1 as a transformed object
      // s=r*!g1 which lives in Y1\X1, the area used but not owned by
      // process 1. Here, !g1 is the group inverse of g1. This implies
      // in the 'graphics' above that the vector from p to q equals the
      // vector from s to r. Notice that r and s both have distance 1
      // from their respective boundary, wheras p and q have distance of
      // 2 units. I'm surprised that the schematic representation above
      // looks clearer to me than my pencil drawings!

   Group g2;
      // only for uniformity. In the PaDyAn program it makes no sense
      // to consider the PC the shell toroidal neighbors
   Group g3;
      // toroidal geometry in 3-direction is even more natural than
      // in 1-direction since it reduces the surface effects from
      // missing neighbors (by providing neighbors)

   Z nTot;
      // total number of space areas that make up the partition
      // nL1*nL2 for 2D and nL1*nL2*nL3 for 3D

   V<Iv> xv1;
      // array of nL1 non-overlapping intervals for angle after map
   V<Iv> xv2;
      // array of nL2 non-overlapping intervals for radius after map
   V<Iv> xv3;
      // array of nL3 non-overlapping intervals for z after map

   V<SubsetOfSpace> sub;
      // The pieces X(i) of space for which the various processes are
      // responsible

   SubsetOfSpace subX;
      // The piece X of space under consideration (set union of the 
      // X(i), but for reasons of efficiency not implemented as such

   VV<Spc> points;
      // For each i \in {1,...nTot} , the point list points[i] will be
      // used for  course graining the sources of the mean field

   V<R> resolutions;

   V<R> coveringRadii;
      // r:=coveringRadii[i] is such that for each x \in X(i)
      // there is a p \in points[i] for which |x-p|<=r holds.
      // coveringRadii[i] and resolutions[i] may differ by
      // a factor of 2 but not much more.

   R smallLength_;
      // a small fraction of the smallest of the resolutions

   V<R> volumes;
      // volumes of the X(i) \subset X.

   Box box;
      // minimum enclosing box

   V<Spc> corners;
      // corners in space

   Spc center;

   void iniPoints();
   void iniResolutions();
   void iniVolumes();
   void iniSub();
   V<Spc> makePoints(Z)const;

   V<ZG> toSend1(Z i, R r)const;
      // as pointed out in the explanation to g1, the messaging
      // between toroidal neighbors involves transformations which
      // get consistently set by calling this function. Although
      // the Group element is the unit element in most cases and
      // then has no effect at all it is not advisable to make use
      // of this and to treat the transformations as an ad hoc or
      // exceptional addition to non-toroidal code. The present
      // implementation of Group and the pairing templates is
      // sufficiently efficient that the computational burden from
      // these additional data is far from being significant.
      // The function will be called for r=map(x)[1]
      // where x is the position of a particle. 
      // There are two possibilities: 
      // (i) the particle position x is in 1-direction (we are
      // considering function toSend1) 'interior' i.e. not near to a
      // boundary. In this we return V<ZG>(0). 
      // (ii) the particle position x is not in the interior
      // so that it has to made known to other parts of the system
      // (which in most cases will be handled by a different process).
      // This transition is described by res[1] which is an 
      // instance of X2oio(Z1,Group).
     
   V<ZG> toSend2(Z i, R r)const;
      // see toSend1
   V<ZG> toSend3(Z i, R r)const;
      // see toSend1
// private methods
   void adjust(R fac);
      // sets the values of the dependent quantities3
      // The overlap factors get set to values which are fac times the
      // ones minmally determined by nLi and nSi
public:
   typedef ParOfSpc Type;
   CPM_IO

   ParOfSpc(const Dat& dat, Z nL1, Z nL2, Z nL3=1,
      Z nS1=2, Z nS2=2, Z nS3=1, R fac=1.);
      // constructor which sets the overlap parameters from nS1,nS2,nS3.
      // The actual overlap is obtained from the minimal one by
      // multiplication with fac
   ParOfSpc(){}
      // trivial, everything zero

   const Dat& getDat()const{ return dat;}
   SubsetOfSpace getX()const{ return subX;}
   SubsetOfSpace getX(Z i)const{ return sub[i];}
   V<Spc> getPoints(Z i)const{ return points[i];}
      // a set of nS1*nS2(*nS3) internal points of X(i) useful
      // for approximating fields and field sources. The primary
      // definition of these points is by giving values to their
      // images under map, so that mapInv has to be used to get
      // these points in Spc.
 //  bool closeToFloor(Z rank, Z i)const;
      // returns true if points[rank][i] is a lowest point in 2-direction
 //  bool closeToCeil(Z rank, Z i)const;
       // returns true if points[rank][i] is a highest point in
       // 2-direction
   bool extremal(Z rank, Z i)const;
      // returns closeToFloor(rank, i)||closeToCeil(rank, i)
   R getRes(Z i)const{ return resolutions[i];}
   R getVol(Z i)const{ return volumes[i];}
      // returns the volume (area in 2D) of X(i) (in real space, not after
      // mapping
      // into the unit cube/square)
   R getR()const{ return dat.getR();}
      // radius of the shell
   R getCovRad(Z i)const{ return coveringRadii[i];}
   R getH()const{ return dat.getH();}
      // return nipHeigt
   R get_xL()const{ return dat.get_xL();}
   R get_xU()const{ return dat.get_xU();}
   R get_xLength()const{ return dat.get_xLength();}
   R get_xLength2()const{ return dat.get_xLength2();}
   R get_yLength()const{ return dat.get_yLength();}
   R get_zLength()const{ return dat.get_zLength();}
   Iv ivPC()const{ return dat.ivPC();}
   V<Spc> testPoints(Z n)const;
      // test points along the PC for testing the values of 
      // the electrical potential 

   bool torus1()const{ return isTorus1;}
   bool torus2()const{ return isTorus2;}
   bool torus3()const{ return isTorus3;}
   bool torus()const{ return torus1()||torus2()||torus3();}

   void setTorus(Z i, const Group& g);
      // In order not to overload the argument list of the constructor,
      // we set toroidal topology (if needed) by the present
      // function. It goes without saying that all dependent
      // quantities get adjusted accordingly.

   Z card()const{ return nTot;}
      // number of subsets

   R resolution(Z rank)const;
      // returns the smallest distance between the coarse graining
      // points in subset rank

   R smallLength()const{ return smallLength_;}

   Rdim map(const Spc& z)const;
      // For a point z in the region RegPar which is to be partitioned
      // the return value belongs to the square [0,1]x[0,1] or the cube
      // [0,1]x[0,1]x[0,1] of Rdim. So, a partition of the region can
      // be achieved by partitioning this square (or cube). This is 
      // done in function index().
      // All points outside RegPar get mapped to points outside
      // the unit square or unit cube.
      // map is a bijective mapping Spc(Rdim\X) --> Rdim\X, where
      // X:={ x \in Rdim | x[1]==0 and x[2]<=-r}

   Spc mapInv(const Rdim& z)const;
      // inverse of map: Rdim/X --> Spc(Rdim\X)

   F<Spc,Rdim> toMap()const;
      // map as autonomous function object. Notice the following:
      // Let par be any instance of the present class. Then
      // f=par.toMap is a function that describes the part of space
      // covered by par. If we change par by e.g. par*=g, (not
      // implemented yet but clear what this would do) then f would still
      // describe the unchanged position (which is perfectly OK!).
      // Function is used to initialize data 'sub'.

   F<Rdim,Spc> toMapInv()const;
      // analogous to the last function

   Z1 linearCoding(const Zdim& z)const;
      // linear index made out of a Z2-or Z3-index.
      
   Z1 linearCodingSub(const Zdim& z)const;
      // just as linearCoding but with nS1,nS2,nS3 playing the role of
      // nL1,nL2,nL3

   Zdim linearCodingInv(Z z)const;
      // inverse of function linearCoding()
      
   Zdim linearCodingSubInv(Z z)const;
      // inverse of function linearCodingSub()

   V<X2oio<Z1,Group> > toSend(Z rank, const Spc& pix)const;
      // This is the function providing the main functionality of the
      // class. It will only be called (only then the behavior is as
      // described) from process rank for values of pix for which
      // owned(rank,pix)==true. Let then res be the return value of
      // the function. Then rd:=res.dim() is the number of processes to
      // which process rank has to send a particle p with p.get_x()==pix.
      // ( since these processes have to use particle p, which is owned
      // by process rank). For i=1,...rd the receiving processes
      // have ranks res[i].c1() and need particle p transformed by
      // Group element res[i].c2(); see explantion to g1 for the meaning
      // these transformations.

   bool owned(Z rank, const Spc& pix)const;
      // returns true if pix belongs to X(rank) and false else

   R dist(Z i, const Spc& pix)const{ return sub[i](pix);}
      // the more negative this quantity the 'deeper inside X(i)'
      // ix point pix. For value 0, pix is on the boundary
      // and for positve values outside. Important tool for 
      // achieving 'unique ownership' of particles in parallel 
      // computing.

   X4<Z,Group,B,B> owner(const Spc& pix)const;
      // Returns res= (i,Group(),false,false) where i is such that pix \in
      // X(i).
      // If not all attributes isTorus... are false, there
      // is the possibilty that there is no i for which
      // pix \in X(i) is valid. Then we return (j,g,b,b4) where
      // pix*=g \in X(j) and g is of the form g1^e1*g1^e2*g3^e3
      // where ei=0 for !isTorusi and \in {1,-1} else.
      // b=false is normal, b=true can only happen in 3D and then
      // indicates that pix is off of any X(i) in lateral
      // direction. Knowing this allows ParDynSys::redistribute(..)
      // to reduce the velocity of the particle in order to 
      // damp lateral motion. 
      // b4 is false if g==unit element.
      // Whether this is the case needs to be known especially for
      // nTot==1, and isTorus1||isTorus2||isTorus3  since then res.c1()
      // is always 1 and gives no signal on whether a point is in X(1) or
      // in Y(1)\X(1). 
      // Evaluation of g==Group() is unnecessarily  complicated.
 
   Z ownerSimple(const Spc& pix)const;
   
   Z ownerSimpleSub(const Spc& pix, Z i)const;

   Z repNegDim()const;
      // dimension of the array to created by the next function

   V<Spc> repNeg(const Spc& x)const;
      // repetition neighbors. Assume that the total space X is only a
      // small part of the real system space (like the slab proposed
      // by Eric Stelter in the model of an electrophotographic 
      // toning station). Then one may want to approximate the huge
      // system by considering X in detail and assume that the system 
      // outside behaves just like the system in X, so that we can think
      // of X as surrounded by copies (repetitions) of X. These copies of
      // X are not assumed to be detailed copies but are only assumed 
      // to agree with X with respect to coarse grained descriptors
      // e.g. of electromagnetic fields. Let s be a field source 
      // located at x in X. Assuming the kind of periodic immersion
      // of X in the larger system, we have to assume that the same
      // source is located also in all points of a lattice generated
      // from x with translation vectors or - more generally - by 
      // transformations as our g1,g2,g3. A reasonable approximation,
      // which will bring about the essential influence from the
      // neigborhood will be to consider only the directly adjacent 
      // copies of X (presently we also take into account the diagonally
      // adjacent ones, which have only edges or points in common with
      // X).
      // The result of shifting the source can be immitated by 
      // considering only one source, evaluating the field not simply in
      // a point y, but in all points which are components of repNeg(y)
      // and add the results. 
      // This involves some approximation in systems
      // like the toning nip when we consider X over which the nip 
      // height cannot be considered constant. It is, however, the
      // intention of this periodic immersion to avoid the need for
      // considering such large pieces of the system.
      // Short explanation: repNeg(x) is the list of points which are 
      // equivalent to x with respect to the kind of periodicity
      // under consideration. See function
      // CPM_TAL
      // void CPM_FCN::prepareStep(...) in cpmpardynsys.h for the only
      // utilitisation of the concept made so far.
      // Since meanwhile I came to distrust this logic (erroneously
      // as it appears to me now) lets make it crystal clear. Formulas
      // have to be understood according to LATEX formatting:
      // Let V be the superposition of shifted terms
      // V(x):=\sum_{i=-n}^{n} V_{i}(x)
      // V_{i}(x):=V_{0}(x-i*s) 
      // Thus
      // V(x)=\sum_{i=-n}^{n} V_{0}(x-i*s)
      // Thus we have in fact to sum up the values that a single
      // function V_{0} gives for a chain of points
      // \{ p=x-i*s | i=-n,...n \}.

   C toColor(const V<Z>& vz)const;
      // a peliminary color coding complex number intended to let the
      // mapping C-->C, z|--> part.toColor(neighbors(z)) be a function
      // which visualizes the Partition part as a pattern in space C.
      // when inserted as f in function
      // void CpmGraphics::Graph::show(const F<C,C>& f, R gamma=0.01);

   void mark(const Camera& gr)const;

   void show(const Frame& fr, bool undistorded=true)const;
      // displays the geometry of the partition on the graphics
      // window fr. If the parameter undistorted is set to 'true'
      // the width of the displayed rectangle is set such that
      // angles are represented correctly (no distortion),
      // otherwise the width of the partition as determined by
      // r, phiL, and phiU just fits the width of the graphics
      // window. The height of the displayed rectangle is
      // determined in relation to h by the parameter
      // const R hightExcess in the implementation of the function.

   void showScalarField(const Camera& ca, const ScalarField& f, 
      Z nx, Z ny, const Response& rp)const;
      // not yet in use

   Record toRecord()const;
      // Record is a data type which allows to hold data of types
      // bool,Z,R, and Word in a convenient manner (if one has to add
      // a new data element of any of those types one does not have
      // to reorganize the yet existing data in any way. So putting
      // into a record is a RAM-based analogy to writing to external
      // storage via printOn(). This will help ma lot in defining the
      // constructor of class Station more densly than before.
      // A present drawback compared to stream writing is that vectors
      // (arrays) can't (not yet) be transfered by a single statement.

//   GroupGraph<Group> groupgraph(const V<Spc>& vs )const;
      // see my experimental file cpmparallel.h for usage
      // and the simple implementation code as explanation.
      // Not yet proven by experience

   V< X2<Spc,Z> > epLatticePrel(const V<X2<R,bool> >& vx, 
      Rdim xL, Rdim xU, R& facCell)const; 
      // ep stands for electro photography
      // vx[i].c1() is a radius of a bounding sphere of a particle,  
      // and vx[i].c2() is true if this particle is a carrier particle.
      // Let res be the return value of the function. For each 
      // j<=res.dim(), res[j].c1() is a point in space, where one has
      // to place a particle with radius and particle type given by
      // vx[res[j].c2()]. The 'particle'-ensemble defined this way 
      // fills, after having transformed positions via map into 
      // the unit cube (square for dimVec=2), the subcube
      // xL.x1<=x[1]<=xU.x1, xL.x2<=x[2]<=xU.x2, ( xL.x3<=x[3]<=xU.x3 ).
      // The last factor controlls the density of the lattice points
      // so that for fast testing also 'dillute' systems can be created
      // (facCell >> 1 very dillute system). In the present version
      // res.dim is not an integer multiple of vx and thus it is not
      // guaranteed that each input 'particle' is used a fixed number of
      // times. As a result it is not guaranteed that the resulting 
      // particles have net charge zero even when vx came from particles
      // of net charge zero. I observed in most cases that |net charge|
      // was arround 10% of the total positive charge. 
      // This is more than I expected. Since it is difficult to build
      // charge neutrality reliably into the present code (without
      // leaving
      // loosely bound particles at the end (which would probably move 
      // away) I will use an explicit neutralization procedure
      // afterwards.
      // In the present context, where the particles are present only
      // in a very abstracted form, this can't be done. 
      // For vx.dim()=1 we define res=V< X2<Spc,Z> >(1) and
      // res[1]=X2<Spc,Z>(xc,1), where xc is the mid-point of the cell
      // defined by xL and xU. This is helpful since it allows to 
      // have a single, well defined particle as a special case of
      // a developer lump.  
      // For vx.dim()=2 the lattice points are just those corresponding
      // to xL and xU.  
      // For dimVec==3 and narrow space in 3-direction 
      // (xU[3]-xL[3])<<(xU[1]-xL[1]) the relative arrangement of
      // toner and carrier is not yet ideal. For densely packed
      // systems as those we are interested in for PaDyAn, this is no
      // concern, however. Severe malfunction now results in
      // program stop, less severe ones create a warning.

V< X2<Spc,Z> > epLattice(const V<X2<R,bool> >& vx, 
      Rdim xL, Rdim xU, R facCel, Z nPar=0)const; 
      // Based on previous function. For nPar<=0 it is just the previous
      // function. For nPar>=1 the value of facCel is taken as initial
      // value for a simple and efficient strategy to let res.dim()
      // approximate nPar
      // by changing facCel from trial to trial by a fixed factor or
      // divisor of 1.1. Even for 10000 particles and very bad initial
      // value for facCell, the execution time is no concern. 

   void autoFrame1(Camera& ca, R autoFrameFactor=1.1);
      // sets the frame of cam such that *this fills the frame
      // with a 'security' determined by the second argument
      
};

namespace{

template <class Dat>
Rdim fmap(const Spc& x, const ParOfSpc<Dat>& ps){ return ps.map(x);}

template <class Dat>
Spc fmapInv(const Rdim& x, const ParOfSpc<Dat>& ps)
{ return ps.mapInv(x);}

}

template <class Dat>
F<Spc,Rdim> ParOfSpc<Dat>::toMap(void)const
{
   return F1<Spc,ParOfSpc<Dat>,Rdim>(*this)(fmap);
}

template <class Dat>
F<Rdim,Spc> ParOfSpc<Dat>::toMapInv(void)const
{
   return F1<Rdim,ParOfSpc<Dat>,Spc>(*this)(fmapInv);
}

template <class Dat>
void ParOfSpc<Dat>::setTorus(Z i, const Group& g)
{
   if (i==1){
      isTorus1=true;
      g1=g;
   }
   else if (i==2){
      isTorus2=true;
      g2=g;
   }
   else if (i==3){
      isTorus3=true;
      g3=g;
   }
   else{
      cpmerror("ParOfSpc<Dat>::setTorus(Z i,...) i not in {1,2,3}");
   }
//   R fac=ov1*nL1*nS1;
//   adjust(fac);
}

template <class Dat>
void ParOfSpc<Dat>::adjust(R fac)
{
   const Z mL=3;
   cpmmessage(mL,"void ParOfSpc<Dat>::adjust() started");
   if (dimVec==2){
      nL3=1;
      nS3=1;
   }
   nTot=nL1*nL2*nL3;
   ov1=fac/(nL1*nS1);
   ov2=fac/(nL2*nS2);
   ov3=fac/(nL3*nS3);
   overlap1=ov1*nL1;
   overlap2=ov2*nL2;
   overlap3=ov3*nL3;
   xv1=Iv(0,1).divide(nL1);
   xv2=Iv(0,1).divide(nL2);
   xv3=Iv(0,1).divide(nL3);
   iniPoints();
   iniVolumes();
   iniResolutions(); // needs volumes
   iniSub(); // needs resolutions
   cpmmessage(mL,
      "void ParOfSpc<Dat>::adjust() done, nTot="&cpmwrite(nTot));
}

template <class Dat>
ParOfSpc<Dat>::ParOfSpc(const Dat& dat_, Z n1, Z n2, Z n3,
   Z nS1_, Z nS2_, Z nS3_, R fac_): dat(dat_),
   nL1(n1),nL2(n2),nL3(n3), nS1(nS1_), nS2(nS2_), nS3(nS3_)
{
   Z mL=2;
   Z mL2=10;
   // 10 is normal value
   Word loc("ParOfSpc<Dat>::ParOfSpc(...)");
   CPM_MA
   adjust(fac_); // main work
   if (cpmverbose>=mL2){ // a visual test structure 
      // to see whether the sub[i] where created
      // correctly
      Word inifile("station.ini");
      RecordHandler rc(inifile);
      Frame fr;
      Word sec="cameracontrol";
      Camera ca(rc,sec,fr);
      for (Z i=1;i<=nTot;i++){
         SubsetOfSpace sos=sub[i];
         if (cpmverbose>mL2){
           // sos.set_dr(5e-6);
            sos.set_r(2000e-6);
          //  sos.set_ds(10e-6); no longer needed
            sos.set_x(Rdim());
            ca.show(sos,Texture(40e-6));
         }
         else{
            ScalarField sf(sos);
            sf.mark(ca);
         }
         cpmwait(10);
      }
   }
   CPM_MZ
}

template <class Dat>
Rdim ParOfSpc<Dat>::map(const Spc& z)const
{
   return dat.f(z.toRdim());
}


template <class Dat>
Spc ParOfSpc<Dat>::mapInv(const Rdim& z)const
{
   return Spc(dat.fInv(z));
}

template <class Dat>
bool ParOfSpc<Dat>::owned(Z rank, const Spc& pix)const
{
   X3<Z,Group,B> zgb=owner(pix);
   return zgb.c1()==rank;
}

template <class Dat>
Z ParOfSpc<Dat>::ownerSimple(const Spc& pix)const
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::ownerSimple(...)");
   CPM_MA
   Iv iv1(0,1),iv2(0,1),iv3(0,1);
   Zdim res0;
   Rdim mz=map(pix);
   bool restrict=true;
   res0[1]=iv1.quant(mz[1],nL1,restrict);
   res0[2]=iv2.quant(mz[2],nL2,restrict);
   if (dimVec==3) res0[3]=iv3.quant(mz[3],nL3,restrict);
   Z res=linearCoding(res0);
   CPM_MZ
   return res;
}

template <class Dat>
Z ParOfSpc<Dat>::ownerSimpleSub(const Spc& pix, Z i)const
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::ownerSimpleSub(...)");
   CPM_MA
   Zdim range=linearCodingInv(i);
   Iv iv1,iv2,iv3;
   iv1=xv1[range[1]];
   iv2=xv1[range[2]];
   if (dimVec==3) iv3=xv1[range[3]];
   Zdim res0;
   Rdim mz=map(pix);
   bool restrict=true;
   res0[1]=iv1.quant(mz[1],nS1,restrict);
   res0[2]=iv2.quant(mz[2],nS2,restrict);
   if (dimVec==3) res0[3]=iv3.quant(mz[3],nS3,restrict);
   Z res=linearCodingSub(res0);
   CPM_MZ
   return res;
}

template <class Dat>
X4<Z,Group,B,B> ParOfSpc<Dat>::owner(const Spc& pix)const
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::owner(...)");
   CPM_MA
   Rdim mz=map(pix);
   Z i1=1,i2=1,i3=1;
   R p;
   Group g;
   B slowDown, res4; // false by default
   Iv iv01(0.,1.);
   p=mz[1];
   
   if (p<0){
      if (isTorus1){
         i1=nL1;
         g*=g1;
         res4=true;
      }
      else{
         i1=1;
      }
   }
   else if (p>1){
      if (isTorus1){
         i1=1;
         g*=!g1;
         res4=true;
      }
      else{
         i1=nL1;
      }
   }
   else{
    //  while (!xv1[i1].contains(p,false,true))
     // { i1++; if (i1>nL1) cpmerror("pBad1="&cpmwrite(p));}
      i1=iv01.quant(p,nL1);
   }

   p=mz[2];
   if (p<0){
      if (isTorus2){
         i2=nL2;
         g*=g2;
         res4=true;
      }
      else{
         i2=1;
      }
   }
   else if (p>1){
      if (isTorus2){
         i2=1;
         g*=!g2;
         res4=true;
      }
      else{
         i2=nL2;
      }
   }
   else{
   //   while (!xv2[i2].contains(p,false,true))
   //   { i2++; if (i2>nL2) cpmerror("pBad2="&cpmwrite(p));}
      i2=iv01.quant(p,nL2);
   }
 
   Zdim res0;
   res0[1]=i1;
   res0[2]=i2;
   if (dimVec==3){
      p=mz[3];
      if (p<0){
         if (isTorus3){
            i3=nL3;
            g*=g3;
            slowDown=true;
            res4=true;
         }
         else{
            i3=1;
         }
      }
      else if (p>1){ 
         if (isTorus3){
            i3=1;
            g*=!g3;
            slowDown=true;
            res4=true;
         }
         else{
            i3=nL3;
         }
      }
      else{
       //  while (!xv3[i3].contains(p,false,true))
       //  { i3++; if (i3>nL3) cpmerror("pBad3="&cpmwrite(p));}
         i3=iv01.quant(p,nL3);
      }
      res0[3]=i3;
   }
   Z res= (nTot==1 && res4==true) ? 0 : linearCoding(res0);
   CPM_MZ
   return X4<Z,Group,B,B>(res,g,slowDown,res4);
}

#define zg(z,g) ZG(Z1(z),g)
#define vzg(z,g) V<ZG>(1,ZG(Z1(z),g))

namespace{

// method to avoid code replication in implementing functions 
// toSend1, toSend2, toSend3

V<X2oio<Z1,Group> > toSendx(Z mL, Word loc, Z i, R r, 
   Z nLx, Iv xvxi, Group gx, R ovx, bool isTorusx)
{
   typedef X2oio<Z1,Group> ZG;
      // would not be needed if we would give the present function
      // the status of a private static member function
   Group u; // unity
   if (nLx==1){
      cpmassert(i==1,loc);
      if (!isTorusx) return V<ZG>(0); 
      else{ // added 03-06-04, also a single interval
            // should be allowed to act as a torus
         if (r<xvxi.inf()+ovx) return vzg(1,gx);
         if (r>xvxi.sup()-ovx) return vzg(1,!gx);
            // no neighborhood relation
      }
   }
   else if (i==1){ // notice nLx>=2
      if (r>xvxi.sup()-ovx) return vzg(2,u);

      if (isTorusx)
         if (r<xvxi.inf()+ovx) return vzg(nLx,gx);
   }
   else if (i==nLx){
      if (isTorusx)
         if (r>xvxi.sup()-ovx) return vzg(1,!gx);

      if (r<xvxi.inf()+ovx) return vzg(nLx-1,u);
   }
   else{
      if (r>xvxi.sup()-ovx) return vzg(i+1,u);
      if (r<xvxi.inf()+ovx) return vzg(i-1,u);
   }
   cpmmessage(mL,loc&" done, V<ZG>(0) returned");
   return V<ZG>(0);
}

} // anon namespace

template <class Dat>
V< X2oio<Z1,Group> > ParOfSpc<Dat>::toSend1(Z i, R r)const
{
   Z mL=3;
   Word loc("ParOfSpc<>::toSend1(...)");
   CPM_MA
   return toSendx(mL,loc,i,r,nL1,xv1[i],g1,ov1,isTorus1);
}

template <class Dat>
V< X2oio<Z1,Group> > ParOfSpc<Dat>::toSend2(Z i, R r)const
{
   Z mL=3;
   Word loc("ParOfSpc<>::toSend2(...)");
   CPM_MA
   return toSendx(mL,loc,i,r,nL2,xv2[i],g2,ov2,isTorus2);
}

template <class Dat>
V< X2oio<Z1,Group> > ParOfSpc<Dat>::toSend3(Z i, R r)const
{
   Z mL=3;
   Word loc("ParOfSpc<>::toSend3(...)");
   CPM_MA
   return toSendx(mL,loc,i,r,nL3,xv3[i],g3,ov3,isTorus3);
}

#undef vzg

template <class Dat>
V<X2oio<Z1,Group> > ParOfSpc<Dat>::toSend(Z rank, const Spc& pix)const
{
   Z mL=3;
   Word loc("ParOfSpc<>::sendTo(...)");
   CPM_MA
   Group u; // unity
   Rdim mpx=map(pix);
   Zdim rk=linearCodingInv(rank);
   R mpx1=mpx.x1;
   R mpx2=mpx.x2;
   V<ZG> zg1=toSend1(rk[1],mpx1);
   V<ZG> zg2=toSend2(rk[2],mpx2);
   V<ZG> zg3(0);
   if (dimVec==3){
      R mpx3=mpx[3];
      zg3=toSend3(rk[3],mpx3);
   }
   
   Z d1=zg1.dim()+1;
   Z d2=zg2.dim()+1;
   Z d3=zg3.dim()+1;

   Z d=d1*d2*d3;
   
   if (d==1){
      cpmmessage(mL,loc&" trivially done, nothing to send");
      return V<ZG>(0);
   }   
   
   V<ZG> s1(d1),s2(d2),s3(d3);

   s1[1]=zg(rk[1],u);
   s2[1]=zg(rk[2],u);
   s3[1]=zg(rk[3],u);
  
   if (d1>1) s1[2]=zg1[1];
   if (d2>1) s2[2]=zg2[1];
   if (d3>1) s3[2]=zg3[1];

   V<ZG> res(d);
   Zdim val;
   Z i,j,k,l=0;
   Group gi,gj,gk,gl;
   // till 2003-06-16 a wrong logic in forming gl created
   // some neighbors at wrong (too distant) places, so the dynamics of
   // some particles may have been spoiled seriously
   if (dimVec==3){
       for (i=1;i<=d1;i++){
         val[1]=s1[i].c1();
         gi=s1[i].c2();
         for (j=1;j<=d2;j++){
            val[2]=s2[j].c1();
            gj=s2[j].c2();
            for (k=1;k<=d3;k++){
               val[3]=s3[k].c1();
               gk=s3[k].c2();
               gl=gi*gj*gk;
               Z il=linearCoding(val);
               if (gl!=u || il!=rank){
                  l++;
                  res[l]=zg(il,gl);
               }
            }
         }
      }
   }
   else{
      for (i=1;i<=d1;i++){
         val[1]=s1[i].c1();
         gi=s1[i].c2();
         for (j=1;j<=d2;j++){
            val[2]=s2[j].c1();
            gj=s2[j].c2();
            Z il=linearCoding(val);
            gl=gi*gj;
            if (gl!=u || il!=rank){
               l++;
               res[l]=zg(il,gl);
            }
         }
      }
   }
   CPM_MZ
   return res.resize(l);
}
#undef zg

template <class Dat>
Z1 ParOfSpc<Dat>::linearCoding(const Zdim& z)const
// i=nL1, j=nL2, k=nL3 should give nL1*nL2*nL3
// i=1, j=1,k=1 should give 1
{
   Z i=z[1]-1;
   Z j=z[2]-1;
   Z res=nL1*j+i;
   if (dimVec==3){
      Z k=z[3]-1;
      res+=nL2*nL1*k;
   }
   Z1 result(res+1);
   if (result>nTot)
      cpmerror("ParOfSpc<Dat>::linearCoding(...): result>nTot");
   if (result<1)
      cpmerror("ParOfSpc<Dat>::linearCoding(...): result<1");
   return result;
}

template <class Dat>
Z1 ParOfSpc<Dat>::linearCodingSub(const Zdim& z)const
// i=nS1, j=nS2, k=nS3 should give nS1*nS2*nS3
// i=1, j=1,k=1 should give 1
{
   Z i=z[1]-1;
   Z j=z[2]-1;
   Z res=nS1*j+i;
   if (dimVec==3){
      Z k=z[3]-1;
      res+=nS2*nS1*k;
   }
   Z1 result(res+1);
   return result;
}

template <class Dat>
Zdim ParOfSpc<Dat>::linearCodingInv(Z z)const
{
   z--;  // now z starts with 0
   Zdim res;
   if (dimVec==2){
      Z i1=z/nL1;
      z-=nL1*i1;
      res[1]=z+1;
      res[2]=i1+1;
   }
   else{ // thus dimVec==3
      Z n1=nL1;
      Z n2=n1*nL2;
      Z i2=z/n2;
      z-=i2*n2;
      Z i1=z/n1;
      z-=i1*n1;
      res[1]=z+1;
      res[2]=i1+1;
      res[3]=i2+1;
   }
   return res;
}

template <class Dat>
Zdim ParOfSpc<Dat>::linearCodingSubInv(Z z)const
{
   z--;  // now z starts with 0
   Zdim res;
   if (dimVec==2){
      Z i1=z/nS1;
      z-=nS1*i1;
      res[1]=z+1;
      res[2]=i1+1;
   }
   else{ // thus dimVec==3
      Z n1=nS1;
      Z n2=n1*nS2;
      Z i2=z/n2;
      z-=i2*n2;
      Z i1=z/n1;
      z-=i1*n1;
      res[1]=z+1;
      res[2]=i1+1;
      res[3]=i2+1;
   }
   return res;
}

template <class Dat>
bool ParOfSpc<Dat>::printOn(ostream& str)const
{
   CpmRoot::writeTitle("ParOfSpc<>",str);
   cpmp(dat);
   cpmp(nL1);
   cpmp(nL2);
   cpmp(nL3);
   cpmp(overlap1);
   cpmp(overlap2);
   cpmp(overlap3);
   cpmp(nS1);
   cpmp(nS2);
   cpmp(nS3);
   cpmp(isTorus1);
   cpmp(isTorus2);
   cpmp(isTorus3);
   cpmp(g1);
   cpmp(g2);
   cpmp(g3);
   return true;
}

template <class Dat>
bool ParOfSpc<Dat>::scanFrom(istream& str)
{
   cpms(dat);
   cpms(nL1);
   cpms(nL2);
   cpms(nL3);
   cpms(overlap1);
   cpms(overlap2);
   cpms(overlap3);
   cpms(nS1);
   cpms(nS2);
   cpms(nS3);
   cpms(isTorus1);
   cpms(isTorus2);
   cpms(isTorus3);
   cpms(g1);
   cpms(g2);
   cpms(g3);
   R fac=ov1*nL1*nS1;
      // it would be natural to make fac a member and to eliminate the
      // members overlap1, overlap2, overlap3. In order not to break the 
      // movie file format, I'll don't do this
   adjust(fac);
   return true;
}

template <class Dat>
C ParOfSpc<Dat>::toColor(const V<Z>& vz)const
{
   Z i,n=vz.dim();
   R r1=1;
   for (i=1;i<=n;i++) r1+=vz[i]*pow(0.13,i);
   R r2=1.14;
   for (i=1;i<=n;i++) r2+=vz[i]*vz[i]*pow(0.57,i);
   R x1=r1*1000;
   R x2=r2*1370;
   R s1=50*(sin(x1)+1);
   s1=5*(s1-floor(s1)-0.5);
   R s2=50*(sin(x2)+1);
   s2=5*(s2-floor(s2)-0.5);
   return C(s1,s2);
}

namespace{ // tool for ParOfSpc<Dat>::show(...)
   template <class Dat>
   C fMap(const C& z, const ParOfSpc<Dat>& p)
   {
      return z;//p.toColor(p.proLis(z));
   }
}

template <class Dat>
void ParOfSpc<Dat>::mark(const Camera& ca)const
{
   Rdim x1,x2;
   x1[2]=0;
   x2[2]=1;
   if (dimVec==3){
      x1[3]=0.5;
      x2[3]=0.5;
   }
   Z i;
   Spc p1,p2;
   R xi;
   Color c(WHITE);
   for (i=1;i<=nL1;i++){
      xi=xv1[i].inf();
      x1[1]=xi;
      x2[1]=xi;
      p1=mapInv(x1);
      p2=mapInv(x2);
      ca.mark(p1,p2,c);
   }
   xi=xv1[nL1].sup();
   x1[1]=xi;
   x2[1]=xi;
   p1=mapInv(x1);
   p2=mapInv(x2);
   ca.mark(p1,p2,c);
}

template <class Dat>
void ParOfSpc<Dat>::show(const Frame& fr, bool undistorded)const
{
   const R hightExcess=5;
   const R gamma=0.75;

   using CpmFunctions::F1;
   using CpmGraphics::Graph;
   R h=dat.getH();
   R r=dat.getR();

   R yL=-hightExcess*h;
   R yU=(1+hightExcess)*h;
   Iv ivy(yL,yU);
   Iv ivx;
   if (1/*undistorded*/){
      R hight=ivy.abs();
      R halfWidth=0.5*hight*fr.getAspRat();
      ivx=Iv(-halfWidth,halfWidth);
   }
   else{
     // ivx=Iv(phiL*r,phiU*r);
   }
   Graph gr(fr);
//   gr.setCol(); // setting colors, not adding them
   gr.setXY(ivx,ivy);
   gr.show(F1<C,ParOfSpc,C>(*this)(fMap<Dat>),gamma);
   gr.vis();
}

template <class Dat>
void ParOfSpc<Dat>::showScalarField(const Camera& ca, 
   const ScalarField& f, Z nx, Z ny, const Response& rsp )const
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::showScalarField(..)");
   cpmmessage(mL,loc&" started");
   Graph gr=ca.getGraph();
   Iv ivx=gr.getIvX();
   Iv ivy=gr.getIvY();
   R unitPlus=1.25;

   Rdim x00(0,0);
   Rdim x01(0,unitPlus);
   Rdim x10(1,0);
   Rdim x11(1,unitPlus);
   Spc s00=mapInv(x00);
   Spc s01=mapInv(x01);
   Spc s10=mapInv(x10);
   Spc s11=mapInv(x11);
   R2 i00=ca.captureMono(s00);
   R2 i01=ca.captureMono(s01);
   R2 i10=ca.captureMono(s10);
   R2 i11=ca.captureMono(s11);

   R xL=cpmsup<R>(ivx.inf(),i00[1],i01[1]); 
   R xU=cpminf<R>(ivx.sup(),i10[1],i11[1]);
   R yL=cpmsup<R>(ivy.inf(),i00[2],i10[2]); 
   R yU=cpminf<R>(ivy.sup(),i01[2],i11[2]); 
   cpmassert(xL<=xU,loc);
   cpmassert(yL<=yU,loc);

   Iv iv_x(xL,xU);
   Iv iv_y(yL,yU);
   V<R> vx=iv_x.centers(nx);
   V<R> vy=iv_y.centers(ny);
   VV<R> val(ny,nx); // indexed like a image matrix
   Z i,j;
   R2 xij;
   for (i=1;i<=nx;i++){
      xij[1]=vx[i];
      for (j=1;j<=ny;j++){
         xij[2]=vy[j];
         Spc sij=ca.inverse(xij,1.);
         val[ny-j+1][i]=f(sij);
      }
   }
   R2_Func r0(val,1,iv_x,iv_y);
   gr.show(r0,rsp,iv_x,iv_y);
   cpmmessage(mL,loc&" done");
}

/*
template <class Dat>
V<Spc> ParOfSpc<Dat>::makePoints(Z ip)const
{
   Zdim zd=linearCodingInv(ip);
   Iv iv1,iv2,iv3;
   iv1=xv1[zd[1]];
   iv2=xv2[zd[2]];
   if (dimVec==3) iv3=xv3[zd[3]];
   V<R> vr1,vr2,vr3;
   vr1=iv1.centers(nS1);
   vr2=iv2.centers(nS2);
   if (dimVec==3) vr3=iv3.centers(nS3);
   Z nRes=nS1*nS2;
   if (dimVec==3) nRes*=nS3;
   V<Spc> res(nRes);
   Z l=0,i,j,k;
   Rdim val;
   if (dimVec==3){
      for (i=1;i<=nS1;i++){
         val[1]=vr1[i];
         for (j=1;j<=nS2;j++){
            val[2]=vr2[j];
            for (k=1;k<=nS3;k++){
               val[3]=vr3[k];
               res[++l]=mapInv(val);
            }
         }
      }
   }
   else{
      for (i=1;i<=nS1;i++){
         val[1]=vr1[i];
         for (j=1;j<=nS2;j++){
            val[2]=vr2[j];
            res[++l]=mapInv(val);
         }
      }
   }
   return res;
}
*/

template <class Dat>
V<Spc> ParOfSpc<Dat>::makePoints(Z ip)const
{
   Zdim zd=linearCodingInv(ip);
   Iv iv1,iv2,iv3;
   iv1=xv1[zd[1]];
   iv2=xv2[zd[2]];
   if (dimVec==3) iv3=xv3[zd[3]];
   V<R> vr1,vr2,vr3;
   vr1=iv1.centers(nS1);
   vr2=iv2.centers(nS2);
   if (dimVec==3) vr3=iv3.centers(nS3);
   Z nRes=nS1*nS2;
   if (dimVec==3) nRes*=nS3;
   V<Spc> res(nRes);
   for (Z i=1;i<=nRes;++i){
      Zdim zd=linearCodingSubInv(i);
      R r1,r2,r3;
      r1=vr1[zd[1]];
      r2=vr2[zd[2]];
      if (dimVec==3) r3=vr3[zd[3]];
      Rdim rd(r1,r2,r3);
      res[i]=mapInv(rd);
   }
   return res;
}

template <class Dat>
bool ParOfSpc<Dat>::extremal(Z rank, Z i)const
{
   rank;
   Zdim zi=linearCodingSubInv(i);
   return (zi[2]==1 || zi[2]==nS2);
}

template <class Dat>
void ParOfSpc<Dat>::iniPoints()
{
   Z i,j,k,l;
   points=VV<Spc>(nTot,0);
   for (l=1;l<=nTot;l++) points[l]=makePoints(l);

   coveringRadii=V<R>(nTot);
   for (l=1;l<=nTot;l++){
      V<Spc> p=points[l];
      Z pd=p.dim();
      Vo<R> mins(pd);
      for (i=1;i<=pd;++i){
         Vo<R> d(pd-1);
         k=0;
         for (j=1;j<=pd;++j){
            if (j==i) continue;
            k++;
            d[k]=(p[i]-p[j]).abs();
         }
         mins[i]=d.inf(); 
      }
      coveringRadii[l]=mins.sup();
   }
}

template <class Dat>
void ParOfSpc<Dat>::iniResolutions()
{
   const R facSmall=1e-3;
   static Word loc("ParOfSpc<Dat>::iniResolutions()");
   resolutions=V<R>(nTot);
   if (dimVec==2) nS3=1;
   Z zCells=nS1*nS2*nS3;
   cpmassert(zCells>0,loc);
   R zCellsInv=1./zCells;
   for (Z i=1;i<=nTot;i++){
      resolutions[i]=Spc::radFrmVol(volumes[i]*zCellsInv);
   }
   smallLength_=R_Vector(resolutions).inf()*facSmall;
}

template <class Dat>
void ParOfSpc<Dat>::iniVolumes()
{
   static Word loc("ParOfSpc<Dat>::iniVolumes()");
   volumes=V<R>(nTot);
   for (Z i=1;i<=nTot;i++){
      Zdim zd=linearCodingInv(i);
      Iv iv1=xv1[zd[1]];
      Iv iv2=xv2[zd[2]];
      if (dimVec==2){

         Rdim z11;
         z11[1]=iv1[1];
         z11[2]=iv2[1];
         Spc x11=mapInv(z11);

         Rdim z12;
         z12[1]=iv1[1];
         z12[2]=iv2[2];
         Spc x12=mapInv(z12);

         Rdim z21;
         z21[1]=iv1[2];
         z21[2]=iv2[1];
         Spc x21=mapInv(z21);

         Vec a=x12-x11;
         Vec b=x21-x11;
         volumes[i]=a.cross(b).abs();
      }
      else{
         Iv iv3=xv3[zd[3]];

         Rdim z111;
         z111[1]=iv1[1];
         z111[2]=iv2[1];
         z111[3]=iv3[1];
         Spc x111=mapInv(z111);

         Rdim z121;
         z121[1]=iv1[1];
         z121[2]=iv2[2];
         z121[3]=iv3[1];
         Spc x121=mapInv(z121);

         Rdim z112;
         z112[1]=iv1[1];
         z112[2]=iv2[1];
         z112[3]=iv3[2];
         Spc x112=mapInv(z112);

         Rdim z211;
         z211[1]=iv1[2];
         z211[2]=iv2[1];
         z211[3]=iv3[1];
         Spc x211=mapInv(z211);

         Vec a=x112-x111;
         Vec b=x121-x111;
         Vec c=x211-x111;
         volumes[i]=cpmabs(a.spat(b,c));
      }
   }
}

template <class Dat>
void ParOfSpc<Dat>::iniSub()
{
   Z mL=3;
   static Word loc("ParOfSpc<Dat>::iniSub()");
   CPM_MA
   DisPar dp(smallLength_,2*smallLength_);
      // 'discretization data'
   cpmassert(smallLength_>0,loc);
   Rdim x0;
   Rdim s0;
   x0[1]=0.5;
   x0[2]=0.5;
   x0[3]=0.5;
   s0[1]=0.5;
   s0[2]=0.5;
   s0[3]=0.5;
   subX=SubsetOfSpace::boxoid(x0,s0,toMap());
   subX.set_dp(dp);
   sub=V<SubsetOfSpace>(nTot);
   Z i;
   for (i=1;i<=nTot;i++){
      Zdim zd=linearCodingInv(i);
      Iv iv1=xv1[zd[1]];
      Iv iv2=xv2[zd[2]];
      Iv iv3;
      if (dimVec==3){
         iv3=xv3[zd[3]];
      }
      x0[1]=iv1.mean();
      x0[2]=iv2.mean();
      x0[3]=iv3.mean();
      s0[1]=iv1.radius();
      s0[2]=iv2.radius();
      s0[3]=iv3.radius();
      sub[i]=SubsetOfSpace::boxoid(x0,s0,toMap());
         // box transformed by toMap()
      sub[i].set_dp(dp);
   }
// defining box
   Z nC=4;
   if (dimVec==3) nC*=2;
   V<Rdim> corn(nC);
   corn[1]=Rdim(0,0); 
   corn[2]=Rdim(0,1);
   corn[3]=Rdim(1,0); 
   corn[4]=Rdim(1,1);
   if (dimVec==3){
      Rdim co=corn[1];
      co[3]=1;
      corn[5]=co;

      co=corn[2];
      co[3]=1;
      corn[6]=co;

      co=corn[3];
      co[3]=1;
      corn[7]=co;

      co=corn[4];
      co[3]=1;
      corn[8]=co;
   }

   corners=V<Spc>(nC);
   for (i=1;i<=nC;i++) corners[i]=mapInv(corn[i]);

   Rdim cen(0.5,0.5);
   if (dimVec==3) cen[3]=0.5;
   center=mapInv(cen);

   RefSys rs0;
   box=Box(corners,rs0);
   CPM_MZ
}

template <class Dat>
Z ParOfSpc<Dat>::repNegDim()const
// we have neighbors only in the direction where we have a torus
// In each direction we consider only one neighbor 
// Thus we have 1+2^count neighbors 
{
   Z res=1;
   if (isTorus1) res*=2;
   if (isTorus2) res*=2;  
   if (isTorus3 && dimVec==3) res*=2;
   if (res==1) return 1;
   else return res+1;
}

template <class Dat>
V<Spc> ParOfSpc<Dat>::repNeg(const Spc& x)const
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::repNeg(const Spc& x)const");
   cpmmessage(mL,loc&" started");
   Z d=repNegDim();
   V<Spc> res(d,x);
   if (d==1) return res;
   Group g_1=!g1;
   Group g_2=!g2;
   Group g_3;
   if (isTorus3 && dimVec==3) g_3=!g3;

   Z l=1;
   if (isTorus1){
      l++;
      res[l]*=g1;
      l++;
      res[l]*=g_1;
   }
   if (isTorus2){
      l++;
      res[l]*=g2;
      l++;
      res[l]*=g_2;
   }
   if (isTorus3 && dimVec==3){
      l++;
      res[l]*=g3;
      l++;
      res[l]*=g_3;
   }

   cpmmessage(mL,loc&" done");
   return res;
}

template <class Dat>
Record ParOfSpc<Dat>::toRecord()const
{
   Record rc;
   Word sec="partition_data";
   rc.set(sec,"xL",dat.get_xL());
   rc.set(sec,"xU",dat.get_xU());
   rc.set(sec,"zL",dat.get_zL());
   rc.set(sec,"zU",dat.get_zU());
   rc.set(sec,"nL1",nL1);
   rc.set(sec,"nL2",nL2);
   rc.set(sec,"nL3",nL3);
   rc.set(sec,"nS1",nS1);
   rc.set(sec,"nS2",nS2);
   rc.set(sec,"nS3",nS3);
   rc.set(sec,"overlap1",overlap1);
   rc.set(sec,"overlap2",overlap2);
   rc.set(sec,"overlap3",overlap3);
   rc.set(sec,"isTorus1",isTorus1);
   rc.set(sec,"isTorus2",isTorus2);
   rc.set(sec,"isTorus3",isTorus3);
   return rc;
}

template <class Dat>
V< X2<Spc,Z> > ParOfSpc<Dat>::epLatticePrel
   (const V<X2<R,bool> >& vx, Rdim xL, Rdim xU, R& facCell)const
{
   Z mL=2; //2
   Z mL2=3; //3
   const R facCellReduc=0.7071; // 1./sqrt(2)
   const Z nTryMax=5;
   Word loc("epLatticePrel(...)");
   Z nTry=0;
again:
   if (nTry>0){
      facCell*=facCellReduc;
   }
   cpmassert(nTry<=nTryMax,loc);
   cpmmessage(1,loc&" started, facCell="&cpmwrite(facCell)&", nTry="&
      cpmwrite(nTry));

   Z n=vx.dim();
   if (n==1){
      Rdim pc=(xL+xU)*0.5;
      Spc xc=mapInv(pc);
      V< X2<Spc,Z> > res(1);
      res[1]=X2<Spc,Z>(xc,1);
      return res;
   }
   if (n==2){
      Spc xc1=mapInv(xL);
      Spc xc2=mapInv(xU);
      V< X2<Spc,Z> > res(2);
      res[1]=X2<Spc,Z>(xc1,1);
      res[2]=X2<Spc,Z>(xc2,1);
      return res;
   }
// Counting the carriers;
   Z i,nCarrier=0;
   for (i=1;i<=n;i++) if (vx[i].c2()==true) nCarrier++;
   Z nToner=n-nCarrier;
   cpmmessage("nCarrier on input: "&cpmwrite(nCarrier));
   cpmmessage("nToner on input: "&cpmwrite(nToner));
   R_Vector vc(nCarrier);
   V<Z> zc(nCarrier);
   R_Vector vt(nToner);
   V<Z> zt(nToner);
   Z it=0,ic=0;
   for (i=1;i<=n;i++){
      R r=vx[i].c1();
      bool c=vx[i].c2();
      if (c){
         ic++;
         vc[ic]=r; 
         zc[ic]=i;
      }
      else{
         it++;
         vt[it]=r;
         zt[it]=i;
      }
   }

// Determining the spacing of the lattice
   R fracTC=(1.*nToner)/nCarrier;

   R_Vector mvc=CpmArrays::moments(vc);
   R_Vector mvt=CpmArrays::moments(vt);
   R rC=(mvc[1]+mvc[3])*facCell;
   R rT=(mvt[1]+mvt[3])*facCell;
      // mv.[3] is the standard deviation
   // These are the radii of bounding sheres to be taken into account
   // in real space. Since we will build the lattice in the unit cube
   // (square) we have to form the corresponding diameters in 
   // the cube parameter space:
   Rdim x0;
   x0.set(0.5);
   Spc sp0=mapInv(x0);
      // central point in Spc controlled by ParOfSpc
      // not of the cell defined by xL,xU

   V<Spc> spC(dimVec);
   V<Spc> spT(dimVec);
   for (i=1;i<=dimVec;i++){
      Rdim dC,dT; // all components 0
      dC[i]=rC;
      dT[i]=rT; // i-th component set
      spC[i]=sp0+Vec(dC);
      spT[i]=sp0+Vec(dT);
   }
   V<Rdim> mapC(dimVec);
   V<Rdim> mapT(dimVec);
   for (i=1;i<=dimVec;i++){
      mapC[i]=map(spC[i]);
      mapT[i]=map(spT[i]);
   }
   Rdim maprC;
   Rdim maprT;
   for (i=1;i<=dimVec;i++){
      maprC[i]=mapC[i][i]-x0[i];
      maprT[i]=mapT[i][i]-x0[i];
   }
   // maprC gives the radii for carrier particles to be used in the unit
   // cube(square) in the 3(or 2) orthogonal space directions;
   // for T correspondingly.
// now we define a lattice for carrier positions first.
// characteristics: Lines in 1-direction, since this is the assumed
// direction of the magnetic field. Spacing in i-direction ( i=2,3):
// 2*(maprC[i]+maprT[i]).
// How many lines?. 
   Rdim width;
   for (i=1;i<=dimVec;i++) width[i]=xU[i]-xL[i];
   V<Z> nC(dimVec);
   nC[1]=(Z)(width[1]/(2*maprC[1]));
   for (i=2;i<=dimVec;i++) nC[i]=(Z)(width[i]/(2*(maprC[i]+maprT[i])));
   for (i=1;i<=dimVec;i++) if (nC[i]<1) nC[i]=1;
   Z nTotC=nC[1]*nC[2];
   if (dimVec==3) nTotC*=nC[3];
      // nTotC is the number of carrier particles to be placed in total
      // if the volume defined by the xU[i] and xL[i] is to be filled
      // according to the carrier line scheme completely.
// Actually, we should use an integer multiple of our initial particle
// supply vx.
   R dy2,dy3,f2,f3;
   dy2=width[2]-nC[2]*2*(maprC[2]+maprT[2]);
      // I had (maprC[i]+maprT[i]) till 03-02-12
   if (dy2<0){
      nTry++;
      goto again;
   }

   f2=width[2]/(width[2]-dy2);
      // for filling the range of width[2], which in
      // the main application is related to the space defined by the
      // fill-in space
   if (dimVec==3){
      
      dy3=width[3]-nC[3]*2*(maprC[3]+maprT[3]);
      if (dy3<0){
         nTry++;
         goto again;
      }
      f3=width[3]/(width[3]-dy3);
   }

// Let us generate the list of carrier positions:
   cpmassert(nTotC>0,loc);
   V<Spc> xC(nTotC);
 
   Z j,k,l;
   Rdim pRun;
   l=0;
   R dsC1=2*maprC[1];
   R dsC2=2*(maprC[2]+maprT[2]);
   R dsC3=(dimVec==2 ? 0 : 2*(maprC[3]+maprT[3]));
      // factor 2 was missing till 03-02-20
   R s0C2=maprC[2]+maprT[2];
   R s0C3=(dimVec==2 ? 0 : maprC[3]+maprT[3]);
   dsC2*=f2; // adjusting to fill width[2]
   s0C2*=f2;
   if (dimVec==3){
      dsC3*=f3; // adjusting to fill width[3]
      s0C3*=f3;
   }
   for (i=1;i<=nC[1];i++){
      pRun[1]=xL[1]+dsC1*(i-0.5);
         // xL[1]+ and corresponding terms 
         // in the interior loops and in the toner loops were
         // missing till 02-09-13
      for (j=1;j<=nC[2];j++){
         pRun[2]=xL[2]+s0C2+dsC2*(j-1);
         if (dimVec==2){
            xC[++l]=mapInv(pRun);
         }
         else{
            for (k=1;k<=nC[3];k++){
               pRun[3]=xL[3]+s0C3+dsC3*(k-1);
               xC[++l]=mapInv(pRun);
            }
         }
      }
   }

   Vio<Z> nT(dimVec);
   nT[1]=cpmsup(1,(Z)(nC[1]*fracTC));
   for (i=2;i<=dimVec;i++){
      nT[i]=cpmsup(1,(nC[i]-1));
   }

   Z nTotT=nT[1]*nT[2];
   if (dimVec==3) nTotT*=nT[3];

   R sLine=dsC1*cpmsup(1,(nC[1]-1));
      // length of a toner line
  
   R dsT1=sLine/nT[1]; // nT[1]>=1 !
      // space for a toner
   R dsT2=dsC2;
   R dsT3=dsC3;
   R s0T2=s0C2+0.5*dsC2;
   R s0T3=(dimVec==2 ? 0 : s0C3+0.5*dsC3);
   cpmassert(nTotT>0,loc);
   V<Spc> xT(nTotT);
   Vec shift2, shift3;;
   shift2[2]=0.75*rT;
   shift3[3]=0.75*rT;
   R r2,r3;
   l=0;
   for (i=1;i<=nT[1];i++){
      pRun[1]=xL[1]+0.5*(dsC1+dsT1)+(i-1)*dsT1;
      r2=2*(i%2)-1; // toggles between +1 and -1
            // this is more effective as a random shift
      for (j=1;j<=nT[2];j++){ // was <
         pRun[2]=xL[2]+s0T2+dsT2*(j-1);
         if (dimVec==2){
            xT[++l]=mapInv(pRun)+shift2*r2;
         }
         else{
            for (k=1;k<=nT[3];k++){  // was <
               pRun[3]=xL[3]+s0T3+dsT3*(k-1);
               r3=2*(k%2)-1;
               xT[++l]=mapInv(pRun)+shift2*r2+shift3*r3;
            }
         }
      }
   }
   Z nTotDev=nTotC+nTotT;
   V<X2<Spc,Z> > res(nTotDev);
   cpmmessage(loc&", nTotC="&cpmwrite(nTotC));
   cpmmessage(loc&", nTotT="&cpmwrite(nTotT));
   cpmmessage(loc&", nTotDev="&cpmwrite(nTotDev));
   
   l=0;
   for (i=1;i<=nTotC;i++){
      res[i].c1()=xC[i];
      cpmmessage(mL2,"i="&cpmwrite(i)&", xC[i][1]="
         &cpmwrite(xC[i][1])&", xC[i][2]="&cpmwrite(xC[i][2]));
      res[i].c2()=zc.cyc(++l);
   }

   l=0;
   Z iAct=nTotC;
   for (i=1;i<=nTotT;i++){
      iAct++;
      res[iAct].c1()=xT[i];
      cpmmessage(mL2,"i="&cpmwrite(i)&", xT[i][1]="
         &cpmwrite(xT[i][1])&", xT[i][2]="&cpmwrite(xT[i][2]));
      res[iAct].c2()=zt.cyc(++l);
   }
   cpmmessage(mL,loc&" done");
   return res;
}

template <class Dat>
V< X2<Spc,Z> > ParOfSpc<Dat>::epLattice(const V<X2<R,bool> >& vx, 
      Rdim xL, Rdim xU, R facCell, Z nPar)const 
// implements a very efficient strategy to reach the 
// prescribed particle number nPar approximately
{
   Z mL=2;
   static Word loc("epLattice(...)");
   CPM_MA
   const R incr=1.1;
   const R decr=1./incr;
   if (nPar<=0){
        // here, only the facCell input is used
      cpmmessage(mL, loc&" to be done for nPar<=0");
      return epLatticePrel(vx,xL,xU,facCell);
   }
   else if (nPar<=2){
      cpmmessage(mL, loc&" to be done for 0<nPar<=2");
      return epLatticePrel(vx.resize(nPar),xL,xU,facCell);
   }
   else ;
   // now the facCell is taken as a starting point
   // to get approximately nPar as the dimension of the 
   // result (return value)
   bool succ=0;
   R fac=facCell;
   V< X2<Spc,Z> > res=epLatticePrel(vx,xL,xU,fac);
   V< X2<Spc,Z> > resLast=res;
   Z d=res.dim();
   Z dLast=resLast.dim();
   Z dDiff=d-nPar; // >0 if we create more lattice points as we need;
   Z dDiffLast=dLast-nPar;
   if (dDiff==0){
      cpmmessage(mL, loc&" done for nPar==res.dim()");
      return res;
   }
      // thus we have to increase particle spacing
   R fid=(dDiff>0 ? incr : decr); 
   while (!succ){
      fac*=fid;
      resLast=res;
      res=epLatticePrel(vx,xL,xU,fac);
      dLast=resLast.dim();
      d=res.dim();
      dDiff=d-nPar;
      dDiffLast=dLast-nPar;
      succ=(dDiff*dDiffLast<=0);
   }
   if (cpmabs(dDiff)<=cpmabs(dDiffLast)){
      cpmmessage(mL,loc&" done exit 1");
      return res;
   }
   else{
      cpmmessage(mL,loc&" done exit 2");
      return resLast;
   }
}

template <class Dat>
V<Spc> ParOfSpc<Dat>::testPoints(Z n)const
{
   V<Rdim> vp=dat.testPoints(n);
   Z vpn=vp.dim();
   V<Spc> res(vpn);
   for (Z i=1;i<=vpn;i++) res[i]=Spc(vp[i]);
   return res;
}

template <class Dat>
void ParOfSpc<Dat>::autoFrame1(Camera& ca, R autoFrameFactor)
{
   Z mL=3;
   Word loc("ParOfSpc<Dat>::autoFrame1(...)");
   CPM_MA
   if (autoFrameFactor<=0);
   else{
      Spc xp=center;
      Spc xc=ca.get_x();
      Vec shift=xp-xc;
      ca*=Group(shift);
      T2<V<C> > pvv=ca.capt(corners);
      V<C> capt=pvv.c1()&pvv.c2();
      Z dc=capt.dim();
      V<R> captx(dc);
      V<R> capty(dc);
      Z i;
      for (i=1;i<=dc;i++){
         captx[i]=capt[i][1];
         capty[i]=capt[i][2];
      }
      Iv ivx(captx);
      Iv ivy(capty);
      ivx.stretch(autoFrameFactor);
      ivy.stretch(autoFrameFactor);
      R dx=ivx.abs();
      R dy=ivy.abs();
      cpmassert(dy>0,loc);
      R aspRatObject=dx/dy;
      R aspRatFrame=ca.getGraph().getAspRat();
      R relFac=aspRatFrame/aspRatObject;
      if (aspRatFrame>=aspRatObject){
         ca.getGraph().setXY(ivx.stretch(relFac),ivy);
      }
      else{
         ca.getGraph().setXY(ivx,ivy.stretch(1./relFac));
      }
   }
   CPM_MZ
}
