//? cpmregzah.h
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_REGZAH_H_
#define CPM_REGZAH_H_
/*
   Description: 'Financial analysis' not regular part of the
      scientific class system. Comments in German. Better ignore.
*/

#include <cpms.h>
#include <cpmtypes.h>
#include <cpmgreg.h>

namespace CpmGeld{

using namespace CpmStd;

using CpmRoot::Word;
using CpmRoot::R;
using CpmRoot::Z;
using CpmRootX::R1;
using CpmArrays::S;
using CpmArrays::X2;
using CpmArrays::V;
using CpmTime::Greg;

const R EUROdurchDM=1.95583;

const R EUR=1.;
   // Waehrungseinheit
const R DM=R(1)/EUROdurchDM;
const R DEM=DM;
const R zinsJahr=360;
const R zinsJahrInv=R(1)/zinsJahr;

R zins(R k, R p, R t);
// k: Kapital, p: Zinsfuss (z. B. 6.5), t: Zahl der Tage
// Ergebnis ist der Zins selbst, nicht k+zins

R zinsPlus(R k, R p, R t);
// k: Kapital, p: Zinsfuss (z. B. 6.5), t: Zahl der Tage
// Ergebnis ist nicht der Zins selbst sondern k+zins

R zinsesZins(R k, R p, R t);
// Wert des Anfangskapitals k, das mit Zinsfuss p
// t Tage verzinst wurde (mit Wiederanlage der Jahreszinsen).
// res=k*(1+q*r)*(1+q)^n
// wobei t/zinsJahr=:n+r, n \in {0,1,2,...}, 0<=r<1, q:=p/100
// Das scheint die Geldvermehrung nach heutiger Bankpraxis
// bei Geldanlage zu sein.

R anLage(R k0, R p, R anl, R dtAnl, R t);
// Wert eines Kapitals das nach t Tagen entsteht, wenn man zum
// Anfangskapital k0 nach jedem Ablauf der Zeitspanne dtAnl*Tag die
// zu dieser Zeitspanne und dem Zinsfuss p gehoerigen Zinsen und
// dann den Wert anl hinzufuegt.

R anLageV(V< X2<Greg,R> > const& v, R p, Greg const& tEnd);
// Geldbetrag der zum Zeitpunkt tEnd aus den Einzahlungen v
// durch Verzinsung mit dem Zinsfuss p entsteht.

R rendite(const Greg& t1, const Greg& t2, R k1, R k2, R anl);
// Seien t1 und t2 zwei Einzahldaten fuer einen Fond (t1<t2), z.B.
// t1 der Tag der ersten Einzahlung.
// k1 und k2 seien der Gesamtwert der Fondanteile zu den ensprechenden
// Zeiten (also z.B. k1=anl).
// anl sei die monatliche (in EUR konstante) Einzahlung.
// Das Resultat ist dann der Zinsfuss mit dem man dieselbe Wertsteigerung
// des Anfangskapitals bei Wiederanlage der Zinsen und regelmaessiger
// Einzahlung erhielte.

R renditeV(V< X2<Greg,R> > const& v, R k, Greg const& t);
// Durch die Einzahlungen v sei zum Zeitpunkt t, der Wert k
// entstanden (natuerlich werden nur die Komponenten in v gezaehlt, deren
// Einzahlungs-Zeitpunkt vor t liegt. Zurueckgegeben wird der Zinsfuss p
// fuer
// den gilt: k==anLageV(v,p,t).

class RegZah{ // Regelmaessige Zahlung (periodic payment)
   Word name;
      // Alle Daten vom Typ Word duerfen Leerzeichen enthalten
      // also Ketten von Woertern herkoemmlicher Deutung sein
   Word typ;
   S<Word> sch;
      // Menge von Schluesselwoertern
      // Damit readFormatted funktioniert, darf kein Schluesselwort
      // Leerzeichen enthalten
   R sum{0.};
      // Geldbetrag, positiv fuer Einkuenfte, negativ fuer zu leistende
      // Zahlungen
   R per{0.};
      // Zahlungsperiode

public:
   typedef RegZah Type;
   CPM_IO_V
   CPM_ORDER
   CPM_NAM_V(RegZah)

   RegZah(){}
      // all members have a default constructor
   RegZah(R sum_, R per_, const Word& name_, const Word& typ_):
      sum(sum_),per(per_),name(name_),typ(typ_){}
   R monZah()const;
      // monatliche Zahlung (verschieden von sum, wenn
      // per verschieden von 1; also z. B. fuer Zahlungen
      // die eimal im Jahr faellig sind (viele Versicherungen)
   Word getName()const{ return name;}
   Word getTyp()const{ return typ;}
   S<Word> getSch()const{ return sch;}
   bool hasSch(const Word& w)const{ return sch.hasElm(w);}
   bool hasTyp(const Word& w)const{ return typ==w;}
   bool writeFormatted(ostream&)const;
      // writes the content of *this  in such a
      // way to a stream, that the following read function can
      // recover it from there.
   bool readFormatted(istream&);
};

} // namespace

#endif
