//? tut2a.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


// tut2a.cpp is an experimental version
// first aim: clarify the unexpected behavior after finishing the
// program run: no console window waiting for explicit closing
// remains.
// first idea: Perhaps Vp does not work well after the reorganization of
// V<>.
/*************************************************************************
   Studying the Kepler oscillator
*********************************************************************/
#include <iomanip> // for setprecision(Z)
#include <cpmbas1.h>
   // Actually, this is a collecion of header files.
#include <cpmvectormore.h>
#include <cpmrmatrix.h>
#include <cpmcmatrix.h>
#include <cpmrmateig.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;

using CpmLinAlg::R2;
using CpmGeo::Iv;
using CpmAlgorithms::solKepEqu;
using CpmAlgorithms::CyclicAlarm;
using CpmImaging::RED;
using CpmImaging::ORANGE;
using CpmImaging::YELLOW;
using CpmImaging::GREEN;
using CpmImaging::CYAN;
using CpmImaging::MAGENTA;
using CpmImaging::darkSky;
using CpmImaging::twiSky;
using CpmImaging::morSky;

namespace CpmKepler{

B stepRed_;

class KepOsc{ // 'Kepler oscillator'
// My name for the system defined as the radial part of the Kepler
// motion arround a fixed central mass. Physical units are used in which
// the moving mass, the gravitational potential at unit distance, and the
// angular momentum have value 1.
// see www.ulrichmutze.de/articles/99-271.pdf, Section 4
// This class implements the dynamics in a naive manner which is meant
// only as a pedagogical device. Classes that implement useful second order
// algorithms for dynamics will be derived from KepOsc.

protected:
// data

   R t_,x_,v_,dt_; // time, position, velocity, time step

// static functions
   static R f(R const& x){R xi=cpminv(x); return (xi-1)*xi*xi;}
      // force law
// member functions
   void init_( R eps, R nPerRev, R E, R t);
      // assumes 0<=eps<1, E arbitrary
      // initializes the common data

 public:
   virtual Word nameOf()const{ return "KepOsc";}
   KepOsc(){init_(0.,32.,0.,0.);}
   explicit KepOsc(R eps, R nPerRev=64., R E=0., R t=0.)
      {init_(eps,nPerRev,E,t);}
   virtual KepOsc* clone()const{ return new KepOsc(*this);}
   R t()const{ return t_;}
   R dt()const{ return dt_;}
   R x()const{ return x_;}
   R v()const{ return v_;}
   R eKin()const{ return v_*v_*0.5;}
      // kinetic energy
   R ePot()const{ R xi=cpminv(x_); return xi*(xi*0.5-1);}
      // potential energy
   R eTot()const{ return eKin()+ePot();}
      //: e total
   R force()const{ return f(x_);}
      //: force
      // The superficially correct name 'for' is not correct since it is
      // a C++ key word.
   R_Vector orbDes()const;
      //: orbit descriptors
      // Returns a list of the usual orbit elements
      // a: major semi-axis, eps: numerical excentricity,
      // n: mean motion, E: excentric anomaly, M: mean anomaly,
      // tP: orbital period (revolution time)
   void rev_(){ v_=-v_;}
      //: reversed
      // Notice name ending in '_' since it is a non-constant function
   void extEvl_(R t, R acc=1e-8);
      //: exact evolution as mutating operation

  // KepOsc extEvl(R t, R acc=1e-8)const;
      //: exact evolution as creating operation

   virtual void step_()
      //: (evolution) step
      // Here only a implementation of the explicit
      // Euler rule. Notice mass is 1 so f(x_) is both force and
      // acceleration. It is meant as a definition of the law of motion
      // which in mathematical style would be written as the
      // first order ODE dx/dt=v, dv/dt=f(x)
   {
      R vm=v_;
      v_+= f(x_)*dt_;
      x_+= (v_+vm)*0.5*dt_;
      t_= t_ + dt_;
   }

   virtual void shf_(R dx, R dv){ x_+=dx; v_+=dv; }
      //: shift
      // virtual since in derived classes which use dependent state
      // variables one has to adjust to the new values of x_ and v_.

   X3<R,Iv,Iv> orbSize()const;
      //: orbit size
      // intervals which contain the t_-range, x_-range , the y_range
      // for a full revolution

   virtual Color col()const{ return ORANGE; }
      //: color for visualization

   void mark(Graph& g)const{ g.mark(R2(x_,v_),col());}
      //: mark

   R2 phsPos()const{ return R2(x_,v_);}
      //: phase (space) position

   R per()const{ return orbDes()[6];}
      //:period

   virtual Word met()const{ return "Euler";}
 };

void KepOsc::init_( R eps, R nPerRev, R E, R t)
// initializing x_ anv v_ from numerical excentricity and excentric
// anomaly. Further, initialization of the dependent quantities.
{
   Word loc("KepOsc::init_(R,R,R,R)");
   cpmassert(eps>=0,loc);
   cpmassert(eps<1,loc);
   t_=t;
   R ai=(1.-eps)*(1+eps);
   R a=1./ai;
   x_=a*(1.-eps*cpmcos(E));
   v_=eps*cpmsqrt(a)*cpmsin(E)/x_;
   R tP=cpm2pi*cpmpow(a,1.5);
   dt_=tP/nPerRev;
}

R_Vector KepOsc::orbDes()const
{
   Word loc("KepOsc::orbDes()");
   Z mL=2;
   CPM_MA
   R e=eTot();
   if (e>=0){
      cpmcerr<<"We have e>=0, un-bound state. x_="<<x_<<" v_="<<v_<<endl;
      cpmdebug(ePot());
      cpmdebug(eKin());
   }
   cpmassert(e<0,loc); // we need a bound state
      // then also x>0.5
   R a=-0.5*cpminv(e);
   R eps=cpmsqrt(1.-1./a);
   R n=cpmpow(a,-1.5);
   R tP=cpm2pi/n;
   R z1=1-x_/a;
   R z2=x_*v_/cpmsqrt(a);
   C z(z1,z2);
   R E=z.arg();
   R M=E-eps*cpmsin(E);
   CPM_MZ
   return R_Vector{a,eps,n,E,M,tP};
}

void KepOsc::extEvl_(R t, R acc)
// exact step from time t_ to time t, which may be arbitrarily large.
// Based on the famous transcendental Kepler equation.
{
   Word loc("KepOsc::extEvl_(R,R)");
   Z mL=2;
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],M=y[5];
   M+=n*(t-t_);
   R E=solKepEqu(M,eps,acc);
   x_=a*(1-eps*cpmcos(E));
   cpmassert(x_>=0.5,loc); // for safty, not needed
   v_=eps*a*a*n*cpmsin(E)/x_;
   t_=t;
   CPM_MZ
}

/***********
KepOsc KepOsc::extEvl(R t, R acc)const
{
   KepOsc res(*this);
   res.extEvl_(t,acc);
   return res;
}
****************/

X3<R,Iv,Iv> KepOsc::orbSize()const
{
   Z mL=2;
   Word loc("KepOsc::orbSize()");
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],tP=y[6];
   R xMin=a*(1-eps);
   R xMax=a*(1+eps);
   R vMax=eps*a*a*n;
   CPM_MZ
   return X3<R,Iv,Iv>(tP,Iv(xMin,xMax),Iv(-vMax,vMax));
}

// A version of KepOsc which does the exact evolution more economicly
// by transfering some work into the constructor and storing the result
// in suitable data elements.

class KepExt{
   KepOsc rep_;
   R a_,eps_,n_,M_,eaan_;
   void ini_()
   {
      R_Vector y=rep_.orbDes();
      a_=y[1];eps_=y[2];n_=y[3];M_=y[5];
      eaan_=eps_*a_*a_*n_; // do this once
   }
public:
   explicit KepExt(KepOsc const& ko):rep_(ko){ini_();}

   explicit KepExt(R eps, R E=0.):rep_(eps,1.,E){ini_();}

   R2 xv(R t, R acc=1.e-12)const
   {
      R M=M_+n_*(t-rep_.t());
      R E=solKepEqu(M,eps_,acc);
      R x=a_*(1.-eps_*cpmcos(E));
      R v=eaan_*cpmsin(E)/x;
      return R2(x,v);
   }
};

// Now the other integration methods are defined a derived classes with
// properly ovewritten virtual functions

// Runge Kutta integrator (2nd order)
// needs only the common data.

class KepRK2 :public KepOsc{
   public:
   Word nameOf()const{ return "KepRK2";}
   KepRK2():KepOsc(){}
   explicit KepRK2(R eps, R nPerRev=64, R E=0):KepOsc(eps,nPerRev,E){}
   KepOsc* clone()const{ return new KepRK2(*this);}
      // all derived classes have to define clone in order to make my
      // polymorphic containers Pp<> and Vp<> work
   void step_()
   {
      R tau=dt_*0.5;
      R a=f(x_);
      R am=f(x_+v_*tau);
      R dv=am*dt_;
      R dx=(v_+a*tau)*dt_;
      x_+=dx;
      v_+=dv;
      t_+=dt_;
   }
   Color col()const{ return WHITE; }
   Word met()const{ return "RK2";}
};

// Direct midpoint integrator (2nd order)
// needs only the common data.

class KepDMI :public KepOsc{
   public:
   Word nameOf()const{ return "KepDMI";}
   KepDMI():KepOsc(){}
   explicit KepDMI(R eps, R nPerRev=64, R E=0):KepOsc(eps,nPerRev,E){}
   KepOsc* clone()const{ return new KepDMI(*this);}
      // all derived classes have to define clone in order to make my
      // polymorphic containers Pp<> and Vp<> work
   void step_()
   {
      R tau=dt_*0.5;
      t_+=tau;
      x_+=(v_*tau);
      v_+=f(x_)*dt_;
      x_+=(v_*tau);
      t_+=tau;
   }
   Color col()const{ return YELLOW; }
   Word met()const{ return "DMI";}
};

class KepALF :public KepOsc{
   R w_,a_;
public:
   Word nameOf()const{ return "KepALF";}
   KepALF():KepOsc(),w_(v_),a_(f(x_)){}
   explicit KepALF(R eps, R nPerRev=64, R E=0):
      KepOsc(eps,nPerRev,E),w_(v_),a_(f(x_)){}
   KepOsc* clone()const{ return new KepALF(*this);}

   void shf_(R dx, R dv)
   { x_+=dx; v_+=dv; w_=v_; a_=f(x_);}

   void steplet_(R dt)
   {
      R tau=dt*0.5;
      R fac=2.;
      t_ += tau;
      x_ += tau*w_;
      v_ += tau*a_;
      w_ = v_*2.-w_;
      a_ = f(x_)*2.-a_;
      v_ += tau*a_;
      x_ += tau*w_;
      t_ += tau;
   }

   void step_()
   {
      if (stepRed_){
         R tau=dt_*0.5;
            steplet_(tau);
            steplet_(tau);
      }
      else steplet_(dt_);
   }

   Color col()const{ return GREEN; }
   Word met()const{ return "ALF";}
};

class KepLF :public KepOsc{
   R tp_,xp_,vp_; // p for previous
   void initLF_()
      // does the additional initializations
      // simplified version, we need not be more accurate than second
      // order
   {
      R h= -dt_;
      if (stepRed_) h*=0.5;
      tp_=t_+h;
      R a=f(x_);
      R dv=a*h*0.5;
      xp_=x_+(v_+dv)*h; // here xp_ is previous x_
      vp_=v_+(a+f(xp_))*h*0.5;
   }
public:
   Word nameOf()const{ return "KepLF";}
   KepLF():KepOsc(){initLF_();}
   explicit KepLF(R eps, R nPerRev=64, R E=0):
      KepOsc(eps,nPerRev,E){initLF_();}
   KepOsc* clone()const{ return new KepLF(*this);}

   void steplet_(R dt)
   {
      R dt2=dt*2;
      R tn=tp_+dt2;
      R vn=vp_+f(x_)*dt2;
      R xn=xp_+v_*dt2;
      tp_=t_;
      vp_=v_;
      xp_=x_;
      t_=tn;
      v_=vn;
      x_=xn;
   }

   void step_()
   {
      if (stepRed_){
         R tau=dt_*0.5;
            steplet_(tau);
            steplet_(tau);
      }
      else steplet_(dt_);
   }

   void shf_(R dx, R dv)
   { x_+=dx; v_+=dv; initLF_();}

   Color col()const{ return RED; }
   Word met()const{ return "LF";}
};

class KepADALF :public KepOsc{
   R w_,a_;
public:
   Word nameOf()const{ return "KepADALF";}
   KepADALF():KepOsc(),w_(v_),a_(f(x_)){}
   explicit KepADALF(R eps, R nPerRev=64, R E=0):
      KepOsc(eps,nPerRev,E),w_(v_),a_(f(x_)){}
   KepOsc* clone()const{ return new KepADALF(*this);}

   void shf_(R dx, R dv)
   { x_+=dx; v_+=dv; w_=v_; a_=f(x_);}

   void step_()
   {
      R tau=dt_*0.25;
      R tau2=dt_*0.5;
      t_ += tau;
      x_ += tau*w_;
      v_ += tau*a_;
      w_ = v_*2.-w_;
      a_ = f(x_)*2.-a_;
      R wm=w_;
      R am=a_;
      t_ += tau2;
      x_ += tau2*w_;
      v_ += tau2*a_;
      w_ = v_*2.-w_;
      a_ = f(x_)*2.-a_;
      v_ += tau*a_;
      x_ += tau*w_;
      t_ += tau;
      a_=(a_+am)*0.5;
      w_=(w_+wm)*0.5;
   }

   Color col()const{ return CYAN; }
   Word met()const{ return "ADALF";}
};

// averaged asynchronous leap-frog (2nd order)
// not densified! New, change 2013/14.
// This is an inferior method that I deviced and tested at the
// very beginning of investigation of second order integrators
// for the time-dependend Schroedinger equation. I discarded the
// method and forgot it. Finally I re-invented it as the simplest
// version of the averaging method.

class KepAALF :public KepOsc{
   R w_,a_;
public:
   Word nameOf()const{ return "KepAALF";}
   KepAALF():KepOsc(),w_(v_),a_(f(x_)){}
   explicit KepAALF(R eps, R nPerRev=64, R E=0):
      KepOsc(eps,nPerRev,E),w_(v_),a_(f(x_)){}
   KepOsc* clone()const{ return new KepAALF(*this);}

   void shf_(R dx, R dv)
   { x_+=dx; v_+=dv; w_=v_; a_=f(x_);}

   void steplet_(R dt)
   {
      R tau=dt*0.5;
      R xm=x_;
      R vm=v_;
      t_ += tau;
      x_ += tau*w_;
      v_ += tau*a_;
      w_ = v_;
      a_ = f(x_);
      v_ = vm + dt*a_;
      x_ = xm + dt*w_;
      t_ += tau;
   }

   void step_()
   {
      if (stepRed_){
         R tau=dt_*0.5;
            steplet_(tau);
            steplet_(tau);
      }
      else steplet_(dt_);
   }

   Color col()const{ return MAGENTA; }
   Word met()const{ return "AALF";}
};

// The file input select integration methods simply by integers; here we
// select the corresponding classes.

Pp<KepOsc> singlet(Z method, R eps, R nPerRev, R E=0.)
{
   Pp<KepOsc> res;
   if (method==-1) res=KepOsc(eps,nPerRev,E);
   else if (method==0) res=KepRK2(eps,nPerRev,E);
   else if (method==1) res=KepDMI(eps,nPerRev,E);
   else if (method==2) res=KepALF(eps,nPerRev,E);
   else if (method==3) res=KepLF(eps,nPerRev,E);
   else if (method==4) res=KepADALF(eps,nPerRev,E);
   else  res=KepAALF(eps,nPerRev,E);
   return res;
}

// trying to replace Vp<KepOsc> by V<KepOsc*>
Vp<KepOsc> generator(S<Z> const& methods, R eps, R nPerRev, R E=0.)
{
   Z d=methods.car();
   Vp<KepOsc> res(d);
   Z i=0;
   S<Z> m{methods}; // to avoid the dubious error on possible
   // conflict with constantness of method.
   for (auto q=m.cbegin(); q!=m.cend(); ++q){
      Z meti=*q;
      i++;
      if (meti==-1) res[i]=KepOsc(eps,nPerRev,E);
      else if (meti==0) res[i]=KepRK2(eps,nPerRev,E);
      else if (meti==1) res[i]=KepDMI(eps,nPerRev,E);
      else if (meti==2) res[i]=KepALF(eps,nPerRev,E);
      else if (meti==3) res[i]=KepLF(eps,nPerRev,E);
      else if (meti==4) res[i]=KepADALF(eps,nPerRev,E);
      else  res[i]=KepAALF(eps,nPerRev,E);
   }
   return res;
}

//////////////////////////////////////////////////////////////////////////

void cpmplot(V<V<R2> > const& dat, Word const& fileName)
{
   Word loc("cpmplot(V<V<R2>>,Word)");
   Z mL=2;
   CPM_MA
   if (fileName.isVoid()){
      CPM_MZ
      return;
   }
   Z i,j,md=dat.dim();
   V<Z> vd(md);
   for (i=1;i<=md;++i){
      vd[i]=dat[i].dim();
   }
   S<Z> sd(vd);
   cpmassert(sd.car()==1,loc);
   Z nd=sd[1];
   OFileStream ofs(fileName);
   ofs()<<std::setprecision(CpmRoot::wrtPrc);
      // otherwise multiple precision mode would not be flexibly be used
   for (i=1;i<=nd;++i){
      ofs()<<endl;
      for (j=1;j<=md;++j){
         R2 y=dat[j][i];
         ofs()<<y[1]<<" "<<y[2]<<"    ";
      }
   }
   CPM_MZ
}

void app1(void) // orbit in phase space (v = momentum since m=1)
// simultaneously for a selected list of integration methods
{
   Word loc("app1");
   Color cb=darkSky; // night sky blue. Each task appi
   // has here its characteristic background color
   RecordHandler rch("app1.ini");
      // This says that we expect data input to be available in file
      // app1.ini in a diretory which is determined by entries in
      // cpmsystemdependencies.h and cpmconfig.ini.
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, fac=1.2, nPerRev=16, tWaitFinal=1;
   Z nTot=1000,displayPeriod=1;
   Word imageName,plotDataName;
   V<Z> methods;
// getting quantities from section 'data' of file app1.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrhf(stepRed_);
   cpmrh(fac);
   cpmrh(tWaitFinal);
   cpmrhf(imageName);
   cpmrhf(plotDataName);
   cpmrh(displayPeriod);
      // now the program has all its data
   cpmrh(methods);
   S<Z> met(methods); // eliminates doublicates
   Vp<KepOsc> vk=generator(met,eps,nPerRev);
   Z md=met.dim();
   X3<R,Iv,Iv> os=vk(1).orbSize(); // cartesian product templates
      // are defined from X2 to X8
   Iv ivx=os.c2(); // Iv = interval is a valuable type
 //  cpmdebug(ivx);
   Iv ivy=os.c3();
 //  cpmdebug(ivy);
   ivx*=fac; // changing the size of an interval, while holding the center
      // fixed
   ivy*=fac; // same factor for the other direction
   V<V<R2> > vvr(md,V<R2>(nTot)); // memory for all phase space positions
      // during run
   Graph g(cb); // basic graphical 'window to the world'
   g.setXY(ivx,ivy);
      // this associates physical sizes (in space) with the
      // 'pixeled' rectangle g
   R nInv=1./nTot; // for use in cpmprogress
   CyclicAlarm cy(displayPeriod); // device for triggering periodic
      // actions
   for (Z i=1;i<=nTot;++i){
      for (Z j=vk.b();j<=vk.e();++j){
         vk(j).mark(g); // 'mark' acts on memory only
         R2 pj=vk(j).phsPos();
         vvr[j][i]=pj;
         vk(j).step_();
      }
      if (cy()||i==nTot){ // cy() becomes 'true' periodically
         g.vis(); // transfers the whole memory frame
            // to screen. This expensive action should probably not be
            // orderd for each integration step. Object cy allows
            // to achieve this.
         cpmprogress("loop",i*nInv); // progress indicator on
            // pane 1 of he status bar
      }
   }
   g.wrt(imageName); // writes the final frame to a ppm image file
      // if imageName is not the void word.
   cpmplot(vvr,Message::extOut(plotDataName)); // does nothing if the name
   // is void
   cpmwait(tWaitFinal,2); // allows us to inspect the last frame
      // for the time (in seconds) put in here. The second argument
      // says that on pane 2 of the status bar there will be a
      // 'countdown' indication of the time left for inspection.
}

void app2(void) // method error of the two integration methods
// represented by means of interaction picture dynamics. For
// applying this terminology, we interpret the dynamics as defined by
// the numerical methods as resulting from the exact Kepler dynamics
// by 'interaction'. This interaction picture motion is much slower for
// the more accurate direct midpoint method. The Runge Kutta motion
// finally explodes and reaches non-bound states for which the exact
// evolution back is no longer defined (by the method KepOsc::extEvl_).
// Therefore the option is implemented to deal only with the stable
// direct midpoint method.
{
   Word loc("app2");
   Color cb=twiSky; // background
   RecordHandler rch("app2.ini");
   Word sec="data";
   R E=0.,eps=0.25, nPerRev=16, acc=1e-8, tWaitFinal=1;
   Z i,j,nTot=1000,nIter=4;
   Word imageName,plotDataName;
   V<Z> methods;
   B backEvl{true};
   cpmrhf(E);
   cpmrh(eps);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(acc);
   cpmrh(tWaitFinal);
   cpmrh(imageName);
   cpmrhf(plotDataName);
   cpmrh(methods);
   cpmrhf(backEvl);
   cpmrhf(stepRed_);

   S<Z> met(methods); // eliminates doublicates
   Z nm=met.dim();
   cpmassert(nm>0,"app2()");
   Vp<KepOsc> vk=generator(met,eps,nPerRev,E);
   KepOsc k0=vk(1);
   R t0=vk(1).t();
   // This defines the initial state.
   X3<R,Iv,Iv> os=k0.orbSize();
   R2 p0=k0.phsPos();
   R xInv=cpminv(os.c2().abs());
   R vInv=cpminv(os.c3().abs());

   V<V<R2> > vvr(vk.dim(),V<R2>(nTot));
   R nInv=1./nTot;
   for (i=1;i<=nTot;++i){ // no graphical actions
      // within the loop. Terribly fast!!!
      KepOsc k1Act=vk(1); // conversion to base class is OK
         // since we call only functions which don't get re-defined
         // in derived classes
      R tAct=k1Act.t(); // the states are aware of the time to which they
         // belong: Schroedinger picture
      if (!backEvl){
         KepExt kExt(k0);
         R2 posExt=kExt.xv(tAct,acc); // storing the exact phase space
                 // position
         for (j=vk.b();j<=vk.e();++j){
            R2 diff=vk(j).phsPos()-posExt;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      else{
         for (j=vk.b();j<=vk.e();++j){
            KepExt vkj(vk(j));
            R2 pj=vkj.xv(t0,acc);
            R2 diff=pj-p0;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      for (j=vk.b();j<=vk.e();++j){
         vk(j).step_();
      }
      cpmprogress("loop",i*nInv,1,2); // only 2 digits for
         // progress indication to avoid 'much adoo about nothing'
   }
   Graph g(cb);
   g.clr_(cb);
   g.setAutoScale(true);
   g.setNoGridText(false);
   g.setNoGrid(true);
   g.setGridTextColor(LIGHTBLUE);
   V<Color> vc(nm);
   for (j=1;j<=nm;++j) vc[j]=vk(j).col();
   g.mark(vvr,vc); // powerful method for graphical
      // representation of polygons; here one or two
      // of those
   g.vis();
   g.wrt(imageName); // does nothing if file name void
   cpmplot(vvr,plotDataName);
   cpmwait(tWaitFinal,2);
}

void app3(void) // interaction picture dynamics for not a single
// point as an initial condition but for an ensemble of
// phase space points which form a closed curve. The deformation
// of this closed curve by interaction picture dynamics is monitored.
// Here the field of view in phase space follows the moving states.
// Otherwise they would slowly drift out. For the asynchronous
// leapfrog method (method 2) it was observed that the loop
// deformation can lead to crossing over. This puzzled me for a while
// since this cannot happen with mappings of the phase space into
// the phase space. However for the asynchronous leapfrog method
// the statespace (which is propery mapped) has additional degrees
// of freedom, which are not shown in the phase space curves. This
// resolves what seemed to be a logical contradiction. The details of
// the process are not clear to me. Finally the time evolution of the
// enclosed phase area is shown.
{
   Word loc("app3");
   Color cb=morSky; // background
   RecordHandler rch("app3.ini");
   Word sec="data";
   R eps=0.25, fac=1.2, nPerRev=16, acc=1e-8, E=0;
   R facX=1, facV=1, tWaitFinal1=1, tWaitFinal2=1, tWaitFrame=0.1;
   Z nTot=1000, nK=12, displayPeriod=1, method=1;
   Word imageName;
   B fileOnShow;
   V<Z> filingLimits;

   cpmrh(eps);
   cpmrh(E);
   cpmrh(fac);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(acc);
   cpmrh(tWaitFinal1);
   cpmrh(tWaitFinal2);
   cpmrh(tWaitFrame);
   cpmrh(nK);
   cpmrh(facX);
   cpmrh(facV);
   cpmrh(displayPeriod);
   cpmrh(method);
   cpmrhf(imageName);
   cpmrhf(stepRed_);
   cpmrhf(fileOnShow);
   cpmrhf(filingLimits);

   IvZ imgInterval; // void
   if (filingLimits.dim()>=2){
      imgInterval=IvZ(filingLimits[1],filingLimits[2]);
   }
   // It turned out to be necessary for being able to document the
   // phenomena that occur in runs to write screen images to file.
   // Since filing an image takes time, it is desirable to have a way
   // to file only a controlled subset of all screen frames.
   // Letting the program run without any filing one can make notes
   // of the frame numbers that contain the interesting events
   // (such as crossing over of loops). In a second run one then may
   // write these frames to file by suitable setting filingLimits.

   Pp<KepOsc> ko=singlet(method,eps,nPerRev,E);

   R x0=ko().x();
   R v0=ko().v();
   R t0=ko().t();
   X3<R,Iv,Iv> os=ko().orbSize();

   Pp<KepOsc> kt=ko; // copy constructor works for Pp<>

   kt().step_();
   R xt=kt().x();
   R vt=kt().v();
   R dx=xt-x0;
   R dv=vt-v0;
   R r=R2(dx,dv).abs();
   R rx=r*facX;
   R rv=r*facV;
   Vp<KepOsc> vk(nK);
   vk.set(ko());
   R phi=0;
   R dphi=cpm2pi/(nK-1);
   R_Vector xx(nK);
   R_Vector vv(nK);

   Z i,j;
   for (i=vk.b();i<=vk.e();++i){
      vk(i).shf_(rx*cpmcos(phi),rv*cpmsin(phi));
      xx[i]=vk(i).x();
      vv[i]=vk(i).v();
      phi+=dphi;
   }
   R_Vector xx0=xx;
   R_Vector vv0=vv;

   R rf=r*fac;
   Iv ivx(-rf,rf);
   Iv ivy=ivx;
   Graph g(cb);
   g.setXY(ivx,ivy);
   R nInv=1./nTot;
   Z nPrev=nTot-1;
   Z imgCounter=0;
   CyclicAlarm cy(displayPeriod);
   R_Vector areas(nTot);
   for (i=1;i<=nTot;++i){
      for (j=vk.b();j<=vk.e();++j){
         vk(j).step_();
         R tAct=vk(j).t();
         KepExt koAct(vk(j));
         R2 xvj=koAct.xv(t0,acc); // time back to time of initial config.
         xx[j]=xvj.x1;
         vv[j]=xvj.x2;
      }
      R_Vector xxs=xx.subMean(); // subtracting the mean
      R_Vector vvs=vv.subMean();
      areas[i]=area(xxs,vvs); // CpmArrays::area(R_Vector,R_Vector)
         // gives the area enclosed by a polygon
      if (i<nPrev){
         if (cy()){
            g.draw(xxs,vvs,ko().col());
            g.setText("frame "&cpm(imgCounter++),0.05,0,ko().col());
            bool fileAct=false;
            if (fileOnShow) fileAct=imgInterval.hasElm(imgCounter);
            g.vis(fileAct);
            cpmprogress("loop",i*nInv);
            cpmwait(tWaitFrame,2);
            g.clr_(cb);
         }
      }
      else if (i==nPrev) g.clr_(cb); // that only a single
         // drawing is on the final display
      else{
         R redFac=0.5; // color of initial configuration should
            // be a bit less bright
         Color ci=ko().col()*redFac;
         g.draw(xxs,vvs,ko().col());
         g.draw(xx0.subMean(),vv0.subMean(),ci); // the final
         // image should also show the initial configuration
         g.setText("frame "&cpm(imgCounter++),0.05,0,ko().col());
         g.vis(fileOnShow); // visualize (display)
         cpmprogress("loop",i*nInv);
      }
   }
   g.wrt(imageName); // does nothing if file name is void
   cpmwait(tWaitFinal1,2);
   if (tWaitFinal2>0){ // shows how the enclosed area evolves over time
      Graph gg(GREEN);
      gg.setWithOrigin(0);
      gg.show(areas,WHITE);
      gg.addText("Time evolution of enclosed area");
      gg.vis(fileOnShow);
      cpmwait(tWaitFinal2,2);
   }
}

R cpmlog10s(R x)
{
   R y=cpmabs(x);
   if (y<1e-20) return -20;
   else return cpmlog10(y);
}

void app4(void)
// Analysis for order. Normal order turns out to be 2.
// However, if order is inferred from a run for which nPerRev is integer
// and nTot is an integer multiple of nPerRev one gets order 4 (and 2.87
// for method RK2). If one changes nTot only by a single step one falls
// to the normal value 2 of the order. This resonance-like behavior of
// order here observed in the case of a strictly periodic motion is not
// yet understood. Since a numerical method can be interpreted as a
// perturbation of the exact method, the KAM theory applies if the orbits
// of the exact method are periodic, just as in our case.
// Notice that so far the error was only recorded at the end of the
// trajectory. The new control 'endOnly' allows to form a error sum over
// the whole trajectory. This never shows orders significantly different
// from 2.

{
   Word loc("app4");
   RecordHandler rch("app4.ini");
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, tWaitFinal=1, nPerRev=16, acc=1e-8;
   Z i,j, nTot=16, nDoubling=2, method=2;
   B endOnly;
 // getting quantities from section 'data' of file app4.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(endOnly);
   cpmrhf(stepRed_);
   cpmrhf(acc);
   cpmrh(tWaitFinal);
   cpmrh(method);
   cpmrhf(cpmverbose);
   Z nData=nDoubling+1;
   R_Vector x(nData);
   R_Vector y(nData);
   Z sp=1; // The sampling period is taken as the number of steps in the
      // first run. Doubling the step number will  n o t  increase the
      // number of samples per revolution. Notice that for endOnly = 1
      // only the error at the end-point of the trajectory is taken into
      // account.
   Word wm; // word for method
   KepExt koExt(eps);
   for (i=1;i<=nData;++i){
      Pp<KepOsc> ko=singlet(method,eps,nPerRev);
      if (i==1) wm=ko().met();
      CyclicAlarm cy(sp);
      R ei=0;
      for (j=1;j<=nTot;++j){
         ko().step_();
         if (endOnly ? j==nTot : cy()){
            R ts=ko().t();
            R xExt=koExt.xv(ts,acc).x1;
            R xs=ko().x();
            ei+=cpmabs(xExt-xs);
         }
      }
      nTot*=2; // implementing halfing the time-step
      nPerRev*=2.0; // implementing halfing the time-step
      sp*=2; // this lets the total number of sampling points constant.
      x[i]=cpmlog10(ko().dt()); // log timestep
      y[i]=cpmlog10s(ei);
         // log absolute position error at tTot or errorsum at more
         // points
      cpmdata<<"x["<<i<<"] = "<<x[i]<<" y["<<i<<"] = "<<y[i]<<endl;
   }
   Z deg=1; // find a straight line
   R_Vector res=polyFit(x,y,deg,0.,1.,false);
   R order=cpmabs(res[4]);
   Graph g;
   g.setWithOrigin(0);
   g.addText("order="&cpm(order)&", method="&wm);
   g.show(x,y);
   cpmwait(tWaitFinal,2);
}

void app5(void)
// Collapsing error curves for different stepsizes by scaling them with
// an input-value for the order.
{
   Word loc("app5");
   RecordHandler rch("app5.ini");
      // This says that we expect data input to be available in file
      // app5.ini in a diretory which is determined by entries in
      // cpmsystemdependencies.h and cpmconfig.ini.
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, tWaitFinal=1, nPerRev=16, order=2;
   Z i,j, nTot=16, nDoubling=2, method=2;
// getting quantities from section 'data' of file app5.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(stepRed_);
   cpmrh(tWaitFinal);
   cpmrhf(order);
   cpmrh(method);
   Z nData=nDoubling+1;
   V< V<R2> > res(nData);
   KepExt koExt(eps);
   R fac=1.;
   Word wm;
   for (i=1;i<=nData;++i){
      Pp<KepOsc> ko=singlet(method,eps,nPerRev);
      if (i==1) wm=ko().met();
      V<R2> resi(nTot);
      for (j=1;j<=nTot;++j){
         ko().step_();
         R tj=ko().t();
         R2 dpos=ko().phsPos()-koExt.xv(tj);
         resi[j]=R2(tj,dpos.x1*fac);
      }
      res[i]=resi;
      nTot*=2;
      nPerRev*=2.0;
      fac*=cpmpow(2.0,order);
   }
   Graph g;
   g.setWithOrigin(0);
   g.addText("method="&wm);
   g.mark(res);
   g.vis();
   cpmwait(tWaitFinal,2);
}

void app6()
// this is a place for playing arround with graphical functions
{
    Frame f;
    f.paint(RED);
    f.addText("This is a trivial experiment");
    f.vis();
    cpmwait(8,2);
    cpmmessage("app6 done");
}

void app7(void)
// Showing the log of the end-of-trajectory error as a dunction of
// log of step size for all the integration methods.
{
   Word loc("app7");
   RecordHandler rch("app7.ini");
   Word sec="data";
   Z metNum=6; // number of integration methods
   R E0=0.,eps=0.25, tWaitFinal=1, nPerRev=16, acc=1e-8;
   Z i,j, nTot=16, nDoubling=2;
   cpmrh(E0);
   cpmrh(eps);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(stepRed_);
   cpmrhf(acc);
   cpmrh(tWaitFinal);
   cpmrhf(cpmverbose);
   Z nData=nDoubling+1;
   R_Vector x(nData);
   R_Matrix y(metNum,nData);
   R prog=0;
   cpmprogress("loop",prog,3);
   R dProg=1./metNum;
   R t0=0.;
   KepOsc koExt(eps,nPerRev,E0,t0);
   R tTot=koExt.dt()*nTot;
   koExt.extEvl_(tTot,acc);
   R2 pfExt=koExt.phsPos();
   for (Z method=0;method<metNum;++method){
      R nPerRevVar=nPerRev;
      Z nTotVar=nTot;
      for (i=1;i<=nData;++i){
         Pp<KepOsc> ko=singlet(method,eps,nPerRevVar,E0);
         for (j=1;j<=nTotVar;++j) ko().step_();
         R2 pf=ko().phsPos();
         nTotVar*=2; // implementing halfing the time-step
         nPerRevVar*=2.0; // implementing halfing the time-step
         x[i]=cpmlog10(ko().dt()); // log timestep
         y[method+1][i]=cpmlog10s(cpmabs(pf.x1-pfExt.x1));
      }
      prog+=dProg;
      cpmprogress("loop",prog,3);
   }
   Graph g;
   g.setWithOrigin(0);
   Word txt("log10 of error (y-axis) vs. log10 of time step ");
   txt&="(x-axis) for 6 integrators";
   g.addText(txt);
   g.setWriteOnShow(true);
   g.show(x,y,V<Color>{BLACK,YELLOW,GREEN,RED,CYAN,MAGENTA});
   cpmwait(tWaitFinal,2);
}

// For n=2000 t = 10.48 with Eigen and t = 57.35 without.
// Thus Eigen accelerates by a factor of 5.5 which should not be ignored.
void app8(void){
   Z n=50;
   R t1 = cpmtime();
   R_Matrix a(n,n,1.);
   R_Matrix x1 = a.ran(1);
   R_Matrix x2 = a.ran(2);
   R_Matrix x3 = a.ran(3);
   R_Matrix r1 = x1*(x2*x3);
   R_Matrix r2 = (x1*x2)*x3;
   R dis = r1.dis(r2);
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"R, n = "<<n<<", dis = "<<dis<<", execTime = "<<t<<endl;
}

void app8c(void){
   Z n=50;
   R t1 = cpmtime();
   C_Matrix a(n,n,C(1.));
   C_Matrix x1 = a.ran(1);
   C_Matrix x2 = a.ran(2);
   C_Matrix x3 = a.ran(3);
   C_Matrix r1 = x1*(x2*x3);
   C_Matrix r2 = (x1*x2)*x3;
   R dis = r1.dis(r2);
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"C, n = "<<n<<", dis = "<<dis<<", execTime = "<<t<<endl;
}

void app9(void){
   Z n=500;
   R t1 = cpmtime();
   R_Matrix a(n,n,1.);
   R_Matrix x1 = a.ran(1);
   SVD svd(x1);
   R err = svd.getError();
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"R, n = "<<n<<", err = "<<err<<", execTime = "<<t<<endl;
}

void app9c(void){
   Z n=353;
   R t1 = cpmtime();
   C_Matrix a(n,n,C(1.));
   C_Matrix x1 = a.ran(1);
   SVDC svd(x1);
   R err = svd.getError();
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"C,  n = "<<n<<", err = "<<err<<", execTime = "<<t<<endl;
}

void app11(void){
   Z n=200;
   R t1 = cpmtime();
   R z0;
   R z1=Root<R>(z0).test(10);
   R z2=Root<R>(z1).ran(1);
   cout<<endl<<" z2 = "<<z2<<endl;
   R_Matrix a(n,n,z2);
   R_Matrix x1 = a.ran(1);
   R_Matrix x2 = a.ran(2);
   R_Matrix x3 = x1*x2;
   SpcObjR spo(x3,true);
   R err = spo.getError();
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"R,  n = "<<n<<", err = "<<err<<", execTime = "<<t<<endl;
}

void app11c(void){
   Z n=200;
   R t1 = cpmtime();
   C z0;
   C z1=z0.test(10);
   C z2=z1.ran(1);
   cout<<endl<<" z2 = "<<z2<<endl;
   C_Matrix a(n,n,z2);
   C_Matrix x1 = a.ran(1);
   C_Matrix x2 = a.ran(2);
  // C_Matrix x3 = x1*x2;
  C_Matrix x3(n,n) ;

   SpcObj spo(x3,true);
   R err = spo.getError();
   R t2 = cpmtime();
   R t = t2-t1;
   cout<<"C,  n = "<<n<<", err = "<<err<<", execTime = "<<t<<endl;
}

void app12(void){
}

} // namespace CpmKepler

void info()
{
   cout<<"Studying the Kepler oscillator"<<endl
   <<" This function takes 0 to 1 integer arguments"<<endl
   <<" values 1 2 3 4 5 6 7"<<endl
   <<" trigger one of the five basic modes, each other number"<<endl
   <<" triggers execution of these modes in succession."<<endl
   <<" No argument is equivalent to an argument of value 1.";
}

/**************************
Z CpmApplication::main_(void)
// Selects the task from the command line, if there is a numerical
// argument following the function name. Further, the program tries to
// find and read tut2.ini. Here the filename is the name of the executable
// with all capital letters removed (so tut2MS.exe would als search for
// tut2.ini).
// Places to search for this are determined by cpmsystemdependencies.h
// and cpmconfig.ini, where the latter enjoys priority.
// If no tut2.ini can be found, the coded default selection takes
// place.
// The functionality of the tasks app1(), app2(), app3(), app4(), app5(),
// app7() is controlled through files ap1.ini, app2.ini, app3.ini,
// app4.ini, app5.ini, app7.ini for which search places are determined by
// the same logic as those of tut2.ini.
{
   using namespace CpmKepler;
   title_="tut2";
      // No longer determining the name of the ini-file
   Z mL=1;
   Z sel=0; // default selection
      // here all
   cpmverbose=1;
   if (args_.valInd(2)){ // arrays allow testing the validity
         // of indexes without running into exceptions
      Word w2=args_[2];
      if (w2=="?"){
         info();
         return 0;
      }
      else sel=w2.toZ(); // conversion to Z (integer)
   }
   Word loc("CpmApplication::main_(), sel="&cpm(sel));
   CPM_MA

   if (sel==1) app1();
   else if (sel==2) app2();
   else if (sel==3) app3();
   else if (sel==4) app4();
   else if (sel==5) app5();
   else if (sel==6) app6(); // for trivial experiments with graphics
   else if (sel==7) app7();
   else { // do everything sequentially
      app1();
      app2();
      app3();
      app4();
      app5();
      app6();
      app7();
   }
   CPM_MZ
   return sel;
}
****************************************/

Z CpmApplication::main_(void)
// Selects the task from the command line, if there is a numerical
// argument following the function name. Further, the program tries to
// find and read tut2.ini. Here the filename is the name of the executable
// with all capital letters removed (so tut2MS.exe would als search for
// tut2.ini).
// Places to search for this are determined by cpmsystemdependencies.h
// and cpmconfig.ini, where the latter enjoys priority.
// If no tut2.ini can be found, the coded default selection takes
// place.
// The functionality of the tasks app1(), app2(), app3(), app4(), app5(),
// app7() is controlled through files ap1.ini, app2.ini, app3.ini,
// app4.ini, app5.ini, app7.ini for which search places are determined by
// the same logic as those of tut2.ini.
{
   using namespace CpmKepler;
   title_="tut2";
      // No longer determining the name of the ini-file
   Z mL=1;
   Z sel=1; // default selection
      // here test on matrix product
   cpmverbose=1;
   if (args_.valInd(2)){ // arrays allow testing the validity
         // of indexes without running into exceptions
      Word w2=args_[2];
      if (w2=="?"){
         info();
         return 0;
      }
      else sel=w2.toZ(); // conversion to Z (integer)
   }
   Word loc("CpmApplication::main_(), sel="&cpm(sel));
   CPM_MA

   if (sel==1) app1();
   else if (sel==2) app2();
   else if (sel==3) app3();
   else if (sel==4) app4();
   else if (sel==5) app5();
   else if (sel==6) app6(); // for trivial experiments with graphics
   else if (sel==7) app7();
   else if (sel==8) {app8(); app8c();}
   else if (sel==9) {app9(); app9c();}
   else if (sel==11) {app11(); app11c();}
   else if (sel==12) app12();
   else { // do everything sequentially
      app1();
      app2();
      app3();
      app4();
      app5();
      app6();
      app7();
   }
   cpmmessage("main_() returns the selection parameter sel = "&cpm(sel)&
   "\n so, a value different from 0 needs not to signal a problem");
   CPM_MZ
   return sel;
}

// end of tut2.cpp
