//? tut2bu.cpp
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


/*************************************************************************
   Studying the Kepler oscillator
*********************************************************************/
#include <iomanip> // for setprecision(Z)
#include <cpmbas1.h>
   // Actually, this is a collecion of header files.
#include <cpmvectormore.h>
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;

using CpmLinAlg::R2;
using CpmGeo::Iv;
using CpmAlgorithms::solKepEqu;
using CpmAlgorithms::CyclicAlarm;
using CpmImaging::YELLOW;
using CpmImaging::GREEN;
using CpmImaging::CYAN;
using CpmImaging::MAGENTA;
using CpmImaging::darkSky;
using CpmImaging::twiSky;
using CpmImaging::morSky;

namespace CpmKepler{

bool stepRed_=false; // step reduction
   // Has no influence for integration method ADALF
R facALF=1.0;
   // An experimental method parameter.
   // Normal value is 1.0

class KepOsc{ // 'Kepler oscillator'
// My name for the system defined as the radial part of the Kepler
// motion arround a fixed central mass. Physical units are used in which
// the moving mass, the gravitational potential at unit distance, and the
// angular momentum have value 1.
// see www.ulrichmutze.de/articles/99-271.pdf, Section 4

// Methodologic remarks:
// The primary aim here is to compare various integrators for the
// equation of motion for the physical system mentioned above.
// There are two strategies for implementing this aim:
// 1. Define for each integration method a separate class.
//    This is the normal way to be followed according to the pure doctrine
//    of object oiented programming. It is also very natural since the
//    various integration methods to be considered need different state
//    data for implementing an integration step as a mapping of the
//    state space into itself.
// 2. Define each integration method as a separate member function
//    within a single class. This requires the state space to contain
//    data which are needed for implementing e.g. the leap-frog integrator
//    and are not needed e.g. for the asynchronous leap-frog integrator.
//    If data belong to a type for which the default constructor is
//    computationally inexpensive such partially relevant data seem to
//    be acceptable and the present class implementation follows this
//    path.

// data
// 1. common data
   R t_,x_,v_,dt_; // time, position, velocity, time step
   Z met_; // method selector for functon step_
// 2. additional data for implementing stepALF_(R dt) and stepADALF_(R dt)
   R1 w_,a_;
// 3. additional data for implementing stepLF_(R dt)
   R tp_,xp_,vp_; // p for previous

// static functions
   static R f(R const& x){R xi=cpminv(x); return (xi-1)*xi*xi;}
      // force law
// member functions
   void init_(Z met, R eps, R nPerRev, R E);
      // assumes 0<=eps<1, E arbitrary
      // initializes the common data
   void init2_(){ w_=v_; a_=f(x_);}
      // does the additional initializations for met_=2,4,5
   void init3_()
      // does the additional initializations for met_=3
      // simplified version, we need not be more accurate than second
      // order
   {
      R h= -dt_;
      if (stepRed_) h*=0.5;
      tp_=t_+h;
      R a=f(x_);
      R dv=a*h*0.5;
      xp_=x_+(v_+dv)*h; // here xp_ is previous x_
      vp_=v_+(a+f(xp_))*h*0.5;
   }

   void initAux_(){
      if (met_==2 || met_==4 || met_==5) init2_();
      if (met_==3) init3_();
   }
      // does the additional initializations for met_=2 and met_=3

   void stepDMI_(R dt); // direct midpoint integrator (2nd order)
      // needs only the common data
   void stepRK_(R dt); // Runge-Kutta (2nd order)
      // needs only the common data
   void stepALF_(R dt); // asynchronous leap-frog (2nd order)
      // needs data w_, a_ for implementation
      // See 'Fortsetzung von UM 10.9.08' in my notebook UM 2008a
   void stepADALF_(R dt); // asynchronous leap-frog (2nd order)
      // averaged densified asynchronous leapfrog
   void stepLF_(R dt); // leap-frog (2nd order)
      // needs tp_,xp_,vp_
   void stepAALF_(R dt); // averaged asynchronous leap-frog (2nd order)
      // not densified! New, change 2013/14.
      // This is an inferior method that I deviced and tested at the
      // very beginning of investigation of second order integrators
      // for the time-dependend Schroedinger equation. I discarded the
      // method and forgot it. Finally I re-invented it as the simplest
      // version of the averaging method.
public:
   Word nameOf()const{ return "KepOsc, mehod="&cpm(met_);}
   KepOsc(){init_(0,0,32,0);}
   explicit KepOsc(Z met, R eps, R nPerRev=64, R E=0)
      {init_(met,eps,nPerRev,E);}
   R t()const{ return t_;}
   R dt()const{ return dt_;}
   R x()const{ return x_;}
   R v()const{ return v_;}
   R eKin()const{ return v_*v_*0.5;}
      // kinetic energy
   R ePot()const{ R xi=cpminv(x_); return xi*(xi*0.5-1);}
      // potential energy
   R eTot()const{ return eKin()+ePot();}
      //: e total
   R force()const{ return f(x_);}
      //: force
      // The superficially correct name 'for' is not correct since it is
      // a C++ key word.
   R_Vector orbDes()const;
      //: orbit descriptors
      // Returns a list of the usual orbit elements
      // a: major semi-axis, eps: numerical excentricity,
      // n: mean motion, E: excentric anomaly, M: mean anomaly,
      // tP: orbital period (revolution time)
   void rev_(){ v_=-v_;}
      //: reversed
      // Notice name ending in '_' since it is a non-constant function
   void extEvl_(R t, R acc=1e-8);
      //: exact evolution

   void step_()
      //: step
      // The actual step algorithm depends on method parameters met_:
      // met_==0 Runge-Kutta 2. order
      // met_==1 direct midpoint integrator
      // met_==2 asynchronous leapfrog integrator
      // met_==3 leapfrog integrator
      // met_==4 averaged densified asynchronous leapfrog integrator
      // met_==5 averaged asynchronous leapfrog integrator (new, only for
                 // completeness, not an acceptable method )
   {
      if (met_==0) stepRK_(dt_);
      else if (met_==1) stepDMI_(dt_);
      else if (met_==2){
         if (stepRed_){
            R tau=dt_*0.5;
            stepALF_(tau);
            stepALF_(tau);
         }
         else stepALF_(dt_);
      }
      else if (met_==3){
         if (stepRed_){
            R tau=dt_*0.5;
            stepLF_(tau);
            stepLF_(tau);
         }
         else stepLF_(dt_);
      }
      else if (met_==4){
            stepADALF_(dt_);
      }
      else if (met_==5){
         if (stepRed_){
            R tau=dt_*0.5;
            stepAALF_(tau);
            stepAALF_(tau);
         }
         else stepAALF_(dt_);
      }
      else cpmerror("bad value of KepOsc::met_");
   }

   void shf_(R dx, R dv)
      //: shift
   {
      x_+=dx; v_+=dv;
      initAux_();
   }

   X3<R,Iv,Iv> orbSize()const;
      //: orbit size
      // intervals which contain the t_-range, x_-range , the y_range
      // for a full revolution

   Color col()const
      //: color for visualization
   {
      if (met_==0) return WHITE; // RK
      else if (met_==1) return YELLOW; // DMI
      else if (met_==2) return GREEN; // ALF
      else if (met_==3) return RED; // LF
      else if (met_==4) return CYAN; // ADALF
      else if (met_==5) return MAGENTA; // AALF
      else cpmerror("bad value of KepOsc::met_ in KepOsc::col()");
   }

   void mark(Graph& g)const{ g.mark(R2(x_,v_),col());}
      //: mark

   R2 phsPos()const{ return R2(x_,v_);}
      //: phase (space) position

   R per()const{ return orbDes()[6];}
      //:period

   Word met()const
   {
      if (met_==0) return "RK2";
      if (met_==1) return "DMI";
      if (met_==2) return "ALF";
      if (met_==3) return "LF";
      if (met_==4) return "ADALF";
      if (met_==5) return "AALF";
      else return "unvalid method";
   }
 };

void KepOsc::init_(Z met, R eps, R nPerRev, R E)
// initializing x_ anv v_ from numerical excentricity and excentric
// anomaly. Further, initialization of the dependent quantities.
{
   Word loc("KepOsc::init_(R,R)");
   cpmassert(eps>=0,loc);
   cpmassert(eps<1,loc);
   met_=met;
   t_=0;
   R ai=(1-eps)*(1+eps);
   R a=1./ai;
   x_=a*(1-eps*cpmcos(E));
   v_=eps*cpmsqrt(a)*cpmsin(E)/x_;
   R tP=cpm2pi*cpmpow(a,1.5);
   dt_=tP/nPerRev;
   initAux_();
}

R_Vector KepOsc::orbDes()const
{
   Word loc("KepOsc::orbDes()");
   Z mL=2;
   CPM_MA
   R e=eTot();
   if (e>=0){
      cpmcerr<<"We have e>=0, un-bound state. x_="<<x_<<" v_="<<v_<<endl;
      cpmdebug(ePot());
      cpmdebug(eKin());
   }
   cpmassert(e<0,loc); // we need a bound state
      // then also x>0.5
   R a=-0.5*cpminv(e);
   R eps=cpmsqrt(1.-1./a);
   R n=cpmpow(a,-1.5);
   R tP=cpm2pi/n;
   R z1=1-x_/a;
   R z2=x_*v_/cpmsqrt(a);
   C z(z1,z2);
   R E=z.arg();
   R M=E-eps*cpmsin(E);
   CPM_MZ
   return R_Vector("",a,eps,n,E,M,tP);
}

void KepOsc::extEvl_(R t, R acc)
// exact step with t arbitrarily large, based on the
// famous transcendental Kepler equation
{
   Word loc("KepOsc::extEvl_(R,R)");
   Z mL=2;
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],M=y[5];
   M+=n*t;
   R E=solKepEqu(M,eps,acc);
   x_=a*(1-eps*cpmcos(E));
   cpmassert(x_>=0.5,loc); // for safty, not needed
   v_=eps*a*a*n*cpmsin(E)/x_;
   t_+=t;
   initAux_();
   CPM_MZ
}

void KepOsc::stepDMI_(R dt)
// nice symmetry, f evaluated only once
{
   R tau=dt*0.5;
   t_+=tau;
   x_+=(v_*tau);
   v_+=f(x_)*dt;
   x_+=(v_*tau);
   t_+=tau;
}

void KepOsc::stepRK_(R dt)
// not so nice, involves two evaluations of f
{
   R tau=dt*0.5;
   R a=f(x_);
   R am=f(x_+v_*tau);
   R dv=am*dt;
   R dx=(v_+a*tau)*dt;
   x_+=dx;
   v_+=dv;
   t_+=dt;
}

void KepOsc::stepALF_(R dt)
// asynchronous leap-frog method with a factor facALF that normally
// should be 1.
{
   R tau=dt*0.5;
   R fac=2*facALF;
   t_ += tau;
   x_ += tau*w_;
   v_ += tau*a_;

   w_ += fac*(v_-w_);
   a_ += fac*(f(x_)-a_);

   v_ += tau*a_;
   x_ += tau*w_;
   t_ += tau;
}

void KepOsc::stepAALF_(R dt)
// averaged asynchronous leap-frog method
{
   R tau=dt*0.5;

   R xm=x_;
   R vm=v_;

   t_ += tau;
   x_ += tau*w_;
   v_ += tau*a_;

   w_ = v_;
   a_ = f(x_);

   v_ = vm + dt*a_;
   x_ = xm + dt*w_;
   t_ += tau;
}

void KepOsc::stepADALF_(R dt)
// averaged densified leap-frog method
{
   R tau=dt*0.25;
   R tau2=dt*0.5;
   t_ += tau;
   x_ += tau*w_;
   v_ += tau*a_;

   w_ = v_*2.-w_;
   a_ = f(x_)*2.-a_;

   R wm=w_;
   R am=a_;

   t_ += tau2;
   x_ += tau2*w_;
   v_ += tau2*a_;

   w_ = v_*2.-w_;
   a_ = f(x_)*2.-a_;

   v_ += tau*a_;
   x_ += tau*w_;
   t_ += tau;

   a_=(a_+am)*0.5;
   w_=(w_+wm)*0.5;

}

void KepOsc::stepLF_(R dt)
{
      R dt2=dt*2;
      R tn=tp_+dt2;
      R vn=vp_+f(x_)*dt2;
      R xn=xp_+v_*dt2;
      tp_=t_;
      vp_=v_;
      xp_=x_;
      t_=tn;
      v_=vn;
      x_=xn;
}

X3<R,Iv,Iv> KepOsc::orbSize()const
{
   Z mL=2;
   Word loc("KepOsc::orbSize()");
   CPM_MA
   R_Vector y=orbDes();
   R a=y[1],eps=y[2],n=y[3],tP=y[6];
   R xMin=a*(1-eps);
   R xMax=a*(1+eps);
   R vMax=eps*a*a*n;
   CPM_MZ
   return X3<R,Iv,Iv>(tP,Iv(xMin,xMax),Iv(-vMax,vMax));
}

//////////////////////////////////////////////////////////////////////////

void cpmplot(V<V<R2> > const& dat, Word const& fileName)
{
   Word loc("cpmplot(V<V<R2>>,Word)");
   Z mL=2;
   CPM_MA
   if (fileName.isVoid()){
      CPM_MZ
      return;
   }
   Z i,j,md=dat.dim();
   V<Z> vd(md);
   for (i=1;i<=md;++i){
      vd[i]=dat[i].dim();
   }
   S<Z> sd(vd);
   cpmassert(sd.car()==1,loc);
   Z nd=sd[1];
   OFileStream ofs(fileName);
   ofs()<<std::setprecision(CpmRoot::wrtPrc);
      // otherwise multiple precision mode would not be flexibly be used
   for (i=1;i<=nd;++i){
      ofs()<<endl;
      for (j=1;j<=md;++j){
         R2 y=dat[j][i];
         ofs()<<y[1]<<" "<<y[2]<<"    ";
      }
   }
   CPM_MZ
}

void app1(void) // orbit in phase space (v = momentum since m=1)
// simultaneously for a selected list of integration methods
{
   Word loc("app1");
   Color cb=darkSky; // night sky blue. Each task appi
   // has here its characteristic background color
   RecordHandler rch("app1.ini");
      // This says that we expect data input to be available in file
      // app1.ini in a diretory which is determined by entries in
      // cpmsystemdependencies.h and cpmconfig.ini.
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, fac=1.2, nPerRev=16, tWaitFinal=1;
   Z nTot=1000,displayPeriod=1;
   Word imageName,plotDataName;
   V<Z> methods;
// getting quantities from section 'data' of file app1.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrhf(stepRed_);
   cpmrh(fac);
   cpmrh(tWaitFinal);
   cpmrhf(imageName);
   cpmrhf(plotDataName);
   cpmrh(displayPeriod);
      // now the program has all its data
   cpmrh(methods);
   cpmrhf(facALF);
   S<Z> met(methods); // eliminates doublicates
   Z nm=met.dim();
   V<KepOsc> vk(nm);
   for (Z i=1;i<=nm;++i){
      vk[i]=KepOsc(met[i],eps,nPerRev);
   }
   X3<R,Iv,Iv> os=vk[1].orbSize(); // cartesian product templates
      // are defined from X2 to X8
   Iv ivx=os.c2(); // Iv = interval is a valuable type
   cpmdebug(ivx);
   Iv ivy=os.c3();
   cpmdebug(ivy);
   ivx*=fac; // changing the size of an interval, while holding the center
      // fixed
   ivy*=fac; // same factor for the other direction
   V<V<R2> > vvr(nm,V<R2>(nTot)); // memory for all phase space positions
      // during run
   Graph g(cb); // basic graphical 'window to the world'
   g.setXY(ivx,ivy);
      // this associates physical sizes (in space) with the
      // 'pixeled' rectangle g
   R nInv=1./nTot; // for use in cpmprogress
   CyclicAlarm cy(displayPeriod); // device for triggering periodic
      // actions
   for (Z i=1;i<=nTot;++i){
      for (Z j=vk.b();j<=vk.e();++j){
         vk[j].mark(g); // 'mark' acts on memory only
         R2 pj=vk[j].phsPos();
         if (i==1) cpmdebug(pj);
         vvr[j][i]=pj;
         vk[j].step_();
      }
      if (cy()||i==nTot){ // cy() becomes 'true' periodically
         g.vis(); // transfers the whole memory frame
            // to screen. This expensive action should probably not be
            // orderd for each integration step. Object cy allows
            // to achieve this.
         cpmprogress("loop",i*nInv); // progress indicator on
            // pane 1 of he status bar
      }
   }
   g.wrt(imageName); // writes the final frame to a ppm image file
      // if imageName is not the void word.
   cpmplot(vvr,Message::extOut(plotDataName)); // does nothing if the name
   // is void
   cpmwait(tWaitFinal,2); // allows us to inspect the last frame
      // for the time (in seconds) put in here. The second argument
      // says that on pane 2 of the status bar there will be a
      // 'countdown' indication of the time left for inspection.
}

void app2(void) // method error of the two integration methods
// represented by means of interaction picture dynamics. For
// applying this terminology, we interpret the dynamics as defined by
// the numerical methods as resulting from the exact Kepler dynamics
// by 'interaction'. This interaction picture motion is much slower for
// the more accurate direct midpoint method. The Runge Kutta motion
// finally explodes and reaches non-bound states for which the exact
// evolution back is no longer defined (by the method KepOsc::extEvl_).
// Therefore the option is implemented to deal only with the stable
// direct midpoint method.
{
   Word loc("app2");
   Color cb=twiSky; // background
   RecordHandler rch("app2.ini");
   Word sec="data";
   R eps=0.25, nPerRev=16, acc=1e-8, tWaitFinal=1;
   Z i,j,nTot=1000,nIter=4;
   Word imageName,plotDataName;
   V<Z> methods;
   bool backEvl=true;
   cpmrh(eps);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(acc);
   cpmrh(tWaitFinal);
   cpmrh(imageName);
   cpmrhf(plotDataName);
   cpmrh(methods);
   cpmrhf(backEvl);
   cpmrhf(stepRed_);
   cpmrhf(facALF);

   S<Z> met(methods); // eliminates doublicates
   Z nm=met.dim();
   cpmassert(nm>0,"app2()");
   V<KepOsc> vk(nm);
   for (Z i=1;i<=nm;++i){
      vk[i]=KepOsc(met[i],eps,nPerRev);
   }
   KepOsc k0=vk[1];
   // This defines the initial state.
   X3<R,Iv,Iv> os=k0.orbSize();
   R2 p0=k0.phsPos();
   R xInv=cpminv(os.c2().abs());
   R vInv=cpminv(os.c3().abs());

   V<V<R2> > vvr(nm,V<R2>(nTot));
   R nInv=1./nTot;
   for (i=1;i<=nTot;++i){ // no graphical actions
      // within the loop. Terribly fast!!!
      KepOsc k1Act=vk[1];
      R tAct=k1Act.t(); // the states are aware of the time to which they
         // belong: Schroedinger picture
      if (!backEvl){
         KepOsc kExt(k0);
         kExt.extEvl_(tAct,acc); // exact evolution
         R2 posExt=kExt.phsPos(); // storing the exact phase space
         // position
         for (j=vk.b();j<=vk.e();++j){
            R2 diff=vk[j].phsPos()-posExt;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      else{
         for (j=vk.b();j<=vk.e();++j){
            KepOsc vkj(vk[j]);
            vkj.rev_();
            vkj.extEvl_(tAct,acc);
            R2 diff=vkj.phsPos()-p0;
            diff[1]*=xInv;
            diff[2]*=vInv;
            vvr[j][i]=diff;
         }
      }
      for (j=vk.b();j<=vk.e();++j){
         vk[j].step_();
      }
      cpmprogress("loop",i*nInv,1,2); // only 2 digits for
         // progress indication to avoid 'much adoo about nothing'
   }
   Graph g(cb);
   g.clr_(cb);
   g.setAutoScale(true);
   g.setNoGridText(false);
   g.setNoGrid(true);
   g.setGridTextColor(LIGHTBLUE);
   V<Color> vc(nm);
   for (j=1;j<=nm;++j) vc[j]=vk[j].col();
   g.mark(vvr,vc); // powerful method for graphical
      // representation of polygons; here one or two
      // of those
   g.vis();
   g.wrt(imageName); // does nothing if file name void
   cpmplot(vvr,plotDataName);
   cpmwait(tWaitFinal,2);
}

void app3(void) // interaction picture dynamics for not a single
// point as an initial condition but for an ensemble of
// phase space points which form a closed curve. The deformation
// of this closed curve by interaction picture dynamics is monitored.
// Here the field of view in phase space follows the moving states.
// Otherwise they would slowly drift out. For the asynchronous
// leapfrog method (method 2) it was observed that the loop
// deformation can lead to crossing over. This puzzled me for a while
// since this cannot happen with mappings of the phase space into
// the phase space. However for the asynchronous leapfrog method
// the statespace (which is propery mapped) has additional degrees
// of freedom, which are not shown in the phase space curves. This
// resolves what seemed to be a logical contradiction. The details of
// the process are not clear to me. Finally the time evolution of the
// enclosed phase area is shown.
{
   Word loc("app3");
   Color cb=morSky; // background
   RecordHandler rch("app3.ini");
   Word sec="data";
   R eps=0.25, fac=1.2, nPerRev=16, acc=1e-8, E=0;
   R facX=1, facV=1, tWaitFinal1=1, tWaitFinal2=1, tWaitFrame=0.1;
   Z nTot=1000, nK=12, displayPeriod=1, method=1;
   Word imageName;
   bool fileOnShow=false;
   V<Z> filingLimits;

   cpmrh(eps);
   cpmrh(E);
   cpmrh(fac);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(acc);
   cpmrh(tWaitFinal1);
   cpmrh(tWaitFinal2);
   cpmrh(tWaitFrame);
   cpmrh(nK);
   cpmrh(facX);
   cpmrh(facV);
   cpmrh(displayPeriod);
   cpmrh(method);
   cpmrhf(imageName);
   cpmrhf(stepRed_);
   cpmrhf(fileOnShow);
   cpmrhf(filingLimits);

   IvZ imgInterval; // void
   if (filingLimits.dim()>=2){
      imgInterval=IvZ(filingLimits[1],filingLimits[2]);
   }
   // It turned out to be necessary for being able to document the
   // phenomena that occur in runs to write screen images to file.
   // Since filing an image takes time, it is desirable to have a way
   // to file only a controlled subset of all screen frames.
   // Letting the program run without any filing one can make notes
   // of the frame numbers that contain the interesting events
   // (such as crossing over of loops). In a second run one then may
   // write these frames to file by suitable setting filingLimits.

   KepOsc ko(method,eps,nPerRev,E);
   R x0=ko.x();
   R v0=ko.v();
   X3<R,Iv,Iv> os=ko.orbSize();

   KepOsc kt(ko);
   kt.step_();
   R xt=kt.x();
   R vt=kt.v();
   R dx=xt-x0;
   R dv=vt-v0;
   R r=R2(dx,dv).abs();
   R rx=r*facX;
   R rv=r*facV;
   V<KepOsc> vk(nK,ko);

   R phi=0;
   R dphi=cpm2pi/(nK-1);
   R_Vector xx(nK);
   R_Vector vv(nK);

   Z i,j;
   for (i=vk.b();i<=vk.e();++i){
      vk[i].shf_(rx*cpmcos(phi),rv*cpmsin(phi));
      xx[i]=vk[i].x();
      vv[i]=vk[i].v();
      phi+=dphi;
   }
   R_Vector xx0=xx;
   R_Vector vv0=vv;

   R rf=r*fac;
   Iv ivx(-rf,rf);
   Iv ivy=ivx;
   Graph g(cb);
   g.setXY(ivx,ivy);
   R nInv=1./nTot;
   Z nPrev=nTot-1;
   Z imgCounter=0;
   CyclicAlarm cy(displayPeriod);
   R_Vector areas(nTot);
   for (i=1;i<=nTot;++i){
      for (j=vk.b();j<=vk.e();++j){
         vk[j].step_();
         R tAct=vk[j].t();
         KepOsc koAct=vk[j];
         koAct.rev_();
         koAct.extEvl_(tAct,acc);
         R xxj=koAct.x();
         R vvj=koAct.v();
         xx[j]=xxj;
         vv[j]=vvj;
      }
      R_Vector xxs=xx.subMean(); // subtracting the mean
      R_Vector vvs=vv.subMean();
      g.draw(xxs,vvs,ko.col());
      areas[i]=area(xxs,vvs); // CpmArrays::area(R_Vector,R_Vector)
         // gives the area enclosed by a polygon
      if (i<nPrev){
         if (cy()){
            g.setText("frame "&cpm(imgCounter++),0.05,0,ko.col());
            bool fileAct=false;
            if (fileOnShow) fileAct=imgInterval.hasElm(imgCounter);
            g.vis(fileAct);
            cpmprogress("loop",i*nInv);
            cpmwait(tWaitFrame);
            g.clr_(cb);
         }
      }
      else if (i==nPrev) g.clr_(cb); // that only a single
         // drawing is on the final display
      else{
         R redFac=0.5; // color of initial configuration should
            // be a bit less bright
         Color ci=ko.col()*redFac;
         g.draw(xx0.subMean(),vv0.subMean(),ci); // the final
         // image should also show the initial configuration
         g.setText("frame "&cpm(imgCounter++),0.05,0,ko.col());
         g.vis(fileOnShow); // visualize (display)
         cpmprogress("loop",i*nInv);
      }
   }
   g.wrt(imageName); // does nothing if file name is void
   cpmwait(tWaitFinal1,2);
   if (tWaitFinal2>0){ // shows how the enclosed area evolves over time
      Graph gg(GREEN);
      gg.show(areas,WHITE);
      gg.addText("Time evolution of enclosed area");
      gg.vis(fileOnShow);
      cpmwait(tWaitFinal2,2);
   }
}

R cpmlog10s(R x)
{
   R y=cpmabs(x);
   if (y<1e-20) return -20;
   else return cpmlog10(y);
}

void app4(void)
// Analysis for order. Normal order turns out to be 2.
// However, if order is inferred from a run for which nPerRev is integer
// and nTot is an integer multiple of nPerRev one gets order 4 (and 2.87
// for method RK2). If one changes nTot only by a single step one falls
// to the normal value 2 of the order. This resonance-like behavior of
// order here observed in the case of a strictly periodic motion is not
// yet understood. Since a numerical method can be interpreted as a
// perturbation of the exact method, the KAM theory applies if the orbits
// of the exact method are periodic, just as in our case.
{
   Word loc("app4");
   RecordHandler rch("app4.ini");
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, tWaitFinal=1, nPerRev=16, acc=1e-8;
   Z i,j, nTot=16, nDoubling=2, method=2;
 // getting quantities from section 'data' of file app4.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(stepRed_);
   cpmrhf(acc);
   cpmrh(tWaitFinal);
   cpmrh(method);
   cpmrhf(facALF);
   cpmrhf(cpmverbose);
   Z nData=nDoubling+1;
   R_Vector x(nData);
   R_Vector y(nData);
   KepOsc koExt(method,eps,nPerRev);
   R tTot=koExt.dt()*nTot; // nTot sets the time span to be considered
      // nTot==nPerRev for considering just one revolution.
   koExt.extEvl_(tTot,acc);
   R2 pfExt=koExt.phsPos(); // Exact phase space position
      // at the end of the trajectory under consideration.
      // Notice that the integration error will always registrated
      // only at time tTot.
   for (i=1;i<=nData;++i){
      KepOsc ko(method,eps,nPerRev);
      for (j=1;j<=nTot;++j) ko.step_();
      R2 pf=ko.phsPos();
      nTot*=2; // implementing halfing the time-step
      nPerRev*=2.0; // implementing halfing the time-step
      x[i]=cpmlog10(ko.dt()); // log timestep
      y[i]=cpmlog10s(cpmabs(pf.x1-pfExt.x1));
         // log absolute position error at tTot
      cpmdata<<"x["<<i<<"] = "<<x[i]<<" y["<<i<<"] = "<<y[i]<<endl;
   }
   R_Vector res=polyFit(x,y,1,0,1,false);
   R order=cpmabs(res[4]);
   Graph g;
   g.setWithOrigin(0);
   g.addText("order="&cpm(order)&", method="&koExt.met());
   g.show(x,y);
   cpmwait(tWaitFinal,2);
}

void app5(void)
{
   Word loc("app5");
   RecordHandler rch("app5.ini");
      // This says that we expect data input to be available in file
      // app5.ini in a diretory which is determined by entries in
      // cpmsystemdependencies.h and cpmconfig.ini.
   Word sec="data"; // names 'rch' and 'sec' are to be employed
      // for the macro cpmrh to work
   R eps=0.25, tWaitFinal=1, nPerRev=16, order=2;
   Z i,j, nTot=16, nDoubling=2, method=2;
// getting quantities from section 'data' of file app5.ini
   cpmrh(eps); // rh: read handler
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(stepRed_);
   cpmrh(tWaitFinal);
   cpmrhf(order);
   cpmrh(method);
   Z nData=nDoubling+1;
   V< V<R2> > res(nData);
   KepOsc koExt(method,eps,nPerRev);
   R fac=1.;
   for (i=1;i<=nData;++i){
      KepOsc ko(method,eps,nPerRev);
      V<R2> resi(nTot);
      for (j=1;j<=nTot;++j){
         ko.step_();
         R tj=ko.t();
         KepOsc kj=koExt;
         kj.extEvl_(tj);
         R2 dpos=ko.phsPos()-kj.phsPos();
         resi[j]=R2(tj,dpos.x1*fac);
      }
      res[i]=resi;
      nTot*=2;
      nPerRev*=2.0;
      fac*=cpmpow(2.0,order);
   }
   Graph g;
   g.setWithOrigin(0);
   g.addText("method="&koExt.met());
   g.mark(res);
   g.vis();
   cpmwait(tWaitFinal,2);
}

void app6()
// this is a place for playing arround with graphical functions
{
    Frame f;
    f.paint(RED);
    f.addText("This is a trivial experiment");
    f.vis();
    cpmwait(8,2);
    cpmmessage("app6 done");
}

void app7(void)
{
   Word loc("app7");
   RecordHandler rch("app7.ini");
   Word sec="data";
   Z metNum=6; // number of integration methods
   R E0=0.,eps=0.25, tWaitFinal=1, nPerRev=16, acc=1e-8;
   Z i,j, nTot=16, nDoubling=2;
   cpmrh(E0);
   cpmrh(eps);
   cpmrh(nPerRev);
   cpmrh(nTot);
   cpmrh(nDoubling);
   cpmrhf(stepRed_);
   cpmrhf(acc);
   cpmrh(tWaitFinal);
   cpmrhf(facALF);
   cpmrhf(cpmverbose);
   Z nData=nDoubling+1;
   R_Vector x(nData);
   R_Matrix y(metNum,nData);
   R prog=0;
   cpmprogress("loop",prog,3);
   R dProg=1./metNum;

   for (Z method=0;method<metNum;++method){

      KepOsc koExt(method,eps,nPerRev,E0);
      R tTot=koExt.dt()*nTot;
      koExt.extEvl_(tTot,acc);
      R2 pfExt=koExt.phsPos(); // exact phase space position
      // at the end of the trajectory under consideration
      R nPerRevVar=nPerRev;
      Z nTotVar=nTot;
      for (i=1;i<=nData;++i){
       //  cpmprogress("loop",prog,3);
         KepOsc ko(method,eps,nPerRevVar,E0);
         for (j=1;j<=nTotVar;++j) ko.step_();
         R2 pf=ko.phsPos();
         nTotVar*=2; // implementing halfing the time-step
         nPerRevVar*=2.0; // implementing halfing the time-step
         x[i]=cpmlog10(ko.dt()); // log timestep
         y[method+1][i]=cpmlog10s(cpmabs(pf.x1-pfExt.x1));
        //y[method+1][i]=cpmlog10s((pf-pfExt).abs());
      }
      prog+=dProg;
      cpmprogress("loop",prog,3);
   }
   Graph g;
   g.setWithOrigin(0);
   Word txt("log10 of error (y-axis) vs. log10 of time step ");
   txt&="(x-axis) for 6 integrators";
   g.addText(txt);
   g.setWriteOnShow(true);
   g.show(x,y,V<Color>("",BLACK,YELLOW,GREEN,RED,CYAN,MAGENTA));
   cpmwait(tWaitFinal,2);
}

} // namespace CpmKepler

void info()
{
   cout<<"Studying the Kepler oscillator"<<endl
   <<" This function takes 0 to 1 integer arguments"<<endl
   <<" values 1 2 3 4 5 6 7"<<endl
   <<" trigger one of the five basic modes, each other number"<<endl
   <<" triggers execution of these modes in succession."<<endl
   <<" No argument is equivalent to an argument of value 1.";
}

Z CpmApplication::main_(void)
// Selects the task from the command line, if there is a numerical
// argument following the function name. Further, the program tries to
// find and read tut2.ini. Here the filename is the name of the executable
// with all capital letters removed (so tut2MS.exe would als search for
// tut2.ini).
// Places to search for this are determined by cpmsystemdependencies.h
// and cpmconfig.ini, where the latter enjoys priority.
// If no tut2.ini can be found, the coded default selection takes
// place.
// The functionality of the tasks app1(), app2(), app3(), app4(), app5(),
// app7() is controlled through files ap1.ini, app2.ini, app3.ini,
// app4.ini, app5.ini, app7.ini for which search places are determined by
// the same logic as those of tut2.ini.
{
   using namespace CpmKepler;
   title_="tut2";
      // No longer determining the name of the ini-file
   Z mL=1;
   Z sel=1; // default selection
   cpmverbose=1;
   if (args_.valInd(2)){ // arrays allow testing the validity
         // of indexes without running into exceptions
      Word w2=args_[2];
      if (w2=="?"){
         info();
         return 0;
      }
      else sel=w2.toZ(); // conversion to Z (integer)
   }
   Word loc("CpmApplication::main_(), sel="&cpm(sel));
   CPM_MA

   if (sel==1) app1();
   else if (sel==2) app2();
   else if (sel==3) app3();
   else if (sel==4) app4();
   else if (sel==5) app5();
   else if (sel==6) app6();
   else if (sel==7) app7(); //// for trivial experiments with graphics
   else { // do everything sequentially
      app1();
      app2();
      app3();
      app4();
      app5();
      app6();
      app7();
   }
   CPM_MZ
   return sel;
}

// end of tut2.cpp
