//? chebyshev.cpp
//? C+- by Ulrich Mutze. Status of work 2022-10-23.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


/*************************************************************************

This program demonstrates the C+- programming style,
particularly the usage of functions as values.
That most functions in the demonstration are Chebyshev polynomials
is not important for this purpose.

Since the present program is intended as a vehicle not only
for demonstration but also for explanation (what I call a tutortial
program), it contains more comments than would be appropriate for a
typical program.

The program is controled by the text file tut3.ini.

Here is a copy of the intoductory part of this file which gives
more information on the topics the program deals with:

/////////////////////////////////////////////////
   about
/////////////////////////////////////////////////
W file=tut3.ini
Ws line1=Graphical representation of Chebyshev polynomials.
Ws line2=Testing orthogonality by Romberg integration.
Ws line3=Finding the proper parameters of the orthogonalizing
Ws line4=weight function by minimization.
Ws line5=Visualization of scalar products.

These are the program specific data that go into the creation
of the start-up screen of the project. Other elements of
this start-up screen are common to all C+- programs which
use class CpmApplication::IniFileBasedApplication
(see file cpminifilebasapp.h) as an application framework.
The definition of function IniFileBasedApplication::announceTheWork()
in file cpminifilebasapp.cpp defines these additional elements.

Notice also the following section, which any ini-file is expected to
have and which allows to influence the program run and the
auto-generated documentation in many ways. Although
ini-files are not seen by the compiler, they read like
C++ code by having type indicators for their data entries.
Also comments are treated as in C++.
Valid type indicators are B, Z, R, W for bool, integer, real, and
character string and the corresponding lists Bs, Zs, Rs, Ws.
There is a further valid type indicator F, which introduces the name
of an ini-file to be included. So ini-files nest to an arbitrary level.
This feature was developed for an industrial project where all data
defining the kinematics of a complex robot were originally held in a
single ini-file.
With the number of data growing to several thousands and with
people from other groups becoming involved in adjusting the
data, it became mandatory to split the data to separatly
editable units and nevertheless ensure integrity of production runs
by having a single ini-file at work.
Several of my C+- programs make use of this facility for similar
reasons.

Text lines in which there is no data type indicator are section
headings. Sections behave very much like namespaces: Althoug
section 'about' (see above) has an entry 'file' a section with
different name may also have an entry named 'file'. Sections
don't nest, so that several F-included files may add to
the same section, which turned out to be very convenient in the
framework for which I developed the scheme.

////////////////////////////////////////////////////////////////////////
   selection
////////////////////////////////////////////////////////////////////////

R tWaitAnnounceScreen=0

Z sel=1
// a selector for the 'depth' of auto-generated documentation files.
// Control of auto-generated documentation files is based on the
// two entries sel, runName. These are the rules:
// 1. for sel<0 no documentation files will be created.
// 2. for sel==0 a light documentation file is created that
//   covers all the input so that the program run can be repeated.
// 3. for sel>0, in addition to the light documentation file
//    a more detailed one will be created that also can made to
//    list computational results.
// 4. All documentation files have names which contain runName
//    appended to the name of the calling program

Z cpmverbose=1
// controls the degree of detail with which program run
// information will be written to log file cpmcerr.txt

Z cpmdbg=2
// controls the reaction of the cpmassert macro:
// 1: warning on failing assertion
// 2: error stop on failing assertion

W runName=070830a

// identifier for the program run that will be appended to
// the names of documentation files.
// If the value of this quantity is "auto" this value will
// be replaced by an autogenerated name.

***********************************************************************/

#include <cpmbas1.h>

using namespace CpmRoot;
using namespace CpmFunctions;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmGraphics;
using namespace CpmImaging;
using namespace CpmApplication;
using namespace std;
using namespace std::placeholders;

class ChebyApp: public IniFileBasedApplication
{
   public:
      ChebyApp();
      void doTheWork();
};

ChebyApp::ChebyApp(void):IniFileBasedApplication("tut3"){}
// This line implies that the program relies on a file named
// chebyshev.ini for getting the data which determine a
// program run. So we are in a position to tell the program
// in advance what it should do. If it is running, our only
// way to influence it is to cancel it.
// The location of this ini file can be given as a second argument to the
// constructor such as
// ChebyApp(void):IniFileBasedApplication("chebyshev","./control"){}
// Or, more flexibly by an entry to the ini-file cpmconfig.ini

namespace{ // Material needed in ChebyApp::doTheWork().
   // What goes on here, is easily understood taking into
   // account the usage there.
   // Functions don't nest in C++, so we have to keep
   // them outside of other functions.
   // C+- has the tools to finally use such outside function code
   // from any place in a very convenient manner. Given these tools,
   // it looks like a happy idea in C++ to prevent functions
   // from nesting. One would need more formal machinery
   // to use nested functions in the same manner as
   // we use top-level functions here.

   R rwsp(R const& x, R const& eps, R const& p)
      // regularized weight (in) scalar product
   {
      R y=cpmabs((1-x)*(1+x))+eps*eps;
      return cpmpow(y,p); // y>=0
   }

   R wsp(R x){ return rwsp(x,0,-0.5);}
      // exact orthogonalizing weight (in) scalar product

   R_Matrix sclPrd( V<R_Func> const& vf, R_Func const& w, Iv const& iv,
      RomCon const& rc=RomCon())
      // building a matrix of scalar products of the functions from list
      // vf. The scalar products employ a weight function w an
      // integration interval iv and depend on the method parameters rc
      // which control the integration method, which is Romberg
      // integration in our case.
   {
      Z done=0;
      Z i,j,n=vf.dim();
      R_Matrix sp(n,n);
      R doneInv=2./(n*(n+1));
      for (i=1;i<=n;++i){
         for (j=i;j<=n;++j){
            R_Func fij=vf[i]*vf[j]*w;
            R rij=fij.romInt(iv[1],iv[2],rc);
            sp[i][j]=rij;
            done++;
            if (i!=j) sp[j][i]=rij;
          }
      }
      return sp;
   }

   R fAux( V<R_Func> const& vf, R_Func const& w, Iv const& iv,
      RomCon const& rc=RomCon())
   // Instead of the matrix of scalar products we return the
   // deviation of this matrix from the unit matrix measured
   // by a real number.
   {
      R_Matrix sp=sclPrd(vf,w,iv,rc);
      Z n=sp.dim();
      R_Vector dia(n,1.);
      R_Matrix diag(dia);
      R_Matrix diff=sp-diag;
      return diff.abs()/(n*n);
   }

   R f2Aux(R_Vector const& p, V<R_Func> const& vf, RomCon const& rc)
      // Just as the previous function, but now function w is
      // parametrized by parameters p[1],p[2] and function code rwsp as
      // given as the first piece in the present namespace. In this form
      // the function will be used in the minmization part of the program.
   {
      R eps=p[1];
      R exponent=p[2];
      Iv iv(-1,1);
#ifdef CPM_Fn
      R_Func w=F2<R,R,R,R>(eps,exponent)(rwsp);
#else
      R_Func w=F<R,R>(bind(rwsp,_1,eps,exponent));
#endif
      R res=fAux(vf,w,iv,rc);
      cpmvalue("eps",eps,2);
      cpmvalue("exp",exponent,3);
      cpmvalue("res",res,4);
         // writing to the status bar
      return res;
   }

   R sin1(R const& x){ return sin(x);}
   R cos1(R const& x){ return cos(x);}
   R acos1(R const& x){ return acos(x);}
      // If multiple precision arithmetic is active
      // the native trigonometric functions have two
      // arguments. For the second argument there is
      // a default value provided so that a normal function call
      // needs only a single argument (as above). For the way we
      // use these functions in the following, the second argument
      // must not be ignored. So, we define functions with the
      // appropriate argument structure here.
}

void ChebyApp::doTheWork()
// does what the name says
{
   Word loc("ChebyApp::doTheWork()");
   Z mL=1;
   CPM_MA
// data input
// It is instructive to inspect file chebyshev.ini from where the
// data come.
bool writeOnShow=(cpmverbose>1);
#if defined(CPM_NOGRAPHICS)
   writeOnShow=true;
#endif
   Word sec("Show curves");
      // We read data for the first section of the program
      // from the first program specific section of the ini-file
   Z nL,nU,graphRes,bgcR,bgcG,bgcB;
   R xL,xR,tWaitGrp,gamma=-1;
   B noGridText, noGrid, autoScale, antiAliasing,
      curvesBlack;
      // Here we declare the quantities which we will 'read in from the
      // ini-file'. To add a new quantity here, and a new corresponding
      // read statement, and the corresponding data entry in the ini-file,
      // is a simple and structured activity which is not likely to
      // produce errors. If it does nevertheless, the program will halt
      // with a clear statement of the problem on the log file
      // cpmcerr.txt.
   cpmrh(nL);
      // Reading from the RecordHandler instance rch, which is the
      // most important data element of class IniFileBasedApplication.
      // rch gets initialized from the ini-file  by means of the
      // constructor of IniFileBasedApplication.
      // RecordHandler is a container for data of various types and
      // has the capability of distributing his data to various processes
      // in MPI-based parallel processing. See file cpmrecordhandler.h
      // for details. For single process programming it is sufficient
      // to think of these read statements as 'reading from the ini-file'
      // and ignoring the fact that here the data actually come out of
      // the container rch.
      // cpmrh(x) stopps as error, if x cannot be found. Either it is not
      // is not there by omission, or the ini-file syntax is violated in
      // in the intended definition of x.
      // cpmrhf(x) (f for 'floppy') gives only a warning on the log file
      // in such a case.
   cpmrh(nU);
   cpmrh(graphRes);
   cpmrh(xL);
   cpmrh(xR);
   cpmrh(tWaitGrp);
   cpmrh(bgcR);
   cpmrh(bgcG);
   cpmrh(bgcB);
   cpmrh(curvesBlack);
   cpmrh(noGridText);
   cpmrh(noGrid);
   cpmrh(autoScale);
   cpmrh(antiAliasing);
   cpmrhf(gamma); // see previous explanation of cpmrhf

// work
   Z n=nU-nL+1;
   V< Fa<R,R> > ch(n);
   Fa<R,R> ACos(acos1);
   Fa<R,R> Cos(cos1);
      // bringing C -functios cos and acos into the F<R,R>
      // framework. To allow for arbitrary precision arithmetic
      // we do not simply write acos and cos here.
   R ni=nL;
   Z i;
   for (i=ch.b();i<=ch.e();++i)
   {
      ch[i]=(ACos*ni)&Cos;
         // building a list of Chebyshev functions. This corresponds to
         // T_n(x)=cos(n*arccos(x)). Notice that the concatenation
         // '&' of functions is the kind which in mathematics is mostly
         // denoted by ';' Mathematical standard concatenation is
         // circ() in C+-, so one could equally well write
         // ch[i]=Cos.circ(ACos*ni);
         // Notice (ACos*n)(x) == ACos(x)*n   and  n o t
         // (ACos*n)(x )== ACos(x*n)
      ni+=1;
   }
   R facN=1/cpmsqrt(cpmpi/2);
   R facN0=1/cpmsqrt(cpmpi);

   V< R_Func > chNor(n); // Nor: normalized
      // to be used in the study of scalar products, the
      // original (not normalized) version will be
      // used for graphical representation.
   for (i=ch.b();i<=ch.e();++i)
   {
      chNor[i]= (i==1&& nL==0) ? ch[i]*facN0 : ch[i]*facN;
   }
   R_Func Wsp(wsp); // use outside function wsp
   Color bgc(bgcR,bgcG,bgcB);
   Graph gr(bgc);
      // If backgroud color does not matter, Graph gr;
      // is sufficient to have a graphical window of maximum size
      // and a wealth of capabilities.
   const Z twp=2; // selects the pane (on status bar) on which the
      // remaining waiting time (if there is one) will be displayed
   if (tWaitGrp>0){
      // it is convenient to switch off a program step by setting the
      // time scheduled for inspection of its result as <=0.
      cpmmessage("Graphics",4); // there ate 4 status panes and we
         // indicate on the last one what we presently are doing
      gr.setX(Iv(xL,xR));
         // setting the x-range to which the horizontal extension of
         // the graphics window corresponds.
         // It is interesting to set e.g. xL=0.999, xR=1
         // nL=50 nU=150 to see the behavior of the curves at the
         // upper end of their natural domain.
      gr.setY(Iv(-1,1));
         // Since all Chebyshev polynomials vary in [-1,1] this is
         // a natural setting for the y-range. Can be overridden
         // by asking for auto scaling.
      gr.setNoGridText(noGridText);
         // grid text gives the actual x,y-range associated with the
         // window. Whether it should be shown can be selected
      gr.setNoGrid(noGrid);
         // showing a grid can be enabled or disabled
      gr.setAutoScale(autoScale);
         // controls auto scaling
      gr.setAntiAliasLines(antiAliasing,gamma);
         // controls drawing of the curves
      gr.setWriteOnShow(writeOnShow);
      if (curvesBlack){ // normally a rainbow color code is used
         // for showing families of curces
         V<Color> cl(5,BLACK);
         gr.show(ch,graphRes,cl);
            // ch is a family of curves. Here, these are
            // the Chebyshev polynomials as selected by the
            // integers nL and nU
      }
      else gr.show(ch,graphRes);
         // graphRes sets the number of x-Values to be used
         // in converting curves to polygons for visualization
      cpmwait(tWaitGrp,twp); // causes the program to wait
         // for as many seconds as the first argument indicates. The
         // remaining time will be displayed on status pane twp
   }

////////////////////////////////////////////////////////////////////////

   sec="Integration";
     // We read data for the second section of the program
     // from the second program specific section of the ini-file
   Z iMethod,iMax,ip,nChebyshev;
   R epsRel,epsAbs,tWaitInt1,tWaitInt2;

   cpmrh(tWaitInt1);
   cpmrh(tWaitInt2);
   cpmrh(epsRel);
   cpmrh(epsAbs);
   cpmrh(iMethod);
   cpmrh(iMax);
   cpmrh(ip);
   cpmrh(nChebyshev);

   RomCon rc(epsRel,epsAbs);
   rc.setMet_(iMethod);
   rc.setCntMax_(iMax);
   rc.setLen_(ip);
   if (tWaitInt1>0){
      cpmmessage("Integration 1",4);
      R norF=nChebyshev==0 ? facN0 : facN;
      R nC=nChebyshev;
      Fa<R,R> fn=(ACos*nC)&Cos;
      fn*=norF;
      R_Func g=Wsp*fn*fn;
      R sL=-1, sU=1;
      R vExact=1;
      R val=g.romInt(sL,sU,rc);
      R err=vExact-val;
      cpmvalue("val",err,3);
      V<Word> lines(0);
      lines<<"Computing the norm of the normalized \
Chebyshev function of index "+cpm(nChebyshev);
         // to avoid long lines in the present source file, not in the
         // screen output
      lines<<"Method is Romberg integration with method parameters";
      lines<<"as given in the section Integration of the ini-file.";
      lines<<"Deviation from the mathematical result is "+cpm(err);
      gr.clr();
      gr.setText(lines);
      gr.vis(writeOnShow);
      cpmwait(tWaitInt1,twp);
   }
   if (tWaitInt2>0){
      cpmmessage("Integration 2",4);
      R_Func g=R_Func(sin1);
      R sL=0, sU=cpmpi;
      R vExact=2; // math knows
      R val=g.romInt(sL,sU,rc);
         // integral of g from sL to sU using Romberg with
         // method parameters rc
      R err=vExact-val;
      cpmvalue("val",err,3);

      V<Word> lines(0);
      lines<<"Integration of function sin from 0 to pi.";
      lines<<"Method is Romberg integration with method parameters";
      lines<<"as given in the section Integration of the ini-file.";
      lines<<"Deviation from the mathematical result is "+cpm(err);
      gr.clr();
      gr.setText(lines);
      gr.vis(writeOnShow);
      cpmwait(tWaitInt2,twp);
   }

////////////////////////////////////////////////////////////////////////
   sec="Minimization";
     // We read data for the third section of the program
     // from the third program specific section of the ini-file
   R tWaitMin,epsStart,expStart,epsMin;

   cpmrh(tWaitMin);
   cpmrh(epsStart);
   cpmrh(expStart);
   cpmrh(epsMin);

   if (tWaitMin>0){
      cpmmessage("Minimization",4);
      gr.clr(); // clear
      V<Word> lines; // preparing the making of screen text
      lines<<"Minimizing the deviation from ortho-normalization";
      lines<<"by adjusting two parameters in the weight function.";
      lines<<"";
      lines<<
         "Observe the values shown on the 4 panes of the status bar.";
// string continuation used since for supporting my present code listing
// procedure, there should not be more than 75 characters in a line.
// There are many ways to achieve this.
      lines<<"pane 1: general messages and warnings";
      lines<<"pane 2: present values of regularizing epsilon, 0 is the\
 exact value";
      lines<<
        "pane 3: present values of the exponent, -0.5 is the exact value"\
        ;
      lines<<"pane 4: present function value, exact minimum value of\
 which is 0";
      lines<<"";
      lines<<"Minimization is running .... ";

      gr.setText(lines);
      gr.vis(writeOnShow); // showing the screen text
      F<R_Vector,R> fMin(bind(f2Aux,_1,chNor,rc)); // this is the function
         // Notice that F<R_Vector,R> is the type expected by class
         // CpmAlgorithms::Mini. All pertinent quantities that are not
         // represented by the R_Vector-typed function argument have
         // to be initialized by quantities which the program has
         // already defined. These are the objects chNor and rc
         // and the piece of function code f2Aux. The syntax that
         // brings these entities together to form a F<R_Vector,R>
         // is simple and clear: f2Aux represents a mapping
         // f2Aux : R_Vector x V<R_Func> x RomCon --> R
         // (consider its declaration abstracted to:
         //       R f2Aux(R_Vector,V<R_Func>,RomCon)
         // )
         // The operation std::bind gives to the second and the third
         // argument of function f2Aux fixed values (namely chNor and rc)
         // and leaves the first argument as a free function argument (the
         // placeholder '_1' indicates this). This leaves us with a
         // function R_Vector --> R for which we introduce the name fMin.
         // It is fun to employ this scheme even with more parameters.
         // Notice that fMin, as defined here, shows strictly the behavior
         // of a F<R_Vector,R>.
         // The values of the parameters that were used in the definition
         // of fMin are completely hidden.

      CpmAlgorithms::Mini mini(fMin,epsMin);
         // instance of a minimizer class
      R_Vector pStart{epsStart,expStart};
      R tStart=cpmtime();
      X2<R_Vector,R> res=mini(pStart);
         // makes the minimizer run, needs a proposal for a starting
         // value. The return value is a pair of values.
      R tEnd=cpmtime();
      Z nEval=mini.nEval();
      R tComp=tEnd-tStart;
      R tPerEval=tComp/nEval;
      R_Vector p=res.c1();
      R yMin=res.c2();
      R eps=p[1];
      R exponent=p[2];
// communicating the result
      cpmvalue("eps",eps,1);
      cpmvalue("exp",exponent,2);
      cpmvalue("yMin",yMin,3);
      gr.clr();
      lines=V<Word>();
      lines<<"Minimization came to an end";
      lines<<"Minimum function value was found as "+cpm(yMin)+
         ", exact value is 0";
      lines<<"eps was found as "+cpm(eps)+", exact value is 0";
      lines<<"exponent was found as "+cpm(exponent)+
         ", exact value is -0.5";
      lines<<"Computation time was "+cpm(tComp)+" seconds";
      lines<<"Number of function evaluations was "+cpm(nEval);
      lines<<"Thus the time per function evaluation was "+cpm(tPerEval)+
         " seconds";
      gr.setText(lines);
      gr.vis(writeOnShow);
      cpmwait(tWaitMin,twp);
   }

   sec="Show orthogonality";
     // We read data for the fourth section of the program
     // from the fourth program specific section of the ini-file
   R tWaitShowOrt,eps=1e-6,exponent=-0.5;
   B interpolate, showDiff;

   cpmrh(tWaitShowOrt);
   cpmrh(showDiff);
   cpmrh(gamma);
   cpmrh(interpolate);
   cpmrh(eps);
   cpmrh(exponent);
   if (tWaitShowOrt>0){
      Viewport::clrPan();
      cpmmessage("Show orthogonality",4);
      Frame fr;
      Frames frs(fr,1,2); // creates a 1x2 matrix of subframes
         // thus 2 frames in a line
      Graph gr1(frs(1,1));
      Graph gr2(frs(1,2));
      V<Word> lines;
      Z nF=chNor.dim();
      lines<<"Showing the mutual scalar products of the "+cpm(nF)+
         " normalized";
      lines<<"Chebyshev polynomials under consideration as a matrix.";
      if (showDiff){
         lines<<"Actually, the unit matrix is subtracted from the";
         lines<<"matrix of scalar products.";
      }
      lines<<"Value of gamma is "+cpm(gamma);
      lines<<"Value of epsilon is "+cpm(eps);
      lines<<"Value of exponent is "+cpm(exponent);
      lines<<"For eps=0 and exponent = -0.5 deviations of the matrix of";
      lines<<"scalar products from the unit matrix are due to numerical";
      lines<<"errors in Romberg integration with method parameters";
      lines<<"chosen as follows:";
      lines<<rc.show();
      lines<<"";
      lines<<"Computing graphics ...";
      gr1.clr();
      gr1.setText(lines);
      gr1.vis(writeOnShow);

      R_Func weight(F<R,R>(bind(rwsp,_1,eps,exponent)));
      R_Matrix ort=sclPrd(chNor,weight,Iv(-1,1),rc);
      if (showDiff){
         R_Vector dia(n,1.);
         R_Matrix diag(dia);
         ort-=diag;
      }
      R absMax=ort.supAbs();
      V< V<C> > vs(n, V<C>(n));
      for (i=1;i<=n;++i) for (Z j=1;j<=n;++j) vs[i][j]=C(ort[i][j]);
      Color::setGamma(gamma);
      gr2.clr();
      gr2.markVVC(vs,false,Iv(),interpolate);
      gr2.setText("Maximum of absolute values is "&cpm(absMax));
      gr2.vis(writeOnShow);

      lines.last()="Computation done";
      gr1.clr();
      gr1.setText(lines);
      gr1.vis(writeOnShow);
      cpmwait(tWaitShowOrt,twp);
   }
   CPM_MZ
}

Z CpmApplication::main_(void)
// In this form the C++ typic function main() (defined in file
// cpmapplication.cpp) expects to see the basic functionality of
// the program.
// Notice that main() in cpmapplication.cpp provides the program
// with its graphical window and a status bar.
// The fact that class ChebyApp is derived from class
// IniFileBasedApplication lets the following short code, although it
// looks like abstract nonsense, do very specific things.
{
   Word loc("Z CpmApplication::main_(void)");
   Z mL=1;
   CPM_MA
   title_=Word("tut3");
   ChebyApp app;
      // creates the application class by reading and deploying
      // tut3.ini
   app.run();
      // brings the application class to live and action
   cout<<"app.tun() in main_() done"<<endl;
   CPM_MZ
   cpmwait(10.,2);
   return 1;
}
