//? vqm2app1.cpp
//? C+- by Ulrich Mutze. Status of work 2022-03-21.
//? Copyright (c) 2020 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


/************************************************************************
   vqm2app1.cpp
***********************************************************************/
#include <vqm2app1.h>
#include <vqm2.h>

#include <cpmframes.h>
#include <cpmgreg.h>
#include <cpmrfuncrec.h>
#include <cpmoperator.h>
#include <cpmtrj.h>
#include <cpmfieldoperator.h>
#include <cpmstate2.h>

#include <stdlib.h> // for system()
#include <unistd.h> // for chdir()

using namespace CpmSystem;
using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmGraphics;
using namespace CpmImaging;
using namespace CpmQuantumMechanics;
using CpmAlgorithms::CyclicAlarm;
using CpmTime::Greg;

////////////////////// App1 /////////////////////////////
// Testing spectral representation of hermitian matrices
// Results 2021-02-27:
// std=c++20 , g++10, d=400
// !defined(CPM_MP) defined(CPM_USE_EIGEN) : 0.599846 for main task
// !defined(CPM_MP) !defined(CPM_USE_EIGEN) : 57.8278 for main task
// Thus using Eigen results in an acceleration by a factor 96.
// now the experience with #define CPM_MP=64
// defined(CPM_USE_EIGEN) -fpermission: 385.38 for main task
// !defined(CPM_USE_EIGEN) -fpermission: even for d=100 stop due to
// too many iterations in tqli. The iteration limit was 1000 till today.
// Even after setting it to 1000000 we got too many iterations for
// d=100. So it is quite clear that the 'Eigen-less' implementation
// of spectral decomposition does not work for multi-precission.
// Last (?) experiment with a less ambitious precision: CPM_MP=32 too
// many iterations, same with CPM_MP=32. However quite fast success (1.92s,
// error 1.2 10^-14 with CPM_LONG (80 bit). For 63 bit double the error
// was ~ 10^-10.

void CpmVQM2::App1::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App2::doTheWork()");
   CPM_MA
   Z d=128;
   Word sec="system data";
////////////////////////////////////////////////////////////////////////
   cpmrh(d);
   R_Vector res=SpcObj::test(d,137);
   cpmcerr<<res;
   CPM_MZ
}

/////////////////////////////// App2 //////////////////////////////
// new version devoted to study of dephasing and localization
// and test of creation and destruction operators.

namespace{

R jitterFunc(C const& c, R_Func const& f, R const& jt, R const& omega )
{
   R x=c[1];
   R t=c[2];
   R res=f(x);
   if (x<=0) return res;
   else{
      R shift=cpmcos(t*omega)*jt;
      return res+shift;
   }
}

} // namespace

void CpmVQM2::App2::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App2::doTheWork()");
   CPM_MA

   B useSI,standardMomentum,cutMomentum,commutator,duplicate,
      buildSpcObj, enfSym,test,showSpcAsBars,
      fileObs,fileSpeObj,filePsi,
      spcObjFromFile,psiFromFile,
      useSplitBoost,useRelCplCon;

   Z d,iterMax,buildT,buildV,buildS,buildR, buildPsi,
      ipolOrd,ipolMet,modeOfDyn;

   R s,mass,kp,ks1,ks2,ks3,kr,boostRel,gamma,eps,eps1,eps2,
      polyFac,bernsteinFac,bezierFac,ipolFac,gaussFac,indFac,
      bernsteinX,bernsteinY,showObs,showSpcObj,showSpc,showPsi,
      phiStep;

   Word file;
   V<Word> line1;
   V<Word> line2;
   V<Word> line3;
   V<Word> line4;

   V<Z> cN;

   V<R> waitTimes,poly,bernstein,bezierX,bezierY,ipolX,ipolY,
      gaussX,gaussSigma,gaussY,indX,indWidth,indY,cRe,cIm;
      // notice that by far the most data come from sections
      // potential, state1, ..., state3 which do not shine up
      // here explicitly

// reading data
// I choose to read all data in succession although for most settings
// in run_control only a part of these data is needed in the code
// to be executed further-on. Letting each control path read only those
// data that it really needs would create complicated interdependencies
// of the various control paths, resulting in the need for duplicating
// code. This also would complicate further extensions.
////////////////////////////////////////////////////////////////////////
   Word sec="about";
////////////////////////////////////////////////////////////////////////
   cpmrh(file);
   cpmrh(line1);
   cpmrh(line2);
   cpmrh(line3);
   cpmrh(line4);

////////////////////////////////////////////////////////////////////////
   sec="run preparation";
////////////////////////////////////////////////////////////////////////
   cpmrh(waitTimes);
   cpmrh(gamma);
   cpmrh(fileObs);
   cpmrh(fileSpeObj);
   cpmrh(filePsi);
   cpmrh(spcObjFromFile);
   cpmrh(psiFromFile);
   cpmrh(showObs);
   cpmrh(showSpcObj);
   cpmrh(showSpc);
   cpmrh(showPsi);
   cpmrh(showSpcAsBars);
   cpmrh(phiStep);
   cpmrh(modeOfDyn);

////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////
   cpmrh(d);
   cpmrh(s);
   cpmrh(mass);
   cpmrh(useSI);
   cpmrh(duplicate);
   cpmrh(buildSpcObj);
   cpmrh(commutator);
   cpmrh(buildT);
   cpmrh(buildV);
   cpmrh(buildS);
   cpmrh(buildR);
   cpmrh(buildPsi);
   cpmrh(useRelCplCon);
   cpmrh(kp);
   cpmrh(ks1);
   cpmrh(ks2);
   cpmrh(ks3);
   cpmrh(kr);
   cpmrh(boostRel);
   cpmrh(useSplitBoost);

////////////////////////////////////////////////////////////////////////
   sec="kinetic energy";
////////////////////////////////////////////////////////////////////////
   cpmrh(standardMomentum);
   cpmrh(cutMomentum);

////////////////////////////////////////////////////////////////////////
   sec="potential";
////////////////////////////////////////////////////////////////////////
   cpmrh(polyFac);
   cpmrh(bernsteinFac);
   cpmrh(bernsteinX);
   cpmrh(bernsteinY);
   cpmrh(bezierFac);
   cpmrh(ipolFac);
   cpmrh(gaussFac);
   cpmrh(indFac);
   cpmrh(poly);
   cpmrh(bernstein);
   cpmrh(bezierX);
   cpmrh(bezierY);
   cpmrh(ipolX);
   cpmrh(ipolY);
   cpmrh(gaussX);
   cpmrh(gaussSigma);
   cpmrh(gaussY);
   cpmrh(indX);
   cpmrh(indWidth);
   cpmrh(indY);
   cpmrh(ipolOrd);
   cpmrh(ipolMet);

////////////////////////////////////////////////////////////////////////
   sec="state1";
////////////////////////////////////////////////////////////////////////
   cpmrh(ipolMet);
   cpmrh(ipolOrd);
   cpmrh(polyFac);
   cpmrh(bernsteinFac);
   cpmrh(bernsteinX);
   cpmrh(bernsteinY);
   cpmrh(bezierFac);
   cpmrh(ipolFac);
   cpmrh(gaussFac);
   cpmrh(indFac);
   cpmrh(poly);
   cpmrh(bernstein);
   cpmrh(bezierX);
   cpmrh(bezierY);
   cpmrh(ipolX);
   cpmrh(ipolY);
   cpmrh(gaussX);
   cpmrh(gaussSigma);
   cpmrh(gaussY);
   cpmrh(indX);
   cpmrh(indWidth);
   cpmrh(indY);

////////////////////////////////////////////////////////////////////////
   sec="state2";
////////////////////////////////////////////////////////////////////////
   cpmrh(ipolMet);
   cpmrh(ipolOrd);
   cpmrh(polyFac);
   cpmrh(bernsteinFac);
   cpmrh(bernsteinX);
   cpmrh(bernsteinY);
   cpmrh(bezierFac);
   cpmrh(ipolFac);
   cpmrh(gaussFac);
   cpmrh(indFac);
   cpmrh(poly);
   cpmrh(bernstein);
   cpmrh(bezierX);
   cpmrh(bezierY);
   cpmrh(ipolX);
   cpmrh(ipolY);
   cpmrh(gaussX);
   cpmrh(gaussSigma);
   cpmrh(gaussY);
   cpmrh(indX);
   cpmrh(indWidth);
   cpmrh(indY);

////////////////////////////////////////////////////////////////////////
   sec="state3";
////////////////////////////////////////////////////////////////////////
   cpmrh(ipolMet);
   cpmrh(ipolOrd);
   cpmrh(polyFac);
   cpmrh(bernsteinFac);
   cpmrh(bernsteinX);
   cpmrh(bernsteinY);
   cpmrh(bezierFac);
   cpmrh(ipolFac);
   cpmrh(gaussFac);
   cpmrh(indFac);
   cpmrh(poly);
   cpmrh(bernstein);
   cpmrh(bezierX);
   cpmrh(bezierY);
   cpmrh(ipolX);
   cpmrh(ipolY);
   cpmrh(gaussX);
   cpmrh(gaussSigma);
   cpmrh(gaussY);
   cpmrh(indX);
   cpmrh(indWidth);
   cpmrh(indY);

////////////////////////////////////////////////////////////////////////
   sec="psi";
////////////////////////////////////////////////////////////////////////
   cpmrh(ipolMet);
   cpmrh(ipolOrd);
   cpmrh(polyFac);
   cpmrh(bernsteinFac);
   cpmrh(bernsteinX);
   cpmrh(bernsteinY);
   cpmrh(bezierFac);
   cpmrh(ipolFac);
   cpmrh(gaussFac);
   cpmrh(indFac);
   cpmrh(poly);
   cpmrh(bernstein);
   cpmrh(bezierX);
   cpmrh(bezierY);
   cpmrh(ipolX);
   cpmrh(ipolY);
   cpmrh(gaussX);
   cpmrh(gaussSigma);
   cpmrh(gaussY);
   cpmrh(indX);
   cpmrh(indWidth);
   cpmrh(indY);

////////////////////////////////////////////////////////////////////////
   sec="linear combinations of eigen-vectors";
////////////////////////////////////////////////////////////////////////
   cpmrh(cN);
   cpmrh(cRe);
   cpmrh(cIm);

////////////////////////////////////////////////////////////////////////
   sec="controling spectral representation";
////////////////////////////////////////////////////////////////////////
   cpmrh(enfSym);
   cpmrh(test);
   cpmrh(iterMax);
   cpmrh(eps);
   cpmrh(eps1);
   cpmrh(eps2);

// An independent test of creation and destruction operators.
// Results get written to cpmcerr

   if (modeOfDyn==13){
      FieldState::test(d);
      return;
   }

   if (modeOfDyn==14){
      FieldState::graphicsTest(d);
      return;
   }

////////////////////////////////////////////////////////////////////////
// notice that there will be a further reading activity from a section
//
//   run control
//
// in the form:
//   Record rtc=(rch.getRecord())("run control","TrjCon");

// setting the stage

   Color::setGamma(gamma);
   Graph g;
   g.setWithOrigin(0);
   Z i;

   if (useSI) QMBase::setUnits("SI");
   else QMBase::setUnits("1");
   QMBase::setMass(mass);
   QMBase::setSize(s);
   QMBase::setPhiStep(phiStep);
   Iv ivi(-0.5,0.5);
   Iv ivf(-s*0.5,s*0.5);
   R lambda0=useRelCplCon ? QMBase::lambdaOsc(d) : R(1);

   R tw2=waitTimes[2];
   Z sP=512;

   // defining the Observable ob and (optionally) its spectral object spo
   Observable ob;
   SpectralObject spo;
   Operator op;
   FieldOperator fop;

   bool spoValid=false;

   cpmmessage("point 1 reached");

   if (spcObjFromFile){
      cpmmessage("Trying to read spectral object from file");
      IFileStream ifs1(inFiles[1]);
      spo.scanFrom(ifs1());
      spoValid=true;
      if(showSpcObj>0){
         g.addText("Spectral object from file");
         spo.show(g);
         cpmwait(showSpcObj);
      }
   }
   else{
      Word sec2="R_FuncRec";
      Record rp,rs1,rs2,rs3;

      if (buildV>0){
         sec="potential";
         rp=(rch.getRecord())(sec,sec2);
      }

      if (buildS>=1){
         sec="state1";
         rs1=(rch.getRecord())(sec,sec2);
      }

      if (buildS>=2){
         sec="state2";
         rs2=(rch.getRecord())(sec,sec2);
      }

      if (buildS>=3){
         sec="state3";
         rs3=(rch.getRecord())(sec,sec2);
      }

// defining operators
      Observable p,T,V_,S_(d),R_(d);
      cpmassert(buildT>=0,loc);
      cpmassert(buildT<=4,loc);
      if (buildT<0||buildT>5) buildT=2;

      if (buildT>0){
         if (standardMomentum)
            p=Observable(Word("momentum"),d);
               // multiplication operator in discrete Fourier space
         else
            p=Observable(Word("pF"),d);
               // finite difference

         if (cutMomentum) p=p.cut();

         if (buildT==1){
            T=p;
         }
         else if (buildT==2){
            if (standardMomentum) T=p*p*(1./(2*mass));
            else T=Observable(Word("eKinF"),d);
         }
         else if (buildT==3){
            T=Observable(Word("eKinRel"),d);
         }
         else if (buildT==4){
            T=Observable(Word("boost"),d);
         }
         else ; // can't happen
      }
      if (buildV>0){
         R_FuncRec fr(rp);
         R_Func fx=fr();
            // notice that a mathematical function of 'the
            // position' does not make sense without normalizing
            // the position to a dimension-less quantity. So it
            // is OK to consider all functions that deal with
            // distributions in space as defined on the normalized
            // interval [-0.5,0.5]
         fx*=(lambda0*kp);
         if (tw2>0){
            g.setX(ivi); // not rescaled
            g.addText("Continuous version of potential");
            g.show(fx,sP);
            cpmwait(tw2);
         }
         V_=Observable::pot(fx,d);
            // this is defined in a way that despite of an actual
            // position range [-size/2, size/2] function fx will be
            // always evaluated in [-1/2, 1/2]

         R2_Func fxy(fx);
         op=Operator::sinParHam(fxy,d);
         R2_Func fpair;
         // trivial interaction between pairs to have a good test
         // by comparing the dynamics from op and fop
         fop=FieldOperator::manParHam(fxy,fpair,d);
      }
      if (buildS>=1){
         R_FuncRec fs1(rs1);
         R_Func gs1=fs1();
         if (tw2>0){
            g.setX(ivi);
            g.addText("State 1");
            g.show(gs1,sP);
            cpmwait(tw2);
         }
         State s1(gs1,d);
         S_+=(Observable("projector",s1)*ks1);

         if (buildS>=2){
            R_FuncRec fs2(rs2);
            R_Func gs2=fs2();
            if (tw2>0){
               g.setX(ivi);
               g.addText("State 2");
               g.show(gs2,sP);
               cpmwait(tw2);
            }
            State s2(gs2,d);
            S_+=(Observable("projector",s2)*ks2);
         }

         if (buildS>=3){
            R_FuncRec fs3(rs3);
            R_Func gs3=fs3();
            if (tw2>0){
               g.setX(ivi);
               g.addText("State 3");
               g.show(gs3,sP);
               cpmwait(tw2);
            }
            State s3(gs3,d);
            S_+=(Observable("projector",s3)*ks3);
         }
         S_*=lambda0;
      }

      if (buildR>0){
         R_=R_.ran()*(kr*lambda0);
      }

      Observable o1(d); // zero operator
      if (buildT>0) o1+=T;

      Observable o2(d); // zero operator
      if (buildV>0) o2+=V_;
      if (buildS>0) o2+=S_;
      if (buildR>0) o2+=R_;

      ob=commutator ? (o1*o2-o2*o1)*C::I : o1+o2 ;

      if (duplicate) ob=ob.prdFock();

      if (buildSpcObj){
         spo=ob.spectralRepresentation(enfSym,test,iterMax,eps,eps1,eps2\
         )\
         ;
         spoValid=true;
      }

      g.clr();

      if (showObs>0&&showSpcObj>0&&spoValid){
         if (ob.dim()<128){
            Frames frs(g,Z(1),Z(2));
            Graph g1(frs(1,1));
            Graph g2(frs(1,2));
            g1.addText("The observable as a complex matrix");
            g2.addText("Spectrum and eigenvectors of the observable");
            ob.show(g1);
            spo.show(g2);
            cpmwait(showObs+showSpcObj);
         }
         else{
            g.addText("The observable as a complex matrix");
            ob.show(g);

            cpmwait(showObs);
            g.clr();
            g.addText("Spectrum and eigenvectors of the observable");
            spo.show(g);
            cpmwait(showSpcObj);
         }
      }
      else if (showObs>0){
         g.addText("The observable as a complex matrix");
         ob.show(g);
         cpmwait(showObs);
      }
      else if (showSpcObj>0&&spoValid){
         g.addText("Spectrum and eigenvectors of the observable");
         spo.show(g);
         cpmwait(showSpcObj);
      }
      else ;

      if (fileObs){
         OFileStream out1(outFiles[1]);
         out1()<<ob;
      }
      if (fileSpeObj&&buildSpcObj){
         OFileStream out2(outFiles[2]);
         out2()<<spo;
      }
      if (showSpc>0&&spoValid){
         R_Vector sp=spo.getSpectrum();
         cpmdata<<"Spectrum of the Observable="<<sp;
         Z bars=showSpcAsBars ? 1 : 0;
         R xPosTitle=0.05;
         g.show(sp,Color(RED),bars,
            "Spectrum of the Observable",xPosTitle);
         cpmwait(showSpc);
      }
   }

   cpmmessage("point 2 reached");

// defining the initial state psi
   State psi;
   bool psiValid=false;

   if (psiFromFile){
      IFileStream ifs2(inFiles[2]);
      psiValid=psi.scanFrom(ifs2());
      if (showPsi>0){
         g.addText("State psi from file");
         psi.showAsCurves(g);
         cpmwait(showPsi);
      }
   }
   else{
      if (spcObjFromFile) d=spo.dim();
      if (buildPsi==1){
         Record rpsi=(rch.getRecord())("psi","R_FuncRec");
         R_FuncRec fpsi(rpsi);
         R_Func gpsi=fpsi();
         if (showPsi>0){
            g.setX(ivi);
            g.addText("Continuous version of state psi");
            g.show(gpsi,sP);
            cpmwait(showPsi);
         }
         psi=State(gpsi,d);
         psiValid=true;
         cpmmessage("point 2.1 reached");
      }
      else if (buildPsi==2){
         cpmassert(spoValid,loc);
         Z cd0=cN.dim();
         Z cd1=cRe.dim();
         Z cd2=cIm.dim();
         cpmassert(cd0==cd1,loc);
         cpmassert(cd1==cd2,loc);
         Z cdm=CpmRootX::inf<Z>(cd0,cd1,cd2);
         psi=State(d);
         Z dm=CpmRootX::inf<Z>(cdm,d);
         for (i=1;i<=dm;++i){
            Z ispo=d-cN[i];
            if (ispo>=0 && ispo<d){
               psi+=(spo[d-cN[i]]*C(cRe[i],cIm[i]));
               // only existing eigen-vectors will be taken
            }
            else{
               cpmwarning(loc&": eigenvector with index "&cpm(ispo)&
                  " does not exist, d="&cpm(d));
            }
         }
         psi.nor_();
         psiValid=true;
         cpmdata<<endl<<"State psi="<<psi;
      }
      else{
         ;
      }

      if (psiValid){
         R vRef=QMBase::vMax(d);
         R vB=vRef*boostRel;
         if (vB!=0){
            if (useSplitBoost)
               psi=psi.splitBoost(vB);
            else
               psi=psi.boost(vB);
         }
         if (showPsi>0){
            g.addText("State psi");
            psi.showAsCurves(g);
            cpmwait(showPsi);
            Z d1(d);
            bool b2=CpmRootX::isPowOfTwo(d1);
            if (b2){
               State psiMom=psi.toMomSpc();
               g.addText("State psi in discrete momentum space");
               psiMom.showAsCurves(g);
            }
            cpmwait(showPsi);
         }
         if (filePsi){
            OFileStream out3(outFiles[3]);
            out3()<<psi;
         }
      }
   }

   cpmmessage("point 3 reached");

// showing the dynamical state evolution if this is
// requested by the input data
   R tw3=waitTimes[3];
   Record rtc=(rch.getRecord())("run control","TrjCon");
   TrjCon tc(rtc);
   Z nSteps=tc.nSteps_;
   Z readTrjFromFile=0;
   Z levOfPed=0; // level of pedantry
   rch.read("run control","readTrjFromFile",readTrjFromFile, levOfPed);
      // if reading failed, readTrjFromFile still has its previous value
      // which is 0. So older ini-files which don't have this parameter
      // (introduced 2004-10-11) continue to be valid for their particlar
      // task
   if (nSteps<=0 &&readTrjFromFile<=0){
      // then we have neither to read or create a trajectory
      CPM_MZ
      return;
   }

   if (readTrjFromFile>0){
      IFileStream in4(inFiles[4]);
      if (readTrjFromFile==1){
         State psiIrr; // Irr for irrelevant, has no function
         SpectralObject spoIrr;
         TrjSR trj1(spoIrr,psiIrr,tc);
         trj1.play(g,in4());
      }
      else if (readTrjFromFile==2){
         State psiIrr;
         Observable obIrr;
         TrjDMI_Obs trj2(obIrr,psiIrr,tc);
         trj2.play(g,in4());
      }
      else if (readTrjFromFile==3){
         State psiIrr;
         Operator opIrr;
         TrjDMI_Opr trj3(opIrr,psiIrr,tc);
         trj3.play(g,in4());
      }
      else if (readTrjFromFile==4){
         FieldState fsIrr;
         FieldOperator foIrr;
         TrjDMI_FilOpr trj3(foIrr,fsIrr,tc);
         trj3.play(g,in4());
      }
      else{
         cpmerror(loc&": readTrjFromFile="
            &cpm(readTrjFromFile)&" not defined");
      }
   }
   else{
      cpmmessage("point 3.2 reached");

      OFileStream out4(outFiles[4]);
         // output file for a single trajectory or the first one
         // in a tandem trajectory
      OFileStream out5(outFiles[5]);
         // output file for the second trajectory of a tandem
         // trajectory
      cpmmessage("point 4 reached");
      Z lim=1;
      if (nSteps>0 && psiValid){
         cpmmessage("point 5 reached");
         if (modeOfDyn==1){
            cpmassert(spoValid,loc);
            TrjSR trj(spo,psi,tc);
            cpmmessage("point 5.1 reached");
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==2){
            TrjDMI_Obs trj(ob,psi,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==3){
            TrjSR trj1=spoValid ? TrjSR(spo,psi,tc): TrjSR(ob,psi,tc);
            TrjDMI_Obs trj2(ob,psi,tc);
            Vp<TrjBase> vtrj(2); // polymorphic array
            vtrj[1]=trj1;
            vtrj[2]=trj2;
         //  Trj2 trj(trj1,trj2); // would also be OK
            Trj2 trj(vtrj);
            trj.run(g,out4(),out5());
            tw3*=4;
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==4){
            TrjDMI_Obs trj1(ob,psi,tc);
            TrjDMI_Opr trj2(op,psi,tc);
            Vp<TrjBase> vtrj(2);
            vtrj[1]=trj1;
            vtrj[2]=trj2;
            Trj2 trj(vtrj);
            trj.run(g,out4(),out5());
            tw3*=4;
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==5){
            TrjDMI_Obs trj(ob,psi,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==6){
            TrjDMI_Opr trj(op,psi,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==7){
            FieldState fs(d,1,lim);
            fs=fs.crt(psi);
            TrjDMI_FilOpr trj(fop,fs,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==8){
            TrjDMI_Opr trj1(op,psi,tc);
            FieldState fs(d,1,lim);
            fs=fs.crt(psi);
            TrjDMI_FilOpr trj2(fop,fs,tc);

            Vp<TrjBase> vtrj(2);
            vtrj[1]=trj1;
            vtrj[2]=trj2;
            Trj2 trj(vtrj);

            trj.run(g,out4(),out5());
            cpmwait(tw3,2);
         }

         else if (modeOfDyn==9){
            V<R> shiftRel{-0.2,0,0.15,0.3 };
            Z nShift=shiftRel.dim();
            V<State> states(nShift);

            for (i=1;i<=nShift;++i){
               R si=QMBase::getSize()*shiftRel[i];
               states[i]=psi.shift(si);
            }

            cpmmessage("states made");
            cpmdebug(nShift);

            Z lim1=1;
            Z lim2=1024;

            FieldState fs1(d,1,lim1);
            for (i=1;i<=nShift;++i){
               fs1=fs1.crt(states[i]);
            }
            fs1.nor_();
            cpmmessage("fs1 done");

            TrjDMI_FilOpr trj1(fop,fs1,tc);

            FieldState fs2(d,1,lim2);
            for (i=1;i<=nShift;++i){
               fs2=fs2.crt(states[i]);
            }
            fs2.nor_();
            cpmmessage("fs2 done");
            TrjDMI_FilOpr trj2(fop,fs2,tc);
         //  TrjDMI_Opr trj2(op,states[1],tc);

            Vp<TrjBase> vtrj(2);
            vtrj[1]=trj1;
            vtrj[2]=trj2;
            Trj2 trj(vtrj);

            trj.run(g,out4(),out5());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==10){
            V<R> shiftRel{-0.05,0.05 };
            Z nShift=shiftRel.dim();
            V<State> states(nShift);

            for (i=1;i<=nShift;++i){
               R si=QMBase::getSize()*shiftRel[i];
               states[i]=psi.shift(si);
            }
            cpmmessage("states made");
            Z lim1=10;
            FieldState fs1(d,1,lim1);
            for (i=1;i<=nShift;++i){
               fs1=fs1.crt(states[i]);
            }
            fs1.nor_();
            TrjDMI_FilOpr trj(fop,fs1,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==11){
            R jitter=0.1;
            R omega=1e9;
            R_Func f1=R_Func::x4*16;
#ifdef CPM_Fn
            R2_Func f2=F3<C,R_Func,R,R,R>(f1,jitter,omega)(jitterFunc);
#else
            R2_Func f2=F<C,R>(bind(jitterFunc,_1,f1,jitter,omega));
#endif
            f2*=lambda0*kp;
            op=Operator::sinParHam(f2,d);
            TrjDMI_Opr trj(op,psi,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else if (modeOfDyn==12){
          //  g.clr(BLACK);
            cpmmessage("modeOfDyn==12 started");
            Operator hf=Operator::freePar(d);
            SpectralObject hole("x",d);
            Iv ivHole(-0.45,-0.4);
            R_Func fHole=R_Func::charFunc(ivHole);
            hole.map_(fHole);
            Z n1,n2,n3;
            sec="system data";
            cpmrhf(n1);
            cpmrhf(n2);
            cpmrhf(n3);
            psi.setBra_(n1,n2,n3);
            Trj2DMI<State,Operator,SpectralObject> trj(psi,hf,hole,tc);
            trj.run(g,out4());
            cpmwait(tw3,2);
         }
         else{
            cpmmessage(loc&": nothing more to do");
         }
      }
   }
   CPM_MZ
}

/////////////////////////////// App3 //////////////////////////////
// not documented

namespace{
   R gauss(const R& x, const R& x0, const R& sigma)
   {
      R y=(x-x0)/sigma;
      return cpmexp(-y*y);
   }
}

void CpmVQM2::App3::doTheWork()
{
   Z mL=2;
   Word loc("CpmVQM2::App3::doTheWork()");
   CPM_MA

   Z n,lim;
   B twoParticles,useSI,useFockSpace;
   R mass,phiStep,x1,x2,sigma1,sigma2,v1Rel,v2Rel;
   R lambda1,lambda2,size;
   Word sec("system data");

   rch.read(sec,"dimension",n);
   rch.read(sec,"phiStep",phiStep);
   rch.read(sec,"useSI",useSI);
   rch.read(sec,"size",size);
   rch.read(sec,"mass",mass);
   rch.read(sec,"lim",lim);
   rch.read(sec,"x1",x1);
   rch.read(sec,"x2",x2);
   rch.read(sec,"sigma1",sigma1);
   rch.read(sec,"sigma2",sigma2);
   rch.read(sec,"v1Rel",v1Rel);
   rch.read(sec,"v2Rel",v2Rel);
   rch.read(sec,"lambda1",lambda1);
   rch.read(sec,"lambda2",lambda2);
   rch.read(sec,"useFockSpace",useFockSpace);
   rch.read(sec,"twoParticles",twoParticles);

   if (useSI) QMBase::setUnits("SI");
   else QMBase::setUnits("1");
   QMBase::setSize(size);
   QMBase::setMass(mass);
   QMBase::setPhiStep(phiStep);
   Frame fr;

#ifdef CPM_Fn
   F<R,R> shape1=F2<R,R,R,R>(x1,sigma1)(gauss);
#else
   F<R,R> shape1(bind(gauss,_1,x1,sigma1));
#endif
   CpmFunctions::R_Func f1(shape1);
   State psi1(f1,n);
   psi1.nor_();

#ifdef CPM_Fn
   F<R,R> shape2=F2<R,R,R,R>(x2,sigma2)(gauss);
#else
   F<R,R> shape2(bind(gauss,_1,x2,sigma2));
#endif
   CpmFunctions::R_Func f2(shape2);

   State psi2(f2,n);
   psi2.nor_();

   Observable boost(Word("n"),n);
   R vRef=QMBase::vMax(n);

   if (v1Rel!=0){
      R v1=v1Rel*vRef;
      psi1=psi1.boost(v1);
   }

   if (twoParticles && v2Rel!=0){
      R v2=v2Rel*vRef;
      psi2=psi2.boost(v2);
   }

   Observable E(Word("eKinF"),n);
   R lambdaRef=QMBase::lambdaOsc(n);
   if (lambda1!=0){
      Observable x(Word("x"),n);
      R lambdaE=lambda1*lambdaRef;
      Observable V=x*x*lambdaE;
      E+=V;
   }

   Record rtc=(rch.getRecord())("run control","TrjCon");
   TrjCon tc(rtc);

   OFileStream out4(outFiles[4]);

   Graph g;

   TrjDMI_Obs trj1(E,psi1,tc);

   if (!useFockSpace){
      trj1.run(g,out4());
   }
   else{
      OFileStream out5(outFiles[5]);

      FieldOperator fop=FieldOperator::cook1(E);
      FieldState fs(n,1,lim);
      fs=fs.crt(psi1);
      if (!twoParticles){
         TrjDMI_FilOpr trj2(fop,fs,tc);
         Vp<TrjBase> vtrj(2);
         vtrj[1]=trj1;
         vtrj[2]=trj2;

         Trj2 trj(vtrj);
         trj.run(g,out4(),out5());
      }
      else{
         fs=fs.crt(psi2);
         if (lambda2!=0.){
            R_Func x2=R_Func::x2;
            R lambdaU=lambda2*lambdaRef;
            Observable U=Observable::pot2(x2,n,lambdaU);
            fop+=FieldOperator::cook2(U);
         }
         TrjDMI_FilOpr trj2(fop,fs,tc);
         trj2.run(g,out5());
      }
   }
   CPM_MZ
}

/////////////////////////////// App4 //////////////////////////////
// scattering

namespace{
   R clbFnc(R const& x, R const& r, R const& fac)
   {
      return fac/cpmsqrt(x*x+r*r);
   }
} // namespace

void CpmVQM2::App4::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App4::doTheWork()");
   CPM_MA

   B useSI, runFromFile;
   Z d,nPar=0;
   R s,mass,lam1Rel,lam2Rel,xRel,wRel,bstRel,
      gamma,showPsi,phiStep;
   V<R> waitTimes;
   V<Z> statistics;

   Word file;
   V<Word> line1;
   V<Word> line2;
   V<Word> line3;
   V<Word> line4;

// reading data
// I choose to read all data in succession although for most settings
// in run_contol only a part of these data is needed in the code
// to be executed furtheron. Letting each control path read only those
// data that it really needs would create complicated interdependencies
// of the various control paths, resulting in the need for doublicating
// code. This also would complicate further extensions.

////////////////////////////////////////////////////////////////////////
   Word sec="about";
////////////////////////////////////////////////////////////////////////
   cpmrh(file);
   cpmrh(line1);
   cpmrh(line2);
   cpmrh(line3);
   cpmrh(line4);

////////////////////////////////////////////////////////////////////////
   sec="run preparation";
////////////////////////////////////////////////////////////////////////
   cpmrh(waitTimes);
   cpmrh(gamma);
   cpmrh(showPsi);
   cpmrh(phiStep);
   cpmrh(statistics);
   cpmrh(runFromFile);

////////////////////////////////////////////////////////////////////////
   sec="run control";
////////////////////////////////////////////////////////////////////////
   B twinProcessing;
   Z viewingPeriod=2, imagingPeriod=100;
   cpmrhf(twinProcessing);
   cpmrhf(viewingPeriod);
   cpmrhf(imagingPeriod);

////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////
   V<R> shifts1;
   V<R> shifts2;

   cpmrh(d);
   cpmrh(s);
   cpmrh(mass);
   cpmrhf(nPar);
   cpmrhf(shifts1);
   cpmrhf(shifts2);
   cpmrh(useSI);
   cpmrh(lam1Rel);
   cpmrh(lam2Rel);

   cpmrh(xRel);
   cpmrh(wRel);
   cpmrh(bstRel);

/////////////////////////////////////////////////////////

// setting the stage

   Color::setGamma(gamma);
   Graph g;
   g.setWithOrigin(0);
   Z i;
   if (useSI) QMBase::setUnits("SI");
   else QMBase::setUnits("1");
   QMBase::setMass(mass);
   QMBase::setSize(s);
   QMBase::setPhiStep(phiStep);
   Iv ivi(-0.5,0.5);
   Iv ivf(-s*0.5,s*0.5);
   R pMax=QMBase::pMax(d);
   R vMax=QMBase::vMax(d);
   R eMax=pMax*pMax/(2*mass);

   R tw1=waitTimes[1];

   Z sP=512;

// defining the FieldOperator fop
   // first the confining potential ~ x^8

   R fac1=eMax*256*lam1Rel;
      // 256 is the reciprocal value of the maximum value
      // of the un-scaled confining potential
   R_Func pot1=R_Func::x4*R_Func::x4*fac1; // ~x^8 rather flat
      // till close to the rims of the domain
   // second the pair potential
   R r0=s/d;
   R fac2=eMax*r0*lam2Rel;
      // r0 is the reciprocal value of the maximum value of the un-scaled
      // Coulomb potential
#ifdef CPM_Fn
   R_Func pot2=F2<R,R,R,R>(r0,fac2)(clbFnc);
#else
   R_Func pot2=F<R,R>(bind(clbFnc,_1,r0,fac2));
#endif

   R2_Func p1(pot1);
   R2_Func p2(pot2);
   // and now the corresponding many-particle Hamiltonian:
   cpmcerr<<"*** fop ahead"<<endl;
   FieldOperator fop=FieldOperator::manParHam(p1,p2,d);
   cpmcerr<<"*** fop done"<<endl;
 // initial state:
   bool unitHeg=true; // aha heg = height
   R_Func gpsi1=R_Func::Gauss(xRel,wRel,unitHeg);
   R_Func gpsi2=R_Func::Gauss(-xRel,wRel,unitHeg);
   if (showPsi>0){
      g.setX(ivi);
      g.addText("Continuous version of state psi1+psi2");
      g.show(gpsi1+gpsi2,sP);
      cpmwait(showPsi);
   }
   State psi1(gpsi1,d);
   State psi2(gpsi2,d);
   R v1=bstRel*vMax;
   // so negative bstRel makes the particles move together
   R v2=-v1;
   psi1=psi1.boost(v1);
   psi2=psi2.boost(v2);

   Record rtc=(rch.getRecord())("run control","TrjCon");
      // Here the section 'run control' comes in and is used to build
      // a TrjCon
   rtc.prnOn(cpmdata);
   TrjCon tc(rtc);

   Z nSteps=tc.nSteps_;
   cpmdebug(nSteps);
   Z sd1=shifts1.dim(),sd2=shifts2.dim();
   Z sds=sd1+sd2;
   V<State> states;
   if (nPar==0 || sds==0 ){ // taking the two prepared states
      states=V<State>(2);
      states[1]=psi1;
      states[2]=psi2;
   }
   else{ // making more states by translation
      nPar=sd1+sd2;
      states=V<State>(nPar);
      for (i=states.b();i<=states.e();++i){
         if (i<=sd1) states[i]=psi1.shift(shifts1[i]);
         else states[i]=psi2.shift(shifts2[i-sd1]);
      }
   }
   Z sd=statistics.dim();
      // This is allowed to be larger than 1; the two runs are done in
      // parallel for method comparison

// creating dynamical state evolution, not reading it from file
   if (!runFromFile){
      Z lim1=statistics[1];
      cpmdebug(lim1);
      FieldState fs1(d,1,lim1); // Fock vaccum
      for (i=states.b();i<=states.e();++i) fs1=fs1.crt(states[i]);
         // creating a multi-particle state by sequential application
         // of creation operators
      fs1.nor_();
      cpmmessage("state fs1 constructed");
      Graph gDebug;
      fs1.mark(gDebug);
      gDebug.vis();
      cpmwait(8,2);

      cpmmessage("trj1 constructed");
      if (sd==1){ // usual case if a single system (many particles,
         // though) is to be run
         OFileStream out1(outFiles[1]);
         TrjDMI_FilOpr trj1(fop,fs1,tc);
         cpmmessage(
         "sd==1 selected, TrjDMI_FilOpr::trj1.run(g,out1()) ahead");
         trj1.run(g,out1());
         cpmwait(tw1);
      }
      else if (sd==2) {
         cpmmessage("sd==2 selected");
            // we have fs1 from above
         Z lim2=statistics[2];
         FieldState fs2(d,1,lim2);
         for (i=states.b();i<=states.e();++i) fs2=fs2.crt(states[i]);
         fs2.nor_();
         cpmmessage("state fs2 constructed");
         if (twinProcessing){
            // more explicit form of trajectory construction
            cpmcerr<<"entered the twinProcessing loop"<<endl;
            R t=0.;
            R dt=tc.dt_;
            R tau=dt*0.5;
            FieldState v1=fop.actOn(fs1,t)*C::Ibar;
            FieldState v2=fop.actOn(fs2,t)*C::Ibar;
               // leapfrog dynamics needs velocities in addition to the
               // states.
            R gamma=tc.gam_;
            Z intMet=tc.intMet_;
            Color::setGamma(gamma);
            Frame fr0;
            Frames frs(fr0,Z(1),Z(2));
            Graph g1(frs(1,1));
            Graph g2(frs(1,2));
               // preparing twin graphics
            CyclicAlarm cview(viewingPeriod);
               // deals with showing screen images
            CyclicAlarm cimg(imagingPeriod);
               // deals with writing image files (in ppm format)
            bool bv=viewingPeriod>0;
            bool bi=imagingPeriod>0;

            cpmdebug(nSteps);
            cpmdebug(dt);
            cpmdebug(viewingPeriod);
            cpmdebug(imagingPeriod);
               // to reliably show which input values vere used. It is not
               // too unlikely that one uses a wrong ini-file!

            if (bv){
               Graph g1(frs(1,1));
               Graph g2(frs(1,2));
               fs1.mark(g1,intMet,gamma);
               fs2.mark(g2,intMet,gamma);
               frs.toFrm().vis(bi);
            } // visualizing the initial state
            cpmcerr<<" dynamic loop ahead"<<endl;
            R nStepsInv=R(1.)/nSteps;
            for (i=1;i<=nSteps;++i){ // time discrete trajectory of
               cpmprogress("dyn. loop",(i-1)*nStepsInv);
               // two states
// the leapfrog step is so simple and symmetric that it needs not be
// converted into a black box
               t+=tau;
               fs1+=v1*tau;
               fs2+=v2*tau;
               v1+=(fop.actOn(fs1,t)*C::Ibar - v1)*2_R;
               v2+=(fop.actOn(fs2,t)*C::Ibar - v2)*2_R;
               fs1+=v1*tau;
               fs2+=v2*tau;
               t+=tau;
                  // end of leapfrog step
// Graphical representation. Here we need states from two trajectories at
// the same time. Unfortunately this cannot be achieved by constructing two
// trajectories of type TrjDMI_FilOpr ( unless one one would modify them in
// a way that they would be able to create series of images as output).

               bool bview=cview();
               bool bimg=cimg();
               if (bview||bimg){
                  Graph g1(frs(1,1));
                  Graph g2(frs(1,2));
                  fs1.mark(g1,intMet,gamma);
                  fs2.mark(g2,intMet,gamma);
                  frs.toFrm().vis(bimg);
               }
            }
// making a gif file by means of GraphicsMagick
            Z returnOfSystem1=chdir(-Message::getOutDir2());
               // chdir ist ein gueltiger C++ Befehl, keine boot-library
               // needed for this (requires #include <unistd.h>)
            Word command("gm convert *.ppm fermi_bose_");
            command&=cpm(d);
            command&=".gif";
            Z returnOfSystem2=system(-command);
            Z returnOfSystem3=chdir("../"); // highly important to go back!
            cpmdebug(returnOfSystem1);
            cpmdebug(returnOfSystem2);
            cpmdebug(returnOfSystem3);
         }
         else {
         cpmcerr<<"entered the !twinProcessing loop"<<endl;
            OFileStream out1(outFiles[1]);
            TrjDMI_FilOpr trj1(fop,fs1,tc);
            OFileStream out2(outFiles[2]);
            TrjDMI_FilOpr trj2(fop,fs2,tc);
            Vp<TrjBase> trj(2);
            trj[1]=trj1;
            trj[2]=trj2;
            Trj2 trj_2(trj);
            trj_2.run(g,out1(),out2());
            cpmwait(tw1);
         }
      }
      else{
         cpmerror("statistics must have one or two entries");
      }
   }
// showing the dynamical state evolution from file
// This is the modality to be used when a series of images is needed
// in order to make a gif-file.
   else{
      Graph g2;
      if (tc.disMode_>=3 && tc.imgPer_>0){
         Frame fr1;
         R ari=fr1.getAspRatInv();
         R ySub=0.45; // 0<ySub<=1, allows to have the images not too
         // large
         Frame fr2(0.1,0.1,ySub*ari,ySub);
            // we make the image frame smaller in order not to overload
            // ImageMagick::convert in making a gif-file
         g2=Graph(fr2);
      }
      FieldState fsIrr;
      FieldOperator fopIrr;
      IFileStream in1(inFiles[1]);
      TrjDMI_FilOpr trj1(fopIrr,fsIrr,tc);
      if (sd==1){
         trj1.play(g2,in1());
      }
      else if (sd==2) {
         IFileStream in2(inFiles[2]);
         TrjDMI_FilOpr trj2(fopIrr,fsIrr,tc);
         Vp<TrjBase> trj(2);
         trj[1]=trj1;
         trj[2]=trj2;
         Trj2 trj_2(trj);
         trj_2.play(g2,in1(),in2());
      }
      else{
         cpmerror("statistics must have one or two entries");
      }
   }
   CPM_MZ
}

/////////////////////////////// App4c //////////////////////////////
// dividing the biotope

void CpmVQM2::App4c::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App4c::doTheWork()");
   CPM_MA

   R gamma,showPsi,phiStep;
   V<R> waitTimes;
   Z statistics;

   Word file;
   V<Word> line1;
   V<Word> line2;
   V<Word> line3;
   V<Word> line4;

////////////////////////////////////////////////////////////////////////
   Word sec="about";
////////////////////////////////////////////////////////////////////////
   cpmrh(file);
   cpmrh(line1);
   cpmrh(line2);
   cpmrh(line3);
   cpmrh(line4);

////////////////////////////////////////////////////////////////////////
   sec="run preparation";
////////////////////////////////////////////////////////////////////////
   cpmrh(waitTimes);
   cpmrhf(gamma);
   cpmrhf(showPsi);
   cpmrhf(phiStep);
   cpmrh(statistics);

////////////////////////////////////////////////////////////////////////
   sec="run control";
////////////////////////////////////////////////////////////////////////

   Z viewingPeriod=2, imagingPeriod=100, docuPeriod=2, intMet=0;
   R dtRel, dt;
   Z method=0; // integration method 0 = midpoint, -3 = DALF
   B dtAbsolute;
   cpmrhf(viewingPeriod);
   cpmrhf(imagingPeriod);
   cpmrhf(docuPeriod);
   cpmrhf(dt);
   cpmrhf(dtRel);
   cpmrhf(method);
   cpmrhf(dtAbsolute);
   cpmrhf(intMet);

////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////

   Z d;
   R mass, lam1Rel, lam2Rel, sigmaWell, x1, x2, xPic, facPic;
   R sigma1, sigma2, v1Rel, v2Rel;
   B fock;

   cpmrh(d);
   cpmrh(mass);
   cpmrh(lam1Rel);
   cpmrh(lam2Rel);
   cpmrh(sigmaWell);

   cpmrh(x1);
   cpmrh(x2);
   cpmrhf(xPic);
   cpmrhf(facPic);

   cpmrh(sigma1);
   cpmrh(sigma2);
   cpmrh(v1Rel);
   cpmrh(v2Rel);

   cpmrhf(fock);

/////////////////////////////////////////////////////////

// setting the stage

   Color::setGamma(gamma);
   Graph gr;
   gr.setWithOrigin(0);
   Z i;

   R s=2.;

   QMBase::setUnits("1");
   QMBase::setMass(mass);
   QMBase::setSize(s);
   QMBase::setPhiStep(phiStep);
   R pMax=QMBase::pMax(d);
   R vMax=QMBase::vMax(d);
   R eMax=pMax*pMax/(2*mass);
   Iv bio=QMBase::getBio();

   R tw1=waitTimes[1];

   Z sP=512;
   bool unitHeg=true; //heg = height

// defining the FieldOperator fop
   // first the confining potential ~ x^8

   R fac1=eMax*lam1Rel;
   R fac2=eMax*lam2Rel;

   R_Func pot1=R_Func::x8*fac1; // ~x^8 rather flat
   R2_Func p(pot1); // tensorproduct pot1 times pot1

   R xw=0.; // wall is in the middle
   R_Func w1=R_Func::Gauss(xw,sigmaWell,unitHeg)*fac2;
   R2_Func w(w1);

 // and now the corresponding many-particle Hamiltonian:
 // FieldOperator fop=FieldOperator::manParHam_w(p,w,tf,d);
 // initial state:

   R_Func gpsi1=R_Func::Gauss(x1,sigma1,unitHeg)*eMax;
   R_Func gpsi2=R_Func::Gauss(x2,sigma2,unitHeg)*eMax;
   cpmassert(x1 <= x2,loc);
   if (showPsi>0){
      gr.setX(bio);
      gr.addText("Continuous version of state psi1+psi2");
      gr.show(pot1+gpsi1+gpsi2+w1,sP);
     // g.show(pot1+w1,sP);
      cpmwait(showPsi,2);
   }
   State psi1(gpsi1,d);
   State psi2(gpsi2,d);
   R v1=v1Rel*vMax;
   R v2=v2Rel*vMax;
   psi1=psi1.boost(v1);
   psi2=psi2.boost(v2);

 //  Record rtc=(rch.getRecord())("run control","TrjCon");
          // Here the section 'run control' comes in and is used to build
          // a TrjCon
//   rtc.prnOn(cpmdata);
//   TrjCon tc(rtc);
 //  R dt=tc.dt_;
   R tPic=(xPic-x2)/v2;

   R tEnd=tPic*facPic;
   R level1=0.;
   R level2=1.;
   R_Func tf=R_Func::smoothTransit(R2(tPic,tEnd),R2(level1,level2));

   State2Data data;
   data.m1=mass;
   data.m2=mass;

   data.size1=s;
   data.size2=s;
   data.box1=1;
   data.box2=1;
   data.alpha=0;
   data.beta=0.;
   data.mu=0.;
   data.eps=0.;
   data.p=0.;
   data.crossway=B(false);
   data.barrierForm=B(true);
   data.autoStep=false;
   data.f=pot1;
   data.g=w1;
   data.ft=tf;

 // creating dynamical state evolution

  Z lim1=statistics;

if (!fock){ // the first comparison of the two methods (fock true and
// false) gave virtually identical runtimes: 1472.62 s for fock and
// 1475.39 for !fock.
   State2::set_d(data);

   State2 s2(psi2,psi1); // now s1 and s2 are no longer needed as
      // State-instances
   State2 s12(psi1,psi2);

   if (lim1==1) s2-=s12;
   if (lim1>1)  s2+=s12;

   s2.nor_();

   bool rel=!dtAbsolute;
   if (dtAbsolute) dtRel=dt;
   R ti=0.;
   R tTest=tEnd;
   State2Trj st(s2,dtRel,rel,tTest);
    // now s2 is no longer needed. st is the active object.
   st.set_t(ti);
   st.autoStep_=false;
   st.stepMet_=method;
   dt=st.dt();
   cout<<"dt = "<<dt<<endl;
   Z nPic=cpmtoz(tPic/dt);
   cout<<"nPic = "<<nPic<<endl;
   Z nSteps=cpmtoz(tEnd/dt);
   cout<<"nSteps = "<<nSteps<<endl;

   RunControl rc;
   rc.tTot=tEnd;
   rc.gamma=gamma; // from section run preparation

   R docuGap=docuPeriod*dt;
   Z nDoc=cpmtoz(R(rc.tTot)/docuGap)+1;
   rc.nDoc=nDoc;

   R viewGap=viewingPeriod*dt;
   Z nView=cpmtoz(R(rc.tTot)/viewGap)+1;
   rc.nView=nView;

   R imgGap=imagingPeriod*dt;
   Z nImg=cpmtoz(R(rc.tTot)/imgGap)+1;
   rc.nImg=nImg;

   rc.ipolMethod=intMet;
   st.run_(rc,gr);
}
else{

   V<State> states{psi1,psi2};

  FieldState fs(d,1,lim1); // Fock vaccum
  for (i=states.b();i<=states.e();++i) fs=fs.crt(states[i]);
         // creating a multi-particle state by sequential application
         // of creation operators
  fs.nor_();
  cpmmessage("state fs constructed");
          // more explicit form of trajectory construction
  R t=0.;
  R tau=dt*0.5;
  R tauh=tau*0.5;
  FieldOperator fop=FieldOperator::manParHam_w(p,w,tf,d);
  FieldState v=fop.actOn(fs,t)*C::Ibar;
     // leapfrog dynamics needs velocities in addition to the states.

  Graph g1;
               // preparing twin graphics
  CyclicAlarm cview(viewingPeriod);
  CyclicAlarm cimg(imagingPeriod);
  bool bv=viewingPeriod>0;
  bool bi=imagingPeriod>0;

  Z nSteps=cpmtoz(tEnd/dt);

  cpmdebug(nSteps);
  cpmdebug(dt);
  cpmdebug(viewingPeriod);
  cpmdebug(imagingPeriod);
     // to reliably show which input values vere used. It is not
     // too unlikely that one uses a wrong ini-file!

   if (bv){
      fs.mark(g1,intMet,gamma);
       g1.vis(bi);
   } // visualizing the initial state

   R nStepsInv=R(1.)/nSteps;

   for (i=1;i<=nSteps;++i){
// the leapfrog step is so simple and symmetric that it needs not be
// converted into a black box
      cpmprogress("",nStepsInv*i);
      t+=tauh;
      fs+=v*tauh;
      v+=(fop.actOn(fs,t)*C::Ibar - v)*2.;
      fs+=v*tau;
      t+=tau;
      v+=(fop.actOn(fs,t)*C::Ibar - v)*2.;
      fs+=v*tauh;
      t+=tauh;
   // end of leapfrog(DALF) step
   // Graphical representation.
      bool bview=cview();
      bool bimg=cimg();
      if (bview||bimg){
         fs.mark(g1,intMet,gamma);
         g1.vis(bimg);
      }
    } // steps loop

}
    CPM_MZ
}

/////////////////////////////// App4b //////////////////////////////
// Hubbard model

void CpmVQM2::App4b::doTheWork()
{
   Z mL=1;
   Word loc("CpmVQM2::App4b::doTheWork()");
   CPM_MA

   Z statistics=1,d,nPar;
   R gamma,showPsi,aHub=-1,bHub=1;
   B pbc{true}; // pbc = periodic boundary conditions
   V<R> waitTimes;

   Word file;
   V<Word> line1;
   V<Word> line2;
   V<Word> line3;
   V<Word> line4;

// reading data
// I choose to read all data in succession although for most settings
// in run_control only a part of these data is needed in the code
// to be executed further on. Letting each control path read only those
// data that it really needs would create complicated interdependencies
// of the various control paths, resulting in the need for duplicating
// code. This also would complicate further extensions.

////////////////////////////////////////////////////////////////////////
   Word sec="about";
////////////////////////////////////////////////////////////////////////
   cpmrh(file);
   cpmrh(line1);
   cpmrh(line2);
   cpmrh(line3);
   cpmrh(line4);

////////////////////////////////////////////////////////////////////////
   sec="run preparation";
////////////////////////////////////////////////////////////////////////
   cpmrh(waitTimes);
   cpmrh(gamma);
   cpmrh(showPsi);
   cpmrh(statistics);

////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////

   cpmrhf(d);
   cpmrhf(nPar);
   cpmrhf(aHub);
   cpmrhf(bHub);
   cpmrhf(pbc);

/////////////////////////////////////////////////////////

// setting the stage

   Color::setGamma(gamma);
   Graph g;
   g.setWithOrigin(0);
   Z i;
   R tw1=waitTimes[1];

 // defining the FieldOperator fop

   FieldOperator fop=FieldOperator::hub(aHub,bHub,d);
   Record rtc=(rch.getRecord())("run control","TrjCon");
          // Here the section 'run control' comes in
   rtc.prnOn(cpmdata);
   TrjCon tc(rtc);

   Z nSteps=tc.nSteps_;
   Z lim=statistics;
      // 1 for fermions, >1 for bosons
   Z test=16;
   R testRes=FieldState::test(test);
   R eps=1e-8;
   if (testRes>eps) cpmerror("There were errors in test");
   State s0(d);
   V<State> vs(nPar);
   for (i=vs.b();i<=vs.e();++i) vs[i]=s0.ran();
   FieldState fs(vs,lim,pbc);
   OFileStream out1(outFiles[1]);
   TrjDMI_FilOpr trj(fop,fs,tc);
   tc.prnOn(cpmcerr);
   trj.run(g,out1());
   cpmwait(tw1,2);
   CPM_MZ
}

////////////////////// App5 ////////////////////////////////////////////

void CpmVQM2::App5::doTheWork()
// First experiences with the new class State2 (started 2006-07)
// Now (2006-09) grown into the most advanced experimental version for
// new research features such as time reversal and 'optical theorem'
{
   Z mL=1;
   Word loc("CpmVQM2::App5::doTheWork()");
   CPM_MA
//////////////////////////////////////////////////////////////////////
   Word sec("run control");
//////////////////////////////////////////////////////////////////////
   Z nSteps,iReversal,reversalMethod,nSmooth,nConcentrate,viewingPeriod,
      docuPeriod,imagingPeriod,ipolMethod=1,optThr=0;
   R tWait=0,gamma=1;
   cpmrh(nSteps);
   cpmrh(iReversal);
   cpmrh(reversalMethod);
   cpmrh(nSmooth);
   cpmrh(nConcentrate);
   cpmrh(viewingPeriod);
   cpmrh(docuPeriod);
   cpmrh(imagingPeriod);
   cpmrh(tWait);
   cpmrh(ipolMethod);
   cpmrh(optThr);
   cpmrh(gamma);

//////////////////////////////////////////////////////////////////////
   sec="system data";
//////////////////////////////////////////////////////////////////////
   Z i;
   Z d1=60,d2=60,stepMet=3;
   cpmrh(d1);
   cpmrh(d2);
   R size1=1.,size2=1.,massRel1,massRel2,vRel1,vRel2;
   R dtRel=0, xRel1=0, xRel2=0, wRel1=0, wRel2=0;
   R enrMagFac=1e9;
   cpmrh(xRel1);
   cpmrh(xRel2);
   cpmrh(wRel1);
   cpmrh(wRel2);
   B autoStep,enfNor,useSI,noRel,crossway{true};
   cpmrh(size1);
   cpmrh(size2);
   cpmrh(noRel);
   cpmrh(massRel1);
   cpmrh(massRel2);
   cpmrh(vRel1);
   cpmrh(vRel2);
   R alpha,beta,mu,eps,p;
   cpmrh(alpha);
   cpmrh(beta);
   cpmrh(mu);
   cpmrh(eps);
   cpmrh(p);
   cpmrh(useSI);
   cpmrh(dtRel);
   cpmrh(enfNor);
   cpmrh(autoStep);
   cpmrh(crossway);
   cpmrh(enrMagFac);
   cpmrh(stepMet);
// start working
   Word units=useSI ? "SI":"1";
   OFileStream out1(outFiles[1]);
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App5::doTheWork()"<<endl;
   out1()<<"units="<<units;
   QMBase::setUnits(units);

   QMBase::setSize(size1);
      // to get the right lattice related data for particle 1
   R m1= noRel ? massRel1 : massRel1*QMBase::mMax(d1);

   QMBase::setSize(size2); // to get the right output from mMax
   R m2=noRel ? massRel2 : massRel2*QMBase::mMax(d2);

   State2Data data;
   data.m1=m1;
   data.m2=m2;
   data.size1=size1;
   data.size2=size2;
   data.alpha=alpha;
   data.beta=beta;
   data.mu=mu;
   data.eps=eps;
   data.p=p;
   data.crossway=crossway;

   State2::set_d(data);

   // initial state:
   B unitHeg{true}; // aha heg = height
   R_Func gpsi1=R_Func::Gauss(xRel1,wRel1,unitHeg);
   R_Func gpsi2=R_Func::Gauss(xRel2,wRel2,unitHeg);

   State s1(gpsi1,d1);
   State s2(gpsi2,d2);
// That boosting instances of State correctly, the class
// QMBase has to be adjusted, needs e.g. member mass
   QMBase::setMass(m1);
   QMBase::setSize(size1);
   R v1=noRel ? vRel1 : vRel1*0.5*QMBase::vMax(d1);
   out1()<<"v1="<<v1<<endl;
      // 0.5 inserted since without this, vRel1=1
      // corresponds to a wave function mainly at
      // Nyquist frequency. Our representation of the Hamilton
      // operator is not very suitable for such strongly
      // varying wave functions
   s1=s1.boost(v1);
 //  s1.reg_(); // no significant influence
   Observable T1(Word("eKinF"),d1);
   R enr1=T1.expVal(s1);
   out1()<<"enr1="<<enr1<<endl;

   QMBase::setMass(m2);
   QMBase::setSize(size2);
   R v2=noRel ? vRel2 :  vRel2*0.5*QMBase::vMax(d2);
   out1()<<"v2="<<v2<<endl;
   s2=s2.boost(v2);
 //  s2.reg_();
   Observable T2(Word("eKinF"),d2);
   R enr2=T2.expVal(s2);
   out1()<<"enr2="<<enr2<<endl;

   Graph gr;
   Z sym=1;
   cpmrh(sym);
   cpmrh(crossway);
      // 0 for non-identical particles
      // -1 for fermions
      // else for bosons
   if (d1!=d2){
      if (sym!=0){ // would result in index errors
         sym=0;
         cpmwarning("sym changed to 0 since d1!=d2");
         cpmwait(1);
      }
   }
   State2 s(s1,s2);
   if (sym!=0){
      State2 sSym(s2,s1);
      if (sym==-1) sSym*=-1;
      s+=sSym;
      s.nor_(1);
   }
   // now s1 and s2 are no longer needed as State-instances
   State2Trj st(s);
   st.revMet_=reversalMethod;
   for (i=0;i<nConcentrate;++i){
      st.concentrate_();
      st.x().mark(gr,ipolMethod, gamma);
      gr.vis();
      cpmwait(tWait+2);
   }
   st.smooth_(3);
   State2Trj stIni(st);
   C enr3=st.enr();
   out1()<<"complex en3 ="<<enr3<<endl;
   R enrInt=enr3.real()-enr1-enr2;
   out1()<<"enrInt ="<<enrInt<<endl;
   out1()<<"This should valish for alpha==beta==0"<<endl;
   R dt0=noRel ? dtRel : dtRel*st.dtLimit();
  // R epsDifferentiation=1e-6;
   R epsDifferentiation=1e-1;
   R dtDiff=dt0*epsDifferentiation;
   out1()<<"dt0="<<dt0<<endl;
   R tTot=nSteps*dt0;
   R turns=tTot*v1/size1;
   out1()<<"turns="<<turns<<endl;
   Z nDoc=docuPeriod>0 ? cpmrnd(1.*nSteps/docuPeriod) : 0;
   R_Matrix docMat(5,nDoc);
   R_Matrix doc1Mat(3,nSteps);
   R_Matrix doc2Mat(3,nSteps);
   R_Matrix doc3Mat(3,nSteps);

   CyclicAlarm mark(viewingPeriod);
   CyclicAlarm docu(docuPeriod);
   CyclicAlarm img(imagingPeriod);

   R t, er0, er0Inv;
   R nInv=nSteps<2 ? 1. : 1./(nSteps-1);
   Z iDoc=1, iImg=1;
   st.stepMet_=stepMet;
   st.enfNor_=enfNor;
   st.autoStep_=autoStep;
   R spAct=0,spPrev,nAct=0,nPrev=0,nPrevPrev=0,mr,nx,xx,spx;
   // mr is an interesting quantity, it is very small for regular
   // time steps and explodes if the norm and the energy do so. Since it
   // involves data from three consecutive time steps its initialization
   // is a bit complicated.
   C xv;
   State2 sIni=st.x();
   Z k;
   R dt=(optThr==0? dt0 : dt0-dtDiff);
   Z kMax=(optThr==0? 1 : 3);
   for (k=1;k<=kMax;++k){
      st=stIni;
      st.set_dt_(dt);
      for (i=0,t=0.5*dt; i<nSteps; ++i,t+=dt){
         R ip=i*nInv;
         cpmprogress("loop",ip,1);
         nx=st.x().abs();
         xx=(st.x()|st.x()).real();
         xv=(st.x()|st.v());
         spx=(st.v()|st.x()).real();
         R dn=nx-1.;
         if (i==0){
            nPrevPrev=nx;
            mr=0;
         }
         else if (i==1){
            nPrev=nx;
            mr=0;
         }
         else if (i==2){
            nAct=nx;
            spAct=spx;
            mr=0;
         }
         else{
            nPrevPrev=nPrev;
            nPrev=nAct;
            nAct=nx;
            spPrev=spAct;
            spAct=spx;
         // mr=(nAct-nPrevPrev)*nPrev-2*dt*spPrev;
            mr=nAct-nPrevPrev-2*dt*spPrev; // best option
         // mr=nAct*nAct-nPrevPrev*nPrevPrev-4*dt*spPrev;
         }
         C ste=st.enr();
         R er=ste.real();
         if (i==0){
            er0=er;
            er0Inv=er0==0 ? R(1)  : 1./er0;
         }
         R der=(er-er0)*er0Inv;
         R ei=ste.imag()*er0Inv;
         cpmvalue("dn",dn,2);
         cpmvalue("der",der,3);
         cpmvalue("ei",ei,4);
         der*=enrMagFac; // for docu graphics
         if (i==iReversal) st.con_();
         if (mark()){
            st.x().mark(gr,ipolMethod, gamma);
            gr.vis();
            cpmwait(tWait);
         }
         if (docu()&& iDoc<=nDoc){
            docMat[1][iDoc]=t;
            docMat[2][iDoc]=dn;
            docMat[3][iDoc]=der;
            docMat[4][iDoc]=ei;
            docMat[5][iDoc]=mr;
            iDoc++;
         }
         if (img()){
            Word imf="app5_";
            gr.write(imf&cpm(iImg++,3));
         }
         if (optThr!=0){
            Z ipp=i+1;
            if (k==1){
               doc1Mat[1][ipp]=xx;
               doc1Mat[2][ipp]=xv.real();
               doc1Mat[3][ipp]=xv.imag();
            }
            else if (k==2){
               doc2Mat[1][ipp]=xx;
               doc2Mat[2][ipp]=xv.real();
               doc2Mat[3][ipp]=xv.imag();
            }
            else if (k==3){
               doc3Mat[1][ipp]=xx;
               doc3Mat[2][ipp]=xv.real();
               doc3Mat[3][ipp]=xv.imag();
            }
            else cpmerror("bad value "&cpm(k)&" of k");
         }
         st.step_();
      }
      if (nSmooth>0){
         for (i=0;i<nSmooth;++i){
            st.smooth_(1);
            st.x().mark(gr,ipolMethod, gamma);
            gr.vis();
            cpmwait(tWait+0.5);
         }
         cpmwait(tWaitWorkScreen_);
      }
      dt+=dtDiff;
   }

   if (optThr==0){
      State2 sFin=st.x();
      sFin*=-1.;
      sFin.con_();
      R diff=(sIni+sFin).norm();
      cpmmessage("diff="&cpm(diff));
      if (docuPeriod>0) gr.show(docMat);
   }
   else{ // curve which shows the validity of the 'optical theorem'
      R_Matrix docX(4,nSteps);
      R magnificationFactor=5e7;
      for (i=1;i<=nSteps;++i){
         R b=(doc2Mat[2][i]-doc2Mat[2][1])*(i-1)*dt0;
         R a1=doc1Mat[1][i]-doc1Mat[1][1];
         R a2=doc2Mat[1][i]-doc2Mat[1][1];
         R a3=doc3Mat[1][i]-doc3Mat[1][1];
         R a=dt0*(a3-a1)/(4*dtDiff)-a2;
         docX[1][i]=i;
        // docX[2][i]=a;
        // docX[3][i]=(a-b)*magnificationFactor;
         docX[2][i]=a1;
         docX[3][i]=a2;
         docX[4][i]=a3;
      }
      gr.show(docX);
   }
   cpmwait(tWaitWorkScreen_);
   CPM_MZ
}

void CpmVQM2::App6::doTheWork()
// Initial state of particle 1 is a Gaussian and
// initial state of particle 2 is eigenstate of
// of the harmonic oscillator built into State2.
// Aim is to see excitation or de-excitation of this
// state as a result of interaction with particle 1.
// This should be used for the image production for the publication.
// 2006-11-03: visualization of the 'reduced' states added.
{
   Z mL=1;
   Word loc("CpmVQM2::App6::doTheWork()");
   CPM_MA

// Getting input from ini file
////////////////////////////////////////////////////////////////////////
   Word sec("run control");
////////////////////////////////////////////////////////////////////////
   Z nSteps,viewingPeriod,docuPeriod,imagingPeriod,ipolMethod=1,
      iViewReducedState=1;
   R tWait=0,gamma=1,beginFadingOut,endFadingOut,
      tWaitDocMat1,tWaitDocMat2,tWaitDocMat3,magViewReducedState;
   V<Z> viewLevels;
   B showAsCurves;
   cpmrh(nSteps);
   cpmrh(beginFadingOut);
   cpmrh(endFadingOut);
   cpmrh(viewingPeriod);
   cpmrh(docuPeriod);
   cpmrh(viewLevels);
   cpmrh(showAsCurves);
   cpmrh(iViewReducedState);
   cpmrh(magViewReducedState);
   cpmrh(tWaitDocMat1);
   cpmrh(tWaitDocMat2);
   cpmrh(tWaitDocMat3);
   cpmrh(imagingPeriod);
   cpmrh(tWait);
   cpmrh(ipolMethod);
   cpmrh(gamma);

////////////////////////////////////////////////////////////////////////
   sec="system data";
////////////////////////////////////////////////////////////////////////

   Z d1, d2, stepMet=3, ePowers=0, box1=1, box2=1;

   R size1=1., size2=1., massRel1, massRel2, vRel1,
      alpha, beta, mu, eps, p, dtRel, xRel1, wRel1,
      misFac=1, tWaitH2=0, tWait_so=0, tWaitV12, enrMagFac=1,
      trnMagFac=1, trnVisExp=1, normTol=-1,
      phaseJitter1=0,phaseJitter2=0;
   V<Z> excLevels;

   B enfNor, useSI, crossway, autoStep;

   cpmrh(d1);
   cpmrh(d2);
   cpmrh(size1);
   cpmrh(size2);
   cpmrh(box1);
   cpmrh(box2);
   cpmrh(massRel1);
   cpmrh(massRel2);
   cpmrh(vRel1);
   cpmrh(xRel1);
   cpmrh(wRel1);
   cpmrh(excLevels);

   cpmrh(alpha);
   cpmrh(beta);
   cpmrh(mu);
   cpmrh(eps);
   cpmrh(p);
   cpmrhf(enfNor);
   cpmrhf(normTol);
   cpmrhf(phaseJitter1);
   cpmrhf(phaseJitter2);
   cpmrh(useSI);
   cpmrh(dtRel);
   cpmrh(ePowers);
   cpmrh(misFac);
   cpmrh(tWaitH2);
   cpmrh(tWait_so);
   cpmrh(tWaitV12);
   cpmrh(crossway);
   cpmrh(enrMagFac);
   cpmrh(trnMagFac);
   cpmrh(trnVisExp);
   cpmrh(stepMet);
   cpmrh(autoStep);
///////////////////// we are in App6::doTheWork() ////////////////////
// Starting the work
   OFileStream out1(outFiles[1]);
      // output during reading data and doing
      // preparing computations
   OFileStream out2(outFiles[2]);
      // data concerning nu and epsilon
   OFileStream out3(outFiles[3]);
      // data concerning reduced states
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App6::doTheWork()"<<endl;
   Word units=useSI ? "SI":"NAT";
   out1()<<"units="<<units;
   QMBase::setUnits(units);
   QMBase::setSize(size1);
      // to get the right lattice related data for particle 1
   out1()<<"size1="<<size1<<endl;
   R m1=massRel1*QMBase::mMax(d1);
   out1()<<"m1="<<m1<<endl;

   QMBase::setSize(size2); // to get the right output from mMax
   out1()<<"size2="<<size2<<endl;
   R m2=massRel2*QMBase::mMax(d2);
   out1()<<"m2="<<m2<<endl;

   State2Data data;
   data.m1=m1;
   data.m2=m2;
   data.size1=size1;
   data.size2=size2;
   data.box1=box1;
   data.box2=box2;
   data.alpha=alpha;
   out1()<<"alpha="<<alpha<<endl;
   data.beta=beta;
   out1()<<"beta="<<beta<<endl;
   data.mu=mu;
   data.eps=eps;
   data.p=p;
   data.crossway=crossway;
///////////////////// we are in App6::doTheWork() ////////////////////
// initial state
// Particle 1 has Gauss shape
   bool unitHeight=true; // avoids normalization factor
   R_Func gpsi0=R_Func::Gauss(0,wRel1,unitHeight);
   R_Func gpsi1=R_Func::Gauss(xRel1,wRel1,unitHeight);
   State s1Prel(gpsi0,d1);
      // will be needed as a state at center,
      // not boosted and not shifted
   State s1(gpsi1,d1);
// That boosting instances of State correctly, the class
// QMBase has to be adjusted, needs e.g. member mass
   QMBase::setMass(m1);
   QMBase::setSize(size1);
   R v1=vRel1*0.5*QMBase::vMax(d1);
   out1()<<"boost velocity of particle 1: "<<v1<<endl;
      // 0.5 inserted since without this vRel1=1
      // corresponds to a wave function mainly at
      // Nyquist frequency. Our representation of the Hamilton
      // operator is not very suitable for such strongly
      // varying wave functions
   Observable T1(Word("eKinF"),d1);
   R enr_s1=T1.expVal(s1);
   out1()<<"energy of particle1 prior to boost: "<<enr_s1<<endl;
   s1=s1.boost(v1);
  // s1=s1.smooth();
   enr_s1=T1.expVal(s1);
   out1()<<"energy of particle1 after boost: "<<enr_s1<<endl;
   R enrKin1=m1*0.5*v1*v1;
   out1()<<"kinetic energy of particle 1 by mass and velocity:\
    "<<enrKin1<<endl;
   R enr1=0.5*(enr_s1+enrKin1); // for stability (paranoia?)
   out1()<<"energy of particle 1 (mean of both): "<<enr1<<endl;

// Particle 2 should be an eigenstate of the harmonic
// oscillator represented by potential

   QMBase::setMass(m2);
   QMBase::setSize(size2);
   Observable T2(Word("eKinF"),d2);
      // F finite difference form
      // class observable --untouched for years-- workes perfectly
      // in the present context.

   R beta0=QMBase::lambdaOsc(d2);
   out1()<<"natural oscillator coupling constant: "<<beta0<<endl;
   bool adjustBeta=beta!=0 && !excLevels.isVoid();
   if (adjustBeta){
      Z dn=excLevels.last()-excLevels.fir();
      cpmassert(dn>0,loc);
      R fac=enr1*QMBase::get_hDiracInv()/dn;
      R betaAdjusted=0.5*m2*fac*fac;
      data.beta=betaAdjusted;
      out1()<<"betaAdjusted="<<betaAdjusted<<endl;
   }
   R_Func pot=R_Func::x2*data.beta*misFac;
      // Potential with beta already set to the final value
      // harmonic oscillator potential, the factor misFac allows to
      // use here a different coupling constant than the one used
      // in the potential of State2

   Observable V2("x",d2,pot);
      // function pot of position operator
   Observable H2=T2+V2;
      // Hamiltonial of a harmonic oscillator

   Graph gr;

   if (tWaitH2>0){
      H2.show(gr);
      cpmwait(tWaitH2);
      gr.clear();
   }
///////////////////// we are in App6::doTheWork() ////////////////////

   SpectralObject so;
   {  // encapsulating the heavy initialization of so
      bool enforceSymmetry=true;
         // our H2 is known to be symmetric
      bool test=false;
         // true would consume a lot of time but worked in my test cases
         // perfectly
      Z iterMax=128;
      R eps0=0;
      R eps1=1e-7,eps2=1e-6; // test values from the past
      so=H2.spectralRepresentation(enforceSymmetry,
      test,iterMax,eps0,eps1,eps2); // this is a big beast
      so=so.order(); // as so is created, its order is
      // counter intuitive (index 0 should correspond to the
      // ground-state). I decided not to change function
      // spectralRepresentation
      // to include ordering, in order not to break old code.
      // So, a new memberfunction of SpectralObject does the job
   }
   so.prnOn(out1());
   if (tWait_so>0){
      so.showSpectrum(gr);
      cpmwait(tWait_so);
      gr.clr();
      so.show(gr);
      cpmwait(tWait_so);
      gr.clr();
   }
   Z iIni=excLevels.b(), iFin=excLevels.e();
   State s2=so[iIni];
   s2=s2.symmetrize(); // experiment
  // s2=s2.smooth();
   Z i,k;
   V<State> eStates(excLevels.dim());
   Z vld=viewLevels.dim();
   V<B> bv(vld,B{false});
   for (i=1;i<=vld;++i) bv[i]=(viewLevels[i]>=0);
   viewLevels=viewLevels.select(bv); // using non-negative levels only
   vld=viewLevels.dim();
   V<State> vStates(viewLevels.dim());
   for (i=eStates.b(); i<=eStates.e(); ++i) eStates[i]=so[1+excLevels[i]];
   cpmmessage("eStates defined");
   for (i=vStates.b(); i<=vStates.e(); ++i) vStates[i]=so[viewLevels[i]];
   cpmmessage("vStates defined");

   if (adjustBeta){
      // only in this case there are two eigenvectors of H2 in the game
      R enrTrn=so(iFin).real()-so(iIni).real();
      // energy gap to the next higher excited state if there is
      // one
      out1()<<"transition energy of particle 2: "<<enrTrn<<endl;
      // this needs to be related to the kinetic energy of incoming
      // particle 1:

      // the enr1!=0
      R enrRelDiff=(enr1-cpmabs(enrTrn))/enr1;
      out1()<<"enrRelDiff="<<enrRelDiff<<
         " this should be small, otherwise the adjustment mechanism\
          failed"<<
         endl;
      State s2Fin=so[iFin];
      State2::set_d(data);
         // we need masses and sizes
      State2 sIni(s1Prel,s2);
      State2 sFin(s1Prel,s2Fin);
      C trnMatElm0=(sIni|sIni.V12());
      C trnMatElm1=(sFin|sIni.V12());
         // transition matrix element. Notice that the interaction
         // operator acts on State2, so that we have to have a
         // 'one particle content' in the state. s1Prel is a good
         // candidate for giving V12 the chance of a high contribution.
      out1()<<"Transition matrix element: "<<trnMatElm1;
      out1()<<"Stay at home matrix element: "<<trnMatElm0;
// We now multiply alpha by such a positve factor so that
// an initial value 1 gives a natural value to the new alpha
      if (alpha!=0){// we need a value in order get the sign
         R tmeR=trnMatElm1.abs();
         cpmassert(tmeR>0,loc);
         R fac=1./tmeR;
         out1()<<"Adjustment factor fac="<<fac<<endl;
         if (fac<1000){
            alpha*=fac;
               // alpha (not data.alpha) will be used as the reference
               // for the fading mechanism.
            data.alpha=alpha;
            out1()<<"Adjusted alpha="<<data.alpha<<endl;
         }
         else{
            out1()<<"Alpha was not adjusted. alpha="<<data.alpha<<endl;
         }
      }
   }
// From now on there are intances of State2 so that 'data' becomes
// effective
///////////////////// we are in App6::doTheWork() ////////////////////

   State2::set_d(data);
   State2 s(s1,s2);
   s.nor_();
   if (tWaitV12>0){ // showing the interaction potential V12
      // by applying it to a flat state
      State s1Flat(d1,C(1.));
      State s2Flat(d2,C(1.));
      State2 sFlat(s1Flat,s2Flat);
      State2 sGraph=sFlat.V12();
      sGraph.mark(gr,ipolMethod, gamma);
      gr.vis();
      cpmwait(tWaitV12);
   }
   // now s1 and s2 are no longer needed as State-instances
   State2Trj st(s);
   st.setPhaseJitter_(phaseJitter1,phaseJitter2);
   R dtLim= ePowers>0 ? s.dtSup(ePowers,tWait,gamma) : st.dtLimit();
      // dtSup is the superior method
   out1()<<"dtLim="<<dtLim<<endl;
   R dt=dtRel*dtLim;
   out1()<<"dt="<<dt<<endl;
   R tTot=nSteps*dt;
   out1()<<"tTot="<<tTot<<endl;
   R turns=tTot*v1/size1;
   out1()<<"turns="<<turns<<endl;

   bool fading=false;
   R tbf=2*tTot, tef=tbf;

   if (beginFadingOut>=0 && endFadingOut>=0 &&
       beginFadingOut<=1 && endFadingOut<=1 &&
       beginFadingOut<endFadingOut)
   {
      fading=true;
      tbf=tTot*beginFadingOut;
      tef=tTot*endFadingOut;
   }

   Z nDocu=docuPeriod>0 ? cpmrnd(1.*nSteps/docuPeriod) : 0;
 //  R_Matrix docMat1(4,nDocu);
   R_Matrix docMat1(2,nDocu);
   R_Matrix docMat2(1+eStates.dim(),nDocu);
   R_Matrix docMat3(3,nDocu);

   CyclicAlarm mark(viewingPeriod);
   CyclicAlarm docu(docuPeriod);
   CyclicAlarm img(imagingPeriod);
///////////////////// we are in App6::doTheWork() ////////////////////
   C xv0;
   R t,xx0,xv0AbsInv,xx0AbsInv;

   Z iDoc=1, iImg=1;
   R nInv=1./nSteps;
   st.stepMet_=stepMet;
   st.autoStep_=autoStep;
   st.enfNor_=enfNor;
   st.normTol_=-1;
   st.set_dt_(dt); //
   R scl=1./(dt*dt);
   Frame frx;
   vld=viewLevels.dim(); // may be 0
   Frames frv;
   V<Graph> grv(vld);
   if (vld>0){
      frv=Frames(frx,Z(vld),Z(1));
      for (i=grv.b();i<=grv.e();++i) grv[i]=Graph(frv(i,1));
   }
   for (i=0,t=0; i<nSteps; ++i,t+=dt){
      R ip=i*nInv;
      cpmprogress("loop",ip,1);
      if (fading && t>=tbf && t<=tef){
         R pf=(t-tbf)/(tef-tbf);
         R fadeFac=R_Func::hill()(pf);
         data.alpha=alpha*fadeFac;
            // fading should be relative to the adjusted value
            // of alpha
         State2::set_d(data);
      }
      R xx=st.xx();
      C xv=st.xv();
      if (i==0){
         xx0=xx;
         xv0=xv;
         R xxa=cpmabs(xx0);
         R xva=xv0.abs();
         xx0AbsInv=(xxa==0 ? R(1) : 1./xxa);
         xv0AbsInv=(xva==0 ? R(1) : 1./xva);
      }
      R nu=xx-xx0;
      C epsilon=xv-xv0;

      R nuRel=nu*xx0AbsInv;
      //C epsilonRel=epsilon*xv0AbsInv;
      C epsilonRel=epsilon; // experiment

      R nuScl=nu*scl;
      C epsilonScl=epsilon*scl;

// since epsilon is complex we divide it into real and imaginary part
      R epsRel1=epsilonRel[1];
      R epsRel2=epsilonRel[2];
      R epsScl1=epsilonScl[1];
      R epsScl2=epsilonScl[2];
      cpmvalue("nuRel",nuRel,2); // nu and epsilon refer to the
      // publication
      cpmvalue("alpha",data.alpha,3);
      epsRel2*=enrMagFac; // for docu graphics
      if (mark()){
         st.x().mark(gr,ipolMethod, gamma);
         gr.vis();
         cpmwait(tWait);
      }
      if (docu()){
      // docMat1,out2()
         docMat1[1][iDoc]=t;
         docMat1[2][iDoc]=nu;
         docMat3[1][iDoc]=t;
         docMat3[2][iDoc]=epsRel1;
         docMat3[3][iDoc]=epsRel2;
      //   docMat1[3][iDoc]=epsRel1;
      //   docMat1[4][iDoc]=epsRel2;
         out2()<<t<<" ";
// relative values
         out2()<<nuRel<<" ";
         out2()<<epsRel1<<" ";
         out2()<<epsRel2<<" ";
// scaled values
// Important new aspect of scaling: It is not necessary to create the
// data for various values of of the time step in a single program run
// since gnuplot allows to represent data from different files in a
// single diagram.
         out2()<<nuScl<<" "; // scl: scaled
         out2()<<epsScl1<<" "; // 1: real part
         out2()<<epsScl2<<" "; // 2: imaginary part
         out2()<<endl;
      // docMat2, out3()
         docMat2[1][iDoc]=t;
// listing the amplidudes of the states associated with
// the oscillator states determined by sList
         out3()<<t<<" ";
         for (k=eStates.b();k<=eStates.e();++k){
            State xredk=st.x().red2(eStates[k].con());
               // 2006-11-14: eStates[k] changed to eStates[k].con()
            R wk=xredk.absSqr();
            if (trnVisExp!=1.0) wk=cpmpow(wk,trnVisExp);
            if (k%2==0){
               wk*=trnMagFac;
            }
            docMat2[1+k][iDoc]=wk;
            out3()<<wk<<" ";
         }
         out3()<<endl;
// graphical part of docu, only active if vStates is not void
         for (k=vStates.b();k<=vStates.e();++k){
            State xredk=st.x().red2(vStates[k].con());
            if (showAsCurves){ // showing the reduced state directly
               if (iDoc==iViewReducedState){
                  // flexible method for avoiding the unsteady
                  // screen presentation resulting from auto-scaling
                  // each frame.
                  for (Z l=grv.b();l<=grv.e();l++){
                     grv[l].setAutoScale(0);
                     grv[l].scaleY(magViewReducedState);
                        // the Y-range at iDoc==iViewReducedState might
                        // be not large enough for the frames to follow.
                        // So we magnify by a suitable 'security factor'
                  }
               }//iDoc==iViewReducedState
               xredk.showAsCurves(grv[k]);
               cpmwait(tWait);
            } // if (showAsCurves)
            else{
               State2 sShow(xredk,vStates[k]);
                  // showing an interesting State2-version of
                  // the reduced state.
               sShow.mark(gr,ipolMethod, gamma);
               gr.vis();
               cpmwait(tWait);
            }
         }
         iDoc++;
      } // if docu()
      if (img()){
         Word imf="app6_";
         Word imn=cpm(iImg,3);
         out1()<<"image # "<<iImg<<" belongs to t="<<t<<endl;
         gr.write(imf&imn);
         iImg++;
      }
      st.step_();
   }
   if (docuPeriod>0 && tWaitDocMat1>0 ){
      gr.show(docMat1);
      cpmwait(tWaitDocMat1);
   }
   if (docuPeriod>0 && tWaitDocMat2>0 ){
      gr.show(docMat2);
      cpmwait(tWaitDocMat2);
   }
   if (docuPeriod>0 && tWaitDocMat3>0 ){
      gr.show(docMat3);
      cpmwait(tWaitDocMat3);
   }
   cpmwait(tWaitWorkScreen_);
   CPM_MZ
}

void CpmVQM2::App6b::doTheWork()
// The original 6b was integrated into 4c. So this is now a placeholder
{

}

void CpmVQM2::App7::doTheWork()
// This is a simple program aiming at testing the consistency
// classes State, Observable, SpectralObject and particularly
// the boost operation. Result: a state at rest gets the correct
// velocity and kinetic energy by the boost operation.
// The reason for making this part was to make sure that the strange
// behavior found in App6 is not caused by State::boost.
// This is clear now!
{
   Z mL=2;
   Word loc("CpmVQM2::App7::doTheWork()");
   CPM_MA
//////////////////////////////////////////////////////////////////////
   Word sec("run control");
//////////////////////////////////////////////////////////////////////
   Z viewingPeriod;
   R travelRel, tWait=0;
   cpmrh(travelRel);
   cpmrh(viewingPeriod);
   cpmrh(tWait);

//////////////////////////////////////////////////////////////////////
   sec="system data";
//////////////////////////////////////////////////////////////////////

   Z d1;
   R size1=1.,massRel1,vRel1;
   R dtRel=0, xRel1=0, wRel1=0;
   B useSI,noRel;

   cpmrh(d1);
   cpmrh(size1);
   cpmrh(noRel);
   cpmrh(massRel1);
   cpmrh(xRel1);
   cpmrh(wRel1);
   cpmrh(vRel1);
   cpmrh(useSI);
   cpmrh(dtRel);
// start working
   Word units=useSI ? "SI":"1";
   OFileStream out1(outFiles[1]);
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App7::doTheWork()"<<endl;
   out1()<<"units="<<units;
   QMBase::setUnits(units);
   QMBase::setSize(size1);
      // to get the right lattice related data for particle 1
   R m1= noRel ? massRel1 : massRel1*QMBase::mMax(d1);
   // That boosting instances of State correctly, the class
// QMBase has to be adjusted, needs e.g. member mass
   QMBase::setMass(m1);
   out1()<<"d1="<<d1<<endl;
   out1()<<"size1="<<size1<<endl;
   out1()<<"m1="<<m1<<endl;

// kinetic energy as spectral object, which is the most efficient
// type of object for implementing dynamics

   bool ipt=CpmRootX::isPowOfTwo(d1);
// two Hamiltonians to select (indirectly via d1)
   Observable ekin= ipt? Observable(Word("eKin"),d1) :
      Observable(Word("eKinF"),d1);
// initial state:
   bool unitHeg=true; // aha heg = height
   R_Func gpsi1=R_Func::Gauss(xRel1,wRel1,unitHeg);
   State s1(gpsi1,d1);
   R enr=(s1|ekin*s1).real();
   out1()<<"Kinetic energy befor boost: "<<enr<<endl;
   R v1=noRel ? vRel1 : vRel1*0.5*QMBase::vMax(d1);
   out1()<<"v1="<<v1<<endl;
      // 0.5 inserted since without this vRel1=1
      // corresponds to a wave function mainly at
      // Nyquist frequency. Our representation of the Hamilton
      // operator is not very suitable for such strongly
      // varying wave functions
   s1=s1.boost(v1);
   enr=(s1|ekin*s1).real();
   out1()<<"Kinetic energy after boost: "<<enr<<endl;
   enr=0.5*m1*v1*v1;
   out1()<<"Kinetic energy from mass and velocity: "<<enr<<endl;
   Graph gr;
   R dt=noRel ? dtRel : dtRel*QMBase::tRef(d1);
   out1()<<"dt direct ="<<dt<<endl;
   R travel=noRel ? travelRel : travelRel*size1;
   cpmassert(v1!=0.,loc);
   R tTot=travel/v1;
   Z nSteps=cpmrnd(tTot/dt);
   out1()<<"nSteps ="<<nSteps<<endl;
   s1.showAsCurves(gr);
   cpmwait(1+tWait);
      // inspecting the initial state
   if (nSteps>0){
      dt=tTot/nSteps;
      out1()<<"dt adjusted ="<<dt<<endl;
      CyclicAlarm cyc(viewingPeriod);
      R nInv=1./nSteps;
      SpectralObject so;
      {  // encapsulating the heavy initialization of so
         bool enforceSymmetry=false;
         // our H2 is known to be symmetric
         bool test=false;
         // true would consume a lot of time but worked in my test cases
         // perfectly
         Z iterMax=128;
         R eps0=0;
         R eps1=1e-7,eps2=1e-6; // test values from the past
         so=ekin.spectralRepresentation(enforceSymmetry,
         test,iterMax,eps0,eps1,eps2); // this is a big beast
      }
      so.expi(-dt);
         // very efficient way of letting free Hamiltonian
         // dynamics act on the boosted state
      for (Z i=0; i<nSteps; ++i){
         R ip=i*nInv;
         cpmprogress("loop",ip,1);
         s1=so*s1; // notice that rounding errors acumulate
         if (cyc()||i==(nSteps-1)){
            s1.showAsCurves(gr);
            cpmwait(tWait);
         }
      }
   }
   cpmwait(tWaitWorkScreen_);
   CPM_MZ
}

////////////////////////// CpmVQM2::App8 ///////////////////////////////

void CpmVQM2::App8::doTheWork()
// Initial states of particle 1 and 2 are Gaussian
// Made from App5 (2006-09-06)
// Covering a time interval with a few runs of different step size.
// Visualizing the scaling behavior.
{
   Z mL=1;
   Word loc("CpmVQM2::App8::doTheWork()");
   CPM_MA
//////////////////////////////////////////////////////////////////////
   Word sec("run control");
//////////////////////////////////////////////////////////////////////
   Z nSteps,viewingPeriod,nRuns,curveRes;
     //,iReversal,reversalMethod,nSmooth,nConcentrate,
     // docuPeriod,imagingPeriod,ipolMethod=1,optThr=0;
   R fac_dt=1,tWait=0,gamma=1;
   V<R> timeRange;
   cpmrh(nSteps);
   cpmrh(nRuns);
   cpmrh(viewingPeriod);
   cpmrh(fac_dt);
   cpmrh(tWait);
   cpmrh(gamma);
   cpmrh(timeRange);
   cpmrh(curveRes);

//////////////////////////////////////////////////////////////////////
   sec="system data";
//////////////////////////////////////////////////////////////////////
   Z d1=60,d2=60,stepMet=3;
   cpmrh(d1);
   cpmrh(d2);
   R size1=1.,size2=1.,massRel1,massRel2,vRel1,vRel2;
   R dtRel=0, xRel1=0, xRel2=0, wRel1=0, wRel2=0;
   cpmrh(xRel1);
   cpmrh(xRel2);
   cpmrh(wRel1);
   cpmrh(wRel2);
   B useSI,noRel,crossway{true};
   B autoStep;
   cpmrh(size1);
   cpmrh(size2);
   cpmrh(noRel);
   cpmrh(massRel1);
   cpmrh(massRel2);
   cpmrh(vRel1);
   cpmrh(vRel2);
   R alpha,beta,mu,eps,p;
   cpmrh(alpha);
   cpmrh(beta);
   cpmrh(mu);
   cpmrh(eps);
   cpmrh(p);
   cpmrh(useSI);
   cpmrh(dtRel);
   cpmrh(autoStep);
   cpmrh(crossway);
   cpmrh(stepMet);
// start working
   Word units=useSI ? "SI":"1";
   OFileStream out1(outFiles[1]);
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App8::doTheWork()"<<endl;
   out1()<<"units="<<units;
   QMBase::setUnits(units);

   QMBase::setSize(size1);
      // to get the right lattice related data for particle 1
   R m1= noRel ? massRel1 : massRel1*QMBase::mMax(d1);

   QMBase::setSize(size2); // to get the right output from mMax
   R m2=noRel ? massRel2 : massRel2*QMBase::mMax(d2);

   State2Data data;
   data.m1=m1;
   data.m2=m2;
   data.size1=size1;
   data.size2=size2;
   data.alpha=alpha;
   data.beta=beta;
   data.mu=mu;
   data.eps=eps;
   data.p=p;
   data.crossway=crossway;

   State2::set_d(data);

   // initial state:
   bool unitHeg=true; // aha heg = height
   R_Func gpsi1=R_Func::Gauss(xRel1,wRel1,unitHeg);
   R_Func gpsi2=R_Func::Gauss(xRel2,wRel2,unitHeg);

   State s1(gpsi1,d1);
   State s2(gpsi2,d2);
// That boosting instances of State correctly, the class
// QMBase has to be adjusted, needs e.g. member mass
   QMBase::setMass(m1);
   QMBase::setSize(size1);
   R v1=noRel ? vRel1 : vRel1*0.5*QMBase::vMax(d1);
   out1()<<"v1="<<v1<<endl;
      // 0.5 inseted since without this vRel1=1
      // corresponds to a wave function mainly at
      // Nyquist frequency. Our representation of the Hamilton
      // operator is not very suitable for such strongly
      // varying wave functions
   s1=s1.boost(v1);
 //  s1.reg_(); // no significant influence
   Observable T1(Word("eKinF"),d1);
   R enr1=T1.expVal(s1);
   out1()<<"enr1="<<enr1<<endl;

   QMBase::setMass(m2);
   QMBase::setSize(size2);
   R v2=noRel ? vRel2 :  vRel2*0.5*QMBase::vMax(d2);
   out1()<<"v2="<<v2<<endl;
   s2=s2.boost(v2);
   Observable T2(Word("eKinF"),d2);
   R enr2=T2.expVal(s2);
   out1()<<"enr2="<<enr2<<endl;

   Graph gr;
   Z sym=1;
   cpmrh(sym);
   cpmrh(crossway);
      // 0 for non-identical particles
      // -1 for fermions
      // else for bosons
   if (d1!=d2){
      if (sym!=0){ // would result in index errors
         sym=0;
         cpmwarning("sym changed to 0 since d1!=d2");
         cpmwait(1);
      }
   }
   State2 s(s1,s2);
   if (sym!=0){
      State2 sSym(s2,s1);
      if (sym==-1) sSym*=-1;
      s+=sSym;
      s.nor_(1);
   }
   // now s1 and s2 are no longer needed as State-instances
   State2Trj st(s);
///////////////// reminder: CpmVQM2::App8 ///////////////////////////////
   R xx0=(st.x()|st.x()).real();
   cpmdebug(xx0);
   C xv0=(st.x()|st.v());
   R xvr0=xv0.real();
   R xvi0=xv0.imag();
   State2Trj st0(st);

   R dt0=noRel ? dtRel : dtRel*st.dtLimit();
   out1()<<"dt0="<<dt0<<endl;
   R tTot=nSteps*dt0;
   R turns=tTot*v1/size1;
   R tTotInv=1./tTot;
   out1()<<"turns="<<turns<<endl;
   st.stepMet_=stepMet;
   st.autoStep_=autoStep;

   R dt=dt0;
   V< F<R,R> > fxx(nRuns);
   V< F<R,R> > fxv(nRuns);
   V< F<R,R> > fxxxv(2*nRuns);

   // for each run (indexed by k) we create two
   // functions fxx[k] and fxv[k]. Creating data arrays
   // would not be usefull since finally we shoud have
   // the curves for different k shown over a common
   // time axis. The method to achieve this is elegant.
   R mult1=1, mult2=1; // scaling factor

   for (Z k=1;k<=nRuns;++k){
      st=st0;
      st.set_dt_(dt);
      // the following vectors serve an elegant method to create a
      // common graphical representation of curces which differ
      // in the number of defining data points
      vector<R> tVal;
      vector<R> xxVal;
      vector<R> xvrVal;
      vector<R> xviVal;
      CyclicAlarm mark(viewingPeriod);
      for (R t=0; t<=tTot; t+=dt){
         R tp=t*tTotInv;
         cpmprogress("loop",tp,1);
         State2 chi=(st.x().dot()-st.v())*dt;
         R xx=(st.x()|st.x()).real()-xx0;
         C xv=(st.x()|st.v());
         R xvr=xv.real()-xvr0;
         R xvi=xv.imag()-xvi0;
         xx*=mult1;
         xvr*=mult2;
         xvi*=mult2;
         R showFac=100;
         xx*=showFac;
         tVal.push_back(t);
         xxVal.push_back(xx);
         xvrVal.push_back(xvr);
         xviVal.push_back(xvi);
         if (k==1){ // show the first nun and not the others
            if (mark()){
               st.x().mark(gr,0, gamma);
               gr.vis();
               cpmwait(tWait);
            }
         }
         st.step_();
      }
      R_Vector t_val(tVal);
      R_Vector xx_val(xxVal);
      R_Vector xvr_val(xvrVal);
      R_Vector xvi_val(xviVal);
      R_Matrix rxx(2);
      R_Matrix rxv(2);
      rxx[1]=t_val;
      rxx[2]=xx_val;
      rxv[1]=t_val;
      rxv[2]=xvr_val;
      fxx[k]=R_Func(rxx);
      fxv[k]=R_Func(rxv);
      fxxxv[k]=R_Func(rxx);
      fxxxv[k+nRuns]=R_Func(rxv);
      dt*=fac_dt;
      mult1*=1./(fac_dt*fac_dt);
      mult2*=1./(fac_dt*fac_dt);
   }
   Iv tr(timeRange[1]*tTot,timeRange[2]*tTot);
   gr.setX(tr);

   gr.addText("fxx");
   gr.show(fxx,curveRes);
   cpmwait(tWaitWorkScreen_);

   gr.addText("fxv");
   gr.show(fxv,curveRes);
   cpmwait(tWaitWorkScreen_);

   gr.addText("fxxxv");
   gr.show(fxxxv,curveRes);
   cpmwait(tWaitWorkScreen_);
   CPM_MZ
}

////////////////////////// CpmVQM2::App9 ///////////////////////////////
namespace{
   R pot2Func(R const& x, R const& alpha, R const& p, R const& mu, R\
    const&\
    eps)
   {
      R r=cpmabs(x);
      return alpha*cpmexp(-mu*r)*cpmpow(r+eps,-p-1);
   }
}

void CpmVQM2::App9::doTheWork()
// Single particle. Exact solution by spectral representation.
// studying the difference between the exact solution and the
// solution obtained by the direct midpoint integrator
// Made from App8 (2006-09-18)
{
   Z mL=1;
   Word loc("CpmVQM2::App9::doTheWork()");
   CPM_MA
//////////////////////////////////////////////////////////////////////
   Word sec("run control");
//////////////////////////////////////////////////////////////////////
   Z verboseQM,nSteps,viewingPeriod,fileData,fileStates,
      nRuns,curveRes,viewingModality;
   R tInspectPotential=0,tInspectCurves1=0,tInspectCurves2=0,
     tInspectSpectrum=0,tInspectEigenstates=0, fac_dt=1,tWait=0;
   B createData{true};
   B scaleCurves{true};
   B showLog{true};
   B showEps;
   V<R> timeRange;
/************** experiment, test *************************
   Vo<R> trO(timeRange);
   Vio<R> trIO(trO);
   cpmdebug(trIO);
************************************ works ! ************/
   cpmrh(verboseQM);
   cpmrh(nSteps);
   cpmrh(nRuns);
   cpmrh(viewingPeriod);
   cpmrh(createData);
   cpmrh(fileData);
   cpmrh(fileStates);
   cpmrh(fac_dt);
   cpmrh(tWait);
   cpmrh(tInspectPotential);
   cpmrh(tInspectCurves1);
   cpmrh(tInspectCurves2);
   cpmrh(tInspectSpectrum);
   cpmrh(tInspectEigenstates);
   cpmrh(scaleCurves);
   cpmrh(showLog);
   cpmrh(showEps);
   cpmrh(timeRange);
   cpmrh(curveRes);
   cpmrh(viewingModality);
//////////////////////////////////////////////////////////////////////
   sec="system data";
//////////////////////////////////////////////////////////////////////
   Z d1=60,stepMet=3;
   cpmrh(d1);
   R size1=1.,massRel1,vRel1;
   R dtRel=0, xRel1=0, wRel1=0;
   cpmrh(xRel1);
   cpmrh(wRel1);
   B useSI, noRel, relToStabilityLimit;
   cpmrh(size1);
   cpmrh(noRel);
   cpmrh(relToStabilityLimit);
   cpmrh(massRel1);
   cpmrh(vRel1);
   R lambdaRel,alpha,mu,eps,p;
   cpmrh(lambdaRel);
   cpmrh(alpha);
   cpmrh(mu);
   cpmrh(eps);
   cpmrh(p);
   cpmrh(useSI);
   cpmrh(dtRel);
   cpmrh(stepMet);
// start working
   if (createData){
   QMBase::setVerbose(verboseQM);
   Word units=useSI ? "SI":"1";
   OFileStream out1(outFiles[1]);
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App9::doTheWork()"<<endl;
   out1()<<"units="<<units;
   QMBase::setUnits(units);
   QMBase::setSize(size1);
      // to get the right lattice related data for particle 1
   R m1= noRel ? massRel1 : massRel1*QMBase::mMax(d1);
   QMBase::setMass(m1);
     // initial state:
   bool unitHeg=true; // aha heg = height
   R_Func gpsi1=R_Func::Gauss(xRel1,wRel1,unitHeg);
   State psi0(gpsi1,d1);
   R v1=noRel ? vRel1 : vRel1*0.5*QMBase::vMax(d1);
   out1()<<"v1="<<v1<<endl;
      // 0.5 inseted since without this vRel1=1
      // corresponds to a wave function mainly at
      // Nyquist frequency. Our representation of the Hamilton
      // operator is not very suitable for such strongly
      // varying wave functions
   psi0=psi0.boost(v1);
   R lambda=QMBase::lambdaOsc(d1)*lambdaRel;
   R_Func pot1=R_Func::x2*lambda; // potential of harmonic oscillator
#ifdef CPM_Fn
   R_Func pot2=F4<R,R,R,R,R,R>(alpha,p,mu,eps)(pot2Func);
#else
   R_Func pot2=F<R,R>(bind(pot2Func,_1,alpha,p,mu,eps));
#endif
   R_Func pot=pot1+pot2;
   R lambdaAll=lambda==0 && alpha==0 ? 0. : 1.;
   Operator op=Operator::sinParHam(R2_Func(pot),d1,lambdaAll);
      // single particle Hamilton operator from pot
   Observable ob=op.toObservable(0.);
   SpectralObject so=ob.spectralRepresentation();
   so=so.order();
   so.writeSpectrum(out1());
   Frame fr0;
   Frames frs(fr0,Z(1),Z(2));
   Graph gr;
   if (tInspectSpectrum>0){
      so.showSpectrum(gr);
      cpmwait(tInspectSpectrum);
      gr.clr();
   }
   if (tInspectEigenstates>0){
      so.show(gr);
      cpmwait(tInspectEigenstates);
      gr.clr();
   }
   Observable dot=ob*QMBase::get_ihDiracInv();
   Observable dot2=dot*dot;
   State phi0=dot*psi0;
 //  phi0*=1.1;
      // initial condition is ready
///////////////// reminder: CpmVQM2::App9 ///////////////////////////////
   if (tInspectPotential>0){
      Iv ivx(-0.5,0.5);
      gr.setX(ivx);
      gr.addText("Potential");
      gr.show(pot);
      cpmwait(tInspectPotential);
   }
   gr.clr();
   Graph gr1(frs(1,1));
   Graph gr2(frs(1,2));

   R dt0=0;
   if (noRel) dt0=dtRel;
   else{
      if (relToStabilityLimit){
         R eLim=so.spectralRadius()*QMBase::get_hDiracInv();
         dt0=dtRel*2*cpminv(eLim);
      }
      else{
         dt0=dtRel*QMBase::dtLimit(d1);
      }
   }
   out1()<<"dt0="<<dt0<<endl;
   R tTot=nSteps*dt0;
   R tTotInv=1./tTot;
   V< F<R,R> > nDiffV(nRuns);
   V< F<R,R> > psiDiffV(nRuns);
   V<Z> sttCon(nRuns); // state count
   Z kRuns=1;
   R dt=dt0;
   R mult=1;
   if ( fileStates>0 && fac_dt!=0.5){
      // adjustment to get sufficiently many synchronous states
      cpmwarning(loc&": fac_dt set to 1/2");
      fac_dt=0.5;
   }
   R fac_mult0=1./(fac_dt*fac_dt);
   R fac_mult=fac_mult0;
   Word fn("states_of_run_");
   R dtFinal=dt;
   while (kRuns<=nRuns){
      OFileStream stf(fn&cpm(kRuns));
      cpmstatus("run "&cpm(kRuns)&" of "&cpm(nRuns),3);
      gr1.setAutoScale(1);
      gr2.setAutoScale(1);
      R t=0;
      R tau=0.5*dt;
      State psi=psi0;
      State phi=phi0;
      CyclicAlarm mark(viewingPeriod);
      vector<R> tVal;
      vector<R> nDiffVal;
      vector<R> psiDiffVal;
      Z frames=0;
      while (t<tTot){
         if (mark()){ // show and document the state if wanted
            frames++;
            SpectralObject sot(so);
            sot.expi(-t);
            State psiExc=sot*psi0; // psi exact
            State diff = showEps ? phi-dot*psi : psiExc-psi;
            R nDiff=psi.abs()-1.;
            R psiDiff=diff.abs();
            nDiff*=mult;
            psiDiff*=mult;
            tVal.push_back(t);
            nDiffVal.push_back(nDiff);
            psiDiffVal.push_back(psiDiff);
            if (fileData>0){
               out1()<<t<<", "<<nDiff<<", "<<psiDiff<<endl;
            }
            if (fileStates>0){
               R tPsi(t);
               Root<R>(tPsi).prnOn(stf());
               psi.prnOn(stf());
            }
///////////////// reminder: CpmVQM2::App9 ///////////////////////////////
            if (viewingModality>0 && kRuns==nRuns){
               // the behavior of the last run can be shown
               if (viewingModality==1){ // three curves
                  // absolute values only in one frame
                  // very informativ
                  V<State> vst{psiExc,psi,diff};
                  State::show(gr,vst);
               }
               else if (viewingModality==2){
                  Z nTog=20;
                  psi.showAsCurves(gr1);
                  //psiExc.showAsCurves(gr1);
                  diff.showAsCurves(gr2);
                  if (viewingModality==2){
                     if (frames==nTog){
                        frames=0;
                        gr1.setAutoScale(1);
                        gr2.setAutoScale(1);
                     }
                     else{
                        gr1.setAutoScale(0);
                        gr2.setAutoScale(0);
                     }
                  }
               }
               else ;
               cpmwait(tWait);
            } // if (viewingModality>0 && kRuns==nRuns)
         } // if (mark())
// advancing the state dynamically
         if (stepMet==3){ // direct midpoint method
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else if (stepMet==2){ // inferior method
            State psiMid=psi+phi*tau;
            t+=tau;
            phi=dot*psiMid;
            psi+=phi*dt;
            t+=tau;
         }
         else if (stepMet==1){ // Euler method, only pedagogical
            phi=dot*psi;
            psi+=phi*dt;
            t+=dt;
         }
         else if (stepMet==4){ // pedagogical variation of the
            // direct midpoint method: re-initialization
            phi=dot*psi; // don't use phi from the last step
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else if (stepMet==5){ // pedagogical variation of the
            // direct midpoint method: normalization and
            // re-initialization
            psi.nor_();
            phi=dot*psi; // don't use phi from the last step
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else if (stepMet==6){ // Askar Cakmak method (????) I can't
            // reconstruct my reasons for this assignment. My understanding
            // of the AC method has always been that AC is the
            // normal leapfrog method. Actually stepMet==6 this is the ALF
            // method (2014-10-01).
            t+=tau;
            psi+=phi*tau;
            phi=dot*psi*2-phi;
            psi+=phi*tau;
            t+=tau;
         }
         else if (stepMet==7){ // midpoint method with smooting
            // no combination of the various conceivable actions of
            // smooting and re-normalization found that looks
            // like working so-la-la beyond the stability limit
            // (2007-06-30)
            // Mere smoothing on ooccasion does not prevent explosion.
            // Smoothing and re-normaizing on occasion freezes the
            // motion.
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
            R nPsi=psi.abs()-1;
            R tol=1e-2;
            if (nPsi>tol){
               psi.smt_();
               phi.smt_();
            }
         }
         else cpmerror(loc&" got unvalid value "&cpm(stepMet)&" for\
          stepMet");
         cpmprogress("loop",t*tTotInv,2);
      }// time loop
      R_Vector t_val(tVal);
      R_Vector nDiff_val(nDiffVal);
      R_Vector psiDiff_val(psiDiffVal);
      R_Matrix nDiffR(2);
      R_Matrix psiDiffR(2);
      nDiffR[1]=t_val;
      psiDiffR[1]=t_val;
      nDiffR[2]=nDiff_val;
      psiDiffR[2]=psiDiff_val;
      nDiffV[kRuns]=R_Func(nDiffR);
      psiDiffV[kRuns]=R_Func(psiDiffR);
      sttCon[kRuns]=t_val.dim();
      dt*=fac_dt;
      dtFinal=dt;
      if (scaleCurves) mult*=fac_mult;
      if (fileStates>0) viewingPeriod*=2;
      kRuns++;
   } // run loop
   if (tInspectCurves1>0){
      Word txt=scaleCurves ?
         "scaled difference of norm from 1" : "difference of norm from\
          1"\
         ;
      Iv tr(timeRange[1]*tTot,timeRange[2]*tTot);
      gr.setX(tr);
      gr.addText(txt);
      gr.show(nDiffV,curveRes,rainBow,showLog);
      cpmwait(tInspectCurves1);
   }
   if (tInspectCurves2>0){
      Iv tr(timeRange[1]*tTot,timeRange[2]*tTot);
      gr.setX(tr);
      Word txt0=showEps ? " epsilon" : " error";
      Word txt=scaleCurves ?
         "scaled norm of"&txt0 : "norm of"&txt0;
      gr.addText(txt);
      gr.show(psiDiffV,curveRes,rainBow,showLog);
      Word img("ip_");
      gr.write(img&getRunName()&"_"&cpm(stepMet));
      cpmwait(tInspectCurves2);
   }

///////////////// reminder: CpmVQM2::App9 ///////////////////////////////

   if (fileStates==1){ // reading synchronous states from files
      // and visulizing their difference
      R scf=1;
       V< V<R2> > cur(nRuns-1);
      for (kRuns=1;kRuns<nRuns;kRuns++){
         IFileStream if1(fn&cpm(kRuns));
         IFileStream if2(fn&cpm(kRuns+1));
         cpmstatus("replay: "&cpm(kRuns)&" of "&cpm(nRuns-1),3);
         Z n1=sttCon[kRuns];
         Z n2=sttCon[kRuns+1];
         cpmassert(n2==n1,loc);
         n1=cpminf<Z>(n1,n2);
         R n1Inv=1./n1;
         R dtLimit=0.25*dtFinal; // factor 100 preliminary
         vector<R2> vtd;
         for (Z j=1;j<=n1;++j){
            R tPsi1,tPsi2;
            State psi1,psi2;
            CpmRoot::scanFrom(tPsi1,if1());
            psi1.scanFrom(if1());
            CpmRoot::scanFrom(tPsi2,if2());
            psi2.scanFrom(if2());
            R tDiff=cpmabs(tPsi1-tPsi2);
            if (tDiff<dtLimit){
               R psi12a=(psi1-psi2).abs();
               if (scaleCurves) psi12a*=scf;
               if (showLog){
                  if (psi12a>0) psi12a=cpmlog10(psi12a);
               }
               R tp1=tPsi1;
               if (showLog && psi12a==0) ;
               else vtd.push_back(R2(tp1,psi12a));
            }
            else{
               cpmwarning(loc&": tDiff="&cpm(tDiff)&" fails to be\
                synchronous");
            }
            cpmprogress("read loop",j*n1Inv,2);
         } // j loop
         cur[kRuns]=V<R2>(1,vtd);
         scf*=4;
      }// kRuns loop
      gr.clr();
      gr.setAntiAliasLines(1);
      gr.mark(cur);
      Word txt1=showLog ? "log10 of " : "";
      Word txt2=scaleCurves ? "scaled " : "";
      Word txt3="norm of difference between synchronous states";
      gr.addText(txt1&txt2&txt3);
      gr.vis();
      cpmwait(tWaitWorkScreen_);
   } // if (fileStates>0)
   } // createData
   else { // thus !createData
      Graph gr;
      gr.clr();
      R aFac=0.25;
      V< V<R2> > cur(nRuns-1);
      Word fn("states_of_run_");
      cpmdebug(nRuns);
      for (Z k=1;k<nRuns;k++){
         R scf=cpmpow(aFac,-nRuns)/(cpmpow(aFac,-k)-1);
         IFileStream if1(fn&cpm(nRuns));
         IFileStream if2(fn&cpm(nRuns-k));
         cpmstatus("reading: "&cpm(k)&" of "&cpm(nRuns-1),3);
         R dtLimit=1e-9;
         vector<R2> vtd;
         Z frameCounter=0;
         bool fail=false;
         for(;;) { // reading from file
            R tPsi1,tPsi2;
            State psi1,psi2;
            fail=!CpmRoot::scanFrom(tPsi1,if1());
            if (fail) break;
            fail=!psi1.scanFrom(if1());
            if (fail) break;
            fail=!CpmRoot::scanFrom(tPsi2,if2());
            if (fail) break;
            fail=!psi2.scanFrom(if2());
            if (fail) break;
            frameCounter++;
            R tDiff=cpmabs(tPsi1-tPsi2);
         //   cpmassert(tDiff<dtLimit,loc);
            if (tDiff<dtLimit){
               State psi12=psi1-psi2;
               R psi12a=psi12.abs();
               psi12.showAsCurves(gr);
               if (scaleCurves) psi12a*=scf;
               if (showLog){
                  if (psi12a>0) psi12a=cpmlog10(psi12a);
               }
               R tp1=tPsi1;
               if (showLog && psi12a==0) ;
               else vtd.push_back(R2(tp1,psi12a));
            }
            else{
               cpmwarning(loc&": tDiff="&cpm(tDiff)&" fails to be\
                synchronous");
            }
         } // for (;;)
         Z nFrames=(Z)vtd.size();
         cpmdebug(nFrames);
         cur[k]=V<R2>(1,vtd);
      }// kRuns loop
  //    gr.clr();
      gr.mark(cur);
      Word txt1=showLog ? "log10 of " : "";
      Word txt2=scaleCurves ? "scaled " : "";
      Word txt3="norm of difference between synchronous states";
      gr.addText(txt1&txt2&txt3);
      gr.vis();
      cpmwait(tWaitWorkScreen_);
   }
   CPM_MZ
}

/////////////////////////////////////////////////////////////////////////

////////////////////////// CpmVQM2::App10 ///////////////////////////////

namespace{
   R log10X(R x, R ex)
   {
      R y=cpmabs(x);
      if (y==0.) return -ex;
      else return cpmlog10(y);
   }
}

void CpmVQM2::App10::doTheWork()
// 1-dimensional case: Euler formla exp(-i omega t)
// studying the difference between the exact solution and the
// solution obtained by the direct midpoint integrator
// Made from App9 (2006-10-13)
{
   Z mL=1;
   Word loc("CpmVQM2::App10::doTheWork()");
   CPM_MA
//////////////////////////////////////////////////////////////////////
   Word sec("run control");
//////////////////////////////////////////////////////////////////////
   Z verboseQM,nSteps,viewingPeriod,fileData,
      nRuns,curveRes,viewingModality;
   R tInspectPotential=0,tInspectCurves1=0,tInspectCurves2=0,
      fac_dt=1,tWait=0,logEx;
   B createData{true};
   B scaleCurves{true};
   B showLog{true};
   B showEps;
   V<R> timeRange;
   cpmrh(verboseQM);
   cpmrh(nSteps);
   cpmrh(nRuns);
   cpmrh(viewingPeriod);
   cpmrh(createData);
   cpmrh(fileData);
   cpmrh(fac_dt);
   cpmrh(tWait);
   cpmrh(tInspectCurves1);
   cpmrh(tInspectCurves2);
   cpmrh(scaleCurves);
   cpmrh(showLog);
   cpmrh(logEx);
   cpmrh(showEps);
   cpmrh(timeRange);
   cpmrh(curveRes);
   cpmrh(viewingModality);
//////////////////////////////////////////////////////////////////////
   sec="system data";
//////////////////////////////////////////////////////////////////////
   Z d1=1,stepMet=3;
   R omega=1,dt0;
   cpmrh(omega);
   cpmrh(dt0);
   cpmrh(stepMet);
// start working
   QMBase::setVerbose(verboseQM);
   OFileStream out1(outFiles[1]);
   Greg date;
   date.now();
   out1()<<"// ";
   date.writeFormatted(out1());
   out1()<<"// output of CpmVQM2::App10::doTheWork()"<<endl;
   QMBase::setUnits("1");
      // to get the right lattice related data for particle 1
   State psi0(d1,0);
   Observable ob(1.,d1);
   SpectralObject so=ob.spectralRepresentation();
   Observable dot=ob*QMBase::get_ihDiracInv();
   Observable dot2=dot*dot;
   State phi0=dot*psi0;
      // initial condition is ready
/////////////// reminder: CpmVQM2::App10 ///////////////////////////////
   Frame fr0;
   Frames frs(fr0,Z(1),Z(2));
   Graph gr;
   gr.clr();
   Graph gr1(frs(1,1));
   Graph gr2(frs(1,2));
   out1()<<"dt0="<<dt0<<endl;
   R tTot=nSteps*dt0;
   R tTotInv=1./tTot;
   V< F<R,R> > nDiffV(nRuns);
   V< F<R,R> > psiDiffV(nRuns);
   V<Z> sttCon(nRuns); // state count
   Z kRuns=1;
   R dt=dt0;
   R mult=1;
   R fac_mult0=1./(fac_dt*fac_dt);
   R fac_mult=fac_mult0;
   Word fn("states_of_run_");
   R dtFinal=dt;
   while (kRuns<=nRuns){
      OFileStream stf(fn&cpm(kRuns));
      cpmstatus("run "&cpm(kRuns)&" of "&cpm(nRuns),3);
      gr1.setAutoScale(1);
      gr2.setAutoScale(1);
      R t=0;
      R tau=0.5*dt;
      State psi=psi0;
      State phi=phi0;
      CyclicAlarm mark(viewingPeriod);
      vector<R> tVal;
      vector<R> nDiffVal;
      vector<R> psiDiffVal;
      Z frames=0;
      while (t<tTot){
         if (mark()){ // show and document the state if wanted
            frames++;
            SpectralObject sot(so);
            sot.expi(-t);
            State psiExc=sot*psi0; // psi exact
            State diff = showEps ? phi-dot*psi : psiExc-psi;
            //R nDiff=psi.abs()-1.;
            //R nDiff=psi[0].imag();
            R nDiff=psi[0].real();
            R psiDiff=diff.abs();
            nDiff*=mult;
            psiDiff*=mult;
            if (showLog){
               nDiff=log10X(nDiff,logEx);
               psiDiff=log10X(psiDiff,logEx);
            }
            tVal.push_back(t);
            nDiffVal.push_back(nDiff);
            psiDiffVal.push_back(psiDiff);
            if (fileData>0){
               out1()<<t<<", "<<nDiff<<", "<<psiDiff<<endl;
            }
/////////////// reminder: CpmVQM2::App10 ///////////////////////////////
            if (viewingModality>0 && kRuns==nRuns){
               if (viewingModality==1){ // three curves
                  // absolute values only in one frame
                  // very informativ
                  V<State> vst{psiExc,psi,diff};
                  gr.setAutoScale(1); //experiment
                  State::show(gr,vst);
               }
               else if (viewingModality==2){
                  Z nTog=10;
               //psi.showAsCurves(gr1);
                  psiExc.showAsCurves(gr1);
                  diff.showAsCurves(gr2);
                  if (viewingModality==2){
                     if (frames==nTog){
                        frames=0;
                        gr1.setAutoScale(1);
                        gr2.setAutoScale(1);
                     }
                     else{
                        gr1.setAutoScale(0);
                        gr2.setAutoScale(0);
                     }
                  }
               }
               else ;
               cpmwait(tWait);
            } // if (viewingModality>0 && kRuns==nRuns)
         } // if (mark())
// advancing the state dynamically
         if (stepMet==3){ // direct midpoint method
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else if (stepMet==2){ // inferior method
            State psiMid=psi+phi*tau;
            t+=tau;
            phi=dot*psiMid;
            psi+=phi*dt;
            t+=tau;
         }
         else if (stepMet==1){ // Euler method, only pedagogical
            phi=dot*psi;
            psi+=phi*dt;
            t+=dt;
         }
         else if (stepMet==4){ // pedagogical variation of the
            // direct midpoint method: re-initialization
            phi=dot*psi; // don't use phi from the last step
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else if (stepMet==5){ // pedagogical variation of the
            // direct midpoint method: normalization and
            // re-initialization
            psi.nor_();
            phi=dot*psi; // don't use phi from the last step
            State psiMid=psi+phi*tau;
            t+=tau;
            phi+=(dot2*psiMid)*dt;
            psi=psiMid+phi*tau;
            t+=tau;
         }
         else cpmerror(loc&" got unvalid value "&cpm(stepMet)&" for\
          stepMet");
         cpmprogress("loop",t*tTotInv,2);
      }// time loop
      R_Vector t_val(tVal);
      R_Vector nDiff_val(nDiffVal);
      R_Vector psiDiff_val(psiDiffVal);
      R_Matrix nDiffR(2);
      R_Matrix psiDiffR(2);
      nDiffR[1]=t_val;
      psiDiffR[1]=t_val;
      nDiffR[2]=nDiff_val;
      psiDiffR[2]=psiDiff_val;
      nDiffV[kRuns]=R_Func(nDiffR);
      psiDiffV[kRuns]=R_Func(psiDiffR);
      sttCon[kRuns]=t_val.dim();
      dt*=fac_dt;
      dtFinal=dt;
      if (scaleCurves) mult*=fac_mult;
      kRuns++;
   } // run loop
   Word txtF=showLog ? "log10 of " : " ";
   if (tInspectCurves1>0){
      Word txt=txtF;
      if (scaleCurves)
         txt&="scaled difference of norm from 1";
      else txt&="difference of norm from 1";
      Iv tr(timeRange[1]*tTot,timeRange[2]*tTot);
      gr.setX(tr);
      gr.addText(txt);
      gr.show(nDiffV,curveRes);
      cpmwait(tInspectCurves1);
   }
   if (tInspectCurves2>0){
      Iv tr(timeRange[1]*tTot,timeRange[2]*tTot);
      gr.setX(tr);
      Word txt=txtF;
      Word txt0=showEps ? " epsilon" : " error";
      if (scaleCurves) txt&="scaled norm of"&txt0;
      else txt&="norm of"&txt0;
      gr.addText(txt);
      gr.show(psiDiffV,curveRes);
      cpmwait(tInspectCurves2);
   }
   CPM_MZ
}

////////////////// void CpmVQM2::function(void) //////////////

// functions of signature void *f(void) for final association with a menu
// selection

void CpmVQM2::app1()
{
   cpmmessage("app1() started");
   App1 a("app1");
   a.run();
}

void CpmVQM2::app2()
{
   cpmmessage("app2() started");
   App2 a("app2");
   a.run();
}

void CpmVQM2::app3()
{
   cpmmessage("app3() started");
   App3 a("app3");
   a.run();
}

void CpmVQM2::app4()
{
   cpmmessage("app4() started");
   App4 a("app4");
   a.run();
}

void CpmVQM2::app4b()
{
   cpmmessage("app4b() started");
   App4b a("app4b");
   a.run();
}

void CpmVQM2::app4c()
{
   cpmmessage("app4c() started");
   App4c a("app4c");
   a.run();
}

void CpmVQM2::app5()
{
   cpmmessage("app5() started");
   App5 a("app5");
   a.run();
}

void CpmVQM2::app6()
{
   cpmmessage("app6() started");
   App6 a("app6");
   a.run();
}

void CpmVQM2::app6b()
{
   cpmmessage("app6b() started");
   App6b a("app6b");
   a.run();
}

void CpmVQM2::app7()
{
   cpmmessage("app7() started");
   App7 a("app7");
   a.run();
}

void CpmVQM2::app8()
{
   cpmmessage("app8() started");
   App8 a("app8");
   a.run();
}

void CpmVQM2::app9()
{
   cpmmessage("app9() started");
   App9 a("app9");
   a.run();
}

void CpmVQM2::app10()
{
   cpmmessage("app10() started");
   App10 a("app10");
   a.run();
}
