//? cpmva.h
//? C+- by Ulrich Mutze. Status of work 2022-05-06.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.


#ifndef CPM_VA_H_
#define CPM_VA_H_
/*
   Purpose: Defining a array class Va<T> which exploits the assumption
   that T supports arithmetics. See cpmv.h

   Recent history: 2012-01-22 function tim_(...) and mean() added
*/
#include <cpmvo.h>
#include <cpmtypes.h>
#include <type_traits>

////////////////////// class Va<> ////////////////////////////

namespace CpmArrays{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::Word;
   using CpmRoot::Root;
   using CpmRootX::inf;

template <typename T>
   // We assume that T provides (explicitely or implicitely)
   // copy constructor, and assignement
   // or that T is a built-in type.
   // T > T, T < T, T += T, -T, T *= T, T /=T, ostream << T, istream >> T

class Va: public Vo<T>{ // version of Vo with arithmetic operations

   typedef Va<T> Type;
   typedef T ScalarType;
   typedef Vo<T> Base;
   typedef V<T> Source;

public:

//CPM_MODCPP(Va)

explicit   Va(Z n=0):Vo<T>(n){}
      // note that Va(0) is defined and is different from Va()

explicit Va(IvZ ivz):Vo<T>(ivz){}

   Va(IvZ ivz, T const& t):Vo<T>(ivz,t){}

// constructors from explicit lists
explicit  Va(std::initializer_list<T> il ):Vo<T>(il){}
   // requires C++11
      // constructors from explicit lists such as
      // Va<Z> v{1,2,4,8};

   Va(Z n,T const& t):Vo<T>(n,t){}
      // initializes all n components with t

   Va(V<T> const& h):Vo<T>(h){}
   // 'down-cast'-constructor

   Va(Vo<T> const& h):Vo<T>(h){}
   // 'down-cast'-constructor
 //  Va(Va<T> const& h):Vo<T>(h){}
   // copy-constructor

/*
#if defined(CPM_USE_MOVE)
   Va(Va<T> && h)
   {
      Source::iv_=h.iv_;
      Source::sz_=h.sz_;
      Source::p_=h.p_;
      cout<<"move constructor Va"<<endl;
      h.iv_=IvZ();
      h.sz_=0;
      h.p_=nullptr;
   }
#endif
*/

   T sum(void)const;
      // returns the sum over all components

   T sumPrd(V<T> const& v)const;
      // returns the sum over (*this)[i]*v[i]. I.e.
      // the scalar product without a conjugation operation for the first
      // factor

   R frac(T const& t)const;
      // returns nt/dim() where nt=card{ j | (*this)[j]==t}
      // For dim()==0 we return 0.

   R frac(const V<T>& v)const;
      // returns nt/dim() where nt=card{ j | (*this)[j]==v[k] for some k}
      // For dim()==0 or v.dim()==0 we return 0.

   virtual Word nameOf()const{
      Word wi="Va<";
      return wi&CpmRoot::Name<T>()(T())&">";
   }

   virtual V<T>* clone(void)const{ return new Va(*this);}

   // discrete Laplacian
   Va<T> laplace(Z os)const; // os: outside but slightly different from
      // enum Outside

   template <class S>
   Va<T>& tim_(S const& s);
      //: times
      // using '*=' instead of 'tim_' causes confusion with existing calls
      // to *=.
      // Needed e.g. to multiply instances of Va<R3> by R as in
      // Va<R3> v=...;
      // v.tim_(3.14); // v*=3.14 would look more natural
  // template <class S>
  // requires !std::is_same<S,T>
  // Va<T>& operator*=(S const& s){ return tim_(s);}


   template <class S>
   Va<T> tim(S const& s)const;
      //: times
      // Non-mutating form of tim_

  // template <class S>
  // requires(typeid(S)!=typeid(T))
   //Va<T> operator *(S const& s)const{ return tim(s);}

   T mean(void)const{ return sum()*cpminv(R(Base::dim()));}
      // returns the sum over all components, divided by
      // their number. This makes sense only if a multiplication
      // T*R is defined

   CPM_SUM_M
   CPM_PRODUCT_M
   CPM_DIFFERENCE
   CPM_DIVISION
   CPM_SCALAR_M
   CPM_DOT_PRODUCT_LEAN
      // does no longer contain a definition of dis. This comes now from
      // V<T>
   CPM_CONJ
   CPM_NORMALIZE
   CPM_IO

private:
   static T sumFunc(T const& t1, T const& t2){ return t1+t2;}
   static T prodFunc(T const& t1, T const& t2){ return t1*t2;}
   //static T prodRFunc(T const& t1, R const& r){ return t1*r;}
   static T negFunc(T const& t) {  return -t;}
   static T idFunc(T const& t) {  return t;}
   static T invFunc(T const& t) {  return Root<T>(t).inv();}
   static T zeroFunc(T const& t)
      {  return Root<T>(t).net(0);}
   static T unitFunc(T const& t)
      {  return Root<T>(t).net(1);}
   static T conjFunc(T const& t){ return Root<T>(t).con();}
   static T dotFunc(T const& t1, T const& t2)
     { return Root<T>(t1).con()*t2;}
   static void accFunc(T& z, T const& zAdd){ z+=zAdd;}
};

// class functions:

template <class T>
T Va<T>::sum(void)const
{
   return Base::fAcc1(idFunc,accFunc);
}

template <class T>
T Va<T>::operator | (Va<T> const& h)const
{
   return Base::fAcc2(dotFunc,accFunc,h);
}

template <class T>
T Va<T>::sumPrd(V<T> const& h)const
{
   return Base::fAcc2(prodFunc,accFunc,h);
}

template <class T>
CpmRoot::R Va<T>::frac(T const& t)const
{
   Z n=Base::dim();
   if (n==0) return 0;
   Z sum=0;
   for (Z i=1;i<=n;i++) if ((*this)[i]==t) sum++;
   return ((CpmRoot::R)(sum))/n;
}

template <class T>
CpmRoot::R Va<T>::frac(const V<T>& v)const
{
   Z n=Base::dim();
   if (n==0) return 0;
   Z nv=v.dim();
   if (nv==0) return 0;
   Z i,j,sum=0;
   for (i=1;i<=n;i++){
      T ti=(*this)[i];
      for (j=1;j<=nv;j++){
         if (ti==v[j]) sum++;
      }
   }
   return ((CpmRoot::R)(sum))/n;
}

// main task

template <class T>
Va<T> Va<T>::neg(void)const
{
   //cout<<"Va::neg called"<<endl;
   V<T> v_Res=Base::fa(negFunc);
   return Va<T>(v_Res);
}

template <class T>
Va<T> Va<T>::inv(void)const
{
   V<T> v_Res=Base::fa(invFunc);
   return Va<T>(v_Res);
}

template <class T>
Va<T> Va<T>::con(void)const
{
   V<T> v_Res=Base::fa(conjFunc);
   return Va<T>(v_Res);
}

template <class T>
Va<T> Va<T>::net(Z i)const
{
   V<T> v_Res= i==1 ? Base::fa(unitFunc) : Base::fa(zeroFunc);
   return Va<T>(v_Res);
}

template <class T>
Va<T> Va<T>::laplace(Z os)const
{
   N d=Base::p_->size();
   N dm=d-1;
   using namespace CpmRoot;
  // using Base::p_;
   Va<T> res(*this);
   for (N i=1;i<dm;++i) res.pi(i)= (*Base::p_)[i]*2_R-(*Base::p_)[i-1]-
      (*Base::p_)[i+1];
   if (os==0){ // toroidal biotope
      res.pi(0)=(*Base::p_)[0]*2_R-(*Base::p_)[dm]-(*Base::p_)[1];
      res.pi(dm)=(*Base::p_)[dm]*2_R-(*Base::p_)[dm-1]-(*Base::p_)[0];
   }
   else if (os==1){ // assuming (*Base::p_)[-1] == (*Base::p_)[0] and
      // (*Base::p_)[d]=(*Base::p_)[dm] for the values of the wavefunction at
      // the sites adjacent to the bioptope's ends.
      res.pi(0)=(*Base::p_)[0]-(*Base::p_)[1];
      res.pi(dm)=(*Base::p_)[dm]-(*Base::p_)[dm-1];

   }
   else if (os==2){ // assuming enforced value 0 of the wavefunction
      // outside the biotope (i.e. (*Base::p_)[-1]=0, (*Base::p_)[d]=0
      res.pi(0)=(*Base::p_)[0]*2_R-(*Base::p_)[1];
      res.pi(dm)=(*Base::p_)[dm]*2_R-(*Base::p_)[dm-1];
   }
   else if (os==3){ // this is the most natural discrete model for
   // potential barriers of infinite height at the ends of the
   // biotope
      res.pi(0)=T();
      res.pi(dm)=T();
   }
   else{
      cpmwarning("never come to here");
   }
   return res;
}

template <class T>
Va<T>& Va<T>::operator +=(const Va<T>& s)
{
   Base::ffl_(sumFunc,s);
   return *this;
}

template <class T>
Va<T>& Va<T>::operator +=(T const& t)
{
   Base::fc_(sumFunc,t);
   return *this;
}

template <class T>
Va<T>& Va<T>::operator *=(const Va<T>& s)
{
   Base::ffl_(prodFunc,s);
   return *this;
}

template <class T>
Va<T>& Va<T>::operator *=(T const& t)
{
   Base::fc_(prodFunc,t);
   return *this;
}

template <class T>
template <class S>
Va<T>& Va<T>::tim_(S const& s)
{
   for (Z i=Base::b();i<=Base::e();++i) Base::cui(i)*=s;
   return *this;
}

template <class T>
template <class S>
Va<T> Va<T>::tim(S const& s)const
{
   Va<T> res=*this;
   return res.tim_(s);
}

template <class T>
bool Va<T>::prnOn(ostream& str)const
{
   return Base::prnOn(str);
}

template <class T>
bool Va<T>::scanFrom(istream& str)
{
   return Base::scanFrom(str);
}

///////////////////////// iterated templates ///////////////////////////
// describing multi-indexed quantities
//////////////////////////// class VVa /////////////////////////////////
// matrices without matrix product
// tensors of rank 2
// We derive VVa<T> from VVo<T> and not from Va<Va<T> >. In the
// latter case Va<T> would have to be the scalar type for VVa<T>
// which is not the case. I tried this 2005-08-19 but came deep
// 'in the woods'. Since thus the arithmetics interface is not
// yet fixed by the definition, we are free to define it anew in
// a more lean way as that of Va. Especially we avoid to
// define friend functions. Also simply taking over the interface
// of Va would not be adequate from the point of view of
// implementation efficiency.
// The implementation of VVa is such that it can be carried over to
// higher numbers of indexes iteratively

template <class T>
class VVa: public VVo<T>{ // version of VVo with arithmetic operations

   typedef VVo<T> Base;
   typedef VVa<T> Type;
   typedef T ScalarType;

public:
   CPM_LIN
      // notice that function con() declared here will be
      // implemented as conjugation of the matrix elements
      // only and not as accompanied by a matrix
      // transposition. The present implementation is natural
      // for an interpretation of VVa's as T-valued areal
      // distributions (as in images or wave functions)
      // Unlike the case for Va<T>, we do not include CPM_IO
      // in the interface of VVa<T>
// constructors
   VVa(Z d1, Z d2):Base(d1,d2){}
   VVa(Z d1, Z d2, T const& t):Base(d1,d2,t){}
      // all components are equal to t
   VVa(void):Base(){}
   VVa(V< V<T> >  const& x):Base(x){}
   VVa(VV<T> const& x):Base(x){}
   VVa(VVo<T> const& x):Base(x){}
};

template <class T>
VVa<T> VVa<T>::operator+(VVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVa<T> res(d,0);
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      Va<T> si=s[i];
      res[i]=ti+si;
   }
   return res;
}

template <class T>
VVa<T> VVa<T>::operator-(VVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVa<T> res(d,0);
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      Va<T> si=s[i];
      res[i]=ti-si;
   }
   return res;
}

template <class T>
VVa<T> VVa<T>::operator*(T const& r)const
{
   Z d=Base::dim();
   VVa<T> res(d,0);
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      res[i]=ti*r;
   }
   return res;
}

template <class T>
VVa<T> VVa<T>::operator-(void)const
{
   Z d=Base::dim();
   VVa<T> res(d,0);
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      res[i]=-ti;
   }
   return res;
}

template <class T>
VVa<T> VVa<T>::con(void)const
{
   Z d=Base::dim();
   VVa<T> res(d,0);
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      res[i]=ti.con();
   }
   return res;
}

template <class T>
T VVa<T>::operator|(VVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   T res=T();
   for (Z i=1;i<=d;i++){
      Va<T> ti=(*this)[i];
      Va<T> si=s[i];
      res+=(ti|si);
   }
   return res;
}

///////////////////////// class VVVa ///////////////////////////////////
// tensors of rank 3

template <class T>
class VVVa: public VVVo<T>{// version of VVVo with arithmetic operations

   typedef VVVo<T> Base;
   typedef VVVa<T> Type;
   typedef T ScalarType;

public:
   CPM_LIN
      // notice that function con() declared here will be
      // implemented as conjugation of the components.
      // The present implementation is natural
      // for an interpretation of VVVa's as T-valued spatial
      // distributions (as in wave functions).
      // Unlike the case for Va<T>, we do not include CPM_IO
      // in the interface of VVVa<T>
// constructors
   VVVa(Z d1,Z d2,Z d3):Base(d1,d2,d3){}
   VVVa(Z d1,Z d2,Z d3,T const& t):Base(d1,d2,d3,t){}
      // all components are equal to t
   VVVa(void):Base(){}
   VVVa( V< V< V<T> > >  const& x):Base(x){}
   VVVa(VVV<T> const& x):Base(x){}
   VVVa(VVVo<T> const& x):Base(x){}
};

template <class T>
VVVa<T> VVVa<T>::operator+(VVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVVa<T> res(d,0,0);
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      VVa<T> si=s[i];
      res[i]=ti+si;
   }
   return res;
}

template <class T>
VVVa<T> VVVa<T>::operator-(VVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVVa<T> res(d,0,0);
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      VVa<T> si=s[i];
      res[i]=ti-si;
   }
   return res;
}

template <class T>
VVVa<T> VVVa<T>::operator*(T const& r)const
{
   Z d=Base::dim();
   VVVa<T> res(d,0,0);
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      res[i]=ti*r;
   }
   return res;
}

template <class T>
VVVa<T> VVVa<T>::operator-(void)const
{
   Z d=Base::dim();
   VVVa<T> res(d,0,0);
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      res[i]=-ti;
   }
   return res;
}

template <class T>
VVVa<T> VVVa<T>::con(void)const
{
   Z d=Base::dim();
   VVVa<T> res(d,0,0);
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      res[i]=ti.con();
   }
   return res;
}

template <class T>
T VVVa<T>::operator|(VVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   T res=T();
   for (Z i=1;i<=d;i++){
      VVa<T> ti=(*this)[i];
      VVa<T> si=s[i];
      res+=(ti|si);
   }
   return res;
}

///////////////////////// class VVVVa ///////////////////////////////////
// tensors of rank 4

template <class T>
class VVVVa: public VVVVo<T>{//version of VVVVo with arithmetic operations

   typedef VVVVo<T> Base;
   typedef VVVVa<T> Type;
   typedef T ScalarType;

public:
   CPM_LIN
      // notice that function con() declared here will be
      // implemented as conjugation of the components.
      // The present implementation is natural
      // for an interpretation of VVVVa's as T-valued spatial
      // distributions (as in wave functions).
      // Unlike the case for Va<T>, we do not include CPM_IO
      // in the interface of VVVVa<T>
// constructors
   VVVVa(Z d1,Z d2,Z d3,Z d4):Base(d1,d2,d3,d4){}
   VVVVa(Z d1,Z d2,Z d3,Z d4,T const& t):Base(d1,d2,d3,d4,t){}
      // all components are equal to t
   VVVVa(void):Base(){}
   VVVVa( V< V< V< V<T> > > > const& x):Base(x){}
   VVVVa(VVVV<T> const& x):Base(x){}
   VVVVa(VVVVo<T> const& x):Base(x){}
};

template <class T>
VVVVa<T> VVVVa<T>::operator+(VVVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVVVa<T> res(d,0,0,0);
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      VVVa<T> si=s[i];
      res[i]=ti+si;
   }
   return res;
}

template <class T>
VVVVa<T> VVVVa<T>::operator-(VVVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   VVVVa<T> res(d,0,0,0);
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      VVVa<T> si=s[i];
      res[i]=ti-si;
   }
   return res;
}

template <class T>
VVVVa<T> VVVVa<T>::operator*(T const& r)const
{
   Z d=Base::dim();
   VVVVa<T> res(d,0,0,0);
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      res[i]=ti*r;
   }
   return res;
}

template <class T>
VVVVa<T> VVVVa<T>::operator-(void)const
{
   Z d=Base::dim();
   VVVVa<T> res(d,0,0,0);
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      res[i]=-ti;
   }
   return res;
}

template <class T>
VVVVa<T> VVVVa<T>::con(void)const
{
   Z d=Base::dim();
   VVVVa<T> res(d,0,0,0);
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      res[i]=ti.con();
   }
   return res;
}

template <class T>
T VVVVa<T>::operator|(VVVVa<T> const& s)const
{
   Z d1=Base::dim(),d2=s.dim();
   Z d=inf<Z>(d1,d2);
   T res=T();
   for (Z i=1;i<=d;i++){
      VVVa<T> ti=(*this)[i];
      VVVa<T> si=s[i];
      res+=(ti|si);
   }
   return res;
}

} // namespace
#endif
