//? cpmvlin.h
//? C+- by Ulrich Mutze. Status of work 2022-05-25.
//? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//?
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//?
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for
//? more details.

#ifndef CPM_VLIN_H_
#define CPM_VLIN_H_
/*
   Purpose: Defining an array class Vlin<T,S> which exploits the assumption
   that T supports just the arithmetics which occurrs in the
   definition of the leapfrog integrator ALF. For more detail see
   template class ALF<X,Y> in file cpmalgorithms.h.
*/

#include <cpmv.h>

////////////////////// class Vlin<> ////////////////////////////

namespace CpmArrays{

   using CpmRoot::Z;
   using CpmRoot::R;
   using CpmRoot::operator""_R;
   using CpmRoot::Word;
   using CpmRoot::Root;

   using namespace std;

template <typename T, typename S>
   // In my leapfrog application I had alwys S=R, but S=C or S=Z
   // could also be useful.
   // In this application T satisfies the user defined concepts
   // affin, rLinear, measurable and the standard concept semiregular.
   // The present definition is made such that Vlin<T,S>
   // also satifies these concepts.

class Vlin: public V<T>{ // version of V with arithmetic operations
   // which, for suitable types T and S make Vlin<T,S> a linear space
   // over the field S.

   typedef V<T> Base;
   using Base::p_;
   using Base::cow_;

public:

// In this class, for simplicity, we define only constuctors for vectors
// for which the first valid index is 1 (index convention)

explicit   Vlin(Z n=0):V<T>(n){}

   Vlin(Z n, T const& t):V<T>(n,t){}

// constructors from explicit lists
explicit  Vlin(std::initializer_list<T> il ):V<T>(il){}
      // constructors from explicit lists such as
      // Va<Z> v{-2,2,4,8}; Notice v[1]== -2, v[4]==8

   Vlin(vector<T> const& v):V<T>(1,v){}

   Vlin(V<T> const& h):V<T>(h)
      {if (h.dim()>0) cpmassert(h.b()==1,"index convention");}
   // 'down-cast'(?)-constructor

   V<T> base()const{ return Base(*this);}

   Word nameOf()const{
      Word wi="Vlin<";
      return wi&CpmRoot::Name<T>()(T())&","&CpmRoot::Name<S>()(S())&">";
   }
// for concept measurable
   R abs2()const;
   R abs()const{ return cpmsqrt(abs2());}

// for concept affine
   Vlin<T,S>& operator +=(Vlin<T,S> const& v);
   Vlin<T,S>& operator -=(Vlin<T,S> const& v);
// for concept rLinear
   Vlin<T,S> operator + (Vlin<T,S> const& v)const;
   Vlin<T,S> operator - (Vlin<T,S> const& v)const;
   Vlin<T,S>& operator *=(S const& s);
   Vlin<T,S> operator * (S const& s)const;
};

template <class T, class S>
R Vlin<T,S>::abs2(void)const
{
   auto res=1_R;
   for (auto val : *p_){
      res*=val.abs2();
   }
   return res;
}


template <class T, class S>
Vlin<T,S>& Vlin<T,S>::operator *=(S const& s)
{
   for (auto val : *p_) val*=s;
   return *this;
}

template <class T, class S>
Vlin<T,S> Vlin<T,S>::operator *(S const& s)const
{
   vector<T> res;
   for (auto val : *p_) res.push_back(val*s);
   return Vlin<T,S>(res);
}

template <class T, class S>
Vlin<T,S>& Vlin<T,S>::operator +=(Vlin<T,S> const& v)
{
   cow_();
   for (auto it1 = begin(*p_), it2=begin(*v.p_);
      it1 != end(*p_); ++it1, ++it2) *it1 += *it2;
   return *this;
}

template <class T, class S>
Vlin<T,S>& Vlin<T,S>::operator -=(Vlin<T,S> const& v)
{
   cow_();
   for (auto it1 = begin(*p_), it2=begin(*v.p_);
      it1 != end(*p_); ++it1, ++it2) *it1 -= *it2;
   return *this;
}

template <class T, class S>
Vlin<T,S> Vlin<T,S>::operator + (Vlin<T,S> const& v)const
{
   vector<T> res;
   for (auto it1 = begin(*p_), it2=begin(*v.p_);
      it1 != end(*p_); ++it1, ++it2) res.push_back(*it1 + *it2);
   return Vlin<T,S>(res);
}

template <class T, class S>
Vlin<T,S> Vlin<T,S>::operator - (Vlin<T,S> const& v)const
{
   vector<T> res;
   for (auto it1 = begin(*p_), it2=begin(*v.p_);
      it1 != end(*p_); ++it1, ++it2) res.push_back(*it1 - *it2);
   return Vlin<T,S>(res);
}

} // namespace
#endif
